/* Generated by Cython 0.29.33 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "yt/utilities/lib/ewahboolarray/boolarray.h",
            "yt/utilities/lib/ewahboolarray/ewah.h",
            "yt/utilities/lib/platform_dep.h",
            "yt/utilities/lib/tsearch.h"
        ],
        "extra_compile_args": [
            "-std=c++14"
        ],
        "include_dirs": [
            "yt/utilities/lib/",
            "yt/utilities/lib/ewahboolarray/"
        ],
        "language": "c++",
        "libraries": [
            "m"
        ],
        "name": "yt.geometry.particle_oct_container",
        "sources": [
            "yt/geometry/particle_oct_container.pyx"
        ]
    },
    "module_name": "yt.geometry.particle_oct_container"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_33"
#define CYTHON_HEX_VERSION 0x001D21F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__yt__geometry__particle_oct_container
#define __PYX_HAVE_API__yt__geometry__particle_oct_container
/* Early includes */
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include <utility>

    #if __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600)
    // move should be defined for these versions of MSVC, but __cplusplus isn't set usefully
    #include <type_traits>

    namespace cython_std {
    template <typename T> typename std::remove_reference<T>::type&& move(T& t) noexcept { return std::move(t); }
    template <typename T> typename std::remove_reference<T>::type&& move(T&& t) noexcept { return std::move(t); }
    }

    #endif
    
#include <map>
#include <vector>
#include <stdint.h>
#include <string>
#include <ostream>
#include <istream>
#include <sstream>
#include "ewah.h"
#include "boolarray.h"
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"
#include <float.h>
#include "tsearch.h"
#include "numpy/npy_math.h"
#include "platform_dep.h"
#include <set>
#include "pythread.h"
#include "pystate.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "yt/geometry/particle_oct_container.pyx",
  "stringsource",
  "__init__.pxd",
  "yt/utilities/lib/fp_utils.pxd",
  "yt/utilities/lib/geometry_utils.pxd",
  "yt/geometry/particle_deposit.pxd",
  "type.pxd",
  "yt/geometry/selection_routines.pxd",
  "yt/utilities/lib/allocation_container.pxd",
  "yt/geometry/oct_container.pxd",
  "yt/geometry/oct_visitors.pxd",
  "yt/utilities/lib/ewah_bool_wrap.pxd",
};
/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;

/* MemviewSliceStruct.proto */
struct __pyx_memoryview_obj;
typedef struct {
  struct __pyx_memoryview_obj *memview;
  char *data;
  Py_ssize_t shape[8];
  Py_ssize_t strides[8];
  Py_ssize_t suboffsets[8];
} __Pyx_memviewslice;
#define __Pyx_MemoryView_Len(m)  (m.shape[0])

/* Atomics.proto */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __pyx_atomic_int_type int
#if CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_incr_aligned(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_aligned(value) __sync_fetch_and_sub(value, 1)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER) && CYTHON_COMPILING_IN_NOGIL
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #pragma intrinsic (_InterlockedExchangeAdd)
    #define __pyx_atomic_incr_aligned(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_aligned(value) _InterlockedExchangeAdd(value, -1)
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif
typedef volatile __pyx_atomic_int_type __pyx_atomic_int;
#if CYTHON_ATOMICS
    #define __pyx_add_acquisition_count(memview)\
             __pyx_atomic_incr_aligned(__pyx_get_slice_count_pointer(memview))
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_atomic_decr_aligned(__pyx_get_slice_count_pointer(memview))
#else
    #define __pyx_add_acquisition_count(memview)\
            __pyx_add_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_sub_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
#endif

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()


/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":775
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":776
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":777
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":778
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":782
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":783
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":784
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":785
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":789
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":790
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":799
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":800
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":801
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":803
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":804
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":805
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":807
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":808
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":810
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":811
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":812
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;

/* "yt/utilities/lib/ewah_bool_array.pxd":97
 *     ctypedef uint32_t ewah_word_type
 * ELSE:
 *     ctypedef np.uint32_t ewah_word_type             # <<<<<<<<<<<<<<
 * ctypedef EWAHBoolArray[ewah_word_type] ewah_bool_array
 * ctypedef EWAHBoolArraySetBitForwardIterator[ewah_word_type] ewah_bool_iterator
 */
typedef __pyx_t_5numpy_uint32_t __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_word_type;

/* "utilities/lib/ewah_bool_wrap.pxd":13
 * )
 * 
 * ctypedef bint bitarrtype             # <<<<<<<<<<<<<<
 * ctypedef pair[np.uint64_t, np.uint64_t] ind_pair
 * 
 */
typedef int __pyx_t_2yt_9utilities_3lib_14ewah_bool_wrap_bitarrtype;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject;
struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector;
struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector;
struct __pyx_obj_2yt_9utilities_3lib_20allocation_container_ObjectPool;
struct __pyx_obj_2yt_8geometry_13oct_container_OctObjectPool;
struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer;
struct __pyx_obj_2yt_8geometry_13oct_container_SparseOctreeContainer;
struct __pyx_obj_2yt_8geometry_13oct_container_RAMSESOctreeContainer;
struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor;
struct __pyx_obj_2yt_8geometry_12oct_visitors_CountTotalOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_CountTotalCells;
struct __pyx_obj_2yt_8geometry_12oct_visitors_MarkOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_MaskOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_IndexOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_MaskedIndexOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_IndexMaskMapOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_ICoordsOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_IResOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_FCoordsOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_FWidthOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_CopyArrayI64;
struct __pyx_obj_2yt_8geometry_12oct_visitors_CopyArrayF64;
struct __pyx_obj_2yt_8geometry_12oct_visitors_CopyFileIndArrayI8;
struct __pyx_obj_2yt_8geometry_12oct_visitors_IdentifyOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_AssignDomainInd;
struct __pyx_obj_2yt_8geometry_12oct_visitors_FillFileIndicesO;
struct __pyx_obj_2yt_8geometry_12oct_visitors_FillFileIndicesR;
struct __pyx_obj_2yt_8geometry_12oct_visitors_CountByDomain;
struct __pyx_obj_2yt_8geometry_12oct_visitors_StoreOctree;
struct __pyx_obj_2yt_8geometry_12oct_visitors_LoadOctree;
struct __pyx_obj_2yt_8geometry_12oct_visitors_MortonIndexOcts;
struct __pyx_obj_2yt_8geometry_12oct_visitors_StoreIndex;
struct __pyx_obj_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor;
struct __pyx_obj_2yt_8geometry_12oct_visitors_NeighbourCellIndexVisitor;
struct __pyx_obj_2yt_8geometry_12oct_visitors_NeighbourCellVisitor;
struct __pyx_obj_2yt_8geometry_16particle_deposit_ParticleDepositOperation;
struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks;
struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection;
struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed;
struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet;
struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector;
struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet;
struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector;
struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer;
struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap;
struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector;
struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer;
struct __pyx_obj_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__;
struct __pyx_array_obj;
struct __pyx_MemviewEnum_obj;
struct __pyx_memoryview_obj;
struct __pyx_memoryviewslice_obj;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":814
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":815
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":816
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":818
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;

/* "yt/utilities/lib/ewah_bool_array.pxd":98
 * ELSE:
 *     ctypedef np.uint32_t ewah_word_type
 * ctypedef EWAHBoolArray[ewah_word_type] ewah_bool_array             # <<<<<<<<<<<<<<
 * ctypedef EWAHBoolArraySetBitForwardIterator[ewah_word_type] ewah_bool_iterator
 * ctypedef vector[size_t] bitset_array
 */
typedef ewah::EWAHBoolArray<__pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_word_type>  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array;

/* "yt/utilities/lib/ewah_bool_array.pxd":99
 *     ctypedef np.uint32_t ewah_word_type
 * ctypedef EWAHBoolArray[ewah_word_type] ewah_bool_array
 * ctypedef EWAHBoolArraySetBitForwardIterator[ewah_word_type] ewah_bool_iterator             # <<<<<<<<<<<<<<
 * ctypedef vector[size_t] bitset_array
 * ctypedef cmap[np.uint64_t, ewah_bool_array] ewah_map
 */
typedef ewah::EWAHBoolArraySetBitForwardIterator<__pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_word_type>  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_iterator;

/* "yt/utilities/lib/ewah_bool_array.pxd":100
 * ctypedef EWAHBoolArray[ewah_word_type] ewah_bool_array
 * ctypedef EWAHBoolArraySetBitForwardIterator[ewah_word_type] ewah_bool_iterator
 * ctypedef vector[size_t] bitset_array             # <<<<<<<<<<<<<<
 * ctypedef cmap[np.uint64_t, ewah_bool_array] ewah_map
 * ctypedef stringstream sstream
 */
typedef std::vector<size_t>  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_bitset_array;

/* "yt/utilities/lib/ewah_bool_array.pxd":101
 * ctypedef EWAHBoolArraySetBitForwardIterator[ewah_word_type] ewah_bool_iterator
 * ctypedef vector[size_t] bitset_array
 * ctypedef cmap[np.uint64_t, ewah_bool_array] ewah_map             # <<<<<<<<<<<<<<
 * ctypedef stringstream sstream
 * ctypedef BoolArray[ewah_word_type] bool_array
 */
typedef std::map<__pyx_t_5numpy_uint64_t,__pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array>  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_map;

/* "yt/utilities/lib/ewah_bool_array.pxd":102
 * ctypedef vector[size_t] bitset_array
 * ctypedef cmap[np.uint64_t, ewah_bool_array] ewah_map
 * ctypedef stringstream sstream             # <<<<<<<<<<<<<<
 * ctypedef BoolArray[ewah_word_type] bool_array
 */
typedef std::stringstream __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_sstream;

/* "yt/utilities/lib/ewah_bool_array.pxd":103
 * ctypedef cmap[np.uint64_t, ewah_bool_array] ewah_map
 * ctypedef stringstream sstream
 * ctypedef BoolArray[ewah_word_type] bool_array             # <<<<<<<<<<<<<<
 */
typedef ewah::BoolArray<__pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_word_type>  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_bool_array;
struct __pyx_t_2yt_8geometry_13grid_visitors_GridTreeNode;
struct __pyx_t_2yt_8geometry_13grid_visitors_GridTreeNodePadded;
struct __pyx_t_2yt_8geometry_13grid_visitors_GridVisitorData;

/* "grid_visitors.pxd":13
 * 
 * 
 * cdef struct GridTreeNode:             # <<<<<<<<<<<<<<
 *     np.int32_t num_children
 *     np.int32_t level
 */
struct __pyx_t_2yt_8geometry_13grid_visitors_GridTreeNode {
  __pyx_t_5numpy_int32_t num_children;
  __pyx_t_5numpy_int32_t level;
  __pyx_t_5numpy_int64_t index;
  __pyx_t_5numpy_float64_t left_edge[3];
  __pyx_t_5numpy_float64_t right_edge[3];
  struct __pyx_t_2yt_8geometry_13grid_visitors_GridTreeNode **children;
  __pyx_t_5numpy_int64_t start_index[3];
  __pyx_t_5numpy_int32_t dims[3];
  __pyx_t_5numpy_float64_t dds[3];
};

/* "grid_visitors.pxd":24
 *     np.float64_t dds[3]
 * 
 * cdef struct GridTreeNodePadded:             # <<<<<<<<<<<<<<
 *     np.int32_t num_children
 *     np.int32_t level
 */
struct __pyx_t_2yt_8geometry_13grid_visitors_GridTreeNodePadded {
  __pyx_t_5numpy_int32_t num_children;
  __pyx_t_5numpy_int32_t level;
  __pyx_t_5numpy_int64_t index;
  __pyx_t_5numpy_float64_t left_edge_x;
  __pyx_t_5numpy_float64_t left_edge_y;
  __pyx_t_5numpy_float64_t left_edge_z;
  __pyx_t_5numpy_float64_t right_edge_x;
  __pyx_t_5numpy_float64_t right_edge_y;
  __pyx_t_5numpy_float64_t right_edge_z;
  __pyx_t_5numpy_int_t children_pointers;
  __pyx_t_5numpy_int64_t start_index_x;
  __pyx_t_5numpy_int64_t start_index_y;
  __pyx_t_5numpy_int64_t start_index_z;
  __pyx_t_5numpy_int32_t dims_x;
  __pyx_t_5numpy_int32_t dims_y;
  __pyx_t_5numpy_int32_t dims_z;
  __pyx_t_5numpy_float64_t dds_x;
  __pyx_t_5numpy_float64_t dds_y;
  __pyx_t_5numpy_float64_t dds_z;
};

/* "grid_visitors.pxd":45
 *     np.float64_t dds_z
 * 
 * cdef struct GridVisitorData:             # <<<<<<<<<<<<<<
 *     GridTreeNode *grid
 *     np.uint64_t index
 */
struct __pyx_t_2yt_8geometry_13grid_visitors_GridVisitorData {
  struct __pyx_t_2yt_8geometry_13grid_visitors_GridTreeNode *grid;
  __pyx_t_5numpy_uint64_t index;
  __pyx_t_5numpy_uint64_t global_index;
  __pyx_t_5numpy_int64_t pos[3];
  int n_tuples;
  int **child_tuples;
  void *array;
  int ref_factor;
};

/* "grid_visitors.pxd":61
 * cdef np.uint8_t check_child_masked(GridVisitorData *data) nogil
 * 
 * ctypedef void grid_visitor_function(GridVisitorData *data,             # <<<<<<<<<<<<<<
 *                                          np.uint8_t selected) nogil
 * # This is similar in spirit to the way oct visitor functions work.  However,
 */
typedef void __pyx_t_2yt_8geometry_13grid_visitors_grid_visitor_function(struct __pyx_t_2yt_8geometry_13grid_visitors_GridVisitorData *, __pyx_t_5numpy_uint8_t);
struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_recursively_visit_octs;
struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid;
struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid_edge;
struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_visit_grid_cells;

/* "selection_routines.pxd":34
 *     cdef np.int64_t _hash
 * 
 *     cdef void recursively_visit_octs(self, Oct *root,             # <<<<<<<<<<<<<<
 *                         np.float64_t pos[3], np.float64_t dds[3],
 *                         int level,
 */
struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_recursively_visit_octs {
  int __pyx_n;
  int visit_covered;
};

/* "selection_routines.pxd":42
 *                               np.float64_t spos[3], np.float64_t sdds[3],
 *                               OctVisitor visitor, int i, int j, int k)
 *     cdef int select_grid(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                np.float64_t right_edge[3],
 *                                np.int32_t level, Oct *o = ?) nogil
 */
struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid {
  int __pyx_n;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *o;
};

/* "selection_routines.pxd":45
 *                                np.float64_t right_edge[3],
 *                                np.int32_t level, Oct *o = ?) nogil
 *     cdef int select_grid_edge(self, np.float64_t left_edge[3],             # <<<<<<<<<<<<<<
 *                                     np.float64_t right_edge[3],
 *                                     np.int32_t level, Oct *o = ?) nogil
 */
struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid_edge {
  int __pyx_n;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *o;
};

/* "selection_routines.pxd":68
 *                                 np.ndarray[np.uint8_t, ndim=3] mask,
 *                                 int level)
 *     cdef void visit_grid_cells(self, GridVisitorData *data,             # <<<<<<<<<<<<<<
 *                     grid_visitor_function *func, np.uint8_t *cached_mask = ?)
 * 
 */
struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_visit_grid_cells {
  int __pyx_n;
  __pyx_t_5numpy_uint8_t *cached_mask;
};
struct __pyx_t_2yt_9utilities_3lib_20allocation_container_AllocationContainer;
struct __pyx_opt_args_2yt_9utilities_3lib_20allocation_container_10ObjectPool_allocate_objs;

/* "yt/utilities/lib/allocation_container.pxd":13
 * 
 * 
 * cdef struct AllocationContainer:             # <<<<<<<<<<<<<<
 *     np.uint64_t n
 *     np.uint64_t n_assigned
 */
struct __pyx_t_2yt_9utilities_3lib_20allocation_container_AllocationContainer {
  __pyx_t_5numpy_uint64_t n;
  __pyx_t_5numpy_uint64_t n_assigned;
  __pyx_t_5numpy_uint64_t offset;
  __pyx_t_5numpy_int64_t con_id;
  void *my_objs;
};

/* "yt/utilities/lib/allocation_container.pxd":24
 *     cdef np.uint64_t n_con
 *     cdef AllocationContainer* containers
 *     cdef void allocate_objs(self, int n_objs, np.int64_t con_id = ?) except *             # <<<<<<<<<<<<<<
 *     cdef void setup_objs(self, void *obj, np.uint64_t count,
 *                          np.uint64_t offset, np.int64_t con_id)
 */
struct __pyx_opt_args_2yt_9utilities_3lib_20allocation_container_10ObjectPool_allocate_objs {
  int __pyx_n;
  __pyx_t_5numpy_int64_t con_id;
};
struct __pyx_t_2yt_8geometry_13oct_container_OctKey;
struct __pyx_t_2yt_8geometry_13oct_container_OctList;
struct __pyx_t_2yt_8geometry_13oct_container_OctAllocationContainer;
struct __pyx_opt_args_2yt_8geometry_13oct_container_15OctreeContainer_get;
struct __pyx_opt_args_2yt_8geometry_13oct_container_15OctreeContainer_visit_all_octs;

/* "oct_container.pxd":25
 * cdef int ORDER_MAX
 * 
 * cdef struct OctKey:             # <<<<<<<<<<<<<<
 *     np.int64_t key
 *     Oct *node
 */
struct __pyx_t_2yt_8geometry_13oct_container_OctKey {
  __pyx_t_5numpy_int64_t key;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *node;
  __pyx_t_5numpy_int64_t *indices;
  __pyx_t_5numpy_int64_t pcount;
};

/* "oct_container.pxd":32
 *     np.int64_t pcount
 * 
 * cdef struct OctList             # <<<<<<<<<<<<<<
 * 
 * cdef struct OctList:
 */
struct __pyx_t_2yt_8geometry_13oct_container_OctList {
  struct __pyx_t_2yt_8geometry_13oct_container_OctList *next;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *o;
};

/* "oct_container.pxd":40
 * # NOTE: This object *has* to be the same size as the AllocationContainer
 * # object.  There's an assert in the __cinit__ function.
 * cdef struct OctAllocationContainer:             # <<<<<<<<<<<<<<
 *     np.uint64_t n
 *     np.uint64_t n_assigned
 */
struct __pyx_t_2yt_8geometry_13oct_container_OctAllocationContainer {
  __pyx_t_5numpy_uint64_t n;
  __pyx_t_5numpy_uint64_t n_assigned;
  __pyx_t_5numpy_uint64_t offset;
  __pyx_t_5numpy_int64_t con_id;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *my_objs;
};

/* "oct_container.pxd":66
 *     cdef public np.int64_t nocts
 *     cdef public int num_domains
 *     cdef Oct *get(self, np.float64_t ppos[3], OctInfo *oinfo = ?,             # <<<<<<<<<<<<<<
 *                   int max_level = ?) nogil
 *     cdef int get_root(self, int ind[3], Oct **o) nogil
 */
struct __pyx_opt_args_2yt_8geometry_13oct_container_15OctreeContainer_get {
  int __pyx_n;
  struct __pyx_t_2yt_8geometry_12oct_visitors_OctInfo *oinfo;
  int max_level;
};

/* "oct_container.pxd":74
 *     # AllocationContainer.offset if such a thing exists.
 *     cdef np.int64_t get_domain_offset(self, int domain_id)
 *     cdef void visit_all_octs(self,             # <<<<<<<<<<<<<<
 *                         selection_routines.SelectorObject selector,
 *                         OctVisitor visitor,
 */
struct __pyx_opt_args_2yt_8geometry_13oct_container_15OctreeContainer_visit_all_octs {
  int __pyx_n;
  int vc;
  __pyx_t_5numpy_int64_t *indices;
};
struct __pyx_t_2yt_8geometry_12oct_visitors_Oct;
struct __pyx_t_2yt_8geometry_12oct_visitors_OctInfo;
struct __pyx_t_2yt_8geometry_12oct_visitors_OctPadded;

/* "oct_visitors.pxd":13
 * 
 * 
 * cdef struct Oct             # <<<<<<<<<<<<<<
 * cdef struct Oct:
 *     np.int64_t file_ind     # index with respect to the order in which it was
 */
struct __pyx_t_2yt_8geometry_12oct_visitors_Oct {
  __pyx_t_5numpy_int64_t file_ind;
  __pyx_t_5numpy_int64_t domain_ind;
  __pyx_t_5numpy_int64_t domain;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct **children;
};

/* "oct_visitors.pxd":21
 *     Oct **children          # Up to 8 long
 * 
 * cdef struct OctInfo:             # <<<<<<<<<<<<<<
 *     np.float64_t left_edge[3]
 *     np.float64_t dds[3]
 */
struct __pyx_t_2yt_8geometry_12oct_visitors_OctInfo {
  __pyx_t_5numpy_float64_t left_edge[3];
  __pyx_t_5numpy_float64_t dds[3];
  __pyx_t_5numpy_int64_t ipos[3];
  __pyx_t_5numpy_int32_t level;
};

/* "oct_visitors.pxd":27
 *     np.int32_t level
 * 
 * cdef struct OctPadded:             # <<<<<<<<<<<<<<
 *     np.int64_t file_ind
 *     np.int64_t domain_ind
 */
struct __pyx_t_2yt_8geometry_12oct_visitors_OctPadded {
  __pyx_t_5numpy_int64_t file_ind;
  __pyx_t_5numpy_int64_t domain_ind;
  __pyx_t_5numpy_int64_t domain;
  __pyx_t_5numpy_int64_t padding;
};

/* "particle_deposit.pxd":119
 * # So in order to mimic a registry functionality,
 * # I manually created a function to lookup the kernel functions.
 * ctypedef np.float64_t (*kernel_func) (np.float64_t) nogil             # <<<<<<<<<<<<<<
 * cdef inline kernel_func get_kernel_func(str kernel_name) nogil:
 *     with gil:
 */
typedef __pyx_t_5numpy_float64_t (*__pyx_t_2yt_8geometry_16particle_deposit_kernel_func)(__pyx_t_5numpy_float64_t);
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__find_collisions;
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__find_collisions_coarse;
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__find_collisions_refined;
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__set;
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__get;
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__select_contaminated;
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__select_uncontaminated;
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_19BoolArrayCollection__set;
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_19BoolArrayCollection__get;
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_19BoolArrayCollection__select_contaminated;
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_19BoolArrayCollection__select_uncontaminated;
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_19BoolArrayCollection__get_ghost_zones;
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_31BoolArrayCollectionUncompressed__set;
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_31BoolArrayCollectionUncompressed__get;

/* "utilities/lib/ewah_bool_wrap.pxd":14
 * 
 * ctypedef bint bitarrtype
 * ctypedef pair[np.uint64_t, np.uint64_t] ind_pair             # <<<<<<<<<<<<<<
 * 
 * cdef class FileBitmasks:
 */
typedef std::pair<__pyx_t_5numpy_uint64_t,__pyx_t_5numpy_uint64_t>  __pyx_t_2yt_9utilities_3lib_14ewah_bool_wrap_ind_pair;

/* "utilities/lib/ewah_bool_wrap.pxd":25
 *     cdef bint _iseq(self, FileBitmasks solf)
 *     cdef BoolArrayCollection _get_bitmask(self, np.uint32_t ifile)
 *     cdef tuple _find_collisions(self, BoolArrayCollection coll, bint verbose=*)             # <<<<<<<<<<<<<<
 *     cdef tuple _find_collisions_coarse(self, BoolArrayCollection coll, bint
 *                 verbose=*, file_list=*)
 */
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__find_collisions {
  int __pyx_n;
  int verbose;
};

/* "utilities/lib/ewah_bool_wrap.pxd":26
 *     cdef BoolArrayCollection _get_bitmask(self, np.uint32_t ifile)
 *     cdef tuple _find_collisions(self, BoolArrayCollection coll, bint verbose=*)
 *     cdef tuple _find_collisions_coarse(self, BoolArrayCollection coll, bint             # <<<<<<<<<<<<<<
 *                 verbose=*, file_list=*)
 *     cdef tuple _find_collisions_refined(self, BoolArrayCollection coll, bint verbose=*)
 */
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__find_collisions_coarse {
  int __pyx_n;
  int verbose;
  PyObject *file_list;
};

/* "utilities/lib/ewah_bool_wrap.pxd":28
 *     cdef tuple _find_collisions_coarse(self, BoolArrayCollection coll, bint
 *                 verbose=*, file_list=*)
 *     cdef tuple _find_collisions_refined(self, BoolArrayCollection coll, bint verbose=*)             # <<<<<<<<<<<<<<
 *     cdef void _set(self, np.uint32_t ifile, np.uint64_t i1, np.uint64_t i2=*)
 *     cdef void _set_coarse(self, np.uint32_t ifile, np.uint64_t i1)
 */
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__find_collisions_refined {
  int __pyx_n;
  int verbose;
};

/* "utilities/lib/ewah_bool_wrap.pxd":29
 *                 verbose=*, file_list=*)
 *     cdef tuple _find_collisions_refined(self, BoolArrayCollection coll, bint verbose=*)
 *     cdef void _set(self, np.uint32_t ifile, np.uint64_t i1, np.uint64_t i2=*)             # <<<<<<<<<<<<<<
 *     cdef void _set_coarse(self, np.uint32_t ifile, np.uint64_t i1)
 *     cdef void _set_refined(self, np.uint32_t ifile, np.uint64_t i1, np.uint64_t i2)
 */
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__set {
  int __pyx_n;
  __pyx_t_5numpy_uint64_t i2;
};

/* "utilities/lib/ewah_bool_wrap.pxd":39
 *     cdef void _set_map(self, np.uint32_t ifile, np.uint64_t i1, np.uint64_t i2)
 *     cdef void _set_refn(self, np.uint32_t ifile, np.uint64_t i1)
 *     cdef bint _get(self, np.uint32_t ifile, np.uint64_t i1, np.uint64_t i2=*)             # <<<<<<<<<<<<<<
 *     cdef bint _get_coarse(self, np.uint32_t ifile, np.uint64_t i1)
 *     cdef void _get_coarse_array(self, np.uint32_t ifile, np.uint64_t imax, np.uint8_t[:] arr) except *
 */
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__get {
  int __pyx_n;
  __pyx_t_5numpy_uint64_t i2;
};

/* "utilities/lib/ewah_bool_wrap.pxd":50
 *     cdef void _logicalxor(self, np.uint32_t ifile, BoolArrayCollection solf, BoolArrayCollection out)
 *     cdef void _logicaland(self, np.uint32_t ifile, BoolArrayCollection solf, BoolArrayCollection out)
 *     cdef void _select_contaminated(self, np.uint32_t ifile, BoolArrayCollection mask, np.uint8_t[:] out,             # <<<<<<<<<<<<<<
 *                np.uint8_t[:] secondary_files, BoolArrayCollection mask2=*)
 *     cdef void _select_uncontaminated(self, np.uint32_t ifile, BoolArrayCollection mask, np.uint8_t[:] out,
 */
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__select_contaminated {
  int __pyx_n;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *mask2;
};

/* "utilities/lib/ewah_bool_wrap.pxd":52
 *     cdef void _select_contaminated(self, np.uint32_t ifile, BoolArrayCollection mask, np.uint8_t[:] out,
 *                np.uint8_t[:] secondary_files, BoolArrayCollection mask2=*)
 *     cdef void _select_uncontaminated(self, np.uint32_t ifile, BoolArrayCollection mask, np.uint8_t[:] out,             # <<<<<<<<<<<<<<
 *                BoolArrayCollection mask2=*)
 *     cdef bytes _dumps(self, np.uint32_t ifile)
 */
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__select_uncontaminated {
  int __pyx_n;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *mask2;
};

/* "utilities/lib/ewah_bool_wrap.pxd":66
 *     cdef void _reset(self)
 *     cdef int _richcmp(self, BoolArrayCollection solf, int op) except -1
 *     cdef void _set(self, np.uint64_t i1, np.uint64_t i2=*)             # <<<<<<<<<<<<<<
 *     cdef void _set_coarse(self, np.uint64_t i1)
 *     cdef void _set_refined(self, np.uint64_t i1, np.uint64_t i2)
 */
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_19BoolArrayCollection__set {
  int __pyx_n;
  __pyx_t_5numpy_uint64_t i2;
};

/* "utilities/lib/ewah_bool_wrap.pxd":73
 *     cdef void _set_map(self, np.uint64_t i1, np.uint64_t i2)
 *     cdef void _set_refn(self, np.uint64_t i1)
 *     cdef bint _get(self, np.uint64_t i1, np.uint64_t i2=*)             # <<<<<<<<<<<<<<
 *     cdef bint _get_coarse(self, np.uint64_t i1)
 *     cdef void _get_coarse_array(self, np.uint64_t imax, np.uint8_t[:] arr) except *
 */
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_19BoolArrayCollection__get {
  int __pyx_n;
  __pyx_t_5numpy_uint64_t i2;
};

/* "utilities/lib/ewah_bool_wrap.pxd":87
 *     cdef void _logicalxor(self, BoolArrayCollection solf, BoolArrayCollection out)
 *     cdef void _logicaland(self, BoolArrayCollection solf, BoolArrayCollection out)
 *     cdef void _select_contaminated(self, BoolArrayCollection mask, np.uint8_t[:] out,             # <<<<<<<<<<<<<<
 *         BoolArrayCollection mask2=*)
 *     cdef void _select_uncontaminated(self, BoolArrayCollection mask, np.uint8_t[:] out,
 */
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_19BoolArrayCollection__select_contaminated {
  int __pyx_n;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *mask2;
};

/* "utilities/lib/ewah_bool_wrap.pxd":89
 *     cdef void _select_contaminated(self, BoolArrayCollection mask, np.uint8_t[:] out,
 *         BoolArrayCollection mask2=*)
 *     cdef void _select_uncontaminated(self, BoolArrayCollection mask, np.uint8_t[:] out,             # <<<<<<<<<<<<<<
 *         BoolArrayCollection mask2=*)
 *     cdef void _get_ghost_zones(self, int ngz, int order1, int order2,
 */
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_19BoolArrayCollection__select_uncontaminated {
  int __pyx_n;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *mask2;
};

/* "utilities/lib/ewah_bool_wrap.pxd":91
 *     cdef void _select_uncontaminated(self, BoolArrayCollection mask, np.uint8_t[:] out,
 *         BoolArrayCollection mask2=*)
 *     cdef void _get_ghost_zones(self, int ngz, int order1, int order2,             # <<<<<<<<<<<<<<
 *                                bint periodicity[3], BoolArrayCollection out_ewah,
 *                                bint coarse_ghosts=*)
 */
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_19BoolArrayCollection__get_ghost_zones {
  int __pyx_n;
  int coarse_ghosts;
};

/* "utilities/lib/ewah_bool_wrap.pxd":105
 *     cdef bitarrtype* ewah_refn
 * 
 *     cdef void _set(self, np.uint64_t i1, np.uint64_t i2=*)             # <<<<<<<<<<<<<<
 *     cdef void _set_coarse(self, np.uint64_t i1)
 *     cdef void _set_refined(self, np.uint64_t i1, np.uint64_t i2)
 */
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_31BoolArrayCollectionUncompressed__set {
  int __pyx_n;
  __pyx_t_5numpy_uint64_t i2;
};

/* "utilities/lib/ewah_bool_wrap.pxd":114
 *     cdef void _set_map(self, np.uint64_t i1, np.uint64_t i2)
 *     cdef void _set_refn(self, np.uint64_t i1)
 *     cdef bint _get(self, np.uint64_t i1, np.uint64_t i2=*)             # <<<<<<<<<<<<<<
 *     cdef bint _get_coarse(self, np.uint64_t i1)
 *     cdef bint _isref(self, np.uint64_t i)
 */
struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_31BoolArrayCollectionUncompressed__get {
  int __pyx_n;
  __pyx_t_5numpy_uint64_t i2;
};
struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_visit;
struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_get_from_index;
struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_coarse;
struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_refined;

/* "yt/geometry/particle_oct_container.pyx":88
 * )
 * 
 * ctypedef cmap[np.uint64_t, bool_array] CoarseRefinedSets             # <<<<<<<<<<<<<<
 * 
 * cdef class ParticleOctreeContainer(OctreeContainer):
 */
typedef std::map<__pyx_t_5numpy_uint64_t,__pyx_t_2yt_9utilities_3lib_15ewah_bool_array_bool_array>  __pyx_t_2yt_8geometry_22particle_oct_container_CoarseRefinedSets;

/* "yt/geometry/particle_oct_container.pyx":325
 *         return level_counts
 * 
 *     cdef visit(self, Oct *o, np.int64_t *counts, level = 0):             # <<<<<<<<<<<<<<
 *         cdef int i, j, k
 *         counts[level] += 1
 */
struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_visit {
  int __pyx_n;
  PyObject *level;
};

/* "yt/geometry/particle_oct_container.pyx":339
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef Oct *get_from_index(self, np.uint64_t mi, np.uint8_t order = ORDER_MAX,             # <<<<<<<<<<<<<<
 *                              int max_level = 99):
 *         cdef Oct *cur
 */
struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_get_from_index {
  int __pyx_n;
  __pyx_t_5numpy_uint8_t order;
  int max_level;
};

/* "yt/geometry/particle_oct_container.pyx":1607
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef void add_coarse(self, np.uint64_t mi1, int bbox = 2):             # <<<<<<<<<<<<<<
 *         self.coarse_select_bool[mi1] = 1
 *         # Neighbors
 */
struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_coarse {
  int __pyx_n;
  int bbox;
};

/* "yt/geometry/particle_oct_container.pyx":1635
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef int add_refined(self, np.uint64_t mi1, np.uint64_t mi2, int bbox = 2) except -1:             # <<<<<<<<<<<<<<
 *         self.refined_select_bool[mi2] = 1
 *         # Neighbors
 */
struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_refined {
  int __pyx_n;
  int bbox;
};

/* "selection_routines.pxd":24
 * 
 * 
 * cdef class SelectorObject:             # <<<<<<<<<<<<<<
 *     cdef public np.int32_t min_level
 *     cdef public np.int32_t max_level
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *__pyx_vtab;
  __pyx_t_5numpy_int32_t min_level;
  __pyx_t_5numpy_int32_t max_level;
  int overlap_cells;
  __pyx_t_5numpy_float64_t domain_width[3];
  __pyx_t_5numpy_float64_t domain_center[3];
  int periodicity[3];
  int _hash_initialized;
  __pyx_t_5numpy_int64_t _hash;
};


/* "selection_routines.pxd":76
 *         self, np.float64_t x1, np.float64_t x2, int d) nogil
 * 
 * cdef class AlwaysSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};


/* "selection_routines.pxd":79
 *     pass
 * 
 * cdef class OctreeSubsetSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef public SelectorObject base_selector
 *     cdef public np.int64_t domain_id
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *base_selector;
  __pyx_t_5numpy_int64_t domain_id;
};


/* "selection_routines.pxd":83
 *     cdef public np.int64_t domain_id
 * 
 * cdef class BooleanSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef public SelectorObject sel1
 *     cdef public SelectorObject sel2
 */
struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *sel1;
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *sel2;
};


/* "yt/utilities/lib/allocation_container.pxd":20
 *     void *my_objs
 * 
 * cdef class ObjectPool:             # <<<<<<<<<<<<<<
 *     cdef public np.uint64_t itemsize
 *     cdef np.uint64_t n_con
 */
struct __pyx_obj_2yt_9utilities_3lib_20allocation_container_ObjectPool {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_9utilities_3lib_20allocation_container_ObjectPool *__pyx_vtab;
  __pyx_t_5numpy_uint64_t itemsize;
  __pyx_t_5numpy_uint64_t n_con;
  struct __pyx_t_2yt_9utilities_3lib_20allocation_container_AllocationContainer *containers;
};


/* "oct_container.pxd":47
 *     Oct *my_objs
 * 
 * cdef class OctObjectPool(ObjectPool):             # <<<<<<<<<<<<<<
 *     cdef inline OctAllocationContainer *get_cont(self, int i):
 *         return <OctAllocationContainer*> (&self.containers[i])
 */
struct __pyx_obj_2yt_8geometry_13oct_container_OctObjectPool {
  struct __pyx_obj_2yt_9utilities_3lib_20allocation_container_ObjectPool __pyx_base;
};


/* "oct_container.pxd":55
 * cdef void OctList_delete(OctList *list)
 * 
 * cdef class OctreeContainer:             # <<<<<<<<<<<<<<
 *     cdef public OctObjectPool domains
 *     cdef Oct ****root_mesh
 */
struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctreeContainer *__pyx_vtab;
  struct __pyx_obj_2yt_8geometry_13oct_container_OctObjectPool *domains;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct ****root_mesh;
  int partial_coverage;
  int level_offset;
  int nn[3];
  __pyx_t_5numpy_uint8_t nz;
  __pyx_t_5numpy_float64_t DLE[3];
  __pyx_t_5numpy_float64_t DRE[3];
  __pyx_t_5numpy_int64_t nocts;
  int num_domains;
  PyObject *fill_style;
};


/* "oct_container.pxd":85
 *     cdef public object fill_style
 * 
 * cdef class SparseOctreeContainer(OctreeContainer):             # <<<<<<<<<<<<<<
 *     cdef OctKey *root_nodes
 *     cdef void *tree_root
 */
struct __pyx_obj_2yt_8geometry_13oct_container_SparseOctreeContainer {
  struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer __pyx_base;
  struct __pyx_t_2yt_8geometry_13oct_container_OctKey *root_nodes;
  void *tree_root;
  int num_root;
  int max_root;
};


/* "oct_container.pxd":93
 *     cdef np.int64_t ipos_to_key(self, int pos[3]) nogil
 * 
 * cdef class RAMSESOctreeContainer(SparseOctreeContainer):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_2yt_8geometry_13oct_container_RAMSESOctreeContainer {
  struct __pyx_obj_2yt_8geometry_13oct_container_SparseOctreeContainer __pyx_base;
};


/* "oct_visitors.pxd":33
 *     np.int64_t padding
 * 
 * cdef class OctVisitor:             # <<<<<<<<<<<<<<
 *     cdef np.uint64_t index
 *     cdef np.uint64_t last
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor *__pyx_vtab;
  __pyx_t_5numpy_uint64_t index;
  __pyx_t_5numpy_uint64_t last;
  __pyx_t_5numpy_int64_t global_index;
  __pyx_t_5numpy_int64_t pos[3];
  __pyx_t_5numpy_uint8_t ind[3];
  int dims;
  __pyx_t_5numpy_int32_t domain;
  __pyx_t_5numpy_int8_t level;
  __pyx_t_5numpy_int8_t nz;
  __pyx_t_5numpy_int32_t nzones;
};


/* "oct_visitors.pxd":59
 *         return (((self.ind[2]*d)+self.ind[1])*d+self.ind[0])
 * 
 * cdef class CountTotalOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_CountTotalOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};


/* "oct_visitors.pxd":62
 *     pass
 * 
 * cdef class CountTotalCells(OctVisitor):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_CountTotalCells {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};


/* "oct_visitors.pxd":65
 *     pass
 * 
 * cdef class MarkOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     # Unused
 *     cdef np.uint8_t[:,:,:,:] mark
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_MarkOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice mark;
};


/* "oct_visitors.pxd":69
 *     cdef np.uint8_t[:,:,:,:] mark
 * 
 * cdef class MaskOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:,:,:,:] mask
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_MaskOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice mask;
};


/* "oct_visitors.pxd":72
 *     cdef np.uint8_t[:,:,:,:] mask
 * 
 * cdef class IndexOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:] oct_index
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_IndexOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice oct_index;
};


/* "oct_visitors.pxd":75
 *     cdef np.int64_t[:] oct_index
 * 
 * cdef class MaskedIndexOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:] oct_index
 *     cdef np.uint8_t[:] oct_mask
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_MaskedIndexOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice oct_index;
  __Pyx_memviewslice oct_mask;
};


/* "oct_visitors.pxd":79
 *     cdef np.uint8_t[:] oct_mask
 * 
 * cdef class IndexMaskMapOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:] oct_index
 *     cdef np.uint8_t[:] oct_mask
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_IndexMaskMapOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice oct_index;
  __Pyx_memviewslice oct_mask;
  __Pyx_memviewslice map_domain_ind;
  __pyx_t_5numpy_uint64_t map_index;
};


/* "oct_visitors.pxd":85
 *     cdef np.uint64_t map_index
 * 
 * cdef class ICoordsOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:,:] icoords
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_ICoordsOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice icoords;
};


/* "oct_visitors.pxd":88
 *     cdef np.int64_t[:,:] icoords
 * 
 * cdef class IResOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:] ires
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_IResOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice ires;
};


/* "oct_visitors.pxd":91
 *     cdef np.int64_t[:] ires
 * 
 * cdef class FCoordsOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t[:,:] fcoords
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_FCoordsOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice fcoords;
};


/* "oct_visitors.pxd":94
 *     cdef np.float64_t[:,:] fcoords
 * 
 * cdef class FWidthOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t[:,:] fwidth
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_FWidthOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice fwidth;
};


/* "oct_visitors.pxd":97
 *     cdef np.float64_t[:,:] fwidth
 * 
 * cdef class CopyArrayI64(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:,:,:,:,:,:] source
 *     cdef np.int64_t[:,:] dest
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_CopyArrayI64 {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice source;
  __Pyx_memviewslice dest;
};


/* "oct_visitors.pxd":101
 *     cdef np.int64_t[:,:] dest
 * 
 * cdef class CopyArrayF64(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t[:,:,:,:,:] source
 *     cdef np.float64_t[:,:] dest
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_CopyArrayF64 {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice source;
  __Pyx_memviewslice dest;
};


/* "oct_visitors.pxd":105
 *     cdef np.float64_t[:,:] dest
 * 
 * cdef class CopyFileIndArrayI8(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t root
 *     cdef np.uint8_t[:] source
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_CopyFileIndArrayI8 {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __pyx_t_5numpy_int64_t root;
  __Pyx_memviewslice source;
  __Pyx_memviewslice dest;
};


/* "oct_visitors.pxd":110
 *     cdef np.uint8_t[:] dest
 * 
 * cdef class IdentifyOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] domain_mask
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_IdentifyOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice domain_mask;
};


/* "oct_visitors.pxd":113
 *     cdef np.uint8_t[:] domain_mask
 * 
 * cdef class AssignDomainInd(OctVisitor):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_AssignDomainInd {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};


/* "oct_visitors.pxd":116
 *     pass
 * 
 * cdef class FillFileIndicesO(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] levels
 *     cdef np.int64_t[:] file_inds
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_FillFileIndicesO {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice levels;
  __Pyx_memviewslice file_inds;
  __Pyx_memviewslice cell_inds;
};


/* "oct_visitors.pxd":121
 *     cdef np.uint8_t[:] cell_inds
 * 
 * cdef class FillFileIndicesR(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] levels
 *     cdef np.int64_t[:] file_inds
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_FillFileIndicesR {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice levels;
  __Pyx_memviewslice file_inds;
  __Pyx_memviewslice cell_inds;
};


/* "oct_visitors.pxd":126
 *     cdef np.uint8_t[:] cell_inds
 * 
 * cdef class CountByDomain(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:] domain_counts
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_CountByDomain {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice domain_counts;
};


/* "oct_visitors.pxd":129
 *     cdef np.int64_t[:] domain_counts
 * 
 * cdef class StoreOctree(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] ref_mask
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_StoreOctree {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice ref_mask;
};


/* "oct_visitors.pxd":132
 *     cdef np.uint8_t[:] ref_mask
 * 
 * cdef class LoadOctree(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] ref_mask
 *     cdef Oct* octs
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_LoadOctree {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice ref_mask;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *octs;
  __pyx_t_5numpy_uint64_t *nocts;
  __pyx_t_5numpy_uint64_t *nfinest;
};


/* "oct_visitors.pxd":138
 *     cdef np.uint64_t *nfinest
 * 
 * cdef class MortonIndexOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] level_arr
 *     cdef np.uint64_t[:] morton_ind
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_MortonIndexOcts {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice level_arr;
  __Pyx_memviewslice morton_ind;
};


/* "oct_visitors.pxd":149
 * 
 * 
 * cdef class StoreIndex(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:,:,:,:] cell_inds
 * 
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_StoreIndex {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  __Pyx_memviewslice cell_inds;
};


/* "oct_visitors.pxd":153
 * 
 * # cimport oct_container
 * cdef class BaseNeighbourVisitor(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef int idim      # 0,1,2 for x,y,z
 *     cdef int direction # +1 for +x, -1 for -x
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  int idim;
  int direction;
  __pyx_t_5numpy_uint8_t neigh_ind[3];
  int other_oct;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *neighbour;
  struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *octree;
  struct __pyx_t_2yt_8geometry_12oct_visitors_OctInfo oi;
  int n_ghost_zones;
};


/* "oct_visitors.pxd":169
 *         return (((self.neigh_ind[2]*d)+self.neigh_ind[1])*d+self.neigh_ind[0])
 * 
 * cdef class NeighbourCellIndexVisitor(BaseNeighbourVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[::1] cell_inds
 *     cdef np.int64_t[::1] domain_inds
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_NeighbourCellIndexVisitor {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor __pyx_base;
  __Pyx_memviewslice cell_inds;
  __Pyx_memviewslice domain_inds;
};


/* "oct_visitors.pxd":173
 *     cdef np.int64_t[::1] domain_inds
 * 
 * cdef class NeighbourCellVisitor(BaseNeighbourVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[::1] levels
 *     cdef np.int64_t[::1] file_inds
 */
struct __pyx_obj_2yt_8geometry_12oct_visitors_NeighbourCellVisitor {
  struct __pyx_obj_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor __pyx_base;
  __Pyx_memviewslice levels;
  __Pyx_memviewslice file_inds;
  __Pyx_memviewslice cell_inds;
  __Pyx_memviewslice domains;
};


/* "particle_deposit.pxd":139
 *             raise NotImplementedError
 * 
 * cdef class ParticleDepositOperation:             # <<<<<<<<<<<<<<
 *     # We assume each will allocate and define their own temporary storage
 *     cdef kernel_func sph_kernel
 */
struct __pyx_obj_2yt_8geometry_16particle_deposit_ParticleDepositOperation {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_8geometry_16particle_deposit_ParticleDepositOperation *__pyx_vtab;
  __pyx_t_2yt_8geometry_16particle_deposit_kernel_func sph_kernel;
  PyObject *nvals;
  int update_values;
};


/* "utilities/lib/ewah_bool_wrap.pxd":16
 * ctypedef pair[np.uint64_t, np.uint64_t] ind_pair
 * 
 * cdef class FileBitmasks:             # <<<<<<<<<<<<<<
 *     cdef np.uint32_t nfiles
 *     cdef ewah_map** ewah_coll
 */
struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *__pyx_vtab;
  __pyx_t_5numpy_uint32_t nfiles;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_map **ewah_coll;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array **ewah_keys;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array **ewah_refn;
};


/* "utilities/lib/ewah_bool_wrap.pxd":58
 *     cdef bint _check(self)
 * 
 * cdef class BoolArrayCollection:             # <<<<<<<<<<<<<<
 *     cdef ewah_map* ewah_coll
 *     cdef ewah_bool_array* ewah_keys
 */
struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_vtab;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_map *ewah_coll;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array *ewah_keys;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array *ewah_refn;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array *ewah_coar;
};


/* "utilities/lib/ewah_bool_wrap.pxd":98
 *     cdef bint _check(self)
 * 
 * cdef class BoolArrayCollectionUncompressed:             # <<<<<<<<<<<<<<
 *     cdef int nele1
 *     cdef int nele2
 */
struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_vtab;
  int nele1;
  int nele2;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_map *ewah_coll;
  __pyx_t_2yt_9utilities_3lib_14ewah_bool_wrap_bitarrtype *ewah_keys;
  __pyx_t_2yt_9utilities_3lib_14ewah_bool_wrap_bitarrtype *ewah_refn;
};


/* "utilities/lib/ewah_bool_wrap.pxd":123
 *     cdef void _compress(self, BoolArrayCollection solf)
 * 
 * cdef class SparseUnorderedBitmaskSet:             # <<<<<<<<<<<<<<
 *     cdef cset[np.uint64_t] entries
 *     cdef void _set(self, np.uint64_t ind)
 */
struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet *__pyx_vtab;
  std::set<__pyx_t_5numpy_uint64_t>  entries;
};


/* "utilities/lib/ewah_bool_wrap.pxd":132
 *     cdef to_array(self)
 * 
 * cdef class SparseUnorderedBitmaskVector:             # <<<<<<<<<<<<<<
 *     cdef int total
 *     cdef vector[np.uint64_t] entries
 */
struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector *__pyx_vtab;
  int total;
  std::vector<__pyx_t_5numpy_uint64_t>  entries;
};


/* "utilities/lib/ewah_bool_wrap.pxd":144
 *     cdef void _prune(self)
 * 
 * cdef class SparseUnorderedRefinedBitmaskSet:             # <<<<<<<<<<<<<<
 *     cdef cset[ind_pair] entries
 *     cdef void _set(self, np.uint64_t ind1, np.uint64_t ind2)
 */
struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet *__pyx_vtab;
  std::set<__pyx_t_2yt_9utilities_3lib_14ewah_bool_wrap_ind_pair>  entries;
};


/* "utilities/lib/ewah_bool_wrap.pxd":153
 *     cdef to_array(self)
 * 
 * cdef class SparseUnorderedRefinedBitmaskVector:             # <<<<<<<<<<<<<<
 *     cdef int total
 *     cdef vector[ind_pair] entries
 */
struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector *__pyx_vtab;
  int total;
  std::vector<__pyx_t_2yt_9utilities_3lib_14ewah_bool_wrap_ind_pair>  entries;
};


/* "yt/geometry/particle_oct_container.pyx":90
 * ctypedef cmap[np.uint64_t, bool_array] CoarseRefinedSets
 * 
 * cdef class ParticleOctreeContainer(OctreeContainer):             # <<<<<<<<<<<<<<
 *     cdef Oct** oct_list
 *     #The starting oct index of each domain
 */
struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer {
  struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer __pyx_base;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct **oct_list;
  __pyx_t_5numpy_int64_t *dom_offsets;
  int max_level;
  int n_ref;
};


/* "yt/geometry/particle_oct_container.pyx":433
 * cdef np.uint64_t FLAG = ~(<np.uint64_t>0)
 * 
 * cdef class ParticleBitmap:             # <<<<<<<<<<<<<<
 *     cdef np.float64_t left_edge[3]
 *     cdef np.float64_t right_edge[3]
 */
struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_vtab;
  __pyx_t_5numpy_float64_t left_edge[3];
  __pyx_t_5numpy_float64_t right_edge[3];
  __pyx_t_5numpy_uint8_t periodicity[3];
  __pyx_t_5numpy_float64_t dds[3];
  __pyx_t_5numpy_float64_t dds_mi1[3];
  __pyx_t_5numpy_float64_t dds_mi2[3];
  __pyx_t_5numpy_float64_t idds[3];
  __pyx_t_5numpy_int32_t dims[3];
  __pyx_t_5numpy_int64_t file_hash;
  __pyx_t_5numpy_uint64_t directional_max2[3];
  __pyx_t_5numpy_uint64_t nfiles;
  __pyx_t_5numpy_int32_t index_order1;
  __pyx_t_5numpy_int32_t index_order2;
  PyObject *masks;
  PyObject *particle_counts;
  PyObject *counts;
  PyObject *max_count;
  PyObject *_last_selector;
  PyObject *_last_return_values;
  PyObject *_cached_octrees;
  PyObject *_last_octree_subset;
  PyObject *_last_oct_handler;
  PyObject *_prev_octree_subset;
  PyObject *_prev_oct_handler;
  __pyx_t_5numpy_uint32_t *file_markers;
  __pyx_t_5numpy_uint64_t n_file_markers;
  __pyx_t_5numpy_uint64_t file_marker_i;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *bitmasks;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *collisions;
  int _used_mi2;
};


/* "yt/geometry/particle_oct_container.pyx":1442
 *         return octree
 * 
 * cdef class ParticleBitmapSelector:             # <<<<<<<<<<<<<<
 *     cdef SelectorObject selector
 *     cdef ParticleBitmap bitmap
 */
struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector {
  PyObject_HEAD
  struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_vtab;
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *selector;
  struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *bitmap;
  __pyx_t_5numpy_uint32_t ngz;
  __pyx_t_5numpy_float64_t DLE[3];
  __pyx_t_5numpy_float64_t DRE[3];
  int periodicity[3];
  __pyx_t_5numpy_uint32_t order1;
  __pyx_t_5numpy_uint32_t order2;
  __pyx_t_5numpy_uint64_t max_index1;
  __pyx_t_5numpy_uint64_t max_index2;
  __pyx_t_5numpy_uint64_t s1;
  __pyx_t_5numpy_uint64_t s2;
  void *pointers[11];
  __Pyx_memviewslice ind1_n;
  __Pyx_memviewslice ind2_n;
  __Pyx_memviewslice neighbors;
  __Pyx_memviewslice neighbor_list1;
  __Pyx_memviewslice neighbor_list2;
  __pyx_t_5numpy_uint32_t nfiles;
  __Pyx_memviewslice file_mask_p;
  __Pyx_memviewslice file_mask_g;
  __Pyx_memviewslice refined_select_bool;
  __Pyx_memviewslice refined_ghosts_bool;
  __Pyx_memviewslice coarse_select_bool;
  __Pyx_memviewslice coarse_ghosts_bool;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet *refined_ghosts_list;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *select_ewah;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *ghosts_ewah;
};


/* "yt/geometry/particle_oct_container.pyx":1965
 *                         self.set_files_refined(mi1,mi2)
 * 
 * cdef class ParticleBitmapOctreeContainer(SparseOctreeContainer):             # <<<<<<<<<<<<<<
 *     cdef Oct** oct_list
 *     cdef public int max_level
 */
struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer {
  struct __pyx_obj_2yt_8geometry_13oct_container_SparseOctreeContainer __pyx_base;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct **oct_list;
  int max_level;
  int n_ref;
  int loaded;
  __pyx_t_5numpy_uint8_t *_ptr_index_base_roots;
  __pyx_t_5numpy_uint8_t *_ptr_index_base_octs;
  __pyx_t_5numpy_uint64_t *_ptr_octs_per_root;
  __Pyx_memviewslice _index_base_roots;
  __Pyx_memviewslice _index_base_octs;
  __Pyx_memviewslice _octs_per_root;
  int overlap_cells;
};


/* "yt/geometry/particle_oct_container.pyx":152
 *                         self.visit_clear(o.children[cind(i,j,k)])
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         #Get the next oct, will traverse domains
 *         #Note that oct containers can be sorted
 */
struct __pyx_obj_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__ {
  PyObject_HEAD
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o;
  int __pyx_v_oi;
  struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self;
  __pyx_t_5numpy_int64_t __pyx_t_0;
  __pyx_t_5numpy_int64_t __pyx_t_1;
  int __pyx_t_2;
};


/* "View.MemoryView":106
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_array_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_array *__pyx_vtab;
  char *data;
  Py_ssize_t len;
  char *format;
  int ndim;
  Py_ssize_t *_shape;
  Py_ssize_t *_strides;
  Py_ssize_t itemsize;
  PyObject *mode;
  PyObject *_format;
  void (*callback_free_data)(void *);
  int free_data;
  int dtype_is_object;
};


/* "View.MemoryView":280
 * 
 * @cname('__pyx_MemviewEnum')
 * cdef class Enum(object):             # <<<<<<<<<<<<<<
 *     cdef object name
 *     def __init__(self, name):
 */
struct __pyx_MemviewEnum_obj {
  PyObject_HEAD
  PyObject *name;
};


/* "View.MemoryView":331
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */
struct __pyx_memoryview_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_memoryview *__pyx_vtab;
  PyObject *obj;
  PyObject *_size;
  PyObject *_array_interface;
  PyThread_type_lock lock;
  __pyx_atomic_int acquisition_count[2];
  __pyx_atomic_int *acquisition_count_aligned_p;
  Py_buffer view;
  int flags;
  int dtype_is_object;
  __Pyx_TypeInfo *typeinfo;
};


/* "View.MemoryView":967
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */
struct __pyx_memoryviewslice_obj {
  struct __pyx_memoryview_obj __pyx_base;
  __Pyx_memviewslice from_slice;
  PyObject *from_object;
  PyObject *(*to_object_func)(char *);
  int (*to_dtype_func)(char *, PyObject *);
};



/* "selection_routines.pxd":24
 * 
 * 
 * cdef class SelectorObject:             # <<<<<<<<<<<<<<
 *     cdef public np.int32_t min_level
 *     cdef public np.int32_t max_level
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject {
  void (*recursively_visit_octs)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, int, struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_recursively_visit_octs *__pyx_optional_args);
  void (*visit_oct_cells)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *, int, int, int);
  int (*select_grid)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_int32_t, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid *__pyx_optional_args);
  int (*select_grid_edge)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_int32_t, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_select_grid_edge *__pyx_optional_args);
  int (*select_cell)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *);
  int (*select_point)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *);
  int (*select_sphere)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t);
  int (*select_bbox)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *);
  int (*select_bbox_edge)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *);
  int (*fill_mask_selector_regular_grid)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, int *, PyArrayObject *, PyArrayObject *, int);
  int (*fill_mask_selector)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t **, int *, PyArrayObject *, PyArrayObject *, int);
  void (*visit_grid_cells)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, struct __pyx_t_2yt_8geometry_13grid_visitors_GridVisitorData *, __pyx_t_2yt_8geometry_13grid_visitors_grid_visitor_function *, struct __pyx_opt_args_2yt_8geometry_18selection_routines_14SelectorObject_visit_grid_cells *__pyx_optional_args);
  __pyx_t_5numpy_float64_t (*periodic_difference)(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t, int);
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject;


/* "selection_routines.pxd":76
 *         self, np.float64_t x1, np.float64_t x2, int d) nogil
 * 
 * cdef class AlwaysSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_AlwaysSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_AlwaysSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_AlwaysSelector;


/* "selection_routines.pxd":79
 *     pass
 * 
 * cdef class OctreeSubsetSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef public SelectorObject base_selector
 *     cdef public np.int64_t domain_id
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_OctreeSubsetSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_OctreeSubsetSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_OctreeSubsetSelector;


/* "selection_routines.pxd":83
 *     cdef public np.int64_t domain_id
 * 
 * cdef class BooleanSelector(SelectorObject):             # <<<<<<<<<<<<<<
 *     cdef public SelectorObject sel1
 *     cdef public SelectorObject sel2
 */

struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanSelector {
  struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanSelector *__pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanSelector;


/* "yt/utilities/lib/allocation_container.pxd":20
 *     void *my_objs
 * 
 * cdef class ObjectPool:             # <<<<<<<<<<<<<<
 *     cdef public np.uint64_t itemsize
 *     cdef np.uint64_t n_con
 */

struct __pyx_vtabstruct_2yt_9utilities_3lib_20allocation_container_ObjectPool {
  void (*allocate_objs)(struct __pyx_obj_2yt_9utilities_3lib_20allocation_container_ObjectPool *, int, struct __pyx_opt_args_2yt_9utilities_3lib_20allocation_container_10ObjectPool_allocate_objs *__pyx_optional_args);
  void (*setup_objs)(struct __pyx_obj_2yt_9utilities_3lib_20allocation_container_ObjectPool *, void *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_int64_t);
  void (*teardown_objs)(struct __pyx_obj_2yt_9utilities_3lib_20allocation_container_ObjectPool *, void *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_int64_t);
};
static struct __pyx_vtabstruct_2yt_9utilities_3lib_20allocation_container_ObjectPool *__pyx_vtabptr_2yt_9utilities_3lib_20allocation_container_ObjectPool;


/* "oct_container.pxd":47
 *     Oct *my_objs
 * 
 * cdef class OctObjectPool(ObjectPool):             # <<<<<<<<<<<<<<
 *     cdef inline OctAllocationContainer *get_cont(self, int i):
 *         return <OctAllocationContainer*> (&self.containers[i])
 */

struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctObjectPool {
  struct __pyx_vtabstruct_2yt_9utilities_3lib_20allocation_container_ObjectPool __pyx_base;
  struct __pyx_t_2yt_8geometry_13oct_container_OctAllocationContainer *(*get_cont)(struct __pyx_obj_2yt_8geometry_13oct_container_OctObjectPool *, int);
};
static struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctObjectPool *__pyx_vtabptr_2yt_8geometry_13oct_container_OctObjectPool;
static CYTHON_INLINE struct __pyx_t_2yt_8geometry_13oct_container_OctAllocationContainer *__pyx_f_2yt_8geometry_13oct_container_13OctObjectPool_get_cont(struct __pyx_obj_2yt_8geometry_13oct_container_OctObjectPool *, int);


/* "oct_container.pxd":55
 * cdef void OctList_delete(OctList *list)
 * 
 * cdef class OctreeContainer:             # <<<<<<<<<<<<<<
 *     cdef public OctObjectPool domains
 *     cdef Oct ****root_mesh
 */

struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctreeContainer {
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *(*get)(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *, __pyx_t_5numpy_float64_t *, struct __pyx_opt_args_2yt_8geometry_13oct_container_15OctreeContainer_get *__pyx_optional_args);
  int (*get_root)(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *, int *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct **);
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct **(*neighbors)(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_OctInfo *, __pyx_t_5numpy_int64_t *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, int *);
  __pyx_t_5numpy_int64_t (*get_domain_offset)(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *, int);
  void (*visit_all_octs)(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *, struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *, struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *, struct __pyx_opt_args_2yt_8geometry_13oct_container_15OctreeContainer_visit_all_octs *__pyx_optional_args);
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *(*next_root)(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *, int, int *);
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *(*next_child)(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *, int, int *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *);
  void (*append_domain)(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *, __pyx_t_5numpy_int64_t);
};
static struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctreeContainer *__pyx_vtabptr_2yt_8geometry_13oct_container_OctreeContainer;


/* "oct_container.pxd":85
 *     cdef public object fill_style
 * 
 * cdef class SparseOctreeContainer(OctreeContainer):             # <<<<<<<<<<<<<<
 *     cdef OctKey *root_nodes
 *     cdef void *tree_root
 */

struct __pyx_vtabstruct_2yt_8geometry_13oct_container_SparseOctreeContainer {
  struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctreeContainer __pyx_base;
  void (*key_to_ipos)(struct __pyx_obj_2yt_8geometry_13oct_container_SparseOctreeContainer *, __pyx_t_5numpy_int64_t, __pyx_t_5numpy_int64_t *);
  __pyx_t_5numpy_int64_t (*ipos_to_key)(struct __pyx_obj_2yt_8geometry_13oct_container_SparseOctreeContainer *, int *);
};
static struct __pyx_vtabstruct_2yt_8geometry_13oct_container_SparseOctreeContainer *__pyx_vtabptr_2yt_8geometry_13oct_container_SparseOctreeContainer;


/* "oct_container.pxd":93
 *     cdef np.int64_t ipos_to_key(self, int pos[3]) nogil
 * 
 * cdef class RAMSESOctreeContainer(SparseOctreeContainer):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_13oct_container_RAMSESOctreeContainer {
  struct __pyx_vtabstruct_2yt_8geometry_13oct_container_SparseOctreeContainer __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_13oct_container_RAMSESOctreeContainer *__pyx_vtabptr_2yt_8geometry_13oct_container_RAMSESOctreeContainer;


/* "oct_visitors.pxd":33
 *     np.int64_t padding
 * 
 * cdef class OctVisitor:             # <<<<<<<<<<<<<<
 *     cdef np.uint64_t index
 *     cdef np.uint64_t last
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor {
  void (*visit)(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_uint8_t);
  int (*oind)(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *);
  int (*rind)(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *);
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor *__pyx_vtabptr_2yt_8geometry_12oct_visitors_OctVisitor;
static CYTHON_INLINE int __pyx_f_2yt_8geometry_12oct_visitors_10OctVisitor_oind(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *);
static CYTHON_INLINE int __pyx_f_2yt_8geometry_12oct_visitors_10OctVisitor_rind(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *);


/* "oct_visitors.pxd":59
 *         return (((self.ind[2]*d)+self.ind[1])*d+self.ind[0])
 * 
 * cdef class CountTotalOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CountTotalOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CountTotalOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_CountTotalOcts;


/* "oct_visitors.pxd":62
 *     pass
 * 
 * cdef class CountTotalCells(OctVisitor):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CountTotalCells {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CountTotalCells *__pyx_vtabptr_2yt_8geometry_12oct_visitors_CountTotalCells;


/* "oct_visitors.pxd":65
 *     pass
 * 
 * cdef class MarkOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     # Unused
 *     cdef np.uint8_t[:,:,:,:] mark
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_MarkOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_MarkOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_MarkOcts;


/* "oct_visitors.pxd":69
 *     cdef np.uint8_t[:,:,:,:] mark
 * 
 * cdef class MaskOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:,:,:,:] mask
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_MaskOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_MaskOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_MaskOcts;


/* "oct_visitors.pxd":72
 *     cdef np.uint8_t[:,:,:,:] mask
 * 
 * cdef class IndexOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:] oct_index
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IndexOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IndexOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_IndexOcts;


/* "oct_visitors.pxd":75
 *     cdef np.int64_t[:] oct_index
 * 
 * cdef class MaskedIndexOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:] oct_index
 *     cdef np.uint8_t[:] oct_mask
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_MaskedIndexOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_MaskedIndexOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_MaskedIndexOcts;


/* "oct_visitors.pxd":79
 *     cdef np.uint8_t[:] oct_mask
 * 
 * cdef class IndexMaskMapOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:] oct_index
 *     cdef np.uint8_t[:] oct_mask
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IndexMaskMapOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IndexMaskMapOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_IndexMaskMapOcts;


/* "oct_visitors.pxd":85
 *     cdef np.uint64_t map_index
 * 
 * cdef class ICoordsOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:,:] icoords
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_ICoordsOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_ICoordsOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_ICoordsOcts;


/* "oct_visitors.pxd":88
 *     cdef np.int64_t[:,:] icoords
 * 
 * cdef class IResOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:] ires
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IResOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IResOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_IResOcts;


/* "oct_visitors.pxd":91
 *     cdef np.int64_t[:] ires
 * 
 * cdef class FCoordsOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t[:,:] fcoords
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FCoordsOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FCoordsOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_FCoordsOcts;


/* "oct_visitors.pxd":94
 *     cdef np.float64_t[:,:] fcoords
 * 
 * cdef class FWidthOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t[:,:] fwidth
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FWidthOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FWidthOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_FWidthOcts;


/* "oct_visitors.pxd":97
 *     cdef np.float64_t[:,:] fwidth
 * 
 * cdef class CopyArrayI64(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:,:,:,:,:,:] source
 *     cdef np.int64_t[:,:] dest
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CopyArrayI64 {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CopyArrayI64 *__pyx_vtabptr_2yt_8geometry_12oct_visitors_CopyArrayI64;


/* "oct_visitors.pxd":101
 *     cdef np.int64_t[:,:] dest
 * 
 * cdef class CopyArrayF64(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t[:,:,:,:,:] source
 *     cdef np.float64_t[:,:] dest
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CopyArrayF64 {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CopyArrayF64 *__pyx_vtabptr_2yt_8geometry_12oct_visitors_CopyArrayF64;


/* "oct_visitors.pxd":105
 *     cdef np.float64_t[:,:] dest
 * 
 * cdef class CopyFileIndArrayI8(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t root
 *     cdef np.uint8_t[:] source
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CopyFileIndArrayI8 {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CopyFileIndArrayI8 *__pyx_vtabptr_2yt_8geometry_12oct_visitors_CopyFileIndArrayI8;


/* "oct_visitors.pxd":110
 *     cdef np.uint8_t[:] dest
 * 
 * cdef class IdentifyOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] domain_mask
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IdentifyOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IdentifyOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_IdentifyOcts;


/* "oct_visitors.pxd":113
 *     cdef np.uint8_t[:] domain_mask
 * 
 * cdef class AssignDomainInd(OctVisitor):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_AssignDomainInd {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_AssignDomainInd *__pyx_vtabptr_2yt_8geometry_12oct_visitors_AssignDomainInd;


/* "oct_visitors.pxd":116
 *     pass
 * 
 * cdef class FillFileIndicesO(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] levels
 *     cdef np.int64_t[:] file_inds
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FillFileIndicesO {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FillFileIndicesO *__pyx_vtabptr_2yt_8geometry_12oct_visitors_FillFileIndicesO;


/* "oct_visitors.pxd":121
 *     cdef np.uint8_t[:] cell_inds
 * 
 * cdef class FillFileIndicesR(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] levels
 *     cdef np.int64_t[:] file_inds
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FillFileIndicesR {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FillFileIndicesR *__pyx_vtabptr_2yt_8geometry_12oct_visitors_FillFileIndicesR;


/* "oct_visitors.pxd":126
 *     cdef np.uint8_t[:] cell_inds
 * 
 * cdef class CountByDomain(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:] domain_counts
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CountByDomain {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CountByDomain *__pyx_vtabptr_2yt_8geometry_12oct_visitors_CountByDomain;


/* "oct_visitors.pxd":129
 *     cdef np.int64_t[:] domain_counts
 * 
 * cdef class StoreOctree(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] ref_mask
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_StoreOctree {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_StoreOctree *__pyx_vtabptr_2yt_8geometry_12oct_visitors_StoreOctree;


/* "oct_visitors.pxd":132
 *     cdef np.uint8_t[:] ref_mask
 * 
 * cdef class LoadOctree(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] ref_mask
 *     cdef Oct* octs
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_LoadOctree {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_LoadOctree *__pyx_vtabptr_2yt_8geometry_12oct_visitors_LoadOctree;


/* "oct_visitors.pxd":138
 *     cdef np.uint64_t *nfinest
 * 
 * cdef class MortonIndexOcts(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[:] level_arr
 *     cdef np.uint64_t[:] morton_ind
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_MortonIndexOcts {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_MortonIndexOcts *__pyx_vtabptr_2yt_8geometry_12oct_visitors_MortonIndexOcts;


/* "oct_visitors.pxd":149
 * 
 * 
 * cdef class StoreIndex(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.int64_t[:,:,:,:] cell_inds
 * 
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_StoreIndex {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_StoreIndex *__pyx_vtabptr_2yt_8geometry_12oct_visitors_StoreIndex;


/* "oct_visitors.pxd":153
 * 
 * # cimport oct_container
 * cdef class BaseNeighbourVisitor(OctVisitor):             # <<<<<<<<<<<<<<
 *     cdef int idim      # 0,1,2 for x,y,z
 *     cdef int direction # +1 for +x, -1 for -x
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor __pyx_base;
  void (*set_neighbour_info)(struct __pyx_obj_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, int *);
  __pyx_t_5numpy_uint8_t (*neighbour_rind)(struct __pyx_obj_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor *);
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor *__pyx_vtabptr_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor;
static CYTHON_INLINE __pyx_t_5numpy_uint8_t __pyx_f_2yt_8geometry_12oct_visitors_20BaseNeighbourVisitor_neighbour_rind(struct __pyx_obj_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor *);


/* "oct_visitors.pxd":169
 *         return (((self.neigh_ind[2]*d)+self.neigh_ind[1])*d+self.neigh_ind[0])
 * 
 * cdef class NeighbourCellIndexVisitor(BaseNeighbourVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[::1] cell_inds
 *     cdef np.int64_t[::1] domain_inds
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_NeighbourCellIndexVisitor {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_NeighbourCellIndexVisitor *__pyx_vtabptr_2yt_8geometry_12oct_visitors_NeighbourCellIndexVisitor;


/* "oct_visitors.pxd":173
 *     cdef np.int64_t[::1] domain_inds
 * 
 * cdef class NeighbourCellVisitor(BaseNeighbourVisitor):             # <<<<<<<<<<<<<<
 *     cdef np.uint8_t[::1] levels
 *     cdef np.int64_t[::1] file_inds
 */

struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_NeighbourCellVisitor {
  struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor __pyx_base;
};
static struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_NeighbourCellVisitor *__pyx_vtabptr_2yt_8geometry_12oct_visitors_NeighbourCellVisitor;


/* "particle_deposit.pxd":139
 *             raise NotImplementedError
 * 
 * cdef class ParticleDepositOperation:             # <<<<<<<<<<<<<<
 *     # We assume each will allocate and define their own temporary storage
 *     cdef kernel_func sph_kernel
 */

struct __pyx_vtabstruct_2yt_8geometry_16particle_deposit_ParticleDepositOperation {
  int (*process)(struct __pyx_obj_2yt_8geometry_16particle_deposit_ParticleDepositOperation *, int *, int, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_int64_t, __pyx_t_5numpy_float64_t *, __Pyx_memviewslice, __pyx_t_5numpy_int64_t);
};
static struct __pyx_vtabstruct_2yt_8geometry_16particle_deposit_ParticleDepositOperation *__pyx_vtabptr_2yt_8geometry_16particle_deposit_ParticleDepositOperation;


/* "utilities/lib/ewah_bool_wrap.pxd":16
 * ctypedef pair[np.uint64_t, np.uint64_t] ind_pair
 * 
 * cdef class FileBitmasks:             # <<<<<<<<<<<<<<
 *     cdef np.uint32_t nfiles
 *     cdef ewah_map** ewah_coll
 */

struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks {
  void (*_reset)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *);
  int (*_iseq)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *);
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *(*_get_bitmask)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t);
  PyObject *(*_find_collisions)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__find_collisions *__pyx_optional_args);
  PyObject *(*_find_collisions_coarse)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__find_collisions_coarse *__pyx_optional_args);
  PyObject *(*_find_collisions_refined)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__find_collisions_refined *__pyx_optional_args);
  void (*_set)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t, __pyx_t_5numpy_uint64_t, struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__set *__pyx_optional_args);
  void (*_set_coarse)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t, __pyx_t_5numpy_uint64_t);
  void (*_set_refined)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t);
  void (*_set_coarse_array)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t, __Pyx_memviewslice);
  void (*_set_refined_array)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t, __pyx_t_5numpy_uint64_t, __Pyx_memviewslice);
  void (*_set_refined_index_array)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t, __pyx_t_5numpy_int64_t, PyArrayObject *, PyArrayObject *);
  void (*_set_map)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t);
  void (*_set_refn)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t, __pyx_t_5numpy_uint64_t);
  int (*_get)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t, __pyx_t_5numpy_uint64_t, struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__get *__pyx_optional_args);
  int (*_get_coarse)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t, __pyx_t_5numpy_uint64_t);
  void (*_get_coarse_array)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t, __pyx_t_5numpy_uint64_t, __Pyx_memviewslice);
  int (*_isref)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t, __pyx_t_5numpy_uint64_t);
  __pyx_t_5numpy_uint64_t (*_count_total)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t);
  __pyx_t_5numpy_uint64_t (*_count_refined)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t);
  __pyx_t_5numpy_uint64_t (*_count_coarse)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t);
  void (*_append)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
  int (*_intersects)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
  void (*_logicalxor)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
  void (*_logicaland)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
  void (*_select_contaminated)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, __Pyx_memviewslice, __Pyx_memviewslice, struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__select_contaminated *__pyx_optional_args);
  void (*_select_uncontaminated)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, __Pyx_memviewslice, struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__select_uncontaminated *__pyx_optional_args);
  PyObject *(*_dumps)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t);
  int (*_loads)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *, __pyx_t_5numpy_uint32_t, PyObject *);
  int (*_check)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *);
};
static struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *__pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks;


/* "utilities/lib/ewah_bool_wrap.pxd":58
 *     cdef bint _check(self)
 * 
 * cdef class BoolArrayCollection:             # <<<<<<<<<<<<<<
 *     cdef ewah_map* ewah_coll
 *     cdef ewah_bool_array* ewah_keys
 */

struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection {
  void (*_reset)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
  int (*_richcmp)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, int);
  void (*_set)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, __pyx_t_5numpy_uint64_t, struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_19BoolArrayCollection__set *__pyx_optional_args);
  void (*_set_coarse)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, __pyx_t_5numpy_uint64_t);
  void (*_set_refined)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t);
  void (*_set_coarse_array)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, __Pyx_memviewslice);
  void (*_set_refined_array)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, __pyx_t_5numpy_uint64_t, __Pyx_memviewslice);
  void (*_set_map)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t);
  void (*_set_refn)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, __pyx_t_5numpy_uint64_t);
  int (*_get)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, __pyx_t_5numpy_uint64_t, struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_19BoolArrayCollection__get *__pyx_optional_args);
  int (*_get_coarse)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, __pyx_t_5numpy_uint64_t);
  void (*_get_coarse_array)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, __pyx_t_5numpy_uint64_t, __Pyx_memviewslice);
  int (*_contains)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, __pyx_t_5numpy_uint64_t);
  int (*_isref)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, __pyx_t_5numpy_uint64_t);
  void (*_ewah_coarse)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
  __pyx_t_5numpy_uint64_t (*_count_total)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
  __pyx_t_5numpy_uint64_t (*_count_refined)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
  __pyx_t_5numpy_uint64_t (*_count_coarse)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
  void (*_append)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
  void (*_logicalor)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
  int (*_intersects)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
  void (*_logicalxor)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
  void (*_logicaland)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
  void (*_select_contaminated)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, __Pyx_memviewslice, struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_19BoolArrayCollection__select_contaminated *__pyx_optional_args);
  void (*_select_uncontaminated)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, __Pyx_memviewslice, struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_19BoolArrayCollection__select_uncontaminated *__pyx_optional_args);
  void (*_get_ghost_zones)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, int, int, int, int *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_19BoolArrayCollection__get_ghost_zones *__pyx_optional_args);
  PyObject *(*_dumps)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
  int (*_loads)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, PyObject *);
  int (*_check)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
};
static struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection;


/* "utilities/lib/ewah_bool_wrap.pxd":98
 *     cdef bint _check(self)
 * 
 * cdef class BoolArrayCollectionUncompressed:             # <<<<<<<<<<<<<<
 *     cdef int nele1
 *     cdef int nele2
 */

struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed {
  void (*_set)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, __pyx_t_5numpy_uint64_t, struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_31BoolArrayCollectionUncompressed__set *__pyx_optional_args);
  void (*_set_coarse)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, __pyx_t_5numpy_uint64_t);
  void (*_set_refined)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t);
  void (*_set_coarse_array)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, __Pyx_memviewslice);
  void (*_set_coarse_array_ptr)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, __pyx_t_5numpy_uint8_t *);
  void (*_set_refined_array)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, __pyx_t_5numpy_uint64_t, __Pyx_memviewslice);
  void (*_set_refined_array_ptr)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint8_t *);
  void (*_set_map)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t);
  void (*_set_refn)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, __pyx_t_5numpy_uint64_t);
  int (*_get)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, __pyx_t_5numpy_uint64_t, struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_31BoolArrayCollectionUncompressed__get *__pyx_optional_args);
  int (*_get_coarse)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, __pyx_t_5numpy_uint64_t);
  int (*_isref)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, __pyx_t_5numpy_uint64_t);
  __pyx_t_5numpy_uint64_t (*_count_total)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *);
  __pyx_t_5numpy_uint64_t (*_count_refined)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *);
  void (*_append)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *);
  int (*_intersects)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *);
  void (*_compress)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
};
static struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed;


/* "utilities/lib/ewah_bool_wrap.pxd":123
 *     cdef void _compress(self, BoolArrayCollection solf)
 * 
 * cdef class SparseUnorderedBitmaskSet:             # <<<<<<<<<<<<<<
 *     cdef cset[np.uint64_t] entries
 *     cdef void _set(self, np.uint64_t ind)
 */

struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet {
  void (*_set)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet *, __pyx_t_5numpy_uint64_t);
  void (*_fill)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet *, __Pyx_memviewslice);
  void (*_fill_ewah)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
  void (*_fill_bool)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *);
  void (*_reset)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet *);
  PyObject *(*to_array)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet *);
};
static struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet *__pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet;


/* "utilities/lib/ewah_bool_wrap.pxd":132
 *     cdef to_array(self)
 * 
 * cdef class SparseUnorderedBitmaskVector:             # <<<<<<<<<<<<<<
 *     cdef int total
 *     cdef vector[np.uint64_t] entries
 */

struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector {
  void (*_set)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector *, __pyx_t_5numpy_uint64_t);
  void (*_fill)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector *, __Pyx_memviewslice);
  void (*_fill_ewah)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
  void (*_fill_bool)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *);
  void (*_reset)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector *);
  PyObject *(*to_array)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector *);
  void (*_remove_duplicates)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector *);
  void (*_prune)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector *);
};
static struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector *__pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector;


/* "utilities/lib/ewah_bool_wrap.pxd":144
 *     cdef void _prune(self)
 * 
 * cdef class SparseUnorderedRefinedBitmaskSet:             # <<<<<<<<<<<<<<
 *     cdef cset[ind_pair] entries
 *     cdef void _set(self, np.uint64_t ind1, np.uint64_t ind2)
 */

struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet {
  void (*_set)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t);
  void (*_fill)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet *, __Pyx_memviewslice, __Pyx_memviewslice);
  void (*_fill_ewah)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
  void (*_fill_bool)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *);
  void (*_reset)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet *);
  PyObject *(*to_array)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet *);
};
static struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet *__pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet;


/* "utilities/lib/ewah_bool_wrap.pxd":153
 *     cdef to_array(self)
 * 
 * cdef class SparseUnorderedRefinedBitmaskVector:             # <<<<<<<<<<<<<<
 *     cdef int total
 *     cdef vector[ind_pair] entries
 */

struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector {
  void (*_set)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t);
  void (*_fill)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector *, __Pyx_memviewslice, __Pyx_memviewslice);
  void (*_fill_ewah)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *);
  void (*_fill_bool)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *);
  void (*_reset)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector *);
  PyObject *(*to_array)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector *);
  void (*_remove_duplicates)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector *);
  void (*_prune)(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector *);
};
static struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector *__pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector;


/* "yt/geometry/particle_oct_container.pyx":90
 * ctypedef cmap[np.uint64_t, bool_array] CoarseRefinedSets
 * 
 * cdef class ParticleOctreeContainer(OctreeContainer):             # <<<<<<<<<<<<<<
 *     cdef Oct** oct_list
 *     #The starting oct index of each domain
 */

struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer {
  struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctreeContainer __pyx_base;
  void (*visit_free)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *);
  void (*visit_clear)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *);
  PyObject *(*visit_assign)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_int64_t *, int, int *);
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *(*allocate_oct)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *);
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *(*refine_oct)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_uint64_t, int, __pyx_t_5numpy_uint8_t);
  void (*filter_particles)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_uint64_t *, __pyx_t_5numpy_int64_t, int, __pyx_t_5numpy_uint8_t);
  PyObject *(*visit)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_int64_t *, struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_visit *__pyx_optional_args);
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *(*get_from_index)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *, __pyx_t_5numpy_uint64_t, struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_get_from_index *__pyx_optional_args);
};
static struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_vtabptr_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer;


/* "yt/geometry/particle_oct_container.pyx":433
 * cdef np.uint64_t FLAG = ~(<np.uint64_t>0)
 * 
 * cdef class ParticleBitmap:             # <<<<<<<<<<<<<<
 *     cdef np.float64_t left_edge[3]
 *     cdef np.float64_t right_edge[3]
 */

struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmap {
  void (*_update_mi2)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *, __pyx_t_5numpy_uint64_t);
  void (*__pyx___set_coarse_index_data_file)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *, __pyx_t_5numpy_uint64_t);
  __pyx_t_5numpy_int64_t (*__pyx___fill_refined_ranges)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_float64_t *, __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_bool_array &);
  void (*__pyx___set_refined_index_data_file)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *, PyArrayObject *, PyArrayObject *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_int64_t);
  void (*__pyx_fuse_0__coarse_index_data_file)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *, PyArrayObject *, PyArrayObject *, __pyx_t_5numpy_uint64_t);
  void (*__pyx_fuse_1__coarse_index_data_file)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *, PyArrayObject *, PyArrayObject *, __pyx_t_5numpy_uint64_t);
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *(*__pyx_fuse_0__refined_index_data_file)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, PyArrayObject *, PyArrayObject *, PyArrayObject *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint8_t);
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *(*__pyx_fuse_1__refined_index_data_file)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, PyArrayObject *, PyArrayObject *, PyArrayObject *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint8_t);
};
static struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_vtabptr_2yt_8geometry_22particle_oct_container_ParticleBitmap;


/* "yt/geometry/particle_oct_container.pyx":1442
 *         return octree
 * 
 * cdef class ParticleBitmapSelector:             # <<<<<<<<<<<<<<
 *     cdef SelectorObject selector
 *     cdef ParticleBitmap bitmap
 */

struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector {
  int (*is_refined)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t);
  int (*is_refined_files)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t);
  void (*add_coarse)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t, struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_coarse *__pyx_optional_args);
  void (*set_files_coarse)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t);
  int (*add_refined)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_refined *__pyx_optional_args);
  void (*set_files_refined)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t);
  void (*add_neighbors_coarse)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t);
  void (*set_files_neighbors_coarse)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t);
  void (*add_neighbors_refined)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t);
  void (*set_files_neighbors_refined)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t);
  void (*set_coarse_list)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *);
  void (*set_refined_list)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *);
  void (*set_coarse_bool)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *);
  void (*set_refined_bool)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *);
  void (*push_refined_bool)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t);
  void (*add_ghost_zones)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *);
  int (*fill_subcells_mi1)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t *);
  int (*fill_subcells_mi2)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t *);
  int (*recursive_morton_mask)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_int32_t, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t *);
  void (*recursive_morton_files)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_int32_t, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_uint64_t);
};
static struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_vtabptr_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector;


/* "yt/geometry/particle_oct_container.pyx":1965
 *                         self.set_files_refined(mi1,mi2)
 * 
 * cdef class ParticleBitmapOctreeContainer(SparseOctreeContainer):             # <<<<<<<<<<<<<<
 *     cdef Oct** oct_list
 *     cdef public int max_level
 */

struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer {
  struct __pyx_vtabstruct_2yt_8geometry_13oct_container_SparseOctreeContainer __pyx_base;
  PyObject *(*visit_assign)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_int64_t *, int, int *, __pyx_t_5numpy_int64_t);
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *(*allocate_oct)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *);
  void (*_get_index_base_octs)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *, __Pyx_memviewslice, __Pyx_memviewslice);
  void (*visit_free)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, int);
  void (*recursive_add)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, PyArrayObject *, int, int *, int, int *);
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *(*refine_oct)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_uint64_t, int);
  void (*filter_particles)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_uint64_t *, __pyx_t_5numpy_int64_t, int);
};
static struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_vtabptr_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer;


/* "View.MemoryView":106
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_array {
  PyObject *(*get_memview)(struct __pyx_array_obj *);
};
static struct __pyx_vtabstruct_array *__pyx_vtabptr_array;


/* "View.MemoryView":331
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */

struct __pyx_vtabstruct_memoryview {
  char *(*get_item_pointer)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*is_slice)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_slice_assignment)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*setitem_slice_assign_scalar)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_indexed)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*convert_item_to_object)(struct __pyx_memoryview_obj *, char *);
  PyObject *(*assign_item_from_object)(struct __pyx_memoryview_obj *, char *, PyObject *);
};
static struct __pyx_vtabstruct_memoryview *__pyx_vtabptr_memoryview;


/* "View.MemoryView":967
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */

struct __pyx_vtabstruct__memoryviewslice {
  struct __pyx_vtabstruct_memoryview __pyx_base;
};
static struct __pyx_vtabstruct__memoryviewslice *__pyx_vtabptr__memoryviewslice;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* BufferGetAndValidate.proto */
#define __Pyx_GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack)\
    ((obj == Py_None || obj == NULL) ?\
    (__Pyx_ZeroBuffer(buf), 0) :\
    __Pyx__GetBufferAndValidate(buf, obj, dtype, flags, nd, cast, stack))
static int  __Pyx__GetBufferAndValidate(Py_buffer* buf, PyObject* obj,
    __Pyx_TypeInfo* dtype, int flags, int nd, int cast, __Pyx_BufFmt_StackElem* stack);
static void __Pyx_ZeroBuffer(Py_buffer* buf);
static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info);
static Py_ssize_t __Pyx_minusones[] = { -1, -1, -1, -1, -1, -1, -1, -1 };
static Py_ssize_t __Pyx_zeros[] = { 0, 0, 0, 0, 0, 0, 0, 0 };

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

#define __Pyx_BufPtrStrided1d(type, buf, i0, s0) (type)((char*)buf + i0 * s0)
/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* BufferFallbackError.proto */
static void __Pyx_RaiseBufferFallbackError(void);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* MemviewSliceInit.proto */
#define __Pyx_BUF_MAX_NDIMS %(BUF_MAX_NDIMS)d
#define __Pyx_MEMVIEW_DIRECT   1
#define __Pyx_MEMVIEW_PTR      2
#define __Pyx_MEMVIEW_FULL     4
#define __Pyx_MEMVIEW_CONTIG   8
#define __Pyx_MEMVIEW_STRIDED  16
#define __Pyx_MEMVIEW_FOLLOW   32
#define __Pyx_IS_C_CONTIG 1
#define __Pyx_IS_F_CONTIG 2
static int __Pyx_init_memviewslice(
                struct __pyx_memoryview_obj *memview,
                int ndim,
                __Pyx_memviewslice *memviewslice,
                int memview_is_new_reference);
static CYTHON_INLINE int __pyx_add_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
static CYTHON_INLINE int __pyx_sub_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
#define __pyx_get_slice_count_pointer(memview) (memview->acquisition_count_aligned_p)
#define __pyx_get_slice_count(memview) (*__pyx_get_slice_count_pointer(memview))
#define __PYX_INC_MEMVIEW(slice, have_gil) __Pyx_INC_MEMVIEW(slice, have_gil, __LINE__)
#define __PYX_XDEC_MEMVIEW(slice, have_gil) __Pyx_XDEC_MEMVIEW(slice, have_gil, __LINE__)
static CYTHON_INLINE void __Pyx_INC_MEMVIEW(__Pyx_memviewslice *, int, int);
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *, int, int);

/* BufferIndexError.proto */
static void __Pyx_RaiseBufferIndexError(int axis);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* UnicodeAsUCS4.proto */
static CYTHON_INLINE Py_UCS4 __Pyx_PyUnicode_AsPy_UCS4(PyObject*);

/* object_ord.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyObject_Ord(c)\
    (likely(PyUnicode_Check(c)) ? (long)__Pyx_PyUnicode_AsPy_UCS4(c) : __Pyx__PyObject_Ord(c))
#else
#define __Pyx_PyObject_Ord(c) __Pyx__PyObject_Ord(c)
#endif
static long __Pyx__PyObject_Ord(PyObject* c);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

#define __Pyx_BufPtrStrided2d(type, buf, i0, s0, i1, s1) (type)((char*)buf + i0 * s0 + i1 * s1)
/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_str(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_str(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* IncludeStringH.proto */
#include <string.h>

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* DivInt[Py_ssize_t].proto */
static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t, Py_ssize_t);

/* UnaryNegOverflows.proto */
#define UNARY_NEG_WOULD_OVERFLOW(x)\
        (((x) < 0) & ((unsigned long)(x) == 0-(unsigned long)(x)))

static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *); /*proto*/
/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* DivInt[long].proto */
static CYTHON_INLINE long __Pyx_div_long(long, long);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* StringJoin.proto */
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION < 3
    #define __Pyx_PyBytes_Join _PyString_Join
    #else
    #define __Pyx_PyBytes_Join _PyBytes_Join
    #endif
#else
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);
#endif

/* PyObject_Unicode.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyObject_Unicode(obj)\
    (likely(PyUnicode_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))
#else
#define __Pyx_PyObject_Unicode(obj)\
    (likely(PyUnicode_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Unicode(obj))
#endif

/* CallNextTpDealloc.proto */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_dealloc);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* FusedFunction.proto */
typedef struct {
    __pyx_CyFunctionObject func;
    PyObject *__signatures__;
    PyObject *type;
    PyObject *self;
} __pyx_FusedFunctionObject;
static PyObject *__pyx_FusedFunction_New(PyMethodDef *ml, int flags,
                                         PyObject *qualname, PyObject *closure,
                                         PyObject *module, PyObject *globals,
                                         PyObject *code);
static int __pyx_FusedFunction_clear(__pyx_FusedFunctionObject *self);
static PyTypeObject *__pyx_FusedFunctionType = NULL;
static int __pyx_FusedFunction_init(void);
#define __Pyx_FusedFunction_USED

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

/* MemviewSliceIsContig.proto */
static int __pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim);

/* OverlappingSlices.proto */
static int __pyx_slices_overlap(__Pyx_memviewslice *slice1,
                                __Pyx_memviewslice *slice2,
                                int ndim, size_t itemsize);

/* Capsule.proto */
static CYTHON_INLINE PyObject *__pyx_capsule_create(void *p, const char *sig);

/* TypeInfoCompare.proto */
static int __pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b);

/* MemviewSliceValidateAndInit.proto */
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_float(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_double(PyObject *, int writable_flag);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint8_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(PyObject *, int writable_flag);

/* Print.proto */
static int __Pyx_Print(PyObject*, PyObject *, int);
#if CYTHON_COMPILING_IN_PYPY || PY_MAJOR_VERSION >= 3
static PyObject* __pyx_print = 0;
static PyObject* __pyx_print_kwargs = 0;
#endif

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5numpy_uint8_t(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5numpy_uint8_t(const char *itemp, PyObject *obj);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_nn___pyx_t_5numpy_uint32_t(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_nn___pyx_t_5numpy_uint64_t(PyObject *, int writable_flag);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5numpy_int64_t(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5numpy_int64_t(const char *itemp, PyObject *obj);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5numpy_uint64_t(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5numpy_uint64_t(const char *itemp, PyObject *obj);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* None.proto */
#include <new>

/* MemviewSliceCopyTemplate.proto */
static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE npy_uint8 __Pyx_PyInt_As_npy_uint8(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE npy_uint64 __Pyx_PyInt_As_npy_uint64(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE npy_int64 __Pyx_PyInt_As_npy_int64(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE npy_uint32 __Pyx_PyInt_As_npy_uint32(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int64(npy_int64 value);

/* CIntFromPy.proto */
static CYTHON_INLINE npy_int32 __Pyx_PyInt_As_npy_int32(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int32(npy_int32 value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_uint64(npy_uint64 value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_Py_intptr_t(Py_intptr_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_uint8(npy_uint8 value);

/* TypeInfoToFormat.proto */
struct __pyx_typeinfo_string {
    char string[3];
};
static struct __pyx_typeinfo_string __Pyx_TypeInfoToFormat(__Pyx_TypeInfo *type);

/* BytesContains.proto */
static CYTHON_INLINE int __Pyx_BytesContains(PyObject* bytes, char character);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_uint32(npy_uint32 value);

/* PrintOne.proto */
static int __Pyx_PrintOne(PyObject* stream, PyObject *o);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value);

/* ImportNumPyArray.proto */
static PyObject *__pyx_numpy_ndarray = NULL;
static PyObject* __Pyx_ImportNumPyArrayTypeIfAvailable(void);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static void __pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_visit_free(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_visit_clear(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o); /* proto*/
static PyObject *__pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_visit_assign(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o, __pyx_t_5numpy_int64_t *__pyx_v_lpos, int __pyx_v_level, int *__pyx_v_max_level); /* proto*/
static __pyx_t_5numpy_int64_t __pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_get_domain_offset(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, CYTHON_UNUSED int __pyx_v_domain_id); /* proto*/
static struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_allocate_oct(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self); /* proto*/
static struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_refine_oct(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o, __pyx_t_5numpy_uint64_t __pyx_v_index, int __pyx_v_level, __pyx_t_5numpy_uint8_t __pyx_v_order); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_filter_particles(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o, __pyx_t_5numpy_uint64_t *__pyx_v_data, __pyx_t_5numpy_int64_t __pyx_v_p, int __pyx_v_level, __pyx_t_5numpy_uint8_t __pyx_v_order); /* proto*/
static PyObject *__pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_visit(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o, __pyx_t_5numpy_int64_t *__pyx_v_counts, struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_visit *__pyx_optional_args); /* proto*/
static struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_get_from_index(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi, struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_get_from_index *__pyx_optional_args); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap__update_mi2(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_index_order2); /* proto*/
static void __pyx_fuse_0__pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___coarse_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyArrayObject *__pyx_v_pos, PyArrayObject *__pyx_v_hsml, __pyx_t_5numpy_uint64_t __pyx_v_file_id); /* proto*/
static void __pyx_fuse_1__pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___coarse_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyArrayObject *__pyx_v_pos, PyArrayObject *__pyx_v_hsml, __pyx_t_5numpy_uint64_t __pyx_v_file_id); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___set_coarse_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_file_id); /* proto*/
static struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_fuse_0__pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___refined_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_in_collection, PyArrayObject *__pyx_v_pos, PyArrayObject *__pyx_v_hsml, PyArrayObject *__pyx_v_mask, __pyx_t_5numpy_uint64_t __pyx_v_count_threshold, __pyx_t_5numpy_uint8_t __pyx_v_mask_threshold); /* proto*/
static struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_fuse_1__pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___refined_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_in_collection, PyArrayObject *__pyx_v_pos, PyArrayObject *__pyx_v_hsml, PyArrayObject *__pyx_v_mask, __pyx_t_5numpy_uint64_t __pyx_v_count_threshold, __pyx_t_5numpy_uint8_t __pyx_v_mask_threshold); /* proto*/
static __pyx_t_5numpy_int64_t __pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___fill_refined_ranges(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_s_ppos, __pyx_t_5numpy_float64_t __pyx_v_radius, __pyx_t_5numpy_float64_t *__pyx_v_LE, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_RE, __pyx_t_5numpy_float64_t *__pyx_v_dds1, __pyx_t_5numpy_uint64_t __pyx_v_xex, __pyx_t_5numpy_uint64_t __pyx_v_yex, __pyx_t_5numpy_uint64_t __pyx_v_zex, __pyx_t_5numpy_float64_t *__pyx_v_dds2, __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_bool_array &__pyx_v_refined_set); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___set_refined_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyArrayObject *__pyx_v_sub_mi1, PyArrayObject *__pyx_v_sub_mi2, __pyx_t_5numpy_uint64_t __pyx_v_file_id, __pyx_t_5numpy_int64_t __pyx_v_nsub_mi); /* proto*/
static int __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_is_refined(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1); /* proto*/
static int __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_is_refined_files(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_coarse(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1, struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_coarse *__pyx_optional_args); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_files_coarse(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1); /* proto*/
static int __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_refined(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1, __pyx_t_5numpy_uint64_t __pyx_v_mi2, struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_refined *__pyx_optional_args); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_files_refined(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1, __pyx_t_5numpy_uint64_t __pyx_v_mi2); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_neighbors_coarse(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_files_neighbors_coarse(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_neighbors_refined(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1, __pyx_t_5numpy_uint64_t __pyx_v_mi2); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_files_neighbors_refined(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1, __pyx_t_5numpy_uint64_t __pyx_v_mi2); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_coarse_list(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_s, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_g); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_refined_list(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_s, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_g); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_coarse_bool(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_s, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_g); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_refined_bool(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_s, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_g); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_push_refined_bool(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_ghost_zones(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_s, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_g); /* proto*/
static int __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_fill_subcells_mi1(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_nlevel, __pyx_t_5numpy_uint64_t *__pyx_v_ind1); /* proto*/
static int __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_fill_subcells_mi2(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_nlevel, __pyx_t_5numpy_uint64_t __pyx_v_mi1, __pyx_t_5numpy_uint64_t *__pyx_v_ind2); /* proto*/
static int __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_recursive_morton_mask(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_int32_t __pyx_v_level, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds, __pyx_t_5numpy_uint64_t __pyx_v_mi1, __pyx_t_5numpy_uint64_t *__pyx_v_cur_ind); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_recursive_morton_files(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_int32_t __pyx_v_level, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds, __pyx_t_5numpy_uint64_t __pyx_v_mi1); /* proto*/
static PyObject *__pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_visit_assign(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o, __pyx_t_5numpy_int64_t *__pyx_v_lpos, int __pyx_v_level, int *__pyx_v_max_level, __pyx_t_5numpy_int64_t __pyx_v_index_root); /* proto*/
static struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_allocate_oct(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer__get_index_base_octs(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, __Pyx_memviewslice __pyx_v_ind, __Pyx_memviewslice __pyx_v_domain_ind); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_visit_free(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o, int __pyx_v_free_this); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_recursive_add(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o, PyArrayObject *__pyx_v_indices, int __pyx_v_level, int *__pyx_v_max_level, int __pyx_v_domain_id, int *__pyx_v_count); /* proto*/
static struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_refine_oct(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o, __pyx_t_5numpy_uint64_t __pyx_v_index, int __pyx_v_level); /* proto*/
static void __pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_filter_particles(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o, __pyx_t_5numpy_uint64_t *__pyx_v_data, __pyx_t_5numpy_int64_t __pyx_v_p, int __pyx_v_level); /* proto*/
static CYTHON_INLINE struct __pyx_t_2yt_8geometry_13oct_container_OctAllocationContainer *__pyx_f_2yt_8geometry_13oct_container_13OctObjectPool_get_cont(struct __pyx_obj_2yt_8geometry_13oct_container_OctObjectPool *__pyx_v_self, int __pyx_v_i); /* proto*/
static CYTHON_INLINE int __pyx_f_2yt_8geometry_12oct_visitors_10OctVisitor_oind(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_2yt_8geometry_12oct_visitors_10OctVisitor_rind(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *__pyx_v_self); /* proto*/
static CYTHON_INLINE __pyx_t_5numpy_uint8_t __pyx_f_2yt_8geometry_12oct_visitors_20BaseNeighbourVisitor_neighbour_rind(struct __pyx_obj_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor *__pyx_v_self); /* proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self); /* proto*/
static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto*/
static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/

/* Module declarations from 'libc.math' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'libcpp.utility' */

/* Module declarations from 'libcpp.map' */

/* Module declarations from 'libcpp.vector' */

/* Module declarations from 'libc.stdint' */

/* Module declarations from 'libcpp' */

/* Module declarations from 'libcpp.string' */

/* Module declarations from 'cython.view' */
static struct __pyx_array_obj *__pyx_array_new(PyObject *, Py_ssize_t, char *, char *, char *); /*proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *, char *, char *, int *); /*proto*/

/* Module declarations from 'yt.utilities.lib.ewah_bool_array' */

/* Module declarations from 'yt.geometry.grid_visitors' */
static __pyx_t_5numpy_uint8_t (*__pyx_f_2yt_8geometry_13grid_visitors_check_child_masked)(struct __pyx_t_2yt_8geometry_13grid_visitors_GridVisitorData *); /*proto*/

/* Module declarations from 'yt.utilities.lib.fp_utils' */
static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_imax(__pyx_t_5numpy_int64_t, __pyx_t_5numpy_int64_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_fmax(__pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_imin(__pyx_t_5numpy_int64_t, __pyx_t_5numpy_int64_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_fmin(__pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_i64min(__pyx_t_5numpy_int64_t, __pyx_t_5numpy_int64_t); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_2yt_9utilities_3lib_8fp_utils__ensure_code(PyObject *); /*proto*/

/* Module declarations from 'libc.float' */

/* Module declarations from 'yt.utilities.lib.geometry_utils' */
static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_ifrexp(__pyx_t_5numpy_float64_t, __pyx_t_5numpy_int64_t *); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_msdb(__pyx_t_5numpy_int64_t, __pyx_t_5numpy_int64_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_xor_msb(__pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_uint64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_spread_64bits_by2(__pyx_t_5numpy_uint64_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_uint64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_compact_64bits_by2(__pyx_t_5numpy_uint64_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_uint64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_encode_morton_64bit(__pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t); /*proto*/
static CYTHON_INLINE void __pyx_f_2yt_9utilities_3lib_14geometry_utils_decode_morton_64bit(__pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t *); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_uint64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton(__pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_int32_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_uint64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_dds(__pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_uint64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_relative_dds(__pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_uint64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_dds(__pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_uint64_t *); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_uint64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_relative_dds(__pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_uint64_t *); /*proto*/
static __pyx_t_5numpy_uint32_t (*__pyx_f_2yt_9utilities_3lib_14geometry_utils_morton_neighbors_coarse)(__pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, int *, __pyx_t_5numpy_uint32_t, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice); /*proto*/
static __pyx_t_5numpy_uint32_t (*__pyx_f_2yt_9utilities_3lib_14geometry_utils_morton_neighbors_refined)(__pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, int *, __pyx_t_5numpy_uint32_t, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice); /*proto*/

/* Module declarations from 'yt.geometry.selection_routines' */
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_AlwaysSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_OctreeSubsetSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_18selection_routines_BooleanSelector = 0;

/* Module declarations from 'yt.utilities.lib.allocation_container' */
static PyTypeObject *__pyx_ptype_2yt_9utilities_3lib_20allocation_container_ObjectPool = 0;

/* Module declarations from 'yt.geometry.oct_container' */
static PyTypeObject *__pyx_ptype_2yt_8geometry_13oct_container_OctObjectPool = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_13oct_container_SparseOctreeContainer = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_13oct_container_RAMSESOctreeContainer = 0;
static int *__pyx_vp_2yt_8geometry_13oct_container_ORDER_MAX = 0;
#define __pyx_v_2yt_8geometry_13oct_container_ORDER_MAX (*__pyx_vp_2yt_8geometry_13oct_container_ORDER_MAX)

/* Module declarations from 'yt.geometry.oct_visitors' */
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_OctVisitor = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_CountTotalOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_CountTotalCells = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_MarkOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_MaskOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_IndexOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_MaskedIndexOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_IndexMaskMapOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_ICoordsOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_IResOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_FCoordsOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_FWidthOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_CopyArrayI64 = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_CopyArrayF64 = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_CopyFileIndArrayI8 = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_IdentifyOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_AssignDomainInd = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_FillFileIndicesO = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_FillFileIndicesR = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_CountByDomain = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_StoreOctree = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_LoadOctree = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_MortonIndexOcts = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_StoreIndex = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_NeighbourCellIndexVisitor = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_12oct_visitors_NeighbourCellVisitor = 0;
static CYTHON_INLINE int __pyx_f_2yt_8geometry_12oct_visitors_cind(int, int, int); /*proto*/

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'numpy.math' */

/* Module declarations from 'yt.geometry.particle_deposit' */
static PyTypeObject *__pyx_ptype_2yt_8geometry_16particle_deposit_ParticleDepositOperation = 0;
static CYTHON_INLINE int __pyx_f_2yt_8geometry_16particle_deposit_gind(int, int, int, int *); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_cubic(__pyx_t_5numpy_float64_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_quartic(__pyx_t_5numpy_float64_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_quintic(__pyx_t_5numpy_float64_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_wendland2(__pyx_t_5numpy_float64_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_wendland4(__pyx_t_5numpy_float64_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_wendland6(__pyx_t_5numpy_float64_t); /*proto*/
static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_dummy(__pyx_t_5numpy_float64_t); /*proto*/

/* Module declarations from 'libcpp.pair' */

/* Module declarations from 'libcpp.set' */

/* Module declarations from 'yt.utilities.lib.ewah_bool_wrap' */
static PyTypeObject *__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks = 0;
static PyTypeObject *__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection = 0;
static PyTypeObject *__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed = 0;
static PyTypeObject *__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet = 0;
static PyTypeObject *__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector = 0;
static PyTypeObject *__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet = 0;
static PyTypeObject *__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector = 0;

/* Module declarations from 'yt.geometry.particle_oct_container' */
static PyTypeObject *__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleBitmap = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer = 0;
static PyTypeObject *__pyx_ptype_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__ = 0;
static PyTypeObject *__pyx_array_type = 0;
static PyTypeObject *__pyx_MemviewEnum_type = 0;
static PyTypeObject *__pyx_memoryview_type = 0;
static PyTypeObject *__pyx_memoryviewslice_type = 0;
static __pyx_t_5numpy_uint64_t __pyx_v_2yt_8geometry_22particle_oct_container_ONEBIT;
static __pyx_t_5numpy_uint64_t __pyx_v_2yt_8geometry_22particle_oct_container_FLAG;
static PyObject *generic = 0;
static PyObject *strided = 0;
static PyObject *indirect = 0;
static PyObject *contiguous = 0;
static PyObject *indirect_contiguous = 0;
static int __pyx_memoryview_thread_locks_used;
static PyThread_type_lock __pyx_memoryview_thread_locks[8];
static void __pyx_f_2yt_8geometry_22particle_oct_container__mask_children(PyArrayObject *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *); /*proto*/
static PyObject *__pyx_convert_vector_to_py_size_t(const std::vector<size_t>  &); /*proto*/
static struct __pyx_array_obj *__pyx_array_new(PyObject *, Py_ssize_t, char *, char *, char *); /*proto*/
static void *__pyx_align_pointer(void *, size_t); /*proto*/
static PyObject *__pyx_memoryview_new(PyObject *, int, int, __Pyx_TypeInfo *); /*proto*/
static CYTHON_INLINE int __pyx_memoryview_check(PyObject *); /*proto*/
static PyObject *_unellipsify(PyObject *, int); /*proto*/
static PyObject *assert_direct_dimensions(Py_ssize_t *, int); /*proto*/
static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *, PyObject *); /*proto*/
static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int, int); /*proto*/
static char *__pyx_pybuffer_index(Py_buffer *, char *, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memslice_transpose(__Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice, int, PyObject *(*)(char *), int (*)(char *, PyObject *), int); /*proto*/
static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *); /*proto*/
static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static Py_ssize_t abs_py_ssize_t(Py_ssize_t); /*proto*/
static char __pyx_get_best_slice_order(__Pyx_memviewslice *, int); /*proto*/
static void _copy_strided_to_strided(char *, Py_ssize_t *, char *, Py_ssize_t *, Py_ssize_t *, Py_ssize_t *, int, size_t); /*proto*/
static void copy_strided_to_strided(__Pyx_memviewslice *, __Pyx_memviewslice *, int, size_t); /*proto*/
static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *, int); /*proto*/
static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *, Py_ssize_t *, Py_ssize_t, int, char); /*proto*/
static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *, __Pyx_memviewslice *, char, int); /*proto*/
static int __pyx_memoryview_err_extents(int, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memoryview_err_dim(PyObject *, char *, int); /*proto*/
static int __pyx_memoryview_err(PyObject *, char *); /*proto*/
static int __pyx_memoryview_copy_contents(__Pyx_memviewslice, __Pyx_memviewslice, int, int, int); /*proto*/
static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *, int, int); /*proto*/
static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *, int, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *, int, size_t, void *, int); /*proto*/
static void __pyx_memoryview__slice_assign_scalar(char *, Py_ssize_t *, Py_ssize_t *, int, size_t, void *); /*proto*/
static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *, PyObject *); /*proto*/
static PyObject *__pyx_format_from_typeinfo(__Pyx_TypeInfo *); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t = { "uint64_t", NULL, sizeof(__pyx_t_5numpy_uint64_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_uint64_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_uint64_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t = { "uint8_t", NULL, sizeof(__pyx_t_5numpy_uint8_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_uint8_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_uint8_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t = { "int64_t", NULL, sizeof(__pyx_t_5numpy_int64_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_int64_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_int64_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_float = { "float", NULL, sizeof(float), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_double = { "double", NULL, sizeof(double), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t = { "float64_t", NULL, sizeof(__pyx_t_5numpy_float64_t), { 0 }, 0, 'R', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_uint32_t = { "uint32_t", NULL, sizeof(__pyx_t_5numpy_uint32_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_uint32_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_uint32_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_object = { "Python object", NULL, sizeof(PyObject *), { 0 }, 0, 'O', 0, 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t = { "int32_t", NULL, sizeof(__pyx_t_5numpy_int32_t), { 0 }, 0, IS_UNSIGNED(__pyx_t_5numpy_int32_t) ? 'U' : 'I', IS_UNSIGNED(__pyx_t_5numpy_int32_t), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t = { "float32_t", NULL, sizeof(__pyx_t_5numpy_float32_t), { 0 }, 0, 'R', 0, 0 };
#define __Pyx_MODULE_NAME "yt.geometry.particle_oct_container"
extern int __pyx_module_is_main_yt__geometry__particle_oct_container;
int __pyx_module_is_main_yt__geometry__particle_oct_container = 0;

/* Implementation of 'yt.geometry.particle_oct_container' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_super;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_open;
static PyObject *__pyx_builtin_OSError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_sum;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_ImportError;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_Ellipsis;
static PyObject *__pyx_builtin_id;
static const char __pyx_k_O[] = "O";
static const char __pyx_k_Q[] = "Q";
static const char __pyx_k_T[] = "T{";
  static const char __pyx_k_c[] = "c";
  static const char __pyx_k_o[] = "o";
  static const char __pyx_k_q[] = "q";
  static const char __pyx_k_s[] = "s";
  static const char __pyx_k__7[] = "()";
  static const char __pyx_k__8[] = "|";
  static const char __pyx_k_id[] = "id";
  static const char __pyx_k_np[] = "np";
  static const char __pyx_k_os[] = "os";
  static const char __pyx_k_rb[] = "rb";
  static const char __pyx_k_u8[] = "u8";
  static const char __pyx_k_wb[] = "wb";
  static const char __pyx_k__12[] = ")";
  static const char __pyx_k__53[] = "^";
  static const char __pyx_k__54[] = "";
  static const char __pyx_k__55[] = ":";
static const char __pyx_k__56[] = "}";
static const char __pyx_k__57[] = ",";
static const char __pyx_k_add[] = "add";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_ngz[] = "ngz";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_out[] = "out";
static const char __pyx_k_pos[] = "pos";
static const char __pyx_k_s_2[] = "(%s)";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_hsml[] = "hsml";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_iter[] = "__iter__";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_kind[] = "kind";
static const char __pyx_k_log2[] = "log2";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mask[] = "mask";
static const char __pyx_k_mm_g[] = "mm_g";
static const char __pyx_k_mm_s[] = "mm_s";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_ndim[] = "ndim";
static const char __pyx_k_none[] = "none";
static const char __pyx_k_open[] = "open";
static const char __pyx_k_pack[] = "pack";
static const char __pyx_k_path[] = "path";
static const char __pyx_k_read[] = "read";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_sort[] = "sort";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_ASCII[] = "ASCII";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_bcoll[] = "bcoll";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_cmask[] = "cmask";
static const char __pyx_k_cubic[] = "cubic";
static const char __pyx_k_dmask[] = "dmask";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_float[] = "float";
static const char __pyx_k_ifile[] = "ifile";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_int32[] = "int32";
static const char __pyx_k_int64[] = "int64";
static const char __pyx_k_mask2[] = "mask2";
static const char __pyx_k_n_ref[] = "n_ref";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_order[] = "order";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_strip[] = "strip";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_uint8[] = "uint8";
static const char __pyx_k_units[] = "units";
static const char __pyx_k_where[] = "where";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_zeros[] = "zeros";
static const char __pyx_k_astype[] = "astype";
static const char __pyx_k_bitmap[] = "bitmap";
static const char __pyx_k_counts[] = "counts";
static const char __pyx_k_double[] = "double";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_finish[] = "finish";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_isfile[] = "isfile";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_nfiles[] = "nfiles";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_order1[] = "order1";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_uint32[] = "uint32";
static const char __pyx_k_uint64[] = "uint64";
static const char __pyx_k_unpack[] = "unpack";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_OSError[] = "OSError";
static const char __pyx_k_argsort[] = "argsort";
static const char __pyx_k_dataset[] = "dataset";
static const char __pyx_k_file_id[] = "file_id";
static const char __pyx_k_float32[] = "float32";
static const char __pyx_k_float64[] = "float64";
static const char __pyx_k_fortran[] = "fortran";
static const char __pyx_k_get_DLE[] = "get_DLE";
static const char __pyx_k_get_DRE[] = "get_DRE";
static const char __pyx_k_in_coll[] = "_in_coll";
static const char __pyx_k_indices[] = "indices";
static const char __pyx_k_memview[] = "memview";
static const char __pyx_k_nsub_mi[] = "nsub_mi";
static const char __pyx_k_quartic[] = "quartic";
static const char __pyx_k_quintic[] = "quintic";
static const char __pyx_k_sub_mi1[] = "sub_mi1";
static const char __pyx_k_sub_mi2[] = "sub_mi2";
static const char __pyx_k_verbose[] = "verbose";
static const char __pyx_k_Ellipsis[] = "Ellipsis";
static const char __pyx_k_File_03d[] = "File: %03d";
static const char __pyx_k_calcsize[] = "calcsize";
static const char __pyx_k_defaults[] = "defaults";
static const char __pyx_k_file_idx[] = "file_idx";
static const char __pyx_k_finalize[] = "finalize";
static const char __pyx_k_get_pbar[] = "get_pbar";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_itemsize[] = "itemsize";
static const char __pyx_k_num_root[] = "num_root";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_selector[] = "selector";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_yt_funcs[] = "yt.funcs";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_base_mask[] = "base_mask";
static const char __pyx_k_domain_id[] = "domain_id";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_file_hash[] = "file_hash";
static const char __pyx_k_file_list[] = "file_list";
static const char __pyx_k_fill_bool[] = "_fill_bool";
static const char __pyx_k_left_edge[] = "left_edge";
static const char __pyx_k_masklevel[] = "masklevel";
static const char __pyx_k_num_zones[] = "num_zones";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_wendland2[] = "wendland2";
static const char __pyx_k_wendland4[] = "wendland4";
static const char __pyx_k_wendland6[] = "wendland6";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_data_files[] = "data_files";
static const char __pyx_k_domain_ind[] = "domain_ind";
static const char __pyx_k_fill_masks[] = "fill_masks";
static const char __pyx_k_io_handler[] = "io_handler";
static const char __pyx_k_print_info[] = "print_info";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_right_edge[] = "right_edge";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_code_length[] = "code_length";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_count_total[] = "count_total";
static const char __pyx_k_defaultdict[] = "defaultdict";
static const char __pyx_k_exceeds_max[] = ") exceeds max (";
static const char __pyx_k_file_mask_g[] = "file_mask_g";
static const char __pyx_k_file_mask_p[] = "file_mask_p";
static const char __pyx_k_periodicity[] = "periodicity";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_count_coarse[] = "count_coarse";
static const char __pyx_k_get_bitmasks[] = "get_bitmasks";
static const char __pyx_k_index_order1[] = "index_order1";
static const char __pyx_k_index_order2[] = "index_order2";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_allocate_root[] = "allocate_root";
static const char __pyx_k_coarse_ghosts[] = "coarse_ghosts";
static const char __pyx_k_count_refined[] = "count_refined";
static const char __pyx_k_in_collection[] = "in_collection";
static const char __pyx_k_mask_to_files[] = "mask_to_files";
static const char __pyx_k_pyx_getbuffer[] = "__pyx_getbuffer";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_save_bitmasks[] = "save_bitmasks";
static const char __pyx_k_selector_mask[] = "selector_mask";
static const char __pyx_k_ParticleBitmap[] = "ParticleBitmap";
static const char __pyx_k_mask_threshold[] = "mask_threshold";
static const char __pyx_k_masks_to_files[] = "masks_to_files";
static const char __pyx_k_out_collection[] = "out_collection";
static const char __pyx_k_View_MemoryView[] = "View.MemoryView";
static const char __pyx_k_allocate_buffer[] = "allocate_buffer";
static const char __pyx_k_bitmask_version[] = "_bitmask_version";
static const char __pyx_k_count_threshold[] = "count_threshold";
static const char __pyx_k_dtype_is_object[] = "dtype_is_object";
static const char __pyx_k_get_periodicity[] = "get_periodicity";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_total_particles[] = "total_particles";
static const char __pyx_k_allocate_domains[] = "allocate_domains";
static const char __pyx_k_convert_to_units[] = "convert_to_units";
static const char __pyx_k_domain_left_edge[] = "domain_left_edge";
static const char __pyx_k_max_index_order2[] = "max_index_order2";
static const char __pyx_k_No_root_found_for[] = "No root found for {},{},{}";
static const char __pyx_k_domain_dimensions[] = "domain_dimensions";
static const char __pyx_k_domain_right_edge[] = "domain_right_edge";
static const char __pyx_k_pyx_unpickle_Enum[] = "__pyx_unpickle_Enum";
static const char __pyx_k_yield_coordinates[] = "_yield_coordinates";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_coarse_ghosts_list[] = "coarse_ghosts_list";
static const char __pyx_k_coarse_select_list[] = "coarse_select_list";
static const char __pyx_k_strided_and_direct[] = "<strided and direct>";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_characteristic_size[] = "characteristic_size";
static const char __pyx_k_strided_and_indirect[] = "<strided and indirect>";
static const char __pyx_k_contiguous_and_direct[] = "<contiguous and direct>";
static const char __pyx_k_Loading_particle_index[] = "Loading particle index";
static const char __pyx_k_MemoryView_of_r_object[] = "<MemoryView of %r object>";
static const char __pyx_k_ParticleBitmapSelector[] = "ParticleBitmapSelector";
static const char __pyx_k_coarse_index_data_file[] = "_coarse_index_data_file";
static const char __pyx_k_is_negative_with_value[] = " is negative with value ";
static const char __pyx_k_MemoryView_of_r_at_0x_x[] = "<MemoryView of %r at 0x%x>";
static const char __pyx_k_ParticleOctreeContainer[] = "ParticleOctreeContainer";
static const char __pyx_k_contiguous_and_indirect[] = "<contiguous and indirect>";
static const char __pyx_k_refined_index_data_file[] = "_refined_index_data_file";
static const char __pyx_k_Cannot_index_with_type_s[] = "Cannot index with type '%s'";
static const char __pyx_k_Invalid_shape_in_axis_d_d[] = "Invalid shape in axis %d: %d.";
static const char __pyx_k_No_matching_signature_found[] = "No matching signature found";
static const char __pyx_k_itemsize_0_for_cython_array[] = "itemsize <= 0 for cython.array";
static const char __pyx_k_ndarray_is_not_C_contiguous[] = "ndarray is not C contiguous";
static const char __pyx_k_Index_for_a_softening_region[] = "Index for a softening region (";
static const char __pyx_k_Total_of_s_s_collisions_3_5f[] = "Total of %s / %s collisions (% 3.5f%%)";
static const char __pyx_k_Child_was_already_initialized[] = "Child was already initialized...";
static const char __pyx_k_ParticleBitmapOctreeContainer[] = "ParticleBitmapOctreeContainer";
static const char __pyx_k_Smoothing_length_for_particle[] = "Smoothing length for particle ";
static const char __pyx_k_unable_to_allocate_array_data[] = "unable to allocate array data.";
static const char __pyx_k_ParticleOctreeContainer___iter[] = "ParticleOctreeContainer.__iter__";
static const char __pyx_k_strided_and_direct_or_indirect[] = "<strided and direct or indirect>";
static const char __pyx_k_Oct_container_tuned_for_Particl[] = "\nOct container tuned for Particles\n\n\n\n\n";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_unknown_dtype_code_in_numpy_pxd[] = "unknown dtype code in numpy.pxd (%d)";
static const char __pyx_k_Buffer_view_does_not_expose_stri[] = "Buffer view does not expose strides";
static const char __pyx_k_Can_only_create_a_buffer_that_is[] = "Can only create a buffer that is contiguous in memory.";
static const char __pyx_k_Cannot_assign_to_read_only_memor[] = "Cannot assign to read-only memoryview";
static const char __pyx_k_Cannot_create_writable_memory_vi[] = "Cannot create writable memory view from read-only memoryview";
static const char __pyx_k_Empty_shape_tuple_for_cython_arr[] = "Empty shape tuple for cython.array";
static const char __pyx_k_Expected_at_least_d_argument_s_g[] = "Expected at least %d argument%s, got %d";
static const char __pyx_k_Format_string_allocated_too_shor[] = "Format string allocated too short, see comment in numpy.pxd";
static const char __pyx_k_Function_call_with_ambiguous_arg[] = "Function call with ambiguous argument types";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))";
static const char __pyx_k_Indirect_dimensions_not_supporte[] = "Indirect dimensions not supported";
static const char __pyx_k_Invalid_mode_expected_c_or_fortr[] = "Invalid mode, expected 'c' or 'fortran', got %s";
static const char __pyx_k_Non_native_byte_order_not_suppor[] = "Non-native byte order not supported";
static const char __pyx_k_Number_of_bitmasks_conflicts_wit[] = "Number of bitmasks ({}) conflicts with number of files ({})";
static const char __pyx_k_Octree_does_not_have_roots_along[] = "Octree does not have roots along dimension {} in a power of 2 ";
static const char __pyx_k_Octree_must_have_the_same_number[] = "Octree must have the same number of roots in each dimension for this.";
static const char __pyx_k_Out_of_bounds_on_buffer_access_a[] = "Out of bounds on buffer access (axis %d)";
static const char __pyx_k_ParticleBitmap__coarse_index_dat[] = "ParticleBitmap._coarse_index_data_file";
static const char __pyx_k_ParticleBitmap__refined_index_da[] = "ParticleBitmap._refined_index_data_file";
static const char __pyx_k_The_file_format_of_the_index_has[] = "The file format of the index has changed since this file was created. It will be replaced with an updated version.";
static const char __pyx_k_Unable_to_convert_item_to_object[] = "Unable to convert item to object";
static const char __pyx_k_got_differing_extents_in_dimensi[] = "got differing extents in dimension %d (got %d and %d)";
static const char __pyx_k_ndarray_is_not_Fortran_contiguou[] = "ndarray is not Fortran contiguous";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_self__ptr_octs_per_root_self_oct[] = "self._ptr_octs_per_root,self.oct_list,self.root_mesh,self.root_nodes,self.tree_root cannot be converted to a Python object for pickling";
static const char __pyx_k_self_dom_offsets_self_oct_list_s[] = "self.dom_offsets,self.oct_list,self.root_mesh cannot be converted to a Python object for pickling";
static const char __pyx_k_self_file_markers_cannot_be_conv[] = "self.file_markers cannot be converted to a Python object for pickling";
static const char __pyx_k_unable_to_allocate_shape_and_str[] = "unable to allocate shape and strides.";
static const char __pyx_k_yt_geometry_particle_oct_contain[] = "yt.geometry.particle_oct_container";
static const char __pyx_k_Format_string_allocated_too_shor_2[] = "Format string allocated too short.";
static const char __pyx_k_yt_geometry_particle_oct_contain_2[] = "yt/geometry/particle_oct_container.pyx";
static PyObject *__pyx_n_s_ASCII;
static PyObject *__pyx_kp_s_Buffer_view_does_not_expose_stri;
static PyObject *__pyx_kp_s_Can_only_create_a_buffer_that_is;
static PyObject *__pyx_kp_s_Cannot_assign_to_read_only_memor;
static PyObject *__pyx_kp_s_Cannot_create_writable_memory_vi;
static PyObject *__pyx_kp_s_Cannot_index_with_type_s;
static PyObject *__pyx_kp_s_Child_was_already_initialized;
static PyObject *__pyx_n_s_Ellipsis;
static PyObject *__pyx_kp_s_Empty_shape_tuple_for_cython_arr;
static PyObject *__pyx_kp_s_Expected_at_least_d_argument_s_g;
static PyObject *__pyx_kp_s_File_03d;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor_2;
static PyObject *__pyx_kp_s_Function_call_with_ambiguous_arg;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_kp_u_Index_for_a_softening_region;
static PyObject *__pyx_kp_s_Indirect_dimensions_not_supporte;
static PyObject *__pyx_kp_s_Invalid_mode_expected_c_or_fortr;
static PyObject *__pyx_kp_s_Invalid_shape_in_axis_d_d;
static PyObject *__pyx_kp_s_Loading_particle_index;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_kp_s_MemoryView_of_r_at_0x_x;
static PyObject *__pyx_kp_s_MemoryView_of_r_object;
static PyObject *__pyx_kp_s_No_matching_signature_found;
static PyObject *__pyx_kp_s_No_root_found_for;
static PyObject *__pyx_kp_u_Non_native_byte_order_not_suppor;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_kp_s_Number_of_bitmasks_conflicts_wit;
static PyObject *__pyx_n_b_O;
static PyObject *__pyx_n_s_OSError;
static PyObject *__pyx_kp_s_Octree_does_not_have_roots_along;
static PyObject *__pyx_kp_s_Octree_must_have_the_same_number;
static PyObject *__pyx_kp_s_Out_of_bounds_on_buffer_access_a;
static PyObject *__pyx_n_s_ParticleBitmap;
static PyObject *__pyx_n_s_ParticleBitmapOctreeContainer;
static PyObject *__pyx_n_s_ParticleBitmapSelector;
static PyObject *__pyx_n_s_ParticleBitmap__coarse_index_dat;
static PyObject *__pyx_n_s_ParticleBitmap__refined_index_da;
static PyObject *__pyx_n_s_ParticleOctreeContainer;
static PyObject *__pyx_n_s_ParticleOctreeContainer___iter;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_Q;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_kp_u_Smoothing_length_for_particle;
static PyObject *__pyx_kp_b_T;
static PyObject *__pyx_kp_s_The_file_format_of_the_index_has;
static PyObject *__pyx_kp_s_Total_of_s_s_collisions_3_5f;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_Unable_to_convert_item_to_object;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_View_MemoryView;
static PyObject *__pyx_kp_u__12;
static PyObject *__pyx_kp_b__53;
static PyObject *__pyx_kp_b__54;
static PyObject *__pyx_kp_b__55;
static PyObject *__pyx_kp_b__56;
static PyObject *__pyx_kp_u__57;
static PyObject *__pyx_kp_s__7;
static PyObject *__pyx_kp_s__8;
static PyObject *__pyx_n_s_add;
static PyObject *__pyx_n_s_allocate_buffer;
static PyObject *__pyx_n_s_allocate_domains;
static PyObject *__pyx_n_s_allocate_root;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_argsort;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_astype;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_s_base_mask;
static PyObject *__pyx_n_s_bcoll;
static PyObject *__pyx_n_s_bitmap;
static PyObject *__pyx_n_s_bitmask_version;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_u_c;
static PyObject *__pyx_n_s_calcsize;
static PyObject *__pyx_n_s_characteristic_size;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_cmask;
static PyObject *__pyx_n_s_coarse_ghosts;
static PyObject *__pyx_n_s_coarse_ghosts_list;
static PyObject *__pyx_n_s_coarse_index_data_file;
static PyObject *__pyx_n_s_coarse_select_list;
static PyObject *__pyx_n_s_code_length;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_kp_s_contiguous_and_direct;
static PyObject *__pyx_kp_s_contiguous_and_indirect;
static PyObject *__pyx_n_s_convert_to_units;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_count_coarse;
static PyObject *__pyx_n_s_count_refined;
static PyObject *__pyx_n_s_count_threshold;
static PyObject *__pyx_n_s_count_total;
static PyObject *__pyx_n_s_counts;
static PyObject *__pyx_n_s_cubic;
static PyObject *__pyx_n_s_data_files;
static PyObject *__pyx_n_s_dataset;
static PyObject *__pyx_n_s_defaultdict;
static PyObject *__pyx_n_s_defaults;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_dmask;
static PyObject *__pyx_n_s_domain_dimensions;
static PyObject *__pyx_n_s_domain_id;
static PyObject *__pyx_n_s_domain_ind;
static PyObject *__pyx_n_s_domain_left_edge;
static PyObject *__pyx_n_s_domain_right_edge;
static PyObject *__pyx_n_s_double;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_dtype_is_object;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_kp_u_exceeds_max;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_file_hash;
static PyObject *__pyx_n_s_file_id;
static PyObject *__pyx_n_s_file_idx;
static PyObject *__pyx_n_s_file_list;
static PyObject *__pyx_n_s_file_mask_g;
static PyObject *__pyx_n_s_file_mask_p;
static PyObject *__pyx_n_s_fill_bool;
static PyObject *__pyx_n_s_fill_masks;
static PyObject *__pyx_n_s_finalize;
static PyObject *__pyx_n_s_finish;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_float;
static PyObject *__pyx_n_s_float32;
static PyObject *__pyx_n_s_float64;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_fortran;
static PyObject *__pyx_n_u_fortran;
static PyObject *__pyx_n_s_get_DLE;
static PyObject *__pyx_n_s_get_DRE;
static PyObject *__pyx_n_s_get_bitmasks;
static PyObject *__pyx_n_s_get_pbar;
static PyObject *__pyx_n_s_get_periodicity;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_kp_s_got_differing_extents_in_dimensi;
static PyObject *__pyx_n_s_hsml;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_ifile;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_in_coll;
static PyObject *__pyx_n_s_in_collection;
static PyObject *__pyx_n_s_index;
static PyObject *__pyx_n_s_index_order1;
static PyObject *__pyx_n_s_index_order2;
static PyObject *__pyx_n_s_indices;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_int32;
static PyObject *__pyx_n_s_int64;
static PyObject *__pyx_n_s_io_handler;
static PyObject *__pyx_kp_u_is_negative_with_value;
static PyObject *__pyx_n_s_isfile;
static PyObject *__pyx_n_s_itemsize;
static PyObject *__pyx_kp_s_itemsize_0_for_cython_array;
static PyObject *__pyx_n_s_iter;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_kind;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_left_edge;
static PyObject *__pyx_n_s_log2;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mask;
static PyObject *__pyx_n_s_mask2;
static PyObject *__pyx_n_s_mask_threshold;
static PyObject *__pyx_n_s_mask_to_files;
static PyObject *__pyx_n_s_masklevel;
static PyObject *__pyx_n_s_masks_to_files;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_max_index_order2;
static PyObject *__pyx_n_s_memview;
static PyObject *__pyx_n_s_mm_g;
static PyObject *__pyx_n_s_mm_s;
static PyObject *__pyx_n_s_mode;
static PyObject *__pyx_n_s_n_ref;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_kp_u_ndarray_is_not_C_contiguous;
static PyObject *__pyx_kp_u_ndarray_is_not_Fortran_contiguou;
static PyObject *__pyx_n_s_ndim;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_nfiles;
static PyObject *__pyx_n_s_ngz;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_none;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_nsub_mi;
static PyObject *__pyx_n_s_num_root;
static PyObject *__pyx_n_s_num_zones;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_o;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_object;
static PyObject *__pyx_n_s_open;
static PyObject *__pyx_n_s_order;
static PyObject *__pyx_n_s_order1;
static PyObject *__pyx_n_s_os;
static PyObject *__pyx_n_s_out;
static PyObject *__pyx_n_s_out_collection;
static PyObject *__pyx_n_s_pack;
static PyObject *__pyx_n_s_path;
static PyObject *__pyx_n_s_periodicity;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pos;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_print_info;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_getbuffer;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_Enum;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_q;
static PyObject *__pyx_n_s_quartic;
static PyObject *__pyx_n_s_quintic;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_rb;
static PyObject *__pyx_n_s_read;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_refined_index_data_file;
static PyObject *__pyx_n_s_right_edge;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_kp_u_s_2;
static PyObject *__pyx_n_s_save_bitmasks;
static PyObject *__pyx_n_s_selector;
static PyObject *__pyx_n_s_selector_mask;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_kp_s_self__ptr_octs_per_root_self_oct;
static PyObject *__pyx_kp_s_self_dom_offsets_self_oct_list_s;
static PyObject *__pyx_kp_s_self_file_markers_cannot_be_conv;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_sort;
static PyObject *__pyx_n_s_split;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_step;
static PyObject *__pyx_n_s_stop;
static PyObject *__pyx_kp_s_strided_and_direct;
static PyObject *__pyx_kp_s_strided_and_direct_or_indirect;
static PyObject *__pyx_kp_s_strided_and_indirect;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_strip;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_sub_mi1;
static PyObject *__pyx_n_s_sub_mi2;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_super;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_total_particles;
static PyObject *__pyx_n_s_u8;
static PyObject *__pyx_n_s_uint32;
static PyObject *__pyx_n_s_uint64;
static PyObject *__pyx_n_s_uint8;
static PyObject *__pyx_kp_s_unable_to_allocate_array_data;
static PyObject *__pyx_kp_s_unable_to_allocate_shape_and_str;
static PyObject *__pyx_n_s_units;
static PyObject *__pyx_kp_u_unknown_dtype_code_in_numpy_pxd;
static PyObject *__pyx_n_s_unpack;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_n_s_verbose;
static PyObject *__pyx_n_s_wb;
static PyObject *__pyx_n_s_wendland2;
static PyObject *__pyx_n_s_wendland4;
static PyObject *__pyx_n_s_wendland6;
static PyObject *__pyx_n_s_where;
static PyObject *__pyx_n_s_write;
static PyObject *__pyx_n_s_yield_coordinates;
static PyObject *__pyx_n_s_yt_funcs;
static PyObject *__pyx_n_s_yt_geometry_particle_oct_contain;
static PyObject *__pyx_kp_s_yt_geometry_particle_oct_contain_2;
static PyObject *__pyx_n_s_zeros;
static PyObject *__pyx_n_s_zip;
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_allocate_root(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self); /* proto */
static void __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_2__dealloc__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_4clear_fileind(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_6__iter__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_9allocate_domains(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_domain_counts); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_11finalize(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, int __pyx_v_domain_id); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_13add(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_uint8_t __pyx_v_order); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_15recursively_count(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_17apply_domain(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, int __pyx_v_domain_id, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mask, int __pyx_v_masklevel); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_19domain_ind(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, PyObject *__pyx_v_selector, int __pyx_v_domain_id, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mask, int __pyx_v_masklevel); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_9max_level___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_9max_level_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_5n_ref___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_5n_ref_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_21__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_23__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap___init__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_left_edge, PyObject *__pyx_v_right_edge, PyObject *__pyx_v_periodicity, PyObject *__pyx_v_file_hash, PyObject *__pyx_v_nfiles, PyObject *__pyx_v_index_order1, PyObject *__pyx_v_index_order2); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_2_bitmask_logicaland(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_ifile, PyObject *__pyx_v_bcoll, PyObject *__pyx_v_out); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_4_bitmask_intersects(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_ifile, PyObject *__pyx_v_bcoll); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6update_mi2(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, __pyx_t_5numpy_float64_t __pyx_v_characteristic_size, __pyx_t_5numpy_uint64_t __pyx_v_max_index_order2); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8_coarse_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_80_coarse_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyArrayObject *__pyx_v_pos, PyArrayObject *__pyx_v_hsml, __pyx_t_5numpy_uint64_t __pyx_v_file_id); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_82_coarse_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyArrayObject *__pyx_v_pos, PyArrayObject *__pyx_v_hsml, __pyx_t_5numpy_uint64_t __pyx_v_file_id); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10_set_coarse_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_file_id); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12_refined_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_86_refined_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_in_collection, PyArrayObject *__pyx_v_pos, PyArrayObject *__pyx_v_hsml, PyArrayObject *__pyx_v_mask, CYTHON_UNUSED PyArrayObject *__pyx_v_sub_mi1, CYTHON_UNUSED PyArrayObject *__pyx_v_sub_mi2, CYTHON_UNUSED __pyx_t_5numpy_uint64_t __pyx_v_file_id, CYTHON_UNUSED __pyx_t_5numpy_int64_t __pyx_v_nsub_mi, __pyx_t_5numpy_uint64_t __pyx_v_count_threshold, __pyx_t_5numpy_uint8_t __pyx_v_mask_threshold); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_88_refined_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_in_collection, PyArrayObject *__pyx_v_pos, PyArrayObject *__pyx_v_hsml, PyArrayObject *__pyx_v_mask, CYTHON_UNUSED PyArrayObject *__pyx_v_sub_mi1, CYTHON_UNUSED PyArrayObject *__pyx_v_sub_mi2, CYTHON_UNUSED __pyx_t_5numpy_uint64_t __pyx_v_file_id, CYTHON_UNUSED __pyx_t_5numpy_int64_t __pyx_v_nsub_mi, __pyx_t_5numpy_uint64_t __pyx_v_count_threshold, __pyx_t_5numpy_uint8_t __pyx_v_mask_threshold); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_set_refined_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyArrayObject *__pyx_v_sub_mi1, PyArrayObject *__pyx_v_sub_mi2, __pyx_t_5numpy_uint64_t __pyx_v_file_id, __pyx_t_5numpy_int64_t __pyx_v_nsub_mi); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_16find_collisions(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_verbose); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_18find_collisions_coarse(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_verbose, PyObject *__pyx_v_file_list); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_20find_uncontaminated(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, __pyx_t_5numpy_uint32_t __pyx_v_ifile, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mask, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mask2); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_22find_contaminated(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, __pyx_t_5numpy_uint32_t __pyx_v_ifile, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mask, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mask2); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_24find_collisions_refined(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_verbose); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_26calcsize_bitmasks(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_28get_bitmasks(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_30iseq_bitmask(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_solf); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_32save_bitmasks(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_fname); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_34check_bitmasks(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_36reset_bitmasks(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_38load_bitmasks(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_fname); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_40print_info(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_42count_coarse(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_ifile); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_44count_refined(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_ifile); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_46count_total(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_ifile); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_48check(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_50primary_indices(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_52file_ownership_mask(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_fid); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_54finalize(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_56get_DLE(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_58get_DRE(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_60get_ghost_zones(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_selector, int __pyx_v_ngz, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_dmask, int __pyx_v_coarse_ghosts); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_62selector2mask(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_selector); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_64mask2files(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_cmask); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_66mask2filemasks(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_cmask, PyArrayObject *__pyx_v_file_idx); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_68filemasks2addfiles(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyArrayObject *__pyx_v_file_masks); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_70identify_file_masks(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_selector); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_72identify_data_files(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_selector, int __pyx_v_ngz); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_74mask_to_files(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mm_s); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_76masks_to_files(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mm_s, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mm_g); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_78construct_octree(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_io_handler, PyObject *__pyx_v_data_files, PyObject *__pyx_v_num_zones, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_selector_mask, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_base_mask); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6nfiles___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6nfiles_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order1___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order1_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order2___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order2_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5masks___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5masks_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5masks_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15particle_counts___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15particle_counts_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15particle_counts_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6counts___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6counts_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6counts_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9max_count___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9max_count_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9max_count_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_last_selector___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_last_selector_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_last_selector_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_return_values___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_return_values_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_return_values_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_cached_octrees___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_cached_octrees_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_cached_octrees_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_octree_subset___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_octree_subset_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_octree_subset_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_last_oct_handler___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_last_oct_handler_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_last_oct_handler_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_prev_octree_subset___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_prev_octree_subset_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_prev_octree_subset_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_prev_oct_handler___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_prev_oct_handler_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_prev_oct_handler_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8bitmasks___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8bitmasks_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8bitmasks_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10collisions___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10collisions_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10collisions_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9_used_mi2___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9_used_mi2_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_92__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_94__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector___cinit__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, PyObject *__pyx_v_selector, PyObject *__pyx_v_bitmap, PyObject *__pyx_v_ngz); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_2fill_masks(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mm_s, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mm_g); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_4find_files(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, PyArrayObject *__pyx_v_file_mask_p, PyArrayObject *__pyx_v_file_mask_g); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer___init__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, PyObject *__pyx_v_domain_dimensions, PyObject *__pyx_v_domain_left_edge, PyObject *__pyx_v_domain_right_edge, int __pyx_v_num_root, PyObject *__pyx_v_num_zones); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_2allocate_domains(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, PyObject *__pyx_v_counts); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_4finalize(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_6get_index_base_octs(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, __Pyx_memviewslice __pyx_v_domain_ind); /* proto */
static void __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_8__dealloc__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_10add(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, PyArrayObject *__pyx_v_indices, CYTHON_UNUSED __pyx_t_5numpy_uint64_t __pyx_v_order1, int __pyx_v_domain_id); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_9max_level___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_9max_level_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_5n_ref___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_5n_ref_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_17_index_base_roots___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_17_index_base_roots_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_16_index_base_octs___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_16_index_base_octs_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_13overlap_cells___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self); /* proto */
static int __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_13overlap_cells_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object); /* proto */
static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_22particle_oct_container_ParticleBitmap(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_9;
static PyObject *__pyx_int_112105877;
static PyObject *__pyx_int_136983863;
static PyObject *__pyx_int_184977713;
static PyObject *__pyx_int_neg_1;
static __pyx_t_5numpy_uint8_t __pyx_k_;
static __pyx_t_5numpy_uint8_t __pyx_k__2;
static __pyx_t_5numpy_uint64_t __pyx_k__13;
static __pyx_t_5numpy_uint8_t __pyx_k__14;
static __pyx_t_5numpy_uint64_t __pyx_k__15;
static __pyx_t_5numpy_uint8_t __pyx_k__16;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__11;
static PyObject *__pyx_slice__17;
static PyObject *__pyx_slice__18;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__48;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_tuple__51;
static PyObject *__pyx_tuple__52;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_tuple__59;
static PyObject *__pyx_tuple__61;
static PyObject *__pyx_tuple__63;
static PyObject *__pyx_tuple__64;
static PyObject *__pyx_tuple__65;
static PyObject *__pyx_tuple__66;
static PyObject *__pyx_tuple__67;
static PyObject *__pyx_tuple__68;
static PyObject *__pyx_codeobj__60;
static PyObject *__pyx_codeobj__62;
static PyObject *__pyx_codeobj__69;
/* Late includes */

/* "yt/geometry/particle_oct_container.pyx":98
 *     cdef public int n_ref
 * 
 *     def allocate_root(self):             # <<<<<<<<<<<<<<
 *         cdef int i, j, k
 *         cdef Oct *cur
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_1allocate_root(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_1allocate_root(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("allocate_root (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_allocate_root(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_allocate_root(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_cur;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("allocate_root", 0);

  /* "yt/geometry/particle_oct_container.pyx":101
 *         cdef int i, j, k
 *         cdef Oct *cur
 *         for i in range(self.nn[0]):             # <<<<<<<<<<<<<<
 *             for j in range(self.nn[1]):
 *                 for k in range(self.nn[2]):
 */
  __pyx_t_1 = (__pyx_v_self->__pyx_base.nn[0]);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":102
 *         cdef Oct *cur
 *         for i in range(self.nn[0]):
 *             for j in range(self.nn[1]):             # <<<<<<<<<<<<<<
 *                 for k in range(self.nn[2]):
 *                     cur = self.allocate_oct()
 */
    __pyx_t_4 = (__pyx_v_self->__pyx_base.nn[1]);
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "yt/geometry/particle_oct_container.pyx":103
 *         for i in range(self.nn[0]):
 *             for j in range(self.nn[1]):
 *                 for k in range(self.nn[2]):             # <<<<<<<<<<<<<<
 *                     cur = self.allocate_oct()
 *                     self.root_mesh[i][j][k] = cur
 */
      __pyx_t_7 = (__pyx_v_self->__pyx_base.nn[2]);
      __pyx_t_8 = __pyx_t_7;
      for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
        __pyx_v_k = __pyx_t_9;

        /* "yt/geometry/particle_oct_container.pyx":104
 *             for j in range(self.nn[1]):
 *                 for k in range(self.nn[2]):
 *                     cur = self.allocate_oct()             # <<<<<<<<<<<<<<
 *                     self.root_mesh[i][j][k] = cur
 * 
 */
        __pyx_v_cur = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_vtab)->allocate_oct(__pyx_v_self);

        /* "yt/geometry/particle_oct_container.pyx":105
 *                 for k in range(self.nn[2]):
 *                     cur = self.allocate_oct()
 *                     self.root_mesh[i][j][k] = cur             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
        (((__pyx_v_self->__pyx_base.root_mesh[__pyx_v_i])[__pyx_v_j])[__pyx_v_k]) = __pyx_v_cur;
      }
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":98
 *     cdef public int n_ref
 * 
 *     def allocate_root(self):             # <<<<<<<<<<<<<<
 *         cdef int i, j, k
 *         cdef Oct *cur
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":107
 *                     self.root_mesh[i][j][k] = cur
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         #Call the freemem ops on every ocy
 *         #of the root mesh recursively
 */

/* Python wrapper */
static void __pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_2__dealloc__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_2__dealloc__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "yt/geometry/particle_oct_container.pyx":111
 *         #of the root mesh recursively
 *         cdef int i, j, k
 *         if self.root_mesh == NULL: return             # <<<<<<<<<<<<<<
 *         for i in range(self.nn[0]):
 *             if self.root_mesh[i] == NULL: continue
 */
  __pyx_t_1 = ((__pyx_v_self->__pyx_base.root_mesh == NULL) != 0);
  if (__pyx_t_1) {
    goto __pyx_L0;
  }

  /* "yt/geometry/particle_oct_container.pyx":112
 *         cdef int i, j, k
 *         if self.root_mesh == NULL: return
 *         for i in range(self.nn[0]):             # <<<<<<<<<<<<<<
 *             if self.root_mesh[i] == NULL: continue
 *             for j in range(self.nn[1]):
 */
  __pyx_t_2 = (__pyx_v_self->__pyx_base.nn[0]);
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "yt/geometry/particle_oct_container.pyx":113
 *         if self.root_mesh == NULL: return
 *         for i in range(self.nn[0]):
 *             if self.root_mesh[i] == NULL: continue             # <<<<<<<<<<<<<<
 *             for j in range(self.nn[1]):
 *                 if self.root_mesh[i][j] == NULL: continue
 */
    __pyx_t_1 = (((__pyx_v_self->__pyx_base.root_mesh[__pyx_v_i]) == NULL) != 0);
    if (__pyx_t_1) {
      goto __pyx_L4_continue;
    }

    /* "yt/geometry/particle_oct_container.pyx":114
 *         for i in range(self.nn[0]):
 *             if self.root_mesh[i] == NULL: continue
 *             for j in range(self.nn[1]):             # <<<<<<<<<<<<<<
 *                 if self.root_mesh[i][j] == NULL: continue
 *                 for k in range(self.nn[2]):
 */
    __pyx_t_5 = (__pyx_v_self->__pyx_base.nn[1]);
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_j = __pyx_t_7;

      /* "yt/geometry/particle_oct_container.pyx":115
 *             if self.root_mesh[i] == NULL: continue
 *             for j in range(self.nn[1]):
 *                 if self.root_mesh[i][j] == NULL: continue             # <<<<<<<<<<<<<<
 *                 for k in range(self.nn[2]):
 *                     if self.root_mesh[i][j][k] == NULL: continue
 */
      __pyx_t_1 = ((((__pyx_v_self->__pyx_base.root_mesh[__pyx_v_i])[__pyx_v_j]) == NULL) != 0);
      if (__pyx_t_1) {
        goto __pyx_L7_continue;
      }

      /* "yt/geometry/particle_oct_container.pyx":116
 *             for j in range(self.nn[1]):
 *                 if self.root_mesh[i][j] == NULL: continue
 *                 for k in range(self.nn[2]):             # <<<<<<<<<<<<<<
 *                     if self.root_mesh[i][j][k] == NULL: continue
 *                     self.visit_free(self.root_mesh[i][j][k])
 */
      __pyx_t_8 = (__pyx_v_self->__pyx_base.nn[2]);
      __pyx_t_9 = __pyx_t_8;
      for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
        __pyx_v_k = __pyx_t_10;

        /* "yt/geometry/particle_oct_container.pyx":117
 *                 if self.root_mesh[i][j] == NULL: continue
 *                 for k in range(self.nn[2]):
 *                     if self.root_mesh[i][j][k] == NULL: continue             # <<<<<<<<<<<<<<
 *                     self.visit_free(self.root_mesh[i][j][k])
 *         free(self.oct_list)
 */
        __pyx_t_1 = (((((__pyx_v_self->__pyx_base.root_mesh[__pyx_v_i])[__pyx_v_j])[__pyx_v_k]) == NULL) != 0);
        if (__pyx_t_1) {
          goto __pyx_L10_continue;
        }

        /* "yt/geometry/particle_oct_container.pyx":118
 *                 for k in range(self.nn[2]):
 *                     if self.root_mesh[i][j][k] == NULL: continue
 *                     self.visit_free(self.root_mesh[i][j][k])             # <<<<<<<<<<<<<<
 *         free(self.oct_list)
 *         free(self.dom_offsets)
 */
        ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_vtab)->visit_free(__pyx_v_self, (((__pyx_v_self->__pyx_base.root_mesh[__pyx_v_i])[__pyx_v_j])[__pyx_v_k]));
        __pyx_L10_continue:;
      }
      __pyx_L7_continue:;
    }
    __pyx_L4_continue:;
  }

  /* "yt/geometry/particle_oct_container.pyx":119
 *                     if self.root_mesh[i][j][k] == NULL: continue
 *                     self.visit_free(self.root_mesh[i][j][k])
 *         free(self.oct_list)             # <<<<<<<<<<<<<<
 *         free(self.dom_offsets)
 * 
 */
  free(__pyx_v_self->oct_list);

  /* "yt/geometry/particle_oct_container.pyx":120
 *                     self.visit_free(self.root_mesh[i][j][k])
 *         free(self.oct_list)
 *         free(self.dom_offsets)             # <<<<<<<<<<<<<<
 * 
 *     cdef void visit_free(self, Oct *o):
 */
  free(__pyx_v_self->dom_offsets);

  /* "yt/geometry/particle_oct_container.pyx":107
 *                     self.root_mesh[i][j][k] = cur
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         #Call the freemem ops on every ocy
 *         #of the root mesh recursively
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":122
 *         free(self.dom_offsets)
 * 
 *     cdef void visit_free(self, Oct *o):             # <<<<<<<<<<<<<<
 *         #Free the memory for this oct recursively
 *         cdef int i, j, k
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_visit_free(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("visit_free", 0);

  /* "yt/geometry/particle_oct_container.pyx":125
 *         #Free the memory for this oct recursively
 *         cdef int i, j, k
 *         for i in range(2):             # <<<<<<<<<<<<<<
 *             for j in range(2):
 *                 for k in range(2):
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 2; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":126
 *         cdef int i, j, k
 *         for i in range(2):
 *             for j in range(2):             # <<<<<<<<<<<<<<
 *                 for k in range(2):
 *                     if o.children != NULL \
 */
    for (__pyx_t_2 = 0; __pyx_t_2 < 2; __pyx_t_2+=1) {
      __pyx_v_j = __pyx_t_2;

      /* "yt/geometry/particle_oct_container.pyx":127
 *         for i in range(2):
 *             for j in range(2):
 *                 for k in range(2):             # <<<<<<<<<<<<<<
 *                     if o.children != NULL \
 *                        and o.children[cind(i,j,k)] != NULL:
 */
      for (__pyx_t_3 = 0; __pyx_t_3 < 2; __pyx_t_3+=1) {
        __pyx_v_k = __pyx_t_3;

        /* "yt/geometry/particle_oct_container.pyx":128
 *             for j in range(2):
 *                 for k in range(2):
 *                     if o.children != NULL \             # <<<<<<<<<<<<<<
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_free(o.children[cind(i,j,k)])
 */
        __pyx_t_5 = ((__pyx_v_o->children != NULL) != 0);
        if (__pyx_t_5) {
        } else {
          __pyx_t_4 = __pyx_t_5;
          goto __pyx_L10_bool_binop_done;
        }

        /* "yt/geometry/particle_oct_container.pyx":129
 *                 for k in range(2):
 *                     if o.children != NULL \
 *                        and o.children[cind(i,j,k)] != NULL:             # <<<<<<<<<<<<<<
 *                         self.visit_free(o.children[cind(i,j,k)])
 *         free(o.children)
 */
        __pyx_t_5 = (((__pyx_v_o->children[__pyx_f_2yt_8geometry_12oct_visitors_cind(__pyx_v_i, __pyx_v_j, __pyx_v_k)]) != NULL) != 0);
        __pyx_t_4 = __pyx_t_5;
        __pyx_L10_bool_binop_done:;

        /* "yt/geometry/particle_oct_container.pyx":128
 *             for j in range(2):
 *                 for k in range(2):
 *                     if o.children != NULL \             # <<<<<<<<<<<<<<
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_free(o.children[cind(i,j,k)])
 */
        if (__pyx_t_4) {

          /* "yt/geometry/particle_oct_container.pyx":130
 *                     if o.children != NULL \
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_free(o.children[cind(i,j,k)])             # <<<<<<<<<<<<<<
 *         free(o.children)
 *         free(o)
 */
          ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_vtab)->visit_free(__pyx_v_self, (__pyx_v_o->children[__pyx_f_2yt_8geometry_12oct_visitors_cind(__pyx_v_i, __pyx_v_j, __pyx_v_k)]));

          /* "yt/geometry/particle_oct_container.pyx":128
 *             for j in range(2):
 *                 for k in range(2):
 *                     if o.children != NULL \             # <<<<<<<<<<<<<<
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_free(o.children[cind(i,j,k)])
 */
        }
      }
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":131
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_free(o.children[cind(i,j,k)])
 *         free(o.children)             # <<<<<<<<<<<<<<
 *         free(o)
 * 
 */
  free(__pyx_v_o->children);

  /* "yt/geometry/particle_oct_container.pyx":132
 *                         self.visit_free(o.children[cind(i,j,k)])
 *         free(o.children)
 *         free(o)             # <<<<<<<<<<<<<<
 * 
 *     def clear_fileind(self):
 */
  free(__pyx_v_o);

  /* "yt/geometry/particle_oct_container.pyx":122
 *         free(self.dom_offsets)
 * 
 *     cdef void visit_free(self, Oct *o):             # <<<<<<<<<<<<<<
 *         #Free the memory for this oct recursively
 *         cdef int i, j, k
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":134
 *         free(o)
 * 
 *     def clear_fileind(self):             # <<<<<<<<<<<<<<
 *         cdef int i, j, k
 *         for i in range(self.nn[0]):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_5clear_fileind(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_5clear_fileind(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear_fileind (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_4clear_fileind(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_4clear_fileind(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("clear_fileind", 0);

  /* "yt/geometry/particle_oct_container.pyx":136
 *     def clear_fileind(self):
 *         cdef int i, j, k
 *         for i in range(self.nn[0]):             # <<<<<<<<<<<<<<
 *             for j in range(self.nn[1]):
 *                 for k in range(self.nn[2]):
 */
  __pyx_t_1 = (__pyx_v_self->__pyx_base.nn[0]);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":137
 *         cdef int i, j, k
 *         for i in range(self.nn[0]):
 *             for j in range(self.nn[1]):             # <<<<<<<<<<<<<<
 *                 for k in range(self.nn[2]):
 *                     self.visit_clear(self.root_mesh[i][j][k])
 */
    __pyx_t_4 = (__pyx_v_self->__pyx_base.nn[1]);
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "yt/geometry/particle_oct_container.pyx":138
 *         for i in range(self.nn[0]):
 *             for j in range(self.nn[1]):
 *                 for k in range(self.nn[2]):             # <<<<<<<<<<<<<<
 *                     self.visit_clear(self.root_mesh[i][j][k])
 * 
 */
      __pyx_t_7 = (__pyx_v_self->__pyx_base.nn[2]);
      __pyx_t_8 = __pyx_t_7;
      for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
        __pyx_v_k = __pyx_t_9;

        /* "yt/geometry/particle_oct_container.pyx":139
 *             for j in range(self.nn[1]):
 *                 for k in range(self.nn[2]):
 *                     self.visit_clear(self.root_mesh[i][j][k])             # <<<<<<<<<<<<<<
 * 
 *     cdef void visit_clear(self, Oct *o):
 */
        ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_vtab)->visit_clear(__pyx_v_self, (((__pyx_v_self->__pyx_base.root_mesh[__pyx_v_i])[__pyx_v_j])[__pyx_v_k]));
      }
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":134
 *         free(o)
 * 
 *     def clear_fileind(self):             # <<<<<<<<<<<<<<
 *         cdef int i, j, k
 *         for i in range(self.nn[0]):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":141
 *                     self.visit_clear(self.root_mesh[i][j][k])
 * 
 *     cdef void visit_clear(self, Oct *o):             # <<<<<<<<<<<<<<
 *         #Free the memory for this oct recursively
 *         cdef int i, j, k
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_visit_clear(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("visit_clear", 0);

  /* "yt/geometry/particle_oct_container.pyx":144
 *         #Free the memory for this oct recursively
 *         cdef int i, j, k
 *         o.file_ind = 0             # <<<<<<<<<<<<<<
 *         for i in range(2):
 *             for j in range(2):
 */
  __pyx_v_o->file_ind = 0;

  /* "yt/geometry/particle_oct_container.pyx":145
 *         cdef int i, j, k
 *         o.file_ind = 0
 *         for i in range(2):             # <<<<<<<<<<<<<<
 *             for j in range(2):
 *                 for k in range(2):
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 2; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":146
 *         o.file_ind = 0
 *         for i in range(2):
 *             for j in range(2):             # <<<<<<<<<<<<<<
 *                 for k in range(2):
 *                     if o.children != NULL \
 */
    for (__pyx_t_2 = 0; __pyx_t_2 < 2; __pyx_t_2+=1) {
      __pyx_v_j = __pyx_t_2;

      /* "yt/geometry/particle_oct_container.pyx":147
 *         for i in range(2):
 *             for j in range(2):
 *                 for k in range(2):             # <<<<<<<<<<<<<<
 *                     if o.children != NULL \
 *                        and o.children[cind(i,j,k)] != NULL:
 */
      for (__pyx_t_3 = 0; __pyx_t_3 < 2; __pyx_t_3+=1) {
        __pyx_v_k = __pyx_t_3;

        /* "yt/geometry/particle_oct_container.pyx":148
 *             for j in range(2):
 *                 for k in range(2):
 *                     if o.children != NULL \             # <<<<<<<<<<<<<<
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_clear(o.children[cind(i,j,k)])
 */
        __pyx_t_5 = ((__pyx_v_o->children != NULL) != 0);
        if (__pyx_t_5) {
        } else {
          __pyx_t_4 = __pyx_t_5;
          goto __pyx_L10_bool_binop_done;
        }

        /* "yt/geometry/particle_oct_container.pyx":149
 *                 for k in range(2):
 *                     if o.children != NULL \
 *                        and o.children[cind(i,j,k)] != NULL:             # <<<<<<<<<<<<<<
 *                         self.visit_clear(o.children[cind(i,j,k)])
 * 
 */
        __pyx_t_5 = (((__pyx_v_o->children[__pyx_f_2yt_8geometry_12oct_visitors_cind(__pyx_v_i, __pyx_v_j, __pyx_v_k)]) != NULL) != 0);
        __pyx_t_4 = __pyx_t_5;
        __pyx_L10_bool_binop_done:;

        /* "yt/geometry/particle_oct_container.pyx":148
 *             for j in range(2):
 *                 for k in range(2):
 *                     if o.children != NULL \             # <<<<<<<<<<<<<<
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_clear(o.children[cind(i,j,k)])
 */
        if (__pyx_t_4) {

          /* "yt/geometry/particle_oct_container.pyx":150
 *                     if o.children != NULL \
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_clear(o.children[cind(i,j,k)])             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
          ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_vtab)->visit_clear(__pyx_v_self, (__pyx_v_o->children[__pyx_f_2yt_8geometry_12oct_visitors_cind(__pyx_v_i, __pyx_v_j, __pyx_v_k)]));

          /* "yt/geometry/particle_oct_container.pyx":148
 *             for j in range(2):
 *                 for k in range(2):
 *                     if o.children != NULL \             # <<<<<<<<<<<<<<
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_clear(o.children[cind(i,j,k)])
 */
        }
      }
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":141
 *                     self.visit_clear(self.root_mesh[i][j][k])
 * 
 *     cdef void visit_clear(self, Oct *o):             # <<<<<<<<<<<<<<
 *         #Free the memory for this oct recursively
 *         cdef int i, j, k
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}
static PyObject *__pyx_gb_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_8generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "yt/geometry/particle_oct_container.pyx":152
 *                         self.visit_clear(o.children[cind(i,j,k)])
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         #Get the next oct, will traverse domains
 *         #Note that oct containers can be sorted
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_7__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_7__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_6__iter__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_6__iter__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self) {
  struct __pyx_obj_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__ *)__pyx_tp_new_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__(__pyx_ptype_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 152, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_8generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_ParticleOctreeContainer___iter, __pyx_n_s_yt_geometry_particle_oct_contain); if (unlikely(!gen)) __PYX_ERR(0, 152, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleOctreeContainer.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_8generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__ *__pyx_cur_scope = ((struct __pyx_obj_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_1;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 152, __pyx_L1_error)

  /* "yt/geometry/particle_oct_container.pyx":158
 *         cdef int oi
 *         cdef Oct *o
 *         for oi in range(self.nocts):             # <<<<<<<<<<<<<<
 *             o = self.oct_list[oi]
 *             yield (o.file_ind, o.domain_ind, o.domain)
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_self->__pyx_base.nocts;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_cur_scope->__pyx_v_oi = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":159
 *         cdef Oct *o
 *         for oi in range(self.nocts):
 *             o = self.oct_list[oi]             # <<<<<<<<<<<<<<
 *             yield (o.file_ind, o.domain_ind, o.domain)
 * 
 */
    __pyx_cur_scope->__pyx_v_o = (__pyx_cur_scope->__pyx_v_self->oct_list[__pyx_cur_scope->__pyx_v_oi]);

    /* "yt/geometry/particle_oct_container.pyx":160
 *         for oi in range(self.nocts):
 *             o = self.oct_list[oi]
 *             yield (o.file_ind, o.domain_ind, o.domain)             # <<<<<<<<<<<<<<
 * 
 *     def allocate_domains(self, domain_counts):
 */
    __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_cur_scope->__pyx_v_o->file_ind); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_From_npy_int64(__pyx_cur_scope->__pyx_v_o->domain_ind); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_cur_scope->__pyx_v_o->domain); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_6);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 160, __pyx_L1_error)
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "yt/geometry/particle_oct_container.pyx":152
 *                         self.visit_clear(o.children[cind(i,j,k)])
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         #Get the next oct, will traverse domains
 *         #Note that oct containers can be sorted
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":162
 *             yield (o.file_ind, o.domain_ind, o.domain)
 * 
 *     def allocate_domains(self, domain_counts):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_10allocate_domains(PyObject *__pyx_v_self, PyObject *__pyx_v_domain_counts); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_10allocate_domains(PyObject *__pyx_v_self, PyObject *__pyx_v_domain_counts) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("allocate_domains (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_9allocate_domains(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self), ((PyObject *)__pyx_v_domain_counts));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_9allocate_domains(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_domain_counts) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("allocate_domains", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":165
 *         pass
 * 
 *     def finalize(self, int domain_id = 0):             # <<<<<<<<<<<<<<
 *         #This will sort the octs in the oct list
 *         #so that domains appear consecutively
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_12finalize(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_12finalize(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_domain_id;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("finalize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_domain_id,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_domain_id);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "finalize") < 0)) __PYX_ERR(0, 165, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_domain_id = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_domain_id == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 165, __pyx_L3_error)
    } else {
      __pyx_v_domain_id = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("finalize", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 165, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleOctreeContainer.finalize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_11finalize(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self), __pyx_v_domain_id);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_11finalize(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, int __pyx_v_domain_id) {
  int __pyx_v_max_level;
  __pyx_t_5numpy_int64_t __pyx_v_i;
  __pyx_t_5numpy_int64_t __pyx_v_lpos;
  int __pyx_v_j;
  int __pyx_v_k;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_11;
  __pyx_t_5numpy_int64_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("finalize", 0);

  /* "yt/geometry/particle_oct_container.pyx":170
 *         #And then find the oct index/offset for
 *         #every domain
 *         cdef int max_level = 0             # <<<<<<<<<<<<<<
 *         self.oct_list = <Oct**> malloc(sizeof(Oct*)*self.nocts)
 *         cdef np.int64_t i = 0, lpos = 0
 */
  __pyx_v_max_level = 0;

  /* "yt/geometry/particle_oct_container.pyx":171
 *         #every domain
 *         cdef int max_level = 0
 *         self.oct_list = <Oct**> malloc(sizeof(Oct*)*self.nocts)             # <<<<<<<<<<<<<<
 *         cdef np.int64_t i = 0, lpos = 0
 *         # Note that we now assign them in the same order they will be visited
 */
  __pyx_v_self->oct_list = ((struct __pyx_t_2yt_8geometry_12oct_visitors_Oct **)malloc(((sizeof(struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *)) * __pyx_v_self->__pyx_base.nocts)));

  /* "yt/geometry/particle_oct_container.pyx":172
 *         cdef int max_level = 0
 *         self.oct_list = <Oct**> malloc(sizeof(Oct*)*self.nocts)
 *         cdef np.int64_t i = 0, lpos = 0             # <<<<<<<<<<<<<<
 *         # Note that we now assign them in the same order they will be visited
 *         # by recursive visitors.
 */
  __pyx_v_i = 0;
  __pyx_v_lpos = 0;

  /* "yt/geometry/particle_oct_container.pyx":175
 *         # Note that we now assign them in the same order they will be visited
 *         # by recursive visitors.
 *         for i in range(self.nn[0]):             # <<<<<<<<<<<<<<
 *             for j in range(self.nn[1]):
 *                 for k in range(self.nn[2]):
 */
  __pyx_t_1 = (__pyx_v_self->__pyx_base.nn[0]);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":176
 *         # by recursive visitors.
 *         for i in range(self.nn[0]):
 *             for j in range(self.nn[1]):             # <<<<<<<<<<<<<<
 *                 for k in range(self.nn[2]):
 *                     self.visit_assign(self.root_mesh[i][j][k], &lpos,
 */
    __pyx_t_4 = (__pyx_v_self->__pyx_base.nn[1]);
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "yt/geometry/particle_oct_container.pyx":177
 *         for i in range(self.nn[0]):
 *             for j in range(self.nn[1]):
 *                 for k in range(self.nn[2]):             # <<<<<<<<<<<<<<
 *                     self.visit_assign(self.root_mesh[i][j][k], &lpos,
 *                                       0, &max_level)
 */
      __pyx_t_7 = (__pyx_v_self->__pyx_base.nn[2]);
      __pyx_t_8 = __pyx_t_7;
      for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
        __pyx_v_k = __pyx_t_9;

        /* "yt/geometry/particle_oct_container.pyx":178
 *             for j in range(self.nn[1]):
 *                 for k in range(self.nn[2]):
 *                     self.visit_assign(self.root_mesh[i][j][k], &lpos,             # <<<<<<<<<<<<<<
 *                                       0, &max_level)
 *         assert(lpos == self.nocts)
 */
        __pyx_t_10 = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_vtab)->visit_assign(__pyx_v_self, (((__pyx_v_self->__pyx_base.root_mesh[__pyx_v_i])[__pyx_v_j])[__pyx_v_k]), (&__pyx_v_lpos), 0, (&__pyx_v_max_level)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 178, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":180
 *                     self.visit_assign(self.root_mesh[i][j][k], &lpos,
 *                                       0, &max_level)
 *         assert(lpos == self.nocts)             # <<<<<<<<<<<<<<
 *         for i in range(self.nocts):
 *             self.oct_list[i].domain_ind = i
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!((__pyx_v_lpos == __pyx_v_self->__pyx_base.nocts) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 180, __pyx_L1_error)
    }
  }
  #endif

  /* "yt/geometry/particle_oct_container.pyx":181
 *                                       0, &max_level)
 *         assert(lpos == self.nocts)
 *         for i in range(self.nocts):             # <<<<<<<<<<<<<<
 *             self.oct_list[i].domain_ind = i
 *             self.oct_list[i].domain = domain_id
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.nocts;
  __pyx_t_11 = __pyx_t_3;
  for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_i = __pyx_t_12;

    /* "yt/geometry/particle_oct_container.pyx":182
 *         assert(lpos == self.nocts)
 *         for i in range(self.nocts):
 *             self.oct_list[i].domain_ind = i             # <<<<<<<<<<<<<<
 *             self.oct_list[i].domain = domain_id
 *         self.max_level = max_level
 */
    (__pyx_v_self->oct_list[__pyx_v_i])->domain_ind = __pyx_v_i;

    /* "yt/geometry/particle_oct_container.pyx":183
 *         for i in range(self.nocts):
 *             self.oct_list[i].domain_ind = i
 *             self.oct_list[i].domain = domain_id             # <<<<<<<<<<<<<<
 *         self.max_level = max_level
 * 
 */
    (__pyx_v_self->oct_list[__pyx_v_i])->domain = __pyx_v_domain_id;
  }

  /* "yt/geometry/particle_oct_container.pyx":184
 *             self.oct_list[i].domain_ind = i
 *             self.oct_list[i].domain = domain_id
 *         self.max_level = max_level             # <<<<<<<<<<<<<<
 * 
 *     cdef visit_assign(self, Oct *o, np.int64_t *lpos, int level, int *max_level):
 */
  __pyx_v_self->max_level = __pyx_v_max_level;

  /* "yt/geometry/particle_oct_container.pyx":165
 *         pass
 * 
 *     def finalize(self, int domain_id = 0):             # <<<<<<<<<<<<<<
 *         #This will sort the octs in the oct list
 *         #so that domains appear consecutively
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleOctreeContainer.finalize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":186
 *         self.max_level = max_level
 * 
 *     cdef visit_assign(self, Oct *o, np.int64_t *lpos, int level, int *max_level):             # <<<<<<<<<<<<<<
 *         cdef int i, j, k
 *         self.oct_list[lpos[0]] = o
 */

static PyObject *__pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_visit_assign(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o, __pyx_t_5numpy_int64_t *__pyx_v_lpos, int __pyx_v_level, int *__pyx_v_max_level) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("visit_assign", 0);

  /* "yt/geometry/particle_oct_container.pyx":188
 *     cdef visit_assign(self, Oct *o, np.int64_t *lpos, int level, int *max_level):
 *         cdef int i, j, k
 *         self.oct_list[lpos[0]] = o             # <<<<<<<<<<<<<<
 *         lpos[0] += 1
 *         max_level[0] = imax(max_level[0], level)
 */
  (__pyx_v_self->oct_list[(__pyx_v_lpos[0])]) = __pyx_v_o;

  /* "yt/geometry/particle_oct_container.pyx":189
 *         cdef int i, j, k
 *         self.oct_list[lpos[0]] = o
 *         lpos[0] += 1             # <<<<<<<<<<<<<<
 *         max_level[0] = imax(max_level[0], level)
 *         for i in range(2):
 */
  __pyx_t_1 = 0;
  (__pyx_v_lpos[__pyx_t_1]) = ((__pyx_v_lpos[__pyx_t_1]) + 1);

  /* "yt/geometry/particle_oct_container.pyx":190
 *         self.oct_list[lpos[0]] = o
 *         lpos[0] += 1
 *         max_level[0] = imax(max_level[0], level)             # <<<<<<<<<<<<<<
 *         for i in range(2):
 *             for j in range(2):
 */
  (__pyx_v_max_level[0]) = __pyx_f_2yt_9utilities_3lib_8fp_utils_imax((__pyx_v_max_level[0]), __pyx_v_level);

  /* "yt/geometry/particle_oct_container.pyx":191
 *         lpos[0] += 1
 *         max_level[0] = imax(max_level[0], level)
 *         for i in range(2):             # <<<<<<<<<<<<<<
 *             for j in range(2):
 *                 for k in range(2):
 */
  for (__pyx_t_2 = 0; __pyx_t_2 < 2; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "yt/geometry/particle_oct_container.pyx":192
 *         max_level[0] = imax(max_level[0], level)
 *         for i in range(2):
 *             for j in range(2):             # <<<<<<<<<<<<<<
 *                 for k in range(2):
 *                     if o.children != NULL \
 */
    for (__pyx_t_3 = 0; __pyx_t_3 < 2; __pyx_t_3+=1) {
      __pyx_v_j = __pyx_t_3;

      /* "yt/geometry/particle_oct_container.pyx":193
 *         for i in range(2):
 *             for j in range(2):
 *                 for k in range(2):             # <<<<<<<<<<<<<<
 *                     if o.children != NULL \
 *                        and o.children[cind(i,j,k)] != NULL:
 */
      for (__pyx_t_4 = 0; __pyx_t_4 < 2; __pyx_t_4+=1) {
        __pyx_v_k = __pyx_t_4;

        /* "yt/geometry/particle_oct_container.pyx":194
 *             for j in range(2):
 *                 for k in range(2):
 *                     if o.children != NULL \             # <<<<<<<<<<<<<<
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_assign(o.children[cind(i,j,k)], lpos,
 */
        __pyx_t_6 = ((__pyx_v_o->children != NULL) != 0);
        if (__pyx_t_6) {
        } else {
          __pyx_t_5 = __pyx_t_6;
          goto __pyx_L10_bool_binop_done;
        }

        /* "yt/geometry/particle_oct_container.pyx":195
 *                 for k in range(2):
 *                     if o.children != NULL \
 *                        and o.children[cind(i,j,k)] != NULL:             # <<<<<<<<<<<<<<
 *                         self.visit_assign(o.children[cind(i,j,k)], lpos,
 *                                 level + 1, max_level)
 */
        __pyx_t_6 = (((__pyx_v_o->children[__pyx_f_2yt_8geometry_12oct_visitors_cind(__pyx_v_i, __pyx_v_j, __pyx_v_k)]) != NULL) != 0);
        __pyx_t_5 = __pyx_t_6;
        __pyx_L10_bool_binop_done:;

        /* "yt/geometry/particle_oct_container.pyx":194
 *             for j in range(2):
 *                 for k in range(2):
 *                     if o.children != NULL \             # <<<<<<<<<<<<<<
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_assign(o.children[cind(i,j,k)], lpos,
 */
        if (__pyx_t_5) {

          /* "yt/geometry/particle_oct_container.pyx":196
 *                     if o.children != NULL \
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_assign(o.children[cind(i,j,k)], lpos,             # <<<<<<<<<<<<<<
 *                                 level + 1, max_level)
 *         return
 */
          __pyx_t_7 = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_vtab)->visit_assign(__pyx_v_self, (__pyx_v_o->children[__pyx_f_2yt_8geometry_12oct_visitors_cind(__pyx_v_i, __pyx_v_j, __pyx_v_k)]), __pyx_v_lpos, (__pyx_v_level + 1), __pyx_v_max_level); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 196, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

          /* "yt/geometry/particle_oct_container.pyx":194
 *             for j in range(2):
 *                 for k in range(2):
 *                     if o.children != NULL \             # <<<<<<<<<<<<<<
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_assign(o.children[cind(i,j,k)], lpos,
 */
        }
      }
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":198
 *                         self.visit_assign(o.children[cind(i,j,k)], lpos,
 *                                 level + 1, max_level)
 *         return             # <<<<<<<<<<<<<<
 * 
 *     cdef np.int64_t get_domain_offset(self, int domain_id):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":186
 *         self.max_level = max_level
 * 
 *     cdef visit_assign(self, Oct *o, np.int64_t *lpos, int level, int *max_level):             # <<<<<<<<<<<<<<
 *         cdef int i, j, k
 *         self.oct_list[lpos[0]] = o
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleOctreeContainer.visit_assign", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":200
 *         return
 * 
 *     cdef np.int64_t get_domain_offset(self, int domain_id):             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */

static __pyx_t_5numpy_int64_t __pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_get_domain_offset(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, CYTHON_UNUSED int __pyx_v_domain_id) {
  __pyx_t_5numpy_int64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_domain_offset", 0);

  /* "yt/geometry/particle_oct_container.pyx":201
 * 
 *     cdef np.int64_t get_domain_offset(self, int domain_id):
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef Oct* allocate_oct(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":200
 *         return
 * 
 *     cdef np.int64_t get_domain_offset(self, int domain_id):             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":203
 *         return 0
 * 
 *     cdef Oct* allocate_oct(self):             # <<<<<<<<<<<<<<
 *         #Allocate the memory, set to NULL or -1
 *         #We reserve space for n_ref particles, but keep
 */

static struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_allocate_oct(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self) {
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_my_oct;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("allocate_oct", 0);

  /* "yt/geometry/particle_oct_container.pyx":207
 *         #We reserve space for n_ref particles, but keep
 *         #track of how many are used with np initially 0
 *         self.nocts += 1             # <<<<<<<<<<<<<<
 *         cdef Oct *my_oct = <Oct*> malloc(sizeof(Oct))
 *         my_oct.domain = -1
 */
  __pyx_v_self->__pyx_base.nocts = (__pyx_v_self->__pyx_base.nocts + 1);

  /* "yt/geometry/particle_oct_container.pyx":208
 *         #track of how many are used with np initially 0
 *         self.nocts += 1
 *         cdef Oct *my_oct = <Oct*> malloc(sizeof(Oct))             # <<<<<<<<<<<<<<
 *         my_oct.domain = -1
 *         my_oct.file_ind = 0
 */
  __pyx_v_my_oct = ((struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *)malloc((sizeof(struct __pyx_t_2yt_8geometry_12oct_visitors_Oct))));

  /* "yt/geometry/particle_oct_container.pyx":209
 *         self.nocts += 1
 *         cdef Oct *my_oct = <Oct*> malloc(sizeof(Oct))
 *         my_oct.domain = -1             # <<<<<<<<<<<<<<
 *         my_oct.file_ind = 0
 *         my_oct.domain_ind = self.nocts - 1
 */
  __pyx_v_my_oct->domain = -1LL;

  /* "yt/geometry/particle_oct_container.pyx":210
 *         cdef Oct *my_oct = <Oct*> malloc(sizeof(Oct))
 *         my_oct.domain = -1
 *         my_oct.file_ind = 0             # <<<<<<<<<<<<<<
 *         my_oct.domain_ind = self.nocts - 1
 *         my_oct.children = NULL
 */
  __pyx_v_my_oct->file_ind = 0;

  /* "yt/geometry/particle_oct_container.pyx":211
 *         my_oct.domain = -1
 *         my_oct.file_ind = 0
 *         my_oct.domain_ind = self.nocts - 1             # <<<<<<<<<<<<<<
 *         my_oct.children = NULL
 *         return my_oct
 */
  __pyx_v_my_oct->domain_ind = (__pyx_v_self->__pyx_base.nocts - 1);

  /* "yt/geometry/particle_oct_container.pyx":212
 *         my_oct.file_ind = 0
 *         my_oct.domain_ind = self.nocts - 1
 *         my_oct.children = NULL             # <<<<<<<<<<<<<<
 *         return my_oct
 * 
 */
  __pyx_v_my_oct->children = NULL;

  /* "yt/geometry/particle_oct_container.pyx":213
 *         my_oct.domain_ind = self.nocts - 1
 *         my_oct.children = NULL
 *         return my_oct             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = __pyx_v_my_oct;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":203
 *         return 0
 * 
 *     cdef Oct* allocate_oct(self):             # <<<<<<<<<<<<<<
 *         #Allocate the memory, set to NULL or -1
 *         #We reserve space for n_ref particles, but keep
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":218
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def add(self, np.ndarray[np.uint64_t, ndim=1] indices,             # <<<<<<<<<<<<<<
 *             np.uint8_t order = ORDER_MAX):
 *         #Add this particle to the root oct
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_14add(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_14add(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_indices = 0;
  __pyx_t_5numpy_uint8_t __pyx_v_order;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_indices,&__pyx_n_s_order,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_order);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "add") < 0)) __PYX_ERR(0, 218, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_indices = ((PyArrayObject *)values[0]);
    if (values[1]) {
      __pyx_v_order = __Pyx_PyInt_As_npy_uint8(values[1]); if (unlikely((__pyx_v_order == ((npy_uint8)-1)) && PyErr_Occurred())) __PYX_ERR(0, 219, __pyx_L3_error)
    } else {
      __pyx_v_order = __pyx_k_;
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 218, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleOctreeContainer.add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(0, 218, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_13add(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self), __pyx_v_indices, __pyx_v_order);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_13add(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, PyArrayObject *__pyx_v_indices, __pyx_t_5numpy_uint8_t __pyx_v_order) {
  __pyx_t_5numpy_int64_t __pyx_v_no;
  __pyx_t_5numpy_int64_t __pyx_v_p;
  __pyx_t_5numpy_uint64_t __pyx_v_index;
  int __pyx_v_i;
  int __pyx_v_level;
  int __pyx_v_ind[3];
  __pyx_t_5numpy_uint64_t *__pyx_v_data;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_cur;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __pyx_t_5numpy_int64_t __pyx_t_6;
  __pyx_t_5numpy_int64_t __pyx_t_7;
  __pyx_t_5numpy_int64_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add", 0);
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 218, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/particle_oct_container.pyx":223
 *         #Then if that oct has children, add it to them recursively
 *         #If the child needs to be refined because of max particles, do so
 *         cdef np.int64_t no = indices.shape[0], p             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t index
 *         cdef int i, level
 */
  __pyx_v_no = (__pyx_v_indices->dimensions[0]);

  /* "yt/geometry/particle_oct_container.pyx":227
 *         cdef int i, level
 *         cdef int ind[3]
 *         if self.root_mesh[0][0][0] == NULL: self.allocate_root()             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t *data = <np.uint64_t *> indices.data
 *         for p in range(no):
 */
  __pyx_t_1 = (((((__pyx_v_self->__pyx_base.root_mesh[0])[0])[0]) == NULL) != 0);
  if (__pyx_t_1) {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_allocate_root); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "yt/geometry/particle_oct_container.pyx":228
 *         cdef int ind[3]
 *         if self.root_mesh[0][0][0] == NULL: self.allocate_root()
 *         cdef np.uint64_t *data = <np.uint64_t *> indices.data             # <<<<<<<<<<<<<<
 *         for p in range(no):
 *             # We have morton indices, which means we choose left and right by
 */
  __pyx_v_data = ((__pyx_t_5numpy_uint64_t *)__pyx_v_indices->data);

  /* "yt/geometry/particle_oct_container.pyx":229
 *         if self.root_mesh[0][0][0] == NULL: self.allocate_root()
 *         cdef np.uint64_t *data = <np.uint64_t *> indices.data
 *         for p in range(no):             # <<<<<<<<<<<<<<
 *             # We have morton indices, which means we choose left and right by
 *             # looking at (MAX_ORDER - level) & with the values 1, 2, 4.
 */
  __pyx_t_5 = __pyx_v_no;
  __pyx_t_6 = __pyx_t_5;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_p = __pyx_t_7;

    /* "yt/geometry/particle_oct_container.pyx":232
 *             # We have morton indices, which means we choose left and right by
 *             # looking at (MAX_ORDER - level) & with the values 1, 2, 4.
 *             level = 0             # <<<<<<<<<<<<<<
 *             index = indices[p]
 *             if index == FLAG:
 */
    __pyx_v_level = 0;

    /* "yt/geometry/particle_oct_container.pyx":233
 *             # looking at (MAX_ORDER - level) & with the values 1, 2, 4.
 *             level = 0
 *             index = indices[p]             # <<<<<<<<<<<<<<
 *             if index == FLAG:
 *                 # This is a marker for the index not being inside the domain
 */
    __pyx_t_8 = __pyx_v_p;
    __pyx_v_index = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_indices.rcbuffer->pybuffer.buf, __pyx_t_8, __pyx_pybuffernd_indices.diminfo[0].strides));

    /* "yt/geometry/particle_oct_container.pyx":234
 *             level = 0
 *             index = indices[p]
 *             if index == FLAG:             # <<<<<<<<<<<<<<
 *                 # This is a marker for the index not being inside the domain
 *                 # we're interested in.
 */
    __pyx_t_1 = ((__pyx_v_index == __pyx_v_2yt_8geometry_22particle_oct_container_FLAG) != 0);
    if (__pyx_t_1) {

      /* "yt/geometry/particle_oct_container.pyx":237
 *                 # This is a marker for the index not being inside the domain
 *                 # we're interested in.
 *                 continue             # <<<<<<<<<<<<<<
 *             # Convert morton index to 3D index of octree root
 *             for i in range(3):
 */
      goto __pyx_L4_continue;

      /* "yt/geometry/particle_oct_container.pyx":234
 *             level = 0
 *             index = indices[p]
 *             if index == FLAG:             # <<<<<<<<<<<<<<
 *                 # This is a marker for the index not being inside the domain
 *                 # we're interested in.
 */
    }

    /* "yt/geometry/particle_oct_container.pyx":239
 *                 continue
 *             # Convert morton index to 3D index of octree root
 *             for i in range(3):             # <<<<<<<<<<<<<<
 *                 ind[i] = (index >> ((order - level)*3 + (2 - i))) & 1
 *             cur = self.root_mesh[ind[0]][ind[1]][ind[2]]
 */
    for (__pyx_t_9 = 0; __pyx_t_9 < 3; __pyx_t_9+=1) {
      __pyx_v_i = __pyx_t_9;

      /* "yt/geometry/particle_oct_container.pyx":240
 *             # Convert morton index to 3D index of octree root
 *             for i in range(3):
 *                 ind[i] = (index >> ((order - level)*3 + (2 - i))) & 1             # <<<<<<<<<<<<<<
 *             cur = self.root_mesh[ind[0]][ind[1]][ind[2]]
 *             if cur == NULL:
 */
      (__pyx_v_ind[__pyx_v_i]) = ((__pyx_v_index >> (((__pyx_v_order - __pyx_v_level) * 3) + (2 - __pyx_v_i))) & 1);
    }

    /* "yt/geometry/particle_oct_container.pyx":241
 *             for i in range(3):
 *                 ind[i] = (index >> ((order - level)*3 + (2 - i))) & 1
 *             cur = self.root_mesh[ind[0]][ind[1]][ind[2]]             # <<<<<<<<<<<<<<
 *             if cur == NULL:
 *                 raise RuntimeError
 */
    __pyx_v_cur = (((__pyx_v_self->__pyx_base.root_mesh[(__pyx_v_ind[0])])[(__pyx_v_ind[1])])[(__pyx_v_ind[2])]);

    /* "yt/geometry/particle_oct_container.pyx":242
 *                 ind[i] = (index >> ((order - level)*3 + (2 - i))) & 1
 *             cur = self.root_mesh[ind[0]][ind[1]][ind[2]]
 *             if cur == NULL:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError
 *             # Continue refining the octree until you reach the level of the
 */
    __pyx_t_1 = ((__pyx_v_cur == NULL) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "yt/geometry/particle_oct_container.pyx":243
 *             cur = self.root_mesh[ind[0]][ind[1]][ind[2]]
 *             if cur == NULL:
 *                 raise RuntimeError             # <<<<<<<<<<<<<<
 *             # Continue refining the octree until you reach the level of the
 *             # morton indexing order. Along the way, use prefix to count
 */
      __Pyx_Raise(__pyx_builtin_RuntimeError, 0, 0, 0);
      __PYX_ERR(0, 243, __pyx_L1_error)

      /* "yt/geometry/particle_oct_container.pyx":242
 *                 ind[i] = (index >> ((order - level)*3 + (2 - i))) & 1
 *             cur = self.root_mesh[ind[0]][ind[1]][ind[2]]
 *             if cur == NULL:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError
 *             # Continue refining the octree until you reach the level of the
 */
    }

    /* "yt/geometry/particle_oct_container.pyx":247
 *             # morton indexing order. Along the way, use prefix to count
 *             # previous indices at levels in the octree?
 *             while (cur.file_ind + 1) > self.n_ref:             # <<<<<<<<<<<<<<
 *                 if level >= order: break # Just dump it here.
 *                 level += 1
 */
    while (1) {
      __pyx_t_1 = (((__pyx_v_cur->file_ind + 1) > __pyx_v_self->n_ref) != 0);
      if (!__pyx_t_1) break;

      /* "yt/geometry/particle_oct_container.pyx":248
 *             # previous indices at levels in the octree?
 *             while (cur.file_ind + 1) > self.n_ref:
 *                 if level >= order: break # Just dump it here.             # <<<<<<<<<<<<<<
 *                 level += 1
 *                 for i in range(3):
 */
      __pyx_t_1 = ((__pyx_v_level >= __pyx_v_order) != 0);
      if (__pyx_t_1) {
        goto __pyx_L11_break;
      }

      /* "yt/geometry/particle_oct_container.pyx":249
 *             while (cur.file_ind + 1) > self.n_ref:
 *                 if level >= order: break # Just dump it here.
 *                 level += 1             # <<<<<<<<<<<<<<
 *                 for i in range(3):
 *                     ind[i] = (index >> ((order - level)*3 + (2 - i))) & 1
 */
      __pyx_v_level = (__pyx_v_level + 1);

      /* "yt/geometry/particle_oct_container.pyx":250
 *                 if level >= order: break # Just dump it here.
 *                 level += 1
 *                 for i in range(3):             # <<<<<<<<<<<<<<
 *                     ind[i] = (index >> ((order - level)*3 + (2 - i))) & 1
 *                 if cur.children == NULL or \
 */
      for (__pyx_t_9 = 0; __pyx_t_9 < 3; __pyx_t_9+=1) {
        __pyx_v_i = __pyx_t_9;

        /* "yt/geometry/particle_oct_container.pyx":251
 *                 level += 1
 *                 for i in range(3):
 *                     ind[i] = (index >> ((order - level)*3 + (2 - i))) & 1             # <<<<<<<<<<<<<<
 *                 if cur.children == NULL or \
 *                    cur.children[cind(ind[0],ind[1],ind[2])] == NULL:
 */
        (__pyx_v_ind[__pyx_v_i]) = ((__pyx_v_index >> (((__pyx_v_order - __pyx_v_level) * 3) + (2 - __pyx_v_i))) & 1);
      }

      /* "yt/geometry/particle_oct_container.pyx":252
 *                 for i in range(3):
 *                     ind[i] = (index >> ((order - level)*3 + (2 - i))) & 1
 *                 if cur.children == NULL or \             # <<<<<<<<<<<<<<
 *                    cur.children[cind(ind[0],ind[1],ind[2])] == NULL:
 *                     cur = self.refine_oct(cur, index, level, order)
 */
      __pyx_t_10 = ((__pyx_v_cur->children == NULL) != 0);
      if (!__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L16_bool_binop_done;
      }

      /* "yt/geometry/particle_oct_container.pyx":253
 *                     ind[i] = (index >> ((order - level)*3 + (2 - i))) & 1
 *                 if cur.children == NULL or \
 *                    cur.children[cind(ind[0],ind[1],ind[2])] == NULL:             # <<<<<<<<<<<<<<
 *                     cur = self.refine_oct(cur, index, level, order)
 *                     self.filter_particles(cur, data, p, level, order)
 */
      __pyx_t_10 = (((__pyx_v_cur->children[__pyx_f_2yt_8geometry_12oct_visitors_cind((__pyx_v_ind[0]), (__pyx_v_ind[1]), (__pyx_v_ind[2]))]) == NULL) != 0);
      __pyx_t_1 = __pyx_t_10;
      __pyx_L16_bool_binop_done:;

      /* "yt/geometry/particle_oct_container.pyx":252
 *                 for i in range(3):
 *                     ind[i] = (index >> ((order - level)*3 + (2 - i))) & 1
 *                 if cur.children == NULL or \             # <<<<<<<<<<<<<<
 *                    cur.children[cind(ind[0],ind[1],ind[2])] == NULL:
 *                     cur = self.refine_oct(cur, index, level, order)
 */
      if (__pyx_t_1) {

        /* "yt/geometry/particle_oct_container.pyx":254
 *                 if cur.children == NULL or \
 *                    cur.children[cind(ind[0],ind[1],ind[2])] == NULL:
 *                     cur = self.refine_oct(cur, index, level, order)             # <<<<<<<<<<<<<<
 *                     self.filter_particles(cur, data, p, level, order)
 *                 else:
 */
        __pyx_v_cur = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_vtab)->refine_oct(__pyx_v_self, __pyx_v_cur, __pyx_v_index, __pyx_v_level, __pyx_v_order);

        /* "yt/geometry/particle_oct_container.pyx":255
 *                    cur.children[cind(ind[0],ind[1],ind[2])] == NULL:
 *                     cur = self.refine_oct(cur, index, level, order)
 *                     self.filter_particles(cur, data, p, level, order)             # <<<<<<<<<<<<<<
 *                 else:
 *                     cur = cur.children[cind(ind[0],ind[1],ind[2])]
 */
        ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_vtab)->filter_particles(__pyx_v_self, __pyx_v_cur, __pyx_v_data, __pyx_v_p, __pyx_v_level, __pyx_v_order);

        /* "yt/geometry/particle_oct_container.pyx":252
 *                 for i in range(3):
 *                     ind[i] = (index >> ((order - level)*3 + (2 - i))) & 1
 *                 if cur.children == NULL or \             # <<<<<<<<<<<<<<
 *                    cur.children[cind(ind[0],ind[1],ind[2])] == NULL:
 *                     cur = self.refine_oct(cur, index, level, order)
 */
        goto __pyx_L15;
      }

      /* "yt/geometry/particle_oct_container.pyx":257
 *                     self.filter_particles(cur, data, p, level, order)
 *                 else:
 *                     cur = cur.children[cind(ind[0],ind[1],ind[2])]             # <<<<<<<<<<<<<<
 *             # If our n_ref is 1, we are always refining, which means we're an
 *             # index octree.  In this case, we should store the index for fast
 */
      /*else*/ {
        __pyx_v_cur = (__pyx_v_cur->children[__pyx_f_2yt_8geometry_12oct_visitors_cind((__pyx_v_ind[0]), (__pyx_v_ind[1]), (__pyx_v_ind[2]))]);
      }
      __pyx_L15:;
    }
    __pyx_L11_break:;

    /* "yt/geometry/particle_oct_container.pyx":261
 *             # index octree.  In this case, we should store the index for fast
 *             # lookup later on when we find neighbors and the like.
 *             if self.n_ref == 1:             # <<<<<<<<<<<<<<
 *                 cur.file_ind = index
 *             else:
 */
    __pyx_t_1 = ((__pyx_v_self->n_ref == 1) != 0);
    if (__pyx_t_1) {

      /* "yt/geometry/particle_oct_container.pyx":262
 *             # lookup later on when we find neighbors and the like.
 *             if self.n_ref == 1:
 *                 cur.file_ind = index             # <<<<<<<<<<<<<<
 *             else:
 *                 cur.file_ind += 1
 */
      __pyx_v_cur->file_ind = __pyx_v_index;

      /* "yt/geometry/particle_oct_container.pyx":261
 *             # index octree.  In this case, we should store the index for fast
 *             # lookup later on when we find neighbors and the like.
 *             if self.n_ref == 1:             # <<<<<<<<<<<<<<
 *                 cur.file_ind = index
 *             else:
 */
      goto __pyx_L18;
    }

    /* "yt/geometry/particle_oct_container.pyx":264
 *                 cur.file_ind = index
 *             else:
 *                 cur.file_ind += 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    /*else*/ {
      __pyx_v_cur->file_ind = (__pyx_v_cur->file_ind + 1);
    }
    __pyx_L18:;
    __pyx_L4_continue:;
  }

  /* "yt/geometry/particle_oct_container.pyx":218
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def add(self, np.ndarray[np.uint64_t, ndim=1] indices,             # <<<<<<<<<<<<<<
 *             np.uint8_t order = ORDER_MAX):
 *         #Add this particle to the root oct
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleOctreeContainer.add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":269
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef Oct *refine_oct(self, Oct *o, np.uint64_t index, int level,             # <<<<<<<<<<<<<<
 *                          np.uint8_t order):
 *         #Allocate and initialize child octs
 */

static struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_refine_oct(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o, __pyx_t_5numpy_uint64_t __pyx_v_index, int __pyx_v_level, __pyx_t_5numpy_uint8_t __pyx_v_order) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_ind[3];
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_noct;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __pyx_t_5numpy_int64_t __pyx_t_4;
  __Pyx_RefNannySetupContext("refine_oct", 0);

  /* "yt/geometry/particle_oct_container.pyx":278
 *         cdef Oct *noct
 *         # TODO: This does not need to be changed.
 *         o.children = <Oct **> malloc(sizeof(Oct *)*8)             # <<<<<<<<<<<<<<
 *         for i in range(2):
 *             for j in range(2):
 */
  __pyx_v_o->children = ((struct __pyx_t_2yt_8geometry_12oct_visitors_Oct **)malloc(((sizeof(struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *)) * 8)));

  /* "yt/geometry/particle_oct_container.pyx":279
 *         # TODO: This does not need to be changed.
 *         o.children = <Oct **> malloc(sizeof(Oct *)*8)
 *         for i in range(2):             # <<<<<<<<<<<<<<
 *             for j in range(2):
 *                 for k in range(2):
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 2; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":280
 *         o.children = <Oct **> malloc(sizeof(Oct *)*8)
 *         for i in range(2):
 *             for j in range(2):             # <<<<<<<<<<<<<<
 *                 for k in range(2):
 *                     noct = self.allocate_oct()
 */
    for (__pyx_t_2 = 0; __pyx_t_2 < 2; __pyx_t_2+=1) {
      __pyx_v_j = __pyx_t_2;

      /* "yt/geometry/particle_oct_container.pyx":281
 *         for i in range(2):
 *             for j in range(2):
 *                 for k in range(2):             # <<<<<<<<<<<<<<
 *                     noct = self.allocate_oct()
 *                     noct.domain = o.domain
 */
      for (__pyx_t_3 = 0; __pyx_t_3 < 2; __pyx_t_3+=1) {
        __pyx_v_k = __pyx_t_3;

        /* "yt/geometry/particle_oct_container.pyx":282
 *             for j in range(2):
 *                 for k in range(2):
 *                     noct = self.allocate_oct()             # <<<<<<<<<<<<<<
 *                     noct.domain = o.domain
 *                     noct.file_ind = 0
 */
        __pyx_v_noct = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_vtab)->allocate_oct(__pyx_v_self);

        /* "yt/geometry/particle_oct_container.pyx":283
 *                 for k in range(2):
 *                     noct = self.allocate_oct()
 *                     noct.domain = o.domain             # <<<<<<<<<<<<<<
 *                     noct.file_ind = 0
 *                     o.children[cind(i,j,k)] = noct
 */
        __pyx_t_4 = __pyx_v_o->domain;
        __pyx_v_noct->domain = __pyx_t_4;

        /* "yt/geometry/particle_oct_container.pyx":284
 *                     noct = self.allocate_oct()
 *                     noct.domain = o.domain
 *                     noct.file_ind = 0             # <<<<<<<<<<<<<<
 *                     o.children[cind(i,j,k)] = noct
 *         o.file_ind = self.n_ref + 1
 */
        __pyx_v_noct->file_ind = 0;

        /* "yt/geometry/particle_oct_container.pyx":285
 *                     noct.domain = o.domain
 *                     noct.file_ind = 0
 *                     o.children[cind(i,j,k)] = noct             # <<<<<<<<<<<<<<
 *         o.file_ind = self.n_ref + 1
 *         for i in range(3):
 */
        (__pyx_v_o->children[__pyx_f_2yt_8geometry_12oct_visitors_cind(__pyx_v_i, __pyx_v_j, __pyx_v_k)]) = __pyx_v_noct;
      }
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":286
 *                     noct.file_ind = 0
 *                     o.children[cind(i,j,k)] = noct
 *         o.file_ind = self.n_ref + 1             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             ind[i] = (index >> ((order - level)*3 + (2 - i))) & 1
 */
  __pyx_v_o->file_ind = (__pyx_v_self->n_ref + 1);

  /* "yt/geometry/particle_oct_container.pyx":287
 *                     o.children[cind(i,j,k)] = noct
 *         o.file_ind = self.n_ref + 1
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             ind[i] = (index >> ((order - level)*3 + (2 - i))) & 1
 *         noct = o.children[cind(ind[0],ind[1],ind[2])]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":288
 *         o.file_ind = self.n_ref + 1
 *         for i in range(3):
 *             ind[i] = (index >> ((order - level)*3 + (2 - i))) & 1             # <<<<<<<<<<<<<<
 *         noct = o.children[cind(ind[0],ind[1],ind[2])]
 *         return noct
 */
    (__pyx_v_ind[__pyx_v_i]) = ((__pyx_v_index >> (((__pyx_v_order - __pyx_v_level) * 3) + (2 - __pyx_v_i))) & 1);
  }

  /* "yt/geometry/particle_oct_container.pyx":289
 *         for i in range(3):
 *             ind[i] = (index >> ((order - level)*3 + (2 - i))) & 1
 *         noct = o.children[cind(ind[0],ind[1],ind[2])]             # <<<<<<<<<<<<<<
 *         return noct
 * 
 */
  __pyx_v_noct = (__pyx_v_o->children[__pyx_f_2yt_8geometry_12oct_visitors_cind((__pyx_v_ind[0]), (__pyx_v_ind[1]), (__pyx_v_ind[2]))]);

  /* "yt/geometry/particle_oct_container.pyx":290
 *             ind[i] = (index >> ((order - level)*3 + (2 - i))) & 1
 *         noct = o.children[cind(ind[0],ind[1],ind[2])]
 *         return noct             # <<<<<<<<<<<<<<
 * 
 *     cdef void filter_particles(self, Oct *o, np.uint64_t *data, np.int64_t p,
 */
  __pyx_r = __pyx_v_noct;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":269
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef Oct *refine_oct(self, Oct *o, np.uint64_t index, int level,             # <<<<<<<<<<<<<<
 *                          np.uint8_t order):
 *         #Allocate and initialize child octs
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":292
 *         return noct
 * 
 *     cdef void filter_particles(self, Oct *o, np.uint64_t *data, np.int64_t p,             # <<<<<<<<<<<<<<
 *                                int level, np.uint8_t order):
 *         # Now we look at the last nref particles to decide where they go.
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_filter_particles(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o, __pyx_t_5numpy_uint64_t *__pyx_v_data, __pyx_t_5numpy_int64_t __pyx_v_p, int __pyx_v_level, __pyx_t_5numpy_uint8_t __pyx_v_order) {
  int __pyx_v_n;
  __pyx_t_5numpy_uint64_t *__pyx_v_arr;
  __pyx_t_5numpy_uint64_t __pyx_v_prefix1;
  __pyx_t_5numpy_uint64_t __pyx_v_prefix2;
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("filter_particles", 0);

  /* "yt/geometry/particle_oct_container.pyx":297
 *         # If p: Loops over all previous morton indices
 *         # If n_ref: Loops over n_ref previous morton indices
 *         cdef int n = imin(p, self.n_ref)             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t *arr = data + imax(p - self.n_ref, 0)
 *         cdef np.uint64_t prefix1, prefix2
 */
  __pyx_v_n = __pyx_f_2yt_9utilities_3lib_8fp_utils_imin(__pyx_v_p, __pyx_v_self->n_ref);

  /* "yt/geometry/particle_oct_container.pyx":298
 *         # If n_ref: Loops over n_ref previous morton indices
 *         cdef int n = imin(p, self.n_ref)
 *         cdef np.uint64_t *arr = data + imax(p - self.n_ref, 0)             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t prefix1, prefix2
 *         # Now we figure out our prefix, which is the oct address at this level.
 */
  __pyx_v_arr = (__pyx_v_data + __pyx_f_2yt_9utilities_3lib_8fp_utils_imax((__pyx_v_p - __pyx_v_self->n_ref), 0));

  /* "yt/geometry/particle_oct_container.pyx":303
 *         # As long as we're actually in Morton order, we do not need to worry
 *         # about *any* of the other children of the oct.
 *         prefix1 = data[p] >> (order - level)*3             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             prefix2 = arr[i] >> (order - level)*3
 */
  __pyx_v_prefix1 = ((__pyx_v_data[__pyx_v_p]) >> ((__pyx_v_order - __pyx_v_level) * 3));

  /* "yt/geometry/particle_oct_container.pyx":304
 *         # about *any* of the other children of the oct.
 *         prefix1 = data[p] >> (order - level)*3
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             prefix2 = arr[i] >> (order - level)*3
 *             if (prefix1 == prefix2):
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":305
 *         prefix1 = data[p] >> (order - level)*3
 *         for i in range(n):
 *             prefix2 = arr[i] >> (order - level)*3             # <<<<<<<<<<<<<<
 *             if (prefix1 == prefix2):
 *                 o.file_ind += 1 # Says how many morton indices are in this octant?
 */
    __pyx_v_prefix2 = ((__pyx_v_arr[__pyx_v_i]) >> ((__pyx_v_order - __pyx_v_level) * 3));

    /* "yt/geometry/particle_oct_container.pyx":306
 *         for i in range(n):
 *             prefix2 = arr[i] >> (order - level)*3
 *             if (prefix1 == prefix2):             # <<<<<<<<<<<<<<
 *                 o.file_ind += 1 # Says how many morton indices are in this octant?
 * 
 */
    __pyx_t_4 = ((__pyx_v_prefix1 == __pyx_v_prefix2) != 0);
    if (__pyx_t_4) {

      /* "yt/geometry/particle_oct_container.pyx":307
 *             prefix2 = arr[i] >> (order - level)*3
 *             if (prefix1 == prefix2):
 *                 o.file_ind += 1 # Says how many morton indices are in this octant?             # <<<<<<<<<<<<<<
 * 
 *     def recursively_count(self):
 */
      __pyx_v_o->file_ind = (__pyx_v_o->file_ind + 1);

      /* "yt/geometry/particle_oct_container.pyx":306
 *         for i in range(n):
 *             prefix2 = arr[i] >> (order - level)*3
 *             if (prefix1 == prefix2):             # <<<<<<<<<<<<<<
 *                 o.file_ind += 1 # Says how many morton indices are in this octant?
 * 
 */
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":292
 *         return noct
 * 
 *     cdef void filter_particles(self, Oct *o, np.uint64_t *data, np.int64_t p,             # <<<<<<<<<<<<<<
 *                                int level, np.uint8_t order):
 *         # Now we look at the last nref particles to decide where they go.
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":309
 *                 o.file_ind += 1 # Says how many morton indices are in this octant?
 * 
 *     def recursively_count(self):             # <<<<<<<<<<<<<<
 *         #Visit every cell, accumulate the # of cells per level
 *         cdef int i, j, k
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_16recursively_count(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_16recursively_count(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("recursively_count (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_15recursively_count(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_15recursively_count(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  __pyx_t_5numpy_int64_t __pyx_v_counts[0x80];
  PyObject *__pyx_v_level_counts = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("recursively_count", 0);

  /* "yt/geometry/particle_oct_container.pyx":313
 *         cdef int i, j, k
 *         cdef np.int64_t counts[128]
 *         for i in range(128): counts[i] = 0             # <<<<<<<<<<<<<<
 *         for i in range(self.nn[0]):
 *             for j in range(self.nn[1]):
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 0x80; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;
    (__pyx_v_counts[__pyx_v_i]) = 0;
  }

  /* "yt/geometry/particle_oct_container.pyx":314
 *         cdef np.int64_t counts[128]
 *         for i in range(128): counts[i] = 0
 *         for i in range(self.nn[0]):             # <<<<<<<<<<<<<<
 *             for j in range(self.nn[1]):
 *                 for k in range(self.nn[2]):
 */
  __pyx_t_1 = (__pyx_v_self->__pyx_base.nn[0]);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":315
 *         for i in range(128): counts[i] = 0
 *         for i in range(self.nn[0]):
 *             for j in range(self.nn[1]):             # <<<<<<<<<<<<<<
 *                 for k in range(self.nn[2]):
 *                     if self.root_mesh[i][j][k] != NULL:
 */
    __pyx_t_4 = (__pyx_v_self->__pyx_base.nn[1]);
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "yt/geometry/particle_oct_container.pyx":316
 *         for i in range(self.nn[0]):
 *             for j in range(self.nn[1]):
 *                 for k in range(self.nn[2]):             # <<<<<<<<<<<<<<
 *                     if self.root_mesh[i][j][k] != NULL:
 *                         self.visit(self.root_mesh[i][j][k], counts)
 */
      __pyx_t_7 = (__pyx_v_self->__pyx_base.nn[2]);
      __pyx_t_8 = __pyx_t_7;
      for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
        __pyx_v_k = __pyx_t_9;

        /* "yt/geometry/particle_oct_container.pyx":317
 *             for j in range(self.nn[1]):
 *                 for k in range(self.nn[2]):
 *                     if self.root_mesh[i][j][k] != NULL:             # <<<<<<<<<<<<<<
 *                         self.visit(self.root_mesh[i][j][k], counts)
 *         level_counts = {}
 */
        __pyx_t_10 = (((((__pyx_v_self->__pyx_base.root_mesh[__pyx_v_i])[__pyx_v_j])[__pyx_v_k]) != NULL) != 0);
        if (__pyx_t_10) {

          /* "yt/geometry/particle_oct_container.pyx":318
 *                 for k in range(self.nn[2]):
 *                     if self.root_mesh[i][j][k] != NULL:
 *                         self.visit(self.root_mesh[i][j][k], counts)             # <<<<<<<<<<<<<<
 *         level_counts = {}
 *         for i in range(128):
 */
          __pyx_t_11 = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_vtab)->visit(__pyx_v_self, (((__pyx_v_self->__pyx_base.root_mesh[__pyx_v_i])[__pyx_v_j])[__pyx_v_k]), __pyx_v_counts, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 318, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

          /* "yt/geometry/particle_oct_container.pyx":317
 *             for j in range(self.nn[1]):
 *                 for k in range(self.nn[2]):
 *                     if self.root_mesh[i][j][k] != NULL:             # <<<<<<<<<<<<<<
 *                         self.visit(self.root_mesh[i][j][k], counts)
 *         level_counts = {}
 */
        }
      }
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":319
 *                     if self.root_mesh[i][j][k] != NULL:
 *                         self.visit(self.root_mesh[i][j][k], counts)
 *         level_counts = {}             # <<<<<<<<<<<<<<
 *         for i in range(128):
 *             if counts[i] == 0: break
 */
  __pyx_t_11 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_v_level_counts = ((PyObject*)__pyx_t_11);
  __pyx_t_11 = 0;

  /* "yt/geometry/particle_oct_container.pyx":320
 *                         self.visit(self.root_mesh[i][j][k], counts)
 *         level_counts = {}
 *         for i in range(128):             # <<<<<<<<<<<<<<
 *             if counts[i] == 0: break
 *             level_counts[i] = counts[i]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 0x80; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":321
 *         level_counts = {}
 *         for i in range(128):
 *             if counts[i] == 0: break             # <<<<<<<<<<<<<<
 *             level_counts[i] = counts[i]
 *         return level_counts
 */
    __pyx_t_10 = (((__pyx_v_counts[__pyx_v_i]) == 0) != 0);
    if (__pyx_t_10) {
      goto __pyx_L13_break;
    }

    /* "yt/geometry/particle_oct_container.pyx":322
 *         for i in range(128):
 *             if counts[i] == 0: break
 *             level_counts[i] = counts[i]             # <<<<<<<<<<<<<<
 *         return level_counts
 * 
 */
    __pyx_t_11 = __Pyx_PyInt_From_npy_int64((__pyx_v_counts[__pyx_v_i])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    if (unlikely(PyDict_SetItem(__pyx_v_level_counts, __pyx_t_12, __pyx_t_11) < 0)) __PYX_ERR(0, 322, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  }
  __pyx_L13_break:;

  /* "yt/geometry/particle_oct_container.pyx":323
 *             if counts[i] == 0: break
 *             level_counts[i] = counts[i]
 *         return level_counts             # <<<<<<<<<<<<<<
 * 
 *     cdef visit(self, Oct *o, np.int64_t *counts, level = 0):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_level_counts);
  __pyx_r = __pyx_v_level_counts;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":309
 *                 o.file_ind += 1 # Says how many morton indices are in this octant?
 * 
 *     def recursively_count(self):             # <<<<<<<<<<<<<<
 *         #Visit every cell, accumulate the # of cells per level
 *         cdef int i, j, k
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleOctreeContainer.recursively_count", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_level_counts);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":325
 *         return level_counts
 * 
 *     cdef visit(self, Oct *o, np.int64_t *counts, level = 0):             # <<<<<<<<<<<<<<
 *         cdef int i, j, k
 *         counts[level] += 1
 */

static PyObject *__pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_visit(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o, __pyx_t_5numpy_int64_t *__pyx_v_counts, struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_visit *__pyx_optional_args) {
  PyObject *__pyx_v_level = ((PyObject *)__pyx_int_0);
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_visit __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("visit", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_level = __pyx_optional_args->level;
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":327
 *     cdef visit(self, Oct *o, np.int64_t *counts, level = 0):
 *         cdef int i, j, k
 *         counts[level] += 1             # <<<<<<<<<<<<<<
 *         for i in range(2):
 *             for j in range(2):
 */
  __pyx_t_1 = __Pyx_PyIndex_AsSsize_t(__pyx_v_level); if (unlikely((__pyx_t_1 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 327, __pyx_L1_error)
  (__pyx_v_counts[__pyx_t_1]) = ((__pyx_v_counts[__pyx_t_1]) + 1);

  /* "yt/geometry/particle_oct_container.pyx":328
 *         cdef int i, j, k
 *         counts[level] += 1
 *         for i in range(2):             # <<<<<<<<<<<<<<
 *             for j in range(2):
 *                 for k in range(2):
 */
  for (__pyx_t_2 = 0; __pyx_t_2 < 2; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "yt/geometry/particle_oct_container.pyx":329
 *         counts[level] += 1
 *         for i in range(2):
 *             for j in range(2):             # <<<<<<<<<<<<<<
 *                 for k in range(2):
 *                     if o.children != NULL \
 */
    for (__pyx_t_3 = 0; __pyx_t_3 < 2; __pyx_t_3+=1) {
      __pyx_v_j = __pyx_t_3;

      /* "yt/geometry/particle_oct_container.pyx":330
 *         for i in range(2):
 *             for j in range(2):
 *                 for k in range(2):             # <<<<<<<<<<<<<<
 *                     if o.children != NULL \
 *                        and o.children[cind(i,j,k)] != NULL:
 */
      for (__pyx_t_4 = 0; __pyx_t_4 < 2; __pyx_t_4+=1) {
        __pyx_v_k = __pyx_t_4;

        /* "yt/geometry/particle_oct_container.pyx":331
 *             for j in range(2):
 *                 for k in range(2):
 *                     if o.children != NULL \             # <<<<<<<<<<<<<<
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit(o.children[cind(i,j,k)], counts, level + 1)
 */
        __pyx_t_6 = ((__pyx_v_o->children != NULL) != 0);
        if (__pyx_t_6) {
        } else {
          __pyx_t_5 = __pyx_t_6;
          goto __pyx_L10_bool_binop_done;
        }

        /* "yt/geometry/particle_oct_container.pyx":332
 *                 for k in range(2):
 *                     if o.children != NULL \
 *                        and o.children[cind(i,j,k)] != NULL:             # <<<<<<<<<<<<<<
 *                         self.visit(o.children[cind(i,j,k)], counts, level + 1)
 *         return
 */
        __pyx_t_6 = (((__pyx_v_o->children[__pyx_f_2yt_8geometry_12oct_visitors_cind(__pyx_v_i, __pyx_v_j, __pyx_v_k)]) != NULL) != 0);
        __pyx_t_5 = __pyx_t_6;
        __pyx_L10_bool_binop_done:;

        /* "yt/geometry/particle_oct_container.pyx":331
 *             for j in range(2):
 *                 for k in range(2):
 *                     if o.children != NULL \             # <<<<<<<<<<<<<<
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit(o.children[cind(i,j,k)], counts, level + 1)
 */
        if (__pyx_t_5) {

          /* "yt/geometry/particle_oct_container.pyx":333
 *                     if o.children != NULL \
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit(o.children[cind(i,j,k)], counts, level + 1)             # <<<<<<<<<<<<<<
 *         return
 * 
 */
          __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_v_level, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 333, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_9.__pyx_n = 1;
          __pyx_t_9.level = __pyx_t_7;
          __pyx_t_8 = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_vtab)->visit(__pyx_v_self, (__pyx_v_o->children[__pyx_f_2yt_8geometry_12oct_visitors_cind(__pyx_v_i, __pyx_v_j, __pyx_v_k)]), __pyx_v_counts, &__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 333, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "yt/geometry/particle_oct_container.pyx":331
 *             for j in range(2):
 *                 for k in range(2):
 *                     if o.children != NULL \             # <<<<<<<<<<<<<<
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit(o.children[cind(i,j,k)], counts, level + 1)
 */
        }
      }
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":334
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit(o.children[cind(i,j,k)], counts, level + 1)
 *         return             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":325
 *         return level_counts
 * 
 *     cdef visit(self, Oct *o, np.int64_t *counts, level = 0):             # <<<<<<<<<<<<<<
 *         cdef int i, j, k
 *         counts[level] += 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleOctreeContainer.visit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":339
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef Oct *get_from_index(self, np.uint64_t mi, np.uint8_t order = ORDER_MAX,             # <<<<<<<<<<<<<<
 *                              int max_level = 99):
 *         cdef Oct *cur
 */

static struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_get_from_index(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi, struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_get_from_index *__pyx_optional_args) {
  __pyx_t_5numpy_uint8_t __pyx_v_order = __pyx_k__2;
  int __pyx_v_max_level = ((int)99);
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_cur;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_next;
  int __pyx_v_i;
  __pyx_t_5numpy_int64_t __pyx_v_level;
  int __pyx_v_ind32[3];
  __pyx_t_5numpy_uint64_t __pyx_v_ind[3];
  __pyx_t_5numpy_uint64_t __pyx_v_index;
  int __pyx_v_level_offset[3];
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __pyx_t_5numpy_uint8_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_from_index", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_order = __pyx_optional_args->order;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_max_level = __pyx_optional_args->max_level;
      }
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":343
 *         cdef Oct *cur
 *         cdef Oct *next
 *         cur = next = NULL             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.int64_t level = -1
 */
  __pyx_v_cur = NULL;
  __pyx_v_next = NULL;

  /* "yt/geometry/particle_oct_container.pyx":345
 *         cur = next = NULL
 *         cdef int i
 *         cdef np.int64_t level = -1             # <<<<<<<<<<<<<<
 *         cdef int ind32[3]
 *         cdef np.uint64_t ind[3]
 */
  __pyx_v_level = -1LL;

  /* "yt/geometry/particle_oct_container.pyx":351
 *         # Get level offset
 *         cdef int level_offset[3]
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             level_offset[i] = np.log2(self.nn[i])
 *             if (1 << level_offset[i]) != self.nn[i]:
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":352
 *         cdef int level_offset[3]
 *         for i in range(3):
 *             level_offset[i] = np.log2(self.nn[i])             # <<<<<<<<<<<<<<
 *             if (1 << level_offset[i]) != self.nn[i]:
 *                 raise Exception("Octree does not have roots along dimension {} in a power of 2 ".format(i))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_log2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_self->__pyx_base.nn[__pyx_v_i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 352, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (__pyx_v_level_offset[__pyx_v_i]) = __pyx_t_6;

    /* "yt/geometry/particle_oct_container.pyx":353
 *         for i in range(3):
 *             level_offset[i] = np.log2(self.nn[i])
 *             if (1 << level_offset[i]) != self.nn[i]:             # <<<<<<<<<<<<<<
 *                 raise Exception("Octree does not have roots along dimension {} in a power of 2 ".format(i))
 *         for i in range(2,3):
 */
    __pyx_t_7 = (((1 << (__pyx_v_level_offset[__pyx_v_i])) != (__pyx_v_self->__pyx_base.nn[__pyx_v_i])) != 0);
    if (unlikely(__pyx_t_7)) {

      /* "yt/geometry/particle_oct_container.pyx":354
 *             level_offset[i] = np.log2(self.nn[i])
 *             if (1 << level_offset[i]) != self.nn[i]:
 *                 raise Exception("Octree does not have roots along dimension {} in a power of 2 ".format(i))             # <<<<<<<<<<<<<<
 *         for i in range(2,3):
 *             if level_offset[i] != level_offset[0]:
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Octree_does_not_have_roots_along, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 354, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 354, __pyx_L1_error)

      /* "yt/geometry/particle_oct_container.pyx":353
 *         for i in range(3):
 *             level_offset[i] = np.log2(self.nn[i])
 *             if (1 << level_offset[i]) != self.nn[i]:             # <<<<<<<<<<<<<<
 *                 raise Exception("Octree does not have roots along dimension {} in a power of 2 ".format(i))
 *         for i in range(2,3):
 */
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":355
 *             if (1 << level_offset[i]) != self.nn[i]:
 *                 raise Exception("Octree does not have roots along dimension {} in a power of 2 ".format(i))
 *         for i in range(2,3):             # <<<<<<<<<<<<<<
 *             if level_offset[i] != level_offset[0]:
 *                 raise Exception("Octree must have the same number of roots in each dimension for this.")
 */
  for (__pyx_t_1 = 2; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":356
 *                 raise Exception("Octree does not have roots along dimension {} in a power of 2 ".format(i))
 *         for i in range(2,3):
 *             if level_offset[i] != level_offset[0]:             # <<<<<<<<<<<<<<
 *                 raise Exception("Octree must have the same number of roots in each dimension for this.")
 *         # Get root for index
 */
    __pyx_t_7 = (((__pyx_v_level_offset[__pyx_v_i]) != (__pyx_v_level_offset[0])) != 0);
    if (unlikely(__pyx_t_7)) {

      /* "yt/geometry/particle_oct_container.pyx":357
 *         for i in range(2,3):
 *             if level_offset[i] != level_offset[0]:
 *                 raise Exception("Octree must have the same number of roots in each dimension for this.")             # <<<<<<<<<<<<<<
 *         # Get root for index
 *         index = (mi >> ((order - level_offset[0])*3))
 */
      __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 357, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 357, __pyx_L1_error)

      /* "yt/geometry/particle_oct_container.pyx":356
 *                 raise Exception("Octree does not have roots along dimension {} in a power of 2 ".format(i))
 *         for i in range(2,3):
 *             if level_offset[i] != level_offset[0]:             # <<<<<<<<<<<<<<
 *                 raise Exception("Octree must have the same number of roots in each dimension for this.")
 *         # Get root for index
 */
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":359
 *                 raise Exception("Octree must have the same number of roots in each dimension for this.")
 *         # Get root for index
 *         index = (mi >> ((order - level_offset[0])*3))             # <<<<<<<<<<<<<<
 *         decode_morton_64bit(index, ind)
 *         for i in range(3):
 */
  __pyx_v_index = (__pyx_v_mi >> ((__pyx_v_order - (__pyx_v_level_offset[0])) * 3));

  /* "yt/geometry/particle_oct_container.pyx":360
 *         # Get root for index
 *         index = (mi >> ((order - level_offset[0])*3))
 *         decode_morton_64bit(index, ind)             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             ind32[i] = ind[i]
 */
  __pyx_f_2yt_9utilities_3lib_14geometry_utils_decode_morton_64bit(__pyx_v_index, __pyx_v_ind);

  /* "yt/geometry/particle_oct_container.pyx":361
 *         index = (mi >> ((order - level_offset[0])*3))
 *         decode_morton_64bit(index, ind)
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             ind32[i] = ind[i]
 *         self.get_root(ind32, &next)
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":362
 *         decode_morton_64bit(index, ind)
 *         for i in range(3):
 *             ind32[i] = ind[i]             # <<<<<<<<<<<<<<
 *         self.get_root(ind32, &next)
 *         # We want to stop recursing when there's nowhere else to go
 */
    (__pyx_v_ind32[__pyx_v_i]) = (__pyx_v_ind[__pyx_v_i]);
  }

  /* "yt/geometry/particle_oct_container.pyx":363
 *         for i in range(3):
 *             ind32[i] = ind[i]
 *         self.get_root(ind32, &next)             # <<<<<<<<<<<<<<
 *         # We want to stop recursing when there's nowhere else to go
 *         level = level_offset[0]
 */
  (void)(((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.get_root(((struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *)__pyx_v_self), __pyx_v_ind32, (&__pyx_v_next)));

  /* "yt/geometry/particle_oct_container.pyx":365
 *         self.get_root(ind32, &next)
 *         # We want to stop recursing when there's nowhere else to go
 *         level = level_offset[0]             # <<<<<<<<<<<<<<
 *         max_level = min(max_level, order)
 *         while next != NULL and level <= max_level:
 */
  __pyx_v_level = (__pyx_v_level_offset[0]);

  /* "yt/geometry/particle_oct_container.pyx":366
 *         # We want to stop recursing when there's nowhere else to go
 *         level = level_offset[0]
 *         max_level = min(max_level, order)             # <<<<<<<<<<<<<<
 *         while next != NULL and level <= max_level:
 *             level += 1
 */
  __pyx_t_8 = __pyx_v_order;
  __pyx_t_1 = __pyx_v_max_level;
  if (((__pyx_t_8 < __pyx_t_1) != 0)) {
    __pyx_t_6 = __pyx_t_8;
  } else {
    __pyx_t_6 = __pyx_t_1;
  }
  __pyx_v_max_level = __pyx_t_6;

  /* "yt/geometry/particle_oct_container.pyx":367
 *         level = level_offset[0]
 *         max_level = min(max_level, order)
 *         while next != NULL and level <= max_level:             # <<<<<<<<<<<<<<
 *             level += 1
 *             for i in range(3):
 */
  while (1) {
    __pyx_t_9 = ((__pyx_v_next != NULL) != 0);
    if (__pyx_t_9) {
    } else {
      __pyx_t_7 = __pyx_t_9;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_9 = ((__pyx_v_level <= __pyx_v_max_level) != 0);
    __pyx_t_7 = __pyx_t_9;
    __pyx_L13_bool_binop_done:;
    if (!__pyx_t_7) break;

    /* "yt/geometry/particle_oct_container.pyx":368
 *         max_level = min(max_level, order)
 *         while next != NULL and level <= max_level:
 *             level += 1             # <<<<<<<<<<<<<<
 *             for i in range(3):
 *                 ind[i] = (mi >> ((order - level)*3 + (2 - i))) & 1
 */
    __pyx_v_level = (__pyx_v_level + 1);

    /* "yt/geometry/particle_oct_container.pyx":369
 *         while next != NULL and level <= max_level:
 *             level += 1
 *             for i in range(3):             # <<<<<<<<<<<<<<
 *                 ind[i] = (mi >> ((order - level)*3 + (2 - i))) & 1
 *             cur = next
 */
    for (__pyx_t_6 = 0; __pyx_t_6 < 3; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "yt/geometry/particle_oct_container.pyx":370
 *             level += 1
 *             for i in range(3):
 *                 ind[i] = (mi >> ((order - level)*3 + (2 - i))) & 1             # <<<<<<<<<<<<<<
 *             cur = next
 *             if cur.children != NULL:
 */
      (__pyx_v_ind[__pyx_v_i]) = ((__pyx_v_mi >> (((__pyx_v_order - __pyx_v_level) * 3) + (2 - __pyx_v_i))) & 1);
    }

    /* "yt/geometry/particle_oct_container.pyx":371
 *             for i in range(3):
 *                 ind[i] = (mi >> ((order - level)*3 + (2 - i))) & 1
 *             cur = next             # <<<<<<<<<<<<<<
 *             if cur.children != NULL:
 *                 next = cur.children[cind(ind[0],ind[1],ind[2])]
 */
    __pyx_v_cur = __pyx_v_next;

    /* "yt/geometry/particle_oct_container.pyx":372
 *                 ind[i] = (mi >> ((order - level)*3 + (2 - i))) & 1
 *             cur = next
 *             if cur.children != NULL:             # <<<<<<<<<<<<<<
 *                 next = cur.children[cind(ind[0],ind[1],ind[2])]
 *             else:
 */
    __pyx_t_7 = ((__pyx_v_cur->children != NULL) != 0);
    if (__pyx_t_7) {

      /* "yt/geometry/particle_oct_container.pyx":373
 *             cur = next
 *             if cur.children != NULL:
 *                 next = cur.children[cind(ind[0],ind[1],ind[2])]             # <<<<<<<<<<<<<<
 *             else:
 *                 next = NULL
 */
      __pyx_v_next = (__pyx_v_cur->children[__pyx_f_2yt_8geometry_12oct_visitors_cind((__pyx_v_ind[0]), (__pyx_v_ind[1]), (__pyx_v_ind[2]))]);

      /* "yt/geometry/particle_oct_container.pyx":372
 *                 ind[i] = (mi >> ((order - level)*3 + (2 - i))) & 1
 *             cur = next
 *             if cur.children != NULL:             # <<<<<<<<<<<<<<
 *                 next = cur.children[cind(ind[0],ind[1],ind[2])]
 *             else:
 */
      goto __pyx_L17;
    }

    /* "yt/geometry/particle_oct_container.pyx":375
 *                 next = cur.children[cind(ind[0],ind[1],ind[2])]
 *             else:
 *                 next = NULL             # <<<<<<<<<<<<<<
 *         return cur
 * 
 */
    /*else*/ {
      __pyx_v_next = NULL;
    }
    __pyx_L17:;
  }

  /* "yt/geometry/particle_oct_container.pyx":376
 *             else:
 *                 next = NULL
 *         return cur             # <<<<<<<<<<<<<<
 * 
 *     def apply_domain(self, int domain_id, BoolArrayCollection mask,
 */
  __pyx_r = __pyx_v_cur;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":339
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef Oct *get_from_index(self, np.uint64_t mi, np.uint8_t order = ORDER_MAX,             # <<<<<<<<<<<<<<
 *                              int max_level = 99):
 *         cdef Oct *cur
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("yt.geometry.particle_oct_container.ParticleOctreeContainer.get_from_index", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":378
 *         return cur
 * 
 *     def apply_domain(self, int domain_id, BoolArrayCollection mask,             # <<<<<<<<<<<<<<
 *                      int masklevel):
 *         cdef SelectorObject selector = AlwaysSelector(None)
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_18apply_domain(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_18apply_domain(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_domain_id;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mask = 0;
  int __pyx_v_masklevel;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("apply_domain (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_domain_id,&__pyx_n_s_mask,&__pyx_n_s_masklevel,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_domain_id)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("apply_domain", 1, 3, 3, 1); __PYX_ERR(0, 378, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_masklevel)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("apply_domain", 1, 3, 3, 2); __PYX_ERR(0, 378, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "apply_domain") < 0)) __PYX_ERR(0, 378, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_domain_id = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_domain_id == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 378, __pyx_L3_error)
    __pyx_v_mask = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)values[1]);
    __pyx_v_masklevel = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_masklevel == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 379, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("apply_domain", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 378, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleOctreeContainer.apply_domain", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mask), __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection, 1, "mask", 0))) __PYX_ERR(0, 378, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_17apply_domain(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self), __pyx_v_domain_id, __pyx_v_mask, __pyx_v_masklevel);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_17apply_domain(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, int __pyx_v_domain_id, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mask, int __pyx_v_masklevel) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_selector = 0;
  PyObject *__pyx_v_ind = NULL;
  __pyx_t_5numpy_int64_t __pyx_v_i;
  CYTHON_UNUSED PyObject *__pyx_v_ind_out = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __pyx_t_5numpy_int64_t __pyx_t_6;
  __pyx_t_5numpy_int64_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("apply_domain", 0);

  /* "yt/geometry/particle_oct_container.pyx":380
 *     def apply_domain(self, int domain_id, BoolArrayCollection mask,
 *                      int masklevel):
 *         cdef SelectorObject selector = AlwaysSelector(None)             # <<<<<<<<<<<<<<
 *         ind = self.domain_ind(selector, mask = mask, masklevel = masklevel)
 *         for i in range(self.nocts):
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_AlwaysSelector), __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_selector = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":381
 *                      int masklevel):
 *         cdef SelectorObject selector = AlwaysSelector(None)
 *         ind = self.domain_ind(selector, mask = mask, masklevel = masklevel)             # <<<<<<<<<<<<<<
 *         for i in range(self.nocts):
 *             if ind[i] < 0: continue
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_domain_ind); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_selector));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_selector));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_selector));
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_mask, ((PyObject *)__pyx_v_mask)) < 0) __PYX_ERR(0, 381, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_masklevel); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_masklevel, __pyx_t_4) < 0) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_ind = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "yt/geometry/particle_oct_container.pyx":382
 *         cdef SelectorObject selector = AlwaysSelector(None)
 *         ind = self.domain_ind(selector, mask = mask, masklevel = masklevel)
 *         for i in range(self.nocts):             # <<<<<<<<<<<<<<
 *             if ind[i] < 0: continue
 *             self.oct_list[i].domain = domain_id
 */
  __pyx_t_5 = __pyx_v_self->__pyx_base.nocts;
  __pyx_t_6 = __pyx_t_5;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "yt/geometry/particle_oct_container.pyx":383
 *         ind = self.domain_ind(selector, mask = mask, masklevel = masklevel)
 *         for i in range(self.nocts):
 *             if ind[i] < 0: continue             # <<<<<<<<<<<<<<
 *             self.oct_list[i].domain = domain_id
 *         ind_out = super(ParticleOctreeContainer,self).domain_ind(selector, domain_id = domain_id)
 */
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_ind, __pyx_v_i, __pyx_t_5numpy_int64_t, 1, __Pyx_PyInt_From_npy_int64, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 383, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_4, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 383, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 383, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_8) {
      goto __pyx_L3_continue;
    }

    /* "yt/geometry/particle_oct_container.pyx":384
 *         for i in range(self.nocts):
 *             if ind[i] < 0: continue
 *             self.oct_list[i].domain = domain_id             # <<<<<<<<<<<<<<
 *         ind_out = super(ParticleOctreeContainer,self).domain_ind(selector, domain_id = domain_id)
 * 
 */
    (__pyx_v_self->oct_list[__pyx_v_i])->domain = __pyx_v_domain_id;
    __pyx_L3_continue:;
  }

  /* "yt/geometry/particle_oct_container.pyx":385
 *             if ind[i] < 0: continue
 *             self.oct_list[i].domain = domain_id
 *         ind_out = super(ParticleOctreeContainer,self).domain_ind(selector, domain_id = domain_id)             # <<<<<<<<<<<<<<
 * 
 *     def domain_ind(self, selector, int domain_id = -1,
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer));
  __Pyx_GIVEREF(((PyObject *)__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_domain_ind); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(((PyObject *)__pyx_v_selector));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_selector));
  PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_selector));
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_domain_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_domain_id, __pyx_t_1) < 0) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_ind_out = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":378
 *         return cur
 * 
 *     def apply_domain(self, int domain_id, BoolArrayCollection mask,             # <<<<<<<<<<<<<<
 *                      int masklevel):
 *         cdef SelectorObject selector = AlwaysSelector(None)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleOctreeContainer.apply_domain", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_selector);
  __Pyx_XDECREF(__pyx_v_ind);
  __Pyx_XDECREF(__pyx_v_ind_out);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":387
 *         ind_out = super(ParticleOctreeContainer,self).domain_ind(selector, domain_id = domain_id)
 * 
 *     def domain_ind(self, selector, int domain_id = -1,             # <<<<<<<<<<<<<<
 *                    BoolArrayCollection mask = None, int masklevel = 99):
 *         if mask is None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_20domain_ind(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_20domain_ind(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_selector = 0;
  int __pyx_v_domain_id;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mask = 0;
  int __pyx_v_masklevel;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("domain_ind (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_selector,&__pyx_n_s_domain_id,&__pyx_n_s_mask,&__pyx_n_s_masklevel,0};
    PyObject* values[4] = {0,0,0,0};

    /* "yt/geometry/particle_oct_container.pyx":388
 * 
 *     def domain_ind(self, selector, int domain_id = -1,
 *                    BoolArrayCollection mask = None, int masklevel = 99):             # <<<<<<<<<<<<<<
 *         if mask is None:
 *             return super(ParticleOctreeContainer,self).domain_ind(selector, domain_id = domain_id)
 */
    values[2] = (PyObject *)((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_selector)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_domain_id);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_masklevel);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "domain_ind") < 0)) __PYX_ERR(0, 387, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_selector = values[0];
    if (values[1]) {
      __pyx_v_domain_id = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_domain_id == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 387, __pyx_L3_error)
    } else {
      __pyx_v_domain_id = ((int)-1);
    }
    __pyx_v_mask = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)values[2]);
    if (values[3]) {
      __pyx_v_masklevel = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_masklevel == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 388, __pyx_L3_error)
    } else {
      __pyx_v_masklevel = ((int)99);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("domain_ind", 0, 1, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 387, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleOctreeContainer.domain_ind", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mask), __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection, 1, "mask", 0))) __PYX_ERR(0, 388, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_19domain_ind(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self), __pyx_v_selector, __pyx_v_domain_id, __pyx_v_mask, __pyx_v_masklevel);

  /* "yt/geometry/particle_oct_container.pyx":387
 *         ind_out = super(ParticleOctreeContainer,self).domain_ind(selector, domain_id = domain_id)
 * 
 *     def domain_ind(self, selector, int domain_id = -1,             # <<<<<<<<<<<<<<
 *                    BoolArrayCollection mask = None, int masklevel = 99):
 *         if mask is None:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_19domain_ind(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, PyObject *__pyx_v_selector, int __pyx_v_domain_id, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mask, int __pyx_v_masklevel) {
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array *__pyx_v_ewah_slct;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_iterator *__pyx_v_iter_set;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_iterator *__pyx_v_iter_end;
  PyArrayObject *__pyx_v_oct_mask = 0;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o;
  int __pyx_v_coct;
  int __pyx_v_cmi;
  size_t __pyx_v_mi;
  PyArrayObject *__pyx_v_ind = 0;
  struct __pyx_obj_2yt_8geometry_12oct_visitors_MaskedIndexOcts *__pyx_v_visitor = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_ind;
  __Pyx_Buffer __pyx_pybuffer_ind;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_oct_mask;
  __Pyx_Buffer __pyx_pybuffer_oct_mask;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyArrayObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_t_13;
  struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_get_from_index __pyx_t_14;
  PyArrayObject *__pyx_t_15 = NULL;
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("domain_ind", 0);
  __pyx_pybuffer_oct_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_oct_mask.refcount = 0;
  __pyx_pybuffernd_oct_mask.data = NULL;
  __pyx_pybuffernd_oct_mask.rcbuffer = &__pyx_pybuffer_oct_mask;
  __pyx_pybuffer_ind.pybuffer.buf = NULL;
  __pyx_pybuffer_ind.refcount = 0;
  __pyx_pybuffernd_ind.data = NULL;
  __pyx_pybuffernd_ind.rcbuffer = &__pyx_pybuffer_ind;

  /* "yt/geometry/particle_oct_container.pyx":389
 *     def domain_ind(self, selector, int domain_id = -1,
 *                    BoolArrayCollection mask = None, int masklevel = 99):
 *         if mask is None:             # <<<<<<<<<<<<<<
 *             return super(ParticleOctreeContainer,self).domain_ind(selector, domain_id = domain_id)
 *         # Create mask for octs that are touched by the mask
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_mask) == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "yt/geometry/particle_oct_container.pyx":390
 *                    BoolArrayCollection mask = None, int masklevel = 99):
 *         if mask is None:
 *             return super(ParticleOctreeContainer,self).domain_ind(selector, domain_id = domain_id)             # <<<<<<<<<<<<<<
 *         # Create mask for octs that are touched by the mask
 *         cdef ewah_bool_array *ewah_slct = <ewah_bool_array *> mask.ewah_keys
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer));
    __Pyx_GIVEREF(((PyObject *)__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer));
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_self));
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_domain_ind); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_selector);
    __Pyx_GIVEREF(__pyx_v_selector);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_selector);
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_domain_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_domain_id, __pyx_t_6) < 0) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 390, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "yt/geometry/particle_oct_container.pyx":389
 *     def domain_ind(self, selector, int domain_id = -1,
 *                    BoolArrayCollection mask = None, int masklevel = 99):
 *         if mask is None:             # <<<<<<<<<<<<<<
 *             return super(ParticleOctreeContainer,self).domain_ind(selector, domain_id = domain_id)
 *         # Create mask for octs that are touched by the mask
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":392
 *             return super(ParticleOctreeContainer,self).domain_ind(selector, domain_id = domain_id)
 *         # Create mask for octs that are touched by the mask
 *         cdef ewah_bool_array *ewah_slct = <ewah_bool_array *> mask.ewah_keys             # <<<<<<<<<<<<<<
 *         cdef ewah_bool_iterator *iter_set = new ewah_bool_iterator(ewah_slct[0].begin())
 *         cdef ewah_bool_iterator *iter_end = new ewah_bool_iterator(ewah_slct[0].end())
 */
  __pyx_v_ewah_slct = ((__pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array *)__pyx_v_mask->ewah_keys);

  /* "yt/geometry/particle_oct_container.pyx":393
 *         # Create mask for octs that are touched by the mask
 *         cdef ewah_bool_array *ewah_slct = <ewah_bool_array *> mask.ewah_keys
 *         cdef ewah_bool_iterator *iter_set = new ewah_bool_iterator(ewah_slct[0].begin())             # <<<<<<<<<<<<<<
 *         cdef ewah_bool_iterator *iter_end = new ewah_bool_iterator(ewah_slct[0].end())
 *         cdef np.ndarray[np.uint8_t, ndim=1] oct_mask
 */
  __pyx_v_iter_set = new __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_iterator((__pyx_v_ewah_slct[0]).begin());

  /* "yt/geometry/particle_oct_container.pyx":394
 *         cdef ewah_bool_array *ewah_slct = <ewah_bool_array *> mask.ewah_keys
 *         cdef ewah_bool_iterator *iter_set = new ewah_bool_iterator(ewah_slct[0].begin())
 *         cdef ewah_bool_iterator *iter_end = new ewah_bool_iterator(ewah_slct[0].end())             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint8_t, ndim=1] oct_mask
 *         oct_mask = np.zeros(self.nocts, 'uint8')
 */
  __pyx_v_iter_end = new __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_iterator((__pyx_v_ewah_slct[0]).end());

  /* "yt/geometry/particle_oct_container.pyx":396
 *         cdef ewah_bool_iterator *iter_end = new ewah_bool_iterator(ewah_slct[0].end())
 *         cdef np.ndarray[np.uint8_t, ndim=1] oct_mask
 *         oct_mask = np.zeros(self.nocts, 'uint8')             # <<<<<<<<<<<<<<
 *         cdef Oct *o
 *         cdef int coct, cmi
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_zeros); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->__pyx_base.nocts); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_5, __pyx_n_s_uint8};
    __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_5, __pyx_n_s_uint8};
    __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  {
    __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_5);
    __Pyx_INCREF(__pyx_n_s_uint8);
    __Pyx_GIVEREF(__pyx_n_s_uint8);
    PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_n_s_uint8);
    __pyx_t_5 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_8, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 396, __pyx_L1_error)
  __pyx_t_9 = ((PyArrayObject *)__pyx_t_6);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_oct_mask.rcbuffer->pybuffer);
    __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_oct_mask.rcbuffer->pybuffer, (PyObject*)__pyx_t_9, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_7 < 0)) {
      PyErr_Fetch(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_oct_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_oct_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_12);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __pyx_t_10 = __pyx_t_11 = __pyx_t_12 = 0;
    }
    __pyx_pybuffernd_oct_mask.diminfo[0].strides = __pyx_pybuffernd_oct_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_oct_mask.diminfo[0].shape = __pyx_pybuffernd_oct_mask.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 396, __pyx_L1_error)
  }
  __pyx_t_9 = 0;
  __pyx_v_oct_mask = ((PyArrayObject *)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "yt/geometry/particle_oct_container.pyx":399
 *         cdef Oct *o
 *         cdef int coct, cmi
 *         coct = cmi = 0             # <<<<<<<<<<<<<<
 *         while iter_set[0] != iter_end[0]:
 *             mi = dereference(iter_set[0])
 */
  __pyx_v_coct = 0;
  __pyx_v_cmi = 0;

  /* "yt/geometry/particle_oct_container.pyx":400
 *         cdef int coct, cmi
 *         coct = cmi = 0
 *         while iter_set[0] != iter_end[0]:             # <<<<<<<<<<<<<<
 *             mi = dereference(iter_set[0])
 *             o = self.get_from_index(mi, order = masklevel)
 */
  while (1) {
    __pyx_t_2 = (((__pyx_v_iter_set[0]) != (__pyx_v_iter_end[0])) != 0);
    if (!__pyx_t_2) break;

    /* "yt/geometry/particle_oct_container.pyx":401
 *         coct = cmi = 0
 *         while iter_set[0] != iter_end[0]:
 *             mi = dereference(iter_set[0])             # <<<<<<<<<<<<<<
 *             o = self.get_from_index(mi, order = masklevel)
 *             if o != NULL:
 */
    __pyx_v_mi = (*(__pyx_v_iter_set[0]));

    /* "yt/geometry/particle_oct_container.pyx":402
 *         while iter_set[0] != iter_end[0]:
 *             mi = dereference(iter_set[0])
 *             o = self.get_from_index(mi, order = masklevel)             # <<<<<<<<<<<<<<
 *             if o != NULL:
 *                 _mask_children(oct_mask, o)
 */
    __pyx_t_14.__pyx_n = 1;
    __pyx_t_14.order = __pyx_v_masklevel;
    __pyx_t_13 = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_vtab)->get_from_index(__pyx_v_self, __pyx_v_mi, &__pyx_t_14); 
    __pyx_v_o = __pyx_t_13;

    /* "yt/geometry/particle_oct_container.pyx":403
 *             mi = dereference(iter_set[0])
 *             o = self.get_from_index(mi, order = masklevel)
 *             if o != NULL:             # <<<<<<<<<<<<<<
 *                 _mask_children(oct_mask, o)
 *                 coct += 1
 */
    __pyx_t_2 = ((__pyx_v_o != NULL) != 0);
    if (__pyx_t_2) {

      /* "yt/geometry/particle_oct_container.pyx":404
 *             o = self.get_from_index(mi, order = masklevel)
 *             if o != NULL:
 *                 _mask_children(oct_mask, o)             # <<<<<<<<<<<<<<
 *                 coct += 1
 *             cmi += 1;
 */
      __pyx_f_2yt_8geometry_22particle_oct_container__mask_children(((PyArrayObject *)__pyx_v_oct_mask), __pyx_v_o);

      /* "yt/geometry/particle_oct_container.pyx":405
 *             if o != NULL:
 *                 _mask_children(oct_mask, o)
 *                 coct += 1             # <<<<<<<<<<<<<<
 *             cmi += 1;
 *             preincrement(iter_set[0])
 */
      __pyx_v_coct = (__pyx_v_coct + 1);

      /* "yt/geometry/particle_oct_container.pyx":403
 *             mi = dereference(iter_set[0])
 *             o = self.get_from_index(mi, order = masklevel)
 *             if o != NULL:             # <<<<<<<<<<<<<<
 *                 _mask_children(oct_mask, o)
 *                 coct += 1
 */
    }

    /* "yt/geometry/particle_oct_container.pyx":406
 *                 _mask_children(oct_mask, o)
 *                 coct += 1
 *             cmi += 1;             # <<<<<<<<<<<<<<
 *             preincrement(iter_set[0])
 *         # Get domain ind
 */
    __pyx_v_cmi = (__pyx_v_cmi + 1);

    /* "yt/geometry/particle_oct_container.pyx":407
 *                 coct += 1
 *             cmi += 1;
 *             preincrement(iter_set[0])             # <<<<<<<<<<<<<<
 *         # Get domain ind
 *         cdef np.ndarray[np.int64_t, ndim=1] ind
 */
    (void)((++(__pyx_v_iter_set[0])));
  }

  /* "yt/geometry/particle_oct_container.pyx":410
 *         # Get domain ind
 *         cdef np.ndarray[np.int64_t, ndim=1] ind
 *         ind = np.zeros(self.nocts, 'int64') - 1             # <<<<<<<<<<<<<<
 *         cdef oct_visitors.MaskedIndexOcts visitor
 *         visitor = oct_visitors.MaskedIndexOcts(self, domain_id)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->__pyx_base.nocts); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_8)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_4, __pyx_n_s_int64};
    __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_4, __pyx_n_s_int64};
    __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_3 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_7, __pyx_t_4);
    __Pyx_INCREF(__pyx_n_s_int64);
    __Pyx_GIVEREF(__pyx_n_s_int64);
    PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_7, __pyx_n_s_int64);
    __pyx_t_4 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_3, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyInt_SubtractObjC(__pyx_t_6, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (!(likely(((__pyx_t_8) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_8, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 410, __pyx_L1_error)
  __pyx_t_15 = ((PyArrayObject *)__pyx_t_8);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_ind.rcbuffer->pybuffer);
    __pyx_t_7 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_ind.rcbuffer->pybuffer, (PyObject*)__pyx_t_15, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_7 < 0)) {
      PyErr_Fetch(&__pyx_t_12, &__pyx_t_11, &__pyx_t_10);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_ind.rcbuffer->pybuffer, (PyObject*)__pyx_v_ind, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_12); Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_12, __pyx_t_11, __pyx_t_10);
      }
      __pyx_t_12 = __pyx_t_11 = __pyx_t_10 = 0;
    }
    __pyx_pybuffernd_ind.diminfo[0].strides = __pyx_pybuffernd_ind.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_ind.diminfo[0].shape = __pyx_pybuffernd_ind.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 410, __pyx_L1_error)
  }
  __pyx_t_15 = 0;
  __pyx_v_ind = ((PyArrayObject *)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "yt/geometry/particle_oct_container.pyx":412
 *         ind = np.zeros(self.nocts, 'int64') - 1
 *         cdef oct_visitors.MaskedIndexOcts visitor
 *         visitor = oct_visitors.MaskedIndexOcts(self, domain_id)             # <<<<<<<<<<<<<<
 *         visitor.oct_index = ind
 *         visitor.oct_mask = oct_mask
 */
  __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_domain_id); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_8);
  __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_2yt_8geometry_12oct_visitors_MaskedIndexOcts), __pyx_t_6, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_visitor = ((struct __pyx_obj_2yt_8geometry_12oct_visitors_MaskedIndexOcts *)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "yt/geometry/particle_oct_container.pyx":413
 *         cdef oct_visitors.MaskedIndexOcts visitor
 *         visitor = oct_visitors.MaskedIndexOcts(self, domain_id)
 *         visitor.oct_index = ind             # <<<<<<<<<<<<<<
 *         visitor.oct_mask = oct_mask
 *         self.visit_all_octs(selector, visitor)
 */
  __pyx_t_16 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(((PyObject *)__pyx_v_ind), PyBUF_WRITABLE); if (unlikely(!__pyx_t_16.memview)) __PYX_ERR(0, 413, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_visitor->oct_index, 0);
  __pyx_v_visitor->oct_index = __pyx_t_16;
  __pyx_t_16.memview = NULL;
  __pyx_t_16.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":414
 *         visitor = oct_visitors.MaskedIndexOcts(self, domain_id)
 *         visitor.oct_index = ind
 *         visitor.oct_mask = oct_mask             # <<<<<<<<<<<<<<
 *         self.visit_all_octs(selector, visitor)
 *         return ind
 */
  __pyx_t_17 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint8_t(((PyObject *)__pyx_v_oct_mask), PyBUF_WRITABLE); if (unlikely(!__pyx_t_17.memview)) __PYX_ERR(0, 414, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_visitor->oct_mask, 0);
  __pyx_v_visitor->oct_mask = __pyx_t_17;
  __pyx_t_17.memview = NULL;
  __pyx_t_17.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":415
 *         visitor.oct_index = ind
 *         visitor.oct_mask = oct_mask
 *         self.visit_all_octs(selector, visitor)             # <<<<<<<<<<<<<<
 *         return ind
 * 
 */
  if (!(likely(((__pyx_v_selector) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_selector, __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject))))) __PYX_ERR(0, 415, __pyx_L1_error)
  ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.visit_all_octs(((struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *)__pyx_v_self), ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_selector), ((struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *)__pyx_v_visitor), NULL);

  /* "yt/geometry/particle_oct_container.pyx":416
 *         visitor.oct_mask = oct_mask
 *         self.visit_all_octs(selector, visitor)
 *         return ind             # <<<<<<<<<<<<<<
 * 
 * cdef void _mask_children(np.ndarray[np.uint8_t] mask, Oct *cur):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_ind));
  __pyx_r = ((PyObject *)__pyx_v_ind);
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":387
 *         ind_out = super(ParticleOctreeContainer,self).domain_ind(selector, domain_id = domain_id)
 * 
 *     def domain_ind(self, selector, int domain_id = -1,             # <<<<<<<<<<<<<<
 *                    BoolArrayCollection mask = None, int masklevel = 99):
 *         if mask is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_ind.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_oct_mask.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleOctreeContainer.domain_ind", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_ind.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_oct_mask.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_oct_mask);
  __Pyx_XDECREF((PyObject *)__pyx_v_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_visitor);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":94
 *     #The starting oct index of each domain
 *     cdef np.int64_t *dom_offsets
 *     cdef public int max_level             # <<<<<<<<<<<<<<
 *     #How many particles do we keep before refining
 *     cdef public int n_ref
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_9max_level_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_9max_level_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_9max_level___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_9max_level___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->max_level); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleOctreeContainer.max_level.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_9max_level_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_9max_level_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_9max_level_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_9max_level_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 94, __pyx_L1_error)
  __pyx_v_self->max_level = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleOctreeContainer.max_level.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":96
 *     cdef public int max_level
 *     #How many particles do we keep before refining
 *     cdef public int n_ref             # <<<<<<<<<<<<<<
 * 
 *     def allocate_root(self):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_5n_ref_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_5n_ref_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_5n_ref___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_5n_ref___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->n_ref); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleOctreeContainer.n_ref.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_5n_ref_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_5n_ref_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_5n_ref_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_5n_ref_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 96, __pyx_L1_error)
  __pyx_v_self->n_ref = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleOctreeContainer.n_ref.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dom_offsets,self.oct_list,self.root_mesh cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_22__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_22__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_21__reduce_cython__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_21__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.dom_offsets,self.oct_list,self.root_mesh cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dom_offsets,self.oct_list,self.root_mesh cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dom_offsets,self.oct_list,self.root_mesh cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleOctreeContainer.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.dom_offsets,self.oct_list,self.root_mesh cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dom_offsets,self.oct_list,self.root_mesh cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_24__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_24__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_23__setstate_cython__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_23__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.dom_offsets,self.oct_list,self.root_mesh cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dom_offsets,self.oct_list,self.root_mesh cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.dom_offsets,self.oct_list,self.root_mesh cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.dom_offsets,self.oct_list,self.root_mesh cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleOctreeContainer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":418
 *         return ind
 * 
 * cdef void _mask_children(np.ndarray[np.uint8_t] mask, Oct *cur):             # <<<<<<<<<<<<<<
 *     cdef int i, j, k
 *     if cur == NULL:
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container__mask_children(PyArrayObject *__pyx_v_mask, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_cur) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_mask_children", 0);
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 418, __pyx_L1_error)
  }
  __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/particle_oct_container.pyx":420
 * cdef void _mask_children(np.ndarray[np.uint8_t] mask, Oct *cur):
 *     cdef int i, j, k
 *     if cur == NULL:             # <<<<<<<<<<<<<<
 *         return
 *     mask[cur.domain_ind] = 1
 */
  __pyx_t_1 = ((__pyx_v_cur == NULL) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/particle_oct_container.pyx":421
 *     cdef int i, j, k
 *     if cur == NULL:
 *         return             # <<<<<<<<<<<<<<
 *     mask[cur.domain_ind] = 1
 *     if cur.children == NULL:
 */
    goto __pyx_L0;

    /* "yt/geometry/particle_oct_container.pyx":420
 * cdef void _mask_children(np.ndarray[np.uint8_t] mask, Oct *cur):
 *     cdef int i, j, k
 *     if cur == NULL:             # <<<<<<<<<<<<<<
 *         return
 *     mask[cur.domain_ind] = 1
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":422
 *     if cur == NULL:
 *         return
 *     mask[cur.domain_ind] = 1             # <<<<<<<<<<<<<<
 *     if cur.children == NULL:
 *         return
 */
  __pyx_t_2 = __pyx_v_cur->domain_ind;
  __pyx_t_3 = -1;
  if (__pyx_t_2 < 0) {
    __pyx_t_2 += __pyx_pybuffernd_mask.diminfo[0].shape;
    if (unlikely(__pyx_t_2 < 0)) __pyx_t_3 = 0;
  } else if (unlikely(__pyx_t_2 >= __pyx_pybuffernd_mask.diminfo[0].shape)) __pyx_t_3 = 0;
  if (unlikely(__pyx_t_3 != -1)) {
    __Pyx_RaiseBufferIndexError(__pyx_t_3);
    __PYX_ERR(0, 422, __pyx_L1_error)
  }
  *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_2, __pyx_pybuffernd_mask.diminfo[0].strides) = 1;

  /* "yt/geometry/particle_oct_container.pyx":423
 *         return
 *     mask[cur.domain_ind] = 1
 *     if cur.children == NULL:             # <<<<<<<<<<<<<<
 *         return
 *     for i in range(2):
 */
  __pyx_t_1 = ((__pyx_v_cur->children == NULL) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/particle_oct_container.pyx":424
 *     mask[cur.domain_ind] = 1
 *     if cur.children == NULL:
 *         return             # <<<<<<<<<<<<<<
 *     for i in range(2):
 *         for j in range(2):
 */
    goto __pyx_L0;

    /* "yt/geometry/particle_oct_container.pyx":423
 *         return
 *     mask[cur.domain_ind] = 1
 *     if cur.children == NULL:             # <<<<<<<<<<<<<<
 *         return
 *     for i in range(2):
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":425
 *     if cur.children == NULL:
 *         return
 *     for i in range(2):             # <<<<<<<<<<<<<<
 *         for j in range(2):
 *             for k in range(2):
 */
  for (__pyx_t_3 = 0; __pyx_t_3 < 2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":426
 *         return
 *     for i in range(2):
 *         for j in range(2):             # <<<<<<<<<<<<<<
 *             for k in range(2):
 *                 _mask_children(mask, cur.children[cind(i,j,k)])
 */
    for (__pyx_t_4 = 0; __pyx_t_4 < 2; __pyx_t_4+=1) {
      __pyx_v_j = __pyx_t_4;

      /* "yt/geometry/particle_oct_container.pyx":427
 *     for i in range(2):
 *         for j in range(2):
 *             for k in range(2):             # <<<<<<<<<<<<<<
 *                 _mask_children(mask, cur.children[cind(i,j,k)])
 * 
 */
      for (__pyx_t_5 = 0; __pyx_t_5 < 2; __pyx_t_5+=1) {
        __pyx_v_k = __pyx_t_5;

        /* "yt/geometry/particle_oct_container.pyx":428
 *         for j in range(2):
 *             for k in range(2):
 *                 _mask_children(mask, cur.children[cind(i,j,k)])             # <<<<<<<<<<<<<<
 * 
 * cdef np.uint64_t ONEBIT=1
 */
        __pyx_f_2yt_8geometry_22particle_oct_container__mask_children(__pyx_v_mask, (__pyx_v_cur->children[__pyx_f_2yt_8geometry_12oct_visitors_cind(__pyx_v_i, __pyx_v_j, __pyx_v_k)]));
      }
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":418
 *         return ind
 * 
 * cdef void _mask_children(np.ndarray[np.uint8_t] mask, Oct *cur):             # <<<<<<<<<<<<<<
 *     cdef int i, j, k
 *     if cur == NULL:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_WriteUnraisable("yt.geometry.particle_oct_container._mask_children", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":465
 *     cdef public int _used_mi2
 * 
 *     def __init__(self, left_edge, right_edge, periodicity, file_hash, nfiles,             # <<<<<<<<<<<<<<
 *                  index_order1, index_order2):
 *         # TODO: Set limit on maximum orders?
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_left_edge = 0;
  PyObject *__pyx_v_right_edge = 0;
  PyObject *__pyx_v_periodicity = 0;
  PyObject *__pyx_v_file_hash = 0;
  PyObject *__pyx_v_nfiles = 0;
  PyObject *__pyx_v_index_order1 = 0;
  PyObject *__pyx_v_index_order2 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left_edge,&__pyx_n_s_right_edge,&__pyx_n_s_periodicity,&__pyx_n_s_file_hash,&__pyx_n_s_nfiles,&__pyx_n_s_index_order1,&__pyx_n_s_index_order2,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left_edge)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right_edge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 7, 7, 1); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_periodicity)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 7, 7, 2); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_file_hash)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 7, 7, 3); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nfiles)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 7, 7, 4); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_index_order1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 7, 7, 5); __PYX_ERR(0, 465, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_index_order2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 7, 7, 6); __PYX_ERR(0, 465, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 465, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_left_edge = values[0];
    __pyx_v_right_edge = values[1];
    __pyx_v_periodicity = values[2];
    __pyx_v_file_hash = values[3];
    __pyx_v_nfiles = values[4];
    __pyx_v_index_order1 = values[5];
    __pyx_v_index_order2 = values[6];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 465, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap___init__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), __pyx_v_left_edge, __pyx_v_right_edge, __pyx_v_periodicity, __pyx_v_file_hash, __pyx_v_nfiles, __pyx_v_index_order1, __pyx_v_index_order2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap___init__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_left_edge, PyObject *__pyx_v_right_edge, PyObject *__pyx_v_periodicity, PyObject *__pyx_v_file_hash, PyObject *__pyx_v_nfiles, PyObject *__pyx_v_index_order1, PyObject *__pyx_v_index_order2) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  __pyx_t_5numpy_uint64_t __pyx_t_3;
  int __pyx_t_4;
  __pyx_t_5numpy_float64_t __pyx_t_5;
  __pyx_t_5numpy_uint8_t __pyx_t_6;
  __pyx_t_5numpy_int32_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/particle_oct_container.pyx":469
 *         # TODO: Set limit on maximum orders?
 *         cdef int i
 *         self._cached_octrees = {}             # <<<<<<<<<<<<<<
 *         self._last_selector = None
 *         self._last_return_values = None
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_cached_octrees);
  __Pyx_DECREF(__pyx_v_self->_cached_octrees);
  __pyx_v_self->_cached_octrees = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":470
 *         cdef int i
 *         self._cached_octrees = {}
 *         self._last_selector = None             # <<<<<<<<<<<<<<
 *         self._last_return_values = None
 *         self._last_octree_subset = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_last_selector);
  __Pyx_DECREF(__pyx_v_self->_last_selector);
  __pyx_v_self->_last_selector = Py_None;

  /* "yt/geometry/particle_oct_container.pyx":471
 *         self._cached_octrees = {}
 *         self._last_selector = None
 *         self._last_return_values = None             # <<<<<<<<<<<<<<
 *         self._last_octree_subset = None
 *         self._last_oct_handler = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_last_return_values);
  __Pyx_DECREF(__pyx_v_self->_last_return_values);
  __pyx_v_self->_last_return_values = Py_None;

  /* "yt/geometry/particle_oct_container.pyx":472
 *         self._last_selector = None
 *         self._last_return_values = None
 *         self._last_octree_subset = None             # <<<<<<<<<<<<<<
 *         self._last_oct_handler = None
 *         self._prev_octree_subset = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_last_octree_subset);
  __Pyx_DECREF(__pyx_v_self->_last_octree_subset);
  __pyx_v_self->_last_octree_subset = Py_None;

  /* "yt/geometry/particle_oct_container.pyx":473
 *         self._last_return_values = None
 *         self._last_octree_subset = None
 *         self._last_oct_handler = None             # <<<<<<<<<<<<<<
 *         self._prev_octree_subset = None
 *         self._prev_oct_handler = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_last_oct_handler);
  __Pyx_DECREF(__pyx_v_self->_last_oct_handler);
  __pyx_v_self->_last_oct_handler = Py_None;

  /* "yt/geometry/particle_oct_container.pyx":474
 *         self._last_octree_subset = None
 *         self._last_oct_handler = None
 *         self._prev_octree_subset = None             # <<<<<<<<<<<<<<
 *         self._prev_oct_handler = None
 *         self.file_hash = file_hash
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_prev_octree_subset);
  __Pyx_DECREF(__pyx_v_self->_prev_octree_subset);
  __pyx_v_self->_prev_octree_subset = Py_None;

  /* "yt/geometry/particle_oct_container.pyx":475
 *         self._last_oct_handler = None
 *         self._prev_octree_subset = None
 *         self._prev_oct_handler = None             # <<<<<<<<<<<<<<
 *         self.file_hash = file_hash
 *         self.nfiles = nfiles
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_prev_oct_handler);
  __Pyx_DECREF(__pyx_v_self->_prev_oct_handler);
  __pyx_v_self->_prev_oct_handler = Py_None;

  /* "yt/geometry/particle_oct_container.pyx":476
 *         self._prev_octree_subset = None
 *         self._prev_oct_handler = None
 *         self.file_hash = file_hash             # <<<<<<<<<<<<<<
 *         self.nfiles = nfiles
 *         for i in range(3):
 */
  __pyx_t_2 = __Pyx_PyInt_As_npy_int64(__pyx_v_file_hash); if (unlikely((__pyx_t_2 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 476, __pyx_L1_error)
  __pyx_v_self->file_hash = __pyx_t_2;

  /* "yt/geometry/particle_oct_container.pyx":477
 *         self._prev_oct_handler = None
 *         self.file_hash = file_hash
 *         self.nfiles = nfiles             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             self.left_edge[i] = left_edge[i]
 */
  __pyx_t_3 = __Pyx_PyInt_As_npy_uint64(__pyx_v_nfiles); if (unlikely((__pyx_t_3 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 477, __pyx_L1_error)
  __pyx_v_self->nfiles = __pyx_t_3;

  /* "yt/geometry/particle_oct_container.pyx":478
 *         self.file_hash = file_hash
 *         self.nfiles = nfiles
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             self.left_edge[i] = left_edge[i]
 *             self.right_edge[i] = right_edge[i]
 */
  for (__pyx_t_4 = 0; __pyx_t_4 < 3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "yt/geometry/particle_oct_container.pyx":479
 *         self.nfiles = nfiles
 *         for i in range(3):
 *             self.left_edge[i] = left_edge[i]             # <<<<<<<<<<<<<<
 *             self.right_edge[i] = right_edge[i]
 *             self.periodicity[i] = <np.uint8_t>periodicity[i]
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_left_edge, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 479, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 479, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    (__pyx_v_self->left_edge[__pyx_v_i]) = __pyx_t_5;

    /* "yt/geometry/particle_oct_container.pyx":480
 *         for i in range(3):
 *             self.left_edge[i] = left_edge[i]
 *             self.right_edge[i] = right_edge[i]             # <<<<<<<<<<<<<<
 *             self.periodicity[i] = <np.uint8_t>periodicity[i]
 *             self.dims[i] = (1<<index_order1)
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_right_edge, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 480, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    (__pyx_v_self->right_edge[__pyx_v_i]) = __pyx_t_5;

    /* "yt/geometry/particle_oct_container.pyx":481
 *             self.left_edge[i] = left_edge[i]
 *             self.right_edge[i] = right_edge[i]
 *             self.periodicity[i] = <np.uint8_t>periodicity[i]             # <<<<<<<<<<<<<<
 *             self.dims[i] = (1<<index_order1)
 *             self.dds[i] = (right_edge[i] - left_edge[i])/self.dims[i]
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_periodicity, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyInt_As_npy_uint8(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_uint8)-1)) && PyErr_Occurred())) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    (__pyx_v_self->periodicity[__pyx_v_i]) = ((__pyx_t_5numpy_uint8_t)__pyx_t_6);

    /* "yt/geometry/particle_oct_container.pyx":482
 *             self.right_edge[i] = right_edge[i]
 *             self.periodicity[i] = <np.uint8_t>periodicity[i]
 *             self.dims[i] = (1<<index_order1)             # <<<<<<<<<<<<<<
 *             self.dds[i] = (right_edge[i] - left_edge[i])/self.dims[i]
 *             self.idds[i] = 1.0/self.dds[i]
 */
    __pyx_t_1 = PyNumber_Lshift(__pyx_int_1, __pyx_v_index_order1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 482, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_7 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 482, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    (__pyx_v_self->dims[__pyx_v_i]) = __pyx_t_7;

    /* "yt/geometry/particle_oct_container.pyx":483
 *             self.periodicity[i] = <np.uint8_t>periodicity[i]
 *             self.dims[i] = (1<<index_order1)
 *             self.dds[i] = (right_edge[i] - left_edge[i])/self.dims[i]             # <<<<<<<<<<<<<<
 *             self.idds[i] = 1.0/self.dds[i]
 *             self.dds_mi1[i] = (right_edge[i] - left_edge[i]) / (1<<index_order1)
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_right_edge, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_left_edge, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyNumber_Subtract(__pyx_t_1, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyInt_From_npy_int32((__pyx_v_self->dims[__pyx_v_i])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_9, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    (__pyx_v_self->dds[__pyx_v_i]) = __pyx_t_5;

    /* "yt/geometry/particle_oct_container.pyx":484
 *             self.dims[i] = (1<<index_order1)
 *             self.dds[i] = (right_edge[i] - left_edge[i])/self.dims[i]
 *             self.idds[i] = 1.0/self.dds[i]             # <<<<<<<<<<<<<<
 *             self.dds_mi1[i] = (right_edge[i] - left_edge[i]) / (1<<index_order1)
 *         # We use 64-bit masks
 */
    if (unlikely((__pyx_v_self->dds[__pyx_v_i]) == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 484, __pyx_L1_error)
    }
    (__pyx_v_self->idds[__pyx_v_i]) = (1.0 / (__pyx_v_self->dds[__pyx_v_i]));

    /* "yt/geometry/particle_oct_container.pyx":485
 *             self.dds[i] = (right_edge[i] - left_edge[i])/self.dims[i]
 *             self.idds[i] = 1.0/self.dds[i]
 *             self.dds_mi1[i] = (right_edge[i] - left_edge[i]) / (1<<index_order1)             # <<<<<<<<<<<<<<
 *         # We use 64-bit masks
 *         self._used_mi2 = 0
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_right_edge, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 485, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_left_edge, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 485, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyNumber_Subtract(__pyx_t_1, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 485, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = PyNumber_Lshift(__pyx_int_1, __pyx_v_index_order1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 485, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_9, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 485, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_5 == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 485, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    (__pyx_v_self->dds_mi1[__pyx_v_i]) = __pyx_t_5;
  }

  /* "yt/geometry/particle_oct_container.pyx":487
 *             self.dds_mi1[i] = (right_edge[i] - left_edge[i]) / (1<<index_order1)
 *         # We use 64-bit masks
 *         self._used_mi2 = 0             # <<<<<<<<<<<<<<
 *         self.index_order1 = index_order1
 *         self._update_mi2(index_order2)
 */
  __pyx_v_self->_used_mi2 = 0;

  /* "yt/geometry/particle_oct_container.pyx":488
 *         # We use 64-bit masks
 *         self._used_mi2 = 0
 *         self.index_order1 = index_order1             # <<<<<<<<<<<<<<
 *         self._update_mi2(index_order2)
 *         # This will be an on/off flag for which morton index values are touched
 */
  __pyx_t_7 = __Pyx_PyInt_As_npy_int32(__pyx_v_index_order1); if (unlikely((__pyx_t_7 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 488, __pyx_L1_error)
  __pyx_v_self->index_order1 = __pyx_t_7;

  /* "yt/geometry/particle_oct_container.pyx":489
 *         self._used_mi2 = 0
 *         self.index_order1 = index_order1
 *         self._update_mi2(index_order2)             # <<<<<<<<<<<<<<
 *         # This will be an on/off flag for which morton index values are touched
 *         # by particles.
 */
  __pyx_t_3 = __Pyx_PyInt_As_npy_uint64(__pyx_v_index_order2); if (unlikely((__pyx_t_3 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 489, __pyx_L1_error)
  ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self->__pyx_vtab)->_update_mi2(__pyx_v_self, __pyx_t_3);

  /* "yt/geometry/particle_oct_container.pyx":493
 *         # by particles.
 *         # This is the simple way, for now.
 *         self.masks = np.zeros((1 << (index_order1 * 3), nfiles), dtype="uint8")             # <<<<<<<<<<<<<<
 *         self.particle_counts = np.zeros(1 << (index_order1 * 3), dtype="uint64")
 *         self.bitmasks = FileBitmasks(self.nfiles)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_v_index_order1, __pyx_int_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = PyNumber_Lshift(__pyx_int_1, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_9);
  __Pyx_INCREF(__pyx_v_nfiles);
  __Pyx_GIVEREF(__pyx_v_nfiles);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_nfiles);
  __pyx_t_9 = 0;
  __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 493, __pyx_L1_error)
  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_9, __pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_10);
  __Pyx_GOTREF(__pyx_v_self->masks);
  __Pyx_DECREF(__pyx_v_self->masks);
  __pyx_v_self->masks = __pyx_t_10;
  __pyx_t_10 = 0;

  /* "yt/geometry/particle_oct_container.pyx":494
 *         # This is the simple way, for now.
 *         self.masks = np.zeros((1 << (index_order1 * 3), nfiles), dtype="uint8")
 *         self.particle_counts = np.zeros(1 << (index_order1 * 3), dtype="uint64")             # <<<<<<<<<<<<<<
 *         self.bitmasks = FileBitmasks(self.nfiles)
 *         self.collisions = BoolArrayCollection()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_np); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = PyNumber_Multiply(__pyx_v_index_order1, __pyx_int_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_9 = PyNumber_Lshift(__pyx_int_1, __pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = PyTuple_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9);
  __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_dtype, __pyx_n_s_uint64) < 0) __PYX_ERR(0, 494, __pyx_L1_error)
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_10, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GIVEREF(__pyx_t_8);
  __Pyx_GOTREF(__pyx_v_self->particle_counts);
  __Pyx_DECREF(__pyx_v_self->particle_counts);
  __pyx_v_self->particle_counts = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "yt/geometry/particle_oct_container.pyx":495
 *         self.masks = np.zeros((1 << (index_order1 * 3), nfiles), dtype="uint8")
 *         self.particle_counts = np.zeros(1 << (index_order1 * 3), dtype="uint64")
 *         self.bitmasks = FileBitmasks(self.nfiles)             # <<<<<<<<<<<<<<
 *         self.collisions = BoolArrayCollection()
 * 
 */
  __pyx_t_8 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->nfiles); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks), __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_GIVEREF(__pyx_t_9);
  __Pyx_GOTREF(__pyx_v_self->bitmasks);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->bitmasks));
  __pyx_v_self->bitmasks = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "yt/geometry/particle_oct_container.pyx":496
 *         self.particle_counts = np.zeros(1 << (index_order1 * 3), dtype="uint64")
 *         self.bitmasks = FileBitmasks(self.nfiles)
 *         self.collisions = BoolArrayCollection()             # <<<<<<<<<<<<<<
 * 
 *     def _bitmask_logicaland(self, ifile, bcoll, out):
 */
  __pyx_t_9 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_9);
  __Pyx_GOTREF(__pyx_v_self->collisions);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->collisions));
  __pyx_v_self->collisions = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_9);
  __pyx_t_9 = 0;

  /* "yt/geometry/particle_oct_container.pyx":465
 *     cdef public int _used_mi2
 * 
 *     def __init__(self, left_edge, right_edge, periodicity, file_hash, nfiles,             # <<<<<<<<<<<<<<
 *                  index_order1, index_order2):
 *         # TODO: Set limit on maximum orders?
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":498
 *         self.collisions = BoolArrayCollection()
 * 
 *     def _bitmask_logicaland(self, ifile, bcoll, out):             # <<<<<<<<<<<<<<
 *         self.bitmasks._logicaland(ifile, bcoll, out)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_3_bitmask_logicaland(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_3_bitmask_logicaland(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_ifile = 0;
  PyObject *__pyx_v_bcoll = 0;
  PyObject *__pyx_v_out = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_bitmask_logicaland (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ifile,&__pyx_n_s_bcoll,&__pyx_n_s_out,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ifile)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bcoll)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_bitmask_logicaland", 1, 3, 3, 1); __PYX_ERR(0, 498, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_out)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_bitmask_logicaland", 1, 3, 3, 2); __PYX_ERR(0, 498, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_bitmask_logicaland") < 0)) __PYX_ERR(0, 498, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_ifile = values[0];
    __pyx_v_bcoll = values[1];
    __pyx_v_out = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_bitmask_logicaland", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 498, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap._bitmask_logicaland", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_2_bitmask_logicaland(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), __pyx_v_ifile, __pyx_v_bcoll, __pyx_v_out);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_2_bitmask_logicaland(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_ifile, PyObject *__pyx_v_bcoll, PyObject *__pyx_v_out) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_uint32_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_bitmask_logicaland", 0);

  /* "yt/geometry/particle_oct_container.pyx":499
 * 
 *     def _bitmask_logicaland(self, ifile, bcoll, out):
 *         self.bitmasks._logicaland(ifile, bcoll, out)             # <<<<<<<<<<<<<<
 * 
 *     def _bitmask_intersects(self, ifile, bcoll):
 */
  __pyx_t_1 = __Pyx_PyInt_As_npy_uint32(__pyx_v_ifile); if (unlikely((__pyx_t_1 == ((npy_uint32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 499, __pyx_L1_error)
  if (!(likely(((__pyx_v_bcoll) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_bcoll, __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection))))) __PYX_ERR(0, 499, __pyx_L1_error)
  if (!(likely(((__pyx_v_out) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_out, __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection))))) __PYX_ERR(0, 499, __pyx_L1_error)
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmasks->__pyx_vtab)->_logicaland(__pyx_v_self->bitmasks, __pyx_t_1, ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_v_bcoll), ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_v_out));

  /* "yt/geometry/particle_oct_container.pyx":498
 *         self.collisions = BoolArrayCollection()
 * 
 *     def _bitmask_logicaland(self, ifile, bcoll, out):             # <<<<<<<<<<<<<<
 *         self.bitmasks._logicaland(ifile, bcoll, out)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap._bitmask_logicaland", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":501
 *         self.bitmasks._logicaland(ifile, bcoll, out)
 * 
 *     def _bitmask_intersects(self, ifile, bcoll):             # <<<<<<<<<<<<<<
 *         return self.bitmasks._intersects(ifile, bcoll)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5_bitmask_intersects(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5_bitmask_intersects(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_ifile = 0;
  PyObject *__pyx_v_bcoll = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_bitmask_intersects (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ifile,&__pyx_n_s_bcoll,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ifile)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bcoll)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_bitmask_intersects", 1, 2, 2, 1); __PYX_ERR(0, 501, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_bitmask_intersects") < 0)) __PYX_ERR(0, 501, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_ifile = values[0];
    __pyx_v_bcoll = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_bitmask_intersects", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 501, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap._bitmask_intersects", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_4_bitmask_intersects(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), __pyx_v_ifile, __pyx_v_bcoll);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_4_bitmask_intersects(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_ifile, PyObject *__pyx_v_bcoll) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_bitmask_intersects", 0);

  /* "yt/geometry/particle_oct_container.pyx":502
 * 
 *     def _bitmask_intersects(self, ifile, bcoll):
 *         return self.bitmasks._intersects(ifile, bcoll)             # <<<<<<<<<<<<<<
 * 
 *     def update_mi2(self, np.float64_t characteristic_size,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_As_npy_uint32(__pyx_v_ifile); if (unlikely((__pyx_t_1 == ((npy_uint32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 502, __pyx_L1_error)
  if (!(likely(((__pyx_v_bcoll) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_bcoll, __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection))))) __PYX_ERR(0, 502, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmasks->__pyx_vtab)->_intersects(__pyx_v_self->bitmasks, __pyx_t_1, ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_v_bcoll))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":501
 *         self.bitmasks._logicaland(ifile, bcoll, out)
 * 
 *     def _bitmask_intersects(self, ifile, bcoll):             # <<<<<<<<<<<<<<
 *         return self.bitmasks._intersects(ifile, bcoll)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap._bitmask_intersects", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":504
 *         return self.bitmasks._intersects(ifile, bcoll)
 * 
 *     def update_mi2(self, np.float64_t characteristic_size,             # <<<<<<<<<<<<<<
 *                    np.uint64_t max_index_order2 = 6):
 *         """
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_7update_mi2(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6update_mi2[] = "\n        mi2 is the *refined* morton index order; mi2 is thus the definition of\n        the size of the refined index objects we stick inside any collisions at\n        the coarse level.  This takes a characteristic size and attempts to\n        compute the mi2 such that the cell is roughly equivalent to the\n        characteristic size.  It will return whether or not it was able to\n        update; if the mi2 has already been used, it does not update.\n        There are cases where the maximum index_order2 that it would compute\n        would be extremely fine, which can do really bad things to memory.  So\n        we allow the setting of a maximum value, which it won't exceed.\n        ";
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_7update_mi2(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __pyx_t_5numpy_float64_t __pyx_v_characteristic_size;
  __pyx_t_5numpy_uint64_t __pyx_v_max_index_order2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update_mi2 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_characteristic_size,&__pyx_n_s_max_index_order2,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_characteristic_size)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_index_order2);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "update_mi2") < 0)) __PYX_ERR(0, 504, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_characteristic_size = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_characteristic_size == ((npy_float64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 504, __pyx_L3_error)
    if (values[1]) {
      __pyx_v_max_index_order2 = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_max_index_order2 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 505, __pyx_L3_error)
    } else {
      __pyx_v_max_index_order2 = ((__pyx_t_5numpy_uint64_t)6);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("update_mi2", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 504, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.update_mi2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6update_mi2(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), __pyx_v_characteristic_size, __pyx_v_max_index_order2);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6update_mi2(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, __pyx_t_5numpy_float64_t __pyx_v_characteristic_size, __pyx_t_5numpy_uint64_t __pyx_v_max_index_order2) {
  __pyx_t_5numpy_uint64_t __pyx_v_index_order2;
  long __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  long __pyx_t_3;
  __pyx_t_5numpy_uint64_t __pyx_t_4;
  __pyx_t_5numpy_uint64_t __pyx_t_5;
  __pyx_t_5numpy_uint64_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update_mi2", 0);

  /* "yt/geometry/particle_oct_container.pyx":517
 *         we allow the setting of a maximum value, which it won't exceed.
 *         """
 *         if self._used_mi2 > 0:             # <<<<<<<<<<<<<<
 *             return self.index_order2
 *         cdef np.uint64_t index_order2 = 2
 */
  __pyx_t_1 = ((__pyx_v_self->_used_mi2 > 0) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/particle_oct_container.pyx":518
 *         """
 *         if self._used_mi2 > 0:
 *             return self.index_order2             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t index_order2 = 2
 *         for i in range(3):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->index_order2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 518, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "yt/geometry/particle_oct_container.pyx":517
 *         we allow the setting of a maximum value, which it won't exceed.
 *         """
 *         if self._used_mi2 > 0:             # <<<<<<<<<<<<<<
 *             return self.index_order2
 *         cdef np.uint64_t index_order2 = 2
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":519
 *         if self._used_mi2 > 0:
 *             return self.index_order2
 *         cdef np.uint64_t index_order2 = 2             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             # Note we're casting to signed here, to avoid negative issues.
 */
  __pyx_v_index_order2 = 2;

  /* "yt/geometry/particle_oct_container.pyx":520
 *             return self.index_order2
 *         cdef np.uint64_t index_order2 = 2
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             # Note we're casting to signed here, to avoid negative issues.
 *             if self.dds_mi1[i] < characteristic_size: continue
 */
  for (__pyx_t_3 = 0; __pyx_t_3 < 3; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":522
 *         for i in range(3):
 *             # Note we're casting to signed here, to avoid negative issues.
 *             if self.dds_mi1[i] < characteristic_size: continue             # <<<<<<<<<<<<<<
 *             index_order2 = max(index_order2, <np.uint64_t> ceil(log2(self.dds_mi1[i] / characteristic_size)))
 *         index_order2 = i64min(max_index_order2, index_order2)
 */
    __pyx_t_1 = (((__pyx_v_self->dds_mi1[__pyx_v_i]) < __pyx_v_characteristic_size) != 0);
    if (__pyx_t_1) {
      goto __pyx_L4_continue;
    }

    /* "yt/geometry/particle_oct_container.pyx":523
 *             # Note we're casting to signed here, to avoid negative issues.
 *             if self.dds_mi1[i] < characteristic_size: continue
 *             index_order2 = max(index_order2, <np.uint64_t> ceil(log2(self.dds_mi1[i] / characteristic_size)))             # <<<<<<<<<<<<<<
 *         index_order2 = i64min(max_index_order2, index_order2)
 *         self._update_mi2(index_order2)
 */
    if (unlikely(__pyx_v_characteristic_size == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 523, __pyx_L1_error)
    }
    __pyx_t_4 = ((__pyx_t_5numpy_uint64_t)ceil(log2(((__pyx_v_self->dds_mi1[__pyx_v_i]) / __pyx_v_characteristic_size))));
    __pyx_t_5 = __pyx_v_index_order2;
    if (((__pyx_t_4 > __pyx_t_5) != 0)) {
      __pyx_t_6 = __pyx_t_4;
    } else {
      __pyx_t_6 = __pyx_t_5;
    }
    __pyx_v_index_order2 = __pyx_t_6;
    __pyx_L4_continue:;
  }

  /* "yt/geometry/particle_oct_container.pyx":524
 *             if self.dds_mi1[i] < characteristic_size: continue
 *             index_order2 = max(index_order2, <np.uint64_t> ceil(log2(self.dds_mi1[i] / characteristic_size)))
 *         index_order2 = i64min(max_index_order2, index_order2)             # <<<<<<<<<<<<<<
 *         self._update_mi2(index_order2)
 *         return self.index_order2
 */
  __pyx_v_index_order2 = __pyx_f_2yt_9utilities_3lib_8fp_utils_i64min(__pyx_v_max_index_order2, __pyx_v_index_order2);

  /* "yt/geometry/particle_oct_container.pyx":525
 *             index_order2 = max(index_order2, <np.uint64_t> ceil(log2(self.dds_mi1[i] / characteristic_size)))
 *         index_order2 = i64min(max_index_order2, index_order2)
 *         self._update_mi2(index_order2)             # <<<<<<<<<<<<<<
 *         return self.index_order2
 * 
 */
  ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self->__pyx_vtab)->_update_mi2(__pyx_v_self, __pyx_v_index_order2);

  /* "yt/geometry/particle_oct_container.pyx":526
 *         index_order2 = i64min(max_index_order2, index_order2)
 *         self._update_mi2(index_order2)
 *         return self.index_order2             # <<<<<<<<<<<<<<
 * 
 *     cdef void _update_mi2(self, np.uint64_t index_order2):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->index_order2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":504
 *         return self.bitmasks._intersects(ifile, bcoll)
 * 
 *     def update_mi2(self, np.float64_t characteristic_size,             # <<<<<<<<<<<<<<
 *                    np.uint64_t max_index_order2 = 6):
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.update_mi2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":528
 *         return self.index_order2
 * 
 *     cdef void _update_mi2(self, np.uint64_t index_order2):             # <<<<<<<<<<<<<<
 *         self.index_order2 = index_order2
 *         mi2_max = (1 << self.index_order2) - 1
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap__update_mi2(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_index_order2) {
  long __pyx_v_mi2_max;
  long __pyx_v_i;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  __pyx_t_5numpy_uint64_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_update_mi2", 0);

  /* "yt/geometry/particle_oct_container.pyx":529
 * 
 *     cdef void _update_mi2(self, np.uint64_t index_order2):
 *         self.index_order2 = index_order2             # <<<<<<<<<<<<<<
 *         mi2_max = (1 << self.index_order2) - 1
 *         self.directional_max2[0] = encode_morton_64bit(mi2_max, 0, 0)
 */
  __pyx_v_self->index_order2 = __pyx_v_index_order2;

  /* "yt/geometry/particle_oct_container.pyx":530
 *     cdef void _update_mi2(self, np.uint64_t index_order2):
 *         self.index_order2 = index_order2
 *         mi2_max = (1 << self.index_order2) - 1             # <<<<<<<<<<<<<<
 *         self.directional_max2[0] = encode_morton_64bit(mi2_max, 0, 0)
 *         self.directional_max2[1] = encode_morton_64bit(0, mi2_max, 0)
 */
  __pyx_v_mi2_max = ((1 << __pyx_v_self->index_order2) - 1);

  /* "yt/geometry/particle_oct_container.pyx":531
 *         self.index_order2 = index_order2
 *         mi2_max = (1 << self.index_order2) - 1
 *         self.directional_max2[0] = encode_morton_64bit(mi2_max, 0, 0)             # <<<<<<<<<<<<<<
 *         self.directional_max2[1] = encode_morton_64bit(0, mi2_max, 0)
 *         self.directional_max2[2] = encode_morton_64bit(0, 0, mi2_max)
 */
  (__pyx_v_self->directional_max2[0]) = __pyx_f_2yt_9utilities_3lib_14geometry_utils_encode_morton_64bit(__pyx_v_mi2_max, 0, 0);

  /* "yt/geometry/particle_oct_container.pyx":532
 *         mi2_max = (1 << self.index_order2) - 1
 *         self.directional_max2[0] = encode_morton_64bit(mi2_max, 0, 0)
 *         self.directional_max2[1] = encode_morton_64bit(0, mi2_max, 0)             # <<<<<<<<<<<<<<
 *         self.directional_max2[2] = encode_morton_64bit(0, 0, mi2_max)
 *         for i in range(3):
 */
  (__pyx_v_self->directional_max2[1]) = __pyx_f_2yt_9utilities_3lib_14geometry_utils_encode_morton_64bit(0, __pyx_v_mi2_max, 0);

  /* "yt/geometry/particle_oct_container.pyx":533
 *         self.directional_max2[0] = encode_morton_64bit(mi2_max, 0, 0)
 *         self.directional_max2[1] = encode_morton_64bit(0, mi2_max, 0)
 *         self.directional_max2[2] = encode_morton_64bit(0, 0, mi2_max)             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             self.dds_mi2[i] = self.dds_mi1[i] / (1<<index_order2)
 */
  (__pyx_v_self->directional_max2[2]) = __pyx_f_2yt_9utilities_3lib_14geometry_utils_encode_morton_64bit(0, 0, __pyx_v_mi2_max);

  /* "yt/geometry/particle_oct_container.pyx":534
 *         self.directional_max2[1] = encode_morton_64bit(0, mi2_max, 0)
 *         self.directional_max2[2] = encode_morton_64bit(0, 0, mi2_max)
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             self.dds_mi2[i] = self.dds_mi1[i] / (1<<index_order2)
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":535
 *         self.directional_max2[2] = encode_morton_64bit(0, 0, mi2_max)
 *         for i in range(3):
 *             self.dds_mi2[i] = self.dds_mi1[i] / (1<<index_order2)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    __pyx_t_2 = (1 << __pyx_v_index_order2);
    if (unlikely(__pyx_t_2 == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 535, __pyx_L1_error)
    }
    (__pyx_v_self->dds_mi2[__pyx_v_i]) = ((__pyx_v_self->dds_mi1[__pyx_v_i]) / __pyx_t_2);
  }

  /* "yt/geometry/particle_oct_container.pyx":528
 *         return self.index_order2
 * 
 *     cdef void _update_mi2(self, np.uint64_t index_order2):             # <<<<<<<<<<<<<<
 *         self.index_order2 = index_order2
 *         mi2_max = (1 << self.index_order2) - 1
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("yt.geometry.particle_oct_container.ParticleBitmap._update_mi2", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":540
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def _coarse_index_data_file(self,             # <<<<<<<<<<<<<<
 *                                 np.ndarray[floating, ndim=2] pos,
 *                                 np.ndarray[floating, ndim=1] hsml,
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9_coarse_index_data_file(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9_coarse_index_data_file = {"_coarse_index_data_file", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9_coarse_index_data_file, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9_coarse_index_data_file(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(0, 540, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(0, 540, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(0, 540, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 540, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8_coarse_index_data_file(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8_coarse_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  PyTypeObject *__pyx_v_ndarray = 0;
  __Pyx_memviewslice __pyx_v_memslice;
  Py_ssize_t __pyx_v_itemsize;
  CYTHON_UNUSED int __pyx_v_dtype_signed;
  char __pyx_v_kind;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_dtype = NULL;
  PyObject *__pyx_v_arg_base = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  __Pyx_memviewslice __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  int __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_coarse_index_data_file", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 540, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 540, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 540, __pyx_L1_error)
  __pyx_t_2 = ((1 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 540, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 540, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_pos, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 540, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 540, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 540, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 540, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_4);
    __Pyx_GIVEREF(__pyx_int_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_4);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 540, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = (__pyx_v_ndarray != ((PyTypeObject*)Py_None));
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_arg, __pyx_v_ndarray); 
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (__pyx_t_2) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_v_dtype = __pyx_t_6;
        __pyx_t_6 = 0;
        goto __pyx_L12;
      }
      __pyx_t_2 = __pyx_memoryview_check(__pyx_v_arg); 
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_base); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_v_arg_base = __pyx_t_6;
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_arg_base, __pyx_v_ndarray); 
        __pyx_t_2 = (__pyx_t_3 != 0);
        if (__pyx_t_2) {
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg_base, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 540, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_v_dtype = __pyx_t_6;
          __pyx_t_6 = 0;
          goto __pyx_L13;
        }
        /*else*/ {
          __Pyx_INCREF(Py_None);
          __pyx_v_dtype = Py_None;
        }
        __pyx_L13:;
        goto __pyx_L12;
      }
      /*else*/ {
        __Pyx_INCREF(Py_None);
        __pyx_v_dtype = Py_None;
      }
      __pyx_L12:;
      __pyx_v_itemsize = -1L;
      __pyx_t_2 = (__pyx_v_dtype != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 540, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_itemsize = __pyx_t_5;
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyObject_Ord(__pyx_t_6); if (unlikely(__pyx_t_7 == ((long)(long)(Py_UCS4)-1))) __PYX_ERR(0, 540, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_kind = __pyx_t_7;
        __pyx_v_dtype_signed = (__pyx_v_kind == 'i');
        switch (__pyx_v_kind) {
          case 'i':
          case 'u':
          break;
          case 'f':
          __pyx_t_2 = (((sizeof(float)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 540, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 540, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 2) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L16_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 540, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          __pyx_t_2 = (((sizeof(double)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L19_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 540, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 540, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 2) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L19_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_double, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 540, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          break;
          case 'c':
          break;
          case 'O':
          break;
          default: break;
        }
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L22_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(float))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L22_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dsds_float(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 540, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L26_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(double))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L26_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_double, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 540, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 540, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 540, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_9), (&__pyx_t_10)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_11 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_9, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_10);
    if (unlikely(__pyx_t_11 == 0)) break;
    if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_12 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_14, __pyx_kp_s__7) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_kp_s__7);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_split); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_12, __pyx_kp_s__8) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_kp_s__8);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_15 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 540, __pyx_L1_error)
    __pyx_t_16 = __pyx_t_15;
    for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
      __pyx_v_i = __pyx_t_17;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_3 = (__pyx_v_dst_type != Py_None);
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (__pyx_t_2) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_13 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 540, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 540, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (__pyx_t_2) {
          __pyx_v_match_found = 1;
          goto __pyx_L34;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L32_break;
        }
        __pyx_L34:;
      }
    }
    __pyx_L32_break:;
    __pyx_t_2 = (__pyx_v_match_found != 0);
    if (__pyx_t_2) {
      __pyx_t_18 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(0, 540, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (__pyx_t_3) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 540, __pyx_L1_error)
  }
  __pyx_t_9 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 540, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_9 > 1) != 0);
  if (__pyx_t_3) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 540, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 540, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_dtype);
  __Pyx_XDECREF(__pyx_v_arg_base);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_0__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_81_coarse_index_data_file(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_2yt_8geometry_22particle_oct_container_14ParticleBitmap_81_coarse_index_data_file = {"__pyx_fuse_0_coarse_index_data_file", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_0__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_81_coarse_index_data_file, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_fuse_0__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_81_coarse_index_data_file(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_pos = 0;
  PyArrayObject *__pyx_v_hsml = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_file_id;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_coarse_index_data_file (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pos,&__pyx_n_s_hsml,&__pyx_n_s_file_id,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pos)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hsml)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_coarse_index_data_file", 1, 3, 3, 1); __PYX_ERR(0, 540, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_file_id)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_coarse_index_data_file", 1, 3, 3, 2); __PYX_ERR(0, 540, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_coarse_index_data_file") < 0)) __PYX_ERR(0, 540, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_pos = ((PyArrayObject *)values[0]);
    __pyx_v_hsml = ((PyArrayObject *)values[1]);
    __pyx_v_file_id = __Pyx_PyInt_As_npy_uint64(values[2]); if (unlikely((__pyx_v_file_id == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 543, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_coarse_index_data_file", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 540, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap._coarse_index_data_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_pos), __pyx_ptype_5numpy_ndarray, 1, "pos", 0))) __PYX_ERR(0, 541, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hsml), __pyx_ptype_5numpy_ndarray, 1, "hsml", 0))) __PYX_ERR(0, 542, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_80_coarse_index_data_file(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), __pyx_v_pos, __pyx_v_hsml, __pyx_v_file_id);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_80_coarse_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyArrayObject *__pyx_v_pos, PyArrayObject *__pyx_v_hsml, __pyx_t_5numpy_uint64_t __pyx_v_file_id) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_hsml;
  __Pyx_Buffer __pyx_pybuffer_hsml;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_pos;
  __Pyx_Buffer __pyx_pybuffer_pos;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_0_coarse_index_data_file", 0);
  __pyx_pybuffer_pos.pybuffer.buf = NULL;
  __pyx_pybuffer_pos.refcount = 0;
  __pyx_pybuffernd_pos.data = NULL;
  __pyx_pybuffernd_pos.rcbuffer = &__pyx_pybuffer_pos;
  __pyx_pybuffer_hsml.pybuffer.buf = NULL;
  __pyx_pybuffer_hsml.refcount = 0;
  __pyx_pybuffernd_hsml.data = NULL;
  __pyx_pybuffernd_hsml.rcbuffer = &__pyx_pybuffer_hsml;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pos.rcbuffer->pybuffer, (PyObject*)__pyx_v_pos, &__Pyx_TypeInfo_float, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 540, __pyx_L1_error)
  }
  __pyx_pybuffernd_pos.diminfo[0].strides = __pyx_pybuffernd_pos.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_pos.diminfo[0].shape = __pyx_pybuffernd_pos.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_pos.diminfo[1].strides = __pyx_pybuffernd_pos.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_pos.diminfo[1].shape = __pyx_pybuffernd_pos.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer, (PyObject*)__pyx_v_hsml, &__Pyx_TypeInfo_float, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 540, __pyx_L1_error)
  }
  __pyx_pybuffernd_hsml.diminfo[0].strides = __pyx_pybuffernd_hsml.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_hsml.diminfo[0].shape = __pyx_pybuffernd_hsml.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/particle_oct_container.pyx":544
 *                                 np.ndarray[floating, ndim=1] hsml,
 *                                 np.uint64_t file_id):
 *         return self.__coarse_index_data_file(pos, hsml, file_id)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self->__pyx_vtab)->__pyx_fuse_0__coarse_index_data_file(__pyx_v_self, ((PyArrayObject *)__pyx_v_pos), ((PyArrayObject *)__pyx_v_hsml), __pyx_v_file_id); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 544, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":540
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def _coarse_index_data_file(self,             # <<<<<<<<<<<<<<
 *                                 np.ndarray[floating, ndim=2] pos,
 *                                 np.ndarray[floating, ndim=1] hsml,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap._coarse_index_data_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_1__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_83_coarse_index_data_file(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_2yt_8geometry_22particle_oct_container_14ParticleBitmap_83_coarse_index_data_file = {"__pyx_fuse_1_coarse_index_data_file", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_1__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_83_coarse_index_data_file, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_fuse_1__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_83_coarse_index_data_file(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_pos = 0;
  PyArrayObject *__pyx_v_hsml = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_file_id;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_coarse_index_data_file (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pos,&__pyx_n_s_hsml,&__pyx_n_s_file_id,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pos)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hsml)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_coarse_index_data_file", 1, 3, 3, 1); __PYX_ERR(0, 540, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_file_id)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_coarse_index_data_file", 1, 3, 3, 2); __PYX_ERR(0, 540, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_coarse_index_data_file") < 0)) __PYX_ERR(0, 540, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_pos = ((PyArrayObject *)values[0]);
    __pyx_v_hsml = ((PyArrayObject *)values[1]);
    __pyx_v_file_id = __Pyx_PyInt_As_npy_uint64(values[2]); if (unlikely((__pyx_v_file_id == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 543, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_coarse_index_data_file", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 540, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap._coarse_index_data_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_pos), __pyx_ptype_5numpy_ndarray, 1, "pos", 0))) __PYX_ERR(0, 541, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hsml), __pyx_ptype_5numpy_ndarray, 1, "hsml", 0))) __PYX_ERR(0, 542, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_82_coarse_index_data_file(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), __pyx_v_pos, __pyx_v_hsml, __pyx_v_file_id);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_82_coarse_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyArrayObject *__pyx_v_pos, PyArrayObject *__pyx_v_hsml, __pyx_t_5numpy_uint64_t __pyx_v_file_id) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_hsml;
  __Pyx_Buffer __pyx_pybuffer_hsml;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_pos;
  __Pyx_Buffer __pyx_pybuffer_pos;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_1_coarse_index_data_file", 0);
  __pyx_pybuffer_pos.pybuffer.buf = NULL;
  __pyx_pybuffer_pos.refcount = 0;
  __pyx_pybuffernd_pos.data = NULL;
  __pyx_pybuffernd_pos.rcbuffer = &__pyx_pybuffer_pos;
  __pyx_pybuffer_hsml.pybuffer.buf = NULL;
  __pyx_pybuffer_hsml.refcount = 0;
  __pyx_pybuffernd_hsml.data = NULL;
  __pyx_pybuffernd_hsml.rcbuffer = &__pyx_pybuffer_hsml;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pos.rcbuffer->pybuffer, (PyObject*)__pyx_v_pos, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 540, __pyx_L1_error)
  }
  __pyx_pybuffernd_pos.diminfo[0].strides = __pyx_pybuffernd_pos.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_pos.diminfo[0].shape = __pyx_pybuffernd_pos.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_pos.diminfo[1].strides = __pyx_pybuffernd_pos.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_pos.diminfo[1].shape = __pyx_pybuffernd_pos.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer, (PyObject*)__pyx_v_hsml, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 540, __pyx_L1_error)
  }
  __pyx_pybuffernd_hsml.diminfo[0].strides = __pyx_pybuffernd_hsml.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_hsml.diminfo[0].shape = __pyx_pybuffernd_hsml.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/particle_oct_container.pyx":544
 *                                 np.ndarray[floating, ndim=1] hsml,
 *                                 np.uint64_t file_id):
 *         return self.__coarse_index_data_file(pos, hsml, file_id)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self->__pyx_vtab)->__pyx_fuse_1__coarse_index_data_file(__pyx_v_self, ((PyArrayObject *)__pyx_v_pos), ((PyArrayObject *)__pyx_v_hsml), __pyx_v_file_id); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 544, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":540
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def _coarse_index_data_file(self,             # <<<<<<<<<<<<<<
 *                                 np.ndarray[floating, ndim=2] pos,
 *                                 np.ndarray[floating, ndim=1] hsml,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap._coarse_index_data_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":549
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void __coarse_index_data_file(self,             # <<<<<<<<<<<<<<
 *                                        np.ndarray[floating, ndim=2] pos,
 *                                        np.ndarray[floating, ndim=1] hsml,
 */

static void __pyx_fuse_0__pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___coarse_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyArrayObject *__pyx_v_pos, PyArrayObject *__pyx_v_hsml, __pyx_t_5numpy_uint64_t __pyx_v_file_id) {
  __pyx_t_5numpy_int64_t __pyx_v_i;
  __pyx_t_5numpy_int64_t __pyx_v_p;
  __pyx_t_5numpy_uint64_t __pyx_v_mi;
  __pyx_t_5numpy_uint64_t __pyx_v_miex;
  __pyx_t_5numpy_uint64_t __pyx_v_mi_split[3];
  __pyx_t_5numpy_float64_t __pyx_v_ppos[3];
  __pyx_t_5numpy_float64_t __pyx_v_s_ppos[3];
  __pyx_t_5numpy_float64_t __pyx_v_clip_pos_l[3];
  __pyx_t_5numpy_float64_t __pyx_v_clip_pos_r[3];
  int __pyx_v_skip;
  __pyx_t_5numpy_uint64_t __pyx_v_bounds[2][3];
  __pyx_t_5numpy_uint64_t __pyx_v_xex;
  __pyx_t_5numpy_uint64_t __pyx_v_yex;
  __pyx_t_5numpy_uint64_t __pyx_v_zex;
  __pyx_t_5numpy_float64_t __pyx_v_LE[3];
  __pyx_t_5numpy_float64_t __pyx_v_RE[3];
  __pyx_t_5numpy_float64_t __pyx_v_DW[3];
  __pyx_t_5numpy_uint8_t __pyx_v_PER[3];
  __pyx_t_5numpy_float64_t __pyx_v_dds[3];
  __pyx_t_5numpy_float64_t __pyx_v_radius;
  __Pyx_memviewslice __pyx_v_mask = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_particle_counts = { 0, 0, { 0 }, { 0 }, { 0 } };
  __pyx_t_5numpy_uint64_t __pyx_v_msize;
  int __pyx_v_axiter[3][2];
  __pyx_t_5numpy_float64_t __pyx_v_axiterv[3][2];
  long __pyx_v_xi;
  long __pyx_v_yi;
  long __pyx_v_zi;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_hsml;
  __Pyx_Buffer __pyx_pybuffer_hsml;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_pos;
  __Pyx_Buffer __pyx_pybuffer_pos;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_4 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __pyx_t_5numpy_int64_t __pyx_t_5;
  npy_intp __pyx_t_6;
  npy_intp __pyx_t_7;
  __pyx_t_5numpy_int64_t __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  float __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  __pyx_t_5numpy_uint64_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_UCS4 __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  long __pyx_t_18;
  long __pyx_t_19;
  long __pyx_t_20;
  __pyx_t_5numpy_uint64_t __pyx_t_21;
  __pyx_t_5numpy_uint64_t __pyx_t_22;
  __pyx_t_5numpy_uint64_t __pyx_t_23;
  __pyx_t_5numpy_uint64_t __pyx_t_24;
  __pyx_t_5numpy_uint64_t __pyx_t_25;
  __pyx_t_5numpy_uint64_t __pyx_t_26;
  __pyx_t_5numpy_uint64_t __pyx_t_27;
  __pyx_t_5numpy_uint64_t __pyx_t_28;
  __pyx_t_5numpy_uint64_t __pyx_t_29;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_0__coarse_index_data_file", 0);
  __pyx_pybuffer_pos.pybuffer.buf = NULL;
  __pyx_pybuffer_pos.refcount = 0;
  __pyx_pybuffernd_pos.data = NULL;
  __pyx_pybuffernd_pos.rcbuffer = &__pyx_pybuffer_pos;
  __pyx_pybuffer_hsml.pybuffer.buf = NULL;
  __pyx_pybuffer_hsml.refcount = 0;
  __pyx_pybuffernd_hsml.data = NULL;
  __pyx_pybuffernd_hsml.rcbuffer = &__pyx_pybuffer_hsml;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pos.rcbuffer->pybuffer, (PyObject*)__pyx_v_pos, &__Pyx_TypeInfo_float, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 549, __pyx_L1_error)
  }
  __pyx_pybuffernd_pos.diminfo[0].strides = __pyx_pybuffernd_pos.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_pos.diminfo[0].shape = __pyx_pybuffernd_pos.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_pos.diminfo[1].strides = __pyx_pybuffernd_pos.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_pos.diminfo[1].shape = __pyx_pybuffernd_pos.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer, (PyObject*)__pyx_v_hsml, &__Pyx_TypeInfo_float, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 549, __pyx_L1_error)
  }
  __pyx_pybuffernd_hsml.diminfo[0].strides = __pyx_pybuffernd_hsml.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_hsml.diminfo[0].shape = __pyx_pybuffernd_hsml.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/particle_oct_container.pyx":570
 *         cdef np.float64_t dds[3]
 *         cdef np.float64_t radius
 *         cdef np.uint8_t[:] mask = self.masks[:, file_id]             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t[:] particle_counts = self.particle_counts
 *         cdef np.uint64_t msize = (1 << (self.index_order1 * 3))
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint64(__pyx_v_file_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_slice__11);
  __Pyx_GIVEREF(__pyx_slice__11);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_slice__11);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_self->masks, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint8_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mask = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":571
 *         cdef np.float64_t radius
 *         cdef np.uint8_t[:] mask = self.masks[:, file_id]
 *         cdef np.uint64_t[:] particle_counts = self.particle_counts             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t msize = (1 << (self.index_order1 * 3))
 *         cdef int axiter[3][2]
 */
  __pyx_t_4 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_v_self->particle_counts, PyBUF_WRITABLE); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(0, 571, __pyx_L1_error)
  __pyx_v_particle_counts = __pyx_t_4;
  __pyx_t_4.memview = NULL;
  __pyx_t_4.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":572
 *         cdef np.uint8_t[:] mask = self.masks[:, file_id]
 *         cdef np.uint64_t[:] particle_counts = self.particle_counts
 *         cdef np.uint64_t msize = (1 << (self.index_order1 * 3))             # <<<<<<<<<<<<<<
 *         cdef int axiter[3][2]
 *         cdef np.float64_t axiterv[3][2]
 */
  __pyx_v_msize = (1 << (__pyx_v_self->index_order1 * 3));

  /* "yt/geometry/particle_oct_container.pyx":576
 *         cdef np.float64_t axiterv[3][2]
 *         # Copy over things for this file (type cast necessary?)
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             LE[i] = self.left_edge[i]
 *             RE[i] = self.right_edge[i]
 */
  for (__pyx_t_5 = 0; __pyx_t_5 < 3; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "yt/geometry/particle_oct_container.pyx":577
 *         # Copy over things for this file (type cast necessary?)
 *         for i in range(3):
 *             LE[i] = self.left_edge[i]             # <<<<<<<<<<<<<<
 *             RE[i] = self.right_edge[i]
 *             PER[i] = self.periodicity[i]
 */
    (__pyx_v_LE[__pyx_v_i]) = (__pyx_v_self->left_edge[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":578
 *         for i in range(3):
 *             LE[i] = self.left_edge[i]
 *             RE[i] = self.right_edge[i]             # <<<<<<<<<<<<<<
 *             PER[i] = self.periodicity[i]
 *             dds[i] = self.dds_mi1[i]
 */
    (__pyx_v_RE[__pyx_v_i]) = (__pyx_v_self->right_edge[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":579
 *             LE[i] = self.left_edge[i]
 *             RE[i] = self.right_edge[i]
 *             PER[i] = self.periodicity[i]             # <<<<<<<<<<<<<<
 *             dds[i] = self.dds_mi1[i]
 *             DW[i] = RE[i] - LE[i]
 */
    (__pyx_v_PER[__pyx_v_i]) = (__pyx_v_self->periodicity[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":580
 *             RE[i] = self.right_edge[i]
 *             PER[i] = self.periodicity[i]
 *             dds[i] = self.dds_mi1[i]             # <<<<<<<<<<<<<<
 *             DW[i] = RE[i] - LE[i]
 *             axiter[i][0] = 0 # We always do an offset of 0
 */
    (__pyx_v_dds[__pyx_v_i]) = (__pyx_v_self->dds_mi1[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":581
 *             PER[i] = self.periodicity[i]
 *             dds[i] = self.dds_mi1[i]
 *             DW[i] = RE[i] - LE[i]             # <<<<<<<<<<<<<<
 *             axiter[i][0] = 0 # We always do an offset of 0
 *             axiterv[i][0] = 0.0
 */
    (__pyx_v_DW[__pyx_v_i]) = ((__pyx_v_RE[__pyx_v_i]) - (__pyx_v_LE[__pyx_v_i]));

    /* "yt/geometry/particle_oct_container.pyx":582
 *             dds[i] = self.dds_mi1[i]
 *             DW[i] = RE[i] - LE[i]
 *             axiter[i][0] = 0 # We always do an offset of 0             # <<<<<<<<<<<<<<
 *             axiterv[i][0] = 0.0
 *         # Mark index of particles that are in this file
 */
    ((__pyx_v_axiter[__pyx_v_i])[0]) = 0;

    /* "yt/geometry/particle_oct_container.pyx":583
 *             DW[i] = RE[i] - LE[i]
 *             axiter[i][0] = 0 # We always do an offset of 0
 *             axiterv[i][0] = 0.0             # <<<<<<<<<<<<<<
 *         # Mark index of particles that are in this file
 *         for p in range(pos.shape[0]):
 */
    ((__pyx_v_axiterv[__pyx_v_i])[0]) = 0.0;
  }

  /* "yt/geometry/particle_oct_container.pyx":585
 *             axiterv[i][0] = 0.0
 *         # Mark index of particles that are in this file
 *         for p in range(pos.shape[0]):             # <<<<<<<<<<<<<<
 *             skip = 0
 *             for i in range(3):
 */
  __pyx_t_6 = (__pyx_v_pos->dimensions[0]);
  __pyx_t_7 = __pyx_t_6;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_7; __pyx_t_5+=1) {
    __pyx_v_p = __pyx_t_5;

    /* "yt/geometry/particle_oct_container.pyx":586
 *         # Mark index of particles that are in this file
 *         for p in range(pos.shape[0]):
 *             skip = 0             # <<<<<<<<<<<<<<
 *             for i in range(3):
 *                 axiter[i][1] = 999
 */
    __pyx_v_skip = 0;

    /* "yt/geometry/particle_oct_container.pyx":587
 *         for p in range(pos.shape[0]):
 *             skip = 0
 *             for i in range(3):             # <<<<<<<<<<<<<<
 *                 axiter[i][1] = 999
 *                 # Skip particles outside the domain
 */
    for (__pyx_t_8 = 0; __pyx_t_8 < 3; __pyx_t_8+=1) {
      __pyx_v_i = __pyx_t_8;

      /* "yt/geometry/particle_oct_container.pyx":588
 *             skip = 0
 *             for i in range(3):
 *                 axiter[i][1] = 999             # <<<<<<<<<<<<<<
 *                 # Skip particles outside the domain
 *                 if not (LE[i] <= pos[p, i] < RE[i]):
 */
      ((__pyx_v_axiter[__pyx_v_i])[1]) = 0x3E7;

      /* "yt/geometry/particle_oct_container.pyx":590
 *                 axiter[i][1] = 999
 *                 # Skip particles outside the domain
 *                 if not (LE[i] <= pos[p, i] < RE[i]):             # <<<<<<<<<<<<<<
 *                     skip = 1
 *                     break
 */
      __pyx_t_9 = __pyx_v_p;
      __pyx_t_10 = __pyx_v_i;
      __pyx_t_11 = (*__Pyx_BufPtrStrided2d(float *, __pyx_pybuffernd_pos.rcbuffer->pybuffer.buf, __pyx_t_9, __pyx_pybuffernd_pos.diminfo[0].strides, __pyx_t_10, __pyx_pybuffernd_pos.diminfo[1].strides));
      __pyx_t_12 = ((__pyx_v_LE[__pyx_v_i]) <= __pyx_t_11);
      if (__pyx_t_12) {
        __pyx_t_12 = (__pyx_t_11 < (__pyx_v_RE[__pyx_v_i]));
      }
      __pyx_t_13 = ((!(__pyx_t_12 != 0)) != 0);
      if (__pyx_t_13) {

        /* "yt/geometry/particle_oct_container.pyx":591
 *                 # Skip particles outside the domain
 *                 if not (LE[i] <= pos[p, i] < RE[i]):
 *                     skip = 1             # <<<<<<<<<<<<<<
 *                     break
 *                 ppos[i] = pos[p,i]
 */
        __pyx_v_skip = 1;

        /* "yt/geometry/particle_oct_container.pyx":592
 *                 if not (LE[i] <= pos[p, i] < RE[i]):
 *                     skip = 1
 *                     break             # <<<<<<<<<<<<<<
 *                 ppos[i] = pos[p,i]
 *             if skip == 1: continue
 */
        goto __pyx_L8_break;

        /* "yt/geometry/particle_oct_container.pyx":590
 *                 axiter[i][1] = 999
 *                 # Skip particles outside the domain
 *                 if not (LE[i] <= pos[p, i] < RE[i]):             # <<<<<<<<<<<<<<
 *                     skip = 1
 *                     break
 */
      }

      /* "yt/geometry/particle_oct_container.pyx":593
 *                     skip = 1
 *                     break
 *                 ppos[i] = pos[p,i]             # <<<<<<<<<<<<<<
 *             if skip == 1: continue
 *             mi = bounded_morton_split_dds(ppos[0], ppos[1], ppos[2], LE,
 */
      __pyx_t_10 = __pyx_v_p;
      __pyx_t_9 = __pyx_v_i;
      (__pyx_v_ppos[__pyx_v_i]) = (*__Pyx_BufPtrStrided2d(float *, __pyx_pybuffernd_pos.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_pos.diminfo[0].strides, __pyx_t_9, __pyx_pybuffernd_pos.diminfo[1].strides));
    }
    __pyx_L8_break:;

    /* "yt/geometry/particle_oct_container.pyx":594
 *                     break
 *                 ppos[i] = pos[p,i]
 *             if skip == 1: continue             # <<<<<<<<<<<<<<
 *             mi = bounded_morton_split_dds(ppos[0], ppos[1], ppos[2], LE,
 *                                           dds, mi_split)
 */
    __pyx_t_13 = ((__pyx_v_skip == 1) != 0);
    if (__pyx_t_13) {
      goto __pyx_L5_continue;
    }

    /* "yt/geometry/particle_oct_container.pyx":595
 *                 ppos[i] = pos[p,i]
 *             if skip == 1: continue
 *             mi = bounded_morton_split_dds(ppos[0], ppos[1], ppos[2], LE,             # <<<<<<<<<<<<<<
 *                                           dds, mi_split)
 *             mask[mi] = 1
 */
    __pyx_v_mi = __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_dds((__pyx_v_ppos[0]), (__pyx_v_ppos[1]), (__pyx_v_ppos[2]), __pyx_v_LE, __pyx_v_dds, __pyx_v_mi_split);

    /* "yt/geometry/particle_oct_container.pyx":597
 *             mi = bounded_morton_split_dds(ppos[0], ppos[1], ppos[2], LE,
 *                                           dds, mi_split)
 *             mask[mi] = 1             # <<<<<<<<<<<<<<
 *             particle_counts[mi] += 1
 *             # Expand mask by softening
 */
    __pyx_t_14 = __pyx_v_mi;
    *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_mask.data + __pyx_t_14 * __pyx_v_mask.strides[0]) )) = 1;

    /* "yt/geometry/particle_oct_container.pyx":598
 *                                           dds, mi_split)
 *             mask[mi] = 1
 *             particle_counts[mi] += 1             # <<<<<<<<<<<<<<
 *             # Expand mask by softening
 *             if hsml is None:
 */
    __pyx_t_14 = __pyx_v_mi;
    *((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_particle_counts.data + __pyx_t_14 * __pyx_v_particle_counts.strides[0]) )) += 1;

    /* "yt/geometry/particle_oct_container.pyx":600
 *             particle_counts[mi] += 1
 *             # Expand mask by softening
 *             if hsml is None:             # <<<<<<<<<<<<<<
 *                 continue
 *             if hsml[p] < 0:
 */
    __pyx_t_13 = (((PyObject *)__pyx_v_hsml) == Py_None);
    __pyx_t_12 = (__pyx_t_13 != 0);
    if (__pyx_t_12) {

      /* "yt/geometry/particle_oct_container.pyx":601
 *             # Expand mask by softening
 *             if hsml is None:
 *                 continue             # <<<<<<<<<<<<<<
 *             if hsml[p] < 0:
 *                 raise RuntimeError(
 */
      goto __pyx_L5_continue;

      /* "yt/geometry/particle_oct_container.pyx":600
 *             particle_counts[mi] += 1
 *             # Expand mask by softening
 *             if hsml is None:             # <<<<<<<<<<<<<<
 *                 continue
 *             if hsml[p] < 0:
 */
    }

    /* "yt/geometry/particle_oct_container.pyx":602
 *             if hsml is None:
 *                 continue
 *             if hsml[p] < 0:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     f"Smoothing length for particle {p} is negative with "
 */
    __pyx_t_8 = __pyx_v_p;
    __pyx_t_12 = (((*__Pyx_BufPtrStrided1d(float *, __pyx_pybuffernd_hsml.rcbuffer->pybuffer.buf, __pyx_t_8, __pyx_pybuffernd_hsml.diminfo[0].strides)) < 0.0) != 0);
    if (unlikely(__pyx_t_12)) {

      /* "yt/geometry/particle_oct_container.pyx":604
 *             if hsml[p] < 0:
 *                 raise RuntimeError(
 *                     f"Smoothing length for particle {p} is negative with "             # <<<<<<<<<<<<<<
 *                     f"value {hsml[p]}")
 *             radius = hsml[p]
 */
      __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 604, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_15 = 0;
      __pyx_t_16 = 127;
      __Pyx_INCREF(__pyx_kp_u_Smoothing_length_for_particle);
      __pyx_t_15 += 30;
      __Pyx_GIVEREF(__pyx_kp_u_Smoothing_length_for_particle);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Smoothing_length_for_particle);
      __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_p); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 604, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_17 = __Pyx_PyObject_FormatSimple(__pyx_t_2, __pyx_empty_unicode); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 604, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_16 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_17) > __pyx_t_16) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_17) : __pyx_t_16;
      __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_17);
      __Pyx_GIVEREF(__pyx_t_17);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_17);
      __pyx_t_17 = 0;
      __Pyx_INCREF(__pyx_kp_u_is_negative_with_value);
      __pyx_t_15 += 24;
      __Pyx_GIVEREF(__pyx_kp_u_is_negative_with_value);
      PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_is_negative_with_value);

      /* "yt/geometry/particle_oct_container.pyx":605
 *                 raise RuntimeError(
 *                     f"Smoothing length for particle {p} is negative with "
 *                     f"value {hsml[p]}")             # <<<<<<<<<<<<<<
 *             radius = hsml[p]
 *             # We first check if we're bounded within the domain; this follows the logic in the
 */
      __pyx_t_8 = __pyx_v_p;
      __pyx_t_17 = PyFloat_FromDouble((*__Pyx_BufPtrStrided1d(float *, __pyx_pybuffernd_hsml.rcbuffer->pybuffer.buf, __pyx_t_8, __pyx_pybuffernd_hsml.diminfo[0].strides))); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 605, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_17, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 605, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __pyx_t_16 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_16) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_16;
      __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "yt/geometry/particle_oct_container.pyx":604
 *             if hsml[p] < 0:
 *                 raise RuntimeError(
 *                     f"Smoothing length for particle {p} is negative with "             # <<<<<<<<<<<<<<
 *                     f"value {hsml[p]}")
 *             radius = hsml[p]
 */
      __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_1, 4, __pyx_t_15, __pyx_t_16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 604, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "yt/geometry/particle_oct_container.pyx":603
 *                 continue
 *             if hsml[p] < 0:
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     f"Smoothing length for particle {p} is negative with "
 *                     f"value {hsml[p]}")
 */
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 603, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 603, __pyx_L1_error)

      /* "yt/geometry/particle_oct_container.pyx":602
 *             if hsml is None:
 *                 continue
 *             if hsml[p] < 0:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     f"Smoothing length for particle {p} is negative with "
 */
    }

    /* "yt/geometry/particle_oct_container.pyx":606
 *                     f"Smoothing length for particle {p} is negative with "
 *                     f"value {hsml[p]}")
 *             radius = hsml[p]             # <<<<<<<<<<<<<<
 *             # We first check if we're bounded within the domain; this follows the logic in the
 *             # pixelize_cartesian routine.  We assume that no smoothing
 */
    __pyx_t_8 = __pyx_v_p;
    __pyx_v_radius = (*__Pyx_BufPtrStrided1d(float *, __pyx_pybuffernd_hsml.rcbuffer->pybuffer.buf, __pyx_t_8, __pyx_pybuffernd_hsml.diminfo[0].strides));

    /* "yt/geometry/particle_oct_container.pyx":610
 *             # pixelize_cartesian routine.  We assume that no smoothing
 *             # length can wrap around both directions.
 *             for i in range(3):             # <<<<<<<<<<<<<<
 *                 if PER[i] and ppos[i] - radius < LE[i]:
 *                     axiter[i][1] = +1
 */
    for (__pyx_t_8 = 0; __pyx_t_8 < 3; __pyx_t_8+=1) {
      __pyx_v_i = __pyx_t_8;

      /* "yt/geometry/particle_oct_container.pyx":611
 *             # length can wrap around both directions.
 *             for i in range(3):
 *                 if PER[i] and ppos[i] - radius < LE[i]:             # <<<<<<<<<<<<<<
 *                     axiter[i][1] = +1
 *                     axiterv[i][1] = DW[i]
 */
      __pyx_t_13 = ((__pyx_v_PER[__pyx_v_i]) != 0);
      if (__pyx_t_13) {
      } else {
        __pyx_t_12 = __pyx_t_13;
        goto __pyx_L16_bool_binop_done;
      }
      __pyx_t_13 = ((((__pyx_v_ppos[__pyx_v_i]) - __pyx_v_radius) < (__pyx_v_LE[__pyx_v_i])) != 0);
      __pyx_t_12 = __pyx_t_13;
      __pyx_L16_bool_binop_done:;
      if (__pyx_t_12) {

        /* "yt/geometry/particle_oct_container.pyx":612
 *             for i in range(3):
 *                 if PER[i] and ppos[i] - radius < LE[i]:
 *                     axiter[i][1] = +1             # <<<<<<<<<<<<<<
 *                     axiterv[i][1] = DW[i]
 *                 elif PER[i] and ppos[i] + radius > RE[i]:
 */
        ((__pyx_v_axiter[__pyx_v_i])[1]) = 1;

        /* "yt/geometry/particle_oct_container.pyx":613
 *                 if PER[i] and ppos[i] - radius < LE[i]:
 *                     axiter[i][1] = +1
 *                     axiterv[i][1] = DW[i]             # <<<<<<<<<<<<<<
 *                 elif PER[i] and ppos[i] + radius > RE[i]:
 *                     axiter[i][1] = -1
 */
        ((__pyx_v_axiterv[__pyx_v_i])[1]) = (__pyx_v_DW[__pyx_v_i]);

        /* "yt/geometry/particle_oct_container.pyx":611
 *             # length can wrap around both directions.
 *             for i in range(3):
 *                 if PER[i] and ppos[i] - radius < LE[i]:             # <<<<<<<<<<<<<<
 *                     axiter[i][1] = +1
 *                     axiterv[i][1] = DW[i]
 */
        goto __pyx_L15;
      }

      /* "yt/geometry/particle_oct_container.pyx":614
 *                     axiter[i][1] = +1
 *                     axiterv[i][1] = DW[i]
 *                 elif PER[i] and ppos[i] + radius > RE[i]:             # <<<<<<<<<<<<<<
 *                     axiter[i][1] = -1
 *                     axiterv[i][1] = -DW[i]
 */
      __pyx_t_13 = ((__pyx_v_PER[__pyx_v_i]) != 0);
      if (__pyx_t_13) {
      } else {
        __pyx_t_12 = __pyx_t_13;
        goto __pyx_L18_bool_binop_done;
      }
      __pyx_t_13 = ((((__pyx_v_ppos[__pyx_v_i]) + __pyx_v_radius) > (__pyx_v_RE[__pyx_v_i])) != 0);
      __pyx_t_12 = __pyx_t_13;
      __pyx_L18_bool_binop_done:;
      if (__pyx_t_12) {

        /* "yt/geometry/particle_oct_container.pyx":615
 *                     axiterv[i][1] = DW[i]
 *                 elif PER[i] and ppos[i] + radius > RE[i]:
 *                     axiter[i][1] = -1             # <<<<<<<<<<<<<<
 *                     axiterv[i][1] = -DW[i]
 *             for xi in range(2):
 */
        ((__pyx_v_axiter[__pyx_v_i])[1]) = -1;

        /* "yt/geometry/particle_oct_container.pyx":616
 *                 elif PER[i] and ppos[i] + radius > RE[i]:
 *                     axiter[i][1] = -1
 *                     axiterv[i][1] = -DW[i]             # <<<<<<<<<<<<<<
 *             for xi in range(2):
 *                 if axiter[0][xi] == 999: continue
 */
        ((__pyx_v_axiterv[__pyx_v_i])[1]) = (-(__pyx_v_DW[__pyx_v_i]));

        /* "yt/geometry/particle_oct_container.pyx":614
 *                     axiter[i][1] = +1
 *                     axiterv[i][1] = DW[i]
 *                 elif PER[i] and ppos[i] + radius > RE[i]:             # <<<<<<<<<<<<<<
 *                     axiter[i][1] = -1
 *                     axiterv[i][1] = -DW[i]
 */
      }
      __pyx_L15:;
    }

    /* "yt/geometry/particle_oct_container.pyx":617
 *                     axiter[i][1] = -1
 *                     axiterv[i][1] = -DW[i]
 *             for xi in range(2):             # <<<<<<<<<<<<<<
 *                 if axiter[0][xi] == 999: continue
 *                 s_ppos[0] = ppos[0] + axiterv[0][xi]
 */
    for (__pyx_t_18 = 0; __pyx_t_18 < 2; __pyx_t_18+=1) {
      __pyx_v_xi = __pyx_t_18;

      /* "yt/geometry/particle_oct_container.pyx":618
 *                     axiterv[i][1] = -DW[i]
 *             for xi in range(2):
 *                 if axiter[0][xi] == 999: continue             # <<<<<<<<<<<<<<
 *                 s_ppos[0] = ppos[0] + axiterv[0][xi]
 *                 for yi in range(2):
 */
      __pyx_t_12 = ((((__pyx_v_axiter[0])[__pyx_v_xi]) == 0x3E7) != 0);
      if (__pyx_t_12) {
        goto __pyx_L20_continue;
      }

      /* "yt/geometry/particle_oct_container.pyx":619
 *             for xi in range(2):
 *                 if axiter[0][xi] == 999: continue
 *                 s_ppos[0] = ppos[0] + axiterv[0][xi]             # <<<<<<<<<<<<<<
 *                 for yi in range(2):
 *                     if axiter[1][yi] == 999: continue
 */
      (__pyx_v_s_ppos[0]) = ((__pyx_v_ppos[0]) + ((__pyx_v_axiterv[0])[__pyx_v_xi]));

      /* "yt/geometry/particle_oct_container.pyx":620
 *                 if axiter[0][xi] == 999: continue
 *                 s_ppos[0] = ppos[0] + axiterv[0][xi]
 *                 for yi in range(2):             # <<<<<<<<<<<<<<
 *                     if axiter[1][yi] == 999: continue
 *                     s_ppos[1] = ppos[1] + axiterv[1][yi]
 */
      for (__pyx_t_19 = 0; __pyx_t_19 < 2; __pyx_t_19+=1) {
        __pyx_v_yi = __pyx_t_19;

        /* "yt/geometry/particle_oct_container.pyx":621
 *                 s_ppos[0] = ppos[0] + axiterv[0][xi]
 *                 for yi in range(2):
 *                     if axiter[1][yi] == 999: continue             # <<<<<<<<<<<<<<
 *                     s_ppos[1] = ppos[1] + axiterv[1][yi]
 *                     for zi in range(2):
 */
        __pyx_t_12 = ((((__pyx_v_axiter[1])[__pyx_v_yi]) == 0x3E7) != 0);
        if (__pyx_t_12) {
          goto __pyx_L23_continue;
        }

        /* "yt/geometry/particle_oct_container.pyx":622
 *                 for yi in range(2):
 *                     if axiter[1][yi] == 999: continue
 *                     s_ppos[1] = ppos[1] + axiterv[1][yi]             # <<<<<<<<<<<<<<
 *                     for zi in range(2):
 *                         if axiter[2][zi] == 999: continue
 */
        (__pyx_v_s_ppos[1]) = ((__pyx_v_ppos[1]) + ((__pyx_v_axiterv[1])[__pyx_v_yi]));

        /* "yt/geometry/particle_oct_container.pyx":623
 *                     if axiter[1][yi] == 999: continue
 *                     s_ppos[1] = ppos[1] + axiterv[1][yi]
 *                     for zi in range(2):             # <<<<<<<<<<<<<<
 *                         if axiter[2][zi] == 999: continue
 *                         s_ppos[2] = ppos[2] + axiterv[2][zi]
 */
        for (__pyx_t_20 = 0; __pyx_t_20 < 2; __pyx_t_20+=1) {
          __pyx_v_zi = __pyx_t_20;

          /* "yt/geometry/particle_oct_container.pyx":624
 *                     s_ppos[1] = ppos[1] + axiterv[1][yi]
 *                     for zi in range(2):
 *                         if axiter[2][zi] == 999: continue             # <<<<<<<<<<<<<<
 *                         s_ppos[2] = ppos[2] + axiterv[2][zi]
 *                         # OK, now we compute the left and right edges for this shift.
 */
          __pyx_t_12 = ((((__pyx_v_axiter[2])[__pyx_v_zi]) == 0x3E7) != 0);
          if (__pyx_t_12) {
            goto __pyx_L26_continue;
          }

          /* "yt/geometry/particle_oct_container.pyx":625
 *                     for zi in range(2):
 *                         if axiter[2][zi] == 999: continue
 *                         s_ppos[2] = ppos[2] + axiterv[2][zi]             # <<<<<<<<<<<<<<
 *                         # OK, now we compute the left and right edges for this shift.
 *                         for i in range(3):
 */
          (__pyx_v_s_ppos[2]) = ((__pyx_v_ppos[2]) + ((__pyx_v_axiterv[2])[__pyx_v_zi]));

          /* "yt/geometry/particle_oct_container.pyx":627
 *                         s_ppos[2] = ppos[2] + axiterv[2][zi]
 *                         # OK, now we compute the left and right edges for this shift.
 *                         for i in range(3):             # <<<<<<<<<<<<<<
 *                             clip_pos_l[i] = fmax(s_ppos[i] - radius, LE[i] + dds[i]/10)
 *                             clip_pos_r[i] = fmin(s_ppos[i] + radius, RE[i] - dds[i]/10)
 */
          for (__pyx_t_8 = 0; __pyx_t_8 < 3; __pyx_t_8+=1) {
            __pyx_v_i = __pyx_t_8;

            /* "yt/geometry/particle_oct_container.pyx":628
 *                         # OK, now we compute the left and right edges for this shift.
 *                         for i in range(3):
 *                             clip_pos_l[i] = fmax(s_ppos[i] - radius, LE[i] + dds[i]/10)             # <<<<<<<<<<<<<<
 *                             clip_pos_r[i] = fmin(s_ppos[i] + radius, RE[i] - dds[i]/10)
 *                         bounded_morton_split_dds(clip_pos_l[0], clip_pos_l[1], clip_pos_l[2], LE, dds, bounds[0])
 */
            (__pyx_v_clip_pos_l[__pyx_v_i]) = __pyx_f_2yt_9utilities_3lib_8fp_utils_fmax(((__pyx_v_s_ppos[__pyx_v_i]) - __pyx_v_radius), ((__pyx_v_LE[__pyx_v_i]) + ((__pyx_v_dds[__pyx_v_i]) / 10.0)));

            /* "yt/geometry/particle_oct_container.pyx":629
 *                         for i in range(3):
 *                             clip_pos_l[i] = fmax(s_ppos[i] - radius, LE[i] + dds[i]/10)
 *                             clip_pos_r[i] = fmin(s_ppos[i] + radius, RE[i] - dds[i]/10)             # <<<<<<<<<<<<<<
 *                         bounded_morton_split_dds(clip_pos_l[0], clip_pos_l[1], clip_pos_l[2], LE, dds, bounds[0])
 *                         bounded_morton_split_dds(clip_pos_r[0], clip_pos_r[1], clip_pos_r[2], LE, dds, bounds[1])
 */
            (__pyx_v_clip_pos_r[__pyx_v_i]) = __pyx_f_2yt_9utilities_3lib_8fp_utils_fmin(((__pyx_v_s_ppos[__pyx_v_i]) + __pyx_v_radius), ((__pyx_v_RE[__pyx_v_i]) - ((__pyx_v_dds[__pyx_v_i]) / 10.0)));
          }

          /* "yt/geometry/particle_oct_container.pyx":630
 *                             clip_pos_l[i] = fmax(s_ppos[i] - radius, LE[i] + dds[i]/10)
 *                             clip_pos_r[i] = fmin(s_ppos[i] + radius, RE[i] - dds[i]/10)
 *                         bounded_morton_split_dds(clip_pos_l[0], clip_pos_l[1], clip_pos_l[2], LE, dds, bounds[0])             # <<<<<<<<<<<<<<
 *                         bounded_morton_split_dds(clip_pos_r[0], clip_pos_r[1], clip_pos_r[2], LE, dds, bounds[1])
 *                         # We go to the upper bound plus one so that we have *inclusive* loops -- the upper bound
 */
          (void)(__pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_dds((__pyx_v_clip_pos_l[0]), (__pyx_v_clip_pos_l[1]), (__pyx_v_clip_pos_l[2]), __pyx_v_LE, __pyx_v_dds, (__pyx_v_bounds[0])));

          /* "yt/geometry/particle_oct_container.pyx":631
 *                             clip_pos_r[i] = fmin(s_ppos[i] + radius, RE[i] - dds[i]/10)
 *                         bounded_morton_split_dds(clip_pos_l[0], clip_pos_l[1], clip_pos_l[2], LE, dds, bounds[0])
 *                         bounded_morton_split_dds(clip_pos_r[0], clip_pos_r[1], clip_pos_r[2], LE, dds, bounds[1])             # <<<<<<<<<<<<<<
 *                         # We go to the upper bound plus one so that we have *inclusive* loops -- the upper bound
 *                         # is the cell *index*, so we want to make sure we include that cell.  This is also why
 */
          (void)(__pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_dds((__pyx_v_clip_pos_r[0]), (__pyx_v_clip_pos_r[1]), (__pyx_v_clip_pos_r[2]), __pyx_v_LE, __pyx_v_dds, (__pyx_v_bounds[1])));

          /* "yt/geometry/particle_oct_container.pyx":635
 *                         # is the cell *index*, so we want to make sure we include that cell.  This is also why
 *                         # we don't need to worry about mi_max being the max index rather than the cell count.
 *                         for xex in range(bounds[0][0], bounds[1][0] + 1):             # <<<<<<<<<<<<<<
 *                             for yex in range(bounds[0][1], bounds[1][1] + 1):
 *                                 for zex in range(bounds[0][2], bounds[1][2] + 1):
 */
          __pyx_t_14 = (((__pyx_v_bounds[1])[0]) + 1);
          __pyx_t_21 = __pyx_t_14;
          for (__pyx_t_22 = ((__pyx_v_bounds[0])[0]); __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
            __pyx_v_xex = __pyx_t_22;

            /* "yt/geometry/particle_oct_container.pyx":636
 *                         # we don't need to worry about mi_max being the max index rather than the cell count.
 *                         for xex in range(bounds[0][0], bounds[1][0] + 1):
 *                             for yex in range(bounds[0][1], bounds[1][1] + 1):             # <<<<<<<<<<<<<<
 *                                 for zex in range(bounds[0][2], bounds[1][2] + 1):
 *                                     miex = encode_morton_64bit(xex, yex, zex)
 */
            __pyx_t_23 = (((__pyx_v_bounds[1])[1]) + 1);
            __pyx_t_24 = __pyx_t_23;
            for (__pyx_t_25 = ((__pyx_v_bounds[0])[1]); __pyx_t_25 < __pyx_t_24; __pyx_t_25+=1) {
              __pyx_v_yex = __pyx_t_25;

              /* "yt/geometry/particle_oct_container.pyx":637
 *                         for xex in range(bounds[0][0], bounds[1][0] + 1):
 *                             for yex in range(bounds[0][1], bounds[1][1] + 1):
 *                                 for zex in range(bounds[0][2], bounds[1][2] + 1):             # <<<<<<<<<<<<<<
 *                                     miex = encode_morton_64bit(xex, yex, zex)
 *                                     mask[miex] = 1
 */
              __pyx_t_26 = (((__pyx_v_bounds[1])[2]) + 1);
              __pyx_t_27 = __pyx_t_26;
              for (__pyx_t_28 = ((__pyx_v_bounds[0])[2]); __pyx_t_28 < __pyx_t_27; __pyx_t_28+=1) {
                __pyx_v_zex = __pyx_t_28;

                /* "yt/geometry/particle_oct_container.pyx":638
 *                             for yex in range(bounds[0][1], bounds[1][1] + 1):
 *                                 for zex in range(bounds[0][2], bounds[1][2] + 1):
 *                                     miex = encode_morton_64bit(xex, yex, zex)             # <<<<<<<<<<<<<<
 *                                     mask[miex] = 1
 *                                     particle_counts[miex] += 1
 */
                __pyx_v_miex = __pyx_f_2yt_9utilities_3lib_14geometry_utils_encode_morton_64bit(__pyx_v_xex, __pyx_v_yex, __pyx_v_zex);

                /* "yt/geometry/particle_oct_container.pyx":639
 *                                 for zex in range(bounds[0][2], bounds[1][2] + 1):
 *                                     miex = encode_morton_64bit(xex, yex, zex)
 *                                     mask[miex] = 1             # <<<<<<<<<<<<<<
 *                                     particle_counts[miex] += 1
 *                                     if miex >= msize:
 */
                __pyx_t_29 = __pyx_v_miex;
                *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_mask.data + __pyx_t_29 * __pyx_v_mask.strides[0]) )) = 1;

                /* "yt/geometry/particle_oct_container.pyx":640
 *                                     miex = encode_morton_64bit(xex, yex, zex)
 *                                     mask[miex] = 1
 *                                     particle_counts[miex] += 1             # <<<<<<<<<<<<<<
 *                                     if miex >= msize:
 *                                         raise IndexError(
 */
                __pyx_t_29 = __pyx_v_miex;
                *((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_particle_counts.data + __pyx_t_29 * __pyx_v_particle_counts.strides[0]) )) += 1;

                /* "yt/geometry/particle_oct_container.pyx":641
 *                                     mask[miex] = 1
 *                                     particle_counts[miex] += 1
 *                                     if miex >= msize:             # <<<<<<<<<<<<<<
 *                                         raise IndexError(
 *                                             "Index for a softening region "
 */
                __pyx_t_12 = ((__pyx_v_miex >= __pyx_v_msize) != 0);
                if (unlikely(__pyx_t_12)) {

                  /* "yt/geometry/particle_oct_container.pyx":643
 *                                     if miex >= msize:
 *                                         raise IndexError(
 *                                             "Index for a softening region "             # <<<<<<<<<<<<<<
 *                                             f"({miex}) exceeds "
 *                                             f"max ({msize})")
 */
                  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 643, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __pyx_t_15 = 0;
                  __pyx_t_16 = 127;
                  __Pyx_INCREF(__pyx_kp_u_Index_for_a_softening_region);
                  __pyx_t_15 += 30;
                  __Pyx_GIVEREF(__pyx_kp_u_Index_for_a_softening_region);
                  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Index_for_a_softening_region);

                  /* "yt/geometry/particle_oct_container.pyx":644
 *                                         raise IndexError(
 *                                             "Index for a softening region "
 *                                             f"({miex}) exceeds "             # <<<<<<<<<<<<<<
 *                                             f"max ({msize})")
 * 
 */
                  __pyx_t_2 = __Pyx_PyInt_From_npy_uint64(__pyx_v_miex); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 644, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_2);
                  __pyx_t_17 = __Pyx_PyObject_FormatSimple(__pyx_t_2, __pyx_empty_unicode); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 644, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_17);
                  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                  __pyx_t_16 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_17) > __pyx_t_16) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_17) : __pyx_t_16;
                  __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_17);
                  __Pyx_GIVEREF(__pyx_t_17);
                  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_17);
                  __pyx_t_17 = 0;
                  __Pyx_INCREF(__pyx_kp_u_exceeds_max);
                  __pyx_t_15 += 15;
                  __Pyx_GIVEREF(__pyx_kp_u_exceeds_max);
                  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_exceeds_max);

                  /* "yt/geometry/particle_oct_container.pyx":645
 *                                             "Index for a softening region "
 *                                             f"({miex}) exceeds "
 *                                             f"max ({msize})")             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
                  __pyx_t_17 = __Pyx_PyInt_From_npy_uint64(__pyx_v_msize); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 645, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_17);
                  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_17, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 645, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_2);
                  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                  __pyx_t_16 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_16) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_16;
                  __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
                  __Pyx_GIVEREF(__pyx_t_2);
                  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_2);
                  __pyx_t_2 = 0;
                  __Pyx_INCREF(__pyx_kp_u__12);
                  __pyx_t_15 += 1;
                  __Pyx_GIVEREF(__pyx_kp_u__12);
                  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u__12);

                  /* "yt/geometry/particle_oct_container.pyx":643
 *                                     if miex >= msize:
 *                                         raise IndexError(
 *                                             "Index for a softening region "             # <<<<<<<<<<<<<<
 *                                             f"({miex}) exceeds "
 *                                             f"max ({msize})")
 */
                  __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_1, 5, __pyx_t_15, __pyx_t_16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 643, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_2);
                  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

                  /* "yt/geometry/particle_oct_container.pyx":642
 *                                     particle_counts[miex] += 1
 *                                     if miex >= msize:
 *                                         raise IndexError(             # <<<<<<<<<<<<<<
 *                                             "Index for a softening region "
 *                                             f"({miex}) exceeds "
 */
                  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
                  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                  __PYX_ERR(0, 642, __pyx_L1_error)

                  /* "yt/geometry/particle_oct_container.pyx":641
 *                                     mask[miex] = 1
 *                                     particle_counts[miex] += 1
 *                                     if miex >= msize:             # <<<<<<<<<<<<<<
 *                                         raise IndexError(
 *                                             "Index for a softening region "
 */
                }
              }
            }
          }
          __pyx_L26_continue:;
        }
        __pyx_L23_continue:;
      }
      __pyx_L20_continue:;
    }
    __pyx_L5_continue:;
  }

  /* "yt/geometry/particle_oct_container.pyx":549
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void __coarse_index_data_file(self,             # <<<<<<<<<<<<<<
 *                                        np.ndarray[floating, ndim=2] pos,
 *                                        np.ndarray[floating, ndim=1] hsml,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_4, 1);
  __Pyx_XDECREF(__pyx_t_17);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.__coarse_index_data_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos.rcbuffer->pybuffer);
  __pyx_L2:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_mask, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_particle_counts, 1);
  __Pyx_RefNannyFinishContext();
}

static void __pyx_fuse_1__pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___coarse_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyArrayObject *__pyx_v_pos, PyArrayObject *__pyx_v_hsml, __pyx_t_5numpy_uint64_t __pyx_v_file_id) {
  __pyx_t_5numpy_int64_t __pyx_v_i;
  __pyx_t_5numpy_int64_t __pyx_v_p;
  __pyx_t_5numpy_uint64_t __pyx_v_mi;
  __pyx_t_5numpy_uint64_t __pyx_v_miex;
  __pyx_t_5numpy_uint64_t __pyx_v_mi_split[3];
  __pyx_t_5numpy_float64_t __pyx_v_ppos[3];
  __pyx_t_5numpy_float64_t __pyx_v_s_ppos[3];
  __pyx_t_5numpy_float64_t __pyx_v_clip_pos_l[3];
  __pyx_t_5numpy_float64_t __pyx_v_clip_pos_r[3];
  int __pyx_v_skip;
  __pyx_t_5numpy_uint64_t __pyx_v_bounds[2][3];
  __pyx_t_5numpy_uint64_t __pyx_v_xex;
  __pyx_t_5numpy_uint64_t __pyx_v_yex;
  __pyx_t_5numpy_uint64_t __pyx_v_zex;
  __pyx_t_5numpy_float64_t __pyx_v_LE[3];
  __pyx_t_5numpy_float64_t __pyx_v_RE[3];
  __pyx_t_5numpy_float64_t __pyx_v_DW[3];
  __pyx_t_5numpy_uint8_t __pyx_v_PER[3];
  __pyx_t_5numpy_float64_t __pyx_v_dds[3];
  __pyx_t_5numpy_float64_t __pyx_v_radius;
  __Pyx_memviewslice __pyx_v_mask = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_particle_counts = { 0, 0, { 0 }, { 0 }, { 0 } };
  __pyx_t_5numpy_uint64_t __pyx_v_msize;
  int __pyx_v_axiter[3][2];
  __pyx_t_5numpy_float64_t __pyx_v_axiterv[3][2];
  long __pyx_v_xi;
  long __pyx_v_yi;
  long __pyx_v_zi;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_hsml;
  __Pyx_Buffer __pyx_pybuffer_hsml;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_pos;
  __Pyx_Buffer __pyx_pybuffer_pos;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_4 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __pyx_t_5numpy_int64_t __pyx_t_5;
  npy_intp __pyx_t_6;
  npy_intp __pyx_t_7;
  __pyx_t_5numpy_int64_t __pyx_t_8;
  __pyx_t_5numpy_int64_t __pyx_t_9;
  __pyx_t_5numpy_int64_t __pyx_t_10;
  double __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  __pyx_t_5numpy_uint64_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_UCS4 __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  long __pyx_t_18;
  long __pyx_t_19;
  long __pyx_t_20;
  __pyx_t_5numpy_uint64_t __pyx_t_21;
  __pyx_t_5numpy_uint64_t __pyx_t_22;
  __pyx_t_5numpy_uint64_t __pyx_t_23;
  __pyx_t_5numpy_uint64_t __pyx_t_24;
  __pyx_t_5numpy_uint64_t __pyx_t_25;
  __pyx_t_5numpy_uint64_t __pyx_t_26;
  __pyx_t_5numpy_uint64_t __pyx_t_27;
  __pyx_t_5numpy_uint64_t __pyx_t_28;
  __pyx_t_5numpy_uint64_t __pyx_t_29;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_1__coarse_index_data_file", 0);
  __pyx_pybuffer_pos.pybuffer.buf = NULL;
  __pyx_pybuffer_pos.refcount = 0;
  __pyx_pybuffernd_pos.data = NULL;
  __pyx_pybuffernd_pos.rcbuffer = &__pyx_pybuffer_pos;
  __pyx_pybuffer_hsml.pybuffer.buf = NULL;
  __pyx_pybuffer_hsml.refcount = 0;
  __pyx_pybuffernd_hsml.data = NULL;
  __pyx_pybuffernd_hsml.rcbuffer = &__pyx_pybuffer_hsml;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pos.rcbuffer->pybuffer, (PyObject*)__pyx_v_pos, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 549, __pyx_L1_error)
  }
  __pyx_pybuffernd_pos.diminfo[0].strides = __pyx_pybuffernd_pos.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_pos.diminfo[0].shape = __pyx_pybuffernd_pos.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_pos.diminfo[1].strides = __pyx_pybuffernd_pos.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_pos.diminfo[1].shape = __pyx_pybuffernd_pos.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer, (PyObject*)__pyx_v_hsml, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 549, __pyx_L1_error)
  }
  __pyx_pybuffernd_hsml.diminfo[0].strides = __pyx_pybuffernd_hsml.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_hsml.diminfo[0].shape = __pyx_pybuffernd_hsml.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/particle_oct_container.pyx":570
 *         cdef np.float64_t dds[3]
 *         cdef np.float64_t radius
 *         cdef np.uint8_t[:] mask = self.masks[:, file_id]             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t[:] particle_counts = self.particle_counts
 *         cdef np.uint64_t msize = (1 << (self.index_order1 * 3))
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint64(__pyx_v_file_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_slice__11);
  __Pyx_GIVEREF(__pyx_slice__11);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_slice__11);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_self->masks, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint8_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_mask = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":571
 *         cdef np.float64_t radius
 *         cdef np.uint8_t[:] mask = self.masks[:, file_id]
 *         cdef np.uint64_t[:] particle_counts = self.particle_counts             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t msize = (1 << (self.index_order1 * 3))
 *         cdef int axiter[3][2]
 */
  __pyx_t_4 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_v_self->particle_counts, PyBUF_WRITABLE); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(0, 571, __pyx_L1_error)
  __pyx_v_particle_counts = __pyx_t_4;
  __pyx_t_4.memview = NULL;
  __pyx_t_4.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":572
 *         cdef np.uint8_t[:] mask = self.masks[:, file_id]
 *         cdef np.uint64_t[:] particle_counts = self.particle_counts
 *         cdef np.uint64_t msize = (1 << (self.index_order1 * 3))             # <<<<<<<<<<<<<<
 *         cdef int axiter[3][2]
 *         cdef np.float64_t axiterv[3][2]
 */
  __pyx_v_msize = (1 << (__pyx_v_self->index_order1 * 3));

  /* "yt/geometry/particle_oct_container.pyx":576
 *         cdef np.float64_t axiterv[3][2]
 *         # Copy over things for this file (type cast necessary?)
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             LE[i] = self.left_edge[i]
 *             RE[i] = self.right_edge[i]
 */
  for (__pyx_t_5 = 0; __pyx_t_5 < 3; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "yt/geometry/particle_oct_container.pyx":577
 *         # Copy over things for this file (type cast necessary?)
 *         for i in range(3):
 *             LE[i] = self.left_edge[i]             # <<<<<<<<<<<<<<
 *             RE[i] = self.right_edge[i]
 *             PER[i] = self.periodicity[i]
 */
    (__pyx_v_LE[__pyx_v_i]) = (__pyx_v_self->left_edge[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":578
 *         for i in range(3):
 *             LE[i] = self.left_edge[i]
 *             RE[i] = self.right_edge[i]             # <<<<<<<<<<<<<<
 *             PER[i] = self.periodicity[i]
 *             dds[i] = self.dds_mi1[i]
 */
    (__pyx_v_RE[__pyx_v_i]) = (__pyx_v_self->right_edge[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":579
 *             LE[i] = self.left_edge[i]
 *             RE[i] = self.right_edge[i]
 *             PER[i] = self.periodicity[i]             # <<<<<<<<<<<<<<
 *             dds[i] = self.dds_mi1[i]
 *             DW[i] = RE[i] - LE[i]
 */
    (__pyx_v_PER[__pyx_v_i]) = (__pyx_v_self->periodicity[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":580
 *             RE[i] = self.right_edge[i]
 *             PER[i] = self.periodicity[i]
 *             dds[i] = self.dds_mi1[i]             # <<<<<<<<<<<<<<
 *             DW[i] = RE[i] - LE[i]
 *             axiter[i][0] = 0 # We always do an offset of 0
 */
    (__pyx_v_dds[__pyx_v_i]) = (__pyx_v_self->dds_mi1[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":581
 *             PER[i] = self.periodicity[i]
 *             dds[i] = self.dds_mi1[i]
 *             DW[i] = RE[i] - LE[i]             # <<<<<<<<<<<<<<
 *             axiter[i][0] = 0 # We always do an offset of 0
 *             axiterv[i][0] = 0.0
 */
    (__pyx_v_DW[__pyx_v_i]) = ((__pyx_v_RE[__pyx_v_i]) - (__pyx_v_LE[__pyx_v_i]));

    /* "yt/geometry/particle_oct_container.pyx":582
 *             dds[i] = self.dds_mi1[i]
 *             DW[i] = RE[i] - LE[i]
 *             axiter[i][0] = 0 # We always do an offset of 0             # <<<<<<<<<<<<<<
 *             axiterv[i][0] = 0.0
 *         # Mark index of particles that are in this file
 */
    ((__pyx_v_axiter[__pyx_v_i])[0]) = 0;

    /* "yt/geometry/particle_oct_container.pyx":583
 *             DW[i] = RE[i] - LE[i]
 *             axiter[i][0] = 0 # We always do an offset of 0
 *             axiterv[i][0] = 0.0             # <<<<<<<<<<<<<<
 *         # Mark index of particles that are in this file
 *         for p in range(pos.shape[0]):
 */
    ((__pyx_v_axiterv[__pyx_v_i])[0]) = 0.0;
  }

  /* "yt/geometry/particle_oct_container.pyx":585
 *             axiterv[i][0] = 0.0
 *         # Mark index of particles that are in this file
 *         for p in range(pos.shape[0]):             # <<<<<<<<<<<<<<
 *             skip = 0
 *             for i in range(3):
 */
  __pyx_t_6 = (__pyx_v_pos->dimensions[0]);
  __pyx_t_7 = __pyx_t_6;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_7; __pyx_t_5+=1) {
    __pyx_v_p = __pyx_t_5;

    /* "yt/geometry/particle_oct_container.pyx":586
 *         # Mark index of particles that are in this file
 *         for p in range(pos.shape[0]):
 *             skip = 0             # <<<<<<<<<<<<<<
 *             for i in range(3):
 *                 axiter[i][1] = 999
 */
    __pyx_v_skip = 0;

    /* "yt/geometry/particle_oct_container.pyx":587
 *         for p in range(pos.shape[0]):
 *             skip = 0
 *             for i in range(3):             # <<<<<<<<<<<<<<
 *                 axiter[i][1] = 999
 *                 # Skip particles outside the domain
 */
    for (__pyx_t_8 = 0; __pyx_t_8 < 3; __pyx_t_8+=1) {
      __pyx_v_i = __pyx_t_8;

      /* "yt/geometry/particle_oct_container.pyx":588
 *             skip = 0
 *             for i in range(3):
 *                 axiter[i][1] = 999             # <<<<<<<<<<<<<<
 *                 # Skip particles outside the domain
 *                 if not (LE[i] <= pos[p, i] < RE[i]):
 */
      ((__pyx_v_axiter[__pyx_v_i])[1]) = 0x3E7;

      /* "yt/geometry/particle_oct_container.pyx":590
 *                 axiter[i][1] = 999
 *                 # Skip particles outside the domain
 *                 if not (LE[i] <= pos[p, i] < RE[i]):             # <<<<<<<<<<<<<<
 *                     skip = 1
 *                     break
 */
      __pyx_t_9 = __pyx_v_p;
      __pyx_t_10 = __pyx_v_i;
      __pyx_t_11 = (*__Pyx_BufPtrStrided2d(double *, __pyx_pybuffernd_pos.rcbuffer->pybuffer.buf, __pyx_t_9, __pyx_pybuffernd_pos.diminfo[0].strides, __pyx_t_10, __pyx_pybuffernd_pos.diminfo[1].strides));
      __pyx_t_12 = ((__pyx_v_LE[__pyx_v_i]) <= __pyx_t_11);
      if (__pyx_t_12) {
        __pyx_t_12 = (__pyx_t_11 < (__pyx_v_RE[__pyx_v_i]));
      }
      __pyx_t_13 = ((!(__pyx_t_12 != 0)) != 0);
      if (__pyx_t_13) {

        /* "yt/geometry/particle_oct_container.pyx":591
 *                 # Skip particles outside the domain
 *                 if not (LE[i] <= pos[p, i] < RE[i]):
 *                     skip = 1             # <<<<<<<<<<<<<<
 *                     break
 *                 ppos[i] = pos[p,i]
 */
        __pyx_v_skip = 1;

        /* "yt/geometry/particle_oct_container.pyx":592
 *                 if not (LE[i] <= pos[p, i] < RE[i]):
 *                     skip = 1
 *                     break             # <<<<<<<<<<<<<<
 *                 ppos[i] = pos[p,i]
 *             if skip == 1: continue
 */
        goto __pyx_L8_break;

        /* "yt/geometry/particle_oct_container.pyx":590
 *                 axiter[i][1] = 999
 *                 # Skip particles outside the domain
 *                 if not (LE[i] <= pos[p, i] < RE[i]):             # <<<<<<<<<<<<<<
 *                     skip = 1
 *                     break
 */
      }

      /* "yt/geometry/particle_oct_container.pyx":593
 *                     skip = 1
 *                     break
 *                 ppos[i] = pos[p,i]             # <<<<<<<<<<<<<<
 *             if skip == 1: continue
 *             mi = bounded_morton_split_dds(ppos[0], ppos[1], ppos[2], LE,
 */
      __pyx_t_10 = __pyx_v_p;
      __pyx_t_9 = __pyx_v_i;
      (__pyx_v_ppos[__pyx_v_i]) = (*__Pyx_BufPtrStrided2d(double *, __pyx_pybuffernd_pos.rcbuffer->pybuffer.buf, __pyx_t_10, __pyx_pybuffernd_pos.diminfo[0].strides, __pyx_t_9, __pyx_pybuffernd_pos.diminfo[1].strides));
    }
    __pyx_L8_break:;

    /* "yt/geometry/particle_oct_container.pyx":594
 *                     break
 *                 ppos[i] = pos[p,i]
 *             if skip == 1: continue             # <<<<<<<<<<<<<<
 *             mi = bounded_morton_split_dds(ppos[0], ppos[1], ppos[2], LE,
 *                                           dds, mi_split)
 */
    __pyx_t_13 = ((__pyx_v_skip == 1) != 0);
    if (__pyx_t_13) {
      goto __pyx_L5_continue;
    }

    /* "yt/geometry/particle_oct_container.pyx":595
 *                 ppos[i] = pos[p,i]
 *             if skip == 1: continue
 *             mi = bounded_morton_split_dds(ppos[0], ppos[1], ppos[2], LE,             # <<<<<<<<<<<<<<
 *                                           dds, mi_split)
 *             mask[mi] = 1
 */
    __pyx_v_mi = __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_dds((__pyx_v_ppos[0]), (__pyx_v_ppos[1]), (__pyx_v_ppos[2]), __pyx_v_LE, __pyx_v_dds, __pyx_v_mi_split);

    /* "yt/geometry/particle_oct_container.pyx":597
 *             mi = bounded_morton_split_dds(ppos[0], ppos[1], ppos[2], LE,
 *                                           dds, mi_split)
 *             mask[mi] = 1             # <<<<<<<<<<<<<<
 *             particle_counts[mi] += 1
 *             # Expand mask by softening
 */
    __pyx_t_14 = __pyx_v_mi;
    *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_mask.data + __pyx_t_14 * __pyx_v_mask.strides[0]) )) = 1;

    /* "yt/geometry/particle_oct_container.pyx":598
 *                                           dds, mi_split)
 *             mask[mi] = 1
 *             particle_counts[mi] += 1             # <<<<<<<<<<<<<<
 *             # Expand mask by softening
 *             if hsml is None:
 */
    __pyx_t_14 = __pyx_v_mi;
    *((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_particle_counts.data + __pyx_t_14 * __pyx_v_particle_counts.strides[0]) )) += 1;

    /* "yt/geometry/particle_oct_container.pyx":600
 *             particle_counts[mi] += 1
 *             # Expand mask by softening
 *             if hsml is None:             # <<<<<<<<<<<<<<
 *                 continue
 *             if hsml[p] < 0:
 */
    __pyx_t_13 = (((PyObject *)__pyx_v_hsml) == Py_None);
    __pyx_t_12 = (__pyx_t_13 != 0);
    if (__pyx_t_12) {

      /* "yt/geometry/particle_oct_container.pyx":601
 *             # Expand mask by softening
 *             if hsml is None:
 *                 continue             # <<<<<<<<<<<<<<
 *             if hsml[p] < 0:
 *                 raise RuntimeError(
 */
      goto __pyx_L5_continue;

      /* "yt/geometry/particle_oct_container.pyx":600
 *             particle_counts[mi] += 1
 *             # Expand mask by softening
 *             if hsml is None:             # <<<<<<<<<<<<<<
 *                 continue
 *             if hsml[p] < 0:
 */
    }

    /* "yt/geometry/particle_oct_container.pyx":602
 *             if hsml is None:
 *                 continue
 *             if hsml[p] < 0:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     f"Smoothing length for particle {p} is negative with "
 */
    __pyx_t_8 = __pyx_v_p;
    __pyx_t_12 = (((*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_hsml.rcbuffer->pybuffer.buf, __pyx_t_8, __pyx_pybuffernd_hsml.diminfo[0].strides)) < 0.0) != 0);
    if (unlikely(__pyx_t_12)) {

      /* "yt/geometry/particle_oct_container.pyx":604
 *             if hsml[p] < 0:
 *                 raise RuntimeError(
 *                     f"Smoothing length for particle {p} is negative with "             # <<<<<<<<<<<<<<
 *                     f"value {hsml[p]}")
 *             radius = hsml[p]
 */
      __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 604, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_15 = 0;
      __pyx_t_16 = 127;
      __Pyx_INCREF(__pyx_kp_u_Smoothing_length_for_particle);
      __pyx_t_15 += 30;
      __Pyx_GIVEREF(__pyx_kp_u_Smoothing_length_for_particle);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Smoothing_length_for_particle);
      __pyx_t_2 = __Pyx_PyInt_From_npy_int64(__pyx_v_p); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 604, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_17 = __Pyx_PyObject_FormatSimple(__pyx_t_2, __pyx_empty_unicode); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 604, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_16 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_17) > __pyx_t_16) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_17) : __pyx_t_16;
      __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_17);
      __Pyx_GIVEREF(__pyx_t_17);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_17);
      __pyx_t_17 = 0;
      __Pyx_INCREF(__pyx_kp_u_is_negative_with_value);
      __pyx_t_15 += 24;
      __Pyx_GIVEREF(__pyx_kp_u_is_negative_with_value);
      PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_is_negative_with_value);

      /* "yt/geometry/particle_oct_container.pyx":605
 *                 raise RuntimeError(
 *                     f"Smoothing length for particle {p} is negative with "
 *                     f"value {hsml[p]}")             # <<<<<<<<<<<<<<
 *             radius = hsml[p]
 *             # We first check if we're bounded within the domain; this follows the logic in the
 */
      __pyx_t_8 = __pyx_v_p;
      __pyx_t_17 = PyFloat_FromDouble((*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_hsml.rcbuffer->pybuffer.buf, __pyx_t_8, __pyx_pybuffernd_hsml.diminfo[0].strides))); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 605, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_17);
      __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_17, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 605, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      __pyx_t_16 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_16) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_16;
      __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "yt/geometry/particle_oct_container.pyx":604
 *             if hsml[p] < 0:
 *                 raise RuntimeError(
 *                     f"Smoothing length for particle {p} is negative with "             # <<<<<<<<<<<<<<
 *                     f"value {hsml[p]}")
 *             radius = hsml[p]
 */
      __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_1, 4, __pyx_t_15, __pyx_t_16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 604, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "yt/geometry/particle_oct_container.pyx":603
 *                 continue
 *             if hsml[p] < 0:
 *                 raise RuntimeError(             # <<<<<<<<<<<<<<
 *                     f"Smoothing length for particle {p} is negative with "
 *                     f"value {hsml[p]}")
 */
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 603, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 603, __pyx_L1_error)

      /* "yt/geometry/particle_oct_container.pyx":602
 *             if hsml is None:
 *                 continue
 *             if hsml[p] < 0:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(
 *                     f"Smoothing length for particle {p} is negative with "
 */
    }

    /* "yt/geometry/particle_oct_container.pyx":606
 *                     f"Smoothing length for particle {p} is negative with "
 *                     f"value {hsml[p]}")
 *             radius = hsml[p]             # <<<<<<<<<<<<<<
 *             # We first check if we're bounded within the domain; this follows the logic in the
 *             # pixelize_cartesian routine.  We assume that no smoothing
 */
    __pyx_t_8 = __pyx_v_p;
    __pyx_v_radius = (*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_hsml.rcbuffer->pybuffer.buf, __pyx_t_8, __pyx_pybuffernd_hsml.diminfo[0].strides));

    /* "yt/geometry/particle_oct_container.pyx":610
 *             # pixelize_cartesian routine.  We assume that no smoothing
 *             # length can wrap around both directions.
 *             for i in range(3):             # <<<<<<<<<<<<<<
 *                 if PER[i] and ppos[i] - radius < LE[i]:
 *                     axiter[i][1] = +1
 */
    for (__pyx_t_8 = 0; __pyx_t_8 < 3; __pyx_t_8+=1) {
      __pyx_v_i = __pyx_t_8;

      /* "yt/geometry/particle_oct_container.pyx":611
 *             # length can wrap around both directions.
 *             for i in range(3):
 *                 if PER[i] and ppos[i] - radius < LE[i]:             # <<<<<<<<<<<<<<
 *                     axiter[i][1] = +1
 *                     axiterv[i][1] = DW[i]
 */
      __pyx_t_13 = ((__pyx_v_PER[__pyx_v_i]) != 0);
      if (__pyx_t_13) {
      } else {
        __pyx_t_12 = __pyx_t_13;
        goto __pyx_L16_bool_binop_done;
      }
      __pyx_t_13 = ((((__pyx_v_ppos[__pyx_v_i]) - __pyx_v_radius) < (__pyx_v_LE[__pyx_v_i])) != 0);
      __pyx_t_12 = __pyx_t_13;
      __pyx_L16_bool_binop_done:;
      if (__pyx_t_12) {

        /* "yt/geometry/particle_oct_container.pyx":612
 *             for i in range(3):
 *                 if PER[i] and ppos[i] - radius < LE[i]:
 *                     axiter[i][1] = +1             # <<<<<<<<<<<<<<
 *                     axiterv[i][1] = DW[i]
 *                 elif PER[i] and ppos[i] + radius > RE[i]:
 */
        ((__pyx_v_axiter[__pyx_v_i])[1]) = 1;

        /* "yt/geometry/particle_oct_container.pyx":613
 *                 if PER[i] and ppos[i] - radius < LE[i]:
 *                     axiter[i][1] = +1
 *                     axiterv[i][1] = DW[i]             # <<<<<<<<<<<<<<
 *                 elif PER[i] and ppos[i] + radius > RE[i]:
 *                     axiter[i][1] = -1
 */
        ((__pyx_v_axiterv[__pyx_v_i])[1]) = (__pyx_v_DW[__pyx_v_i]);

        /* "yt/geometry/particle_oct_container.pyx":611
 *             # length can wrap around both directions.
 *             for i in range(3):
 *                 if PER[i] and ppos[i] - radius < LE[i]:             # <<<<<<<<<<<<<<
 *                     axiter[i][1] = +1
 *                     axiterv[i][1] = DW[i]
 */
        goto __pyx_L15;
      }

      /* "yt/geometry/particle_oct_container.pyx":614
 *                     axiter[i][1] = +1
 *                     axiterv[i][1] = DW[i]
 *                 elif PER[i] and ppos[i] + radius > RE[i]:             # <<<<<<<<<<<<<<
 *                     axiter[i][1] = -1
 *                     axiterv[i][1] = -DW[i]
 */
      __pyx_t_13 = ((__pyx_v_PER[__pyx_v_i]) != 0);
      if (__pyx_t_13) {
      } else {
        __pyx_t_12 = __pyx_t_13;
        goto __pyx_L18_bool_binop_done;
      }
      __pyx_t_13 = ((((__pyx_v_ppos[__pyx_v_i]) + __pyx_v_radius) > (__pyx_v_RE[__pyx_v_i])) != 0);
      __pyx_t_12 = __pyx_t_13;
      __pyx_L18_bool_binop_done:;
      if (__pyx_t_12) {

        /* "yt/geometry/particle_oct_container.pyx":615
 *                     axiterv[i][1] = DW[i]
 *                 elif PER[i] and ppos[i] + radius > RE[i]:
 *                     axiter[i][1] = -1             # <<<<<<<<<<<<<<
 *                     axiterv[i][1] = -DW[i]
 *             for xi in range(2):
 */
        ((__pyx_v_axiter[__pyx_v_i])[1]) = -1;

        /* "yt/geometry/particle_oct_container.pyx":616
 *                 elif PER[i] and ppos[i] + radius > RE[i]:
 *                     axiter[i][1] = -1
 *                     axiterv[i][1] = -DW[i]             # <<<<<<<<<<<<<<
 *             for xi in range(2):
 *                 if axiter[0][xi] == 999: continue
 */
        ((__pyx_v_axiterv[__pyx_v_i])[1]) = (-(__pyx_v_DW[__pyx_v_i]));

        /* "yt/geometry/particle_oct_container.pyx":614
 *                     axiter[i][1] = +1
 *                     axiterv[i][1] = DW[i]
 *                 elif PER[i] and ppos[i] + radius > RE[i]:             # <<<<<<<<<<<<<<
 *                     axiter[i][1] = -1
 *                     axiterv[i][1] = -DW[i]
 */
      }
      __pyx_L15:;
    }

    /* "yt/geometry/particle_oct_container.pyx":617
 *                     axiter[i][1] = -1
 *                     axiterv[i][1] = -DW[i]
 *             for xi in range(2):             # <<<<<<<<<<<<<<
 *                 if axiter[0][xi] == 999: continue
 *                 s_ppos[0] = ppos[0] + axiterv[0][xi]
 */
    for (__pyx_t_18 = 0; __pyx_t_18 < 2; __pyx_t_18+=1) {
      __pyx_v_xi = __pyx_t_18;

      /* "yt/geometry/particle_oct_container.pyx":618
 *                     axiterv[i][1] = -DW[i]
 *             for xi in range(2):
 *                 if axiter[0][xi] == 999: continue             # <<<<<<<<<<<<<<
 *                 s_ppos[0] = ppos[0] + axiterv[0][xi]
 *                 for yi in range(2):
 */
      __pyx_t_12 = ((((__pyx_v_axiter[0])[__pyx_v_xi]) == 0x3E7) != 0);
      if (__pyx_t_12) {
        goto __pyx_L20_continue;
      }

      /* "yt/geometry/particle_oct_container.pyx":619
 *             for xi in range(2):
 *                 if axiter[0][xi] == 999: continue
 *                 s_ppos[0] = ppos[0] + axiterv[0][xi]             # <<<<<<<<<<<<<<
 *                 for yi in range(2):
 *                     if axiter[1][yi] == 999: continue
 */
      (__pyx_v_s_ppos[0]) = ((__pyx_v_ppos[0]) + ((__pyx_v_axiterv[0])[__pyx_v_xi]));

      /* "yt/geometry/particle_oct_container.pyx":620
 *                 if axiter[0][xi] == 999: continue
 *                 s_ppos[0] = ppos[0] + axiterv[0][xi]
 *                 for yi in range(2):             # <<<<<<<<<<<<<<
 *                     if axiter[1][yi] == 999: continue
 *                     s_ppos[1] = ppos[1] + axiterv[1][yi]
 */
      for (__pyx_t_19 = 0; __pyx_t_19 < 2; __pyx_t_19+=1) {
        __pyx_v_yi = __pyx_t_19;

        /* "yt/geometry/particle_oct_container.pyx":621
 *                 s_ppos[0] = ppos[0] + axiterv[0][xi]
 *                 for yi in range(2):
 *                     if axiter[1][yi] == 999: continue             # <<<<<<<<<<<<<<
 *                     s_ppos[1] = ppos[1] + axiterv[1][yi]
 *                     for zi in range(2):
 */
        __pyx_t_12 = ((((__pyx_v_axiter[1])[__pyx_v_yi]) == 0x3E7) != 0);
        if (__pyx_t_12) {
          goto __pyx_L23_continue;
        }

        /* "yt/geometry/particle_oct_container.pyx":622
 *                 for yi in range(2):
 *                     if axiter[1][yi] == 999: continue
 *                     s_ppos[1] = ppos[1] + axiterv[1][yi]             # <<<<<<<<<<<<<<
 *                     for zi in range(2):
 *                         if axiter[2][zi] == 999: continue
 */
        (__pyx_v_s_ppos[1]) = ((__pyx_v_ppos[1]) + ((__pyx_v_axiterv[1])[__pyx_v_yi]));

        /* "yt/geometry/particle_oct_container.pyx":623
 *                     if axiter[1][yi] == 999: continue
 *                     s_ppos[1] = ppos[1] + axiterv[1][yi]
 *                     for zi in range(2):             # <<<<<<<<<<<<<<
 *                         if axiter[2][zi] == 999: continue
 *                         s_ppos[2] = ppos[2] + axiterv[2][zi]
 */
        for (__pyx_t_20 = 0; __pyx_t_20 < 2; __pyx_t_20+=1) {
          __pyx_v_zi = __pyx_t_20;

          /* "yt/geometry/particle_oct_container.pyx":624
 *                     s_ppos[1] = ppos[1] + axiterv[1][yi]
 *                     for zi in range(2):
 *                         if axiter[2][zi] == 999: continue             # <<<<<<<<<<<<<<
 *                         s_ppos[2] = ppos[2] + axiterv[2][zi]
 *                         # OK, now we compute the left and right edges for this shift.
 */
          __pyx_t_12 = ((((__pyx_v_axiter[2])[__pyx_v_zi]) == 0x3E7) != 0);
          if (__pyx_t_12) {
            goto __pyx_L26_continue;
          }

          /* "yt/geometry/particle_oct_container.pyx":625
 *                     for zi in range(2):
 *                         if axiter[2][zi] == 999: continue
 *                         s_ppos[2] = ppos[2] + axiterv[2][zi]             # <<<<<<<<<<<<<<
 *                         # OK, now we compute the left and right edges for this shift.
 *                         for i in range(3):
 */
          (__pyx_v_s_ppos[2]) = ((__pyx_v_ppos[2]) + ((__pyx_v_axiterv[2])[__pyx_v_zi]));

          /* "yt/geometry/particle_oct_container.pyx":627
 *                         s_ppos[2] = ppos[2] + axiterv[2][zi]
 *                         # OK, now we compute the left and right edges for this shift.
 *                         for i in range(3):             # <<<<<<<<<<<<<<
 *                             clip_pos_l[i] = fmax(s_ppos[i] - radius, LE[i] + dds[i]/10)
 *                             clip_pos_r[i] = fmin(s_ppos[i] + radius, RE[i] - dds[i]/10)
 */
          for (__pyx_t_8 = 0; __pyx_t_8 < 3; __pyx_t_8+=1) {
            __pyx_v_i = __pyx_t_8;

            /* "yt/geometry/particle_oct_container.pyx":628
 *                         # OK, now we compute the left and right edges for this shift.
 *                         for i in range(3):
 *                             clip_pos_l[i] = fmax(s_ppos[i] - radius, LE[i] + dds[i]/10)             # <<<<<<<<<<<<<<
 *                             clip_pos_r[i] = fmin(s_ppos[i] + radius, RE[i] - dds[i]/10)
 *                         bounded_morton_split_dds(clip_pos_l[0], clip_pos_l[1], clip_pos_l[2], LE, dds, bounds[0])
 */
            (__pyx_v_clip_pos_l[__pyx_v_i]) = __pyx_f_2yt_9utilities_3lib_8fp_utils_fmax(((__pyx_v_s_ppos[__pyx_v_i]) - __pyx_v_radius), ((__pyx_v_LE[__pyx_v_i]) + ((__pyx_v_dds[__pyx_v_i]) / 10.0)));

            /* "yt/geometry/particle_oct_container.pyx":629
 *                         for i in range(3):
 *                             clip_pos_l[i] = fmax(s_ppos[i] - radius, LE[i] + dds[i]/10)
 *                             clip_pos_r[i] = fmin(s_ppos[i] + radius, RE[i] - dds[i]/10)             # <<<<<<<<<<<<<<
 *                         bounded_morton_split_dds(clip_pos_l[0], clip_pos_l[1], clip_pos_l[2], LE, dds, bounds[0])
 *                         bounded_morton_split_dds(clip_pos_r[0], clip_pos_r[1], clip_pos_r[2], LE, dds, bounds[1])
 */
            (__pyx_v_clip_pos_r[__pyx_v_i]) = __pyx_f_2yt_9utilities_3lib_8fp_utils_fmin(((__pyx_v_s_ppos[__pyx_v_i]) + __pyx_v_radius), ((__pyx_v_RE[__pyx_v_i]) - ((__pyx_v_dds[__pyx_v_i]) / 10.0)));
          }

          /* "yt/geometry/particle_oct_container.pyx":630
 *                             clip_pos_l[i] = fmax(s_ppos[i] - radius, LE[i] + dds[i]/10)
 *                             clip_pos_r[i] = fmin(s_ppos[i] + radius, RE[i] - dds[i]/10)
 *                         bounded_morton_split_dds(clip_pos_l[0], clip_pos_l[1], clip_pos_l[2], LE, dds, bounds[0])             # <<<<<<<<<<<<<<
 *                         bounded_morton_split_dds(clip_pos_r[0], clip_pos_r[1], clip_pos_r[2], LE, dds, bounds[1])
 *                         # We go to the upper bound plus one so that we have *inclusive* loops -- the upper bound
 */
          (void)(__pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_dds((__pyx_v_clip_pos_l[0]), (__pyx_v_clip_pos_l[1]), (__pyx_v_clip_pos_l[2]), __pyx_v_LE, __pyx_v_dds, (__pyx_v_bounds[0])));

          /* "yt/geometry/particle_oct_container.pyx":631
 *                             clip_pos_r[i] = fmin(s_ppos[i] + radius, RE[i] - dds[i]/10)
 *                         bounded_morton_split_dds(clip_pos_l[0], clip_pos_l[1], clip_pos_l[2], LE, dds, bounds[0])
 *                         bounded_morton_split_dds(clip_pos_r[0], clip_pos_r[1], clip_pos_r[2], LE, dds, bounds[1])             # <<<<<<<<<<<<<<
 *                         # We go to the upper bound plus one so that we have *inclusive* loops -- the upper bound
 *                         # is the cell *index*, so we want to make sure we include that cell.  This is also why
 */
          (void)(__pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_dds((__pyx_v_clip_pos_r[0]), (__pyx_v_clip_pos_r[1]), (__pyx_v_clip_pos_r[2]), __pyx_v_LE, __pyx_v_dds, (__pyx_v_bounds[1])));

          /* "yt/geometry/particle_oct_container.pyx":635
 *                         # is the cell *index*, so we want to make sure we include that cell.  This is also why
 *                         # we don't need to worry about mi_max being the max index rather than the cell count.
 *                         for xex in range(bounds[0][0], bounds[1][0] + 1):             # <<<<<<<<<<<<<<
 *                             for yex in range(bounds[0][1], bounds[1][1] + 1):
 *                                 for zex in range(bounds[0][2], bounds[1][2] + 1):
 */
          __pyx_t_14 = (((__pyx_v_bounds[1])[0]) + 1);
          __pyx_t_21 = __pyx_t_14;
          for (__pyx_t_22 = ((__pyx_v_bounds[0])[0]); __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
            __pyx_v_xex = __pyx_t_22;

            /* "yt/geometry/particle_oct_container.pyx":636
 *                         # we don't need to worry about mi_max being the max index rather than the cell count.
 *                         for xex in range(bounds[0][0], bounds[1][0] + 1):
 *                             for yex in range(bounds[0][1], bounds[1][1] + 1):             # <<<<<<<<<<<<<<
 *                                 for zex in range(bounds[0][2], bounds[1][2] + 1):
 *                                     miex = encode_morton_64bit(xex, yex, zex)
 */
            __pyx_t_23 = (((__pyx_v_bounds[1])[1]) + 1);
            __pyx_t_24 = __pyx_t_23;
            for (__pyx_t_25 = ((__pyx_v_bounds[0])[1]); __pyx_t_25 < __pyx_t_24; __pyx_t_25+=1) {
              __pyx_v_yex = __pyx_t_25;

              /* "yt/geometry/particle_oct_container.pyx":637
 *                         for xex in range(bounds[0][0], bounds[1][0] + 1):
 *                             for yex in range(bounds[0][1], bounds[1][1] + 1):
 *                                 for zex in range(bounds[0][2], bounds[1][2] + 1):             # <<<<<<<<<<<<<<
 *                                     miex = encode_morton_64bit(xex, yex, zex)
 *                                     mask[miex] = 1
 */
              __pyx_t_26 = (((__pyx_v_bounds[1])[2]) + 1);
              __pyx_t_27 = __pyx_t_26;
              for (__pyx_t_28 = ((__pyx_v_bounds[0])[2]); __pyx_t_28 < __pyx_t_27; __pyx_t_28+=1) {
                __pyx_v_zex = __pyx_t_28;

                /* "yt/geometry/particle_oct_container.pyx":638
 *                             for yex in range(bounds[0][1], bounds[1][1] + 1):
 *                                 for zex in range(bounds[0][2], bounds[1][2] + 1):
 *                                     miex = encode_morton_64bit(xex, yex, zex)             # <<<<<<<<<<<<<<
 *                                     mask[miex] = 1
 *                                     particle_counts[miex] += 1
 */
                __pyx_v_miex = __pyx_f_2yt_9utilities_3lib_14geometry_utils_encode_morton_64bit(__pyx_v_xex, __pyx_v_yex, __pyx_v_zex);

                /* "yt/geometry/particle_oct_container.pyx":639
 *                                 for zex in range(bounds[0][2], bounds[1][2] + 1):
 *                                     miex = encode_morton_64bit(xex, yex, zex)
 *                                     mask[miex] = 1             # <<<<<<<<<<<<<<
 *                                     particle_counts[miex] += 1
 *                                     if miex >= msize:
 */
                __pyx_t_29 = __pyx_v_miex;
                *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_mask.data + __pyx_t_29 * __pyx_v_mask.strides[0]) )) = 1;

                /* "yt/geometry/particle_oct_container.pyx":640
 *                                     miex = encode_morton_64bit(xex, yex, zex)
 *                                     mask[miex] = 1
 *                                     particle_counts[miex] += 1             # <<<<<<<<<<<<<<
 *                                     if miex >= msize:
 *                                         raise IndexError(
 */
                __pyx_t_29 = __pyx_v_miex;
                *((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_particle_counts.data + __pyx_t_29 * __pyx_v_particle_counts.strides[0]) )) += 1;

                /* "yt/geometry/particle_oct_container.pyx":641
 *                                     mask[miex] = 1
 *                                     particle_counts[miex] += 1
 *                                     if miex >= msize:             # <<<<<<<<<<<<<<
 *                                         raise IndexError(
 *                                             "Index for a softening region "
 */
                __pyx_t_12 = ((__pyx_v_miex >= __pyx_v_msize) != 0);
                if (unlikely(__pyx_t_12)) {

                  /* "yt/geometry/particle_oct_container.pyx":643
 *                                     if miex >= msize:
 *                                         raise IndexError(
 *                                             "Index for a softening region "             # <<<<<<<<<<<<<<
 *                                             f"({miex}) exceeds "
 *                                             f"max ({msize})")
 */
                  __pyx_t_1 = PyTuple_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 643, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __pyx_t_15 = 0;
                  __pyx_t_16 = 127;
                  __Pyx_INCREF(__pyx_kp_u_Index_for_a_softening_region);
                  __pyx_t_15 += 30;
                  __Pyx_GIVEREF(__pyx_kp_u_Index_for_a_softening_region);
                  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_Index_for_a_softening_region);

                  /* "yt/geometry/particle_oct_container.pyx":644
 *                                         raise IndexError(
 *                                             "Index for a softening region "
 *                                             f"({miex}) exceeds "             # <<<<<<<<<<<<<<
 *                                             f"max ({msize})")
 * 
 */
                  __pyx_t_2 = __Pyx_PyInt_From_npy_uint64(__pyx_v_miex); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 644, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_2);
                  __pyx_t_17 = __Pyx_PyObject_FormatSimple(__pyx_t_2, __pyx_empty_unicode); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 644, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_17);
                  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                  __pyx_t_16 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_17) > __pyx_t_16) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_17) : __pyx_t_16;
                  __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_17);
                  __Pyx_GIVEREF(__pyx_t_17);
                  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_17);
                  __pyx_t_17 = 0;
                  __Pyx_INCREF(__pyx_kp_u_exceeds_max);
                  __pyx_t_15 += 15;
                  __Pyx_GIVEREF(__pyx_kp_u_exceeds_max);
                  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_exceeds_max);

                  /* "yt/geometry/particle_oct_container.pyx":645
 *                                             "Index for a softening region "
 *                                             f"({miex}) exceeds "
 *                                             f"max ({msize})")             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
                  __pyx_t_17 = __Pyx_PyInt_From_npy_uint64(__pyx_v_msize); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 645, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_17);
                  __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_17, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 645, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_2);
                  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                  __pyx_t_16 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_16) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_16;
                  __pyx_t_15 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
                  __Pyx_GIVEREF(__pyx_t_2);
                  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_2);
                  __pyx_t_2 = 0;
                  __Pyx_INCREF(__pyx_kp_u__12);
                  __pyx_t_15 += 1;
                  __Pyx_GIVEREF(__pyx_kp_u__12);
                  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u__12);

                  /* "yt/geometry/particle_oct_container.pyx":643
 *                                     if miex >= msize:
 *                                         raise IndexError(
 *                                             "Index for a softening region "             # <<<<<<<<<<<<<<
 *                                             f"({miex}) exceeds "
 *                                             f"max ({msize})")
 */
                  __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_1, 5, __pyx_t_15, __pyx_t_16); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 643, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_2);
                  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

                  /* "yt/geometry/particle_oct_container.pyx":642
 *                                     particle_counts[miex] += 1
 *                                     if miex >= msize:
 *                                         raise IndexError(             # <<<<<<<<<<<<<<
 *                                             "Index for a softening region "
 *                                             f"({miex}) exceeds "
 */
                  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
                  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                  __PYX_ERR(0, 642, __pyx_L1_error)

                  /* "yt/geometry/particle_oct_container.pyx":641
 *                                     mask[miex] = 1
 *                                     particle_counts[miex] += 1
 *                                     if miex >= msize:             # <<<<<<<<<<<<<<
 *                                         raise IndexError(
 *                                             "Index for a softening region "
 */
                }
              }
            }
          }
          __pyx_L26_continue:;
        }
        __pyx_L23_continue:;
      }
      __pyx_L20_continue:;
    }
    __pyx_L5_continue:;
  }

  /* "yt/geometry/particle_oct_container.pyx":549
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void __coarse_index_data_file(self,             # <<<<<<<<<<<<<<
 *                                        np.ndarray[floating, ndim=2] pos,
 *                                        np.ndarray[floating, ndim=1] hsml,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_4, 1);
  __Pyx_XDECREF(__pyx_t_17);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.__coarse_index_data_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos.rcbuffer->pybuffer);
  __pyx_L2:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_mask, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_particle_counts, 1);
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":650
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def _set_coarse_index_data_file(self, np.uint64_t file_id):             # <<<<<<<<<<<<<<
 *         return self.__set_coarse_index_data_file(file_id)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_11_set_coarse_index_data_file(PyObject *__pyx_v_self, PyObject *__pyx_arg_file_id); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_11_set_coarse_index_data_file(PyObject *__pyx_v_self, PyObject *__pyx_arg_file_id) {
  __pyx_t_5numpy_uint64_t __pyx_v_file_id;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set_coarse_index_data_file (wrapper)", 0);
  assert(__pyx_arg_file_id); {
    __pyx_v_file_id = __Pyx_PyInt_As_npy_uint64(__pyx_arg_file_id); if (unlikely((__pyx_v_file_id == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 650, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap._set_coarse_index_data_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10_set_coarse_index_data_file(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((__pyx_t_5numpy_uint64_t)__pyx_v_file_id));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10_set_coarse_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_file_id) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_coarse_index_data_file", 0);

  /* "yt/geometry/particle_oct_container.pyx":651
 *     @cython.cdivision(True)
 *     def _set_coarse_index_data_file(self, np.uint64_t file_id):
 *         return self.__set_coarse_index_data_file(file_id)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self->__pyx_vtab)->__pyx___set_coarse_index_data_file(__pyx_v_self, __pyx_v_file_id)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":650
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def _set_coarse_index_data_file(self, np.uint64_t file_id):             # <<<<<<<<<<<<<<
 *         return self.__set_coarse_index_data_file(file_id)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap._set_coarse_index_data_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":656
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void __set_coarse_index_data_file(self, np.uint64_t file_id):             # <<<<<<<<<<<<<<
 *         cdef np.int64_t i
 *         cdef FileBitmasks bitmasks = self.bitmasks
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___set_coarse_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_file_id) {
  __pyx_t_5numpy_int64_t __pyx_v_i;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *__pyx_v_bitmasks = 0;
  PyArrayObject *__pyx_v_mask = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyArrayObject *__pyx_t_3 = NULL;
  npy_intp __pyx_t_4;
  npy_intp __pyx_t_5;
  __pyx_t_5numpy_int64_t __pyx_t_6;
  __pyx_t_5numpy_int64_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set_coarse_index_data_file", 0);
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;

  /* "yt/geometry/particle_oct_container.pyx":658
 *     cdef void __set_coarse_index_data_file(self, np.uint64_t file_id):
 *         cdef np.int64_t i
 *         cdef FileBitmasks bitmasks = self.bitmasks             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint8_t, ndim=1] mask = self.masks[:,file_id]
 *         # Add in order
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->bitmasks);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_bitmasks = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":659
 *         cdef np.int64_t i
 *         cdef FileBitmasks bitmasks = self.bitmasks
 *         cdef np.ndarray[np.uint8_t, ndim=1] mask = self.masks[:,file_id]             # <<<<<<<<<<<<<<
 *         # Add in order
 *         for i in range(mask.shape[0]):
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint64(__pyx_v_file_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_slice__11);
  __Pyx_GIVEREF(__pyx_slice__11);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_slice__11);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_self->masks, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 659, __pyx_L1_error)
  __pyx_t_3 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_t_3, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_mask = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 659, __pyx_L1_error)
    } else {__pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_3 = 0;
  __pyx_v_mask = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":661
 *         cdef np.ndarray[np.uint8_t, ndim=1] mask = self.masks[:,file_id]
 *         # Add in order
 *         for i in range(mask.shape[0]):             # <<<<<<<<<<<<<<
 *             if mask[i] == 1:
 *                 bitmasks._set_coarse(file_id, i)
 */
  __pyx_t_4 = (__pyx_v_mask->dimensions[0]);
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "yt/geometry/particle_oct_container.pyx":662
 *         # Add in order
 *         for i in range(mask.shape[0]):
 *             if mask[i] == 1:             # <<<<<<<<<<<<<<
 *                 bitmasks._set_coarse(file_id, i)
 * 
 */
    __pyx_t_7 = __pyx_v_i;
    __pyx_t_8 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_mask.diminfo[0].strides)) == 1) != 0);
    if (__pyx_t_8) {

      /* "yt/geometry/particle_oct_container.pyx":663
 *         for i in range(mask.shape[0]):
 *             if mask[i] == 1:
 *                 bitmasks._set_coarse(file_id, i)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
      ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_bitmasks->__pyx_vtab)->_set_coarse(__pyx_v_bitmasks, __pyx_v_file_id, __pyx_v_i);

      /* "yt/geometry/particle_oct_container.pyx":662
 *         # Add in order
 *         for i in range(mask.shape[0]):
 *             if mask[i] == 1:             # <<<<<<<<<<<<<<
 *                 bitmasks._set_coarse(file_id, i)
 * 
 */
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":656
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void __set_coarse_index_data_file(self, np.uint64_t file_id):             # <<<<<<<<<<<<<<
 *         cdef np.int64_t i
 *         cdef FileBitmasks bitmasks = self.bitmasks
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_WriteUnraisable("yt.geometry.particle_oct_container.ParticleBitmap.__set_coarse_index_data_file", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_bitmasks);
  __Pyx_XDECREF((PyObject *)__pyx_v_mask);
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":669
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     def _refined_index_data_file(self,             # <<<<<<<<<<<<<<
 *                                  BoolArrayCollection in_collection,
 *                                  np.ndarray[floating, ndim=2] pos,
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_13_refined_index_data_file(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_2yt_8geometry_22particle_oct_container_14ParticleBitmap_13_refined_index_data_file = {"_refined_index_data_file", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_13_refined_index_data_file, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_13_refined_index_data_file(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(0, 669, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(0, 669, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(0, 669, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 669, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12_refined_index_data_file(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12_refined_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  PyTypeObject *__pyx_v_ndarray = 0;
  __Pyx_memviewslice __pyx_v_memslice;
  Py_ssize_t __pyx_v_itemsize;
  CYTHON_UNUSED int __pyx_v_dtype_signed;
  char __pyx_v_kind;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_dtype = NULL;
  PyObject *__pyx_v_arg_base = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  long __pyx_t_7;
  __Pyx_memviewslice __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  int __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_refined_index_data_file", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 669, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 669, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 669, __pyx_L1_error)
  __pyx_t_2 = ((2 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 669, __pyx_L1_error)
    }
    __pyx_t_1 = PyTuple_GET_ITEM(((PyObject*)__pyx_v_args), 2);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 669, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_pos, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 669, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 669, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 669, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 669, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_9);
    __Pyx_GIVEREF(__pyx_int_9);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_9);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 669, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = (__pyx_v_ndarray != ((PyTypeObject*)Py_None));
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_arg, __pyx_v_ndarray); 
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (__pyx_t_2) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_v_dtype = __pyx_t_6;
        __pyx_t_6 = 0;
        goto __pyx_L12;
      }
      __pyx_t_2 = __pyx_memoryview_check(__pyx_v_arg); 
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_base); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_v_arg_base = __pyx_t_6;
        __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_arg_base, __pyx_v_ndarray); 
        __pyx_t_2 = (__pyx_t_3 != 0);
        if (__pyx_t_2) {
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg_base, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 669, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_v_dtype = __pyx_t_6;
          __pyx_t_6 = 0;
          goto __pyx_L13;
        }
        /*else*/ {
          __Pyx_INCREF(Py_None);
          __pyx_v_dtype = Py_None;
        }
        __pyx_L13:;
        goto __pyx_L12;
      }
      /*else*/ {
        __Pyx_INCREF(Py_None);
        __pyx_v_dtype = Py_None;
      }
      __pyx_L12:;
      __pyx_v_itemsize = -1L;
      __pyx_t_2 = (__pyx_v_dtype != Py_None);
      __pyx_t_3 = (__pyx_t_2 != 0);
      if (__pyx_t_3) {
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_itemsize); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_itemsize = __pyx_t_5;
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_dtype, __pyx_n_s_kind); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyObject_Ord(__pyx_t_6); if (unlikely(__pyx_t_7 == ((long)(long)(Py_UCS4)-1))) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_v_kind = __pyx_t_7;
        __pyx_v_dtype_signed = (__pyx_v_kind == 'i');
        switch (__pyx_v_kind) {
          case 'i':
          case 'u':
          break;
          case 'f':
          __pyx_t_2 = (((sizeof(float)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L16_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 669, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 669, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 2) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L16_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 669, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          __pyx_t_2 = (((sizeof(double)) == __pyx_v_itemsize) != 0);
          if (__pyx_t_2) {
          } else {
            __pyx_t_3 = __pyx_t_2;
            goto __pyx_L19_bool_binop_done;
          }
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_arg, __pyx_n_s_ndim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 669, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_t_6); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 669, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_2 = ((((Py_ssize_t)__pyx_t_5) == 2) != 0);
          __pyx_t_3 = __pyx_t_2;
          __pyx_L19_bool_binop_done:;
          if (__pyx_t_3) {
            if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_double, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 669, __pyx_L1_error)
            goto __pyx_L10_break;
          }
          break;
          case 'c':
          break;
          case 'O':
          break;
          default: break;
        }
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L22_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(float))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L22_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dsds_float(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_float, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 669, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    __pyx_t_2 = ((__pyx_v_itemsize == -1L) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L26_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_itemsize == (sizeof(double))) != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L26_bool_binop_done:;
    if (__pyx_t_3) {
      __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_dsds_double(__pyx_v_arg, 0); 
      __pyx_v_memslice = __pyx_t_8;
      __pyx_t_3 = (__pyx_v_memslice.memview != 0);
      if (__pyx_t_3) {
        __PYX_XDEC_MEMVIEW((&__pyx_v_memslice), 1); 
        if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_double, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 669, __pyx_L1_error)
        goto __pyx_L10_break;
      }
      /*else*/ {
        PyErr_Clear(); 
      }
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, Py_None, long, 1, __Pyx_PyInt_From_long, 1, 0, 0) < 0)) __PYX_ERR(0, 669, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 669, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_9), (&__pyx_t_10)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_11 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_9, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_10);
    if (unlikely(__pyx_t_11 == 0)) break;
    if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_12 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_14, __pyx_kp_s__7) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_kp_s__7);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_split); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_12, __pyx_kp_s__8) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_kp_s__8);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_15 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_15 == ((Py_ssize_t)-1))) __PYX_ERR(0, 669, __pyx_L1_error)
    __pyx_t_16 = __pyx_t_15;
    for (__pyx_t_17 = 0; __pyx_t_17 < __pyx_t_16; __pyx_t_17+=1) {
      __pyx_v_i = __pyx_t_17;
      __pyx_t_1 = PyList_GET_ITEM(__pyx_v_dest_sig, __pyx_v_i);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_3 = (__pyx_v_dst_type != Py_None);
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (__pyx_t_2) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_13 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_13); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_13); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 669, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (__pyx_t_2) {
          __pyx_v_match_found = 1;
          goto __pyx_L34;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L32_break;
        }
        __pyx_L34:;
      }
    }
    __pyx_L32_break:;
    __pyx_t_2 = (__pyx_v_match_found != 0);
    if (__pyx_t_2) {
      __pyx_t_18 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(0, 669, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (__pyx_t_3) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 669, __pyx_L1_error)
  }
  __pyx_t_9 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 669, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_9 > 1) != 0);
  if (__pyx_t_3) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 669, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 669, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), PyList_GET_ITEM(__pyx_v_candidates, 0)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_dtype);
  __Pyx_XDECREF(__pyx_v_arg_base);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_0__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_87_refined_index_data_file(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_2yt_8geometry_22particle_oct_container_14ParticleBitmap_87_refined_index_data_file = {"__pyx_fuse_0_refined_index_data_file", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_0__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_87_refined_index_data_file, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_fuse_0__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_87_refined_index_data_file(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_in_collection = 0;
  PyArrayObject *__pyx_v_pos = 0;
  PyArrayObject *__pyx_v_hsml = 0;
  PyArrayObject *__pyx_v_mask = 0;
  CYTHON_UNUSED PyArrayObject *__pyx_v_sub_mi1 = 0;
  CYTHON_UNUSED PyArrayObject *__pyx_v_sub_mi2 = 0;
  CYTHON_UNUSED __pyx_t_5numpy_uint64_t __pyx_v_file_id;
  CYTHON_UNUSED __pyx_t_5numpy_int64_t __pyx_v_nsub_mi;
  __pyx_t_5numpy_uint64_t __pyx_v_count_threshold;
  __pyx_t_5numpy_uint8_t __pyx_v_mask_threshold;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_refined_index_data_file (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_in_collection,&__pyx_n_s_pos,&__pyx_n_s_hsml,&__pyx_n_s_mask,&__pyx_n_s_sub_mi1,&__pyx_n_s_sub_mi2,&__pyx_n_s_file_id,&__pyx_n_s_nsub_mi,&__pyx_n_s_count_threshold,&__pyx_n_s_mask_threshold,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_in_collection)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pos)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_refined_index_data_file", 0, 8, 10, 1); __PYX_ERR(0, 669, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hsml)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_refined_index_data_file", 0, 8, 10, 2); __PYX_ERR(0, 669, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_refined_index_data_file", 0, 8, 10, 3); __PYX_ERR(0, 669, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sub_mi1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_refined_index_data_file", 0, 8, 10, 4); __PYX_ERR(0, 669, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sub_mi2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_refined_index_data_file", 0, 8, 10, 5); __PYX_ERR(0, 669, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_file_id)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_refined_index_data_file", 0, 8, 10, 6); __PYX_ERR(0, 669, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nsub_mi)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_refined_index_data_file", 0, 8, 10, 7); __PYX_ERR(0, 669, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_count_threshold);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask_threshold);
          if (value) { values[9] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_refined_index_data_file") < 0)) __PYX_ERR(0, 669, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_in_collection = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)values[0]);
    __pyx_v_pos = ((PyArrayObject *)values[1]);
    __pyx_v_hsml = ((PyArrayObject *)values[2]);
    __pyx_v_mask = ((PyArrayObject *)values[3]);
    __pyx_v_sub_mi1 = ((PyArrayObject *)values[4]);
    __pyx_v_sub_mi2 = ((PyArrayObject *)values[5]);
    __pyx_v_file_id = __Pyx_PyInt_As_npy_uint64(values[6]); if (unlikely((__pyx_v_file_id == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 676, __pyx_L3_error)
    __pyx_v_nsub_mi = __Pyx_PyInt_As_npy_int64(values[7]); if (unlikely((__pyx_v_nsub_mi == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 676, __pyx_L3_error)
    if (values[8]) {
      __pyx_v_count_threshold = __Pyx_PyInt_As_npy_uint64(values[8]); if (unlikely((__pyx_v_count_threshold == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 677, __pyx_L3_error)
    } else {
      __pyx_v_count_threshold = __pyx_k__13;
    }
    if (values[9]) {
      __pyx_v_mask_threshold = __Pyx_PyInt_As_npy_uint8(values[9]); if (unlikely((__pyx_v_mask_threshold == ((npy_uint8)-1)) && PyErr_Occurred())) __PYX_ERR(0, 678, __pyx_L3_error)
    } else {
      __pyx_v_mask_threshold = __pyx_k__14;
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_refined_index_data_file", 0, 8, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 669, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap._refined_index_data_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_in_collection), __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection, 1, "in_collection", 0))) __PYX_ERR(0, 670, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_pos), __pyx_ptype_5numpy_ndarray, 1, "pos", 0))) __PYX_ERR(0, 671, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hsml), __pyx_ptype_5numpy_ndarray, 1, "hsml", 0))) __PYX_ERR(0, 672, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mask), __pyx_ptype_5numpy_ndarray, 1, "mask", 0))) __PYX_ERR(0, 673, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sub_mi1), __pyx_ptype_5numpy_ndarray, 1, "sub_mi1", 0))) __PYX_ERR(0, 674, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sub_mi2), __pyx_ptype_5numpy_ndarray, 1, "sub_mi2", 0))) __PYX_ERR(0, 675, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_86_refined_index_data_file(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), __pyx_v_in_collection, __pyx_v_pos, __pyx_v_hsml, __pyx_v_mask, __pyx_v_sub_mi1, __pyx_v_sub_mi2, __pyx_v_file_id, __pyx_v_nsub_mi, __pyx_v_count_threshold, __pyx_v_mask_threshold);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_86_refined_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_in_collection, PyArrayObject *__pyx_v_pos, PyArrayObject *__pyx_v_hsml, PyArrayObject *__pyx_v_mask, CYTHON_UNUSED PyArrayObject *__pyx_v_sub_mi1, CYTHON_UNUSED PyArrayObject *__pyx_v_sub_mi2, CYTHON_UNUSED __pyx_t_5numpy_uint64_t __pyx_v_file_id, CYTHON_UNUSED __pyx_t_5numpy_int64_t __pyx_v_nsub_mi, __pyx_t_5numpy_uint64_t __pyx_v_count_threshold, __pyx_t_5numpy_uint8_t __pyx_v_mask_threshold) {
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v__in_coll = 0;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_out_collection = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_hsml;
  __Pyx_Buffer __pyx_pybuffer_hsml;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_pos;
  __Pyx_Buffer __pyx_pybuffer_pos;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_sub_mi1;
  __Pyx_Buffer __pyx_pybuffer_sub_mi1;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_sub_mi2;
  __Pyx_Buffer __pyx_pybuffer_sub_mi2;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_0_refined_index_data_file", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_in_collection);
  __pyx_pybuffer_pos.pybuffer.buf = NULL;
  __pyx_pybuffer_pos.refcount = 0;
  __pyx_pybuffernd_pos.data = NULL;
  __pyx_pybuffernd_pos.rcbuffer = &__pyx_pybuffer_pos;
  __pyx_pybuffer_hsml.pybuffer.buf = NULL;
  __pyx_pybuffer_hsml.refcount = 0;
  __pyx_pybuffernd_hsml.data = NULL;
  __pyx_pybuffernd_hsml.rcbuffer = &__pyx_pybuffer_hsml;
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  __pyx_pybuffer_sub_mi1.pybuffer.buf = NULL;
  __pyx_pybuffer_sub_mi1.refcount = 0;
  __pyx_pybuffernd_sub_mi1.data = NULL;
  __pyx_pybuffernd_sub_mi1.rcbuffer = &__pyx_pybuffer_sub_mi1;
  __pyx_pybuffer_sub_mi2.pybuffer.buf = NULL;
  __pyx_pybuffer_sub_mi2.refcount = 0;
  __pyx_pybuffernd_sub_mi2.data = NULL;
  __pyx_pybuffernd_sub_mi2.rcbuffer = &__pyx_pybuffer_sub_mi2;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pos.rcbuffer->pybuffer, (PyObject*)__pyx_v_pos, &__Pyx_TypeInfo_float, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 669, __pyx_L1_error)
  }
  __pyx_pybuffernd_pos.diminfo[0].strides = __pyx_pybuffernd_pos.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_pos.diminfo[0].shape = __pyx_pybuffernd_pos.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_pos.diminfo[1].strides = __pyx_pybuffernd_pos.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_pos.diminfo[1].shape = __pyx_pybuffernd_pos.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer, (PyObject*)__pyx_v_hsml, &__Pyx_TypeInfo_float, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 669, __pyx_L1_error)
  }
  __pyx_pybuffernd_hsml.diminfo[0].strides = __pyx_pybuffernd_hsml.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_hsml.diminfo[0].shape = __pyx_pybuffernd_hsml.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 669, __pyx_L1_error)
  }
  __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer, (PyObject*)__pyx_v_sub_mi1, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 669, __pyx_L1_error)
  }
  __pyx_pybuffernd_sub_mi1.diminfo[0].strides = __pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_sub_mi1.diminfo[0].shape = __pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer, (PyObject*)__pyx_v_sub_mi2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 669, __pyx_L1_error)
  }
  __pyx_pybuffernd_sub_mi2.diminfo[0].strides = __pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_sub_mi2.diminfo[0].shape = __pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/particle_oct_container.pyx":679
 *                                  np.uint64_t count_threshold = 128,
 *                                  np.uint8_t mask_threshold = 2):
 *         self._used_mi2 = 1             # <<<<<<<<<<<<<<
 *         if in_collection is None:
 *             in_collection = BoolArrayCollection()
 */
  __pyx_v_self->_used_mi2 = 1;

  /* "yt/geometry/particle_oct_container.pyx":680
 *                                  np.uint8_t mask_threshold = 2):
 *         self._used_mi2 = 1
 *         if in_collection is None:             # <<<<<<<<<<<<<<
 *             in_collection = BoolArrayCollection()
 *         cdef BoolArrayCollection _in_coll = in_collection
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_in_collection) == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "yt/geometry/particle_oct_container.pyx":681
 *         self._used_mi2 = 1
 *         if in_collection is None:
 *             in_collection = BoolArrayCollection()             # <<<<<<<<<<<<<<
 *         cdef BoolArrayCollection _in_coll = in_collection
 *         out_collection = self.__refined_index_data_file(_in_coll, pos, hsml, mask,
 */
    __pyx_t_3 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_in_collection, ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "yt/geometry/particle_oct_container.pyx":680
 *                                  np.uint8_t mask_threshold = 2):
 *         self._used_mi2 = 1
 *         if in_collection is None:             # <<<<<<<<<<<<<<
 *             in_collection = BoolArrayCollection()
 *         cdef BoolArrayCollection _in_coll = in_collection
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":682
 *         if in_collection is None:
 *             in_collection = BoolArrayCollection()
 *         cdef BoolArrayCollection _in_coll = in_collection             # <<<<<<<<<<<<<<
 *         out_collection = self.__refined_index_data_file(_in_coll, pos, hsml, mask,
 *                                               count_threshold, mask_threshold)
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_in_collection));
  __pyx_v__in_coll = __pyx_v_in_collection;

  /* "yt/geometry/particle_oct_container.pyx":683
 *             in_collection = BoolArrayCollection()
 *         cdef BoolArrayCollection _in_coll = in_collection
 *         out_collection = self.__refined_index_data_file(_in_coll, pos, hsml, mask,             # <<<<<<<<<<<<<<
 *                                               count_threshold, mask_threshold)
 *         return 0, out_collection
 */
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self->__pyx_vtab)->__pyx_fuse_0__refined_index_data_file(__pyx_v_self, __pyx_v__in_coll, ((PyArrayObject *)__pyx_v_pos), ((PyArrayObject *)__pyx_v_hsml), ((PyArrayObject *)__pyx_v_mask), __pyx_v_count_threshold, __pyx_v_mask_threshold)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 683, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_out_collection = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":685
 *         out_collection = self.__refined_index_data_file(_in_coll, pos, hsml, mask,
 *                                               count_threshold, mask_threshold)
 *         return 0, out_collection             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_int_0);
  __Pyx_INCREF(((PyObject *)__pyx_v_out_collection));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_out_collection));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_out_collection));
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":669
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     def _refined_index_data_file(self,             # <<<<<<<<<<<<<<
 *                                  BoolArrayCollection in_collection,
 *                                  np.ndarray[floating, ndim=2] pos,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap._refined_index_data_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v__in_coll);
  __Pyx_XDECREF((PyObject *)__pyx_v_out_collection);
  __Pyx_XDECREF((PyObject *)__pyx_v_in_collection);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_fuse_1__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_89_refined_index_data_file(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_2yt_8geometry_22particle_oct_container_14ParticleBitmap_89_refined_index_data_file = {"__pyx_fuse_1_refined_index_data_file", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_1__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_89_refined_index_data_file, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_fuse_1__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_89_refined_index_data_file(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_in_collection = 0;
  PyArrayObject *__pyx_v_pos = 0;
  PyArrayObject *__pyx_v_hsml = 0;
  PyArrayObject *__pyx_v_mask = 0;
  CYTHON_UNUSED PyArrayObject *__pyx_v_sub_mi1 = 0;
  CYTHON_UNUSED PyArrayObject *__pyx_v_sub_mi2 = 0;
  CYTHON_UNUSED __pyx_t_5numpy_uint64_t __pyx_v_file_id;
  CYTHON_UNUSED __pyx_t_5numpy_int64_t __pyx_v_nsub_mi;
  __pyx_t_5numpy_uint64_t __pyx_v_count_threshold;
  __pyx_t_5numpy_uint8_t __pyx_v_mask_threshold;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_refined_index_data_file (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_in_collection,&__pyx_n_s_pos,&__pyx_n_s_hsml,&__pyx_n_s_mask,&__pyx_n_s_sub_mi1,&__pyx_n_s_sub_mi2,&__pyx_n_s_file_id,&__pyx_n_s_nsub_mi,&__pyx_n_s_count_threshold,&__pyx_n_s_mask_threshold,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_in_collection)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pos)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_refined_index_data_file", 0, 8, 10, 1); __PYX_ERR(0, 669, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hsml)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_refined_index_data_file", 0, 8, 10, 2); __PYX_ERR(0, 669, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_refined_index_data_file", 0, 8, 10, 3); __PYX_ERR(0, 669, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sub_mi1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_refined_index_data_file", 0, 8, 10, 4); __PYX_ERR(0, 669, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sub_mi2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_refined_index_data_file", 0, 8, 10, 5); __PYX_ERR(0, 669, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_file_id)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_refined_index_data_file", 0, 8, 10, 6); __PYX_ERR(0, 669, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nsub_mi)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_refined_index_data_file", 0, 8, 10, 7); __PYX_ERR(0, 669, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_count_threshold);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask_threshold);
          if (value) { values[9] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_refined_index_data_file") < 0)) __PYX_ERR(0, 669, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_in_collection = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)values[0]);
    __pyx_v_pos = ((PyArrayObject *)values[1]);
    __pyx_v_hsml = ((PyArrayObject *)values[2]);
    __pyx_v_mask = ((PyArrayObject *)values[3]);
    __pyx_v_sub_mi1 = ((PyArrayObject *)values[4]);
    __pyx_v_sub_mi2 = ((PyArrayObject *)values[5]);
    __pyx_v_file_id = __Pyx_PyInt_As_npy_uint64(values[6]); if (unlikely((__pyx_v_file_id == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 676, __pyx_L3_error)
    __pyx_v_nsub_mi = __Pyx_PyInt_As_npy_int64(values[7]); if (unlikely((__pyx_v_nsub_mi == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 676, __pyx_L3_error)
    if (values[8]) {
      __pyx_v_count_threshold = __Pyx_PyInt_As_npy_uint64(values[8]); if (unlikely((__pyx_v_count_threshold == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 677, __pyx_L3_error)
    } else {
      __pyx_v_count_threshold = __pyx_k__15;
    }
    if (values[9]) {
      __pyx_v_mask_threshold = __Pyx_PyInt_As_npy_uint8(values[9]); if (unlikely((__pyx_v_mask_threshold == ((npy_uint8)-1)) && PyErr_Occurred())) __PYX_ERR(0, 678, __pyx_L3_error)
    } else {
      __pyx_v_mask_threshold = __pyx_k__16;
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_refined_index_data_file", 0, 8, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 669, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap._refined_index_data_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_in_collection), __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection, 1, "in_collection", 0))) __PYX_ERR(0, 670, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_pos), __pyx_ptype_5numpy_ndarray, 1, "pos", 0))) __PYX_ERR(0, 671, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hsml), __pyx_ptype_5numpy_ndarray, 1, "hsml", 0))) __PYX_ERR(0, 672, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mask), __pyx_ptype_5numpy_ndarray, 1, "mask", 0))) __PYX_ERR(0, 673, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sub_mi1), __pyx_ptype_5numpy_ndarray, 1, "sub_mi1", 0))) __PYX_ERR(0, 674, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sub_mi2), __pyx_ptype_5numpy_ndarray, 1, "sub_mi2", 0))) __PYX_ERR(0, 675, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_88_refined_index_data_file(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), __pyx_v_in_collection, __pyx_v_pos, __pyx_v_hsml, __pyx_v_mask, __pyx_v_sub_mi1, __pyx_v_sub_mi2, __pyx_v_file_id, __pyx_v_nsub_mi, __pyx_v_count_threshold, __pyx_v_mask_threshold);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_88_refined_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_in_collection, PyArrayObject *__pyx_v_pos, PyArrayObject *__pyx_v_hsml, PyArrayObject *__pyx_v_mask, CYTHON_UNUSED PyArrayObject *__pyx_v_sub_mi1, CYTHON_UNUSED PyArrayObject *__pyx_v_sub_mi2, CYTHON_UNUSED __pyx_t_5numpy_uint64_t __pyx_v_file_id, CYTHON_UNUSED __pyx_t_5numpy_int64_t __pyx_v_nsub_mi, __pyx_t_5numpy_uint64_t __pyx_v_count_threshold, __pyx_t_5numpy_uint8_t __pyx_v_mask_threshold) {
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v__in_coll = 0;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_out_collection = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_hsml;
  __Pyx_Buffer __pyx_pybuffer_hsml;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_pos;
  __Pyx_Buffer __pyx_pybuffer_pos;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_sub_mi1;
  __Pyx_Buffer __pyx_pybuffer_sub_mi1;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_sub_mi2;
  __Pyx_Buffer __pyx_pybuffer_sub_mi2;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_1_refined_index_data_file", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_in_collection);
  __pyx_pybuffer_pos.pybuffer.buf = NULL;
  __pyx_pybuffer_pos.refcount = 0;
  __pyx_pybuffernd_pos.data = NULL;
  __pyx_pybuffernd_pos.rcbuffer = &__pyx_pybuffer_pos;
  __pyx_pybuffer_hsml.pybuffer.buf = NULL;
  __pyx_pybuffer_hsml.refcount = 0;
  __pyx_pybuffernd_hsml.data = NULL;
  __pyx_pybuffernd_hsml.rcbuffer = &__pyx_pybuffer_hsml;
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  __pyx_pybuffer_sub_mi1.pybuffer.buf = NULL;
  __pyx_pybuffer_sub_mi1.refcount = 0;
  __pyx_pybuffernd_sub_mi1.data = NULL;
  __pyx_pybuffernd_sub_mi1.rcbuffer = &__pyx_pybuffer_sub_mi1;
  __pyx_pybuffer_sub_mi2.pybuffer.buf = NULL;
  __pyx_pybuffer_sub_mi2.refcount = 0;
  __pyx_pybuffernd_sub_mi2.data = NULL;
  __pyx_pybuffernd_sub_mi2.rcbuffer = &__pyx_pybuffer_sub_mi2;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pos.rcbuffer->pybuffer, (PyObject*)__pyx_v_pos, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 669, __pyx_L1_error)
  }
  __pyx_pybuffernd_pos.diminfo[0].strides = __pyx_pybuffernd_pos.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_pos.diminfo[0].shape = __pyx_pybuffernd_pos.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_pos.diminfo[1].strides = __pyx_pybuffernd_pos.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_pos.diminfo[1].shape = __pyx_pybuffernd_pos.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer, (PyObject*)__pyx_v_hsml, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 669, __pyx_L1_error)
  }
  __pyx_pybuffernd_hsml.diminfo[0].strides = __pyx_pybuffernd_hsml.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_hsml.diminfo[0].shape = __pyx_pybuffernd_hsml.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 669, __pyx_L1_error)
  }
  __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer, (PyObject*)__pyx_v_sub_mi1, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 669, __pyx_L1_error)
  }
  __pyx_pybuffernd_sub_mi1.diminfo[0].strides = __pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_sub_mi1.diminfo[0].shape = __pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer, (PyObject*)__pyx_v_sub_mi2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 669, __pyx_L1_error)
  }
  __pyx_pybuffernd_sub_mi2.diminfo[0].strides = __pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_sub_mi2.diminfo[0].shape = __pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/particle_oct_container.pyx":679
 *                                  np.uint64_t count_threshold = 128,
 *                                  np.uint8_t mask_threshold = 2):
 *         self._used_mi2 = 1             # <<<<<<<<<<<<<<
 *         if in_collection is None:
 *             in_collection = BoolArrayCollection()
 */
  __pyx_v_self->_used_mi2 = 1;

  /* "yt/geometry/particle_oct_container.pyx":680
 *                                  np.uint8_t mask_threshold = 2):
 *         self._used_mi2 = 1
 *         if in_collection is None:             # <<<<<<<<<<<<<<
 *             in_collection = BoolArrayCollection()
 *         cdef BoolArrayCollection _in_coll = in_collection
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_in_collection) == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "yt/geometry/particle_oct_container.pyx":681
 *         self._used_mi2 = 1
 *         if in_collection is None:
 *             in_collection = BoolArrayCollection()             # <<<<<<<<<<<<<<
 *         cdef BoolArrayCollection _in_coll = in_collection
 *         out_collection = self.__refined_index_data_file(_in_coll, pos, hsml, mask,
 */
    __pyx_t_3 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_in_collection, ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "yt/geometry/particle_oct_container.pyx":680
 *                                  np.uint8_t mask_threshold = 2):
 *         self._used_mi2 = 1
 *         if in_collection is None:             # <<<<<<<<<<<<<<
 *             in_collection = BoolArrayCollection()
 *         cdef BoolArrayCollection _in_coll = in_collection
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":682
 *         if in_collection is None:
 *             in_collection = BoolArrayCollection()
 *         cdef BoolArrayCollection _in_coll = in_collection             # <<<<<<<<<<<<<<
 *         out_collection = self.__refined_index_data_file(_in_coll, pos, hsml, mask,
 *                                               count_threshold, mask_threshold)
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_in_collection));
  __pyx_v__in_coll = __pyx_v_in_collection;

  /* "yt/geometry/particle_oct_container.pyx":683
 *             in_collection = BoolArrayCollection()
 *         cdef BoolArrayCollection _in_coll = in_collection
 *         out_collection = self.__refined_index_data_file(_in_coll, pos, hsml, mask,             # <<<<<<<<<<<<<<
 *                                               count_threshold, mask_threshold)
 *         return 0, out_collection
 */
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self->__pyx_vtab)->__pyx_fuse_1__refined_index_data_file(__pyx_v_self, __pyx_v__in_coll, ((PyArrayObject *)__pyx_v_pos), ((PyArrayObject *)__pyx_v_hsml), ((PyArrayObject *)__pyx_v_mask), __pyx_v_count_threshold, __pyx_v_mask_threshold)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 683, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_out_collection = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":685
 *         out_collection = self.__refined_index_data_file(_in_coll, pos, hsml, mask,
 *                                               count_threshold, mask_threshold)
 *         return 0, out_collection             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_int_0);
  __Pyx_INCREF(((PyObject *)__pyx_v_out_collection));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_out_collection));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_out_collection));
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":669
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     def _refined_index_data_file(self,             # <<<<<<<<<<<<<<
 *                                  BoolArrayCollection in_collection,
 *                                  np.ndarray[floating, ndim=2] pos,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap._refined_index_data_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v__in_coll);
  __Pyx_XDECREF((PyObject *)__pyx_v_out_collection);
  __Pyx_XDECREF((PyObject *)__pyx_v_in_collection);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":691
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef BoolArrayCollection __refined_index_data_file(             # <<<<<<<<<<<<<<
 *         self,
 *         BoolArrayCollection in_collection,
 */

static struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_fuse_0__pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___refined_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_in_collection, PyArrayObject *__pyx_v_pos, PyArrayObject *__pyx_v_hsml, PyArrayObject *__pyx_v_mask, __pyx_t_5numpy_uint64_t __pyx_v_count_threshold, __pyx_t_5numpy_uint8_t __pyx_v_mask_threshold) {
  __pyx_t_5numpy_int64_t __pyx_v_p;
  __pyx_t_5numpy_int64_t __pyx_v_sorted_ind;
  __pyx_t_5numpy_uint64_t __pyx_v_i;
  __pyx_t_5numpy_uint64_t __pyx_v_mi1;
  __pyx_t_5numpy_uint64_t __pyx_v_mi2;
  __pyx_t_5numpy_float64_t __pyx_v_ppos[3];
  __pyx_t_5numpy_float64_t __pyx_v_s_ppos[3];
  int __pyx_v_skip;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_this_collection = 0;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_out_collection = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_bounds[2][3];
  __pyx_t_5numpy_uint8_t __pyx_v_fully_enclosed;
  __pyx_t_5numpy_float64_t __pyx_v_LE[3];
  __pyx_t_5numpy_float64_t __pyx_v_RE[3];
  __pyx_t_5numpy_float64_t __pyx_v_DW[3];
  __pyx_t_5numpy_uint8_t __pyx_v_PER[3];
  __pyx_t_5numpy_float64_t __pyx_v_dds1[3];
  __pyx_t_5numpy_float64_t __pyx_v_dds2[3];
  __pyx_t_5numpy_float64_t __pyx_v_radius;
  __pyx_t_5numpy_uint64_t __pyx_v_mi_split1[3];
  __pyx_t_5numpy_uint64_t __pyx_v_mi_split2[3];
  __pyx_t_5numpy_uint64_t __pyx_v_miex1;
  __Pyx_memviewslice __pyx_v_particle_counts = { 0, 0, { 0 }, { 0 }, { 0 } };
  __pyx_t_5numpy_uint64_t __pyx_v_xex;
  __pyx_t_5numpy_uint64_t __pyx_v_yex;
  __pyx_t_5numpy_uint64_t __pyx_v_zex;
  __pyx_t_5numpy_float64_t __pyx_v_clip_pos_l[3];
  __pyx_t_5numpy_float64_t __pyx_v_clip_pos_r[3];
  int __pyx_v_axiter[3][2];
  __pyx_t_5numpy_float64_t __pyx_v_axiterv[3][2];
  __pyx_t_2yt_8geometry_22particle_oct_container_CoarseRefinedSets __pyx_v_coarse_refined_map;
  __pyx_t_5numpy_uint64_t __pyx_v_nfully_enclosed;
  __pyx_t_5numpy_uint64_t __pyx_v_n_calls;
  CYTHON_UNUSED long __pyx_v_mi1_max;
  CYTHON_UNUSED long __pyx_v_mi2_max;
  __pyx_t_5numpy_uint64_t __pyx_v_max_mi1_elements;
  __pyx_t_5numpy_uint64_t __pyx_v_max_mi2_elements;
  PyArrayObject *__pyx_v_refined_count = 0;
  PyArrayObject *__pyx_v_morton_indices = 0;
  PyArrayObject *__pyx_v_sorted_order = 0;
  long __pyx_v_xi;
  long __pyx_v_yi;
  long __pyx_v_zi;
  __pyx_t_5numpy_uint64_t __pyx_v_vec_i;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_bool_array *__pyx_v_buf;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_word_type __pyx_v_w;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array *__pyx_v_refined_arr;
  std::pair<__pyx_t_5numpy_uint64_t,__pyx_t_2yt_9utilities_3lib_15ewah_bool_array_bool_array>  __pyx_v_it1;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_hsml;
  __Pyx_Buffer __pyx_pybuffer_hsml;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_morton_indices;
  __Pyx_Buffer __pyx_pybuffer_morton_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_pos;
  __Pyx_Buffer __pyx_pybuffer_pos;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_refined_count;
  __Pyx_Buffer __pyx_pybuffer_refined_count;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_sorted_order;
  __Pyx_Buffer __pyx_pybuffer_sorted_order;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyArrayObject *__pyx_t_6 = NULL;
  __pyx_t_5numpy_uint64_t __pyx_t_7;
  PyArrayObject *__pyx_t_8 = NULL;
  npy_intp __pyx_t_9;
  npy_intp __pyx_t_10;
  __pyx_t_5numpy_int64_t __pyx_t_11;
  __pyx_t_5numpy_int64_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  __pyx_t_5numpy_int64_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __pyx_t_5numpy_int64_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  __pyx_t_5numpy_int64_t __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  PyArrayObject *__pyx_t_21 = NULL;
  int __pyx_t_22;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  __pyx_t_5numpy_uint64_t __pyx_t_26;
  float __pyx_t_27;
  long __pyx_t_28;
  long __pyx_t_29;
  long __pyx_t_30;
  __pyx_t_5numpy_uint64_t __pyx_t_31;
  __pyx_t_5numpy_uint64_t __pyx_t_32;
  __pyx_t_5numpy_uint64_t __pyx_t_33;
  __pyx_t_5numpy_uint64_t __pyx_t_34;
  __pyx_t_5numpy_uint64_t __pyx_t_35;
  __pyx_t_5numpy_uint64_t __pyx_t_36;
  __pyx_t_5numpy_uint64_t __pyx_t_37;
  __pyx_t_5numpy_uint64_t __pyx_t_38;
  std::map<__pyx_t_5numpy_uint64_t,__pyx_t_2yt_9utilities_3lib_15ewah_bool_array_bool_array> ::iterator __pyx_t_39;
  std::pair<__pyx_t_5numpy_uint64_t,__pyx_t_2yt_9utilities_3lib_15ewah_bool_array_bool_array>  __pyx_t_40;
  size_t __pyx_t_41;
  size_t __pyx_t_42;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_0__refined_index_data_file", 0);
  __pyx_pybuffer_refined_count.pybuffer.buf = NULL;
  __pyx_pybuffer_refined_count.refcount = 0;
  __pyx_pybuffernd_refined_count.data = NULL;
  __pyx_pybuffernd_refined_count.rcbuffer = &__pyx_pybuffer_refined_count;
  __pyx_pybuffer_morton_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_morton_indices.refcount = 0;
  __pyx_pybuffernd_morton_indices.data = NULL;
  __pyx_pybuffernd_morton_indices.rcbuffer = &__pyx_pybuffer_morton_indices;
  __pyx_pybuffer_sorted_order.pybuffer.buf = NULL;
  __pyx_pybuffer_sorted_order.refcount = 0;
  __pyx_pybuffernd_sorted_order.data = NULL;
  __pyx_pybuffernd_sorted_order.rcbuffer = &__pyx_pybuffer_sorted_order;
  __pyx_pybuffer_pos.pybuffer.buf = NULL;
  __pyx_pybuffer_pos.refcount = 0;
  __pyx_pybuffernd_pos.data = NULL;
  __pyx_pybuffernd_pos.rcbuffer = &__pyx_pybuffer_pos;
  __pyx_pybuffer_hsml.pybuffer.buf = NULL;
  __pyx_pybuffer_hsml.refcount = 0;
  __pyx_pybuffernd_hsml.data = NULL;
  __pyx_pybuffernd_hsml.rcbuffer = &__pyx_pybuffer_hsml;
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pos.rcbuffer->pybuffer, (PyObject*)__pyx_v_pos, &__Pyx_TypeInfo_float, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 691, __pyx_L1_error)
  }
  __pyx_pybuffernd_pos.diminfo[0].strides = __pyx_pybuffernd_pos.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_pos.diminfo[0].shape = __pyx_pybuffernd_pos.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_pos.diminfo[1].strides = __pyx_pybuffernd_pos.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_pos.diminfo[1].shape = __pyx_pybuffernd_pos.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer, (PyObject*)__pyx_v_hsml, &__Pyx_TypeInfo_float, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 691, __pyx_L1_error)
  }
  __pyx_pybuffernd_hsml.diminfo[0].strides = __pyx_pybuffernd_hsml.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_hsml.diminfo[0].shape = __pyx_pybuffernd_hsml.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 691, __pyx_L1_error)
  }
  __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/particle_oct_container.pyx":719
 *         cdef np.uint64_t mi_split2[3]
 *         cdef np.uint64_t miex1
 *         cdef np.uint64_t[:] particle_counts = self.particle_counts             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t xex, yex, zex
 *         cdef np.float64_t clip_pos_l[3]
 */
  __pyx_t_1 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_v_self->particle_counts, PyBUF_WRITABLE); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(0, 719, __pyx_L1_error)
  __pyx_v_particle_counts = __pyx_t_1;
  __pyx_t_1.memview = NULL;
  __pyx_t_1.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":726
 *         cdef np.float64_t axiterv[3][2]
 *         cdef CoarseRefinedSets coarse_refined_map
 *         cdef np.uint64_t nfully_enclosed = 0, n_calls = 0             # <<<<<<<<<<<<<<
 *         mi1_max = (1 << self.index_order1) - 1
 *         mi2_max = (1 << self.index_order2) - 1
 */
  __pyx_v_nfully_enclosed = 0;
  __pyx_v_n_calls = 0;

  /* "yt/geometry/particle_oct_container.pyx":727
 *         cdef CoarseRefinedSets coarse_refined_map
 *         cdef np.uint64_t nfully_enclosed = 0, n_calls = 0
 *         mi1_max = (1 << self.index_order1) - 1             # <<<<<<<<<<<<<<
 *         mi2_max = (1 << self.index_order2) - 1
 *         cdef np.uint64_t max_mi1_elements = 1 << (3*self.index_order1)
 */
  __pyx_v_mi1_max = ((1 << __pyx_v_self->index_order1) - 1);

  /* "yt/geometry/particle_oct_container.pyx":728
 *         cdef np.uint64_t nfully_enclosed = 0, n_calls = 0
 *         mi1_max = (1 << self.index_order1) - 1
 *         mi2_max = (1 << self.index_order2) - 1             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t max_mi1_elements = 1 << (3*self.index_order1)
 *         cdef np.uint64_t max_mi2_elements = 1 << (3*self.index_order2)
 */
  __pyx_v_mi2_max = ((1 << __pyx_v_self->index_order2) - 1);

  /* "yt/geometry/particle_oct_container.pyx":729
 *         mi1_max = (1 << self.index_order1) - 1
 *         mi2_max = (1 << self.index_order2) - 1
 *         cdef np.uint64_t max_mi1_elements = 1 << (3*self.index_order1)             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t max_mi2_elements = 1 << (3*self.index_order2)
 *         cdef np.ndarray[np.uint64_t, ndim=1] refined_count = np.zeros(max_mi1_elements, dtype="uint64")
 */
  __pyx_v_max_mi1_elements = (1 << (3 * __pyx_v_self->index_order1));

  /* "yt/geometry/particle_oct_container.pyx":730
 *         mi2_max = (1 << self.index_order2) - 1
 *         cdef np.uint64_t max_mi1_elements = 1 << (3*self.index_order1)
 *         cdef np.uint64_t max_mi2_elements = 1 << (3*self.index_order2)             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint64_t, ndim=1] refined_count = np.zeros(max_mi1_elements, dtype="uint64")
 *         # Copy things from structure (type cast)
 */
  __pyx_v_max_mi2_elements = (1 << (3 * __pyx_v_self->index_order2));

  /* "yt/geometry/particle_oct_container.pyx":731
 *         cdef np.uint64_t max_mi1_elements = 1 << (3*self.index_order1)
 *         cdef np.uint64_t max_mi2_elements = 1 << (3*self.index_order2)
 *         cdef np.ndarray[np.uint64_t, ndim=1] refined_count = np.zeros(max_mi1_elements, dtype="uint64")             # <<<<<<<<<<<<<<
 *         # Copy things from structure (type cast)
 *         for i in range(3):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_npy_uint64(__pyx_v_max_mi1_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_n_s_uint64) < 0) __PYX_ERR(0, 731, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 731, __pyx_L1_error)
  __pyx_t_6 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_refined_count.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_refined_count = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_refined_count.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 731, __pyx_L1_error)
    } else {__pyx_pybuffernd_refined_count.diminfo[0].strides = __pyx_pybuffernd_refined_count.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_refined_count.diminfo[0].shape = __pyx_pybuffernd_refined_count.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_6 = 0;
  __pyx_v_refined_count = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "yt/geometry/particle_oct_container.pyx":733
 *         cdef np.ndarray[np.uint64_t, ndim=1] refined_count = np.zeros(max_mi1_elements, dtype="uint64")
 *         # Copy things from structure (type cast)
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             LE[i] = self.left_edge[i]
 *             RE[i] = self.right_edge[i]
 */
  for (__pyx_t_7 = 0; __pyx_t_7 < 3; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "yt/geometry/particle_oct_container.pyx":734
 *         # Copy things from structure (type cast)
 *         for i in range(3):
 *             LE[i] = self.left_edge[i]             # <<<<<<<<<<<<<<
 *             RE[i] = self.right_edge[i]
 *             PER[i] = self.periodicity[i]
 */
    (__pyx_v_LE[__pyx_v_i]) = (__pyx_v_self->left_edge[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":735
 *         for i in range(3):
 *             LE[i] = self.left_edge[i]
 *             RE[i] = self.right_edge[i]             # <<<<<<<<<<<<<<
 *             PER[i] = self.periodicity[i]
 *             dds1[i] = self.dds_mi1[i]
 */
    (__pyx_v_RE[__pyx_v_i]) = (__pyx_v_self->right_edge[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":736
 *             LE[i] = self.left_edge[i]
 *             RE[i] = self.right_edge[i]
 *             PER[i] = self.periodicity[i]             # <<<<<<<<<<<<<<
 *             dds1[i] = self.dds_mi1[i]
 *             dds2[i] = self.dds_mi2[i]
 */
    (__pyx_v_PER[__pyx_v_i]) = (__pyx_v_self->periodicity[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":737
 *             RE[i] = self.right_edge[i]
 *             PER[i] = self.periodicity[i]
 *             dds1[i] = self.dds_mi1[i]             # <<<<<<<<<<<<<<
 *             dds2[i] = self.dds_mi2[i]
 *             DW[i] = RE[i] - LE[i]
 */
    (__pyx_v_dds1[__pyx_v_i]) = (__pyx_v_self->dds_mi1[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":738
 *             PER[i] = self.periodicity[i]
 *             dds1[i] = self.dds_mi1[i]
 *             dds2[i] = self.dds_mi2[i]             # <<<<<<<<<<<<<<
 *             DW[i] = RE[i] - LE[i]
 *             axiter[i][0] = 0 # We always do an offset of 0
 */
    (__pyx_v_dds2[__pyx_v_i]) = (__pyx_v_self->dds_mi2[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":739
 *             dds1[i] = self.dds_mi1[i]
 *             dds2[i] = self.dds_mi2[i]
 *             DW[i] = RE[i] - LE[i]             # <<<<<<<<<<<<<<
 *             axiter[i][0] = 0 # We always do an offset of 0
 *             axiterv[i][0] = 0.0
 */
    (__pyx_v_DW[__pyx_v_i]) = ((__pyx_v_RE[__pyx_v_i]) - (__pyx_v_LE[__pyx_v_i]));

    /* "yt/geometry/particle_oct_container.pyx":740
 *             dds2[i] = self.dds_mi2[i]
 *             DW[i] = RE[i] - LE[i]
 *             axiter[i][0] = 0 # We always do an offset of 0             # <<<<<<<<<<<<<<
 *             axiterv[i][0] = 0.0
 *         cdef np.ndarray[np.uint64_t, ndim=1] morton_indices = np.empty(pos.shape[0], dtype="u8")
 */
    ((__pyx_v_axiter[__pyx_v_i])[0]) = 0;

    /* "yt/geometry/particle_oct_container.pyx":741
 *             DW[i] = RE[i] - LE[i]
 *             axiter[i][0] = 0 # We always do an offset of 0
 *             axiterv[i][0] = 0.0             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint64_t, ndim=1] morton_indices = np.empty(pos.shape[0], dtype="u8")
 *         for p in range(pos.shape[0]):
 */
    ((__pyx_v_axiterv[__pyx_v_i])[0]) = 0.0;
  }

  /* "yt/geometry/particle_oct_container.pyx":742
 *             axiter[i][0] = 0 # We always do an offset of 0
 *             axiterv[i][0] = 0.0
 *         cdef np.ndarray[np.uint64_t, ndim=1] morton_indices = np.empty(pos.shape[0], dtype="u8")             # <<<<<<<<<<<<<<
 *         for p in range(pos.shape[0]):
 *             morton_indices[p] = bounded_morton(pos[p, 0], pos[p, 1], pos[p, 2],
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_Py_intptr_t((__pyx_v_pos->dimensions[0])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_n_s_u8) < 0) __PYX_ERR(0, 742, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 742, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_morton_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_morton_indices = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_morton_indices.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 742, __pyx_L1_error)
    } else {__pyx_pybuffernd_morton_indices.diminfo[0].strides = __pyx_pybuffernd_morton_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_morton_indices.diminfo[0].shape = __pyx_pybuffernd_morton_indices.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_8 = 0;
  __pyx_v_morton_indices = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":743
 *             axiterv[i][0] = 0.0
 *         cdef np.ndarray[np.uint64_t, ndim=1] morton_indices = np.empty(pos.shape[0], dtype="u8")
 *         for p in range(pos.shape[0]):             # <<<<<<<<<<<<<<
 *             morton_indices[p] = bounded_morton(pos[p, 0], pos[p, 1], pos[p, 2],
 *                                                LE, RE, self.index_order1)
 */
  __pyx_t_9 = (__pyx_v_pos->dimensions[0]);
  __pyx_t_10 = __pyx_t_9;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_p = __pyx_t_11;

    /* "yt/geometry/particle_oct_container.pyx":744
 *         cdef np.ndarray[np.uint64_t, ndim=1] morton_indices = np.empty(pos.shape[0], dtype="u8")
 *         for p in range(pos.shape[0]):
 *             morton_indices[p] = bounded_morton(pos[p, 0], pos[p, 1], pos[p, 2],             # <<<<<<<<<<<<<<
 *                                                LE, RE, self.index_order1)
 *         # Loop over positions skipping those outside the domain
 */
    __pyx_t_12 = __pyx_v_p;
    __pyx_t_13 = 0;
    __pyx_t_14 = __pyx_v_p;
    __pyx_t_15 = 1;
    __pyx_t_16 = __pyx_v_p;
    __pyx_t_17 = 2;

    /* "yt/geometry/particle_oct_container.pyx":745
 *         for p in range(pos.shape[0]):
 *             morton_indices[p] = bounded_morton(pos[p, 0], pos[p, 1], pos[p, 2],
 *                                                LE, RE, self.index_order1)             # <<<<<<<<<<<<<<
 *         # Loop over positions skipping those outside the domain
 *         cdef np.ndarray[np.uint64_t, ndim=1, cast=True] sorted_order
 */
    __pyx_t_18 = __pyx_v_p;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_morton_indices.rcbuffer->pybuffer.buf, __pyx_t_18, __pyx_pybuffernd_morton_indices.diminfo[0].strides) = __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton((*__Pyx_BufPtrStrided2d(float *, __pyx_pybuffernd_pos.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_pos.diminfo[0].strides, __pyx_t_13, __pyx_pybuffernd_pos.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(float *, __pyx_pybuffernd_pos.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_pos.diminfo[0].strides, __pyx_t_15, __pyx_pybuffernd_pos.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(float *, __pyx_pybuffernd_pos.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_pos.diminfo[0].strides, __pyx_t_17, __pyx_pybuffernd_pos.diminfo[1].strides)), __pyx_v_LE, __pyx_v_RE, __pyx_v_self->index_order1);
  }

  /* "yt/geometry/particle_oct_container.pyx":748
 *         # Loop over positions skipping those outside the domain
 *         cdef np.ndarray[np.uint64_t, ndim=1, cast=True] sorted_order
 *         if hsml is None:             # <<<<<<<<<<<<<<
 *             # casting to uint64 for compatibility with 32 bits systems
 *             # see https://github.com/yt-project/yt/issues/3656
 */
  __pyx_t_19 = (((PyObject *)__pyx_v_hsml) == Py_None);
  __pyx_t_20 = (__pyx_t_19 != 0);
  if (__pyx_t_20) {

    /* "yt/geometry/particle_oct_container.pyx":751
 *             # casting to uint64 for compatibility with 32 bits systems
 *             # see https://github.com/yt-project/yt/issues/3656
 *             sorted_order = np.argsort(morton_indices).astype(np.uint64, copy=False)             # <<<<<<<<<<<<<<
 *         else:
 *             sorted_order = np.argsort(hsml)[::-1].astype(np.uint64, copy=False)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_argsort); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, ((PyObject *)__pyx_v_morton_indices)) : __Pyx_PyObject_CallOneArg(__pyx_t_4, ((PyObject *)__pyx_v_morton_indices));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_astype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_uint64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_copy, Py_False) < 0) __PYX_ERR(0, 751, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 751, __pyx_L1_error)
    __pyx_t_21 = ((PyArrayObject *)__pyx_t_2);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sorted_order.rcbuffer->pybuffer);
      __pyx_t_22 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_sorted_order.rcbuffer->pybuffer, (PyObject*)__pyx_t_21, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 1, __pyx_stack);
      if (unlikely(__pyx_t_22 < 0)) {
        PyErr_Fetch(&__pyx_t_23, &__pyx_t_24, &__pyx_t_25);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_sorted_order.rcbuffer->pybuffer, (PyObject*)__pyx_v_sorted_order, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 1, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_23); Py_XDECREF(__pyx_t_24); Py_XDECREF(__pyx_t_25);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_23, __pyx_t_24, __pyx_t_25);
        }
        __pyx_t_23 = __pyx_t_24 = __pyx_t_25 = 0;
      }
      __pyx_pybuffernd_sorted_order.diminfo[0].strides = __pyx_pybuffernd_sorted_order.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_sorted_order.diminfo[0].shape = __pyx_pybuffernd_sorted_order.rcbuffer->pybuffer.shape[0];
      if (unlikely(__pyx_t_22 < 0)) __PYX_ERR(0, 751, __pyx_L1_error)
    }
    __pyx_t_21 = 0;
    __pyx_v_sorted_order = ((PyArrayObject *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "yt/geometry/particle_oct_container.pyx":748
 *         # Loop over positions skipping those outside the domain
 *         cdef np.ndarray[np.uint64_t, ndim=1, cast=True] sorted_order
 *         if hsml is None:             # <<<<<<<<<<<<<<
 *             # casting to uint64 for compatibility with 32 bits systems
 *             # see https://github.com/yt-project/yt/issues/3656
 */
    goto __pyx_L7;
  }

  /* "yt/geometry/particle_oct_container.pyx":753
 *             sorted_order = np.argsort(morton_indices).astype(np.uint64, copy=False)
 *         else:
 *             sorted_order = np.argsort(hsml)[::-1].astype(np.uint64, copy=False)             # <<<<<<<<<<<<<<
 *         for sorted_ind in range(sorted_order.shape[0]):
 *             p = sorted_order[sorted_ind]
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_argsort); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, ((PyObject *)__pyx_v_hsml)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_hsml));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_slice__17); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_astype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_uint64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_copy, Py_False) < 0) __PYX_ERR(0, 753, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 753, __pyx_L1_error)
    __pyx_t_21 = ((PyArrayObject *)__pyx_t_4);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sorted_order.rcbuffer->pybuffer);
      __pyx_t_22 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_sorted_order.rcbuffer->pybuffer, (PyObject*)__pyx_t_21, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 1, __pyx_stack);
      if (unlikely(__pyx_t_22 < 0)) {
        PyErr_Fetch(&__pyx_t_25, &__pyx_t_24, &__pyx_t_23);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_sorted_order.rcbuffer->pybuffer, (PyObject*)__pyx_v_sorted_order, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 1, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_25); Py_XDECREF(__pyx_t_24); Py_XDECREF(__pyx_t_23);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_25, __pyx_t_24, __pyx_t_23);
        }
        __pyx_t_25 = __pyx_t_24 = __pyx_t_23 = 0;
      }
      __pyx_pybuffernd_sorted_order.diminfo[0].strides = __pyx_pybuffernd_sorted_order.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_sorted_order.diminfo[0].shape = __pyx_pybuffernd_sorted_order.rcbuffer->pybuffer.shape[0];
      if (unlikely(__pyx_t_22 < 0)) __PYX_ERR(0, 753, __pyx_L1_error)
    }
    __pyx_t_21 = 0;
    __pyx_v_sorted_order = ((PyArrayObject *)__pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L7:;

  /* "yt/geometry/particle_oct_container.pyx":754
 *         else:
 *             sorted_order = np.argsort(hsml)[::-1].astype(np.uint64, copy=False)
 *         for sorted_ind in range(sorted_order.shape[0]):             # <<<<<<<<<<<<<<
 *             p = sorted_order[sorted_ind]
 *             skip = 0
 */
  __pyx_t_9 = (__pyx_v_sorted_order->dimensions[0]);
  __pyx_t_10 = __pyx_t_9;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_sorted_ind = __pyx_t_11;

    /* "yt/geometry/particle_oct_container.pyx":755
 *             sorted_order = np.argsort(hsml)[::-1].astype(np.uint64, copy=False)
 *         for sorted_ind in range(sorted_order.shape[0]):
 *             p = sorted_order[sorted_ind]             # <<<<<<<<<<<<<<
 *             skip = 0
 *             for i in range(3):
 */
    __pyx_t_16 = __pyx_v_sorted_ind;
    __pyx_v_p = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_sorted_order.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_sorted_order.diminfo[0].strides));

    /* "yt/geometry/particle_oct_container.pyx":756
 *         for sorted_ind in range(sorted_order.shape[0]):
 *             p = sorted_order[sorted_ind]
 *             skip = 0             # <<<<<<<<<<<<<<
 *             for i in range(3):
 *                 axiter[i][1] = 999
 */
    __pyx_v_skip = 0;

    /* "yt/geometry/particle_oct_container.pyx":757
 *             p = sorted_order[sorted_ind]
 *             skip = 0
 *             for i in range(3):             # <<<<<<<<<<<<<<
 *                 axiter[i][1] = 999
 *                 if not (LE[i] <= pos[p, i] < RE[i]):
 */
    for (__pyx_t_7 = 0; __pyx_t_7 < 3; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "yt/geometry/particle_oct_container.pyx":758
 *             skip = 0
 *             for i in range(3):
 *                 axiter[i][1] = 999             # <<<<<<<<<<<<<<
 *                 if not (LE[i] <= pos[p, i] < RE[i]):
 *                     skip = 1
 */
      ((__pyx_v_axiter[__pyx_v_i])[1]) = 0x3E7;

      /* "yt/geometry/particle_oct_container.pyx":759
 *             for i in range(3):
 *                 axiter[i][1] = 999
 *                 if not (LE[i] <= pos[p, i] < RE[i]):             # <<<<<<<<<<<<<<
 *                     skip = 1
 *                     break
 */
      __pyx_t_16 = __pyx_v_p;
      __pyx_t_26 = __pyx_v_i;
      __pyx_t_27 = (*__Pyx_BufPtrStrided2d(float *, __pyx_pybuffernd_pos.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_pos.diminfo[0].strides, __pyx_t_26, __pyx_pybuffernd_pos.diminfo[1].strides));
      __pyx_t_20 = ((__pyx_v_LE[__pyx_v_i]) <= __pyx_t_27);
      if (__pyx_t_20) {
        __pyx_t_20 = (__pyx_t_27 < (__pyx_v_RE[__pyx_v_i]));
      }
      __pyx_t_19 = ((!(__pyx_t_20 != 0)) != 0);
      if (__pyx_t_19) {

        /* "yt/geometry/particle_oct_container.pyx":760
 *                 axiter[i][1] = 999
 *                 if not (LE[i] <= pos[p, i] < RE[i]):
 *                     skip = 1             # <<<<<<<<<<<<<<
 *                     break
 *                 ppos[i] = pos[p,i]
 */
        __pyx_v_skip = 1;

        /* "yt/geometry/particle_oct_container.pyx":761
 *                 if not (LE[i] <= pos[p, i] < RE[i]):
 *                     skip = 1
 *                     break             # <<<<<<<<<<<<<<
 *                 ppos[i] = pos[p,i]
 *             if skip == 1: continue
 */
        goto __pyx_L11_break;

        /* "yt/geometry/particle_oct_container.pyx":759
 *             for i in range(3):
 *                 axiter[i][1] = 999
 *                 if not (LE[i] <= pos[p, i] < RE[i]):             # <<<<<<<<<<<<<<
 *                     skip = 1
 *                     break
 */
      }

      /* "yt/geometry/particle_oct_container.pyx":762
 *                     skip = 1
 *                     break
 *                 ppos[i] = pos[p,i]             # <<<<<<<<<<<<<<
 *             if skip == 1: continue
 *             # Only look if collision at coarse index
 */
      __pyx_t_16 = __pyx_v_p;
      __pyx_t_26 = __pyx_v_i;
      (__pyx_v_ppos[__pyx_v_i]) = (*__Pyx_BufPtrStrided2d(float *, __pyx_pybuffernd_pos.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_pos.diminfo[0].strides, __pyx_t_26, __pyx_pybuffernd_pos.diminfo[1].strides));
    }
    __pyx_L11_break:;

    /* "yt/geometry/particle_oct_container.pyx":763
 *                     break
 *                 ppos[i] = pos[p,i]
 *             if skip == 1: continue             # <<<<<<<<<<<<<<
 *             # Only look if collision at coarse index
 *             mi1 = bounded_morton_split_dds(ppos[0], ppos[1], ppos[2], LE,
 */
    __pyx_t_19 = ((__pyx_v_skip == 1) != 0);
    if (__pyx_t_19) {
      goto __pyx_L8_continue;
    }

    /* "yt/geometry/particle_oct_container.pyx":765
 *             if skip == 1: continue
 *             # Only look if collision at coarse index
 *             mi1 = bounded_morton_split_dds(ppos[0], ppos[1], ppos[2], LE,             # <<<<<<<<<<<<<<
 *                                            dds1, mi_split1)
 *             if hsml is None:
 */
    __pyx_v_mi1 = __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_dds((__pyx_v_ppos[0]), (__pyx_v_ppos[1]), (__pyx_v_ppos[2]), __pyx_v_LE, __pyx_v_dds1, __pyx_v_mi_split1);

    /* "yt/geometry/particle_oct_container.pyx":767
 *             mi1 = bounded_morton_split_dds(ppos[0], ppos[1], ppos[2], LE,
 *                                            dds1, mi_split1)
 *             if hsml is None:             # <<<<<<<<<<<<<<
 *                 if mask[mi1] < mask_threshold \
 *                         or particle_counts[mi1] < count_threshold:
 */
    __pyx_t_19 = (((PyObject *)__pyx_v_hsml) == Py_None);
    __pyx_t_20 = (__pyx_t_19 != 0);
    if (__pyx_t_20) {

      /* "yt/geometry/particle_oct_container.pyx":768
 *                                            dds1, mi_split1)
 *             if hsml is None:
 *                 if mask[mi1] < mask_threshold \             # <<<<<<<<<<<<<<
 *                         or particle_counts[mi1] < count_threshold:
 *                     continue
 */
      __pyx_t_7 = __pyx_v_mi1;
      __pyx_t_19 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_mask.diminfo[0].strides)) < __pyx_v_mask_threshold) != 0);
      if (!__pyx_t_19) {
      } else {
        __pyx_t_20 = __pyx_t_19;
        goto __pyx_L16_bool_binop_done;
      }

      /* "yt/geometry/particle_oct_container.pyx":769
 *             if hsml is None:
 *                 if mask[mi1] < mask_threshold \
 *                         or particle_counts[mi1] < count_threshold:             # <<<<<<<<<<<<<<
 *                     continue
 *                 # Determine sub index within cell of primary index
 */
      __pyx_t_7 = __pyx_v_mi1;
      __pyx_t_19 = (((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_particle_counts.data + __pyx_t_7 * __pyx_v_particle_counts.strides[0]) ))) < __pyx_v_count_threshold) != 0);
      __pyx_t_20 = __pyx_t_19;
      __pyx_L16_bool_binop_done:;

      /* "yt/geometry/particle_oct_container.pyx":768
 *                                            dds1, mi_split1)
 *             if hsml is None:
 *                 if mask[mi1] < mask_threshold \             # <<<<<<<<<<<<<<
 *                         or particle_counts[mi1] < count_threshold:
 *                     continue
 */
      if (__pyx_t_20) {

        /* "yt/geometry/particle_oct_container.pyx":770
 *                 if mask[mi1] < mask_threshold \
 *                         or particle_counts[mi1] < count_threshold:
 *                     continue             # <<<<<<<<<<<<<<
 *                 # Determine sub index within cell of primary index
 *                 mi2 = bounded_morton_split_relative_dds(
 */
        goto __pyx_L8_continue;

        /* "yt/geometry/particle_oct_container.pyx":768
 *                                            dds1, mi_split1)
 *             if hsml is None:
 *                 if mask[mi1] < mask_threshold \             # <<<<<<<<<<<<<<
 *                         or particle_counts[mi1] < count_threshold:
 *                     continue
 */
      }

      /* "yt/geometry/particle_oct_container.pyx":772
 *                     continue
 *                 # Determine sub index within cell of primary index
 *                 mi2 = bounded_morton_split_relative_dds(             # <<<<<<<<<<<<<<
 *                     ppos[0], ppos[1], ppos[2], LE, dds1, dds2, mi_split2)
 *                 if refined_count[mi1] == 0:
 */
      __pyx_v_mi2 = __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_relative_dds((__pyx_v_ppos[0]), (__pyx_v_ppos[1]), (__pyx_v_ppos[2]), __pyx_v_LE, __pyx_v_dds1, __pyx_v_dds2, __pyx_v_mi_split2);

      /* "yt/geometry/particle_oct_container.pyx":774
 *                 mi2 = bounded_morton_split_relative_dds(
 *                     ppos[0], ppos[1], ppos[2], LE, dds1, dds2, mi_split2)
 *                 if refined_count[mi1] == 0:             # <<<<<<<<<<<<<<
 *                     coarse_refined_map[mi1].padWithZeroes(max_mi2_elements)
 *                 if not coarse_refined_map[mi1].get(mi2):
 */
      __pyx_t_7 = __pyx_v_mi1;
      __pyx_t_20 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_refined_count.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_refined_count.diminfo[0].strides)) == 0) != 0);
      if (__pyx_t_20) {

        /* "yt/geometry/particle_oct_container.pyx":775
 *                     ppos[0], ppos[1], ppos[2], LE, dds1, dds2, mi_split2)
 *                 if refined_count[mi1] == 0:
 *                     coarse_refined_map[mi1].padWithZeroes(max_mi2_elements)             # <<<<<<<<<<<<<<
 *                 if not coarse_refined_map[mi1].get(mi2):
 *                     coarse_refined_map[mi1].set(mi2)
 */
        (void)((__pyx_v_coarse_refined_map[__pyx_v_mi1]).padWithZeroes(__pyx_v_max_mi2_elements));

        /* "yt/geometry/particle_oct_container.pyx":774
 *                 mi2 = bounded_morton_split_relative_dds(
 *                     ppos[0], ppos[1], ppos[2], LE, dds1, dds2, mi_split2)
 *                 if refined_count[mi1] == 0:             # <<<<<<<<<<<<<<
 *                     coarse_refined_map[mi1].padWithZeroes(max_mi2_elements)
 *                 if not coarse_refined_map[mi1].get(mi2):
 */
      }

      /* "yt/geometry/particle_oct_container.pyx":776
 *                 if refined_count[mi1] == 0:
 *                     coarse_refined_map[mi1].padWithZeroes(max_mi2_elements)
 *                 if not coarse_refined_map[mi1].get(mi2):             # <<<<<<<<<<<<<<
 *                     coarse_refined_map[mi1].set(mi2)
 *                     refined_count[mi1] += 1
 */
      __pyx_t_20 = ((!((__pyx_v_coarse_refined_map[__pyx_v_mi1]).get(__pyx_v_mi2) != 0)) != 0);
      if (__pyx_t_20) {

        /* "yt/geometry/particle_oct_container.pyx":777
 *                     coarse_refined_map[mi1].padWithZeroes(max_mi2_elements)
 *                 if not coarse_refined_map[mi1].get(mi2):
 *                     coarse_refined_map[mi1].set(mi2)             # <<<<<<<<<<<<<<
 *                     refined_count[mi1] += 1
 *             else: # only hit if we have smoothing lengths.
 */
        (__pyx_v_coarse_refined_map[__pyx_v_mi1]).set(__pyx_v_mi2);

        /* "yt/geometry/particle_oct_container.pyx":778
 *                 if not coarse_refined_map[mi1].get(mi2):
 *                     coarse_refined_map[mi1].set(mi2)
 *                     refined_count[mi1] += 1             # <<<<<<<<<<<<<<
 *             else: # only hit if we have smoothing lengths.
 *                 # We have to do essentially the identical process to in the coarse indexing,
 */
        __pyx_t_7 = __pyx_v_mi1;
        *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_refined_count.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_refined_count.diminfo[0].strides) += 1;

        /* "yt/geometry/particle_oct_container.pyx":776
 *                 if refined_count[mi1] == 0:
 *                     coarse_refined_map[mi1].padWithZeroes(max_mi2_elements)
 *                 if not coarse_refined_map[mi1].get(mi2):             # <<<<<<<<<<<<<<
 *                     coarse_refined_map[mi1].set(mi2)
 *                     refined_count[mi1] += 1
 */
      }

      /* "yt/geometry/particle_oct_container.pyx":767
 *             mi1 = bounded_morton_split_dds(ppos[0], ppos[1], ppos[2], LE,
 *                                            dds1, mi_split1)
 *             if hsml is None:             # <<<<<<<<<<<<<<
 *                 if mask[mi1] < mask_threshold \
 *                         or particle_counts[mi1] < count_threshold:
 */
      goto __pyx_L14;
    }

    /* "yt/geometry/particle_oct_container.pyx":783
 *                 # except here we need to fill in all the subranges as well as the coarse ranges
 *                 # Note that we are also doing the null case, where we do no shifting
 *                 radius = hsml[p]             # <<<<<<<<<<<<<<
 *                 #if mask[mi1] <= 4: # only one thing in this area
 *                 #    continue
 */
    /*else*/ {
      __pyx_t_16 = __pyx_v_p;
      __pyx_v_radius = (*__Pyx_BufPtrStrided1d(float *, __pyx_pybuffernd_hsml.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_hsml.diminfo[0].strides));

      /* "yt/geometry/particle_oct_container.pyx":786
 *                 #if mask[mi1] <= 4: # only one thing in this area
 *                 #    continue
 *                 for i in range(3):             # <<<<<<<<<<<<<<
 *                     if PER[i] and ppos[i] - radius < LE[i]:
 *                         axiter[i][1] = +1
 */
      for (__pyx_t_7 = 0; __pyx_t_7 < 3; __pyx_t_7+=1) {
        __pyx_v_i = __pyx_t_7;

        /* "yt/geometry/particle_oct_container.pyx":787
 *                 #    continue
 *                 for i in range(3):
 *                     if PER[i] and ppos[i] - radius < LE[i]:             # <<<<<<<<<<<<<<
 *                         axiter[i][1] = +1
 *                         axiterv[i][1] = DW[i]
 */
        __pyx_t_19 = ((__pyx_v_PER[__pyx_v_i]) != 0);
        if (__pyx_t_19) {
        } else {
          __pyx_t_20 = __pyx_t_19;
          goto __pyx_L23_bool_binop_done;
        }
        __pyx_t_19 = ((((__pyx_v_ppos[__pyx_v_i]) - __pyx_v_radius) < (__pyx_v_LE[__pyx_v_i])) != 0);
        __pyx_t_20 = __pyx_t_19;
        __pyx_L23_bool_binop_done:;
        if (__pyx_t_20) {

          /* "yt/geometry/particle_oct_container.pyx":788
 *                 for i in range(3):
 *                     if PER[i] and ppos[i] - radius < LE[i]:
 *                         axiter[i][1] = +1             # <<<<<<<<<<<<<<
 *                         axiterv[i][1] = DW[i]
 *                     elif PER[i] and ppos[i] + radius > RE[i]:
 */
          ((__pyx_v_axiter[__pyx_v_i])[1]) = 1;

          /* "yt/geometry/particle_oct_container.pyx":789
 *                     if PER[i] and ppos[i] - radius < LE[i]:
 *                         axiter[i][1] = +1
 *                         axiterv[i][1] = DW[i]             # <<<<<<<<<<<<<<
 *                     elif PER[i] and ppos[i] + radius > RE[i]:
 *                         axiter[i][1] = -1
 */
          ((__pyx_v_axiterv[__pyx_v_i])[1]) = (__pyx_v_DW[__pyx_v_i]);

          /* "yt/geometry/particle_oct_container.pyx":787
 *                 #    continue
 *                 for i in range(3):
 *                     if PER[i] and ppos[i] - radius < LE[i]:             # <<<<<<<<<<<<<<
 *                         axiter[i][1] = +1
 *                         axiterv[i][1] = DW[i]
 */
          goto __pyx_L22;
        }

        /* "yt/geometry/particle_oct_container.pyx":790
 *                         axiter[i][1] = +1
 *                         axiterv[i][1] = DW[i]
 *                     elif PER[i] and ppos[i] + radius > RE[i]:             # <<<<<<<<<<<<<<
 *                         axiter[i][1] = -1
 *                         axiterv[i][1] = -DW[i]
 */
        __pyx_t_19 = ((__pyx_v_PER[__pyx_v_i]) != 0);
        if (__pyx_t_19) {
        } else {
          __pyx_t_20 = __pyx_t_19;
          goto __pyx_L25_bool_binop_done;
        }
        __pyx_t_19 = ((((__pyx_v_ppos[__pyx_v_i]) + __pyx_v_radius) > (__pyx_v_RE[__pyx_v_i])) != 0);
        __pyx_t_20 = __pyx_t_19;
        __pyx_L25_bool_binop_done:;
        if (__pyx_t_20) {

          /* "yt/geometry/particle_oct_container.pyx":791
 *                         axiterv[i][1] = DW[i]
 *                     elif PER[i] and ppos[i] + radius > RE[i]:
 *                         axiter[i][1] = -1             # <<<<<<<<<<<<<<
 *                         axiterv[i][1] = -DW[i]
 *                 for xi in range(2):
 */
          ((__pyx_v_axiter[__pyx_v_i])[1]) = -1;

          /* "yt/geometry/particle_oct_container.pyx":792
 *                     elif PER[i] and ppos[i] + radius > RE[i]:
 *                         axiter[i][1] = -1
 *                         axiterv[i][1] = -DW[i]             # <<<<<<<<<<<<<<
 *                 for xi in range(2):
 *                     if axiter[0][xi] == 999: continue
 */
          ((__pyx_v_axiterv[__pyx_v_i])[1]) = (-(__pyx_v_DW[__pyx_v_i]));

          /* "yt/geometry/particle_oct_container.pyx":790
 *                         axiter[i][1] = +1
 *                         axiterv[i][1] = DW[i]
 *                     elif PER[i] and ppos[i] + radius > RE[i]:             # <<<<<<<<<<<<<<
 *                         axiter[i][1] = -1
 *                         axiterv[i][1] = -DW[i]
 */
        }
        __pyx_L22:;
      }

      /* "yt/geometry/particle_oct_container.pyx":793
 *                         axiter[i][1] = -1
 *                         axiterv[i][1] = -DW[i]
 *                 for xi in range(2):             # <<<<<<<<<<<<<<
 *                     if axiter[0][xi] == 999: continue
 *                     s_ppos[0] = ppos[0] + axiterv[0][xi]
 */
      for (__pyx_t_28 = 0; __pyx_t_28 < 2; __pyx_t_28+=1) {
        __pyx_v_xi = __pyx_t_28;

        /* "yt/geometry/particle_oct_container.pyx":794
 *                         axiterv[i][1] = -DW[i]
 *                 for xi in range(2):
 *                     if axiter[0][xi] == 999: continue             # <<<<<<<<<<<<<<
 *                     s_ppos[0] = ppos[0] + axiterv[0][xi]
 *                     for yi in range(2):
 */
        __pyx_t_20 = ((((__pyx_v_axiter[0])[__pyx_v_xi]) == 0x3E7) != 0);
        if (__pyx_t_20) {
          goto __pyx_L27_continue;
        }

        /* "yt/geometry/particle_oct_container.pyx":795
 *                 for xi in range(2):
 *                     if axiter[0][xi] == 999: continue
 *                     s_ppos[0] = ppos[0] + axiterv[0][xi]             # <<<<<<<<<<<<<<
 *                     for yi in range(2):
 *                         if axiter[1][yi] == 999: continue
 */
        (__pyx_v_s_ppos[0]) = ((__pyx_v_ppos[0]) + ((__pyx_v_axiterv[0])[__pyx_v_xi]));

        /* "yt/geometry/particle_oct_container.pyx":796
 *                     if axiter[0][xi] == 999: continue
 *                     s_ppos[0] = ppos[0] + axiterv[0][xi]
 *                     for yi in range(2):             # <<<<<<<<<<<<<<
 *                         if axiter[1][yi] == 999: continue
 *                         s_ppos[1] = ppos[1] + axiterv[1][yi]
 */
        for (__pyx_t_29 = 0; __pyx_t_29 < 2; __pyx_t_29+=1) {
          __pyx_v_yi = __pyx_t_29;

          /* "yt/geometry/particle_oct_container.pyx":797
 *                     s_ppos[0] = ppos[0] + axiterv[0][xi]
 *                     for yi in range(2):
 *                         if axiter[1][yi] == 999: continue             # <<<<<<<<<<<<<<
 *                         s_ppos[1] = ppos[1] + axiterv[1][yi]
 *                         for zi in range(2):
 */
          __pyx_t_20 = ((((__pyx_v_axiter[1])[__pyx_v_yi]) == 0x3E7) != 0);
          if (__pyx_t_20) {
            goto __pyx_L30_continue;
          }

          /* "yt/geometry/particle_oct_container.pyx":798
 *                     for yi in range(2):
 *                         if axiter[1][yi] == 999: continue
 *                         s_ppos[1] = ppos[1] + axiterv[1][yi]             # <<<<<<<<<<<<<<
 *                         for zi in range(2):
 *                             if axiter[2][zi] == 999: continue
 */
          (__pyx_v_s_ppos[1]) = ((__pyx_v_ppos[1]) + ((__pyx_v_axiterv[1])[__pyx_v_yi]));

          /* "yt/geometry/particle_oct_container.pyx":799
 *                         if axiter[1][yi] == 999: continue
 *                         s_ppos[1] = ppos[1] + axiterv[1][yi]
 *                         for zi in range(2):             # <<<<<<<<<<<<<<
 *                             if axiter[2][zi] == 999: continue
 *                             s_ppos[2] = ppos[2] + axiterv[2][zi]
 */
          for (__pyx_t_30 = 0; __pyx_t_30 < 2; __pyx_t_30+=1) {
            __pyx_v_zi = __pyx_t_30;

            /* "yt/geometry/particle_oct_container.pyx":800
 *                         s_ppos[1] = ppos[1] + axiterv[1][yi]
 *                         for zi in range(2):
 *                             if axiter[2][zi] == 999: continue             # <<<<<<<<<<<<<<
 *                             s_ppos[2] = ppos[2] + axiterv[2][zi]
 *                             # OK, now we compute the left and right edges for this shift.
 */
            __pyx_t_20 = ((((__pyx_v_axiter[2])[__pyx_v_zi]) == 0x3E7) != 0);
            if (__pyx_t_20) {
              goto __pyx_L33_continue;
            }

            /* "yt/geometry/particle_oct_container.pyx":801
 *                         for zi in range(2):
 *                             if axiter[2][zi] == 999: continue
 *                             s_ppos[2] = ppos[2] + axiterv[2][zi]             # <<<<<<<<<<<<<<
 *                             # OK, now we compute the left and right edges for this shift.
 *                             for i in range(3):
 */
            (__pyx_v_s_ppos[2]) = ((__pyx_v_ppos[2]) + ((__pyx_v_axiterv[2])[__pyx_v_zi]));

            /* "yt/geometry/particle_oct_container.pyx":803
 *                             s_ppos[2] = ppos[2] + axiterv[2][zi]
 *                             # OK, now we compute the left and right edges for this shift.
 *                             for i in range(3):             # <<<<<<<<<<<<<<
 *                                 # casting to int64 is not nice but is so we can have negative values we clip
 *                                 clip_pos_l[i] = fmax(s_ppos[i] - radius, LE[i] + dds1[i]/10)
 */
            for (__pyx_t_7 = 0; __pyx_t_7 < 3; __pyx_t_7+=1) {
              __pyx_v_i = __pyx_t_7;

              /* "yt/geometry/particle_oct_container.pyx":805
 *                             for i in range(3):
 *                                 # casting to int64 is not nice but is so we can have negative values we clip
 *                                 clip_pos_l[i] = fmax(s_ppos[i] - radius, LE[i] + dds1[i]/10)             # <<<<<<<<<<<<<<
 *                                 clip_pos_r[i] = fmin(s_ppos[i] + radius, RE[i] - dds1[i]/10)
 * 
 */
              (__pyx_v_clip_pos_l[__pyx_v_i]) = __pyx_f_2yt_9utilities_3lib_8fp_utils_fmax(((__pyx_v_s_ppos[__pyx_v_i]) - __pyx_v_radius), ((__pyx_v_LE[__pyx_v_i]) + ((__pyx_v_dds1[__pyx_v_i]) / 10.0)));

              /* "yt/geometry/particle_oct_container.pyx":806
 *                                 # casting to int64 is not nice but is so we can have negative values we clip
 *                                 clip_pos_l[i] = fmax(s_ppos[i] - radius, LE[i] + dds1[i]/10)
 *                                 clip_pos_r[i] = fmin(s_ppos[i] + radius, RE[i] - dds1[i]/10)             # <<<<<<<<<<<<<<
 * 
 *                             bounded_morton_split_dds(clip_pos_l[0], clip_pos_l[1], clip_pos_l[2], LE, dds1, bounds[0])
 */
              (__pyx_v_clip_pos_r[__pyx_v_i]) = __pyx_f_2yt_9utilities_3lib_8fp_utils_fmin(((__pyx_v_s_ppos[__pyx_v_i]) + __pyx_v_radius), ((__pyx_v_RE[__pyx_v_i]) - ((__pyx_v_dds1[__pyx_v_i]) / 10.0)));
            }

            /* "yt/geometry/particle_oct_container.pyx":808
 *                                 clip_pos_r[i] = fmin(s_ppos[i] + radius, RE[i] - dds1[i]/10)
 * 
 *                             bounded_morton_split_dds(clip_pos_l[0], clip_pos_l[1], clip_pos_l[2], LE, dds1, bounds[0])             # <<<<<<<<<<<<<<
 *                             bounded_morton_split_dds(clip_pos_r[0], clip_pos_r[1], clip_pos_r[2], LE, dds1, bounds[1])
 * 
 */
            (void)(__pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_dds((__pyx_v_clip_pos_l[0]), (__pyx_v_clip_pos_l[1]), (__pyx_v_clip_pos_l[2]), __pyx_v_LE, __pyx_v_dds1, (__pyx_v_bounds[0])));

            /* "yt/geometry/particle_oct_container.pyx":809
 * 
 *                             bounded_morton_split_dds(clip_pos_l[0], clip_pos_l[1], clip_pos_l[2], LE, dds1, bounds[0])
 *                             bounded_morton_split_dds(clip_pos_r[0], clip_pos_r[1], clip_pos_r[2], LE, dds1, bounds[1])             # <<<<<<<<<<<<<<
 * 
 *                             # We go to the upper bound plus one so that we have *inclusive* loops -- the upper bound
 */
            (void)(__pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_dds((__pyx_v_clip_pos_r[0]), (__pyx_v_clip_pos_r[1]), (__pyx_v_clip_pos_r[2]), __pyx_v_LE, __pyx_v_dds1, (__pyx_v_bounds[1])));

            /* "yt/geometry/particle_oct_container.pyx":818
 *                             # greater than the left edge and less than the
 *                             # right edge, we are fully enclosed.
 *                             for xex in range(bounds[0][0], bounds[1][0] + 1):             # <<<<<<<<<<<<<<
 *                                 for yex in range(bounds[0][1], bounds[1][1] + 1):
 *                                     for zex in range(bounds[0][2], bounds[1][2] + 1):
 */
            __pyx_t_7 = (((__pyx_v_bounds[1])[0]) + 1);
            __pyx_t_26 = __pyx_t_7;
            for (__pyx_t_31 = ((__pyx_v_bounds[0])[0]); __pyx_t_31 < __pyx_t_26; __pyx_t_31+=1) {
              __pyx_v_xex = __pyx_t_31;

              /* "yt/geometry/particle_oct_container.pyx":819
 *                             # right edge, we are fully enclosed.
 *                             for xex in range(bounds[0][0], bounds[1][0] + 1):
 *                                 for yex in range(bounds[0][1], bounds[1][1] + 1):             # <<<<<<<<<<<<<<
 *                                     for zex in range(bounds[0][2], bounds[1][2] + 1):
 *                                         miex1 = encode_morton_64bit(xex, yex, zex)
 */
              __pyx_t_32 = (((__pyx_v_bounds[1])[1]) + 1);
              __pyx_t_33 = __pyx_t_32;
              for (__pyx_t_34 = ((__pyx_v_bounds[0])[1]); __pyx_t_34 < __pyx_t_33; __pyx_t_34+=1) {
                __pyx_v_yex = __pyx_t_34;

                /* "yt/geometry/particle_oct_container.pyx":820
 *                             for xex in range(bounds[0][0], bounds[1][0] + 1):
 *                                 for yex in range(bounds[0][1], bounds[1][1] + 1):
 *                                     for zex in range(bounds[0][2], bounds[1][2] + 1):             # <<<<<<<<<<<<<<
 *                                         miex1 = encode_morton_64bit(xex, yex, zex)
 *                                         if mask[miex1] < mask_threshold or \
 */
                __pyx_t_35 = (((__pyx_v_bounds[1])[2]) + 1);
                __pyx_t_36 = __pyx_t_35;
                for (__pyx_t_37 = ((__pyx_v_bounds[0])[2]); __pyx_t_37 < __pyx_t_36; __pyx_t_37+=1) {
                  __pyx_v_zex = __pyx_t_37;

                  /* "yt/geometry/particle_oct_container.pyx":821
 *                                 for yex in range(bounds[0][1], bounds[1][1] + 1):
 *                                     for zex in range(bounds[0][2], bounds[1][2] + 1):
 *                                         miex1 = encode_morton_64bit(xex, yex, zex)             # <<<<<<<<<<<<<<
 *                                         if mask[miex1] < mask_threshold or \
 *                                                 particle_counts[miex1] < count_threshold:
 */
                  __pyx_v_miex1 = __pyx_f_2yt_9utilities_3lib_14geometry_utils_encode_morton_64bit(__pyx_v_xex, __pyx_v_yex, __pyx_v_zex);

                  /* "yt/geometry/particle_oct_container.pyx":822
 *                                     for zex in range(bounds[0][2], bounds[1][2] + 1):
 *                                         miex1 = encode_morton_64bit(xex, yex, zex)
 *                                         if mask[miex1] < mask_threshold or \             # <<<<<<<<<<<<<<
 *                                                 particle_counts[miex1] < count_threshold:
 *                                             continue
 */
                  __pyx_t_38 = __pyx_v_miex1;
                  __pyx_t_19 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_38, __pyx_pybuffernd_mask.diminfo[0].strides)) < __pyx_v_mask_threshold) != 0);
                  if (!__pyx_t_19) {
                  } else {
                    __pyx_t_20 = __pyx_t_19;
                    goto __pyx_L45_bool_binop_done;
                  }

                  /* "yt/geometry/particle_oct_container.pyx":823
 *                                         miex1 = encode_morton_64bit(xex, yex, zex)
 *                                         if mask[miex1] < mask_threshold or \
 *                                                 particle_counts[miex1] < count_threshold:             # <<<<<<<<<<<<<<
 *                                             continue
 *                                         # this explicitly requires that it be *between*
 */
                  __pyx_t_38 = __pyx_v_miex1;
                  __pyx_t_19 = (((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_particle_counts.data + __pyx_t_38 * __pyx_v_particle_counts.strides[0]) ))) < __pyx_v_count_threshold) != 0);
                  __pyx_t_20 = __pyx_t_19;
                  __pyx_L45_bool_binop_done:;

                  /* "yt/geometry/particle_oct_container.pyx":822
 *                                     for zex in range(bounds[0][2], bounds[1][2] + 1):
 *                                         miex1 = encode_morton_64bit(xex, yex, zex)
 *                                         if mask[miex1] < mask_threshold or \             # <<<<<<<<<<<<<<
 *                                                 particle_counts[miex1] < count_threshold:
 *                                             continue
 */
                  if (__pyx_t_20) {

                    /* "yt/geometry/particle_oct_container.pyx":824
 *                                         if mask[miex1] < mask_threshold or \
 *                                                 particle_counts[miex1] < count_threshold:
 *                                             continue             # <<<<<<<<<<<<<<
 *                                         # this explicitly requires that it be *between*
 *                                         # them, not overlapping
 */
                    goto __pyx_L42_continue;

                    /* "yt/geometry/particle_oct_container.pyx":822
 *                                     for zex in range(bounds[0][2], bounds[1][2] + 1):
 *                                         miex1 = encode_morton_64bit(xex, yex, zex)
 *                                         if mask[miex1] < mask_threshold or \             # <<<<<<<<<<<<<<
 *                                                 particle_counts[miex1] < count_threshold:
 *                                             continue
 */
                  }

                  /* "yt/geometry/particle_oct_container.pyx":827
 *                                         # this explicitly requires that it be *between*
 *                                         # them, not overlapping
 *                                         if xex > bounds[0][0] and xex < bounds[1][0] and \             # <<<<<<<<<<<<<<
 *                                            yex > bounds[0][1] and yex < bounds[1][1] and \
 *                                            zex > bounds[0][2] and zex < bounds[1][2]:
 */
                  __pyx_t_19 = ((__pyx_v_xex > ((__pyx_v_bounds[0])[0])) != 0);
                  if (__pyx_t_19) {
                  } else {
                    __pyx_t_20 = __pyx_t_19;
                    goto __pyx_L48_bool_binop_done;
                  }
                  __pyx_t_19 = ((__pyx_v_xex < ((__pyx_v_bounds[1])[0])) != 0);
                  if (__pyx_t_19) {
                  } else {
                    __pyx_t_20 = __pyx_t_19;
                    goto __pyx_L48_bool_binop_done;
                  }

                  /* "yt/geometry/particle_oct_container.pyx":828
 *                                         # them, not overlapping
 *                                         if xex > bounds[0][0] and xex < bounds[1][0] and \
 *                                            yex > bounds[0][1] and yex < bounds[1][1] and \             # <<<<<<<<<<<<<<
 *                                            zex > bounds[0][2] and zex < bounds[1][2]:
 *                                             fully_enclosed = 1
 */
                  __pyx_t_19 = ((__pyx_v_yex > ((__pyx_v_bounds[0])[1])) != 0);
                  if (__pyx_t_19) {
                  } else {
                    __pyx_t_20 = __pyx_t_19;
                    goto __pyx_L48_bool_binop_done;
                  }
                  __pyx_t_19 = ((__pyx_v_yex < ((__pyx_v_bounds[1])[1])) != 0);
                  if (__pyx_t_19) {
                  } else {
                    __pyx_t_20 = __pyx_t_19;
                    goto __pyx_L48_bool_binop_done;
                  }

                  /* "yt/geometry/particle_oct_container.pyx":829
 *                                         if xex > bounds[0][0] and xex < bounds[1][0] and \
 *                                            yex > bounds[0][1] and yex < bounds[1][1] and \
 *                                            zex > bounds[0][2] and zex < bounds[1][2]:             # <<<<<<<<<<<<<<
 *                                             fully_enclosed = 1
 *                                         else:
 */
                  __pyx_t_19 = ((__pyx_v_zex > ((__pyx_v_bounds[0])[2])) != 0);
                  if (__pyx_t_19) {
                  } else {
                    __pyx_t_20 = __pyx_t_19;
                    goto __pyx_L48_bool_binop_done;
                  }
                  __pyx_t_19 = ((__pyx_v_zex < ((__pyx_v_bounds[1])[2])) != 0);
                  __pyx_t_20 = __pyx_t_19;
                  __pyx_L48_bool_binop_done:;

                  /* "yt/geometry/particle_oct_container.pyx":827
 *                                         # this explicitly requires that it be *between*
 *                                         # them, not overlapping
 *                                         if xex > bounds[0][0] and xex < bounds[1][0] and \             # <<<<<<<<<<<<<<
 *                                            yex > bounds[0][1] and yex < bounds[1][1] and \
 *                                            zex > bounds[0][2] and zex < bounds[1][2]:
 */
                  if (__pyx_t_20) {

                    /* "yt/geometry/particle_oct_container.pyx":830
 *                                            yex > bounds[0][1] and yex < bounds[1][1] and \
 *                                            zex > bounds[0][2] and zex < bounds[1][2]:
 *                                             fully_enclosed = 1             # <<<<<<<<<<<<<<
 *                                         else:
 *                                             fully_enclosed = 0
 */
                    __pyx_v_fully_enclosed = 1;

                    /* "yt/geometry/particle_oct_container.pyx":827
 *                                         # this explicitly requires that it be *between*
 *                                         # them, not overlapping
 *                                         if xex > bounds[0][0] and xex < bounds[1][0] and \             # <<<<<<<<<<<<<<
 *                                            yex > bounds[0][1] and yex < bounds[1][1] and \
 *                                            zex > bounds[0][2] and zex < bounds[1][2]:
 */
                    goto __pyx_L47;
                  }

                  /* "yt/geometry/particle_oct_container.pyx":832
 *                                             fully_enclosed = 1
 *                                         else:
 *                                             fully_enclosed = 0             # <<<<<<<<<<<<<<
 *                                         # Now we need to fill our sub-range
 *                                         if refined_count[miex1] == 0:
 */
                  /*else*/ {
                    __pyx_v_fully_enclosed = 0;
                  }
                  __pyx_L47:;

                  /* "yt/geometry/particle_oct_container.pyx":834
 *                                             fully_enclosed = 0
 *                                         # Now we need to fill our sub-range
 *                                         if refined_count[miex1] == 0:             # <<<<<<<<<<<<<<
 *                                             coarse_refined_map[miex1].padWithZeroes(max_mi2_elements)
 *                                         elif refined_count[miex1] >= max_mi2_elements:
 */
                  __pyx_t_38 = __pyx_v_miex1;
                  __pyx_t_20 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_refined_count.rcbuffer->pybuffer.buf, __pyx_t_38, __pyx_pybuffernd_refined_count.diminfo[0].strides)) == 0) != 0);
                  if (__pyx_t_20) {

                    /* "yt/geometry/particle_oct_container.pyx":835
 *                                         # Now we need to fill our sub-range
 *                                         if refined_count[miex1] == 0:
 *                                             coarse_refined_map[miex1].padWithZeroes(max_mi2_elements)             # <<<<<<<<<<<<<<
 *                                         elif refined_count[miex1] >= max_mi2_elements:
 *                                             continue
 */
                    (void)((__pyx_v_coarse_refined_map[__pyx_v_miex1]).padWithZeroes(__pyx_v_max_mi2_elements));

                    /* "yt/geometry/particle_oct_container.pyx":834
 *                                             fully_enclosed = 0
 *                                         # Now we need to fill our sub-range
 *                                         if refined_count[miex1] == 0:             # <<<<<<<<<<<<<<
 *                                             coarse_refined_map[miex1].padWithZeroes(max_mi2_elements)
 *                                         elif refined_count[miex1] >= max_mi2_elements:
 */
                    goto __pyx_L54;
                  }

                  /* "yt/geometry/particle_oct_container.pyx":836
 *                                         if refined_count[miex1] == 0:
 *                                             coarse_refined_map[miex1].padWithZeroes(max_mi2_elements)
 *                                         elif refined_count[miex1] >= max_mi2_elements:             # <<<<<<<<<<<<<<
 *                                             continue
 *                                         if fully_enclosed == 1:
 */
                  __pyx_t_38 = __pyx_v_miex1;
                  __pyx_t_20 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_refined_count.rcbuffer->pybuffer.buf, __pyx_t_38, __pyx_pybuffernd_refined_count.diminfo[0].strides)) >= __pyx_v_max_mi2_elements) != 0);
                  if (__pyx_t_20) {

                    /* "yt/geometry/particle_oct_container.pyx":837
 *                                             coarse_refined_map[miex1].padWithZeroes(max_mi2_elements)
 *                                         elif refined_count[miex1] >= max_mi2_elements:
 *                                             continue             # <<<<<<<<<<<<<<
 *                                         if fully_enclosed == 1:
 *                                             nfully_enclosed += 1
 */
                    goto __pyx_L42_continue;

                    /* "yt/geometry/particle_oct_container.pyx":836
 *                                         if refined_count[miex1] == 0:
 *                                             coarse_refined_map[miex1].padWithZeroes(max_mi2_elements)
 *                                         elif refined_count[miex1] >= max_mi2_elements:             # <<<<<<<<<<<<<<
 *                                             continue
 *                                         if fully_enclosed == 1:
 */
                  }
                  __pyx_L54:;

                  /* "yt/geometry/particle_oct_container.pyx":838
 *                                         elif refined_count[miex1] >= max_mi2_elements:
 *                                             continue
 *                                         if fully_enclosed == 1:             # <<<<<<<<<<<<<<
 *                                             nfully_enclosed += 1
 *                                             coarse_refined_map[miex1].inplace_logicalxor(
 */
                  __pyx_t_20 = ((__pyx_v_fully_enclosed == 1) != 0);
                  if (__pyx_t_20) {

                    /* "yt/geometry/particle_oct_container.pyx":839
 *                                             continue
 *                                         if fully_enclosed == 1:
 *                                             nfully_enclosed += 1             # <<<<<<<<<<<<<<
 *                                             coarse_refined_map[miex1].inplace_logicalxor(
 *                                                 coarse_refined_map[miex1])
 */
                    __pyx_v_nfully_enclosed = (__pyx_v_nfully_enclosed + 1);

                    /* "yt/geometry/particle_oct_container.pyx":840
 *                                         if fully_enclosed == 1:
 *                                             nfully_enclosed += 1
 *                                             coarse_refined_map[miex1].inplace_logicalxor(             # <<<<<<<<<<<<<<
 *                                                 coarse_refined_map[miex1])
 *                                             coarse_refined_map[miex1].inplace_logicalnot()
 */
                    (__pyx_v_coarse_refined_map[__pyx_v_miex1]).inplace_logicalxor((__pyx_v_coarse_refined_map[__pyx_v_miex1]));

                    /* "yt/geometry/particle_oct_container.pyx":842
 *                                             coarse_refined_map[miex1].inplace_logicalxor(
 *                                                 coarse_refined_map[miex1])
 *                                             coarse_refined_map[miex1].inplace_logicalnot()             # <<<<<<<<<<<<<<
 *                                             refined_count[miex1] = max_mi2_elements
 *                                             continue
 */
                    (__pyx_v_coarse_refined_map[__pyx_v_miex1]).inplace_logicalnot();

                    /* "yt/geometry/particle_oct_container.pyx":843
 *                                                 coarse_refined_map[miex1])
 *                                             coarse_refined_map[miex1].inplace_logicalnot()
 *                                             refined_count[miex1] = max_mi2_elements             # <<<<<<<<<<<<<<
 *                                             continue
 *                                         n_calls += 1
 */
                    __pyx_t_38 = __pyx_v_miex1;
                    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_refined_count.rcbuffer->pybuffer.buf, __pyx_t_38, __pyx_pybuffernd_refined_count.diminfo[0].strides) = __pyx_v_max_mi2_elements;

                    /* "yt/geometry/particle_oct_container.pyx":844
 *                                             coarse_refined_map[miex1].inplace_logicalnot()
 *                                             refined_count[miex1] = max_mi2_elements
 *                                             continue             # <<<<<<<<<<<<<<
 *                                         n_calls += 1
 *                                         refined_count[miex1] += self.__fill_refined_ranges(s_ppos, radius, LE, RE,
 */
                    goto __pyx_L42_continue;

                    /* "yt/geometry/particle_oct_container.pyx":838
 *                                         elif refined_count[miex1] >= max_mi2_elements:
 *                                             continue
 *                                         if fully_enclosed == 1:             # <<<<<<<<<<<<<<
 *                                             nfully_enclosed += 1
 *                                             coarse_refined_map[miex1].inplace_logicalxor(
 */
                  }

                  /* "yt/geometry/particle_oct_container.pyx":845
 *                                             refined_count[miex1] = max_mi2_elements
 *                                             continue
 *                                         n_calls += 1             # <<<<<<<<<<<<<<
 *                                         refined_count[miex1] += self.__fill_refined_ranges(s_ppos, radius, LE, RE,
 *                                                                    dds1, xex, yex, zex,
 */
                  __pyx_v_n_calls = (__pyx_v_n_calls + 1);

                  /* "yt/geometry/particle_oct_container.pyx":846
 *                                             continue
 *                                         n_calls += 1
 *                                         refined_count[miex1] += self.__fill_refined_ranges(s_ppos, radius, LE, RE,             # <<<<<<<<<<<<<<
 *                                                                    dds1, xex, yex, zex,
 *                                                                    dds2,
 */
                  __pyx_t_16 = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self->__pyx_vtab)->__pyx___fill_refined_ranges(__pyx_v_self, __pyx_v_s_ppos, __pyx_v_radius, __pyx_v_LE, __pyx_v_RE, __pyx_v_dds1, __pyx_v_xex, __pyx_v_yex, __pyx_v_zex, __pyx_v_dds2, (__pyx_v_coarse_refined_map[__pyx_v_miex1])); if (unlikely(__pyx_t_16 == ((__pyx_t_5numpy_int64_t)-1LL))) __PYX_ERR(0, 846, __pyx_L1_error)
                  __pyx_t_38 = __pyx_v_miex1;
                  *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_refined_count.rcbuffer->pybuffer.buf, __pyx_t_38, __pyx_pybuffernd_refined_count.diminfo[0].strides) += __pyx_t_16;
                  __pyx_L42_continue:;
                }
              }
            }
            __pyx_L33_continue:;
          }
          __pyx_L30_continue:;
        }
        __pyx_L27_continue:;
      }
    }
    __pyx_L14:;
    __pyx_L8_continue:;
  }

  /* "yt/geometry/particle_oct_container.pyx":851
 *                                                                    coarse_refined_map[miex1])
 *         cdef np.uint64_t vec_i
 *         cdef bool_array *buf = NULL             # <<<<<<<<<<<<<<
 *         cdef ewah_word_type w
 *         this_collection = BoolArrayCollection()
 */
  __pyx_v_buf = NULL;

  /* "yt/geometry/particle_oct_container.pyx":853
 *         cdef bool_array *buf = NULL
 *         cdef ewah_word_type w
 *         this_collection = BoolArrayCollection()             # <<<<<<<<<<<<<<
 *         cdef ewah_bool_array *refined_arr = NULL
 *         for it1 in coarse_refined_map:
 */
  __pyx_t_4 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_this_collection = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/particle_oct_container.pyx":854
 *         cdef ewah_word_type w
 *         this_collection = BoolArrayCollection()
 *         cdef ewah_bool_array *refined_arr = NULL             # <<<<<<<<<<<<<<
 *         for it1 in coarse_refined_map:
 *             mi1 = it1.first
 */
  __pyx_v_refined_arr = NULL;

  /* "yt/geometry/particle_oct_container.pyx":855
 *         this_collection = BoolArrayCollection()
 *         cdef ewah_bool_array *refined_arr = NULL
 *         for it1 in coarse_refined_map:             # <<<<<<<<<<<<<<
 *             mi1 = it1.first
 *             refined_arr = &this_collection.ewah_coll[0][mi1]
 */
  __pyx_t_39 = __pyx_v_coarse_refined_map.begin();
  for (;;) {
    if (!(__pyx_t_39 != __pyx_v_coarse_refined_map.end())) break;
    __pyx_t_40 = *__pyx_t_39;
    ++__pyx_t_39;
    __pyx_v_it1 = __pyx_t_40;

    /* "yt/geometry/particle_oct_container.pyx":856
 *         cdef ewah_bool_array *refined_arr = NULL
 *         for it1 in coarse_refined_map:
 *             mi1 = it1.first             # <<<<<<<<<<<<<<
 *             refined_arr = &this_collection.ewah_coll[0][mi1]
 *             this_collection.ewah_keys[0].set(mi1)
 */
    __pyx_t_7 = __pyx_v_it1.first;
    __pyx_v_mi1 = __pyx_t_7;

    /* "yt/geometry/particle_oct_container.pyx":857
 *         for it1 in coarse_refined_map:
 *             mi1 = it1.first
 *             refined_arr = &this_collection.ewah_coll[0][mi1]             # <<<<<<<<<<<<<<
 *             this_collection.ewah_keys[0].set(mi1)
 *             this_collection.ewah_refn[0].set(mi1)
 */
    __pyx_v_refined_arr = (&((__pyx_v_this_collection->ewah_coll[0])[__pyx_v_mi1]));

    /* "yt/geometry/particle_oct_container.pyx":858
 *             mi1 = it1.first
 *             refined_arr = &this_collection.ewah_coll[0][mi1]
 *             this_collection.ewah_keys[0].set(mi1)             # <<<<<<<<<<<<<<
 *             this_collection.ewah_refn[0].set(mi1)
 *             buf = &it1.second
 */
    (void)((__pyx_v_this_collection->ewah_keys[0]).set(__pyx_v_mi1));

    /* "yt/geometry/particle_oct_container.pyx":859
 *             refined_arr = &this_collection.ewah_coll[0][mi1]
 *             this_collection.ewah_keys[0].set(mi1)
 *             this_collection.ewah_refn[0].set(mi1)             # <<<<<<<<<<<<<<
 *             buf = &it1.second
 *             for vec_i in range(buf.sizeInBytes() / sizeof(ewah_word_type)):
 */
    (void)((__pyx_v_this_collection->ewah_refn[0]).set(__pyx_v_mi1));

    /* "yt/geometry/particle_oct_container.pyx":860
 *             this_collection.ewah_keys[0].set(mi1)
 *             this_collection.ewah_refn[0].set(mi1)
 *             buf = &it1.second             # <<<<<<<<<<<<<<
 *             for vec_i in range(buf.sizeInBytes() / sizeof(ewah_word_type)):
 *                 w = buf.getWord(vec_i)
 */
    __pyx_v_buf = (&__pyx_v_it1.second);

    /* "yt/geometry/particle_oct_container.pyx":861
 *             this_collection.ewah_refn[0].set(mi1)
 *             buf = &it1.second
 *             for vec_i in range(buf.sizeInBytes() / sizeof(ewah_word_type)):             # <<<<<<<<<<<<<<
 *                 w = buf.getWord(vec_i)
 *                 refined_arr.addWord(w)
 */
    __pyx_t_41 = (__pyx_v_buf->sizeInBytes() / (sizeof(__pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_word_type)));
    __pyx_t_42 = __pyx_t_41;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_42; __pyx_t_7+=1) {
      __pyx_v_vec_i = __pyx_t_7;

      /* "yt/geometry/particle_oct_container.pyx":862
 *             buf = &it1.second
 *             for vec_i in range(buf.sizeInBytes() / sizeof(ewah_word_type)):
 *                 w = buf.getWord(vec_i)             # <<<<<<<<<<<<<<
 *                 refined_arr.addWord(w)
 *         out_collection = BoolArrayCollection()
 */
      __pyx_v_w = __pyx_v_buf->getWord(__pyx_v_vec_i);

      /* "yt/geometry/particle_oct_container.pyx":863
 *             for vec_i in range(buf.sizeInBytes() / sizeof(ewah_word_type)):
 *                 w = buf.getWord(vec_i)
 *                 refined_arr.addWord(w)             # <<<<<<<<<<<<<<
 *         out_collection = BoolArrayCollection()
 *         in_collection._logicalor(this_collection, out_collection)
 */
      (void)(__pyx_v_refined_arr->addWord(__pyx_v_w));
    }

    /* "yt/geometry/particle_oct_container.pyx":855
 *         this_collection = BoolArrayCollection()
 *         cdef ewah_bool_array *refined_arr = NULL
 *         for it1 in coarse_refined_map:             # <<<<<<<<<<<<<<
 *             mi1 = it1.first
 *             refined_arr = &this_collection.ewah_coll[0][mi1]
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":864
 *                 w = buf.getWord(vec_i)
 *                 refined_arr.addWord(w)
 *         out_collection = BoolArrayCollection()             # <<<<<<<<<<<<<<
 *         in_collection._logicalor(this_collection, out_collection)
 *         return out_collection
 */
  __pyx_t_4 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_out_collection = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/particle_oct_container.pyx":865
 *                 refined_arr.addWord(w)
 *         out_collection = BoolArrayCollection()
 *         in_collection._logicalor(this_collection, out_collection)             # <<<<<<<<<<<<<<
 *         return out_collection
 * 
 */
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_v_in_collection->__pyx_vtab)->_logicalor(__pyx_v_in_collection, __pyx_v_this_collection, __pyx_v_out_collection);

  /* "yt/geometry/particle_oct_container.pyx":866
 *         out_collection = BoolArrayCollection()
 *         in_collection._logicalor(this_collection, out_collection)
 *         return out_collection             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_out_collection));
  __pyx_r = __pyx_v_out_collection;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":691
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef BoolArrayCollection __refined_index_data_file(             # <<<<<<<<<<<<<<
 *         self,
 *         BoolArrayCollection in_collection,
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_1, 1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_morton_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_refined_count.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sorted_order.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.__refined_index_data_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_morton_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_refined_count.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sorted_order.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_this_collection);
  __Pyx_XDECREF((PyObject *)__pyx_v_out_collection);
  __PYX_XDEC_MEMVIEW(&__pyx_v_particle_counts, 1);
  __Pyx_XDECREF((PyObject *)__pyx_v_refined_count);
  __Pyx_XDECREF((PyObject *)__pyx_v_morton_indices);
  __Pyx_XDECREF((PyObject *)__pyx_v_sorted_order);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_fuse_1__pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___refined_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_in_collection, PyArrayObject *__pyx_v_pos, PyArrayObject *__pyx_v_hsml, PyArrayObject *__pyx_v_mask, __pyx_t_5numpy_uint64_t __pyx_v_count_threshold, __pyx_t_5numpy_uint8_t __pyx_v_mask_threshold) {
  __pyx_t_5numpy_int64_t __pyx_v_p;
  __pyx_t_5numpy_int64_t __pyx_v_sorted_ind;
  __pyx_t_5numpy_uint64_t __pyx_v_i;
  __pyx_t_5numpy_uint64_t __pyx_v_mi1;
  __pyx_t_5numpy_uint64_t __pyx_v_mi2;
  __pyx_t_5numpy_float64_t __pyx_v_ppos[3];
  __pyx_t_5numpy_float64_t __pyx_v_s_ppos[3];
  int __pyx_v_skip;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_this_collection = 0;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_out_collection = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_bounds[2][3];
  __pyx_t_5numpy_uint8_t __pyx_v_fully_enclosed;
  __pyx_t_5numpy_float64_t __pyx_v_LE[3];
  __pyx_t_5numpy_float64_t __pyx_v_RE[3];
  __pyx_t_5numpy_float64_t __pyx_v_DW[3];
  __pyx_t_5numpy_uint8_t __pyx_v_PER[3];
  __pyx_t_5numpy_float64_t __pyx_v_dds1[3];
  __pyx_t_5numpy_float64_t __pyx_v_dds2[3];
  __pyx_t_5numpy_float64_t __pyx_v_radius;
  __pyx_t_5numpy_uint64_t __pyx_v_mi_split1[3];
  __pyx_t_5numpy_uint64_t __pyx_v_mi_split2[3];
  __pyx_t_5numpy_uint64_t __pyx_v_miex1;
  __Pyx_memviewslice __pyx_v_particle_counts = { 0, 0, { 0 }, { 0 }, { 0 } };
  __pyx_t_5numpy_uint64_t __pyx_v_xex;
  __pyx_t_5numpy_uint64_t __pyx_v_yex;
  __pyx_t_5numpy_uint64_t __pyx_v_zex;
  __pyx_t_5numpy_float64_t __pyx_v_clip_pos_l[3];
  __pyx_t_5numpy_float64_t __pyx_v_clip_pos_r[3];
  int __pyx_v_axiter[3][2];
  __pyx_t_5numpy_float64_t __pyx_v_axiterv[3][2];
  __pyx_t_2yt_8geometry_22particle_oct_container_CoarseRefinedSets __pyx_v_coarse_refined_map;
  __pyx_t_5numpy_uint64_t __pyx_v_nfully_enclosed;
  __pyx_t_5numpy_uint64_t __pyx_v_n_calls;
  CYTHON_UNUSED long __pyx_v_mi1_max;
  CYTHON_UNUSED long __pyx_v_mi2_max;
  __pyx_t_5numpy_uint64_t __pyx_v_max_mi1_elements;
  __pyx_t_5numpy_uint64_t __pyx_v_max_mi2_elements;
  PyArrayObject *__pyx_v_refined_count = 0;
  PyArrayObject *__pyx_v_morton_indices = 0;
  PyArrayObject *__pyx_v_sorted_order = 0;
  long __pyx_v_xi;
  long __pyx_v_yi;
  long __pyx_v_zi;
  __pyx_t_5numpy_uint64_t __pyx_v_vec_i;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_bool_array *__pyx_v_buf;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_word_type __pyx_v_w;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array *__pyx_v_refined_arr;
  std::pair<__pyx_t_5numpy_uint64_t,__pyx_t_2yt_9utilities_3lib_15ewah_bool_array_bool_array>  __pyx_v_it1;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_hsml;
  __Pyx_Buffer __pyx_pybuffer_hsml;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_mask;
  __Pyx_Buffer __pyx_pybuffer_mask;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_morton_indices;
  __Pyx_Buffer __pyx_pybuffer_morton_indices;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_pos;
  __Pyx_Buffer __pyx_pybuffer_pos;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_refined_count;
  __Pyx_Buffer __pyx_pybuffer_refined_count;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_sorted_order;
  __Pyx_Buffer __pyx_pybuffer_sorted_order;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyArrayObject *__pyx_t_6 = NULL;
  __pyx_t_5numpy_uint64_t __pyx_t_7;
  PyArrayObject *__pyx_t_8 = NULL;
  npy_intp __pyx_t_9;
  npy_intp __pyx_t_10;
  __pyx_t_5numpy_int64_t __pyx_t_11;
  __pyx_t_5numpy_int64_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  __pyx_t_5numpy_int64_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  __pyx_t_5numpy_int64_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  __pyx_t_5numpy_int64_t __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  PyArrayObject *__pyx_t_21 = NULL;
  int __pyx_t_22;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  __pyx_t_5numpy_uint64_t __pyx_t_26;
  double __pyx_t_27;
  long __pyx_t_28;
  long __pyx_t_29;
  long __pyx_t_30;
  __pyx_t_5numpy_uint64_t __pyx_t_31;
  __pyx_t_5numpy_uint64_t __pyx_t_32;
  __pyx_t_5numpy_uint64_t __pyx_t_33;
  __pyx_t_5numpy_uint64_t __pyx_t_34;
  __pyx_t_5numpy_uint64_t __pyx_t_35;
  __pyx_t_5numpy_uint64_t __pyx_t_36;
  __pyx_t_5numpy_uint64_t __pyx_t_37;
  __pyx_t_5numpy_uint64_t __pyx_t_38;
  std::map<__pyx_t_5numpy_uint64_t,__pyx_t_2yt_9utilities_3lib_15ewah_bool_array_bool_array> ::iterator __pyx_t_39;
  std::pair<__pyx_t_5numpy_uint64_t,__pyx_t_2yt_9utilities_3lib_15ewah_bool_array_bool_array>  __pyx_t_40;
  size_t __pyx_t_41;
  size_t __pyx_t_42;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_1__refined_index_data_file", 0);
  __pyx_pybuffer_refined_count.pybuffer.buf = NULL;
  __pyx_pybuffer_refined_count.refcount = 0;
  __pyx_pybuffernd_refined_count.data = NULL;
  __pyx_pybuffernd_refined_count.rcbuffer = &__pyx_pybuffer_refined_count;
  __pyx_pybuffer_morton_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_morton_indices.refcount = 0;
  __pyx_pybuffernd_morton_indices.data = NULL;
  __pyx_pybuffernd_morton_indices.rcbuffer = &__pyx_pybuffer_morton_indices;
  __pyx_pybuffer_sorted_order.pybuffer.buf = NULL;
  __pyx_pybuffer_sorted_order.refcount = 0;
  __pyx_pybuffernd_sorted_order.data = NULL;
  __pyx_pybuffernd_sorted_order.rcbuffer = &__pyx_pybuffer_sorted_order;
  __pyx_pybuffer_pos.pybuffer.buf = NULL;
  __pyx_pybuffer_pos.refcount = 0;
  __pyx_pybuffernd_pos.data = NULL;
  __pyx_pybuffernd_pos.rcbuffer = &__pyx_pybuffer_pos;
  __pyx_pybuffer_hsml.pybuffer.buf = NULL;
  __pyx_pybuffer_hsml.refcount = 0;
  __pyx_pybuffernd_hsml.data = NULL;
  __pyx_pybuffernd_hsml.rcbuffer = &__pyx_pybuffer_hsml;
  __pyx_pybuffer_mask.pybuffer.buf = NULL;
  __pyx_pybuffer_mask.refcount = 0;
  __pyx_pybuffernd_mask.data = NULL;
  __pyx_pybuffernd_mask.rcbuffer = &__pyx_pybuffer_mask;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pos.rcbuffer->pybuffer, (PyObject*)__pyx_v_pos, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) __PYX_ERR(0, 691, __pyx_L1_error)
  }
  __pyx_pybuffernd_pos.diminfo[0].strides = __pyx_pybuffernd_pos.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_pos.diminfo[0].shape = __pyx_pybuffernd_pos.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_pos.diminfo[1].strides = __pyx_pybuffernd_pos.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_pos.diminfo[1].shape = __pyx_pybuffernd_pos.rcbuffer->pybuffer.shape[1];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer, (PyObject*)__pyx_v_hsml, &__Pyx_TypeInfo_double, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 691, __pyx_L1_error)
  }
  __pyx_pybuffernd_hsml.diminfo[0].strides = __pyx_pybuffernd_hsml.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_hsml.diminfo[0].shape = __pyx_pybuffernd_hsml.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_mask.rcbuffer->pybuffer, (PyObject*)__pyx_v_mask, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 691, __pyx_L1_error)
  }
  __pyx_pybuffernd_mask.diminfo[0].strides = __pyx_pybuffernd_mask.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_mask.diminfo[0].shape = __pyx_pybuffernd_mask.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/particle_oct_container.pyx":719
 *         cdef np.uint64_t mi_split2[3]
 *         cdef np.uint64_t miex1
 *         cdef np.uint64_t[:] particle_counts = self.particle_counts             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t xex, yex, zex
 *         cdef np.float64_t clip_pos_l[3]
 */
  __pyx_t_1 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_v_self->particle_counts, PyBUF_WRITABLE); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(0, 719, __pyx_L1_error)
  __pyx_v_particle_counts = __pyx_t_1;
  __pyx_t_1.memview = NULL;
  __pyx_t_1.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":726
 *         cdef np.float64_t axiterv[3][2]
 *         cdef CoarseRefinedSets coarse_refined_map
 *         cdef np.uint64_t nfully_enclosed = 0, n_calls = 0             # <<<<<<<<<<<<<<
 *         mi1_max = (1 << self.index_order1) - 1
 *         mi2_max = (1 << self.index_order2) - 1
 */
  __pyx_v_nfully_enclosed = 0;
  __pyx_v_n_calls = 0;

  /* "yt/geometry/particle_oct_container.pyx":727
 *         cdef CoarseRefinedSets coarse_refined_map
 *         cdef np.uint64_t nfully_enclosed = 0, n_calls = 0
 *         mi1_max = (1 << self.index_order1) - 1             # <<<<<<<<<<<<<<
 *         mi2_max = (1 << self.index_order2) - 1
 *         cdef np.uint64_t max_mi1_elements = 1 << (3*self.index_order1)
 */
  __pyx_v_mi1_max = ((1 << __pyx_v_self->index_order1) - 1);

  /* "yt/geometry/particle_oct_container.pyx":728
 *         cdef np.uint64_t nfully_enclosed = 0, n_calls = 0
 *         mi1_max = (1 << self.index_order1) - 1
 *         mi2_max = (1 << self.index_order2) - 1             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t max_mi1_elements = 1 << (3*self.index_order1)
 *         cdef np.uint64_t max_mi2_elements = 1 << (3*self.index_order2)
 */
  __pyx_v_mi2_max = ((1 << __pyx_v_self->index_order2) - 1);

  /* "yt/geometry/particle_oct_container.pyx":729
 *         mi1_max = (1 << self.index_order1) - 1
 *         mi2_max = (1 << self.index_order2) - 1
 *         cdef np.uint64_t max_mi1_elements = 1 << (3*self.index_order1)             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t max_mi2_elements = 1 << (3*self.index_order2)
 *         cdef np.ndarray[np.uint64_t, ndim=1] refined_count = np.zeros(max_mi1_elements, dtype="uint64")
 */
  __pyx_v_max_mi1_elements = (1 << (3 * __pyx_v_self->index_order1));

  /* "yt/geometry/particle_oct_container.pyx":730
 *         mi2_max = (1 << self.index_order2) - 1
 *         cdef np.uint64_t max_mi1_elements = 1 << (3*self.index_order1)
 *         cdef np.uint64_t max_mi2_elements = 1 << (3*self.index_order2)             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint64_t, ndim=1] refined_count = np.zeros(max_mi1_elements, dtype="uint64")
 *         # Copy things from structure (type cast)
 */
  __pyx_v_max_mi2_elements = (1 << (3 * __pyx_v_self->index_order2));

  /* "yt/geometry/particle_oct_container.pyx":731
 *         cdef np.uint64_t max_mi1_elements = 1 << (3*self.index_order1)
 *         cdef np.uint64_t max_mi2_elements = 1 << (3*self.index_order2)
 *         cdef np.ndarray[np.uint64_t, ndim=1] refined_count = np.zeros(max_mi1_elements, dtype="uint64")             # <<<<<<<<<<<<<<
 *         # Copy things from structure (type cast)
 *         for i in range(3):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_npy_uint64(__pyx_v_max_mi1_elements); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_n_s_uint64) < 0) __PYX_ERR(0, 731, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 731, __pyx_L1_error)
  __pyx_t_6 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_refined_count.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_refined_count = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_refined_count.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 731, __pyx_L1_error)
    } else {__pyx_pybuffernd_refined_count.diminfo[0].strides = __pyx_pybuffernd_refined_count.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_refined_count.diminfo[0].shape = __pyx_pybuffernd_refined_count.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_6 = 0;
  __pyx_v_refined_count = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "yt/geometry/particle_oct_container.pyx":733
 *         cdef np.ndarray[np.uint64_t, ndim=1] refined_count = np.zeros(max_mi1_elements, dtype="uint64")
 *         # Copy things from structure (type cast)
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             LE[i] = self.left_edge[i]
 *             RE[i] = self.right_edge[i]
 */
  for (__pyx_t_7 = 0; __pyx_t_7 < 3; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "yt/geometry/particle_oct_container.pyx":734
 *         # Copy things from structure (type cast)
 *         for i in range(3):
 *             LE[i] = self.left_edge[i]             # <<<<<<<<<<<<<<
 *             RE[i] = self.right_edge[i]
 *             PER[i] = self.periodicity[i]
 */
    (__pyx_v_LE[__pyx_v_i]) = (__pyx_v_self->left_edge[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":735
 *         for i in range(3):
 *             LE[i] = self.left_edge[i]
 *             RE[i] = self.right_edge[i]             # <<<<<<<<<<<<<<
 *             PER[i] = self.periodicity[i]
 *             dds1[i] = self.dds_mi1[i]
 */
    (__pyx_v_RE[__pyx_v_i]) = (__pyx_v_self->right_edge[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":736
 *             LE[i] = self.left_edge[i]
 *             RE[i] = self.right_edge[i]
 *             PER[i] = self.periodicity[i]             # <<<<<<<<<<<<<<
 *             dds1[i] = self.dds_mi1[i]
 *             dds2[i] = self.dds_mi2[i]
 */
    (__pyx_v_PER[__pyx_v_i]) = (__pyx_v_self->periodicity[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":737
 *             RE[i] = self.right_edge[i]
 *             PER[i] = self.periodicity[i]
 *             dds1[i] = self.dds_mi1[i]             # <<<<<<<<<<<<<<
 *             dds2[i] = self.dds_mi2[i]
 *             DW[i] = RE[i] - LE[i]
 */
    (__pyx_v_dds1[__pyx_v_i]) = (__pyx_v_self->dds_mi1[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":738
 *             PER[i] = self.periodicity[i]
 *             dds1[i] = self.dds_mi1[i]
 *             dds2[i] = self.dds_mi2[i]             # <<<<<<<<<<<<<<
 *             DW[i] = RE[i] - LE[i]
 *             axiter[i][0] = 0 # We always do an offset of 0
 */
    (__pyx_v_dds2[__pyx_v_i]) = (__pyx_v_self->dds_mi2[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":739
 *             dds1[i] = self.dds_mi1[i]
 *             dds2[i] = self.dds_mi2[i]
 *             DW[i] = RE[i] - LE[i]             # <<<<<<<<<<<<<<
 *             axiter[i][0] = 0 # We always do an offset of 0
 *             axiterv[i][0] = 0.0
 */
    (__pyx_v_DW[__pyx_v_i]) = ((__pyx_v_RE[__pyx_v_i]) - (__pyx_v_LE[__pyx_v_i]));

    /* "yt/geometry/particle_oct_container.pyx":740
 *             dds2[i] = self.dds_mi2[i]
 *             DW[i] = RE[i] - LE[i]
 *             axiter[i][0] = 0 # We always do an offset of 0             # <<<<<<<<<<<<<<
 *             axiterv[i][0] = 0.0
 *         cdef np.ndarray[np.uint64_t, ndim=1] morton_indices = np.empty(pos.shape[0], dtype="u8")
 */
    ((__pyx_v_axiter[__pyx_v_i])[0]) = 0;

    /* "yt/geometry/particle_oct_container.pyx":741
 *             DW[i] = RE[i] - LE[i]
 *             axiter[i][0] = 0 # We always do an offset of 0
 *             axiterv[i][0] = 0.0             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint64_t, ndim=1] morton_indices = np.empty(pos.shape[0], dtype="u8")
 *         for p in range(pos.shape[0]):
 */
    ((__pyx_v_axiterv[__pyx_v_i])[0]) = 0.0;
  }

  /* "yt/geometry/particle_oct_container.pyx":742
 *             axiter[i][0] = 0 # We always do an offset of 0
 *             axiterv[i][0] = 0.0
 *         cdef np.ndarray[np.uint64_t, ndim=1] morton_indices = np.empty(pos.shape[0], dtype="u8")             # <<<<<<<<<<<<<<
 *         for p in range(pos.shape[0]):
 *             morton_indices[p] = bounded_morton(pos[p, 0], pos[p, 1], pos[p, 2],
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_Py_intptr_t((__pyx_v_pos->dimensions[0])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_n_s_u8) < 0) __PYX_ERR(0, 742, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 742, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_morton_indices.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_morton_indices = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_morton_indices.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 742, __pyx_L1_error)
    } else {__pyx_pybuffernd_morton_indices.diminfo[0].strides = __pyx_pybuffernd_morton_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_morton_indices.diminfo[0].shape = __pyx_pybuffernd_morton_indices.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_8 = 0;
  __pyx_v_morton_indices = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":743
 *             axiterv[i][0] = 0.0
 *         cdef np.ndarray[np.uint64_t, ndim=1] morton_indices = np.empty(pos.shape[0], dtype="u8")
 *         for p in range(pos.shape[0]):             # <<<<<<<<<<<<<<
 *             morton_indices[p] = bounded_morton(pos[p, 0], pos[p, 1], pos[p, 2],
 *                                                LE, RE, self.index_order1)
 */
  __pyx_t_9 = (__pyx_v_pos->dimensions[0]);
  __pyx_t_10 = __pyx_t_9;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_p = __pyx_t_11;

    /* "yt/geometry/particle_oct_container.pyx":744
 *         cdef np.ndarray[np.uint64_t, ndim=1] morton_indices = np.empty(pos.shape[0], dtype="u8")
 *         for p in range(pos.shape[0]):
 *             morton_indices[p] = bounded_morton(pos[p, 0], pos[p, 1], pos[p, 2],             # <<<<<<<<<<<<<<
 *                                                LE, RE, self.index_order1)
 *         # Loop over positions skipping those outside the domain
 */
    __pyx_t_12 = __pyx_v_p;
    __pyx_t_13 = 0;
    __pyx_t_14 = __pyx_v_p;
    __pyx_t_15 = 1;
    __pyx_t_16 = __pyx_v_p;
    __pyx_t_17 = 2;

    /* "yt/geometry/particle_oct_container.pyx":745
 *         for p in range(pos.shape[0]):
 *             morton_indices[p] = bounded_morton(pos[p, 0], pos[p, 1], pos[p, 2],
 *                                                LE, RE, self.index_order1)             # <<<<<<<<<<<<<<
 *         # Loop over positions skipping those outside the domain
 *         cdef np.ndarray[np.uint64_t, ndim=1, cast=True] sorted_order
 */
    __pyx_t_18 = __pyx_v_p;
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_morton_indices.rcbuffer->pybuffer.buf, __pyx_t_18, __pyx_pybuffernd_morton_indices.diminfo[0].strides) = __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton((*__Pyx_BufPtrStrided2d(double *, __pyx_pybuffernd_pos.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_pos.diminfo[0].strides, __pyx_t_13, __pyx_pybuffernd_pos.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(double *, __pyx_pybuffernd_pos.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_pos.diminfo[0].strides, __pyx_t_15, __pyx_pybuffernd_pos.diminfo[1].strides)), (*__Pyx_BufPtrStrided2d(double *, __pyx_pybuffernd_pos.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_pos.diminfo[0].strides, __pyx_t_17, __pyx_pybuffernd_pos.diminfo[1].strides)), __pyx_v_LE, __pyx_v_RE, __pyx_v_self->index_order1);
  }

  /* "yt/geometry/particle_oct_container.pyx":748
 *         # Loop over positions skipping those outside the domain
 *         cdef np.ndarray[np.uint64_t, ndim=1, cast=True] sorted_order
 *         if hsml is None:             # <<<<<<<<<<<<<<
 *             # casting to uint64 for compatibility with 32 bits systems
 *             # see https://github.com/yt-project/yt/issues/3656
 */
  __pyx_t_19 = (((PyObject *)__pyx_v_hsml) == Py_None);
  __pyx_t_20 = (__pyx_t_19 != 0);
  if (__pyx_t_20) {

    /* "yt/geometry/particle_oct_container.pyx":751
 *             # casting to uint64 for compatibility with 32 bits systems
 *             # see https://github.com/yt-project/yt/issues/3656
 *             sorted_order = np.argsort(morton_indices).astype(np.uint64, copy=False)             # <<<<<<<<<<<<<<
 *         else:
 *             sorted_order = np.argsort(hsml)[::-1].astype(np.uint64, copy=False)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_argsort); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, ((PyObject *)__pyx_v_morton_indices)) : __Pyx_PyObject_CallOneArg(__pyx_t_4, ((PyObject *)__pyx_v_morton_indices));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_astype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_uint64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_copy, Py_False) < 0) __PYX_ERR(0, 751, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 751, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 751, __pyx_L1_error)
    __pyx_t_21 = ((PyArrayObject *)__pyx_t_2);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sorted_order.rcbuffer->pybuffer);
      __pyx_t_22 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_sorted_order.rcbuffer->pybuffer, (PyObject*)__pyx_t_21, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 1, __pyx_stack);
      if (unlikely(__pyx_t_22 < 0)) {
        PyErr_Fetch(&__pyx_t_23, &__pyx_t_24, &__pyx_t_25);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_sorted_order.rcbuffer->pybuffer, (PyObject*)__pyx_v_sorted_order, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 1, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_23); Py_XDECREF(__pyx_t_24); Py_XDECREF(__pyx_t_25);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_23, __pyx_t_24, __pyx_t_25);
        }
        __pyx_t_23 = __pyx_t_24 = __pyx_t_25 = 0;
      }
      __pyx_pybuffernd_sorted_order.diminfo[0].strides = __pyx_pybuffernd_sorted_order.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_sorted_order.diminfo[0].shape = __pyx_pybuffernd_sorted_order.rcbuffer->pybuffer.shape[0];
      if (unlikely(__pyx_t_22 < 0)) __PYX_ERR(0, 751, __pyx_L1_error)
    }
    __pyx_t_21 = 0;
    __pyx_v_sorted_order = ((PyArrayObject *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "yt/geometry/particle_oct_container.pyx":748
 *         # Loop over positions skipping those outside the domain
 *         cdef np.ndarray[np.uint64_t, ndim=1, cast=True] sorted_order
 *         if hsml is None:             # <<<<<<<<<<<<<<
 *             # casting to uint64 for compatibility with 32 bits systems
 *             # see https://github.com/yt-project/yt/issues/3656
 */
    goto __pyx_L7;
  }

  /* "yt/geometry/particle_oct_container.pyx":753
 *             sorted_order = np.argsort(morton_indices).astype(np.uint64, copy=False)
 *         else:
 *             sorted_order = np.argsort(hsml)[::-1].astype(np.uint64, copy=False)             # <<<<<<<<<<<<<<
 *         for sorted_ind in range(sorted_order.shape[0]):
 *             p = sorted_order[sorted_ind]
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_argsort); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, ((PyObject *)__pyx_v_hsml)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_hsml));
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_slice__18); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_astype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_uint64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_copy, Py_False) < 0) __PYX_ERR(0, 753, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 753, __pyx_L1_error)
    __pyx_t_21 = ((PyArrayObject *)__pyx_t_4);
    {
      __Pyx_BufFmt_StackElem __pyx_stack[1];
      __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sorted_order.rcbuffer->pybuffer);
      __pyx_t_22 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_sorted_order.rcbuffer->pybuffer, (PyObject*)__pyx_t_21, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 1, __pyx_stack);
      if (unlikely(__pyx_t_22 < 0)) {
        PyErr_Fetch(&__pyx_t_25, &__pyx_t_24, &__pyx_t_23);
        if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_sorted_order.rcbuffer->pybuffer, (PyObject*)__pyx_v_sorted_order, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 1, __pyx_stack) == -1)) {
          Py_XDECREF(__pyx_t_25); Py_XDECREF(__pyx_t_24); Py_XDECREF(__pyx_t_23);
          __Pyx_RaiseBufferFallbackError();
        } else {
          PyErr_Restore(__pyx_t_25, __pyx_t_24, __pyx_t_23);
        }
        __pyx_t_25 = __pyx_t_24 = __pyx_t_23 = 0;
      }
      __pyx_pybuffernd_sorted_order.diminfo[0].strides = __pyx_pybuffernd_sorted_order.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_sorted_order.diminfo[0].shape = __pyx_pybuffernd_sorted_order.rcbuffer->pybuffer.shape[0];
      if (unlikely(__pyx_t_22 < 0)) __PYX_ERR(0, 753, __pyx_L1_error)
    }
    __pyx_t_21 = 0;
    __pyx_v_sorted_order = ((PyArrayObject *)__pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L7:;

  /* "yt/geometry/particle_oct_container.pyx":754
 *         else:
 *             sorted_order = np.argsort(hsml)[::-1].astype(np.uint64, copy=False)
 *         for sorted_ind in range(sorted_order.shape[0]):             # <<<<<<<<<<<<<<
 *             p = sorted_order[sorted_ind]
 *             skip = 0
 */
  __pyx_t_9 = (__pyx_v_sorted_order->dimensions[0]);
  __pyx_t_10 = __pyx_t_9;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_sorted_ind = __pyx_t_11;

    /* "yt/geometry/particle_oct_container.pyx":755
 *             sorted_order = np.argsort(hsml)[::-1].astype(np.uint64, copy=False)
 *         for sorted_ind in range(sorted_order.shape[0]):
 *             p = sorted_order[sorted_ind]             # <<<<<<<<<<<<<<
 *             skip = 0
 *             for i in range(3):
 */
    __pyx_t_16 = __pyx_v_sorted_ind;
    __pyx_v_p = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_sorted_order.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_sorted_order.diminfo[0].strides));

    /* "yt/geometry/particle_oct_container.pyx":756
 *         for sorted_ind in range(sorted_order.shape[0]):
 *             p = sorted_order[sorted_ind]
 *             skip = 0             # <<<<<<<<<<<<<<
 *             for i in range(3):
 *                 axiter[i][1] = 999
 */
    __pyx_v_skip = 0;

    /* "yt/geometry/particle_oct_container.pyx":757
 *             p = sorted_order[sorted_ind]
 *             skip = 0
 *             for i in range(3):             # <<<<<<<<<<<<<<
 *                 axiter[i][1] = 999
 *                 if not (LE[i] <= pos[p, i] < RE[i]):
 */
    for (__pyx_t_7 = 0; __pyx_t_7 < 3; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "yt/geometry/particle_oct_container.pyx":758
 *             skip = 0
 *             for i in range(3):
 *                 axiter[i][1] = 999             # <<<<<<<<<<<<<<
 *                 if not (LE[i] <= pos[p, i] < RE[i]):
 *                     skip = 1
 */
      ((__pyx_v_axiter[__pyx_v_i])[1]) = 0x3E7;

      /* "yt/geometry/particle_oct_container.pyx":759
 *             for i in range(3):
 *                 axiter[i][1] = 999
 *                 if not (LE[i] <= pos[p, i] < RE[i]):             # <<<<<<<<<<<<<<
 *                     skip = 1
 *                     break
 */
      __pyx_t_16 = __pyx_v_p;
      __pyx_t_26 = __pyx_v_i;
      __pyx_t_27 = (*__Pyx_BufPtrStrided2d(double *, __pyx_pybuffernd_pos.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_pos.diminfo[0].strides, __pyx_t_26, __pyx_pybuffernd_pos.diminfo[1].strides));
      __pyx_t_20 = ((__pyx_v_LE[__pyx_v_i]) <= __pyx_t_27);
      if (__pyx_t_20) {
        __pyx_t_20 = (__pyx_t_27 < (__pyx_v_RE[__pyx_v_i]));
      }
      __pyx_t_19 = ((!(__pyx_t_20 != 0)) != 0);
      if (__pyx_t_19) {

        /* "yt/geometry/particle_oct_container.pyx":760
 *                 axiter[i][1] = 999
 *                 if not (LE[i] <= pos[p, i] < RE[i]):
 *                     skip = 1             # <<<<<<<<<<<<<<
 *                     break
 *                 ppos[i] = pos[p,i]
 */
        __pyx_v_skip = 1;

        /* "yt/geometry/particle_oct_container.pyx":761
 *                 if not (LE[i] <= pos[p, i] < RE[i]):
 *                     skip = 1
 *                     break             # <<<<<<<<<<<<<<
 *                 ppos[i] = pos[p,i]
 *             if skip == 1: continue
 */
        goto __pyx_L11_break;

        /* "yt/geometry/particle_oct_container.pyx":759
 *             for i in range(3):
 *                 axiter[i][1] = 999
 *                 if not (LE[i] <= pos[p, i] < RE[i]):             # <<<<<<<<<<<<<<
 *                     skip = 1
 *                     break
 */
      }

      /* "yt/geometry/particle_oct_container.pyx":762
 *                     skip = 1
 *                     break
 *                 ppos[i] = pos[p,i]             # <<<<<<<<<<<<<<
 *             if skip == 1: continue
 *             # Only look if collision at coarse index
 */
      __pyx_t_16 = __pyx_v_p;
      __pyx_t_26 = __pyx_v_i;
      (__pyx_v_ppos[__pyx_v_i]) = (*__Pyx_BufPtrStrided2d(double *, __pyx_pybuffernd_pos.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_pos.diminfo[0].strides, __pyx_t_26, __pyx_pybuffernd_pos.diminfo[1].strides));
    }
    __pyx_L11_break:;

    /* "yt/geometry/particle_oct_container.pyx":763
 *                     break
 *                 ppos[i] = pos[p,i]
 *             if skip == 1: continue             # <<<<<<<<<<<<<<
 *             # Only look if collision at coarse index
 *             mi1 = bounded_morton_split_dds(ppos[0], ppos[1], ppos[2], LE,
 */
    __pyx_t_19 = ((__pyx_v_skip == 1) != 0);
    if (__pyx_t_19) {
      goto __pyx_L8_continue;
    }

    /* "yt/geometry/particle_oct_container.pyx":765
 *             if skip == 1: continue
 *             # Only look if collision at coarse index
 *             mi1 = bounded_morton_split_dds(ppos[0], ppos[1], ppos[2], LE,             # <<<<<<<<<<<<<<
 *                                            dds1, mi_split1)
 *             if hsml is None:
 */
    __pyx_v_mi1 = __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_dds((__pyx_v_ppos[0]), (__pyx_v_ppos[1]), (__pyx_v_ppos[2]), __pyx_v_LE, __pyx_v_dds1, __pyx_v_mi_split1);

    /* "yt/geometry/particle_oct_container.pyx":767
 *             mi1 = bounded_morton_split_dds(ppos[0], ppos[1], ppos[2], LE,
 *                                            dds1, mi_split1)
 *             if hsml is None:             # <<<<<<<<<<<<<<
 *                 if mask[mi1] < mask_threshold \
 *                         or particle_counts[mi1] < count_threshold:
 */
    __pyx_t_19 = (((PyObject *)__pyx_v_hsml) == Py_None);
    __pyx_t_20 = (__pyx_t_19 != 0);
    if (__pyx_t_20) {

      /* "yt/geometry/particle_oct_container.pyx":768
 *                                            dds1, mi_split1)
 *             if hsml is None:
 *                 if mask[mi1] < mask_threshold \             # <<<<<<<<<<<<<<
 *                         or particle_counts[mi1] < count_threshold:
 *                     continue
 */
      __pyx_t_7 = __pyx_v_mi1;
      __pyx_t_19 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_mask.diminfo[0].strides)) < __pyx_v_mask_threshold) != 0);
      if (!__pyx_t_19) {
      } else {
        __pyx_t_20 = __pyx_t_19;
        goto __pyx_L16_bool_binop_done;
      }

      /* "yt/geometry/particle_oct_container.pyx":769
 *             if hsml is None:
 *                 if mask[mi1] < mask_threshold \
 *                         or particle_counts[mi1] < count_threshold:             # <<<<<<<<<<<<<<
 *                     continue
 *                 # Determine sub index within cell of primary index
 */
      __pyx_t_7 = __pyx_v_mi1;
      __pyx_t_19 = (((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_particle_counts.data + __pyx_t_7 * __pyx_v_particle_counts.strides[0]) ))) < __pyx_v_count_threshold) != 0);
      __pyx_t_20 = __pyx_t_19;
      __pyx_L16_bool_binop_done:;

      /* "yt/geometry/particle_oct_container.pyx":768
 *                                            dds1, mi_split1)
 *             if hsml is None:
 *                 if mask[mi1] < mask_threshold \             # <<<<<<<<<<<<<<
 *                         or particle_counts[mi1] < count_threshold:
 *                     continue
 */
      if (__pyx_t_20) {

        /* "yt/geometry/particle_oct_container.pyx":770
 *                 if mask[mi1] < mask_threshold \
 *                         or particle_counts[mi1] < count_threshold:
 *                     continue             # <<<<<<<<<<<<<<
 *                 # Determine sub index within cell of primary index
 *                 mi2 = bounded_morton_split_relative_dds(
 */
        goto __pyx_L8_continue;

        /* "yt/geometry/particle_oct_container.pyx":768
 *                                            dds1, mi_split1)
 *             if hsml is None:
 *                 if mask[mi1] < mask_threshold \             # <<<<<<<<<<<<<<
 *                         or particle_counts[mi1] < count_threshold:
 *                     continue
 */
      }

      /* "yt/geometry/particle_oct_container.pyx":772
 *                     continue
 *                 # Determine sub index within cell of primary index
 *                 mi2 = bounded_morton_split_relative_dds(             # <<<<<<<<<<<<<<
 *                     ppos[0], ppos[1], ppos[2], LE, dds1, dds2, mi_split2)
 *                 if refined_count[mi1] == 0:
 */
      __pyx_v_mi2 = __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_relative_dds((__pyx_v_ppos[0]), (__pyx_v_ppos[1]), (__pyx_v_ppos[2]), __pyx_v_LE, __pyx_v_dds1, __pyx_v_dds2, __pyx_v_mi_split2);

      /* "yt/geometry/particle_oct_container.pyx":774
 *                 mi2 = bounded_morton_split_relative_dds(
 *                     ppos[0], ppos[1], ppos[2], LE, dds1, dds2, mi_split2)
 *                 if refined_count[mi1] == 0:             # <<<<<<<<<<<<<<
 *                     coarse_refined_map[mi1].padWithZeroes(max_mi2_elements)
 *                 if not coarse_refined_map[mi1].get(mi2):
 */
      __pyx_t_7 = __pyx_v_mi1;
      __pyx_t_20 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_refined_count.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_refined_count.diminfo[0].strides)) == 0) != 0);
      if (__pyx_t_20) {

        /* "yt/geometry/particle_oct_container.pyx":775
 *                     ppos[0], ppos[1], ppos[2], LE, dds1, dds2, mi_split2)
 *                 if refined_count[mi1] == 0:
 *                     coarse_refined_map[mi1].padWithZeroes(max_mi2_elements)             # <<<<<<<<<<<<<<
 *                 if not coarse_refined_map[mi1].get(mi2):
 *                     coarse_refined_map[mi1].set(mi2)
 */
        (void)((__pyx_v_coarse_refined_map[__pyx_v_mi1]).padWithZeroes(__pyx_v_max_mi2_elements));

        /* "yt/geometry/particle_oct_container.pyx":774
 *                 mi2 = bounded_morton_split_relative_dds(
 *                     ppos[0], ppos[1], ppos[2], LE, dds1, dds2, mi_split2)
 *                 if refined_count[mi1] == 0:             # <<<<<<<<<<<<<<
 *                     coarse_refined_map[mi1].padWithZeroes(max_mi2_elements)
 *                 if not coarse_refined_map[mi1].get(mi2):
 */
      }

      /* "yt/geometry/particle_oct_container.pyx":776
 *                 if refined_count[mi1] == 0:
 *                     coarse_refined_map[mi1].padWithZeroes(max_mi2_elements)
 *                 if not coarse_refined_map[mi1].get(mi2):             # <<<<<<<<<<<<<<
 *                     coarse_refined_map[mi1].set(mi2)
 *                     refined_count[mi1] += 1
 */
      __pyx_t_20 = ((!((__pyx_v_coarse_refined_map[__pyx_v_mi1]).get(__pyx_v_mi2) != 0)) != 0);
      if (__pyx_t_20) {

        /* "yt/geometry/particle_oct_container.pyx":777
 *                     coarse_refined_map[mi1].padWithZeroes(max_mi2_elements)
 *                 if not coarse_refined_map[mi1].get(mi2):
 *                     coarse_refined_map[mi1].set(mi2)             # <<<<<<<<<<<<<<
 *                     refined_count[mi1] += 1
 *             else: # only hit if we have smoothing lengths.
 */
        (__pyx_v_coarse_refined_map[__pyx_v_mi1]).set(__pyx_v_mi2);

        /* "yt/geometry/particle_oct_container.pyx":778
 *                 if not coarse_refined_map[mi1].get(mi2):
 *                     coarse_refined_map[mi1].set(mi2)
 *                     refined_count[mi1] += 1             # <<<<<<<<<<<<<<
 *             else: # only hit if we have smoothing lengths.
 *                 # We have to do essentially the identical process to in the coarse indexing,
 */
        __pyx_t_7 = __pyx_v_mi1;
        *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_refined_count.rcbuffer->pybuffer.buf, __pyx_t_7, __pyx_pybuffernd_refined_count.diminfo[0].strides) += 1;

        /* "yt/geometry/particle_oct_container.pyx":776
 *                 if refined_count[mi1] == 0:
 *                     coarse_refined_map[mi1].padWithZeroes(max_mi2_elements)
 *                 if not coarse_refined_map[mi1].get(mi2):             # <<<<<<<<<<<<<<
 *                     coarse_refined_map[mi1].set(mi2)
 *                     refined_count[mi1] += 1
 */
      }

      /* "yt/geometry/particle_oct_container.pyx":767
 *             mi1 = bounded_morton_split_dds(ppos[0], ppos[1], ppos[2], LE,
 *                                            dds1, mi_split1)
 *             if hsml is None:             # <<<<<<<<<<<<<<
 *                 if mask[mi1] < mask_threshold \
 *                         or particle_counts[mi1] < count_threshold:
 */
      goto __pyx_L14;
    }

    /* "yt/geometry/particle_oct_container.pyx":783
 *                 # except here we need to fill in all the subranges as well as the coarse ranges
 *                 # Note that we are also doing the null case, where we do no shifting
 *                 radius = hsml[p]             # <<<<<<<<<<<<<<
 *                 #if mask[mi1] <= 4: # only one thing in this area
 *                 #    continue
 */
    /*else*/ {
      __pyx_t_16 = __pyx_v_p;
      __pyx_v_radius = (*__Pyx_BufPtrStrided1d(double *, __pyx_pybuffernd_hsml.rcbuffer->pybuffer.buf, __pyx_t_16, __pyx_pybuffernd_hsml.diminfo[0].strides));

      /* "yt/geometry/particle_oct_container.pyx":786
 *                 #if mask[mi1] <= 4: # only one thing in this area
 *                 #    continue
 *                 for i in range(3):             # <<<<<<<<<<<<<<
 *                     if PER[i] and ppos[i] - radius < LE[i]:
 *                         axiter[i][1] = +1
 */
      for (__pyx_t_7 = 0; __pyx_t_7 < 3; __pyx_t_7+=1) {
        __pyx_v_i = __pyx_t_7;

        /* "yt/geometry/particle_oct_container.pyx":787
 *                 #    continue
 *                 for i in range(3):
 *                     if PER[i] and ppos[i] - radius < LE[i]:             # <<<<<<<<<<<<<<
 *                         axiter[i][1] = +1
 *                         axiterv[i][1] = DW[i]
 */
        __pyx_t_19 = ((__pyx_v_PER[__pyx_v_i]) != 0);
        if (__pyx_t_19) {
        } else {
          __pyx_t_20 = __pyx_t_19;
          goto __pyx_L23_bool_binop_done;
        }
        __pyx_t_19 = ((((__pyx_v_ppos[__pyx_v_i]) - __pyx_v_radius) < (__pyx_v_LE[__pyx_v_i])) != 0);
        __pyx_t_20 = __pyx_t_19;
        __pyx_L23_bool_binop_done:;
        if (__pyx_t_20) {

          /* "yt/geometry/particle_oct_container.pyx":788
 *                 for i in range(3):
 *                     if PER[i] and ppos[i] - radius < LE[i]:
 *                         axiter[i][1] = +1             # <<<<<<<<<<<<<<
 *                         axiterv[i][1] = DW[i]
 *                     elif PER[i] and ppos[i] + radius > RE[i]:
 */
          ((__pyx_v_axiter[__pyx_v_i])[1]) = 1;

          /* "yt/geometry/particle_oct_container.pyx":789
 *                     if PER[i] and ppos[i] - radius < LE[i]:
 *                         axiter[i][1] = +1
 *                         axiterv[i][1] = DW[i]             # <<<<<<<<<<<<<<
 *                     elif PER[i] and ppos[i] + radius > RE[i]:
 *                         axiter[i][1] = -1
 */
          ((__pyx_v_axiterv[__pyx_v_i])[1]) = (__pyx_v_DW[__pyx_v_i]);

          /* "yt/geometry/particle_oct_container.pyx":787
 *                 #    continue
 *                 for i in range(3):
 *                     if PER[i] and ppos[i] - radius < LE[i]:             # <<<<<<<<<<<<<<
 *                         axiter[i][1] = +1
 *                         axiterv[i][1] = DW[i]
 */
          goto __pyx_L22;
        }

        /* "yt/geometry/particle_oct_container.pyx":790
 *                         axiter[i][1] = +1
 *                         axiterv[i][1] = DW[i]
 *                     elif PER[i] and ppos[i] + radius > RE[i]:             # <<<<<<<<<<<<<<
 *                         axiter[i][1] = -1
 *                         axiterv[i][1] = -DW[i]
 */
        __pyx_t_19 = ((__pyx_v_PER[__pyx_v_i]) != 0);
        if (__pyx_t_19) {
        } else {
          __pyx_t_20 = __pyx_t_19;
          goto __pyx_L25_bool_binop_done;
        }
        __pyx_t_19 = ((((__pyx_v_ppos[__pyx_v_i]) + __pyx_v_radius) > (__pyx_v_RE[__pyx_v_i])) != 0);
        __pyx_t_20 = __pyx_t_19;
        __pyx_L25_bool_binop_done:;
        if (__pyx_t_20) {

          /* "yt/geometry/particle_oct_container.pyx":791
 *                         axiterv[i][1] = DW[i]
 *                     elif PER[i] and ppos[i] + radius > RE[i]:
 *                         axiter[i][1] = -1             # <<<<<<<<<<<<<<
 *                         axiterv[i][1] = -DW[i]
 *                 for xi in range(2):
 */
          ((__pyx_v_axiter[__pyx_v_i])[1]) = -1;

          /* "yt/geometry/particle_oct_container.pyx":792
 *                     elif PER[i] and ppos[i] + radius > RE[i]:
 *                         axiter[i][1] = -1
 *                         axiterv[i][1] = -DW[i]             # <<<<<<<<<<<<<<
 *                 for xi in range(2):
 *                     if axiter[0][xi] == 999: continue
 */
          ((__pyx_v_axiterv[__pyx_v_i])[1]) = (-(__pyx_v_DW[__pyx_v_i]));

          /* "yt/geometry/particle_oct_container.pyx":790
 *                         axiter[i][1] = +1
 *                         axiterv[i][1] = DW[i]
 *                     elif PER[i] and ppos[i] + radius > RE[i]:             # <<<<<<<<<<<<<<
 *                         axiter[i][1] = -1
 *                         axiterv[i][1] = -DW[i]
 */
        }
        __pyx_L22:;
      }

      /* "yt/geometry/particle_oct_container.pyx":793
 *                         axiter[i][1] = -1
 *                         axiterv[i][1] = -DW[i]
 *                 for xi in range(2):             # <<<<<<<<<<<<<<
 *                     if axiter[0][xi] == 999: continue
 *                     s_ppos[0] = ppos[0] + axiterv[0][xi]
 */
      for (__pyx_t_28 = 0; __pyx_t_28 < 2; __pyx_t_28+=1) {
        __pyx_v_xi = __pyx_t_28;

        /* "yt/geometry/particle_oct_container.pyx":794
 *                         axiterv[i][1] = -DW[i]
 *                 for xi in range(2):
 *                     if axiter[0][xi] == 999: continue             # <<<<<<<<<<<<<<
 *                     s_ppos[0] = ppos[0] + axiterv[0][xi]
 *                     for yi in range(2):
 */
        __pyx_t_20 = ((((__pyx_v_axiter[0])[__pyx_v_xi]) == 0x3E7) != 0);
        if (__pyx_t_20) {
          goto __pyx_L27_continue;
        }

        /* "yt/geometry/particle_oct_container.pyx":795
 *                 for xi in range(2):
 *                     if axiter[0][xi] == 999: continue
 *                     s_ppos[0] = ppos[0] + axiterv[0][xi]             # <<<<<<<<<<<<<<
 *                     for yi in range(2):
 *                         if axiter[1][yi] == 999: continue
 */
        (__pyx_v_s_ppos[0]) = ((__pyx_v_ppos[0]) + ((__pyx_v_axiterv[0])[__pyx_v_xi]));

        /* "yt/geometry/particle_oct_container.pyx":796
 *                     if axiter[0][xi] == 999: continue
 *                     s_ppos[0] = ppos[0] + axiterv[0][xi]
 *                     for yi in range(2):             # <<<<<<<<<<<<<<
 *                         if axiter[1][yi] == 999: continue
 *                         s_ppos[1] = ppos[1] + axiterv[1][yi]
 */
        for (__pyx_t_29 = 0; __pyx_t_29 < 2; __pyx_t_29+=1) {
          __pyx_v_yi = __pyx_t_29;

          /* "yt/geometry/particle_oct_container.pyx":797
 *                     s_ppos[0] = ppos[0] + axiterv[0][xi]
 *                     for yi in range(2):
 *                         if axiter[1][yi] == 999: continue             # <<<<<<<<<<<<<<
 *                         s_ppos[1] = ppos[1] + axiterv[1][yi]
 *                         for zi in range(2):
 */
          __pyx_t_20 = ((((__pyx_v_axiter[1])[__pyx_v_yi]) == 0x3E7) != 0);
          if (__pyx_t_20) {
            goto __pyx_L30_continue;
          }

          /* "yt/geometry/particle_oct_container.pyx":798
 *                     for yi in range(2):
 *                         if axiter[1][yi] == 999: continue
 *                         s_ppos[1] = ppos[1] + axiterv[1][yi]             # <<<<<<<<<<<<<<
 *                         for zi in range(2):
 *                             if axiter[2][zi] == 999: continue
 */
          (__pyx_v_s_ppos[1]) = ((__pyx_v_ppos[1]) + ((__pyx_v_axiterv[1])[__pyx_v_yi]));

          /* "yt/geometry/particle_oct_container.pyx":799
 *                         if axiter[1][yi] == 999: continue
 *                         s_ppos[1] = ppos[1] + axiterv[1][yi]
 *                         for zi in range(2):             # <<<<<<<<<<<<<<
 *                             if axiter[2][zi] == 999: continue
 *                             s_ppos[2] = ppos[2] + axiterv[2][zi]
 */
          for (__pyx_t_30 = 0; __pyx_t_30 < 2; __pyx_t_30+=1) {
            __pyx_v_zi = __pyx_t_30;

            /* "yt/geometry/particle_oct_container.pyx":800
 *                         s_ppos[1] = ppos[1] + axiterv[1][yi]
 *                         for zi in range(2):
 *                             if axiter[2][zi] == 999: continue             # <<<<<<<<<<<<<<
 *                             s_ppos[2] = ppos[2] + axiterv[2][zi]
 *                             # OK, now we compute the left and right edges for this shift.
 */
            __pyx_t_20 = ((((__pyx_v_axiter[2])[__pyx_v_zi]) == 0x3E7) != 0);
            if (__pyx_t_20) {
              goto __pyx_L33_continue;
            }

            /* "yt/geometry/particle_oct_container.pyx":801
 *                         for zi in range(2):
 *                             if axiter[2][zi] == 999: continue
 *                             s_ppos[2] = ppos[2] + axiterv[2][zi]             # <<<<<<<<<<<<<<
 *                             # OK, now we compute the left and right edges for this shift.
 *                             for i in range(3):
 */
            (__pyx_v_s_ppos[2]) = ((__pyx_v_ppos[2]) + ((__pyx_v_axiterv[2])[__pyx_v_zi]));

            /* "yt/geometry/particle_oct_container.pyx":803
 *                             s_ppos[2] = ppos[2] + axiterv[2][zi]
 *                             # OK, now we compute the left and right edges for this shift.
 *                             for i in range(3):             # <<<<<<<<<<<<<<
 *                                 # casting to int64 is not nice but is so we can have negative values we clip
 *                                 clip_pos_l[i] = fmax(s_ppos[i] - radius, LE[i] + dds1[i]/10)
 */
            for (__pyx_t_7 = 0; __pyx_t_7 < 3; __pyx_t_7+=1) {
              __pyx_v_i = __pyx_t_7;

              /* "yt/geometry/particle_oct_container.pyx":805
 *                             for i in range(3):
 *                                 # casting to int64 is not nice but is so we can have negative values we clip
 *                                 clip_pos_l[i] = fmax(s_ppos[i] - radius, LE[i] + dds1[i]/10)             # <<<<<<<<<<<<<<
 *                                 clip_pos_r[i] = fmin(s_ppos[i] + radius, RE[i] - dds1[i]/10)
 * 
 */
              (__pyx_v_clip_pos_l[__pyx_v_i]) = __pyx_f_2yt_9utilities_3lib_8fp_utils_fmax(((__pyx_v_s_ppos[__pyx_v_i]) - __pyx_v_radius), ((__pyx_v_LE[__pyx_v_i]) + ((__pyx_v_dds1[__pyx_v_i]) / 10.0)));

              /* "yt/geometry/particle_oct_container.pyx":806
 *                                 # casting to int64 is not nice but is so we can have negative values we clip
 *                                 clip_pos_l[i] = fmax(s_ppos[i] - radius, LE[i] + dds1[i]/10)
 *                                 clip_pos_r[i] = fmin(s_ppos[i] + radius, RE[i] - dds1[i]/10)             # <<<<<<<<<<<<<<
 * 
 *                             bounded_morton_split_dds(clip_pos_l[0], clip_pos_l[1], clip_pos_l[2], LE, dds1, bounds[0])
 */
              (__pyx_v_clip_pos_r[__pyx_v_i]) = __pyx_f_2yt_9utilities_3lib_8fp_utils_fmin(((__pyx_v_s_ppos[__pyx_v_i]) + __pyx_v_radius), ((__pyx_v_RE[__pyx_v_i]) - ((__pyx_v_dds1[__pyx_v_i]) / 10.0)));
            }

            /* "yt/geometry/particle_oct_container.pyx":808
 *                                 clip_pos_r[i] = fmin(s_ppos[i] + radius, RE[i] - dds1[i]/10)
 * 
 *                             bounded_morton_split_dds(clip_pos_l[0], clip_pos_l[1], clip_pos_l[2], LE, dds1, bounds[0])             # <<<<<<<<<<<<<<
 *                             bounded_morton_split_dds(clip_pos_r[0], clip_pos_r[1], clip_pos_r[2], LE, dds1, bounds[1])
 * 
 */
            (void)(__pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_dds((__pyx_v_clip_pos_l[0]), (__pyx_v_clip_pos_l[1]), (__pyx_v_clip_pos_l[2]), __pyx_v_LE, __pyx_v_dds1, (__pyx_v_bounds[0])));

            /* "yt/geometry/particle_oct_container.pyx":809
 * 
 *                             bounded_morton_split_dds(clip_pos_l[0], clip_pos_l[1], clip_pos_l[2], LE, dds1, bounds[0])
 *                             bounded_morton_split_dds(clip_pos_r[0], clip_pos_r[1], clip_pos_r[2], LE, dds1, bounds[1])             # <<<<<<<<<<<<<<
 * 
 *                             # We go to the upper bound plus one so that we have *inclusive* loops -- the upper bound
 */
            (void)(__pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_dds((__pyx_v_clip_pos_r[0]), (__pyx_v_clip_pos_r[1]), (__pyx_v_clip_pos_r[2]), __pyx_v_LE, __pyx_v_dds1, (__pyx_v_bounds[1])));

            /* "yt/geometry/particle_oct_container.pyx":818
 *                             # greater than the left edge and less than the
 *                             # right edge, we are fully enclosed.
 *                             for xex in range(bounds[0][0], bounds[1][0] + 1):             # <<<<<<<<<<<<<<
 *                                 for yex in range(bounds[0][1], bounds[1][1] + 1):
 *                                     for zex in range(bounds[0][2], bounds[1][2] + 1):
 */
            __pyx_t_7 = (((__pyx_v_bounds[1])[0]) + 1);
            __pyx_t_26 = __pyx_t_7;
            for (__pyx_t_31 = ((__pyx_v_bounds[0])[0]); __pyx_t_31 < __pyx_t_26; __pyx_t_31+=1) {
              __pyx_v_xex = __pyx_t_31;

              /* "yt/geometry/particle_oct_container.pyx":819
 *                             # right edge, we are fully enclosed.
 *                             for xex in range(bounds[0][0], bounds[1][0] + 1):
 *                                 for yex in range(bounds[0][1], bounds[1][1] + 1):             # <<<<<<<<<<<<<<
 *                                     for zex in range(bounds[0][2], bounds[1][2] + 1):
 *                                         miex1 = encode_morton_64bit(xex, yex, zex)
 */
              __pyx_t_32 = (((__pyx_v_bounds[1])[1]) + 1);
              __pyx_t_33 = __pyx_t_32;
              for (__pyx_t_34 = ((__pyx_v_bounds[0])[1]); __pyx_t_34 < __pyx_t_33; __pyx_t_34+=1) {
                __pyx_v_yex = __pyx_t_34;

                /* "yt/geometry/particle_oct_container.pyx":820
 *                             for xex in range(bounds[0][0], bounds[1][0] + 1):
 *                                 for yex in range(bounds[0][1], bounds[1][1] + 1):
 *                                     for zex in range(bounds[0][2], bounds[1][2] + 1):             # <<<<<<<<<<<<<<
 *                                         miex1 = encode_morton_64bit(xex, yex, zex)
 *                                         if mask[miex1] < mask_threshold or \
 */
                __pyx_t_35 = (((__pyx_v_bounds[1])[2]) + 1);
                __pyx_t_36 = __pyx_t_35;
                for (__pyx_t_37 = ((__pyx_v_bounds[0])[2]); __pyx_t_37 < __pyx_t_36; __pyx_t_37+=1) {
                  __pyx_v_zex = __pyx_t_37;

                  /* "yt/geometry/particle_oct_container.pyx":821
 *                                 for yex in range(bounds[0][1], bounds[1][1] + 1):
 *                                     for zex in range(bounds[0][2], bounds[1][2] + 1):
 *                                         miex1 = encode_morton_64bit(xex, yex, zex)             # <<<<<<<<<<<<<<
 *                                         if mask[miex1] < mask_threshold or \
 *                                                 particle_counts[miex1] < count_threshold:
 */
                  __pyx_v_miex1 = __pyx_f_2yt_9utilities_3lib_14geometry_utils_encode_morton_64bit(__pyx_v_xex, __pyx_v_yex, __pyx_v_zex);

                  /* "yt/geometry/particle_oct_container.pyx":822
 *                                     for zex in range(bounds[0][2], bounds[1][2] + 1):
 *                                         miex1 = encode_morton_64bit(xex, yex, zex)
 *                                         if mask[miex1] < mask_threshold or \             # <<<<<<<<<<<<<<
 *                                                 particle_counts[miex1] < count_threshold:
 *                                             continue
 */
                  __pyx_t_38 = __pyx_v_miex1;
                  __pyx_t_19 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_mask.rcbuffer->pybuffer.buf, __pyx_t_38, __pyx_pybuffernd_mask.diminfo[0].strides)) < __pyx_v_mask_threshold) != 0);
                  if (!__pyx_t_19) {
                  } else {
                    __pyx_t_20 = __pyx_t_19;
                    goto __pyx_L45_bool_binop_done;
                  }

                  /* "yt/geometry/particle_oct_container.pyx":823
 *                                         miex1 = encode_morton_64bit(xex, yex, zex)
 *                                         if mask[miex1] < mask_threshold or \
 *                                                 particle_counts[miex1] < count_threshold:             # <<<<<<<<<<<<<<
 *                                             continue
 *                                         # this explicitly requires that it be *between*
 */
                  __pyx_t_38 = __pyx_v_miex1;
                  __pyx_t_19 = (((*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_particle_counts.data + __pyx_t_38 * __pyx_v_particle_counts.strides[0]) ))) < __pyx_v_count_threshold) != 0);
                  __pyx_t_20 = __pyx_t_19;
                  __pyx_L45_bool_binop_done:;

                  /* "yt/geometry/particle_oct_container.pyx":822
 *                                     for zex in range(bounds[0][2], bounds[1][2] + 1):
 *                                         miex1 = encode_morton_64bit(xex, yex, zex)
 *                                         if mask[miex1] < mask_threshold or \             # <<<<<<<<<<<<<<
 *                                                 particle_counts[miex1] < count_threshold:
 *                                             continue
 */
                  if (__pyx_t_20) {

                    /* "yt/geometry/particle_oct_container.pyx":824
 *                                         if mask[miex1] < mask_threshold or \
 *                                                 particle_counts[miex1] < count_threshold:
 *                                             continue             # <<<<<<<<<<<<<<
 *                                         # this explicitly requires that it be *between*
 *                                         # them, not overlapping
 */
                    goto __pyx_L42_continue;

                    /* "yt/geometry/particle_oct_container.pyx":822
 *                                     for zex in range(bounds[0][2], bounds[1][2] + 1):
 *                                         miex1 = encode_morton_64bit(xex, yex, zex)
 *                                         if mask[miex1] < mask_threshold or \             # <<<<<<<<<<<<<<
 *                                                 particle_counts[miex1] < count_threshold:
 *                                             continue
 */
                  }

                  /* "yt/geometry/particle_oct_container.pyx":827
 *                                         # this explicitly requires that it be *between*
 *                                         # them, not overlapping
 *                                         if xex > bounds[0][0] and xex < bounds[1][0] and \             # <<<<<<<<<<<<<<
 *                                            yex > bounds[0][1] and yex < bounds[1][1] and \
 *                                            zex > bounds[0][2] and zex < bounds[1][2]:
 */
                  __pyx_t_19 = ((__pyx_v_xex > ((__pyx_v_bounds[0])[0])) != 0);
                  if (__pyx_t_19) {
                  } else {
                    __pyx_t_20 = __pyx_t_19;
                    goto __pyx_L48_bool_binop_done;
                  }
                  __pyx_t_19 = ((__pyx_v_xex < ((__pyx_v_bounds[1])[0])) != 0);
                  if (__pyx_t_19) {
                  } else {
                    __pyx_t_20 = __pyx_t_19;
                    goto __pyx_L48_bool_binop_done;
                  }

                  /* "yt/geometry/particle_oct_container.pyx":828
 *                                         # them, not overlapping
 *                                         if xex > bounds[0][0] and xex < bounds[1][0] and \
 *                                            yex > bounds[0][1] and yex < bounds[1][1] and \             # <<<<<<<<<<<<<<
 *                                            zex > bounds[0][2] and zex < bounds[1][2]:
 *                                             fully_enclosed = 1
 */
                  __pyx_t_19 = ((__pyx_v_yex > ((__pyx_v_bounds[0])[1])) != 0);
                  if (__pyx_t_19) {
                  } else {
                    __pyx_t_20 = __pyx_t_19;
                    goto __pyx_L48_bool_binop_done;
                  }
                  __pyx_t_19 = ((__pyx_v_yex < ((__pyx_v_bounds[1])[1])) != 0);
                  if (__pyx_t_19) {
                  } else {
                    __pyx_t_20 = __pyx_t_19;
                    goto __pyx_L48_bool_binop_done;
                  }

                  /* "yt/geometry/particle_oct_container.pyx":829
 *                                         if xex > bounds[0][0] and xex < bounds[1][0] and \
 *                                            yex > bounds[0][1] and yex < bounds[1][1] and \
 *                                            zex > bounds[0][2] and zex < bounds[1][2]:             # <<<<<<<<<<<<<<
 *                                             fully_enclosed = 1
 *                                         else:
 */
                  __pyx_t_19 = ((__pyx_v_zex > ((__pyx_v_bounds[0])[2])) != 0);
                  if (__pyx_t_19) {
                  } else {
                    __pyx_t_20 = __pyx_t_19;
                    goto __pyx_L48_bool_binop_done;
                  }
                  __pyx_t_19 = ((__pyx_v_zex < ((__pyx_v_bounds[1])[2])) != 0);
                  __pyx_t_20 = __pyx_t_19;
                  __pyx_L48_bool_binop_done:;

                  /* "yt/geometry/particle_oct_container.pyx":827
 *                                         # this explicitly requires that it be *between*
 *                                         # them, not overlapping
 *                                         if xex > bounds[0][0] and xex < bounds[1][0] and \             # <<<<<<<<<<<<<<
 *                                            yex > bounds[0][1] and yex < bounds[1][1] and \
 *                                            zex > bounds[0][2] and zex < bounds[1][2]:
 */
                  if (__pyx_t_20) {

                    /* "yt/geometry/particle_oct_container.pyx":830
 *                                            yex > bounds[0][1] and yex < bounds[1][1] and \
 *                                            zex > bounds[0][2] and zex < bounds[1][2]:
 *                                             fully_enclosed = 1             # <<<<<<<<<<<<<<
 *                                         else:
 *                                             fully_enclosed = 0
 */
                    __pyx_v_fully_enclosed = 1;

                    /* "yt/geometry/particle_oct_container.pyx":827
 *                                         # this explicitly requires that it be *between*
 *                                         # them, not overlapping
 *                                         if xex > bounds[0][0] and xex < bounds[1][0] and \             # <<<<<<<<<<<<<<
 *                                            yex > bounds[0][1] and yex < bounds[1][1] and \
 *                                            zex > bounds[0][2] and zex < bounds[1][2]:
 */
                    goto __pyx_L47;
                  }

                  /* "yt/geometry/particle_oct_container.pyx":832
 *                                             fully_enclosed = 1
 *                                         else:
 *                                             fully_enclosed = 0             # <<<<<<<<<<<<<<
 *                                         # Now we need to fill our sub-range
 *                                         if refined_count[miex1] == 0:
 */
                  /*else*/ {
                    __pyx_v_fully_enclosed = 0;
                  }
                  __pyx_L47:;

                  /* "yt/geometry/particle_oct_container.pyx":834
 *                                             fully_enclosed = 0
 *                                         # Now we need to fill our sub-range
 *                                         if refined_count[miex1] == 0:             # <<<<<<<<<<<<<<
 *                                             coarse_refined_map[miex1].padWithZeroes(max_mi2_elements)
 *                                         elif refined_count[miex1] >= max_mi2_elements:
 */
                  __pyx_t_38 = __pyx_v_miex1;
                  __pyx_t_20 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_refined_count.rcbuffer->pybuffer.buf, __pyx_t_38, __pyx_pybuffernd_refined_count.diminfo[0].strides)) == 0) != 0);
                  if (__pyx_t_20) {

                    /* "yt/geometry/particle_oct_container.pyx":835
 *                                         # Now we need to fill our sub-range
 *                                         if refined_count[miex1] == 0:
 *                                             coarse_refined_map[miex1].padWithZeroes(max_mi2_elements)             # <<<<<<<<<<<<<<
 *                                         elif refined_count[miex1] >= max_mi2_elements:
 *                                             continue
 */
                    (void)((__pyx_v_coarse_refined_map[__pyx_v_miex1]).padWithZeroes(__pyx_v_max_mi2_elements));

                    /* "yt/geometry/particle_oct_container.pyx":834
 *                                             fully_enclosed = 0
 *                                         # Now we need to fill our sub-range
 *                                         if refined_count[miex1] == 0:             # <<<<<<<<<<<<<<
 *                                             coarse_refined_map[miex1].padWithZeroes(max_mi2_elements)
 *                                         elif refined_count[miex1] >= max_mi2_elements:
 */
                    goto __pyx_L54;
                  }

                  /* "yt/geometry/particle_oct_container.pyx":836
 *                                         if refined_count[miex1] == 0:
 *                                             coarse_refined_map[miex1].padWithZeroes(max_mi2_elements)
 *                                         elif refined_count[miex1] >= max_mi2_elements:             # <<<<<<<<<<<<<<
 *                                             continue
 *                                         if fully_enclosed == 1:
 */
                  __pyx_t_38 = __pyx_v_miex1;
                  __pyx_t_20 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_refined_count.rcbuffer->pybuffer.buf, __pyx_t_38, __pyx_pybuffernd_refined_count.diminfo[0].strides)) >= __pyx_v_max_mi2_elements) != 0);
                  if (__pyx_t_20) {

                    /* "yt/geometry/particle_oct_container.pyx":837
 *                                             coarse_refined_map[miex1].padWithZeroes(max_mi2_elements)
 *                                         elif refined_count[miex1] >= max_mi2_elements:
 *                                             continue             # <<<<<<<<<<<<<<
 *                                         if fully_enclosed == 1:
 *                                             nfully_enclosed += 1
 */
                    goto __pyx_L42_continue;

                    /* "yt/geometry/particle_oct_container.pyx":836
 *                                         if refined_count[miex1] == 0:
 *                                             coarse_refined_map[miex1].padWithZeroes(max_mi2_elements)
 *                                         elif refined_count[miex1] >= max_mi2_elements:             # <<<<<<<<<<<<<<
 *                                             continue
 *                                         if fully_enclosed == 1:
 */
                  }
                  __pyx_L54:;

                  /* "yt/geometry/particle_oct_container.pyx":838
 *                                         elif refined_count[miex1] >= max_mi2_elements:
 *                                             continue
 *                                         if fully_enclosed == 1:             # <<<<<<<<<<<<<<
 *                                             nfully_enclosed += 1
 *                                             coarse_refined_map[miex1].inplace_logicalxor(
 */
                  __pyx_t_20 = ((__pyx_v_fully_enclosed == 1) != 0);
                  if (__pyx_t_20) {

                    /* "yt/geometry/particle_oct_container.pyx":839
 *                                             continue
 *                                         if fully_enclosed == 1:
 *                                             nfully_enclosed += 1             # <<<<<<<<<<<<<<
 *                                             coarse_refined_map[miex1].inplace_logicalxor(
 *                                                 coarse_refined_map[miex1])
 */
                    __pyx_v_nfully_enclosed = (__pyx_v_nfully_enclosed + 1);

                    /* "yt/geometry/particle_oct_container.pyx":840
 *                                         if fully_enclosed == 1:
 *                                             nfully_enclosed += 1
 *                                             coarse_refined_map[miex1].inplace_logicalxor(             # <<<<<<<<<<<<<<
 *                                                 coarse_refined_map[miex1])
 *                                             coarse_refined_map[miex1].inplace_logicalnot()
 */
                    (__pyx_v_coarse_refined_map[__pyx_v_miex1]).inplace_logicalxor((__pyx_v_coarse_refined_map[__pyx_v_miex1]));

                    /* "yt/geometry/particle_oct_container.pyx":842
 *                                             coarse_refined_map[miex1].inplace_logicalxor(
 *                                                 coarse_refined_map[miex1])
 *                                             coarse_refined_map[miex1].inplace_logicalnot()             # <<<<<<<<<<<<<<
 *                                             refined_count[miex1] = max_mi2_elements
 *                                             continue
 */
                    (__pyx_v_coarse_refined_map[__pyx_v_miex1]).inplace_logicalnot();

                    /* "yt/geometry/particle_oct_container.pyx":843
 *                                                 coarse_refined_map[miex1])
 *                                             coarse_refined_map[miex1].inplace_logicalnot()
 *                                             refined_count[miex1] = max_mi2_elements             # <<<<<<<<<<<<<<
 *                                             continue
 *                                         n_calls += 1
 */
                    __pyx_t_38 = __pyx_v_miex1;
                    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_refined_count.rcbuffer->pybuffer.buf, __pyx_t_38, __pyx_pybuffernd_refined_count.diminfo[0].strides) = __pyx_v_max_mi2_elements;

                    /* "yt/geometry/particle_oct_container.pyx":844
 *                                             coarse_refined_map[miex1].inplace_logicalnot()
 *                                             refined_count[miex1] = max_mi2_elements
 *                                             continue             # <<<<<<<<<<<<<<
 *                                         n_calls += 1
 *                                         refined_count[miex1] += self.__fill_refined_ranges(s_ppos, radius, LE, RE,
 */
                    goto __pyx_L42_continue;

                    /* "yt/geometry/particle_oct_container.pyx":838
 *                                         elif refined_count[miex1] >= max_mi2_elements:
 *                                             continue
 *                                         if fully_enclosed == 1:             # <<<<<<<<<<<<<<
 *                                             nfully_enclosed += 1
 *                                             coarse_refined_map[miex1].inplace_logicalxor(
 */
                  }

                  /* "yt/geometry/particle_oct_container.pyx":845
 *                                             refined_count[miex1] = max_mi2_elements
 *                                             continue
 *                                         n_calls += 1             # <<<<<<<<<<<<<<
 *                                         refined_count[miex1] += self.__fill_refined_ranges(s_ppos, radius, LE, RE,
 *                                                                    dds1, xex, yex, zex,
 */
                  __pyx_v_n_calls = (__pyx_v_n_calls + 1);

                  /* "yt/geometry/particle_oct_container.pyx":846
 *                                             continue
 *                                         n_calls += 1
 *                                         refined_count[miex1] += self.__fill_refined_ranges(s_ppos, radius, LE, RE,             # <<<<<<<<<<<<<<
 *                                                                    dds1, xex, yex, zex,
 *                                                                    dds2,
 */
                  __pyx_t_16 = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self->__pyx_vtab)->__pyx___fill_refined_ranges(__pyx_v_self, __pyx_v_s_ppos, __pyx_v_radius, __pyx_v_LE, __pyx_v_RE, __pyx_v_dds1, __pyx_v_xex, __pyx_v_yex, __pyx_v_zex, __pyx_v_dds2, (__pyx_v_coarse_refined_map[__pyx_v_miex1])); if (unlikely(__pyx_t_16 == ((__pyx_t_5numpy_int64_t)-1LL))) __PYX_ERR(0, 846, __pyx_L1_error)
                  __pyx_t_38 = __pyx_v_miex1;
                  *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_refined_count.rcbuffer->pybuffer.buf, __pyx_t_38, __pyx_pybuffernd_refined_count.diminfo[0].strides) += __pyx_t_16;
                  __pyx_L42_continue:;
                }
              }
            }
            __pyx_L33_continue:;
          }
          __pyx_L30_continue:;
        }
        __pyx_L27_continue:;
      }
    }
    __pyx_L14:;
    __pyx_L8_continue:;
  }

  /* "yt/geometry/particle_oct_container.pyx":851
 *                                                                    coarse_refined_map[miex1])
 *         cdef np.uint64_t vec_i
 *         cdef bool_array *buf = NULL             # <<<<<<<<<<<<<<
 *         cdef ewah_word_type w
 *         this_collection = BoolArrayCollection()
 */
  __pyx_v_buf = NULL;

  /* "yt/geometry/particle_oct_container.pyx":853
 *         cdef bool_array *buf = NULL
 *         cdef ewah_word_type w
 *         this_collection = BoolArrayCollection()             # <<<<<<<<<<<<<<
 *         cdef ewah_bool_array *refined_arr = NULL
 *         for it1 in coarse_refined_map:
 */
  __pyx_t_4 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_this_collection = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/particle_oct_container.pyx":854
 *         cdef ewah_word_type w
 *         this_collection = BoolArrayCollection()
 *         cdef ewah_bool_array *refined_arr = NULL             # <<<<<<<<<<<<<<
 *         for it1 in coarse_refined_map:
 *             mi1 = it1.first
 */
  __pyx_v_refined_arr = NULL;

  /* "yt/geometry/particle_oct_container.pyx":855
 *         this_collection = BoolArrayCollection()
 *         cdef ewah_bool_array *refined_arr = NULL
 *         for it1 in coarse_refined_map:             # <<<<<<<<<<<<<<
 *             mi1 = it1.first
 *             refined_arr = &this_collection.ewah_coll[0][mi1]
 */
  __pyx_t_39 = __pyx_v_coarse_refined_map.begin();
  for (;;) {
    if (!(__pyx_t_39 != __pyx_v_coarse_refined_map.end())) break;
    __pyx_t_40 = *__pyx_t_39;
    ++__pyx_t_39;
    __pyx_v_it1 = __pyx_t_40;

    /* "yt/geometry/particle_oct_container.pyx":856
 *         cdef ewah_bool_array *refined_arr = NULL
 *         for it1 in coarse_refined_map:
 *             mi1 = it1.first             # <<<<<<<<<<<<<<
 *             refined_arr = &this_collection.ewah_coll[0][mi1]
 *             this_collection.ewah_keys[0].set(mi1)
 */
    __pyx_t_7 = __pyx_v_it1.first;
    __pyx_v_mi1 = __pyx_t_7;

    /* "yt/geometry/particle_oct_container.pyx":857
 *         for it1 in coarse_refined_map:
 *             mi1 = it1.first
 *             refined_arr = &this_collection.ewah_coll[0][mi1]             # <<<<<<<<<<<<<<
 *             this_collection.ewah_keys[0].set(mi1)
 *             this_collection.ewah_refn[0].set(mi1)
 */
    __pyx_v_refined_arr = (&((__pyx_v_this_collection->ewah_coll[0])[__pyx_v_mi1]));

    /* "yt/geometry/particle_oct_container.pyx":858
 *             mi1 = it1.first
 *             refined_arr = &this_collection.ewah_coll[0][mi1]
 *             this_collection.ewah_keys[0].set(mi1)             # <<<<<<<<<<<<<<
 *             this_collection.ewah_refn[0].set(mi1)
 *             buf = &it1.second
 */
    (void)((__pyx_v_this_collection->ewah_keys[0]).set(__pyx_v_mi1));

    /* "yt/geometry/particle_oct_container.pyx":859
 *             refined_arr = &this_collection.ewah_coll[0][mi1]
 *             this_collection.ewah_keys[0].set(mi1)
 *             this_collection.ewah_refn[0].set(mi1)             # <<<<<<<<<<<<<<
 *             buf = &it1.second
 *             for vec_i in range(buf.sizeInBytes() / sizeof(ewah_word_type)):
 */
    (void)((__pyx_v_this_collection->ewah_refn[0]).set(__pyx_v_mi1));

    /* "yt/geometry/particle_oct_container.pyx":860
 *             this_collection.ewah_keys[0].set(mi1)
 *             this_collection.ewah_refn[0].set(mi1)
 *             buf = &it1.second             # <<<<<<<<<<<<<<
 *             for vec_i in range(buf.sizeInBytes() / sizeof(ewah_word_type)):
 *                 w = buf.getWord(vec_i)
 */
    __pyx_v_buf = (&__pyx_v_it1.second);

    /* "yt/geometry/particle_oct_container.pyx":861
 *             this_collection.ewah_refn[0].set(mi1)
 *             buf = &it1.second
 *             for vec_i in range(buf.sizeInBytes() / sizeof(ewah_word_type)):             # <<<<<<<<<<<<<<
 *                 w = buf.getWord(vec_i)
 *                 refined_arr.addWord(w)
 */
    __pyx_t_41 = (__pyx_v_buf->sizeInBytes() / (sizeof(__pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_word_type)));
    __pyx_t_42 = __pyx_t_41;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_42; __pyx_t_7+=1) {
      __pyx_v_vec_i = __pyx_t_7;

      /* "yt/geometry/particle_oct_container.pyx":862
 *             buf = &it1.second
 *             for vec_i in range(buf.sizeInBytes() / sizeof(ewah_word_type)):
 *                 w = buf.getWord(vec_i)             # <<<<<<<<<<<<<<
 *                 refined_arr.addWord(w)
 *         out_collection = BoolArrayCollection()
 */
      __pyx_v_w = __pyx_v_buf->getWord(__pyx_v_vec_i);

      /* "yt/geometry/particle_oct_container.pyx":863
 *             for vec_i in range(buf.sizeInBytes() / sizeof(ewah_word_type)):
 *                 w = buf.getWord(vec_i)
 *                 refined_arr.addWord(w)             # <<<<<<<<<<<<<<
 *         out_collection = BoolArrayCollection()
 *         in_collection._logicalor(this_collection, out_collection)
 */
      (void)(__pyx_v_refined_arr->addWord(__pyx_v_w));
    }

    /* "yt/geometry/particle_oct_container.pyx":855
 *         this_collection = BoolArrayCollection()
 *         cdef ewah_bool_array *refined_arr = NULL
 *         for it1 in coarse_refined_map:             # <<<<<<<<<<<<<<
 *             mi1 = it1.first
 *             refined_arr = &this_collection.ewah_coll[0][mi1]
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":864
 *                 w = buf.getWord(vec_i)
 *                 refined_arr.addWord(w)
 *         out_collection = BoolArrayCollection()             # <<<<<<<<<<<<<<
 *         in_collection._logicalor(this_collection, out_collection)
 *         return out_collection
 */
  __pyx_t_4 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_out_collection = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/particle_oct_container.pyx":865
 *                 refined_arr.addWord(w)
 *         out_collection = BoolArrayCollection()
 *         in_collection._logicalor(this_collection, out_collection)             # <<<<<<<<<<<<<<
 *         return out_collection
 * 
 */
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_v_in_collection->__pyx_vtab)->_logicalor(__pyx_v_in_collection, __pyx_v_this_collection, __pyx_v_out_collection);

  /* "yt/geometry/particle_oct_container.pyx":866
 *         out_collection = BoolArrayCollection()
 *         in_collection._logicalor(this_collection, out_collection)
 *         return out_collection             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_out_collection));
  __pyx_r = __pyx_v_out_collection;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":691
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef BoolArrayCollection __refined_index_data_file(             # <<<<<<<<<<<<<<
 *         self,
 *         BoolArrayCollection in_collection,
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_1, 1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_morton_indices.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_refined_count.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sorted_order.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.__refined_index_data_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_hsml.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_mask.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_morton_indices.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_refined_count.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sorted_order.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_this_collection);
  __Pyx_XDECREF((PyObject *)__pyx_v_out_collection);
  __PYX_XDEC_MEMVIEW(&__pyx_v_particle_counts, 1);
  __Pyx_XDECREF((PyObject *)__pyx_v_refined_count);
  __Pyx_XDECREF((PyObject *)__pyx_v_morton_indices);
  __Pyx_XDECREF((PyObject *)__pyx_v_sorted_order);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":872
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef np.int64_t __fill_refined_ranges(self, np.float64_t s_ppos[3], np.float64_t radius,             # <<<<<<<<<<<<<<
 *                                            np.float64_t LE[3], np.float64_t RE[3],
 *                                            np.float64_t dds1[3], np.uint64_t xex, np.uint64_t yex, np.uint64_t zex,
 */

static __pyx_t_5numpy_int64_t __pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___fill_refined_ranges(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, __pyx_t_5numpy_float64_t *__pyx_v_s_ppos, __pyx_t_5numpy_float64_t __pyx_v_radius, __pyx_t_5numpy_float64_t *__pyx_v_LE, CYTHON_UNUSED __pyx_t_5numpy_float64_t *__pyx_v_RE, __pyx_t_5numpy_float64_t *__pyx_v_dds1, __pyx_t_5numpy_uint64_t __pyx_v_xex, __pyx_t_5numpy_uint64_t __pyx_v_yex, __pyx_t_5numpy_uint64_t __pyx_v_zex, __pyx_t_5numpy_float64_t *__pyx_v_dds2, __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_bool_array &__pyx_v_refined_set) {
  int __pyx_v_i;
  __pyx_t_5numpy_uint64_t __pyx_v_bounds_l[3];
  __pyx_t_5numpy_uint64_t __pyx_v_bounds_r[3];
  __pyx_t_5numpy_uint64_t __pyx_v_miex2;
  __pyx_t_5numpy_uint64_t __pyx_v_miex2_min;
  __pyx_t_5numpy_uint64_t __pyx_v_miex2_max;
  __pyx_t_5numpy_float64_t __pyx_v_clip_pos_l[3];
  __pyx_t_5numpy_float64_t __pyx_v_clip_pos_r[3];
  __pyx_t_5numpy_float64_t __pyx_v_cell_edge_l;
  __pyx_t_5numpy_float64_t __pyx_v_cell_edge_r;
  __pyx_t_5numpy_uint64_t __pyx_v_ex1[3];
  __pyx_t_5numpy_uint64_t __pyx_v_xiex_min;
  __pyx_t_5numpy_uint64_t __pyx_v_yiex_min;
  __pyx_t_5numpy_uint64_t __pyx_v_ziex_min;
  __pyx_t_5numpy_uint64_t __pyx_v_xiex_max;
  __pyx_t_5numpy_uint64_t __pyx_v_yiex_max;
  __pyx_t_5numpy_uint64_t __pyx_v_ziex_max;
  __pyx_t_5numpy_uint64_t __pyx_v_old_nsub;
  __pyx_t_5numpy_uint64_t __pyx_v_xex_max;
  __pyx_t_5numpy_uint64_t __pyx_v_yex_max;
  __pyx_t_5numpy_uint64_t __pyx_v_zex_max;
  __pyx_t_5numpy_int64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __pyx_t_5numpy_uint64_t __pyx_t_4;
  __pyx_t_5numpy_uint64_t __pyx_t_5;
  __pyx_t_5numpy_uint64_t __pyx_t_6;
  __Pyx_RefNannySetupContext("__fill_refined_ranges", 0);

  /* "yt/geometry/particle_oct_container.pyx":885
 *         cdef np.uint64_t xiex_min, yiex_min, ziex_min
 *         cdef np.uint64_t xiex_max, yiex_max, ziex_max
 *         cdef np.uint64_t old_nsub = refined_set.numberOfOnes()             # <<<<<<<<<<<<<<
 *         ex1[0] = xex
 *         ex1[1] = yex
 */
  __pyx_v_old_nsub = __pyx_v_refined_set.numberOfOnes();

  /* "yt/geometry/particle_oct_container.pyx":886
 *         cdef np.uint64_t xiex_max, yiex_max, ziex_max
 *         cdef np.uint64_t old_nsub = refined_set.numberOfOnes()
 *         ex1[0] = xex             # <<<<<<<<<<<<<<
 *         ex1[1] = yex
 *         ex1[2] = zex
 */
  (__pyx_v_ex1[0]) = __pyx_v_xex;

  /* "yt/geometry/particle_oct_container.pyx":887
 *         cdef np.uint64_t old_nsub = refined_set.numberOfOnes()
 *         ex1[0] = xex
 *         ex1[1] = yex             # <<<<<<<<<<<<<<
 *         ex1[2] = zex
 *         # Check a few special cases
 */
  (__pyx_v_ex1[1]) = __pyx_v_yex;

  /* "yt/geometry/particle_oct_container.pyx":888
 *         ex1[0] = xex
 *         ex1[1] = yex
 *         ex1[2] = zex             # <<<<<<<<<<<<<<
 *         # Check a few special cases
 *         for i in range(3):
 */
  (__pyx_v_ex1[2]) = __pyx_v_zex;

  /* "yt/geometry/particle_oct_container.pyx":890
 *         ex1[2] = zex
 *         # Check a few special cases
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             # Figure out our bounds inside our coarse cell, in the space of the
 *             # full domain
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":893
 *             # Figure out our bounds inside our coarse cell, in the space of the
 *             # full domain
 *             cell_edge_l = ex1[i] * dds1[i] + LE[i]             # <<<<<<<<<<<<<<
 *             cell_edge_r = cell_edge_l + dds1[i]
 *             if s_ppos[i] + radius < cell_edge_l or s_ppos[i] - radius > cell_edge_r:
 */
    __pyx_v_cell_edge_l = (((__pyx_v_ex1[__pyx_v_i]) * (__pyx_v_dds1[__pyx_v_i])) + (__pyx_v_LE[__pyx_v_i]));

    /* "yt/geometry/particle_oct_container.pyx":894
 *             # full domain
 *             cell_edge_l = ex1[i] * dds1[i] + LE[i]
 *             cell_edge_r = cell_edge_l + dds1[i]             # <<<<<<<<<<<<<<
 *             if s_ppos[i] + radius < cell_edge_l or s_ppos[i] - radius > cell_edge_r:
 *                 return 0
 */
    __pyx_v_cell_edge_r = (__pyx_v_cell_edge_l + (__pyx_v_dds1[__pyx_v_i]));

    /* "yt/geometry/particle_oct_container.pyx":895
 *             cell_edge_l = ex1[i] * dds1[i] + LE[i]
 *             cell_edge_r = cell_edge_l + dds1[i]
 *             if s_ppos[i] + radius < cell_edge_l or s_ppos[i] - radius > cell_edge_r:             # <<<<<<<<<<<<<<
 *                 return 0
 *             clip_pos_l[i] = fmax(s_ppos[i] - radius, cell_edge_l + dds2[i]/2.0)
 */
    __pyx_t_3 = ((((__pyx_v_s_ppos[__pyx_v_i]) + __pyx_v_radius) < __pyx_v_cell_edge_l) != 0);
    if (!__pyx_t_3) {
    } else {
      __pyx_t_2 = __pyx_t_3;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_3 = ((((__pyx_v_s_ppos[__pyx_v_i]) - __pyx_v_radius) > __pyx_v_cell_edge_r) != 0);
    __pyx_t_2 = __pyx_t_3;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_2) {

      /* "yt/geometry/particle_oct_container.pyx":896
 *             cell_edge_r = cell_edge_l + dds1[i]
 *             if s_ppos[i] + radius < cell_edge_l or s_ppos[i] - radius > cell_edge_r:
 *                 return 0             # <<<<<<<<<<<<<<
 *             clip_pos_l[i] = fmax(s_ppos[i] - radius, cell_edge_l + dds2[i]/2.0)
 *             clip_pos_r[i] = fmin(s_ppos[i] + radius, cell_edge_r - dds2[i]/2.0)
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "yt/geometry/particle_oct_container.pyx":895
 *             cell_edge_l = ex1[i] * dds1[i] + LE[i]
 *             cell_edge_r = cell_edge_l + dds1[i]
 *             if s_ppos[i] + radius < cell_edge_l or s_ppos[i] - radius > cell_edge_r:             # <<<<<<<<<<<<<<
 *                 return 0
 *             clip_pos_l[i] = fmax(s_ppos[i] - radius, cell_edge_l + dds2[i]/2.0)
 */
    }

    /* "yt/geometry/particle_oct_container.pyx":897
 *             if s_ppos[i] + radius < cell_edge_l or s_ppos[i] - radius > cell_edge_r:
 *                 return 0
 *             clip_pos_l[i] = fmax(s_ppos[i] - radius, cell_edge_l + dds2[i]/2.0)             # <<<<<<<<<<<<<<
 *             clip_pos_r[i] = fmin(s_ppos[i] + radius, cell_edge_r - dds2[i]/2.0)
 *         miex2_min = bounded_morton_split_relative_dds(clip_pos_l[0], clip_pos_l[1], clip_pos_l[2],
 */
    (__pyx_v_clip_pos_l[__pyx_v_i]) = __pyx_f_2yt_9utilities_3lib_8fp_utils_fmax(((__pyx_v_s_ppos[__pyx_v_i]) - __pyx_v_radius), (__pyx_v_cell_edge_l + ((__pyx_v_dds2[__pyx_v_i]) / 2.0)));

    /* "yt/geometry/particle_oct_container.pyx":898
 *                 return 0
 *             clip_pos_l[i] = fmax(s_ppos[i] - radius, cell_edge_l + dds2[i]/2.0)
 *             clip_pos_r[i] = fmin(s_ppos[i] + radius, cell_edge_r - dds2[i]/2.0)             # <<<<<<<<<<<<<<
 *         miex2_min = bounded_morton_split_relative_dds(clip_pos_l[0], clip_pos_l[1], clip_pos_l[2],
 *                                                 LE, dds1, dds2, bounds_l)
 */
    (__pyx_v_clip_pos_r[__pyx_v_i]) = __pyx_f_2yt_9utilities_3lib_8fp_utils_fmin(((__pyx_v_s_ppos[__pyx_v_i]) + __pyx_v_radius), (__pyx_v_cell_edge_r - ((__pyx_v_dds2[__pyx_v_i]) / 2.0)));
  }

  /* "yt/geometry/particle_oct_container.pyx":899
 *             clip_pos_l[i] = fmax(s_ppos[i] - radius, cell_edge_l + dds2[i]/2.0)
 *             clip_pos_r[i] = fmin(s_ppos[i] + radius, cell_edge_r - dds2[i]/2.0)
 *         miex2_min = bounded_morton_split_relative_dds(clip_pos_l[0], clip_pos_l[1], clip_pos_l[2],             # <<<<<<<<<<<<<<
 *                                                 LE, dds1, dds2, bounds_l)
 *         miex2_max = bounded_morton_split_relative_dds(clip_pos_r[0], clip_pos_r[1], clip_pos_r[2],
 */
  __pyx_v_miex2_min = __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_relative_dds((__pyx_v_clip_pos_l[0]), (__pyx_v_clip_pos_l[1]), (__pyx_v_clip_pos_l[2]), __pyx_v_LE, __pyx_v_dds1, __pyx_v_dds2, __pyx_v_bounds_l);

  /* "yt/geometry/particle_oct_container.pyx":901
 *         miex2_min = bounded_morton_split_relative_dds(clip_pos_l[0], clip_pos_l[1], clip_pos_l[2],
 *                                                 LE, dds1, dds2, bounds_l)
 *         miex2_max = bounded_morton_split_relative_dds(clip_pos_r[0], clip_pos_r[1], clip_pos_r[2],             # <<<<<<<<<<<<<<
 *                                                 LE, dds1, dds2, bounds_r)
 *         xex_max = self.directional_max2[0]
 */
  __pyx_v_miex2_max = __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_relative_dds((__pyx_v_clip_pos_r[0]), (__pyx_v_clip_pos_r[1]), (__pyx_v_clip_pos_r[2]), __pyx_v_LE, __pyx_v_dds1, __pyx_v_dds2, __pyx_v_bounds_r);

  /* "yt/geometry/particle_oct_container.pyx":903
 *         miex2_max = bounded_morton_split_relative_dds(clip_pos_r[0], clip_pos_r[1], clip_pos_r[2],
 *                                                 LE, dds1, dds2, bounds_r)
 *         xex_max = self.directional_max2[0]             # <<<<<<<<<<<<<<
 *         yex_max = self.directional_max2[1]
 *         zex_max = self.directional_max2[2]
 */
  __pyx_v_xex_max = (__pyx_v_self->directional_max2[0]);

  /* "yt/geometry/particle_oct_container.pyx":904
 *                                                 LE, dds1, dds2, bounds_r)
 *         xex_max = self.directional_max2[0]
 *         yex_max = self.directional_max2[1]             # <<<<<<<<<<<<<<
 *         zex_max = self.directional_max2[2]
 *         xiex_min = miex2_min & xex_max
 */
  __pyx_v_yex_max = (__pyx_v_self->directional_max2[1]);

  /* "yt/geometry/particle_oct_container.pyx":905
 *         xex_max = self.directional_max2[0]
 *         yex_max = self.directional_max2[1]
 *         zex_max = self.directional_max2[2]             # <<<<<<<<<<<<<<
 *         xiex_min = miex2_min & xex_max
 *         yiex_min = miex2_min & yex_max
 */
  __pyx_v_zex_max = (__pyx_v_self->directional_max2[2]);

  /* "yt/geometry/particle_oct_container.pyx":906
 *         yex_max = self.directional_max2[1]
 *         zex_max = self.directional_max2[2]
 *         xiex_min = miex2_min & xex_max             # <<<<<<<<<<<<<<
 *         yiex_min = miex2_min & yex_max
 *         ziex_min = miex2_min & zex_max
 */
  __pyx_v_xiex_min = (__pyx_v_miex2_min & __pyx_v_xex_max);

  /* "yt/geometry/particle_oct_container.pyx":907
 *         zex_max = self.directional_max2[2]
 *         xiex_min = miex2_min & xex_max
 *         yiex_min = miex2_min & yex_max             # <<<<<<<<<<<<<<
 *         ziex_min = miex2_min & zex_max
 *         xiex_max = miex2_max & xex_max
 */
  __pyx_v_yiex_min = (__pyx_v_miex2_min & __pyx_v_yex_max);

  /* "yt/geometry/particle_oct_container.pyx":908
 *         xiex_min = miex2_min & xex_max
 *         yiex_min = miex2_min & yex_max
 *         ziex_min = miex2_min & zex_max             # <<<<<<<<<<<<<<
 *         xiex_max = miex2_max & xex_max
 *         yiex_max = miex2_max & yex_max
 */
  __pyx_v_ziex_min = (__pyx_v_miex2_min & __pyx_v_zex_max);

  /* "yt/geometry/particle_oct_container.pyx":909
 *         yiex_min = miex2_min & yex_max
 *         ziex_min = miex2_min & zex_max
 *         xiex_max = miex2_max & xex_max             # <<<<<<<<<<<<<<
 *         yiex_max = miex2_max & yex_max
 *         ziex_max = miex2_max & zex_max
 */
  __pyx_v_xiex_max = (__pyx_v_miex2_max & __pyx_v_xex_max);

  /* "yt/geometry/particle_oct_container.pyx":910
 *         ziex_min = miex2_min & zex_max
 *         xiex_max = miex2_max & xex_max
 *         yiex_max = miex2_max & yex_max             # <<<<<<<<<<<<<<
 *         ziex_max = miex2_max & zex_max
 *         # This could *probably* be sped up by iterating over words.
 */
  __pyx_v_yiex_max = (__pyx_v_miex2_max & __pyx_v_yex_max);

  /* "yt/geometry/particle_oct_container.pyx":911
 *         xiex_max = miex2_max & xex_max
 *         yiex_max = miex2_max & yex_max
 *         ziex_max = miex2_max & zex_max             # <<<<<<<<<<<<<<
 *         # This could *probably* be sped up by iterating over words.
 *         for miex2 in range(miex2_min, miex2_max + 1):
 */
  __pyx_v_ziex_max = (__pyx_v_miex2_max & __pyx_v_zex_max);

  /* "yt/geometry/particle_oct_container.pyx":913
 *         ziex_max = miex2_max & zex_max
 *         # This could *probably* be sped up by iterating over words.
 *         for miex2 in range(miex2_min, miex2_max + 1):             # <<<<<<<<<<<<<<
 *             #miex2 = encode_morton_64bit(xex2, yex2, zex2)
 *             #decode_morton_64bit(miex2, ex2)
 */
  __pyx_t_4 = (__pyx_v_miex2_max + 1);
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = __pyx_v_miex2_min; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_miex2 = __pyx_t_6;

    /* "yt/geometry/particle_oct_container.pyx":917
 *             #decode_morton_64bit(miex2, ex2)
 *             # Let's check all our cases here
 *             if (miex2 & xex_max) < (xiex_min): continue             # <<<<<<<<<<<<<<
 *             if (miex2 & xex_max) > (xiex_max): continue
 *             if (miex2 & yex_max) < (yiex_min): continue
 */
    __pyx_t_2 = (((__pyx_v_miex2 & __pyx_v_xex_max) < __pyx_v_xiex_min) != 0);
    if (__pyx_t_2) {
      goto __pyx_L8_continue;
    }

    /* "yt/geometry/particle_oct_container.pyx":918
 *             # Let's check all our cases here
 *             if (miex2 & xex_max) < (xiex_min): continue
 *             if (miex2 & xex_max) > (xiex_max): continue             # <<<<<<<<<<<<<<
 *             if (miex2 & yex_max) < (yiex_min): continue
 *             if (miex2 & yex_max) > (yiex_max): continue
 */
    __pyx_t_2 = (((__pyx_v_miex2 & __pyx_v_xex_max) > __pyx_v_xiex_max) != 0);
    if (__pyx_t_2) {
      goto __pyx_L8_continue;
    }

    /* "yt/geometry/particle_oct_container.pyx":919
 *             if (miex2 & xex_max) < (xiex_min): continue
 *             if (miex2 & xex_max) > (xiex_max): continue
 *             if (miex2 & yex_max) < (yiex_min): continue             # <<<<<<<<<<<<<<
 *             if (miex2 & yex_max) > (yiex_max): continue
 *             if (miex2 & zex_max) < (ziex_min): continue
 */
    __pyx_t_2 = (((__pyx_v_miex2 & __pyx_v_yex_max) < __pyx_v_yiex_min) != 0);
    if (__pyx_t_2) {
      goto __pyx_L8_continue;
    }

    /* "yt/geometry/particle_oct_container.pyx":920
 *             if (miex2 & xex_max) > (xiex_max): continue
 *             if (miex2 & yex_max) < (yiex_min): continue
 *             if (miex2 & yex_max) > (yiex_max): continue             # <<<<<<<<<<<<<<
 *             if (miex2 & zex_max) < (ziex_min): continue
 *             if (miex2 & zex_max) > (ziex_max): continue
 */
    __pyx_t_2 = (((__pyx_v_miex2 & __pyx_v_yex_max) > __pyx_v_yiex_max) != 0);
    if (__pyx_t_2) {
      goto __pyx_L8_continue;
    }

    /* "yt/geometry/particle_oct_container.pyx":921
 *             if (miex2 & yex_max) < (yiex_min): continue
 *             if (miex2 & yex_max) > (yiex_max): continue
 *             if (miex2 & zex_max) < (ziex_min): continue             # <<<<<<<<<<<<<<
 *             if (miex2 & zex_max) > (ziex_max): continue
 *             refined_set.set(miex2)
 */
    __pyx_t_2 = (((__pyx_v_miex2 & __pyx_v_zex_max) < __pyx_v_ziex_min) != 0);
    if (__pyx_t_2) {
      goto __pyx_L8_continue;
    }

    /* "yt/geometry/particle_oct_container.pyx":922
 *             if (miex2 & yex_max) > (yiex_max): continue
 *             if (miex2 & zex_max) < (ziex_min): continue
 *             if (miex2 & zex_max) > (ziex_max): continue             # <<<<<<<<<<<<<<
 *             refined_set.set(miex2)
 *         return refined_set.numberOfOnes() - old_nsub
 */
    __pyx_t_2 = (((__pyx_v_miex2 & __pyx_v_zex_max) > __pyx_v_ziex_max) != 0);
    if (__pyx_t_2) {
      goto __pyx_L8_continue;
    }

    /* "yt/geometry/particle_oct_container.pyx":923
 *             if (miex2 & zex_max) < (ziex_min): continue
 *             if (miex2 & zex_max) > (ziex_max): continue
 *             refined_set.set(miex2)             # <<<<<<<<<<<<<<
 *         return refined_set.numberOfOnes() - old_nsub
 * 
 */
    __pyx_v_refined_set.set(__pyx_v_miex2);
    __pyx_L8_continue:;
  }

  /* "yt/geometry/particle_oct_container.pyx":924
 *             if (miex2 & zex_max) > (ziex_max): continue
 *             refined_set.set(miex2)
 *         return refined_set.numberOfOnes() - old_nsub             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = (__pyx_v_refined_set.numberOfOnes() - __pyx_v_old_nsub);
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":872
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef np.int64_t __fill_refined_ranges(self, np.float64_t s_ppos[3], np.float64_t radius,             # <<<<<<<<<<<<<<
 *                                            np.float64_t LE[3], np.float64_t RE[3],
 *                                            np.float64_t dds1[3], np.uint64_t xex, np.uint64_t yex, np.uint64_t zex,
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":930
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     def _set_refined_index_data_file(self,             # <<<<<<<<<<<<<<
 *                                      np.ndarray[np.uint64_t, ndim=1] sub_mi1,
 *                                      np.ndarray[np.uint64_t, ndim=1] sub_mi2,
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_set_refined_index_data_file(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_set_refined_index_data_file(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_sub_mi1 = 0;
  PyArrayObject *__pyx_v_sub_mi2 = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_file_id;
  __pyx_t_5numpy_int64_t __pyx_v_nsub_mi;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set_refined_index_data_file (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_sub_mi1,&__pyx_n_s_sub_mi2,&__pyx_n_s_file_id,&__pyx_n_s_nsub_mi,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sub_mi1)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sub_mi2)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_set_refined_index_data_file", 1, 4, 4, 1); __PYX_ERR(0, 930, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_file_id)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_set_refined_index_data_file", 1, 4, 4, 2); __PYX_ERR(0, 930, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nsub_mi)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_set_refined_index_data_file", 1, 4, 4, 3); __PYX_ERR(0, 930, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_set_refined_index_data_file") < 0)) __PYX_ERR(0, 930, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_sub_mi1 = ((PyArrayObject *)values[0]);
    __pyx_v_sub_mi2 = ((PyArrayObject *)values[1]);
    __pyx_v_file_id = __Pyx_PyInt_As_npy_uint64(values[2]); if (unlikely((__pyx_v_file_id == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 933, __pyx_L3_error)
    __pyx_v_nsub_mi = __Pyx_PyInt_As_npy_int64(values[3]); if (unlikely((__pyx_v_nsub_mi == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 933, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_set_refined_index_data_file", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 930, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap._set_refined_index_data_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sub_mi1), __pyx_ptype_5numpy_ndarray, 1, "sub_mi1", 0))) __PYX_ERR(0, 931, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_sub_mi2), __pyx_ptype_5numpy_ndarray, 1, "sub_mi2", 0))) __PYX_ERR(0, 932, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_set_refined_index_data_file(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), __pyx_v_sub_mi1, __pyx_v_sub_mi2, __pyx_v_file_id, __pyx_v_nsub_mi);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_set_refined_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyArrayObject *__pyx_v_sub_mi1, PyArrayObject *__pyx_v_sub_mi2, __pyx_t_5numpy_uint64_t __pyx_v_file_id, __pyx_t_5numpy_int64_t __pyx_v_nsub_mi) {
  __Pyx_LocalBuf_ND __pyx_pybuffernd_sub_mi1;
  __Pyx_Buffer __pyx_pybuffer_sub_mi1;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_sub_mi2;
  __Pyx_Buffer __pyx_pybuffer_sub_mi2;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_set_refined_index_data_file", 0);
  __pyx_pybuffer_sub_mi1.pybuffer.buf = NULL;
  __pyx_pybuffer_sub_mi1.refcount = 0;
  __pyx_pybuffernd_sub_mi1.data = NULL;
  __pyx_pybuffernd_sub_mi1.rcbuffer = &__pyx_pybuffer_sub_mi1;
  __pyx_pybuffer_sub_mi2.pybuffer.buf = NULL;
  __pyx_pybuffer_sub_mi2.refcount = 0;
  __pyx_pybuffernd_sub_mi2.data = NULL;
  __pyx_pybuffernd_sub_mi2.rcbuffer = &__pyx_pybuffer_sub_mi2;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer, (PyObject*)__pyx_v_sub_mi1, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 930, __pyx_L1_error)
  }
  __pyx_pybuffernd_sub_mi1.diminfo[0].strides = __pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_sub_mi1.diminfo[0].shape = __pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer, (PyObject*)__pyx_v_sub_mi2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 930, __pyx_L1_error)
  }
  __pyx_pybuffernd_sub_mi2.diminfo[0].strides = __pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_sub_mi2.diminfo[0].shape = __pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/particle_oct_container.pyx":934
 *                                      np.ndarray[np.uint64_t, ndim=1] sub_mi2,
 *                                      np.uint64_t file_id, np.int64_t nsub_mi):
 *         return self.__set_refined_index_data_file(sub_mi1, sub_mi2,             # <<<<<<<<<<<<<<
 *                                                   file_id, nsub_mi)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "yt/geometry/particle_oct_container.pyx":935
 *                                      np.uint64_t file_id, np.int64_t nsub_mi):
 *         return self.__set_refined_index_data_file(sub_mi1, sub_mi2,
 *                                                   file_id, nsub_mi)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_t_1 = __Pyx_void_to_None(((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self->__pyx_vtab)->__pyx___set_refined_index_data_file(__pyx_v_self, ((PyArrayObject *)__pyx_v_sub_mi1), ((PyArrayObject *)__pyx_v_sub_mi2), __pyx_v_file_id, __pyx_v_nsub_mi)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":930
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     def _set_refined_index_data_file(self,             # <<<<<<<<<<<<<<
 *                                      np.ndarray[np.uint64_t, ndim=1] sub_mi1,
 *                                      np.ndarray[np.uint64_t, ndim=1] sub_mi2,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap._set_refined_index_data_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":941
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef void __set_refined_index_data_file(self,             # <<<<<<<<<<<<<<
 *                                             np.ndarray[np.uint64_t, ndim=1] sub_mi1,
 *                                             np.ndarray[np.uint64_t, ndim=1] sub_mi2,
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___set_refined_index_data_file(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyArrayObject *__pyx_v_sub_mi1, PyArrayObject *__pyx_v_sub_mi2, __pyx_t_5numpy_uint64_t __pyx_v_file_id, __pyx_t_5numpy_int64_t __pyx_v_nsub_mi) {
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *__pyx_v_bitmasks = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_sub_mi1;
  __Pyx_Buffer __pyx_pybuffer_sub_mi1;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_sub_mi2;
  __Pyx_Buffer __pyx_pybuffer_sub_mi2;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set_refined_index_data_file", 0);
  __pyx_pybuffer_sub_mi1.pybuffer.buf = NULL;
  __pyx_pybuffer_sub_mi1.refcount = 0;
  __pyx_pybuffernd_sub_mi1.data = NULL;
  __pyx_pybuffernd_sub_mi1.rcbuffer = &__pyx_pybuffer_sub_mi1;
  __pyx_pybuffer_sub_mi2.pybuffer.buf = NULL;
  __pyx_pybuffer_sub_mi2.refcount = 0;
  __pyx_pybuffernd_sub_mi2.data = NULL;
  __pyx_pybuffernd_sub_mi2.rcbuffer = &__pyx_pybuffer_sub_mi2;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer, (PyObject*)__pyx_v_sub_mi1, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 941, __pyx_L1_error)
  }
  __pyx_pybuffernd_sub_mi1.diminfo[0].strides = __pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_sub_mi1.diminfo[0].shape = __pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer, (PyObject*)__pyx_v_sub_mi2, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 941, __pyx_L1_error)
  }
  __pyx_pybuffernd_sub_mi2.diminfo[0].strides = __pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_sub_mi2.diminfo[0].shape = __pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/particle_oct_container.pyx":946
 *                                             np.uint64_t file_id, np.int64_t nsub_mi):
 *         cdef np.int64_t i, p
 *         cdef FileBitmasks bitmasks = self.bitmasks             # <<<<<<<<<<<<<<
 *         bitmasks._set_refined_index_array(file_id, nsub_mi, sub_mi1, sub_mi2)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->bitmasks);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_bitmasks = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":947
 *         cdef np.int64_t i, p
 *         cdef FileBitmasks bitmasks = self.bitmasks
 *         bitmasks._set_refined_index_array(file_id, nsub_mi, sub_mi1, sub_mi2)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_bitmasks->__pyx_vtab)->_set_refined_index_array(__pyx_v_bitmasks, __pyx_v_file_id, __pyx_v_nsub_mi, ((PyArrayObject *)__pyx_v_sub_mi1), ((PyArrayObject *)__pyx_v_sub_mi2));

  /* "yt/geometry/particle_oct_container.pyx":941
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef void __set_refined_index_data_file(self,             # <<<<<<<<<<<<<<
 *                                             np.ndarray[np.uint64_t, ndim=1] sub_mi1,
 *                                             np.ndarray[np.uint64_t, ndim=1] sub_mi2,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_WriteUnraisable("yt.geometry.particle_oct_container.ParticleBitmap.__set_refined_index_data_file", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sub_mi1.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sub_mi2.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_bitmasks);
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":953
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     def find_collisions(self, verbose=False):             # <<<<<<<<<<<<<<
 *         cdef tuple cc, rc
 *         cc, rc = self.bitmasks._find_collisions(self.collisions,verbose)
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17find_collisions(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17find_collisions(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_verbose = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_collisions (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_verbose,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_verbose);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "find_collisions") < 0)) __PYX_ERR(0, 953, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_verbose = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("find_collisions", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 953, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.find_collisions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_16find_collisions(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), __pyx_v_verbose);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_16find_collisions(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_verbose) {
  PyObject *__pyx_v_cc = 0;
  PyObject *__pyx_v_rc = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__find_collisions __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_collisions", 0);

  /* "yt/geometry/particle_oct_container.pyx":955
 *     def find_collisions(self, verbose=False):
 *         cdef tuple cc, rc
 *         cc, rc = self.bitmasks._find_collisions(self.collisions,verbose)             # <<<<<<<<<<<<<<
 *         return cc, rc
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->collisions);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_verbose); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 955, __pyx_L1_error)
  __pyx_t_4.__pyx_n = 1;
  __pyx_t_4.verbose = __pyx_t_2;
  __pyx_t_3 = ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmasks->__pyx_vtab)->_find_collisions(__pyx_v_self->bitmasks, ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_1), &__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(__pyx_t_3 != Py_None)) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 955, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_5);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 955, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 955, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 955, __pyx_L1_error)
  }
  if (!(likely(PyTuple_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 955, __pyx_L1_error)
  if (!(likely(PyTuple_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(0, 955, __pyx_L1_error)
  __pyx_v_cc = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_rc = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "yt/geometry/particle_oct_container.pyx":956
 *         cdef tuple cc, rc
 *         cc, rc = self.bitmasks._find_collisions(self.collisions,verbose)
 *         return cc, rc             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 956, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_cc);
  __Pyx_GIVEREF(__pyx_v_cc);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_cc);
  __Pyx_INCREF(__pyx_v_rc);
  __Pyx_GIVEREF(__pyx_v_rc);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_rc);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":953
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     def find_collisions(self, verbose=False):             # <<<<<<<<<<<<<<
 *         cdef tuple cc, rc
 *         cc, rc = self.bitmasks._find_collisions(self.collisions,verbose)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.find_collisions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cc);
  __Pyx_XDECREF(__pyx_v_rc);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":962
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     def find_collisions_coarse(self, verbose=False, file_list = None):             # <<<<<<<<<<<<<<
 *         cdef int nc, nm
 *         nc, nm = self.bitmasks._find_collisions_coarse(self.collisions, verbose, file_list)
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19find_collisions_coarse(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19find_collisions_coarse(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_verbose = 0;
  PyObject *__pyx_v_file_list = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_collisions_coarse (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_verbose,&__pyx_n_s_file_list,0};
    PyObject* values[2] = {0,0};
    values[0] = ((PyObject *)Py_False);
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_verbose);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_file_list);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "find_collisions_coarse") < 0)) __PYX_ERR(0, 962, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_verbose = values[0];
    __pyx_v_file_list = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("find_collisions_coarse", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 962, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.find_collisions_coarse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_18find_collisions_coarse(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), __pyx_v_verbose, __pyx_v_file_list);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_18find_collisions_coarse(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_verbose, PyObject *__pyx_v_file_list) {
  int __pyx_v_nc;
  int __pyx_v_nm;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__find_collisions_coarse __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_collisions_coarse", 0);

  /* "yt/geometry/particle_oct_container.pyx":964
 *     def find_collisions_coarse(self, verbose=False, file_list = None):
 *         cdef int nc, nm
 *         nc, nm = self.bitmasks._find_collisions_coarse(self.collisions, verbose, file_list)             # <<<<<<<<<<<<<<
 *         return nc, nm
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->collisions);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_verbose); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 964, __pyx_L1_error)
  __pyx_t_4.__pyx_n = 2;
  __pyx_t_4.verbose = __pyx_t_2;
  __pyx_t_4.file_list = __pyx_v_file_list;
  __pyx_t_3 = ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmasks->__pyx_vtab)->_find_collisions_coarse(__pyx_v_self->bitmasks, ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_1), &__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(__pyx_t_3 != Py_None)) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 964, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_5);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 964, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 964, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 964, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_nc = __pyx_t_6;
  __pyx_v_nm = __pyx_t_7;

  /* "yt/geometry/particle_oct_container.pyx":965
 *         cdef int nc, nm
 *         nc, nm = self.bitmasks._find_collisions_coarse(self.collisions, verbose, file_list)
 *         return nc, nm             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_nc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_nm); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 965, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":962
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     def find_collisions_coarse(self, verbose=False, file_list = None):             # <<<<<<<<<<<<<<
 *         cdef int nc, nm
 *         nc, nm = self.bitmasks._find_collisions_coarse(self.collisions, verbose, file_list)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.find_collisions_coarse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":971
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     def find_uncontaminated(self, np.uint32_t ifile, BoolArrayCollection mask,             # <<<<<<<<<<<<<<
 *                             BoolArrayCollection mask2 = None):
 *         cdef np.ndarray[np.uint8_t, ndim=1] arr = np.zeros((1 << (self.index_order1 * 3)),'uint8')
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_21find_uncontaminated(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_21find_uncontaminated(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __pyx_t_5numpy_uint32_t __pyx_v_ifile;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mask = 0;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mask2 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_uncontaminated (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ifile,&__pyx_n_s_mask,&__pyx_n_s_mask2,0};
    PyObject* values[3] = {0,0,0};

    /* "yt/geometry/particle_oct_container.pyx":972
 *     @cython.initializedcheck(False)
 *     def find_uncontaminated(self, np.uint32_t ifile, BoolArrayCollection mask,
 *                             BoolArrayCollection mask2 = None):             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint8_t, ndim=1] arr = np.zeros((1 << (self.index_order1 * 3)),'uint8')
 *         cdef np.uint8_t[:] arr_view = arr
 */
    values[2] = (PyObject *)((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ifile)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("find_uncontaminated", 0, 2, 3, 1); __PYX_ERR(0, 971, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask2);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "find_uncontaminated") < 0)) __PYX_ERR(0, 971, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_ifile = __Pyx_PyInt_As_npy_uint32(values[0]); if (unlikely((__pyx_v_ifile == ((npy_uint32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 971, __pyx_L3_error)
    __pyx_v_mask = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)values[1]);
    __pyx_v_mask2 = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("find_uncontaminated", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 971, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.find_uncontaminated", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mask), __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection, 1, "mask", 0))) __PYX_ERR(0, 971, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mask2), __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection, 1, "mask2", 0))) __PYX_ERR(0, 972, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_20find_uncontaminated(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), __pyx_v_ifile, __pyx_v_mask, __pyx_v_mask2);

  /* "yt/geometry/particle_oct_container.pyx":971
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     def find_uncontaminated(self, np.uint32_t ifile, BoolArrayCollection mask,             # <<<<<<<<<<<<<<
 *                             BoolArrayCollection mask2 = None):
 *         cdef np.ndarray[np.uint8_t, ndim=1] arr = np.zeros((1 << (self.index_order1 * 3)),'uint8')
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_20find_uncontaminated(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, __pyx_t_5numpy_uint32_t __pyx_v_ifile, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mask, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mask2) {
  PyArrayObject *__pyx_v_arr = 0;
  __Pyx_memviewslice __pyx_v_arr_view = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_LocalBuf_ND __pyx_pybuffernd_arr;
  __Pyx_Buffer __pyx_pybuffer_arr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__select_uncontaminated __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_uncontaminated", 0);
  __pyx_pybuffer_arr.pybuffer.buf = NULL;
  __pyx_pybuffer_arr.refcount = 0;
  __pyx_pybuffernd_arr.data = NULL;
  __pyx_pybuffernd_arr.rcbuffer = &__pyx_pybuffer_arr;

  /* "yt/geometry/particle_oct_container.pyx":973
 *     def find_uncontaminated(self, np.uint32_t ifile, BoolArrayCollection mask,
 *                             BoolArrayCollection mask2 = None):
 *         cdef np.ndarray[np.uint8_t, ndim=1] arr = np.zeros((1 << (self.index_order1 * 3)),'uint8')             # <<<<<<<<<<<<<<
 *         cdef np.uint8_t[:] arr_view = arr
 *         self.bitmasks._select_uncontaminated(ifile, mask, arr_view, mask2)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_long((1 << (__pyx_v_self->index_order1 * 3))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_n_s_uint8};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 973, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_n_s_uint8};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 973, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 973, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_uint8);
    __Pyx_GIVEREF(__pyx_n_s_uint8);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_n_s_uint8);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 973, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 973, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_arr.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_arr = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_arr.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 973, __pyx_L1_error)
    } else {__pyx_pybuffernd_arr.diminfo[0].strides = __pyx_pybuffernd_arr.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_arr.diminfo[0].shape = __pyx_pybuffernd_arr.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_7 = 0;
  __pyx_v_arr = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":974
 *                             BoolArrayCollection mask2 = None):
 *         cdef np.ndarray[np.uint8_t, ndim=1] arr = np.zeros((1 << (self.index_order1 * 3)),'uint8')
 *         cdef np.uint8_t[:] arr_view = arr             # <<<<<<<<<<<<<<
 *         self.bitmasks._select_uncontaminated(ifile, mask, arr_view, mask2)
 *         return arr
 */
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint8_t(((PyObject *)__pyx_v_arr), PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 974, __pyx_L1_error)
  __pyx_v_arr_view = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":975
 *         cdef np.ndarray[np.uint8_t, ndim=1] arr = np.zeros((1 << (self.index_order1 * 3)),'uint8')
 *         cdef np.uint8_t[:] arr_view = arr
 *         self.bitmasks._select_uncontaminated(ifile, mask, arr_view, mask2)             # <<<<<<<<<<<<<<
 *         return arr
 * 
 */
  __pyx_t_9.__pyx_n = 1;
  __pyx_t_9.mask2 = __pyx_v_mask2;
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmasks->__pyx_vtab)->_select_uncontaminated(__pyx_v_self->bitmasks, __pyx_v_ifile, __pyx_v_mask, __pyx_v_arr_view, &__pyx_t_9); 

  /* "yt/geometry/particle_oct_container.pyx":976
 *         cdef np.uint8_t[:] arr_view = arr
 *         self.bitmasks._select_uncontaminated(ifile, mask, arr_view, mask2)
 *         return arr             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_arr));
  __pyx_r = ((PyObject *)__pyx_v_arr);
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":971
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     def find_uncontaminated(self, np.uint32_t ifile, BoolArrayCollection mask,             # <<<<<<<<<<<<<<
 *                             BoolArrayCollection mask2 = None):
 *         cdef np.ndarray[np.uint8_t, ndim=1] arr = np.zeros((1 << (self.index_order1 * 3)),'uint8')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_arr.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.find_uncontaminated", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_arr.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_arr);
  __PYX_XDEC_MEMVIEW(&__pyx_v_arr_view, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":982
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     def find_contaminated(self, np.uint32_t ifile, BoolArrayCollection mask,             # <<<<<<<<<<<<<<
 *                           BoolArrayCollection mask2 = None):
 *         cdef np.ndarray[np.uint8_t, ndim=1] arr = np.zeros((1 << (self.index_order1 * 3)),'uint8')
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_23find_contaminated(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_23find_contaminated(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __pyx_t_5numpy_uint32_t __pyx_v_ifile;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mask = 0;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mask2 = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_contaminated (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ifile,&__pyx_n_s_mask,&__pyx_n_s_mask2,0};
    PyObject* values[3] = {0,0,0};

    /* "yt/geometry/particle_oct_container.pyx":983
 *     @cython.initializedcheck(False)
 *     def find_contaminated(self, np.uint32_t ifile, BoolArrayCollection mask,
 *                           BoolArrayCollection mask2 = None):             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint8_t, ndim=1] arr = np.zeros((1 << (self.index_order1 * 3)),'uint8')
 *         cdef np.uint8_t[:] arr_view = arr
 */
    values[2] = (PyObject *)((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ifile)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("find_contaminated", 0, 2, 3, 1); __PYX_ERR(0, 982, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask2);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "find_contaminated") < 0)) __PYX_ERR(0, 982, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_ifile = __Pyx_PyInt_As_npy_uint32(values[0]); if (unlikely((__pyx_v_ifile == ((npy_uint32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 982, __pyx_L3_error)
    __pyx_v_mask = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)values[1]);
    __pyx_v_mask2 = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("find_contaminated", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 982, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.find_contaminated", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mask), __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection, 1, "mask", 0))) __PYX_ERR(0, 982, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mask2), __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection, 1, "mask2", 0))) __PYX_ERR(0, 983, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_22find_contaminated(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), __pyx_v_ifile, __pyx_v_mask, __pyx_v_mask2);

  /* "yt/geometry/particle_oct_container.pyx":982
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     def find_contaminated(self, np.uint32_t ifile, BoolArrayCollection mask,             # <<<<<<<<<<<<<<
 *                           BoolArrayCollection mask2 = None):
 *         cdef np.ndarray[np.uint8_t, ndim=1] arr = np.zeros((1 << (self.index_order1 * 3)),'uint8')
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_22find_contaminated(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, __pyx_t_5numpy_uint32_t __pyx_v_ifile, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mask, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mask2) {
  PyArrayObject *__pyx_v_arr = 0;
  __Pyx_memviewslice __pyx_v_arr_view = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyArrayObject *__pyx_v_sfiles = 0;
  __Pyx_memviewslice __pyx_v_sfiles_view = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_LocalBuf_ND __pyx_pybuffernd_arr;
  __Pyx_Buffer __pyx_pybuffer_arr;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_sfiles;
  __Pyx_Buffer __pyx_pybuffer_sfiles;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyArrayObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyArrayObject *__pyx_t_9 = NULL;
  struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__select_contaminated __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_contaminated", 0);
  __pyx_pybuffer_arr.pybuffer.buf = NULL;
  __pyx_pybuffer_arr.refcount = 0;
  __pyx_pybuffernd_arr.data = NULL;
  __pyx_pybuffernd_arr.rcbuffer = &__pyx_pybuffer_arr;
  __pyx_pybuffer_sfiles.pybuffer.buf = NULL;
  __pyx_pybuffer_sfiles.refcount = 0;
  __pyx_pybuffernd_sfiles.data = NULL;
  __pyx_pybuffernd_sfiles.rcbuffer = &__pyx_pybuffer_sfiles;

  /* "yt/geometry/particle_oct_container.pyx":984
 *     def find_contaminated(self, np.uint32_t ifile, BoolArrayCollection mask,
 *                           BoolArrayCollection mask2 = None):
 *         cdef np.ndarray[np.uint8_t, ndim=1] arr = np.zeros((1 << (self.index_order1 * 3)),'uint8')             # <<<<<<<<<<<<<<
 *         cdef np.uint8_t[:] arr_view = arr
 *         cdef np.ndarray[np.uint8_t, ndim=1] sfiles = np.zeros(self.nfiles,'uint8')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_long((1 << (__pyx_v_self->index_order1 * 3))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_n_s_uint8};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 984, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_n_s_uint8};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 984, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 984, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_uint8);
    __Pyx_GIVEREF(__pyx_n_s_uint8);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_n_s_uint8);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 984, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 984, __pyx_L1_error)
  __pyx_t_7 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_arr.rcbuffer->pybuffer, (PyObject*)__pyx_t_7, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_arr = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_arr.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 984, __pyx_L1_error)
    } else {__pyx_pybuffernd_arr.diminfo[0].strides = __pyx_pybuffernd_arr.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_arr.diminfo[0].shape = __pyx_pybuffernd_arr.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_7 = 0;
  __pyx_v_arr = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":985
 *                           BoolArrayCollection mask2 = None):
 *         cdef np.ndarray[np.uint8_t, ndim=1] arr = np.zeros((1 << (self.index_order1 * 3)),'uint8')
 *         cdef np.uint8_t[:] arr_view = arr             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint8_t, ndim=1] sfiles = np.zeros(self.nfiles,'uint8')
 *         cdef np.uint8_t[:] sfiles_view = sfiles
 */
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint8_t(((PyObject *)__pyx_v_arr), PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 985, __pyx_L1_error)
  __pyx_v_arr_view = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":986
 *         cdef np.ndarray[np.uint8_t, ndim=1] arr = np.zeros((1 << (self.index_order1 * 3)),'uint8')
 *         cdef np.uint8_t[:] arr_view = arr
 *         cdef np.ndarray[np.uint8_t, ndim=1] sfiles = np.zeros(self.nfiles,'uint8')             # <<<<<<<<<<<<<<
 *         cdef np.uint8_t[:] sfiles_view = sfiles
 *         self.bitmasks._select_contaminated(ifile, mask, arr_view, sfiles_view, mask2)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->nfiles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 986, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_t_3, __pyx_n_s_uint8};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 986, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_t_3, __pyx_n_s_uint8};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 986, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_4 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 986, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_t_3);
    __Pyx_INCREF(__pyx_n_s_uint8);
    __Pyx_GIVEREF(__pyx_n_s_uint8);
    PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_n_s_uint8);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 986, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 986, __pyx_L1_error)
  __pyx_t_9 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_sfiles.rcbuffer->pybuffer, (PyObject*)__pyx_t_9, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_sfiles = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_sfiles.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 986, __pyx_L1_error)
    } else {__pyx_pybuffernd_sfiles.diminfo[0].strides = __pyx_pybuffernd_sfiles.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_sfiles.diminfo[0].shape = __pyx_pybuffernd_sfiles.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_9 = 0;
  __pyx_v_sfiles = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":987
 *         cdef np.uint8_t[:] arr_view = arr
 *         cdef np.ndarray[np.uint8_t, ndim=1] sfiles = np.zeros(self.nfiles,'uint8')
 *         cdef np.uint8_t[:] sfiles_view = sfiles             # <<<<<<<<<<<<<<
 *         self.bitmasks._select_contaminated(ifile, mask, arr_view, sfiles_view, mask2)
 *         return arr, np.where(sfiles)[0].astype('uint32')
 */
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint8_t(((PyObject *)__pyx_v_sfiles), PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 987, __pyx_L1_error)
  __pyx_v_sfiles_view = __pyx_t_8;
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":988
 *         cdef np.ndarray[np.uint8_t, ndim=1] sfiles = np.zeros(self.nfiles,'uint8')
 *         cdef np.uint8_t[:] sfiles_view = sfiles
 *         self.bitmasks._select_contaminated(ifile, mask, arr_view, sfiles_view, mask2)             # <<<<<<<<<<<<<<
 *         return arr, np.where(sfiles)[0].astype('uint32')
 * 
 */
  __pyx_t_10.__pyx_n = 1;
  __pyx_t_10.mask2 = __pyx_v_mask2;
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmasks->__pyx_vtab)->_select_contaminated(__pyx_v_self->bitmasks, __pyx_v_ifile, __pyx_v_mask, __pyx_v_arr_view, __pyx_v_sfiles_view, &__pyx_t_10); 

  /* "yt/geometry/particle_oct_container.pyx":989
 *         cdef np.uint8_t[:] sfiles_view = sfiles
 *         self.bitmasks._select_contaminated(ifile, mask, arr_view, sfiles_view, mask2)
 *         return arr, np.where(sfiles)[0].astype('uint32')             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_where); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_sfiles)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_sfiles));
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_6, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_astype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_3, __pyx_n_s_uint32) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_n_s_uint32);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(((PyObject *)__pyx_v_arr));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_arr));
  PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_v_arr));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":982
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     def find_contaminated(self, np.uint32_t ifile, BoolArrayCollection mask,             # <<<<<<<<<<<<<<
 *                           BoolArrayCollection mask2 = None):
 *         cdef np.ndarray[np.uint8_t, ndim=1] arr = np.zeros((1 << (self.index_order1 * 3)),'uint8')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_arr.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sfiles.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.find_contaminated", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_arr.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_sfiles.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_arr);
  __PYX_XDEC_MEMVIEW(&__pyx_v_arr_view, 1);
  __Pyx_XDECREF((PyObject *)__pyx_v_sfiles);
  __PYX_XDEC_MEMVIEW(&__pyx_v_sfiles_view, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":995
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     def find_collisions_refined(self, verbose=False):             # <<<<<<<<<<<<<<
 *         cdef np.int32_t nc, nm
 *         nc, nm = self.bitmasks._find_collisions_refined(self.collisions,verbose)
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_25find_collisions_refined(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_25find_collisions_refined(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_verbose = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_collisions_refined (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_verbose,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_verbose);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "find_collisions_refined") < 0)) __PYX_ERR(0, 995, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_verbose = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("find_collisions_refined", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 995, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.find_collisions_refined", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_24find_collisions_refined(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), __pyx_v_verbose);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_24find_collisions_refined(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_verbose) {
  __pyx_t_5numpy_int32_t __pyx_v_nc;
  __pyx_t_5numpy_int32_t __pyx_v_nm;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__find_collisions_refined __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __pyx_t_5numpy_int32_t __pyx_t_6;
  __pyx_t_5numpy_int32_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_collisions_refined", 0);

  /* "yt/geometry/particle_oct_container.pyx":997
 *     def find_collisions_refined(self, verbose=False):
 *         cdef np.int32_t nc, nm
 *         nc, nm = self.bitmasks._find_collisions_refined(self.collisions,verbose)             # <<<<<<<<<<<<<<
 *         return nc, nm
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->collisions);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_verbose); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 997, __pyx_L1_error)
  __pyx_t_4.__pyx_n = 1;
  __pyx_t_4.verbose = __pyx_t_2;
  __pyx_t_3 = ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmasks->__pyx_vtab)->_find_collisions_refined(__pyx_v_self->bitmasks, ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_1), &__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(__pyx_t_3 != Py_None)) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 997, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_5);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(0, 997, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyInt_As_npy_int32(__pyx_t_1); if (unlikely((__pyx_t_6 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __Pyx_PyInt_As_npy_int32(__pyx_t_5); if (unlikely((__pyx_t_7 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_nc = __pyx_t_6;
  __pyx_v_nm = __pyx_t_7;

  /* "yt/geometry/particle_oct_container.pyx":998
 *         cdef np.int32_t nc, nm
 *         nc, nm = self.bitmasks._find_collisions_refined(self.collisions,verbose)
 *         return nc, nm             # <<<<<<<<<<<<<<
 * 
 *     def calcsize_bitmasks(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_npy_int32(__pyx_v_nc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_From_npy_int32(__pyx_v_nm); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":995
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     def find_collisions_refined(self, verbose=False):             # <<<<<<<<<<<<<<
 *         cdef np.int32_t nc, nm
 *         nc, nm = self.bitmasks._find_collisions_refined(self.collisions,verbose)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.find_collisions_refined", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1000
 *         return nc, nm
 * 
 *     def calcsize_bitmasks(self):             # <<<<<<<<<<<<<<
 *         # TODO: All cython
 *         cdef bytes serial_BAC
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_27calcsize_bitmasks(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_27calcsize_bitmasks(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calcsize_bitmasks (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_26calcsize_bitmasks(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_26calcsize_bitmasks(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_v_serial_BAC = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_ifile;
  int __pyx_v_out;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __pyx_t_5numpy_uint64_t __pyx_t_6;
  __pyx_t_5numpy_uint64_t __pyx_t_7;
  __pyx_t_5numpy_uint64_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calcsize_bitmasks", 0);

  /* "yt/geometry/particle_oct_container.pyx":1004
 *         cdef bytes serial_BAC
 *         cdef np.uint64_t ifile
 *         cdef int out = 0             # <<<<<<<<<<<<<<
 *         out += struct.calcsize('Q')
 *         # Bitmaps for each file
 */
  __pyx_v_out = 0;

  /* "yt/geometry/particle_oct_container.pyx":1005
 *         cdef np.uint64_t ifile
 *         cdef int out = 0
 *         out += struct.calcsize('Q')             # <<<<<<<<<<<<<<
 *         # Bitmaps for each file
 *         for ifile in range(self.nfiles):
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_out); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_struct); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_calcsize); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_n_s_Q) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_n_s_Q);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_out = __pyx_t_5;

  /* "yt/geometry/particle_oct_container.pyx":1007
 *         out += struct.calcsize('Q')
 *         # Bitmaps for each file
 *         for ifile in range(self.nfiles):             # <<<<<<<<<<<<<<
 *             serial_BAC = self.bitmasks._dumps(ifile)
 *             out += struct.calcsize('Q')
 */
  __pyx_t_6 = __pyx_v_self->nfiles;
  __pyx_t_7 = __pyx_t_6;
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_ifile = __pyx_t_8;

    /* "yt/geometry/particle_oct_container.pyx":1008
 *         # Bitmaps for each file
 *         for ifile in range(self.nfiles):
 *             serial_BAC = self.bitmasks._dumps(ifile)             # <<<<<<<<<<<<<<
 *             out += struct.calcsize('Q')
 *             out += len(serial_BAC)
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmasks->__pyx_vtab)->_dumps(__pyx_v_self->bitmasks, __pyx_v_ifile); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1008, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_serial_BAC, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "yt/geometry/particle_oct_container.pyx":1009
 *         for ifile in range(self.nfiles):
 *             serial_BAC = self.bitmasks._dumps(ifile)
 *             out += struct.calcsize('Q')             # <<<<<<<<<<<<<<
 *             out += len(serial_BAC)
 *         # Bitmap for collisions
 */
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_out); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1009, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_struct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1009, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_calcsize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1009, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_n_s_Q) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_s_Q);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1009, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1009, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1009, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_out = __pyx_t_5;

    /* "yt/geometry/particle_oct_container.pyx":1010
 *             serial_BAC = self.bitmasks._dumps(ifile)
 *             out += struct.calcsize('Q')
 *             out += len(serial_BAC)             # <<<<<<<<<<<<<<
 *         # Bitmap for collisions
 *         serial_BAC = self.collisions._dumps()
 */
    if (unlikely(__pyx_v_serial_BAC == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 1010, __pyx_L1_error)
    }
    __pyx_t_9 = PyBytes_GET_SIZE(__pyx_v_serial_BAC); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1010, __pyx_L1_error)
    __pyx_v_out = (__pyx_v_out + __pyx_t_9);
  }

  /* "yt/geometry/particle_oct_container.pyx":1012
 *             out += len(serial_BAC)
 *         # Bitmap for collisions
 *         serial_BAC = self.collisions._dumps()             # <<<<<<<<<<<<<<
 *         out += struct.calcsize('Q')
 *         out += len(serial_BAC)
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_v_self->collisions->__pyx_vtab)->_dumps(__pyx_v_self->collisions); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XDECREF_SET(__pyx_v_serial_BAC, ((PyObject*)__pyx_t_3));
  __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1013
 *         # Bitmap for collisions
 *         serial_BAC = self.collisions._dumps()
 *         out += struct.calcsize('Q')             # <<<<<<<<<<<<<<
 *         out += len(serial_BAC)
 *         return out
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_out); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_struct); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_calcsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_n_s_Q) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_n_s_Q);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1013, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_out = __pyx_t_5;

  /* "yt/geometry/particle_oct_container.pyx":1014
 *         serial_BAC = self.collisions._dumps()
 *         out += struct.calcsize('Q')
 *         out += len(serial_BAC)             # <<<<<<<<<<<<<<
 *         return out
 * 
 */
  if (unlikely(__pyx_v_serial_BAC == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1014, __pyx_L1_error)
  }
  __pyx_t_9 = PyBytes_GET_SIZE(__pyx_v_serial_BAC); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1014, __pyx_L1_error)
  __pyx_v_out = (__pyx_v_out + __pyx_t_9);

  /* "yt/geometry/particle_oct_container.pyx":1015
 *         out += struct.calcsize('Q')
 *         out += len(serial_BAC)
 *         return out             # <<<<<<<<<<<<<<
 * 
 *     def get_bitmasks(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_out); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1000
 *         return nc, nm
 * 
 *     def calcsize_bitmasks(self):             # <<<<<<<<<<<<<<
 *         # TODO: All cython
 *         cdef bytes serial_BAC
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.calcsize_bitmasks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_serial_BAC);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1017
 *         return out
 * 
 *     def get_bitmasks(self):             # <<<<<<<<<<<<<<
 *         return self.bitmasks
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_29get_bitmasks(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_29get_bitmasks(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_bitmasks (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_28get_bitmasks(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_28get_bitmasks(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_bitmasks", 0);

  /* "yt/geometry/particle_oct_container.pyx":1018
 * 
 *     def get_bitmasks(self):
 *         return self.bitmasks             # <<<<<<<<<<<<<<
 * 
 *     def iseq_bitmask(self, solf):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->bitmasks));
  __pyx_r = ((PyObject *)__pyx_v_self->bitmasks);
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1017
 *         return out
 * 
 *     def get_bitmasks(self):             # <<<<<<<<<<<<<<
 *         return self.bitmasks
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1020
 *         return self.bitmasks
 * 
 *     def iseq_bitmask(self, solf):             # <<<<<<<<<<<<<<
 *         return self.bitmasks._iseq(solf.get_bitmasks())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_31iseq_bitmask(PyObject *__pyx_v_self, PyObject *__pyx_v_solf); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_31iseq_bitmask(PyObject *__pyx_v_self, PyObject *__pyx_v_solf) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iseq_bitmask (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_30iseq_bitmask(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_solf));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_30iseq_bitmask(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_solf) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("iseq_bitmask", 0);

  /* "yt/geometry/particle_oct_container.pyx":1021
 * 
 *     def iseq_bitmask(self, solf):
 *         return self.bitmasks._iseq(solf.get_bitmasks())             # <<<<<<<<<<<<<<
 * 
 *     def save_bitmasks(self, fname):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_solf, __pyx_n_s_get_bitmasks); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks))))) __PYX_ERR(0, 1021, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmasks->__pyx_vtab)->_iseq(__pyx_v_self->bitmasks, ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_t_1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1020
 *         return self.bitmasks
 * 
 *     def iseq_bitmask(self, solf):             # <<<<<<<<<<<<<<
 *         return self.bitmasks._iseq(solf.get_bitmasks())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.iseq_bitmask", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1023
 *         return self.bitmasks._iseq(solf.get_bitmasks())
 * 
 *     def save_bitmasks(self, fname):             # <<<<<<<<<<<<<<
 *         cdef bytes serial_BAC
 *         cdef np.uint64_t ifile
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_33save_bitmasks(PyObject *__pyx_v_self, PyObject *__pyx_v_fname); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_33save_bitmasks(PyObject *__pyx_v_self, PyObject *__pyx_v_fname) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("save_bitmasks (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_32save_bitmasks(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_fname));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_32save_bitmasks(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_fname) {
  PyObject *__pyx_v_serial_BAC = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_ifile;
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  __pyx_t_5numpy_uint64_t __pyx_t_9;
  __pyx_t_5numpy_uint64_t __pyx_t_10;
  __pyx_t_5numpy_uint64_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("save_bitmasks", 0);

  /* "yt/geometry/particle_oct_container.pyx":1026
 *         cdef bytes serial_BAC
 *         cdef np.uint64_t ifile
 *         f = open(fname,'wb')             # <<<<<<<<<<<<<<
 *         # Header
 *         f.write(struct.pack('Q', _bitmask_version))
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1026, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_fname);
  __Pyx_GIVEREF(__pyx_v_fname);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_fname);
  __Pyx_INCREF(__pyx_n_s_wb);
  __Pyx_GIVEREF(__pyx_n_s_wb);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_wb);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1026, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_f = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1028
 *         f = open(fname,'wb')
 *         # Header
 *         f.write(struct.pack('Q', _bitmask_version))             # <<<<<<<<<<<<<<
 *         f.write(struct.pack('q', self.file_hash))
 *         f.write(struct.pack('Q', self.nfiles))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1028, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_struct); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1028, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_pack); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1028, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_bitmask_version); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1028, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_n_s_Q, __pyx_t_4};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1028, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_n_s_Q, __pyx_t_4};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1028, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1028, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_INCREF(__pyx_n_s_Q);
    __Pyx_GIVEREF(__pyx_n_s_Q);
    PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_n_s_Q);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1028, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1028, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1029
 *         # Header
 *         f.write(struct.pack('Q', _bitmask_version))
 *         f.write(struct.pack('q', self.file_hash))             # <<<<<<<<<<<<<<
 *         f.write(struct.pack('Q', self.nfiles))
 *         # Bitmap for each file
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1029, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_struct); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1029, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_pack); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1029, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->file_hash); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1029, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_8)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_n_s_q, __pyx_t_5};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_n_s_q, __pyx_t_5};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_n_s_q);
    __Pyx_GIVEREF(__pyx_n_s_q);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_7, __pyx_n_s_q);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_7, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1029, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1029, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1030
 *         f.write(struct.pack('Q', _bitmask_version))
 *         f.write(struct.pack('q', self.file_hash))
 *         f.write(struct.pack('Q', self.nfiles))             # <<<<<<<<<<<<<<
 *         # Bitmap for each file
 *         for ifile in range(self.nfiles):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_struct); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_pack); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->nfiles); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_5 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_n_s_Q, __pyx_t_8};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1030, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_n_s_Q, __pyx_t_8};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1030, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  } else
  #endif
  {
    __pyx_t_4 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1030, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_n_s_Q);
    __Pyx_GIVEREF(__pyx_n_s_Q);
    PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_7, __pyx_n_s_Q);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_7, __pyx_t_8);
    __pyx_t_8 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1030, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1032
 *         f.write(struct.pack('Q', self.nfiles))
 *         # Bitmap for each file
 *         for ifile in range(self.nfiles):             # <<<<<<<<<<<<<<
 *             serial_BAC = self.bitmasks._dumps(ifile)
 *             f.write(struct.pack('Q', len(serial_BAC)))
 */
  __pyx_t_9 = __pyx_v_self->nfiles;
  __pyx_t_10 = __pyx_t_9;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_ifile = __pyx_t_11;

    /* "yt/geometry/particle_oct_container.pyx":1033
 *         # Bitmap for each file
 *         for ifile in range(self.nfiles):
 *             serial_BAC = self.bitmasks._dumps(ifile)             # <<<<<<<<<<<<<<
 *             f.write(struct.pack('Q', len(serial_BAC)))
 *             f.write(serial_BAC)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmasks->__pyx_vtab)->_dumps(__pyx_v_self->bitmasks, __pyx_v_ifile); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1033, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_serial_BAC, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "yt/geometry/particle_oct_container.pyx":1034
 *         for ifile in range(self.nfiles):
 *             serial_BAC = self.bitmasks._dumps(ifile)
 *             f.write(struct.pack('Q', len(serial_BAC)))             # <<<<<<<<<<<<<<
 *             f.write(serial_BAC)
 *         # Collisions
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_struct); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_pack); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v_serial_BAC == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 1034, __pyx_L1_error)
    }
    __pyx_t_12 = PyBytes_GET_SIZE(__pyx_v_serial_BAC); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1034, __pyx_L1_error)
    __pyx_t_6 = PyInt_FromSsize_t(__pyx_t_12); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_n_s_Q, __pyx_t_6};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1034, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_n_s_Q, __pyx_t_6};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1034, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1034, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(__pyx_n_s_Q);
      __Pyx_GIVEREF(__pyx_n_s_Q);
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_7, __pyx_n_s_Q);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_7, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1034, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "yt/geometry/particle_oct_container.pyx":1035
 *             serial_BAC = self.bitmasks._dumps(ifile)
 *             f.write(struct.pack('Q', len(serial_BAC)))
 *             f.write(serial_BAC)             # <<<<<<<<<<<<<<
 *         # Collisions
 *         serial_BAC = self.collisions._dumps()
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1035, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_serial_BAC) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_serial_BAC);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1035, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "yt/geometry/particle_oct_container.pyx":1037
 *             f.write(serial_BAC)
 *         # Collisions
 *         serial_BAC = self.collisions._dumps()             # <<<<<<<<<<<<<<
 *         f.write(struct.pack('Q', len(serial_BAC)))
 *         f.write(serial_BAC)
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_v_self->collisions->__pyx_vtab)->_dumps(__pyx_v_self->collisions); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XDECREF_SET(__pyx_v_serial_BAC, ((PyObject*)__pyx_t_2));
  __pyx_t_2 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1038
 *         # Collisions
 *         serial_BAC = self.collisions._dumps()
 *         f.write(struct.pack('Q', len(serial_BAC)))             # <<<<<<<<<<<<<<
 *         f.write(serial_BAC)
 *         f.close()
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_struct); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_pack); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(__pyx_v_serial_BAC == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1038, __pyx_L1_error)
  }
  __pyx_t_12 = PyBytes_GET_SIZE(__pyx_v_serial_BAC); if (unlikely(__pyx_t_12 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1038, __pyx_L1_error)
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_n_s_Q, __pyx_t_4};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1038, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_n_s_Q, __pyx_t_4};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1038, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1038, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_INCREF(__pyx_n_s_Q);
    __Pyx_GIVEREF(__pyx_n_s_Q);
    PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_n_s_Q);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1038, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1038, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1039
 *         serial_BAC = self.collisions._dumps()
 *         f.write(struct.pack('Q', len(serial_BAC)))
 *         f.write(serial_BAC)             # <<<<<<<<<<<<<<
 *         f.close()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_serial_BAC) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_serial_BAC);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1039, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1040
 *         f.write(struct.pack('Q', len(serial_BAC)))
 *         f.write(serial_BAC)
 *         f.close()             # <<<<<<<<<<<<<<
 * 
 *     def check_bitmasks(self):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_close); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1040, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1040, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1023
 *         return self.bitmasks._iseq(solf.get_bitmasks())
 * 
 *     def save_bitmasks(self, fname):             # <<<<<<<<<<<<<<
 *         cdef bytes serial_BAC
 *         cdef np.uint64_t ifile
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.save_bitmasks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_serial_BAC);
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1042
 *         f.close()
 * 
 *     def check_bitmasks(self):             # <<<<<<<<<<<<<<
 *         return self.bitmasks._check()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_35check_bitmasks(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_35check_bitmasks(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("check_bitmasks (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_34check_bitmasks(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_34check_bitmasks(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check_bitmasks", 0);

  /* "yt/geometry/particle_oct_container.pyx":1043
 * 
 *     def check_bitmasks(self):
 *         return self.bitmasks._check()             # <<<<<<<<<<<<<<
 * 
 *     def reset_bitmasks(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmasks->__pyx_vtab)->_check(__pyx_v_self->bitmasks)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1042
 *         f.close()
 * 
 *     def check_bitmasks(self):             # <<<<<<<<<<<<<<
 *         return self.bitmasks._check()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.check_bitmasks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1045
 *         return self.bitmasks._check()
 * 
 *     def reset_bitmasks(self):             # <<<<<<<<<<<<<<
 *         self.bitmasks._reset()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_37reset_bitmasks(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_37reset_bitmasks(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset_bitmasks (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_36reset_bitmasks(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_36reset_bitmasks(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset_bitmasks", 0);

  /* "yt/geometry/particle_oct_container.pyx":1046
 * 
 *     def reset_bitmasks(self):
 *         self.bitmasks._reset()             # <<<<<<<<<<<<<<
 * 
 *     def load_bitmasks(self, fname):
 */
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmasks->__pyx_vtab)->_reset(__pyx_v_self->bitmasks);

  /* "yt/geometry/particle_oct_container.pyx":1045
 *         return self.bitmasks._check()
 * 
 *     def reset_bitmasks(self):             # <<<<<<<<<<<<<<
 *         self.bitmasks._reset()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1048
 *         self.bitmasks._reset()
 * 
 *     def load_bitmasks(self, fname):             # <<<<<<<<<<<<<<
 *         cdef bint read_flag = 1
 *         cdef bint irflag
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_39load_bitmasks(PyObject *__pyx_v_self, PyObject *__pyx_v_fname); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_39load_bitmasks(PyObject *__pyx_v_self, PyObject *__pyx_v_fname) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("load_bitmasks (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_38load_bitmasks(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_fname));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_38load_bitmasks(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_fname) {
  int __pyx_v_read_flag;
  int __pyx_v_irflag;
  __pyx_t_5numpy_uint64_t __pyx_v_ver;
  __pyx_t_5numpy_uint64_t __pyx_v_nfiles;
  __pyx_t_5numpy_int64_t __pyx_v_file_hash;
  __pyx_t_5numpy_uint64_t __pyx_v_size_serial;
  int __pyx_v_overwrite;
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_v_pb = NULL;
  PyObject *__pyx_v_ifile = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  __pyx_t_5numpy_uint64_t __pyx_t_12;
  __pyx_t_5numpy_int64_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  PyObject *(*__pyx_t_15)(PyObject *);
  PyObject *__pyx_t_16 = NULL;
  __pyx_t_5numpy_uint32_t __pyx_t_17;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("load_bitmasks", 0);

  /* "yt/geometry/particle_oct_container.pyx":1049
 * 
 *     def load_bitmasks(self, fname):
 *         cdef bint read_flag = 1             # <<<<<<<<<<<<<<
 *         cdef bint irflag
 *         cdef np.uint64_t ver
 */
  __pyx_v_read_flag = 1;

  /* "yt/geometry/particle_oct_container.pyx":1052
 *         cdef bint irflag
 *         cdef np.uint64_t ver
 *         cdef np.uint64_t nfiles = 0             # <<<<<<<<<<<<<<
 *         cdef np.int64_t file_hash
 *         cdef np.uint64_t size_serial
 */
  __pyx_v_nfiles = 0;

  /* "yt/geometry/particle_oct_container.pyx":1055
 *         cdef np.int64_t file_hash
 *         cdef np.uint64_t size_serial
 *         cdef bint overwrite = 0             # <<<<<<<<<<<<<<
 *         # Verify that file is correct version
 *         if not os.path.isfile(fname):
 */
  __pyx_v_overwrite = 0;

  /* "yt/geometry/particle_oct_container.pyx":1057
 *         cdef bint overwrite = 0
 *         # Verify that file is correct version
 *         if not os.path.isfile(fname):             # <<<<<<<<<<<<<<
 *             raise OSError
 *         f = open(fname,'rb')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_os); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_path); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_isfile); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_fname) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_fname);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1057, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1057, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "yt/geometry/particle_oct_container.pyx":1058
 *         # Verify that file is correct version
 *         if not os.path.isfile(fname):
 *             raise OSError             # <<<<<<<<<<<<<<
 *         f = open(fname,'rb')
 *         ver, = struct.unpack('Q',f.read(struct.calcsize('Q')))
 */
    __Pyx_Raise(__pyx_builtin_OSError, 0, 0, 0);
    __PYX_ERR(0, 1058, __pyx_L1_error)

    /* "yt/geometry/particle_oct_container.pyx":1057
 *         cdef bint overwrite = 0
 *         # Verify that file is correct version
 *         if not os.path.isfile(fname):             # <<<<<<<<<<<<<<
 *             raise OSError
 *         f = open(fname,'rb')
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":1059
 *         if not os.path.isfile(fname):
 *             raise OSError
 *         f = open(fname,'rb')             # <<<<<<<<<<<<<<
 *         ver, = struct.unpack('Q',f.read(struct.calcsize('Q')))
 *         if ver == self.nfiles and ver != _bitmask_version:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_fname);
  __Pyx_GIVEREF(__pyx_v_fname);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_fname);
  __Pyx_INCREF(__pyx_n_s_rb);
  __Pyx_GIVEREF(__pyx_n_s_rb);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_rb);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1059, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_f = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1060
 *             raise OSError
 *         f = open(fname,'rb')
 *         ver, = struct.unpack('Q',f.read(struct.calcsize('Q')))             # <<<<<<<<<<<<<<
 *         if ver == self.nfiles and ver != _bitmask_version:
 *             overwrite = 1
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_struct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_unpack); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_read); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_struct); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_calcsize); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_7 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_8, __pyx_n_s_Q) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_n_s_Q);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  __pyx_t_10 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_10 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_n_s_Q, __pyx_t_1};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1060, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_n_s_Q, __pyx_t_1};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1060, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1060, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_INCREF(__pyx_n_s_Q);
    __Pyx_GIVEREF(__pyx_n_s_Q);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_10, __pyx_n_s_Q);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_10, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1060, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 1)) {
      if (size > 1) __Pyx_RaiseTooManyValuesError(1);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1060, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
    } else {
      __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
    }
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1060, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_7 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1060, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_11 = Py_TYPE(__pyx_t_7)->tp_iternext;
    index = 0; __pyx_t_3 = __pyx_t_11(__pyx_t_7); if (unlikely(!__pyx_t_3)) goto __pyx_L4_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_7), 1) < 0) __PYX_ERR(0, 1060, __pyx_L1_error)
    __pyx_t_11 = NULL;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L5_unpacking_done;
    __pyx_L4_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_11 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1060, __pyx_L1_error)
    __pyx_L5_unpacking_done:;
  }
  __pyx_t_12 = __Pyx_PyInt_As_npy_uint64(__pyx_t_3); if (unlikely((__pyx_t_12 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_ver = __pyx_t_12;

  /* "yt/geometry/particle_oct_container.pyx":1061
 *         f = open(fname,'rb')
 *         ver, = struct.unpack('Q',f.read(struct.calcsize('Q')))
 *         if ver == self.nfiles and ver != _bitmask_version:             # <<<<<<<<<<<<<<
 *             overwrite = 1
 *             nfiles = ver
 */
  __pyx_t_4 = ((__pyx_v_ver == __pyx_v_self->nfiles) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_5 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyInt_From_npy_uint64(__pyx_v_ver); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_bitmask_version); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1061, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_5 = __pyx_t_4;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_5) {

    /* "yt/geometry/particle_oct_container.pyx":1062
 *         ver, = struct.unpack('Q',f.read(struct.calcsize('Q')))
 *         if ver == self.nfiles and ver != _bitmask_version:
 *             overwrite = 1             # <<<<<<<<<<<<<<
 *             nfiles = ver
 *             ver = 0 # Original bitmaps had number of files first
 */
    __pyx_v_overwrite = 1;

    /* "yt/geometry/particle_oct_container.pyx":1063
 *         if ver == self.nfiles and ver != _bitmask_version:
 *             overwrite = 1
 *             nfiles = ver             # <<<<<<<<<<<<<<
 *             ver = 0 # Original bitmaps had number of files first
 *         if ver != _bitmask_version:
 */
    __pyx_v_nfiles = __pyx_v_ver;

    /* "yt/geometry/particle_oct_container.pyx":1064
 *             overwrite = 1
 *             nfiles = ver
 *             ver = 0 # Original bitmaps had number of files first             # <<<<<<<<<<<<<<
 *         if ver != _bitmask_version:
 *             raise OSError("The file format of the index has changed since "
 */
    __pyx_v_ver = 0;

    /* "yt/geometry/particle_oct_container.pyx":1061
 *         f = open(fname,'rb')
 *         ver, = struct.unpack('Q',f.read(struct.calcsize('Q')))
 *         if ver == self.nfiles and ver != _bitmask_version:             # <<<<<<<<<<<<<<
 *             overwrite = 1
 *             nfiles = ver
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":1065
 *             nfiles = ver
 *             ver = 0 # Original bitmaps had number of files first
 *         if ver != _bitmask_version:             # <<<<<<<<<<<<<<
 *             raise OSError("The file format of the index has changed since "
 *                           "this file was created. It will be replaced with an "
 */
  __pyx_t_7 = __Pyx_PyInt_From_npy_uint64(__pyx_v_ver); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_bitmask_version); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1065, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_7, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1065, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1065, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_t_5)) {

    /* "yt/geometry/particle_oct_container.pyx":1066
 *             ver = 0 # Original bitmaps had number of files first
 *         if ver != _bitmask_version:
 *             raise OSError("The file format of the index has changed since "             # <<<<<<<<<<<<<<
 *                           "this file was created. It will be replaced with an "
 *                           "updated version.")
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_OSError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1066, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1066, __pyx_L1_error)

    /* "yt/geometry/particle_oct_container.pyx":1065
 *             nfiles = ver
 *             ver = 0 # Original bitmaps had number of files first
 *         if ver != _bitmask_version:             # <<<<<<<<<<<<<<
 *             raise OSError("The file format of the index has changed since "
 *                           "this file was created. It will be replaced with an "
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":1070
 *                           "updated version.")
 *         # Read file hash
 *         file_hash, = struct.unpack('q', f.read(struct.calcsize('q')))             # <<<<<<<<<<<<<<
 *         if file_hash != self.file_hash:
 *             raise OSError
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_struct); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_unpack); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_read); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_struct); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_calcsize); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_6 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_n_s_q) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_n_s_q);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  __pyx_t_10 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_10 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_7)) {
    PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_n_s_q, __pyx_t_3};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1070, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
    PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_n_s_q, __pyx_t_3};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1070, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1070, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_1) {
      __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1); __pyx_t_1 = NULL;
    }
    __Pyx_INCREF(__pyx_n_s_q);
    __Pyx_GIVEREF(__pyx_n_s_q);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_10, __pyx_n_s_q);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_10, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1070, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 1)) {
      if (size > 1) __Pyx_RaiseTooManyValuesError(1);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1070, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
    } else {
      __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
    }
    __Pyx_INCREF(__pyx_t_7);
    #else
    __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1070, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_6 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1070, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_11 = Py_TYPE(__pyx_t_6)->tp_iternext;
    index = 0; __pyx_t_7 = __pyx_t_11(__pyx_t_6); if (unlikely(!__pyx_t_7)) goto __pyx_L10_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_7);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_6), 1) < 0) __PYX_ERR(0, 1070, __pyx_L1_error)
    __pyx_t_11 = NULL;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L11_unpacking_done;
    __pyx_L10_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_11 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1070, __pyx_L1_error)
    __pyx_L11_unpacking_done:;
  }
  __pyx_t_13 = __Pyx_PyInt_As_npy_int64(__pyx_t_7); if (unlikely((__pyx_t_13 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_file_hash = __pyx_t_13;

  /* "yt/geometry/particle_oct_container.pyx":1071
 *         # Read file hash
 *         file_hash, = struct.unpack('q', f.read(struct.calcsize('q')))
 *         if file_hash != self.file_hash:             # <<<<<<<<<<<<<<
 *             raise OSError
 *         # Read number of bitmaps
 */
  __pyx_t_5 = ((__pyx_v_file_hash != __pyx_v_self->file_hash) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "yt/geometry/particle_oct_container.pyx":1072
 *         file_hash, = struct.unpack('q', f.read(struct.calcsize('q')))
 *         if file_hash != self.file_hash:
 *             raise OSError             # <<<<<<<<<<<<<<
 *         # Read number of bitmaps
 *         if nfiles == 0:
 */
    __Pyx_Raise(__pyx_builtin_OSError, 0, 0, 0);
    __PYX_ERR(0, 1072, __pyx_L1_error)

    /* "yt/geometry/particle_oct_container.pyx":1071
 *         # Read file hash
 *         file_hash, = struct.unpack('q', f.read(struct.calcsize('q')))
 *         if file_hash != self.file_hash:             # <<<<<<<<<<<<<<
 *             raise OSError
 *         # Read number of bitmaps
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":1074
 *             raise OSError
 *         # Read number of bitmaps
 *         if nfiles == 0:             # <<<<<<<<<<<<<<
 *             nfiles, = struct.unpack('Q', f.read(struct.calcsize('Q')))
 *             if nfiles != self.nfiles:
 */
  __pyx_t_5 = ((__pyx_v_nfiles == 0) != 0);
  if (__pyx_t_5) {

    /* "yt/geometry/particle_oct_container.pyx":1075
 *         # Read number of bitmaps
 *         if nfiles == 0:
 *             nfiles, = struct.unpack('Q', f.read(struct.calcsize('Q')))             # <<<<<<<<<<<<<<
 *             if nfiles != self.nfiles:
 *                 raise OSError(
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_struct); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1075, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_unpack); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1075, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_read); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1075, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_struct); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1075, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_calcsize); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1075, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_8, __pyx_n_s_Q) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_n_s_Q);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1075, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_7 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_9, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1075, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_n_s_Q, __pyx_t_7};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1075, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_n_s_Q, __pyx_t_7};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1075, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_1 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1075, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_INCREF(__pyx_n_s_Q);
      __Pyx_GIVEREF(__pyx_n_s_Q);
      PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_10, __pyx_n_s_Q);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_10, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1075, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 1)) {
        if (size > 1) __Pyx_RaiseTooManyValuesError(1);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1075, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
      } else {
        __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
      }
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1075, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1075, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_1)->tp_iternext;
      index = 0; __pyx_t_6 = __pyx_t_11(__pyx_t_1); if (unlikely(!__pyx_t_6)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_1), 1) < 0) __PYX_ERR(0, 1075, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L15_unpacking_done;
      __pyx_L14_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1075, __pyx_L1_error)
      __pyx_L15_unpacking_done:;
    }
    __pyx_t_12 = __Pyx_PyInt_As_npy_uint64(__pyx_t_6); if (unlikely((__pyx_t_12 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1075, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_nfiles = __pyx_t_12;

    /* "yt/geometry/particle_oct_container.pyx":1076
 *         if nfiles == 0:
 *             nfiles, = struct.unpack('Q', f.read(struct.calcsize('Q')))
 *             if nfiles != self.nfiles:             # <<<<<<<<<<<<<<
 *                 raise OSError(
 *                     "Number of bitmasks ({}) conflicts with number of files "
 */
    __pyx_t_5 = ((__pyx_v_nfiles != __pyx_v_self->nfiles) != 0);
    if (unlikely(__pyx_t_5)) {

      /* "yt/geometry/particle_oct_container.pyx":1079
 *                 raise OSError(
 *                     "Number of bitmasks ({}) conflicts with number of files "
 *                     "({})".format(nfiles, self.nfiles))             # <<<<<<<<<<<<<<
 *         # Read bitmap for each file
 *         pb = get_pbar("Loading particle index", nfiles)
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_Number_of_bitmasks_conflicts_wit, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1079, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __Pyx_PyInt_From_npy_uint64(__pyx_v_nfiles); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1079, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->nfiles); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1079, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_3 = NULL;
      __pyx_t_10 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_10 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_1, __pyx_t_7};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1079, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_1, __pyx_t_7};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1079, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1079, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_3) {
          __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_3); __pyx_t_3 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_10, __pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_10, __pyx_t_7);
        __pyx_t_1 = 0;
        __pyx_t_7 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1079, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "yt/geometry/particle_oct_container.pyx":1077
 *             nfiles, = struct.unpack('Q', f.read(struct.calcsize('Q')))
 *             if nfiles != self.nfiles:
 *                 raise OSError(             # <<<<<<<<<<<<<<
 *                     "Number of bitmasks ({}) conflicts with number of files "
 *                     "({})".format(nfiles, self.nfiles))
 */
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_OSError, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1077, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 1077, __pyx_L1_error)

      /* "yt/geometry/particle_oct_container.pyx":1076
 *         if nfiles == 0:
 *             nfiles, = struct.unpack('Q', f.read(struct.calcsize('Q')))
 *             if nfiles != self.nfiles:             # <<<<<<<<<<<<<<
 *                 raise OSError(
 *                     "Number of bitmasks ({}) conflicts with number of files "
 */
    }

    /* "yt/geometry/particle_oct_container.pyx":1074
 *             raise OSError
 *         # Read number of bitmaps
 *         if nfiles == 0:             # <<<<<<<<<<<<<<
 *             nfiles, = struct.unpack('Q', f.read(struct.calcsize('Q')))
 *             if nfiles != self.nfiles:
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":1081
 *                     "({})".format(nfiles, self.nfiles))
 *         # Read bitmap for each file
 *         pb = get_pbar("Loading particle index", nfiles)             # <<<<<<<<<<<<<<
 *         for ifile in range(nfiles):
 *             pb.update(ifile+1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_get_pbar); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = __Pyx_PyInt_From_npy_uint64(__pyx_v_nfiles); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = NULL;
  __pyx_t_10 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_10 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_kp_s_Loading_particle_index, __pyx_t_9};
    __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1081, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_kp_s_Loading_particle_index, __pyx_t_9};
    __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1081, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  } else
  #endif
  {
    __pyx_t_1 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1081, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_s_Loading_particle_index);
    __Pyx_GIVEREF(__pyx_kp_s_Loading_particle_index);
    PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_10, __pyx_kp_s_Loading_particle_index);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_10, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1081, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_pb = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1082
 *         # Read bitmap for each file
 *         pb = get_pbar("Loading particle index", nfiles)
 *         for ifile in range(nfiles):             # <<<<<<<<<<<<<<
 *             pb.update(ifile+1)
 *             size_serial, = struct.unpack('Q', f.read(struct.calcsize('Q')))
 */
  __pyx_t_6 = __Pyx_PyInt_From_npy_uint64(__pyx_v_nfiles); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1082, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1082, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_6 = __pyx_t_2; __Pyx_INCREF(__pyx_t_6); __pyx_t_14 = 0;
    __pyx_t_15 = NULL;
  } else {
    __pyx_t_14 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1082, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_15 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1082, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_15)) {
      if (likely(PyList_CheckExact(__pyx_t_6))) {
        if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_14); __Pyx_INCREF(__pyx_t_2); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1082, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_6, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1082, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_14); __Pyx_INCREF(__pyx_t_2); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 1082, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_6, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1082, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_15(__pyx_t_6);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1082, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_ifile, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "yt/geometry/particle_oct_container.pyx":1083
 *         pb = get_pbar("Loading particle index", nfiles)
 *         for ifile in range(nfiles):
 *             pb.update(ifile+1)             # <<<<<<<<<<<<<<
 *             size_serial, = struct.unpack('Q', f.read(struct.calcsize('Q')))
 *             irflag = self.bitmasks._loads(ifile, f.read(size_serial))
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_pb, __pyx_n_s_update); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1083, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyInt_AddObjC(__pyx_v_ifile, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1083, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_7, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1083, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "yt/geometry/particle_oct_container.pyx":1084
 *         for ifile in range(nfiles):
 *             pb.update(ifile+1)
 *             size_serial, = struct.unpack('Q', f.read(struct.calcsize('Q')))             # <<<<<<<<<<<<<<
 *             irflag = self.bitmasks._loads(ifile, f.read(size_serial))
 *             if irflag == 0:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_struct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_unpack); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_read); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_struct); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_calcsize); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_16))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_16);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_16, function);
      }
    }
    __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_16, __pyx_t_8, __pyx_n_s_Q) : __Pyx_PyObject_CallOneArg(__pyx_t_16, __pyx_n_s_Q);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    __pyx_t_16 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_16)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_16);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_16, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_9)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_n_s_Q, __pyx_t_1};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1084, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_n_s_Q, __pyx_t_1};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1084, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_3 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1084, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_INCREF(__pyx_n_s_Q);
      __Pyx_GIVEREF(__pyx_n_s_Q);
      PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_10, __pyx_n_s_Q);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_10, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1084, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 1)) {
        if (size > 1) __Pyx_RaiseTooManyValuesError(1);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1084, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
      } else {
        __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
      }
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1084, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1084, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_11 = Py_TYPE(__pyx_t_3)->tp_iternext;
      index = 0; __pyx_t_9 = __pyx_t_11(__pyx_t_3); if (unlikely(!__pyx_t_9)) goto __pyx_L19_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_3), 1) < 0) __PYX_ERR(0, 1084, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L20_unpacking_done;
      __pyx_L19_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1084, __pyx_L1_error)
      __pyx_L20_unpacking_done:;
    }
    __pyx_t_12 = __Pyx_PyInt_As_npy_uint64(__pyx_t_9); if (unlikely((__pyx_t_12 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1084, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_v_size_serial = __pyx_t_12;

    /* "yt/geometry/particle_oct_container.pyx":1085
 *             pb.update(ifile+1)
 *             size_serial, = struct.unpack('Q', f.read(struct.calcsize('Q')))
 *             irflag = self.bitmasks._loads(ifile, f.read(size_serial))             # <<<<<<<<<<<<<<
 *             if irflag == 0:
 *                 read_flag = 0
 */
    __pyx_t_17 = __Pyx_PyInt_As_npy_uint32(__pyx_v_ifile); if (unlikely((__pyx_t_17 == ((npy_uint32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1085, __pyx_L1_error)
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_read); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1085, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_size_serial); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1085, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1085, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 1085, __pyx_L1_error)
    __pyx_v_irflag = ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmasks->__pyx_vtab)->_loads(__pyx_v_self->bitmasks, __pyx_t_17, ((PyObject*)__pyx_t_2));
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "yt/geometry/particle_oct_container.pyx":1086
 *             size_serial, = struct.unpack('Q', f.read(struct.calcsize('Q')))
 *             irflag = self.bitmasks._loads(ifile, f.read(size_serial))
 *             if irflag == 0:             # <<<<<<<<<<<<<<
 *                 read_flag = 0
 *         pb.finish()
 */
    __pyx_t_5 = ((__pyx_v_irflag == 0) != 0);
    if (__pyx_t_5) {

      /* "yt/geometry/particle_oct_container.pyx":1087
 *             irflag = self.bitmasks._loads(ifile, f.read(size_serial))
 *             if irflag == 0:
 *                 read_flag = 0             # <<<<<<<<<<<<<<
 *         pb.finish()
 *         # Collisions
 */
      __pyx_v_read_flag = 0;

      /* "yt/geometry/particle_oct_container.pyx":1086
 *             size_serial, = struct.unpack('Q', f.read(struct.calcsize('Q')))
 *             irflag = self.bitmasks._loads(ifile, f.read(size_serial))
 *             if irflag == 0:             # <<<<<<<<<<<<<<
 *                 read_flag = 0
 *         pb.finish()
 */
    }

    /* "yt/geometry/particle_oct_container.pyx":1082
 *         # Read bitmap for each file
 *         pb = get_pbar("Loading particle index", nfiles)
 *         for ifile in range(nfiles):             # <<<<<<<<<<<<<<
 *             pb.update(ifile+1)
 *             size_serial, = struct.unpack('Q', f.read(struct.calcsize('Q')))
 */
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1088
 *             if irflag == 0:
 *                 read_flag = 0
 *         pb.finish()             # <<<<<<<<<<<<<<
 *         # Collisions
 *         size_serial, = struct.unpack('Q',f.read(struct.calcsize('Q')))
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_pb, __pyx_n_s_finish); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_6 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1090
 *         pb.finish()
 *         # Collisions
 *         size_serial, = struct.unpack('Q',f.read(struct.calcsize('Q')))             # <<<<<<<<<<<<<<
 *         irflag = self.collisions._loads(f.read(size_serial))
 *         f.close()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_struct); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_unpack); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_read); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_struct); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_calcsize); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_16))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_16);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_16, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_16, __pyx_t_7, __pyx_n_s_Q) : __Pyx_PyObject_CallOneArg(__pyx_t_16, __pyx_n_s_Q);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  __pyx_t_16 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_16)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_16);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_16, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_10 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
      __pyx_t_10 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_9)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_n_s_Q, __pyx_t_2};
    __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1090, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_n_s_Q, __pyx_t_2};
    __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1090, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_1 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1090, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_n_s_Q);
    __Pyx_GIVEREF(__pyx_n_s_Q);
    PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_10, __pyx_n_s_Q);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_10, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_1, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1090, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
    PyObject* sequence = __pyx_t_6;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 1)) {
      if (size > 1) __Pyx_RaiseTooManyValuesError(1);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1090, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
    } else {
      __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
    }
    __Pyx_INCREF(__pyx_t_9);
    #else
    __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1090, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    #endif
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_1 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1090, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_11 = Py_TYPE(__pyx_t_1)->tp_iternext;
    index = 0; __pyx_t_9 = __pyx_t_11(__pyx_t_1); if (unlikely(!__pyx_t_9)) goto __pyx_L22_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_9);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_1), 1) < 0) __PYX_ERR(0, 1090, __pyx_L1_error)
    __pyx_t_11 = NULL;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    goto __pyx_L23_unpacking_done;
    __pyx_L22_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_11 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1090, __pyx_L1_error)
    __pyx_L23_unpacking_done:;
  }
  __pyx_t_12 = __Pyx_PyInt_As_npy_uint64(__pyx_t_9); if (unlikely((__pyx_t_12 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_v_size_serial = __pyx_t_12;

  /* "yt/geometry/particle_oct_container.pyx":1091
 *         # Collisions
 *         size_serial, = struct.unpack('Q',f.read(struct.calcsize('Q')))
 *         irflag = self.collisions._loads(f.read(size_serial))             # <<<<<<<<<<<<<<
 *         f.close()
 *         # Save in correct format
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_read); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1091, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint64(__pyx_v_size_serial); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1091, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1091, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (!(likely(PyBytes_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(0, 1091, __pyx_L1_error)
  __pyx_v_irflag = ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_v_self->collisions->__pyx_vtab)->_loads(__pyx_v_self->collisions, ((PyObject*)__pyx_t_6));
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1092
 *         size_serial, = struct.unpack('Q',f.read(struct.calcsize('Q')))
 *         irflag = self.collisions._loads(f.read(size_serial))
 *         f.close()             # <<<<<<<<<<<<<<
 *         # Save in correct format
 *         if overwrite == 1:
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_close); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_6 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1092, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1094
 *         f.close()
 *         # Save in correct format
 *         if overwrite == 1:             # <<<<<<<<<<<<<<
 *             self.save_bitmasks(fname)
 *         return read_flag
 */
  __pyx_t_5 = ((__pyx_v_overwrite == 1) != 0);
  if (__pyx_t_5) {

    /* "yt/geometry/particle_oct_container.pyx":1095
 *         # Save in correct format
 *         if overwrite == 1:
 *             self.save_bitmasks(fname)             # <<<<<<<<<<<<<<
 *         return read_flag
 * 
 */
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_save_bitmasks); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1095, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_6 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_1, __pyx_v_fname) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_fname);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1095, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "yt/geometry/particle_oct_container.pyx":1094
 *         f.close()
 *         # Save in correct format
 *         if overwrite == 1:             # <<<<<<<<<<<<<<
 *             self.save_bitmasks(fname)
 *         return read_flag
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":1096
 *         if overwrite == 1:
 *             self.save_bitmasks(fname)
 *         return read_flag             # <<<<<<<<<<<<<<
 * 
 *     def print_info(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_read_flag); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1096, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1048
 *         self.bitmasks._reset()
 * 
 *     def load_bitmasks(self, fname):             # <<<<<<<<<<<<<<
 *         cdef bint read_flag = 1
 *         cdef bint irflag
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.load_bitmasks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_pb);
  __Pyx_XDECREF(__pyx_v_ifile);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1098
 *         return read_flag
 * 
 *     def print_info(self):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t ifile
 *         for ifile in range(self.nfiles):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_41print_info(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_41print_info(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("print_info (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_40print_info(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_40print_info(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  __pyx_t_5numpy_uint64_t __pyx_v_ifile;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_uint64_t __pyx_t_1;
  __pyx_t_5numpy_uint64_t __pyx_t_2;
  __pyx_t_5numpy_uint64_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("print_info", 0);

  /* "yt/geometry/particle_oct_container.pyx":1100
 *     def print_info(self):
 *         cdef np.uint64_t ifile
 *         for ifile in range(self.nfiles):             # <<<<<<<<<<<<<<
 *             self.bitmasks.print_info(ifile, "File: %03d" % ifile)
 * 
 */
  __pyx_t_1 = __pyx_v_self->nfiles;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_ifile = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":1101
 *         cdef np.uint64_t ifile
 *         for ifile in range(self.nfiles):
 *             self.bitmasks.print_info(ifile, "File: %03d" % ifile)             # <<<<<<<<<<<<<<
 * 
 *     def count_coarse(self, ifile):
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->bitmasks), __pyx_n_s_print_info); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_npy_uint64(__pyx_v_ifile); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_npy_uint64(__pyx_v_ifile); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyString_Format(__pyx_kp_s_File_03d, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_t_8};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1101, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_t_8};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1101, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_8);
      __pyx_t_6 = 0;
      __pyx_t_8 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "yt/geometry/particle_oct_container.pyx":1098
 *         return read_flag
 * 
 *     def print_info(self):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t ifile
 *         for ifile in range(self.nfiles):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.print_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1103
 *             self.bitmasks.print_info(ifile, "File: %03d" % ifile)
 * 
 *     def count_coarse(self, ifile):             # <<<<<<<<<<<<<<
 *         r"""Get the number of coarse cells set for a file."""
 *         return self.bitmasks.count_coarse(ifile)
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_43count_coarse(PyObject *__pyx_v_self, PyObject *__pyx_v_ifile); /*proto*/
static char __pyx_doc_2yt_8geometry_22particle_oct_container_14ParticleBitmap_42count_coarse[] = "Get the number of coarse cells set for a file.";
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_43count_coarse(PyObject *__pyx_v_self, PyObject *__pyx_v_ifile) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("count_coarse (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_42count_coarse(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_ifile));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_42count_coarse(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_ifile) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_coarse", 0);

  /* "yt/geometry/particle_oct_container.pyx":1105
 *     def count_coarse(self, ifile):
 *         r"""Get the number of coarse cells set for a file."""
 *         return self.bitmasks.count_coarse(ifile)             # <<<<<<<<<<<<<<
 * 
 *     def count_refined(self, ifile):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->bitmasks), __pyx_n_s_count_coarse); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_ifile) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_ifile);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1103
 *             self.bitmasks.print_info(ifile, "File: %03d" % ifile)
 * 
 *     def count_coarse(self, ifile):             # <<<<<<<<<<<<<<
 *         r"""Get the number of coarse cells set for a file."""
 *         return self.bitmasks.count_coarse(ifile)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.count_coarse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1107
 *         return self.bitmasks.count_coarse(ifile)
 * 
 *     def count_refined(self, ifile):             # <<<<<<<<<<<<<<
 *         r"""Get the number of cells refined for a file."""
 *         return self.bitmasks.count_refined(ifile)
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_45count_refined(PyObject *__pyx_v_self, PyObject *__pyx_v_ifile); /*proto*/
static char __pyx_doc_2yt_8geometry_22particle_oct_container_14ParticleBitmap_44count_refined[] = "Get the number of cells refined for a file.";
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_45count_refined(PyObject *__pyx_v_self, PyObject *__pyx_v_ifile) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("count_refined (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_44count_refined(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_ifile));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_44count_refined(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_ifile) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_refined", 0);

  /* "yt/geometry/particle_oct_container.pyx":1109
 *     def count_refined(self, ifile):
 *         r"""Get the number of cells refined for a file."""
 *         return self.bitmasks.count_refined(ifile)             # <<<<<<<<<<<<<<
 * 
 *     def count_total(self, ifile):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->bitmasks), __pyx_n_s_count_refined); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_ifile) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_ifile);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1107
 *         return self.bitmasks.count_coarse(ifile)
 * 
 *     def count_refined(self, ifile):             # <<<<<<<<<<<<<<
 *         r"""Get the number of cells refined for a file."""
 *         return self.bitmasks.count_refined(ifile)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.count_refined", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1111
 *         return self.bitmasks.count_refined(ifile)
 * 
 *     def count_total(self, ifile):             # <<<<<<<<<<<<<<
 *         r"""Get the total number of cells set for a file."""
 *         return self.bitmasks.count_total(ifile)
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_47count_total(PyObject *__pyx_v_self, PyObject *__pyx_v_ifile); /*proto*/
static char __pyx_doc_2yt_8geometry_22particle_oct_container_14ParticleBitmap_46count_total[] = "Get the total number of cells set for a file.";
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_47count_total(PyObject *__pyx_v_self, PyObject *__pyx_v_ifile) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("count_total (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_46count_total(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_ifile));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_46count_total(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_ifile) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_total", 0);

  /* "yt/geometry/particle_oct_container.pyx":1113
 *     def count_total(self, ifile):
 *         r"""Get the total number of cells set for a file."""
 *         return self.bitmasks.count_total(ifile)             # <<<<<<<<<<<<<<
 * 
 *     def check(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->bitmasks), __pyx_n_s_count_total); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_ifile) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_ifile);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1111
 *         return self.bitmasks.count_refined(ifile)
 * 
 *     def count_total(self, ifile):             # <<<<<<<<<<<<<<
 *         r"""Get the total number of cells set for a file."""
 *         return self.bitmasks.count_total(ifile)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.count_total", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1115
 *         return self.bitmasks.count_total(ifile)
 * 
 *     def check(self):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t mi1
 *         cdef ewah_bool_array arr_totref, arr_tottwo
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_49check(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_49check(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("check (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_48check(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_48check(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  __pyx_t_5numpy_uint64_t __pyx_v_mi1;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array __pyx_v_arr_totref;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array __pyx_v_arr_tottwo;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array __pyx_v_arr;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array __pyx_v_arr_any;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array __pyx_v_arr_two;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array __pyx_v_arr_swap;
  std::vector<size_t>  __pyx_v_vec_totref;
  std::vector<size_t> ::iterator __pyx_v_it_mi1;
  int __pyx_v_nm;
  int __pyx_v_nc;
  __pyx_t_5numpy_uint64_t __pyx_v_ifile;
  __pyx_t_5numpy_uint64_t __pyx_v_nbitmasks;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  __pyx_t_5numpy_uint64_t __pyx_t_3;
  __pyx_t_5numpy_uint64_t __pyx_t_4;
  __pyx_t_5numpy_uint64_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  double __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check", 0);

  /* "yt/geometry/particle_oct_container.pyx":1121
 *         cdef vector[size_t] vec_totref
 *         cdef vector[size_t].iterator it_mi1
 *         cdef int nm = 0, nc = 0             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t ifile, nbitmasks
 *         nbitmasks = len(self.bitmasks)
 */
  __pyx_v_nm = 0;
  __pyx_v_nc = 0;

  /* "yt/geometry/particle_oct_container.pyx":1123
 *         cdef int nm = 0, nc = 0
 *         cdef np.uint64_t ifile, nbitmasks
 *         nbitmasks = len(self.bitmasks)             # <<<<<<<<<<<<<<
 *         # Locate all indices with second level refinement
 *         for ifile in range(self.nfiles):
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->bitmasks);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_nbitmasks = __pyx_t_2;

  /* "yt/geometry/particle_oct_container.pyx":1125
 *         nbitmasks = len(self.bitmasks)
 *         # Locate all indices with second level refinement
 *         for ifile in range(self.nfiles):             # <<<<<<<<<<<<<<
 *             arr = (<ewah_bool_array**> self.bitmasks.ewah_refn)[ifile][0]
 *             arr_totref.logicalor(arr,arr_totref)
 */
  __pyx_t_3 = __pyx_v_self->nfiles;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_ifile = __pyx_t_5;

    /* "yt/geometry/particle_oct_container.pyx":1126
 *         # Locate all indices with second level refinement
 *         for ifile in range(self.nfiles):
 *             arr = (<ewah_bool_array**> self.bitmasks.ewah_refn)[ifile][0]             # <<<<<<<<<<<<<<
 *             arr_totref.logicalor(arr,arr_totref)
 *         # Count collections & second level indices
 */
    __pyx_v_arr = ((((__pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array **)__pyx_v_self->bitmasks->ewah_refn)[__pyx_v_ifile])[0]);

    /* "yt/geometry/particle_oct_container.pyx":1127
 *         for ifile in range(self.nfiles):
 *             arr = (<ewah_bool_array**> self.bitmasks.ewah_refn)[ifile][0]
 *             arr_totref.logicalor(arr,arr_totref)             # <<<<<<<<<<<<<<
 *         # Count collections & second level indices
 *         vec_totref = arr_totref.toArray()
 */
    __pyx_v_arr_totref.logicalor(__pyx_v_arr, __pyx_v_arr_totref);
  }

  /* "yt/geometry/particle_oct_container.pyx":1129
 *             arr_totref.logicalor(arr,arr_totref)
 *         # Count collections & second level indices
 *         vec_totref = arr_totref.toArray()             # <<<<<<<<<<<<<<
 *         it_mi1 = vec_totref.begin()
 *         while it_mi1 != vec_totref.end():
 */
  __pyx_v_vec_totref = __pyx_v_arr_totref.toArray();

  /* "yt/geometry/particle_oct_container.pyx":1130
 *         # Count collections & second level indices
 *         vec_totref = arr_totref.toArray()
 *         it_mi1 = vec_totref.begin()             # <<<<<<<<<<<<<<
 *         while it_mi1 != vec_totref.end():
 *             mi1 = dereference(it_mi1)
 */
  __pyx_v_it_mi1 = __pyx_v_vec_totref.begin();

  /* "yt/geometry/particle_oct_container.pyx":1131
 *         vec_totref = arr_totref.toArray()
 *         it_mi1 = vec_totref.begin()
 *         while it_mi1 != vec_totref.end():             # <<<<<<<<<<<<<<
 *             mi1 = dereference(it_mi1)
 *             arr_any.reset()
 */
  while (1) {
    __pyx_t_6 = ((__pyx_v_it_mi1 != __pyx_v_vec_totref.end()) != 0);
    if (!__pyx_t_6) break;

    /* "yt/geometry/particle_oct_container.pyx":1132
 *         it_mi1 = vec_totref.begin()
 *         while it_mi1 != vec_totref.end():
 *             mi1 = dereference(it_mi1)             # <<<<<<<<<<<<<<
 *             arr_any.reset()
 *             arr_two.reset()
 */
    __pyx_v_mi1 = (*__pyx_v_it_mi1);

    /* "yt/geometry/particle_oct_container.pyx":1133
 *         while it_mi1 != vec_totref.end():
 *             mi1 = dereference(it_mi1)
 *             arr_any.reset()             # <<<<<<<<<<<<<<
 *             arr_two.reset()
 *             for ifile in range(nbitmasks):
 */
    __pyx_v_arr_any.reset();

    /* "yt/geometry/particle_oct_container.pyx":1134
 *             mi1 = dereference(it_mi1)
 *             arr_any.reset()
 *             arr_two.reset()             # <<<<<<<<<<<<<<
 *             for ifile in range(nbitmasks):
 *                 if self.bitmasks._isref(ifile, mi1) == 1:
 */
    __pyx_v_arr_two.reset();

    /* "yt/geometry/particle_oct_container.pyx":1135
 *             arr_any.reset()
 *             arr_two.reset()
 *             for ifile in range(nbitmasks):             # <<<<<<<<<<<<<<
 *                 if self.bitmasks._isref(ifile, mi1) == 1:
 *                     arr = (<cmap[np.int64_t, ewah_bool_array]**> self.bitmasks.ewah_coll)[ifile][0][mi1]
 */
    __pyx_t_3 = __pyx_v_nbitmasks;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_ifile = __pyx_t_5;

      /* "yt/geometry/particle_oct_container.pyx":1136
 *             arr_two.reset()
 *             for ifile in range(nbitmasks):
 *                 if self.bitmasks._isref(ifile, mi1) == 1:             # <<<<<<<<<<<<<<
 *                     arr = (<cmap[np.int64_t, ewah_bool_array]**> self.bitmasks.ewah_coll)[ifile][0][mi1]
 *                     arr_any.logicaland(arr, arr_two) # Indices in previous files
 */
      __pyx_t_6 = ((((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmasks->__pyx_vtab)->_isref(__pyx_v_self->bitmasks, __pyx_v_ifile, __pyx_v_mi1) == 1) != 0);
      if (__pyx_t_6) {

        /* "yt/geometry/particle_oct_container.pyx":1137
 *             for ifile in range(nbitmasks):
 *                 if self.bitmasks._isref(ifile, mi1) == 1:
 *                     arr = (<cmap[np.int64_t, ewah_bool_array]**> self.bitmasks.ewah_coll)[ifile][0][mi1]             # <<<<<<<<<<<<<<
 *                     arr_any.logicaland(arr, arr_two) # Indices in previous files
 *                     arr_any.logicalor(arr, arr_swap) # All second level indices
 */
        __pyx_v_arr = (((((std::map<__pyx_t_5numpy_int64_t,__pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array>  **)__pyx_v_self->bitmasks->ewah_coll)[__pyx_v_ifile])[0])[__pyx_v_mi1]);

        /* "yt/geometry/particle_oct_container.pyx":1138
 *                 if self.bitmasks._isref(ifile, mi1) == 1:
 *                     arr = (<cmap[np.int64_t, ewah_bool_array]**> self.bitmasks.ewah_coll)[ifile][0][mi1]
 *                     arr_any.logicaland(arr, arr_two) # Indices in previous files             # <<<<<<<<<<<<<<
 *                     arr_any.logicalor(arr, arr_swap) # All second level indices
 *                     arr_any = arr_swap
 */
        __pyx_v_arr_any.logicaland(__pyx_v_arr, __pyx_v_arr_two);

        /* "yt/geometry/particle_oct_container.pyx":1139
 *                     arr = (<cmap[np.int64_t, ewah_bool_array]**> self.bitmasks.ewah_coll)[ifile][0][mi1]
 *                     arr_any.logicaland(arr, arr_two) # Indices in previous files
 *                     arr_any.logicalor(arr, arr_swap) # All second level indices             # <<<<<<<<<<<<<<
 *                     arr_any = arr_swap
 *                     arr_two.logicalor(arr_tottwo,arr_tottwo)
 */
        __pyx_v_arr_any.logicalor(__pyx_v_arr, __pyx_v_arr_swap);

        /* "yt/geometry/particle_oct_container.pyx":1140
 *                     arr_any.logicaland(arr, arr_two) # Indices in previous files
 *                     arr_any.logicalor(arr, arr_swap) # All second level indices
 *                     arr_any = arr_swap             # <<<<<<<<<<<<<<
 *                     arr_two.logicalor(arr_tottwo,arr_tottwo)
 *             nc += arr_tottwo.numberOfOnes()
 */
        __pyx_v_arr_any = __pyx_v_arr_swap;

        /* "yt/geometry/particle_oct_container.pyx":1141
 *                     arr_any.logicalor(arr, arr_swap) # All second level indices
 *                     arr_any = arr_swap
 *                     arr_two.logicalor(arr_tottwo,arr_tottwo)             # <<<<<<<<<<<<<<
 *             nc += arr_tottwo.numberOfOnes()
 *             nm += arr_any.numberOfOnes()
 */
        __pyx_v_arr_two.logicalor(__pyx_v_arr_tottwo, __pyx_v_arr_tottwo);

        /* "yt/geometry/particle_oct_container.pyx":1136
 *             arr_two.reset()
 *             for ifile in range(nbitmasks):
 *                 if self.bitmasks._isref(ifile, mi1) == 1:             # <<<<<<<<<<<<<<
 *                     arr = (<cmap[np.int64_t, ewah_bool_array]**> self.bitmasks.ewah_coll)[ifile][0][mi1]
 *                     arr_any.logicaland(arr, arr_two) # Indices in previous files
 */
      }
    }

    /* "yt/geometry/particle_oct_container.pyx":1142
 *                     arr_any = arr_swap
 *                     arr_two.logicalor(arr_tottwo,arr_tottwo)
 *             nc += arr_tottwo.numberOfOnes()             # <<<<<<<<<<<<<<
 *             nm += arr_any.numberOfOnes()
 *             preincrement(it_mi1)
 */
    __pyx_v_nc = (__pyx_v_nc + __pyx_v_arr_tottwo.numberOfOnes());

    /* "yt/geometry/particle_oct_container.pyx":1143
 *                     arr_two.logicalor(arr_tottwo,arr_tottwo)
 *             nc += arr_tottwo.numberOfOnes()
 *             nm += arr_any.numberOfOnes()             # <<<<<<<<<<<<<<
 *             preincrement(it_mi1)
 *         # nc: total number of second level morton indices that are repeated
 */
    __pyx_v_nm = (__pyx_v_nm + __pyx_v_arr_any.numberOfOnes());

    /* "yt/geometry/particle_oct_container.pyx":1144
 *             nc += arr_tottwo.numberOfOnes()
 *             nm += arr_any.numberOfOnes()
 *             preincrement(it_mi1)             # <<<<<<<<<<<<<<
 *         # nc: total number of second level morton indices that are repeated
 *         # nm: total number of second level morton indices
 */
    (void)((++__pyx_v_it_mi1));
  }

  /* "yt/geometry/particle_oct_container.pyx":1147
 *         # nc: total number of second level morton indices that are repeated
 *         # nm: total number of second level morton indices
 *         print("Total of %s / %s collisions (% 3.5f%%)" % (nc, nm, 100.0*float(nc)/nm))             # <<<<<<<<<<<<<<
 * 
 *     def primary_indices(self):
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_nc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_nm); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = (100.0 * ((double)__pyx_v_nc));
  if (unlikely(__pyx_v_nm == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 1147, __pyx_L1_error)
  }
  __pyx_t_9 = PyFloat_FromDouble((__pyx_t_8 / __pyx_v_nm)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = PyTuple_New(3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_9);
  __pyx_t_1 = 0;
  __pyx_t_7 = 0;
  __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyString_Format(__pyx_kp_s_Total_of_s_s_collisions_3_5f, __pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (__Pyx_PrintOne(0, __pyx_t_9) < 0) __PYX_ERR(0, 1147, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1115
 *         return self.bitmasks.count_total(ifile)
 * 
 *     def check(self):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t mi1
 *         cdef ewah_bool_array arr_totref, arr_tottwo
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.check", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1149
 *         print("Total of %s / %s collisions (% 3.5f%%)" % (nc, nm, 100.0*float(nc)/nm))
 * 
 *     def primary_indices(self):             # <<<<<<<<<<<<<<
 *         mi = (<ewah_bool_array*> self.collisions.ewah_keys)[0].toArray()
 *         return np.array(mi,'uint64')
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_51primary_indices(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_51primary_indices(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("primary_indices (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_50primary_indices(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_50primary_indices(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  std::vector<size_t>  __pyx_v_mi;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("primary_indices", 0);

  /* "yt/geometry/particle_oct_container.pyx":1150
 * 
 *     def primary_indices(self):
 *         mi = (<ewah_bool_array*> self.collisions.ewah_keys)[0].toArray()             # <<<<<<<<<<<<<<
 *         return np.array(mi,'uint64')
 * 
 */
  __pyx_v_mi = (((__pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array *)__pyx_v_self->collisions->ewah_keys)[0]).toArray();

  /* "yt/geometry/particle_oct_container.pyx":1151
 *     def primary_indices(self):
 *         mi = (<ewah_bool_array*> self.collisions.ewah_keys)[0].toArray()
 *         return np.array(mi,'uint64')             # <<<<<<<<<<<<<<
 * 
 *     def file_ownership_mask(self, fid):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_convert_vector_to_py_size_t(__pyx_v_mi); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_n_s_uint64};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1151, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_n_s_uint64};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1151, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_uint64);
    __Pyx_GIVEREF(__pyx_n_s_uint64);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_n_s_uint64);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1149
 *         print("Total of %s / %s collisions (% 3.5f%%)" % (nc, nm, 100.0*float(nc)/nm))
 * 
 *     def primary_indices(self):             # <<<<<<<<<<<<<<
 *         mi = (<ewah_bool_array*> self.collisions.ewah_keys)[0].toArray()
 *         return np.array(mi,'uint64')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.primary_indices", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1153
 *         return np.array(mi,'uint64')
 * 
 *     def file_ownership_mask(self, fid):             # <<<<<<<<<<<<<<
 *         cdef BoolArrayCollection out
 *         out = self.bitmasks._get_bitmask(<np.uint32_t> fid)
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_53file_ownership_mask(PyObject *__pyx_v_self, PyObject *__pyx_v_fid); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_53file_ownership_mask(PyObject *__pyx_v_self, PyObject *__pyx_v_fid) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("file_ownership_mask (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_52file_ownership_mask(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_fid));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_52file_ownership_mask(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_fid) {
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_out = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_uint32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("file_ownership_mask", 0);

  /* "yt/geometry/particle_oct_container.pyx":1155
 *     def file_ownership_mask(self, fid):
 *         cdef BoolArrayCollection out
 *         out = self.bitmasks._get_bitmask(<np.uint32_t> fid)             # <<<<<<<<<<<<<<
 *         return out
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_As_npy_uint32(__pyx_v_fid); if (unlikely((__pyx_t_1 == ((npy_uint32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1155, __pyx_L1_error)
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmasks->__pyx_vtab)->_get_bitmask(__pyx_v_self->bitmasks, ((__pyx_t_5numpy_uint32_t)__pyx_t_1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_out = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1156
 *         cdef BoolArrayCollection out
 *         out = self.bitmasks._get_bitmask(<np.uint32_t> fid)
 *         return out             # <<<<<<<<<<<<<<
 * 
 *     def finalize(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_out));
  __pyx_r = ((PyObject *)__pyx_v_out);
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1153
 *         return np.array(mi,'uint64')
 * 
 *     def file_ownership_mask(self, fid):             # <<<<<<<<<<<<<<
 *         cdef BoolArrayCollection out
 *         out = self.bitmasks._get_bitmask(<np.uint32_t> fid)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.file_ownership_mask", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_out);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1158
 *         return out
 * 
 *     def finalize(self):             # <<<<<<<<<<<<<<
 *         return
 *         # self.index_octree = ParticleOctreeContainer([1,1,1],
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_55finalize(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_55finalize(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("finalize (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_54finalize(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_54finalize(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("finalize", 0);

  /* "yt/geometry/particle_oct_container.pyx":1159
 * 
 *     def finalize(self):
 *         return             # <<<<<<<<<<<<<<
 *         # self.index_octree = ParticleOctreeContainer([1,1,1],
 *         #     [self.left_edge[0], self.left_edge[1], self.left_edge[2]],
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1158
 *         return out
 * 
 *     def finalize(self):             # <<<<<<<<<<<<<<
 *         return
 *         # self.index_octree = ParticleOctreeContainer([1,1,1],
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1172
 *         # self.index_octree.finalize()
 * 
 *     def get_DLE(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.ndarray[np.float64_t, ndim=1] DLE
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_57get_DLE(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_57get_DLE(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_DLE (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_56get_DLE(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_56get_DLE(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  int __pyx_v_i;
  PyArrayObject *__pyx_v_DLE = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_DLE;
  __Pyx_Buffer __pyx_pybuffer_DLE;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyArrayObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_DLE", 0);
  __pyx_pybuffer_DLE.pybuffer.buf = NULL;
  __pyx_pybuffer_DLE.refcount = 0;
  __pyx_pybuffernd_DLE.data = NULL;
  __pyx_pybuffernd_DLE.rcbuffer = &__pyx_pybuffer_DLE;

  /* "yt/geometry/particle_oct_container.pyx":1175
 *         cdef int i
 *         cdef np.ndarray[np.float64_t, ndim=1] DLE
 *         DLE = np.zeros(3, dtype='float64')             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             DLE[i] = self.left_edge[i]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_float64) < 0) __PYX_ERR(0, 1175, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__20, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1175, __pyx_L1_error)
  __pyx_t_4 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_DLE.rcbuffer->pybuffer);
    __pyx_t_5 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_DLE.rcbuffer->pybuffer, (PyObject*)__pyx_t_4, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_5 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_DLE.rcbuffer->pybuffer, (PyObject*)__pyx_v_DLE, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      }
      __pyx_t_6 = __pyx_t_7 = __pyx_t_8 = 0;
    }
    __pyx_pybuffernd_DLE.diminfo[0].strides = __pyx_pybuffernd_DLE.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_DLE.diminfo[0].shape = __pyx_pybuffernd_DLE.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1175, __pyx_L1_error)
  }
  __pyx_t_4 = 0;
  __pyx_v_DLE = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1176
 *         cdef np.ndarray[np.float64_t, ndim=1] DLE
 *         DLE = np.zeros(3, dtype='float64')
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             DLE[i] = self.left_edge[i]
 *         return DLE
 */
  for (__pyx_t_5 = 0; __pyx_t_5 < 3; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "yt/geometry/particle_oct_container.pyx":1177
 *         DLE = np.zeros(3, dtype='float64')
 *         for i in range(3):
 *             DLE[i] = self.left_edge[i]             # <<<<<<<<<<<<<<
 *         return DLE
 *     def get_DRE(self):
 */
    __pyx_t_9 = __pyx_v_i;
    __pyx_t_10 = -1;
    if (__pyx_t_9 < 0) {
      __pyx_t_9 += __pyx_pybuffernd_DLE.diminfo[0].shape;
      if (unlikely(__pyx_t_9 < 0)) __pyx_t_10 = 0;
    } else if (unlikely(__pyx_t_9 >= __pyx_pybuffernd_DLE.diminfo[0].shape)) __pyx_t_10 = 0;
    if (unlikely(__pyx_t_10 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_10);
      __PYX_ERR(0, 1177, __pyx_L1_error)
    }
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_DLE.rcbuffer->pybuffer.buf, __pyx_t_9, __pyx_pybuffernd_DLE.diminfo[0].strides) = (__pyx_v_self->left_edge[__pyx_v_i]);
  }

  /* "yt/geometry/particle_oct_container.pyx":1178
 *         for i in range(3):
 *             DLE[i] = self.left_edge[i]
 *         return DLE             # <<<<<<<<<<<<<<
 *     def get_DRE(self):
 *         cdef int i
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_DLE));
  __pyx_r = ((PyObject *)__pyx_v_DLE);
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1172
 *         # self.index_octree.finalize()
 * 
 *     def get_DLE(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.ndarray[np.float64_t, ndim=1] DLE
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_DLE.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.get_DLE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_DLE.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_DLE);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1179
 *             DLE[i] = self.left_edge[i]
 *         return DLE
 *     def get_DRE(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.ndarray[np.float64_t, ndim=1] DRE
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_59get_DRE(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_59get_DRE(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_DRE (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_58get_DRE(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_58get_DRE(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  int __pyx_v_i;
  PyArrayObject *__pyx_v_DRE = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_DRE;
  __Pyx_Buffer __pyx_pybuffer_DRE;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyArrayObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_DRE", 0);
  __pyx_pybuffer_DRE.pybuffer.buf = NULL;
  __pyx_pybuffer_DRE.refcount = 0;
  __pyx_pybuffernd_DRE.data = NULL;
  __pyx_pybuffernd_DRE.rcbuffer = &__pyx_pybuffer_DRE;

  /* "yt/geometry/particle_oct_container.pyx":1182
 *         cdef int i
 *         cdef np.ndarray[np.float64_t, ndim=1] DRE
 *         DRE = np.zeros(3, dtype='float64')             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             DRE[i] = self.right_edge[i]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_float64) < 0) __PYX_ERR(0, 1182, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__20, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1182, __pyx_L1_error)
  __pyx_t_4 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_DRE.rcbuffer->pybuffer);
    __pyx_t_5 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_DRE.rcbuffer->pybuffer, (PyObject*)__pyx_t_4, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_5 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_DRE.rcbuffer->pybuffer, (PyObject*)__pyx_v_DRE, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      }
      __pyx_t_6 = __pyx_t_7 = __pyx_t_8 = 0;
    }
    __pyx_pybuffernd_DRE.diminfo[0].strides = __pyx_pybuffernd_DRE.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_DRE.diminfo[0].shape = __pyx_pybuffernd_DRE.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1182, __pyx_L1_error)
  }
  __pyx_t_4 = 0;
  __pyx_v_DRE = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1183
 *         cdef np.ndarray[np.float64_t, ndim=1] DRE
 *         DRE = np.zeros(3, dtype='float64')
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             DRE[i] = self.right_edge[i]
 *         return DRE
 */
  for (__pyx_t_5 = 0; __pyx_t_5 < 3; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "yt/geometry/particle_oct_container.pyx":1184
 *         DRE = np.zeros(3, dtype='float64')
 *         for i in range(3):
 *             DRE[i] = self.right_edge[i]             # <<<<<<<<<<<<<<
 *         return DRE
 * 
 */
    __pyx_t_9 = __pyx_v_i;
    __pyx_t_10 = -1;
    if (__pyx_t_9 < 0) {
      __pyx_t_9 += __pyx_pybuffernd_DRE.diminfo[0].shape;
      if (unlikely(__pyx_t_9 < 0)) __pyx_t_10 = 0;
    } else if (unlikely(__pyx_t_9 >= __pyx_pybuffernd_DRE.diminfo[0].shape)) __pyx_t_10 = 0;
    if (unlikely(__pyx_t_10 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_10);
      __PYX_ERR(0, 1184, __pyx_L1_error)
    }
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_DRE.rcbuffer->pybuffer.buf, __pyx_t_9, __pyx_pybuffernd_DRE.diminfo[0].strides) = (__pyx_v_self->right_edge[__pyx_v_i]);
  }

  /* "yt/geometry/particle_oct_container.pyx":1185
 *         for i in range(3):
 *             DRE[i] = self.right_edge[i]
 *         return DRE             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_DRE));
  __pyx_r = ((PyObject *)__pyx_v_DRE);
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1179
 *             DLE[i] = self.left_edge[i]
 *         return DLE
 *     def get_DRE(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.ndarray[np.float64_t, ndim=1] DRE
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_DRE.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.get_DRE", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_DRE.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_DRE);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1190
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def get_ghost_zones(self, SelectorObject selector, int ngz,             # <<<<<<<<<<<<<<
 *                         BoolArrayCollection dmask = None, bint coarse_ghosts = False):
 *         cdef BoolArrayCollection gmask, gmask2, out
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_61get_ghost_zones(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_61get_ghost_zones(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_selector = 0;
  int __pyx_v_ngz;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_dmask = 0;
  int __pyx_v_coarse_ghosts;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_ghost_zones (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_selector,&__pyx_n_s_ngz,&__pyx_n_s_dmask,&__pyx_n_s_coarse_ghosts,0};
    PyObject* values[4] = {0,0,0,0};

    /* "yt/geometry/particle_oct_container.pyx":1191
 *     @cython.cdivision(True)
 *     def get_ghost_zones(self, SelectorObject selector, int ngz,
 *                         BoolArrayCollection dmask = None, bint coarse_ghosts = False):             # <<<<<<<<<<<<<<
 *         cdef BoolArrayCollection gmask, gmask2, out
 *         cdef np.ndarray[np.uint8_t, ndim=1] periodic = selector.get_periodicity()
 */
    values[2] = (PyObject *)((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_selector)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ngz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_ghost_zones", 0, 2, 4, 1); __PYX_ERR(0, 1190, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dmask);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coarse_ghosts);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_ghost_zones") < 0)) __PYX_ERR(0, 1190, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_selector = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)values[0]);
    __pyx_v_ngz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_ngz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1190, __pyx_L3_error)
    __pyx_v_dmask = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)values[2]);
    if (values[3]) {
      __pyx_v_coarse_ghosts = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_coarse_ghosts == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1191, __pyx_L3_error)
    } else {
      __pyx_v_coarse_ghosts = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_ghost_zones", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1190, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.get_ghost_zones", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_selector), __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject, 1, "selector", 0))) __PYX_ERR(0, 1190, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dmask), __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection, 1, "dmask", 0))) __PYX_ERR(0, 1191, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_60get_ghost_zones(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), __pyx_v_selector, __pyx_v_ngz, __pyx_v_dmask, __pyx_v_coarse_ghosts);

  /* "yt/geometry/particle_oct_container.pyx":1190
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def get_ghost_zones(self, SelectorObject selector, int ngz,             # <<<<<<<<<<<<<<
 *                         BoolArrayCollection dmask = None, bint coarse_ghosts = False):
 *         cdef BoolArrayCollection gmask, gmask2, out
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_60get_ghost_zones(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_selector, int __pyx_v_ngz, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_dmask, int __pyx_v_coarse_ghosts) {
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_gmask = 0;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_gmask2 = 0;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_out = 0;
  PyArrayObject *__pyx_v_periodic = 0;
  int __pyx_v_periodicity[3];
  int __pyx_v_i;
  struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_morton_selector = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_dfiles = NULL;
  PyObject *__pyx_v_gfiles = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_periodic;
  __Pyx_Buffer __pyx_pybuffer_periodic;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyArrayObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_19BoolArrayCollection__get_ghost_zones __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_ghost_zones", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_dmask);
  __pyx_pybuffer_periodic.pybuffer.buf = NULL;
  __pyx_pybuffer_periodic.refcount = 0;
  __pyx_pybuffernd_periodic.data = NULL;
  __pyx_pybuffernd_periodic.rcbuffer = &__pyx_pybuffer_periodic;

  /* "yt/geometry/particle_oct_container.pyx":1193
 *                         BoolArrayCollection dmask = None, bint coarse_ghosts = False):
 *         cdef BoolArrayCollection gmask, gmask2, out
 *         cdef np.ndarray[np.uint8_t, ndim=1] periodic = selector.get_periodicity()             # <<<<<<<<<<<<<<
 *         cdef bint periodicity[3]
 *         cdef int i
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_selector), __pyx_n_s_get_periodicity); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1193, __pyx_L1_error)
  __pyx_t_4 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_periodic.rcbuffer->pybuffer, (PyObject*)__pyx_t_4, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_periodic = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_periodic.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 1193, __pyx_L1_error)
    } else {__pyx_pybuffernd_periodic.diminfo[0].strides = __pyx_pybuffernd_periodic.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_periodic.diminfo[0].shape = __pyx_pybuffernd_periodic.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_4 = 0;
  __pyx_v_periodic = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1196
 *         cdef bint periodicity[3]
 *         cdef int i
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             periodicity[i] = periodic[i]
 *         if dmask is None:
 */
  for (__pyx_t_5 = 0; __pyx_t_5 < 3; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "yt/geometry/particle_oct_container.pyx":1197
 *         cdef int i
 *         for i in range(3):
 *             periodicity[i] = periodic[i]             # <<<<<<<<<<<<<<
 *         if dmask is None:
 *             dmask = BoolArrayCollection()
 */
    __pyx_t_6 = __pyx_v_i;
    (__pyx_v_periodicity[__pyx_v_i]) = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_periodic.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_periodic.diminfo[0].strides));
  }

  /* "yt/geometry/particle_oct_container.pyx":1198
 *         for i in range(3):
 *             periodicity[i] = periodic[i]
 *         if dmask is None:             # <<<<<<<<<<<<<<
 *             dmask = BoolArrayCollection()
 *             gmask2 = BoolArrayCollection()
 */
  __pyx_t_7 = (((PyObject *)__pyx_v_dmask) == Py_None);
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "yt/geometry/particle_oct_container.pyx":1199
 *             periodicity[i] = periodic[i]
 *         if dmask is None:
 *             dmask = BoolArrayCollection()             # <<<<<<<<<<<<<<
 *             gmask2 = BoolArrayCollection()
 *             morton_selector = ParticleBitmapSelector(selector,self,ngz=0)
 */
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1199, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_dmask, ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "yt/geometry/particle_oct_container.pyx":1200
 *         if dmask is None:
 *             dmask = BoolArrayCollection()
 *             gmask2 = BoolArrayCollection()             # <<<<<<<<<<<<<<
 *             morton_selector = ParticleBitmapSelector(selector,self,ngz=0)
 *             morton_selector.fill_masks(dmask, gmask2)
 */
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1200, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_gmask2 = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "yt/geometry/particle_oct_container.pyx":1201
 *             dmask = BoolArrayCollection()
 *             gmask2 = BoolArrayCollection()
 *             morton_selector = ParticleBitmapSelector(selector,self,ngz=0)             # <<<<<<<<<<<<<<
 *             morton_selector.fill_masks(dmask, gmask2)
 *         gmask = BoolArrayCollection()
 */
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)__pyx_v_selector));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_selector));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_selector));
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self));
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_ngz, __pyx_int_0) < 0) __PYX_ERR(0, 1201, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector), __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_morton_selector = ((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "yt/geometry/particle_oct_container.pyx":1202
 *             gmask2 = BoolArrayCollection()
 *             morton_selector = ParticleBitmapSelector(selector,self,ngz=0)
 *             morton_selector.fill_masks(dmask, gmask2)             # <<<<<<<<<<<<<<
 *         gmask = BoolArrayCollection()
 *         dmask._get_ghost_zones(ngz, self.index_order1, self.index_order2,
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_morton_selector), __pyx_n_s_fill_masks); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, ((PyObject *)__pyx_v_dmask), ((PyObject *)__pyx_v_gmask2)};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1202, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_1, ((PyObject *)__pyx_v_dmask), ((PyObject *)__pyx_v_gmask2)};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1202, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_1) {
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_1); __pyx_t_1 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_dmask));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_dmask));
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_5, ((PyObject *)__pyx_v_dmask));
      __Pyx_INCREF(((PyObject *)__pyx_v_gmask2));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_gmask2));
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_5, ((PyObject *)__pyx_v_gmask2));
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "yt/geometry/particle_oct_container.pyx":1198
 *         for i in range(3):
 *             periodicity[i] = periodic[i]
 *         if dmask is None:             # <<<<<<<<<<<<<<
 *             dmask = BoolArrayCollection()
 *             gmask2 = BoolArrayCollection()
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":1203
 *             morton_selector = ParticleBitmapSelector(selector,self,ngz=0)
 *             morton_selector.fill_masks(dmask, gmask2)
 *         gmask = BoolArrayCollection()             # <<<<<<<<<<<<<<
 *         dmask._get_ghost_zones(ngz, self.index_order1, self.index_order2,
 *                                periodicity, gmask, <bint>coarse_ghosts)
 */
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_gmask = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1204
 *             morton_selector.fill_masks(dmask, gmask2)
 *         gmask = BoolArrayCollection()
 *         dmask._get_ghost_zones(ngz, self.index_order1, self.index_order2,             # <<<<<<<<<<<<<<
 *                                periodicity, gmask, <bint>coarse_ghosts)
 *         dfiles, gfiles = self.masks_to_files(dmask, gmask)
 */
  __pyx_t_10.__pyx_n = 1;
  __pyx_t_10.coarse_ghosts = (__pyx_v_coarse_ghosts != 0);
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_v_dmask->__pyx_vtab)->_get_ghost_zones(__pyx_v_dmask, __pyx_v_ngz, __pyx_v_self->index_order1, __pyx_v_self->index_order2, __pyx_v_periodicity, __pyx_v_gmask, &__pyx_t_10); 

  /* "yt/geometry/particle_oct_container.pyx":1206
 *         dmask._get_ghost_zones(ngz, self.index_order1, self.index_order2,
 *                                periodicity, gmask, <bint>coarse_ghosts)
 *         dfiles, gfiles = self.masks_to_files(dmask, gmask)             # <<<<<<<<<<<<<<
 *         out = BoolArrayCollection()
 *         gmask._logicalor(dmask, out)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_masks_to_files); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_dmask), ((PyObject *)__pyx_v_gmask)};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1206, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_9, ((PyObject *)__pyx_v_dmask), ((PyObject *)__pyx_v_gmask)};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1206, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  {
    __pyx_t_1 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__pyx_t_9) {
      __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_9); __pyx_t_9 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_dmask));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_dmask));
    PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_5, ((PyObject *)__pyx_v_dmask));
    __Pyx_INCREF(((PyObject *)__pyx_v_gmask));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_gmask));
    PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_5, ((PyObject *)__pyx_v_gmask));
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 1206, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_1);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_9 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_11 = Py_TYPE(__pyx_t_9)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_11(__pyx_t_9); if (unlikely(!__pyx_t_2)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_1 = __pyx_t_11(__pyx_t_9); if (unlikely(!__pyx_t_1)) goto __pyx_L6_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_9), 2) < 0) __PYX_ERR(0, 1206, __pyx_L1_error)
    __pyx_t_11 = NULL;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    goto __pyx_L7_unpacking_done;
    __pyx_L6_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_11 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 1206, __pyx_L1_error)
    __pyx_L7_unpacking_done:;
  }
  __pyx_v_dfiles = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_gfiles = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1207
 *                                periodicity, gmask, <bint>coarse_ghosts)
 *         dfiles, gfiles = self.masks_to_files(dmask, gmask)
 *         out = BoolArrayCollection()             # <<<<<<<<<<<<<<
 *         gmask._logicalor(dmask, out)
 *         return gfiles, out
 */
  __pyx_t_3 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_out = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1208
 *         dfiles, gfiles = self.masks_to_files(dmask, gmask)
 *         out = BoolArrayCollection()
 *         gmask._logicalor(dmask, out)             # <<<<<<<<<<<<<<
 *         return gfiles, out
 * 
 */
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_v_gmask->__pyx_vtab)->_logicalor(__pyx_v_gmask, __pyx_v_dmask, __pyx_v_out);

  /* "yt/geometry/particle_oct_container.pyx":1209
 *         out = BoolArrayCollection()
 *         gmask._logicalor(dmask, out)
 *         return gfiles, out             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_gfiles);
  __Pyx_GIVEREF(__pyx_v_gfiles);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_gfiles);
  __Pyx_INCREF(((PyObject *)__pyx_v_out));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_out));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_out));
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1190
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def get_ghost_zones(self, SelectorObject selector, int ngz,             # <<<<<<<<<<<<<<
 *                         BoolArrayCollection dmask = None, bint coarse_ghosts = False):
 *         cdef BoolArrayCollection gmask, gmask2, out
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_periodic.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.get_ghost_zones", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_periodic.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_gmask);
  __Pyx_XDECREF((PyObject *)__pyx_v_gmask2);
  __Pyx_XDECREF((PyObject *)__pyx_v_out);
  __Pyx_XDECREF((PyObject *)__pyx_v_periodic);
  __Pyx_XDECREF((PyObject *)__pyx_v_morton_selector);
  __Pyx_XDECREF(__pyx_v_dfiles);
  __Pyx_XDECREF(__pyx_v_gfiles);
  __Pyx_XDECREF((PyObject *)__pyx_v_dmask);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1214
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def selector2mask(self, SelectorObject selector):             # <<<<<<<<<<<<<<
 *         cdef BoolArrayCollection cmask = BoolArrayCollection()
 *         cdef ParticleBitmapSelector morton_selector
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_63selector2mask(PyObject *__pyx_v_self, PyObject *__pyx_v_selector); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_63selector2mask(PyObject *__pyx_v_self, PyObject *__pyx_v_selector) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("selector2mask (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_selector), __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject, 1, "selector", 0))) __PYX_ERR(0, 1214, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_62selector2mask(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_selector));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_62selector2mask(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_selector) {
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_cmask = 0;
  struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_morton_selector = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("selector2mask", 0);

  /* "yt/geometry/particle_oct_container.pyx":1215
 *     @cython.cdivision(True)
 *     def selector2mask(self, SelectorObject selector):
 *         cdef BoolArrayCollection cmask = BoolArrayCollection()             # <<<<<<<<<<<<<<
 *         cdef ParticleBitmapSelector morton_selector
 *         morton_selector = ParticleBitmapSelector(selector,self,ngz=0)
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_cmask = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1217
 *         cdef BoolArrayCollection cmask = BoolArrayCollection()
 *         cdef ParticleBitmapSelector morton_selector
 *         morton_selector = ParticleBitmapSelector(selector,self,ngz=0)             # <<<<<<<<<<<<<<
 *         morton_selector.fill_masks(cmask)
 *         return cmask
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_v_selector));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_selector));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_selector));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_ngz, __pyx_int_0) < 0) __PYX_ERR(0, 1217, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector), __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_morton_selector = ((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1218
 *         cdef ParticleBitmapSelector morton_selector
 *         morton_selector = ParticleBitmapSelector(selector,self,ngz=0)
 *         morton_selector.fill_masks(cmask)             # <<<<<<<<<<<<<<
 *         return cmask
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_morton_selector), __pyx_n_s_fill_masks); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, ((PyObject *)__pyx_v_cmask)) : __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_cmask));
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1219
 *         morton_selector = ParticleBitmapSelector(selector,self,ngz=0)
 *         morton_selector.fill_masks(cmask)
 *         return cmask             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_cmask));
  __pyx_r = ((PyObject *)__pyx_v_cmask);
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1214
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def selector2mask(self, SelectorObject selector):             # <<<<<<<<<<<<<<
 *         cdef BoolArrayCollection cmask = BoolArrayCollection()
 *         cdef ParticleBitmapSelector morton_selector
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.selector2mask", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cmask);
  __Pyx_XDECREF((PyObject *)__pyx_v_morton_selector);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1224
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def mask2files(self, BoolArrayCollection cmask):             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint32_t, ndim=1] file_idx
 *         file_idx = self.mask_to_files(cmask)
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_65mask2files(PyObject *__pyx_v_self, PyObject *__pyx_v_cmask); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_65mask2files(PyObject *__pyx_v_self, PyObject *__pyx_v_cmask) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mask2files (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_cmask), __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection, 1, "cmask", 0))) __PYX_ERR(0, 1224, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_64mask2files(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_v_cmask));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_64mask2files(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_cmask) {
  PyArrayObject *__pyx_v_file_idx = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_file_idx;
  __Pyx_Buffer __pyx_pybuffer_file_idx;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyArrayObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mask2files", 0);
  __pyx_pybuffer_file_idx.pybuffer.buf = NULL;
  __pyx_pybuffer_file_idx.refcount = 0;
  __pyx_pybuffernd_file_idx.data = NULL;
  __pyx_pybuffernd_file_idx.rcbuffer = &__pyx_pybuffer_file_idx;

  /* "yt/geometry/particle_oct_container.pyx":1226
 *     def mask2files(self, BoolArrayCollection cmask):
 *         cdef np.ndarray[np.uint32_t, ndim=1] file_idx
 *         file_idx = self.mask_to_files(cmask)             # <<<<<<<<<<<<<<
 *         return file_idx
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_mask_to_files); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, ((PyObject *)__pyx_v_cmask)) : __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_cmask));
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1226, __pyx_L1_error)
  __pyx_t_4 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_idx.rcbuffer->pybuffer);
    __pyx_t_5 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_idx.rcbuffer->pybuffer, (PyObject*)__pyx_t_4, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_5 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_idx.rcbuffer->pybuffer, (PyObject*)__pyx_v_file_idx, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      }
      __pyx_t_6 = __pyx_t_7 = __pyx_t_8 = 0;
    }
    __pyx_pybuffernd_file_idx.diminfo[0].strides = __pyx_pybuffernd_file_idx.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_file_idx.diminfo[0].shape = __pyx_pybuffernd_file_idx.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1226, __pyx_L1_error)
  }
  __pyx_t_4 = 0;
  __pyx_v_file_idx = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1227
 *         cdef np.ndarray[np.uint32_t, ndim=1] file_idx
 *         file_idx = self.mask_to_files(cmask)
 *         return file_idx             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_file_idx));
  __pyx_r = ((PyObject *)__pyx_v_file_idx);
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1224
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def mask2files(self, BoolArrayCollection cmask):             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint32_t, ndim=1] file_idx
 *         file_idx = self.mask_to_files(cmask)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_idx.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.mask2files", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_idx.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_file_idx);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1232
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def mask2filemasks(self, BoolArrayCollection cmask, np.ndarray[np.uint32_t, ndim=1] file_idx):             # <<<<<<<<<<<<<<
 *         cdef BoolArrayCollection fmask
 *         cdef np.int32_t fid
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_67mask2filemasks(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_67mask2filemasks(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_cmask = 0;
  PyArrayObject *__pyx_v_file_idx = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mask2filemasks (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cmask,&__pyx_n_s_file_idx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cmask)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_file_idx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("mask2filemasks", 1, 2, 2, 1); __PYX_ERR(0, 1232, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "mask2filemasks") < 0)) __PYX_ERR(0, 1232, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_cmask = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)values[0]);
    __pyx_v_file_idx = ((PyArrayObject *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("mask2filemasks", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1232, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.mask2filemasks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_cmask), __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection, 1, "cmask", 0))) __PYX_ERR(0, 1232, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_file_idx), __pyx_ptype_5numpy_ndarray, 1, "file_idx", 0))) __PYX_ERR(0, 1232, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_66mask2filemasks(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), __pyx_v_cmask, __pyx_v_file_idx);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_66mask2filemasks(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_cmask, PyArrayObject *__pyx_v_file_idx) {
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_fmask = 0;
  __pyx_t_5numpy_int32_t __pyx_v_fid;
  PyArrayObject *__pyx_v_file_masks = 0;
  CYTHON_UNUSED int __pyx_v_i;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_file_idx;
  __Pyx_Buffer __pyx_pybuffer_file_idx;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_file_masks;
  __Pyx_Buffer __pyx_pybuffer_file_masks;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyArrayObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *(*__pyx_t_12)(PyObject *);
  PyObject *__pyx_t_13 = NULL;
  PyObject *(*__pyx_t_14)(PyObject *);
  __pyx_t_5numpy_int32_t __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mask2filemasks", 0);
  __pyx_pybuffer_file_masks.pybuffer.buf = NULL;
  __pyx_pybuffer_file_masks.refcount = 0;
  __pyx_pybuffernd_file_masks.data = NULL;
  __pyx_pybuffernd_file_masks.rcbuffer = &__pyx_pybuffer_file_masks;
  __pyx_pybuffer_file_idx.pybuffer.buf = NULL;
  __pyx_pybuffer_file_idx.refcount = 0;
  __pyx_pybuffernd_file_idx.data = NULL;
  __pyx_pybuffernd_file_idx.rcbuffer = &__pyx_pybuffer_file_idx;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_idx.rcbuffer->pybuffer, (PyObject*)__pyx_v_file_idx, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1232, __pyx_L1_error)
  }
  __pyx_pybuffernd_file_idx.diminfo[0].strides = __pyx_pybuffernd_file_idx.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_file_idx.diminfo[0].shape = __pyx_pybuffernd_file_idx.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/particle_oct_container.pyx":1238
 *         cdef int i
 *         # Get bitmasks for parts of files touching the selector
 *         file_masks = np.array([BoolArrayCollection() for i in range(len(file_idx))],             # <<<<<<<<<<<<<<
 *                               dtype="object")
 *         for i, (fid, fmask) in enumerate(zip(file_idx,file_masks)):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_Length(((PyObject *)__pyx_v_file_idx)); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1238, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;
    __pyx_t_6 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1238, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 1238, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1239
 *         # Get bitmasks for parts of files touching the selector
 *         file_masks = np.array([BoolArrayCollection() for i in range(len(file_idx))],
 *                               dtype="object")             # <<<<<<<<<<<<<<
 *         for i, (fid, fmask) in enumerate(zip(file_idx,file_masks)):
 *             self.bitmasks._logicaland(<np.uint32_t> fid, cmask, fmask)
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_object) < 0) __PYX_ERR(0, 1239, __pyx_L1_error)

  /* "yt/geometry/particle_oct_container.pyx":1238
 *         cdef int i
 *         # Get bitmasks for parts of files touching the selector
 *         file_masks = np.array([BoolArrayCollection() for i in range(len(file_idx))],             # <<<<<<<<<<<<<<
 *                               dtype="object")
 *         for i, (fid, fmask) in enumerate(zip(file_idx,file_masks)):
 */
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1238, __pyx_L1_error)
  __pyx_t_8 = ((PyArrayObject *)__pyx_t_7);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_masks.rcbuffer->pybuffer);
    __pyx_t_5 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_masks.rcbuffer->pybuffer, (PyObject*)__pyx_t_8, &__Pyx_TypeInfo_object, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_5 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_masks.rcbuffer->pybuffer, (PyObject*)__pyx_v_file_masks, &__Pyx_TypeInfo_object, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
      __pyx_t_9 = __pyx_t_10 = __pyx_t_11 = 0;
    }
    __pyx_pybuffernd_file_masks.diminfo[0].strides = __pyx_pybuffernd_file_masks.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_file_masks.diminfo[0].shape = __pyx_pybuffernd_file_masks.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1238, __pyx_L1_error)
  }
  __pyx_t_8 = 0;
  __pyx_v_file_masks = ((PyArrayObject *)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1240
 *         file_masks = np.array([BoolArrayCollection() for i in range(len(file_idx))],
 *                               dtype="object")
 *         for i, (fid, fmask) in enumerate(zip(file_idx,file_masks)):             # <<<<<<<<<<<<<<
 *             self.bitmasks._logicaland(<np.uint32_t> fid, cmask, fmask)
 *         return file_masks
 */
  __pyx_t_5 = 0;
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(((PyObject *)__pyx_v_file_idx));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_file_idx));
  PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)__pyx_v_file_idx));
  __Pyx_INCREF(((PyObject *)__pyx_v_file_masks));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_file_masks));
  PyTuple_SET_ITEM(__pyx_t_7, 1, ((PyObject *)__pyx_v_file_masks));
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_7 = __pyx_t_1; __Pyx_INCREF(__pyx_t_7); __pyx_t_3 = 0;
    __pyx_t_12 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1240, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_12 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1240, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_12)) {
      if (likely(PyList_CheckExact(__pyx_t_7))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 1240, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_7, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1240, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 1240, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_7, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1240, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_12(__pyx_t_7);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1240, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1240, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_2);
      #else
      __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1240, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1240, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_13 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1240, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_14 = Py_TYPE(__pyx_t_13)->tp_iternext;
      index = 0; __pyx_t_6 = __pyx_t_14(__pyx_t_13); if (unlikely(!__pyx_t_6)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      index = 1; __pyx_t_2 = __pyx_t_14(__pyx_t_13); if (unlikely(!__pyx_t_2)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_13), 2) < 0) __PYX_ERR(0, 1240, __pyx_L1_error)
      __pyx_t_14 = NULL;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_14 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1240, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }
    __pyx_t_15 = __Pyx_PyInt_As_npy_int32(__pyx_t_6); if (unlikely((__pyx_t_15 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1240, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection))))) __PYX_ERR(0, 1240, __pyx_L1_error)
    __pyx_v_fid = __pyx_t_15;
    __Pyx_XDECREF_SET(__pyx_v_fmask, ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_2));
    __pyx_t_2 = 0;
    __pyx_v_i = __pyx_t_5;
    __pyx_t_5 = (__pyx_t_5 + 1);

    /* "yt/geometry/particle_oct_container.pyx":1241
 *                               dtype="object")
 *         for i, (fid, fmask) in enumerate(zip(file_idx,file_masks)):
 *             self.bitmasks._logicaland(<np.uint32_t> fid, cmask, fmask)             # <<<<<<<<<<<<<<
 *         return file_masks
 * 
 */
    ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmasks->__pyx_vtab)->_logicaland(__pyx_v_self->bitmasks, ((__pyx_t_5numpy_uint32_t)__pyx_v_fid), __pyx_v_cmask, __pyx_v_fmask);

    /* "yt/geometry/particle_oct_container.pyx":1240
 *         file_masks = np.array([BoolArrayCollection() for i in range(len(file_idx))],
 *                               dtype="object")
 *         for i, (fid, fmask) in enumerate(zip(file_idx,file_masks)):             # <<<<<<<<<<<<<<
 *             self.bitmasks._logicaland(<np.uint32_t> fid, cmask, fmask)
 *         return file_masks
 */
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1242
 *         for i, (fid, fmask) in enumerate(zip(file_idx,file_masks)):
 *             self.bitmasks._logicaland(<np.uint32_t> fid, cmask, fmask)
 *         return file_masks             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_file_masks));
  __pyx_r = ((PyObject *)__pyx_v_file_masks);
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1232
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def mask2filemasks(self, BoolArrayCollection cmask, np.ndarray[np.uint32_t, ndim=1] file_idx):             # <<<<<<<<<<<<<<
 *         cdef BoolArrayCollection fmask
 *         cdef np.int32_t fid
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_13);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_idx.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_masks.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.mask2filemasks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_idx.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_masks.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_fmask);
  __Pyx_XDECREF((PyObject *)__pyx_v_file_masks);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1247
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def filemasks2addfiles(self, np.ndarray[object, ndim=1] file_masks):             # <<<<<<<<<<<<<<
 *         cdef list addfile_idx
 *         addfile_idx = len(file_masks)*[None]
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_69filemasks2addfiles(PyObject *__pyx_v_self, PyObject *__pyx_v_file_masks); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_69filemasks2addfiles(PyObject *__pyx_v_self, PyObject *__pyx_v_file_masks) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("filemasks2addfiles (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_file_masks), __pyx_ptype_5numpy_ndarray, 1, "file_masks", 0))) __PYX_ERR(0, 1247, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_68filemasks2addfiles(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyArrayObject *)__pyx_v_file_masks));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_68filemasks2addfiles(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyArrayObject *__pyx_v_file_masks) {
  PyObject *__pyx_v_addfile_idx = 0;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_fmask = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_file_masks;
  __Pyx_Buffer __pyx_pybuffer_file_masks;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("filemasks2addfiles", 0);
  __pyx_pybuffer_file_masks.pybuffer.buf = NULL;
  __pyx_pybuffer_file_masks.refcount = 0;
  __pyx_pybuffernd_file_masks.data = NULL;
  __pyx_pybuffernd_file_masks.rcbuffer = &__pyx_pybuffer_file_masks;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_masks.rcbuffer->pybuffer, (PyObject*)__pyx_v_file_masks, &__Pyx_TypeInfo_object, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1247, __pyx_L1_error)
  }
  __pyx_pybuffernd_file_masks.diminfo[0].strides = __pyx_pybuffernd_file_masks.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_file_masks.diminfo[0].shape = __pyx_pybuffernd_file_masks.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/particle_oct_container.pyx":1249
 *     def filemasks2addfiles(self, np.ndarray[object, ndim=1] file_masks):
 *         cdef list addfile_idx
 *         addfile_idx = len(file_masks)*[None]             # <<<<<<<<<<<<<<
 *         for i, fmask in enumerate(file_masks):
 *             addfile_idx[i] = self.mask_to_files(fmask).astype('uint32')
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_file_masks)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1249, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_3, 0, Py_None);
  __pyx_t_4 = PyNumber_Multiply(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PyList_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(0, 1249, __pyx_L1_error)
  __pyx_v_addfile_idx = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1250
 *         cdef list addfile_idx
 *         addfile_idx = len(file_masks)*[None]
 *         for i, fmask in enumerate(file_masks):             # <<<<<<<<<<<<<<
 *             addfile_idx[i] = self.mask_to_files(fmask).astype('uint32')
 *         return addfile_idx
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_4 = __pyx_int_0;
  if (likely(PyList_CheckExact(((PyObject *)__pyx_v_file_masks))) || PyTuple_CheckExact(((PyObject *)__pyx_v_file_masks))) {
    __pyx_t_3 = ((PyObject *)__pyx_v_file_masks); __Pyx_INCREF(__pyx_t_3); __pyx_t_1 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_1 = -1; __pyx_t_3 = PyObject_GetIter(((PyObject *)__pyx_v_file_masks)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1250, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_1 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_2); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 1250, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1250, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_2); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 1250, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1250, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_5(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1250, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_fmask, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_4);
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_4, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4);
    __pyx_t_4 = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "yt/geometry/particle_oct_container.pyx":1251
 *         addfile_idx = len(file_masks)*[None]
 *         for i, fmask in enumerate(file_masks):
 *             addfile_idx[i] = self.mask_to_files(fmask).astype('uint32')             # <<<<<<<<<<<<<<
 *         return addfile_idx
 * 
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_mask_to_files); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_v_fmask) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_fmask);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_astype); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_2 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_n_s_uint32) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_n_s_uint32);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v_addfile_idx == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 1251, __pyx_L1_error)
    }
    if (unlikely(PyObject_SetItem(__pyx_v_addfile_idx, __pyx_v_i, __pyx_t_2) < 0)) __PYX_ERR(0, 1251, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "yt/geometry/particle_oct_container.pyx":1250
 *         cdef list addfile_idx
 *         addfile_idx = len(file_masks)*[None]
 *         for i, fmask in enumerate(file_masks):             # <<<<<<<<<<<<<<
 *             addfile_idx[i] = self.mask_to_files(fmask).astype('uint32')
 *         return addfile_idx
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1252
 *         for i, fmask in enumerate(file_masks):
 *             addfile_idx[i] = self.mask_to_files(fmask).astype('uint32')
 *         return addfile_idx             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_addfile_idx);
  __pyx_r = __pyx_v_addfile_idx;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1247
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def filemasks2addfiles(self, np.ndarray[object, ndim=1] file_masks):             # <<<<<<<<<<<<<<
 *         cdef list addfile_idx
 *         addfile_idx = len(file_masks)*[None]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_masks.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.filemasks2addfiles", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_masks.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF(__pyx_v_addfile_idx);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_fmask);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1257
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def identify_file_masks(self, SelectorObject selector):             # <<<<<<<<<<<<<<
 *         cdef BoolArrayCollection cmask = BoolArrayCollection()
 *         cdef BoolArrayCollection fmask
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_71identify_file_masks(PyObject *__pyx_v_self, PyObject *__pyx_v_selector); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_71identify_file_masks(PyObject *__pyx_v_self, PyObject *__pyx_v_selector) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("identify_file_masks (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_selector), __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject, 1, "selector", 0))) __PYX_ERR(0, 1257, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_70identify_file_masks(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_selector));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_70identify_file_masks(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_selector) {
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_cmask = 0;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_fmask = 0;
  __pyx_t_5numpy_int32_t __pyx_v_fid;
  PyArrayObject *__pyx_v_file_masks = 0;
  PyArrayObject *__pyx_v_file_idx = 0;
  PyObject *__pyx_v_addfile_idx = 0;
  struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_morton_selector = 0;
  PyObject *__pyx_v_i = NULL;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_file_idx;
  __Pyx_Buffer __pyx_pybuffer_file_idx;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_file_masks;
  __Pyx_Buffer __pyx_pybuffer_file_masks;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyArrayObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyArrayObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *(*__pyx_t_15)(PyObject *);
  __pyx_t_5numpy_int32_t __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("identify_file_masks", 0);
  __pyx_pybuffer_file_masks.pybuffer.buf = NULL;
  __pyx_pybuffer_file_masks.refcount = 0;
  __pyx_pybuffernd_file_masks.data = NULL;
  __pyx_pybuffernd_file_masks.rcbuffer = &__pyx_pybuffer_file_masks;
  __pyx_pybuffer_file_idx.pybuffer.buf = NULL;
  __pyx_pybuffer_file_idx.refcount = 0;
  __pyx_pybuffernd_file_idx.data = NULL;
  __pyx_pybuffernd_file_idx.rcbuffer = &__pyx_pybuffer_file_idx;

  /* "yt/geometry/particle_oct_container.pyx":1258
 *     @cython.cdivision(True)
 *     def identify_file_masks(self, SelectorObject selector):
 *         cdef BoolArrayCollection cmask = BoolArrayCollection()             # <<<<<<<<<<<<<<
 *         cdef BoolArrayCollection fmask
 *         cdef np.int32_t fid
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_cmask = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1266
 *         # Get bitmask for selector
 *         cdef ParticleBitmapSelector morton_selector
 *         morton_selector = ParticleBitmapSelector(selector, self, ngz=0)             # <<<<<<<<<<<<<<
 *         morton_selector.fill_masks(cmask)
 *         # Get bitmasks for parts of files touching the selector
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_v_selector));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_selector));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_selector));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_ngz, __pyx_int_0) < 0) __PYX_ERR(0, 1266, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector), __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_morton_selector = ((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1267
 *         cdef ParticleBitmapSelector morton_selector
 *         morton_selector = ParticleBitmapSelector(selector, self, ngz=0)
 *         morton_selector.fill_masks(cmask)             # <<<<<<<<<<<<<<
 *         # Get bitmasks for parts of files touching the selector
 *         file_idx = self.mask_to_files(cmask)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_morton_selector), __pyx_n_s_fill_masks); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, ((PyObject *)__pyx_v_cmask)) : __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_cmask));
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1269
 *         morton_selector.fill_masks(cmask)
 *         # Get bitmasks for parts of files touching the selector
 *         file_idx = self.mask_to_files(cmask)             # <<<<<<<<<<<<<<
 *         file_masks = np.array([BoolArrayCollection() for i in range(len(file_idx))],
 *                               dtype="object")
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_mask_to_files); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, ((PyObject *)__pyx_v_cmask)) : __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_cmask));
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1269, __pyx_L1_error)
  __pyx_t_4 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_idx.rcbuffer->pybuffer);
    __pyx_t_5 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_idx.rcbuffer->pybuffer, (PyObject*)__pyx_t_4, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_5 < 0)) {
      PyErr_Fetch(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_idx.rcbuffer->pybuffer, (PyObject*)__pyx_v_file_idx, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_6); Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      }
      __pyx_t_6 = __pyx_t_7 = __pyx_t_8 = 0;
    }
    __pyx_pybuffernd_file_idx.diminfo[0].strides = __pyx_pybuffernd_file_idx.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_file_idx.diminfo[0].shape = __pyx_pybuffernd_file_idx.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1269, __pyx_L1_error)
  }
  __pyx_t_4 = 0;
  __pyx_v_file_idx = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1270
 *         # Get bitmasks for parts of files touching the selector
 *         file_idx = self.mask_to_files(cmask)
 *         file_masks = np.array([BoolArrayCollection() for i in range(len(file_idx))],             # <<<<<<<<<<<<<<
 *                               dtype="object")
 *         addfile_idx = len(file_idx)*[None]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_9 = PyObject_Length(((PyObject *)__pyx_v_file_idx)); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1270, __pyx_L1_error)
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_10)) || PyTuple_CheckExact(__pyx_t_10)) {
    __pyx_t_1 = __pyx_t_10; __Pyx_INCREF(__pyx_t_1); __pyx_t_9 = 0;
    __pyx_t_11 = NULL;
  } else {
    __pyx_t_9 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1270, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  for (;;) {
    if (likely(!__pyx_t_11)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_10 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_10); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 1270, __pyx_L1_error)
        #else
        __pyx_t_10 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1270, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        #endif
      } else {
        if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_10); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 1270, __pyx_L1_error)
        #else
        __pyx_t_10 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1270, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        #endif
      }
    } else {
      __pyx_t_10 = __pyx_t_11(__pyx_t_1);
      if (unlikely(!__pyx_t_10)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1270, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_10);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_10);
    __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_10))) __PYX_ERR(0, 1270, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1271
 *         file_idx = self.mask_to_files(cmask)
 *         file_masks = np.array([BoolArrayCollection() for i in range(len(file_idx))],
 *                               dtype="object")             # <<<<<<<<<<<<<<
 *         addfile_idx = len(file_idx)*[None]
 *         for i, (fid, fmask) in enumerate(zip(file_idx,file_masks)):
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_n_s_object) < 0) __PYX_ERR(0, 1271, __pyx_L1_error)

  /* "yt/geometry/particle_oct_container.pyx":1270
 *         # Get bitmasks for parts of files touching the selector
 *         file_idx = self.mask_to_files(cmask)
 *         file_masks = np.array([BoolArrayCollection() for i in range(len(file_idx))],             # <<<<<<<<<<<<<<
 *                               dtype="object")
 *         addfile_idx = len(file_idx)*[None]
 */
  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_10) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_10, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1270, __pyx_L1_error)
  __pyx_t_12 = ((PyArrayObject *)__pyx_t_10);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_masks.rcbuffer->pybuffer);
    __pyx_t_5 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_masks.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_object, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_5 < 0)) {
      PyErr_Fetch(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_masks.rcbuffer->pybuffer, (PyObject*)__pyx_v_file_masks, &__Pyx_TypeInfo_object, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_6);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_8, __pyx_t_7, __pyx_t_6);
      }
      __pyx_t_8 = __pyx_t_7 = __pyx_t_6 = 0;
    }
    __pyx_pybuffernd_file_masks.diminfo[0].strides = __pyx_pybuffernd_file_masks.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_file_masks.diminfo[0].shape = __pyx_pybuffernd_file_masks.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1270, __pyx_L1_error)
  }
  __pyx_t_12 = 0;
  __pyx_v_file_masks = ((PyArrayObject *)__pyx_t_10);
  __pyx_t_10 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1272
 *         file_masks = np.array([BoolArrayCollection() for i in range(len(file_idx))],
 *                               dtype="object")
 *         addfile_idx = len(file_idx)*[None]             # <<<<<<<<<<<<<<
 *         for i, (fid, fmask) in enumerate(zip(file_idx,file_masks)):
 *             self.bitmasks._logicaland(<np.uint32_t> fid, cmask, fmask)
 */
  __pyx_t_9 = PyObject_Length(((PyObject *)__pyx_v_file_idx)); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1272, __pyx_L1_error)
  __pyx_t_10 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_3, 0, Py_None);
  __pyx_t_1 = PyNumber_Multiply(__pyx_t_10, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 1272, __pyx_L1_error)
  __pyx_v_addfile_idx = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1273
 *                               dtype="object")
 *         addfile_idx = len(file_idx)*[None]
 *         for i, (fid, fmask) in enumerate(zip(file_idx,file_masks)):             # <<<<<<<<<<<<<<
 *             self.bitmasks._logicaland(<np.uint32_t> fid, cmask, fmask)
 *             addfile_idx[i] = self.mask_to_files(fmask).astype('uint32')
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_1 = __pyx_int_0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_file_idx));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_file_idx));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_file_idx));
  __Pyx_INCREF(((PyObject *)__pyx_v_file_masks));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_file_masks));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_file_masks));
  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_3, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_10)) || PyTuple_CheckExact(__pyx_t_10)) {
    __pyx_t_3 = __pyx_t_10; __Pyx_INCREF(__pyx_t_3); __pyx_t_9 = 0;
    __pyx_t_11 = NULL;
  } else {
    __pyx_t_9 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_11 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1273, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  for (;;) {
    if (likely(!__pyx_t_11)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_10 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_9); __Pyx_INCREF(__pyx_t_10); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 1273, __pyx_L1_error)
        #else
        __pyx_t_10 = PySequence_ITEM(__pyx_t_3, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1273, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        #endif
      } else {
        if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_9); __Pyx_INCREF(__pyx_t_10); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 1273, __pyx_L1_error)
        #else
        __pyx_t_10 = PySequence_ITEM(__pyx_t_3, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1273, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        #endif
      }
    } else {
      __pyx_t_10 = __pyx_t_11(__pyx_t_3);
      if (unlikely(!__pyx_t_10)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1273, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_10);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_10))) || (PyList_CheckExact(__pyx_t_10))) {
      PyObject* sequence = __pyx_t_10;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1273, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_13 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_13 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_13);
      #else
      __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1273, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_13 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1273, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      #endif
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_14 = PyObject_GetIter(__pyx_t_10); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1273, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_15 = Py_TYPE(__pyx_t_14)->tp_iternext;
      index = 0; __pyx_t_2 = __pyx_t_15(__pyx_t_14); if (unlikely(!__pyx_t_2)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      index = 1; __pyx_t_13 = __pyx_t_15(__pyx_t_14); if (unlikely(!__pyx_t_13)) goto __pyx_L7_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_13);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_15(__pyx_t_14), 2) < 0) __PYX_ERR(0, 1273, __pyx_L1_error)
      __pyx_t_15 = NULL;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      goto __pyx_L8_unpacking_done;
      __pyx_L7_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_15 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1273, __pyx_L1_error)
      __pyx_L8_unpacking_done:;
    }
    __pyx_t_16 = __Pyx_PyInt_As_npy_int32(__pyx_t_2); if (unlikely((__pyx_t_16 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1273, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(((__pyx_t_13) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_13, __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection))))) __PYX_ERR(0, 1273, __pyx_L1_error)
    __pyx_v_fid = __pyx_t_16;
    __Pyx_XDECREF_SET(__pyx_v_fmask, ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_13));
    __pyx_t_13 = 0;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_10 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_10;
    __pyx_t_10 = 0;

    /* "yt/geometry/particle_oct_container.pyx":1274
 *         addfile_idx = len(file_idx)*[None]
 *         for i, (fid, fmask) in enumerate(zip(file_idx,file_masks)):
 *             self.bitmasks._logicaland(<np.uint32_t> fid, cmask, fmask)             # <<<<<<<<<<<<<<
 *             addfile_idx[i] = self.mask_to_files(fmask).astype('uint32')
 *         return file_idx.astype('uint32'), file_masks, addfile_idx
 */
    ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmasks->__pyx_vtab)->_logicaland(__pyx_v_self->bitmasks, ((__pyx_t_5numpy_uint32_t)__pyx_v_fid), __pyx_v_cmask, __pyx_v_fmask);

    /* "yt/geometry/particle_oct_container.pyx":1275
 *         for i, (fid, fmask) in enumerate(zip(file_idx,file_masks)):
 *             self.bitmasks._logicaland(<np.uint32_t> fid, cmask, fmask)
 *             addfile_idx[i] = self.mask_to_files(fmask).astype('uint32')             # <<<<<<<<<<<<<<
 *         return file_idx.astype('uint32'), file_masks, addfile_idx
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_mask_to_files); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_14 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_14)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_14);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_13 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_14, ((PyObject *)__pyx_v_fmask)) : __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_fmask));
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_astype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_10 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_13, __pyx_n_s_uint32) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_n_s_uint32);
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__pyx_v_addfile_idx == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 1275, __pyx_L1_error)
    }
    if (unlikely(PyObject_SetItem(__pyx_v_addfile_idx, __pyx_v_i, __pyx_t_10) < 0)) __PYX_ERR(0, 1275, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "yt/geometry/particle_oct_container.pyx":1273
 *                               dtype="object")
 *         addfile_idx = len(file_idx)*[None]
 *         for i, (fid, fmask) in enumerate(zip(file_idx,file_masks)):             # <<<<<<<<<<<<<<
 *             self.bitmasks._logicaland(<np.uint32_t> fid, cmask, fmask)
 *             addfile_idx[i] = self.mask_to_files(fmask).astype('uint32')
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1276
 *             self.bitmasks._logicaland(<np.uint32_t> fid, cmask, fmask)
 *             addfile_idx[i] = self.mask_to_files(fmask).astype('uint32')
 *         return file_idx.astype('uint32'), file_masks, addfile_idx             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_file_idx), __pyx_n_s_astype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_10 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_10, __pyx_n_s_uint32) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_s_uint32);
  __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_v_file_masks));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_file_masks));
  PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)__pyx_v_file_masks));
  __Pyx_INCREF(__pyx_v_addfile_idx);
  __Pyx_GIVEREF(__pyx_v_addfile_idx);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_addfile_idx);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1257
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def identify_file_masks(self, SelectorObject selector):             # <<<<<<<<<<<<<<
 *         cdef BoolArrayCollection cmask = BoolArrayCollection()
 *         cdef BoolArrayCollection fmask
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_idx.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_masks.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.identify_file_masks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_idx.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_masks.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cmask);
  __Pyx_XDECREF((PyObject *)__pyx_v_fmask);
  __Pyx_XDECREF((PyObject *)__pyx_v_file_masks);
  __Pyx_XDECREF((PyObject *)__pyx_v_file_idx);
  __Pyx_XDECREF(__pyx_v_addfile_idx);
  __Pyx_XDECREF((PyObject *)__pyx_v_morton_selector);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1281
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def identify_data_files(self, SelectorObject selector, int ngz = 0):             # <<<<<<<<<<<<<<
 *         cdef BoolArrayCollection cmask_s = BoolArrayCollection()
 *         cdef BoolArrayCollection cmask_g = BoolArrayCollection()
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_73identify_data_files(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_73identify_data_files(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_selector = 0;
  int __pyx_v_ngz;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("identify_data_files (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_selector,&__pyx_n_s_ngz,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_selector)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ngz);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "identify_data_files") < 0)) __PYX_ERR(0, 1281, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_selector = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)values[0]);
    if (values[1]) {
      __pyx_v_ngz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_ngz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1281, __pyx_L3_error)
    } else {
      __pyx_v_ngz = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("identify_data_files", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1281, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.identify_data_files", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_selector), __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject, 1, "selector", 0))) __PYX_ERR(0, 1281, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_72identify_data_files(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), __pyx_v_selector, __pyx_v_ngz);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_72identify_data_files(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_selector, int __pyx_v_ngz) {
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_cmask_s = 0;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_cmask_g = 0;
  struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_morton_selector = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("identify_data_files", 0);

  /* "yt/geometry/particle_oct_container.pyx":1282
 *     @cython.cdivision(True)
 *     def identify_data_files(self, SelectorObject selector, int ngz = 0):
 *         cdef BoolArrayCollection cmask_s = BoolArrayCollection()             # <<<<<<<<<<<<<<
 *         cdef BoolArrayCollection cmask_g = BoolArrayCollection()
 *         # Find mask of selected morton indices
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_cmask_s = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1283
 *     def identify_data_files(self, SelectorObject selector, int ngz = 0):
 *         cdef BoolArrayCollection cmask_s = BoolArrayCollection()
 *         cdef BoolArrayCollection cmask_g = BoolArrayCollection()             # <<<<<<<<<<<<<<
 *         # Find mask of selected morton indices
 *         cdef ParticleBitmapSelector morton_selector
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_cmask_g = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1286
 *         # Find mask of selected morton indices
 *         cdef ParticleBitmapSelector morton_selector
 *         morton_selector = ParticleBitmapSelector(selector, self, ngz=ngz)             # <<<<<<<<<<<<<<
 *         morton_selector.fill_masks(cmask_s, cmask_g)
 *         return self.masks_to_files(cmask_s, cmask_g), (cmask_s, cmask_g)
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_v_selector));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_selector));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_selector));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_ngz); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_ngz, __pyx_t_3) < 0) __PYX_ERR(0, 1286, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector), __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_morton_selector = ((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1287
 *         cdef ParticleBitmapSelector morton_selector
 *         morton_selector = ParticleBitmapSelector(selector, self, ngz=ngz)
 *         morton_selector.fill_masks(cmask_s, cmask_g)             # <<<<<<<<<<<<<<
 *         return self.masks_to_files(cmask_s, cmask_g), (cmask_s, cmask_g)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_morton_selector), __pyx_n_s_fill_masks); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_1, ((PyObject *)__pyx_v_cmask_s), ((PyObject *)__pyx_v_cmask_g)};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1287, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_1, ((PyObject *)__pyx_v_cmask_s), ((PyObject *)__pyx_v_cmask_g)};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1287, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_1) {
      __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1); __pyx_t_1 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_cmask_s));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_cmask_s));
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, ((PyObject *)__pyx_v_cmask_s));
    __Pyx_INCREF(((PyObject *)__pyx_v_cmask_g));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_cmask_g));
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, ((PyObject *)__pyx_v_cmask_g));
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1288
 *         morton_selector = ParticleBitmapSelector(selector, self, ngz=ngz)
 *         morton_selector.fill_masks(cmask_s, cmask_g)
 *         return self.masks_to_files(cmask_s, cmask_g), (cmask_s, cmask_g)             # <<<<<<<<<<<<<<
 * 
 *     def mask_to_files(self, BoolArrayCollection mm_s):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_masks_to_files); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_cmask_s), ((PyObject *)__pyx_v_cmask_g)};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1288, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_cmask_s), ((PyObject *)__pyx_v_cmask_g)};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1288, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  {
    __pyx_t_1 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_cmask_s));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_cmask_s));
    PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_4, ((PyObject *)__pyx_v_cmask_s));
    __Pyx_INCREF(((PyObject *)__pyx_v_cmask_g));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_cmask_g));
    PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_4, ((PyObject *)__pyx_v_cmask_g));
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_cmask_s));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_cmask_s));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_cmask_s));
  __Pyx_INCREF(((PyObject *)__pyx_v_cmask_g));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_cmask_g));
  PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_cmask_g));
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1281
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def identify_data_files(self, SelectorObject selector, int ngz = 0):             # <<<<<<<<<<<<<<
 *         cdef BoolArrayCollection cmask_s = BoolArrayCollection()
 *         cdef BoolArrayCollection cmask_g = BoolArrayCollection()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.identify_data_files", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_cmask_s);
  __Pyx_XDECREF((PyObject *)__pyx_v_cmask_g);
  __Pyx_XDECREF((PyObject *)__pyx_v_morton_selector);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1290
 *         return self.masks_to_files(cmask_s, cmask_g), (cmask_s, cmask_g)
 * 
 *     def mask_to_files(self, BoolArrayCollection mm_s):             # <<<<<<<<<<<<<<
 *         cdef FileBitmasks mm_d = self.bitmasks
 *         cdef np.uint32_t ifile
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_75mask_to_files(PyObject *__pyx_v_self, PyObject *__pyx_v_mm_s); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_75mask_to_files(PyObject *__pyx_v_self, PyObject *__pyx_v_mm_s) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mask_to_files (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mm_s), __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection, 1, "mm_s", 0))) __PYX_ERR(0, 1290, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_74mask_to_files(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_v_mm_s));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_74mask_to_files(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mm_s) {
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *__pyx_v_mm_d = 0;
  __pyx_t_5numpy_uint32_t __pyx_v_ifile;
  PyArrayObject *__pyx_v_file_mask_p = 0;
  PyArrayObject *__pyx_v_file_idx_p = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_file_idx_p;
  __Pyx_Buffer __pyx_pybuffer_file_idx_p;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_file_mask_p;
  __Pyx_Buffer __pyx_pybuffer_file_mask_p;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyArrayObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __pyx_t_5numpy_uint64_t __pyx_t_10;
  __pyx_t_5numpy_uint64_t __pyx_t_11;
  __pyx_t_5numpy_uint32_t __pyx_t_12;
  size_t __pyx_t_13;
  int __pyx_t_14;
  PyArrayObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mask_to_files", 0);
  __pyx_pybuffer_file_mask_p.pybuffer.buf = NULL;
  __pyx_pybuffer_file_mask_p.refcount = 0;
  __pyx_pybuffernd_file_mask_p.data = NULL;
  __pyx_pybuffernd_file_mask_p.rcbuffer = &__pyx_pybuffer_file_mask_p;
  __pyx_pybuffer_file_idx_p.pybuffer.buf = NULL;
  __pyx_pybuffer_file_idx_p.refcount = 0;
  __pyx_pybuffernd_file_idx_p.data = NULL;
  __pyx_pybuffernd_file_idx_p.rcbuffer = &__pyx_pybuffer_file_idx_p;

  /* "yt/geometry/particle_oct_container.pyx":1291
 * 
 *     def mask_to_files(self, BoolArrayCollection mm_s):
 *         cdef FileBitmasks mm_d = self.bitmasks             # <<<<<<<<<<<<<<
 *         cdef np.uint32_t ifile
 *         cdef np.ndarray[np.uint8_t, ndim=1] file_mask_p
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->bitmasks);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_mm_d = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1294
 *         cdef np.uint32_t ifile
 *         cdef np.ndarray[np.uint8_t, ndim=1] file_mask_p
 *         file_mask_p = np.zeros(self.nfiles, dtype="uint8")             # <<<<<<<<<<<<<<
 *         # Compare with mask of particles
 *         for ifile in range(self.nfiles):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->nfiles); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 1294, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1294, __pyx_L1_error)
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer);
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_6 < 0)) {
      PyErr_Fetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer, (PyObject*)__pyx_v_file_mask_p, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      }
      __pyx_t_7 = __pyx_t_8 = __pyx_t_9 = 0;
    }
    __pyx_pybuffernd_file_mask_p.diminfo[0].strides = __pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_file_mask_p.diminfo[0].shape = __pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1294, __pyx_L1_error)
  }
  __pyx_t_5 = 0;
  __pyx_v_file_mask_p = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1296
 *         file_mask_p = np.zeros(self.nfiles, dtype="uint8")
 *         # Compare with mask of particles
 *         for ifile in range(self.nfiles):             # <<<<<<<<<<<<<<
 *             # Only continue if the file is not already selected
 *             if file_mask_p[ifile] == 0:
 */
  __pyx_t_10 = __pyx_v_self->nfiles;
  __pyx_t_11 = __pyx_t_10;
  for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_ifile = __pyx_t_12;

    /* "yt/geometry/particle_oct_container.pyx":1298
 *         for ifile in range(self.nfiles):
 *             # Only continue if the file is not already selected
 *             if file_mask_p[ifile] == 0:             # <<<<<<<<<<<<<<
 *                 if mm_d._intersects(ifile, mm_s):
 *                     file_mask_p[ifile] = 1
 */
    __pyx_t_13 = __pyx_v_ifile;
    __pyx_t_6 = -1;
    if (unlikely(__pyx_t_13 >= (size_t)__pyx_pybuffernd_file_mask_p.diminfo[0].shape)) __pyx_t_6 = 0;
    if (unlikely(__pyx_t_6 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_6);
      __PYX_ERR(0, 1298, __pyx_L1_error)
    }
    __pyx_t_14 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_file_mask_p.diminfo[0].strides)) == 0) != 0);
    if (__pyx_t_14) {

      /* "yt/geometry/particle_oct_container.pyx":1299
 *             # Only continue if the file is not already selected
 *             if file_mask_p[ifile] == 0:
 *                 if mm_d._intersects(ifile, mm_s):             # <<<<<<<<<<<<<<
 *                     file_mask_p[ifile] = 1
 *         cdef np.ndarray[np.int32_t, ndim=1] file_idx_p
 */
      __pyx_t_14 = (((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_mm_d->__pyx_vtab)->_intersects(__pyx_v_mm_d, __pyx_v_ifile, __pyx_v_mm_s) != 0);
      if (__pyx_t_14) {

        /* "yt/geometry/particle_oct_container.pyx":1300
 *             if file_mask_p[ifile] == 0:
 *                 if mm_d._intersects(ifile, mm_s):
 *                     file_mask_p[ifile] = 1             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.int32_t, ndim=1] file_idx_p
 *         file_idx_p = np.where(file_mask_p)[0].astype('int32')
 */
        __pyx_t_13 = __pyx_v_ifile;
        __pyx_t_6 = -1;
        if (unlikely(__pyx_t_13 >= (size_t)__pyx_pybuffernd_file_mask_p.diminfo[0].shape)) __pyx_t_6 = 0;
        if (unlikely(__pyx_t_6 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_6);
          __PYX_ERR(0, 1300, __pyx_L1_error)
        }
        *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer.buf, __pyx_t_13, __pyx_pybuffernd_file_mask_p.diminfo[0].strides) = 1;

        /* "yt/geometry/particle_oct_container.pyx":1299
 *             # Only continue if the file is not already selected
 *             if file_mask_p[ifile] == 0:
 *                 if mm_d._intersects(ifile, mm_s):             # <<<<<<<<<<<<<<
 *                     file_mask_p[ifile] = 1
 *         cdef np.ndarray[np.int32_t, ndim=1] file_idx_p
 */
      }

      /* "yt/geometry/particle_oct_container.pyx":1298
 *         for ifile in range(self.nfiles):
 *             # Only continue if the file is not already selected
 *             if file_mask_p[ifile] == 0:             # <<<<<<<<<<<<<<
 *                 if mm_d._intersects(ifile, mm_s):
 *                     file_mask_p[ifile] = 1
 */
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":1302
 *                     file_mask_p[ifile] = 1
 *         cdef np.ndarray[np.int32_t, ndim=1] file_idx_p
 *         file_idx_p = np.where(file_mask_p)[0].astype('int32')             # <<<<<<<<<<<<<<
 *         return file_idx_p.astype('uint32')
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_where); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, ((PyObject *)__pyx_v_file_mask_p)) : __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_file_mask_p));
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_astype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_n_s_int32) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_n_s_int32);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1302, __pyx_L1_error)
  __pyx_t_15 = ((PyArrayObject *)__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_idx_p.rcbuffer->pybuffer);
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_idx_p.rcbuffer->pybuffer, (PyObject*)__pyx_t_15, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_6 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_idx_p.rcbuffer->pybuffer, (PyObject*)__pyx_v_file_idx_p, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_7);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_8, __pyx_t_7);
      }
      __pyx_t_9 = __pyx_t_8 = __pyx_t_7 = 0;
    }
    __pyx_pybuffernd_file_idx_p.diminfo[0].strides = __pyx_pybuffernd_file_idx_p.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_file_idx_p.diminfo[0].shape = __pyx_pybuffernd_file_idx_p.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1302, __pyx_L1_error)
  }
  __pyx_t_15 = 0;
  __pyx_v_file_idx_p = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1303
 *         cdef np.ndarray[np.int32_t, ndim=1] file_idx_p
 *         file_idx_p = np.where(file_mask_p)[0].astype('int32')
 *         return file_idx_p.astype('uint32')             # <<<<<<<<<<<<<<
 * 
 *     def masks_to_files(self, BoolArrayCollection mm_s, BoolArrayCollection mm_g):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_file_idx_p), __pyx_n_s_astype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_n_s_uint32) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_n_s_uint32);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1290
 *         return self.masks_to_files(cmask_s, cmask_g), (cmask_s, cmask_g)
 * 
 *     def mask_to_files(self, BoolArrayCollection mm_s):             # <<<<<<<<<<<<<<
 *         cdef FileBitmasks mm_d = self.bitmasks
 *         cdef np.uint32_t ifile
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_idx_p.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.mask_to_files", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_idx_p.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mm_d);
  __Pyx_XDECREF((PyObject *)__pyx_v_file_mask_p);
  __Pyx_XDECREF((PyObject *)__pyx_v_file_idx_p);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1305
 *         return file_idx_p.astype('uint32')
 * 
 *     def masks_to_files(self, BoolArrayCollection mm_s, BoolArrayCollection mm_g):             # <<<<<<<<<<<<<<
 *         cdef FileBitmasks mm_d = self.bitmasks
 *         cdef np.uint32_t ifile
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_77masks_to_files(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_77masks_to_files(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mm_s = 0;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mm_g = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("masks_to_files (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mm_s,&__pyx_n_s_mm_g,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mm_s)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mm_g)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("masks_to_files", 1, 2, 2, 1); __PYX_ERR(0, 1305, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "masks_to_files") < 0)) __PYX_ERR(0, 1305, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_mm_s = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)values[0]);
    __pyx_v_mm_g = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("masks_to_files", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1305, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.masks_to_files", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mm_s), __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection, 1, "mm_s", 0))) __PYX_ERR(0, 1305, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mm_g), __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection, 1, "mm_g", 0))) __PYX_ERR(0, 1305, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_76masks_to_files(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), __pyx_v_mm_s, __pyx_v_mm_g);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_76masks_to_files(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mm_s, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mm_g) {
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *__pyx_v_mm_d = 0;
  __pyx_t_5numpy_uint32_t __pyx_v_ifile;
  PyArrayObject *__pyx_v_file_mask_p = 0;
  PyArrayObject *__pyx_v_file_mask_g = 0;
  PyArrayObject *__pyx_v_file_idx_p = 0;
  PyArrayObject *__pyx_v_file_idx_g = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_file_idx_g;
  __Pyx_Buffer __pyx_pybuffer_file_idx_g;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_file_idx_p;
  __Pyx_Buffer __pyx_pybuffer_file_idx_p;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_file_mask_g;
  __Pyx_Buffer __pyx_pybuffer_file_mask_g;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_file_mask_p;
  __Pyx_Buffer __pyx_pybuffer_file_mask_p;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyArrayObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyArrayObject *__pyx_t_10 = NULL;
  __pyx_t_5numpy_uint64_t __pyx_t_11;
  __pyx_t_5numpy_uint64_t __pyx_t_12;
  __pyx_t_5numpy_uint32_t __pyx_t_13;
  size_t __pyx_t_14;
  int __pyx_t_15;
  PyArrayObject *__pyx_t_16 = NULL;
  PyArrayObject *__pyx_t_17 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("masks_to_files", 0);
  __pyx_pybuffer_file_mask_p.pybuffer.buf = NULL;
  __pyx_pybuffer_file_mask_p.refcount = 0;
  __pyx_pybuffernd_file_mask_p.data = NULL;
  __pyx_pybuffernd_file_mask_p.rcbuffer = &__pyx_pybuffer_file_mask_p;
  __pyx_pybuffer_file_mask_g.pybuffer.buf = NULL;
  __pyx_pybuffer_file_mask_g.refcount = 0;
  __pyx_pybuffernd_file_mask_g.data = NULL;
  __pyx_pybuffernd_file_mask_g.rcbuffer = &__pyx_pybuffer_file_mask_g;
  __pyx_pybuffer_file_idx_p.pybuffer.buf = NULL;
  __pyx_pybuffer_file_idx_p.refcount = 0;
  __pyx_pybuffernd_file_idx_p.data = NULL;
  __pyx_pybuffernd_file_idx_p.rcbuffer = &__pyx_pybuffer_file_idx_p;
  __pyx_pybuffer_file_idx_g.pybuffer.buf = NULL;
  __pyx_pybuffer_file_idx_g.refcount = 0;
  __pyx_pybuffernd_file_idx_g.data = NULL;
  __pyx_pybuffernd_file_idx_g.rcbuffer = &__pyx_pybuffer_file_idx_g;

  /* "yt/geometry/particle_oct_container.pyx":1306
 * 
 *     def masks_to_files(self, BoolArrayCollection mm_s, BoolArrayCollection mm_g):
 *         cdef FileBitmasks mm_d = self.bitmasks             # <<<<<<<<<<<<<<
 *         cdef np.uint32_t ifile
 *         cdef np.ndarray[np.uint8_t, ndim=1] file_mask_p
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->bitmasks);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_mm_d = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1310
 *         cdef np.ndarray[np.uint8_t, ndim=1] file_mask_p
 *         cdef np.ndarray[np.uint8_t, ndim=1] file_mask_g
 *         file_mask_p = np.zeros(self.nfiles, dtype="uint8")             # <<<<<<<<<<<<<<
 *         file_mask_g = np.zeros(self.nfiles, dtype="uint8")
 *         # Compare with mask of particles
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->nfiles); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 1310, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1310, __pyx_L1_error)
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_4);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer);
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_6 < 0)) {
      PyErr_Fetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer, (PyObject*)__pyx_v_file_mask_p, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      }
      __pyx_t_7 = __pyx_t_8 = __pyx_t_9 = 0;
    }
    __pyx_pybuffernd_file_mask_p.diminfo[0].strides = __pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_file_mask_p.diminfo[0].shape = __pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1310, __pyx_L1_error)
  }
  __pyx_t_5 = 0;
  __pyx_v_file_mask_p = ((PyArrayObject *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1311
 *         cdef np.ndarray[np.uint8_t, ndim=1] file_mask_g
 *         file_mask_p = np.zeros(self.nfiles, dtype="uint8")
 *         file_mask_g = np.zeros(self.nfiles, dtype="uint8")             # <<<<<<<<<<<<<<
 *         # Compare with mask of particles
 *         for ifile in range(self.nfiles):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->nfiles); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_dtype, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 1311, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1311, __pyx_L1_error)
  __pyx_t_10 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_mask_g.rcbuffer->pybuffer);
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_mask_g.rcbuffer->pybuffer, (PyObject*)__pyx_t_10, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_6 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_mask_g.rcbuffer->pybuffer, (PyObject*)__pyx_v_file_mask_g, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_7);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_8, __pyx_t_7);
      }
      __pyx_t_9 = __pyx_t_8 = __pyx_t_7 = 0;
    }
    __pyx_pybuffernd_file_mask_g.diminfo[0].strides = __pyx_pybuffernd_file_mask_g.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_file_mask_g.diminfo[0].shape = __pyx_pybuffernd_file_mask_g.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1311, __pyx_L1_error)
  }
  __pyx_t_10 = 0;
  __pyx_v_file_mask_g = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1313
 *         file_mask_g = np.zeros(self.nfiles, dtype="uint8")
 *         # Compare with mask of particles
 *         for ifile in range(self.nfiles):             # <<<<<<<<<<<<<<
 *             # Only continue if the file is not already selected
 *             if file_mask_p[ifile] == 0:
 */
  __pyx_t_11 = __pyx_v_self->nfiles;
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = 0; __pyx_t_13 < __pyx_t_12; __pyx_t_13+=1) {
    __pyx_v_ifile = __pyx_t_13;

    /* "yt/geometry/particle_oct_container.pyx":1315
 *         for ifile in range(self.nfiles):
 *             # Only continue if the file is not already selected
 *             if file_mask_p[ifile] == 0:             # <<<<<<<<<<<<<<
 *                 if mm_d._intersects(ifile, mm_s):
 *                     file_mask_p[ifile] = 1
 */
    __pyx_t_14 = __pyx_v_ifile;
    __pyx_t_6 = -1;
    if (unlikely(__pyx_t_14 >= (size_t)__pyx_pybuffernd_file_mask_p.diminfo[0].shape)) __pyx_t_6 = 0;
    if (unlikely(__pyx_t_6 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_6);
      __PYX_ERR(0, 1315, __pyx_L1_error)
    }
    __pyx_t_15 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_file_mask_p.diminfo[0].strides)) == 0) != 0);
    if (__pyx_t_15) {

      /* "yt/geometry/particle_oct_container.pyx":1316
 *             # Only continue if the file is not already selected
 *             if file_mask_p[ifile] == 0:
 *                 if mm_d._intersects(ifile, mm_s):             # <<<<<<<<<<<<<<
 *                     file_mask_p[ifile] = 1
 *                     file_mask_g[ifile] = 0 # No intersection
 */
      __pyx_t_15 = (((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_mm_d->__pyx_vtab)->_intersects(__pyx_v_mm_d, __pyx_v_ifile, __pyx_v_mm_s) != 0);
      if (__pyx_t_15) {

        /* "yt/geometry/particle_oct_container.pyx":1317
 *             if file_mask_p[ifile] == 0:
 *                 if mm_d._intersects(ifile, mm_s):
 *                     file_mask_p[ifile] = 1             # <<<<<<<<<<<<<<
 *                     file_mask_g[ifile] = 0 # No intersection
 *                 elif mm_d._intersects(ifile, mm_g):
 */
        __pyx_t_14 = __pyx_v_ifile;
        __pyx_t_6 = -1;
        if (unlikely(__pyx_t_14 >= (size_t)__pyx_pybuffernd_file_mask_p.diminfo[0].shape)) __pyx_t_6 = 0;
        if (unlikely(__pyx_t_6 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_6);
          __PYX_ERR(0, 1317, __pyx_L1_error)
        }
        *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_file_mask_p.diminfo[0].strides) = 1;

        /* "yt/geometry/particle_oct_container.pyx":1318
 *                 if mm_d._intersects(ifile, mm_s):
 *                     file_mask_p[ifile] = 1
 *                     file_mask_g[ifile] = 0 # No intersection             # <<<<<<<<<<<<<<
 *                 elif mm_d._intersects(ifile, mm_g):
 *                     file_mask_g[ifile] = 1
 */
        __pyx_t_14 = __pyx_v_ifile;
        __pyx_t_6 = -1;
        if (unlikely(__pyx_t_14 >= (size_t)__pyx_pybuffernd_file_mask_g.diminfo[0].shape)) __pyx_t_6 = 0;
        if (unlikely(__pyx_t_6 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_6);
          __PYX_ERR(0, 1318, __pyx_L1_error)
        }
        *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_file_mask_g.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_file_mask_g.diminfo[0].strides) = 0;

        /* "yt/geometry/particle_oct_container.pyx":1316
 *             # Only continue if the file is not already selected
 *             if file_mask_p[ifile] == 0:
 *                 if mm_d._intersects(ifile, mm_s):             # <<<<<<<<<<<<<<
 *                     file_mask_p[ifile] = 1
 *                     file_mask_g[ifile] = 0 # No intersection
 */
        goto __pyx_L6;
      }

      /* "yt/geometry/particle_oct_container.pyx":1319
 *                     file_mask_p[ifile] = 1
 *                     file_mask_g[ifile] = 0 # No intersection
 *                 elif mm_d._intersects(ifile, mm_g):             # <<<<<<<<<<<<<<
 *                     file_mask_g[ifile] = 1
 *         cdef np.ndarray[np.int32_t, ndim=1] file_idx_p
 */
      __pyx_t_15 = (((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_mm_d->__pyx_vtab)->_intersects(__pyx_v_mm_d, __pyx_v_ifile, __pyx_v_mm_g) != 0);
      if (__pyx_t_15) {

        /* "yt/geometry/particle_oct_container.pyx":1320
 *                     file_mask_g[ifile] = 0 # No intersection
 *                 elif mm_d._intersects(ifile, mm_g):
 *                     file_mask_g[ifile] = 1             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.int32_t, ndim=1] file_idx_p
 *         cdef np.ndarray[np.int32_t, ndim=1] file_idx_g
 */
        __pyx_t_14 = __pyx_v_ifile;
        __pyx_t_6 = -1;
        if (unlikely(__pyx_t_14 >= (size_t)__pyx_pybuffernd_file_mask_g.diminfo[0].shape)) __pyx_t_6 = 0;
        if (unlikely(__pyx_t_6 != -1)) {
          __Pyx_RaiseBufferIndexError(__pyx_t_6);
          __PYX_ERR(0, 1320, __pyx_L1_error)
        }
        *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_file_mask_g.rcbuffer->pybuffer.buf, __pyx_t_14, __pyx_pybuffernd_file_mask_g.diminfo[0].strides) = 1;

        /* "yt/geometry/particle_oct_container.pyx":1319
 *                     file_mask_p[ifile] = 1
 *                     file_mask_g[ifile] = 0 # No intersection
 *                 elif mm_d._intersects(ifile, mm_g):             # <<<<<<<<<<<<<<
 *                     file_mask_g[ifile] = 1
 *         cdef np.ndarray[np.int32_t, ndim=1] file_idx_p
 */
      }
      __pyx_L6:;

      /* "yt/geometry/particle_oct_container.pyx":1315
 *         for ifile in range(self.nfiles):
 *             # Only continue if the file is not already selected
 *             if file_mask_p[ifile] == 0:             # <<<<<<<<<<<<<<
 *                 if mm_d._intersects(ifile, mm_s):
 *                     file_mask_p[ifile] = 1
 */
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":1323
 *         cdef np.ndarray[np.int32_t, ndim=1] file_idx_p
 *         cdef np.ndarray[np.int32_t, ndim=1] file_idx_g
 *         file_idx_p = np.where(file_mask_p)[0].astype('int32')             # <<<<<<<<<<<<<<
 *         file_idx_g = np.where(file_mask_g)[0].astype('int32')
 *         return file_idx_p.astype('uint32'), file_idx_g.astype('uint32')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_where); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, ((PyObject *)__pyx_v_file_mask_p)) : __Pyx_PyObject_CallOneArg(__pyx_t_1, ((PyObject *)__pyx_v_file_mask_p));
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_4, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_astype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_n_s_int32) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_n_s_int32);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1323, __pyx_L1_error)
  __pyx_t_16 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_idx_p.rcbuffer->pybuffer);
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_idx_p.rcbuffer->pybuffer, (PyObject*)__pyx_t_16, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_6 < 0)) {
      PyErr_Fetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_idx_p.rcbuffer->pybuffer, (PyObject*)__pyx_v_file_idx_p, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_7); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      }
      __pyx_t_7 = __pyx_t_8 = __pyx_t_9 = 0;
    }
    __pyx_pybuffernd_file_idx_p.diminfo[0].strides = __pyx_pybuffernd_file_idx_p.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_file_idx_p.diminfo[0].shape = __pyx_pybuffernd_file_idx_p.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1323, __pyx_L1_error)
  }
  __pyx_t_16 = 0;
  __pyx_v_file_idx_p = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1324
 *         cdef np.ndarray[np.int32_t, ndim=1] file_idx_g
 *         file_idx_p = np.where(file_mask_p)[0].astype('int32')
 *         file_idx_g = np.where(file_mask_g)[0].astype('int32')             # <<<<<<<<<<<<<<
 *         return file_idx_p.astype('uint32'), file_idx_g.astype('uint32')
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_where); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, ((PyObject *)__pyx_v_file_mask_g)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_file_mask_g));
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_4, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_astype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_n_s_int32) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_n_s_int32);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1324, __pyx_L1_error)
  __pyx_t_17 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_idx_g.rcbuffer->pybuffer);
    __pyx_t_6 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_idx_g.rcbuffer->pybuffer, (PyObject*)__pyx_t_17, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_6 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_idx_g.rcbuffer->pybuffer, (PyObject*)__pyx_v_file_idx_g, &__Pyx_TypeInfo_nn___pyx_t_5numpy_int32_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_8); Py_XDECREF(__pyx_t_7);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_8, __pyx_t_7);
      }
      __pyx_t_9 = __pyx_t_8 = __pyx_t_7 = 0;
    }
    __pyx_pybuffernd_file_idx_g.diminfo[0].strides = __pyx_pybuffernd_file_idx_g.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_file_idx_g.diminfo[0].shape = __pyx_pybuffernd_file_idx_g.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1324, __pyx_L1_error)
  }
  __pyx_t_17 = 0;
  __pyx_v_file_idx_g = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1325
 *         file_idx_p = np.where(file_mask_p)[0].astype('int32')
 *         file_idx_g = np.where(file_mask_g)[0].astype('int32')
 *         return file_idx_p.astype('uint32'), file_idx_g.astype('uint32')             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_file_idx_p), __pyx_n_s_astype); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_n_s_uint32) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_n_s_uint32);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_file_idx_g), __pyx_n_s_astype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_n_s_uint32) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_s_uint32);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1305
 *         return file_idx_p.astype('uint32')
 * 
 *     def masks_to_files(self, BoolArrayCollection mm_s, BoolArrayCollection mm_g):             # <<<<<<<<<<<<<<
 *         cdef FileBitmasks mm_d = self.bitmasks
 *         cdef np.uint32_t ifile
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_idx_g.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_idx_p.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_mask_g.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.masks_to_files", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_idx_g.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_idx_p.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_mask_g.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mm_d);
  __Pyx_XDECREF((PyObject *)__pyx_v_file_mask_p);
  __Pyx_XDECREF((PyObject *)__pyx_v_file_mask_g);
  __Pyx_XDECREF((PyObject *)__pyx_v_file_idx_p);
  __Pyx_XDECREF((PyObject *)__pyx_v_file_idx_g);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1330
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def construct_octree(self, index, io_handler, data_files,             # <<<<<<<<<<<<<<
 *                          num_zones,
 *                          BoolArrayCollection selector_mask,
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_79construct_octree(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_79construct_octree(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_index = 0;
  PyObject *__pyx_v_io_handler = 0;
  PyObject *__pyx_v_data_files = 0;
  PyObject *__pyx_v_num_zones = 0;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_selector_mask = 0;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_base_mask = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("construct_octree (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_index,&__pyx_n_s_io_handler,&__pyx_n_s_data_files,&__pyx_n_s_num_zones,&__pyx_n_s_selector_mask,&__pyx_n_s_base_mask,0};
    PyObject* values[6] = {0,0,0,0,0,0};

    /* "yt/geometry/particle_oct_container.pyx":1333
 *                          num_zones,
 *                          BoolArrayCollection selector_mask,
 *                          BoolArrayCollection base_mask = None):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t total_pcount
 *         cdef np.uint64_t i, j, k, n
 */
    values[5] = (PyObject *)((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_index)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_io_handler)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("construct_octree", 0, 5, 6, 1); __PYX_ERR(0, 1330, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data_files)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("construct_octree", 0, 5, 6, 2); __PYX_ERR(0, 1330, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_zones)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("construct_octree", 0, 5, 6, 3); __PYX_ERR(0, 1330, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_selector_mask)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("construct_octree", 0, 5, 6, 4); __PYX_ERR(0, 1330, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_base_mask);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "construct_octree") < 0)) __PYX_ERR(0, 1330, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_index = values[0];
    __pyx_v_io_handler = values[1];
    __pyx_v_data_files = values[2];
    __pyx_v_num_zones = values[3];
    __pyx_v_selector_mask = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)values[4]);
    __pyx_v_base_mask = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)values[5]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("construct_octree", 0, 5, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1330, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.construct_octree", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_selector_mask), __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection, 1, "selector_mask", 0))) __PYX_ERR(0, 1332, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_base_mask), __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection, 1, "base_mask", 0))) __PYX_ERR(0, 1333, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_78construct_octree(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), __pyx_v_index, __pyx_v_io_handler, __pyx_v_data_files, __pyx_v_num_zones, __pyx_v_selector_mask, __pyx_v_base_mask);

  /* "yt/geometry/particle_oct_container.pyx":1330
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def construct_octree(self, index, io_handler, data_files,             # <<<<<<<<<<<<<<
 *                          num_zones,
 *                          BoolArrayCollection selector_mask,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_78construct_octree(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_io_handler, PyObject *__pyx_v_data_files, PyObject *__pyx_v_num_zones, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_selector_mask, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_base_mask) {
  __pyx_t_5numpy_uint64_t __pyx_v_total_pcount;
  __pyx_t_5numpy_uint64_t __pyx_v_i;
  __pyx_t_5numpy_uint64_t __pyx_v_j;
  __pyx_t_5numpy_uint64_t __pyx_v_k;
  int __pyx_v_ind[3];
  __pyx_t_5numpy_uint64_t __pyx_v_ind64[3];
  struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_octree = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_mi;
  __pyx_t_5numpy_uint64_t __pyx_v_mi_root;
  PyArrayObject *__pyx_v_pos = 0;
  PyArrayObject *__pyx_v_pos32 = 0;
  PyArrayObject *__pyx_v_pos64 = 0;
  __pyx_t_5numpy_float64_t __pyx_v_ppos[3];
  __pyx_t_5numpy_float64_t __pyx_v_DLE[3];
  __pyx_t_5numpy_float64_t __pyx_v_DRE[3];
  CYTHON_UNUSED int __pyx_v_bitsize;
  PyArrayObject *__pyx_v_morton_ind = 0;
  __pyx_t_5numpy_uint64_t __pyx_v_nroot;
  __pyx_t_5numpy_uint64_t __pyx_v_croot;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array *__pyx_v_ewah_slct;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array *__pyx_v_ewah_base;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_iterator *__pyx_v_iter_set;
  __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_iterator *__pyx_v_iter_end;
  PyArrayObject *__pyx_v_slct_arr = 0;
  PyObject *__pyx_v_files_touched = NULL;
  PyObject *__pyx_v_data_file = NULL;
  __pyx_t_5numpy_uint64_t __pyx_v_base_pcount;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_morton_ind;
  __Pyx_Buffer __pyx_pybuffer_morton_ind;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_pos32;
  __Pyx_Buffer __pyx_pybuffer_pos32;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_pos64;
  __Pyx_Buffer __pyx_pybuffer_pos64;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_slct_arr;
  __Pyx_Buffer __pyx_pybuffer_slct_arr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_uint64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __pyx_t_5numpy_int32_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  PyArrayObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  Py_ssize_t __pyx_t_16;
  PyObject *(*__pyx_t_17)(PyObject *);
  PyArrayObject *__pyx_t_18 = NULL;
  Py_ssize_t __pyx_t_19;
  PyObject *(*__pyx_t_20)(PyObject *);
  PyArrayObject *__pyx_t_21 = NULL;
  PyArrayObject *__pyx_t_22 = NULL;
  npy_intp __pyx_t_23;
  npy_intp __pyx_t_24;
  __pyx_t_5numpy_uint64_t __pyx_t_25;
  __pyx_t_5numpy_uint64_t __pyx_t_26;
  __pyx_t_5numpy_uint64_t __pyx_t_27;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("construct_octree", 0);
  __pyx_pybuffer_pos32.pybuffer.buf = NULL;
  __pyx_pybuffer_pos32.refcount = 0;
  __pyx_pybuffernd_pos32.data = NULL;
  __pyx_pybuffernd_pos32.rcbuffer = &__pyx_pybuffer_pos32;
  __pyx_pybuffer_pos64.pybuffer.buf = NULL;
  __pyx_pybuffer_pos64.refcount = 0;
  __pyx_pybuffernd_pos64.data = NULL;
  __pyx_pybuffernd_pos64.rcbuffer = &__pyx_pybuffer_pos64;
  __pyx_pybuffer_morton_ind.pybuffer.buf = NULL;
  __pyx_pybuffer_morton_ind.refcount = 0;
  __pyx_pybuffernd_morton_ind.data = NULL;
  __pyx_pybuffernd_morton_ind.rcbuffer = &__pyx_pybuffer_morton_ind;
  __pyx_pybuffer_slct_arr.pybuffer.buf = NULL;
  __pyx_pybuffer_slct_arr.refcount = 0;
  __pyx_pybuffernd_slct_arr.data = NULL;
  __pyx_pybuffernd_slct_arr.rcbuffer = &__pyx_pybuffer_slct_arr;

  /* "yt/geometry/particle_oct_container.pyx":1346
 *         cdef np.float64_t DLE[3]
 *         cdef np.float64_t DRE[3]
 *         cdef int bitsize = 0             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             DLE[i] = self.left_edge[i]
 */
  __pyx_v_bitsize = 0;

  /* "yt/geometry/particle_oct_container.pyx":1347
 *         cdef np.float64_t DRE[3]
 *         cdef int bitsize = 0
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             DLE[i] = self.left_edge[i]
 *             DRE[i] = self.right_edge[i]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":1348
 *         cdef int bitsize = 0
 *         for i in range(3):
 *             DLE[i] = self.left_edge[i]             # <<<<<<<<<<<<<<
 *             DRE[i] = self.right_edge[i]
 *         cdef np.ndarray[np.uint64_t, ndim=1] morton_ind
 */
    (__pyx_v_DLE[__pyx_v_i]) = (__pyx_v_self->left_edge[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":1349
 *         for i in range(3):
 *             DLE[i] = self.left_edge[i]
 *             DRE[i] = self.right_edge[i]             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint64_t, ndim=1] morton_ind
 *         # Determine cells that need to be added to the octree
 */
    (__pyx_v_DRE[__pyx_v_i]) = (__pyx_v_self->right_edge[__pyx_v_i]);
  }

  /* "yt/geometry/particle_oct_container.pyx":1354
 *         cdef np.ndarray[np.uint32_t, ndim=1] file_idx_p
 *         cdef np.ndarray[np.uint32_t, ndim=1] file_idx_g
 *         cdef np.uint64_t nroot = selector_mask._count_total()             # <<<<<<<<<<<<<<
 *         # Now we can actually create a sparse octree.
 *         octree = ParticleBitmapOctreeContainer(
 */
  __pyx_v_nroot = ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_v_selector_mask->__pyx_vtab)->_count_total(__pyx_v_selector_mask);

  /* "yt/geometry/particle_oct_container.pyx":1357
 *         # Now we can actually create a sparse octree.
 *         octree = ParticleBitmapOctreeContainer(
 *             (self.dims[0], self.dims[1], self.dims[2]),             # <<<<<<<<<<<<<<
 *             (self.left_edge[0], self.left_edge[1], self.left_edge[2]),
 *             (self.right_edge[0], self.right_edge[1], self.right_edge[2]),
 */
  __pyx_t_2 = __Pyx_PyInt_From_npy_int32((__pyx_v_self->dims[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_npy_int32((__pyx_v_self->dims[1])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_npy_int32((__pyx_v_self->dims[2])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1358
 *         octree = ParticleBitmapOctreeContainer(
 *             (self.dims[0], self.dims[1], self.dims[2]),
 *             (self.left_edge[0], self.left_edge[1], self.left_edge[2]),             # <<<<<<<<<<<<<<
 *             (self.right_edge[0], self.right_edge[1], self.right_edge[2]),
 *             nroot, num_zones)
 */
  __pyx_t_4 = PyFloat_FromDouble((__pyx_v_self->left_edge[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyFloat_FromDouble((__pyx_v_self->left_edge[1])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyFloat_FromDouble((__pyx_v_self->left_edge[2])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_2);
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1359
 *             (self.dims[0], self.dims[1], self.dims[2]),
 *             (self.left_edge[0], self.left_edge[1], self.left_edge[2]),
 *             (self.right_edge[0], self.right_edge[1], self.right_edge[2]),             # <<<<<<<<<<<<<<
 *             nroot, num_zones)
 *         octree.n_ref = index.dataset.n_ref
 */
  __pyx_t_2 = PyFloat_FromDouble((__pyx_v_self->right_edge[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble((__pyx_v_self->right_edge[1])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble((__pyx_v_self->right_edge[2])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1360
 *             (self.left_edge[0], self.left_edge[1], self.left_edge[2]),
 *             (self.right_edge[0], self.right_edge[1], self.right_edge[2]),
 *             nroot, num_zones)             # <<<<<<<<<<<<<<
 *         octree.n_ref = index.dataset.n_ref
 *         octree.level_offset = self.index_order1
 */
  __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_nroot); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "yt/geometry/particle_oct_container.pyx":1356
 *         cdef np.uint64_t nroot = selector_mask._count_total()
 *         # Now we can actually create a sparse octree.
 *         octree = ParticleBitmapOctreeContainer(             # <<<<<<<<<<<<<<
 *             (self.dims[0], self.dims[1], self.dims[2]),
 *             (self.left_edge[0], self.left_edge[1], self.left_edge[2]),
 */
  __pyx_t_3 = PyTuple_New(5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_4);
  __Pyx_INCREF(__pyx_v_num_zones);
  __Pyx_GIVEREF(__pyx_v_num_zones);
  PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_v_num_zones);
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer), __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_octree = ((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1361
 *             (self.right_edge[0], self.right_edge[1], self.right_edge[2]),
 *             nroot, num_zones)
 *         octree.n_ref = index.dataset.n_ref             # <<<<<<<<<<<<<<
 *         octree.level_offset = self.index_order1
 *         octree.allocate_domains()
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_dataset); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_n_ref); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1361, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_octree->n_ref = __pyx_t_8;

  /* "yt/geometry/particle_oct_container.pyx":1362
 *             nroot, num_zones)
 *         octree.n_ref = index.dataset.n_ref
 *         octree.level_offset = self.index_order1             # <<<<<<<<<<<<<<
 *         octree.allocate_domains()
 *         # Add roots based on the mask
 */
  __pyx_t_9 = __pyx_v_self->index_order1;
  __pyx_v_octree->__pyx_base.__pyx_base.level_offset = __pyx_t_9;

  /* "yt/geometry/particle_oct_container.pyx":1363
 *         octree.n_ref = index.dataset.n_ref
 *         octree.level_offset = self.index_order1
 *         octree.allocate_domains()             # <<<<<<<<<<<<<<
 *         # Add roots based on the mask
 *         cdef np.uint64_t croot = 0
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_octree), __pyx_n_s_allocate_domains); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1365
 *         octree.allocate_domains()
 *         # Add roots based on the mask
 *         cdef np.uint64_t croot = 0             # <<<<<<<<<<<<<<
 *         cdef ewah_bool_array *ewah_slct = <ewah_bool_array *> selector_mask.ewah_keys
 *         cdef ewah_bool_array *ewah_base
 */
  __pyx_v_croot = 0;

  /* "yt/geometry/particle_oct_container.pyx":1366
 *         # Add roots based on the mask
 *         cdef np.uint64_t croot = 0
 *         cdef ewah_bool_array *ewah_slct = <ewah_bool_array *> selector_mask.ewah_keys             # <<<<<<<<<<<<<<
 *         cdef ewah_bool_array *ewah_base
 *         if base_mask is not None:
 */
  __pyx_v_ewah_slct = ((__pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array *)__pyx_v_selector_mask->ewah_keys);

  /* "yt/geometry/particle_oct_container.pyx":1368
 *         cdef ewah_bool_array *ewah_slct = <ewah_bool_array *> selector_mask.ewah_keys
 *         cdef ewah_bool_array *ewah_base
 *         if base_mask is not None:             # <<<<<<<<<<<<<<
 *             ewah_base = <ewah_bool_array *> base_mask.ewah_keys
 *         else:
 */
  __pyx_t_10 = (((PyObject *)__pyx_v_base_mask) != Py_None);
  __pyx_t_11 = (__pyx_t_10 != 0);
  if (__pyx_t_11) {

    /* "yt/geometry/particle_oct_container.pyx":1369
 *         cdef ewah_bool_array *ewah_base
 *         if base_mask is not None:
 *             ewah_base = <ewah_bool_array *> base_mask.ewah_keys             # <<<<<<<<<<<<<<
 *         else:
 *             ewah_base = NULL
 */
    __pyx_v_ewah_base = ((__pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_array *)__pyx_v_base_mask->ewah_keys);

    /* "yt/geometry/particle_oct_container.pyx":1368
 *         cdef ewah_bool_array *ewah_slct = <ewah_bool_array *> selector_mask.ewah_keys
 *         cdef ewah_bool_array *ewah_base
 *         if base_mask is not None:             # <<<<<<<<<<<<<<
 *             ewah_base = <ewah_bool_array *> base_mask.ewah_keys
 *         else:
 */
    goto __pyx_L5;
  }

  /* "yt/geometry/particle_oct_container.pyx":1371
 *             ewah_base = <ewah_bool_array *> base_mask.ewah_keys
 *         else:
 *             ewah_base = NULL             # <<<<<<<<<<<<<<
 *         cdef ewah_bool_iterator *iter_set = new ewah_bool_iterator(ewah_slct[0].begin())
 *         cdef ewah_bool_iterator *iter_end = new ewah_bool_iterator(ewah_slct[0].end())
 */
  /*else*/ {
    __pyx_v_ewah_base = NULL;
  }
  __pyx_L5:;

  /* "yt/geometry/particle_oct_container.pyx":1372
 *         else:
 *             ewah_base = NULL
 *         cdef ewah_bool_iterator *iter_set = new ewah_bool_iterator(ewah_slct[0].begin())             # <<<<<<<<<<<<<<
 *         cdef ewah_bool_iterator *iter_end = new ewah_bool_iterator(ewah_slct[0].end())
 *         cdef np.ndarray[np.uint8_t, ndim=1] slct_arr
 */
  __pyx_v_iter_set = new __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_iterator((__pyx_v_ewah_slct[0]).begin());

  /* "yt/geometry/particle_oct_container.pyx":1373
 *             ewah_base = NULL
 *         cdef ewah_bool_iterator *iter_set = new ewah_bool_iterator(ewah_slct[0].begin())
 *         cdef ewah_bool_iterator *iter_end = new ewah_bool_iterator(ewah_slct[0].end())             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.uint8_t, ndim=1] slct_arr
 *         slct_arr = np.zeros((1 << (self.index_order1 * 3)),'uint8')
 */
  __pyx_v_iter_end = new __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_ewah_bool_iterator((__pyx_v_ewah_slct[0]).end());

  /* "yt/geometry/particle_oct_container.pyx":1375
 *         cdef ewah_bool_iterator *iter_end = new ewah_bool_iterator(ewah_slct[0].end())
 *         cdef np.ndarray[np.uint8_t, ndim=1] slct_arr
 *         slct_arr = np.zeros((1 << (self.index_order1 * 3)),'uint8')             # <<<<<<<<<<<<<<
 *         while iter_set[0] != iter_end[0]:
 *             mi = dereference(iter_set[0])
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_From_long((1 << (__pyx_v_self->index_order1 * 3))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_7)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_4, __pyx_n_s_uint8};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1375, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
    PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_4, __pyx_n_s_uint8};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1375, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_8, __pyx_t_4);
    __Pyx_INCREF(__pyx_n_s_uint8);
    __Pyx_GIVEREF(__pyx_n_s_uint8);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_8, __pyx_n_s_uint8);
    __pyx_t_4 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1375, __pyx_L1_error)
  __pyx_t_12 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_slct_arr.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_slct_arr.rcbuffer->pybuffer, (PyObject*)__pyx_t_12, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_slct_arr.rcbuffer->pybuffer, (PyObject*)__pyx_v_slct_arr, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_13); Py_XDECREF(__pyx_t_14); Py_XDECREF(__pyx_t_15);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      }
      __pyx_t_13 = __pyx_t_14 = __pyx_t_15 = 0;
    }
    __pyx_pybuffernd_slct_arr.diminfo[0].strides = __pyx_pybuffernd_slct_arr.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_slct_arr.diminfo[0].shape = __pyx_pybuffernd_slct_arr.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 1375, __pyx_L1_error)
  }
  __pyx_t_12 = 0;
  __pyx_v_slct_arr = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1376
 *         cdef np.ndarray[np.uint8_t, ndim=1] slct_arr
 *         slct_arr = np.zeros((1 << (self.index_order1 * 3)),'uint8')
 *         while iter_set[0] != iter_end[0]:             # <<<<<<<<<<<<<<
 *             mi = dereference(iter_set[0])
 *             if ewah_base != NULL and ewah_base[0].get(mi) == 0:
 */
  while (1) {
    __pyx_t_11 = (((__pyx_v_iter_set[0]) != (__pyx_v_iter_end[0])) != 0);
    if (!__pyx_t_11) break;

    /* "yt/geometry/particle_oct_container.pyx":1377
 *         slct_arr = np.zeros((1 << (self.index_order1 * 3)),'uint8')
 *         while iter_set[0] != iter_end[0]:
 *             mi = dereference(iter_set[0])             # <<<<<<<<<<<<<<
 *             if ewah_base != NULL and ewah_base[0].get(mi) == 0:
 *                 octree._index_base_roots[croot] = 0
 */
    __pyx_v_mi = (*(__pyx_v_iter_set[0]));

    /* "yt/geometry/particle_oct_container.pyx":1378
 *         while iter_set[0] != iter_end[0]:
 *             mi = dereference(iter_set[0])
 *             if ewah_base != NULL and ewah_base[0].get(mi) == 0:             # <<<<<<<<<<<<<<
 *                 octree._index_base_roots[croot] = 0
 *                 slct_arr[mi] = 2
 */
    __pyx_t_10 = ((__pyx_v_ewah_base != NULL) != 0);
    if (__pyx_t_10) {
    } else {
      __pyx_t_11 = __pyx_t_10;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_10 = (((__pyx_v_ewah_base[0]).get(__pyx_v_mi) == 0) != 0);
    __pyx_t_11 = __pyx_t_10;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_11) {

      /* "yt/geometry/particle_oct_container.pyx":1379
 *             mi = dereference(iter_set[0])
 *             if ewah_base != NULL and ewah_base[0].get(mi) == 0:
 *                 octree._index_base_roots[croot] = 0             # <<<<<<<<<<<<<<
 *                 slct_arr[mi] = 2
 *             else:
 */
      if (unlikely(!__pyx_v_octree->_index_base_roots.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1379, __pyx_L1_error)}
      __pyx_t_1 = __pyx_v_croot;
      *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_octree->_index_base_roots.data + __pyx_t_1 * __pyx_v_octree->_index_base_roots.strides[0]) )) = 0;

      /* "yt/geometry/particle_oct_container.pyx":1380
 *             if ewah_base != NULL and ewah_base[0].get(mi) == 0:
 *                 octree._index_base_roots[croot] = 0
 *                 slct_arr[mi] = 2             # <<<<<<<<<<<<<<
 *             else:
 *                 slct_arr[mi] = 1
 */
      __pyx_t_1 = __pyx_v_mi;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_slct_arr.rcbuffer->pybuffer.buf, __pyx_t_1, __pyx_pybuffernd_slct_arr.diminfo[0].strides) = 2;

      /* "yt/geometry/particle_oct_container.pyx":1378
 *         while iter_set[0] != iter_end[0]:
 *             mi = dereference(iter_set[0])
 *             if ewah_base != NULL and ewah_base[0].get(mi) == 0:             # <<<<<<<<<<<<<<
 *                 octree._index_base_roots[croot] = 0
 *                 slct_arr[mi] = 2
 */
      goto __pyx_L8;
    }

    /* "yt/geometry/particle_oct_container.pyx":1382
 *                 slct_arr[mi] = 2
 *             else:
 *                 slct_arr[mi] = 1             # <<<<<<<<<<<<<<
 *             decode_morton_64bit(mi, ind64)
 *             for j in range(3):
 */
    /*else*/ {
      __pyx_t_1 = __pyx_v_mi;
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_slct_arr.rcbuffer->pybuffer.buf, __pyx_t_1, __pyx_pybuffernd_slct_arr.diminfo[0].strides) = 1;
    }
    __pyx_L8:;

    /* "yt/geometry/particle_oct_container.pyx":1383
 *             else:
 *                 slct_arr[mi] = 1
 *             decode_morton_64bit(mi, ind64)             # <<<<<<<<<<<<<<
 *             for j in range(3):
 *                 ind[j] = ind64[j]
 */
    __pyx_f_2yt_9utilities_3lib_14geometry_utils_decode_morton_64bit(__pyx_v_mi, __pyx_v_ind64);

    /* "yt/geometry/particle_oct_container.pyx":1384
 *                 slct_arr[mi] = 1
 *             decode_morton_64bit(mi, ind64)
 *             for j in range(3):             # <<<<<<<<<<<<<<
 *                 ind[j] = ind64[j]
 *             octree.next_root(1, ind)
 */
    for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
      __pyx_v_j = __pyx_t_1;

      /* "yt/geometry/particle_oct_container.pyx":1385
 *             decode_morton_64bit(mi, ind64)
 *             for j in range(3):
 *                 ind[j] = ind64[j]             # <<<<<<<<<<<<<<
 *             octree.next_root(1, ind)
 *             croot += 1
 */
      (__pyx_v_ind[__pyx_v_j]) = (__pyx_v_ind64[__pyx_v_j]);
    }

    /* "yt/geometry/particle_oct_container.pyx":1386
 *             for j in range(3):
 *                 ind[j] = ind64[j]
 *             octree.next_root(1, ind)             # <<<<<<<<<<<<<<
 *             croot += 1
 *             preincrement(iter_set[0])
 */
    (void)(((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_octree->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.next_root(((struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *)__pyx_v_octree), 1, __pyx_v_ind));

    /* "yt/geometry/particle_oct_container.pyx":1387
 *                 ind[j] = ind64[j]
 *             octree.next_root(1, ind)
 *             croot += 1             # <<<<<<<<<<<<<<
 *             preincrement(iter_set[0])
 *         assert(croot == nroot)
 */
    __pyx_v_croot = (__pyx_v_croot + 1);

    /* "yt/geometry/particle_oct_container.pyx":1388
 *             octree.next_root(1, ind)
 *             croot += 1
 *             preincrement(iter_set[0])             # <<<<<<<<<<<<<<
 *         assert(croot == nroot)
 *         if ewah_base != NULL:
 */
    (void)((++(__pyx_v_iter_set[0])));
  }

  /* "yt/geometry/particle_oct_container.pyx":1389
 *             croot += 1
 *             preincrement(iter_set[0])
 *         assert(croot == nroot)             # <<<<<<<<<<<<<<
 *         if ewah_base != NULL:
 *             assert(np.sum(octree._index_base_roots) == ewah_base[0].numberOfOnes())
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!((__pyx_v_croot == __pyx_v_nroot) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 1389, __pyx_L1_error)
    }
  }
  #endif

  /* "yt/geometry/particle_oct_container.pyx":1390
 *             preincrement(iter_set[0])
 *         assert(croot == nroot)
 *         if ewah_base != NULL:             # <<<<<<<<<<<<<<
 *             assert(np.sum(octree._index_base_roots) == ewah_base[0].numberOfOnes())
 *         # Get morton indices for all particles in this file and those
 */
  __pyx_t_11 = ((__pyx_v_ewah_base != NULL) != 0);
  if (__pyx_t_11) {

    /* "yt/geometry/particle_oct_container.pyx":1391
 *         assert(croot == nroot)
 *         if ewah_base != NULL:
 *             assert(np.sum(octree._index_base_roots) == ewah_base[0].numberOfOnes())             # <<<<<<<<<<<<<<
 *         # Get morton indices for all particles in this file and those
 *         # contaminating cells it has majority control of.
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_sum); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_v_octree->_index_base_roots.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1391, __pyx_L1_error)}
      __pyx_t_7 = __pyx_memoryview_fromslice(__pyx_v_octree->_index_base_roots, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint8_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint8_t, 0);; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_7);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyInt_FromSize_t((__pyx_v_ewah_base[0]).numberOfOnes()); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = PyObject_RichCompare(__pyx_t_3, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1391, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_11)) {
        PyErr_SetNone(PyExc_AssertionError);
        __PYX_ERR(0, 1391, __pyx_L1_error)
      }
    }
    #endif

    /* "yt/geometry/particle_oct_container.pyx":1390
 *             preincrement(iter_set[0])
 *         assert(croot == nroot)
 *         if ewah_base != NULL:             # <<<<<<<<<<<<<<
 *             assert(np.sum(octree._index_base_roots) == ewah_base[0].numberOfOnes())
 *         # Get morton indices for all particles in this file and those
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":1394
 *         # Get morton indices for all particles in this file and those
 *         # contaminating cells it has majority control of.
 *         files_touched = data_files #+ buffer_files  # datafile object from ID goes here             # <<<<<<<<<<<<<<
 *         total_pcount = 0
 *         for data_file in files_touched:
 */
  __Pyx_INCREF(__pyx_v_data_files);
  __pyx_v_files_touched = __pyx_v_data_files;

  /* "yt/geometry/particle_oct_container.pyx":1395
 *         # contaminating cells it has majority control of.
 *         files_touched = data_files #+ buffer_files  # datafile object from ID goes here
 *         total_pcount = 0             # <<<<<<<<<<<<<<
 *         for data_file in files_touched:
 *             total_pcount += sum(data_file.total_particles.values())
 */
  __pyx_v_total_pcount = 0;

  /* "yt/geometry/particle_oct_container.pyx":1396
 *         files_touched = data_files #+ buffer_files  # datafile object from ID goes here
 *         total_pcount = 0
 *         for data_file in files_touched:             # <<<<<<<<<<<<<<
 *             total_pcount += sum(data_file.total_particles.values())
 *         morton_ind = np.empty(total_pcount, dtype='uint64')
 */
  if (likely(PyList_CheckExact(__pyx_v_files_touched)) || PyTuple_CheckExact(__pyx_v_files_touched)) {
    __pyx_t_7 = __pyx_v_files_touched; __Pyx_INCREF(__pyx_t_7); __pyx_t_16 = 0;
    __pyx_t_17 = NULL;
  } else {
    __pyx_t_16 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_v_files_touched); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1396, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_17 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1396, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_17)) {
      if (likely(PyList_CheckExact(__pyx_t_7))) {
        if (__pyx_t_16 >= PyList_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_16); __Pyx_INCREF(__pyx_t_5); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 1396, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_7, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1396, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_16 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_16); __Pyx_INCREF(__pyx_t_5); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 1396, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_7, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1396, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_17(__pyx_t_7);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1396, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_data_file, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "yt/geometry/particle_oct_container.pyx":1397
 *         total_pcount = 0
 *         for data_file in files_touched:
 *             total_pcount += sum(data_file.total_particles.values())             # <<<<<<<<<<<<<<
 *         morton_ind = np.empty(total_pcount, dtype='uint64')
 *         total_pcount = 0
 */
    __pyx_t_5 = __Pyx_PyInt_From_npy_uint64(__pyx_v_total_pcount); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data_file, __pyx_n_s_total_particles); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_values); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyInt_As_npy_uint64(__pyx_t_3); if (unlikely((__pyx_t_1 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1397, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_total_pcount = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":1396
 *         files_touched = data_files #+ buffer_files  # datafile object from ID goes here
 *         total_pcount = 0
 *         for data_file in files_touched:             # <<<<<<<<<<<<<<
 *             total_pcount += sum(data_file.total_particles.values())
 *         morton_ind = np.empty(total_pcount, dtype='uint64')
 */
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1398
 *         for data_file in files_touched:
 *             total_pcount += sum(data_file.total_particles.values())
 *         morton_ind = np.empty(total_pcount, dtype='uint64')             # <<<<<<<<<<<<<<
 *         total_pcount = 0
 *         cdef np.uint64_t base_pcount = 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_From_npy_uint64(__pyx_v_total_pcount); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7);
  __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_dtype, __pyx_n_s_uint64) < 0) __PYX_ERR(0, 1398, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1398, __pyx_L1_error)
  __pyx_t_18 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_morton_ind.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_morton_ind.rcbuffer->pybuffer, (PyObject*)__pyx_t_18, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_15, &__pyx_t_14, &__pyx_t_13);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_morton_ind.rcbuffer->pybuffer, (PyObject*)__pyx_v_morton_ind, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_15); Py_XDECREF(__pyx_t_14); Py_XDECREF(__pyx_t_13);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_15, __pyx_t_14, __pyx_t_13);
      }
      __pyx_t_15 = __pyx_t_14 = __pyx_t_13 = 0;
    }
    __pyx_pybuffernd_morton_ind.diminfo[0].strides = __pyx_pybuffernd_morton_ind.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_morton_ind.diminfo[0].shape = __pyx_pybuffernd_morton_ind.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 1398, __pyx_L1_error)
  }
  __pyx_t_18 = 0;
  __pyx_v_morton_ind = ((PyArrayObject *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1399
 *             total_pcount += sum(data_file.total_particles.values())
 *         morton_ind = np.empty(total_pcount, dtype='uint64')
 *         total_pcount = 0             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t base_pcount = 0
 *         for data_file in files_touched:
 */
  __pyx_v_total_pcount = 0;

  /* "yt/geometry/particle_oct_container.pyx":1400
 *         morton_ind = np.empty(total_pcount, dtype='uint64')
 *         total_pcount = 0
 *         cdef np.uint64_t base_pcount = 0             # <<<<<<<<<<<<<<
 *         for data_file in files_touched:
 *             # We now get our particle positions
 */
  __pyx_v_base_pcount = 0;

  /* "yt/geometry/particle_oct_container.pyx":1401
 *         total_pcount = 0
 *         cdef np.uint64_t base_pcount = 0
 *         for data_file in files_touched:             # <<<<<<<<<<<<<<
 *             # We now get our particle positions
 *             for pos in io_handler._yield_coordinates(data_file):
 */
  if (likely(PyList_CheckExact(__pyx_v_files_touched)) || PyTuple_CheckExact(__pyx_v_files_touched)) {
    __pyx_t_5 = __pyx_v_files_touched; __Pyx_INCREF(__pyx_t_5); __pyx_t_16 = 0;
    __pyx_t_17 = NULL;
  } else {
    __pyx_t_16 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_files_touched); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_17 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1401, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_17)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_16 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_16); __Pyx_INCREF(__pyx_t_7); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 1401, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_5, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1401, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_16 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_16); __Pyx_INCREF(__pyx_t_7); __pyx_t_16++; if (unlikely(0 < 0)) __PYX_ERR(0, 1401, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_5, __pyx_t_16); __pyx_t_16++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1401, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_17(__pyx_t_5);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1401, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    __Pyx_XDECREF_SET(__pyx_v_data_file, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "yt/geometry/particle_oct_container.pyx":1403
 *         for data_file in files_touched:
 *             # We now get our particle positions
 *             for pos in io_handler._yield_coordinates(data_file):             # <<<<<<<<<<<<<<
 *                 pos32 = pos64 = None
 *                 bitsize = 0
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_io_handler, __pyx_n_s_yield_coordinates); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_7 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_3, __pyx_v_data_file) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_data_file);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
      __pyx_t_6 = __pyx_t_7; __Pyx_INCREF(__pyx_t_6); __pyx_t_19 = 0;
      __pyx_t_20 = NULL;
    } else {
      __pyx_t_19 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1403, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_20 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 1403, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    for (;;) {
      if (likely(!__pyx_t_20)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_19 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_19); __Pyx_INCREF(__pyx_t_7); __pyx_t_19++; if (unlikely(0 < 0)) __PYX_ERR(0, 1403, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_6, __pyx_t_19); __pyx_t_19++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1403, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        } else {
          if (__pyx_t_19 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_19); __Pyx_INCREF(__pyx_t_7); __pyx_t_19++; if (unlikely(0 < 0)) __PYX_ERR(0, 1403, __pyx_L1_error)
          #else
          __pyx_t_7 = PySequence_ITEM(__pyx_t_6, __pyx_t_19); __pyx_t_19++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1403, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
        }
      } else {
        __pyx_t_7 = __pyx_t_20(__pyx_t_6);
        if (unlikely(!__pyx_t_7)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1403, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_7);
      }
      if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1403, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_pos, ((PyArrayObject *)__pyx_t_7));
      __pyx_t_7 = 0;

      /* "yt/geometry/particle_oct_container.pyx":1404
 *             # We now get our particle positions
 *             for pos in io_handler._yield_coordinates(data_file):
 *                 pos32 = pos64 = None             # <<<<<<<<<<<<<<
 *                 bitsize = 0
 *                 if pos.dtype == np.float32:
 */
      __pyx_t_21 = ((PyArrayObject *)Py_None);
      {
        __Pyx_BufFmt_StackElem __pyx_stack[1];
        __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos32.rcbuffer->pybuffer);
        __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pos32.rcbuffer->pybuffer, (PyObject*)__pyx_t_21, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
        if (unlikely(__pyx_t_8 < 0)) {
          PyErr_Fetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
          if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pos32.rcbuffer->pybuffer, (PyObject*)__pyx_v_pos32, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
            Py_XDECREF(__pyx_t_13); Py_XDECREF(__pyx_t_14); Py_XDECREF(__pyx_t_15);
            __Pyx_RaiseBufferFallbackError();
          } else {
            PyErr_Restore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
          }
          __pyx_t_13 = __pyx_t_14 = __pyx_t_15 = 0;
        }
        __pyx_pybuffernd_pos32.diminfo[0].strides = __pyx_pybuffernd_pos32.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_pos32.diminfo[0].shape = __pyx_pybuffernd_pos32.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_pos32.diminfo[1].strides = __pyx_pybuffernd_pos32.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_pos32.diminfo[1].shape = __pyx_pybuffernd_pos32.rcbuffer->pybuffer.shape[1];
        if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 1404, __pyx_L1_error)
      }
      __pyx_t_21 = 0;
      __Pyx_INCREF(Py_None);
      __Pyx_XDECREF_SET(__pyx_v_pos32, ((PyArrayObject *)Py_None));
      __pyx_t_22 = ((PyArrayObject *)Py_None);
      {
        __Pyx_BufFmt_StackElem __pyx_stack[1];
        __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos64.rcbuffer->pybuffer);
        __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pos64.rcbuffer->pybuffer, (PyObject*)__pyx_t_22, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
        if (unlikely(__pyx_t_8 < 0)) {
          PyErr_Fetch(&__pyx_t_15, &__pyx_t_14, &__pyx_t_13);
          if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pos64.rcbuffer->pybuffer, (PyObject*)__pyx_v_pos64, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
            Py_XDECREF(__pyx_t_15); Py_XDECREF(__pyx_t_14); Py_XDECREF(__pyx_t_13);
            __Pyx_RaiseBufferFallbackError();
          } else {
            PyErr_Restore(__pyx_t_15, __pyx_t_14, __pyx_t_13);
          }
          __pyx_t_15 = __pyx_t_14 = __pyx_t_13 = 0;
        }
        __pyx_pybuffernd_pos64.diminfo[0].strides = __pyx_pybuffernd_pos64.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_pos64.diminfo[0].shape = __pyx_pybuffernd_pos64.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_pos64.diminfo[1].strides = __pyx_pybuffernd_pos64.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_pos64.diminfo[1].shape = __pyx_pybuffernd_pos64.rcbuffer->pybuffer.shape[1];
        if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 1404, __pyx_L1_error)
      }
      __pyx_t_22 = 0;
      __Pyx_INCREF(Py_None);
      __Pyx_XDECREF_SET(__pyx_v_pos64, ((PyArrayObject *)Py_None));

      /* "yt/geometry/particle_oct_container.pyx":1405
 *             for pos in io_handler._yield_coordinates(data_file):
 *                 pos32 = pos64 = None
 *                 bitsize = 0             # <<<<<<<<<<<<<<
 *                 if pos.dtype == np.float32:
 *                     pos32 = pos
 */
      __pyx_v_bitsize = 0;

      /* "yt/geometry/particle_oct_container.pyx":1406
 *                 pos32 = pos64 = None
 *                 bitsize = 0
 *                 if pos.dtype == np.float32:             # <<<<<<<<<<<<<<
 *                     pos32 = pos
 *                     bitsize = 32
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_pos), __pyx_n_s_dtype); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1406, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1406, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float32); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1406, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyObject_RichCompare(__pyx_t_7, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1406, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1406, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_11) {

        /* "yt/geometry/particle_oct_container.pyx":1407
 *                 bitsize = 0
 *                 if pos.dtype == np.float32:
 *                     pos32 = pos             # <<<<<<<<<<<<<<
 *                     bitsize = 32
 *                     for j in range(pos.shape[0]):
 */
        {
          __Pyx_BufFmt_StackElem __pyx_stack[1];
          __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos32.rcbuffer->pybuffer);
          __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pos32.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_v_pos), &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
          if (unlikely(__pyx_t_8 < 0)) {
            PyErr_Fetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
            if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pos32.rcbuffer->pybuffer, (PyObject*)__pyx_v_pos32, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float32_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
              Py_XDECREF(__pyx_t_13); Py_XDECREF(__pyx_t_14); Py_XDECREF(__pyx_t_15);
              __Pyx_RaiseBufferFallbackError();
            } else {
              PyErr_Restore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
            }
            __pyx_t_13 = __pyx_t_14 = __pyx_t_15 = 0;
          }
          __pyx_pybuffernd_pos32.diminfo[0].strides = __pyx_pybuffernd_pos32.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_pos32.diminfo[0].shape = __pyx_pybuffernd_pos32.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_pos32.diminfo[1].strides = __pyx_pybuffernd_pos32.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_pos32.diminfo[1].shape = __pyx_pybuffernd_pos32.rcbuffer->pybuffer.shape[1];
          if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 1407, __pyx_L1_error)
        }
        __Pyx_INCREF(((PyObject *)__pyx_v_pos));
        __Pyx_DECREF_SET(__pyx_v_pos32, ((PyArrayObject *)__pyx_v_pos));

        /* "yt/geometry/particle_oct_container.pyx":1408
 *                 if pos.dtype == np.float32:
 *                     pos32 = pos
 *                     bitsize = 32             # <<<<<<<<<<<<<<
 *                     for j in range(pos.shape[0]):
 *                         for k in range(3):
 */
        __pyx_v_bitsize = 32;

        /* "yt/geometry/particle_oct_container.pyx":1409
 *                     pos32 = pos
 *                     bitsize = 32
 *                     for j in range(pos.shape[0]):             # <<<<<<<<<<<<<<
 *                         for k in range(3):
 *                             ppos[k] = pos32[j,k]
 */
        __pyx_t_23 = (__pyx_v_pos->dimensions[0]);
        __pyx_t_24 = __pyx_t_23;
        for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_24; __pyx_t_1+=1) {
          __pyx_v_j = __pyx_t_1;

          /* "yt/geometry/particle_oct_container.pyx":1410
 *                     bitsize = 32
 *                     for j in range(pos.shape[0]):
 *                         for k in range(3):             # <<<<<<<<<<<<<<
 *                             ppos[k] = pos32[j,k]
 *                         mi = bounded_morton(ppos[0], ppos[1], ppos[2],
 */
          for (__pyx_t_25 = 0; __pyx_t_25 < 3; __pyx_t_25+=1) {
            __pyx_v_k = __pyx_t_25;

            /* "yt/geometry/particle_oct_container.pyx":1411
 *                     for j in range(pos.shape[0]):
 *                         for k in range(3):
 *                             ppos[k] = pos32[j,k]             # <<<<<<<<<<<<<<
 *                         mi = bounded_morton(ppos[0], ppos[1], ppos[2],
 *                                             DLE, DRE, ORDER_MAX)
 */
            __pyx_t_26 = __pyx_v_j;
            __pyx_t_27 = __pyx_v_k;
            (__pyx_v_ppos[__pyx_v_k]) = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float32_t *, __pyx_pybuffernd_pos32.rcbuffer->pybuffer.buf, __pyx_t_26, __pyx_pybuffernd_pos32.diminfo[0].strides, __pyx_t_27, __pyx_pybuffernd_pos32.diminfo[1].strides));
          }

          /* "yt/geometry/particle_oct_container.pyx":1412
 *                         for k in range(3):
 *                             ppos[k] = pos32[j,k]
 *                         mi = bounded_morton(ppos[0], ppos[1], ppos[2],             # <<<<<<<<<<<<<<
 *                                             DLE, DRE, ORDER_MAX)
 *                         mi_root = mi >> (3*(ORDER_MAX-self.index_order1))
 */
          __pyx_v_mi = __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton((__pyx_v_ppos[0]), (__pyx_v_ppos[1]), (__pyx_v_ppos[2]), __pyx_v_DLE, __pyx_v_DRE, __pyx_v_2yt_8geometry_13oct_container_ORDER_MAX);

          /* "yt/geometry/particle_oct_container.pyx":1414
 *                         mi = bounded_morton(ppos[0], ppos[1], ppos[2],
 *                                             DLE, DRE, ORDER_MAX)
 *                         mi_root = mi >> (3*(ORDER_MAX-self.index_order1))             # <<<<<<<<<<<<<<
 *                         if slct_arr[mi_root] > 0:
 *                             morton_ind[total_pcount] = mi
 */
          __pyx_v_mi_root = (__pyx_v_mi >> (3 * (__pyx_v_2yt_8geometry_13oct_container_ORDER_MAX - __pyx_v_self->index_order1)));

          /* "yt/geometry/particle_oct_container.pyx":1415
 *                                             DLE, DRE, ORDER_MAX)
 *                         mi_root = mi >> (3*(ORDER_MAX-self.index_order1))
 *                         if slct_arr[mi_root] > 0:             # <<<<<<<<<<<<<<
 *                             morton_ind[total_pcount] = mi
 *                             total_pcount += 1
 */
          __pyx_t_25 = __pyx_v_mi_root;
          __pyx_t_11 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_slct_arr.rcbuffer->pybuffer.buf, __pyx_t_25, __pyx_pybuffernd_slct_arr.diminfo[0].strides)) > 0) != 0);
          if (__pyx_t_11) {

            /* "yt/geometry/particle_oct_container.pyx":1416
 *                         mi_root = mi >> (3*(ORDER_MAX-self.index_order1))
 *                         if slct_arr[mi_root] > 0:
 *                             morton_ind[total_pcount] = mi             # <<<<<<<<<<<<<<
 *                             total_pcount += 1
 *                             if slct_arr[mi_root] == 1:
 */
            __pyx_t_25 = __pyx_v_total_pcount;
            *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_morton_ind.rcbuffer->pybuffer.buf, __pyx_t_25, __pyx_pybuffernd_morton_ind.diminfo[0].strides) = __pyx_v_mi;

            /* "yt/geometry/particle_oct_container.pyx":1417
 *                         if slct_arr[mi_root] > 0:
 *                             morton_ind[total_pcount] = mi
 *                             total_pcount += 1             # <<<<<<<<<<<<<<
 *                             if slct_arr[mi_root] == 1:
 *                                 base_pcount += 1
 */
            __pyx_v_total_pcount = (__pyx_v_total_pcount + 1);

            /* "yt/geometry/particle_oct_container.pyx":1418
 *                             morton_ind[total_pcount] = mi
 *                             total_pcount += 1
 *                             if slct_arr[mi_root] == 1:             # <<<<<<<<<<<<<<
 *                                 base_pcount += 1
 *                 elif pos.dtype == np.float64:
 */
            __pyx_t_25 = __pyx_v_mi_root;
            __pyx_t_11 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_slct_arr.rcbuffer->pybuffer.buf, __pyx_t_25, __pyx_pybuffernd_slct_arr.diminfo[0].strides)) == 1) != 0);
            if (__pyx_t_11) {

              /* "yt/geometry/particle_oct_container.pyx":1419
 *                             total_pcount += 1
 *                             if slct_arr[mi_root] == 1:
 *                                 base_pcount += 1             # <<<<<<<<<<<<<<
 *                 elif pos.dtype == np.float64:
 *                     pos64 = pos
 */
              __pyx_v_base_pcount = (__pyx_v_base_pcount + 1);

              /* "yt/geometry/particle_oct_container.pyx":1418
 *                             morton_ind[total_pcount] = mi
 *                             total_pcount += 1
 *                             if slct_arr[mi_root] == 1:             # <<<<<<<<<<<<<<
 *                                 base_pcount += 1
 *                 elif pos.dtype == np.float64:
 */
            }

            /* "yt/geometry/particle_oct_container.pyx":1415
 *                                             DLE, DRE, ORDER_MAX)
 *                         mi_root = mi >> (3*(ORDER_MAX-self.index_order1))
 *                         if slct_arr[mi_root] > 0:             # <<<<<<<<<<<<<<
 *                             morton_ind[total_pcount] = mi
 *                             total_pcount += 1
 */
          }
        }

        /* "yt/geometry/particle_oct_container.pyx":1406
 *                 pos32 = pos64 = None
 *                 bitsize = 0
 *                 if pos.dtype == np.float32:             # <<<<<<<<<<<<<<
 *                     pos32 = pos
 *                     bitsize = 32
 */
        goto __pyx_L20;
      }

      /* "yt/geometry/particle_oct_container.pyx":1420
 *                             if slct_arr[mi_root] == 1:
 *                                 base_pcount += 1
 *                 elif pos.dtype == np.float64:             # <<<<<<<<<<<<<<
 *                     pos64 = pos
 *                     bitsize = 64
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_pos), __pyx_n_s_dtype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1420, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1420, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_float64); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1420, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1420, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1420, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (likely(__pyx_t_11)) {

        /* "yt/geometry/particle_oct_container.pyx":1421
 *                                 base_pcount += 1
 *                 elif pos.dtype == np.float64:
 *                     pos64 = pos             # <<<<<<<<<<<<<<
 *                     bitsize = 64
 *                     for j in range(pos.shape[0]):
 */
        {
          __Pyx_BufFmt_StackElem __pyx_stack[1];
          __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos64.rcbuffer->pybuffer);
          __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pos64.rcbuffer->pybuffer, (PyObject*)((PyArrayObject *)__pyx_v_pos), &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack);
          if (unlikely(__pyx_t_8 < 0)) {
            PyErr_Fetch(&__pyx_t_15, &__pyx_t_14, &__pyx_t_13);
            if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_pos64.rcbuffer->pybuffer, (PyObject*)__pyx_v_pos64, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 2, 0, __pyx_stack) == -1)) {
              Py_XDECREF(__pyx_t_15); Py_XDECREF(__pyx_t_14); Py_XDECREF(__pyx_t_13);
              __Pyx_RaiseBufferFallbackError();
            } else {
              PyErr_Restore(__pyx_t_15, __pyx_t_14, __pyx_t_13);
            }
            __pyx_t_15 = __pyx_t_14 = __pyx_t_13 = 0;
          }
          __pyx_pybuffernd_pos64.diminfo[0].strides = __pyx_pybuffernd_pos64.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_pos64.diminfo[0].shape = __pyx_pybuffernd_pos64.rcbuffer->pybuffer.shape[0]; __pyx_pybuffernd_pos64.diminfo[1].strides = __pyx_pybuffernd_pos64.rcbuffer->pybuffer.strides[1]; __pyx_pybuffernd_pos64.diminfo[1].shape = __pyx_pybuffernd_pos64.rcbuffer->pybuffer.shape[1];
          if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 1421, __pyx_L1_error)
        }
        __Pyx_INCREF(((PyObject *)__pyx_v_pos));
        __Pyx_DECREF_SET(__pyx_v_pos64, ((PyArrayObject *)__pyx_v_pos));

        /* "yt/geometry/particle_oct_container.pyx":1422
 *                 elif pos.dtype == np.float64:
 *                     pos64 = pos
 *                     bitsize = 64             # <<<<<<<<<<<<<<
 *                     for j in range(pos.shape[0]):
 *                         for k in range(3):
 */
        __pyx_v_bitsize = 64;

        /* "yt/geometry/particle_oct_container.pyx":1423
 *                     pos64 = pos
 *                     bitsize = 64
 *                     for j in range(pos.shape[0]):             # <<<<<<<<<<<<<<
 *                         for k in range(3):
 *                             ppos[k] = pos64[j,k]
 */
        __pyx_t_23 = (__pyx_v_pos->dimensions[0]);
        __pyx_t_24 = __pyx_t_23;
        for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_24; __pyx_t_1+=1) {
          __pyx_v_j = __pyx_t_1;

          /* "yt/geometry/particle_oct_container.pyx":1424
 *                     bitsize = 64
 *                     for j in range(pos.shape[0]):
 *                         for k in range(3):             # <<<<<<<<<<<<<<
 *                             ppos[k] = pos64[j,k]
 *                         mi = bounded_morton(ppos[0], ppos[1], ppos[2],
 */
          for (__pyx_t_25 = 0; __pyx_t_25 < 3; __pyx_t_25+=1) {
            __pyx_v_k = __pyx_t_25;

            /* "yt/geometry/particle_oct_container.pyx":1425
 *                     for j in range(pos.shape[0]):
 *                         for k in range(3):
 *                             ppos[k] = pos64[j,k]             # <<<<<<<<<<<<<<
 *                         mi = bounded_morton(ppos[0], ppos[1], ppos[2],
 *                                             DLE, DRE, ORDER_MAX)
 */
            __pyx_t_27 = __pyx_v_j;
            __pyx_t_26 = __pyx_v_k;
            (__pyx_v_ppos[__pyx_v_k]) = (*__Pyx_BufPtrStrided2d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_pos64.rcbuffer->pybuffer.buf, __pyx_t_27, __pyx_pybuffernd_pos64.diminfo[0].strides, __pyx_t_26, __pyx_pybuffernd_pos64.diminfo[1].strides));
          }

          /* "yt/geometry/particle_oct_container.pyx":1426
 *                         for k in range(3):
 *                             ppos[k] = pos64[j,k]
 *                         mi = bounded_morton(ppos[0], ppos[1], ppos[2],             # <<<<<<<<<<<<<<
 *                                             DLE, DRE, ORDER_MAX)
 *                         mi_root = mi >> (3*(ORDER_MAX-self.index_order1))
 */
          __pyx_v_mi = __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton((__pyx_v_ppos[0]), (__pyx_v_ppos[1]), (__pyx_v_ppos[2]), __pyx_v_DLE, __pyx_v_DRE, __pyx_v_2yt_8geometry_13oct_container_ORDER_MAX);

          /* "yt/geometry/particle_oct_container.pyx":1428
 *                         mi = bounded_morton(ppos[0], ppos[1], ppos[2],
 *                                             DLE, DRE, ORDER_MAX)
 *                         mi_root = mi >> (3*(ORDER_MAX-self.index_order1))             # <<<<<<<<<<<<<<
 *                         if slct_arr[mi_root] > 0:
 *                             morton_ind[total_pcount] = mi
 */
          __pyx_v_mi_root = (__pyx_v_mi >> (3 * (__pyx_v_2yt_8geometry_13oct_container_ORDER_MAX - __pyx_v_self->index_order1)));

          /* "yt/geometry/particle_oct_container.pyx":1429
 *                                             DLE, DRE, ORDER_MAX)
 *                         mi_root = mi >> (3*(ORDER_MAX-self.index_order1))
 *                         if slct_arr[mi_root] > 0:             # <<<<<<<<<<<<<<
 *                             morton_ind[total_pcount] = mi
 *                             total_pcount += 1
 */
          __pyx_t_25 = __pyx_v_mi_root;
          __pyx_t_11 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_slct_arr.rcbuffer->pybuffer.buf, __pyx_t_25, __pyx_pybuffernd_slct_arr.diminfo[0].strides)) > 0) != 0);
          if (__pyx_t_11) {

            /* "yt/geometry/particle_oct_container.pyx":1430
 *                         mi_root = mi >> (3*(ORDER_MAX-self.index_order1))
 *                         if slct_arr[mi_root] > 0:
 *                             morton_ind[total_pcount] = mi             # <<<<<<<<<<<<<<
 *                             total_pcount += 1
 *                             if slct_arr[mi_root] == 1:
 */
            __pyx_t_25 = __pyx_v_total_pcount;
            *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_morton_ind.rcbuffer->pybuffer.buf, __pyx_t_25, __pyx_pybuffernd_morton_ind.diminfo[0].strides) = __pyx_v_mi;

            /* "yt/geometry/particle_oct_container.pyx":1431
 *                         if slct_arr[mi_root] > 0:
 *                             morton_ind[total_pcount] = mi
 *                             total_pcount += 1             # <<<<<<<<<<<<<<
 *                             if slct_arr[mi_root] == 1:
 *                                 base_pcount += 1
 */
            __pyx_v_total_pcount = (__pyx_v_total_pcount + 1);

            /* "yt/geometry/particle_oct_container.pyx":1432
 *                             morton_ind[total_pcount] = mi
 *                             total_pcount += 1
 *                             if slct_arr[mi_root] == 1:             # <<<<<<<<<<<<<<
 *                                 base_pcount += 1
 *                 else:
 */
            __pyx_t_25 = __pyx_v_mi_root;
            __pyx_t_11 = (((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_slct_arr.rcbuffer->pybuffer.buf, __pyx_t_25, __pyx_pybuffernd_slct_arr.diminfo[0].strides)) == 1) != 0);
            if (__pyx_t_11) {

              /* "yt/geometry/particle_oct_container.pyx":1433
 *                             total_pcount += 1
 *                             if slct_arr[mi_root] == 1:
 *                                 base_pcount += 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise RuntimeError
 */
              __pyx_v_base_pcount = (__pyx_v_base_pcount + 1);

              /* "yt/geometry/particle_oct_container.pyx":1432
 *                             morton_ind[total_pcount] = mi
 *                             total_pcount += 1
 *                             if slct_arr[mi_root] == 1:             # <<<<<<<<<<<<<<
 *                                 base_pcount += 1
 *                 else:
 */
            }

            /* "yt/geometry/particle_oct_container.pyx":1429
 *                                             DLE, DRE, ORDER_MAX)
 *                         mi_root = mi >> (3*(ORDER_MAX-self.index_order1))
 *                         if slct_arr[mi_root] > 0:             # <<<<<<<<<<<<<<
 *                             morton_ind[total_pcount] = mi
 *                             total_pcount += 1
 */
          }
        }

        /* "yt/geometry/particle_oct_container.pyx":1420
 *                             if slct_arr[mi_root] == 1:
 *                                 base_pcount += 1
 *                 elif pos.dtype == np.float64:             # <<<<<<<<<<<<<<
 *                     pos64 = pos
 *                     bitsize = 64
 */
        goto __pyx_L20;
      }

      /* "yt/geometry/particle_oct_container.pyx":1435
 *                                 base_pcount += 1
 *                 else:
 *                     raise RuntimeError             # <<<<<<<<<<<<<<
 *         morton_ind = morton_ind[:total_pcount]
 *         morton_ind.sort()
 */
      /*else*/ {
        __Pyx_Raise(__pyx_builtin_RuntimeError, 0, 0, 0);
        __PYX_ERR(0, 1435, __pyx_L1_error)
      }
      __pyx_L20:;

      /* "yt/geometry/particle_oct_container.pyx":1403
 *         for data_file in files_touched:
 *             # We now get our particle positions
 *             for pos in io_handler._yield_coordinates(data_file):             # <<<<<<<<<<<<<<
 *                 pos32 = pos64 = None
 *                 bitsize = 0
 */
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "yt/geometry/particle_oct_container.pyx":1401
 *         total_pcount = 0
 *         cdef np.uint64_t base_pcount = 0
 *         for data_file in files_touched:             # <<<<<<<<<<<<<<
 *             # We now get our particle positions
 *             for pos in io_handler._yield_coordinates(data_file):
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1436
 *                 else:
 *                     raise RuntimeError
 *         morton_ind = morton_ind[:total_pcount]             # <<<<<<<<<<<<<<
 *         morton_ind.sort()
 *         octree.add(morton_ind, self.index_order1)
 */
  __pyx_t_5 = __Pyx_PyInt_From_npy_uint64(__pyx_v_total_pcount); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PySlice_New(Py_None, __pyx_t_5, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_morton_ind), __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1436, __pyx_L1_error)
  __pyx_t_18 = ((PyArrayObject *)__pyx_t_5);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_morton_ind.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_morton_ind.rcbuffer->pybuffer, (PyObject*)__pyx_t_18, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_morton_ind.rcbuffer->pybuffer, (PyObject*)__pyx_v_morton_ind, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_13); Py_XDECREF(__pyx_t_14); Py_XDECREF(__pyx_t_15);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      }
      __pyx_t_13 = __pyx_t_14 = __pyx_t_15 = 0;
    }
    __pyx_pybuffernd_morton_ind.diminfo[0].strides = __pyx_pybuffernd_morton_ind.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_morton_ind.diminfo[0].shape = __pyx_pybuffernd_morton_ind.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 1436, __pyx_L1_error)
  }
  __pyx_t_18 = 0;
  __Pyx_DECREF_SET(__pyx_v_morton_ind, ((PyArrayObject *)__pyx_t_5));
  __pyx_t_5 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1437
 *                     raise RuntimeError
 *         morton_ind = morton_ind[:total_pcount]
 *         morton_ind.sort()             # <<<<<<<<<<<<<<
 *         octree.add(morton_ind, self.index_order1)
 *         octree.finalize()
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_morton_ind), __pyx_n_s_sort); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1438
 *         morton_ind = morton_ind[:total_pcount]
 *         morton_ind.sort()
 *         octree.add(morton_ind, self.index_order1)             # <<<<<<<<<<<<<<
 *         octree.finalize()
 *         return octree
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_octree), __pyx_n_s_add); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->index_order1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, ((PyObject *)__pyx_v_morton_ind), __pyx_t_4};
    __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1438, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
    PyObject *__pyx_temp[3] = {__pyx_t_7, ((PyObject *)__pyx_v_morton_ind), __pyx_t_4};
    __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1438, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_3 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_7) {
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_morton_ind));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_morton_ind));
    PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_8, ((PyObject *)__pyx_v_morton_ind));
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_8, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_3, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1439
 *         morton_ind.sort()
 *         octree.add(morton_ind, self.index_order1)
 *         octree.finalize()             # <<<<<<<<<<<<<<
 *         return octree
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_octree), __pyx_n_s_finalize); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1440
 *         octree.add(morton_ind, self.index_order1)
 *         octree.finalize()
 *         return octree             # <<<<<<<<<<<<<<
 * 
 * cdef class ParticleBitmapSelector:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_octree));
  __pyx_r = ((PyObject *)__pyx_v_octree);
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1330
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def construct_octree(self, index, io_handler, data_files,             # <<<<<<<<<<<<<<
 *                          num_zones,
 *                          BoolArrayCollection selector_mask,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_morton_ind.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos32.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos64.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_slct_arr.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.construct_octree", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_morton_ind.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos32.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_pos64.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_slct_arr.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_octree);
  __Pyx_XDECREF((PyObject *)__pyx_v_pos);
  __Pyx_XDECREF((PyObject *)__pyx_v_pos32);
  __Pyx_XDECREF((PyObject *)__pyx_v_pos64);
  __Pyx_XDECREF((PyObject *)__pyx_v_morton_ind);
  __Pyx_XDECREF((PyObject *)__pyx_v_slct_arr);
  __Pyx_XDECREF(__pyx_v_files_touched);
  __Pyx_XDECREF(__pyx_v_data_file);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":444
 *     cdef np.int64_t file_hash
 *     cdef np.uint64_t directional_max2[3]
 *     cdef public np.uint64_t nfiles             # <<<<<<<<<<<<<<
 *     cdef public np.int32_t index_order1
 *     cdef public np.int32_t index_order2
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6nfiles_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6nfiles_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6nfiles___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6nfiles___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->nfiles); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.nfiles.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6nfiles_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6nfiles_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6nfiles_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6nfiles_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_uint64_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_npy_uint64(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 444, __pyx_L1_error)
  __pyx_v_self->nfiles = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.nfiles.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":445
 *     cdef np.uint64_t directional_max2[3]
 *     cdef public np.uint64_t nfiles
 *     cdef public np.int32_t index_order1             # <<<<<<<<<<<<<<
 *     cdef public np.int32_t index_order2
 *     cdef public object masks
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order1_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order1_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order1___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order1___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->index_order1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.index_order1.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order1_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order1_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order1_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order1_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_int32_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_npy_int32(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 445, __pyx_L1_error)
  __pyx_v_self->index_order1 = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.index_order1.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":446
 *     cdef public np.uint64_t nfiles
 *     cdef public np.int32_t index_order1
 *     cdef public np.int32_t index_order2             # <<<<<<<<<<<<<<
 *     cdef public object masks
 *     cdef public object particle_counts
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order2_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order2_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order2___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order2___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_npy_int32(__pyx_v_self->index_order2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.index_order2.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order2_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order2_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order2_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order2_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_int32_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_npy_int32(__pyx_v_value); if (unlikely((__pyx_t_1 == ((npy_int32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 446, __pyx_L1_error)
  __pyx_v_self->index_order2 = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.index_order2.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":447
 *     cdef public np.int32_t index_order1
 *     cdef public np.int32_t index_order2
 *     cdef public object masks             # <<<<<<<<<<<<<<
 *     cdef public object particle_counts
 *     cdef public object counts
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5masks_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5masks_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5masks___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5masks___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->masks);
  __pyx_r = __pyx_v_self->masks;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5masks_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5masks_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5masks_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5masks_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->masks);
  __Pyx_DECREF(__pyx_v_self->masks);
  __pyx_v_self->masks = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5masks_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5masks_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5masks_4__del__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5masks_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->masks);
  __Pyx_DECREF(__pyx_v_self->masks);
  __pyx_v_self->masks = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":448
 *     cdef public np.int32_t index_order2
 *     cdef public object masks
 *     cdef public object particle_counts             # <<<<<<<<<<<<<<
 *     cdef public object counts
 *     cdef public object max_count
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15particle_counts_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15particle_counts_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15particle_counts___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15particle_counts___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->particle_counts);
  __pyx_r = __pyx_v_self->particle_counts;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15particle_counts_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15particle_counts_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15particle_counts_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15particle_counts_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->particle_counts);
  __Pyx_DECREF(__pyx_v_self->particle_counts);
  __pyx_v_self->particle_counts = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15particle_counts_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15particle_counts_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15particle_counts_4__del__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15particle_counts_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->particle_counts);
  __Pyx_DECREF(__pyx_v_self->particle_counts);
  __pyx_v_self->particle_counts = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":449
 *     cdef public object masks
 *     cdef public object particle_counts
 *     cdef public object counts             # <<<<<<<<<<<<<<
 *     cdef public object max_count
 *     cdef public object _last_selector
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6counts_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6counts_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6counts___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6counts___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->counts);
  __pyx_r = __pyx_v_self->counts;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6counts_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6counts_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6counts_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6counts_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->counts);
  __Pyx_DECREF(__pyx_v_self->counts);
  __pyx_v_self->counts = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6counts_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6counts_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6counts_4__del__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6counts_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->counts);
  __Pyx_DECREF(__pyx_v_self->counts);
  __pyx_v_self->counts = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":450
 *     cdef public object particle_counts
 *     cdef public object counts
 *     cdef public object max_count             # <<<<<<<<<<<<<<
 *     cdef public object _last_selector
 *     cdef public object _last_return_values
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9max_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9max_count_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9max_count___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9max_count___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->max_count);
  __pyx_r = __pyx_v_self->max_count;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9max_count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9max_count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9max_count_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9max_count_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->max_count);
  __Pyx_DECREF(__pyx_v_self->max_count);
  __pyx_v_self->max_count = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9max_count_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9max_count_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9max_count_4__del__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9max_count_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->max_count);
  __Pyx_DECREF(__pyx_v_self->max_count);
  __pyx_v_self->max_count = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":451
 *     cdef public object counts
 *     cdef public object max_count
 *     cdef public object _last_selector             # <<<<<<<<<<<<<<
 *     cdef public object _last_return_values
 *     cdef public object _cached_octrees
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_last_selector_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_last_selector_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_last_selector___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_last_selector___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_last_selector);
  __pyx_r = __pyx_v_self->_last_selector;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_last_selector_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_last_selector_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_last_selector_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_last_selector_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_last_selector);
  __Pyx_DECREF(__pyx_v_self->_last_selector);
  __pyx_v_self->_last_selector = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_last_selector_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_last_selector_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_last_selector_4__del__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_last_selector_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_last_selector);
  __Pyx_DECREF(__pyx_v_self->_last_selector);
  __pyx_v_self->_last_selector = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":452
 *     cdef public object max_count
 *     cdef public object _last_selector
 *     cdef public object _last_return_values             # <<<<<<<<<<<<<<
 *     cdef public object _cached_octrees
 *     cdef public object _last_octree_subset
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_return_values_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_return_values_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_return_values___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_return_values___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_last_return_values);
  __pyx_r = __pyx_v_self->_last_return_values;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_return_values_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_return_values_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_return_values_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_return_values_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_last_return_values);
  __Pyx_DECREF(__pyx_v_self->_last_return_values);
  __pyx_v_self->_last_return_values = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_return_values_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_return_values_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_return_values_4__del__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_return_values_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_last_return_values);
  __Pyx_DECREF(__pyx_v_self->_last_return_values);
  __pyx_v_self->_last_return_values = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":453
 *     cdef public object _last_selector
 *     cdef public object _last_return_values
 *     cdef public object _cached_octrees             # <<<<<<<<<<<<<<
 *     cdef public object _last_octree_subset
 *     cdef public object _last_oct_handler
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_cached_octrees_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_cached_octrees_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_cached_octrees___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_cached_octrees___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_cached_octrees);
  __pyx_r = __pyx_v_self->_cached_octrees;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_cached_octrees_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_cached_octrees_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_cached_octrees_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_cached_octrees_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_cached_octrees);
  __Pyx_DECREF(__pyx_v_self->_cached_octrees);
  __pyx_v_self->_cached_octrees = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_cached_octrees_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_cached_octrees_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_cached_octrees_4__del__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_cached_octrees_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_cached_octrees);
  __Pyx_DECREF(__pyx_v_self->_cached_octrees);
  __pyx_v_self->_cached_octrees = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":454
 *     cdef public object _last_return_values
 *     cdef public object _cached_octrees
 *     cdef public object _last_octree_subset             # <<<<<<<<<<<<<<
 *     cdef public object _last_oct_handler
 *     cdef public object _prev_octree_subset
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_octree_subset_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_octree_subset_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_octree_subset___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_octree_subset___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_last_octree_subset);
  __pyx_r = __pyx_v_self->_last_octree_subset;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_octree_subset_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_octree_subset_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_octree_subset_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_octree_subset_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_last_octree_subset);
  __Pyx_DECREF(__pyx_v_self->_last_octree_subset);
  __pyx_v_self->_last_octree_subset = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_octree_subset_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_octree_subset_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_octree_subset_4__del__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_octree_subset_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_last_octree_subset);
  __Pyx_DECREF(__pyx_v_self->_last_octree_subset);
  __pyx_v_self->_last_octree_subset = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":455
 *     cdef public object _cached_octrees
 *     cdef public object _last_octree_subset
 *     cdef public object _last_oct_handler             # <<<<<<<<<<<<<<
 *     cdef public object _prev_octree_subset
 *     cdef public object _prev_oct_handler
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_last_oct_handler_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_last_oct_handler_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_last_oct_handler___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_last_oct_handler___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_last_oct_handler);
  __pyx_r = __pyx_v_self->_last_oct_handler;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_last_oct_handler_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_last_oct_handler_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_last_oct_handler_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_last_oct_handler_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_last_oct_handler);
  __Pyx_DECREF(__pyx_v_self->_last_oct_handler);
  __pyx_v_self->_last_oct_handler = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_last_oct_handler_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_last_oct_handler_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_last_oct_handler_4__del__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_last_oct_handler_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_last_oct_handler);
  __Pyx_DECREF(__pyx_v_self->_last_oct_handler);
  __pyx_v_self->_last_oct_handler = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":456
 *     cdef public object _last_octree_subset
 *     cdef public object _last_oct_handler
 *     cdef public object _prev_octree_subset             # <<<<<<<<<<<<<<
 *     cdef public object _prev_oct_handler
 *     cdef np.uint32_t *file_markers
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_prev_octree_subset_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_prev_octree_subset_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_prev_octree_subset___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_prev_octree_subset___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_prev_octree_subset);
  __pyx_r = __pyx_v_self->_prev_octree_subset;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_prev_octree_subset_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_prev_octree_subset_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_prev_octree_subset_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_prev_octree_subset_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_prev_octree_subset);
  __Pyx_DECREF(__pyx_v_self->_prev_octree_subset);
  __pyx_v_self->_prev_octree_subset = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_prev_octree_subset_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_prev_octree_subset_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_prev_octree_subset_4__del__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_prev_octree_subset_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_prev_octree_subset);
  __Pyx_DECREF(__pyx_v_self->_prev_octree_subset);
  __pyx_v_self->_prev_octree_subset = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":457
 *     cdef public object _last_oct_handler
 *     cdef public object _prev_octree_subset
 *     cdef public object _prev_oct_handler             # <<<<<<<<<<<<<<
 *     cdef np.uint32_t *file_markers
 *     cdef np.uint64_t n_file_markers
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_prev_oct_handler_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_prev_oct_handler_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_prev_oct_handler___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_prev_oct_handler___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_prev_oct_handler);
  __pyx_r = __pyx_v_self->_prev_oct_handler;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_prev_oct_handler_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_prev_oct_handler_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_prev_oct_handler_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_prev_oct_handler_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_prev_oct_handler);
  __Pyx_DECREF(__pyx_v_self->_prev_oct_handler);
  __pyx_v_self->_prev_oct_handler = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_prev_oct_handler_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_prev_oct_handler_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_prev_oct_handler_4__del__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_prev_oct_handler_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_prev_oct_handler);
  __Pyx_DECREF(__pyx_v_self->_prev_oct_handler);
  __pyx_v_self->_prev_oct_handler = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":461
 *     cdef np.uint64_t n_file_markers
 *     cdef np.uint64_t file_marker_i
 *     cdef public FileBitmasks bitmasks             # <<<<<<<<<<<<<<
 *     cdef public BoolArrayCollection collisions
 *     cdef public int _used_mi2
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8bitmasks_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8bitmasks_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8bitmasks___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8bitmasks___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->bitmasks));
  __pyx_r = ((PyObject *)__pyx_v_self->bitmasks);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8bitmasks_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8bitmasks_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8bitmasks_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8bitmasks_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks))))) __PYX_ERR(0, 461, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->bitmasks);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->bitmasks));
  __pyx_v_self->bitmasks = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.bitmasks.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8bitmasks_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8bitmasks_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8bitmasks_4__del__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8bitmasks_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->bitmasks);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->bitmasks));
  __pyx_v_self->bitmasks = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":462
 *     cdef np.uint64_t file_marker_i
 *     cdef public FileBitmasks bitmasks
 *     cdef public BoolArrayCollection collisions             # <<<<<<<<<<<<<<
 *     cdef public int _used_mi2
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10collisions_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10collisions_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10collisions___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10collisions___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->collisions));
  __pyx_r = ((PyObject *)__pyx_v_self->collisions);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10collisions_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10collisions_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10collisions_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10collisions_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection))))) __PYX_ERR(0, 462, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->collisions);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->collisions));
  __pyx_v_self->collisions = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.collisions.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10collisions_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10collisions_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10collisions_4__del__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10collisions_4__del__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->collisions);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->collisions));
  __pyx_v_self->collisions = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":463
 *     cdef public FileBitmasks bitmasks
 *     cdef public BoolArrayCollection collisions
 *     cdef public int _used_mi2             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, left_edge, right_edge, periodicity, file_hash, nfiles,
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9_used_mi2_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9_used_mi2_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9_used_mi2___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9_used_mi2___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_used_mi2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap._used_mi2.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9_used_mi2_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9_used_mi2_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9_used_mi2_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9_used_mi2_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 463, __pyx_L1_error)
  __pyx_v_self->_used_mi2 = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap._used_mi2.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.file_markers cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_93__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_93__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_92__reduce_cython__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_92__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.file_markers cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.file_markers cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.file_markers cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.file_markers cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.file_markers cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_95__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_95__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_94__setstate_cython__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_14ParticleBitmap_94__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.file_markers cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.file_markers cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.file_markers cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.file_markers cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmap.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1473
 *     cdef BoolArrayColl ghosts_ewah
 * 
 *     def __cinit__(self, selector, bitmap, ngz=0):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.ndarray[np.uint8_t, ndim=1] periodicity = np.zeros(3, dtype='uint8')
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_selector = 0;
  PyObject *__pyx_v_bitmap = 0;
  PyObject *__pyx_v_ngz = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_selector,&__pyx_n_s_bitmap,&__pyx_n_s_ngz,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_selector)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bitmap)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, 1); __PYX_ERR(0, 1473, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ngz);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 1473, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_selector = values[0];
    __pyx_v_bitmap = values[1];
    __pyx_v_ngz = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1473, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapSelector.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector___cinit__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self), __pyx_v_selector, __pyx_v_bitmap, __pyx_v_ngz);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector___cinit__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, PyObject *__pyx_v_selector, PyObject *__pyx_v_bitmap, PyObject *__pyx_v_ngz) {
  int __pyx_v_i;
  PyArrayObject *__pyx_v_periodicity = 0;
  PyArrayObject *__pyx_v_DLE = 0;
  PyArrayObject *__pyx_v_DRE = 0;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_DLE;
  __Pyx_Buffer __pyx_pybuffer_DLE;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_DRE;
  __Pyx_Buffer __pyx_pybuffer_DRE;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_periodicity;
  __Pyx_Buffer __pyx_pybuffer_periodicity;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyArrayObject *__pyx_t_4 = NULL;
  PyArrayObject *__pyx_t_5 = NULL;
  PyArrayObject *__pyx_t_6 = NULL;
  __pyx_t_5numpy_uint32_t __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  Py_ssize_t __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  __Pyx_memviewslice __pyx_t_15 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_16 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_17 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_18 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __pyx_pybuffer_periodicity.pybuffer.buf = NULL;
  __pyx_pybuffer_periodicity.refcount = 0;
  __pyx_pybuffernd_periodicity.data = NULL;
  __pyx_pybuffernd_periodicity.rcbuffer = &__pyx_pybuffer_periodicity;
  __pyx_pybuffer_DLE.pybuffer.buf = NULL;
  __pyx_pybuffer_DLE.refcount = 0;
  __pyx_pybuffernd_DLE.data = NULL;
  __pyx_pybuffernd_DLE.rcbuffer = &__pyx_pybuffer_DLE;
  __pyx_pybuffer_DRE.pybuffer.buf = NULL;
  __pyx_pybuffer_DRE.refcount = 0;
  __pyx_pybuffernd_DRE.data = NULL;
  __pyx_pybuffernd_DRE.rcbuffer = &__pyx_pybuffer_DRE;

  /* "yt/geometry/particle_oct_container.pyx":1475
 *     def __cinit__(self, selector, bitmap, ngz=0):
 *         cdef int i
 *         cdef np.ndarray[np.uint8_t, ndim=1] periodicity = np.zeros(3, dtype='uint8')             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.float64_t, ndim=1] DLE = np.zeros(3, dtype='float64')
 *         cdef np.ndarray[np.float64_t, ndim=1] DRE = np.zeros(3, dtype='float64')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 1475, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__20, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1475, __pyx_L1_error)
  __pyx_t_4 = ((PyArrayObject *)__pyx_t_3);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_periodicity.rcbuffer->pybuffer, (PyObject*)__pyx_t_4, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_periodicity = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_periodicity.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 1475, __pyx_L1_error)
    } else {__pyx_pybuffernd_periodicity.diminfo[0].strides = __pyx_pybuffernd_periodicity.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_periodicity.diminfo[0].shape = __pyx_pybuffernd_periodicity.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_4 = 0;
  __pyx_v_periodicity = ((PyArrayObject *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1476
 *         cdef int i
 *         cdef np.ndarray[np.uint8_t, ndim=1] periodicity = np.zeros(3, dtype='uint8')
 *         cdef np.ndarray[np.float64_t, ndim=1] DLE = np.zeros(3, dtype='float64')             # <<<<<<<<<<<<<<
 *         cdef np.ndarray[np.float64_t, ndim=1] DRE = np.zeros(3, dtype='float64')
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_n_s_float64) < 0) __PYX_ERR(0, 1476, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__20, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1476, __pyx_L1_error)
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_2);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_DLE.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_DLE = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_DLE.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 1476, __pyx_L1_error)
    } else {__pyx_pybuffernd_DLE.diminfo[0].strides = __pyx_pybuffernd_DLE.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_DLE.diminfo[0].shape = __pyx_pybuffernd_DLE.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_5 = 0;
  __pyx_v_DLE = ((PyArrayObject *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1477
 *         cdef np.ndarray[np.uint8_t, ndim=1] periodicity = np.zeros(3, dtype='uint8')
 *         cdef np.ndarray[np.float64_t, ndim=1] DLE = np.zeros(3, dtype='float64')
 *         cdef np.ndarray[np.float64_t, ndim=1] DRE = np.zeros(3, dtype='float64')             # <<<<<<<<<<<<<<
 * 
 *         self.selector = selector
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1477, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1477, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1477, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_n_s_float64) < 0) __PYX_ERR(0, 1477, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__20, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1477, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1477, __pyx_L1_error)
  __pyx_t_6 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_DRE.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
      __pyx_v_DRE = ((PyArrayObject *)Py_None); __Pyx_INCREF(Py_None); __pyx_pybuffernd_DRE.rcbuffer->pybuffer.buf = NULL;
      __PYX_ERR(0, 1477, __pyx_L1_error)
    } else {__pyx_pybuffernd_DRE.diminfo[0].strides = __pyx_pybuffernd_DRE.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_DRE.diminfo[0].shape = __pyx_pybuffernd_DRE.rcbuffer->pybuffer.shape[0];
    }
  }
  __pyx_t_6 = 0;
  __pyx_v_DRE = ((PyArrayObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1479
 *         cdef np.ndarray[np.float64_t, ndim=1] DRE = np.zeros(3, dtype='float64')
 * 
 *         self.selector = selector             # <<<<<<<<<<<<<<
 *         self.bitmap = bitmap
 *         self.ngz = ngz
 */
  if (!(likely(((__pyx_v_selector) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_selector, __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject))))) __PYX_ERR(0, 1479, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_selector;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->selector);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->selector));
  __pyx_v_self->selector = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1480
 * 
 *         self.selector = selector
 *         self.bitmap = bitmap             # <<<<<<<<<<<<<<
 *         self.ngz = ngz
 *         # Things from the bitmap & selector
 */
  if (!(likely(((__pyx_v_bitmap) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_bitmap, __pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleBitmap))))) __PYX_ERR(0, 1480, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_bitmap;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->bitmap);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->bitmap));
  __pyx_v_self->bitmap = ((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1481
 *         self.selector = selector
 *         self.bitmap = bitmap
 *         self.ngz = ngz             # <<<<<<<<<<<<<<
 *         # Things from the bitmap & selector
 *         periodicity = selector.get_periodicity()
 */
  __pyx_t_7 = __Pyx_PyInt_As_npy_uint32(__pyx_v_ngz); if (unlikely((__pyx_t_7 == ((npy_uint32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1481, __pyx_L1_error)
  __pyx_v_self->ngz = __pyx_t_7;

  /* "yt/geometry/particle_oct_container.pyx":1483
 *         self.ngz = ngz
 *         # Things from the bitmap & selector
 *         periodicity = selector.get_periodicity()             # <<<<<<<<<<<<<<
 *         DLE = bitmap.get_DLE()
 *         DRE = bitmap.get_DRE()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_selector, __pyx_n_s_get_periodicity); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1483, __pyx_L1_error)
  __pyx_t_4 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_periodicity.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_periodicity.rcbuffer->pybuffer, (PyObject*)__pyx_t_4, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_periodicity.rcbuffer->pybuffer, (PyObject*)__pyx_v_periodicity, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
      __pyx_t_9 = __pyx_t_10 = __pyx_t_11 = 0;
    }
    __pyx_pybuffernd_periodicity.diminfo[0].strides = __pyx_pybuffernd_periodicity.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_periodicity.diminfo[0].shape = __pyx_pybuffernd_periodicity.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 1483, __pyx_L1_error)
  }
  __pyx_t_4 = 0;
  __Pyx_DECREF_SET(__pyx_v_periodicity, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1484
 *         # Things from the bitmap & selector
 *         periodicity = selector.get_periodicity()
 *         DLE = bitmap.get_DLE()             # <<<<<<<<<<<<<<
 *         DRE = bitmap.get_DRE()
 *         for i in range(3):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_bitmap, __pyx_n_s_get_DLE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1484, __pyx_L1_error)
  __pyx_t_5 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_DLE.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_DLE.rcbuffer->pybuffer, (PyObject*)__pyx_t_5, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_11, &__pyx_t_10, &__pyx_t_9);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_DLE.rcbuffer->pybuffer, (PyObject*)__pyx_v_DLE, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_11); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_9);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_11, __pyx_t_10, __pyx_t_9);
      }
      __pyx_t_11 = __pyx_t_10 = __pyx_t_9 = 0;
    }
    __pyx_pybuffernd_DLE.diminfo[0].strides = __pyx_pybuffernd_DLE.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_DLE.diminfo[0].shape = __pyx_pybuffernd_DLE.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 1484, __pyx_L1_error)
  }
  __pyx_t_5 = 0;
  __Pyx_DECREF_SET(__pyx_v_DLE, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1485
 *         periodicity = selector.get_periodicity()
 *         DLE = bitmap.get_DLE()
 *         DRE = bitmap.get_DRE()             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             self.DLE[i] = DLE[i]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_bitmap, __pyx_n_s_get_DRE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 1485, __pyx_L1_error)
  __pyx_t_6 = ((PyArrayObject *)__pyx_t_1);
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_DRE.rcbuffer->pybuffer);
    __pyx_t_8 = __Pyx_GetBufferAndValidate(&__pyx_pybuffernd_DRE.rcbuffer->pybuffer, (PyObject*)__pyx_t_6, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack);
    if (unlikely(__pyx_t_8 < 0)) {
      PyErr_Fetch(&__pyx_t_9, &__pyx_t_10, &__pyx_t_11);
      if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_DRE.rcbuffer->pybuffer, (PyObject*)__pyx_v_DRE, &__Pyx_TypeInfo_nn___pyx_t_5numpy_float64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) {
        Py_XDECREF(__pyx_t_9); Py_XDECREF(__pyx_t_10); Py_XDECREF(__pyx_t_11);
        __Pyx_RaiseBufferFallbackError();
      } else {
        PyErr_Restore(__pyx_t_9, __pyx_t_10, __pyx_t_11);
      }
      __pyx_t_9 = __pyx_t_10 = __pyx_t_11 = 0;
    }
    __pyx_pybuffernd_DRE.diminfo[0].strides = __pyx_pybuffernd_DRE.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_DRE.diminfo[0].shape = __pyx_pybuffernd_DRE.rcbuffer->pybuffer.shape[0];
    if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 1485, __pyx_L1_error)
  }
  __pyx_t_6 = 0;
  __Pyx_DECREF_SET(__pyx_v_DRE, ((PyArrayObject *)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1486
 *         DLE = bitmap.get_DLE()
 *         DRE = bitmap.get_DRE()
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             self.DLE[i] = DLE[i]
 *             self.DRE[i] = DRE[i]
 */
  for (__pyx_t_8 = 0; __pyx_t_8 < 3; __pyx_t_8+=1) {
    __pyx_v_i = __pyx_t_8;

    /* "yt/geometry/particle_oct_container.pyx":1487
 *         DRE = bitmap.get_DRE()
 *         for i in range(3):
 *             self.DLE[i] = DLE[i]             # <<<<<<<<<<<<<<
 *             self.DRE[i] = DRE[i]
 *             self.periodicity[i] = periodicity[i]
 */
    __pyx_t_12 = __pyx_v_i;
    __pyx_t_13 = -1;
    if (__pyx_t_12 < 0) {
      __pyx_t_12 += __pyx_pybuffernd_DLE.diminfo[0].shape;
      if (unlikely(__pyx_t_12 < 0)) __pyx_t_13 = 0;
    } else if (unlikely(__pyx_t_12 >= __pyx_pybuffernd_DLE.diminfo[0].shape)) __pyx_t_13 = 0;
    if (unlikely(__pyx_t_13 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_13);
      __PYX_ERR(0, 1487, __pyx_L1_error)
    }
    (__pyx_v_self->DLE[__pyx_v_i]) = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_DLE.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_DLE.diminfo[0].strides));

    /* "yt/geometry/particle_oct_container.pyx":1488
 *         for i in range(3):
 *             self.DLE[i] = DLE[i]
 *             self.DRE[i] = DRE[i]             # <<<<<<<<<<<<<<
 *             self.periodicity[i] = periodicity[i]
 *         self.order1 = bitmap.index_order1
 */
    __pyx_t_12 = __pyx_v_i;
    __pyx_t_13 = -1;
    if (__pyx_t_12 < 0) {
      __pyx_t_12 += __pyx_pybuffernd_DRE.diminfo[0].shape;
      if (unlikely(__pyx_t_12 < 0)) __pyx_t_13 = 0;
    } else if (unlikely(__pyx_t_12 >= __pyx_pybuffernd_DRE.diminfo[0].shape)) __pyx_t_13 = 0;
    if (unlikely(__pyx_t_13 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_13);
      __PYX_ERR(0, 1488, __pyx_L1_error)
    }
    (__pyx_v_self->DRE[__pyx_v_i]) = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_float64_t *, __pyx_pybuffernd_DRE.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_DRE.diminfo[0].strides));

    /* "yt/geometry/particle_oct_container.pyx":1489
 *             self.DLE[i] = DLE[i]
 *             self.DRE[i] = DRE[i]
 *             self.periodicity[i] = periodicity[i]             # <<<<<<<<<<<<<<
 *         self.order1 = bitmap.index_order1
 *         self.order2 = bitmap.index_order2
 */
    __pyx_t_12 = __pyx_v_i;
    __pyx_t_13 = -1;
    if (__pyx_t_12 < 0) {
      __pyx_t_12 += __pyx_pybuffernd_periodicity.diminfo[0].shape;
      if (unlikely(__pyx_t_12 < 0)) __pyx_t_13 = 0;
    } else if (unlikely(__pyx_t_12 >= __pyx_pybuffernd_periodicity.diminfo[0].shape)) __pyx_t_13 = 0;
    if (unlikely(__pyx_t_13 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_13);
      __PYX_ERR(0, 1489, __pyx_L1_error)
    }
    (__pyx_v_self->periodicity[__pyx_v_i]) = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_periodicity.rcbuffer->pybuffer.buf, __pyx_t_12, __pyx_pybuffernd_periodicity.diminfo[0].strides));
  }

  /* "yt/geometry/particle_oct_container.pyx":1490
 *             self.DRE[i] = DRE[i]
 *             self.periodicity[i] = periodicity[i]
 *         self.order1 = bitmap.index_order1             # <<<<<<<<<<<<<<
 *         self.order2 = bitmap.index_order2
 *         self.nfiles = bitmap.nfiles
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_bitmap, __pyx_n_s_index_order1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyInt_As_npy_uint32(__pyx_t_1); if (unlikely((__pyx_t_7 == ((npy_uint32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1490, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->order1 = __pyx_t_7;

  /* "yt/geometry/particle_oct_container.pyx":1491
 *             self.periodicity[i] = periodicity[i]
 *         self.order1 = bitmap.index_order1
 *         self.order2 = bitmap.index_order2             # <<<<<<<<<<<<<<
 *         self.nfiles = bitmap.nfiles
 *         self.max_index1 = <np.uint64_t>(1 << self.order1)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_bitmap, __pyx_n_s_index_order2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyInt_As_npy_uint32(__pyx_t_1); if (unlikely((__pyx_t_7 == ((npy_uint32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1491, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->order2 = __pyx_t_7;

  /* "yt/geometry/particle_oct_container.pyx":1492
 *         self.order1 = bitmap.index_order1
 *         self.order2 = bitmap.index_order2
 *         self.nfiles = bitmap.nfiles             # <<<<<<<<<<<<<<
 *         self.max_index1 = <np.uint64_t>(1 << self.order1)
 *         self.max_index2 = <np.uint64_t>(1 << self.order2)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_bitmap, __pyx_n_s_nfiles); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyInt_As_npy_uint32(__pyx_t_1); if (unlikely((__pyx_t_7 == ((npy_uint32)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1492, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->nfiles = __pyx_t_7;

  /* "yt/geometry/particle_oct_container.pyx":1493
 *         self.order2 = bitmap.index_order2
 *         self.nfiles = bitmap.nfiles
 *         self.max_index1 = <np.uint64_t>(1 << self.order1)             # <<<<<<<<<<<<<<
 *         self.max_index2 = <np.uint64_t>(1 << self.order2)
 *         self.s1 = <np.uint64_t>(1 << (self.order1*3))
 */
  __pyx_v_self->max_index1 = ((__pyx_t_5numpy_uint64_t)(1 << __pyx_v_self->order1));

  /* "yt/geometry/particle_oct_container.pyx":1494
 *         self.nfiles = bitmap.nfiles
 *         self.max_index1 = <np.uint64_t>(1 << self.order1)
 *         self.max_index2 = <np.uint64_t>(1 << self.order2)             # <<<<<<<<<<<<<<
 *         self.s1 = <np.uint64_t>(1 << (self.order1*3))
 *         self.s2 = <np.uint64_t>(1 << (self.order2*3))
 */
  __pyx_v_self->max_index2 = ((__pyx_t_5numpy_uint64_t)(1 << __pyx_v_self->order2));

  /* "yt/geometry/particle_oct_container.pyx":1495
 *         self.max_index1 = <np.uint64_t>(1 << self.order1)
 *         self.max_index2 = <np.uint64_t>(1 << self.order2)
 *         self.s1 = <np.uint64_t>(1 << (self.order1*3))             # <<<<<<<<<<<<<<
 *         self.s2 = <np.uint64_t>(1 << (self.order2*3))
 * 
 */
  __pyx_v_self->s1 = ((__pyx_t_5numpy_uint64_t)(1 << (__pyx_v_self->order1 * 3)));

  /* "yt/geometry/particle_oct_container.pyx":1496
 *         self.max_index2 = <np.uint64_t>(1 << self.order2)
 *         self.s1 = <np.uint64_t>(1 << (self.order1*3))
 *         self.s2 = <np.uint64_t>(1 << (self.order2*3))             # <<<<<<<<<<<<<<
 * 
 *         self.neighbors = np.zeros((2*ngz+1, 3), dtype='uint32')
 */
  __pyx_v_self->s2 = ((__pyx_t_5numpy_uint64_t)(1 << (__pyx_v_self->order2 * 3)));

  /* "yt/geometry/particle_oct_container.pyx":1498
 *         self.s2 = <np.uint64_t>(1 << (self.order2*3))
 * 
 *         self.neighbors = np.zeros((2*ngz+1, 3), dtype='uint32')             # <<<<<<<<<<<<<<
 *         self.ind1_n = np.zeros((2*ngz+1, 3), dtype='uint64')
 *         self.ind2_n = np.zeros((2*ngz+1, 3), dtype='uint64')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_int_2, __pyx_v_ngz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_uint32) < 0) __PYX_ERR(0, 1498, __pyx_L1_error)
  __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_15 = __Pyx_PyObject_to_MemoryviewSlice_dsds_nn___pyx_t_5numpy_uint32_t(__pyx_t_14, PyBUF_WRITABLE); if (unlikely(!__pyx_t_15.memview)) __PYX_ERR(0, 1498, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->neighbors, 0);
  __pyx_v_self->neighbors = __pyx_t_15;
  __pyx_t_15.memview = NULL;
  __pyx_t_15.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":1499
 * 
 *         self.neighbors = np.zeros((2*ngz+1, 3), dtype='uint32')
 *         self.ind1_n = np.zeros((2*ngz+1, 3), dtype='uint64')             # <<<<<<<<<<<<<<
 *         self.ind2_n = np.zeros((2*ngz+1, 3), dtype='uint64')
 *         self.neighbor_list1 = np.zeros((2*ngz+1)**3, dtype='uint64')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_np); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = PyNumber_Multiply(__pyx_int_2, __pyx_v_ngz); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_t_14, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_3);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_int_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_14);
  __pyx_t_14 = 0;
  __pyx_t_14 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  if (PyDict_SetItem(__pyx_t_14, __pyx_n_s_dtype, __pyx_n_s_uint64) < 0) __PYX_ERR(0, 1499, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_16 = __Pyx_PyObject_to_MemoryviewSlice_dsds_nn___pyx_t_5numpy_uint64_t(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_16.memview)) __PYX_ERR(0, 1499, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->ind1_n, 0);
  __pyx_v_self->ind1_n = __pyx_t_16;
  __pyx_t_16.memview = NULL;
  __pyx_t_16.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":1500
 *         self.neighbors = np.zeros((2*ngz+1, 3), dtype='uint32')
 *         self.ind1_n = np.zeros((2*ngz+1, 3), dtype='uint64')
 *         self.ind2_n = np.zeros((2*ngz+1, 3), dtype='uint64')             # <<<<<<<<<<<<<<
 *         self.neighbor_list1 = np.zeros((2*ngz+1)**3, dtype='uint64')
 *         self.neighbor_list2 = np.zeros((2*ngz+1)**3, dtype='uint64')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Multiply(__pyx_int_2, __pyx_v_ngz); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_t_2, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_n_s_uint64) < 0) __PYX_ERR(0, 1500, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_16 = __Pyx_PyObject_to_MemoryviewSlice_dsds_nn___pyx_t_5numpy_uint64_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_16.memview)) __PYX_ERR(0, 1500, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->ind2_n, 0);
  __pyx_v_self->ind2_n = __pyx_t_16;
  __pyx_t_16.memview = NULL;
  __pyx_t_16.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":1501
 *         self.ind1_n = np.zeros((2*ngz+1, 3), dtype='uint64')
 *         self.ind2_n = np.zeros((2*ngz+1, 3), dtype='uint64')
 *         self.neighbor_list1 = np.zeros((2*ngz+1)**3, dtype='uint64')             # <<<<<<<<<<<<<<
 *         self.neighbor_list2 = np.zeros((2*ngz+1)**3, dtype='uint64')
 *         self.file_mask_p = np.zeros(bitmap.nfiles, dtype='uint8')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_int_2, __pyx_v_ngz); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Power(__pyx_t_3, __pyx_int_3, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_uint64) < 0) __PYX_ERR(0, 1501, __pyx_L1_error)
  __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_17 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_14, PyBUF_WRITABLE); if (unlikely(!__pyx_t_17.memview)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->neighbor_list1, 0);
  __pyx_v_self->neighbor_list1 = __pyx_t_17;
  __pyx_t_17.memview = NULL;
  __pyx_t_17.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":1502
 *         self.ind2_n = np.zeros((2*ngz+1, 3), dtype='uint64')
 *         self.neighbor_list1 = np.zeros((2*ngz+1)**3, dtype='uint64')
 *         self.neighbor_list2 = np.zeros((2*ngz+1)**3, dtype='uint64')             # <<<<<<<<<<<<<<
 *         self.file_mask_p = np.zeros(bitmap.nfiles, dtype='uint8')
 *         self.file_mask_g = np.zeros(bitmap.nfiles, dtype='uint8')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_np); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = PyNumber_Multiply(__pyx_int_2, __pyx_v_ngz); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_t_14, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_14 = PyNumber_Power(__pyx_t_3, __pyx_int_3, Py_None); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_14);
  __pyx_t_14 = 0;
  __pyx_t_14 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  if (PyDict_SetItem(__pyx_t_14, __pyx_n_s_dtype, __pyx_n_s_uint64) < 0) __PYX_ERR(0, 1502, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_14); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __pyx_t_17 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_17.memview)) __PYX_ERR(0, 1502, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->neighbor_list2, 0);
  __pyx_v_self->neighbor_list2 = __pyx_t_17;
  __pyx_t_17.memview = NULL;
  __pyx_t_17.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":1503
 *         self.neighbor_list1 = np.zeros((2*ngz+1)**3, dtype='uint64')
 *         self.neighbor_list2 = np.zeros((2*ngz+1)**3, dtype='uint64')
 *         self.file_mask_p = np.zeros(bitmap.nfiles, dtype='uint8')             # <<<<<<<<<<<<<<
 *         self.file_mask_g = np.zeros(bitmap.nfiles, dtype='uint8')
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_bitmap, __pyx_n_s_nfiles); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 1503, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_18 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint8_t(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_18.memview)) __PYX_ERR(0, 1503, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->file_mask_p, 0);
  __pyx_v_self->file_mask_p = __pyx_t_18;
  __pyx_t_18.memview = NULL;
  __pyx_t_18.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":1504
 *         self.neighbor_list2 = np.zeros((2*ngz+1)**3, dtype='uint64')
 *         self.file_mask_p = np.zeros(bitmap.nfiles, dtype='uint8')
 *         self.file_mask_g = np.zeros(bitmap.nfiles, dtype='uint8')             # <<<<<<<<<<<<<<
 * 
 *         self.refined_select_bool = np.zeros(self.s2, 'uint8')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_bitmap, __pyx_n_s_nfiles); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_n_s_uint8) < 0) __PYX_ERR(0, 1504, __pyx_L1_error)
  __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_18 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint8_t(__pyx_t_14, PyBUF_WRITABLE); if (unlikely(!__pyx_t_18.memview)) __PYX_ERR(0, 1504, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->file_mask_g, 0);
  __pyx_v_self->file_mask_g = __pyx_t_18;
  __pyx_t_18.memview = NULL;
  __pyx_t_18.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":1506
 *         self.file_mask_g = np.zeros(bitmap.nfiles, dtype='uint8')
 * 
 *         self.refined_select_bool = np.zeros(self.s2, 'uint8')             # <<<<<<<<<<<<<<
 *         self.refined_ghosts_bool = np.zeros(self.s2, 'uint8')
 *         self.coarse_select_bool = np.zeros(self.s1, 'uint8')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->s2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_t_1, __pyx_n_s_uint8};
    __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1506, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_t_1, __pyx_n_s_uint8};
    __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1506, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  {
    __pyx_t_19 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1506, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_19);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_19, 0+__pyx_t_8, __pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_uint8);
    __Pyx_GIVEREF(__pyx_n_s_uint8);
    PyTuple_SET_ITEM(__pyx_t_19, 1+__pyx_t_8, __pyx_n_s_uint8);
    __pyx_t_1 = 0;
    __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_19, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1506, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_18 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint8_t(__pyx_t_14, PyBUF_WRITABLE); if (unlikely(!__pyx_t_18.memview)) __PYX_ERR(0, 1506, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->refined_select_bool, 0);
  __pyx_v_self->refined_select_bool = __pyx_t_18;
  __pyx_t_18.memview = NULL;
  __pyx_t_18.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":1507
 * 
 *         self.refined_select_bool = np.zeros(self.s2, 'uint8')
 *         self.refined_ghosts_bool = np.zeros(self.s2, 'uint8')             # <<<<<<<<<<<<<<
 *         self.coarse_select_bool = np.zeros(self.s1, 'uint8')
 *         self.coarse_ghosts_bool = np.zeros(self.s1, 'uint8')
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->s2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_19))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_19);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_19);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_19, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_19)) {
    PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_t_3, __pyx_n_s_uint8};
    __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_19, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1507, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_19)) {
    PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_t_3, __pyx_n_s_uint8};
    __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_19, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1507, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_2 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1507, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_1) {
      __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1); __pyx_t_1 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_8, __pyx_t_3);
    __Pyx_INCREF(__pyx_n_s_uint8);
    __Pyx_GIVEREF(__pyx_n_s_uint8);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_8, __pyx_n_s_uint8);
    __pyx_t_3 = 0;
    __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_19, __pyx_t_2, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1507, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_18 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint8_t(__pyx_t_14, PyBUF_WRITABLE); if (unlikely(!__pyx_t_18.memview)) __PYX_ERR(0, 1507, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->refined_ghosts_bool, 0);
  __pyx_v_self->refined_ghosts_bool = __pyx_t_18;
  __pyx_t_18.memview = NULL;
  __pyx_t_18.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":1508
 *         self.refined_select_bool = np.zeros(self.s2, 'uint8')
 *         self.refined_ghosts_bool = np.zeros(self.s2, 'uint8')
 *         self.coarse_select_bool = np.zeros(self.s1, 'uint8')             # <<<<<<<<<<<<<<
 *         self.coarse_ghosts_bool = np.zeros(self.s1, 'uint8')
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_19, __pyx_n_s_np); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_19, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  __pyx_t_19 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->s1); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_3 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_19, __pyx_n_s_uint8};
    __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1508, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_t_19, __pyx_n_s_uint8};
    __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1508, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  } else
  #endif
  {
    __pyx_t_1 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1508, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_19);
    PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_8, __pyx_t_19);
    __Pyx_INCREF(__pyx_n_s_uint8);
    __Pyx_GIVEREF(__pyx_n_s_uint8);
    PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_8, __pyx_n_s_uint8);
    __pyx_t_19 = 0;
    __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1508, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_18 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint8_t(__pyx_t_14, PyBUF_WRITABLE); if (unlikely(!__pyx_t_18.memview)) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->coarse_select_bool, 0);
  __pyx_v_self->coarse_select_bool = __pyx_t_18;
  __pyx_t_18.memview = NULL;
  __pyx_t_18.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":1509
 *         self.refined_ghosts_bool = np.zeros(self.s2, 'uint8')
 *         self.coarse_select_bool = np.zeros(self.s1, 'uint8')
 *         self.coarse_ghosts_bool = np.zeros(self.s1, 'uint8')             # <<<<<<<<<<<<<<
 * 
 *         self.refined_ghosts_list = SparseUnorderedRefinedBitmask()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_zeros); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->s1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_19 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_19 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_19)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_19);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_19, __pyx_t_2, __pyx_n_s_uint8};
    __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1509, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
    PyObject *__pyx_temp[3] = {__pyx_t_19, __pyx_t_2, __pyx_n_s_uint8};
    __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1509, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_3 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_19) {
      __Pyx_GIVEREF(__pyx_t_19); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_19); __pyx_t_19 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_8, __pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_uint8);
    __Pyx_GIVEREF(__pyx_n_s_uint8);
    PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_8, __pyx_n_s_uint8);
    __pyx_t_2 = 0;
    __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_18 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint8_t(__pyx_t_14, PyBUF_WRITABLE); if (unlikely(!__pyx_t_18.memview)) __PYX_ERR(0, 1509, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->coarse_ghosts_bool, 0);
  __pyx_v_self->coarse_ghosts_bool = __pyx_t_18;
  __pyx_t_18.memview = NULL;
  __pyx_t_18.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":1511
 *         self.coarse_ghosts_bool = np.zeros(self.s1, 'uint8')
 * 
 *         self.refined_ghosts_list = SparseUnorderedRefinedBitmask()             # <<<<<<<<<<<<<<
 *         self.select_ewah = BoolArrayColl(self.s1, self.s2)
 *         self.ghosts_ewah = BoolArrayColl(self.s1, self.s2)
 */
  __pyx_t_14 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet)); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_14);
  __Pyx_GOTREF(__pyx_v_self->refined_ghosts_list);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->refined_ghosts_list));
  __pyx_v_self->refined_ghosts_list = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet *)__pyx_t_14);
  __pyx_t_14 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1512
 * 
 *         self.refined_ghosts_list = SparseUnorderedRefinedBitmask()
 *         self.select_ewah = BoolArrayColl(self.s1, self.s2)             # <<<<<<<<<<<<<<
 *         self.ghosts_ewah = BoolArrayColl(self.s1, self.s2)
 * 
 */
  __pyx_t_14 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->s1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->s2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_14 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed), __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->select_ewah);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->select_ewah));
  __pyx_v_self->select_ewah = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1513
 *         self.refined_ghosts_list = SparseUnorderedRefinedBitmask()
 *         self.select_ewah = BoolArrayColl(self.s1, self.s2)
 *         self.ghosts_ewah = BoolArrayColl(self.s1, self.s2)             # <<<<<<<<<<<<<<
 * 
 *     def fill_masks(self, BoolArrayCollection mm_s, BoolArrayCollection mm_g = None):
 */
  __pyx_t_1 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->s1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->s2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_14 = PyTuple_New(2); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_14, 1, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed), __pyx_t_14, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->ghosts_ewah);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->ghosts_ewah));
  __pyx_v_self->ghosts_ewah = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1473
 *     cdef BoolArrayColl ghosts_ewah
 * 
 *     def __cinit__(self, selector, bitmap, ngz=0):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef np.ndarray[np.uint8_t, ndim=1] periodicity = np.zeros(3, dtype='uint8')
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_14);
  __PYX_XDEC_MEMVIEW(&__pyx_t_15, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_16, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_17, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_18, 1);
  __Pyx_XDECREF(__pyx_t_19);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_DLE.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_DRE.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_periodicity.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapSelector.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_DLE.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_DRE.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_periodicity.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_periodicity);
  __Pyx_XDECREF((PyObject *)__pyx_v_DLE);
  __Pyx_XDECREF((PyObject *)__pyx_v_DRE);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1515
 *         self.ghosts_ewah = BoolArrayColl(self.s1, self.s2)
 * 
 *     def fill_masks(self, BoolArrayCollection mm_s, BoolArrayCollection mm_g = None):             # <<<<<<<<<<<<<<
 *         # Normal variables
 *         cdef int i
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_3fill_masks(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_3fill_masks(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mm_s = 0;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mm_g = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fill_masks (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_mm_s,&__pyx_n_s_mm_g,0};
    PyObject* values[2] = {0,0};
    values[1] = (PyObject *)((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mm_s)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mm_g);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "fill_masks") < 0)) __PYX_ERR(0, 1515, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_mm_s = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)values[0]);
    __pyx_v_mm_g = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("fill_masks", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1515, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapSelector.fill_masks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mm_s), __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection, 1, "mm_s", 0))) __PYX_ERR(0, 1515, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_mm_g), __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection, 1, "mm_g", 0))) __PYX_ERR(0, 1515, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_2fill_masks(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self), __pyx_v_mm_s, __pyx_v_mm_g);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_2fill_masks(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mm_s, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *__pyx_v_mm_g) {
  int __pyx_v_i;
  __pyx_t_5numpy_int32_t __pyx_v_level;
  __pyx_t_5numpy_uint64_t __pyx_v_mi1;
  __pyx_t_5numpy_float64_t __pyx_v_pos[3];
  __pyx_t_5numpy_float64_t __pyx_v_dds[3];
  __pyx_t_5numpy_uint64_t __pyx_v_cur_ind[3];
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_s0 = 0;
  struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_g0 = 0;
  __pyx_t_5numpy_float64_t __pyx_v_rpos[3];
  int __pyx_v_sbbox;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __pyx_t_5numpy_uint64_t __pyx_t_7;
  __pyx_t_5numpy_uint64_t __pyx_t_8;
  __pyx_t_5numpy_uint64_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fill_masks", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_mm_g);

  /* "yt/geometry/particle_oct_container.pyx":1518
 *         # Normal variables
 *         cdef int i
 *         cdef np.int32_t level = 0             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t mi1
 *         mi1 = ~(<np.uint64_t>0)
 */
  __pyx_v_level = 0;

  /* "yt/geometry/particle_oct_container.pyx":1520
 *         cdef np.int32_t level = 0
 *         cdef np.uint64_t mi1
 *         mi1 = ~(<np.uint64_t>0)             # <<<<<<<<<<<<<<
 *         cdef np.float64_t pos[3]
 *         cdef np.float64_t dds[3]
 */
  __pyx_v_mi1 = (~((__pyx_t_5numpy_uint64_t)0));

  /* "yt/geometry/particle_oct_container.pyx":1524
 *         cdef np.float64_t dds[3]
 *         cdef np.uint64_t cur_ind[3]
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             cur_ind[i] = 0
 *             pos[i] = self.DLE[i]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":1525
 *         cdef np.uint64_t cur_ind[3]
 *         for i in range(3):
 *             cur_ind[i] = 0             # <<<<<<<<<<<<<<
 *             pos[i] = self.DLE[i]
 *             dds[i] = self.DRE[i] - self.DLE[i]
 */
    (__pyx_v_cur_ind[__pyx_v_i]) = 0;

    /* "yt/geometry/particle_oct_container.pyx":1526
 *         for i in range(3):
 *             cur_ind[i] = 0
 *             pos[i] = self.DLE[i]             # <<<<<<<<<<<<<<
 *             dds[i] = self.DRE[i] - self.DLE[i]
 *         if mm_g is None:
 */
    (__pyx_v_pos[__pyx_v_i]) = (__pyx_v_self->DLE[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":1527
 *             cur_ind[i] = 0
 *             pos[i] = self.DLE[i]
 *             dds[i] = self.DRE[i] - self.DLE[i]             # <<<<<<<<<<<<<<
 *         if mm_g is None:
 *             mm_g = BoolArrayCollection()
 */
    (__pyx_v_dds[__pyx_v_i]) = ((__pyx_v_self->DRE[__pyx_v_i]) - (__pyx_v_self->DLE[__pyx_v_i]));
  }

  /* "yt/geometry/particle_oct_container.pyx":1528
 *             pos[i] = self.DLE[i]
 *             dds[i] = self.DRE[i] - self.DLE[i]
 *         if mm_g is None:             # <<<<<<<<<<<<<<
 *             mm_g = BoolArrayCollection()
 *         # Uncompressed version
 */
  __pyx_t_2 = (((PyObject *)__pyx_v_mm_g) == Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "yt/geometry/particle_oct_container.pyx":1529
 *             dds[i] = self.DRE[i] - self.DLE[i]
 *         if mm_g is None:
 *             mm_g = BoolArrayCollection()             # <<<<<<<<<<<<<<
 *         # Uncompressed version
 *         cdef BoolArrayColl mm_s0
 */
    __pyx_t_4 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF_SET(__pyx_v_mm_g, ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "yt/geometry/particle_oct_container.pyx":1528
 *             pos[i] = self.DLE[i]
 *             dds[i] = self.DRE[i] - self.DLE[i]
 *         if mm_g is None:             # <<<<<<<<<<<<<<
 *             mm_g = BoolArrayCollection()
 *         # Uncompressed version
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":1533
 *         cdef BoolArrayColl mm_s0
 *         cdef BoolArrayColl mm_g0
 *         mm_s0 = BoolArrayColl(self.s1, self.s2)             # <<<<<<<<<<<<<<
 *         mm_g0 = BoolArrayColl(self.s1, self.s2)
 *         # Recurse
 */
  __pyx_t_4 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->s1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->s2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed), __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_mm_s0 = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1534
 *         cdef BoolArrayColl mm_g0
 *         mm_s0 = BoolArrayColl(self.s1, self.s2)
 *         mm_g0 = BoolArrayColl(self.s1, self.s2)             # <<<<<<<<<<<<<<
 *         # Recurse
 *         cdef np.float64_t rpos[3]
 */
  __pyx_t_5 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->s1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_npy_uint64(__pyx_v_self->s2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_6);
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed), __pyx_t_4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_mm_g0 = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1537
 *         # Recurse
 *         cdef np.float64_t rpos[3]
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             rpos[i] = self.DRE[i] - self.bitmap.dds_mi2[i]/2.0
 *         sbbox = self.selector.select_bbox_edge(pos, rpos)
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":1538
 *         cdef np.float64_t rpos[3]
 *         for i in range(3):
 *             rpos[i] = self.DRE[i] - self.bitmap.dds_mi2[i]/2.0             # <<<<<<<<<<<<<<
 *         sbbox = self.selector.select_bbox_edge(pos, rpos)
 *         if sbbox == 1:
 */
    (__pyx_v_rpos[__pyx_v_i]) = ((__pyx_v_self->DRE[__pyx_v_i]) - ((__pyx_v_self->bitmap->dds_mi2[__pyx_v_i]) / 2.0));
  }

  /* "yt/geometry/particle_oct_container.pyx":1539
 *         for i in range(3):
 *             rpos[i] = self.DRE[i] - self.bitmap.dds_mi2[i]/2.0
 *         sbbox = self.selector.select_bbox_edge(pos, rpos)             # <<<<<<<<<<<<<<
 *         if sbbox == 1:
 *             for mi1 in range(<np.uint64_t>self.s1):
 */
  __pyx_v_sbbox = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->selector->__pyx_vtab)->select_bbox_edge(__pyx_v_self->selector, __pyx_v_pos, __pyx_v_rpos);

  /* "yt/geometry/particle_oct_container.pyx":1540
 *             rpos[i] = self.DRE[i] - self.bitmap.dds_mi2[i]/2.0
 *         sbbox = self.selector.select_bbox_edge(pos, rpos)
 *         if sbbox == 1:             # <<<<<<<<<<<<<<
 *             for mi1 in range(<np.uint64_t>self.s1):
 *                 mm_s0._set_coarse(mi1)
 */
  __pyx_t_3 = ((__pyx_v_sbbox == 1) != 0);
  if (__pyx_t_3) {

    /* "yt/geometry/particle_oct_container.pyx":1541
 *         sbbox = self.selector.select_bbox_edge(pos, rpos)
 *         if sbbox == 1:
 *             for mi1 in range(<np.uint64_t>self.s1):             # <<<<<<<<<<<<<<
 *                 mm_s0._set_coarse(mi1)
 *             mm_s0._compress(mm_s)
 */
    __pyx_t_7 = ((__pyx_t_5numpy_uint64_t)__pyx_v_self->s1);
    __pyx_t_8 = __pyx_t_7;
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
      __pyx_v_mi1 = __pyx_t_9;

      /* "yt/geometry/particle_oct_container.pyx":1542
 *         if sbbox == 1:
 *             for mi1 in range(<np.uint64_t>self.s1):
 *                 mm_s0._set_coarse(mi1)             # <<<<<<<<<<<<<<
 *             mm_s0._compress(mm_s)
 *             return
 */
      ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_v_mm_s0->__pyx_vtab)->_set_coarse(__pyx_v_mm_s0, __pyx_v_mi1);
    }

    /* "yt/geometry/particle_oct_container.pyx":1543
 *             for mi1 in range(<np.uint64_t>self.s1):
 *                 mm_s0._set_coarse(mi1)
 *             mm_s0._compress(mm_s)             # <<<<<<<<<<<<<<
 *             return
 *         else:
 */
    ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_v_mm_s0->__pyx_vtab)->_compress(__pyx_v_mm_s0, __pyx_v_mm_s);

    /* "yt/geometry/particle_oct_container.pyx":1544
 *                 mm_s0._set_coarse(mi1)
 *             mm_s0._compress(mm_s)
 *             return             # <<<<<<<<<<<<<<
 *         else:
 *             self.recursive_morton_mask(level, pos, dds, mi1, cur_ind)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "yt/geometry/particle_oct_container.pyx":1540
 *             rpos[i] = self.DRE[i] - self.bitmap.dds_mi2[i]/2.0
 *         sbbox = self.selector.select_bbox_edge(pos, rpos)
 *         if sbbox == 1:             # <<<<<<<<<<<<<<
 *             for mi1 in range(<np.uint64_t>self.s1):
 *                 mm_s0._set_coarse(mi1)
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":1546
 *             return
 *         else:
 *             self.recursive_morton_mask(level, pos, dds, mi1, cur_ind)             # <<<<<<<<<<<<<<
 *         # Set coarse morton indices in order
 *         self.set_coarse_bool(mm_s0, mm_g0)
 */
  /*else*/ {
    __pyx_t_1 = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->recursive_morton_mask(__pyx_v_self, __pyx_v_level, __pyx_v_pos, __pyx_v_dds, __pyx_v_mi1, __pyx_v_cur_ind); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 1546, __pyx_L1_error)
  }

  /* "yt/geometry/particle_oct_container.pyx":1548
 *             self.recursive_morton_mask(level, pos, dds, mi1, cur_ind)
 *         # Set coarse morton indices in order
 *         self.set_coarse_bool(mm_s0, mm_g0)             # <<<<<<<<<<<<<<
 *         self.set_refined_list(mm_s0, mm_g0)
 *         self.set_refined_bool(mm_s0, mm_g0)
 */
  ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->set_coarse_bool(__pyx_v_self, __pyx_v_mm_s0, __pyx_v_mm_g0);

  /* "yt/geometry/particle_oct_container.pyx":1549
 *         # Set coarse morton indices in order
 *         self.set_coarse_bool(mm_s0, mm_g0)
 *         self.set_refined_list(mm_s0, mm_g0)             # <<<<<<<<<<<<<<
 *         self.set_refined_bool(mm_s0, mm_g0)
 *         # Compress
 */
  ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->set_refined_list(__pyx_v_self, __pyx_v_mm_s0, __pyx_v_mm_g0);

  /* "yt/geometry/particle_oct_container.pyx":1550
 *         self.set_coarse_bool(mm_s0, mm_g0)
 *         self.set_refined_list(mm_s0, mm_g0)
 *         self.set_refined_bool(mm_s0, mm_g0)             # <<<<<<<<<<<<<<
 *         # Compress
 *         mm_s0._compress(mm_s)
 */
  ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->set_refined_bool(__pyx_v_self, __pyx_v_mm_s0, __pyx_v_mm_g0);

  /* "yt/geometry/particle_oct_container.pyx":1552
 *         self.set_refined_bool(mm_s0, mm_g0)
 *         # Compress
 *         mm_s0._compress(mm_s)             # <<<<<<<<<<<<<<
 *         mm_g0._compress(mm_g)
 * 
 */
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_v_mm_s0->__pyx_vtab)->_compress(__pyx_v_mm_s0, __pyx_v_mm_s);

  /* "yt/geometry/particle_oct_container.pyx":1553
 *         # Compress
 *         mm_s0._compress(mm_s)
 *         mm_g0._compress(mm_g)             # <<<<<<<<<<<<<<
 * 
 *     def find_files(self,
 */
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_v_mm_g0->__pyx_vtab)->_compress(__pyx_v_mm_g0, __pyx_v_mm_g);

  /* "yt/geometry/particle_oct_container.pyx":1515
 *         self.ghosts_ewah = BoolArrayColl(self.s1, self.s2)
 * 
 *     def fill_masks(self, BoolArrayCollection mm_s, BoolArrayCollection mm_g = None):             # <<<<<<<<<<<<<<
 *         # Normal variables
 *         cdef int i
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapSelector.fill_masks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mm_s0);
  __Pyx_XDECREF((PyObject *)__pyx_v_mm_g0);
  __Pyx_XDECREF((PyObject *)__pyx_v_mm_g);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1555
 *         mm_g0._compress(mm_g)
 * 
 *     def find_files(self,             # <<<<<<<<<<<<<<
 *                    np.ndarray[np.uint8_t, ndim=1] file_mask_p,
 *                    np.ndarray[np.uint8_t, ndim=1] file_mask_g):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_5find_files(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_5find_files(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_file_mask_p = 0;
  PyArrayObject *__pyx_v_file_mask_g = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("find_files (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_file_mask_p,&__pyx_n_s_file_mask_g,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_file_mask_p)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_file_mask_g)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("find_files", 1, 2, 2, 1); __PYX_ERR(0, 1555, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "find_files") < 0)) __PYX_ERR(0, 1555, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_file_mask_p = ((PyArrayObject *)values[0]);
    __pyx_v_file_mask_g = ((PyArrayObject *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("find_files", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1555, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapSelector.find_files", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_file_mask_p), __pyx_ptype_5numpy_ndarray, 1, "file_mask_p", 0))) __PYX_ERR(0, 1556, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_file_mask_g), __pyx_ptype_5numpy_ndarray, 1, "file_mask_g", 0))) __PYX_ERR(0, 1557, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_4find_files(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self), __pyx_v_file_mask_p, __pyx_v_file_mask_g);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_4find_files(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, PyArrayObject *__pyx_v_file_mask_p, PyArrayObject *__pyx_v_file_mask_g) {
  __pyx_t_5numpy_uint64_t __pyx_v_i;
  __pyx_t_5numpy_int32_t __pyx_v_level;
  __pyx_t_5numpy_uint64_t __pyx_v_mi1;
  __pyx_t_5numpy_float64_t __pyx_v_pos[3];
  __pyx_t_5numpy_float64_t __pyx_v_dds[3];
  __Pyx_LocalBuf_ND __pyx_pybuffernd_file_mask_g;
  __Pyx_Buffer __pyx_pybuffer_file_mask_g;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_file_mask_p;
  __Pyx_Buffer __pyx_pybuffer_file_mask_p;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_uint64_t __pyx_t_1;
  __pyx_t_5numpy_uint32_t __pyx_t_2;
  __pyx_t_5numpy_uint32_t __pyx_t_3;
  __pyx_t_5numpy_uint64_t __pyx_t_4;
  int __pyx_t_5;
  __pyx_t_5numpy_uint64_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("find_files", 0);
  __pyx_pybuffer_file_mask_p.pybuffer.buf = NULL;
  __pyx_pybuffer_file_mask_p.refcount = 0;
  __pyx_pybuffernd_file_mask_p.data = NULL;
  __pyx_pybuffernd_file_mask_p.rcbuffer = &__pyx_pybuffer_file_mask_p;
  __pyx_pybuffer_file_mask_g.pybuffer.buf = NULL;
  __pyx_pybuffer_file_mask_g.refcount = 0;
  __pyx_pybuffernd_file_mask_g.data = NULL;
  __pyx_pybuffernd_file_mask_g.rcbuffer = &__pyx_pybuffer_file_mask_g;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer, (PyObject*)__pyx_v_file_mask_p, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1555, __pyx_L1_error)
  }
  __pyx_pybuffernd_file_mask_p.diminfo[0].strides = __pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_file_mask_p.diminfo[0].shape = __pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer.shape[0];
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_file_mask_g.rcbuffer->pybuffer, (PyObject*)__pyx_v_file_mask_g, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, PyBUF_FORMAT| PyBUF_STRIDES| PyBUF_WRITABLE, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 1555, __pyx_L1_error)
  }
  __pyx_pybuffernd_file_mask_g.diminfo[0].strides = __pyx_pybuffernd_file_mask_g.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_file_mask_g.diminfo[0].shape = __pyx_pybuffernd_file_mask_g.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/particle_oct_container.pyx":1559
 *                    np.ndarray[np.uint8_t, ndim=1] file_mask_g):
 *         cdef np.uint64_t i
 *         cdef np.int32_t level = 0             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t mi1
 *         mi1 = ~(<np.uint64_t>0)
 */
  __pyx_v_level = 0;

  /* "yt/geometry/particle_oct_container.pyx":1561
 *         cdef np.int32_t level = 0
 *         cdef np.uint64_t mi1
 *         mi1 = ~(<np.uint64_t>0)             # <<<<<<<<<<<<<<
 *         cdef np.float64_t pos[3]
 *         cdef np.float64_t dds[3]
 */
  __pyx_v_mi1 = (~((__pyx_t_5numpy_uint64_t)0));

  /* "yt/geometry/particle_oct_container.pyx":1564
 *         cdef np.float64_t pos[3]
 *         cdef np.float64_t dds[3]
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             pos[i] = self.DLE[i]
 *             dds[i] = self.DRE[i] - self.DLE[i]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":1565
 *         cdef np.float64_t dds[3]
 *         for i in range(3):
 *             pos[i] = self.DLE[i]             # <<<<<<<<<<<<<<
 *             dds[i] = self.DRE[i] - self.DLE[i]
 *         # Fill with input
 */
    (__pyx_v_pos[__pyx_v_i]) = (__pyx_v_self->DLE[__pyx_v_i]);

    /* "yt/geometry/particle_oct_container.pyx":1566
 *         for i in range(3):
 *             pos[i] = self.DLE[i]
 *             dds[i] = self.DRE[i] - self.DLE[i]             # <<<<<<<<<<<<<<
 *         # Fill with input
 *         for i in range(self.nfiles):
 */
    (__pyx_v_dds[__pyx_v_i]) = ((__pyx_v_self->DRE[__pyx_v_i]) - (__pyx_v_self->DLE[__pyx_v_i]));
  }

  /* "yt/geometry/particle_oct_container.pyx":1568
 *             dds[i] = self.DRE[i] - self.DLE[i]
 *         # Fill with input
 *         for i in range(self.nfiles):             # <<<<<<<<<<<<<<
 *             self.file_mask_p[i] = file_mask_p[i]
 *             self.file_mask_g[i] = file_mask_g[i]
 */
  __pyx_t_2 = __pyx_v_self->nfiles;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":1569
 *         # Fill with input
 *         for i in range(self.nfiles):
 *             self.file_mask_p[i] = file_mask_p[i]             # <<<<<<<<<<<<<<
 *             self.file_mask_g[i] = file_mask_g[i]
 *         # Recurse
 */
    __pyx_t_4 = __pyx_v_i;
    __pyx_t_5 = -1;
    if (unlikely(__pyx_t_4 >= (size_t)__pyx_pybuffernd_file_mask_p.diminfo[0].shape)) __pyx_t_5 = 0;
    if (unlikely(__pyx_t_5 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_5);
      __PYX_ERR(0, 1569, __pyx_L1_error)
    }
    if (unlikely(!__pyx_v_self->file_mask_p.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1569, __pyx_L1_error)}
    __pyx_t_6 = __pyx_v_i;
    __pyx_t_5 = -1;
    if (unlikely(__pyx_t_6 >= (size_t)__pyx_v_self->file_mask_p.shape[0])) __pyx_t_5 = 0;
    if (unlikely(__pyx_t_5 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_5);
      __PYX_ERR(0, 1569, __pyx_L1_error)
    }
    *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->file_mask_p.data + __pyx_t_6 * __pyx_v_self->file_mask_p.strides[0]) )) = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_file_mask_p.diminfo[0].strides));

    /* "yt/geometry/particle_oct_container.pyx":1570
 *         for i in range(self.nfiles):
 *             self.file_mask_p[i] = file_mask_p[i]
 *             self.file_mask_g[i] = file_mask_g[i]             # <<<<<<<<<<<<<<
 *         # Recurse
 *         self.recursive_morton_files(level, pos, dds, mi1)
 */
    __pyx_t_4 = __pyx_v_i;
    __pyx_t_5 = -1;
    if (unlikely(__pyx_t_4 >= (size_t)__pyx_pybuffernd_file_mask_g.diminfo[0].shape)) __pyx_t_5 = 0;
    if (unlikely(__pyx_t_5 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_5);
      __PYX_ERR(0, 1570, __pyx_L1_error)
    }
    if (unlikely(!__pyx_v_self->file_mask_g.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1570, __pyx_L1_error)}
    __pyx_t_6 = __pyx_v_i;
    __pyx_t_5 = -1;
    if (unlikely(__pyx_t_6 >= (size_t)__pyx_v_self->file_mask_g.shape[0])) __pyx_t_5 = 0;
    if (unlikely(__pyx_t_5 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_5);
      __PYX_ERR(0, 1570, __pyx_L1_error)
    }
    *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->file_mask_g.data + __pyx_t_6 * __pyx_v_self->file_mask_g.strides[0]) )) = (*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_file_mask_g.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_file_mask_g.diminfo[0].strides));
  }

  /* "yt/geometry/particle_oct_container.pyx":1572
 *             self.file_mask_g[i] = file_mask_g[i]
 *         # Recurse
 *         self.recursive_morton_files(level, pos, dds, mi1)             # <<<<<<<<<<<<<<
 *         # Fill with results
 *         for i in range(self.nfiles):
 */
  ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->recursive_morton_files(__pyx_v_self, __pyx_v_level, __pyx_v_pos, __pyx_v_dds, __pyx_v_mi1);

  /* "yt/geometry/particle_oct_container.pyx":1574
 *         self.recursive_morton_files(level, pos, dds, mi1)
 *         # Fill with results
 *         for i in range(self.nfiles):             # <<<<<<<<<<<<<<
 *             file_mask_p[i] = self.file_mask_p[i]
 *             if file_mask_p[i]:
 */
  __pyx_t_2 = __pyx_v_self->nfiles;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":1575
 *         # Fill with results
 *         for i in range(self.nfiles):
 *             file_mask_p[i] = self.file_mask_p[i]             # <<<<<<<<<<<<<<
 *             if file_mask_p[i]:
 *                 file_mask_g[i] = 0
 */
    if (unlikely(!__pyx_v_self->file_mask_p.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1575, __pyx_L1_error)}
    __pyx_t_4 = __pyx_v_i;
    __pyx_t_5 = -1;
    if (unlikely(__pyx_t_4 >= (size_t)__pyx_v_self->file_mask_p.shape[0])) __pyx_t_5 = 0;
    if (unlikely(__pyx_t_5 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_5);
      __PYX_ERR(0, 1575, __pyx_L1_error)
    }
    __pyx_t_6 = __pyx_v_i;
    __pyx_t_5 = -1;
    if (unlikely(__pyx_t_6 >= (size_t)__pyx_pybuffernd_file_mask_p.diminfo[0].shape)) __pyx_t_5 = 0;
    if (unlikely(__pyx_t_5 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_5);
      __PYX_ERR(0, 1575, __pyx_L1_error)
    }
    *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_file_mask_p.diminfo[0].strides) = (*((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->file_mask_p.data + __pyx_t_4 * __pyx_v_self->file_mask_p.strides[0]) )));

    /* "yt/geometry/particle_oct_container.pyx":1576
 *         for i in range(self.nfiles):
 *             file_mask_p[i] = self.file_mask_p[i]
 *             if file_mask_p[i]:             # <<<<<<<<<<<<<<
 *                 file_mask_g[i] = 0
 *             else:
 */
    __pyx_t_4 = __pyx_v_i;
    __pyx_t_5 = -1;
    if (unlikely(__pyx_t_4 >= (size_t)__pyx_pybuffernd_file_mask_p.diminfo[0].shape)) __pyx_t_5 = 0;
    if (unlikely(__pyx_t_5 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_5);
      __PYX_ERR(0, 1576, __pyx_L1_error)
    }
    __pyx_t_7 = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_file_mask_p.diminfo[0].strides)) != 0);
    if (__pyx_t_7) {

      /* "yt/geometry/particle_oct_container.pyx":1577
 *             file_mask_p[i] = self.file_mask_p[i]
 *             if file_mask_p[i]:
 *                 file_mask_g[i] = 0             # <<<<<<<<<<<<<<
 *             else:
 *                 file_mask_g[i] = self.file_mask_g[i]
 */
      __pyx_t_4 = __pyx_v_i;
      __pyx_t_5 = -1;
      if (unlikely(__pyx_t_4 >= (size_t)__pyx_pybuffernd_file_mask_g.diminfo[0].shape)) __pyx_t_5 = 0;
      if (unlikely(__pyx_t_5 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_5);
        __PYX_ERR(0, 1577, __pyx_L1_error)
      }
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_file_mask_g.rcbuffer->pybuffer.buf, __pyx_t_4, __pyx_pybuffernd_file_mask_g.diminfo[0].strides) = 0;

      /* "yt/geometry/particle_oct_container.pyx":1576
 *         for i in range(self.nfiles):
 *             file_mask_p[i] = self.file_mask_p[i]
 *             if file_mask_p[i]:             # <<<<<<<<<<<<<<
 *                 file_mask_g[i] = 0
 *             else:
 */
      goto __pyx_L9;
    }

    /* "yt/geometry/particle_oct_container.pyx":1579
 *                 file_mask_g[i] = 0
 *             else:
 *                 file_mask_g[i] = self.file_mask_g[i]             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    /*else*/ {
      if (unlikely(!__pyx_v_self->file_mask_g.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1579, __pyx_L1_error)}
      __pyx_t_4 = __pyx_v_i;
      __pyx_t_5 = -1;
      if (unlikely(__pyx_t_4 >= (size_t)__pyx_v_self->file_mask_g.shape[0])) __pyx_t_5 = 0;
      if (unlikely(__pyx_t_5 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_5);
        __PYX_ERR(0, 1579, __pyx_L1_error)
      }
      __pyx_t_6 = __pyx_v_i;
      __pyx_t_5 = -1;
      if (unlikely(__pyx_t_6 >= (size_t)__pyx_pybuffernd_file_mask_g.diminfo[0].shape)) __pyx_t_5 = 0;
      if (unlikely(__pyx_t_5 != -1)) {
        __Pyx_RaiseBufferIndexError(__pyx_t_5);
        __PYX_ERR(0, 1579, __pyx_L1_error)
      }
      *__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint8_t *, __pyx_pybuffernd_file_mask_g.rcbuffer->pybuffer.buf, __pyx_t_6, __pyx_pybuffernd_file_mask_g.diminfo[0].strides) = (*((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->file_mask_g.data + __pyx_t_4 * __pyx_v_self->file_mask_g.strides[0]) )));
    }
    __pyx_L9:;
  }

  /* "yt/geometry/particle_oct_container.pyx":1555
 *         mm_g0._compress(mm_g)
 * 
 *     def find_files(self,             # <<<<<<<<<<<<<<
 *                    np.ndarray[np.uint8_t, ndim=1] file_mask_p,
 *                    np.ndarray[np.uint8_t, ndim=1] file_mask_g):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_mask_g.rcbuffer->pybuffer);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapSelector.find_files", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_mask_g.rcbuffer->pybuffer);
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_file_mask_p.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1585
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef bint is_refined(self, np.uint64_t mi1):             # <<<<<<<<<<<<<<
 *         return self.bitmap.collisions._isref(mi1)
 * 
 */

static int __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_is_refined(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_refined", 0);

  /* "yt/geometry/particle_oct_container.pyx":1586
 *     @cython.initializedcheck(False)
 *     cdef bint is_refined(self, np.uint64_t mi1):
 *         return self.bitmap.collisions._isref(mi1)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_v_self->bitmap->collisions->__pyx_vtab)->_isref(__pyx_v_self->bitmap->collisions, __pyx_v_mi1);
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1585
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef bint is_refined(self, np.uint64_t mi1):             # <<<<<<<<<<<<<<
 *         return self.bitmap.collisions._isref(mi1)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1591
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef bint is_refined_files(self, np.uint64_t mi1):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t i
 *         if self.bitmap.collisions._isref(mi1):
 */

static int __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_is_refined_files(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1) {
  __pyx_t_5numpy_uint64_t __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __pyx_t_5numpy_uint32_t __pyx_t_2;
  __pyx_t_5numpy_uint32_t __pyx_t_3;
  __pyx_t_5numpy_uint64_t __pyx_t_4;
  __pyx_t_5numpy_uint64_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_refined_files", 0);

  /* "yt/geometry/particle_oct_container.pyx":1593
 *     cdef bint is_refined_files(self, np.uint64_t mi1):
 *         cdef np.uint64_t i
 *         if self.bitmap.collisions._isref(mi1):             # <<<<<<<<<<<<<<
 *             # Don't refine if files all selected already
 *             for i in range(self.nfiles):
 */
  __pyx_t_1 = (((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)__pyx_v_self->bitmap->collisions->__pyx_vtab)->_isref(__pyx_v_self->bitmap->collisions, __pyx_v_mi1) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/particle_oct_container.pyx":1595
 *         if self.bitmap.collisions._isref(mi1):
 *             # Don't refine if files all selected already
 *             for i in range(self.nfiles):             # <<<<<<<<<<<<<<
 *                 if self.file_mask_p[i] == 0:
 *                     if self.bitmap.bitmasks._isref(i, mi1) == 1:
 */
    __pyx_t_2 = __pyx_v_self->nfiles;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "yt/geometry/particle_oct_container.pyx":1596
 *             # Don't refine if files all selected already
 *             for i in range(self.nfiles):
 *                 if self.file_mask_p[i] == 0:             # <<<<<<<<<<<<<<
 *                     if self.bitmap.bitmasks._isref(i, mi1) == 1:
 *                         return 1
 */
      if (unlikely(!__pyx_v_self->file_mask_p.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1596, __pyx_L1_error)}
      __pyx_t_5 = __pyx_v_i;
      __pyx_t_1 = (((*((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->file_mask_p.data + __pyx_t_5 * __pyx_v_self->file_mask_p.strides[0]) ))) == 0) != 0);
      if (__pyx_t_1) {

        /* "yt/geometry/particle_oct_container.pyx":1597
 *             for i in range(self.nfiles):
 *                 if self.file_mask_p[i] == 0:
 *                     if self.bitmap.bitmasks._isref(i, mi1) == 1:             # <<<<<<<<<<<<<<
 *                         return 1
 *             return 0
 */
        __pyx_t_1 = ((((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmap->bitmasks->__pyx_vtab)->_isref(__pyx_v_self->bitmap->bitmasks, __pyx_v_i, __pyx_v_mi1) == 1) != 0);
        if (__pyx_t_1) {

          /* "yt/geometry/particle_oct_container.pyx":1598
 *                 if self.file_mask_p[i] == 0:
 *                     if self.bitmap.bitmasks._isref(i, mi1) == 1:
 *                         return 1             # <<<<<<<<<<<<<<
 *             return 0
 *         else:
 */
          __pyx_r = 1;
          goto __pyx_L0;

          /* "yt/geometry/particle_oct_container.pyx":1597
 *             for i in range(self.nfiles):
 *                 if self.file_mask_p[i] == 0:
 *                     if self.bitmap.bitmasks._isref(i, mi1) == 1:             # <<<<<<<<<<<<<<
 *                         return 1
 *             return 0
 */
        }

        /* "yt/geometry/particle_oct_container.pyx":1596
 *             # Don't refine if files all selected already
 *             for i in range(self.nfiles):
 *                 if self.file_mask_p[i] == 0:             # <<<<<<<<<<<<<<
 *                     if self.bitmap.bitmasks._isref(i, mi1) == 1:
 *                         return 1
 */
      }
    }

    /* "yt/geometry/particle_oct_container.pyx":1599
 *                     if self.bitmap.bitmasks._isref(i, mi1) == 1:
 *                         return 1
 *             return 0             # <<<<<<<<<<<<<<
 *         else:
 *             return 0
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "yt/geometry/particle_oct_container.pyx":1593
 *     cdef bint is_refined_files(self, np.uint64_t mi1):
 *         cdef np.uint64_t i
 *         if self.bitmap.collisions._isref(mi1):             # <<<<<<<<<<<<<<
 *             # Don't refine if files all selected already
 *             for i in range(self.nfiles):
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":1601
 *             return 0
 *         else:
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  /*else*/ {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "yt/geometry/particle_oct_container.pyx":1591
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef bint is_refined_files(self, np.uint64_t mi1):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t i
 *         if self.bitmap.collisions._isref(mi1):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("yt.geometry.particle_oct_container.ParticleBitmapSelector.is_refined_files", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1607
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef void add_coarse(self, np.uint64_t mi1, int bbox = 2):             # <<<<<<<<<<<<<<
 *         self.coarse_select_bool[mi1] = 1
 *         # Neighbors
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_coarse(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1, struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_coarse *__pyx_optional_args) {
  int __pyx_v_bbox = ((int)2);
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_uint64_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("add_coarse", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_bbox = __pyx_optional_args->bbox;
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":1608
 *     @cython.initializedcheck(False)
 *     cdef void add_coarse(self, np.uint64_t mi1, int bbox = 2):
 *         self.coarse_select_bool[mi1] = 1             # <<<<<<<<<<<<<<
 *         # Neighbors
 *         if (self.ngz > 0) and (bbox == 2):
 */
  __pyx_t_1 = __pyx_v_mi1;
  *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->coarse_select_bool.data + __pyx_t_1 * __pyx_v_self->coarse_select_bool.strides[0]) )) = 1;

  /* "yt/geometry/particle_oct_container.pyx":1610
 *         self.coarse_select_bool[mi1] = 1
 *         # Neighbors
 *         if (self.ngz > 0) and (bbox == 2):             # <<<<<<<<<<<<<<
 *             if not self.is_refined(mi1):
 *                 self.add_neighbors_coarse(mi1)
 */
  __pyx_t_3 = ((__pyx_v_self->ngz > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = ((__pyx_v_bbox == 2) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "yt/geometry/particle_oct_container.pyx":1611
 *         # Neighbors
 *         if (self.ngz > 0) and (bbox == 2):
 *             if not self.is_refined(mi1):             # <<<<<<<<<<<<<<
 *                 self.add_neighbors_coarse(mi1)
 * 
 */
    __pyx_t_2 = ((!(((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->is_refined(__pyx_v_self, __pyx_v_mi1) != 0)) != 0);
    if (__pyx_t_2) {

      /* "yt/geometry/particle_oct_container.pyx":1612
 *         if (self.ngz > 0) and (bbox == 2):
 *             if not self.is_refined(mi1):
 *                 self.add_neighbors_coarse(mi1)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
      ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->add_neighbors_coarse(__pyx_v_self, __pyx_v_mi1);

      /* "yt/geometry/particle_oct_container.pyx":1611
 *         # Neighbors
 *         if (self.ngz > 0) and (bbox == 2):
 *             if not self.is_refined(mi1):             # <<<<<<<<<<<<<<
 *                 self.add_neighbors_coarse(mi1)
 * 
 */
    }

    /* "yt/geometry/particle_oct_container.pyx":1610
 *         self.coarse_select_bool[mi1] = 1
 *         # Neighbors
 *         if (self.ngz > 0) and (bbox == 2):             # <<<<<<<<<<<<<<
 *             if not self.is_refined(mi1):
 *                 self.add_neighbors_coarse(mi1)
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":1607
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef void add_coarse(self, np.uint64_t mi1, int bbox = 2):             # <<<<<<<<<<<<<<
 *         self.coarse_select_bool[mi1] = 1
 *         # Neighbors
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":1618
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef void set_files_coarse(self, np.uint64_t mi1):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t i
 *         cdef bint flag_ref = self.is_refined(mi1)
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_files_coarse(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1) {
  __pyx_t_5numpy_uint64_t __pyx_v_i;
  int __pyx_v_flag_ref;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __pyx_t_5numpy_uint32_t __pyx_t_2;
  __pyx_t_5numpy_uint32_t __pyx_t_3;
  __pyx_t_5numpy_uint64_t __pyx_t_4;
  __pyx_t_5numpy_uint64_t __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("set_files_coarse", 0);

  /* "yt/geometry/particle_oct_container.pyx":1620
 *     cdef void set_files_coarse(self, np.uint64_t mi1):
 *         cdef np.uint64_t i
 *         cdef bint flag_ref = self.is_refined(mi1)             # <<<<<<<<<<<<<<
 *         # Flag files at coarse level
 *         if flag_ref == 0:
 */
  __pyx_v_flag_ref = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->is_refined(__pyx_v_self, __pyx_v_mi1);

  /* "yt/geometry/particle_oct_container.pyx":1622
 *         cdef bint flag_ref = self.is_refined(mi1)
 *         # Flag files at coarse level
 *         if flag_ref == 0:             # <<<<<<<<<<<<<<
 *             for i in range(self.nfiles):
 *                 if self.file_mask_p[i] == 0:
 */
  __pyx_t_1 = ((__pyx_v_flag_ref == 0) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/particle_oct_container.pyx":1623
 *         # Flag files at coarse level
 *         if flag_ref == 0:
 *             for i in range(self.nfiles):             # <<<<<<<<<<<<<<
 *                 if self.file_mask_p[i] == 0:
 *                     if self.bitmap.bitmasks._get_coarse(i, mi1) == 1:
 */
    __pyx_t_2 = __pyx_v_self->nfiles;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "yt/geometry/particle_oct_container.pyx":1624
 *         if flag_ref == 0:
 *             for i in range(self.nfiles):
 *                 if self.file_mask_p[i] == 0:             # <<<<<<<<<<<<<<
 *                     if self.bitmap.bitmasks._get_coarse(i, mi1) == 1:
 *                         self.file_mask_p[i] = 1
 */
      __pyx_t_5 = __pyx_v_i;
      __pyx_t_1 = (((*((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->file_mask_p.data + __pyx_t_5 * __pyx_v_self->file_mask_p.strides[0]) ))) == 0) != 0);
      if (__pyx_t_1) {

        /* "yt/geometry/particle_oct_container.pyx":1625
 *             for i in range(self.nfiles):
 *                 if self.file_mask_p[i] == 0:
 *                     if self.bitmap.bitmasks._get_coarse(i, mi1) == 1:             # <<<<<<<<<<<<<<
 *                         self.file_mask_p[i] = 1
 *         # Neighbors
 */
        __pyx_t_1 = ((((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmap->bitmasks->__pyx_vtab)->_get_coarse(__pyx_v_self->bitmap->bitmasks, __pyx_v_i, __pyx_v_mi1) == 1) != 0);
        if (__pyx_t_1) {

          /* "yt/geometry/particle_oct_container.pyx":1626
 *                 if self.file_mask_p[i] == 0:
 *                     if self.bitmap.bitmasks._get_coarse(i, mi1) == 1:
 *                         self.file_mask_p[i] = 1             # <<<<<<<<<<<<<<
 *         # Neighbors
 *         if (flag_ref == 0) and (self.ngz > 0):
 */
          __pyx_t_5 = __pyx_v_i;
          *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->file_mask_p.data + __pyx_t_5 * __pyx_v_self->file_mask_p.strides[0]) )) = 1;

          /* "yt/geometry/particle_oct_container.pyx":1625
 *             for i in range(self.nfiles):
 *                 if self.file_mask_p[i] == 0:
 *                     if self.bitmap.bitmasks._get_coarse(i, mi1) == 1:             # <<<<<<<<<<<<<<
 *                         self.file_mask_p[i] = 1
 *         # Neighbors
 */
        }

        /* "yt/geometry/particle_oct_container.pyx":1624
 *         if flag_ref == 0:
 *             for i in range(self.nfiles):
 *                 if self.file_mask_p[i] == 0:             # <<<<<<<<<<<<<<
 *                     if self.bitmap.bitmasks._get_coarse(i, mi1) == 1:
 *                         self.file_mask_p[i] = 1
 */
      }
    }

    /* "yt/geometry/particle_oct_container.pyx":1622
 *         cdef bint flag_ref = self.is_refined(mi1)
 *         # Flag files at coarse level
 *         if flag_ref == 0:             # <<<<<<<<<<<<<<
 *             for i in range(self.nfiles):
 *                 if self.file_mask_p[i] == 0:
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":1628
 *                         self.file_mask_p[i] = 1
 *         # Neighbors
 *         if (flag_ref == 0) and (self.ngz > 0):             # <<<<<<<<<<<<<<
 *             self.set_files_neighbors_coarse(mi1)
 * 
 */
  __pyx_t_6 = ((__pyx_v_flag_ref == 0) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_1 = __pyx_t_6;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_6 = ((__pyx_v_self->ngz > 0) != 0);
  __pyx_t_1 = __pyx_t_6;
  __pyx_L9_bool_binop_done:;
  if (__pyx_t_1) {

    /* "yt/geometry/particle_oct_container.pyx":1629
 *         # Neighbors
 *         if (flag_ref == 0) and (self.ngz > 0):
 *             self.set_files_neighbors_coarse(mi1)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->set_files_neighbors_coarse(__pyx_v_self, __pyx_v_mi1);

    /* "yt/geometry/particle_oct_container.pyx":1628
 *                         self.file_mask_p[i] = 1
 *         # Neighbors
 *         if (flag_ref == 0) and (self.ngz > 0):             # <<<<<<<<<<<<<<
 *             self.set_files_neighbors_coarse(mi1)
 * 
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":1618
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef void set_files_coarse(self, np.uint64_t mi1):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t i
 *         cdef bint flag_ref = self.is_refined(mi1)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":1635
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef int add_refined(self, np.uint64_t mi1, np.uint64_t mi2, int bbox = 2) except -1:             # <<<<<<<<<<<<<<
 *         self.refined_select_bool[mi2] = 1
 *         # Neighbors
 */

static int __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_refined(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1, __pyx_t_5numpy_uint64_t __pyx_v_mi2, struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_refined *__pyx_optional_args) {
  int __pyx_v_bbox = ((int)2);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_uint64_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("add_refined", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_bbox = __pyx_optional_args->bbox;
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":1636
 *     @cython.initializedcheck(False)
 *     cdef int add_refined(self, np.uint64_t mi1, np.uint64_t mi2, int bbox = 2) except -1:
 *         self.refined_select_bool[mi2] = 1             # <<<<<<<<<<<<<<
 *         # Neighbors
 *         if (self.ngz > 0) and (bbox == 2):
 */
  __pyx_t_1 = __pyx_v_mi2;
  *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->refined_select_bool.data + __pyx_t_1 * __pyx_v_self->refined_select_bool.strides[0]) )) = 1;

  /* "yt/geometry/particle_oct_container.pyx":1638
 *         self.refined_select_bool[mi2] = 1
 *         # Neighbors
 *         if (self.ngz > 0) and (bbox == 2):             # <<<<<<<<<<<<<<
 *             self.add_neighbors_refined(mi1, mi2)
 * 
 */
  __pyx_t_3 = ((__pyx_v_self->ngz > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = ((__pyx_v_bbox == 2) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "yt/geometry/particle_oct_container.pyx":1639
 *         # Neighbors
 *         if (self.ngz > 0) and (bbox == 2):
 *             self.add_neighbors_refined(mi1, mi2)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->add_neighbors_refined(__pyx_v_self, __pyx_v_mi1, __pyx_v_mi2);

    /* "yt/geometry/particle_oct_container.pyx":1638
 *         self.refined_select_bool[mi2] = 1
 *         # Neighbors
 *         if (self.ngz > 0) and (bbox == 2):             # <<<<<<<<<<<<<<
 *             self.add_neighbors_refined(mi1, mi2)
 * 
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":1635
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef int add_refined(self, np.uint64_t mi1, np.uint64_t mi2, int bbox = 2) except -1:             # <<<<<<<<<<<<<<
 *         self.refined_select_bool[mi2] = 1
 *         # Neighbors
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1645
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef void set_files_refined(self, np.uint64_t mi1, np.uint64_t mi2):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t i
 *         # Flag files
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_files_refined(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1, __pyx_t_5numpy_uint64_t __pyx_v_mi2) {
  __pyx_t_5numpy_uint64_t __pyx_v_i;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_uint32_t __pyx_t_1;
  __pyx_t_5numpy_uint32_t __pyx_t_2;
  __pyx_t_5numpy_uint64_t __pyx_t_3;
  __pyx_t_5numpy_uint64_t __pyx_t_4;
  int __pyx_t_5;
  struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__get __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("set_files_refined", 0);

  /* "yt/geometry/particle_oct_container.pyx":1648
 *         cdef np.uint64_t i
 *         # Flag files
 *         for i in range(self.nfiles):             # <<<<<<<<<<<<<<
 *             if self.file_mask_p[i] == 0:
 *                 if self.bitmap.bitmasks._get(i, mi1, mi2):
 */
  __pyx_t_1 = __pyx_v_self->nfiles;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":1649
 *         # Flag files
 *         for i in range(self.nfiles):
 *             if self.file_mask_p[i] == 0:             # <<<<<<<<<<<<<<
 *                 if self.bitmap.bitmasks._get(i, mi1, mi2):
 *                     self.file_mask_p[i] = 1
 */
    __pyx_t_4 = __pyx_v_i;
    __pyx_t_5 = (((*((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->file_mask_p.data + __pyx_t_4 * __pyx_v_self->file_mask_p.strides[0]) ))) == 0) != 0);
    if (__pyx_t_5) {

      /* "yt/geometry/particle_oct_container.pyx":1650
 *         for i in range(self.nfiles):
 *             if self.file_mask_p[i] == 0:
 *                 if self.bitmap.bitmasks._get(i, mi1, mi2):             # <<<<<<<<<<<<<<
 *                     self.file_mask_p[i] = 1
 *         # Neighbors
 */
      __pyx_t_6.__pyx_n = 1;
      __pyx_t_6.i2 = __pyx_v_mi2;
      __pyx_t_5 = ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmap->bitmasks->__pyx_vtab)->_get(__pyx_v_self->bitmap->bitmasks, __pyx_v_i, __pyx_v_mi1, &__pyx_t_6); 
      __pyx_t_7 = (__pyx_t_5 != 0);
      if (__pyx_t_7) {

        /* "yt/geometry/particle_oct_container.pyx":1651
 *             if self.file_mask_p[i] == 0:
 *                 if self.bitmap.bitmasks._get(i, mi1, mi2):
 *                     self.file_mask_p[i] = 1             # <<<<<<<<<<<<<<
 *         # Neighbors
 *         if (self.ngz > 0):
 */
        __pyx_t_4 = __pyx_v_i;
        *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->file_mask_p.data + __pyx_t_4 * __pyx_v_self->file_mask_p.strides[0]) )) = 1;

        /* "yt/geometry/particle_oct_container.pyx":1650
 *         for i in range(self.nfiles):
 *             if self.file_mask_p[i] == 0:
 *                 if self.bitmap.bitmasks._get(i, mi1, mi2):             # <<<<<<<<<<<<<<
 *                     self.file_mask_p[i] = 1
 *         # Neighbors
 */
      }

      /* "yt/geometry/particle_oct_container.pyx":1649
 *         # Flag files
 *         for i in range(self.nfiles):
 *             if self.file_mask_p[i] == 0:             # <<<<<<<<<<<<<<
 *                 if self.bitmap.bitmasks._get(i, mi1, mi2):
 *                     self.file_mask_p[i] = 1
 */
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":1653
 *                     self.file_mask_p[i] = 1
 *         # Neighbors
 *         if (self.ngz > 0):             # <<<<<<<<<<<<<<
 *             self.set_files_neighbors_refined(mi1, mi2)
 * 
 */
  __pyx_t_7 = ((__pyx_v_self->ngz > 0) != 0);
  if (__pyx_t_7) {

    /* "yt/geometry/particle_oct_container.pyx":1654
 *         # Neighbors
 *         if (self.ngz > 0):
 *             self.set_files_neighbors_refined(mi1, mi2)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->set_files_neighbors_refined(__pyx_v_self, __pyx_v_mi1, __pyx_v_mi2);

    /* "yt/geometry/particle_oct_container.pyx":1653
 *                     self.file_mask_p[i] = 1
 *         # Neighbors
 *         if (self.ngz > 0):             # <<<<<<<<<<<<<<
 *             self.set_files_neighbors_refined(mi1, mi2)
 * 
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":1645
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef void set_files_refined(self, np.uint64_t mi1, np.uint64_t mi2):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t i
 *         # Flag files
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":1660
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef void add_neighbors_coarse(self, np.uint64_t mi1):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t m
 *         cdef np.uint32_t ntot
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_neighbors_coarse(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1) {
  __pyx_t_5numpy_uint64_t __pyx_v_m;
  __pyx_t_5numpy_uint32_t __pyx_v_ntot;
  __pyx_t_5numpy_uint64_t __pyx_v_mi1_n;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_uint32_t __pyx_t_1;
  __pyx_t_5numpy_uint32_t __pyx_t_2;
  __pyx_t_5numpy_uint64_t __pyx_t_3;
  __pyx_t_5numpy_uint64_t __pyx_t_4;
  __Pyx_RefNannySetupContext("add_neighbors_coarse", 0);

  /* "yt/geometry/particle_oct_container.pyx":1664
 *         cdef np.uint32_t ntot
 *         cdef np.uint64_t mi1_n
 *         ntot = morton_neighbors_coarse(mi1, self.max_index1,             # <<<<<<<<<<<<<<
 *                                        self.periodicity,
 *                                        self.ngz, self.neighbors,
 */
  __pyx_v_ntot = __pyx_f_2yt_9utilities_3lib_14geometry_utils_morton_neighbors_coarse(__pyx_v_mi1, __pyx_v_self->max_index1, __pyx_v_self->periodicity, __pyx_v_self->ngz, __pyx_v_self->neighbors, __pyx_v_self->ind1_n, __pyx_v_self->neighbor_list1);

  /* "yt/geometry/particle_oct_container.pyx":1668
 *                                        self.ngz, self.neighbors,
 *                                        self.ind1_n, self.neighbor_list1)
 *         for m in range(ntot):             # <<<<<<<<<<<<<<
 *             mi1_n = self.neighbor_list1[m]
 *             self.coarse_ghosts_bool[mi1_n] = 1
 */
  __pyx_t_1 = __pyx_v_ntot;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_m = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":1669
 *                                        self.ind1_n, self.neighbor_list1)
 *         for m in range(ntot):
 *             mi1_n = self.neighbor_list1[m]             # <<<<<<<<<<<<<<
 *             self.coarse_ghosts_bool[mi1_n] = 1
 * 
 */
    __pyx_t_4 = __pyx_v_m;
    __pyx_v_mi1_n = (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->neighbor_list1.data + __pyx_t_4 * __pyx_v_self->neighbor_list1.strides[0]) )));

    /* "yt/geometry/particle_oct_container.pyx":1670
 *         for m in range(ntot):
 *             mi1_n = self.neighbor_list1[m]
 *             self.coarse_ghosts_bool[mi1_n] = 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    __pyx_t_4 = __pyx_v_mi1_n;
    *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->coarse_ghosts_bool.data + __pyx_t_4 * __pyx_v_self->coarse_ghosts_bool.strides[0]) )) = 1;
  }

  /* "yt/geometry/particle_oct_container.pyx":1660
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef void add_neighbors_coarse(self, np.uint64_t mi1):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t m
 *         cdef np.uint32_t ntot
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":1676
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef void set_files_neighbors_coarse(self, np.uint64_t mi1):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t i, m
 *         cdef np.uint32_t ntot
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_files_neighbors_coarse(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1) {
  __pyx_t_5numpy_uint64_t __pyx_v_i;
  __pyx_t_5numpy_uint64_t __pyx_v_m;
  __pyx_t_5numpy_uint32_t __pyx_v_ntot;
  __pyx_t_5numpy_uint64_t __pyx_v_mi1_n;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_uint32_t __pyx_t_1;
  __pyx_t_5numpy_uint32_t __pyx_t_2;
  __pyx_t_5numpy_uint64_t __pyx_t_3;
  __pyx_t_5numpy_uint64_t __pyx_t_4;
  __pyx_t_5numpy_uint32_t __pyx_t_5;
  __pyx_t_5numpy_uint32_t __pyx_t_6;
  __pyx_t_5numpy_uint64_t __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("set_files_neighbors_coarse", 0);

  /* "yt/geometry/particle_oct_container.pyx":1680
 *         cdef np.uint32_t ntot
 *         cdef np.uint64_t mi1_n
 *         ntot = morton_neighbors_coarse(mi1, self.max_index1,             # <<<<<<<<<<<<<<
 *                                        self.periodicity,
 *                                        self.ngz, self.neighbors,
 */
  __pyx_v_ntot = __pyx_f_2yt_9utilities_3lib_14geometry_utils_morton_neighbors_coarse(__pyx_v_mi1, __pyx_v_self->max_index1, __pyx_v_self->periodicity, __pyx_v_self->ngz, __pyx_v_self->neighbors, __pyx_v_self->ind1_n, __pyx_v_self->neighbor_list1);

  /* "yt/geometry/particle_oct_container.pyx":1684
 *                                        self.ngz, self.neighbors,
 *                                        self.ind1_n, self.neighbor_list1)
 *         for m in range(ntot):             # <<<<<<<<<<<<<<
 *             mi1_n = self.neighbor_list1[m]
 *             for i in range(self.nfiles):
 */
  __pyx_t_1 = __pyx_v_ntot;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_m = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":1685
 *                                        self.ind1_n, self.neighbor_list1)
 *         for m in range(ntot):
 *             mi1_n = self.neighbor_list1[m]             # <<<<<<<<<<<<<<
 *             for i in range(self.nfiles):
 *                 if self.file_mask_g[i] == 0:
 */
    __pyx_t_4 = __pyx_v_m;
    __pyx_v_mi1_n = (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->neighbor_list1.data + __pyx_t_4 * __pyx_v_self->neighbor_list1.strides[0]) )));

    /* "yt/geometry/particle_oct_container.pyx":1686
 *         for m in range(ntot):
 *             mi1_n = self.neighbor_list1[m]
 *             for i in range(self.nfiles):             # <<<<<<<<<<<<<<
 *                 if self.file_mask_g[i] == 0:
 *                     if self.bitmap.bitmasks._get_coarse(i, mi1_n):
 */
    __pyx_t_5 = __pyx_v_self->nfiles;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_6; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "yt/geometry/particle_oct_container.pyx":1687
 *             mi1_n = self.neighbor_list1[m]
 *             for i in range(self.nfiles):
 *                 if self.file_mask_g[i] == 0:             # <<<<<<<<<<<<<<
 *                     if self.bitmap.bitmasks._get_coarse(i, mi1_n):
 *                         self.file_mask_g[i] = 1
 */
      __pyx_t_7 = __pyx_v_i;
      __pyx_t_8 = (((*((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->file_mask_g.data + __pyx_t_7 * __pyx_v_self->file_mask_g.strides[0]) ))) == 0) != 0);
      if (__pyx_t_8) {

        /* "yt/geometry/particle_oct_container.pyx":1688
 *             for i in range(self.nfiles):
 *                 if self.file_mask_g[i] == 0:
 *                     if self.bitmap.bitmasks._get_coarse(i, mi1_n):             # <<<<<<<<<<<<<<
 *                         self.file_mask_g[i] = 1
 * 
 */
        __pyx_t_8 = (((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmap->bitmasks->__pyx_vtab)->_get_coarse(__pyx_v_self->bitmap->bitmasks, __pyx_v_i, __pyx_v_mi1_n) != 0);
        if (__pyx_t_8) {

          /* "yt/geometry/particle_oct_container.pyx":1689
 *                 if self.file_mask_g[i] == 0:
 *                     if self.bitmap.bitmasks._get_coarse(i, mi1_n):
 *                         self.file_mask_g[i] = 1             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
          __pyx_t_7 = __pyx_v_i;
          *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->file_mask_g.data + __pyx_t_7 * __pyx_v_self->file_mask_g.strides[0]) )) = 1;

          /* "yt/geometry/particle_oct_container.pyx":1688
 *             for i in range(self.nfiles):
 *                 if self.file_mask_g[i] == 0:
 *                     if self.bitmap.bitmasks._get_coarse(i, mi1_n):             # <<<<<<<<<<<<<<
 *                         self.file_mask_g[i] = 1
 * 
 */
        }

        /* "yt/geometry/particle_oct_container.pyx":1687
 *             mi1_n = self.neighbor_list1[m]
 *             for i in range(self.nfiles):
 *                 if self.file_mask_g[i] == 0:             # <<<<<<<<<<<<<<
 *                     if self.bitmap.bitmasks._get_coarse(i, mi1_n):
 *                         self.file_mask_g[i] = 1
 */
      }
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":1676
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef void set_files_neighbors_coarse(self, np.uint64_t mi1):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t i, m
 *         cdef np.uint32_t ntot
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":1695
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef void add_neighbors_refined(self, np.uint64_t mi1, np.uint64_t mi2):             # <<<<<<<<<<<<<<
 *         cdef int m
 *         cdef np.uint32_t ntot
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_neighbors_refined(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1, __pyx_t_5numpy_uint64_t __pyx_v_mi2) {
  int __pyx_v_m;
  __pyx_t_5numpy_uint32_t __pyx_v_ntot;
  __pyx_t_5numpy_uint64_t __pyx_v_mi1_n;
  __pyx_t_5numpy_uint64_t __pyx_v_mi2_n;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_int32_t __pyx_t_1;
  __pyx_t_5numpy_int32_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  __pyx_t_5numpy_uint64_t __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("add_neighbors_refined", 0);

  /* "yt/geometry/particle_oct_container.pyx":1699
 *         cdef np.uint32_t ntot
 *         cdef np.uint64_t mi1_n, mi2_n
 *         ntot = morton_neighbors_refined(mi1, mi2,             # <<<<<<<<<<<<<<
 *                                         self.max_index1, self.max_index2,
 *                                         self.periodicity, self.ngz,
 */
  __pyx_v_ntot = __pyx_f_2yt_9utilities_3lib_14geometry_utils_morton_neighbors_refined(__pyx_v_mi1, __pyx_v_mi2, __pyx_v_self->max_index1, __pyx_v_self->max_index2, __pyx_v_self->periodicity, __pyx_v_self->ngz, __pyx_v_self->neighbors, __pyx_v_self->ind1_n, __pyx_v_self->ind2_n, __pyx_v_self->neighbor_list1, __pyx_v_self->neighbor_list2);

  /* "yt/geometry/particle_oct_container.pyx":1704
 *                                         self.neighbors, self.ind1_n, self.ind2_n,
 *                                         self.neighbor_list1, self.neighbor_list2)
 *         for m in range(<np.int32_t>ntot):             # <<<<<<<<<<<<<<
 *             mi1_n = self.neighbor_list1[m]
 *             mi2_n = self.neighbor_list2[m]
 */
  __pyx_t_1 = ((__pyx_t_5numpy_int32_t)__pyx_v_ntot);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_m = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":1705
 *                                         self.neighbor_list1, self.neighbor_list2)
 *         for m in range(<np.int32_t>ntot):
 *             mi1_n = self.neighbor_list1[m]             # <<<<<<<<<<<<<<
 *             mi2_n = self.neighbor_list2[m]
 *             self.coarse_ghosts_bool[mi1_n] = 1
 */
    __pyx_t_4 = __pyx_v_m;
    __pyx_v_mi1_n = (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->neighbor_list1.data + __pyx_t_4 * __pyx_v_self->neighbor_list1.strides[0]) )));

    /* "yt/geometry/particle_oct_container.pyx":1706
 *         for m in range(<np.int32_t>ntot):
 *             mi1_n = self.neighbor_list1[m]
 *             mi2_n = self.neighbor_list2[m]             # <<<<<<<<<<<<<<
 *             self.coarse_ghosts_bool[mi1_n] = 1
 *             IF RefinedExternalGhosts == 1:
 */
    __pyx_t_4 = __pyx_v_m;
    __pyx_v_mi2_n = (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->neighbor_list2.data + __pyx_t_4 * __pyx_v_self->neighbor_list2.strides[0]) )));

    /* "yt/geometry/particle_oct_container.pyx":1707
 *             mi1_n = self.neighbor_list1[m]
 *             mi2_n = self.neighbor_list2[m]
 *             self.coarse_ghosts_bool[mi1_n] = 1             # <<<<<<<<<<<<<<
 *             IF RefinedExternalGhosts == 1:
 *                 if mi1_n == mi1:
 */
    __pyx_t_5 = __pyx_v_mi1_n;
    *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->coarse_ghosts_bool.data + __pyx_t_5 * __pyx_v_self->coarse_ghosts_bool.strides[0]) )) = 1;

    /* "yt/geometry/particle_oct_container.pyx":1709
 *             self.coarse_ghosts_bool[mi1_n] = 1
 *             IF RefinedExternalGhosts == 1:
 *                 if mi1_n == mi1:             # <<<<<<<<<<<<<<
 *                     self.refined_ghosts_bool[mi2_n] = 1
 *                 else:
 */
    __pyx_t_6 = ((__pyx_v_mi1_n == __pyx_v_mi1) != 0);
    if (__pyx_t_6) {

      /* "yt/geometry/particle_oct_container.pyx":1710
 *             IF RefinedExternalGhosts == 1:
 *                 if mi1_n == mi1:
 *                     self.refined_ghosts_bool[mi2_n] = 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     self.refined_ghosts_list._set(mi1_n, mi2_n)
 */
      __pyx_t_5 = __pyx_v_mi2_n;
      *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->refined_ghosts_bool.data + __pyx_t_5 * __pyx_v_self->refined_ghosts_bool.strides[0]) )) = 1;

      /* "yt/geometry/particle_oct_container.pyx":1709
 *             self.coarse_ghosts_bool[mi1_n] = 1
 *             IF RefinedExternalGhosts == 1:
 *                 if mi1_n == mi1:             # <<<<<<<<<<<<<<
 *                     self.refined_ghosts_bool[mi2_n] = 1
 *                 else:
 */
      goto __pyx_L5;
    }

    /* "yt/geometry/particle_oct_container.pyx":1712
 *                     self.refined_ghosts_bool[mi2_n] = 1
 *                 else:
 *                     self.refined_ghosts_list._set(mi1_n, mi2_n)             # <<<<<<<<<<<<<<
 *             ELSE:
 *                 if mi1_n == mi1:
 */
    /*else*/ {
      ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet *)__pyx_v_self->refined_ghosts_list->__pyx_vtab)->_set(__pyx_v_self->refined_ghosts_list, __pyx_v_mi1_n, __pyx_v_mi2_n);
    }
    __pyx_L5:;
  }

  /* "yt/geometry/particle_oct_container.pyx":1695
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef void add_neighbors_refined(self, np.uint64_t mi1, np.uint64_t mi2):             # <<<<<<<<<<<<<<
 *         cdef int m
 *         cdef np.uint32_t ntot
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":1723
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef void set_files_neighbors_refined(self, np.uint64_t mi1, np.uint64_t mi2):             # <<<<<<<<<<<<<<
 *         cdef int i, m
 *         cdef np.uint32_t ntot
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_files_neighbors_refined(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1, __pyx_t_5numpy_uint64_t __pyx_v_mi2) {
  int __pyx_v_i;
  int __pyx_v_m;
  __pyx_t_5numpy_uint32_t __pyx_v_ntot;
  __pyx_t_5numpy_uint64_t __pyx_v_mi1_n;
  __pyx_t_5numpy_uint64_t __pyx_v_mi2_n;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_int32_t __pyx_t_1;
  __pyx_t_5numpy_int32_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  __pyx_t_5numpy_uint32_t __pyx_t_6;
  __pyx_t_5numpy_uint32_t __pyx_t_7;
  int __pyx_t_8;
  struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_12FileBitmasks__get __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("set_files_neighbors_refined", 0);

  /* "yt/geometry/particle_oct_container.pyx":1727
 *         cdef np.uint32_t ntot
 *         cdef np.uint64_t mi1_n, mi2_n
 *         ntot = morton_neighbors_refined(mi1, mi2,             # <<<<<<<<<<<<<<
 *                                         self.max_index1, self.max_index2,
 *                                         self.periodicity, self.ngz,
 */
  __pyx_v_ntot = __pyx_f_2yt_9utilities_3lib_14geometry_utils_morton_neighbors_refined(__pyx_v_mi1, __pyx_v_mi2, __pyx_v_self->max_index1, __pyx_v_self->max_index2, __pyx_v_self->periodicity, __pyx_v_self->ngz, __pyx_v_self->neighbors, __pyx_v_self->ind1_n, __pyx_v_self->ind2_n, __pyx_v_self->neighbor_list1, __pyx_v_self->neighbor_list2);

  /* "yt/geometry/particle_oct_container.pyx":1732
 *                                         self.neighbors, self.ind1_n, self.ind2_n,
 *                                         self.neighbor_list1, self.neighbor_list2)
 *         for m in range(<np.int32_t>ntot):             # <<<<<<<<<<<<<<
 *             mi1_n = self.neighbor_list1[m]
 *             mi2_n = self.neighbor_list2[m]
 */
  __pyx_t_1 = ((__pyx_t_5numpy_int32_t)__pyx_v_ntot);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_m = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":1733
 *                                         self.neighbor_list1, self.neighbor_list2)
 *         for m in range(<np.int32_t>ntot):
 *             mi1_n = self.neighbor_list1[m]             # <<<<<<<<<<<<<<
 *             mi2_n = self.neighbor_list2[m]
 *             if self.is_refined(mi1_n) == 1:
 */
    __pyx_t_4 = __pyx_v_m;
    __pyx_v_mi1_n = (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->neighbor_list1.data + __pyx_t_4 * __pyx_v_self->neighbor_list1.strides[0]) )));

    /* "yt/geometry/particle_oct_container.pyx":1734
 *         for m in range(<np.int32_t>ntot):
 *             mi1_n = self.neighbor_list1[m]
 *             mi2_n = self.neighbor_list2[m]             # <<<<<<<<<<<<<<
 *             if self.is_refined(mi1_n) == 1:
 *                 for i in range(self.nfiles):
 */
    __pyx_t_4 = __pyx_v_m;
    __pyx_v_mi2_n = (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->neighbor_list2.data + __pyx_t_4 * __pyx_v_self->neighbor_list2.strides[0]) )));

    /* "yt/geometry/particle_oct_container.pyx":1735
 *             mi1_n = self.neighbor_list1[m]
 *             mi2_n = self.neighbor_list2[m]
 *             if self.is_refined(mi1_n) == 1:             # <<<<<<<<<<<<<<
 *                 for i in range(self.nfiles):
 *                     if self.file_mask_g[i] == 0:
 */
    __pyx_t_5 = ((((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->is_refined(__pyx_v_self, __pyx_v_mi1_n) == 1) != 0);
    if (__pyx_t_5) {

      /* "yt/geometry/particle_oct_container.pyx":1736
 *             mi2_n = self.neighbor_list2[m]
 *             if self.is_refined(mi1_n) == 1:
 *                 for i in range(self.nfiles):             # <<<<<<<<<<<<<<
 *                     if self.file_mask_g[i] == 0:
 *                         if self.bitmap.bitmasks._get(i, mi1_n, mi2_n) == 1:
 */
      __pyx_t_6 = __pyx_v_self->nfiles;
      __pyx_t_7 = __pyx_t_6;
      for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
        __pyx_v_i = __pyx_t_8;

        /* "yt/geometry/particle_oct_container.pyx":1737
 *             if self.is_refined(mi1_n) == 1:
 *                 for i in range(self.nfiles):
 *                     if self.file_mask_g[i] == 0:             # <<<<<<<<<<<<<<
 *                         if self.bitmap.bitmasks._get(i, mi1_n, mi2_n) == 1:
 *                             self.file_mask_g[i] = 1
 */
        __pyx_t_4 = __pyx_v_i;
        __pyx_t_5 = (((*((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->file_mask_g.data + __pyx_t_4 * __pyx_v_self->file_mask_g.strides[0]) ))) == 0) != 0);
        if (__pyx_t_5) {

          /* "yt/geometry/particle_oct_container.pyx":1738
 *                 for i in range(self.nfiles):
 *                     if self.file_mask_g[i] == 0:
 *                         if self.bitmap.bitmasks._get(i, mi1_n, mi2_n) == 1:             # <<<<<<<<<<<<<<
 *                             self.file_mask_g[i] = 1
 *             else:
 */
          __pyx_t_9.__pyx_n = 1;
          __pyx_t_9.i2 = __pyx_v_mi2_n;
          __pyx_t_5 = ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmap->bitmasks->__pyx_vtab)->_get(__pyx_v_self->bitmap->bitmasks, __pyx_v_i, __pyx_v_mi1_n, &__pyx_t_9); 
          __pyx_t_10 = ((__pyx_t_5 == 1) != 0);
          if (__pyx_t_10) {

            /* "yt/geometry/particle_oct_container.pyx":1739
 *                     if self.file_mask_g[i] == 0:
 *                         if self.bitmap.bitmasks._get(i, mi1_n, mi2_n) == 1:
 *                             self.file_mask_g[i] = 1             # <<<<<<<<<<<<<<
 *             else:
 *                 for i in range(self.nfiles):
 */
            __pyx_t_4 = __pyx_v_i;
            *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->file_mask_g.data + __pyx_t_4 * __pyx_v_self->file_mask_g.strides[0]) )) = 1;

            /* "yt/geometry/particle_oct_container.pyx":1738
 *                 for i in range(self.nfiles):
 *                     if self.file_mask_g[i] == 0:
 *                         if self.bitmap.bitmasks._get(i, mi1_n, mi2_n) == 1:             # <<<<<<<<<<<<<<
 *                             self.file_mask_g[i] = 1
 *             else:
 */
          }

          /* "yt/geometry/particle_oct_container.pyx":1737
 *             if self.is_refined(mi1_n) == 1:
 *                 for i in range(self.nfiles):
 *                     if self.file_mask_g[i] == 0:             # <<<<<<<<<<<<<<
 *                         if self.bitmap.bitmasks._get(i, mi1_n, mi2_n) == 1:
 *                             self.file_mask_g[i] = 1
 */
        }
      }

      /* "yt/geometry/particle_oct_container.pyx":1735
 *             mi1_n = self.neighbor_list1[m]
 *             mi2_n = self.neighbor_list2[m]
 *             if self.is_refined(mi1_n) == 1:             # <<<<<<<<<<<<<<
 *                 for i in range(self.nfiles):
 *                     if self.file_mask_g[i] == 0:
 */
      goto __pyx_L5;
    }

    /* "yt/geometry/particle_oct_container.pyx":1741
 *                             self.file_mask_g[i] = 1
 *             else:
 *                 for i in range(self.nfiles):             # <<<<<<<<<<<<<<
 *                     if self.file_mask_g[i] == 0:
 *                         if self.bitmap.bitmasks._get_coarse(i, mi1_n) == 1:
 */
    /*else*/ {
      __pyx_t_6 = __pyx_v_self->nfiles;
      __pyx_t_7 = __pyx_t_6;
      for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
        __pyx_v_i = __pyx_t_8;

        /* "yt/geometry/particle_oct_container.pyx":1742
 *             else:
 *                 for i in range(self.nfiles):
 *                     if self.file_mask_g[i] == 0:             # <<<<<<<<<<<<<<
 *                         if self.bitmap.bitmasks._get_coarse(i, mi1_n) == 1:
 *                             self.file_mask_g[i] = 1
 */
        __pyx_t_4 = __pyx_v_i;
        __pyx_t_10 = (((*((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->file_mask_g.data + __pyx_t_4 * __pyx_v_self->file_mask_g.strides[0]) ))) == 0) != 0);
        if (__pyx_t_10) {

          /* "yt/geometry/particle_oct_container.pyx":1743
 *                 for i in range(self.nfiles):
 *                     if self.file_mask_g[i] == 0:
 *                         if self.bitmap.bitmasks._get_coarse(i, mi1_n) == 1:             # <<<<<<<<<<<<<<
 *                             self.file_mask_g[i] = 1
 *                             break # If not refined, only one file should be selected
 */
          __pyx_t_10 = ((((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)__pyx_v_self->bitmap->bitmasks->__pyx_vtab)->_get_coarse(__pyx_v_self->bitmap->bitmasks, __pyx_v_i, __pyx_v_mi1_n) == 1) != 0);
          if (__pyx_t_10) {

            /* "yt/geometry/particle_oct_container.pyx":1744
 *                     if self.file_mask_g[i] == 0:
 *                         if self.bitmap.bitmasks._get_coarse(i, mi1_n) == 1:
 *                             self.file_mask_g[i] = 1             # <<<<<<<<<<<<<<
 *                             break # If not refined, only one file should be selected
 * 
 */
            __pyx_t_4 = __pyx_v_i;
            *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->file_mask_g.data + __pyx_t_4 * __pyx_v_self->file_mask_g.strides[0]) )) = 1;

            /* "yt/geometry/particle_oct_container.pyx":1745
 *                         if self.bitmap.bitmasks._get_coarse(i, mi1_n) == 1:
 *                             self.file_mask_g[i] = 1
 *                             break # If not refined, only one file should be selected             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
            goto __pyx_L11_break;

            /* "yt/geometry/particle_oct_container.pyx":1743
 *                 for i in range(self.nfiles):
 *                     if self.file_mask_g[i] == 0:
 *                         if self.bitmap.bitmasks._get_coarse(i, mi1_n) == 1:             # <<<<<<<<<<<<<<
 *                             self.file_mask_g[i] = 1
 *                             break # If not refined, only one file should be selected
 */
          }

          /* "yt/geometry/particle_oct_container.pyx":1742
 *             else:
 *                 for i in range(self.nfiles):
 *                     if self.file_mask_g[i] == 0:             # <<<<<<<<<<<<<<
 *                         if self.bitmap.bitmasks._get_coarse(i, mi1_n) == 1:
 *                             self.file_mask_g[i] = 1
 */
        }
      }
      __pyx_L11_break:;
    }
    __pyx_L5:;
  }

  /* "yt/geometry/particle_oct_container.pyx":1723
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef void set_files_neighbors_refined(self, np.uint64_t mi1, np.uint64_t mi2):             # <<<<<<<<<<<<<<
 *         cdef int i, m
 *         cdef np.uint32_t ntot
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":1750
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void set_coarse_list(self, BoolArrayColl mm_s, BoolArrayColl mm_g):             # <<<<<<<<<<<<<<
 *         self.coarse_select_list._fill_bool(mm_s)
 *         self.coarse_ghosts_list._fill_bool(mm_g)
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_coarse_list(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_s, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_g) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_coarse_list", 0);

  /* "yt/geometry/particle_oct_container.pyx":1751
 *     @cython.cdivision(True)
 *     cdef void set_coarse_list(self, BoolArrayColl mm_s, BoolArrayColl mm_g):
 *         self.coarse_select_list._fill_bool(mm_s)             # <<<<<<<<<<<<<<
 *         self.coarse_ghosts_list._fill_bool(mm_g)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_coarse_select_list); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1751, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_fill_bool); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1751, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, ((PyObject *)__pyx_v_mm_s)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_mm_s));
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1751, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1752
 *     cdef void set_coarse_list(self, BoolArrayColl mm_s, BoolArrayColl mm_g):
 *         self.coarse_select_list._fill_bool(mm_s)
 *         self.coarse_ghosts_list._fill_bool(mm_g)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_coarse_ghosts_list); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_fill_bool); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, ((PyObject *)__pyx_v_mm_g)) : __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_mm_g));
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1750
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void set_coarse_list(self, BoolArrayColl mm_s, BoolArrayColl mm_g):             # <<<<<<<<<<<<<<
 *         self.coarse_select_list._fill_bool(mm_s)
 *         self.coarse_ghosts_list._fill_bool(mm_g)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("yt.geometry.particle_oct_container.ParticleBitmapSelector.set_coarse_list", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":1757
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void set_refined_list(self, BoolArrayColl mm_s, BoolArrayColl mm_g):             # <<<<<<<<<<<<<<
 *         self.refined_ghosts_list._fill_bool(mm_g)
 * 
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_refined_list(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_s, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_g) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_refined_list", 0);

  /* "yt/geometry/particle_oct_container.pyx":1758
 *     @cython.cdivision(True)
 *     cdef void set_refined_list(self, BoolArrayColl mm_s, BoolArrayColl mm_g):
 *         self.refined_ghosts_list._fill_bool(mm_g)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet *)__pyx_v_self->refined_ghosts_list->__pyx_vtab)->_fill_bool(__pyx_v_self->refined_ghosts_list, __pyx_v_mm_g);

  /* "yt/geometry/particle_oct_container.pyx":1757
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void set_refined_list(self, BoolArrayColl mm_s, BoolArrayColl mm_g):             # <<<<<<<<<<<<<<
 *         self.refined_ghosts_list._fill_bool(mm_g)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":1763
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void set_coarse_bool(self, BoolArrayColl mm_s, BoolArrayColl mm_g):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t mi1
 *         mm_s._set_coarse_array_ptr(&self.coarse_select_bool[0])
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_coarse_bool(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_s, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_g) {
  __pyx_t_5numpy_uint64_t __pyx_v_mi1;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  __pyx_t_5numpy_uint64_t __pyx_t_2;
  __pyx_t_5numpy_uint64_t __pyx_t_3;
  __pyx_t_5numpy_uint64_t __pyx_t_4;
  __pyx_t_5numpy_uint64_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_coarse_bool", 0);

  /* "yt/geometry/particle_oct_container.pyx":1765
 *     cdef void set_coarse_bool(self, BoolArrayColl mm_s, BoolArrayColl mm_g):
 *         cdef np.uint64_t mi1
 *         mm_s._set_coarse_array_ptr(&self.coarse_select_bool[0])             # <<<<<<<<<<<<<<
 *         for mi1 in range(self.s1):
 *             self.coarse_select_bool[mi1] = 0
 */
  if (unlikely(!__pyx_v_self->coarse_select_bool.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1765, __pyx_L1_error)}
  __pyx_t_1 = 0;
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_v_mm_s->__pyx_vtab)->_set_coarse_array_ptr(__pyx_v_mm_s, (&(*((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->coarse_select_bool.data + __pyx_t_1 * __pyx_v_self->coarse_select_bool.strides[0]) )))));

  /* "yt/geometry/particle_oct_container.pyx":1766
 *         cdef np.uint64_t mi1
 *         mm_s._set_coarse_array_ptr(&self.coarse_select_bool[0])
 *         for mi1 in range(self.s1):             # <<<<<<<<<<<<<<
 *             self.coarse_select_bool[mi1] = 0
 *         mm_g._set_coarse_array_ptr(&self.coarse_ghosts_bool[0])
 */
  __pyx_t_2 = __pyx_v_self->s1;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_mi1 = __pyx_t_4;

    /* "yt/geometry/particle_oct_container.pyx":1767
 *         mm_s._set_coarse_array_ptr(&self.coarse_select_bool[0])
 *         for mi1 in range(self.s1):
 *             self.coarse_select_bool[mi1] = 0             # <<<<<<<<<<<<<<
 *         mm_g._set_coarse_array_ptr(&self.coarse_ghosts_bool[0])
 *         for mi1 in range(self.s1):
 */
    if (unlikely(!__pyx_v_self->coarse_select_bool.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1767, __pyx_L1_error)}
    __pyx_t_5 = __pyx_v_mi1;
    *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->coarse_select_bool.data + __pyx_t_5 * __pyx_v_self->coarse_select_bool.strides[0]) )) = 0;
  }

  /* "yt/geometry/particle_oct_container.pyx":1768
 *         for mi1 in range(self.s1):
 *             self.coarse_select_bool[mi1] = 0
 *         mm_g._set_coarse_array_ptr(&self.coarse_ghosts_bool[0])             # <<<<<<<<<<<<<<
 *         for mi1 in range(self.s1):
 *             self.coarse_ghosts_bool[mi1] = 0
 */
  if (unlikely(!__pyx_v_self->coarse_ghosts_bool.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1768, __pyx_L1_error)}
  __pyx_t_1 = 0;
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_v_mm_g->__pyx_vtab)->_set_coarse_array_ptr(__pyx_v_mm_g, (&(*((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->coarse_ghosts_bool.data + __pyx_t_1 * __pyx_v_self->coarse_ghosts_bool.strides[0]) )))));

  /* "yt/geometry/particle_oct_container.pyx":1769
 *             self.coarse_select_bool[mi1] = 0
 *         mm_g._set_coarse_array_ptr(&self.coarse_ghosts_bool[0])
 *         for mi1 in range(self.s1):             # <<<<<<<<<<<<<<
 *             self.coarse_ghosts_bool[mi1] = 0
 * 
 */
  __pyx_t_2 = __pyx_v_self->s1;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_mi1 = __pyx_t_4;

    /* "yt/geometry/particle_oct_container.pyx":1770
 *         mm_g._set_coarse_array_ptr(&self.coarse_ghosts_bool[0])
 *         for mi1 in range(self.s1):
 *             self.coarse_ghosts_bool[mi1] = 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    if (unlikely(!__pyx_v_self->coarse_ghosts_bool.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1770, __pyx_L1_error)}
    __pyx_t_5 = __pyx_v_mi1;
    *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->coarse_ghosts_bool.data + __pyx_t_5 * __pyx_v_self->coarse_ghosts_bool.strides[0]) )) = 0;
  }

  /* "yt/geometry/particle_oct_container.pyx":1763
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void set_coarse_bool(self, BoolArrayColl mm_s, BoolArrayColl mm_g):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t mi1
 *         mm_s._set_coarse_array_ptr(&self.coarse_select_bool[0])
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("yt.geometry.particle_oct_container.ParticleBitmapSelector.set_coarse_bool", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":1775
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void set_refined_bool(self, BoolArrayColl mm_s, BoolArrayColl mm_g):             # <<<<<<<<<<<<<<
 *         mm_s._append(self.select_ewah)
 *         mm_g._append(self.ghosts_ewah)
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_refined_bool(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_s, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_g) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("set_refined_bool", 0);

  /* "yt/geometry/particle_oct_container.pyx":1776
 *     @cython.cdivision(True)
 *     cdef void set_refined_bool(self, BoolArrayColl mm_s, BoolArrayColl mm_g):
 *         mm_s._append(self.select_ewah)             # <<<<<<<<<<<<<<
 *         mm_g._append(self.ghosts_ewah)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->select_ewah);
  __Pyx_INCREF(__pyx_t_1);
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_v_mm_s->__pyx_vtab)->_append(__pyx_v_mm_s, ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_t_1));
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1777
 *     cdef void set_refined_bool(self, BoolArrayColl mm_s, BoolArrayColl mm_g):
 *         mm_s._append(self.select_ewah)
 *         mm_g._append(self.ghosts_ewah)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->ghosts_ewah);
  __Pyx_INCREF(__pyx_t_1);
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_v_mm_g->__pyx_vtab)->_append(__pyx_v_mm_g, ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_t_1));
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1775
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void set_refined_bool(self, BoolArrayColl mm_s, BoolArrayColl mm_g):             # <<<<<<<<<<<<<<
 *         mm_s._append(self.select_ewah)
 *         mm_g._append(self.ghosts_ewah)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":1783
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef void push_refined_bool(self, np.uint64_t mi1):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t mi2
 *         self.select_ewah._set_refined_array_ptr(mi1, &self.refined_select_bool[0])
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_push_refined_bool(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_mi1) {
  __pyx_t_5numpy_uint64_t __pyx_v_mi2;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  __pyx_t_5numpy_uint64_t __pyx_t_2;
  __pyx_t_5numpy_uint64_t __pyx_t_3;
  __pyx_t_5numpy_uint64_t __pyx_t_4;
  __pyx_t_5numpy_uint64_t __pyx_t_5;
  __Pyx_RefNannySetupContext("push_refined_bool", 0);

  /* "yt/geometry/particle_oct_container.pyx":1785
 *     cdef void push_refined_bool(self, np.uint64_t mi1):
 *         cdef np.uint64_t mi2
 *         self.select_ewah._set_refined_array_ptr(mi1, &self.refined_select_bool[0])             # <<<<<<<<<<<<<<
 *         for mi2 in range(self.s2):
 *             self.refined_select_bool[mi2] = 0
 */
  __pyx_t_1 = 0;
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_v_self->select_ewah->__pyx_vtab)->_set_refined_array_ptr(__pyx_v_self->select_ewah, __pyx_v_mi1, (&(*((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->refined_select_bool.data + __pyx_t_1 * __pyx_v_self->refined_select_bool.strides[0]) )))));

  /* "yt/geometry/particle_oct_container.pyx":1786
 *         cdef np.uint64_t mi2
 *         self.select_ewah._set_refined_array_ptr(mi1, &self.refined_select_bool[0])
 *         for mi2 in range(self.s2):             # <<<<<<<<<<<<<<
 *             self.refined_select_bool[mi2] = 0
 *         self.ghosts_ewah._set_refined_array_ptr(mi1, &self.refined_ghosts_bool[0])
 */
  __pyx_t_2 = __pyx_v_self->s2;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_mi2 = __pyx_t_4;

    /* "yt/geometry/particle_oct_container.pyx":1787
 *         self.select_ewah._set_refined_array_ptr(mi1, &self.refined_select_bool[0])
 *         for mi2 in range(self.s2):
 *             self.refined_select_bool[mi2] = 0             # <<<<<<<<<<<<<<
 *         self.ghosts_ewah._set_refined_array_ptr(mi1, &self.refined_ghosts_bool[0])
 *         for mi2 in range(self.s2):
 */
    __pyx_t_5 = __pyx_v_mi2;
    *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->refined_select_bool.data + __pyx_t_5 * __pyx_v_self->refined_select_bool.strides[0]) )) = 0;
  }

  /* "yt/geometry/particle_oct_container.pyx":1788
 *         for mi2 in range(self.s2):
 *             self.refined_select_bool[mi2] = 0
 *         self.ghosts_ewah._set_refined_array_ptr(mi1, &self.refined_ghosts_bool[0])             # <<<<<<<<<<<<<<
 *         for mi2 in range(self.s2):
 *             self.refined_ghosts_bool[mi2] = 0
 */
  __pyx_t_1 = 0;
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_v_self->ghosts_ewah->__pyx_vtab)->_set_refined_array_ptr(__pyx_v_self->ghosts_ewah, __pyx_v_mi1, (&(*((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->refined_ghosts_bool.data + __pyx_t_1 * __pyx_v_self->refined_ghosts_bool.strides[0]) )))));

  /* "yt/geometry/particle_oct_container.pyx":1789
 *             self.refined_select_bool[mi2] = 0
 *         self.ghosts_ewah._set_refined_array_ptr(mi1, &self.refined_ghosts_bool[0])
 *         for mi2 in range(self.s2):             # <<<<<<<<<<<<<<
 *             self.refined_ghosts_bool[mi2] = 0
 * 
 */
  __pyx_t_2 = __pyx_v_self->s2;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_mi2 = __pyx_t_4;

    /* "yt/geometry/particle_oct_container.pyx":1790
 *         self.ghosts_ewah._set_refined_array_ptr(mi1, &self.refined_ghosts_bool[0])
 *         for mi2 in range(self.s2):
 *             self.refined_ghosts_bool[mi2] = 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    __pyx_t_5 = __pyx_v_mi2;
    *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->refined_ghosts_bool.data + __pyx_t_5 * __pyx_v_self->refined_ghosts_bool.strides[0]) )) = 0;
  }

  /* "yt/geometry/particle_oct_container.pyx":1783
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     cdef void push_refined_bool(self, np.uint64_t mi1):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t mi2
 *         self.select_ewah._set_refined_array_ptr(mi1, &self.refined_select_bool[0])
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":1795
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void add_ghost_zones(self, BoolArrayColl mm_s, BoolArrayColl mm_g):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t mi1, mi2, mi1_n, mi2_n
 *         # Get ghost zones, unordered
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_ghost_zones(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_s, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *__pyx_v_mm_g) {
  __pyx_t_5numpy_uint64_t __pyx_v_mi1;
  __pyx_t_5numpy_uint64_t __pyx_v_mi2;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_uint64_t __pyx_t_1;
  __pyx_t_5numpy_uint64_t __pyx_t_2;
  __pyx_t_5numpy_uint64_t __pyx_t_3;
  int __pyx_t_4;
  __pyx_t_5numpy_uint64_t __pyx_t_5;
  __pyx_t_5numpy_uint64_t __pyx_t_6;
  __pyx_t_5numpy_uint64_t __pyx_t_7;
  struct __pyx_opt_args_2yt_9utilities_3lib_14ewah_bool_wrap_31BoolArrayCollectionUncompressed__get __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  __pyx_t_5numpy_uint64_t __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_ghost_zones", 0);

  /* "yt/geometry/particle_oct_container.pyx":1798
 *         cdef np.uint64_t mi1, mi2, mi1_n, mi2_n
 *         # Get ghost zones, unordered
 *         for mi1 in range(self.s1):             # <<<<<<<<<<<<<<
 *             if mm_s._get_coarse(mi1):
 *                 if self.is_refined(mi1):
 */
  __pyx_t_1 = __pyx_v_self->s1;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_mi1 = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":1799
 *         # Get ghost zones, unordered
 *         for mi1 in range(self.s1):
 *             if mm_s._get_coarse(mi1):             # <<<<<<<<<<<<<<
 *                 if self.is_refined(mi1):
 *                     for mi2 in range(self.s2):
 */
    __pyx_t_4 = (((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_v_mm_s->__pyx_vtab)->_get_coarse(__pyx_v_mm_s, __pyx_v_mi1) != 0);
    if (__pyx_t_4) {

      /* "yt/geometry/particle_oct_container.pyx":1800
 *         for mi1 in range(self.s1):
 *             if mm_s._get_coarse(mi1):
 *                 if self.is_refined(mi1):             # <<<<<<<<<<<<<<
 *                     for mi2 in range(self.s2):
 *                         if mm_s._get(mi1, mi2):
 */
      __pyx_t_4 = (((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->is_refined(__pyx_v_self, __pyx_v_mi1) != 0);
      if (__pyx_t_4) {

        /* "yt/geometry/particle_oct_container.pyx":1801
 *             if mm_s._get_coarse(mi1):
 *                 if self.is_refined(mi1):
 *                     for mi2 in range(self.s2):             # <<<<<<<<<<<<<<
 *                         if mm_s._get(mi1, mi2):
 *                             self.add_neighbors_refined(mi1, mi2)
 */
        __pyx_t_5 = __pyx_v_self->s2;
        __pyx_t_6 = __pyx_t_5;
        for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
          __pyx_v_mi2 = __pyx_t_7;

          /* "yt/geometry/particle_oct_container.pyx":1802
 *                 if self.is_refined(mi1):
 *                     for mi2 in range(self.s2):
 *                         if mm_s._get(mi1, mi2):             # <<<<<<<<<<<<<<
 *                             self.add_neighbors_refined(mi1, mi2)
 *                     # self.push_refined_bool(mi1)
 */
          __pyx_t_8.__pyx_n = 1;
          __pyx_t_8.i2 = __pyx_v_mi2;
          __pyx_t_4 = ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_v_mm_s->__pyx_vtab)->_get(__pyx_v_mm_s, __pyx_v_mi1, &__pyx_t_8); 
          __pyx_t_9 = (__pyx_t_4 != 0);
          if (__pyx_t_9) {

            /* "yt/geometry/particle_oct_container.pyx":1803
 *                     for mi2 in range(self.s2):
 *                         if mm_s._get(mi1, mi2):
 *                             self.add_neighbors_refined(mi1, mi2)             # <<<<<<<<<<<<<<
 *                     # self.push_refined_bool(mi1)
 *                     self.ghosts_ewah._set_refined_array_ptr(mi1,
 */
            ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->add_neighbors_refined(__pyx_v_self, __pyx_v_mi1, __pyx_v_mi2);

            /* "yt/geometry/particle_oct_container.pyx":1802
 *                 if self.is_refined(mi1):
 *                     for mi2 in range(self.s2):
 *                         if mm_s._get(mi1, mi2):             # <<<<<<<<<<<<<<
 *                             self.add_neighbors_refined(mi1, mi2)
 *                     # self.push_refined_bool(mi1)
 */
          }
        }

        /* "yt/geometry/particle_oct_container.pyx":1806
 *                     # self.push_refined_bool(mi1)
 *                     self.ghosts_ewah._set_refined_array_ptr(mi1,
 *                                                             &self.refined_ghosts_bool[0])             # <<<<<<<<<<<<<<
 *                     for mi2 in range(self.s2):
 *                         self.refined_ghosts_bool[mi2] = 0
 */
        if (unlikely(!__pyx_v_self->refined_ghosts_bool.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1806, __pyx_L1_error)}
        __pyx_t_10 = 0;

        /* "yt/geometry/particle_oct_container.pyx":1805
 *                             self.add_neighbors_refined(mi1, mi2)
 *                     # self.push_refined_bool(mi1)
 *                     self.ghosts_ewah._set_refined_array_ptr(mi1,             # <<<<<<<<<<<<<<
 *                                                             &self.refined_ghosts_bool[0])
 *                     for mi2 in range(self.s2):
 */
        ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_v_self->ghosts_ewah->__pyx_vtab)->_set_refined_array_ptr(__pyx_v_self->ghosts_ewah, __pyx_v_mi1, (&(*((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->refined_ghosts_bool.data + __pyx_t_10 * __pyx_v_self->refined_ghosts_bool.strides[0]) )))));

        /* "yt/geometry/particle_oct_container.pyx":1807
 *                     self.ghosts_ewah._set_refined_array_ptr(mi1,
 *                                                             &self.refined_ghosts_bool[0])
 *                     for mi2 in range(self.s2):             # <<<<<<<<<<<<<<
 *                         self.refined_ghosts_bool[mi2] = 0
 *                 else:
 */
        __pyx_t_5 = __pyx_v_self->s2;
        __pyx_t_6 = __pyx_t_5;
        for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
          __pyx_v_mi2 = __pyx_t_7;

          /* "yt/geometry/particle_oct_container.pyx":1808
 *                                                             &self.refined_ghosts_bool[0])
 *                     for mi2 in range(self.s2):
 *                         self.refined_ghosts_bool[mi2] = 0             # <<<<<<<<<<<<<<
 *                 else:
 *                     self.add_neighbors_coarse(mi1)
 */
          if (unlikely(!__pyx_v_self->refined_ghosts_bool.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1808, __pyx_L1_error)}
          __pyx_t_11 = __pyx_v_mi2;
          *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->refined_ghosts_bool.data + __pyx_t_11 * __pyx_v_self->refined_ghosts_bool.strides[0]) )) = 0;
        }

        /* "yt/geometry/particle_oct_container.pyx":1800
 *         for mi1 in range(self.s1):
 *             if mm_s._get_coarse(mi1):
 *                 if self.is_refined(mi1):             # <<<<<<<<<<<<<<
 *                     for mi2 in range(self.s2):
 *                         if mm_s._get(mi1, mi2):
 */
        goto __pyx_L6;
      }

      /* "yt/geometry/particle_oct_container.pyx":1810
 *                         self.refined_ghosts_bool[mi2] = 0
 *                 else:
 *                     self.add_neighbors_coarse(mi1)             # <<<<<<<<<<<<<<
 *         # Add ghost zones to bool array in order
 *         mm_g._set_coarse_array_ptr(&self.coarse_ghosts_bool[0])
 */
      /*else*/ {
        ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->add_neighbors_coarse(__pyx_v_self, __pyx_v_mi1);
      }
      __pyx_L6:;

      /* "yt/geometry/particle_oct_container.pyx":1799
 *         # Get ghost zones, unordered
 *         for mi1 in range(self.s1):
 *             if mm_s._get_coarse(mi1):             # <<<<<<<<<<<<<<
 *                 if self.is_refined(mi1):
 *                     for mi2 in range(self.s2):
 */
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":1812
 *                     self.add_neighbors_coarse(mi1)
 *         # Add ghost zones to bool array in order
 *         mm_g._set_coarse_array_ptr(&self.coarse_ghosts_bool[0])             # <<<<<<<<<<<<<<
 *         for mi1 in range(self.s1):
 *             self.coarse_ghosts_bool[mi1] = 0
 */
  if (unlikely(!__pyx_v_self->coarse_ghosts_bool.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1812, __pyx_L1_error)}
  __pyx_t_10 = 0;
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_v_mm_g->__pyx_vtab)->_set_coarse_array_ptr(__pyx_v_mm_g, (&(*((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->coarse_ghosts_bool.data + __pyx_t_10 * __pyx_v_self->coarse_ghosts_bool.strides[0]) )))));

  /* "yt/geometry/particle_oct_container.pyx":1813
 *         # Add ghost zones to bool array in order
 *         mm_g._set_coarse_array_ptr(&self.coarse_ghosts_bool[0])
 *         for mi1 in range(self.s1):             # <<<<<<<<<<<<<<
 *             self.coarse_ghosts_bool[mi1] = 0
 *         self.refined_ghosts_list._fill_bool(mm_g)
 */
  __pyx_t_1 = __pyx_v_self->s1;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_mi1 = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":1814
 *         mm_g._set_coarse_array_ptr(&self.coarse_ghosts_bool[0])
 *         for mi1 in range(self.s1):
 *             self.coarse_ghosts_bool[mi1] = 0             # <<<<<<<<<<<<<<
 *         self.refined_ghosts_list._fill_bool(mm_g)
 *         mm_g._append(self.ghosts_ewah)
 */
    if (unlikely(!__pyx_v_self->coarse_ghosts_bool.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1814, __pyx_L1_error)}
    __pyx_t_5 = __pyx_v_mi1;
    *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->coarse_ghosts_bool.data + __pyx_t_5 * __pyx_v_self->coarse_ghosts_bool.strides[0]) )) = 0;
  }

  /* "yt/geometry/particle_oct_container.pyx":1815
 *         for mi1 in range(self.s1):
 *             self.coarse_ghosts_bool[mi1] = 0
 *         self.refined_ghosts_list._fill_bool(mm_g)             # <<<<<<<<<<<<<<
 *         mm_g._append(self.ghosts_ewah)
 * 
 */
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet *)__pyx_v_self->refined_ghosts_list->__pyx_vtab)->_fill_bool(__pyx_v_self->refined_ghosts_list, __pyx_v_mm_g);

  /* "yt/geometry/particle_oct_container.pyx":1816
 *             self.coarse_ghosts_bool[mi1] = 0
 *         self.refined_ghosts_list._fill_bool(mm_g)
 *         mm_g._append(self.ghosts_ewah)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_t_12 = ((PyObject *)__pyx_v_self->ghosts_ewah);
  __Pyx_INCREF(__pyx_t_12);
  ((struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_v_mm_g->__pyx_vtab)->_append(__pyx_v_mm_g, ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)__pyx_t_12));
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1795
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void add_ghost_zones(self, BoolArrayColl mm_s, BoolArrayColl mm_g):             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t mi1, mi2, mi1_n, mi2_n
 *         # Get ghost zones, unordered
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_WriteUnraisable("yt.geometry.particle_oct_container.ParticleBitmapSelector.add_ghost_zones", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":1821
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int fill_subcells_mi1(self,             # <<<<<<<<<<<<<<
 *                                np.uint64_t nlevel,
 *                                np.uint64_t ind1[3]) except -1:
 */

static int __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_fill_subcells_mi1(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_nlevel, __pyx_t_5numpy_uint64_t *__pyx_v_ind1) {
  __pyx_t_5numpy_uint64_t __pyx_v_imi;
  __pyx_t_5numpy_uint64_t __pyx_v_fmi;
  __pyx_t_5numpy_uint64_t __pyx_v_mi;
  __pyx_t_5numpy_uint64_t __pyx_v_shift_by;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_uint64_t __pyx_t_1;
  __pyx_t_5numpy_uint64_t __pyx_t_2;
  __pyx_t_5numpy_uint64_t __pyx_t_3;
  struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_coarse __pyx_t_4;
  __Pyx_RefNannySetupContext("fill_subcells_mi1", 0);

  /* "yt/geometry/particle_oct_container.pyx":1826
 *         cdef np.uint64_t imi, fmi
 *         cdef np.uint64_t mi
 *         cdef np.uint64_t shift_by = 3 * (self.bitmap.index_order1 - nlevel)             # <<<<<<<<<<<<<<
 *         imi = encode_morton_64bit(ind1[0], ind1[1], ind1[2]) << shift_by
 *         fmi = imi + (1 << shift_by)
 */
  __pyx_v_shift_by = (3 * (__pyx_v_self->bitmap->index_order1 - __pyx_v_nlevel));

  /* "yt/geometry/particle_oct_container.pyx":1827
 *         cdef np.uint64_t mi
 *         cdef np.uint64_t shift_by = 3 * (self.bitmap.index_order1 - nlevel)
 *         imi = encode_morton_64bit(ind1[0], ind1[1], ind1[2]) << shift_by             # <<<<<<<<<<<<<<
 *         fmi = imi + (1 << shift_by)
 *         for mi in range(imi, fmi):
 */
  __pyx_v_imi = (__pyx_f_2yt_9utilities_3lib_14geometry_utils_encode_morton_64bit((__pyx_v_ind1[0]), (__pyx_v_ind1[1]), (__pyx_v_ind1[2])) << __pyx_v_shift_by);

  /* "yt/geometry/particle_oct_container.pyx":1828
 *         cdef np.uint64_t shift_by = 3 * (self.bitmap.index_order1 - nlevel)
 *         imi = encode_morton_64bit(ind1[0], ind1[1], ind1[2]) << shift_by
 *         fmi = imi + (1 << shift_by)             # <<<<<<<<<<<<<<
 *         for mi in range(imi, fmi):
 *             self.add_coarse(mi, 1)
 */
  __pyx_v_fmi = (__pyx_v_imi + (1 << __pyx_v_shift_by));

  /* "yt/geometry/particle_oct_container.pyx":1829
 *         imi = encode_morton_64bit(ind1[0], ind1[1], ind1[2]) << shift_by
 *         fmi = imi + (1 << shift_by)
 *         for mi in range(imi, fmi):             # <<<<<<<<<<<<<<
 *             self.add_coarse(mi, 1)
 * 
 */
  __pyx_t_1 = __pyx_v_fmi;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = __pyx_v_imi; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_mi = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":1830
 *         fmi = imi + (1 << shift_by)
 *         for mi in range(imi, fmi):
 *             self.add_coarse(mi, 1)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    __pyx_t_4.__pyx_n = 1;
    __pyx_t_4.bbox = 1;
    ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->add_coarse(__pyx_v_self, __pyx_v_mi, &__pyx_t_4); 
  }

  /* "yt/geometry/particle_oct_container.pyx":1821
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int fill_subcells_mi1(self,             # <<<<<<<<<<<<<<
 *                                np.uint64_t nlevel,
 *                                np.uint64_t ind1[3]) except -1:
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1835
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int fill_subcells_mi2(self,             # <<<<<<<<<<<<<<
 *                                np.uint64_t nlevel,
 *                                np.uint64_t mi1,
 */

static int __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_fill_subcells_mi2(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_uint64_t __pyx_v_nlevel, __pyx_t_5numpy_uint64_t __pyx_v_mi1, __pyx_t_5numpy_uint64_t *__pyx_v_ind2) {
  __pyx_t_5numpy_uint64_t __pyx_v_imi;
  __pyx_t_5numpy_uint64_t __pyx_v_fmi;
  __pyx_t_5numpy_uint64_t __pyx_v_shift_by;
  __pyx_t_5numpy_uint64_t __pyx_v_mi2;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_uint64_t __pyx_t_1;
  __pyx_t_5numpy_uint64_t __pyx_t_2;
  __pyx_t_5numpy_uint64_t __pyx_t_3;
  int __pyx_t_4;
  struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_refined __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fill_subcells_mi2", 0);

  /* "yt/geometry/particle_oct_container.pyx":1840
 *                                np.uint64_t ind2[3]) except -1:
 *         cdef np.uint64_t imi, fmi
 *         cdef np.uint64_t shift_by = 3 * ((self.bitmap.index_order2 +             # <<<<<<<<<<<<<<
 *                                           self.bitmap.index_order1) - nlevel)
 *         imi = encode_morton_64bit(ind2[0], ind2[1], ind2[2]) << shift_by
 */
  __pyx_v_shift_by = (3 * ((__pyx_v_self->bitmap->index_order2 + __pyx_v_self->bitmap->index_order1) - __pyx_v_nlevel));

  /* "yt/geometry/particle_oct_container.pyx":1842
 *         cdef np.uint64_t shift_by = 3 * ((self.bitmap.index_order2 +
 *                                           self.bitmap.index_order1) - nlevel)
 *         imi = encode_morton_64bit(ind2[0], ind2[1], ind2[2]) << shift_by             # <<<<<<<<<<<<<<
 *         fmi = imi + (1 << shift_by)
 *         for mi2 in range(imi, fmi):
 */
  __pyx_v_imi = (__pyx_f_2yt_9utilities_3lib_14geometry_utils_encode_morton_64bit((__pyx_v_ind2[0]), (__pyx_v_ind2[1]), (__pyx_v_ind2[2])) << __pyx_v_shift_by);

  /* "yt/geometry/particle_oct_container.pyx":1843
 *                                           self.bitmap.index_order1) - nlevel)
 *         imi = encode_morton_64bit(ind2[0], ind2[1], ind2[2]) << shift_by
 *         fmi = imi + (1 << shift_by)             # <<<<<<<<<<<<<<
 *         for mi2 in range(imi, fmi):
 *             self.add_refined(mi1, mi2, 1)
 */
  __pyx_v_fmi = (__pyx_v_imi + (1 << __pyx_v_shift_by));

  /* "yt/geometry/particle_oct_container.pyx":1844
 *         imi = encode_morton_64bit(ind2[0], ind2[1], ind2[2]) << shift_by
 *         fmi = imi + (1 << shift_by)
 *         for mi2 in range(imi, fmi):             # <<<<<<<<<<<<<<
 *             self.add_refined(mi1, mi2, 1)
 * 
 */
  __pyx_t_1 = __pyx_v_fmi;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = __pyx_v_imi; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_mi2 = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":1845
 *         fmi = imi + (1 << shift_by)
 *         for mi2 in range(imi, fmi):
 *             self.add_refined(mi1, mi2, 1)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    __pyx_t_5.__pyx_n = 1;
    __pyx_t_5.bbox = 1;
    __pyx_t_4 = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->add_refined(__pyx_v_self, __pyx_v_mi1, __pyx_v_mi2, &__pyx_t_5); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 1845, __pyx_L1_error)
  }

  /* "yt/geometry/particle_oct_container.pyx":1835
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int fill_subcells_mi2(self,             # <<<<<<<<<<<<<<
 *                                np.uint64_t nlevel,
 *                                np.uint64_t mi1,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapSelector.fill_subcells_mi2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1850
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int recursive_morton_mask(             # <<<<<<<<<<<<<<
 *         self, np.int32_t level, np.float64_t pos[3],
 *         np.float64_t dds[3], np.uint64_t mi1, np.uint64_t cur_ind[3]) except -1:
 */

static int __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_recursive_morton_mask(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_int32_t __pyx_v_level, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds, __pyx_t_5numpy_uint64_t __pyx_v_mi1, __pyx_t_5numpy_uint64_t *__pyx_v_cur_ind) {
  __pyx_t_5numpy_uint64_t __pyx_v_mi2;
  __pyx_t_5numpy_float64_t __pyx_v_npos[3];
  __pyx_t_5numpy_float64_t __pyx_v_rpos[3];
  __pyx_t_5numpy_float64_t __pyx_v_ndds[3];
  __pyx_t_5numpy_uint64_t __pyx_v_nlevel;
  __pyx_t_5numpy_uint64_t __pyx_v_ncur_ind[3];
  __pyx_t_5numpy_uint64_t *__pyx_v_zeros;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_sbbox;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_uint64_t __pyx_t_1[3];
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_coarse __pyx_t_7;
  struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_refined __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("recursive_morton_mask", 0);

  /* "yt/geometry/particle_oct_container.pyx":1861
 *         cdef np.uint64_t ind2[3]
 *         cdef np.uint64_t ncur_ind[3]
 *         cdef np.uint64_t* zeros = [0, 0, 0]             # <<<<<<<<<<<<<<
 *         cdef int i, j, k, m, sbbox
 *         PyErr_CheckSignals()
 */
  __pyx_t_1[0] = 0;
  __pyx_t_1[1] = 0;
  __pyx_t_1[2] = 0;
  __pyx_v_zeros = __pyx_t_1;

  /* "yt/geometry/particle_oct_container.pyx":1863
 *         cdef np.uint64_t* zeros = [0, 0, 0]
 *         cdef int i, j, k, m, sbbox
 *         PyErr_CheckSignals()             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             ndds[i] = dds[i]/2
 */
  __pyx_t_2 = PyErr_CheckSignals(); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 1863, __pyx_L1_error)

  /* "yt/geometry/particle_oct_container.pyx":1864
 *         cdef int i, j, k, m, sbbox
 *         PyErr_CheckSignals()
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             ndds[i] = dds[i]/2
 *         nlevel = level + 1
 */
  for (__pyx_t_2 = 0; __pyx_t_2 < 3; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "yt/geometry/particle_oct_container.pyx":1865
 *         PyErr_CheckSignals()
 *         for i in range(3):
 *             ndds[i] = dds[i]/2             # <<<<<<<<<<<<<<
 *         nlevel = level + 1
 *         # Loop over octs
 */
    (__pyx_v_ndds[__pyx_v_i]) = ((__pyx_v_dds[__pyx_v_i]) / 2.0);
  }

  /* "yt/geometry/particle_oct_container.pyx":1866
 *         for i in range(3):
 *             ndds[i] = dds[i]/2
 *         nlevel = level + 1             # <<<<<<<<<<<<<<
 *         # Loop over octs
 *         for i in range(2):
 */
  __pyx_v_nlevel = (__pyx_v_level + 1);

  /* "yt/geometry/particle_oct_container.pyx":1868
 *         nlevel = level + 1
 *         # Loop over octs
 *         for i in range(2):             # <<<<<<<<<<<<<<
 *             npos[0] = pos[0] + i*ndds[0]
 *             rpos[0] = npos[0] + ndds[0]
 */
  for (__pyx_t_2 = 0; __pyx_t_2 < 2; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "yt/geometry/particle_oct_container.pyx":1869
 *         # Loop over octs
 *         for i in range(2):
 *             npos[0] = pos[0] + i*ndds[0]             # <<<<<<<<<<<<<<
 *             rpos[0] = npos[0] + ndds[0]
 *             ncur_ind[0] = (cur_ind[0] << 1) + i
 */
    (__pyx_v_npos[0]) = ((__pyx_v_pos[0]) + (__pyx_v_i * (__pyx_v_ndds[0])));

    /* "yt/geometry/particle_oct_container.pyx":1870
 *         for i in range(2):
 *             npos[0] = pos[0] + i*ndds[0]
 *             rpos[0] = npos[0] + ndds[0]             # <<<<<<<<<<<<<<
 *             ncur_ind[0] = (cur_ind[0] << 1) + i
 *             for j in range(2):
 */
    (__pyx_v_rpos[0]) = ((__pyx_v_npos[0]) + (__pyx_v_ndds[0]));

    /* "yt/geometry/particle_oct_container.pyx":1871
 *             npos[0] = pos[0] + i*ndds[0]
 *             rpos[0] = npos[0] + ndds[0]
 *             ncur_ind[0] = (cur_ind[0] << 1) + i             # <<<<<<<<<<<<<<
 *             for j in range(2):
 *                 npos[1] = pos[1] + j*ndds[1]
 */
    (__pyx_v_ncur_ind[0]) = (((__pyx_v_cur_ind[0]) << 1) + __pyx_v_i);

    /* "yt/geometry/particle_oct_container.pyx":1872
 *             rpos[0] = npos[0] + ndds[0]
 *             ncur_ind[0] = (cur_ind[0] << 1) + i
 *             for j in range(2):             # <<<<<<<<<<<<<<
 *                 npos[1] = pos[1] + j*ndds[1]
 *                 rpos[1] = npos[1] + ndds[1]
 */
    for (__pyx_t_3 = 0; __pyx_t_3 < 2; __pyx_t_3+=1) {
      __pyx_v_j = __pyx_t_3;

      /* "yt/geometry/particle_oct_container.pyx":1873
 *             ncur_ind[0] = (cur_ind[0] << 1) + i
 *             for j in range(2):
 *                 npos[1] = pos[1] + j*ndds[1]             # <<<<<<<<<<<<<<
 *                 rpos[1] = npos[1] + ndds[1]
 *                 ncur_ind[1] = (cur_ind[1] << 1) + j
 */
      (__pyx_v_npos[1]) = ((__pyx_v_pos[1]) + (__pyx_v_j * (__pyx_v_ndds[1])));

      /* "yt/geometry/particle_oct_container.pyx":1874
 *             for j in range(2):
 *                 npos[1] = pos[1] + j*ndds[1]
 *                 rpos[1] = npos[1] + ndds[1]             # <<<<<<<<<<<<<<
 *                 ncur_ind[1] = (cur_ind[1] << 1) + j
 *                 for k in range(2):
 */
      (__pyx_v_rpos[1]) = ((__pyx_v_npos[1]) + (__pyx_v_ndds[1]));

      /* "yt/geometry/particle_oct_container.pyx":1875
 *                 npos[1] = pos[1] + j*ndds[1]
 *                 rpos[1] = npos[1] + ndds[1]
 *                 ncur_ind[1] = (cur_ind[1] << 1) + j             # <<<<<<<<<<<<<<
 *                 for k in range(2):
 *                     npos[2] = pos[2] + k*ndds[2]
 */
      (__pyx_v_ncur_ind[1]) = (((__pyx_v_cur_ind[1]) << 1) + __pyx_v_j);

      /* "yt/geometry/particle_oct_container.pyx":1876
 *                 rpos[1] = npos[1] + ndds[1]
 *                 ncur_ind[1] = (cur_ind[1] << 1) + j
 *                 for k in range(2):             # <<<<<<<<<<<<<<
 *                     npos[2] = pos[2] + k*ndds[2]
 *                     rpos[2] = npos[2] + ndds[2]
 */
      for (__pyx_t_4 = 0; __pyx_t_4 < 2; __pyx_t_4+=1) {
        __pyx_v_k = __pyx_t_4;

        /* "yt/geometry/particle_oct_container.pyx":1877
 *                 ncur_ind[1] = (cur_ind[1] << 1) + j
 *                 for k in range(2):
 *                     npos[2] = pos[2] + k*ndds[2]             # <<<<<<<<<<<<<<
 *                     rpos[2] = npos[2] + ndds[2]
 *                     ncur_ind[2] = (cur_ind[2] << 1) + k
 */
        (__pyx_v_npos[2]) = ((__pyx_v_pos[2]) + (__pyx_v_k * (__pyx_v_ndds[2])));

        /* "yt/geometry/particle_oct_container.pyx":1878
 *                 for k in range(2):
 *                     npos[2] = pos[2] + k*ndds[2]
 *                     rpos[2] = npos[2] + ndds[2]             # <<<<<<<<<<<<<<
 *                     ncur_ind[2] = (cur_ind[2] << 1) + k
 *                     # Only recurse into selected cells
 */
        (__pyx_v_rpos[2]) = ((__pyx_v_npos[2]) + (__pyx_v_ndds[2]));

        /* "yt/geometry/particle_oct_container.pyx":1879
 *                     npos[2] = pos[2] + k*ndds[2]
 *                     rpos[2] = npos[2] + ndds[2]
 *                     ncur_ind[2] = (cur_ind[2] << 1) + k             # <<<<<<<<<<<<<<
 *                     # Only recurse into selected cells
 *                     sbbox = self.selector.select_bbox_edge(npos, rpos)
 */
        (__pyx_v_ncur_ind[2]) = (((__pyx_v_cur_ind[2]) << 1) + __pyx_v_k);

        /* "yt/geometry/particle_oct_container.pyx":1881
 *                     ncur_ind[2] = (cur_ind[2] << 1) + k
 *                     # Only recurse into selected cells
 *                     sbbox = self.selector.select_bbox_edge(npos, rpos)             # <<<<<<<<<<<<<<
 *                     if sbbox == 0:
 *                         continue
 */
        __pyx_v_sbbox = ((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->selector->__pyx_vtab)->select_bbox_edge(__pyx_v_self->selector, __pyx_v_npos, __pyx_v_rpos);

        /* "yt/geometry/particle_oct_container.pyx":1882
 *                     # Only recurse into selected cells
 *                     sbbox = self.selector.select_bbox_edge(npos, rpos)
 *                     if sbbox == 0:             # <<<<<<<<<<<<<<
 *                         continue
 *                     if nlevel < self.order1:
 */
        __pyx_t_5 = ((__pyx_v_sbbox == 0) != 0);
        if (__pyx_t_5) {

          /* "yt/geometry/particle_oct_container.pyx":1883
 *                     sbbox = self.selector.select_bbox_edge(npos, rpos)
 *                     if sbbox == 0:
 *                         continue             # <<<<<<<<<<<<<<
 *                     if nlevel < self.order1:
 *                         if sbbox == 1:
 */
          goto __pyx_L9_continue;

          /* "yt/geometry/particle_oct_container.pyx":1882
 *                     # Only recurse into selected cells
 *                     sbbox = self.selector.select_bbox_edge(npos, rpos)
 *                     if sbbox == 0:             # <<<<<<<<<<<<<<
 *                         continue
 *                     if nlevel < self.order1:
 */
        }

        /* "yt/geometry/particle_oct_container.pyx":1884
 *                     if sbbox == 0:
 *                         continue
 *                     if nlevel < self.order1:             # <<<<<<<<<<<<<<
 *                         if sbbox == 1:
 *                             self.fill_subcells_mi1(nlevel, ncur_ind)
 */
        __pyx_t_5 = ((__pyx_v_nlevel < __pyx_v_self->order1) != 0);
        if (__pyx_t_5) {

          /* "yt/geometry/particle_oct_container.pyx":1885
 *                         continue
 *                     if nlevel < self.order1:
 *                         if sbbox == 1:             # <<<<<<<<<<<<<<
 *                             self.fill_subcells_mi1(nlevel, ncur_ind)
 *                         else:
 */
          __pyx_t_5 = ((__pyx_v_sbbox == 1) != 0);
          if (__pyx_t_5) {

            /* "yt/geometry/particle_oct_container.pyx":1886
 *                     if nlevel < self.order1:
 *                         if sbbox == 1:
 *                             self.fill_subcells_mi1(nlevel, ncur_ind)             # <<<<<<<<<<<<<<
 *                         else:
 *                             self.recursive_morton_mask(
 */
            __pyx_t_6 = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->fill_subcells_mi1(__pyx_v_self, __pyx_v_nlevel, __pyx_v_ncur_ind); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1886, __pyx_L1_error)

            /* "yt/geometry/particle_oct_container.pyx":1885
 *                         continue
 *                     if nlevel < self.order1:
 *                         if sbbox == 1:             # <<<<<<<<<<<<<<
 *                             self.fill_subcells_mi1(nlevel, ncur_ind)
 *                         else:
 */
            goto __pyx_L13;
          }

          /* "yt/geometry/particle_oct_container.pyx":1888
 *                             self.fill_subcells_mi1(nlevel, ncur_ind)
 *                         else:
 *                             self.recursive_morton_mask(             # <<<<<<<<<<<<<<
 *                                 nlevel, npos, ndds, mi1, ncur_ind)
 *                     elif nlevel == self.order1:
 */
          /*else*/ {

            /* "yt/geometry/particle_oct_container.pyx":1889
 *                         else:
 *                             self.recursive_morton_mask(
 *                                 nlevel, npos, ndds, mi1, ncur_ind)             # <<<<<<<<<<<<<<
 *                     elif nlevel == self.order1:
 *                         mi1 = encode_morton_64bit(
 */
            __pyx_t_6 = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->recursive_morton_mask(__pyx_v_self, __pyx_v_nlevel, __pyx_v_npos, __pyx_v_ndds, __pyx_v_mi1, __pyx_v_ncur_ind); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1888, __pyx_L1_error)
          }
          __pyx_L13:;

          /* "yt/geometry/particle_oct_container.pyx":1884
 *                     if sbbox == 0:
 *                         continue
 *                     if nlevel < self.order1:             # <<<<<<<<<<<<<<
 *                         if sbbox == 1:
 *                             self.fill_subcells_mi1(nlevel, ncur_ind)
 */
          goto __pyx_L12;
        }

        /* "yt/geometry/particle_oct_container.pyx":1890
 *                             self.recursive_morton_mask(
 *                                 nlevel, npos, ndds, mi1, ncur_ind)
 *                     elif nlevel == self.order1:             # <<<<<<<<<<<<<<
 *                         mi1 = encode_morton_64bit(
 *                             ncur_ind[0], ncur_ind[1], ncur_ind[2])
 */
        __pyx_t_5 = ((__pyx_v_nlevel == __pyx_v_self->order1) != 0);
        if (__pyx_t_5) {

          /* "yt/geometry/particle_oct_container.pyx":1891
 *                                 nlevel, npos, ndds, mi1, ncur_ind)
 *                     elif nlevel == self.order1:
 *                         mi1 = encode_morton_64bit(             # <<<<<<<<<<<<<<
 *                             ncur_ind[0], ncur_ind[1], ncur_ind[2])
 *                         if sbbox == 2: # an edge cell
 */
          __pyx_v_mi1 = __pyx_f_2yt_9utilities_3lib_14geometry_utils_encode_morton_64bit((__pyx_v_ncur_ind[0]), (__pyx_v_ncur_ind[1]), (__pyx_v_ncur_ind[2]));

          /* "yt/geometry/particle_oct_container.pyx":1893
 *                         mi1 = encode_morton_64bit(
 *                             ncur_ind[0], ncur_ind[1], ncur_ind[2])
 *                         if sbbox == 2: # an edge cell             # <<<<<<<<<<<<<<
 *                             if self.is_refined(mi1) == 1:
 *                                 # note we pass zeros here in the last argument
 */
          __pyx_t_5 = ((__pyx_v_sbbox == 2) != 0);
          if (__pyx_t_5) {

            /* "yt/geometry/particle_oct_container.pyx":1894
 *                             ncur_ind[0], ncur_ind[1], ncur_ind[2])
 *                         if sbbox == 2: # an edge cell
 *                             if self.is_refined(mi1) == 1:             # <<<<<<<<<<<<<<
 *                                 # note we pass zeros here in the last argument
 *                                 # this is because we now need to generate
 */
            __pyx_t_5 = ((((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->is_refined(__pyx_v_self, __pyx_v_mi1) == 1) != 0);
            if (__pyx_t_5) {

              /* "yt/geometry/particle_oct_container.pyx":1904
 *                                 # calculated in each stack frame) so this is
 *                                 # safe
 *                                 self.recursive_morton_mask(             # <<<<<<<<<<<<<<
 *                                     nlevel, npos, ndds, mi1, zeros)
 *                         self.add_coarse(mi1, sbbox)
 */
              __pyx_t_6 = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->recursive_morton_mask(__pyx_v_self, __pyx_v_nlevel, __pyx_v_npos, __pyx_v_ndds, __pyx_v_mi1, __pyx_v_zeros); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1904, __pyx_L1_error)

              /* "yt/geometry/particle_oct_container.pyx":1894
 *                             ncur_ind[0], ncur_ind[1], ncur_ind[2])
 *                         if sbbox == 2: # an edge cell
 *                             if self.is_refined(mi1) == 1:             # <<<<<<<<<<<<<<
 *                                 # note we pass zeros here in the last argument
 *                                 # this is because we now need to generate
 */
            }

            /* "yt/geometry/particle_oct_container.pyx":1893
 *                         mi1 = encode_morton_64bit(
 *                             ncur_ind[0], ncur_ind[1], ncur_ind[2])
 *                         if sbbox == 2: # an edge cell             # <<<<<<<<<<<<<<
 *                             if self.is_refined(mi1) == 1:
 *                                 # note we pass zeros here in the last argument
 */
          }

          /* "yt/geometry/particle_oct_container.pyx":1906
 *                                 self.recursive_morton_mask(
 *                                     nlevel, npos, ndds, mi1, zeros)
 *                         self.add_coarse(mi1, sbbox)             # <<<<<<<<<<<<<<
 *                         self.push_refined_bool(mi1)
 *                     elif nlevel < (self.order1 + self.order2):
 */
          __pyx_t_7.__pyx_n = 1;
          __pyx_t_7.bbox = __pyx_v_sbbox;
          ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->add_coarse(__pyx_v_self, __pyx_v_mi1, &__pyx_t_7); 

          /* "yt/geometry/particle_oct_container.pyx":1907
 *                                     nlevel, npos, ndds, mi1, zeros)
 *                         self.add_coarse(mi1, sbbox)
 *                         self.push_refined_bool(mi1)             # <<<<<<<<<<<<<<
 *                     elif nlevel < (self.order1 + self.order2):
 *                         if sbbox == 1:
 */
          ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->push_refined_bool(__pyx_v_self, __pyx_v_mi1);

          /* "yt/geometry/particle_oct_container.pyx":1890
 *                             self.recursive_morton_mask(
 *                                 nlevel, npos, ndds, mi1, ncur_ind)
 *                     elif nlevel == self.order1:             # <<<<<<<<<<<<<<
 *                         mi1 = encode_morton_64bit(
 *                             ncur_ind[0], ncur_ind[1], ncur_ind[2])
 */
          goto __pyx_L12;
        }

        /* "yt/geometry/particle_oct_container.pyx":1908
 *                         self.add_coarse(mi1, sbbox)
 *                         self.push_refined_bool(mi1)
 *                     elif nlevel < (self.order1 + self.order2):             # <<<<<<<<<<<<<<
 *                         if sbbox == 1:
 *                             self.fill_subcells_mi2(nlevel, mi1, ncur_ind)
 */
        __pyx_t_5 = ((__pyx_v_nlevel < (__pyx_v_self->order1 + __pyx_v_self->order2)) != 0);
        if (__pyx_t_5) {

          /* "yt/geometry/particle_oct_container.pyx":1909
 *                         self.push_refined_bool(mi1)
 *                     elif nlevel < (self.order1 + self.order2):
 *                         if sbbox == 1:             # <<<<<<<<<<<<<<
 *                             self.fill_subcells_mi2(nlevel, mi1, ncur_ind)
 *                         else:
 */
          __pyx_t_5 = ((__pyx_v_sbbox == 1) != 0);
          if (__pyx_t_5) {

            /* "yt/geometry/particle_oct_container.pyx":1910
 *                     elif nlevel < (self.order1 + self.order2):
 *                         if sbbox == 1:
 *                             self.fill_subcells_mi2(nlevel, mi1, ncur_ind)             # <<<<<<<<<<<<<<
 *                         else:
 *                             self.recursive_morton_mask(
 */
            __pyx_t_6 = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->fill_subcells_mi2(__pyx_v_self, __pyx_v_nlevel, __pyx_v_mi1, __pyx_v_ncur_ind); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1910, __pyx_L1_error)

            /* "yt/geometry/particle_oct_container.pyx":1909
 *                         self.push_refined_bool(mi1)
 *                     elif nlevel < (self.order1 + self.order2):
 *                         if sbbox == 1:             # <<<<<<<<<<<<<<
 *                             self.fill_subcells_mi2(nlevel, mi1, ncur_ind)
 *                         else:
 */
            goto __pyx_L16;
          }

          /* "yt/geometry/particle_oct_container.pyx":1912
 *                             self.fill_subcells_mi2(nlevel, mi1, ncur_ind)
 *                         else:
 *                             self.recursive_morton_mask(             # <<<<<<<<<<<<<<
 *                                 nlevel, npos, ndds, mi1, ncur_ind)
 *                     elif nlevel == (self.order1 + self.order2):
 */
          /*else*/ {

            /* "yt/geometry/particle_oct_container.pyx":1913
 *                         else:
 *                             self.recursive_morton_mask(
 *                                 nlevel, npos, ndds, mi1, ncur_ind)             # <<<<<<<<<<<<<<
 *                     elif nlevel == (self.order1 + self.order2):
 *                         mi2 = encode_morton_64bit(
 */
            __pyx_t_6 = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->recursive_morton_mask(__pyx_v_self, __pyx_v_nlevel, __pyx_v_npos, __pyx_v_ndds, __pyx_v_mi1, __pyx_v_ncur_ind); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1912, __pyx_L1_error)
          }
          __pyx_L16:;

          /* "yt/geometry/particle_oct_container.pyx":1908
 *                         self.add_coarse(mi1, sbbox)
 *                         self.push_refined_bool(mi1)
 *                     elif nlevel < (self.order1 + self.order2):             # <<<<<<<<<<<<<<
 *                         if sbbox == 1:
 *                             self.fill_subcells_mi2(nlevel, mi1, ncur_ind)
 */
          goto __pyx_L12;
        }

        /* "yt/geometry/particle_oct_container.pyx":1914
 *                             self.recursive_morton_mask(
 *                                 nlevel, npos, ndds, mi1, ncur_ind)
 *                     elif nlevel == (self.order1 + self.order2):             # <<<<<<<<<<<<<<
 *                         mi2 = encode_morton_64bit(
 *                             ncur_ind[0], ncur_ind[1], ncur_ind[2])
 */
        __pyx_t_5 = ((__pyx_v_nlevel == (__pyx_v_self->order1 + __pyx_v_self->order2)) != 0);
        if (__pyx_t_5) {

          /* "yt/geometry/particle_oct_container.pyx":1915
 *                                 nlevel, npos, ndds, mi1, ncur_ind)
 *                     elif nlevel == (self.order1 + self.order2):
 *                         mi2 = encode_morton_64bit(             # <<<<<<<<<<<<<<
 *                             ncur_ind[0], ncur_ind[1], ncur_ind[2])
 *                         self.add_refined(mi1, mi2, sbbox)
 */
          __pyx_v_mi2 = __pyx_f_2yt_9utilities_3lib_14geometry_utils_encode_morton_64bit((__pyx_v_ncur_ind[0]), (__pyx_v_ncur_ind[1]), (__pyx_v_ncur_ind[2]));

          /* "yt/geometry/particle_oct_container.pyx":1917
 *                         mi2 = encode_morton_64bit(
 *                             ncur_ind[0], ncur_ind[1], ncur_ind[2])
 *                         self.add_refined(mi1, mi2, sbbox)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
          __pyx_t_8.__pyx_n = 1;
          __pyx_t_8.bbox = __pyx_v_sbbox;
          __pyx_t_6 = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->add_refined(__pyx_v_self, __pyx_v_mi1, __pyx_v_mi2, &__pyx_t_8); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1917, __pyx_L1_error)

          /* "yt/geometry/particle_oct_container.pyx":1914
 *                             self.recursive_morton_mask(
 *                                 nlevel, npos, ndds, mi1, ncur_ind)
 *                     elif nlevel == (self.order1 + self.order2):             # <<<<<<<<<<<<<<
 *                         mi2 = encode_morton_64bit(
 *                             ncur_ind[0], ncur_ind[1], ncur_ind[2])
 */
        }
        __pyx_L12:;
        __pyx_L9_continue:;
      }
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":1918
 *                             ncur_ind[0], ncur_ind[1], ncur_ind[2])
 *                         self.add_refined(mi1, mi2, sbbox)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":1850
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef int recursive_morton_mask(             # <<<<<<<<<<<<<<
 *         self, np.int32_t level, np.float64_t pos[3],
 *         np.float64_t dds[3], np.uint64_t mi1, np.uint64_t cur_ind[3]) except -1:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapSelector.recursive_morton_mask", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1923
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void recursive_morton_files(self, np.int32_t level, np.float64_t pos[3],             # <<<<<<<<<<<<<<
 *                                      np.float64_t dds[3], np.uint64_t mi1):
 *         cdef np.uint64_t mi2
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_recursive_morton_files(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, __pyx_t_5numpy_int32_t __pyx_v_level, __pyx_t_5numpy_float64_t *__pyx_v_pos, __pyx_t_5numpy_float64_t *__pyx_v_dds, __pyx_t_5numpy_uint64_t __pyx_v_mi1) {
  __pyx_t_5numpy_uint64_t __pyx_v_mi2;
  __pyx_t_5numpy_float64_t __pyx_v_npos[3];
  __pyx_t_5numpy_float64_t __pyx_v_rpos[3];
  __pyx_t_5numpy_float64_t __pyx_v_ndds[3];
  __pyx_t_5numpy_uint64_t __pyx_v_nlevel;
  __pyx_t_5numpy_float64_t __pyx_v_DLE[3];
  __pyx_t_5numpy_uint64_t __pyx_v_ind1[3];
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_m;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("recursive_morton_files", 0);

  /* "yt/geometry/particle_oct_container.pyx":1934
 *         cdef np.uint64_t ind2[3]
 *         cdef int i, j, k, m
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             ndds[i] = dds[i]/2
 *         nlevel = level + 1
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":1935
 *         cdef int i, j, k, m
 *         for i in range(3):
 *             ndds[i] = dds[i]/2             # <<<<<<<<<<<<<<
 *         nlevel = level + 1
 *         # Loop over octs
 */
    (__pyx_v_ndds[__pyx_v_i]) = ((__pyx_v_dds[__pyx_v_i]) / 2.0);
  }

  /* "yt/geometry/particle_oct_container.pyx":1936
 *         for i in range(3):
 *             ndds[i] = dds[i]/2
 *         nlevel = level + 1             # <<<<<<<<<<<<<<
 *         # Loop over octs
 *         for i in range(2):
 */
  __pyx_v_nlevel = (__pyx_v_level + 1);

  /* "yt/geometry/particle_oct_container.pyx":1938
 *         nlevel = level + 1
 *         # Loop over octs
 *         for i in range(2):             # <<<<<<<<<<<<<<
 *             npos[0] = pos[0] + i*ndds[0]
 *             rpos[0] = npos[0] + ndds[0]
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 2; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":1939
 *         # Loop over octs
 *         for i in range(2):
 *             npos[0] = pos[0] + i*ndds[0]             # <<<<<<<<<<<<<<
 *             rpos[0] = npos[0] + ndds[0]
 *             for j in range(2):
 */
    (__pyx_v_npos[0]) = ((__pyx_v_pos[0]) + (__pyx_v_i * (__pyx_v_ndds[0])));

    /* "yt/geometry/particle_oct_container.pyx":1940
 *         for i in range(2):
 *             npos[0] = pos[0] + i*ndds[0]
 *             rpos[0] = npos[0] + ndds[0]             # <<<<<<<<<<<<<<
 *             for j in range(2):
 *                 npos[1] = pos[1] + j*ndds[1]
 */
    (__pyx_v_rpos[0]) = ((__pyx_v_npos[0]) + (__pyx_v_ndds[0]));

    /* "yt/geometry/particle_oct_container.pyx":1941
 *             npos[0] = pos[0] + i*ndds[0]
 *             rpos[0] = npos[0] + ndds[0]
 *             for j in range(2):             # <<<<<<<<<<<<<<
 *                 npos[1] = pos[1] + j*ndds[1]
 *                 rpos[1] = npos[1] + ndds[1]
 */
    for (__pyx_t_2 = 0; __pyx_t_2 < 2; __pyx_t_2+=1) {
      __pyx_v_j = __pyx_t_2;

      /* "yt/geometry/particle_oct_container.pyx":1942
 *             rpos[0] = npos[0] + ndds[0]
 *             for j in range(2):
 *                 npos[1] = pos[1] + j*ndds[1]             # <<<<<<<<<<<<<<
 *                 rpos[1] = npos[1] + ndds[1]
 *                 for k in range(2):
 */
      (__pyx_v_npos[1]) = ((__pyx_v_pos[1]) + (__pyx_v_j * (__pyx_v_ndds[1])));

      /* "yt/geometry/particle_oct_container.pyx":1943
 *             for j in range(2):
 *                 npos[1] = pos[1] + j*ndds[1]
 *                 rpos[1] = npos[1] + ndds[1]             # <<<<<<<<<<<<<<
 *                 for k in range(2):
 *                     npos[2] = pos[2] + k*ndds[2]
 */
      (__pyx_v_rpos[1]) = ((__pyx_v_npos[1]) + (__pyx_v_ndds[1]));

      /* "yt/geometry/particle_oct_container.pyx":1944
 *                 npos[1] = pos[1] + j*ndds[1]
 *                 rpos[1] = npos[1] + ndds[1]
 *                 for k in range(2):             # <<<<<<<<<<<<<<
 *                     npos[2] = pos[2] + k*ndds[2]
 *                     rpos[2] = npos[2] + ndds[2]
 */
      for (__pyx_t_3 = 0; __pyx_t_3 < 2; __pyx_t_3+=1) {
        __pyx_v_k = __pyx_t_3;

        /* "yt/geometry/particle_oct_container.pyx":1945
 *                 rpos[1] = npos[1] + ndds[1]
 *                 for k in range(2):
 *                     npos[2] = pos[2] + k*ndds[2]             # <<<<<<<<<<<<<<
 *                     rpos[2] = npos[2] + ndds[2]
 *                     # Only recurse into selected cells
 */
        (__pyx_v_npos[2]) = ((__pyx_v_pos[2]) + (__pyx_v_k * (__pyx_v_ndds[2])));

        /* "yt/geometry/particle_oct_container.pyx":1946
 *                 for k in range(2):
 *                     npos[2] = pos[2] + k*ndds[2]
 *                     rpos[2] = npos[2] + ndds[2]             # <<<<<<<<<<<<<<
 *                     # Only recurse into selected cells
 *                     if not self.selector.select_bbox(npos, rpos): continue
 */
        (__pyx_v_rpos[2]) = ((__pyx_v_npos[2]) + (__pyx_v_ndds[2]));

        /* "yt/geometry/particle_oct_container.pyx":1948
 *                     rpos[2] = npos[2] + ndds[2]
 *                     # Only recurse into selected cells
 *                     if not self.selector.select_bbox(npos, rpos): continue             # <<<<<<<<<<<<<<
 *                     if nlevel < self.order1:
 *                         self.recursive_morton_files(nlevel, npos, ndds, mi1)
 */
        __pyx_t_4 = ((!(((struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_v_self->selector->__pyx_vtab)->select_bbox(__pyx_v_self->selector, __pyx_v_npos, __pyx_v_rpos) != 0)) != 0);
        if (__pyx_t_4) {
          goto __pyx_L9_continue;
        }

        /* "yt/geometry/particle_oct_container.pyx":1949
 *                     # Only recurse into selected cells
 *                     if not self.selector.select_bbox(npos, rpos): continue
 *                     if nlevel < self.order1:             # <<<<<<<<<<<<<<
 *                         self.recursive_morton_files(nlevel, npos, ndds, mi1)
 *                     elif nlevel == self.order1:
 */
        __pyx_t_4 = ((__pyx_v_nlevel < __pyx_v_self->order1) != 0);
        if (__pyx_t_4) {

          /* "yt/geometry/particle_oct_container.pyx":1950
 *                     if not self.selector.select_bbox(npos, rpos): continue
 *                     if nlevel < self.order1:
 *                         self.recursive_morton_files(nlevel, npos, ndds, mi1)             # <<<<<<<<<<<<<<
 *                     elif nlevel == self.order1:
 *                         mi1 = bounded_morton_dds(npos[0], npos[1], npos[2], self.DLE, ndds)
 */
          ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->recursive_morton_files(__pyx_v_self, __pyx_v_nlevel, __pyx_v_npos, __pyx_v_ndds, __pyx_v_mi1);

          /* "yt/geometry/particle_oct_container.pyx":1949
 *                     # Only recurse into selected cells
 *                     if not self.selector.select_bbox(npos, rpos): continue
 *                     if nlevel < self.order1:             # <<<<<<<<<<<<<<
 *                         self.recursive_morton_files(nlevel, npos, ndds, mi1)
 *                     elif nlevel == self.order1:
 */
          goto __pyx_L12;
        }

        /* "yt/geometry/particle_oct_container.pyx":1951
 *                     if nlevel < self.order1:
 *                         self.recursive_morton_files(nlevel, npos, ndds, mi1)
 *                     elif nlevel == self.order1:             # <<<<<<<<<<<<<<
 *                         mi1 = bounded_morton_dds(npos[0], npos[1], npos[2], self.DLE, ndds)
 *                         if self.is_refined_files(mi1):
 */
        __pyx_t_4 = ((__pyx_v_nlevel == __pyx_v_self->order1) != 0);
        if (__pyx_t_4) {

          /* "yt/geometry/particle_oct_container.pyx":1952
 *                         self.recursive_morton_files(nlevel, npos, ndds, mi1)
 *                     elif nlevel == self.order1:
 *                         mi1 = bounded_morton_dds(npos[0], npos[1], npos[2], self.DLE, ndds)             # <<<<<<<<<<<<<<
 *                         if self.is_refined_files(mi1):
 *                             self.recursive_morton_files(nlevel, npos, ndds, mi1)
 */
          __pyx_v_mi1 = __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_dds((__pyx_v_npos[0]), (__pyx_v_npos[1]), (__pyx_v_npos[2]), __pyx_v_self->DLE, __pyx_v_ndds);

          /* "yt/geometry/particle_oct_container.pyx":1953
 *                     elif nlevel == self.order1:
 *                         mi1 = bounded_morton_dds(npos[0], npos[1], npos[2], self.DLE, ndds)
 *                         if self.is_refined_files(mi1):             # <<<<<<<<<<<<<<
 *                             self.recursive_morton_files(nlevel, npos, ndds, mi1)
 *                         self.set_files_coarse(mi1)
 */
          __pyx_t_4 = (((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->is_refined_files(__pyx_v_self, __pyx_v_mi1) != 0);
          if (__pyx_t_4) {

            /* "yt/geometry/particle_oct_container.pyx":1954
 *                         mi1 = bounded_morton_dds(npos[0], npos[1], npos[2], self.DLE, ndds)
 *                         if self.is_refined_files(mi1):
 *                             self.recursive_morton_files(nlevel, npos, ndds, mi1)             # <<<<<<<<<<<<<<
 *                         self.set_files_coarse(mi1)
 *                     elif nlevel < (self.order1 + self.order2):
 */
            ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->recursive_morton_files(__pyx_v_self, __pyx_v_nlevel, __pyx_v_npos, __pyx_v_ndds, __pyx_v_mi1);

            /* "yt/geometry/particle_oct_container.pyx":1953
 *                     elif nlevel == self.order1:
 *                         mi1 = bounded_morton_dds(npos[0], npos[1], npos[2], self.DLE, ndds)
 *                         if self.is_refined_files(mi1):             # <<<<<<<<<<<<<<
 *                             self.recursive_morton_files(nlevel, npos, ndds, mi1)
 *                         self.set_files_coarse(mi1)
 */
          }

          /* "yt/geometry/particle_oct_container.pyx":1955
 *                         if self.is_refined_files(mi1):
 *                             self.recursive_morton_files(nlevel, npos, ndds, mi1)
 *                         self.set_files_coarse(mi1)             # <<<<<<<<<<<<<<
 *                     elif nlevel < (self.order1 + self.order2):
 *                         self.recursive_morton_files(nlevel, npos, ndds, mi1)
 */
          ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->set_files_coarse(__pyx_v_self, __pyx_v_mi1);

          /* "yt/geometry/particle_oct_container.pyx":1951
 *                     if nlevel < self.order1:
 *                         self.recursive_morton_files(nlevel, npos, ndds, mi1)
 *                     elif nlevel == self.order1:             # <<<<<<<<<<<<<<
 *                         mi1 = bounded_morton_dds(npos[0], npos[1], npos[2], self.DLE, ndds)
 *                         if self.is_refined_files(mi1):
 */
          goto __pyx_L12;
        }

        /* "yt/geometry/particle_oct_container.pyx":1956
 *                             self.recursive_morton_files(nlevel, npos, ndds, mi1)
 *                         self.set_files_coarse(mi1)
 *                     elif nlevel < (self.order1 + self.order2):             # <<<<<<<<<<<<<<
 *                         self.recursive_morton_files(nlevel, npos, ndds, mi1)
 *                     elif nlevel == (self.order1 + self.order2):
 */
        __pyx_t_4 = ((__pyx_v_nlevel < (__pyx_v_self->order1 + __pyx_v_self->order2)) != 0);
        if (__pyx_t_4) {

          /* "yt/geometry/particle_oct_container.pyx":1957
 *                         self.set_files_coarse(mi1)
 *                     elif nlevel < (self.order1 + self.order2):
 *                         self.recursive_morton_files(nlevel, npos, ndds, mi1)             # <<<<<<<<<<<<<<
 *                     elif nlevel == (self.order1 + self.order2):
 *                         decode_morton_64bit(mi1,ind1)
 */
          ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->recursive_morton_files(__pyx_v_self, __pyx_v_nlevel, __pyx_v_npos, __pyx_v_ndds, __pyx_v_mi1);

          /* "yt/geometry/particle_oct_container.pyx":1956
 *                             self.recursive_morton_files(nlevel, npos, ndds, mi1)
 *                         self.set_files_coarse(mi1)
 *                     elif nlevel < (self.order1 + self.order2):             # <<<<<<<<<<<<<<
 *                         self.recursive_morton_files(nlevel, npos, ndds, mi1)
 *                     elif nlevel == (self.order1 + self.order2):
 */
          goto __pyx_L12;
        }

        /* "yt/geometry/particle_oct_container.pyx":1958
 *                     elif nlevel < (self.order1 + self.order2):
 *                         self.recursive_morton_files(nlevel, npos, ndds, mi1)
 *                     elif nlevel == (self.order1 + self.order2):             # <<<<<<<<<<<<<<
 *                         decode_morton_64bit(mi1,ind1)
 *                         for m in range(3):
 */
        __pyx_t_4 = ((__pyx_v_nlevel == (__pyx_v_self->order1 + __pyx_v_self->order2)) != 0);
        if (__pyx_t_4) {

          /* "yt/geometry/particle_oct_container.pyx":1959
 *                         self.recursive_morton_files(nlevel, npos, ndds, mi1)
 *                     elif nlevel == (self.order1 + self.order2):
 *                         decode_morton_64bit(mi1,ind1)             # <<<<<<<<<<<<<<
 *                         for m in range(3):
 *                             DLE[m] = self.DLE[m] + ndds[m]*ind1[m]*self.max_index2
 */
          __pyx_f_2yt_9utilities_3lib_14geometry_utils_decode_morton_64bit(__pyx_v_mi1, __pyx_v_ind1);

          /* "yt/geometry/particle_oct_container.pyx":1960
 *                     elif nlevel == (self.order1 + self.order2):
 *                         decode_morton_64bit(mi1,ind1)
 *                         for m in range(3):             # <<<<<<<<<<<<<<
 *                             DLE[m] = self.DLE[m] + ndds[m]*ind1[m]*self.max_index2
 *                         mi2 = bounded_morton_dds(npos[0], npos[1], npos[2], DLE, ndds)
 */
          for (__pyx_t_5 = 0; __pyx_t_5 < 3; __pyx_t_5+=1) {
            __pyx_v_m = __pyx_t_5;

            /* "yt/geometry/particle_oct_container.pyx":1961
 *                         decode_morton_64bit(mi1,ind1)
 *                         for m in range(3):
 *                             DLE[m] = self.DLE[m] + ndds[m]*ind1[m]*self.max_index2             # <<<<<<<<<<<<<<
 *                         mi2 = bounded_morton_dds(npos[0], npos[1], npos[2], DLE, ndds)
 *                         self.set_files_refined(mi1,mi2)
 */
            (__pyx_v_DLE[__pyx_v_m]) = ((__pyx_v_self->DLE[__pyx_v_m]) + (((__pyx_v_ndds[__pyx_v_m]) * (__pyx_v_ind1[__pyx_v_m])) * __pyx_v_self->max_index2));
          }

          /* "yt/geometry/particle_oct_container.pyx":1962
 *                         for m in range(3):
 *                             DLE[m] = self.DLE[m] + ndds[m]*ind1[m]*self.max_index2
 *                         mi2 = bounded_morton_dds(npos[0], npos[1], npos[2], DLE, ndds)             # <<<<<<<<<<<<<<
 *                         self.set_files_refined(mi1,mi2)
 * 
 */
          __pyx_v_mi2 = __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_dds((__pyx_v_npos[0]), (__pyx_v_npos[1]), (__pyx_v_npos[2]), __pyx_v_DLE, __pyx_v_ndds);

          /* "yt/geometry/particle_oct_container.pyx":1963
 *                             DLE[m] = self.DLE[m] + ndds[m]*ind1[m]*self.max_index2
 *                         mi2 = bounded_morton_dds(npos[0], npos[1], npos[2], DLE, ndds)
 *                         self.set_files_refined(mi1,mi2)             # <<<<<<<<<<<<<<
 * 
 * cdef class ParticleBitmapOctreeContainer(SparseOctreeContainer):
 */
          ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self->__pyx_vtab)->set_files_refined(__pyx_v_self, __pyx_v_mi1, __pyx_v_mi2);

          /* "yt/geometry/particle_oct_container.pyx":1958
 *                     elif nlevel < (self.order1 + self.order2):
 *                         self.recursive_morton_files(nlevel, npos, ndds, mi1)
 *                     elif nlevel == (self.order1 + self.order2):             # <<<<<<<<<<<<<<
 *                         decode_morton_64bit(mi1,ind1)
 *                         for m in range(3):
 */
        }
        __pyx_L12:;
        __pyx_L9_continue:;
      }
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":1923
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void recursive_morton_files(self, np.int32_t level, np.float64_t pos[3],             # <<<<<<<<<<<<<<
 *                                      np.float64_t dds[3], np.uint64_t mi1):
 *         cdef np.uint64_t mi2
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_6__reduce_cython__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapSelector.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_8__setstate_cython__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapSelector.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1977
 *     cdef np.uint64_t[:] _octs_per_root
 *     cdef public int overlap_cells
 *     def __init__(self, domain_dimensions, domain_left_edge, domain_right_edge,             # <<<<<<<<<<<<<<
 *                  int num_root, num_zones = 2):
 *         super(ParticleBitmapOctreeContainer, self).__init__(
 */

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_domain_dimensions = 0;
  PyObject *__pyx_v_domain_left_edge = 0;
  PyObject *__pyx_v_domain_right_edge = 0;
  int __pyx_v_num_root;
  PyObject *__pyx_v_num_zones = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_domain_dimensions,&__pyx_n_s_domain_left_edge,&__pyx_n_s_domain_right_edge,&__pyx_n_s_num_root,&__pyx_n_s_num_zones,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[4] = ((PyObject *)__pyx_int_2);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_domain_dimensions)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_domain_left_edge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 5, 1); __PYX_ERR(0, 1977, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_domain_right_edge)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 5, 2); __PYX_ERR(0, 1977, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_root)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 5, 3); __PYX_ERR(0, 1977, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_zones);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1977, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_domain_dimensions = values[0];
    __pyx_v_domain_left_edge = values[1];
    __pyx_v_domain_right_edge = values[2];
    __pyx_v_num_root = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_num_root == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1978, __pyx_L3_error)
    __pyx_v_num_zones = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1977, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer___init__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self), __pyx_v_domain_dimensions, __pyx_v_domain_left_edge, __pyx_v_domain_right_edge, __pyx_v_num_root, __pyx_v_num_zones);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer___init__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, PyObject *__pyx_v_domain_dimensions, PyObject *__pyx_v_domain_left_edge, PyObject *__pyx_v_domain_right_edge, int __pyx_v_num_root, PyObject *__pyx_v_num_zones) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __pyx_t_5numpy_uint8_t *__pyx_t_8;
  struct __pyx_array_obj *__pyx_t_9 = NULL;
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __pyx_t_5numpy_uint64_t *__pyx_t_11;
  __Pyx_memviewslice __pyx_t_12 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "yt/geometry/particle_oct_container.pyx":1979
 *     def __init__(self, domain_dimensions, domain_left_edge, domain_right_edge,
 *                  int num_root, num_zones = 2):
 *         super(ParticleBitmapOctreeContainer, self).__init__(             # <<<<<<<<<<<<<<
 *             domain_dimensions, domain_left_edge, domain_right_edge,
 *             num_zones)
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer));
  __Pyx_GIVEREF(((PyObject *)__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1981
 *         super(ParticleBitmapOctreeContainer, self).__init__(
 *             domain_dimensions, domain_left_edge, domain_right_edge,
 *             num_zones)             # <<<<<<<<<<<<<<
 *         self.loaded = 0
 *         self.fill_style = "o"
 */
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_domain_dimensions, __pyx_v_domain_left_edge, __pyx_v_domain_right_edge, __pyx_v_num_zones};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1979, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[5] = {__pyx_t_3, __pyx_v_domain_dimensions, __pyx_v_domain_left_edge, __pyx_v_domain_right_edge, __pyx_v_num_zones};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 4+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1979, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(4+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1979, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_domain_dimensions);
    __Pyx_GIVEREF(__pyx_v_domain_dimensions);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_domain_dimensions);
    __Pyx_INCREF(__pyx_v_domain_left_edge);
    __Pyx_GIVEREF(__pyx_v_domain_left_edge);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_domain_left_edge);
    __Pyx_INCREF(__pyx_v_domain_right_edge);
    __Pyx_GIVEREF(__pyx_v_domain_right_edge);
    PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_4, __pyx_v_domain_right_edge);
    __Pyx_INCREF(__pyx_v_num_zones);
    __Pyx_GIVEREF(__pyx_v_num_zones);
    PyTuple_SET_ITEM(__pyx_t_5, 3+__pyx_t_4, __pyx_v_num_zones);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1979, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1982
 *             domain_dimensions, domain_left_edge, domain_right_edge,
 *             num_zones)
 *         self.loaded = 0             # <<<<<<<<<<<<<<
 *         self.fill_style = "o"
 *         self.partial_coverage = 2
 */
  __pyx_v_self->loaded = 0;

  /* "yt/geometry/particle_oct_container.pyx":1983
 *             num_zones)
 *         self.loaded = 0
 *         self.fill_style = "o"             # <<<<<<<<<<<<<<
 *         self.partial_coverage = 2
 *         self.overlap_cells = 0
 */
  __Pyx_INCREF(__pyx_n_s_o);
  __Pyx_GIVEREF(__pyx_n_s_o);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.__pyx_base.fill_style);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.__pyx_base.fill_style);
  __pyx_v_self->__pyx_base.__pyx_base.fill_style = __pyx_n_s_o;

  /* "yt/geometry/particle_oct_container.pyx":1984
 *         self.loaded = 0
 *         self.fill_style = "o"
 *         self.partial_coverage = 2             # <<<<<<<<<<<<<<
 *         self.overlap_cells = 0
 *         # Now the overrides
 */
  __pyx_v_self->__pyx_base.__pyx_base.partial_coverage = 2;

  /* "yt/geometry/particle_oct_container.pyx":1985
 *         self.fill_style = "o"
 *         self.partial_coverage = 2
 *         self.overlap_cells = 0             # <<<<<<<<<<<<<<
 *         # Now the overrides
 *         self.max_level = -1
 */
  __pyx_v_self->overlap_cells = 0;

  /* "yt/geometry/particle_oct_container.pyx":1987
 *         self.overlap_cells = 0
 *         # Now the overrides
 *         self.max_level = -1             # <<<<<<<<<<<<<<
 *         self.max_root = num_root
 *         self.root_nodes = <OctKey*> malloc(sizeof(OctKey) * num_root)
 */
  __pyx_v_self->max_level = -1;

  /* "yt/geometry/particle_oct_container.pyx":1988
 *         # Now the overrides
 *         self.max_level = -1
 *         self.max_root = num_root             # <<<<<<<<<<<<<<
 *         self.root_nodes = <OctKey*> malloc(sizeof(OctKey) * num_root)
 *         self._ptr_index_base_roots = <np.uint8_t*> malloc(sizeof(np.uint8_t) * num_root)
 */
  __pyx_v_self->__pyx_base.max_root = __pyx_v_num_root;

  /* "yt/geometry/particle_oct_container.pyx":1989
 *         self.max_level = -1
 *         self.max_root = num_root
 *         self.root_nodes = <OctKey*> malloc(sizeof(OctKey) * num_root)             # <<<<<<<<<<<<<<
 *         self._ptr_index_base_roots = <np.uint8_t*> malloc(sizeof(np.uint8_t) * num_root)
 *         self._ptr_octs_per_root = <np.uint64_t*> malloc(sizeof(np.uint64_t) * num_root)
 */
  __pyx_v_self->__pyx_base.root_nodes = ((struct __pyx_t_2yt_8geometry_13oct_container_OctKey *)malloc(((sizeof(struct __pyx_t_2yt_8geometry_13oct_container_OctKey)) * __pyx_v_num_root)));

  /* "yt/geometry/particle_oct_container.pyx":1990
 *         self.max_root = num_root
 *         self.root_nodes = <OctKey*> malloc(sizeof(OctKey) * num_root)
 *         self._ptr_index_base_roots = <np.uint8_t*> malloc(sizeof(np.uint8_t) * num_root)             # <<<<<<<<<<<<<<
 *         self._ptr_octs_per_root = <np.uint64_t*> malloc(sizeof(np.uint64_t) * num_root)
 *         for i in range(num_root):
 */
  __pyx_v_self->_ptr_index_base_roots = ((__pyx_t_5numpy_uint8_t *)malloc(((sizeof(__pyx_t_5numpy_uint8_t)) * __pyx_v_num_root)));

  /* "yt/geometry/particle_oct_container.pyx":1991
 *         self.root_nodes = <OctKey*> malloc(sizeof(OctKey) * num_root)
 *         self._ptr_index_base_roots = <np.uint8_t*> malloc(sizeof(np.uint8_t) * num_root)
 *         self._ptr_octs_per_root = <np.uint64_t*> malloc(sizeof(np.uint64_t) * num_root)             # <<<<<<<<<<<<<<
 *         for i in range(num_root):
 *             self.root_nodes[i].key = -1
 */
  __pyx_v_self->_ptr_octs_per_root = ((__pyx_t_5numpy_uint64_t *)malloc(((sizeof(__pyx_t_5numpy_uint64_t)) * __pyx_v_num_root)));

  /* "yt/geometry/particle_oct_container.pyx":1992
 *         self._ptr_index_base_roots = <np.uint8_t*> malloc(sizeof(np.uint8_t) * num_root)
 *         self._ptr_octs_per_root = <np.uint64_t*> malloc(sizeof(np.uint64_t) * num_root)
 *         for i in range(num_root):             # <<<<<<<<<<<<<<
 *             self.root_nodes[i].key = -1
 *             self.root_nodes[i].node = NULL
 */
  __pyx_t_4 = __pyx_v_num_root;
  __pyx_t_6 = __pyx_t_4;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "yt/geometry/particle_oct_container.pyx":1993
 *         self._ptr_octs_per_root = <np.uint64_t*> malloc(sizeof(np.uint64_t) * num_root)
 *         for i in range(num_root):
 *             self.root_nodes[i].key = -1             # <<<<<<<<<<<<<<
 *             self.root_nodes[i].node = NULL
 *             self._ptr_index_base_roots[i] = 1
 */
    (__pyx_v_self->__pyx_base.root_nodes[__pyx_v_i]).key = -1LL;

    /* "yt/geometry/particle_oct_container.pyx":1994
 *         for i in range(num_root):
 *             self.root_nodes[i].key = -1
 *             self.root_nodes[i].node = NULL             # <<<<<<<<<<<<<<
 *             self._ptr_index_base_roots[i] = 1
 *             self._ptr_octs_per_root[i] = 0
 */
    (__pyx_v_self->__pyx_base.root_nodes[__pyx_v_i]).node = NULL;

    /* "yt/geometry/particle_oct_container.pyx":1995
 *             self.root_nodes[i].key = -1
 *             self.root_nodes[i].node = NULL
 *             self._ptr_index_base_roots[i] = 1             # <<<<<<<<<<<<<<
 *             self._ptr_octs_per_root[i] = 0
 *         self._index_base_roots = <np.uint8_t[:num_root]> self._ptr_index_base_roots
 */
    (__pyx_v_self->_ptr_index_base_roots[__pyx_v_i]) = 1;

    /* "yt/geometry/particle_oct_container.pyx":1996
 *             self.root_nodes[i].node = NULL
 *             self._ptr_index_base_roots[i] = 1
 *             self._ptr_octs_per_root[i] = 0             # <<<<<<<<<<<<<<
 *         self._index_base_roots = <np.uint8_t[:num_root]> self._ptr_index_base_roots
 *         self._octs_per_root = <np.uint64_t[:num_root]> self._ptr_octs_per_root
 */
    (__pyx_v_self->_ptr_octs_per_root[__pyx_v_i]) = 0;
  }

  /* "yt/geometry/particle_oct_container.pyx":1997
 *             self._ptr_index_base_roots[i] = 1
 *             self._ptr_octs_per_root[i] = 0
 *         self._index_base_roots = <np.uint8_t[:num_root]> self._ptr_index_base_roots             # <<<<<<<<<<<<<<
 *         self._octs_per_root = <np.uint64_t[:num_root]> self._ptr_octs_per_root
 * 
 */
  __pyx_t_8 = __pyx_v_self->_ptr_index_base_roots;
  if (!__pyx_t_8) {
    PyErr_SetString(PyExc_ValueError,"Cannot create cython.array from NULL pointer");
    __PYX_ERR(0, 1997, __pyx_L1_error)
  }
  __pyx_t_2 = __pyx_format_from_typeinfo(&__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = Py_BuildValue((char*) "("  __PYX_BUILD_PY_SSIZE_T  ")", ((Py_ssize_t)__pyx_v_num_root)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __pyx_array_new(__pyx_t_1, sizeof(__pyx_t_5numpy_uint8_t), PyBytes_AS_STRING(__pyx_t_2), (char *) "c", (char *) __pyx_t_8);
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint8_t(((PyObject *)__pyx_t_9), PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 1997, __pyx_L1_error)
  __Pyx_DECREF(((PyObject *)__pyx_t_9)); __pyx_t_9 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->_index_base_roots, 0);
  __pyx_v_self->_index_base_roots = __pyx_t_10;
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":1998
 *             self._ptr_octs_per_root[i] = 0
 *         self._index_base_roots = <np.uint8_t[:num_root]> self._ptr_index_base_roots
 *         self._octs_per_root = <np.uint64_t[:num_root]> self._ptr_octs_per_root             # <<<<<<<<<<<<<<
 * 
 *     def allocate_domains(self, counts = None):
 */
  __pyx_t_11 = __pyx_v_self->_ptr_octs_per_root;
  if (!__pyx_t_11) {
    PyErr_SetString(PyExc_ValueError,"Cannot create cython.array from NULL pointer");
    __PYX_ERR(0, 1998, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_format_from_typeinfo(&__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = Py_BuildValue((char*) "("  __PYX_BUILD_PY_SSIZE_T  ")", ((Py_ssize_t)__pyx_v_num_root)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = __pyx_array_new(__pyx_t_2, sizeof(__pyx_t_5numpy_uint64_t), PyBytes_AS_STRING(__pyx_t_1), (char *) "c", (char *) __pyx_t_11);
  if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_12 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(((PyObject *)__pyx_t_9), PyBUF_WRITABLE); if (unlikely(!__pyx_t_12.memview)) __PYX_ERR(0, 1998, __pyx_L1_error)
  __Pyx_DECREF(((PyObject *)__pyx_t_9)); __pyx_t_9 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->_octs_per_root, 0);
  __pyx_v_self->_octs_per_root = __pyx_t_12;
  __pyx_t_12.memview = NULL;
  __pyx_t_12.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":1977
 *     cdef np.uint64_t[:] _octs_per_root
 *     cdef public int overlap_cells
 *     def __init__(self, domain_dimensions, domain_left_edge, domain_right_edge,             # <<<<<<<<<<<<<<
 *                  int num_root, num_zones = 2):
 *         super(ParticleBitmapOctreeContainer, self).__init__(
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(((PyObject *)__pyx_t_9));
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_12, 1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":2000
 *         self._octs_per_root = <np.uint64_t[:num_root]> self._ptr_octs_per_root
 * 
 *     def allocate_domains(self, counts = None):             # <<<<<<<<<<<<<<
 *         if counts is None:
 *             counts = [self.max_root]
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_3allocate_domains(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_3allocate_domains(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_counts = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("allocate_domains (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_counts,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_counts);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "allocate_domains") < 0)) __PYX_ERR(0, 2000, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_counts = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("allocate_domains", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2000, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer.allocate_domains", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_2allocate_domains(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self), __pyx_v_counts);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_2allocate_domains(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, PyObject *__pyx_v_counts) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("allocate_domains", 0);
  __Pyx_INCREF(__pyx_v_counts);

  /* "yt/geometry/particle_oct_container.pyx":2001
 * 
 *     def allocate_domains(self, counts = None):
 *         if counts is None:             # <<<<<<<<<<<<<<
 *             counts = [self.max_root]
 *         OctreeContainer.allocate_domains(self, counts)
 */
  __pyx_t_1 = (__pyx_v_counts == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "yt/geometry/particle_oct_container.pyx":2002
 *     def allocate_domains(self, counts = None):
 *         if counts is None:
 *             counts = [self.max_root]             # <<<<<<<<<<<<<<
 *         OctreeContainer.allocate_domains(self, counts)
 * 
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->__pyx_base.max_root); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2002, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2002, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyList_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_counts, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "yt/geometry/particle_oct_container.pyx":2001
 * 
 *     def allocate_domains(self, counts = None):
 *         if counts is None:             # <<<<<<<<<<<<<<
 *             counts = [self.max_root]
 *         OctreeContainer.allocate_domains(self, counts)
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":2003
 *         if counts is None:
 *             counts = [self.max_root]
 *         OctreeContainer.allocate_domains(self, counts)             # <<<<<<<<<<<<<<
 * 
 *     def finalize(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer), __pyx_n_s_allocate_domains); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_self), __pyx_v_counts};
    __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2003, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_self), __pyx_v_counts};
    __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2003, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2003, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_self));
    __Pyx_INCREF(__pyx_v_counts);
    __Pyx_GIVEREF(__pyx_v_counts);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_counts);
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2003, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "yt/geometry/particle_oct_container.pyx":2000
 *         self._octs_per_root = <np.uint64_t[:num_root]> self._ptr_octs_per_root
 * 
 *     def allocate_domains(self, counts = None):             # <<<<<<<<<<<<<<
 *         if counts is None:
 *             counts = [self.max_root]
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer.allocate_domains", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_counts);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":2005
 *         OctreeContainer.allocate_domains(self, counts)
 * 
 *     def finalize(self):             # <<<<<<<<<<<<<<
 *         # Assign domain ind
 *         cdef SelectorObject selector = AlwaysSelector(None)
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_5finalize(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_5finalize(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("finalize (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_4finalize(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_4finalize(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_selector = 0;
  struct __pyx_obj_2yt_8geometry_12oct_visitors_AssignDomainInd *__pyx_v_visitor = 0;
  __pyx_t_5numpy_int64_t __pyx_v_nprev_octs;
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __pyx_t_5numpy_uint8_t *__pyx_t_3;
  struct __pyx_array_obj *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("finalize", 0);

  /* "yt/geometry/particle_oct_container.pyx":2007
 *     def finalize(self):
 *         # Assign domain ind
 *         cdef SelectorObject selector = AlwaysSelector(None)             # <<<<<<<<<<<<<<
 *         selector.overlap_cells = self.overlap_cells
 *         cdef oct_visitors.AssignDomainInd visitor
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_AlwaysSelector), __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_selector = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":2008
 *         # Assign domain ind
 *         cdef SelectorObject selector = AlwaysSelector(None)
 *         selector.overlap_cells = self.overlap_cells             # <<<<<<<<<<<<<<
 *         cdef oct_visitors.AssignDomainInd visitor
 *         visitor = oct_visitors.AssignDomainInd(self)
 */
  __pyx_t_2 = __pyx_v_self->overlap_cells;
  __pyx_v_selector->overlap_cells = __pyx_t_2;

  /* "yt/geometry/particle_oct_container.pyx":2010
 *         selector.overlap_cells = self.overlap_cells
 *         cdef oct_visitors.AssignDomainInd visitor
 *         visitor = oct_visitors.AssignDomainInd(self)             # <<<<<<<<<<<<<<
 *         self.visit_all_octs(selector, visitor)
 *         assert ((visitor.global_index+1)*visitor.nz == visitor.index)
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_2yt_8geometry_12oct_visitors_AssignDomainInd), ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_visitor = ((struct __pyx_obj_2yt_8geometry_12oct_visitors_AssignDomainInd *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":2011
 *         cdef oct_visitors.AssignDomainInd visitor
 *         visitor = oct_visitors.AssignDomainInd(self)
 *         self.visit_all_octs(selector, visitor)             # <<<<<<<<<<<<<<
 *         assert ((visitor.global_index+1)*visitor.nz == visitor.index)
 *         # Copy indexes
 */
  ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.visit_all_octs(((struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *)__pyx_v_self), __pyx_v_selector, ((struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *)__pyx_v_visitor), NULL);

  /* "yt/geometry/particle_oct_container.pyx":2012
 *         visitor = oct_visitors.AssignDomainInd(self)
 *         self.visit_all_octs(selector, visitor)
 *         assert ((visitor.global_index+1)*visitor.nz == visitor.index)             # <<<<<<<<<<<<<<
 *         # Copy indexes
 *         self._ptr_index_base_octs = <np.uint8_t*> malloc(sizeof(np.uint8_t)*self.nocts)
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!((((__pyx_v_visitor->__pyx_base.global_index + 1) * __pyx_v_visitor->__pyx_base.nz) == __pyx_v_visitor->__pyx_base.index) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 2012, __pyx_L1_error)
    }
  }
  #endif

  /* "yt/geometry/particle_oct_container.pyx":2014
 *         assert ((visitor.global_index+1)*visitor.nz == visitor.index)
 *         # Copy indexes
 *         self._ptr_index_base_octs = <np.uint8_t*> malloc(sizeof(np.uint8_t)*self.nocts)             # <<<<<<<<<<<<<<
 *         self._index_base_octs = <np.uint8_t[:self.nocts]> self._ptr_index_base_octs
 *         cdef np.int64_t nprev_octs = 0
 */
  __pyx_v_self->_ptr_index_base_octs = ((__pyx_t_5numpy_uint8_t *)malloc(((sizeof(__pyx_t_5numpy_uint8_t)) * __pyx_v_self->__pyx_base.__pyx_base.nocts)));

  /* "yt/geometry/particle_oct_container.pyx":2015
 *         # Copy indexes
 *         self._ptr_index_base_octs = <np.uint8_t*> malloc(sizeof(np.uint8_t)*self.nocts)
 *         self._index_base_octs = <np.uint8_t[:self.nocts]> self._ptr_index_base_octs             # <<<<<<<<<<<<<<
 *         cdef np.int64_t nprev_octs = 0
 *         cdef int i
 */
  __pyx_t_3 = __pyx_v_self->_ptr_index_base_octs;
  if (!__pyx_t_3) {
    PyErr_SetString(PyExc_ValueError,"Cannot create cython.array from NULL pointer");
    __PYX_ERR(0, 2015, __pyx_L1_error)
  }
  __pyx_t_5 = __pyx_format_from_typeinfo(&__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = Py_BuildValue((char*) "("  __PYX_BUILD_PY_SSIZE_T  ")", ((Py_ssize_t)__pyx_v_self->__pyx_base.__pyx_base.nocts)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __pyx_array_new(__pyx_t_1, sizeof(__pyx_t_5numpy_uint8_t), PyBytes_AS_STRING(__pyx_t_5), (char *) "c", (char *) __pyx_t_3);
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint8_t(((PyObject *)__pyx_t_4), PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 2015, __pyx_L1_error)
  __Pyx_DECREF(((PyObject *)__pyx_t_4)); __pyx_t_4 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->_index_base_octs, 0);
  __pyx_v_self->_index_base_octs = __pyx_t_6;
  __pyx_t_6.memview = NULL;
  __pyx_t_6.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":2016
 *         self._ptr_index_base_octs = <np.uint8_t*> malloc(sizeof(np.uint8_t)*self.nocts)
 *         self._index_base_octs = <np.uint8_t[:self.nocts]> self._ptr_index_base_octs
 *         cdef np.int64_t nprev_octs = 0             # <<<<<<<<<<<<<<
 *         cdef int i
 *         for i in range(self.num_root):
 */
  __pyx_v_nprev_octs = 0;

  /* "yt/geometry/particle_oct_container.pyx":2018
 *         cdef np.int64_t nprev_octs = 0
 *         cdef int i
 *         for i in range(self.num_root):             # <<<<<<<<<<<<<<
 *             self._index_base_octs[nprev_octs:(nprev_octs+self._octs_per_root[i])] = self._index_base_roots[i]
 *             nprev_octs += self._octs_per_root[i]
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.num_root;
  __pyx_t_7 = __pyx_t_2;
  for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
    __pyx_v_i = __pyx_t_8;

    /* "yt/geometry/particle_oct_container.pyx":2019
 *         cdef int i
 *         for i in range(self.num_root):
 *             self._index_base_octs[nprev_octs:(nprev_octs+self._octs_per_root[i])] = self._index_base_roots[i]             # <<<<<<<<<<<<<<
 *             nprev_octs += self._octs_per_root[i]
 * 
 */
    if (unlikely(!__pyx_v_self->_index_base_roots.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 2019, __pyx_L1_error)}
    __pyx_t_9 = __pyx_v_i;
    __pyx_t_10 = -1;
    if (__pyx_t_9 < 0) {
      __pyx_t_9 += __pyx_v_self->_index_base_roots.shape[0];
      if (unlikely(__pyx_t_9 < 0)) __pyx_t_10 = 0;
    } else if (unlikely(__pyx_t_9 >= __pyx_v_self->_index_base_roots.shape[0])) __pyx_t_10 = 0;
    if (unlikely(__pyx_t_10 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_10);
      __PYX_ERR(0, 2019, __pyx_L1_error)
    }
    if (unlikely(!__pyx_v_self->_index_base_octs.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 2019, __pyx_L1_error)}
    if (unlikely(!__pyx_v_self->_octs_per_root.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 2019, __pyx_L1_error)}
    __pyx_t_11 = __pyx_v_i;
    __pyx_t_10 = -1;
    if (__pyx_t_11 < 0) {
      __pyx_t_11 += __pyx_v_self->_octs_per_root.shape[0];
      if (unlikely(__pyx_t_11 < 0)) __pyx_t_10 = 0;
    } else if (unlikely(__pyx_t_11 >= __pyx_v_self->_octs_per_root.shape[0])) __pyx_t_10 = 0;
    if (unlikely(__pyx_t_10 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_10);
      __PYX_ERR(0, 2019, __pyx_L1_error)
    }
    __pyx_t_6.data = __pyx_v_self->_index_base_octs.data;
    __pyx_t_6.memview = __pyx_v_self->_index_base_octs.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_6, 0);
    __pyx_t_10 = -1;
    if (unlikely(__pyx_memoryview_slice_memviewslice(
    &__pyx_t_6,
    __pyx_v_self->_index_base_octs.shape[0], __pyx_v_self->_index_base_octs.strides[0], __pyx_v_self->_index_base_octs.suboffsets[0],
    0,
    0,
    &__pyx_t_10,
    __pyx_v_nprev_octs,
    (__pyx_v_nprev_octs + (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->_octs_per_root.data + __pyx_t_11 * __pyx_v_self->_octs_per_root.strides[0]) )))),
    0,
    1,
    1,
    0,
    1) < 0))
{
    __PYX_ERR(0, 2019, __pyx_L1_error)
}

{
        __pyx_t_5numpy_uint8_t __pyx_temp_scalar = (*((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->_index_base_roots.data + __pyx_t_9 * __pyx_v_self->_index_base_roots.strides[0]) )));
        {
            Py_ssize_t __pyx_temp_extent_0 = __pyx_t_6.shape[0];
            Py_ssize_t __pyx_temp_stride_0 = __pyx_t_6.strides[0];
            char *__pyx_temp_pointer_0;
            Py_ssize_t __pyx_temp_idx_0;
            __pyx_temp_pointer_0 = __pyx_t_6.data;
            for (__pyx_temp_idx_0 = 0; __pyx_temp_idx_0 < __pyx_temp_extent_0; __pyx_temp_idx_0++) {
              *((__pyx_t_5numpy_uint8_t *) __pyx_temp_pointer_0) = __pyx_temp_scalar;
              __pyx_temp_pointer_0 += __pyx_temp_stride_0;
            }
        }
    }
    __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
    __pyx_t_6.memview = NULL;
    __pyx_t_6.data = NULL;

    /* "yt/geometry/particle_oct_container.pyx":2020
 *         for i in range(self.num_root):
 *             self._index_base_octs[nprev_octs:(nprev_octs+self._octs_per_root[i])] = self._index_base_roots[i]
 *             nprev_octs += self._octs_per_root[i]             # <<<<<<<<<<<<<<
 * 
 *     cdef visit_assign(self, Oct *o, np.int64_t *lpos, int level, int *max_level,
 */
    if (unlikely(!__pyx_v_self->_octs_per_root.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 2020, __pyx_L1_error)}
    __pyx_t_9 = __pyx_v_i;
    __pyx_t_10 = -1;
    if (__pyx_t_9 < 0) {
      __pyx_t_9 += __pyx_v_self->_octs_per_root.shape[0];
      if (unlikely(__pyx_t_9 < 0)) __pyx_t_10 = 0;
    } else if (unlikely(__pyx_t_9 >= __pyx_v_self->_octs_per_root.shape[0])) __pyx_t_10 = 0;
    if (unlikely(__pyx_t_10 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_10);
      __PYX_ERR(0, 2020, __pyx_L1_error)
    }
    __pyx_v_nprev_octs = (__pyx_v_nprev_octs + (*((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->_octs_per_root.data + __pyx_t_9 * __pyx_v_self->_octs_per_root.strides[0]) ))));
  }

  /* "yt/geometry/particle_oct_container.pyx":2005
 *         OctreeContainer.allocate_domains(self, counts)
 * 
 *     def finalize(self):             # <<<<<<<<<<<<<<
 *         # Assign domain ind
 *         cdef SelectorObject selector = AlwaysSelector(None)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(((PyObject *)__pyx_t_4));
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer.finalize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_selector);
  __Pyx_XDECREF((PyObject *)__pyx_v_visitor);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":2022
 *             nprev_octs += self._octs_per_root[i]
 * 
 *     cdef visit_assign(self, Oct *o, np.int64_t *lpos, int level, int *max_level,             # <<<<<<<<<<<<<<
 *                       np.int64_t index_root):
 *         cdef int i, j, k
 */

static PyObject *__pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_visit_assign(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o, __pyx_t_5numpy_int64_t *__pyx_v_lpos, int __pyx_v_level, int *__pyx_v_max_level, __pyx_t_5numpy_int64_t __pyx_v_index_root) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __pyx_t_5numpy_int64_t __pyx_t_2;
  int __pyx_t_3;
  __pyx_t_5numpy_int64_t __pyx_t_4;
  long __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("visit_assign", 0);

  /* "yt/geometry/particle_oct_container.pyx":2025
 *                       np.int64_t index_root):
 *         cdef int i, j, k
 *         if o.children == NULL:             # <<<<<<<<<<<<<<
 *             self.oct_list[lpos[0]] = o
 *             self._index_base_octs[lpos[0]] = self._index_base_roots[index_root]
 */
  __pyx_t_1 = ((__pyx_v_o->children == NULL) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/particle_oct_container.pyx":2026
 *         cdef int i, j, k
 *         if o.children == NULL:
 *             self.oct_list[lpos[0]] = o             # <<<<<<<<<<<<<<
 *             self._index_base_octs[lpos[0]] = self._index_base_roots[index_root]
 *             lpos[0] += 1
 */
    (__pyx_v_self->oct_list[(__pyx_v_lpos[0])]) = __pyx_v_o;

    /* "yt/geometry/particle_oct_container.pyx":2027
 *         if o.children == NULL:
 *             self.oct_list[lpos[0]] = o
 *             self._index_base_octs[lpos[0]] = self._index_base_roots[index_root]             # <<<<<<<<<<<<<<
 *             lpos[0] += 1
 *         max_level[0] = imax(max_level[0], level)
 */
    if (unlikely(!__pyx_v_self->_index_base_roots.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 2027, __pyx_L1_error)}
    __pyx_t_2 = __pyx_v_index_root;
    __pyx_t_3 = -1;
    if (__pyx_t_2 < 0) {
      __pyx_t_2 += __pyx_v_self->_index_base_roots.shape[0];
      if (unlikely(__pyx_t_2 < 0)) __pyx_t_3 = 0;
    } else if (unlikely(__pyx_t_2 >= __pyx_v_self->_index_base_roots.shape[0])) __pyx_t_3 = 0;
    if (unlikely(__pyx_t_3 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_3);
      __PYX_ERR(0, 2027, __pyx_L1_error)
    }
    if (unlikely(!__pyx_v_self->_index_base_octs.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 2027, __pyx_L1_error)}
    __pyx_t_4 = (__pyx_v_lpos[0]);
    __pyx_t_3 = -1;
    if (__pyx_t_4 < 0) {
      __pyx_t_4 += __pyx_v_self->_index_base_octs.shape[0];
      if (unlikely(__pyx_t_4 < 0)) __pyx_t_3 = 0;
    } else if (unlikely(__pyx_t_4 >= __pyx_v_self->_index_base_octs.shape[0])) __pyx_t_3 = 0;
    if (unlikely(__pyx_t_3 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_3);
      __PYX_ERR(0, 2027, __pyx_L1_error)
    }
    *((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->_index_base_octs.data + __pyx_t_4 * __pyx_v_self->_index_base_octs.strides[0]) )) = (*((__pyx_t_5numpy_uint8_t *) ( /* dim=0 */ (__pyx_v_self->_index_base_roots.data + __pyx_t_2 * __pyx_v_self->_index_base_roots.strides[0]) )));

    /* "yt/geometry/particle_oct_container.pyx":2028
 *             self.oct_list[lpos[0]] = o
 *             self._index_base_octs[lpos[0]] = self._index_base_roots[index_root]
 *             lpos[0] += 1             # <<<<<<<<<<<<<<
 *         max_level[0] = imax(max_level[0], level)
 *         for i in range(2):
 */
    __pyx_t_5 = 0;
    (__pyx_v_lpos[__pyx_t_5]) = ((__pyx_v_lpos[__pyx_t_5]) + 1);

    /* "yt/geometry/particle_oct_container.pyx":2025
 *                       np.int64_t index_root):
 *         cdef int i, j, k
 *         if o.children == NULL:             # <<<<<<<<<<<<<<
 *             self.oct_list[lpos[0]] = o
 *             self._index_base_octs[lpos[0]] = self._index_base_roots[index_root]
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":2029
 *             self._index_base_octs[lpos[0]] = self._index_base_roots[index_root]
 *             lpos[0] += 1
 *         max_level[0] = imax(max_level[0], level)             # <<<<<<<<<<<<<<
 *         for i in range(2):
 *             for j in range(2):
 */
  (__pyx_v_max_level[0]) = __pyx_f_2yt_9utilities_3lib_8fp_utils_imax((__pyx_v_max_level[0]), __pyx_v_level);

  /* "yt/geometry/particle_oct_container.pyx":2030
 *             lpos[0] += 1
 *         max_level[0] = imax(max_level[0], level)
 *         for i in range(2):             # <<<<<<<<<<<<<<
 *             for j in range(2):
 *                 for k in range(2):
 */
  for (__pyx_t_3 = 0; __pyx_t_3 < 2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":2031
 *         max_level[0] = imax(max_level[0], level)
 *         for i in range(2):
 *             for j in range(2):             # <<<<<<<<<<<<<<
 *                 for k in range(2):
 *                     if o.children != NULL \
 */
    for (__pyx_t_6 = 0; __pyx_t_6 < 2; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "yt/geometry/particle_oct_container.pyx":2032
 *         for i in range(2):
 *             for j in range(2):
 *                 for k in range(2):             # <<<<<<<<<<<<<<
 *                     if o.children != NULL \
 *                        and o.children[cind(i,j,k)] != NULL:
 */
      for (__pyx_t_7 = 0; __pyx_t_7 < 2; __pyx_t_7+=1) {
        __pyx_v_k = __pyx_t_7;

        /* "yt/geometry/particle_oct_container.pyx":2033
 *             for j in range(2):
 *                 for k in range(2):
 *                     if o.children != NULL \             # <<<<<<<<<<<<<<
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_assign(o.children[cind(i,j,k)], lpos,
 */
        __pyx_t_8 = ((__pyx_v_o->children != NULL) != 0);
        if (__pyx_t_8) {
        } else {
          __pyx_t_1 = __pyx_t_8;
          goto __pyx_L11_bool_binop_done;
        }

        /* "yt/geometry/particle_oct_container.pyx":2034
 *                 for k in range(2):
 *                     if o.children != NULL \
 *                        and o.children[cind(i,j,k)] != NULL:             # <<<<<<<<<<<<<<
 *                         self.visit_assign(o.children[cind(i,j,k)], lpos,
 *                                           level + 1, max_level, index_root)
 */
        __pyx_t_8 = (((__pyx_v_o->children[__pyx_f_2yt_8geometry_12oct_visitors_cind(__pyx_v_i, __pyx_v_j, __pyx_v_k)]) != NULL) != 0);
        __pyx_t_1 = __pyx_t_8;
        __pyx_L11_bool_binop_done:;

        /* "yt/geometry/particle_oct_container.pyx":2033
 *             for j in range(2):
 *                 for k in range(2):
 *                     if o.children != NULL \             # <<<<<<<<<<<<<<
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_assign(o.children[cind(i,j,k)], lpos,
 */
        if (__pyx_t_1) {

          /* "yt/geometry/particle_oct_container.pyx":2035
 *                     if o.children != NULL \
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_assign(o.children[cind(i,j,k)], lpos,             # <<<<<<<<<<<<<<
 *                                           level + 1, max_level, index_root)
 *         return
 */
          __pyx_t_9 = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->visit_assign(__pyx_v_self, (__pyx_v_o->children[__pyx_f_2yt_8geometry_12oct_visitors_cind(__pyx_v_i, __pyx_v_j, __pyx_v_k)]), __pyx_v_lpos, (__pyx_v_level + 1), __pyx_v_max_level, __pyx_v_index_root); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2035, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

          /* "yt/geometry/particle_oct_container.pyx":2033
 *             for j in range(2):
 *                 for k in range(2):
 *                     if o.children != NULL \             # <<<<<<<<<<<<<<
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_assign(o.children[cind(i,j,k)], lpos,
 */
        }
      }
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":2037
 *                         self.visit_assign(o.children[cind(i,j,k)], lpos,
 *                                           level + 1, max_level, index_root)
 *         return             # <<<<<<<<<<<<<<
 * 
 *     cdef Oct* allocate_oct(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":2022
 *             nprev_octs += self._octs_per_root[i]
 * 
 *     cdef visit_assign(self, Oct *o, np.int64_t *lpos, int level, int *max_level,             # <<<<<<<<<<<<<<
 *                       np.int64_t index_root):
 *         cdef int i, j, k
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer.visit_assign", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":2039
 *         return
 * 
 *     cdef Oct* allocate_oct(self):             # <<<<<<<<<<<<<<
 *         #Allocate the memory, set to NULL or -1
 *         #We reserve space for n_ref particles, but keep
 */

static struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_allocate_oct(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self) {
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_my_oct;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("allocate_oct", 0);

  /* "yt/geometry/particle_oct_container.pyx":2043
 *         #We reserve space for n_ref particles, but keep
 *         #track of how many are used with np initially 0
 *         self.nocts += 1             # <<<<<<<<<<<<<<
 *         cdef Oct *my_oct = <Oct*> malloc(sizeof(Oct))
 *         my_oct.domain = -1
 */
  __pyx_v_self->__pyx_base.__pyx_base.nocts = (__pyx_v_self->__pyx_base.__pyx_base.nocts + 1);

  /* "yt/geometry/particle_oct_container.pyx":2044
 *         #track of how many are used with np initially 0
 *         self.nocts += 1
 *         cdef Oct *my_oct = <Oct*> malloc(sizeof(Oct))             # <<<<<<<<<<<<<<
 *         my_oct.domain = -1
 *         my_oct.file_ind = 0
 */
  __pyx_v_my_oct = ((struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *)malloc((sizeof(struct __pyx_t_2yt_8geometry_12oct_visitors_Oct))));

  /* "yt/geometry/particle_oct_container.pyx":2045
 *         self.nocts += 1
 *         cdef Oct *my_oct = <Oct*> malloc(sizeof(Oct))
 *         my_oct.domain = -1             # <<<<<<<<<<<<<<
 *         my_oct.file_ind = 0
 *         my_oct.domain_ind = self.nocts - 1
 */
  __pyx_v_my_oct->domain = -1LL;

  /* "yt/geometry/particle_oct_container.pyx":2046
 *         cdef Oct *my_oct = <Oct*> malloc(sizeof(Oct))
 *         my_oct.domain = -1
 *         my_oct.file_ind = 0             # <<<<<<<<<<<<<<
 *         my_oct.domain_ind = self.nocts - 1
 *         my_oct.children = NULL
 */
  __pyx_v_my_oct->file_ind = 0;

  /* "yt/geometry/particle_oct_container.pyx":2047
 *         my_oct.domain = -1
 *         my_oct.file_ind = 0
 *         my_oct.domain_ind = self.nocts - 1             # <<<<<<<<<<<<<<
 *         my_oct.children = NULL
 *         return my_oct
 */
  __pyx_v_my_oct->domain_ind = (__pyx_v_self->__pyx_base.__pyx_base.nocts - 1);

  /* "yt/geometry/particle_oct_container.pyx":2048
 *         my_oct.file_ind = 0
 *         my_oct.domain_ind = self.nocts - 1
 *         my_oct.children = NULL             # <<<<<<<<<<<<<<
 *         return my_oct
 * 
 */
  __pyx_v_my_oct->children = NULL;

  /* "yt/geometry/particle_oct_container.pyx":2049
 *         my_oct.domain_ind = self.nocts - 1
 *         my_oct.children = NULL
 *         return my_oct             # <<<<<<<<<<<<<<
 * 
 *     def get_index_base_octs(self, np.int64_t[:] domain_ind):
 */
  __pyx_r = __pyx_v_my_oct;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":2039
 *         return
 * 
 *     cdef Oct* allocate_oct(self):             # <<<<<<<<<<<<<<
 *         #Allocate the memory, set to NULL or -1
 *         #We reserve space for n_ref particles, but keep
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":2051
 *         return my_oct
 * 
 *     def get_index_base_octs(self, np.int64_t[:] domain_ind):             # <<<<<<<<<<<<<<
 *         cdef np.int64_t ndst = np.max(domain_ind) + 1
 *         ind = np.zeros(ndst, 'int64') - 1
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_7get_index_base_octs(PyObject *__pyx_v_self, PyObject *__pyx_arg_domain_ind); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_7get_index_base_octs(PyObject *__pyx_v_self, PyObject *__pyx_arg_domain_ind) {
  __Pyx_memviewslice __pyx_v_domain_ind = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_index_base_octs (wrapper)", 0);
  assert(__pyx_arg_domain_ind); {
    __pyx_v_domain_ind = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_arg_domain_ind, PyBUF_WRITABLE); if (unlikely(!__pyx_v_domain_ind.memview)) __PYX_ERR(0, 2051, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer.get_index_base_octs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_6get_index_base_octs(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self), __pyx_v_domain_ind);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_6get_index_base_octs(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, __Pyx_memviewslice __pyx_v_domain_ind) {
  __pyx_t_5numpy_int64_t __pyx_v_ndst;
  PyObject *__pyx_v_ind = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_memviewslice __pyx_t_8 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_index_base_octs", 0);

  /* "yt/geometry/particle_oct_container.pyx":2052
 * 
 *     def get_index_base_octs(self, np.int64_t[:] domain_ind):
 *         cdef np.int64_t ndst = np.max(domain_ind) + 1             # <<<<<<<<<<<<<<
 *         ind = np.zeros(ndst, 'int64') - 1
 *         self._get_index_base_octs(ind, domain_ind)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_max); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_domain_ind, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_int64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_int64_t, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_npy_int64(__pyx_t_3); if (unlikely((__pyx_t_5 == ((npy_int64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2052, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_ndst = __pyx_t_5;

  /* "yt/geometry/particle_oct_container.pyx":2053
 *     def get_index_base_octs(self, np.int64_t[:] domain_ind):
 *         cdef np.int64_t ndst = np.max(domain_ind) + 1
 *         ind = np.zeros(ndst, 'int64') - 1             # <<<<<<<<<<<<<<
 *         self._get_index_base_octs(ind, domain_ind)
 *         return ind[ind >= 0]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_zeros); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_npy_int64(__pyx_v_ndst); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_1, __pyx_n_s_int64};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2053, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_1, __pyx_n_s_int64};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2053, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2053, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_int64);
    __Pyx_GIVEREF(__pyx_n_s_int64);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_n_s_int64);
    __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2053, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2053, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_ind = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "yt/geometry/particle_oct_container.pyx":2054
 *         cdef np.int64_t ndst = np.max(domain_ind) + 1
 *         ind = np.zeros(ndst, 'int64') - 1
 *         self._get_index_base_octs(ind, domain_ind)             # <<<<<<<<<<<<<<
 *         return ind[ind >= 0]
 * 
 */
  __pyx_t_8 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(__pyx_v_ind, PyBUF_WRITABLE); if (unlikely(!__pyx_t_8.memview)) __PYX_ERR(0, 2054, __pyx_L1_error)
  ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->_get_index_base_octs(__pyx_v_self, __pyx_t_8, __pyx_v_domain_ind);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __pyx_t_8.memview = NULL;
  __pyx_t_8.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":2055
 *         ind = np.zeros(ndst, 'int64') - 1
 *         self._get_index_base_octs(ind, domain_ind)
 *         return ind[ind >= 0]             # <<<<<<<<<<<<<<
 * 
 *     cdef void _get_index_base_octs(self, np.int64_t[:] ind, np.int64_t[:] domain_ind):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_ind, __pyx_int_0, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2055, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_ind, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":2051
 *         return my_oct
 * 
 *     def get_index_base_octs(self, np.int64_t[:] domain_ind):             # <<<<<<<<<<<<<<
 *         cdef np.int64_t ndst = np.max(domain_ind) + 1
 *         ind = np.zeros(ndst, 'int64') - 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __PYX_XDEC_MEMVIEW(&__pyx_t_8, 1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer.get_index_base_octs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_domain_ind, 1);
  __Pyx_XDECREF(__pyx_v_ind);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":2057
 *         return ind[ind >= 0]
 * 
 *     cdef void _get_index_base_octs(self, np.int64_t[:] ind, np.int64_t[:] domain_ind):             # <<<<<<<<<<<<<<
 *         cdef SelectorObject selector = AlwaysSelector(None)
 *         selector.overlap_cells = self.overlap_cells
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer__get_index_base_octs(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, __Pyx_memviewslice __pyx_v_ind, __Pyx_memviewslice __pyx_v_domain_ind) {
  struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *__pyx_v_selector = 0;
  struct __pyx_obj_2yt_8geometry_12oct_visitors_IndexMaskMapOcts *__pyx_v_visitor = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_index_base_octs", 0);

  /* "yt/geometry/particle_oct_container.pyx":2058
 * 
 *     cdef void _get_index_base_octs(self, np.int64_t[:] ind, np.int64_t[:] domain_ind):
 *         cdef SelectorObject selector = AlwaysSelector(None)             # <<<<<<<<<<<<<<
 *         selector.overlap_cells = self.overlap_cells
 *         cdef oct_visitors.IndexMaskMapOcts visitor
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_2yt_8geometry_18selection_routines_AlwaysSelector), __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_selector = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":2059
 *     cdef void _get_index_base_octs(self, np.int64_t[:] ind, np.int64_t[:] domain_ind):
 *         cdef SelectorObject selector = AlwaysSelector(None)
 *         selector.overlap_cells = self.overlap_cells             # <<<<<<<<<<<<<<
 *         cdef oct_visitors.IndexMaskMapOcts visitor
 *         visitor = oct_visitors.IndexMaskMapOcts(self)
 */
  __pyx_t_2 = __pyx_v_self->overlap_cells;
  __pyx_v_selector->overlap_cells = __pyx_t_2;

  /* "yt/geometry/particle_oct_container.pyx":2061
 *         selector.overlap_cells = self.overlap_cells
 *         cdef oct_visitors.IndexMaskMapOcts visitor
 *         visitor = oct_visitors.IndexMaskMapOcts(self)             # <<<<<<<<<<<<<<
 *         visitor.oct_mask = self._index_base_octs
 *         visitor.oct_index = ind
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_2yt_8geometry_12oct_visitors_IndexMaskMapOcts), ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2061, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_visitor = ((struct __pyx_obj_2yt_8geometry_12oct_visitors_IndexMaskMapOcts *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":2062
 *         cdef oct_visitors.IndexMaskMapOcts visitor
 *         visitor = oct_visitors.IndexMaskMapOcts(self)
 *         visitor.oct_mask = self._index_base_octs             # <<<<<<<<<<<<<<
 *         visitor.oct_index = ind
 *         visitor.map_domain_ind = domain_ind
 */
  if (unlikely(!__pyx_v_self->_index_base_octs.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 2062, __pyx_L1_error)}
  __pyx_t_3 = __pyx_v_self->_index_base_octs;
  __PYX_INC_MEMVIEW(&__pyx_t_3, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_visitor->oct_mask, 0);
  __pyx_v_visitor->oct_mask = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;

  /* "yt/geometry/particle_oct_container.pyx":2063
 *         visitor = oct_visitors.IndexMaskMapOcts(self)
 *         visitor.oct_mask = self._index_base_octs
 *         visitor.oct_index = ind             # <<<<<<<<<<<<<<
 *         visitor.map_domain_ind = domain_ind
 *         self.visit_all_octs(selector, visitor)
 */
  __PYX_XDEC_MEMVIEW(&__pyx_v_visitor->oct_index, 0);
  __PYX_INC_MEMVIEW(&__pyx_v_ind, 0);
  __pyx_v_visitor->oct_index = __pyx_v_ind;

  /* "yt/geometry/particle_oct_container.pyx":2064
 *         visitor.oct_mask = self._index_base_octs
 *         visitor.oct_index = ind
 *         visitor.map_domain_ind = domain_ind             # <<<<<<<<<<<<<<
 *         self.visit_all_octs(selector, visitor)
 * 
 */
  __PYX_XDEC_MEMVIEW(&__pyx_v_visitor->map_domain_ind, 0);
  __PYX_INC_MEMVIEW(&__pyx_v_domain_ind, 0);
  __pyx_v_visitor->map_domain_ind = __pyx_v_domain_ind;

  /* "yt/geometry/particle_oct_container.pyx":2065
 *         visitor.oct_index = ind
 *         visitor.map_domain_ind = domain_ind
 *         self.visit_all_octs(selector, visitor)             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.visit_all_octs(((struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *)__pyx_v_self), __pyx_v_selector, ((struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *)__pyx_v_visitor), NULL);

  /* "yt/geometry/particle_oct_container.pyx":2057
 *         return ind[ind >= 0]
 * 
 *     cdef void _get_index_base_octs(self, np.int64_t[:] ind, np.int64_t[:] domain_ind):             # <<<<<<<<<<<<<<
 *         cdef SelectorObject selector = AlwaysSelector(None)
 *         selector.overlap_cells = self.overlap_cells
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_WriteUnraisable("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer._get_index_base_octs", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_selector);
  __Pyx_XDECREF((PyObject *)__pyx_v_visitor);
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":2067
 *         self.visit_all_octs(selector, visitor)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         #Call the freemem ops on every ocy
 *         #of the root mesh recursively
 */

/* Python wrapper */
static void __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_9__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_9__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_8__dealloc__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_8__dealloc__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self) {
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "yt/geometry/particle_oct_container.pyx":2071
 *         #of the root mesh recursively
 *         cdef int i
 *         if self.root_nodes== NULL: return             # <<<<<<<<<<<<<<
 *         if self.loaded == 0:
 *             for i in range(self.max_root):
 */
  __pyx_t_1 = ((__pyx_v_self->__pyx_base.root_nodes == NULL) != 0);
  if (__pyx_t_1) {
    goto __pyx_L0;
  }

  /* "yt/geometry/particle_oct_container.pyx":2072
 *         cdef int i
 *         if self.root_nodes== NULL: return
 *         if self.loaded == 0:             # <<<<<<<<<<<<<<
 *             for i in range(self.max_root):
 *                 if self.root_nodes[i].node == NULL: continue
 */
  __pyx_t_1 = ((__pyx_v_self->loaded == 0) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/particle_oct_container.pyx":2073
 *         if self.root_nodes== NULL: return
 *         if self.loaded == 0:
 *             for i in range(self.max_root):             # <<<<<<<<<<<<<<
 *                 if self.root_nodes[i].node == NULL: continue
 *                 self.visit_free(&self.root_nodes.node[i], 0)
 */
    __pyx_t_2 = __pyx_v_self->__pyx_base.max_root;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "yt/geometry/particle_oct_container.pyx":2074
 *         if self.loaded == 0:
 *             for i in range(self.max_root):
 *                 if self.root_nodes[i].node == NULL: continue             # <<<<<<<<<<<<<<
 *                 self.visit_free(&self.root_nodes.node[i], 0)
 *             self.root_nodes = NULL
 */
      __pyx_t_1 = (((__pyx_v_self->__pyx_base.root_nodes[__pyx_v_i]).node == NULL) != 0);
      if (__pyx_t_1) {
        goto __pyx_L5_continue;
      }

      /* "yt/geometry/particle_oct_container.pyx":2075
 *             for i in range(self.max_root):
 *                 if self.root_nodes[i].node == NULL: continue
 *                 self.visit_free(&self.root_nodes.node[i], 0)             # <<<<<<<<<<<<<<
 *             self.root_nodes = NULL
 *         free(self.oct_list)
 */
      ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->visit_free(__pyx_v_self, (&(__pyx_v_self->__pyx_base.root_nodes->node[__pyx_v_i])), 0);
      __pyx_L5_continue:;
    }

    /* "yt/geometry/particle_oct_container.pyx":2076
 *                 if self.root_nodes[i].node == NULL: continue
 *                 self.visit_free(&self.root_nodes.node[i], 0)
 *             self.root_nodes = NULL             # <<<<<<<<<<<<<<
 *         free(self.oct_list)
 *         free(self._ptr_index_base_roots)
 */
    __pyx_v_self->__pyx_base.root_nodes = NULL;

    /* "yt/geometry/particle_oct_container.pyx":2072
 *         cdef int i
 *         if self.root_nodes== NULL: return
 *         if self.loaded == 0:             # <<<<<<<<<<<<<<
 *             for i in range(self.max_root):
 *                 if self.root_nodes[i].node == NULL: continue
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":2077
 *                 self.visit_free(&self.root_nodes.node[i], 0)
 *             self.root_nodes = NULL
 *         free(self.oct_list)             # <<<<<<<<<<<<<<
 *         free(self._ptr_index_base_roots)
 *         free(self._ptr_index_base_octs)
 */
  free(__pyx_v_self->oct_list);

  /* "yt/geometry/particle_oct_container.pyx":2078
 *             self.root_nodes = NULL
 *         free(self.oct_list)
 *         free(self._ptr_index_base_roots)             # <<<<<<<<<<<<<<
 *         free(self._ptr_index_base_octs)
 *         free(self._ptr_octs_per_root)
 */
  free(__pyx_v_self->_ptr_index_base_roots);

  /* "yt/geometry/particle_oct_container.pyx":2079
 *         free(self.oct_list)
 *         free(self._ptr_index_base_roots)
 *         free(self._ptr_index_base_octs)             # <<<<<<<<<<<<<<
 *         free(self._ptr_octs_per_root)
 *         self.oct_list = NULL
 */
  free(__pyx_v_self->_ptr_index_base_octs);

  /* "yt/geometry/particle_oct_container.pyx":2080
 *         free(self._ptr_index_base_roots)
 *         free(self._ptr_index_base_octs)
 *         free(self._ptr_octs_per_root)             # <<<<<<<<<<<<<<
 *         self.oct_list = NULL
 * 
 */
  free(__pyx_v_self->_ptr_octs_per_root);

  /* "yt/geometry/particle_oct_container.pyx":2081
 *         free(self._ptr_index_base_octs)
 *         free(self._ptr_octs_per_root)
 *         self.oct_list = NULL             # <<<<<<<<<<<<<<
 * 
 *     cdef void visit_free(self, Oct *o, int free_this):
 */
  __pyx_v_self->oct_list = NULL;

  /* "yt/geometry/particle_oct_container.pyx":2067
 *         self.visit_all_octs(selector, visitor)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         #Call the freemem ops on every ocy
 *         #of the root mesh recursively
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":2083
 *         self.oct_list = NULL
 * 
 *     cdef void visit_free(self, Oct *o, int free_this):             # <<<<<<<<<<<<<<
 *         #Free the memory for this oct recursively
 *         cdef int i, j, k
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_visit_free(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o, int __pyx_v_free_this) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("visit_free", 0);

  /* "yt/geometry/particle_oct_container.pyx":2086
 *         #Free the memory for this oct recursively
 *         cdef int i, j, k
 *         for i in range(2):             # <<<<<<<<<<<<<<
 *             for j in range(2):
 *                 for k in range(2):
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 2; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/geometry/particle_oct_container.pyx":2087
 *         cdef int i, j, k
 *         for i in range(2):
 *             for j in range(2):             # <<<<<<<<<<<<<<
 *                 for k in range(2):
 *                     if o.children != NULL \
 */
    for (__pyx_t_2 = 0; __pyx_t_2 < 2; __pyx_t_2+=1) {
      __pyx_v_j = __pyx_t_2;

      /* "yt/geometry/particle_oct_container.pyx":2088
 *         for i in range(2):
 *             for j in range(2):
 *                 for k in range(2):             # <<<<<<<<<<<<<<
 *                     if o.children != NULL \
 *                        and o.children[cind(i,j,k)] != NULL:
 */
      for (__pyx_t_3 = 0; __pyx_t_3 < 2; __pyx_t_3+=1) {
        __pyx_v_k = __pyx_t_3;

        /* "yt/geometry/particle_oct_container.pyx":2089
 *             for j in range(2):
 *                 for k in range(2):
 *                     if o.children != NULL \             # <<<<<<<<<<<<<<
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_free(o.children[cind(i,j,k)], 1)
 */
        __pyx_t_5 = ((__pyx_v_o->children != NULL) != 0);
        if (__pyx_t_5) {
        } else {
          __pyx_t_4 = __pyx_t_5;
          goto __pyx_L10_bool_binop_done;
        }

        /* "yt/geometry/particle_oct_container.pyx":2090
 *                 for k in range(2):
 *                     if o.children != NULL \
 *                        and o.children[cind(i,j,k)] != NULL:             # <<<<<<<<<<<<<<
 *                         self.visit_free(o.children[cind(i,j,k)], 1)
 *         if o.children != NULL:
 */
        __pyx_t_5 = (((__pyx_v_o->children[__pyx_f_2yt_8geometry_12oct_visitors_cind(__pyx_v_i, __pyx_v_j, __pyx_v_k)]) != NULL) != 0);
        __pyx_t_4 = __pyx_t_5;
        __pyx_L10_bool_binop_done:;

        /* "yt/geometry/particle_oct_container.pyx":2089
 *             for j in range(2):
 *                 for k in range(2):
 *                     if o.children != NULL \             # <<<<<<<<<<<<<<
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_free(o.children[cind(i,j,k)], 1)
 */
        if (__pyx_t_4) {

          /* "yt/geometry/particle_oct_container.pyx":2091
 *                     if o.children != NULL \
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_free(o.children[cind(i,j,k)], 1)             # <<<<<<<<<<<<<<
 *         if o.children != NULL:
 *             free(o.children)
 */
          ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->visit_free(__pyx_v_self, (__pyx_v_o->children[__pyx_f_2yt_8geometry_12oct_visitors_cind(__pyx_v_i, __pyx_v_j, __pyx_v_k)]), 1);

          /* "yt/geometry/particle_oct_container.pyx":2089
 *             for j in range(2):
 *                 for k in range(2):
 *                     if o.children != NULL \             # <<<<<<<<<<<<<<
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_free(o.children[cind(i,j,k)], 1)
 */
        }
      }
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":2092
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_free(o.children[cind(i,j,k)], 1)
 *         if o.children != NULL:             # <<<<<<<<<<<<<<
 *             free(o.children)
 *         if free_this == 1:
 */
  __pyx_t_4 = ((__pyx_v_o->children != NULL) != 0);
  if (__pyx_t_4) {

    /* "yt/geometry/particle_oct_container.pyx":2093
 *                         self.visit_free(o.children[cind(i,j,k)], 1)
 *         if o.children != NULL:
 *             free(o.children)             # <<<<<<<<<<<<<<
 *         if free_this == 1:
 *             free(o)
 */
    free(__pyx_v_o->children);

    /* "yt/geometry/particle_oct_container.pyx":2092
 *                        and o.children[cind(i,j,k)] != NULL:
 *                         self.visit_free(o.children[cind(i,j,k)], 1)
 *         if o.children != NULL:             # <<<<<<<<<<<<<<
 *             free(o.children)
 *         if free_this == 1:
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":2094
 *         if o.children != NULL:
 *             free(o.children)
 *         if free_this == 1:             # <<<<<<<<<<<<<<
 *             free(o)
 * 
 */
  __pyx_t_4 = ((__pyx_v_free_this == 1) != 0);
  if (__pyx_t_4) {

    /* "yt/geometry/particle_oct_container.pyx":2095
 *             free(o.children)
 *         if free_this == 1:
 *             free(o)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
    free(__pyx_v_o);

    /* "yt/geometry/particle_oct_container.pyx":2094
 *         if o.children != NULL:
 *             free(o.children)
 *         if free_this == 1:             # <<<<<<<<<<<<<<
 *             free(o)
 * 
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":2083
 *         self.oct_list = NULL
 * 
 *     cdef void visit_free(self, Oct *o, int free_this):             # <<<<<<<<<<<<<<
 *         #Free the memory for this oct recursively
 *         cdef int i, j, k
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":2100
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void recursive_add(self, Oct *o, np.ndarray[np.uint64_t, ndim=1] indices,             # <<<<<<<<<<<<<<
 *                             int level, int *max_level, int domain_id, int *count):
 *         cdef np.int64_t no = indices.shape[0], beg, end, nind
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_recursive_add(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o, PyArrayObject *__pyx_v_indices, int __pyx_v_level, int *__pyx_v_max_level, int __pyx_v_domain_id, int *__pyx_v_count) {
  __pyx_t_5numpy_int64_t __pyx_v_no;
  __pyx_t_5numpy_int64_t __pyx_v_beg;
  __pyx_t_5numpy_int64_t __pyx_v_end;
  __pyx_t_5numpy_int64_t __pyx_v_nind;
  __pyx_t_5numpy_int64_t __pyx_v_index;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_ind[3];
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_noct;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  long __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("recursive_add", 0);
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 2100, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/particle_oct_container.pyx":2102
 *     cdef void recursive_add(self, Oct *o, np.ndarray[np.uint64_t, ndim=1] indices,
 *                             int level, int *max_level, int domain_id, int *count):
 *         cdef np.int64_t no = indices.shape[0], beg, end, nind             # <<<<<<<<<<<<<<
 *         cdef np.int64_t index
 *         cdef int i, j, k
 */
  __pyx_v_no = (__pyx_v_indices->dimensions[0]);

  /* "yt/geometry/particle_oct_container.pyx":2108
 *         cdef Oct *noct
 *         cdef Oct *noct_ch
 *         beg = end = 0             # <<<<<<<<<<<<<<
 *         if level > max_level[0]: max_level[0] = level
 *         # Initialize children
 */
  __pyx_v_beg = 0;
  __pyx_v_end = 0;

  /* "yt/geometry/particle_oct_container.pyx":2109
 *         cdef Oct *noct_ch
 *         beg = end = 0
 *         if level > max_level[0]: max_level[0] = level             # <<<<<<<<<<<<<<
 *         # Initialize children
 *         if o.children == NULL:
 */
  __pyx_t_1 = ((__pyx_v_level > (__pyx_v_max_level[0])) != 0);
  if (__pyx_t_1) {
    (__pyx_v_max_level[0]) = __pyx_v_level;
  }

  /* "yt/geometry/particle_oct_container.pyx":2111
 *         if level > max_level[0]: max_level[0] = level
 *         # Initialize children
 *         if o.children == NULL:             # <<<<<<<<<<<<<<
 *             o.children = <Oct **> malloc(sizeof(Oct *)*8)
 *             for i in range(2):
 */
  __pyx_t_1 = ((__pyx_v_o->children == NULL) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/particle_oct_container.pyx":2112
 *         # Initialize children
 *         if o.children == NULL:
 *             o.children = <Oct **> malloc(sizeof(Oct *)*8)             # <<<<<<<<<<<<<<
 *             for i in range(2):
 *                 for j in range(2):
 */
    __pyx_v_o->children = ((struct __pyx_t_2yt_8geometry_12oct_visitors_Oct **)malloc(((sizeof(struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *)) * 8)));

    /* "yt/geometry/particle_oct_container.pyx":2113
 *         if o.children == NULL:
 *             o.children = <Oct **> malloc(sizeof(Oct *)*8)
 *             for i in range(2):             # <<<<<<<<<<<<<<
 *                 for j in range(2):
 *                     for k in range(2):
 */
    for (__pyx_t_2 = 0; __pyx_t_2 < 2; __pyx_t_2+=1) {
      __pyx_v_i = __pyx_t_2;

      /* "yt/geometry/particle_oct_container.pyx":2114
 *             o.children = <Oct **> malloc(sizeof(Oct *)*8)
 *             for i in range(2):
 *                 for j in range(2):             # <<<<<<<<<<<<<<
 *                     for k in range(2):
 *                         o.children[cind(i,j,k)] = NULL
 */
      for (__pyx_t_3 = 0; __pyx_t_3 < 2; __pyx_t_3+=1) {
        __pyx_v_j = __pyx_t_3;

        /* "yt/geometry/particle_oct_container.pyx":2115
 *             for i in range(2):
 *                 for j in range(2):
 *                     for k in range(2):             # <<<<<<<<<<<<<<
 *                         o.children[cind(i,j,k)] = NULL
 *                         # noct = self.allocate_oct()
 */
        for (__pyx_t_4 = 0; __pyx_t_4 < 2; __pyx_t_4+=1) {
          __pyx_v_k = __pyx_t_4;

          /* "yt/geometry/particle_oct_container.pyx":2116
 *                 for j in range(2):
 *                     for k in range(2):
 *                         o.children[cind(i,j,k)] = NULL             # <<<<<<<<<<<<<<
 *                         # noct = self.allocate_oct()
 *                         # noct.domain = o.domain
 */
          (__pyx_v_o->children[__pyx_f_2yt_8geometry_12oct_visitors_cind(__pyx_v_i, __pyx_v_j, __pyx_v_k)]) = NULL;
        }
      }
    }

    /* "yt/geometry/particle_oct_container.pyx":2111
 *         if level > max_level[0]: max_level[0] = level
 *         # Initialize children
 *         if o.children == NULL:             # <<<<<<<<<<<<<<
 *             o.children = <Oct **> malloc(sizeof(Oct *)*8)
 *             for i in range(2):
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":2122
 *                         # o.children[cind(i,j,k)] = noct
 *         # Loop through sets of particles with matching prefix at this level
 *         while end < no:             # <<<<<<<<<<<<<<
 *             beg = end
 *             index = (indices[beg] >> ((ORDER_MAX - level)*3))
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_end < __pyx_v_no) != 0);
    if (!__pyx_t_1) break;

    /* "yt/geometry/particle_oct_container.pyx":2123
 *         # Loop through sets of particles with matching prefix at this level
 *         while end < no:
 *             beg = end             # <<<<<<<<<<<<<<
 *             index = (indices[beg] >> ((ORDER_MAX - level)*3))
 *             while (end < no) and (index == (indices[end] >> ((ORDER_MAX - level)*3))):
 */
    __pyx_v_beg = __pyx_v_end;

    /* "yt/geometry/particle_oct_container.pyx":2124
 *         while end < no:
 *             beg = end
 *             index = (indices[beg] >> ((ORDER_MAX - level)*3))             # <<<<<<<<<<<<<<
 *             while (end < no) and (index == (indices[end] >> ((ORDER_MAX - level)*3))):
 *                 end += 1
 */
    __pyx_t_5 = __pyx_v_beg;
    __pyx_v_index = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_indices.rcbuffer->pybuffer.buf, __pyx_t_5, __pyx_pybuffernd_indices.diminfo[0].strides)) >> ((__pyx_v_2yt_8geometry_13oct_container_ORDER_MAX - __pyx_v_level) * 3));

    /* "yt/geometry/particle_oct_container.pyx":2125
 *             beg = end
 *             index = (indices[beg] >> ((ORDER_MAX - level)*3))
 *             while (end < no) and (index == (indices[end] >> ((ORDER_MAX - level)*3))):             # <<<<<<<<<<<<<<
 *                 end += 1
 *             nind = (end - beg)
 */
    while (1) {
      __pyx_t_6 = ((__pyx_v_end < __pyx_v_no) != 0);
      if (__pyx_t_6) {
      } else {
        __pyx_t_1 = __pyx_t_6;
        goto __pyx_L15_bool_binop_done;
      }
      __pyx_t_5 = __pyx_v_end;
      __pyx_t_6 = ((__pyx_v_index == ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_indices.rcbuffer->pybuffer.buf, __pyx_t_5, __pyx_pybuffernd_indices.diminfo[0].strides)) >> ((__pyx_v_2yt_8geometry_13oct_container_ORDER_MAX - __pyx_v_level) * 3))) != 0);
      __pyx_t_1 = __pyx_t_6;
      __pyx_L15_bool_binop_done:;
      if (!__pyx_t_1) break;

      /* "yt/geometry/particle_oct_container.pyx":2126
 *             index = (indices[beg] >> ((ORDER_MAX - level)*3))
 *             while (end < no) and (index == (indices[end] >> ((ORDER_MAX - level)*3))):
 *                 end += 1             # <<<<<<<<<<<<<<
 *             nind = (end - beg)
 *             # Add oct
 */
      __pyx_v_end = (__pyx_v_end + 1);
    }

    /* "yt/geometry/particle_oct_container.pyx":2127
 *             while (end < no) and (index == (indices[end] >> ((ORDER_MAX - level)*3))):
 *                 end += 1
 *             nind = (end - beg)             # <<<<<<<<<<<<<<
 *             # Add oct
 *             for i in range(3):
 */
    __pyx_v_nind = (__pyx_v_end - __pyx_v_beg);

    /* "yt/geometry/particle_oct_container.pyx":2129
 *             nind = (end - beg)
 *             # Add oct
 *             for i in range(3):             # <<<<<<<<<<<<<<
 *                 ind[i] = ((index >> (2 - i)) & 1)
 *             # noct = o.children[cind(ind[0],ind[1],ind[2])]
 */
    for (__pyx_t_2 = 0; __pyx_t_2 < 3; __pyx_t_2+=1) {
      __pyx_v_i = __pyx_t_2;

      /* "yt/geometry/particle_oct_container.pyx":2130
 *             # Add oct
 *             for i in range(3):
 *                 ind[i] = ((index >> (2 - i)) & 1)             # <<<<<<<<<<<<<<
 *             # noct = o.children[cind(ind[0],ind[1],ind[2])]
 *             if o.children[cind(ind[0],ind[1],ind[2])] != NULL:
 */
      (__pyx_v_ind[__pyx_v_i]) = ((__pyx_v_index >> (2 - __pyx_v_i)) & 1);
    }

    /* "yt/geometry/particle_oct_container.pyx":2132
 *                 ind[i] = ((index >> (2 - i)) & 1)
 *             # noct = o.children[cind(ind[0],ind[1],ind[2])]
 *             if o.children[cind(ind[0],ind[1],ind[2])] != NULL:             # <<<<<<<<<<<<<<
 *                 raise Exception('Child was already initialized...')
 *             noct = self.allocate_oct()
 */
    __pyx_t_1 = (((__pyx_v_o->children[__pyx_f_2yt_8geometry_12oct_visitors_cind((__pyx_v_ind[0]), (__pyx_v_ind[1]), (__pyx_v_ind[2]))]) != NULL) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "yt/geometry/particle_oct_container.pyx":2133
 *             # noct = o.children[cind(ind[0],ind[1],ind[2])]
 *             if o.children[cind(ind[0],ind[1],ind[2])] != NULL:
 *                 raise Exception('Child was already initialized...')             # <<<<<<<<<<<<<<
 *             noct = self.allocate_oct()
 *             noct.domain = o.domain
 */
      __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_Raise(__pyx_t_7, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __PYX_ERR(0, 2133, __pyx_L1_error)

      /* "yt/geometry/particle_oct_container.pyx":2132
 *                 ind[i] = ((index >> (2 - i)) & 1)
 *             # noct = o.children[cind(ind[0],ind[1],ind[2])]
 *             if o.children[cind(ind[0],ind[1],ind[2])] != NULL:             # <<<<<<<<<<<<<<
 *                 raise Exception('Child was already initialized...')
 *             noct = self.allocate_oct()
 */
    }

    /* "yt/geometry/particle_oct_container.pyx":2134
 *             if o.children[cind(ind[0],ind[1],ind[2])] != NULL:
 *                 raise Exception('Child was already initialized...')
 *             noct = self.allocate_oct()             # <<<<<<<<<<<<<<
 *             noct.domain = o.domain
 *             o.children[cind(ind[0],ind[1],ind[2])] = noct
 */
    __pyx_v_noct = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->allocate_oct(__pyx_v_self);

    /* "yt/geometry/particle_oct_container.pyx":2135
 *                 raise Exception('Child was already initialized...')
 *             noct = self.allocate_oct()
 *             noct.domain = o.domain             # <<<<<<<<<<<<<<
 *             o.children[cind(ind[0],ind[1],ind[2])] = noct
 *             # Don't add it to the list if it will be refined
 */
    __pyx_t_5 = __pyx_v_o->domain;
    __pyx_v_noct->domain = __pyx_t_5;

    /* "yt/geometry/particle_oct_container.pyx":2136
 *             noct = self.allocate_oct()
 *             noct.domain = o.domain
 *             o.children[cind(ind[0],ind[1],ind[2])] = noct             # <<<<<<<<<<<<<<
 *             # Don't add it to the list if it will be refined
 *             if nind > self.n_ref and level < ORDER_MAX:
 */
    (__pyx_v_o->children[__pyx_f_2yt_8geometry_12oct_visitors_cind((__pyx_v_ind[0]), (__pyx_v_ind[1]), (__pyx_v_ind[2]))]) = __pyx_v_noct;

    /* "yt/geometry/particle_oct_container.pyx":2138
 *             o.children[cind(ind[0],ind[1],ind[2])] = noct
 *             # Don't add it to the list if it will be refined
 *             if nind > self.n_ref and level < ORDER_MAX:             # <<<<<<<<<<<<<<
 *                 self.nocts -= 1
 *                 noct.domain_ind = -1 # overwritten by finalize
 */
    __pyx_t_6 = ((__pyx_v_nind > __pyx_v_self->n_ref) != 0);
    if (__pyx_t_6) {
    } else {
      __pyx_t_1 = __pyx_t_6;
      goto __pyx_L21_bool_binop_done;
    }
    __pyx_t_6 = ((__pyx_v_level < __pyx_v_2yt_8geometry_13oct_container_ORDER_MAX) != 0);
    __pyx_t_1 = __pyx_t_6;
    __pyx_L21_bool_binop_done:;
    if (__pyx_t_1) {

      /* "yt/geometry/particle_oct_container.pyx":2139
 *             # Don't add it to the list if it will be refined
 *             if nind > self.n_ref and level < ORDER_MAX:
 *                 self.nocts -= 1             # <<<<<<<<<<<<<<
 *                 noct.domain_ind = -1 # overwritten by finalize
 *             else:
 */
      __pyx_v_self->__pyx_base.__pyx_base.nocts = (__pyx_v_self->__pyx_base.__pyx_base.nocts - 1);

      /* "yt/geometry/particle_oct_container.pyx":2140
 *             if nind > self.n_ref and level < ORDER_MAX:
 *                 self.nocts -= 1
 *                 noct.domain_ind = -1 # overwritten by finalize             # <<<<<<<<<<<<<<
 *             else:
 *                 count[0] += 1
 */
      __pyx_v_noct->domain_ind = -1LL;

      /* "yt/geometry/particle_oct_container.pyx":2138
 *             o.children[cind(ind[0],ind[1],ind[2])] = noct
 *             # Don't add it to the list if it will be refined
 *             if nind > self.n_ref and level < ORDER_MAX:             # <<<<<<<<<<<<<<
 *                 self.nocts -= 1
 *                 noct.domain_ind = -1 # overwritten by finalize
 */
      goto __pyx_L20;
    }

    /* "yt/geometry/particle_oct_container.pyx":2142
 *                 noct.domain_ind = -1 # overwritten by finalize
 *             else:
 *                 count[0] += 1             # <<<<<<<<<<<<<<
 *             noct.file_ind = o.file_ind
 *             # noct.file_ind = nind
 */
    /*else*/ {
      __pyx_t_8 = 0;
      (__pyx_v_count[__pyx_t_8]) = ((__pyx_v_count[__pyx_t_8]) + 1);
    }
    __pyx_L20:;

    /* "yt/geometry/particle_oct_container.pyx":2143
 *             else:
 *                 count[0] += 1
 *             noct.file_ind = o.file_ind             # <<<<<<<<<<<<<<
 *             # noct.file_ind = nind
 *             # o.file_ind = self.n_ref + 1
 */
    __pyx_t_5 = __pyx_v_o->file_ind;
    __pyx_v_noct->file_ind = __pyx_t_5;

    /* "yt/geometry/particle_oct_container.pyx":2147
 *             # o.file_ind = self.n_ref + 1
 *             # Refine oct or add its children
 *             if nind > self.n_ref and level < ORDER_MAX:             # <<<<<<<<<<<<<<
 *                 self.recursive_add(noct, indices[beg:end], level+1,
 *                                    max_level, domain_id, count)
 */
    __pyx_t_6 = ((__pyx_v_nind > __pyx_v_self->n_ref) != 0);
    if (__pyx_t_6) {
    } else {
      __pyx_t_1 = __pyx_t_6;
      goto __pyx_L24_bool_binop_done;
    }
    __pyx_t_6 = ((__pyx_v_level < __pyx_v_2yt_8geometry_13oct_container_ORDER_MAX) != 0);
    __pyx_t_1 = __pyx_t_6;
    __pyx_L24_bool_binop_done:;
    if (__pyx_t_1) {

      /* "yt/geometry/particle_oct_container.pyx":2148
 *             # Refine oct or add its children
 *             if nind > self.n_ref and level < ORDER_MAX:
 *                 self.recursive_add(noct, indices[beg:end], level+1,             # <<<<<<<<<<<<<<
 *                                    max_level, domain_id, count)
 * 
 */
      __pyx_t_7 = __Pyx_PyInt_From_npy_int64(__pyx_v_beg); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2148, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_9 = __Pyx_PyInt_From_npy_int64(__pyx_v_end); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2148, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PySlice_New(__pyx_t_7, __pyx_t_9, Py_None); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2148, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_indices), __pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2148, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (!(likely(((__pyx_t_9) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_9, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 2148, __pyx_L1_error)

      /* "yt/geometry/particle_oct_container.pyx":2149
 *             if nind > self.n_ref and level < ORDER_MAX:
 *                 self.recursive_add(noct, indices[beg:end], level+1,
 *                                    max_level, domain_id, count)             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
      ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->recursive_add(__pyx_v_self, __pyx_v_noct, ((PyArrayObject *)__pyx_t_9), (__pyx_v_level + 1), __pyx_v_max_level, __pyx_v_domain_id, __pyx_v_count);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "yt/geometry/particle_oct_container.pyx":2147
 *             # o.file_ind = self.n_ref + 1
 *             # Refine oct or add its children
 *             if nind > self.n_ref and level < ORDER_MAX:             # <<<<<<<<<<<<<<
 *                 self.recursive_add(noct, indices[beg:end], level+1,
 *                                    max_level, domain_id, count)
 */
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":2100
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef void recursive_add(self, Oct *o, np.ndarray[np.uint64_t, ndim=1] indices,             # <<<<<<<<<<<<<<
 *                             int level, int *max_level, int domain_id, int *count):
 *         cdef np.int64_t no = indices.shape[0], beg, end, nind
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_WriteUnraisable("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer.recursive_add", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":2154
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def add(self, np.ndarray[np.uint64_t, ndim=1] indices,             # <<<<<<<<<<<<<<
 *              np.uint64_t order1, int domain_id = -1):
 *         #Add this particle to the root oct
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_11add(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_11add(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyArrayObject *__pyx_v_indices = 0;
  CYTHON_UNUSED __pyx_t_5numpy_uint64_t __pyx_v_order1;
  int __pyx_v_domain_id;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_indices,&__pyx_n_s_order1,&__pyx_n_s_domain_id,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_indices)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_order1)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("add", 0, 2, 3, 1); __PYX_ERR(0, 2154, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_domain_id);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "add") < 0)) __PYX_ERR(0, 2154, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_indices = ((PyArrayObject *)values[0]);
    __pyx_v_order1 = __Pyx_PyInt_As_npy_uint64(values[1]); if (unlikely((__pyx_v_order1 == ((npy_uint64)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2155, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_domain_id = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_domain_id == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2155, __pyx_L3_error)
    } else {
      __pyx_v_domain_id = ((int)-1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2154, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer.add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_indices), __pyx_ptype_5numpy_ndarray, 1, "indices", 0))) __PYX_ERR(0, 2154, __pyx_L1_error)
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_10add(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self), __pyx_v_indices, __pyx_v_order1, __pyx_v_domain_id);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_10add(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, PyArrayObject *__pyx_v_indices, CYTHON_UNUSED __pyx_t_5numpy_uint64_t __pyx_v_order1, int __pyx_v_domain_id) {
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_root;
  __pyx_t_5numpy_int64_t __pyx_v_no;
  __pyx_t_5numpy_int64_t __pyx_v_beg;
  __pyx_t_5numpy_int64_t __pyx_v_end;
  __pyx_t_5numpy_int64_t __pyx_v_index;
  CYTHON_UNUSED __pyx_t_5numpy_int64_t __pyx_v_nind;
  int __pyx_v_i;
  int __pyx_v_ind[3];
  __pyx_t_5numpy_uint64_t __pyx_v_ind64[3];
  int __pyx_v_max_level;
  __pyx_t_5numpy_int64_t __pyx_v_index_root;
  int __pyx_v_root_count;
  __Pyx_LocalBuf_ND __pyx_pybuffernd_indices;
  __Pyx_Buffer __pyx_pybuffer_indices;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __pyx_t_5numpy_int64_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add", 0);
  __pyx_pybuffer_indices.pybuffer.buf = NULL;
  __pyx_pybuffer_indices.refcount = 0;
  __pyx_pybuffernd_indices.data = NULL;
  __pyx_pybuffernd_indices.rcbuffer = &__pyx_pybuffer_indices;
  {
    __Pyx_BufFmt_StackElem __pyx_stack[1];
    if (unlikely(__Pyx_GetBufferAndValidate(&__pyx_pybuffernd_indices.rcbuffer->pybuffer, (PyObject*)__pyx_v_indices, &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, PyBUF_FORMAT| PyBUF_STRIDES, 1, 0, __pyx_stack) == -1)) __PYX_ERR(0, 2154, __pyx_L1_error)
  }
  __pyx_pybuffernd_indices.diminfo[0].strides = __pyx_pybuffernd_indices.rcbuffer->pybuffer.strides[0]; __pyx_pybuffernd_indices.diminfo[0].shape = __pyx_pybuffernd_indices.rcbuffer->pybuffer.shape[0];

  /* "yt/geometry/particle_oct_container.pyx":2159
 *         #Then if that oct has children, add it to them recursively
 *         #If the child needs to be refined because of max particles, do so
 *         cdef Oct *root = NULL             # <<<<<<<<<<<<<<
 *         cdef np.int64_t no = indices.shape[0], beg, end, index, nind
 *         cdef int i, level
 */
  __pyx_v_root = NULL;

  /* "yt/geometry/particle_oct_container.pyx":2160
 *         #If the child needs to be refined because of max particles, do so
 *         cdef Oct *root = NULL
 *         cdef np.int64_t no = indices.shape[0], beg, end, index, nind             # <<<<<<<<<<<<<<
 *         cdef int i, level
 *         cdef int ind[3]
 */
  __pyx_v_no = (__pyx_v_indices->dimensions[0]);

  /* "yt/geometry/particle_oct_container.pyx":2164
 *         cdef int ind[3]
 *         cdef np.uint64_t ind64[3]
 *         cdef int max_level = self.max_level             # <<<<<<<<<<<<<<
 *         # Note what we're doing here: we have decided the root will always be
 *         # zero, since we're in a forest of octrees, where the root_mesh node is
 */
  __pyx_t_1 = __pyx_v_self->max_level;
  __pyx_v_max_level = __pyx_t_1;

  /* "yt/geometry/particle_oct_container.pyx":2170
 *         # respect to that, which means we need to keep a few different arrays
 *         # of them.
 *         cdef np.int64_t index_root = 0             # <<<<<<<<<<<<<<
 *         cdef int root_count
 *         beg = end = 0
 */
  __pyx_v_index_root = 0;

  /* "yt/geometry/particle_oct_container.pyx":2172
 *         cdef np.int64_t index_root = 0
 *         cdef int root_count
 *         beg = end = 0             # <<<<<<<<<<<<<<
 *         self._octs_per_root[:] = 1 # Roots count regardless
 *         while end < no:
 */
  __pyx_v_beg = 0;
  __pyx_v_end = 0;

  /* "yt/geometry/particle_oct_container.pyx":2173
 *         cdef int root_count
 *         beg = end = 0
 *         self._octs_per_root[:] = 1 # Roots count regardless             # <<<<<<<<<<<<<<
 *         while end < no:
 *             # Determine number of octs with this prefix
 */
  if (unlikely(!__pyx_v_self->_octs_per_root.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 2173, __pyx_L1_error)}
  {
      __pyx_t_5numpy_uint64_t __pyx_temp_scalar = 1;
      {
          Py_ssize_t __pyx_temp_extent_0 = __pyx_v_self->_octs_per_root.shape[0];
          Py_ssize_t __pyx_temp_stride_0 = __pyx_v_self->_octs_per_root.strides[0];
          char *__pyx_temp_pointer_0;
          Py_ssize_t __pyx_temp_idx_0;
          __pyx_temp_pointer_0 = __pyx_v_self->_octs_per_root.data;
          for (__pyx_temp_idx_0 = 0; __pyx_temp_idx_0 < __pyx_temp_extent_0; __pyx_temp_idx_0++) {
            *((__pyx_t_5numpy_uint64_t *) __pyx_temp_pointer_0) = __pyx_temp_scalar;
            __pyx_temp_pointer_0 += __pyx_temp_stride_0;
          }
      }
  }

  /* "yt/geometry/particle_oct_container.pyx":2174
 *         beg = end = 0
 *         self._octs_per_root[:] = 1 # Roots count regardless
 *         while end < no:             # <<<<<<<<<<<<<<
 *             # Determine number of octs with this prefix
 *             beg = end
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_end < __pyx_v_no) != 0);
    if (!__pyx_t_2) break;

    /* "yt/geometry/particle_oct_container.pyx":2176
 *         while end < no:
 *             # Determine number of octs with this prefix
 *             beg = end             # <<<<<<<<<<<<<<
 *             index = (indices[beg] >> ((ORDER_MAX - self.level_offset)*3))
 *             while (end < no) and (index == (indices[end] >> ((ORDER_MAX - self.level_offset)*3))):
 */
    __pyx_v_beg = __pyx_v_end;

    /* "yt/geometry/particle_oct_container.pyx":2177
 *             # Determine number of octs with this prefix
 *             beg = end
 *             index = (indices[beg] >> ((ORDER_MAX - self.level_offset)*3))             # <<<<<<<<<<<<<<
 *             while (end < no) and (index == (indices[end] >> ((ORDER_MAX - self.level_offset)*3))):
 *                 end += 1
 */
    __pyx_t_3 = __pyx_v_beg;
    __pyx_v_index = ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_indices.rcbuffer->pybuffer.buf, __pyx_t_3, __pyx_pybuffernd_indices.diminfo[0].strides)) >> ((__pyx_v_2yt_8geometry_13oct_container_ORDER_MAX - __pyx_v_self->__pyx_base.__pyx_base.level_offset) * 3));

    /* "yt/geometry/particle_oct_container.pyx":2178
 *             beg = end
 *             index = (indices[beg] >> ((ORDER_MAX - self.level_offset)*3))
 *             while (end < no) and (index == (indices[end] >> ((ORDER_MAX - self.level_offset)*3))):             # <<<<<<<<<<<<<<
 *                 end += 1
 *             nind = (end - beg)
 */
    while (1) {
      __pyx_t_4 = ((__pyx_v_end < __pyx_v_no) != 0);
      if (__pyx_t_4) {
      } else {
        __pyx_t_2 = __pyx_t_4;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_3 = __pyx_v_end;
      __pyx_t_4 = ((__pyx_v_index == ((*__Pyx_BufPtrStrided1d(__pyx_t_5numpy_uint64_t *, __pyx_pybuffernd_indices.rcbuffer->pybuffer.buf, __pyx_t_3, __pyx_pybuffernd_indices.diminfo[0].strides)) >> ((__pyx_v_2yt_8geometry_13oct_container_ORDER_MAX - __pyx_v_self->__pyx_base.__pyx_base.level_offset) * 3))) != 0);
      __pyx_t_2 = __pyx_t_4;
      __pyx_L7_bool_binop_done:;
      if (!__pyx_t_2) break;

      /* "yt/geometry/particle_oct_container.pyx":2179
 *             index = (indices[beg] >> ((ORDER_MAX - self.level_offset)*3))
 *             while (end < no) and (index == (indices[end] >> ((ORDER_MAX - self.level_offset)*3))):
 *                 end += 1             # <<<<<<<<<<<<<<
 *             nind = (end - beg)
 *             # Find root for prefix
 */
      __pyx_v_end = (__pyx_v_end + 1);
    }

    /* "yt/geometry/particle_oct_container.pyx":2180
 *             while (end < no) and (index == (indices[end] >> ((ORDER_MAX - self.level_offset)*3))):
 *                 end += 1
 *             nind = (end - beg)             # <<<<<<<<<<<<<<
 *             # Find root for prefix
 *             decode_morton_64bit(index, ind64)
 */
    __pyx_v_nind = (__pyx_v_end - __pyx_v_beg);

    /* "yt/geometry/particle_oct_container.pyx":2182
 *             nind = (end - beg)
 *             # Find root for prefix
 *             decode_morton_64bit(index, ind64)             # <<<<<<<<<<<<<<
 *             for i in range(3):
 *                 ind[i] = ind64[i]
 */
    __pyx_f_2yt_9utilities_3lib_14geometry_utils_decode_morton_64bit(__pyx_v_index, __pyx_v_ind64);

    /* "yt/geometry/particle_oct_container.pyx":2183
 *             # Find root for prefix
 *             decode_morton_64bit(index, ind64)
 *             for i in range(3):             # <<<<<<<<<<<<<<
 *                 ind[i] = ind64[i]
 *             while (index_root < self.num_root) and \
 */
    for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
      __pyx_v_i = __pyx_t_1;

      /* "yt/geometry/particle_oct_container.pyx":2184
 *             decode_morton_64bit(index, ind64)
 *             for i in range(3):
 *                 ind[i] = ind64[i]             # <<<<<<<<<<<<<<
 *             while (index_root < self.num_root) and \
 *                   (self.ipos_to_key(ind) != self.root_nodes[index_root].key):
 */
      (__pyx_v_ind[__pyx_v_i]) = (__pyx_v_ind64[__pyx_v_i]);
    }

    /* "yt/geometry/particle_oct_container.pyx":2185
 *             for i in range(3):
 *                 ind[i] = ind64[i]
 *             while (index_root < self.num_root) and \             # <<<<<<<<<<<<<<
 *                   (self.ipos_to_key(ind) != self.root_nodes[index_root].key):
 *                 index_root += 1
 */
    while (1) {
      __pyx_t_4 = ((__pyx_v_index_root < __pyx_v_self->__pyx_base.num_root) != 0);
      if (__pyx_t_4) {
      } else {
        __pyx_t_2 = __pyx_t_4;
        goto __pyx_L13_bool_binop_done;
      }

      /* "yt/geometry/particle_oct_container.pyx":2186
 *                 ind[i] = ind64[i]
 *             while (index_root < self.num_root) and \
 *                   (self.ipos_to_key(ind) != self.root_nodes[index_root].key):             # <<<<<<<<<<<<<<
 *                 index_root += 1
 *             if index_root >= self.num_root:
 */
      __pyx_t_4 = ((((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.ipos_to_key(((struct __pyx_obj_2yt_8geometry_13oct_container_SparseOctreeContainer *)__pyx_v_self), __pyx_v_ind) != (__pyx_v_self->__pyx_base.root_nodes[__pyx_v_index_root]).key) != 0);
      __pyx_t_2 = __pyx_t_4;
      __pyx_L13_bool_binop_done:;
      if (!__pyx_t_2) break;

      /* "yt/geometry/particle_oct_container.pyx":2187
 *             while (index_root < self.num_root) and \
 *                   (self.ipos_to_key(ind) != self.root_nodes[index_root].key):
 *                 index_root += 1             # <<<<<<<<<<<<<<
 *             if index_root >= self.num_root:
 *                 raise Exception('No root found for {},{},{}'.format(ind[0],ind[1],ind[2]))
 */
      __pyx_v_index_root = (__pyx_v_index_root + 1);
    }

    /* "yt/geometry/particle_oct_container.pyx":2188
 *                   (self.ipos_to_key(ind) != self.root_nodes[index_root].key):
 *                 index_root += 1
 *             if index_root >= self.num_root:             # <<<<<<<<<<<<<<
 *                 raise Exception('No root found for {},{},{}'.format(ind[0],ind[1],ind[2]))
 *             root = self.root_nodes[index_root].node
 */
    __pyx_t_2 = ((__pyx_v_index_root >= __pyx_v_self->__pyx_base.num_root) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "yt/geometry/particle_oct_container.pyx":2189
 *                 index_root += 1
 *             if index_root >= self.num_root:
 *                 raise Exception('No root found for {},{},{}'.format(ind[0],ind[1],ind[2]))             # <<<<<<<<<<<<<<
 *             root = self.root_nodes[index_root].node
 *             # self.get_root(ind, &root)
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_No_root_found_for, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyInt_From_int((__pyx_v_ind[0])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyInt_From_int((__pyx_v_ind[1])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyInt_From_int((__pyx_v_ind[2])); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = NULL;
      __pyx_t_1 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_1 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_t_7, __pyx_t_8, __pyx_t_9};
        __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_1, 3+__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2189, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_t_7, __pyx_t_8, __pyx_t_9};
        __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_1, 3+__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2189, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else
      #endif
      {
        __pyx_t_11 = PyTuple_New(3+__pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2189, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (__pyx_t_10) {
          __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_10); __pyx_t_10 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_1, __pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_1, __pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_1, __pyx_t_9);
        __pyx_t_7 = 0;
        __pyx_t_8 = 0;
        __pyx_t_9 = 0;
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_11, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2189, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 2189, __pyx_L1_error)

      /* "yt/geometry/particle_oct_container.pyx":2188
 *                   (self.ipos_to_key(ind) != self.root_nodes[index_root].key):
 *                 index_root += 1
 *             if index_root >= self.num_root:             # <<<<<<<<<<<<<<
 *                 raise Exception('No root found for {},{},{}'.format(ind[0],ind[1],ind[2]))
 *             root = self.root_nodes[index_root].node
 */
    }

    /* "yt/geometry/particle_oct_container.pyx":2190
 *             if index_root >= self.num_root:
 *                 raise Exception('No root found for {},{},{}'.format(ind[0],ind[1],ind[2]))
 *             root = self.root_nodes[index_root].node             # <<<<<<<<<<<<<<
 *             # self.get_root(ind, &root)
 *             # if root == NULL:
 */
    __pyx_t_12 = (__pyx_v_self->__pyx_base.root_nodes[__pyx_v_index_root]).node;
    __pyx_v_root = __pyx_t_12;

    /* "yt/geometry/particle_oct_container.pyx":2194
 *             # if root == NULL:
 *             #     raise Exception('No root found for {},{},{}'.format(ind[0],ind[1],ind[2]))
 *             root.file_ind = index_root             # <<<<<<<<<<<<<<
 *             # Refine root as necessary
 *             if (end - beg) > self.n_ref:
 */
    __pyx_v_root->file_ind = __pyx_v_index_root;

    /* "yt/geometry/particle_oct_container.pyx":2196
 *             root.file_ind = index_root
 *             # Refine root as necessary
 *             if (end - beg) > self.n_ref:             # <<<<<<<<<<<<<<
 *                 root_count = 0
 *                 self.nocts -= 1
 */
    __pyx_t_2 = (((__pyx_v_end - __pyx_v_beg) > __pyx_v_self->n_ref) != 0);
    if (__pyx_t_2) {

      /* "yt/geometry/particle_oct_container.pyx":2197
 *             # Refine root as necessary
 *             if (end - beg) > self.n_ref:
 *                 root_count = 0             # <<<<<<<<<<<<<<
 *                 self.nocts -= 1
 *                 self.recursive_add(root, indices[beg:end], self.level_offset+1,
 */
      __pyx_v_root_count = 0;

      /* "yt/geometry/particle_oct_container.pyx":2198
 *             if (end - beg) > self.n_ref:
 *                 root_count = 0
 *                 self.nocts -= 1             # <<<<<<<<<<<<<<
 *                 self.recursive_add(root, indices[beg:end], self.level_offset+1,
 *                                    &max_level, domain_id, &root_count)
 */
      __pyx_v_self->__pyx_base.__pyx_base.nocts = (__pyx_v_self->__pyx_base.__pyx_base.nocts - 1);

      /* "yt/geometry/particle_oct_container.pyx":2199
 *                 root_count = 0
 *                 self.nocts -= 1
 *                 self.recursive_add(root, indices[beg:end], self.level_offset+1,             # <<<<<<<<<<<<<<
 *                                    &max_level, domain_id, &root_count)
 *                 self._octs_per_root[index_root] = <np.uint64_t>root_count
 */
      __pyx_t_6 = __Pyx_PyInt_From_npy_int64(__pyx_v_beg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2199, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = __Pyx_PyInt_From_npy_int64(__pyx_v_end); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2199, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_11 = PySlice_New(__pyx_t_6, __pyx_t_5, Py_None); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2199, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_indices), __pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2199, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_5numpy_ndarray))))) __PYX_ERR(0, 2199, __pyx_L1_error)

      /* "yt/geometry/particle_oct_container.pyx":2200
 *                 self.nocts -= 1
 *                 self.recursive_add(root, indices[beg:end], self.level_offset+1,
 *                                    &max_level, domain_id, &root_count)             # <<<<<<<<<<<<<<
 *                 self._octs_per_root[index_root] = <np.uint64_t>root_count
 *         self.max_level = max_level
 */
      ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->recursive_add(__pyx_v_self, __pyx_v_root, ((PyArrayObject *)__pyx_t_5), (__pyx_v_self->__pyx_base.__pyx_base.level_offset + 1), (&__pyx_v_max_level), __pyx_v_domain_id, (&__pyx_v_root_count));
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "yt/geometry/particle_oct_container.pyx":2201
 *                 self.recursive_add(root, indices[beg:end], self.level_offset+1,
 *                                    &max_level, domain_id, &root_count)
 *                 self._octs_per_root[index_root] = <np.uint64_t>root_count             # <<<<<<<<<<<<<<
 *         self.max_level = max_level
 *         assert(self.nocts == np.sum(self._octs_per_root))
 */
      if (unlikely(!__pyx_v_self->_octs_per_root.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 2201, __pyx_L1_error)}
      __pyx_t_3 = __pyx_v_index_root;
      *((__pyx_t_5numpy_uint64_t *) ( /* dim=0 */ (__pyx_v_self->_octs_per_root.data + __pyx_t_3 * __pyx_v_self->_octs_per_root.strides[0]) )) = ((__pyx_t_5numpy_uint64_t)__pyx_v_root_count);

      /* "yt/geometry/particle_oct_container.pyx":2196
 *             root.file_ind = index_root
 *             # Refine root as necessary
 *             if (end - beg) > self.n_ref:             # <<<<<<<<<<<<<<
 *                 root_count = 0
 *                 self.nocts -= 1
 */
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":2202
 *                                    &max_level, domain_id, &root_count)
 *                 self._octs_per_root[index_root] = <np.uint64_t>root_count
 *         self.max_level = max_level             # <<<<<<<<<<<<<<
 *         assert(self.nocts == np.sum(self._octs_per_root))
 * 
 */
  __pyx_v_self->max_level = __pyx_v_max_level;

  /* "yt/geometry/particle_oct_container.pyx":2203
 *                 self._octs_per_root[index_root] = <np.uint64_t>root_count
 *         self.max_level = max_level
 *         assert(self.nocts == np.sum(self._octs_per_root))             # <<<<<<<<<<<<<<
 * 
 *     @cython.boundscheck(False)
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_5 = __Pyx_PyInt_From_npy_int64(__pyx_v_self->__pyx_base.__pyx_base.nocts); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_sum); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_v_self->_octs_per_root.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 2203, __pyx_L1_error)}
    __pyx_t_6 = __pyx_memoryview_fromslice(__pyx_v_self->_octs_per_root, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint64_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint64_t, 0);; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_11 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 2203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyObject_RichCompare(__pyx_t_5, __pyx_t_11, Py_EQ); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 2203, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 2203, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_2)) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 2203, __pyx_L1_error)
    }
  }
  #endif

  /* "yt/geometry/particle_oct_container.pyx":2154
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def add(self, np.ndarray[np.uint64_t, ndim=1] indices,             # <<<<<<<<<<<<<<
 *              np.uint64_t order1, int domain_id = -1):
 *         #Add this particle to the root oct
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  { PyObject *__pyx_type, *__pyx_value, *__pyx_tb;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&__pyx_type, &__pyx_value, &__pyx_tb);
    __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __Pyx_ErrRestore(__pyx_type, __pyx_value, __pyx_tb);}
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer.add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  goto __pyx_L2;
  __pyx_L0:;
  __Pyx_SafeReleaseBuffer(&__pyx_pybuffernd_indices.rcbuffer->pybuffer);
  __pyx_L2:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":2208
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef Oct *refine_oct(self, Oct *o, np.uint64_t index, int level):             # <<<<<<<<<<<<<<
 *         #Allocate and initialize child octs
 *         #Attach particles to child octs
 */

static struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_refine_oct(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o, __pyx_t_5numpy_uint64_t __pyx_v_index, int __pyx_v_level) {
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_v_k;
  int __pyx_v_ind[3];
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_noct;
  struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __pyx_t_5numpy_int64_t __pyx_t_5;
  __Pyx_RefNannySetupContext("refine_oct", 0);

  /* "yt/geometry/particle_oct_container.pyx":2217
 * 
 *         # Initialize empty children
 *         if o.children == NULL:             # <<<<<<<<<<<<<<
 *             o.children = <Oct **> malloc(sizeof(Oct *)*8)
 * 
 */
  __pyx_t_1 = ((__pyx_v_o->children == NULL) != 0);
  if (__pyx_t_1) {

    /* "yt/geometry/particle_oct_container.pyx":2218
 *         # Initialize empty children
 *         if o.children == NULL:
 *             o.children = <Oct **> malloc(sizeof(Oct *)*8)             # <<<<<<<<<<<<<<
 * 
 *         # This version can be used to just add the child containing the index
 */
    __pyx_v_o->children = ((struct __pyx_t_2yt_8geometry_12oct_visitors_Oct **)malloc(((sizeof(struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *)) * 8)));

    /* "yt/geometry/particle_oct_container.pyx":2217
 * 
 *         # Initialize empty children
 *         if o.children == NULL:             # <<<<<<<<<<<<<<
 *             o.children = <Oct **> malloc(sizeof(Oct *)*8)
 * 
 */
  }

  /* "yt/geometry/particle_oct_container.pyx":2236
 * 
 * 
 *         for i in range(2):             # <<<<<<<<<<<<<<
 *             for j in range(2):
 *                 for k in range(2):
 */
  for (__pyx_t_2 = 0; __pyx_t_2 < 2; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "yt/geometry/particle_oct_container.pyx":2237
 * 
 *         for i in range(2):
 *             for j in range(2):             # <<<<<<<<<<<<<<
 *                 for k in range(2):
 *                     noct = self.allocate_oct()
 */
    for (__pyx_t_3 = 0; __pyx_t_3 < 2; __pyx_t_3+=1) {
      __pyx_v_j = __pyx_t_3;

      /* "yt/geometry/particle_oct_container.pyx":2238
 *         for i in range(2):
 *             for j in range(2):
 *                 for k in range(2):             # <<<<<<<<<<<<<<
 *                     noct = self.allocate_oct()
 *                     noct.domain = o.domain
 */
      for (__pyx_t_4 = 0; __pyx_t_4 < 2; __pyx_t_4+=1) {
        __pyx_v_k = __pyx_t_4;

        /* "yt/geometry/particle_oct_container.pyx":2239
 *             for j in range(2):
 *                 for k in range(2):
 *                     noct = self.allocate_oct()             # <<<<<<<<<<<<<<
 *                     noct.domain = o.domain
 *                     noct.file_ind = 0
 */
        __pyx_v_noct = ((struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->allocate_oct(__pyx_v_self);

        /* "yt/geometry/particle_oct_container.pyx":2240
 *                 for k in range(2):
 *                     noct = self.allocate_oct()
 *                     noct.domain = o.domain             # <<<<<<<<<<<<<<
 *                     noct.file_ind = 0
 *                     o.children[cind(i,j,k)] = noct
 */
        __pyx_t_5 = __pyx_v_o->domain;
        __pyx_v_noct->domain = __pyx_t_5;

        /* "yt/geometry/particle_oct_container.pyx":2241
 *                     noct = self.allocate_oct()
 *                     noct.domain = o.domain
 *                     noct.file_ind = 0             # <<<<<<<<<<<<<<
 *                     o.children[cind(i,j,k)] = noct
 *         o.file_ind = self.n_ref + 1
 */
        __pyx_v_noct->file_ind = 0;

        /* "yt/geometry/particle_oct_container.pyx":2242
 *                     noct.domain = o.domain
 *                     noct.file_ind = 0
 *                     o.children[cind(i,j,k)] = noct             # <<<<<<<<<<<<<<
 *         o.file_ind = self.n_ref + 1
 *         for i in range(3):
 */
        (__pyx_v_o->children[__pyx_f_2yt_8geometry_12oct_visitors_cind(__pyx_v_i, __pyx_v_j, __pyx_v_k)]) = __pyx_v_noct;
      }
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":2243
 *                     noct.file_ind = 0
 *                     o.children[cind(i,j,k)] = noct
 *         o.file_ind = self.n_ref + 1             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             ind[i] = (index >> ((ORDER_MAX - level)*3 + (2 - i))) & 1
 */
  __pyx_v_o->file_ind = (__pyx_v_self->n_ref + 1);

  /* "yt/geometry/particle_oct_container.pyx":2244
 *                     o.children[cind(i,j,k)] = noct
 *         o.file_ind = self.n_ref + 1
 *         for i in range(3):             # <<<<<<<<<<<<<<
 *             ind[i] = (index >> ((ORDER_MAX - level)*3 + (2 - i))) & 1
 *         noct = o.children[cind(ind[0],ind[1],ind[2])]
 */
  for (__pyx_t_2 = 0; __pyx_t_2 < 3; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "yt/geometry/particle_oct_container.pyx":2245
 *         o.file_ind = self.n_ref + 1
 *         for i in range(3):
 *             ind[i] = (index >> ((ORDER_MAX - level)*3 + (2 - i))) & 1             # <<<<<<<<<<<<<<
 *         noct = o.children[cind(ind[0],ind[1],ind[2])]
 *         return noct
 */
    (__pyx_v_ind[__pyx_v_i]) = ((__pyx_v_index >> (((__pyx_v_2yt_8geometry_13oct_container_ORDER_MAX - __pyx_v_level) * 3) + (2 - __pyx_v_i))) & 1);
  }

  /* "yt/geometry/particle_oct_container.pyx":2246
 *         for i in range(3):
 *             ind[i] = (index >> ((ORDER_MAX - level)*3 + (2 - i))) & 1
 *         noct = o.children[cind(ind[0],ind[1],ind[2])]             # <<<<<<<<<<<<<<
 *         return noct
 * 
 */
  __pyx_v_noct = (__pyx_v_o->children[__pyx_f_2yt_8geometry_12oct_visitors_cind((__pyx_v_ind[0]), (__pyx_v_ind[1]), (__pyx_v_ind[2]))]);

  /* "yt/geometry/particle_oct_container.pyx":2247
 *             ind[i] = (index >> ((ORDER_MAX - level)*3 + (2 - i))) & 1
 *         noct = o.children[cind(ind[0],ind[1],ind[2])]
 *         return noct             # <<<<<<<<<<<<<<
 * 
 *     cdef void filter_particles(self, Oct *o, np.uint64_t *data, np.int64_t p,
 */
  __pyx_r = __pyx_v_noct;
  goto __pyx_L0;

  /* "yt/geometry/particle_oct_container.pyx":2208
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef Oct *refine_oct(self, Oct *o, np.uint64_t index, int level):             # <<<<<<<<<<<<<<
 *         #Allocate and initialize child octs
 *         #Attach particles to child octs
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":2249
 *         return noct
 * 
 *     cdef void filter_particles(self, Oct *o, np.uint64_t *data, np.int64_t p,             # <<<<<<<<<<<<<<
 *                                int level):
 *         # Now we look at the last nref particles to decide where they go.
 */

static void __pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_filter_particles(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *__pyx_v_o, __pyx_t_5numpy_uint64_t *__pyx_v_data, __pyx_t_5numpy_int64_t __pyx_v_p, int __pyx_v_level) {
  int __pyx_v_n;
  __pyx_t_5numpy_uint64_t *__pyx_v_arr;
  __pyx_t_5numpy_uint64_t __pyx_v_prefix1;
  int __pyx_v_i;
  __pyx_t_5numpy_uint64_t __pyx_v_prefix2;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("filter_particles", 0);

  /* "yt/geometry/particle_oct_container.pyx":2252
 *                                int level):
 *         # Now we look at the last nref particles to decide where they go.
 *         cdef int n = imin(p, self.n_ref)             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t *arr = data + imax(p - self.n_ref, 0)
 *         # Now we figure out our prefix, which is the oct address at this level.
 */
  __pyx_v_n = __pyx_f_2yt_9utilities_3lib_8fp_utils_imin(__pyx_v_p, __pyx_v_self->n_ref);

  /* "yt/geometry/particle_oct_container.pyx":2253
 *         # Now we look at the last nref particles to decide where they go.
 *         cdef int n = imin(p, self.n_ref)
 *         cdef np.uint64_t *arr = data + imax(p - self.n_ref, 0)             # <<<<<<<<<<<<<<
 *         # Now we figure out our prefix, which is the oct address at this level.
 *         # As long as we're actually in Morton order, we do not need to worry
 */
  __pyx_v_arr = (__pyx_v_data + __pyx_f_2yt_9utilities_3lib_8fp_utils_imax((__pyx_v_p - __pyx_v_self->n_ref), 0));

  /* "yt/geometry/particle_oct_container.pyx":2257
 *         # As long as we're actually in Morton order, we do not need to worry
 *         # about *any* of the other children of the oct.
 *         prefix1 = data[p] >> (ORDER_MAX - level)*3             # <<<<<<<<<<<<<<
 *         for i in range(n):
 *             prefix2 = arr[i] >> (ORDER_MAX - level)*3
 */
  __pyx_v_prefix1 = ((__pyx_v_data[__pyx_v_p]) >> ((__pyx_v_2yt_8geometry_13oct_container_ORDER_MAX - __pyx_v_level) * 3));

  /* "yt/geometry/particle_oct_container.pyx":2258
 *         # about *any* of the other children of the oct.
 *         prefix1 = data[p] >> (ORDER_MAX - level)*3
 *         for i in range(n):             # <<<<<<<<<<<<<<
 *             prefix2 = arr[i] >> (ORDER_MAX - level)*3
 *             if (prefix1 == prefix2):
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "yt/geometry/particle_oct_container.pyx":2259
 *         prefix1 = data[p] >> (ORDER_MAX - level)*3
 *         for i in range(n):
 *             prefix2 = arr[i] >> (ORDER_MAX - level)*3             # <<<<<<<<<<<<<<
 *             if (prefix1 == prefix2):
 *                 o.file_ind += 1
 */
    __pyx_v_prefix2 = ((__pyx_v_arr[__pyx_v_i]) >> ((__pyx_v_2yt_8geometry_13oct_container_ORDER_MAX - __pyx_v_level) * 3));

    /* "yt/geometry/particle_oct_container.pyx":2260
 *         for i in range(n):
 *             prefix2 = arr[i] >> (ORDER_MAX - level)*3
 *             if (prefix1 == prefix2):             # <<<<<<<<<<<<<<
 *                 o.file_ind += 1
 */
    __pyx_t_4 = ((__pyx_v_prefix1 == __pyx_v_prefix2) != 0);
    if (__pyx_t_4) {

      /* "yt/geometry/particle_oct_container.pyx":2261
 *             prefix2 = arr[i] >> (ORDER_MAX - level)*3
 *             if (prefix1 == prefix2):
 *                 o.file_ind += 1             # <<<<<<<<<<<<<<
 */
      __pyx_v_o->file_ind = (__pyx_v_o->file_ind + 1);

      /* "yt/geometry/particle_oct_container.pyx":2260
 *         for i in range(n):
 *             prefix2 = arr[i] >> (ORDER_MAX - level)*3
 *             if (prefix1 == prefix2):             # <<<<<<<<<<<<<<
 *                 o.file_ind += 1
 */
    }
  }

  /* "yt/geometry/particle_oct_container.pyx":2249
 *         return noct
 * 
 *     cdef void filter_particles(self, Oct *o, np.uint64_t *data, np.int64_t p,             # <<<<<<<<<<<<<<
 *                                int level):
 *         # Now we look at the last nref particles to decide where they go.
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "yt/geometry/particle_oct_container.pyx":1967
 * cdef class ParticleBitmapOctreeContainer(SparseOctreeContainer):
 *     cdef Oct** oct_list
 *     cdef public int max_level             # <<<<<<<<<<<<<<
 *     cdef public int n_ref
 *     cdef int loaded # Loaded with load_octree?
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_9max_level_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_9max_level_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_9max_level___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_9max_level___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->max_level); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer.max_level.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_9max_level_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_9max_level_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_9max_level_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_9max_level_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1967, __pyx_L1_error)
  __pyx_v_self->max_level = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer.max_level.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1968
 *     cdef Oct** oct_list
 *     cdef public int max_level
 *     cdef public int n_ref             # <<<<<<<<<<<<<<
 *     cdef int loaded # Loaded with load_octree?
 *     cdef np.uint8_t* _ptr_index_base_roots
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_5n_ref_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_5n_ref_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_5n_ref___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_5n_ref___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->n_ref); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1968, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer.n_ref.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_5n_ref_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_5n_ref_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_5n_ref_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_5n_ref_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1968, __pyx_L1_error)
  __pyx_v_self->n_ref = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer.n_ref.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1973
 *     cdef np.uint8_t* _ptr_index_base_octs
 *     cdef np.uint64_t* _ptr_octs_per_root
 *     cdef public np.uint8_t[:] _index_base_roots             # <<<<<<<<<<<<<<
 *     cdef public np.uint8_t[:] _index_base_octs
 *     cdef np.uint64_t[:] _octs_per_root
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_17_index_base_roots_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_17_index_base_roots_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_17_index_base_roots___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_17_index_base_roots___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_self->_index_base_roots.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1973, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->_index_base_roots, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint8_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint8_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer._index_base_roots.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_17_index_base_roots_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_17_index_base_roots_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_17_index_base_roots_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_17_index_base_roots_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint8_t(__pyx_v_value, PyBUF_WRITABLE); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(0, 1973, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->_index_base_roots, 0);
  __pyx_v_self->_index_base_roots = __pyx_t_1;
  __pyx_t_1.memview = NULL;
  __pyx_t_1.data = NULL;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_1, 1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer._index_base_roots.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1974
 *     cdef np.uint64_t* _ptr_octs_per_root
 *     cdef public np.uint8_t[:] _index_base_roots
 *     cdef public np.uint8_t[:] _index_base_octs             # <<<<<<<<<<<<<<
 *     cdef np.uint64_t[:] _octs_per_root
 *     cdef public int overlap_cells
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_16_index_base_octs_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_16_index_base_octs_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_16_index_base_octs___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_16_index_base_octs___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_self->_index_base_octs.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1974, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->_index_base_octs, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn___pyx_t_5numpy_uint8_t, (int (*)(char *, PyObject *)) __pyx_memview_set_nn___pyx_t_5numpy_uint8_t, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer._index_base_octs.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_16_index_base_octs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_16_index_base_octs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_16_index_base_octs_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_16_index_base_octs_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint8_t(__pyx_v_value, PyBUF_WRITABLE); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(0, 1974, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->_index_base_octs, 0);
  __pyx_v_self->_index_base_octs = __pyx_t_1;
  __pyx_t_1.memview = NULL;
  __pyx_t_1.data = NULL;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_1, 1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer._index_base_octs.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/geometry/particle_oct_container.pyx":1976
 *     cdef public np.uint8_t[:] _index_base_octs
 *     cdef np.uint64_t[:] _octs_per_root
 *     cdef public int overlap_cells             # <<<<<<<<<<<<<<
 *     def __init__(self, domain_dimensions, domain_left_edge, domain_right_edge,
 *                  int num_root, num_zones = 2):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_13overlap_cells_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_13overlap_cells_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_13overlap_cells___get__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_13overlap_cells___get__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->overlap_cells); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer.overlap_cells.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_13overlap_cells_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_13overlap_cells_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_13overlap_cells_2__set__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_13overlap_cells_2__set__(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1976, __pyx_L1_error)
  __pyx_v_self->overlap_cells = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer.overlap_cells.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._ptr_octs_per_root,self.oct_list,self.root_mesh,self.root_nodes,self.tree_root cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_13__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_13__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_12__reduce_cython__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self._ptr_octs_per_root,self.oct_list,self.root_mesh,self.root_nodes,self.tree_root cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._ptr_octs_per_root,self.oct_list,self.root_mesh,self.root_nodes,self.tree_root cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._ptr_octs_per_root,self.oct_list,self.root_mesh,self.root_nodes,self.tree_root cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self._ptr_octs_per_root,self.oct_list,self.root_mesh,self.root_nodes,self.tree_root cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._ptr_octs_per_root,self.oct_list,self.root_mesh,self.root_nodes,self.tree_root cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_15__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_15__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_14__setstate_cython__(((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self._ptr_octs_per_root,self.oct_list,self.root_mesh,self.root_nodes,self.tree_root cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._ptr_octs_per_root,self.oct_list,self.root_mesh,self.root_nodes,self.tree_root cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self._ptr_octs_per_root,self.oct_list,self.root_mesh,self.root_nodes,self.tree_root cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._ptr_octs_per_root,self.oct_list,self.root_mesh,self.root_nodes,self.tree_root cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5numpy_7ndarray___getbuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_i;
  int __pyx_v_ndim;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  int __pyx_v_t;
  char *__pyx_v_f;
  PyArray_Descr *__pyx_v_descr = 0;
  int __pyx_v_offset;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyArray_Descr *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  char *__pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":265
 * 
 *             cdef int i, ndim
 *             cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 */
  __pyx_v_endian_detector = 1;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":266
 *             cdef int i, ndim
 *             cdef int endian_detector = 1
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 * 
 *             ndim = PyArray_NDIM(self)
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":268
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 *             ndim = PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_self);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":271
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_C_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 272, __pyx_L1_error)

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *             ndim = PyArray_NDIM(self)
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":275
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_ARRAY_F_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  if (unlikely(__pyx_t_1)) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 276, __pyx_L1_error)

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":278
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 *             info.buf = PyArray_DATA(self)             # <<<<<<<<<<<<<<
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_v_info->buf = PyArray_DATA(__pyx_v_self);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":279
 * 
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 # Allocate new buffer for strides and shape info.
 */
  __pyx_v_info->ndim = __pyx_v_ndim;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":283
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)             # <<<<<<<<<<<<<<
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 */
    __pyx_v_info->strides = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * 2) * ((size_t)__pyx_v_ndim))));

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":284
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim             # <<<<<<<<<<<<<<
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 */
    __pyx_v_info->shape = (__pyx_v_info->strides + __pyx_v_ndim);

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":285
 *                 info.strides = <Py_ssize_t*>PyObject_Malloc(sizeof(Py_ssize_t) * 2 * <size_t>ndim)
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):             # <<<<<<<<<<<<<<
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 */
    __pyx_t_4 = __pyx_v_ndim;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":286
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]             # <<<<<<<<<<<<<<
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 */
      (__pyx_v_info->strides[__pyx_v_i]) = (PyArray_STRIDES(__pyx_v_self)[__pyx_v_i]);

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":287
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 */
      (__pyx_v_info->shape[__pyx_v_i]) = (PyArray_DIMS(__pyx_v_self)[__pyx_v_i]);
    }

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
    goto __pyx_L9;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":289
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 */
  /*else*/ {
    __pyx_v_info->strides = ((Py_ssize_t *)PyArray_STRIDES(__pyx_v_self));

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":290
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 */
    __pyx_v_info->shape = ((Py_ssize_t *)PyArray_DIMS(__pyx_v_self));
  }
  __pyx_L9:;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":291
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 */
  __pyx_v_info->suboffsets = NULL;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":292
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)             # <<<<<<<<<<<<<<
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 * 
 */
  __pyx_v_info->itemsize = PyArray_ITEMSIZE(__pyx_v_self);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":293
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)             # <<<<<<<<<<<<<<
 * 
 *             cdef int t
 */
  __pyx_v_info->readonly = (!(PyArray_ISWRITEABLE(__pyx_v_self) != 0));

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":296
 * 
 *             cdef int t
 *             cdef char* f = NULL             # <<<<<<<<<<<<<<
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)
 *             cdef int offset
 */
  __pyx_v_f = NULL;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":297
 *             cdef int t
 *             cdef char* f = NULL
 *             cdef dtype descr = <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 *             cdef int offset
 * 
 */
  __pyx_t_7 = PyArray_DESCR(__pyx_v_self);
  __pyx_t_3 = ((PyObject *)__pyx_t_7);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_descr = ((PyArray_Descr *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":300
 *             cdef int offset
 * 
 *             info.obj = self             # <<<<<<<<<<<<<<
 * 
 *             if not PyDataType_HASFIELDS(descr):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  __pyx_t_1 = ((!(PyDataType_HASFIELDS(__pyx_v_descr) != 0)) != 0);
  if (__pyx_t_1) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":303
 * 
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num             # <<<<<<<<<<<<<<
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 */
    __pyx_t_4 = __pyx_v_descr->type_num;
    __pyx_v_t = __pyx_t_4;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '>') != 0);
    if (!__pyx_t_2) {
      goto __pyx_L15_next_or;
    } else {
    }
    __pyx_t_2 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_L15_next_or:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":305
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '<') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_2 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L14_bool_binop_done:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_1)) {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 306, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 306, __pyx_L1_error)

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":304
 *             if not PyDataType_HASFIELDS(descr):
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":307
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 */
    switch (__pyx_v_t) {
      case NPY_BYTE:
      __pyx_v_f = ((char *)"b");
      break;
      case NPY_UBYTE:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":308
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 */
      __pyx_v_f = ((char *)"B");
      break;
      case NPY_SHORT:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":309
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 */
      __pyx_v_f = ((char *)"h");
      break;
      case NPY_USHORT:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":310
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 */
      __pyx_v_f = ((char *)"H");
      break;
      case NPY_INT:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":311
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 */
      __pyx_v_f = ((char *)"i");
      break;
      case NPY_UINT:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":312
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 */
      __pyx_v_f = ((char *)"I");
      break;
      case NPY_LONG:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":313
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 */
      __pyx_v_f = ((char *)"l");
      break;
      case NPY_ULONG:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":314
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 */
      __pyx_v_f = ((char *)"L");
      break;
      case NPY_LONGLONG:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":315
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 */
      __pyx_v_f = ((char *)"q");
      break;
      case NPY_ULONGLONG:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":316
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 */
      __pyx_v_f = ((char *)"Q");
      break;
      case NPY_FLOAT:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":317
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 */
      __pyx_v_f = ((char *)"f");
      break;
      case NPY_DOUBLE:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":318
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 */
      __pyx_v_f = ((char *)"d");
      break;
      case NPY_LONGDOUBLE:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":319
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 */
      __pyx_v_f = ((char *)"g");
      break;
      case NPY_CFLOAT:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":320
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 */
      __pyx_v_f = ((char *)"Zf");
      break;
      case NPY_CDOUBLE:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":321
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"
 */
      __pyx_v_f = ((char *)"Zd");
      break;
      case NPY_CLONGDOUBLE:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":322
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 */
      __pyx_v_f = ((char *)"Zg");
      break;
      case NPY_OBJECT:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":323
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_v_f = ((char *)"O");
      break;
      default:

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":325
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *                 info.format = f
 *                 return
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 325, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 325, __pyx_L1_error)
      break;
    }

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":326
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
    __pyx_v_info->format = __pyx_v_f;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":327
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":302
 *             info.obj = self
 * 
 *             if not PyDataType_HASFIELDS(descr):             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":329
 *                 return
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)             # <<<<<<<<<<<<<<
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 */
  /*else*/ {
    __pyx_v_info->format = ((char *)PyObject_Malloc(0xFF));

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":330
 *             else:
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment             # <<<<<<<<<<<<<<
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,
 */
    (__pyx_v_info->format[0]) = '^';

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":331
 *                 info.format = <char*>PyObject_Malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0             # <<<<<<<<<<<<<<
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 */
    __pyx_v_offset = 0;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":332
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,             # <<<<<<<<<<<<<<
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 */
    __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_descr, (__pyx_v_info->format + 1), (__pyx_v_info->format + 0xFF), (&__pyx_v_offset)); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(2, 332, __pyx_L1_error)
    __pyx_v_f = __pyx_t_9;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":335
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 *                 f[0] = c'\0' # Terminate format string             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 */
    (__pyx_v_f[0]) = '\x00';
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fulfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.ndarray.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_descr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_5numpy_7ndarray_2__releasebuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_t_1 = (PyArray_HASFIELDS(__pyx_v_self) != 0);
  if (__pyx_t_1) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":339
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)
 */
    PyObject_Free(__pyx_v_info->format);

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":338
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":341
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 PyObject_Free(info.strides)             # <<<<<<<<<<<<<<
 *                 # info.shape was stored after info.strides in the same block
 * 
 */
    PyObject_Free(__pyx_v_info->strides);

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":340
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 PyObject_Free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":337
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 PyObject_Free(info.format)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":820
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":820
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":823
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":824
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":823
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":826
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":827
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":826
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":829
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":830
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":829
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":832
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":833
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":832
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":835
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":836
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":837
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":836
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":839
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":835
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":841
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *__pyx_v_descr, char *__pyx_v_f, char *__pyx_v_end, int *__pyx_v_offset) {
  PyArray_Descr *__pyx_v_child = 0;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_childname = NULL;
  PyObject *__pyx_v_new_offset = NULL;
  PyObject *__pyx_v_t = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  long __pyx_t_8;
  char *__pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_util_dtypestring", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":846
 * 
 *     cdef dtype child
 *     cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 *     cdef tuple fields
 */
  __pyx_v_endian_detector = 1;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":847
 *     cdef dtype child
 *     cdef int endian_detector = 1
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 *     cdef tuple fields
 * 
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":850
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  if (unlikely(__pyx_v_descr->names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(2, 850, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_descr->names; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(2, 850, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 850, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_childname, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":851
 * 
 *     for childname in descr.names:
 *         fields = descr.fields[childname]             # <<<<<<<<<<<<<<
 *         child, new_offset = fields
 * 
 */
    if (unlikely(__pyx_v_descr->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 851, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_descr->fields, __pyx_v_childname); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 851, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyTuple_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(2, 851, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_fields, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":852
 *     for childname in descr.names:
 *         fields = descr.fields[childname]
 *         child, new_offset = fields             # <<<<<<<<<<<<<<
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 */
    if (likely(__pyx_v_fields != Py_None)) {
      PyObject* sequence = __pyx_v_fields;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(2, 852, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 852, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 852, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(2, 852, __pyx_L1_error)
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_dtype))))) __PYX_ERR(2, 852, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_child, ((PyArray_Descr *)__pyx_t_3));
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_new_offset, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":854
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 854, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_new_offset, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 854, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 854, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = ((((__pyx_v_end - __pyx_v_f) - ((int)__pyx_t_5)) < 15) != 0);
    if (unlikely(__pyx_t_6)) {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":855
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 855, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 855, __pyx_L1_error)

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":854
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    }

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":857
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '>') != 0);
    if (!__pyx_t_7) {
      goto __pyx_L8_next_or;
    } else {
    }
    __pyx_t_7 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_L8_next_or:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":858
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *             raise ValueError(u"Non-native byte order not supported")
 *             # One could encode it in the format string and have Cython
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '<') != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_7 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L7_bool_binop_done:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":857
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    if (unlikely(__pyx_t_6)) {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":859
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 859, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(2, 859, __pyx_L1_error)

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":857
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":869
 * 
 *         # Output padding bytes
 *         while offset[0] < new_offset:             # <<<<<<<<<<<<<<
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 */
    while (1) {
      __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 869, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_v_new_offset, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 869, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 869, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!__pyx_t_6) break;

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":870
 *         # Output padding bytes
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte             # <<<<<<<<<<<<<<
 *             f += 1
 *             offset[0] += 1
 */
      (__pyx_v_f[0]) = 0x78;

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":871
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte
 *             f += 1             # <<<<<<<<<<<<<<
 *             offset[0] += 1
 * 
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":872
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 *             offset[0] += 1             # <<<<<<<<<<<<<<
 * 
 *         offset[0] += child.itemsize
 */
      __pyx_t_8 = 0;
      (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + 1);
    }

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":874
 *             offset[0] += 1
 * 
 *         offset[0] += child.itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not PyDataType_HASFIELDS(child):
 */
    __pyx_t_8 = 0;
    (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + __pyx_v_child->elsize);

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":876
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
    __pyx_t_6 = ((!(PyDataType_HASFIELDS(__pyx_v_child) != 0)) != 0);
    if (__pyx_t_6) {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":877
 * 
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num             # <<<<<<<<<<<<<<
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_child->type_num); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 877, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":878
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      __pyx_t_6 = (((__pyx_v_end - __pyx_v_f) < 5) != 0);
      if (unlikely(__pyx_t_6)) {

        /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 879, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(2, 879, __pyx_L1_error)

        /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":878
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":882
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_BYTE); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 882, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 882, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 882, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 98;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":883
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"             # <<<<<<<<<<<<<<
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UBYTE); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 883, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 883, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 66;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":884
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"             # <<<<<<<<<<<<<<
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_SHORT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 884, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x68;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":885
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"             # <<<<<<<<<<<<<<
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_USHORT); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 885, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 885, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 72;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":886
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_INT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 886, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 886, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x69;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":887
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UINT); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 887, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 887, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 73;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":888
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 888, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x6C;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":889
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 889, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 76;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":890
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGLONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 890, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 890, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x71;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":891
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONGLONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 891, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 891, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 81;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":892
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"             # <<<<<<<<<<<<<<
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_FLOAT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 892, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x66;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":893
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_DOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x64;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":894
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"             # <<<<<<<<<<<<<<
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 894, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 894, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x67;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":895
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf             # <<<<<<<<<<<<<<
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CFLOAT); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 895, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 895, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x66;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":896
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd             # <<<<<<<<<<<<<<
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 896, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 896, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x64;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":897
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg             # <<<<<<<<<<<<<<
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CLONGDOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 897, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 897, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x67;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":898
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_OBJECT); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 898, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(2, 898, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (likely(__pyx_t_6)) {
        (__pyx_v_f[0]) = 79;
        goto __pyx_L15;
      }

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":900
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *             f += 1
 *         else:
 */
      /*else*/ {
        __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 900, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 900, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(2, 900, __pyx_L1_error)
      }
      __pyx_L15:;

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":901
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *             f += 1             # <<<<<<<<<<<<<<
 *         else:
 *             # Cython ignores struct boundary information ("T{...}"),
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":876
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
      goto __pyx_L13;
    }

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":905
 *             # Cython ignores struct boundary information ("T{...}"),
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)             # <<<<<<<<<<<<<<
 *     return f
 * 
 */
    /*else*/ {
      __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_child, __pyx_v_f, __pyx_v_end, __pyx_v_offset); if (unlikely(__pyx_t_9 == ((char *)NULL))) __PYX_ERR(2, 905, __pyx_L1_error)
      __pyx_v_f = __pyx_t_9;
    }
    __pyx_L13:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":850
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":906
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":841
 *         return ()
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("numpy._util_dtypestring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_childname);
  __Pyx_XDECREF(__pyx_v_new_offset);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1021
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1022
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1023
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1021
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1025
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1026
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1027
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1028
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1027
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1029
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1025
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1033
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1035
 * cdef inline int import_array() except -1:
 *     try:
 *         _import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1035, __pyx_L3_error)

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1036
 *     try:
 *         _import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1036, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1037
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1037, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1037, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1034
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1033
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1039
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1041
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1041, __pyx_L3_error)

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1042
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1042, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1043
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1043, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1043, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1040
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1039
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1045
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1047
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 1047, __pyx_L3_error)

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1048
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 1048, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1049
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 1049, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 1049, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1046
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1045
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":13
 * 
 * 
 * cdef inline np.int64_t imax(np.int64_t i0, np.int64_t i1) nogil:             # <<<<<<<<<<<<<<
 *     if i0 > i1: return i0
 *     return i1
 */

static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_imax(__pyx_t_5numpy_int64_t __pyx_v_i0, __pyx_t_5numpy_int64_t __pyx_v_i1) {
  __pyx_t_5numpy_int64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/utilities/lib/fp_utils.pxd":14
 * 
 * cdef inline np.int64_t imax(np.int64_t i0, np.int64_t i1) nogil:
 *     if i0 > i1: return i0             # <<<<<<<<<<<<<<
 *     return i1
 * 
 */
  __pyx_t_1 = ((__pyx_v_i0 > __pyx_v_i1) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_i0;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":15
 * cdef inline np.int64_t imax(np.int64_t i0, np.int64_t i1) nogil:
 *     if i0 > i1: return i0
 *     return i1             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.float64_t fmax(np.float64_t f0, np.float64_t f1) nogil:
 */
  __pyx_r = __pyx_v_i1;
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":13
 * 
 * 
 * cdef inline np.int64_t imax(np.int64_t i0, np.int64_t i1) nogil:             # <<<<<<<<<<<<<<
 *     if i0 > i1: return i0
 *     return i1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":17
 *     return i1
 * 
 * cdef inline np.float64_t fmax(np.float64_t f0, np.float64_t f1) nogil:             # <<<<<<<<<<<<<<
 *     if f0 > f1: return f0
 *     return f1
 */

static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_fmax(__pyx_t_5numpy_float64_t __pyx_v_f0, __pyx_t_5numpy_float64_t __pyx_v_f1) {
  __pyx_t_5numpy_float64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/utilities/lib/fp_utils.pxd":18
 * 
 * cdef inline np.float64_t fmax(np.float64_t f0, np.float64_t f1) nogil:
 *     if f0 > f1: return f0             # <<<<<<<<<<<<<<
 *     return f1
 * 
 */
  __pyx_t_1 = ((__pyx_v_f0 > __pyx_v_f1) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_f0;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":19
 * cdef inline np.float64_t fmax(np.float64_t f0, np.float64_t f1) nogil:
 *     if f0 > f1: return f0
 *     return f1             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.int64_t imin(np.int64_t i0, np.int64_t i1) nogil:
 */
  __pyx_r = __pyx_v_f1;
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":17
 *     return i1
 * 
 * cdef inline np.float64_t fmax(np.float64_t f0, np.float64_t f1) nogil:             # <<<<<<<<<<<<<<
 *     if f0 > f1: return f0
 *     return f1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":21
 *     return f1
 * 
 * cdef inline np.int64_t imin(np.int64_t i0, np.int64_t i1) nogil:             # <<<<<<<<<<<<<<
 *     if i0 < i1: return i0
 *     return i1
 */

static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_imin(__pyx_t_5numpy_int64_t __pyx_v_i0, __pyx_t_5numpy_int64_t __pyx_v_i1) {
  __pyx_t_5numpy_int64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/utilities/lib/fp_utils.pxd":22
 * 
 * cdef inline np.int64_t imin(np.int64_t i0, np.int64_t i1) nogil:
 *     if i0 < i1: return i0             # <<<<<<<<<<<<<<
 *     return i1
 * 
 */
  __pyx_t_1 = ((__pyx_v_i0 < __pyx_v_i1) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_i0;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":23
 * cdef inline np.int64_t imin(np.int64_t i0, np.int64_t i1) nogil:
 *     if i0 < i1: return i0
 *     return i1             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.float64_t fmin(np.float64_t f0, np.float64_t f1) nogil:
 */
  __pyx_r = __pyx_v_i1;
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":21
 *     return f1
 * 
 * cdef inline np.int64_t imin(np.int64_t i0, np.int64_t i1) nogil:             # <<<<<<<<<<<<<<
 *     if i0 < i1: return i0
 *     return i1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":25
 *     return i1
 * 
 * cdef inline np.float64_t fmin(np.float64_t f0, np.float64_t f1) nogil:             # <<<<<<<<<<<<<<
 *     if f0 < f1: return f0
 *     return f1
 */

static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_fmin(__pyx_t_5numpy_float64_t __pyx_v_f0, __pyx_t_5numpy_float64_t __pyx_v_f1) {
  __pyx_t_5numpy_float64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/utilities/lib/fp_utils.pxd":26
 * 
 * cdef inline np.float64_t fmin(np.float64_t f0, np.float64_t f1) nogil:
 *     if f0 < f1: return f0             # <<<<<<<<<<<<<<
 *     return f1
 * 
 */
  __pyx_t_1 = ((__pyx_v_f0 < __pyx_v_f1) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_f0;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":27
 * cdef inline np.float64_t fmin(np.float64_t f0, np.float64_t f1) nogil:
 *     if f0 < f1: return f0
 *     return f1             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.float64_t fabs(np.float64_t f0) nogil:
 */
  __pyx_r = __pyx_v_f1;
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":25
 *     return i1
 * 
 * cdef inline np.float64_t fmin(np.float64_t f0, np.float64_t f1) nogil:             # <<<<<<<<<<<<<<
 *     if f0 < f1: return f0
 *     return f1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":29
 *     return f1
 * 
 * cdef inline np.float64_t fabs(np.float64_t f0) nogil:             # <<<<<<<<<<<<<<
 *     if f0 < 0.0: return -f0
 *     return f0
 */

static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_fabs(__pyx_t_5numpy_float64_t __pyx_v_f0) {
  __pyx_t_5numpy_float64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/utilities/lib/fp_utils.pxd":30
 * 
 * cdef inline np.float64_t fabs(np.float64_t f0) nogil:
 *     if f0 < 0.0: return -f0             # <<<<<<<<<<<<<<
 *     return f0
 * 
 */
  __pyx_t_1 = ((__pyx_v_f0 < 0.0) != 0);
  if (__pyx_t_1) {
    __pyx_r = (-__pyx_v_f0);
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":31
 * cdef inline np.float64_t fabs(np.float64_t f0) nogil:
 *     if f0 < 0.0: return -f0
 *     return f0             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.int64_t iclip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:
 */
  __pyx_r = __pyx_v_f0;
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":29
 *     return f1
 * 
 * cdef inline np.float64_t fabs(np.float64_t f0) nogil:             # <<<<<<<<<<<<<<
 *     if f0 < 0.0: return -f0
 *     return f0
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":33
 *     return f0
 * 
 * cdef inline np.int64_t iclip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:             # <<<<<<<<<<<<<<
 *     if i < a: return a
 *     if i > b: return b
 */

static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_iclip(__pyx_t_5numpy_int64_t __pyx_v_i, __pyx_t_5numpy_int64_t __pyx_v_a, __pyx_t_5numpy_int64_t __pyx_v_b) {
  __pyx_t_5numpy_int64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/utilities/lib/fp_utils.pxd":34
 * 
 * cdef inline np.int64_t iclip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:
 *     if i < a: return a             # <<<<<<<<<<<<<<
 *     if i > b: return b
 *     return i
 */
  __pyx_t_1 = ((__pyx_v_i < __pyx_v_a) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_a;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":35
 * cdef inline np.int64_t iclip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:
 *     if i < a: return a
 *     if i > b: return b             # <<<<<<<<<<<<<<
 *     return i
 * 
 */
  __pyx_t_1 = ((__pyx_v_i > __pyx_v_b) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_b;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":36
 *     if i < a: return a
 *     if i > b: return b
 *     return i             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.int64_t i64clip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:
 */
  __pyx_r = __pyx_v_i;
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":33
 *     return f0
 * 
 * cdef inline np.int64_t iclip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:             # <<<<<<<<<<<<<<
 *     if i < a: return a
 *     if i > b: return b
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":38
 *     return i
 * 
 * cdef inline np.int64_t i64clip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:             # <<<<<<<<<<<<<<
 *     if i < a: return a
 *     if i > b: return b
 */

static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_i64clip(__pyx_t_5numpy_int64_t __pyx_v_i, __pyx_t_5numpy_int64_t __pyx_v_a, __pyx_t_5numpy_int64_t __pyx_v_b) {
  __pyx_t_5numpy_int64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/utilities/lib/fp_utils.pxd":39
 * 
 * cdef inline np.int64_t i64clip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:
 *     if i < a: return a             # <<<<<<<<<<<<<<
 *     if i > b: return b
 *     return i
 */
  __pyx_t_1 = ((__pyx_v_i < __pyx_v_a) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_a;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":40
 * cdef inline np.int64_t i64clip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:
 *     if i < a: return a
 *     if i > b: return b             # <<<<<<<<<<<<<<
 *     return i
 * 
 */
  __pyx_t_1 = ((__pyx_v_i > __pyx_v_b) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_b;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":41
 *     if i < a: return a
 *     if i > b: return b
 *     return i             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.float64_t fclip(np.float64_t f,
 */
  __pyx_r = __pyx_v_i;
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":38
 *     return i
 * 
 * cdef inline np.int64_t i64clip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:             # <<<<<<<<<<<<<<
 *     if i < a: return a
 *     if i > b: return b
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":43
 *     return i
 * 
 * cdef inline np.float64_t fclip(np.float64_t f,             # <<<<<<<<<<<<<<
 *                       np.float64_t a, np.float64_t b) nogil:
 *     return fmin(fmax(f, a), b)
 */

static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_fclip(__pyx_t_5numpy_float64_t __pyx_v_f, __pyx_t_5numpy_float64_t __pyx_v_a, __pyx_t_5numpy_float64_t __pyx_v_b) {
  __pyx_t_5numpy_float64_t __pyx_r;

  /* "yt/utilities/lib/fp_utils.pxd":45
 * cdef inline np.float64_t fclip(np.float64_t f,
 *                       np.float64_t a, np.float64_t b) nogil:
 *     return fmin(fmax(f, a), b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.int64_t i64max(np.int64_t i0, np.int64_t i1) nogil:
 */
  __pyx_r = __pyx_f_2yt_9utilities_3lib_8fp_utils_fmin(__pyx_f_2yt_9utilities_3lib_8fp_utils_fmax(__pyx_v_f, __pyx_v_a), __pyx_v_b);
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":43
 *     return i
 * 
 * cdef inline np.float64_t fclip(np.float64_t f,             # <<<<<<<<<<<<<<
 *                       np.float64_t a, np.float64_t b) nogil:
 *     return fmin(fmax(f, a), b)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":47
 *     return fmin(fmax(f, a), b)
 * 
 * cdef inline np.int64_t i64max(np.int64_t i0, np.int64_t i1) nogil:             # <<<<<<<<<<<<<<
 *     if i0 > i1: return i0
 *     return i1
 */

static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_i64max(__pyx_t_5numpy_int64_t __pyx_v_i0, __pyx_t_5numpy_int64_t __pyx_v_i1) {
  __pyx_t_5numpy_int64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/utilities/lib/fp_utils.pxd":48
 * 
 * cdef inline np.int64_t i64max(np.int64_t i0, np.int64_t i1) nogil:
 *     if i0 > i1: return i0             # <<<<<<<<<<<<<<
 *     return i1
 * 
 */
  __pyx_t_1 = ((__pyx_v_i0 > __pyx_v_i1) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_i0;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":49
 * cdef inline np.int64_t i64max(np.int64_t i0, np.int64_t i1) nogil:
 *     if i0 > i1: return i0
 *     return i1             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.int64_t i64min(np.int64_t i0, np.int64_t i1) nogil:
 */
  __pyx_r = __pyx_v_i1;
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":47
 *     return fmin(fmax(f, a), b)
 * 
 * cdef inline np.int64_t i64max(np.int64_t i0, np.int64_t i1) nogil:             # <<<<<<<<<<<<<<
 *     if i0 > i1: return i0
 *     return i1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":51
 *     return i1
 * 
 * cdef inline np.int64_t i64min(np.int64_t i0, np.int64_t i1) nogil:             # <<<<<<<<<<<<<<
 *     if i0 < i1: return i0
 *     return i1
 */

static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_8fp_utils_i64min(__pyx_t_5numpy_int64_t __pyx_v_i0, __pyx_t_5numpy_int64_t __pyx_v_i1) {
  __pyx_t_5numpy_int64_t __pyx_r;
  int __pyx_t_1;

  /* "yt/utilities/lib/fp_utils.pxd":52
 * 
 * cdef inline np.int64_t i64min(np.int64_t i0, np.int64_t i1) nogil:
 *     if i0 < i1: return i0             # <<<<<<<<<<<<<<
 *     return i1
 * 
 */
  __pyx_t_1 = ((__pyx_v_i0 < __pyx_v_i1) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_i0;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/fp_utils.pxd":53
 * cdef inline np.int64_t i64min(np.int64_t i0, np.int64_t i1) nogil:
 *     if i0 < i1: return i0
 *     return i1             # <<<<<<<<<<<<<<
 * 
 * cdef inline _ensure_code(arr):
 */
  __pyx_r = __pyx_v_i1;
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":51
 *     return i1
 * 
 * cdef inline np.int64_t i64min(np.int64_t i0, np.int64_t i1) nogil:             # <<<<<<<<<<<<<<
 *     if i0 < i1: return i0
 *     return i1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "yt/utilities/lib/fp_utils.pxd":55
 *     return i1
 * 
 * cdef inline _ensure_code(arr):             # <<<<<<<<<<<<<<
 *     if hasattr(arr, "units"):
 *         if "code_length" == str(arr.units):
 */

static CYTHON_INLINE PyObject *__pyx_f_2yt_9utilities_3lib_8fp_utils__ensure_code(PyObject *__pyx_v_arr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_ensure_code", 0);

  /* "yt/utilities/lib/fp_utils.pxd":56
 * 
 * cdef inline _ensure_code(arr):
 *     if hasattr(arr, "units"):             # <<<<<<<<<<<<<<
 *         if "code_length" == str(arr.units):
 *             return arr
 */
  __pyx_t_1 = __Pyx_HasAttr(__pyx_v_arr, __pyx_n_s_units); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(3, 56, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "yt/utilities/lib/fp_utils.pxd":57
 * cdef inline _ensure_code(arr):
 *     if hasattr(arr, "units"):
 *         if "code_length" == str(arr.units):             # <<<<<<<<<<<<<<
 *             return arr
 *         arr.convert_to_units("code_length")
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_arr, __pyx_n_s_units); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 57, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_n_s_code_length, __pyx_t_4, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(3, 57, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_2) {

      /* "yt/utilities/lib/fp_utils.pxd":58
 *     if hasattr(arr, "units"):
 *         if "code_length" == str(arr.units):
 *             return arr             # <<<<<<<<<<<<<<
 *         arr.convert_to_units("code_length")
 *     return arr
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_arr);
      __pyx_r = __pyx_v_arr;
      goto __pyx_L0;

      /* "yt/utilities/lib/fp_utils.pxd":57
 * cdef inline _ensure_code(arr):
 *     if hasattr(arr, "units"):
 *         if "code_length" == str(arr.units):             # <<<<<<<<<<<<<<
 *             return arr
 *         arr.convert_to_units("code_length")
 */
    }

    /* "yt/utilities/lib/fp_utils.pxd":59
 *         if "code_length" == str(arr.units):
 *             return arr
 *         arr.convert_to_units("code_length")             # <<<<<<<<<<<<<<
 *     return arr
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_arr, __pyx_n_s_convert_to_units); if (unlikely(!__pyx_t_3)) __PYX_ERR(3, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_n_s_code_length) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_n_s_code_length);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(3, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "yt/utilities/lib/fp_utils.pxd":56
 * 
 * cdef inline _ensure_code(arr):
 *     if hasattr(arr, "units"):             # <<<<<<<<<<<<<<
 *         if "code_length" == str(arr.units):
 *             return arr
 */
  }

  /* "yt/utilities/lib/fp_utils.pxd":60
 *             return arr
 *         arr.convert_to_units("code_length")
 *     return arr             # <<<<<<<<<<<<<<
 * 
 * ctypedef fused any_float:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_arr);
  __pyx_r = __pyx_v_arr;
  goto __pyx_L0;

  /* "yt/utilities/lib/fp_utils.pxd":55
 *     return i1
 * 
 * cdef inline _ensure_code(arr):             # <<<<<<<<<<<<<<
 *     if hasattr(arr, "units"):
 *         if "code_length" == str(arr.units):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("yt.utilities.lib.fp_utils._ensure_code", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/geometry_utils.pxd":24
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.int64_t ifrexp(np.float64_t x, np.int64_t *e):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t m
 *     cdef int e0 = 0
 */

static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_ifrexp(__pyx_t_5numpy_float64_t __pyx_v_x, __pyx_t_5numpy_int64_t *__pyx_v_e) {
  __pyx_t_5numpy_float64_t __pyx_v_m;
  int __pyx_v_e0;
  __pyx_t_5numpy_int64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ifrexp", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":26
 * cdef inline np.int64_t ifrexp(np.float64_t x, np.int64_t *e):
 *     cdef np.float64_t m
 *     cdef int e0 = 0             # <<<<<<<<<<<<<<
 *     m = frexp(x,&e0)
 *     e[0] = <np.int64_t>e0
 */
  __pyx_v_e0 = 0;

  /* "yt/utilities/lib/geometry_utils.pxd":27
 *     cdef np.float64_t m
 *     cdef int e0 = 0
 *     m = frexp(x,&e0)             # <<<<<<<<<<<<<<
 *     e[0] = <np.int64_t>e0
 *     return <np.int64_t>ldexp(m,<int>DBL_MANT_DIG)
 */
  __pyx_v_m = frexp(__pyx_v_x, (&__pyx_v_e0));

  /* "yt/utilities/lib/geometry_utils.pxd":28
 *     cdef int e0 = 0
 *     m = frexp(x,&e0)
 *     e[0] = <np.int64_t>e0             # <<<<<<<<<<<<<<
 *     return <np.int64_t>ldexp(m,<int>DBL_MANT_DIG)
 * 
 */
  (__pyx_v_e[0]) = ((__pyx_t_5numpy_int64_t)__pyx_v_e0);

  /* "yt/utilities/lib/geometry_utils.pxd":29
 *     m = frexp(x,&e0)
 *     e[0] = <np.int64_t>e0
 *     return <np.int64_t>ldexp(m,<int>DBL_MANT_DIG)             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  __pyx_r = ((__pyx_t_5numpy_int64_t)ldexp(__pyx_v_m, ((int)DBL_MANT_DIG)));
  goto __pyx_L0;

  /* "yt/utilities/lib/geometry_utils.pxd":24
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.int64_t ifrexp(np.float64_t x, np.int64_t *e):             # <<<<<<<<<<<<<<
 *     cdef np.float64_t m
 *     cdef int e0 = 0
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/geometry_utils.pxd":34
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.int64_t msdb(np.int64_t a, np.int64_t b):             # <<<<<<<<<<<<<<
 *     """Get the most significant differing bit between a and b."""
 *     cdef np.int64_t c, ndx
 */

static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_msdb(__pyx_t_5numpy_int64_t __pyx_v_a, __pyx_t_5numpy_int64_t __pyx_v_b) {
  __pyx_t_5numpy_int64_t __pyx_v_c;
  __pyx_t_5numpy_int64_t __pyx_v_ndx;
  __pyx_t_5numpy_int64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("msdb", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":37
 *     """Get the most significant differing bit between a and b."""
 *     cdef np.int64_t c, ndx
 *     c = a ^ b             # <<<<<<<<<<<<<<
 *     ndx = 0
 *     while (0 < c):
 */
  __pyx_v_c = (__pyx_v_a ^ __pyx_v_b);

  /* "yt/utilities/lib/geometry_utils.pxd":38
 *     cdef np.int64_t c, ndx
 *     c = a ^ b
 *     ndx = 0             # <<<<<<<<<<<<<<
 *     while (0 < c):
 *         c = (c >> 1)
 */
  __pyx_v_ndx = 0;

  /* "yt/utilities/lib/geometry_utils.pxd":39
 *     c = a ^ b
 *     ndx = 0
 *     while (0 < c):             # <<<<<<<<<<<<<<
 *         c = (c >> 1)
 *         ndx+=1
 */
  while (1) {
    __pyx_t_1 = ((0 < __pyx_v_c) != 0);
    if (!__pyx_t_1) break;

    /* "yt/utilities/lib/geometry_utils.pxd":40
 *     ndx = 0
 *     while (0 < c):
 *         c = (c >> 1)             # <<<<<<<<<<<<<<
 *         ndx+=1
 *     return ndx
 */
    __pyx_v_c = (__pyx_v_c >> 1);

    /* "yt/utilities/lib/geometry_utils.pxd":41
 *     while (0 < c):
 *         c = (c >> 1)
 *         ndx+=1             # <<<<<<<<<<<<<<
 *     return ndx
 * 
 */
    __pyx_v_ndx = (__pyx_v_ndx + 1);
  }

  /* "yt/utilities/lib/geometry_utils.pxd":42
 *         c = (c >> 1)
 *         ndx+=1
 *     return ndx             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  __pyx_r = __pyx_v_ndx;
  goto __pyx_L0;

  /* "yt/utilities/lib/geometry_utils.pxd":34
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.int64_t msdb(np.int64_t a, np.int64_t b):             # <<<<<<<<<<<<<<
 *     """Get the most significant differing bit between a and b."""
 *     cdef np.int64_t c, ndx
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/geometry_utils.pxd":47
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.int64_t xor_msb(np.float64_t a, np.float64_t b):             # <<<<<<<<<<<<<<
 *     """Get the exponent of the highest differing bit between a and b"""
 *     # Get mantissa and exponents for each number
 */

static CYTHON_INLINE __pyx_t_5numpy_int64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_xor_msb(__pyx_t_5numpy_float64_t __pyx_v_a, __pyx_t_5numpy_float64_t __pyx_v_b) {
  __pyx_t_5numpy_int64_t __pyx_v_a_m;
  __pyx_t_5numpy_int64_t __pyx_v_a_e;
  __pyx_t_5numpy_int64_t __pyx_v_b_m;
  __pyx_t_5numpy_int64_t __pyx_v_b_e;
  __pyx_t_5numpy_int64_t __pyx_v_x;
  __pyx_t_5numpy_int64_t __pyx_v_y;
  __pyx_t_5numpy_int64_t __pyx_v_z;
  __pyx_t_5numpy_int64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("xor_msb", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":51
 *     # Get mantissa and exponents for each number
 *     cdef np.int64_t a_m, a_e, b_m, b_e, x, y, z
 *     b_e = 0             # <<<<<<<<<<<<<<
 *     a_e = 0
 *     a_m = ifrexp(a,&a_e)
 */
  __pyx_v_b_e = 0;

  /* "yt/utilities/lib/geometry_utils.pxd":52
 *     cdef np.int64_t a_m, a_e, b_m, b_e, x, y, z
 *     b_e = 0
 *     a_e = 0             # <<<<<<<<<<<<<<
 *     a_m = ifrexp(a,&a_e)
 *     b_m = ifrexp(b,&b_e)
 */
  __pyx_v_a_e = 0;

  /* "yt/utilities/lib/geometry_utils.pxd":53
 *     b_e = 0
 *     a_e = 0
 *     a_m = ifrexp(a,&a_e)             # <<<<<<<<<<<<<<
 *     b_m = ifrexp(b,&b_e)
 *     x = <np.int64_t> ((a_e+1)*DBL_MANT_DIG)
 */
  __pyx_v_a_m = __pyx_f_2yt_9utilities_3lib_14geometry_utils_ifrexp(__pyx_v_a, (&__pyx_v_a_e));

  /* "yt/utilities/lib/geometry_utils.pxd":54
 *     a_e = 0
 *     a_m = ifrexp(a,&a_e)
 *     b_m = ifrexp(b,&b_e)             # <<<<<<<<<<<<<<
 *     x = <np.int64_t> ((a_e+1)*DBL_MANT_DIG)
 *     y = <np.int64_t> ((b_e+1)*DBL_MANT_DIG)
 */
  __pyx_v_b_m = __pyx_f_2yt_9utilities_3lib_14geometry_utils_ifrexp(__pyx_v_b, (&__pyx_v_b_e));

  /* "yt/utilities/lib/geometry_utils.pxd":55
 *     a_m = ifrexp(a,&a_e)
 *     b_m = ifrexp(b,&b_e)
 *     x = <np.int64_t> ((a_e+1)*DBL_MANT_DIG)             # <<<<<<<<<<<<<<
 *     y = <np.int64_t> ((b_e+1)*DBL_MANT_DIG)
 *     # Compare mantissa if exponents equal
 */
  __pyx_v_x = ((__pyx_t_5numpy_int64_t)((__pyx_v_a_e + 1) * DBL_MANT_DIG));

  /* "yt/utilities/lib/geometry_utils.pxd":56
 *     b_m = ifrexp(b,&b_e)
 *     x = <np.int64_t> ((a_e+1)*DBL_MANT_DIG)
 *     y = <np.int64_t> ((b_e+1)*DBL_MANT_DIG)             # <<<<<<<<<<<<<<
 *     # Compare mantissa if exponents equal
 *     if x == y:
 */
  __pyx_v_y = ((__pyx_t_5numpy_int64_t)((__pyx_v_b_e + 1) * DBL_MANT_DIG));

  /* "yt/utilities/lib/geometry_utils.pxd":58
 *     y = <np.int64_t> ((b_e+1)*DBL_MANT_DIG)
 *     # Compare mantissa if exponents equal
 *     if x == y:             # <<<<<<<<<<<<<<
 *         if a_m == b_m: return 0
 *         z = msdb(a_m,b_m)
 */
  __pyx_t_1 = ((__pyx_v_x == __pyx_v_y) != 0);
  if (__pyx_t_1) {

    /* "yt/utilities/lib/geometry_utils.pxd":59
 *     # Compare mantissa if exponents equal
 *     if x == y:
 *         if a_m == b_m: return 0             # <<<<<<<<<<<<<<
 *         z = msdb(a_m,b_m)
 *         #if 1: return z
 */
    __pyx_t_1 = ((__pyx_v_a_m == __pyx_v_b_m) != 0);
    if (__pyx_t_1) {
      __pyx_r = 0;
      goto __pyx_L0;
    }

    /* "yt/utilities/lib/geometry_utils.pxd":60
 *     if x == y:
 *         if a_m == b_m: return 0
 *         z = msdb(a_m,b_m)             # <<<<<<<<<<<<<<
 *         #if 1: return z
 *         x = x - z
 */
    __pyx_v_z = __pyx_f_2yt_9utilities_3lib_14geometry_utils_msdb(__pyx_v_a_m, __pyx_v_b_m);

    /* "yt/utilities/lib/geometry_utils.pxd":62
 *         z = msdb(a_m,b_m)
 *         #if 1: return z
 *         x = x - z             # <<<<<<<<<<<<<<
 *         return x-1 # required so that xor_msb(0.0,1.0)!=xor_msb(1.0,1.0)
 *     # Otherwise return largest exponent
 */
    __pyx_v_x = (__pyx_v_x - __pyx_v_z);

    /* "yt/utilities/lib/geometry_utils.pxd":63
 *         #if 1: return z
 *         x = x - z
 *         return x-1 # required so that xor_msb(0.0,1.0)!=xor_msb(1.0,1.0)             # <<<<<<<<<<<<<<
 *     # Otherwise return largest exponent
 *     if y < x:
 */
    __pyx_r = (__pyx_v_x - 1);
    goto __pyx_L0;

    /* "yt/utilities/lib/geometry_utils.pxd":58
 *     y = <np.int64_t> ((b_e+1)*DBL_MANT_DIG)
 *     # Compare mantissa if exponents equal
 *     if x == y:             # <<<<<<<<<<<<<<
 *         if a_m == b_m: return 0
 *         z = msdb(a_m,b_m)
 */
  }

  /* "yt/utilities/lib/geometry_utils.pxd":65
 *         return x-1 # required so that xor_msb(0.0,1.0)!=xor_msb(1.0,1.0)
 *     # Otherwise return largest exponent
 *     if y < x:             # <<<<<<<<<<<<<<
 *         return x
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_y < __pyx_v_x) != 0);
  if (__pyx_t_1) {

    /* "yt/utilities/lib/geometry_utils.pxd":66
 *     # Otherwise return largest exponent
 *     if y < x:
 *         return x             # <<<<<<<<<<<<<<
 *     else:
 *         return y
 */
    __pyx_r = __pyx_v_x;
    goto __pyx_L0;

    /* "yt/utilities/lib/geometry_utils.pxd":65
 *         return x-1 # required so that xor_msb(0.0,1.0)!=xor_msb(1.0,1.0)
 *     # Otherwise return largest exponent
 *     if y < x:             # <<<<<<<<<<<<<<
 *         return x
 *     else:
 */
  }

  /* "yt/utilities/lib/geometry_utils.pxd":68
 *         return x
 *     else:
 *         return y             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  /*else*/ {
    __pyx_r = __pyx_v_y;
    goto __pyx_L0;
  }

  /* "yt/utilities/lib/geometry_utils.pxd":47
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.int64_t xor_msb(np.float64_t a, np.float64_t b):             # <<<<<<<<<<<<<<
 *     """Get the exponent of the highest differing bit between a and b"""
 *     # Get mantissa and exponents for each number
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/geometry_utils.pxd":73
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline int compare_floats_morton(np.float64_t p[3], np.float64_t q[3]):             # <<<<<<<<<<<<<<
 *     cdef int j, out, dim
 *     cdef np.int64_t x, y
 */

static CYTHON_INLINE int __pyx_f_2yt_9utilities_3lib_14geometry_utils_compare_floats_morton(__pyx_t_5numpy_float64_t *__pyx_v_p, __pyx_t_5numpy_float64_t *__pyx_v_q) {
  int __pyx_v_j;
  int __pyx_v_out;
  int __pyx_v_dim;
  __pyx_t_5numpy_int64_t __pyx_v_x;
  __pyx_t_5numpy_int64_t __pyx_v_y;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("compare_floats_morton", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":76
 *     cdef int j, out, dim
 *     cdef np.int64_t x, y
 *     x = -9999999999             # <<<<<<<<<<<<<<
 *     y = 0
 *     dim = 0
 */
  __pyx_v_x = -9999999999LL;

  /* "yt/utilities/lib/geometry_utils.pxd":77
 *     cdef np.int64_t x, y
 *     x = -9999999999
 *     y = 0             # <<<<<<<<<<<<<<
 *     dim = 0
 *     for j in range(3):#[::-1]:
 */
  __pyx_v_y = 0;

  /* "yt/utilities/lib/geometry_utils.pxd":78
 *     x = -9999999999
 *     y = 0
 *     dim = 0             # <<<<<<<<<<<<<<
 *     for j in range(3):#[::-1]:
 *         y = xor_msb(p[j],q[j])
 */
  __pyx_v_dim = 0;

  /* "yt/utilities/lib/geometry_utils.pxd":79
 *     y = 0
 *     dim = 0
 *     for j in range(3):#[::-1]:             # <<<<<<<<<<<<<<
 *         y = xor_msb(p[j],q[j])
 *         if x < y:
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_j = __pyx_t_1;

    /* "yt/utilities/lib/geometry_utils.pxd":80
 *     dim = 0
 *     for j in range(3):#[::-1]:
 *         y = xor_msb(p[j],q[j])             # <<<<<<<<<<<<<<
 *         if x < y:
 *            x = y
 */
    __pyx_v_y = __pyx_f_2yt_9utilities_3lib_14geometry_utils_xor_msb((__pyx_v_p[__pyx_v_j]), (__pyx_v_q[__pyx_v_j]));

    /* "yt/utilities/lib/geometry_utils.pxd":81
 *     for j in range(3):#[::-1]:
 *         y = xor_msb(p[j],q[j])
 *         if x < y:             # <<<<<<<<<<<<<<
 *            x = y
 *            dim = j
 */
    __pyx_t_2 = ((__pyx_v_x < __pyx_v_y) != 0);
    if (__pyx_t_2) {

      /* "yt/utilities/lib/geometry_utils.pxd":82
 *         y = xor_msb(p[j],q[j])
 *         if x < y:
 *            x = y             # <<<<<<<<<<<<<<
 *            dim = j
 *     if p[dim] < q[dim]:
 */
      __pyx_v_x = __pyx_v_y;

      /* "yt/utilities/lib/geometry_utils.pxd":83
 *         if x < y:
 *            x = y
 *            dim = j             # <<<<<<<<<<<<<<
 *     if p[dim] < q[dim]:
 *         out = 1
 */
      __pyx_v_dim = __pyx_v_j;

      /* "yt/utilities/lib/geometry_utils.pxd":81
 *     for j in range(3):#[::-1]:
 *         y = xor_msb(p[j],q[j])
 *         if x < y:             # <<<<<<<<<<<<<<
 *            x = y
 *            dim = j
 */
    }
  }

  /* "yt/utilities/lib/geometry_utils.pxd":84
 *            x = y
 *            dim = j
 *     if p[dim] < q[dim]:             # <<<<<<<<<<<<<<
 *         out = 1
 *     else:
 */
  __pyx_t_2 = (((__pyx_v_p[__pyx_v_dim]) < (__pyx_v_q[__pyx_v_dim])) != 0);
  if (__pyx_t_2) {

    /* "yt/utilities/lib/geometry_utils.pxd":85
 *            dim = j
 *     if p[dim] < q[dim]:
 *         out = 1             # <<<<<<<<<<<<<<
 *     else:
 *         out = 0
 */
    __pyx_v_out = 1;

    /* "yt/utilities/lib/geometry_utils.pxd":84
 *            x = y
 *            dim = j
 *     if p[dim] < q[dim]:             # <<<<<<<<<<<<<<
 *         out = 1
 *     else:
 */
    goto __pyx_L6;
  }

  /* "yt/utilities/lib/geometry_utils.pxd":87
 *         out = 1
 *     else:
 *         out = 0             # <<<<<<<<<<<<<<
 *     return out
 * 
 */
  /*else*/ {
    __pyx_v_out = 0;
  }
  __pyx_L6:;

  /* "yt/utilities/lib/geometry_utils.pxd":88
 *     else:
 *         out = 0
 *     return out             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  __pyx_r = __pyx_v_out;
  goto __pyx_L0;

  /* "yt/utilities/lib/geometry_utils.pxd":73
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline int compare_floats_morton(np.float64_t p[3], np.float64_t q[3]):             # <<<<<<<<<<<<<<
 *     cdef int j, out, dim
 *     cdef np.int64_t x, y
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/geometry_utils.pxd":93
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.float64_t euclidean_distance(np.float64_t[:] p, np.float64_t[:] q):             # <<<<<<<<<<<<<<
 *     cdef int j
 *     cdef np.float64_t d
 */

static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_euclidean_distance(__Pyx_memviewslice __pyx_v_p, __Pyx_memviewslice __pyx_v_q) {
  int __pyx_v_j;
  __pyx_t_5numpy_float64_t __pyx_v_d;
  __pyx_t_5numpy_float64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("euclidean_distance", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":96
 *     cdef int j
 *     cdef np.float64_t d
 *     d = 0.0             # <<<<<<<<<<<<<<
 *     for j in range(3):
 *         d+=(p[j]-q[j])**2
 */
  __pyx_v_d = 0.0;

  /* "yt/utilities/lib/geometry_utils.pxd":97
 *     cdef np.float64_t d
 *     d = 0.0
 *     for j in range(3):             # <<<<<<<<<<<<<<
 *         d+=(p[j]-q[j])**2
 *     return sqrt(d)
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_j = __pyx_t_1;

    /* "yt/utilities/lib/geometry_utils.pxd":98
 *     d = 0.0
 *     for j in range(3):
 *         d+=(p[j]-q[j])**2             # <<<<<<<<<<<<<<
 *     return sqrt(d)
 * 
 */
    __pyx_t_2 = __pyx_v_j;
    __pyx_t_3 = -1;
    if (__pyx_t_2 < 0) {
      __pyx_t_2 += __pyx_v_p.shape[0];
      if (unlikely(__pyx_t_2 < 0)) __pyx_t_3 = 0;
    } else if (unlikely(__pyx_t_2 >= __pyx_v_p.shape[0])) __pyx_t_3 = 0;
    if (unlikely(__pyx_t_3 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_3);
      __PYX_ERR(4, 98, __pyx_L1_error)
    }
    __pyx_t_4 = __pyx_v_j;
    __pyx_t_3 = -1;
    if (__pyx_t_4 < 0) {
      __pyx_t_4 += __pyx_v_q.shape[0];
      if (unlikely(__pyx_t_4 < 0)) __pyx_t_3 = 0;
    } else if (unlikely(__pyx_t_4 >= __pyx_v_q.shape[0])) __pyx_t_3 = 0;
    if (unlikely(__pyx_t_3 != -1)) {
      __Pyx_RaiseBufferIndexError(__pyx_t_3);
      __PYX_ERR(4, 98, __pyx_L1_error)
    }
    __pyx_v_d = (__pyx_v_d + pow(((*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_p.data + __pyx_t_2 * __pyx_v_p.strides[0]) ))) - (*((__pyx_t_5numpy_float64_t *) ( /* dim=0 */ (__pyx_v_q.data + __pyx_t_4 * __pyx_v_q.strides[0]) )))), 2.0));
  }

  /* "yt/utilities/lib/geometry_utils.pxd":99
 *     for j in range(3):
 *         d+=(p[j]-q[j])**2
 *     return sqrt(d)             # <<<<<<<<<<<<<<
 * 
 * # Todo: allow radius reported independently in each dimension for rectangular domain
 */
  __pyx_r = sqrt(__pyx_v_d);
  goto __pyx_L0;

  /* "yt/utilities/lib/geometry_utils.pxd":93
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.float64_t euclidean_distance(np.float64_t[:] p, np.float64_t[:] q):             # <<<<<<<<<<<<<<
 *     cdef int j
 *     cdef np.float64_t d
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("yt.utilities.lib.geometry_utils.euclidean_distance", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/geometry_utils.pxd":105
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.float64_t smallest_quadtree_box(np.float64_t p[3], np.float64_t q[3], np.int32_t order,             # <<<<<<<<<<<<<<
 *                                                np.float64_t DLE[3], np.float64_t DRE[3],
 *                                                np.float64_t *cx, np.float64_t *cy, np.float64_t *cz):
 */

static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_smallest_quadtree_box(__pyx_t_5numpy_float64_t *__pyx_v_p, __pyx_t_5numpy_float64_t *__pyx_v_q, __pyx_t_5numpy_int32_t __pyx_v_order, __pyx_t_5numpy_float64_t *__pyx_v_DLE, __pyx_t_5numpy_float64_t *__pyx_v_DRE, __pyx_t_5numpy_float64_t *__pyx_v_cx, __pyx_t_5numpy_float64_t *__pyx_v_cy, __pyx_t_5numpy_float64_t *__pyx_v_cz) {
  int __pyx_v_j;
  __pyx_t_5numpy_float64_t __pyx_v_c[3];
  __pyx_t_5numpy_uint64_t __pyx_v_pidx[3];
  __pyx_t_5numpy_uint64_t __pyx_v_pidx_next[3];
  __pyx_t_5numpy_uint64_t __pyx_v_qidx_next[3];
  __pyx_t_5numpy_float64_t __pyx_v_dds[3];
  __pyx_t_5numpy_float64_t __pyx_v_rad;
  int __pyx_v_lvl;
  int __pyx_v_done;
  __pyx_t_5numpy_float64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("smallest_quadtree_box", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":112
 *     cdef np.uint64_t pidx[3]
 *     # cdef np.uint64_t qidx[3]
 *     for j in range(3):             # <<<<<<<<<<<<<<
 *         pidx[j] = 0
 *         # qidx[j] = 0
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_j = __pyx_t_1;

    /* "yt/utilities/lib/geometry_utils.pxd":113
 *     # cdef np.uint64_t qidx[3]
 *     for j in range(3):
 *         pidx[j] = 0             # <<<<<<<<<<<<<<
 *         # qidx[j] = 0
 *     cdef np.uint64_t pidx_next[3]
 */
    (__pyx_v_pidx[__pyx_v_j]) = 0;
  }

  /* "yt/utilities/lib/geometry_utils.pxd":119
 *     cdef np.float64_t dds[3]
 *     cdef np.float64_t rad
 *     cdef int lvl = 0             # <<<<<<<<<<<<<<
 *     cdef int done = 0
 *     while not done:
 */
  __pyx_v_lvl = 0;

  /* "yt/utilities/lib/geometry_utils.pxd":120
 *     cdef np.float64_t rad
 *     cdef int lvl = 0
 *     cdef int done = 0             # <<<<<<<<<<<<<<
 *     while not done:
 *         if (lvl+1 >= order):
 */
  __pyx_v_done = 0;

  /* "yt/utilities/lib/geometry_utils.pxd":121
 *     cdef int lvl = 0
 *     cdef int done = 0
 *     while not done:             # <<<<<<<<<<<<<<
 *         if (lvl+1 >= order):
 *             done = 1
 */
  while (1) {
    __pyx_t_2 = ((!(__pyx_v_done != 0)) != 0);
    if (!__pyx_t_2) break;

    /* "yt/utilities/lib/geometry_utils.pxd":122
 *     cdef int done = 0
 *     while not done:
 *         if (lvl+1 >= order):             # <<<<<<<<<<<<<<
 *             done = 1
 *         for j in range(3):
 */
    __pyx_t_2 = (((__pyx_v_lvl + 1) >= __pyx_v_order) != 0);
    if (__pyx_t_2) {

      /* "yt/utilities/lib/geometry_utils.pxd":123
 *     while not done:
 *         if (lvl+1 >= order):
 *             done = 1             # <<<<<<<<<<<<<<
 *         for j in range(3):
 *             dds[j] = (DRE[j] - DLE[j])/(1 << (<int> lvl+1))
 */
      __pyx_v_done = 1;

      /* "yt/utilities/lib/geometry_utils.pxd":122
 *     cdef int done = 0
 *     while not done:
 *         if (lvl+1 >= order):             # <<<<<<<<<<<<<<
 *             done = 1
 *         for j in range(3):
 */
    }

    /* "yt/utilities/lib/geometry_utils.pxd":124
 *         if (lvl+1 >= order):
 *             done = 1
 *         for j in range(3):             # <<<<<<<<<<<<<<
 *             dds[j] = (DRE[j] - DLE[j])/(1 << (<int> lvl+1))
 *             pidx_next[j] = <np.uint64_t>((p[j] - DLE[j])/dds[j])
 */
    for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
      __pyx_v_j = __pyx_t_1;

      /* "yt/utilities/lib/geometry_utils.pxd":125
 *             done = 1
 *         for j in range(3):
 *             dds[j] = (DRE[j] - DLE[j])/(1 << (<int> lvl+1))             # <<<<<<<<<<<<<<
 *             pidx_next[j] = <np.uint64_t>((p[j] - DLE[j])/dds[j])
 *             qidx_next[j] = <np.uint64_t>((q[j] - DLE[j])/dds[j])
 */
      (__pyx_v_dds[__pyx_v_j]) = (((__pyx_v_DRE[__pyx_v_j]) - (__pyx_v_DLE[__pyx_v_j])) / (1 << (((int)__pyx_v_lvl) + 1)));

      /* "yt/utilities/lib/geometry_utils.pxd":126
 *         for j in range(3):
 *             dds[j] = (DRE[j] - DLE[j])/(1 << (<int> lvl+1))
 *             pidx_next[j] = <np.uint64_t>((p[j] - DLE[j])/dds[j])             # <<<<<<<<<<<<<<
 *             qidx_next[j] = <np.uint64_t>((q[j] - DLE[j])/dds[j])
 *         for j in range(3):
 */
      (__pyx_v_pidx_next[__pyx_v_j]) = ((__pyx_t_5numpy_uint64_t)(((__pyx_v_p[__pyx_v_j]) - (__pyx_v_DLE[__pyx_v_j])) / (__pyx_v_dds[__pyx_v_j])));

      /* "yt/utilities/lib/geometry_utils.pxd":127
 *             dds[j] = (DRE[j] - DLE[j])/(1 << (<int> lvl+1))
 *             pidx_next[j] = <np.uint64_t>((p[j] - DLE[j])/dds[j])
 *             qidx_next[j] = <np.uint64_t>((q[j] - DLE[j])/dds[j])             # <<<<<<<<<<<<<<
 *         for j in range(3):
 *             if pidx_next[j]!=qidx_next[j]:
 */
      (__pyx_v_qidx_next[__pyx_v_j]) = ((__pyx_t_5numpy_uint64_t)(((__pyx_v_q[__pyx_v_j]) - (__pyx_v_DLE[__pyx_v_j])) / (__pyx_v_dds[__pyx_v_j])));
    }

    /* "yt/utilities/lib/geometry_utils.pxd":128
 *             pidx_next[j] = <np.uint64_t>((p[j] - DLE[j])/dds[j])
 *             qidx_next[j] = <np.uint64_t>((q[j] - DLE[j])/dds[j])
 *         for j in range(3):             # <<<<<<<<<<<<<<
 *             if pidx_next[j]!=qidx_next[j]:
 *                 done = 1
 */
    for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
      __pyx_v_j = __pyx_t_1;

      /* "yt/utilities/lib/geometry_utils.pxd":129
 *             qidx_next[j] = <np.uint64_t>((q[j] - DLE[j])/dds[j])
 *         for j in range(3):
 *             if pidx_next[j]!=qidx_next[j]:             # <<<<<<<<<<<<<<
 *                 done = 1
 *                 break
 */
      __pyx_t_2 = (((__pyx_v_pidx_next[__pyx_v_j]) != (__pyx_v_qidx_next[__pyx_v_j])) != 0);
      if (__pyx_t_2) {

        /* "yt/utilities/lib/geometry_utils.pxd":130
 *         for j in range(3):
 *             if pidx_next[j]!=qidx_next[j]:
 *                 done = 1             # <<<<<<<<<<<<<<
 *                 break
 *         if not done:
 */
        __pyx_v_done = 1;

        /* "yt/utilities/lib/geometry_utils.pxd":131
 *             if pidx_next[j]!=qidx_next[j]:
 *                 done = 1
 *                 break             # <<<<<<<<<<<<<<
 *         if not done:
 *             for j in range(3):
 */
        goto __pyx_L11_break;

        /* "yt/utilities/lib/geometry_utils.pxd":129
 *             qidx_next[j] = <np.uint64_t>((q[j] - DLE[j])/dds[j])
 *         for j in range(3):
 *             if pidx_next[j]!=qidx_next[j]:             # <<<<<<<<<<<<<<
 *                 done = 1
 *                 break
 */
      }
    }
    __pyx_L11_break:;

    /* "yt/utilities/lib/geometry_utils.pxd":132
 *                 done = 1
 *                 break
 *         if not done:             # <<<<<<<<<<<<<<
 *             for j in range(3):
 *                 pidx[j] = pidx_next[j]
 */
    __pyx_t_2 = ((!(__pyx_v_done != 0)) != 0);
    if (__pyx_t_2) {

      /* "yt/utilities/lib/geometry_utils.pxd":133
 *                 break
 *         if not done:
 *             for j in range(3):             # <<<<<<<<<<<<<<
 *                 pidx[j] = pidx_next[j]
 *                 # qidx[j] = qidx_next[j]
 */
      for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
        __pyx_v_j = __pyx_t_1;

        /* "yt/utilities/lib/geometry_utils.pxd":134
 *         if not done:
 *             for j in range(3):
 *                 pidx[j] = pidx_next[j]             # <<<<<<<<<<<<<<
 *                 # qidx[j] = qidx_next[j]
 *             lvl+=1
 */
        (__pyx_v_pidx[__pyx_v_j]) = (__pyx_v_pidx_next[__pyx_v_j]);
      }

      /* "yt/utilities/lib/geometry_utils.pxd":136
 *                 pidx[j] = pidx_next[j]
 *                 # qidx[j] = qidx_next[j]
 *             lvl+=1             # <<<<<<<<<<<<<<
 *     rad = 0.0
 *     for j in range(3):
 */
      __pyx_v_lvl = (__pyx_v_lvl + 1);

      /* "yt/utilities/lib/geometry_utils.pxd":132
 *                 done = 1
 *                 break
 *         if not done:             # <<<<<<<<<<<<<<
 *             for j in range(3):
 *                 pidx[j] = pidx_next[j]
 */
    }
  }

  /* "yt/utilities/lib/geometry_utils.pxd":137
 *                 # qidx[j] = qidx_next[j]
 *             lvl+=1
 *     rad = 0.0             # <<<<<<<<<<<<<<
 *     for j in range(3):
 *         dds[j] = (DRE[j] - DLE[j])/(1 << lvl)
 */
  __pyx_v_rad = 0.0;

  /* "yt/utilities/lib/geometry_utils.pxd":138
 *             lvl+=1
 *     rad = 0.0
 *     for j in range(3):             # <<<<<<<<<<<<<<
 *         dds[j] = (DRE[j] - DLE[j])/(1 << lvl)
 *         c[j] = dds[j]*(<np.float64_t>pidx[j]+0.5)
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_j = __pyx_t_1;

    /* "yt/utilities/lib/geometry_utils.pxd":139
 *     rad = 0.0
 *     for j in range(3):
 *         dds[j] = (DRE[j] - DLE[j])/(1 << lvl)             # <<<<<<<<<<<<<<
 *         c[j] = dds[j]*(<np.float64_t>pidx[j]+0.5)
 *         rad+=((dds[j]/2.0)**2)
 */
    (__pyx_v_dds[__pyx_v_j]) = (((__pyx_v_DRE[__pyx_v_j]) - (__pyx_v_DLE[__pyx_v_j])) / (1 << __pyx_v_lvl));

    /* "yt/utilities/lib/geometry_utils.pxd":140
 *     for j in range(3):
 *         dds[j] = (DRE[j] - DLE[j])/(1 << lvl)
 *         c[j] = dds[j]*(<np.float64_t>pidx[j]+0.5)             # <<<<<<<<<<<<<<
 *         rad+=((dds[j]/2.0)**2)
 *     cx[0] = c[0]
 */
    (__pyx_v_c[__pyx_v_j]) = ((__pyx_v_dds[__pyx_v_j]) * (((__pyx_t_5numpy_float64_t)(__pyx_v_pidx[__pyx_v_j])) + 0.5));

    /* "yt/utilities/lib/geometry_utils.pxd":141
 *         dds[j] = (DRE[j] - DLE[j])/(1 << lvl)
 *         c[j] = dds[j]*(<np.float64_t>pidx[j]+0.5)
 *         rad+=((dds[j]/2.0)**2)             # <<<<<<<<<<<<<<
 *     cx[0] = c[0]
 *     cy[0] = c[1]
 */
    __pyx_v_rad = (__pyx_v_rad + pow(((__pyx_v_dds[__pyx_v_j]) / 2.0), 2.0));
  }

  /* "yt/utilities/lib/geometry_utils.pxd":142
 *         c[j] = dds[j]*(<np.float64_t>pidx[j]+0.5)
 *         rad+=((dds[j]/2.0)**2)
 *     cx[0] = c[0]             # <<<<<<<<<<<<<<
 *     cy[0] = c[1]
 *     cz[0] = c[2]
 */
  (__pyx_v_cx[0]) = (__pyx_v_c[0]);

  /* "yt/utilities/lib/geometry_utils.pxd":143
 *         rad+=((dds[j]/2.0)**2)
 *     cx[0] = c[0]
 *     cy[0] = c[1]             # <<<<<<<<<<<<<<
 *     cz[0] = c[2]
 *     return sqrt(rad)
 */
  (__pyx_v_cy[0]) = (__pyx_v_c[1]);

  /* "yt/utilities/lib/geometry_utils.pxd":144
 *     cx[0] = c[0]
 *     cy[0] = c[1]
 *     cz[0] = c[2]             # <<<<<<<<<<<<<<
 *     return sqrt(rad)
 * 
 */
  (__pyx_v_cz[0]) = (__pyx_v_c[2]);

  /* "yt/utilities/lib/geometry_utils.pxd":145
 *     cy[0] = c[1]
 *     cz[0] = c[2]
 *     return sqrt(rad)             # <<<<<<<<<<<<<<
 * 
 * #-----------------------------------------------------------------------------
 */
  __pyx_r = sqrt(__pyx_v_rad);
  goto __pyx_L0;

  /* "yt/utilities/lib/geometry_utils.pxd":105
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.float64_t smallest_quadtree_box(np.float64_t p[3], np.float64_t q[3], np.int32_t order,             # <<<<<<<<<<<<<<
 *                                                np.float64_t DLE[3], np.float64_t DRE[3],
 *                                                np.float64_t *cx, np.float64_t *cy, np.float64_t *cz):
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/geometry_utils.pxd":152
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.uint64_t spread_64bits_by3(np.uint64_t x):             # <<<<<<<<<<<<<<
 *     x=(x&(<np.uint64_t>0x00000000001FFFFF))
 *     x=(x|(x<<20))*(<np.uint64_t>0x000001FFC00003FF)
 */

static CYTHON_INLINE __pyx_t_5numpy_uint64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_spread_64bits_by3(__pyx_t_5numpy_uint64_t __pyx_v_x) {
  __pyx_t_5numpy_uint64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("spread_64bits_by3", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":153
 * @cython.wraparound(False)
 * cdef inline np.uint64_t spread_64bits_by3(np.uint64_t x):
 *     x=(x&(<np.uint64_t>0x00000000001FFFFF))             # <<<<<<<<<<<<<<
 *     x=(x|(x<<20))*(<np.uint64_t>0x000001FFC00003FF)
 * 
 */
  __pyx_v_x = (__pyx_v_x & ((__pyx_t_5numpy_uint64_t)0x00000000001FFFFF));

  /* "yt/utilities/lib/geometry_utils.pxd":154
 * cdef inline np.uint64_t spread_64bits_by3(np.uint64_t x):
 *     x=(x&(<np.uint64_t>0x00000000001FFFFF))
 *     x=(x|(x<<20))*(<np.uint64_t>0x000001FFC00003FF)             # <<<<<<<<<<<<<<
 * 
 * #-----------------------------------------------------------------------------
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x << 20)) * ((__pyx_t_5numpy_uint64_t)0x000001FFC00003FF));

  /* "yt/utilities/lib/geometry_utils.pxd":152
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.uint64_t spread_64bits_by3(np.uint64_t x):             # <<<<<<<<<<<<<<
 *     x=(x&(<np.uint64_t>0x00000000001FFFFF))
 *     x=(x|(x<<20))*(<np.uint64_t>0x000001FFC00003FF)
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/geometry_utils.pxd":161
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.uint64_t spread_64bits_by2(np.uint64_t x):             # <<<<<<<<<<<<<<
 *     # This magic comes from http://stackoverflow.com/questions/1024754/how-to-compute-a-3d-morton-number-interleave-the-bits-of-3-ints
 *     # Only reversible up to 2097151
 */

static CYTHON_INLINE __pyx_t_5numpy_uint64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_spread_64bits_by2(__pyx_t_5numpy_uint64_t __pyx_v_x) {
  __pyx_t_5numpy_uint64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("spread_64bits_by2", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":166
 *     # Select highest 21 bits (Required to be reversible to 21st bit)
 *     # x = ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---k jihg fedc ba98 7654 3210
 *     x=(x&(<np.uint64_t>0x00000000001FFFFF))             # <<<<<<<<<<<<<<
 *     # x = ---- ---- ---- ---- ---- ---k jihg fedc ba-- ---- ---- ---- ---- --98 7654 3210
 *     x=(x|(x<<20))&(<np.uint64_t>0x000001FFC00003FF)
 */
  __pyx_v_x = (__pyx_v_x & ((__pyx_t_5numpy_uint64_t)0x00000000001FFFFF));

  /* "yt/utilities/lib/geometry_utils.pxd":168
 *     x=(x&(<np.uint64_t>0x00000000001FFFFF))
 *     # x = ---- ---- ---- ---- ---- ---k jihg fedc ba-- ---- ---- ---- ---- --98 7654 3210
 *     x=(x|(x<<20))&(<np.uint64_t>0x000001FFC00003FF)             # <<<<<<<<<<<<<<
 *     # x = ---- ---- ---- -kji hgf- ---- ---- -edc ba-- ---- ---- 9876 5--- ---- ---4 3210
 *     x=(x|(x<<10))&(<np.uint64_t>0x0007E007C00F801F)
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x << 20)) & ((__pyx_t_5numpy_uint64_t)0x000001FFC00003FF));

  /* "yt/utilities/lib/geometry_utils.pxd":170
 *     x=(x|(x<<20))&(<np.uint64_t>0x000001FFC00003FF)
 *     # x = ---- ---- ---- -kji hgf- ---- ---- -edc ba-- ---- ---- 9876 5--- ---- ---4 3210
 *     x=(x|(x<<10))&(<np.uint64_t>0x0007E007C00F801F)             # <<<<<<<<<<<<<<
 *     # x = ---- ---- -kji h--- -gf- ---- -edc ---- ba-- ---- 987- ---6 5--- ---4 32-- --10
 *     x=(x|(x<<4))&(<np.uint64_t>0x00786070C0E181C3)
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x << 10)) & ((__pyx_t_5numpy_uint64_t)0x0007E007C00F801F));

  /* "yt/utilities/lib/geometry_utils.pxd":172
 *     x=(x|(x<<10))&(<np.uint64_t>0x0007E007C00F801F)
 *     # x = ---- ---- -kji h--- -gf- ---- -edc ---- ba-- ---- 987- ---6 5--- ---4 32-- --10
 *     x=(x|(x<<4))&(<np.uint64_t>0x00786070C0E181C3)             # <<<<<<<<<<<<<<
 *     # x = ---- ---k ji-- h--g --f- ---e d--c --b- -a-- --98 --7- -6-- 5--- -43- -2-- 1--0
 *     x=(x|(x<<2))&(<np.uint64_t>0x0199219243248649)
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x << 4)) & ((__pyx_t_5numpy_uint64_t)0x00786070C0E181C3));

  /* "yt/utilities/lib/geometry_utils.pxd":174
 *     x=(x|(x<<4))&(<np.uint64_t>0x00786070C0E181C3)
 *     # x = ---- ---k ji-- h--g --f- ---e d--c --b- -a-- --98 --7- -6-- 5--- -43- -2-- 1--0
 *     x=(x|(x<<2))&(<np.uint64_t>0x0199219243248649)             # <<<<<<<<<<<<<<
 *     # x = ---- -kj- -i-- h--g --f- -e-- d--c --b- -a-- 9--8 --7- -6-- 5--4 --3- -2-- 1--0
 *     x=(x|(x<<2))&(<np.uint64_t>0x0649249249249249)
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x << 2)) & ((__pyx_t_5numpy_uint64_t)0x0199219243248649));

  /* "yt/utilities/lib/geometry_utils.pxd":176
 *     x=(x|(x<<2))&(<np.uint64_t>0x0199219243248649)
 *     # x = ---- -kj- -i-- h--g --f- -e-- d--c --b- -a-- 9--8 --7- -6-- 5--4 --3- -2-- 1--0
 *     x=(x|(x<<2))&(<np.uint64_t>0x0649249249249249)             # <<<<<<<<<<<<<<
 *     # x = ---k --j- -i-- h--g --f- -e-- d--c --b- -a-- 9--8 --7- -6-- 5--4 --3- -2-- 1--0
 *     x=(x|(x<<2))&(<np.uint64_t>0x1249249249249249)
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x << 2)) & ((__pyx_t_5numpy_uint64_t)0x0649249249249249));

  /* "yt/utilities/lib/geometry_utils.pxd":178
 *     x=(x|(x<<2))&(<np.uint64_t>0x0649249249249249)
 *     # x = ---k --j- -i-- h--g --f- -e-- d--c --b- -a-- 9--8 --7- -6-- 5--4 --3- -2-- 1--0
 *     x=(x|(x<<2))&(<np.uint64_t>0x1249249249249249)             # <<<<<<<<<<<<<<
 *     return x
 * 
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x << 2)) & ((__pyx_t_5numpy_uint64_t)0x1249249249249249));

  /* "yt/utilities/lib/geometry_utils.pxd":179
 *     # x = ---k --j- -i-- h--g --f- -e-- d--c --b- -a-- 9--8 --7- -6-- 5--4 --3- -2-- 1--0
 *     x=(x|(x<<2))&(<np.uint64_t>0x1249249249249249)
 *     return x             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  __pyx_r = __pyx_v_x;
  goto __pyx_L0;

  /* "yt/utilities/lib/geometry_utils.pxd":161
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.uint64_t spread_64bits_by2(np.uint64_t x):             # <<<<<<<<<<<<<<
 *     # This magic comes from http://stackoverflow.com/questions/1024754/how-to-compute-a-3d-morton-number-interleave-the-bits-of-3-ints
 *     # Only reversible up to 2097151
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/geometry_utils.pxd":184
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.uint64_t compact_64bits_by2(np.uint64_t x):             # <<<<<<<<<<<<<<
 *     # Reversed magic
 *     x=x&(<np.uint64_t>0x1249249249249249)
 */

static CYTHON_INLINE __pyx_t_5numpy_uint64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_compact_64bits_by2(__pyx_t_5numpy_uint64_t __pyx_v_x) {
  __pyx_t_5numpy_uint64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compact_64bits_by2", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":186
 * cdef inline np.uint64_t compact_64bits_by2(np.uint64_t x):
 *     # Reversed magic
 *     x=x&(<np.uint64_t>0x1249249249249249)             # <<<<<<<<<<<<<<
 *     x=(x|(x>>2))&(<np.uint64_t>0x0649249249249249)
 *     x=(x|(x>>2))&(<np.uint64_t>0x0199219243248649)
 */
  __pyx_v_x = (__pyx_v_x & ((__pyx_t_5numpy_uint64_t)0x1249249249249249));

  /* "yt/utilities/lib/geometry_utils.pxd":187
 *     # Reversed magic
 *     x=x&(<np.uint64_t>0x1249249249249249)
 *     x=(x|(x>>2))&(<np.uint64_t>0x0649249249249249)             # <<<<<<<<<<<<<<
 *     x=(x|(x>>2))&(<np.uint64_t>0x0199219243248649)
 *     x=(x|(x>>2))&(<np.uint64_t>0x00786070C0E181C3)
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x >> 2)) & ((__pyx_t_5numpy_uint64_t)0x0649249249249249));

  /* "yt/utilities/lib/geometry_utils.pxd":188
 *     x=x&(<np.uint64_t>0x1249249249249249)
 *     x=(x|(x>>2))&(<np.uint64_t>0x0649249249249249)
 *     x=(x|(x>>2))&(<np.uint64_t>0x0199219243248649)             # <<<<<<<<<<<<<<
 *     x=(x|(x>>2))&(<np.uint64_t>0x00786070C0E181C3)
 *     x=(x|(x>>4))&(<np.uint64_t>0x0007E007C00F801F)
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x >> 2)) & ((__pyx_t_5numpy_uint64_t)0x0199219243248649));

  /* "yt/utilities/lib/geometry_utils.pxd":189
 *     x=(x|(x>>2))&(<np.uint64_t>0x0649249249249249)
 *     x=(x|(x>>2))&(<np.uint64_t>0x0199219243248649)
 *     x=(x|(x>>2))&(<np.uint64_t>0x00786070C0E181C3)             # <<<<<<<<<<<<<<
 *     x=(x|(x>>4))&(<np.uint64_t>0x0007E007C00F801F)
 *     x=(x|(x>>10))&(<np.uint64_t>0x000001FFC00003FF)
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x >> 2)) & ((__pyx_t_5numpy_uint64_t)0x00786070C0E181C3));

  /* "yt/utilities/lib/geometry_utils.pxd":190
 *     x=(x|(x>>2))&(<np.uint64_t>0x0199219243248649)
 *     x=(x|(x>>2))&(<np.uint64_t>0x00786070C0E181C3)
 *     x=(x|(x>>4))&(<np.uint64_t>0x0007E007C00F801F)             # <<<<<<<<<<<<<<
 *     x=(x|(x>>10))&(<np.uint64_t>0x000001FFC00003FF)
 *     x=(x|(x>>20))&(<np.uint64_t>0x00000000001FFFFF)
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x >> 4)) & ((__pyx_t_5numpy_uint64_t)0x0007E007C00F801F));

  /* "yt/utilities/lib/geometry_utils.pxd":191
 *     x=(x|(x>>2))&(<np.uint64_t>0x00786070C0E181C3)
 *     x=(x|(x>>4))&(<np.uint64_t>0x0007E007C00F801F)
 *     x=(x|(x>>10))&(<np.uint64_t>0x000001FFC00003FF)             # <<<<<<<<<<<<<<
 *     x=(x|(x>>20))&(<np.uint64_t>0x00000000001FFFFF)
 *     return x
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x >> 10)) & ((__pyx_t_5numpy_uint64_t)0x000001FFC00003FF));

  /* "yt/utilities/lib/geometry_utils.pxd":192
 *     x=(x|(x>>4))&(<np.uint64_t>0x0007E007C00F801F)
 *     x=(x|(x>>10))&(<np.uint64_t>0x000001FFC00003FF)
 *     x=(x|(x>>20))&(<np.uint64_t>0x00000000001FFFFF)             # <<<<<<<<<<<<<<
 *     return x
 * 
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x >> 20)) & ((__pyx_t_5numpy_uint64_t)0x00000000001FFFFF));

  /* "yt/utilities/lib/geometry_utils.pxd":193
 *     x=(x|(x>>10))&(<np.uint64_t>0x000001FFC00003FF)
 *     x=(x|(x>>20))&(<np.uint64_t>0x00000000001FFFFF)
 *     return x             # <<<<<<<<<<<<<<
 * 
 * #-----------------------------------------------------------------------------
 */
  __pyx_r = __pyx_v_x;
  goto __pyx_L0;

  /* "yt/utilities/lib/geometry_utils.pxd":184
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.uint64_t compact_64bits_by2(np.uint64_t x):             # <<<<<<<<<<<<<<
 *     # Reversed magic
 *     x=x&(<np.uint64_t>0x1249249249249249)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/geometry_utils.pxd":200
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.uint32_t spread_32bits_by2(np.uint32_t x):             # <<<<<<<<<<<<<<
 *     # Only reversible up to 1023
 *     # Select highest 10 bits (Required to be reversible to 10st bit)
 */

static CYTHON_INLINE __pyx_t_5numpy_uint32_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_spread_32bits_by2(__pyx_t_5numpy_uint32_t __pyx_v_x) {
  __pyx_t_5numpy_uint32_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("spread_32bits_by2", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":204
 *     # Select highest 10 bits (Required to be reversible to 10st bit)
 *     # x = ---- ---- ---- ---- ---- --98 7654 3210
 *     x=(x&(<np.uint32_t>0x000003FF))             # <<<<<<<<<<<<<<
 *     # x = ---- --98 ---- ---- ---- ---- 7654 3210
 *     x=(x|(x<<16))&(<np.uint32_t>0xFF0000FF)
 */
  __pyx_v_x = (__pyx_v_x & ((__pyx_t_5numpy_uint32_t)0x000003FF));

  /* "yt/utilities/lib/geometry_utils.pxd":206
 *     x=(x&(<np.uint32_t>0x000003FF))
 *     # x = ---- --98 ---- ---- ---- ---- 7654 3210
 *     x=(x|(x<<16))&(<np.uint32_t>0xFF0000FF)             # <<<<<<<<<<<<<<
 *     # x = ---- --98 ---- ---- 7654 ---- ---- 3210
 *     x=(x|(x<<8))&(<np.uint32_t>0x0300F00F)
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x << 16)) & ((__pyx_t_5numpy_uint32_t)0xFF0000FF));

  /* "yt/utilities/lib/geometry_utils.pxd":208
 *     x=(x|(x<<16))&(<np.uint32_t>0xFF0000FF)
 *     # x = ---- --98 ---- ---- 7654 ---- ---- 3210
 *     x=(x|(x<<8))&(<np.uint32_t>0x0300F00F)             # <<<<<<<<<<<<<<
 *     # x = ---- --98 ---- 76-- --54 ---- 32-- --10
 *     x=(x|(x<<4))&(<np.uint32_t>0x030C30C3)
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x << 8)) & ((__pyx_t_5numpy_uint32_t)0x0300F00F));

  /* "yt/utilities/lib/geometry_utils.pxd":210
 *     x=(x|(x<<8))&(<np.uint32_t>0x0300F00F)
 *     # x = ---- --98 ---- 76-- --54 ---- 32-- --10
 *     x=(x|(x<<4))&(<np.uint32_t>0x030C30C3)             # <<<<<<<<<<<<<<
 *     # x = ---- 9--8 --7- -6-- 5--4 --3- -2-- 1--0
 *     x=(x|(x<<2))&(<np.uint32_t>0x09249249)
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x << 4)) & ((__pyx_t_5numpy_uint32_t)0x030C30C3));

  /* "yt/utilities/lib/geometry_utils.pxd":212
 *     x=(x|(x<<4))&(<np.uint32_t>0x030C30C3)
 *     # x = ---- 9--8 --7- -6-- 5--4 --3- -2-- 1--0
 *     x=(x|(x<<2))&(<np.uint32_t>0x09249249)             # <<<<<<<<<<<<<<
 *     return x
 * 
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x << 2)) & ((__pyx_t_5numpy_uint32_t)0x09249249));

  /* "yt/utilities/lib/geometry_utils.pxd":213
 *     # x = ---- 9--8 --7- -6-- 5--4 --3- -2-- 1--0
 *     x=(x|(x<<2))&(<np.uint32_t>0x09249249)
 *     return x             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  __pyx_r = __pyx_v_x;
  goto __pyx_L0;

  /* "yt/utilities/lib/geometry_utils.pxd":200
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.uint32_t spread_32bits_by2(np.uint32_t x):             # <<<<<<<<<<<<<<
 *     # Only reversible up to 1023
 *     # Select highest 10 bits (Required to be reversible to 10st bit)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/geometry_utils.pxd":218
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.uint32_t compact_32bits_by2(np.uint32_t x):             # <<<<<<<<<<<<<<
 *     # Reversed magic
 *     x=x&(<np.uint32_t>0x09249249)
 */

static CYTHON_INLINE __pyx_t_5numpy_uint32_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_compact_32bits_by2(__pyx_t_5numpy_uint32_t __pyx_v_x) {
  __pyx_t_5numpy_uint32_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compact_32bits_by2", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":220
 * cdef inline np.uint32_t compact_32bits_by2(np.uint32_t x):
 *     # Reversed magic
 *     x=x&(<np.uint32_t>0x09249249)             # <<<<<<<<<<<<<<
 *     x=(x|(x>>2))&(<np.uint32_t>0x030C30C3)
 *     x=(x|(x>>4))&(<np.uint32_t>0x0300F00F)
 */
  __pyx_v_x = (__pyx_v_x & ((__pyx_t_5numpy_uint32_t)0x09249249));

  /* "yt/utilities/lib/geometry_utils.pxd":221
 *     # Reversed magic
 *     x=x&(<np.uint32_t>0x09249249)
 *     x=(x|(x>>2))&(<np.uint32_t>0x030C30C3)             # <<<<<<<<<<<<<<
 *     x=(x|(x>>4))&(<np.uint32_t>0x0300F00F)
 *     x=(x|(x>>8))&(<np.uint32_t>0xFF0000FF)
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x >> 2)) & ((__pyx_t_5numpy_uint32_t)0x030C30C3));

  /* "yt/utilities/lib/geometry_utils.pxd":222
 *     x=x&(<np.uint32_t>0x09249249)
 *     x=(x|(x>>2))&(<np.uint32_t>0x030C30C3)
 *     x=(x|(x>>4))&(<np.uint32_t>0x0300F00F)             # <<<<<<<<<<<<<<
 *     x=(x|(x>>8))&(<np.uint32_t>0xFF0000FF)
 *     x=(x|(x>>16))&(<np.uint32_t>0x000003FF)
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x >> 4)) & ((__pyx_t_5numpy_uint32_t)0x0300F00F));

  /* "yt/utilities/lib/geometry_utils.pxd":223
 *     x=(x|(x>>2))&(<np.uint32_t>0x030C30C3)
 *     x=(x|(x>>4))&(<np.uint32_t>0x0300F00F)
 *     x=(x|(x>>8))&(<np.uint32_t>0xFF0000FF)             # <<<<<<<<<<<<<<
 *     x=(x|(x>>16))&(<np.uint32_t>0x000003FF)
 *     return x
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x >> 8)) & ((__pyx_t_5numpy_uint32_t)0xFF0000FF));

  /* "yt/utilities/lib/geometry_utils.pxd":224
 *     x=(x|(x>>4))&(<np.uint32_t>0x0300F00F)
 *     x=(x|(x>>8))&(<np.uint32_t>0xFF0000FF)
 *     x=(x|(x>>16))&(<np.uint32_t>0x000003FF)             # <<<<<<<<<<<<<<
 *     return x
 * 
 */
  __pyx_v_x = ((__pyx_v_x | (__pyx_v_x >> 16)) & ((__pyx_t_5numpy_uint32_t)0x000003FF));

  /* "yt/utilities/lib/geometry_utils.pxd":225
 *     x=(x|(x>>8))&(<np.uint32_t>0xFF0000FF)
 *     x=(x|(x>>16))&(<np.uint32_t>0x000003FF)
 *     return x             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  __pyx_r = __pyx_v_x;
  goto __pyx_L0;

  /* "yt/utilities/lib/geometry_utils.pxd":218
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.uint32_t compact_32bits_by2(np.uint32_t x):             # <<<<<<<<<<<<<<
 *     # Reversed magic
 *     x=x&(<np.uint32_t>0x09249249)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/geometry_utils.pxd":230
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.uint64_t masked_merge_64bit(np.uint64_t a, np.uint64_t b, np.uint64_t mask):             # <<<<<<<<<<<<<<
 *     # https://graphics.stanford.edu/~seander/bithacks.html#MaskedMerge
 *     return a ^ ((a ^ b) & mask)
 */

static CYTHON_INLINE __pyx_t_5numpy_uint64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_masked_merge_64bit(__pyx_t_5numpy_uint64_t __pyx_v_a, __pyx_t_5numpy_uint64_t __pyx_v_b, __pyx_t_5numpy_uint64_t __pyx_v_mask) {
  __pyx_t_5numpy_uint64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("masked_merge_64bit", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":232
 * cdef inline np.uint64_t masked_merge_64bit(np.uint64_t a, np.uint64_t b, np.uint64_t mask):
 *     # https://graphics.stanford.edu/~seander/bithacks.html#MaskedMerge
 *     return a ^ ((a ^ b) & mask)             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  __pyx_r = (__pyx_v_a ^ ((__pyx_v_a ^ __pyx_v_b) & __pyx_v_mask));
  goto __pyx_L0;

  /* "yt/utilities/lib/geometry_utils.pxd":230
 * @cython.boundscheck(False)
 * @cython.wraparound(False)
 * cdef inline np.uint64_t masked_merge_64bit(np.uint64_t a, np.uint64_t b, np.uint64_t mask):             # <<<<<<<<<<<<<<
 *     # https://graphics.stanford.edu/~seander/bithacks.html#MaskedMerge
 *     return a ^ ((a ^ b) & mask)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/geometry_utils.pxd":235
 * 
 * @cython.cdivision(True)
 * cdef inline np.uint64_t encode_morton_64bit(np.uint64_t x_ind, np.uint64_t y_ind, np.uint64_t z_ind):             # <<<<<<<<<<<<<<
 *     cdef np.uint64_t mi
 *     mi = 0
 */

static CYTHON_INLINE __pyx_t_5numpy_uint64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_encode_morton_64bit(__pyx_t_5numpy_uint64_t __pyx_v_x_ind, __pyx_t_5numpy_uint64_t __pyx_v_y_ind, __pyx_t_5numpy_uint64_t __pyx_v_z_ind) {
  __pyx_t_5numpy_uint64_t __pyx_v_mi;
  __pyx_t_5numpy_uint64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("encode_morton_64bit", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":237
 * cdef inline np.uint64_t encode_morton_64bit(np.uint64_t x_ind, np.uint64_t y_ind, np.uint64_t z_ind):
 *     cdef np.uint64_t mi
 *     mi = 0             # <<<<<<<<<<<<<<
 *     mi |= spread_64bits_by2(z_ind)<<ZSHIFT
 *     mi |= spread_64bits_by2(y_ind)<<YSHIFT
 */
  __pyx_v_mi = 0;

  /* "yt/utilities/lib/geometry_utils.pxd":238
 *     cdef np.uint64_t mi
 *     mi = 0
 *     mi |= spread_64bits_by2(z_ind)<<ZSHIFT             # <<<<<<<<<<<<<<
 *     mi |= spread_64bits_by2(y_ind)<<YSHIFT
 *     mi |= spread_64bits_by2(x_ind)<<XSHIFT
 */
  __pyx_v_mi = (__pyx_v_mi | (__pyx_f_2yt_9utilities_3lib_14geometry_utils_spread_64bits_by2(__pyx_v_z_ind) << 0));

  /* "yt/utilities/lib/geometry_utils.pxd":239
 *     mi = 0
 *     mi |= spread_64bits_by2(z_ind)<<ZSHIFT
 *     mi |= spread_64bits_by2(y_ind)<<YSHIFT             # <<<<<<<<<<<<<<
 *     mi |= spread_64bits_by2(x_ind)<<XSHIFT
 *     return mi
 */
  __pyx_v_mi = (__pyx_v_mi | (__pyx_f_2yt_9utilities_3lib_14geometry_utils_spread_64bits_by2(__pyx_v_y_ind) << 1));

  /* "yt/utilities/lib/geometry_utils.pxd":240
 *     mi |= spread_64bits_by2(z_ind)<<ZSHIFT
 *     mi |= spread_64bits_by2(y_ind)<<YSHIFT
 *     mi |= spread_64bits_by2(x_ind)<<XSHIFT             # <<<<<<<<<<<<<<
 *     return mi
 * 
 */
  __pyx_v_mi = (__pyx_v_mi | (__pyx_f_2yt_9utilities_3lib_14geometry_utils_spread_64bits_by2(__pyx_v_x_ind) << 2));

  /* "yt/utilities/lib/geometry_utils.pxd":241
 *     mi |= spread_64bits_by2(y_ind)<<YSHIFT
 *     mi |= spread_64bits_by2(x_ind)<<XSHIFT
 *     return mi             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  __pyx_r = __pyx_v_mi;
  goto __pyx_L0;

  /* "yt/utilities/lib/geometry_utils.pxd":235
 * 
 * @cython.cdivision(True)
 * cdef inline np.uint64_t encode_morton_64bit(np.uint64_t x_ind, np.uint64_t y_ind, np.uint64_t z_ind):             # <<<<<<<<<<<<<<
 *     cdef np.uint64_t mi
 *     mi = 0
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/geometry_utils.pxd":244
 * 
 * @cython.cdivision(True)
 * cdef inline void decode_morton_64bit(np.uint64_t mi, np.uint64_t *p):             # <<<<<<<<<<<<<<
 *     p[0] = compact_64bits_by2(mi>>XSHIFT)
 *     p[1] = compact_64bits_by2(mi>>YSHIFT)
 */

static CYTHON_INLINE void __pyx_f_2yt_9utilities_3lib_14geometry_utils_decode_morton_64bit(__pyx_t_5numpy_uint64_t __pyx_v_mi, __pyx_t_5numpy_uint64_t *__pyx_v_p) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("decode_morton_64bit", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":245
 * @cython.cdivision(True)
 * cdef inline void decode_morton_64bit(np.uint64_t mi, np.uint64_t *p):
 *     p[0] = compact_64bits_by2(mi>>XSHIFT)             # <<<<<<<<<<<<<<
 *     p[1] = compact_64bits_by2(mi>>YSHIFT)
 *     p[2] = compact_64bits_by2(mi>>ZSHIFT)
 */
  (__pyx_v_p[0]) = __pyx_f_2yt_9utilities_3lib_14geometry_utils_compact_64bits_by2((__pyx_v_mi >> 2));

  /* "yt/utilities/lib/geometry_utils.pxd":246
 * cdef inline void decode_morton_64bit(np.uint64_t mi, np.uint64_t *p):
 *     p[0] = compact_64bits_by2(mi>>XSHIFT)
 *     p[1] = compact_64bits_by2(mi>>YSHIFT)             # <<<<<<<<<<<<<<
 *     p[2] = compact_64bits_by2(mi>>ZSHIFT)
 * 
 */
  (__pyx_v_p[1]) = __pyx_f_2yt_9utilities_3lib_14geometry_utils_compact_64bits_by2((__pyx_v_mi >> 1));

  /* "yt/utilities/lib/geometry_utils.pxd":247
 *     p[0] = compact_64bits_by2(mi>>XSHIFT)
 *     p[1] = compact_64bits_by2(mi>>YSHIFT)
 *     p[2] = compact_64bits_by2(mi>>ZSHIFT)             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  (__pyx_v_p[2]) = __pyx_f_2yt_9utilities_3lib_14geometry_utils_compact_64bits_by2((__pyx_v_mi >> 0));

  /* "yt/utilities/lib/geometry_utils.pxd":244
 * 
 * @cython.cdivision(True)
 * cdef inline void decode_morton_64bit(np.uint64_t mi, np.uint64_t *p):             # <<<<<<<<<<<<<<
 *     p[0] = compact_64bits_by2(mi>>XSHIFT)
 *     p[1] = compact_64bits_by2(mi>>YSHIFT)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "yt/utilities/lib/geometry_utils.pxd":250
 * 
 * @cython.cdivision(True)
 * cdef inline np.uint64_t bounded_morton(np.float64_t x, np.float64_t y, np.float64_t z,             # <<<<<<<<<<<<<<
 *                                np.float64_t *DLE, np.float64_t *DRE, np.int32_t order):
 *     cdef int i
 */

static CYTHON_INLINE __pyx_t_5numpy_uint64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton(__pyx_t_5numpy_float64_t __pyx_v_x, __pyx_t_5numpy_float64_t __pyx_v_y, __pyx_t_5numpy_float64_t __pyx_v_z, __pyx_t_5numpy_float64_t *__pyx_v_DLE, __pyx_t_5numpy_float64_t *__pyx_v_DRE, __pyx_t_5numpy_int32_t __pyx_v_order) {
  int __pyx_v_i;
  __pyx_t_5numpy_float64_t __pyx_v_dds[3];
  __pyx_t_5numpy_uint64_t __pyx_v_x_ind;
  __pyx_t_5numpy_uint64_t __pyx_v_y_ind;
  __pyx_t_5numpy_uint64_t __pyx_v_z_ind;
  __pyx_t_5numpy_uint64_t __pyx_v_mi;
  __pyx_t_5numpy_uint64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("bounded_morton", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":256
 *     cdef np.uint64_t x_ind, y_ind, z_ind
 *     cdef np.uint64_t mi
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         dds[i] = (DRE[i] - DLE[i]) / (1 << order)
 *     x_ind = <np.uint64_t> ((x - DLE[0])/dds[0])
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/geometry_utils.pxd":257
 *     cdef np.uint64_t mi
 *     for i in range(3):
 *         dds[i] = (DRE[i] - DLE[i]) / (1 << order)             # <<<<<<<<<<<<<<
 *     x_ind = <np.uint64_t> ((x - DLE[0])/dds[0])
 *     y_ind = <np.uint64_t> ((y - DLE[1])/dds[1])
 */
    (__pyx_v_dds[__pyx_v_i]) = (((__pyx_v_DRE[__pyx_v_i]) - (__pyx_v_DLE[__pyx_v_i])) / (1 << __pyx_v_order));
  }

  /* "yt/utilities/lib/geometry_utils.pxd":258
 *     for i in range(3):
 *         dds[i] = (DRE[i] - DLE[i]) / (1 << order)
 *     x_ind = <np.uint64_t> ((x - DLE[0])/dds[0])             # <<<<<<<<<<<<<<
 *     y_ind = <np.uint64_t> ((y - DLE[1])/dds[1])
 *     z_ind = <np.uint64_t> ((z - DLE[2])/dds[2])
 */
  __pyx_v_x_ind = ((__pyx_t_5numpy_uint64_t)((__pyx_v_x - (__pyx_v_DLE[0])) / (__pyx_v_dds[0])));

  /* "yt/utilities/lib/geometry_utils.pxd":259
 *         dds[i] = (DRE[i] - DLE[i]) / (1 << order)
 *     x_ind = <np.uint64_t> ((x - DLE[0])/dds[0])
 *     y_ind = <np.uint64_t> ((y - DLE[1])/dds[1])             # <<<<<<<<<<<<<<
 *     z_ind = <np.uint64_t> ((z - DLE[2])/dds[2])
 *     mi = encode_morton_64bit(x_ind,y_ind,z_ind)
 */
  __pyx_v_y_ind = ((__pyx_t_5numpy_uint64_t)((__pyx_v_y - (__pyx_v_DLE[1])) / (__pyx_v_dds[1])));

  /* "yt/utilities/lib/geometry_utils.pxd":260
 *     x_ind = <np.uint64_t> ((x - DLE[0])/dds[0])
 *     y_ind = <np.uint64_t> ((y - DLE[1])/dds[1])
 *     z_ind = <np.uint64_t> ((z - DLE[2])/dds[2])             # <<<<<<<<<<<<<<
 *     mi = encode_morton_64bit(x_ind,y_ind,z_ind)
 *     return mi
 */
  __pyx_v_z_ind = ((__pyx_t_5numpy_uint64_t)((__pyx_v_z - (__pyx_v_DLE[2])) / (__pyx_v_dds[2])));

  /* "yt/utilities/lib/geometry_utils.pxd":261
 *     y_ind = <np.uint64_t> ((y - DLE[1])/dds[1])
 *     z_ind = <np.uint64_t> ((z - DLE[2])/dds[2])
 *     mi = encode_morton_64bit(x_ind,y_ind,z_ind)             # <<<<<<<<<<<<<<
 *     return mi
 * 
 */
  __pyx_v_mi = __pyx_f_2yt_9utilities_3lib_14geometry_utils_encode_morton_64bit(__pyx_v_x_ind, __pyx_v_y_ind, __pyx_v_z_ind);

  /* "yt/utilities/lib/geometry_utils.pxd":262
 *     z_ind = <np.uint64_t> ((z - DLE[2])/dds[2])
 *     mi = encode_morton_64bit(x_ind,y_ind,z_ind)
 *     return mi             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  __pyx_r = __pyx_v_mi;
  goto __pyx_L0;

  /* "yt/utilities/lib/geometry_utils.pxd":250
 * 
 * @cython.cdivision(True)
 * cdef inline np.uint64_t bounded_morton(np.float64_t x, np.float64_t y, np.float64_t z,             # <<<<<<<<<<<<<<
 *                                np.float64_t *DLE, np.float64_t *DRE, np.int32_t order):
 *     cdef int i
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/geometry_utils.pxd":265
 * 
 * @cython.cdivision(True)
 * cdef inline np.uint64_t bounded_morton_relative(np.float64_t x, np.float64_t y, np.float64_t z,             # <<<<<<<<<<<<<<
 *                                np.float64_t *DLE, np.float64_t *DRE,
 *                                np.int32_t order1, np.int32_t order2):
 */

static CYTHON_INLINE __pyx_t_5numpy_uint64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_relative(__pyx_t_5numpy_float64_t __pyx_v_x, __pyx_t_5numpy_float64_t __pyx_v_y, __pyx_t_5numpy_float64_t __pyx_v_z, __pyx_t_5numpy_float64_t *__pyx_v_DLE, __pyx_t_5numpy_float64_t *__pyx_v_DRE, __pyx_t_5numpy_int32_t __pyx_v_order1, __pyx_t_5numpy_int32_t __pyx_v_order2) {
  int __pyx_v_i;
  __pyx_t_5numpy_float64_t __pyx_v_dds1[3];
  __pyx_t_5numpy_float64_t __pyx_v_dds2[3];
  __pyx_t_5numpy_float64_t __pyx_v_DLE2[3];
  __pyx_t_5numpy_uint64_t __pyx_v_x_ind;
  __pyx_t_5numpy_uint64_t __pyx_v_y_ind;
  __pyx_t_5numpy_uint64_t __pyx_v_z_ind;
  __pyx_t_5numpy_uint64_t __pyx_v_mi2;
  __pyx_t_5numpy_uint64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("bounded_morton_relative", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":274
 *     cdef np.uint64_t x_ind, y_ind, z_ind
 *     cdef np.uint64_t mi2
 *     for i in range(3):             # <<<<<<<<<<<<<<
 *         dds1[i] = (DRE[i] - DLE[i]) / (1 << order1)
 *         dds2[i] = dds1[i] / (1 << order2)
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "yt/utilities/lib/geometry_utils.pxd":275
 *     cdef np.uint64_t mi2
 *     for i in range(3):
 *         dds1[i] = (DRE[i] - DLE[i]) / (1 << order1)             # <<<<<<<<<<<<<<
 *         dds2[i] = dds1[i] / (1 << order2)
 *     DLE2[0] = <np.float64_t> (<np.uint64_t> ((x - DLE[0])/dds1[0])) * dds1[0]
 */
    (__pyx_v_dds1[__pyx_v_i]) = (((__pyx_v_DRE[__pyx_v_i]) - (__pyx_v_DLE[__pyx_v_i])) / (1 << __pyx_v_order1));

    /* "yt/utilities/lib/geometry_utils.pxd":276
 *     for i in range(3):
 *         dds1[i] = (DRE[i] - DLE[i]) / (1 << order1)
 *         dds2[i] = dds1[i] / (1 << order2)             # <<<<<<<<<<<<<<
 *     DLE2[0] = <np.float64_t> (<np.uint64_t> ((x - DLE[0])/dds1[0])) * dds1[0]
 *     DLE2[1] = <np.float64_t> (<np.uint64_t> ((y - DLE[1])/dds1[1])) * dds1[1]
 */
    (__pyx_v_dds2[__pyx_v_i]) = ((__pyx_v_dds1[__pyx_v_i]) / (1 << __pyx_v_order2));
  }

  /* "yt/utilities/lib/geometry_utils.pxd":277
 *         dds1[i] = (DRE[i] - DLE[i]) / (1 << order1)
 *         dds2[i] = dds1[i] / (1 << order2)
 *     DLE2[0] = <np.float64_t> (<np.uint64_t> ((x - DLE[0])/dds1[0])) * dds1[0]             # <<<<<<<<<<<<<<
 *     DLE2[1] = <np.float64_t> (<np.uint64_t> ((y - DLE[1])/dds1[1])) * dds1[1]
 *     DLE2[2] = <np.float64_t> (<np.uint64_t> ((z - DLE[2])/dds1[2])) * dds1[2]
 */
  (__pyx_v_DLE2[0]) = (((__pyx_t_5numpy_float64_t)((__pyx_t_5numpy_uint64_t)((__pyx_v_x - (__pyx_v_DLE[0])) / (__pyx_v_dds1[0])))) * (__pyx_v_dds1[0]));

  /* "yt/utilities/lib/geometry_utils.pxd":278
 *         dds2[i] = dds1[i] / (1 << order2)
 *     DLE2[0] = <np.float64_t> (<np.uint64_t> ((x - DLE[0])/dds1[0])) * dds1[0]
 *     DLE2[1] = <np.float64_t> (<np.uint64_t> ((y - DLE[1])/dds1[1])) * dds1[1]             # <<<<<<<<<<<<<<
 *     DLE2[2] = <np.float64_t> (<np.uint64_t> ((z - DLE[2])/dds1[2])) * dds1[2]
 *     x_ind = <np.uint64_t> ((x - DLE2[0])/dds2[0])
 */
  (__pyx_v_DLE2[1]) = (((__pyx_t_5numpy_float64_t)((__pyx_t_5numpy_uint64_t)((__pyx_v_y - (__pyx_v_DLE[1])) / (__pyx_v_dds1[1])))) * (__pyx_v_dds1[1]));

  /* "yt/utilities/lib/geometry_utils.pxd":279
 *     DLE2[0] = <np.float64_t> (<np.uint64_t> ((x - DLE[0])/dds1[0])) * dds1[0]
 *     DLE2[1] = <np.float64_t> (<np.uint64_t> ((y - DLE[1])/dds1[1])) * dds1[1]
 *     DLE2[2] = <np.float64_t> (<np.uint64_t> ((z - DLE[2])/dds1[2])) * dds1[2]             # <<<<<<<<<<<<<<
 *     x_ind = <np.uint64_t> ((x - DLE2[0])/dds2[0])
 *     y_ind = <np.uint64_t> ((y - DLE2[1])/dds2[1])
 */
  (__pyx_v_DLE2[2]) = (((__pyx_t_5numpy_float64_t)((__pyx_t_5numpy_uint64_t)((__pyx_v_z - (__pyx_v_DLE[2])) / (__pyx_v_dds1[2])))) * (__pyx_v_dds1[2]));

  /* "yt/utilities/lib/geometry_utils.pxd":280
 *     DLE2[1] = <np.float64_t> (<np.uint64_t> ((y - DLE[1])/dds1[1])) * dds1[1]
 *     DLE2[2] = <np.float64_t> (<np.uint64_t> ((z - DLE[2])/dds1[2])) * dds1[2]
 *     x_ind = <np.uint64_t> ((x - DLE2[0])/dds2[0])             # <<<<<<<<<<<<<<
 *     y_ind = <np.uint64_t> ((y - DLE2[1])/dds2[1])
 *     z_ind = <np.uint64_t> ((z - DLE2[2])/dds2[2])
 */
  __pyx_v_x_ind = ((__pyx_t_5numpy_uint64_t)((__pyx_v_x - (__pyx_v_DLE2[0])) / (__pyx_v_dds2[0])));

  /* "yt/utilities/lib/geometry_utils.pxd":281
 *     DLE2[2] = <np.float64_t> (<np.uint64_t> ((z - DLE[2])/dds1[2])) * dds1[2]
 *     x_ind = <np.uint64_t> ((x - DLE2[0])/dds2[0])
 *     y_ind = <np.uint64_t> ((y - DLE2[1])/dds2[1])             # <<<<<<<<<<<<<<
 *     z_ind = <np.uint64_t> ((z - DLE2[2])/dds2[2])
 *     mi2 = encode_morton_64bit(x_ind,y_ind,z_ind)
 */
  __pyx_v_y_ind = ((__pyx_t_5numpy_uint64_t)((__pyx_v_y - (__pyx_v_DLE2[1])) / (__pyx_v_dds2[1])));

  /* "yt/utilities/lib/geometry_utils.pxd":282
 *     x_ind = <np.uint64_t> ((x - DLE2[0])/dds2[0])
 *     y_ind = <np.uint64_t> ((y - DLE2[1])/dds2[1])
 *     z_ind = <np.uint64_t> ((z - DLE2[2])/dds2[2])             # <<<<<<<<<<<<<<
 *     mi2 = encode_morton_64bit(x_ind,y_ind,z_ind)
 *     return mi2
 */
  __pyx_v_z_ind = ((__pyx_t_5numpy_uint64_t)((__pyx_v_z - (__pyx_v_DLE2[2])) / (__pyx_v_dds2[2])));

  /* "yt/utilities/lib/geometry_utils.pxd":283
 *     y_ind = <np.uint64_t> ((y - DLE2[1])/dds2[1])
 *     z_ind = <np.uint64_t> ((z - DLE2[2])/dds2[2])
 *     mi2 = encode_morton_64bit(x_ind,y_ind,z_ind)             # <<<<<<<<<<<<<<
 *     return mi2
 * 
 */
  __pyx_v_mi2 = __pyx_f_2yt_9utilities_3lib_14geometry_utils_encode_morton_64bit(__pyx_v_x_ind, __pyx_v_y_ind, __pyx_v_z_ind);

  /* "yt/utilities/lib/geometry_utils.pxd":284
 *     z_ind = <np.uint64_t> ((z - DLE2[2])/dds2[2])
 *     mi2 = encode_morton_64bit(x_ind,y_ind,z_ind)
 *     return mi2             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_mi2;
  goto __pyx_L0;

  /* "yt/utilities/lib/geometry_utils.pxd":265
 * 
 * @cython.cdivision(True)
 * cdef inline np.uint64_t bounded_morton_relative(np.float64_t x, np.float64_t y, np.float64_t z,             # <<<<<<<<<<<<<<
 *                                np.float64_t *DLE, np.float64_t *DRE,
 *                                np.int32_t order1, np.int32_t order2):
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/geometry_utils.pxd":289
 * # This doesn't seem to be much, if at all, faster...
 * @cython.cdivision(True)
 * cdef inline np.uint64_t bounded_morton_dds(np.float64_t x, np.float64_t y, np.float64_t z,             # <<<<<<<<<<<<<<
 *                                np.float64_t *DLE, np.float64_t *dds):
 *     cdef np.uint64_t x_ind, y_ind, z_ind
 */

static CYTHON_INLINE __pyx_t_5numpy_uint64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_dds(__pyx_t_5numpy_float64_t __pyx_v_x, __pyx_t_5numpy_float64_t __pyx_v_y, __pyx_t_5numpy_float64_t __pyx_v_z, __pyx_t_5numpy_float64_t *__pyx_v_DLE, __pyx_t_5numpy_float64_t *__pyx_v_dds) {
  __pyx_t_5numpy_uint64_t __pyx_v_x_ind;
  __pyx_t_5numpy_uint64_t __pyx_v_y_ind;
  __pyx_t_5numpy_uint64_t __pyx_v_z_ind;
  __pyx_t_5numpy_uint64_t __pyx_v_mi;
  __pyx_t_5numpy_uint64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bounded_morton_dds", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":293
 *     cdef np.uint64_t x_ind, y_ind, z_ind
 *     cdef np.uint64_t mi
 *     x_ind = <np.uint64_t> ((x - DLE[0])/dds[0])             # <<<<<<<<<<<<<<
 *     y_ind = <np.uint64_t> ((y - DLE[1])/dds[1])
 *     z_ind = <np.uint64_t> ((z - DLE[2])/dds[2])
 */
  __pyx_v_x_ind = ((__pyx_t_5numpy_uint64_t)((__pyx_v_x - (__pyx_v_DLE[0])) / (__pyx_v_dds[0])));

  /* "yt/utilities/lib/geometry_utils.pxd":294
 *     cdef np.uint64_t mi
 *     x_ind = <np.uint64_t> ((x - DLE[0])/dds[0])
 *     y_ind = <np.uint64_t> ((y - DLE[1])/dds[1])             # <<<<<<<<<<<<<<
 *     z_ind = <np.uint64_t> ((z - DLE[2])/dds[2])
 *     mi = encode_morton_64bit(x_ind,y_ind,z_ind)
 */
  __pyx_v_y_ind = ((__pyx_t_5numpy_uint64_t)((__pyx_v_y - (__pyx_v_DLE[1])) / (__pyx_v_dds[1])));

  /* "yt/utilities/lib/geometry_utils.pxd":295
 *     x_ind = <np.uint64_t> ((x - DLE[0])/dds[0])
 *     y_ind = <np.uint64_t> ((y - DLE[1])/dds[1])
 *     z_ind = <np.uint64_t> ((z - DLE[2])/dds[2])             # <<<<<<<<<<<<<<
 *     mi = encode_morton_64bit(x_ind,y_ind,z_ind)
 *     return mi
 */
  __pyx_v_z_ind = ((__pyx_t_5numpy_uint64_t)((__pyx_v_z - (__pyx_v_DLE[2])) / (__pyx_v_dds[2])));

  /* "yt/utilities/lib/geometry_utils.pxd":296
 *     y_ind = <np.uint64_t> ((y - DLE[1])/dds[1])
 *     z_ind = <np.uint64_t> ((z - DLE[2])/dds[2])
 *     mi = encode_morton_64bit(x_ind,y_ind,z_ind)             # <<<<<<<<<<<<<<
 *     return mi
 * 
 */
  __pyx_v_mi = __pyx_f_2yt_9utilities_3lib_14geometry_utils_encode_morton_64bit(__pyx_v_x_ind, __pyx_v_y_ind, __pyx_v_z_ind);

  /* "yt/utilities/lib/geometry_utils.pxd":297
 *     z_ind = <np.uint64_t> ((z - DLE[2])/dds[2])
 *     mi = encode_morton_64bit(x_ind,y_ind,z_ind)
 *     return mi             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  __pyx_r = __pyx_v_mi;
  goto __pyx_L0;

  /* "yt/utilities/lib/geometry_utils.pxd":289
 * # This doesn't seem to be much, if at all, faster...
 * @cython.cdivision(True)
 * cdef inline np.uint64_t bounded_morton_dds(np.float64_t x, np.float64_t y, np.float64_t z,             # <<<<<<<<<<<<<<
 *                                np.float64_t *DLE, np.float64_t *dds):
 *     cdef np.uint64_t x_ind, y_ind, z_ind
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/geometry_utils.pxd":300
 * 
 * @cython.cdivision(True)
 * cdef inline np.uint64_t bounded_morton_relative_dds(np.float64_t x, np.float64_t y, np.float64_t z,             # <<<<<<<<<<<<<<
 *                                np.float64_t *DLE, np.float64_t *dds1, np.float64_t *dds2):
 *     cdef np.float64_t DLE2[3]
 */

static CYTHON_INLINE __pyx_t_5numpy_uint64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_relative_dds(__pyx_t_5numpy_float64_t __pyx_v_x, __pyx_t_5numpy_float64_t __pyx_v_y, __pyx_t_5numpy_float64_t __pyx_v_z, __pyx_t_5numpy_float64_t *__pyx_v_DLE, __pyx_t_5numpy_float64_t *__pyx_v_dds1, __pyx_t_5numpy_float64_t *__pyx_v_dds2) {
  __pyx_t_5numpy_float64_t __pyx_v_DLE2[3];
  __pyx_t_5numpy_uint64_t __pyx_v_x_ind;
  __pyx_t_5numpy_uint64_t __pyx_v_y_ind;
  __pyx_t_5numpy_uint64_t __pyx_v_z_ind;
  __pyx_t_5numpy_uint64_t __pyx_v_mi2;
  __pyx_t_5numpy_uint64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bounded_morton_relative_dds", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":305
 *     cdef np.uint64_t x_ind, y_ind, z_ind
 *     cdef np.uint64_t mi2
 *     DLE2[0] = <np.float64_t> (<np.uint64_t> ((x - DLE[0])/dds1[0])) * dds1[0]             # <<<<<<<<<<<<<<
 *     DLE2[1] = <np.float64_t> (<np.uint64_t> ((y - DLE[1])/dds1[1])) * dds1[1]
 *     DLE2[2] = <np.float64_t> (<np.uint64_t> ((z - DLE[2])/dds1[2])) * dds1[2]
 */
  (__pyx_v_DLE2[0]) = (((__pyx_t_5numpy_float64_t)((__pyx_t_5numpy_uint64_t)((__pyx_v_x - (__pyx_v_DLE[0])) / (__pyx_v_dds1[0])))) * (__pyx_v_dds1[0]));

  /* "yt/utilities/lib/geometry_utils.pxd":306
 *     cdef np.uint64_t mi2
 *     DLE2[0] = <np.float64_t> (<np.uint64_t> ((x - DLE[0])/dds1[0])) * dds1[0]
 *     DLE2[1] = <np.float64_t> (<np.uint64_t> ((y - DLE[1])/dds1[1])) * dds1[1]             # <<<<<<<<<<<<<<
 *     DLE2[2] = <np.float64_t> (<np.uint64_t> ((z - DLE[2])/dds1[2])) * dds1[2]
 *     x_ind = <np.uint64_t> ((x - DLE2[0])/dds2[0])
 */
  (__pyx_v_DLE2[1]) = (((__pyx_t_5numpy_float64_t)((__pyx_t_5numpy_uint64_t)((__pyx_v_y - (__pyx_v_DLE[1])) / (__pyx_v_dds1[1])))) * (__pyx_v_dds1[1]));

  /* "yt/utilities/lib/geometry_utils.pxd":307
 *     DLE2[0] = <np.float64_t> (<np.uint64_t> ((x - DLE[0])/dds1[0])) * dds1[0]
 *     DLE2[1] = <np.float64_t> (<np.uint64_t> ((y - DLE[1])/dds1[1])) * dds1[1]
 *     DLE2[2] = <np.float64_t> (<np.uint64_t> ((z - DLE[2])/dds1[2])) * dds1[2]             # <<<<<<<<<<<<<<
 *     x_ind = <np.uint64_t> ((x - DLE2[0])/dds2[0])
 *     y_ind = <np.uint64_t> ((y - DLE2[1])/dds2[1])
 */
  (__pyx_v_DLE2[2]) = (((__pyx_t_5numpy_float64_t)((__pyx_t_5numpy_uint64_t)((__pyx_v_z - (__pyx_v_DLE[2])) / (__pyx_v_dds1[2])))) * (__pyx_v_dds1[2]));

  /* "yt/utilities/lib/geometry_utils.pxd":308
 *     DLE2[1] = <np.float64_t> (<np.uint64_t> ((y - DLE[1])/dds1[1])) * dds1[1]
 *     DLE2[2] = <np.float64_t> (<np.uint64_t> ((z - DLE[2])/dds1[2])) * dds1[2]
 *     x_ind = <np.uint64_t> ((x - DLE2[0])/dds2[0])             # <<<<<<<<<<<<<<
 *     y_ind = <np.uint64_t> ((y - DLE2[1])/dds2[1])
 *     z_ind = <np.uint64_t> ((z - DLE2[2])/dds2[2])
 */
  __pyx_v_x_ind = ((__pyx_t_5numpy_uint64_t)((__pyx_v_x - (__pyx_v_DLE2[0])) / (__pyx_v_dds2[0])));

  /* "yt/utilities/lib/geometry_utils.pxd":309
 *     DLE2[2] = <np.float64_t> (<np.uint64_t> ((z - DLE[2])/dds1[2])) * dds1[2]
 *     x_ind = <np.uint64_t> ((x - DLE2[0])/dds2[0])
 *     y_ind = <np.uint64_t> ((y - DLE2[1])/dds2[1])             # <<<<<<<<<<<<<<
 *     z_ind = <np.uint64_t> ((z - DLE2[2])/dds2[2])
 *     mi2 = encode_morton_64bit(x_ind,y_ind,z_ind)
 */
  __pyx_v_y_ind = ((__pyx_t_5numpy_uint64_t)((__pyx_v_y - (__pyx_v_DLE2[1])) / (__pyx_v_dds2[1])));

  /* "yt/utilities/lib/geometry_utils.pxd":310
 *     x_ind = <np.uint64_t> ((x - DLE2[0])/dds2[0])
 *     y_ind = <np.uint64_t> ((y - DLE2[1])/dds2[1])
 *     z_ind = <np.uint64_t> ((z - DLE2[2])/dds2[2])             # <<<<<<<<<<<<<<
 *     mi2 = encode_morton_64bit(x_ind,y_ind,z_ind)
 *     return mi2
 */
  __pyx_v_z_ind = ((__pyx_t_5numpy_uint64_t)((__pyx_v_z - (__pyx_v_DLE2[2])) / (__pyx_v_dds2[2])));

  /* "yt/utilities/lib/geometry_utils.pxd":311
 *     y_ind = <np.uint64_t> ((y - DLE2[1])/dds2[1])
 *     z_ind = <np.uint64_t> ((z - DLE2[2])/dds2[2])
 *     mi2 = encode_morton_64bit(x_ind,y_ind,z_ind)             # <<<<<<<<<<<<<<
 *     return mi2
 * 
 */
  __pyx_v_mi2 = __pyx_f_2yt_9utilities_3lib_14geometry_utils_encode_morton_64bit(__pyx_v_x_ind, __pyx_v_y_ind, __pyx_v_z_ind);

  /* "yt/utilities/lib/geometry_utils.pxd":312
 *     z_ind = <np.uint64_t> ((z - DLE2[2])/dds2[2])
 *     mi2 = encode_morton_64bit(x_ind,y_ind,z_ind)
 *     return mi2             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_mi2;
  goto __pyx_L0;

  /* "yt/utilities/lib/geometry_utils.pxd":300
 * 
 * @cython.cdivision(True)
 * cdef inline np.uint64_t bounded_morton_relative_dds(np.float64_t x, np.float64_t y, np.float64_t z,             # <<<<<<<<<<<<<<
 *                                np.float64_t *DLE, np.float64_t *dds1, np.float64_t *dds2):
 *     cdef np.float64_t DLE2[3]
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/geometry_utils.pxd":316
 * 
 * @cython.cdivision(True)
 * cdef inline np.uint64_t bounded_morton_split_dds(np.float64_t x, np.float64_t y, np.float64_t z,             # <<<<<<<<<<<<<<
 *                                np.float64_t *DLE, np.float64_t *dds, np.uint64_t *p):
 *     cdef np.uint64_t mi
 */

static CYTHON_INLINE __pyx_t_5numpy_uint64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_dds(__pyx_t_5numpy_float64_t __pyx_v_x, __pyx_t_5numpy_float64_t __pyx_v_y, __pyx_t_5numpy_float64_t __pyx_v_z, __pyx_t_5numpy_float64_t *__pyx_v_DLE, __pyx_t_5numpy_float64_t *__pyx_v_dds, __pyx_t_5numpy_uint64_t *__pyx_v_p) {
  __pyx_t_5numpy_uint64_t __pyx_v_mi;
  __pyx_t_5numpy_uint64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bounded_morton_split_dds", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":319
 *                                np.float64_t *DLE, np.float64_t *dds, np.uint64_t *p):
 *     cdef np.uint64_t mi
 *     p[0] = <np.uint64_t> ((x - DLE[0])/dds[0])             # <<<<<<<<<<<<<<
 *     p[1] = <np.uint64_t> ((y - DLE[1])/dds[1])
 *     p[2] = <np.uint64_t> ((z - DLE[2])/dds[2])
 */
  (__pyx_v_p[0]) = ((__pyx_t_5numpy_uint64_t)((__pyx_v_x - (__pyx_v_DLE[0])) / (__pyx_v_dds[0])));

  /* "yt/utilities/lib/geometry_utils.pxd":320
 *     cdef np.uint64_t mi
 *     p[0] = <np.uint64_t> ((x - DLE[0])/dds[0])
 *     p[1] = <np.uint64_t> ((y - DLE[1])/dds[1])             # <<<<<<<<<<<<<<
 *     p[2] = <np.uint64_t> ((z - DLE[2])/dds[2])
 *     mi = encode_morton_64bit(p[0], p[1], p[2])
 */
  (__pyx_v_p[1]) = ((__pyx_t_5numpy_uint64_t)((__pyx_v_y - (__pyx_v_DLE[1])) / (__pyx_v_dds[1])));

  /* "yt/utilities/lib/geometry_utils.pxd":321
 *     p[0] = <np.uint64_t> ((x - DLE[0])/dds[0])
 *     p[1] = <np.uint64_t> ((y - DLE[1])/dds[1])
 *     p[2] = <np.uint64_t> ((z - DLE[2])/dds[2])             # <<<<<<<<<<<<<<
 *     mi = encode_morton_64bit(p[0], p[1], p[2])
 *     return mi
 */
  (__pyx_v_p[2]) = ((__pyx_t_5numpy_uint64_t)((__pyx_v_z - (__pyx_v_DLE[2])) / (__pyx_v_dds[2])));

  /* "yt/utilities/lib/geometry_utils.pxd":322
 *     p[1] = <np.uint64_t> ((y - DLE[1])/dds[1])
 *     p[2] = <np.uint64_t> ((z - DLE[2])/dds[2])
 *     mi = encode_morton_64bit(p[0], p[1], p[2])             # <<<<<<<<<<<<<<
 *     return mi
 * 
 */
  __pyx_v_mi = __pyx_f_2yt_9utilities_3lib_14geometry_utils_encode_morton_64bit((__pyx_v_p[0]), (__pyx_v_p[1]), (__pyx_v_p[2]));

  /* "yt/utilities/lib/geometry_utils.pxd":323
 *     p[2] = <np.uint64_t> ((z - DLE[2])/dds[2])
 *     mi = encode_morton_64bit(p[0], p[1], p[2])
 *     return mi             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  __pyx_r = __pyx_v_mi;
  goto __pyx_L0;

  /* "yt/utilities/lib/geometry_utils.pxd":316
 * 
 * @cython.cdivision(True)
 * cdef inline np.uint64_t bounded_morton_split_dds(np.float64_t x, np.float64_t y, np.float64_t z,             # <<<<<<<<<<<<<<
 *                                np.float64_t *DLE, np.float64_t *dds, np.uint64_t *p):
 *     cdef np.uint64_t mi
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "yt/utilities/lib/geometry_utils.pxd":326
 * 
 * @cython.cdivision(True)
 * cdef inline np.uint64_t bounded_morton_split_relative_dds(np.float64_t x, np.float64_t y, np.float64_t z,             # <<<<<<<<<<<<<<
 *                                np.float64_t *DLE, np.float64_t *dds1, np.float64_t *dds2,
 *                                np.uint64_t *p2):
 */

static CYTHON_INLINE __pyx_t_5numpy_uint64_t __pyx_f_2yt_9utilities_3lib_14geometry_utils_bounded_morton_split_relative_dds(__pyx_t_5numpy_float64_t __pyx_v_x, __pyx_t_5numpy_float64_t __pyx_v_y, __pyx_t_5numpy_float64_t __pyx_v_z, __pyx_t_5numpy_float64_t *__pyx_v_DLE, __pyx_t_5numpy_float64_t *__pyx_v_dds1, __pyx_t_5numpy_float64_t *__pyx_v_dds2, __pyx_t_5numpy_uint64_t *__pyx_v_p2) {
  __pyx_t_5numpy_float64_t __pyx_v_DLE2[3];
  __pyx_t_5numpy_uint64_t __pyx_v_mi2;
  __pyx_t_5numpy_uint64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bounded_morton_split_relative_dds", 0);

  /* "yt/utilities/lib/geometry_utils.pxd":331
 *     cdef np.float64_t DLE2[3]
 *     cdef np.uint64_t mi2
 *     DLE2[0] = DLE[0] + <np.float64_t> (<np.uint64_t> ((x - DLE[0])/dds1[0])) * dds1[0]             # <<<<<<<<<<<<<<
 *     DLE2[1] = DLE[1] + <np.float64_t> (<np.uint64_t> ((y - DLE[1])/dds1[1])) * dds1[1]
 *     DLE2[2] = DLE[2] + <np.float64_t> (<np.uint64_t> ((z - DLE[2])/dds1[2])) * dds1[2]
 */
  (__pyx_v_DLE2[0]) = ((__pyx_v_DLE[0]) + (((__pyx_t_5numpy_float64_t)((__pyx_t_5numpy_uint64_t)((__pyx_v_x - (__pyx_v_DLE[0])) / (__pyx_v_dds1[0])))) * (__pyx_v_dds1[0])));

  /* "yt/utilities/lib/geometry_utils.pxd":332
 *     cdef np.uint64_t mi2
 *     DLE2[0] = DLE[0] + <np.float64_t> (<np.uint64_t> ((x - DLE[0])/dds1[0])) * dds1[0]
 *     DLE2[1] = DLE[1] + <np.float64_t> (<np.uint64_t> ((y - DLE[1])/dds1[1])) * dds1[1]             # <<<<<<<<<<<<<<
 *     DLE2[2] = DLE[2] + <np.float64_t> (<np.uint64_t> ((z - DLE[2])/dds1[2])) * dds1[2]
 *     p2[0] = <np.uint64_t> ((x - DLE2[0])/dds2[0])
 */
  (__pyx_v_DLE2[1]) = ((__pyx_v_DLE[1]) + (((__pyx_t_5numpy_float64_t)((__pyx_t_5numpy_uint64_t)((__pyx_v_y - (__pyx_v_DLE[1])) / (__pyx_v_dds1[1])))) * (__pyx_v_dds1[1])));

  /* "yt/utilities/lib/geometry_utils.pxd":333
 *     DLE2[0] = DLE[0] + <np.float64_t> (<np.uint64_t> ((x - DLE[0])/dds1[0])) * dds1[0]
 *     DLE2[1] = DLE[1] + <np.float64_t> (<np.uint64_t> ((y - DLE[1])/dds1[1])) * dds1[1]
 *     DLE2[2] = DLE[2] + <np.float64_t> (<np.uint64_t> ((z - DLE[2])/dds1[2])) * dds1[2]             # <<<<<<<<<<<<<<
 *     p2[0] = <np.uint64_t> ((x - DLE2[0])/dds2[0])
 *     p2[1] = <np.uint64_t> ((y - DLE2[1])/dds2[1])
 */
  (__pyx_v_DLE2[2]) = ((__pyx_v_DLE[2]) + (((__pyx_t_5numpy_float64_t)((__pyx_t_5numpy_uint64_t)((__pyx_v_z - (__pyx_v_DLE[2])) / (__pyx_v_dds1[2])))) * (__pyx_v_dds1[2])));

  /* "yt/utilities/lib/geometry_utils.pxd":334
 *     DLE2[1] = DLE[1] + <np.float64_t> (<np.uint64_t> ((y - DLE[1])/dds1[1])) * dds1[1]
 *     DLE2[2] = DLE[2] + <np.float64_t> (<np.uint64_t> ((z - DLE[2])/dds1[2])) * dds1[2]
 *     p2[0] = <np.uint64_t> ((x - DLE2[0])/dds2[0])             # <<<<<<<<<<<<<<
 *     p2[1] = <np.uint64_t> ((y - DLE2[1])/dds2[1])
 *     p2[2] = <np.uint64_t> ((z - DLE2[2])/dds2[2])
 */
  (__pyx_v_p2[0]) = ((__pyx_t_5numpy_uint64_t)((__pyx_v_x - (__pyx_v_DLE2[0])) / (__pyx_v_dds2[0])));

  /* "yt/utilities/lib/geometry_utils.pxd":335
 *     DLE2[2] = DLE[2] + <np.float64_t> (<np.uint64_t> ((z - DLE[2])/dds1[2])) * dds1[2]
 *     p2[0] = <np.uint64_t> ((x - DLE2[0])/dds2[0])
 *     p2[1] = <np.uint64_t> ((y - DLE2[1])/dds2[1])             # <<<<<<<<<<<<<<
 *     p2[2] = <np.uint64_t> ((z - DLE2[2])/dds2[2])
 *     mi2 = encode_morton_64bit(p2[0], p2[1], p2[2])
 */
  (__pyx_v_p2[1]) = ((__pyx_t_5numpy_uint64_t)((__pyx_v_y - (__pyx_v_DLE2[1])) / (__pyx_v_dds2[1])));

  /* "yt/utilities/lib/geometry_utils.pxd":336
 *     p2[0] = <np.uint64_t> ((x - DLE2[0])/dds2[0])
 *     p2[1] = <np.uint64_t> ((y - DLE2[1])/dds2[1])
 *     p2[2] = <np.uint64_t> ((z - DLE2[2])/dds2[2])             # <<<<<<<<<<<<<<
 *     mi2 = encode_morton_64bit(p2[0], p2[1], p2[2])
 *     return mi2
 */
  (__pyx_v_p2[2]) = ((__pyx_t_5numpy_uint64_t)((__pyx_v_z - (__pyx_v_DLE2[2])) / (__pyx_v_dds2[2])));

  /* "yt/utilities/lib/geometry_utils.pxd":337
 *     p2[1] = <np.uint64_t> ((y - DLE2[1])/dds2[1])
 *     p2[2] = <np.uint64_t> ((z - DLE2[2])/dds2[2])
 *     mi2 = encode_morton_64bit(p2[0], p2[1], p2[2])             # <<<<<<<<<<<<<<
 *     return mi2
 * 
 */
  __pyx_v_mi2 = __pyx_f_2yt_9utilities_3lib_14geometry_utils_encode_morton_64bit((__pyx_v_p2[0]), (__pyx_v_p2[1]), (__pyx_v_p2[2]));

  /* "yt/utilities/lib/geometry_utils.pxd":338
 *     p2[2] = <np.uint64_t> ((z - DLE2[2])/dds2[2])
 *     mi2 = encode_morton_64bit(p2[0], p2[1], p2[2])
 *     return mi2             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_mi2;
  goto __pyx_L0;

  /* "yt/utilities/lib/geometry_utils.pxd":326
 * 
 * @cython.cdivision(True)
 * cdef inline np.uint64_t bounded_morton_split_relative_dds(np.float64_t x, np.float64_t y, np.float64_t z,             # <<<<<<<<<<<<<<
 *                                np.float64_t *DLE, np.float64_t *dds1, np.float64_t *dds2,
 *                                np.uint64_t *p2):
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "selection_routines.pxd":87
 *     cdef public SelectorObject sel2
 * 
 * cdef inline np.float64_t _periodic_dist(np.float64_t x1, np.float64_t x2,             # <<<<<<<<<<<<<<
 *                                         np.float64_t dw, bint periodic) nogil:
 *     cdef np.float64_t rel = x1 - x2
 */

static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_8geometry_18selection_routines__periodic_dist(__pyx_t_5numpy_float64_t __pyx_v_x1, __pyx_t_5numpy_float64_t __pyx_v_x2, __pyx_t_5numpy_float64_t __pyx_v_dw, int __pyx_v_periodic) {
  __pyx_t_5numpy_float64_t __pyx_v_rel;
  __pyx_t_5numpy_float64_t __pyx_r;
  int __pyx_t_1;

  /* "selection_routines.pxd":89
 * cdef inline np.float64_t _periodic_dist(np.float64_t x1, np.float64_t x2,
 *                                         np.float64_t dw, bint periodic) nogil:
 *     cdef np.float64_t rel = x1 - x2             # <<<<<<<<<<<<<<
 *     if not periodic: return rel
 *     if rel > dw * 0.5:
 */
  __pyx_v_rel = (__pyx_v_x1 - __pyx_v_x2);

  /* "selection_routines.pxd":90
 *                                         np.float64_t dw, bint periodic) nogil:
 *     cdef np.float64_t rel = x1 - x2
 *     if not periodic: return rel             # <<<<<<<<<<<<<<
 *     if rel > dw * 0.5:
 *         rel -= dw
 */
  __pyx_t_1 = ((!(__pyx_v_periodic != 0)) != 0);
  if (__pyx_t_1) {
    __pyx_r = __pyx_v_rel;
    goto __pyx_L0;
  }

  /* "selection_routines.pxd":91
 *     cdef np.float64_t rel = x1 - x2
 *     if not periodic: return rel
 *     if rel > dw * 0.5:             # <<<<<<<<<<<<<<
 *         rel -= dw
 *     elif rel < -dw * 0.5:
 */
  __pyx_t_1 = ((__pyx_v_rel > (__pyx_v_dw * 0.5)) != 0);
  if (__pyx_t_1) {

    /* "selection_routines.pxd":92
 *     if not periodic: return rel
 *     if rel > dw * 0.5:
 *         rel -= dw             # <<<<<<<<<<<<<<
 *     elif rel < -dw * 0.5:
 *         rel += dw
 */
    __pyx_v_rel = (__pyx_v_rel - __pyx_v_dw);

    /* "selection_routines.pxd":91
 *     cdef np.float64_t rel = x1 - x2
 *     if not periodic: return rel
 *     if rel > dw * 0.5:             # <<<<<<<<<<<<<<
 *         rel -= dw
 *     elif rel < -dw * 0.5:
 */
    goto __pyx_L4;
  }

  /* "selection_routines.pxd":93
 *     if rel > dw * 0.5:
 *         rel -= dw
 *     elif rel < -dw * 0.5:             # <<<<<<<<<<<<<<
 *         rel += dw
 *     return rel
 */
  __pyx_t_1 = ((__pyx_v_rel < ((-__pyx_v_dw) * 0.5)) != 0);
  if (__pyx_t_1) {

    /* "selection_routines.pxd":94
 *         rel -= dw
 *     elif rel < -dw * 0.5:
 *         rel += dw             # <<<<<<<<<<<<<<
 *     return rel
 */
    __pyx_v_rel = (__pyx_v_rel + __pyx_v_dw);

    /* "selection_routines.pxd":93
 *     if rel > dw * 0.5:
 *         rel -= dw
 *     elif rel < -dw * 0.5:             # <<<<<<<<<<<<<<
 *         rel += dw
 *     return rel
 */
  }
  __pyx_L4:;

  /* "selection_routines.pxd":95
 *     elif rel < -dw * 0.5:
 *         rel += dw
 *     return rel             # <<<<<<<<<<<<<<
 */
  __pyx_r = __pyx_v_rel;
  goto __pyx_L0;

  /* "selection_routines.pxd":87
 *     cdef public SelectorObject sel2
 * 
 * cdef inline np.float64_t _periodic_dist(np.float64_t x1, np.float64_t x2,             # <<<<<<<<<<<<<<
 *                                         np.float64_t dw, bint periodic) nogil:
 *     cdef np.float64_t rel = x1 - x2
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "oct_container.pxd":48
 * 
 * cdef class OctObjectPool(ObjectPool):
 *     cdef inline OctAllocationContainer *get_cont(self, int i):             # <<<<<<<<<<<<<<
 *         return <OctAllocationContainer*> (&self.containers[i])
 * 
 */

static CYTHON_INLINE struct __pyx_t_2yt_8geometry_13oct_container_OctAllocationContainer *__pyx_f_2yt_8geometry_13oct_container_13OctObjectPool_get_cont(struct __pyx_obj_2yt_8geometry_13oct_container_OctObjectPool *__pyx_v_self, int __pyx_v_i) {
  struct __pyx_t_2yt_8geometry_13oct_container_OctAllocationContainer *__pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_cont", 0);

  /* "oct_container.pxd":49
 * cdef class OctObjectPool(ObjectPool):
 *     cdef inline OctAllocationContainer *get_cont(self, int i):
 *         return <OctAllocationContainer*> (&self.containers[i])             # <<<<<<<<<<<<<<
 * 
 * cdef OctList *OctList_append(OctList *list, Oct *o)
 */
  __pyx_r = ((struct __pyx_t_2yt_8geometry_13oct_container_OctAllocationContainer *)(&(__pyx_v_self->__pyx_base.containers[__pyx_v_i])));
  goto __pyx_L0;

  /* "oct_container.pxd":48
 * 
 * cdef class OctObjectPool(ObjectPool):
 *     cdef inline OctAllocationContainer *get_cont(self, int i):             # <<<<<<<<<<<<<<
 *         return <OctAllocationContainer*> (&self.containers[i])
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "oct_visitors.pxd":51
 *     cdef void visit(self, Oct*, np.uint8_t selected)
 * 
 *     cdef inline int oind(self):             # <<<<<<<<<<<<<<
 *         cdef int d = self.nz
 *         return (((self.ind[0]*d)+self.ind[1])*d+self.ind[2])
 */

static CYTHON_INLINE int __pyx_f_2yt_8geometry_12oct_visitors_10OctVisitor_oind(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *__pyx_v_self) {
  int __pyx_v_d;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_int8_t __pyx_t_1;
  __Pyx_RefNannySetupContext("oind", 0);

  /* "oct_visitors.pxd":52
 * 
 *     cdef inline int oind(self):
 *         cdef int d = self.nz             # <<<<<<<<<<<<<<
 *         return (((self.ind[0]*d)+self.ind[1])*d+self.ind[2])
 * 
 */
  __pyx_t_1 = __pyx_v_self->nz;
  __pyx_v_d = __pyx_t_1;

  /* "oct_visitors.pxd":53
 *     cdef inline int oind(self):
 *         cdef int d = self.nz
 *         return (((self.ind[0]*d)+self.ind[1])*d+self.ind[2])             # <<<<<<<<<<<<<<
 * 
 *     cdef inline int rind(self):
 */
  __pyx_r = (((((__pyx_v_self->ind[0]) * __pyx_v_d) + (__pyx_v_self->ind[1])) * __pyx_v_d) + (__pyx_v_self->ind[2]));
  goto __pyx_L0;

  /* "oct_visitors.pxd":51
 *     cdef void visit(self, Oct*, np.uint8_t selected)
 * 
 *     cdef inline int oind(self):             # <<<<<<<<<<<<<<
 *         cdef int d = self.nz
 *         return (((self.ind[0]*d)+self.ind[1])*d+self.ind[2])
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "oct_visitors.pxd":55
 *         return (((self.ind[0]*d)+self.ind[1])*d+self.ind[2])
 * 
 *     cdef inline int rind(self):             # <<<<<<<<<<<<<<
 *         cdef int d = self.nz
 *         return (((self.ind[2]*d)+self.ind[1])*d+self.ind[0])
 */

static CYTHON_INLINE int __pyx_f_2yt_8geometry_12oct_visitors_10OctVisitor_rind(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor *__pyx_v_self) {
  int __pyx_v_d;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_int8_t __pyx_t_1;
  __Pyx_RefNannySetupContext("rind", 0);

  /* "oct_visitors.pxd":56
 * 
 *     cdef inline int rind(self):
 *         cdef int d = self.nz             # <<<<<<<<<<<<<<
 *         return (((self.ind[2]*d)+self.ind[1])*d+self.ind[0])
 * 
 */
  __pyx_t_1 = __pyx_v_self->nz;
  __pyx_v_d = __pyx_t_1;

  /* "oct_visitors.pxd":57
 *     cdef inline int rind(self):
 *         cdef int d = self.nz
 *         return (((self.ind[2]*d)+self.ind[1])*d+self.ind[0])             # <<<<<<<<<<<<<<
 * 
 * cdef class CountTotalOcts(OctVisitor):
 */
  __pyx_r = (((((__pyx_v_self->ind[2]) * __pyx_v_d) + (__pyx_v_self->ind[1])) * __pyx_v_d) + (__pyx_v_self->ind[0]));
  goto __pyx_L0;

  /* "oct_visitors.pxd":55
 *         return (((self.ind[0]*d)+self.ind[1])*d+self.ind[2])
 * 
 *     cdef inline int rind(self):             # <<<<<<<<<<<<<<
 *         cdef int d = self.nz
 *         return (((self.ind[2]*d)+self.ind[1])*d+self.ind[0])
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "oct_visitors.pxd":142
 *     cdef np.uint64_t[:] morton_ind
 * 
 * cdef inline int cind(int i, int j, int k) nogil:             # <<<<<<<<<<<<<<
 *     # THIS ONLY WORKS FOR CHILDREN.  It is not general for zones.
 *     return (((i*2)+j)*2+k)
 */

static CYTHON_INLINE int __pyx_f_2yt_8geometry_12oct_visitors_cind(int __pyx_v_i, int __pyx_v_j, int __pyx_v_k) {
  int __pyx_r;

  /* "oct_visitors.pxd":144
 * cdef inline int cind(int i, int j, int k) nogil:
 *     # THIS ONLY WORKS FOR CHILDREN.  It is not general for zones.
 *     return (((i*2)+j)*2+k)             # <<<<<<<<<<<<<<
 * 
 * from oct_container cimport OctreeContainer
 */
  __pyx_r = ((((__pyx_v_i * 2) + __pyx_v_j) * 2) + __pyx_v_k);
  goto __pyx_L0;

  /* "oct_visitors.pxd":142
 *     cdef np.uint64_t[:] morton_ind
 * 
 * cdef inline int cind(int i, int j, int k) nogil:             # <<<<<<<<<<<<<<
 *     # THIS ONLY WORKS FOR CHILDREN.  It is not general for zones.
 *     return (((i*2)+j)*2+k)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "oct_visitors.pxd":165
 *     cdef void set_neighbour_info(self, Oct *o, int ishift[3])
 * 
 *     cdef inline np.uint8_t neighbour_rind(self):             # <<<<<<<<<<<<<<
 *         cdef int d = self.nz
 *         return (((self.neigh_ind[2]*d)+self.neigh_ind[1])*d+self.neigh_ind[0])
 */

static CYTHON_INLINE __pyx_t_5numpy_uint8_t __pyx_f_2yt_8geometry_12oct_visitors_20BaseNeighbourVisitor_neighbour_rind(struct __pyx_obj_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor *__pyx_v_self) {
  int __pyx_v_d;
  __pyx_t_5numpy_uint8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __pyx_t_5numpy_int8_t __pyx_t_1;
  __Pyx_RefNannySetupContext("neighbour_rind", 0);

  /* "oct_visitors.pxd":166
 * 
 *     cdef inline np.uint8_t neighbour_rind(self):
 *         cdef int d = self.nz             # <<<<<<<<<<<<<<
 *         return (((self.neigh_ind[2]*d)+self.neigh_ind[1])*d+self.neigh_ind[0])
 * 
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.nz;
  __pyx_v_d = __pyx_t_1;

  /* "oct_visitors.pxd":167
 *     cdef inline np.uint8_t neighbour_rind(self):
 *         cdef int d = self.nz
 *         return (((self.neigh_ind[2]*d)+self.neigh_ind[1])*d+self.neigh_ind[0])             # <<<<<<<<<<<<<<
 * 
 * cdef class NeighbourCellIndexVisitor(BaseNeighbourVisitor):
 */
  __pyx_r = (((((__pyx_v_self->neigh_ind[2]) * __pyx_v_d) + (__pyx_v_self->neigh_ind[1])) * __pyx_v_d) + (__pyx_v_self->neigh_ind[0]));
  goto __pyx_L0;

  /* "oct_visitors.pxd":165
 *     cdef void set_neighbour_info(self, Oct *o, int ishift[3])
 * 
 *     cdef inline np.uint8_t neighbour_rind(self):             # <<<<<<<<<<<<<<
 *         cdef int d = self.nz
 *         return (((self.neigh_ind[2]*d)+self.neigh_ind[1])*d+self.neigh_ind[0])
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "particle_deposit.pxd":27
 *     void *alloca(int)
 * 
 * cdef inline int gind(int i, int j, int k, int dims[3]):             # <<<<<<<<<<<<<<
 *     # The ordering is such that we want i to vary the slowest in this instance,
 *     # even though in other instances it varies the fastest.  To see this in
 */

static CYTHON_INLINE int __pyx_f_2yt_8geometry_16particle_deposit_gind(int __pyx_v_i, int __pyx_v_j, int __pyx_v_k, int *__pyx_v_dims) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("gind", 0);

  /* "particle_deposit.pxd":32
 *     # action, try looking at the results of an n_ref=256 particle CIC plot,
 *     # which shows it the most clearly.
 *     return ((i*dims[1])+j)*dims[2]+k             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((((__pyx_v_i * (__pyx_v_dims[1])) + __pyx_v_j) * (__pyx_v_dims[2])) + __pyx_v_k);
  goto __pyx_L0;

  /* "particle_deposit.pxd":27
 *     void *alloca(int)
 * 
 * cdef inline int gind(int i, int j, int k, int dims[3]):             # <<<<<<<<<<<<<<
 *     # The ordering is such that we want i to vary the slowest in this instance,
 *     # even though in other instances it varies the fastest.  To see this in
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "particle_deposit.pxd":39
 * ####################################################
 * 
 * cdef inline np.float64_t sph_kernel_cubic(np.float64_t x) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.float64_t kernel
 *     # C is 8/pi
 */

static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_cubic(__pyx_t_5numpy_float64_t __pyx_v_x) {
  __pyx_t_5numpy_float64_t __pyx_v_kernel;
  __pyx_t_5numpy_float64_t __pyx_v_C;
  __pyx_t_5numpy_float64_t __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "particle_deposit.pxd":42
 *     cdef np.float64_t kernel
 *     # C is 8/pi
 *     cdef np.float64_t C = 2.5464790894703255             # <<<<<<<<<<<<<<
 *     if x <= 0.5:
 *         kernel = 1.-6.*x*x*(1.-x)
 */
  __pyx_v_C = 2.5464790894703255;

  /* "particle_deposit.pxd":43
 *     # C is 8/pi
 *     cdef np.float64_t C = 2.5464790894703255
 *     if x <= 0.5:             # <<<<<<<<<<<<<<
 *         kernel = 1.-6.*x*x*(1.-x)
 *     elif x>0.5 and x<=1.0:
 */
  __pyx_t_1 = ((__pyx_v_x <= 0.5) != 0);
  if (__pyx_t_1) {

    /* "particle_deposit.pxd":44
 *     cdef np.float64_t C = 2.5464790894703255
 *     if x <= 0.5:
 *         kernel = 1.-6.*x*x*(1.-x)             # <<<<<<<<<<<<<<
 *     elif x>0.5 and x<=1.0:
 *         kernel = 2.*(1.-x)*(1.-x)*(1.-x)
 */
    __pyx_v_kernel = (1. - (((6. * __pyx_v_x) * __pyx_v_x) * (1. - __pyx_v_x)));

    /* "particle_deposit.pxd":43
 *     # C is 8/pi
 *     cdef np.float64_t C = 2.5464790894703255
 *     if x <= 0.5:             # <<<<<<<<<<<<<<
 *         kernel = 1.-6.*x*x*(1.-x)
 *     elif x>0.5 and x<=1.0:
 */
    goto __pyx_L3;
  }

  /* "particle_deposit.pxd":45
 *     if x <= 0.5:
 *         kernel = 1.-6.*x*x*(1.-x)
 *     elif x>0.5 and x<=1.0:             # <<<<<<<<<<<<<<
 *         kernel = 2.*(1.-x)*(1.-x)*(1.-x)
 *     else:
 */
  __pyx_t_2 = ((__pyx_v_x > 0.5) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_x <= 1.0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "particle_deposit.pxd":46
 *         kernel = 1.-6.*x*x*(1.-x)
 *     elif x>0.5 and x<=1.0:
 *         kernel = 2.*(1.-x)*(1.-x)*(1.-x)             # <<<<<<<<<<<<<<
 *     else:
 *         kernel = 0.
 */
    __pyx_v_kernel = (((2. * (1. - __pyx_v_x)) * (1. - __pyx_v_x)) * (1. - __pyx_v_x));

    /* "particle_deposit.pxd":45
 *     if x <= 0.5:
 *         kernel = 1.-6.*x*x*(1.-x)
 *     elif x>0.5 and x<=1.0:             # <<<<<<<<<<<<<<
 *         kernel = 2.*(1.-x)*(1.-x)*(1.-x)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "particle_deposit.pxd":48
 *         kernel = 2.*(1.-x)*(1.-x)*(1.-x)
 *     else:
 *         kernel = 0.             # <<<<<<<<<<<<<<
 *     return kernel * C
 * 
 */
  /*else*/ {
    __pyx_v_kernel = 0.;
  }
  __pyx_L3:;

  /* "particle_deposit.pxd":49
 *     else:
 *         kernel = 0.
 *     return kernel * C             # <<<<<<<<<<<<<<
 * 
 * ########################################################
 */
  __pyx_r = (__pyx_v_kernel * __pyx_v_C);
  goto __pyx_L0;

  /* "particle_deposit.pxd":39
 * ####################################################
 * 
 * cdef inline np.float64_t sph_kernel_cubic(np.float64_t x) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.float64_t kernel
 *     # C is 8/pi
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "particle_deposit.pxd":56
 * 
 * # quartic spline
 * cdef inline np.float64_t sph_kernel_quartic(np.float64_t x) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 9.71404681957369  # 5.**6/512/np.pi
 */

static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_quartic(__pyx_t_5numpy_float64_t __pyx_v_x) {
  __pyx_t_5numpy_float64_t __pyx_v_kernel;
  __pyx_t_5numpy_float64_t __pyx_v_C;
  __pyx_t_5numpy_float64_t __pyx_r;
  int __pyx_t_1;

  /* "particle_deposit.pxd":58
 * cdef inline np.float64_t sph_kernel_quartic(np.float64_t x) nogil:
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 9.71404681957369  # 5.**6/512/np.pi             # <<<<<<<<<<<<<<
 *     if x < 1:
 *         kernel = (1.-x)**4
 */
  __pyx_v_C = 9.71404681957369;

  /* "particle_deposit.pxd":59
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 9.71404681957369  # 5.**6/512/np.pi
 *     if x < 1:             # <<<<<<<<<<<<<<
 *         kernel = (1.-x)**4
 *         if x < 3./5:
 */
  __pyx_t_1 = ((__pyx_v_x < 1.0) != 0);
  if (__pyx_t_1) {

    /* "particle_deposit.pxd":60
 *     cdef np.float64_t C = 9.71404681957369  # 5.**6/512/np.pi
 *     if x < 1:
 *         kernel = (1.-x)**4             # <<<<<<<<<<<<<<
 *         if x < 3./5:
 *             kernel -= 5*(3./5-x)**4
 */
    __pyx_v_kernel = pow((1. - __pyx_v_x), 4.0);

    /* "particle_deposit.pxd":61
 *     if x < 1:
 *         kernel = (1.-x)**4
 *         if x < 3./5:             # <<<<<<<<<<<<<<
 *             kernel -= 5*(3./5-x)**4
 *             if x < 1./5:
 */
    __pyx_t_1 = ((__pyx_v_x < (3. / 5.0)) != 0);
    if (__pyx_t_1) {

      /* "particle_deposit.pxd":62
 *         kernel = (1.-x)**4
 *         if x < 3./5:
 *             kernel -= 5*(3./5-x)**4             # <<<<<<<<<<<<<<
 *             if x < 1./5:
 *                 kernel += 10*(1./5-x)**4
 */
      __pyx_v_kernel = (__pyx_v_kernel - (5.0 * pow(((3. / 5.0) - __pyx_v_x), 4.0)));

      /* "particle_deposit.pxd":63
 *         if x < 3./5:
 *             kernel -= 5*(3./5-x)**4
 *             if x < 1./5:             # <<<<<<<<<<<<<<
 *                 kernel += 10*(1./5-x)**4
 *     else:
 */
      __pyx_t_1 = ((__pyx_v_x < (1. / 5.0)) != 0);
      if (__pyx_t_1) {

        /* "particle_deposit.pxd":64
 *             kernel -= 5*(3./5-x)**4
 *             if x < 1./5:
 *                 kernel += 10*(1./5-x)**4             # <<<<<<<<<<<<<<
 *     else:
 *         kernel = 0.
 */
        __pyx_v_kernel = (__pyx_v_kernel + (10.0 * pow(((1. / 5.0) - __pyx_v_x), 4.0)));

        /* "particle_deposit.pxd":63
 *         if x < 3./5:
 *             kernel -= 5*(3./5-x)**4
 *             if x < 1./5:             # <<<<<<<<<<<<<<
 *                 kernel += 10*(1./5-x)**4
 *     else:
 */
      }

      /* "particle_deposit.pxd":61
 *     if x < 1:
 *         kernel = (1.-x)**4
 *         if x < 3./5:             # <<<<<<<<<<<<<<
 *             kernel -= 5*(3./5-x)**4
 *             if x < 1./5:
 */
    }

    /* "particle_deposit.pxd":59
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 9.71404681957369  # 5.**6/512/np.pi
 *     if x < 1:             # <<<<<<<<<<<<<<
 *         kernel = (1.-x)**4
 *         if x < 3./5:
 */
    goto __pyx_L3;
  }

  /* "particle_deposit.pxd":66
 *                 kernel += 10*(1./5-x)**4
 *     else:
 *         kernel = 0.             # <<<<<<<<<<<<<<
 *     return kernel * C
 * 
 */
  /*else*/ {
    __pyx_v_kernel = 0.;
  }
  __pyx_L3:;

  /* "particle_deposit.pxd":67
 *     else:
 *         kernel = 0.
 *     return kernel * C             # <<<<<<<<<<<<<<
 * 
 * # quintic spline
 */
  __pyx_r = (__pyx_v_kernel * __pyx_v_C);
  goto __pyx_L0;

  /* "particle_deposit.pxd":56
 * 
 * # quartic spline
 * cdef inline np.float64_t sph_kernel_quartic(np.float64_t x) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 9.71404681957369  # 5.**6/512/np.pi
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "particle_deposit.pxd":70
 * 
 * # quintic spline
 * cdef inline np.float64_t sph_kernel_quintic(np.float64_t x) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 17.403593027098754  # 3.**7/40/np.pi
 */

static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_quintic(__pyx_t_5numpy_float64_t __pyx_v_x) {
  __pyx_t_5numpy_float64_t __pyx_v_kernel;
  __pyx_t_5numpy_float64_t __pyx_v_C;
  __pyx_t_5numpy_float64_t __pyx_r;
  int __pyx_t_1;

  /* "particle_deposit.pxd":72
 * cdef inline np.float64_t sph_kernel_quintic(np.float64_t x) nogil:
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 17.403593027098754  # 3.**7/40/np.pi             # <<<<<<<<<<<<<<
 *     if x < 1:
 *         kernel = (1.-x)**5
 */
  __pyx_v_C = 17.403593027098754;

  /* "particle_deposit.pxd":73
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 17.403593027098754  # 3.**7/40/np.pi
 *     if x < 1:             # <<<<<<<<<<<<<<
 *         kernel = (1.-x)**5
 *         if x < 2./3:
 */
  __pyx_t_1 = ((__pyx_v_x < 1.0) != 0);
  if (__pyx_t_1) {

    /* "particle_deposit.pxd":74
 *     cdef np.float64_t C = 17.403593027098754  # 3.**7/40/np.pi
 *     if x < 1:
 *         kernel = (1.-x)**5             # <<<<<<<<<<<<<<
 *         if x < 2./3:
 *             kernel -= 6*(2./3-x)**5
 */
    __pyx_v_kernel = pow((1. - __pyx_v_x), 5.0);

    /* "particle_deposit.pxd":75
 *     if x < 1:
 *         kernel = (1.-x)**5
 *         if x < 2./3:             # <<<<<<<<<<<<<<
 *             kernel -= 6*(2./3-x)**5
 *             if x < 1./3:
 */
    __pyx_t_1 = ((__pyx_v_x < (2. / 3.0)) != 0);
    if (__pyx_t_1) {

      /* "particle_deposit.pxd":76
 *         kernel = (1.-x)**5
 *         if x < 2./3:
 *             kernel -= 6*(2./3-x)**5             # <<<<<<<<<<<<<<
 *             if x < 1./3:
 *                 kernel += 15*(1./3-x)**5
 */
      __pyx_v_kernel = (__pyx_v_kernel - (6.0 * pow(((2. / 3.0) - __pyx_v_x), 5.0)));

      /* "particle_deposit.pxd":77
 *         if x < 2./3:
 *             kernel -= 6*(2./3-x)**5
 *             if x < 1./3:             # <<<<<<<<<<<<<<
 *                 kernel += 15*(1./3-x)**5
 *     else:
 */
      __pyx_t_1 = ((__pyx_v_x < (1. / 3.0)) != 0);
      if (__pyx_t_1) {

        /* "particle_deposit.pxd":78
 *             kernel -= 6*(2./3-x)**5
 *             if x < 1./3:
 *                 kernel += 15*(1./3-x)**5             # <<<<<<<<<<<<<<
 *     else:
 *         kernel = 0.
 */
        __pyx_v_kernel = (__pyx_v_kernel + (15.0 * pow(((1. / 3.0) - __pyx_v_x), 5.0)));

        /* "particle_deposit.pxd":77
 *         if x < 2./3:
 *             kernel -= 6*(2./3-x)**5
 *             if x < 1./3:             # <<<<<<<<<<<<<<
 *                 kernel += 15*(1./3-x)**5
 *     else:
 */
      }

      /* "particle_deposit.pxd":75
 *     if x < 1:
 *         kernel = (1.-x)**5
 *         if x < 2./3:             # <<<<<<<<<<<<<<
 *             kernel -= 6*(2./3-x)**5
 *             if x < 1./3:
 */
    }

    /* "particle_deposit.pxd":73
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 17.403593027098754  # 3.**7/40/np.pi
 *     if x < 1:             # <<<<<<<<<<<<<<
 *         kernel = (1.-x)**5
 *         if x < 2./3:
 */
    goto __pyx_L3;
  }

  /* "particle_deposit.pxd":80
 *                 kernel += 15*(1./3-x)**5
 *     else:
 *         kernel = 0.             # <<<<<<<<<<<<<<
 *     return kernel * C
 * 
 */
  /*else*/ {
    __pyx_v_kernel = 0.;
  }
  __pyx_L3:;

  /* "particle_deposit.pxd":81
 *     else:
 *         kernel = 0.
 *     return kernel * C             # <<<<<<<<<<<<<<
 * 
 * # Wendland C2
 */
  __pyx_r = (__pyx_v_kernel * __pyx_v_C);
  goto __pyx_L0;

  /* "particle_deposit.pxd":70
 * 
 * # quintic spline
 * cdef inline np.float64_t sph_kernel_quintic(np.float64_t x) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 17.403593027098754  # 3.**7/40/np.pi
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "particle_deposit.pxd":84
 * 
 * # Wendland C2
 * cdef inline np.float64_t sph_kernel_wendland2(np.float64_t x) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 3.3422538049298023  # 21./2/np.pi
 */

static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_wendland2(__pyx_t_5numpy_float64_t __pyx_v_x) {
  __pyx_t_5numpy_float64_t __pyx_v_kernel;
  __pyx_t_5numpy_float64_t __pyx_v_C;
  __pyx_t_5numpy_float64_t __pyx_r;
  int __pyx_t_1;

  /* "particle_deposit.pxd":86
 * cdef inline np.float64_t sph_kernel_wendland2(np.float64_t x) nogil:
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 3.3422538049298023  # 21./2/np.pi             # <<<<<<<<<<<<<<
 *     if x < 1:
 *         kernel = (1.-x)**4 * (1+4*x)
 */
  __pyx_v_C = 3.3422538049298023;

  /* "particle_deposit.pxd":87
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 3.3422538049298023  # 21./2/np.pi
 *     if x < 1:             # <<<<<<<<<<<<<<
 *         kernel = (1.-x)**4 * (1+4*x)
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_x < 1.0) != 0);
  if (__pyx_t_1) {

    /* "particle_deposit.pxd":88
 *     cdef np.float64_t C = 3.3422538049298023  # 21./2/np.pi
 *     if x < 1:
 *         kernel = (1.-x)**4 * (1+4*x)             # <<<<<<<<<<<<<<
 *     else:
 *         kernel = 0.
 */
    __pyx_v_kernel = (pow((1. - __pyx_v_x), 4.0) * (1.0 + (4.0 * __pyx_v_x)));

    /* "particle_deposit.pxd":87
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 3.3422538049298023  # 21./2/np.pi
 *     if x < 1:             # <<<<<<<<<<<<<<
 *         kernel = (1.-x)**4 * (1+4*x)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "particle_deposit.pxd":90
 *         kernel = (1.-x)**4 * (1+4*x)
 *     else:
 *         kernel = 0.             # <<<<<<<<<<<<<<
 *     return kernel * C
 * 
 */
  /*else*/ {
    __pyx_v_kernel = 0.;
  }
  __pyx_L3:;

  /* "particle_deposit.pxd":91
 *     else:
 *         kernel = 0.
 *     return kernel * C             # <<<<<<<<<<<<<<
 * 
 * # Wendland C4
 */
  __pyx_r = (__pyx_v_kernel * __pyx_v_C);
  goto __pyx_L0;

  /* "particle_deposit.pxd":84
 * 
 * # Wendland C2
 * cdef inline np.float64_t sph_kernel_wendland2(np.float64_t x) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 3.3422538049298023  # 21./2/np.pi
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "particle_deposit.pxd":94
 * 
 * # Wendland C4
 * cdef inline np.float64_t sph_kernel_wendland4(np.float64_t x) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 4.923856051905513  # 495./32/np.pi
 */

static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_wendland4(__pyx_t_5numpy_float64_t __pyx_v_x) {
  __pyx_t_5numpy_float64_t __pyx_v_kernel;
  __pyx_t_5numpy_float64_t __pyx_v_C;
  __pyx_t_5numpy_float64_t __pyx_r;
  int __pyx_t_1;

  /* "particle_deposit.pxd":96
 * cdef inline np.float64_t sph_kernel_wendland4(np.float64_t x) nogil:
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 4.923856051905513  # 495./32/np.pi             # <<<<<<<<<<<<<<
 *     if x < 1:
 *         kernel = (1.-x)**6 * (1+6*x+35./3*x**2)
 */
  __pyx_v_C = 4.923856051905513;

  /* "particle_deposit.pxd":97
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 4.923856051905513  # 495./32/np.pi
 *     if x < 1:             # <<<<<<<<<<<<<<
 *         kernel = (1.-x)**6 * (1+6*x+35./3*x**2)
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_x < 1.0) != 0);
  if (__pyx_t_1) {

    /* "particle_deposit.pxd":98
 *     cdef np.float64_t C = 4.923856051905513  # 495./32/np.pi
 *     if x < 1:
 *         kernel = (1.-x)**6 * (1+6*x+35./3*x**2)             # <<<<<<<<<<<<<<
 *     else:
 *         kernel = 0.
 */
    __pyx_v_kernel = (pow((1. - __pyx_v_x), 6.0) * ((1.0 + (6.0 * __pyx_v_x)) + ((35. / 3.0) * pow(__pyx_v_x, 2.0))));

    /* "particle_deposit.pxd":97
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 4.923856051905513  # 495./32/np.pi
 *     if x < 1:             # <<<<<<<<<<<<<<
 *         kernel = (1.-x)**6 * (1+6*x+35./3*x**2)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "particle_deposit.pxd":100
 *         kernel = (1.-x)**6 * (1+6*x+35./3*x**2)
 *     else:
 *         kernel = 0.             # <<<<<<<<<<<<<<
 *     return kernel * C
 * 
 */
  /*else*/ {
    __pyx_v_kernel = 0.;
  }
  __pyx_L3:;

  /* "particle_deposit.pxd":101
 *     else:
 *         kernel = 0.
 *     return kernel * C             # <<<<<<<<<<<<<<
 * 
 * # Wendland C6
 */
  __pyx_r = (__pyx_v_kernel * __pyx_v_C);
  goto __pyx_L0;

  /* "particle_deposit.pxd":94
 * 
 * # Wendland C4
 * cdef inline np.float64_t sph_kernel_wendland4(np.float64_t x) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 4.923856051905513  # 495./32/np.pi
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "particle_deposit.pxd":104
 * 
 * # Wendland C6
 * cdef inline np.float64_t sph_kernel_wendland6(np.float64_t x) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 6.78895304126366  # 1365./64/np.pi
 */

static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_wendland6(__pyx_t_5numpy_float64_t __pyx_v_x) {
  __pyx_t_5numpy_float64_t __pyx_v_kernel;
  __pyx_t_5numpy_float64_t __pyx_v_C;
  __pyx_t_5numpy_float64_t __pyx_r;
  int __pyx_t_1;

  /* "particle_deposit.pxd":106
 * cdef inline np.float64_t sph_kernel_wendland6(np.float64_t x) nogil:
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 6.78895304126366  # 1365./64/np.pi             # <<<<<<<<<<<<<<
 *     if x < 1:
 *         kernel = (1.-x)**8 * (1+8*x+25*x**2+32*x**3)
 */
  __pyx_v_C = 6.78895304126366;

  /* "particle_deposit.pxd":107
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 6.78895304126366  # 1365./64/np.pi
 *     if x < 1:             # <<<<<<<<<<<<<<
 *         kernel = (1.-x)**8 * (1+8*x+25*x**2+32*x**3)
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_x < 1.0) != 0);
  if (__pyx_t_1) {

    /* "particle_deposit.pxd":108
 *     cdef np.float64_t C = 6.78895304126366  # 1365./64/np.pi
 *     if x < 1:
 *         kernel = (1.-x)**8 * (1+8*x+25*x**2+32*x**3)             # <<<<<<<<<<<<<<
 *     else:
 *         kernel = 0.
 */
    __pyx_v_kernel = (pow((1. - __pyx_v_x), 8.0) * (((1.0 + (8.0 * __pyx_v_x)) + (25.0 * pow(__pyx_v_x, 2.0))) + (32.0 * pow(__pyx_v_x, 3.0))));

    /* "particle_deposit.pxd":107
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 6.78895304126366  # 1365./64/np.pi
 *     if x < 1:             # <<<<<<<<<<<<<<
 *         kernel = (1.-x)**8 * (1+8*x+25*x**2+32*x**3)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "particle_deposit.pxd":110
 *         kernel = (1.-x)**8 * (1+8*x+25*x**2+32*x**3)
 *     else:
 *         kernel = 0.             # <<<<<<<<<<<<<<
 *     return kernel * C
 * 
 */
  /*else*/ {
    __pyx_v_kernel = 0.;
  }
  __pyx_L3:;

  /* "particle_deposit.pxd":111
 *     else:
 *         kernel = 0.
 *     return kernel * C             # <<<<<<<<<<<<<<
 * 
 * cdef inline np.float64_t sph_kernel_dummy(np.float64_t x) nogil:
 */
  __pyx_r = (__pyx_v_kernel * __pyx_v_C);
  goto __pyx_L0;

  /* "particle_deposit.pxd":104
 * 
 * # Wendland C6
 * cdef inline np.float64_t sph_kernel_wendland6(np.float64_t x) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.float64_t kernel
 *     cdef np.float64_t C = 6.78895304126366  # 1365./64/np.pi
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "particle_deposit.pxd":113
 *     return kernel * C
 * 
 * cdef inline np.float64_t sph_kernel_dummy(np.float64_t x) nogil:             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */

static CYTHON_INLINE __pyx_t_5numpy_float64_t __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_dummy(CYTHON_UNUSED __pyx_t_5numpy_float64_t __pyx_v_x) {
  __pyx_t_5numpy_float64_t __pyx_r;

  /* "particle_deposit.pxd":114
 * 
 * cdef inline np.float64_t sph_kernel_dummy(np.float64_t x) nogil:
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * # I don't know the way to use a dict in a cdef class.
 */
  __pyx_r = 0.0;
  goto __pyx_L0;

  /* "particle_deposit.pxd":113
 *     return kernel * C
 * 
 * cdef inline np.float64_t sph_kernel_dummy(np.float64_t x) nogil:             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "particle_deposit.pxd":120
 * # I manually created a function to lookup the kernel functions.
 * ctypedef np.float64_t (*kernel_func) (np.float64_t) nogil
 * cdef inline kernel_func get_kernel_func(str kernel_name) nogil:             # <<<<<<<<<<<<<<
 *     with gil:
 *         if kernel_name == 'cubic':
 */

static CYTHON_INLINE __pyx_t_2yt_8geometry_16particle_deposit_kernel_func __pyx_f_2yt_8geometry_16particle_deposit_get_kernel_func(PyObject *__pyx_v_kernel_name) {
  __pyx_t_2yt_8geometry_16particle_deposit_kernel_func __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("get_kernel_func", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "particle_deposit.pxd":121
 * ctypedef np.float64_t (*kernel_func) (np.float64_t) nogil
 * cdef inline kernel_func get_kernel_func(str kernel_name) nogil:
 *     with gil:             # <<<<<<<<<<<<<<
 *         if kernel_name == 'cubic':
 *             return sph_kernel_cubic
 */
  /*try:*/ {
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "particle_deposit.pxd":122
 * cdef inline kernel_func get_kernel_func(str kernel_name) nogil:
 *     with gil:
 *         if kernel_name == 'cubic':             # <<<<<<<<<<<<<<
 *             return sph_kernel_cubic
 *         elif kernel_name == 'quartic':
 */
          __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_v_kernel_name, __pyx_n_s_cubic, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(5, 122, __pyx_L7_error)
          __pyx_t_2 = (__pyx_t_1 != 0);
          if (__pyx_t_2) {

            /* "particle_deposit.pxd":123
 *     with gil:
 *         if kernel_name == 'cubic':
 *             return sph_kernel_cubic             # <<<<<<<<<<<<<<
 *         elif kernel_name == 'quartic':
 *             return sph_kernel_quartic
 */
            __pyx_r = __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_cubic;
            goto __pyx_L6_return;

            /* "particle_deposit.pxd":122
 * cdef inline kernel_func get_kernel_func(str kernel_name) nogil:
 *     with gil:
 *         if kernel_name == 'cubic':             # <<<<<<<<<<<<<<
 *             return sph_kernel_cubic
 *         elif kernel_name == 'quartic':
 */
          }

          /* "particle_deposit.pxd":124
 *         if kernel_name == 'cubic':
 *             return sph_kernel_cubic
 *         elif kernel_name == 'quartic':             # <<<<<<<<<<<<<<
 *             return sph_kernel_quartic
 *         elif kernel_name == 'quintic':
 */
          __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_kernel_name, __pyx_n_s_quartic, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(5, 124, __pyx_L7_error)
          __pyx_t_1 = (__pyx_t_2 != 0);
          if (__pyx_t_1) {

            /* "particle_deposit.pxd":125
 *             return sph_kernel_cubic
 *         elif kernel_name == 'quartic':
 *             return sph_kernel_quartic             # <<<<<<<<<<<<<<
 *         elif kernel_name == 'quintic':
 *             return sph_kernel_quintic
 */
            __pyx_r = __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_quartic;
            goto __pyx_L6_return;

            /* "particle_deposit.pxd":124
 *         if kernel_name == 'cubic':
 *             return sph_kernel_cubic
 *         elif kernel_name == 'quartic':             # <<<<<<<<<<<<<<
 *             return sph_kernel_quartic
 *         elif kernel_name == 'quintic':
 */
          }

          /* "particle_deposit.pxd":126
 *         elif kernel_name == 'quartic':
 *             return sph_kernel_quartic
 *         elif kernel_name == 'quintic':             # <<<<<<<<<<<<<<
 *             return sph_kernel_quintic
 *         elif kernel_name == 'wendland2':
 */
          __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_v_kernel_name, __pyx_n_s_quintic, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(5, 126, __pyx_L7_error)
          __pyx_t_2 = (__pyx_t_1 != 0);
          if (__pyx_t_2) {

            /* "particle_deposit.pxd":127
 *             return sph_kernel_quartic
 *         elif kernel_name == 'quintic':
 *             return sph_kernel_quintic             # <<<<<<<<<<<<<<
 *         elif kernel_name == 'wendland2':
 *             return sph_kernel_wendland2
 */
            __pyx_r = __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_quintic;
            goto __pyx_L6_return;

            /* "particle_deposit.pxd":126
 *         elif kernel_name == 'quartic':
 *             return sph_kernel_quartic
 *         elif kernel_name == 'quintic':             # <<<<<<<<<<<<<<
 *             return sph_kernel_quintic
 *         elif kernel_name == 'wendland2':
 */
          }

          /* "particle_deposit.pxd":128
 *         elif kernel_name == 'quintic':
 *             return sph_kernel_quintic
 *         elif kernel_name == 'wendland2':             # <<<<<<<<<<<<<<
 *             return sph_kernel_wendland2
 *         elif kernel_name == 'wendland4':
 */
          __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_kernel_name, __pyx_n_s_wendland2, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(5, 128, __pyx_L7_error)
          __pyx_t_1 = (__pyx_t_2 != 0);
          if (__pyx_t_1) {

            /* "particle_deposit.pxd":129
 *             return sph_kernel_quintic
 *         elif kernel_name == 'wendland2':
 *             return sph_kernel_wendland2             # <<<<<<<<<<<<<<
 *         elif kernel_name == 'wendland4':
 *             return sph_kernel_wendland4
 */
            __pyx_r = __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_wendland2;
            goto __pyx_L6_return;

            /* "particle_deposit.pxd":128
 *         elif kernel_name == 'quintic':
 *             return sph_kernel_quintic
 *         elif kernel_name == 'wendland2':             # <<<<<<<<<<<<<<
 *             return sph_kernel_wendland2
 *         elif kernel_name == 'wendland4':
 */
          }

          /* "particle_deposit.pxd":130
 *         elif kernel_name == 'wendland2':
 *             return sph_kernel_wendland2
 *         elif kernel_name == 'wendland4':             # <<<<<<<<<<<<<<
 *             return sph_kernel_wendland4
 *         elif kernel_name == 'wendland6':
 */
          __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_v_kernel_name, __pyx_n_s_wendland4, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(5, 130, __pyx_L7_error)
          __pyx_t_2 = (__pyx_t_1 != 0);
          if (__pyx_t_2) {

            /* "particle_deposit.pxd":131
 *             return sph_kernel_wendland2
 *         elif kernel_name == 'wendland4':
 *             return sph_kernel_wendland4             # <<<<<<<<<<<<<<
 *         elif kernel_name == 'wendland6':
 *             return sph_kernel_wendland6
 */
            __pyx_r = __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_wendland4;
            goto __pyx_L6_return;

            /* "particle_deposit.pxd":130
 *         elif kernel_name == 'wendland2':
 *             return sph_kernel_wendland2
 *         elif kernel_name == 'wendland4':             # <<<<<<<<<<<<<<
 *             return sph_kernel_wendland4
 *         elif kernel_name == 'wendland6':
 */
          }

          /* "particle_deposit.pxd":132
 *         elif kernel_name == 'wendland4':
 *             return sph_kernel_wendland4
 *         elif kernel_name == 'wendland6':             # <<<<<<<<<<<<<<
 *             return sph_kernel_wendland6
 *         elif kernel_name == 'none':
 */
          __pyx_t_2 = (__Pyx_PyString_Equals(__pyx_v_kernel_name, __pyx_n_s_wendland6, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(5, 132, __pyx_L7_error)
          __pyx_t_1 = (__pyx_t_2 != 0);
          if (__pyx_t_1) {

            /* "particle_deposit.pxd":133
 *             return sph_kernel_wendland4
 *         elif kernel_name == 'wendland6':
 *             return sph_kernel_wendland6             # <<<<<<<<<<<<<<
 *         elif kernel_name == 'none':
 *             return sph_kernel_dummy
 */
            __pyx_r = __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_wendland6;
            goto __pyx_L6_return;

            /* "particle_deposit.pxd":132
 *         elif kernel_name == 'wendland4':
 *             return sph_kernel_wendland4
 *         elif kernel_name == 'wendland6':             # <<<<<<<<<<<<<<
 *             return sph_kernel_wendland6
 *         elif kernel_name == 'none':
 */
          }

          /* "particle_deposit.pxd":134
 *         elif kernel_name == 'wendland6':
 *             return sph_kernel_wendland6
 *         elif kernel_name == 'none':             # <<<<<<<<<<<<<<
 *             return sph_kernel_dummy
 *         else:
 */
          __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_v_kernel_name, __pyx_n_s_none, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(5, 134, __pyx_L7_error)
          __pyx_t_2 = (__pyx_t_1 != 0);
          if (likely(__pyx_t_2)) {

            /* "particle_deposit.pxd":135
 *             return sph_kernel_wendland6
 *         elif kernel_name == 'none':
 *             return sph_kernel_dummy             # <<<<<<<<<<<<<<
 *         else:
 *             raise NotImplementedError
 */
            __pyx_r = __pyx_f_2yt_8geometry_16particle_deposit_sph_kernel_dummy;
            goto __pyx_L6_return;

            /* "particle_deposit.pxd":134
 *         elif kernel_name == 'wendland6':
 *             return sph_kernel_wendland6
 *         elif kernel_name == 'none':             # <<<<<<<<<<<<<<
 *             return sph_kernel_dummy
 *         else:
 */
          }

          /* "particle_deposit.pxd":137
 *             return sph_kernel_dummy
 *         else:
 *             raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 * cdef class ParticleDepositOperation:
 */
          /*else*/ {
            __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
            __PYX_ERR(5, 137, __pyx_L7_error)
          }
        }

        /* "particle_deposit.pxd":121
 * ctypedef np.float64_t (*kernel_func) (np.float64_t) nogil
 * cdef inline kernel_func get_kernel_func(str kernel_name) nogil:
 *     with gil:             # <<<<<<<<<<<<<<
 *         if kernel_name == 'cubic':
 *             return sph_kernel_cubic
 */
        /*finally:*/ {
          __pyx_L6_return: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L3_return;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
        }
    }
  }
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "particle_deposit.pxd":120
 * # I manually created a function to lookup the kernel functions.
 * ctypedef np.float64_t (*kernel_func) (np.float64_t) nogil
 * cdef inline kernel_func get_kernel_func(str kernel_name) nogil:             # <<<<<<<<<<<<<<
 *     with gil:
 *         if kernel_name == 'cubic':
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("yt.geometry.particle_deposit.get_kernel_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_size_t")
 * cdef object __pyx_convert_vector_to_py_size_t(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

static PyObject *__pyx_convert_vector_to_py_size_t(const std::vector<size_t>  &__pyx_v_v) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py_size_t", 0);

  /* "vector.to_py":61
 * @cname("__pyx_convert_vector_to_py_size_t")
 * cdef object __pyx_convert_vector_to_py_size_t(vector[X]& v):
 *     return [v[i] for i in range(v.size())]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_v.size();
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;
    __pyx_t_5 = __Pyx_PyInt_FromSize_t((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_size_t")
 * cdef object __pyx_convert_vector_to_py_size_t(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py_size_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":123
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

/* Python wrapper */
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_shape = 0;
  Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_format = 0;
  PyObject *__pyx_v_mode = 0;
  int __pyx_v_allocate_buffer;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_shape,&__pyx_n_s_itemsize,&__pyx_n_s_format,&__pyx_n_s_mode,&__pyx_n_s_allocate_buffer,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[3] = ((PyObject *)__pyx_n_s_c);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_itemsize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 1); __PYX_ERR(1, 123, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_format)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 2); __PYX_ERR(1, 123, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mode);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_allocate_buffer);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(1, 123, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_shape = ((PyObject*)values[0]);
    __pyx_v_itemsize = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_itemsize == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 123, __pyx_L3_error)
    __pyx_v_format = values[2];
    __pyx_v_mode = values[3];
    if (values[4]) {
      __pyx_v_allocate_buffer = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_allocate_buffer == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 124, __pyx_L3_error)
    } else {

      /* "View.MemoryView":124
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,
 *                   mode="c", bint allocate_buffer=True):             # <<<<<<<<<<<<<<
 * 
 *         cdef int idx
 */
      __pyx_v_allocate_buffer = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 123, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_shape), (&PyTuple_Type), 1, "shape", 1))) __PYX_ERR(1, 123, __pyx_L1_error)
  if (unlikely(((PyObject *)__pyx_v_format) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "format"); __PYX_ERR(1, 123, __pyx_L1_error)
  }
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(((struct __pyx_array_obj *)__pyx_v_self), __pyx_v_shape, __pyx_v_itemsize, __pyx_v_format, __pyx_v_mode, __pyx_v_allocate_buffer);

  /* "View.MemoryView":123
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_dim;
  PyObject **__pyx_v_p;
  char __pyx_v_order;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  char *__pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF(__pyx_v_format);

  /* "View.MemoryView":130
 *         cdef PyObject **p
 * 
 *         self.ndim = <int> len(shape)             # <<<<<<<<<<<<<<
 *         self.itemsize = itemsize
 * 
 */
  if (unlikely(__pyx_v_shape == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 130, __pyx_L1_error)
  }
  __pyx_t_1 = PyTuple_GET_SIZE(__pyx_v_shape); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 130, __pyx_L1_error)
  __pyx_v_self->ndim = ((int)__pyx_t_1);

  /* "View.MemoryView":131
 * 
 *         self.ndim = <int> len(shape)
 *         self.itemsize = itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not self.ndim:
 */
  __pyx_v_self->itemsize = __pyx_v_itemsize;

  /* "View.MemoryView":133
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_self->ndim != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":134
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__35, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 134, __pyx_L1_error)

    /* "View.MemoryView":133
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */
  }

  /* "View.MemoryView":136
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */
  __pyx_t_2 = ((__pyx_v_itemsize <= 0) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":137
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__36, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 137, __pyx_L1_error)

    /* "View.MemoryView":136
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */
  }

  /* "View.MemoryView":139
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  __pyx_t_2 = PyBytes_Check(__pyx_v_format); 
  __pyx_t_4 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_4) {

    /* "View.MemoryView":140
 * 
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')             # <<<<<<<<<<<<<<
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_format, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_n_s_ASCII) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_s_ASCII);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_format, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":139
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  }

  /* "View.MemoryView":141
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string             # <<<<<<<<<<<<<<
 *         self.format = self._format
 * 
 */
  if (!(likely(PyBytes_CheckExact(__pyx_v_format))||((__pyx_v_format) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_format)->tp_name), 0))) __PYX_ERR(1, 141, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_format;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_format);
  __Pyx_DECREF(__pyx_v_self->_format);
  __pyx_v_self->_format = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":142
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_self->_format == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(1, 142, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyBytes_AsWritableString(__pyx_v_self->_format); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(1, 142, __pyx_L1_error)
  __pyx_v_self->format = __pyx_t_7;

  /* "View.MemoryView":145
 * 
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)             # <<<<<<<<<<<<<<
 *         self._strides = self._shape + self.ndim
 * 
 */
  __pyx_v_self->_shape = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * __pyx_v_self->ndim) * 2)));

  /* "View.MemoryView":146
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)
 *         self._strides = self._shape + self.ndim             # <<<<<<<<<<<<<<
 * 
 *         if not self._shape:
 */
  __pyx_v_self->_strides = (__pyx_v_self->_shape + __pyx_v_self->ndim);

  /* "View.MemoryView":148
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */
  __pyx_t_4 = ((!(__pyx_v_self->_shape != 0)) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "View.MemoryView":149
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__37, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 149, __pyx_L1_error)

    /* "View.MemoryView":148
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */
  }

  /* "View.MemoryView":152
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */
  __pyx_t_8 = 0;
  __pyx_t_3 = __pyx_v_shape; __Pyx_INCREF(__pyx_t_3); __pyx_t_1 = 0;
  for (;;) {
    if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_5); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(1, 152, __pyx_L1_error)
    #else
    __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 152, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_9;
    __pyx_v_idx = __pyx_t_8;
    __pyx_t_8 = (__pyx_t_8 + 1);

    /* "View.MemoryView":153
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */
    __pyx_t_4 = ((__pyx_v_dim <= 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":154
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))             # <<<<<<<<<<<<<<
 *             self._shape[idx] = dim
 * 
 */
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_6);
      __pyx_t_5 = 0;
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(1, 154, __pyx_L1_error)

      /* "View.MemoryView":153
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */
    }

    /* "View.MemoryView":155
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim             # <<<<<<<<<<<<<<
 * 
 *         cdef char order
 */
    (__pyx_v_self->_shape[__pyx_v_idx]) = __pyx_v_dim;

    /* "View.MemoryView":152
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":158
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_fortran, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 158, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "View.MemoryView":159
 *         cdef char order
 *         if mode == 'fortran':
 *             order = b'F'             # <<<<<<<<<<<<<<
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 */
    __pyx_v_order = 'F';

    /* "View.MemoryView":160
 *         if mode == 'fortran':
 *             order = b'F'
 *             self.mode = u'fortran'             # <<<<<<<<<<<<<<
 *         elif mode == 'c':
 *             order = b'C'
 */
    __Pyx_INCREF(__pyx_n_u_fortran);
    __Pyx_GIVEREF(__pyx_n_u_fortran);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_fortran;

    /* "View.MemoryView":158
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":161
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_c, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 161, __pyx_L1_error)
  if (likely(__pyx_t_4)) {

    /* "View.MemoryView":162
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 *             order = b'C'             # <<<<<<<<<<<<<<
 *             self.mode = u'c'
 *         else:
 */
    __pyx_v_order = 'C';

    /* "View.MemoryView":163
 *         elif mode == 'c':
 *             order = b'C'
 *             self.mode = u'c'             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 */
    __Pyx_INCREF(__pyx_n_u_c);
    __Pyx_GIVEREF(__pyx_n_u_c);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_c;

    /* "View.MemoryView":161
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":165
 *             self.mode = u'c'
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)             # <<<<<<<<<<<<<<
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_v_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_10, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_ERR(1, 165, __pyx_L1_error)
  }
  __pyx_L10:;

  /* "View.MemoryView":167
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,             # <<<<<<<<<<<<<<
 *                                              itemsize, self.ndim, order)
 * 
 */
  __pyx_v_self->len = __pyx_fill_contig_strides_array(__pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_itemsize, __pyx_v_self->ndim, __pyx_v_order);

  /* "View.MemoryView":170
 *                                              itemsize, self.ndim, order)
 * 
 *         self.free_data = allocate_buffer             # <<<<<<<<<<<<<<
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:
 */
  __pyx_v_self->free_data = __pyx_v_allocate_buffer;

  /* "View.MemoryView":171
 * 
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'             # <<<<<<<<<<<<<<
 *         if allocate_buffer:
 * 
 */
  __pyx_t_10 = PyObject_RichCompare(__pyx_v_format, __pyx_n_b_O, Py_EQ); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 171, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 171, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_self->dtype_is_object = __pyx_t_4;

  /* "View.MemoryView":172
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = (__pyx_v_allocate_buffer != 0);
  if (__pyx_t_4) {

    /* "View.MemoryView":175
 * 
 * 
 *             self.data = <char *>malloc(self.len)             # <<<<<<<<<<<<<<
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")
 */
    __pyx_v_self->data = ((char *)malloc(__pyx_v_self->len));

    /* "View.MemoryView":176
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */
    __pyx_t_4 = ((!(__pyx_v_self->data != 0)) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":177
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__38, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 177, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(1, 177, __pyx_L1_error)

      /* "View.MemoryView":176
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */
    }

    /* "View.MemoryView":179
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */
    __pyx_t_4 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_4) {

      /* "View.MemoryView":180
 * 
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data             # <<<<<<<<<<<<<<
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 */
      __pyx_v_p = ((PyObject **)__pyx_v_self->data);

      /* "View.MemoryView":181
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):             # <<<<<<<<<<<<<<
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)
 */
      if (unlikely(__pyx_v_itemsize == 0)) {
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
        __PYX_ERR(1, 181, __pyx_L1_error)
      }
      else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_self->len))) {
        PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
        __PYX_ERR(1, 181, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_div_Py_ssize_t(__pyx_v_self->len, __pyx_v_itemsize);
      __pyx_t_9 = __pyx_t_1;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_9; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "View.MemoryView":182
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None             # <<<<<<<<<<<<<<
 *                     Py_INCREF(Py_None)
 * 
 */
        (__pyx_v_p[__pyx_v_i]) = Py_None;

        /* "View.MemoryView":183
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
        Py_INCREF(Py_None);
      }

      /* "View.MemoryView":179
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */
    }

    /* "View.MemoryView":172
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":123
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_format);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":186
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(((struct __pyx_array_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_bufmode;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":187
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1             # <<<<<<<<<<<<<<
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
  __pyx_v_bufmode = -1;

  /* "View.MemoryView":188
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_c, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 188, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":189
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
    __pyx_v_bufmode = (PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":188
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":190
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_fortran, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 190, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":191
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 */
    __pyx_v_bufmode = (PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":190
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":192
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */
  __pyx_t_1 = ((!((__pyx_v_flags & __pyx_v_bufmode) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":193
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__39, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 193, __pyx_L1_error)

    /* "View.MemoryView":192
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */
  }

  /* "View.MemoryView":194
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data             # <<<<<<<<<<<<<<
 *         info.len = self.len
 *         info.ndim = self.ndim
 */
  __pyx_t_4 = __pyx_v_self->data;
  __pyx_v_info->buf = __pyx_t_4;

  /* "View.MemoryView":195
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 *         info.len = self.len             # <<<<<<<<<<<<<<
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 */
  __pyx_t_5 = __pyx_v_self->len;
  __pyx_v_info->len = __pyx_t_5;

  /* "View.MemoryView":196
 *         info.buf = self.data
 *         info.len = self.len
 *         info.ndim = self.ndim             # <<<<<<<<<<<<<<
 *         info.shape = self._shape
 *         info.strides = self._strides
 */
  __pyx_t_6 = __pyx_v_self->ndim;
  __pyx_v_info->ndim = __pyx_t_6;

  /* "View.MemoryView":197
 *         info.len = self.len
 *         info.ndim = self.ndim
 *         info.shape = self._shape             # <<<<<<<<<<<<<<
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 */
  __pyx_t_7 = __pyx_v_self->_shape;
  __pyx_v_info->shape = __pyx_t_7;

  /* "View.MemoryView":198
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 *         info.strides = self._strides             # <<<<<<<<<<<<<<
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 */
  __pyx_t_7 = __pyx_v_self->_strides;
  __pyx_v_info->strides = __pyx_t_7;

  /* "View.MemoryView":199
 *         info.shape = self._shape
 *         info.strides = self._strides
 *         info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *         info.itemsize = self.itemsize
 *         info.readonly = 0
 */
  __pyx_v_info->suboffsets = NULL;

  /* "View.MemoryView":200
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize             # <<<<<<<<<<<<<<
 *         info.readonly = 0
 * 
 */
  __pyx_t_5 = __pyx_v_self->itemsize;
  __pyx_v_info->itemsize = __pyx_t_5;

  /* "View.MemoryView":201
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 *         info.readonly = 0             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  __pyx_v_info->readonly = 0;

  /* "View.MemoryView":203
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":204
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_4 = __pyx_v_self->format;
    __pyx_v_info->format = __pyx_t_4;

    /* "View.MemoryView":203
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
    goto __pyx_L5;
  }

  /* "View.MemoryView":206
 *             info.format = self.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.obj = self
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L5:;

  /* "View.MemoryView":208
 *             info.format = NULL
 * 
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":186
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":212
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

/* Python wrapper */
static void __pyx_array___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_array___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":213
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */
  __pyx_t_1 = ((__pyx_v_self->callback_free_data != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":214
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)             # <<<<<<<<<<<<<<
 *         elif self.free_data:
 *             if self.dtype_is_object:
 */
    __pyx_v_self->callback_free_data(__pyx_v_self->data);

    /* "View.MemoryView":213
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":215
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */
  __pyx_t_1 = (__pyx_v_self->free_data != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":216
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */
    __pyx_t_1 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":217
 *         elif self.free_data:
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,             # <<<<<<<<<<<<<<
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 */
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_self->data, __pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_self->ndim, 0);

      /* "View.MemoryView":216
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */
    }

    /* "View.MemoryView":219
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 *             free(self.data)             # <<<<<<<<<<<<<<
 *         PyObject_Free(self._shape)
 * 
 */
    free(__pyx_v_self->data);

    /* "View.MemoryView":215
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":220
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 *         PyObject_Free(self._shape)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  PyObject_Free(__pyx_v_self->_shape);

  /* "View.MemoryView":212
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":223
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_5array_7memview___get__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":224
 *     @property
 *     def memview(self):
 *         return self.get_memview()             # <<<<<<<<<<<<<<
 * 
 *     @cname('get_memview')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_array *)__pyx_v_self->__pyx_vtab)->get_memview(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":223
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.memview.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":227
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self) {
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_memview", 0);

  /* "View.MemoryView":228
 *     @cname('get_memview')
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE             # <<<<<<<<<<<<<<
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 */
  __pyx_v_flags = ((PyBUF_ANY_CONTIGUOUS | PyBUF_FORMAT) | PyBUF_WRITABLE);

  /* "View.MemoryView":229
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":227
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.get_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":231
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":232
 * 
 *     def __len__(self):
 *         return self._shape[0]             # <<<<<<<<<<<<<<
 * 
 *     def __getattr__(self, attr):
 */
  __pyx_r = (__pyx_v_self->_shape[0]);
  goto __pyx_L0;

  /* "View.MemoryView":231
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":234
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr); /*proto*/
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattr__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_attr));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getattr__", 0);

  /* "View.MemoryView":235
 * 
 *     def __getattr__(self, attr):
 *         return getattr(self.memview, attr)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, item):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetAttr(__pyx_t_1, __pyx_v_attr); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":234
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":237
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":238
 * 
 *     def __getitem__(self, item):
 *         return self.memview[item]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, item, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_item); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":237
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":240
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

/* Python wrapper */
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /*proto*/
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "View.MemoryView":241
 * 
 *     def __setitem__(self, item, value):
 *         self.memview[item] = value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_v_item, __pyx_v_value) < 0)) __PYX_ERR(1, 241, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":240
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array___reduce_cython__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__40, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array_2__setstate_cython__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__41, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":245
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

static struct __pyx_array_obj *__pyx_array_new(PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, char *__pyx_v_format, char *__pyx_v_mode, char *__pyx_v_buf) {
  struct __pyx_array_obj *__pyx_v_result = 0;
  struct __pyx_array_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("array_cwrapper", 0);

  /* "View.MemoryView":249
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_buf == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":250
 * 
 *     if buf == NULL:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))             # <<<<<<<<<<<<<<
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":249
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":252
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */
  /*else*/ {
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_3 = 0;

    /* "View.MemoryView":253
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)             # <<<<<<<<<<<<<<
 *         result.data = buf
 * 
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_allocate_buffer, Py_False) < 0) __PYX_ERR(1, 253, __pyx_L1_error)

    /* "View.MemoryView":252
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "View.MemoryView":254
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)
 *         result.data = buf             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
    __pyx_v_result->data = __pyx_v_buf;
  }
  __pyx_L3:;

  /* "View.MemoryView":256
 *         result.data = buf
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":245
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.array_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":282
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

/* Python wrapper */
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 282, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_name = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 282, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.Enum.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "View.MemoryView":283
 *     cdef object name
 *     def __init__(self, name):
 *         self.name = name             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return self.name
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "View.MemoryView":282
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":284
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":285
 *         self.name = name
 *     def __repr__(self):
 *         return self.name             # <<<<<<<<<<<<<<
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* "View.MemoryView":284
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum___reduce_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->name != Py_None);
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.Enum.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum_2__setstate_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_unpickle_Enum__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.Enum.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":299
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

static void *__pyx_align_pointer(void *__pyx_v_memory, size_t __pyx_v_alignment) {
  Py_intptr_t __pyx_v_aligned_p;
  size_t __pyx_v_offset;
  void *__pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":301
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory             # <<<<<<<<<<<<<<
 *     cdef size_t offset
 * 
 */
  __pyx_v_aligned_p = ((Py_intptr_t)__pyx_v_memory);

  /* "View.MemoryView":305
 * 
 *     with cython.cdivision(True):
 *         offset = aligned_p % alignment             # <<<<<<<<<<<<<<
 * 
 *     if offset > 0:
 */
  __pyx_v_offset = (__pyx_v_aligned_p % __pyx_v_alignment);

  /* "View.MemoryView":307
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  __pyx_t_1 = ((__pyx_v_offset > 0) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":308
 * 
 *     if offset > 0:
 *         aligned_p += alignment - offset             # <<<<<<<<<<<<<<
 * 
 *     return <void *> aligned_p
 */
    __pyx_v_aligned_p = (__pyx_v_aligned_p + (__pyx_v_alignment - __pyx_v_offset));

    /* "View.MemoryView":307
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  }

  /* "View.MemoryView":310
 *         aligned_p += alignment - offset
 * 
 *     return <void *> aligned_p             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((void *)__pyx_v_aligned_p);
  goto __pyx_L0;

  /* "View.MemoryView":299
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":346
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

/* Python wrapper */
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  int __pyx_v_flags;
  int __pyx_v_dtype_is_object;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_flags,&__pyx_n_s_dtype_is_object,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flags)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, 1); __PYX_ERR(1, 346, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dtype_is_object);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(1, 346, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_obj = values[0];
    __pyx_v_flags = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 346, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_dtype_is_object = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_dtype_is_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 346, __pyx_L3_error)
    } else {
      __pyx_v_dtype_is_object = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 346, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_obj, __pyx_v_flags, __pyx_v_dtype_is_object);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "View.MemoryView":347
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj             # <<<<<<<<<<<<<<
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 */
  __Pyx_INCREF(__pyx_v_obj);
  __Pyx_GIVEREF(__pyx_v_obj);
  __Pyx_GOTREF(__pyx_v_self->obj);
  __Pyx_DECREF(__pyx_v_self->obj);
  __pyx_v_self->obj = __pyx_v_obj;

  /* "View.MemoryView":348
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj
 *         self.flags = flags             # <<<<<<<<<<<<<<
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 */
  __pyx_v_self->flags = __pyx_v_flags;

  /* "View.MemoryView":349
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  __pyx_t_2 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))) == ((PyObject *)__pyx_memoryview_type));
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_obj != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "View.MemoryView":350
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)             # <<<<<<<<<<<<<<
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 */
    __pyx_t_4 = __Pyx_GetBuffer(__pyx_v_obj, (&__pyx_v_self->view), __pyx_v_flags); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 350, __pyx_L1_error)

    /* "View.MemoryView":351
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    __pyx_t_1 = ((((PyObject *)__pyx_v_self->view.obj) == NULL) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":352
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None             # <<<<<<<<<<<<<<
 *                 Py_INCREF(Py_None)
 * 
 */
      ((Py_buffer *)(&__pyx_v_self->view))->obj = Py_None;

      /* "View.MemoryView":353
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 */
      Py_INCREF(Py_None);

      /* "View.MemoryView":351
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    }

    /* "View.MemoryView":349
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  }

  /* "View.MemoryView":355
 *                 Py_INCREF(Py_None)
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():             # <<<<<<<<<<<<<<
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 */
  __pyx_t_1 = ((!(__PYX_CYTHON_ATOMICS_ENABLED() != 0)) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":357
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 */
    __pyx_t_1 = ((__pyx_memoryview_thread_locks_used < 8) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":358
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]             # <<<<<<<<<<<<<<
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:
 */
      __pyx_v_self->lock = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);

      /* "View.MemoryView":359
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1             # <<<<<<<<<<<<<<
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 */
      __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used + 1);

      /* "View.MemoryView":357
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 */
    }

    /* "View.MemoryView":360
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 */
    __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":361
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()             # <<<<<<<<<<<<<<
 *                 if self.lock is NULL:
 *                     raise MemoryError
 */
      __pyx_v_self->lock = PyThread_allocate_lock();

      /* "View.MemoryView":362
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 * 
 */
      __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
      if (unlikely(__pyx_t_1)) {

        /* "View.MemoryView":363
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 *                     raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
        PyErr_NoMemory(); __PYX_ERR(1, 363, __pyx_L1_error)

        /* "View.MemoryView":362
 *             if self.lock is NULL:
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 * 
 */
      }

      /* "View.MemoryView":360
 *                 self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *                 __pyx_memoryview_thread_locks_used += 1
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 self.lock = PyThread_allocate_lock()
 *                 if self.lock is NULL:
 */
    }

    /* "View.MemoryView":355
 *                 Py_INCREF(Py_None)
 * 
 *         if not __PYX_CYTHON_ATOMICS_ENABLED():             # <<<<<<<<<<<<<<
 *             global __pyx_memoryview_thread_locks_used
 *             if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 */
  }

  /* "View.MemoryView":365
 *                     raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":366
 * 
 *         if flags & PyBUF_FORMAT:
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')             # <<<<<<<<<<<<<<
 *         else:
 *             self.dtype_is_object = dtype_is_object
 */
    __pyx_t_2 = (((__pyx_v_self->view.format[0]) == 'O') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L12_bool_binop_done;
    }
    __pyx_t_2 = (((__pyx_v_self->view.format[1]) == '\x00') != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L12_bool_binop_done:;
    __pyx_v_self->dtype_is_object = __pyx_t_1;

    /* "View.MemoryView":365
 *                     raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
    goto __pyx_L11;
  }

  /* "View.MemoryView":368
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 *             self.dtype_is_object = dtype_is_object             # <<<<<<<<<<<<<<
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 */
  /*else*/ {
    __pyx_v_self->dtype_is_object = __pyx_v_dtype_is_object;
  }
  __pyx_L11:;

  /* "View.MemoryView":370
 *             self.dtype_is_object = dtype_is_object
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(             # <<<<<<<<<<<<<<
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL
 */
  __pyx_v_self->acquisition_count_aligned_p = ((__pyx_atomic_int *)__pyx_align_pointer(((void *)(&(__pyx_v_self->acquisition_count[0]))), (sizeof(__pyx_atomic_int))));

  /* "View.MemoryView":372
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(memoryview self):
 */
  __pyx_v_self->typeinfo = NULL;

  /* "View.MemoryView":346
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":374
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

/* Python wrapper */
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self) {
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyThread_type_lock __pyx_t_6;
  PyThread_type_lock __pyx_t_7;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":375
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
  __pyx_t_1 = (__pyx_v_self->obj != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":376
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)             # <<<<<<<<<<<<<<
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 */
    __Pyx_ReleaseBuffer((&__pyx_v_self->view));

    /* "View.MemoryView":375
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":377
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  __pyx_t_2 = ((((Py_buffer *)(&__pyx_v_self->view))->obj == Py_None) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":379
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL             # <<<<<<<<<<<<<<
 *             Py_DECREF(Py_None)
 * 
 */
    ((Py_buffer *)(&__pyx_v_self->view))->obj = NULL;

    /* "View.MemoryView":380
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 *             Py_DECREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         cdef int i
 */
    Py_DECREF(Py_None);

    /* "View.MemoryView":377
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":384
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  __pyx_t_2 = ((__pyx_v_self->lock != NULL) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":385
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):             # <<<<<<<<<<<<<<
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 */
    __pyx_t_3 = __pyx_memoryview_thread_locks_used;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "View.MemoryView":386
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      __pyx_t_2 = (((__pyx_memoryview_thread_locks[__pyx_v_i]) == __pyx_v_self->lock) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":387
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1             # <<<<<<<<<<<<<<
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 */
        __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used - 1);

        /* "View.MemoryView":388
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        __pyx_t_2 = ((__pyx_v_i != __pyx_memoryview_thread_locks_used) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":390
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])             # <<<<<<<<<<<<<<
 *                     break
 *             else:
 */
          __pyx_t_6 = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);
          __pyx_t_7 = (__pyx_memoryview_thread_locks[__pyx_v_i]);

          /* "View.MemoryView":389
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (             # <<<<<<<<<<<<<<
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break
 */
          (__pyx_memoryview_thread_locks[__pyx_v_i]) = __pyx_t_6;
          (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]) = __pyx_t_7;

          /* "View.MemoryView":388
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        }

        /* "View.MemoryView":391
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break             # <<<<<<<<<<<<<<
 *             else:
 *                 PyThread_free_lock(self.lock)
 */
        goto __pyx_L6_break;

        /* "View.MemoryView":386
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      }
    }
    /*else*/ {

      /* "View.MemoryView":393
 *                     break
 *             else:
 *                 PyThread_free_lock(self.lock)             # <<<<<<<<<<<<<<
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 */
      PyThread_free_lock(__pyx_v_self->lock);
    }
    __pyx_L6_break:;

    /* "View.MemoryView":384
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  }

  /* "View.MemoryView":374
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":395
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  Py_ssize_t __pyx_v_dim;
  char *__pyx_v_itemp;
  PyObject *__pyx_v_idx = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  char *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_item_pointer", 0);

  /* "View.MemoryView":397
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf             # <<<<<<<<<<<<<<
 * 
 *         for dim, idx in enumerate(index):
 */
  __pyx_v_itemp = ((char *)__pyx_v_self->view.buf);

  /* "View.MemoryView":399
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_v_index)) || PyTuple_CheckExact(__pyx_v_index)) {
    __pyx_t_2 = __pyx_v_index; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 399, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(1, 399, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 399, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(1, 399, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 399, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 399, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_1;
    __pyx_t_1 = (__pyx_t_1 + 1);

    /* "View.MemoryView":400
 * 
 *         for dim, idx in enumerate(index):
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)             # <<<<<<<<<<<<<<
 * 
 *         return itemp
 */
    __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_v_idx); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 400, __pyx_L1_error)
    __pyx_t_7 = __pyx_pybuffer_index((&__pyx_v_self->view), __pyx_v_itemp, __pyx_t_6, __pyx_v_dim); if (unlikely(__pyx_t_7 == ((char *)NULL))) __PYX_ERR(1, 400, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_7;

    /* "View.MemoryView":399
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":402
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 *         return itemp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_itemp;
  goto __pyx_L0;

  /* "View.MemoryView":395
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.get_item_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":405
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index); /*proto*/
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_indices = NULL;
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  char *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":406
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  __pyx_t_1 = (__pyx_v_index == __pyx_builtin_Ellipsis);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":407
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:
 *             return self             # <<<<<<<<<<<<<<
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __pyx_r = ((PyObject *)__pyx_v_self);
    goto __pyx_L0;

    /* "View.MemoryView":406
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  }

  /* "View.MemoryView":409
 *             return self
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         cdef char *itemp
 */
  __pyx_t_3 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(__pyx_t_3 != Py_None)) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 409, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 409, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 409, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(1, 409, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_indices = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "View.MemoryView":412
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 412, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "View.MemoryView":413
 *         cdef char *itemp
 *         if have_slices:
 *             return memview_slice(self, indices)             # <<<<<<<<<<<<<<
 *         else:
 *             itemp = self.get_item_pointer(indices)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((PyObject *)__pyx_memview_slice(__pyx_v_self, __pyx_v_indices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":412
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  }

  /* "View.MemoryView":415
 *             return memview_slice(self, indices)
 *         else:
 *             itemp = self.get_item_pointer(indices)             # <<<<<<<<<<<<<<
 *             return self.convert_item_to_object(itemp)
 * 
 */
  /*else*/ {
    __pyx_t_6 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_indices); if (unlikely(__pyx_t_6 == ((char *)NULL))) __PYX_ERR(1, 415, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_6;

    /* "View.MemoryView":416
 *         else:
 *             itemp = self.get_item_pointer(indices)
 *             return self.convert_item_to_object(itemp)             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(memoryview self, object index, object value):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->convert_item_to_object(__pyx_v_self, __pyx_v_itemp); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":405
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":418
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

/* Python wrapper */
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /*proto*/
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_obj = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);
  __Pyx_INCREF(__pyx_v_index);

  /* "View.MemoryView":419
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */
  __pyx_t_1 = (__pyx_v_self->view.readonly != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":420
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__42, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 420, __pyx_L1_error)

    /* "View.MemoryView":419
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */
  }

  /* "View.MemoryView":422
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         if have_slices:
 */
  __pyx_t_2 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(__pyx_t_2 != Py_None)) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 422, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(1, 422, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_3;
  __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_index, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "View.MemoryView":424
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 424, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "View.MemoryView":425
 * 
 *         if have_slices:
 *             obj = self.is_slice(value)             # <<<<<<<<<<<<<<
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->is_slice(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_obj = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "View.MemoryView":426
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_obj); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 426, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "View.MemoryView":427
 *             obj = self.is_slice(value)
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)
 */
      __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assignment(__pyx_v_self, __pyx_t_2, __pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "View.MemoryView":426
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":429
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)             # <<<<<<<<<<<<<<
 *         else:
 *             self.setitem_indexed(index, value)
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 429, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_memoryview_type))))) __PYX_ERR(1, 429, __pyx_L1_error)
      __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assign_scalar(__pyx_v_self, ((struct __pyx_memoryview_obj *)__pyx_t_4), __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 429, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L5:;

    /* "View.MemoryView":424
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":431
 *                 self.setitem_slice_assign_scalar(self[index], value)
 *         else:
 *             self.setitem_indexed(index, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef is_slice(self, obj):
 */
  /*else*/ {
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_indexed(__pyx_v_self, __pyx_v_index, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L4:;

  /* "View.MemoryView":418
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":433
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_slice", 0);
  __Pyx_INCREF(__pyx_v_obj);

  /* "View.MemoryView":434
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_memoryview_type); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":435
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "View.MemoryView":436
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_6 = __Pyx_PyInt_From_int(((__pyx_v_self->flags & (~PyBUF_WRITABLE)) | PyBUF_ANY_CONTIGUOUS)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 436, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":437
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)             # <<<<<<<<<<<<<<
 *             except TypeError:
 *                 return None
 */
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 437, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);

        /* "View.MemoryView":436
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 436, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_v_obj);
        __Pyx_GIVEREF(__pyx_v_obj);
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_obj);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_7);
        __pyx_t_6 = 0;
        __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 436, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF_SET(__pyx_v_obj, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "View.MemoryView":435
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "View.MemoryView":438
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 *             except TypeError:             # <<<<<<<<<<<<<<
 *                 return None
 * 
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_6) < 0) __PYX_ERR(1, 438, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":439
 *                                  self.dtype_is_object)
 *             except TypeError:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *         return obj
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L7_except_return;
      }
      goto __pyx_L6_except_error;
      __pyx_L6_except_error:;

      /* "View.MemoryView":435
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L7_except_return:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L0;
      __pyx_L9_try_end:;
    }

    /* "View.MemoryView":434
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  }

  /* "View.MemoryView":441
 *                 return None
 * 
 *         return obj             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assignment(self, dst, src):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_obj);
  __pyx_r = __pyx_v_obj;
  goto __pyx_L0;

  /* "View.MemoryView":433
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":443
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src) {
  __Pyx_memviewslice __pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_src_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  __Pyx_memviewslice *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_slice_assignment", 0);

  /* "View.MemoryView":447
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */
  if (!(likely(((__pyx_v_src) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_src, __pyx_memoryview_type))))) __PYX_ERR(1, 447, __pyx_L1_error)
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_src), (&__pyx_v_src_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 447, __pyx_L1_error)

  /* "View.MemoryView":448
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],             # <<<<<<<<<<<<<<
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 */
  if (!(likely(((__pyx_v_dst) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_dst, __pyx_memoryview_type))))) __PYX_ERR(1, 448, __pyx_L1_error)
  __pyx_t_2 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_dst), (&__pyx_v_dst_slice)); if (unlikely(__pyx_t_2 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 448, __pyx_L1_error)

  /* "View.MemoryView":449
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 449, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 449, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":447
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */
  __pyx_t_6 = __pyx_memoryview_copy_contents((__pyx_t_1[0]), (__pyx_t_2[0]), __pyx_t_4, __pyx_t_5, __pyx_v_self->dtype_is_object); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 447, __pyx_L1_error)

  /* "View.MemoryView":443
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assignment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":451
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value) {
  int __pyx_v_array[0x80];
  void *__pyx_v_tmp;
  void *__pyx_v_item;
  __Pyx_memviewslice *__pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_tmp_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  char const *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_slice_assign_scalar", 0);

  /* "View.MemoryView":453
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 *         cdef int array[128]
 *         cdef void *tmp = NULL             # <<<<<<<<<<<<<<
 *         cdef void *item
 * 
 */
  __pyx_v_tmp = NULL;

  /* "View.MemoryView":458
 *         cdef __Pyx_memviewslice *dst_slice
 *         cdef __Pyx_memviewslice tmp_slice
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)             # <<<<<<<<<<<<<<
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_dst, (&__pyx_v_tmp_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 458, __pyx_L1_error)
  __pyx_v_dst_slice = __pyx_t_1;

  /* "View.MemoryView":460
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
  __pyx_t_2 = ((((size_t)__pyx_v_self->view.itemsize) > (sizeof(__pyx_v_array))) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":461
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)             # <<<<<<<<<<<<<<
 *             if tmp == NULL:
 *                 raise MemoryError
 */
    __pyx_v_tmp = PyMem_Malloc(__pyx_v_self->view.itemsize);

    /* "View.MemoryView":462
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    __pyx_t_2 = ((__pyx_v_tmp == NULL) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":463
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 *             item = tmp
 *         else:
 */
      PyErr_NoMemory(); __PYX_ERR(1, 463, __pyx_L1_error)

      /* "View.MemoryView":462
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    }

    /* "View.MemoryView":464
 *             if tmp == NULL:
 *                 raise MemoryError
 *             item = tmp             # <<<<<<<<<<<<<<
 *         else:
 *             item = <void *> array
 */
    __pyx_v_item = __pyx_v_tmp;

    /* "View.MemoryView":460
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":466
 *             item = tmp
 *         else:
 *             item = <void *> array             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  /*else*/ {
    __pyx_v_item = ((void *)__pyx_v_array);
  }
  __pyx_L3:;

  /* "View.MemoryView":468
 *             item = <void *> array
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value
 */
  /*try:*/ {

    /* "View.MemoryView":469
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
    __pyx_t_2 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":470
 *         try:
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value             # <<<<<<<<<<<<<<
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)
 */
      (((PyObject **)__pyx_v_item)[0]) = ((PyObject *)__pyx_v_value);

      /* "View.MemoryView":469
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
      goto __pyx_L8;
    }

    /* "View.MemoryView":472
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*else*/ {
      __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, ((char *)__pyx_v_item), __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 472, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_L8:;

    /* "View.MemoryView":476
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    __pyx_t_2 = ((__pyx_v_self->view.suboffsets != NULL) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":477
 * 
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)             # <<<<<<<<<<<<<<
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 *                                 item, self.dtype_is_object)
 */
      __pyx_t_3 = assert_direct_dimensions(__pyx_v_self->view.suboffsets, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 477, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "View.MemoryView":476
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    }

    /* "View.MemoryView":478
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,             # <<<<<<<<<<<<<<
 *                                 item, self.dtype_is_object)
 *         finally:
 */
    __pyx_memoryview_slice_assign_scalar(__pyx_v_dst_slice, __pyx_v_dst->view.ndim, __pyx_v_self->view.itemsize, __pyx_v_item, __pyx_v_self->dtype_is_object);
  }

  /* "View.MemoryView":481
 *                                 item, self.dtype_is_object)
 *         finally:
 *             PyMem_Free(tmp)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_indexed(self, index, value):
 */
  /*finally:*/ {
    /*normal exit:*/{
      PyMem_Free(__pyx_v_tmp);
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __pyx_t_4 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_6 = __pyx_filename;
      {
        PyMem_Free(__pyx_v_tmp);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __pyx_lineno = __pyx_t_4; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_6;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "View.MemoryView":451
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assign_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":483
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_indexed", 0);

  /* "View.MemoryView":484
 * 
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)             # <<<<<<<<<<<<<<
 *         self.assign_item_from_object(itemp, value)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_index); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(1, 484, __pyx_L1_error)
  __pyx_v_itemp = __pyx_t_1;

  /* "View.MemoryView":485
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":483
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_indexed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":487
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_bytesitem = 0;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":490
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef bytes bytesitem
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":493
 *         cdef bytes bytesitem
 * 
 *         bytesitem = itemp[:self.view.itemsize]             # <<<<<<<<<<<<<<
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 */
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_itemp + 0, __pyx_v_self->view.itemsize - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_bytesitem = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":494
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "View.MemoryView":495
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)             # <<<<<<<<<<<<<<
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_unpack); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 495, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 495, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 495, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 495, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 495, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_7) {
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_6);
        __Pyx_INCREF(__pyx_v_bytesitem);
        __Pyx_GIVEREF(__pyx_v_bytesitem);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_bytesitem);
        __pyx_t_6 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 495, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_result = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "View.MemoryView":494
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    }

    /* "View.MemoryView":499
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
    /*else:*/ {
      __pyx_t_10 = strlen(__pyx_v_self->view.format); 
      __pyx_t_11 = ((__pyx_t_10 == 1) != 0);
      if (__pyx_t_11) {

        /* "View.MemoryView":500
 *         else:
 *             if len(self.view.format) == 1:
 *                 return result[0]             # <<<<<<<<<<<<<<
 *             return result
 * 
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_result, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 500, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L6_except_return;

        /* "View.MemoryView":499
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
      }

      /* "View.MemoryView":501
 *             if len(self.view.format) == 1:
 *                 return result[0]
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_result);
      __pyx_r = __pyx_v_result;
      goto __pyx_L6_except_return;
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "View.MemoryView":496
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unable to convert item to object")
 *         else:
 */
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_5, &__pyx_t_9);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_error); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 496, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_5, __pyx_t_9);
    __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_9 = 0;
    if (__pyx_t_8) {
      __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_5, &__pyx_t_1) < 0) __PYX_ERR(1, 496, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_1);

      /* "View.MemoryView":497
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__43, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 497, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(1, 497, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "View.MemoryView":494
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L0;
  }

  /* "View.MemoryView":487
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesitem);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":503
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_struct = NULL;
  char __pyx_v_c;
  PyObject *__pyx_v_bytesvalue = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  char *__pyx_t_11;
  char *__pyx_t_12;
  char *__pyx_t_13;
  char *__pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":506
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef char c
 *         cdef bytes bytesvalue
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":511
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
  __pyx_t_2 = PyTuple_Check(__pyx_v_value); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "View.MemoryView":512
 * 
 *         if isinstance(value, tuple):
 *             bytesvalue = struct.pack(self.view.format, *value)             # <<<<<<<<<<<<<<
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyNumber_Add(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(1, 512, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":511
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":514
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)             # <<<<<<<<<<<<<<
 * 
 *         for i, c in enumerate(bytesvalue):
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 514, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 514, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 514, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_1);
      __Pyx_INCREF(__pyx_v_value);
      __Pyx_GIVEREF(__pyx_v_value);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_value);
      __pyx_t_1 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 514, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(1, 514, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":516
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
  __pyx_t_9 = 0;
  if (unlikely(__pyx_v_bytesvalue == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");
    __PYX_ERR(1, 516, __pyx_L1_error)
  }
  __Pyx_INCREF(__pyx_v_bytesvalue);
  __pyx_t_10 = __pyx_v_bytesvalue;
  __pyx_t_12 = PyBytes_AS_STRING(__pyx_t_10);
  __pyx_t_13 = (__pyx_t_12 + PyBytes_GET_SIZE(__pyx_t_10));
  for (__pyx_t_14 = __pyx_t_12; __pyx_t_14 < __pyx_t_13; __pyx_t_14++) {
    __pyx_t_11 = __pyx_t_14;
    __pyx_v_c = (__pyx_t_11[0]);

    /* "View.MemoryView":517
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    __pyx_v_i = __pyx_t_9;

    /* "View.MemoryView":516
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
    __pyx_t_9 = (__pyx_t_9 + 1);

    /* "View.MemoryView":517
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    (__pyx_v_itemp[__pyx_v_i]) = __pyx_v_c;
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "View.MemoryView":503
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("View.MemoryView.memoryview.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":520
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t *__pyx_t_4;
  char *__pyx_t_5;
  void *__pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":521
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */
  __pyx_t_2 = ((__pyx_v_flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->view.readonly != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":522
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__44, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 522, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 522, __pyx_L1_error)

    /* "View.MemoryView":521
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */
  }

  /* "View.MemoryView":524
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_ND) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":525
 * 
 *         if flags & PyBUF_ND:
 *             info.shape = self.view.shape             # <<<<<<<<<<<<<<
 *         else:
 *             info.shape = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.shape;
    __pyx_v_info->shape = __pyx_t_4;

    /* "View.MemoryView":524
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
    goto __pyx_L6;
  }

  /* "View.MemoryView":527
 *             info.shape = self.view.shape
 *         else:
 *             info.shape = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_STRIDES:
 */
  /*else*/ {
    __pyx_v_info->shape = NULL;
  }
  __pyx_L6:;

  /* "View.MemoryView":529
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_STRIDES) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":530
 * 
 *         if flags & PyBUF_STRIDES:
 *             info.strides = self.view.strides             # <<<<<<<<<<<<<<
 *         else:
 *             info.strides = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.strides;
    __pyx_v_info->strides = __pyx_t_4;

    /* "View.MemoryView":529
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
    goto __pyx_L7;
  }

  /* "View.MemoryView":532
 *             info.strides = self.view.strides
 *         else:
 *             info.strides = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_INDIRECT:
 */
  /*else*/ {
    __pyx_v_info->strides = NULL;
  }
  __pyx_L7:;

  /* "View.MemoryView":534
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_INDIRECT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":535
 * 
 *         if flags & PyBUF_INDIRECT:
 *             info.suboffsets = self.view.suboffsets             # <<<<<<<<<<<<<<
 *         else:
 *             info.suboffsets = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.suboffsets;
    __pyx_v_info->suboffsets = __pyx_t_4;

    /* "View.MemoryView":534
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
    goto __pyx_L8;
  }

  /* "View.MemoryView":537
 *             info.suboffsets = self.view.suboffsets
 *         else:
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  /*else*/ {
    __pyx_v_info->suboffsets = NULL;
  }
  __pyx_L8:;

  /* "View.MemoryView":539
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":540
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.view.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_5 = __pyx_v_self->view.format;
    __pyx_v_info->format = __pyx_t_5;

    /* "View.MemoryView":539
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":542
 *             info.format = self.view.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.buf = self.view.buf
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L9:;

  /* "View.MemoryView":544
 *             info.format = NULL
 * 
 *         info.buf = self.view.buf             # <<<<<<<<<<<<<<
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 */
  __pyx_t_6 = __pyx_v_self->view.buf;
  __pyx_v_info->buf = __pyx_t_6;

  /* "View.MemoryView":545
 * 
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim             # <<<<<<<<<<<<<<
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 */
  __pyx_t_7 = __pyx_v_self->view.ndim;
  __pyx_v_info->ndim = __pyx_t_7;

  /* "View.MemoryView":546
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize             # <<<<<<<<<<<<<<
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 */
  __pyx_t_8 = __pyx_v_self->view.itemsize;
  __pyx_v_info->itemsize = __pyx_t_8;

  /* "View.MemoryView":547
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len             # <<<<<<<<<<<<<<
 *         info.readonly = self.view.readonly
 *         info.obj = self
 */
  __pyx_t_8 = __pyx_v_self->view.len;
  __pyx_v_info->len = __pyx_t_8;

  /* "View.MemoryView":548
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly             # <<<<<<<<<<<<<<
 *         info.obj = self
 * 
 */
  __pyx_t_1 = __pyx_v_self->view.readonly;
  __pyx_v_info->readonly = __pyx_t_1;

  /* "View.MemoryView":549
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":520
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":555
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":556
 *     @property
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&result.from_slice)
 *         return result
 */
  __pyx_t_1 = __pyx_memoryview_copy_object(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_memoryviewslice_type))))) __PYX_ERR(1, 556, __pyx_L1_error)
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":557
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  __pyx_t_2 = __pyx_memslice_transpose((&__pyx_v_result->from_slice)); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(1, 557, __pyx_L1_error)

  /* "View.MemoryView":558
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":555
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.T.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":561
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":562
 *     @property
 *     def base(self):
 *         return self.obj             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->obj);
  __pyx_r = __pyx_v_self->obj;
  goto __pyx_L0;

  /* "View.MemoryView":561
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":565
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":566
 *     @property
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
  for (__pyx_t_4 = __pyx_v_self->view.shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
    __pyx_t_2 = __pyx_t_4;
    __pyx_v_length = (__pyx_t_2[0]);
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(1, 566, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = PyList_AsTuple(((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":565
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.shape.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":569
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_stride;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":570
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */
  __pyx_t_1 = ((__pyx_v_self->view.strides == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":572
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__45, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 572, __pyx_L1_error)

    /* "View.MemoryView":570
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */
  }

  /* "View.MemoryView":574
 *             raise ValueError("Buffer view does not expose strides")
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = (__pyx_v_self->view.strides + __pyx_v_self->view.ndim);
  for (__pyx_t_5 = __pyx_v_self->view.strides; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
    __pyx_t_3 = __pyx_t_5;
    __pyx_v_stride = (__pyx_t_3[0]);
    __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_stride); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(1, 574, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_t_6 = PyList_AsTuple(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":569
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.strides.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":577
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":578
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->view.suboffsets == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":579
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Multiply(__pyx_tuple__46, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":578
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  }

  /* "View.MemoryView":581
 *             return (-1,) * self.view.ndim
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = (__pyx_v_self->view.suboffsets + __pyx_v_self->view.ndim);
  for (__pyx_t_6 = __pyx_v_self->view.suboffsets; __pyx_t_6 < __pyx_t_5; __pyx_t_6++) {
    __pyx_t_4 = __pyx_t_6;
    __pyx_v_suboffset = (__pyx_t_4[0]);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_suboffset); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 581, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(1, 581, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = PyList_AsTuple(((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":577
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.suboffsets.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":584
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":585
 *     @property
 *     def ndim(self):
 *         return self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":584
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.ndim.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":588
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":589
 *     @property
 *     def itemsize(self):
 *         return self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":588
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.itemsize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":592
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":593
 *     @property
 *     def nbytes(self):
 *         return self.size * self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":592
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.nbytes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":596
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":597
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_size == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":598
 *     def size(self):
 *         if self._size is None:
 *             result = 1             # <<<<<<<<<<<<<<
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 */
    __Pyx_INCREF(__pyx_int_1);
    __pyx_v_result = __pyx_int_1;

    /* "View.MemoryView":600
 *             result = 1
 * 
 *             for length in self.view.shape[:self.view.ndim]:             # <<<<<<<<<<<<<<
 *                 result *= length
 * 
 */
    __pyx_t_4 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
    for (__pyx_t_5 = __pyx_v_self->view.shape; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
      __pyx_t_3 = __pyx_t_5;
      __pyx_t_6 = PyInt_FromSsize_t((__pyx_t_3[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 600, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "View.MemoryView":601
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 *                 result *= length             # <<<<<<<<<<<<<<
 * 
 *             self._size = result
 */
      __pyx_t_6 = PyNumber_InPlaceMultiply(__pyx_v_result, __pyx_v_length); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 601, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_6);
      __pyx_t_6 = 0;
    }

    /* "View.MemoryView":603
 *                 result *= length
 * 
 *             self._size = result             # <<<<<<<<<<<<<<
 * 
 *         return self._size
 */
    __Pyx_INCREF(__pyx_v_result);
    __Pyx_GIVEREF(__pyx_v_result);
    __Pyx_GOTREF(__pyx_v_self->_size);
    __Pyx_DECREF(__pyx_v_self->_size);
    __pyx_v_self->_size = __pyx_v_result;

    /* "View.MemoryView":597
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  }

  /* "View.MemoryView":605
 *             self._size = result
 * 
 *         return self._size             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_size);
  __pyx_r = __pyx_v_self->_size;
  goto __pyx_L0;

  /* "View.MemoryView":596
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":607
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

/* Python wrapper */
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":608
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->view.ndim >= 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":609
 *     def __len__(self):
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         return 0
 */
    __pyx_r = (__pyx_v_self->view.shape[0]);
    goto __pyx_L0;

    /* "View.MemoryView":608
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  }

  /* "View.MemoryView":611
 *             return self.view.shape[0]
 * 
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":607
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":613
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":614
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":615
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "View.MemoryView":614
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":613
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":617
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "View.MemoryView":618
 * 
 *     def __str__(self):
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_object, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":617
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":621
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_c_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_c_contig", 0);

  /* "View.MemoryView":624
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 624, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":625
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def is_f_contig(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'C', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":621
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_c_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":627
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_f_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_f_contig", 0);

  /* "View.MemoryView":630
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 630, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":631
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'F', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":627
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_f_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":633
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_mslice;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "View.MemoryView":635
 *     def copy(self):
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &mslice)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_F_CONTIGUOUS));

  /* "View.MemoryView":637
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 * 
 *         slice_copy(self, &mslice)             # <<<<<<<<<<<<<<
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,
 *                                    self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_mslice));

  /* "View.MemoryView":638
 * 
 *         slice_copy(self, &mslice)
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                    self.view.itemsize,
 *                                    flags|PyBUF_C_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_mslice), ((char *)"c"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_C_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 638, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":643
 *                                    self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &mslice)             # <<<<<<<<<<<<<<
 * 
 *     def copy_fortran(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_mslice)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":633
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":645
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy_fortran (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy_fortran", 0);

  /* "View.MemoryView":647
 *     def copy_fortran(self):
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &src)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_C_CONTIGUOUS));

  /* "View.MemoryView":649
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 * 
 *         slice_copy(self, &src)             # <<<<<<<<<<<<<<
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,
 *                                 self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_src));

  /* "View.MemoryView":650
 * 
 *         slice_copy(self, &src)
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                 self.view.itemsize,
 *                                 flags|PyBUF_F_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_src), ((char *)"fortran"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_F_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 650, __pyx_L1_error)
  __pyx_v_dst = __pyx_t_1;

  /* "View.MemoryView":655
 *                                 self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &dst)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_dst)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":645
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy_fortran", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview___reduce_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__47, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview_2__setstate_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__48, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":659
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

static PyObject *__pyx_memoryview_new(PyObject *__pyx_v_o, int __pyx_v_flags, int __pyx_v_dtype_is_object, __Pyx_TypeInfo *__pyx_v_typeinfo) {
  struct __pyx_memoryview_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_cwrapper", 0);

  /* "View.MemoryView":660
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)             # <<<<<<<<<<<<<<
 *     result.typeinfo = typeinfo
 *     return result
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_o);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryview_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":661
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_v_result->typeinfo = __pyx_v_typeinfo;

  /* "View.MemoryView":662
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_check')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":659
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":665
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

static CYTHON_INLINE int __pyx_memoryview_check(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("memoryview_check", 0);

  /* "View.MemoryView":666
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):
 *     return isinstance(o, memoryview)             # <<<<<<<<<<<<<<
 * 
 * cdef tuple _unellipsify(object index, int ndim):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_memoryview_type); 
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "View.MemoryView":665
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":668
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

static PyObject *_unellipsify(PyObject *__pyx_v_index, int __pyx_v_ndim) {
  PyObject *__pyx_v_tup = NULL;
  PyObject *__pyx_v_result = NULL;
  int __pyx_v_have_slices;
  int __pyx_v_seen_ellipsis;
  CYTHON_UNUSED PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_item = NULL;
  Py_ssize_t __pyx_v_nslices;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unellipsify", 0);

  /* "View.MemoryView":673
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */
  __pyx_t_1 = PyTuple_Check(__pyx_v_index); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":674
 *     """
 *     if not isinstance(index, tuple):
 *         tup = (index,)             # <<<<<<<<<<<<<<
 *     else:
 *         tup = index
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_index);
    __Pyx_GIVEREF(__pyx_v_index);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_index);
    __pyx_v_tup = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "View.MemoryView":673
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":676
 *         tup = (index,)
 *     else:
 *         tup = index             # <<<<<<<<<<<<<<
 * 
 *     result = []
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_index);
    __pyx_v_tup = __pyx_v_index;
  }
  __pyx_L3:;

  /* "View.MemoryView":678
 *         tup = index
 * 
 *     result = []             # <<<<<<<<<<<<<<
 *     have_slices = False
 *     seen_ellipsis = False
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_result = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":679
 * 
 *     result = []
 *     have_slices = False             # <<<<<<<<<<<<<<
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 */
  __pyx_v_have_slices = 0;

  /* "View.MemoryView":680
 *     result = []
 *     have_slices = False
 *     seen_ellipsis = False             # <<<<<<<<<<<<<<
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 */
  __pyx_v_seen_ellipsis = 0;

  /* "View.MemoryView":681
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_3 = __pyx_int_0;
  if (likely(PyList_CheckExact(__pyx_v_tup)) || PyTuple_CheckExact(__pyx_v_tup)) {
    __pyx_t_4 = __pyx_v_tup; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_tup); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 681, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(1, 681, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 681, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(1, 681, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 681, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_6(__pyx_t_4);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 681, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_3);
    __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "View.MemoryView":682
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */
    __pyx_t_2 = (__pyx_v_item == __pyx_builtin_Ellipsis);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":683
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */
      __pyx_t_1 = ((!(__pyx_v_seen_ellipsis != 0)) != 0);
      if (__pyx_t_1) {

        /* "View.MemoryView":684
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */
        __pyx_t_8 = PyObject_Length(__pyx_v_tup); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(1, 684, __pyx_L1_error)
        __pyx_t_7 = PyList_New(1 * ((((__pyx_v_ndim - __pyx_t_8) + 1)<0) ? 0:((__pyx_v_ndim - __pyx_t_8) + 1))); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 684, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        { Py_ssize_t __pyx_temp;
          for (__pyx_temp=0; __pyx_temp < ((__pyx_v_ndim - __pyx_t_8) + 1); __pyx_temp++) {
            __Pyx_INCREF(__pyx_slice__11);
            __Pyx_GIVEREF(__pyx_slice__11);
            PyList_SET_ITEM(__pyx_t_7, __pyx_temp, __pyx_slice__11);
          }
        }
        __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_7); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 684, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "View.MemoryView":685
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True             # <<<<<<<<<<<<<<
 *             else:
 *                 result.append(slice(None))
 */
        __pyx_v_seen_ellipsis = 1;

        /* "View.MemoryView":683
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":687
 *                 seen_ellipsis = True
 *             else:
 *                 result.append(slice(None))             # <<<<<<<<<<<<<<
 *             have_slices = True
 *         else:
 */
      /*else*/ {
        __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_slice__11); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 687, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":688
 *             else:
 *                 result.append(slice(None))
 *             have_slices = True             # <<<<<<<<<<<<<<
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 */
      __pyx_v_have_slices = 1;

      /* "View.MemoryView":682
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":690
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */
    /*else*/ {
      __pyx_t_2 = PySlice_Check(__pyx_v_item); 
      __pyx_t_10 = ((!(__pyx_t_2 != 0)) != 0);
      if (__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_10 = ((!(PyIndex_Check(__pyx_v_item) != 0)) != 0);
      __pyx_t_1 = __pyx_t_10;
      __pyx_L9_bool_binop_done:;
      if (unlikely(__pyx_t_1)) {

        /* "View.MemoryView":691
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 *                 raise TypeError("Cannot index with type '%s'" % type(item))             # <<<<<<<<<<<<<<
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 */
        __pyx_t_7 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_index_with_type_s, ((PyObject *)Py_TYPE(__pyx_v_item))); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 691, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 691, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_Raise(__pyx_t_11, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __PYX_ERR(1, 691, __pyx_L1_error)

        /* "View.MemoryView":690
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */
      }

      /* "View.MemoryView":693
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 *             have_slices = have_slices or isinstance(item, slice)             # <<<<<<<<<<<<<<
 *             result.append(item)
 * 
 */
      __pyx_t_10 = (__pyx_v_have_slices != 0);
      if (!__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_10 = PySlice_Check(__pyx_v_item); 
      __pyx_t_2 = (__pyx_t_10 != 0);
      __pyx_t_1 = __pyx_t_2;
      __pyx_L11_bool_binop_done:;
      __pyx_v_have_slices = __pyx_t_1;

      /* "View.MemoryView":694
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 *             result.append(item)             # <<<<<<<<<<<<<<
 * 
 *     nslices = ndim - len(result)
 */
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_v_item); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 694, __pyx_L1_error)
    }
    __pyx_L6:;

    /* "View.MemoryView":681
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":696
 *             result.append(item)
 * 
 *     nslices = ndim - len(result)             # <<<<<<<<<<<<<<
 *     if nslices:
 *         result.extend([slice(None)] * nslices)
 */
  __pyx_t_5 = PyList_GET_SIZE(__pyx_v_result); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 696, __pyx_L1_error)
  __pyx_v_nslices = (__pyx_v_ndim - __pyx_t_5);

  /* "View.MemoryView":697
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */
  __pyx_t_1 = (__pyx_v_nslices != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":698
 *     nslices = ndim - len(result)
 *     if nslices:
 *         result.extend([slice(None)] * nslices)             # <<<<<<<<<<<<<<
 * 
 *     return have_slices or nslices, tuple(result)
 */
    __pyx_t_3 = PyList_New(1 * ((__pyx_v_nslices<0) ? 0:__pyx_v_nslices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    { Py_ssize_t __pyx_temp;
      for (__pyx_temp=0; __pyx_temp < __pyx_v_nslices; __pyx_temp++) {
        __Pyx_INCREF(__pyx_slice__11);
        __Pyx_GIVEREF(__pyx_slice__11);
        PyList_SET_ITEM(__pyx_t_3, __pyx_temp, __pyx_slice__11);
      }
    }
    __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_3); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 698, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "View.MemoryView":697
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */
  }

  /* "View.MemoryView":700
 *         result.extend([slice(None)] * nslices)
 * 
 *     return have_slices or nslices, tuple(result)             # <<<<<<<<<<<<<<
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 */
  __Pyx_XDECREF(__pyx_r);
  if (!__pyx_v_have_slices) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_have_slices); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L14_bool_binop_done;
  }
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_nslices); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_L14_bool_binop_done:;
  __pyx_t_4 = PyList_AsTuple(__pyx_v_result); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = ((PyObject*)__pyx_t_11);
  __pyx_t_11 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":668
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("View.MemoryView._unellipsify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tup);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":702
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

static PyObject *assert_direct_dimensions(Py_ssize_t *__pyx_v_suboffsets, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assert_direct_dimensions", 0);

  /* "View.MemoryView":703
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")
 */
  __pyx_t_2 = (__pyx_v_suboffsets + __pyx_v_ndim);
  for (__pyx_t_3 = __pyx_v_suboffsets; __pyx_t_3 < __pyx_t_2; __pyx_t_3++) {
    __pyx_t_1 = __pyx_t_3;
    __pyx_v_suboffset = (__pyx_t_1[0]);

    /* "View.MemoryView":704
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */
    __pyx_t_4 = ((__pyx_v_suboffset >= 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":705
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__49, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 705, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(1, 705, __pyx_L1_error)

      /* "View.MemoryView":704
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */
    }
  }

  /* "View.MemoryView":702
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.assert_direct_dimensions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":712
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *__pyx_v_memview, PyObject *__pyx_v_indices) {
  int __pyx_v_new_ndim;
  int __pyx_v_suboffset_dim;
  int __pyx_v_dim;
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  __Pyx_memviewslice *__pyx_v_p_src;
  struct __pyx_memoryviewslice_obj *__pyx_v_memviewsliceobj = 0;
  __Pyx_memviewslice *__pyx_v_p_dst;
  int *__pyx_v_p_suboffset_dim;
  Py_ssize_t __pyx_v_start;
  Py_ssize_t __pyx_v_stop;
  Py_ssize_t __pyx_v_step;
  int __pyx_v_have_start;
  int __pyx_v_have_stop;
  int __pyx_v_have_step;
  PyObject *__pyx_v_index = NULL;
  struct __pyx_memoryview_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_memoryview_obj *__pyx_t_4;
  char *__pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memview_slice", 0);

  /* "View.MemoryView":713
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim             # <<<<<<<<<<<<<<
 *     cdef bint negative_step
 *     cdef __Pyx_memviewslice src, dst
 */
  __pyx_v_new_ndim = 0;
  __pyx_v_suboffset_dim = -1;

  /* "View.MemoryView":720
 * 
 * 
 *     memset(&dst, 0, sizeof(dst))             # <<<<<<<<<<<<<<
 * 
 *     cdef _memoryviewslice memviewsliceobj
 */
  (void)(memset((&__pyx_v_dst), 0, (sizeof(__pyx_v_dst))));

  /* "View.MemoryView":724
 *     cdef _memoryviewslice memviewsliceobj
 * 
 *     assert memview.view.ndim > 0             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!((__pyx_v_memview->view.ndim > 0) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(1, 724, __pyx_L1_error)
    }
  }
  #endif

  /* "View.MemoryView":726
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":727
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview             # <<<<<<<<<<<<<<
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(1, 727, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_memviewsliceobj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":728
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, &src)
 */
    __pyx_v_p_src = (&__pyx_v_memviewsliceobj->from_slice);

    /* "View.MemoryView":726
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":730
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 *         slice_copy(memview, &src)             # <<<<<<<<<<<<<<
 *         p_src = &src
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_src));

    /* "View.MemoryView":731
 *     else:
 *         slice_copy(memview, &src)
 *         p_src = &src             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_p_src = (&__pyx_v_src);
  }
  __pyx_L3:;

  /* "View.MemoryView":737
 * 
 * 
 *     dst.memview = p_src.memview             # <<<<<<<<<<<<<<
 *     dst.data = p_src.data
 * 
 */
  __pyx_t_4 = __pyx_v_p_src->memview;
  __pyx_v_dst.memview = __pyx_t_4;

  /* "View.MemoryView":738
 * 
 *     dst.memview = p_src.memview
 *     dst.data = p_src.data             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __pyx_v_p_src->data;
  __pyx_v_dst.data = __pyx_t_5;

  /* "View.MemoryView":743
 * 
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst             # <<<<<<<<<<<<<<
 *     cdef int *p_suboffset_dim = &suboffset_dim
 *     cdef Py_ssize_t start, stop, step
 */
  __pyx_v_p_dst = (&__pyx_v_dst);

  /* "View.MemoryView":744
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst
 *     cdef int *p_suboffset_dim = &suboffset_dim             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t start, stop, step
 *     cdef bint have_start, have_stop, have_step
 */
  __pyx_v_p_suboffset_dim = (&__pyx_v_suboffset_dim);

  /* "View.MemoryView":748
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */
  __pyx_t_6 = 0;
  if (likely(PyList_CheckExact(__pyx_v_indices)) || PyTuple_CheckExact(__pyx_v_indices)) {
    __pyx_t_3 = __pyx_v_indices; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_indices); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 748, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 748, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(1, 748, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 748, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(1, 748, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 748, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      }
    } else {
      __pyx_t_9 = __pyx_t_8(__pyx_t_3);
      if (unlikely(!__pyx_t_9)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 748, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_9);
    }
    __Pyx_XDECREF_SET(__pyx_v_index, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_v_dim = __pyx_t_6;
    __pyx_t_6 = (__pyx_t_6 + 1);

    /* "View.MemoryView":749
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
    __pyx_t_2 = (PyIndex_Check(__pyx_v_index) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":753
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 *                 index, 0, 0, # start, stop, step             # <<<<<<<<<<<<<<
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 */
      __pyx_t_10 = __Pyx_PyIndex_AsSsize_t(__pyx_v_index); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 753, __pyx_L1_error)

      /* "View.MemoryView":750
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_t_10, 0, 0, 0, 0, 0, 0); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(1, 750, __pyx_L1_error)

      /* "View.MemoryView":749
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":756
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
    __pyx_t_2 = (__pyx_v_index == Py_None);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":757
 *                 False)
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1             # <<<<<<<<<<<<<<
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 */
      (__pyx_v_p_dst->shape[__pyx_v_new_ndim]) = 1;

      /* "View.MemoryView":758
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0             # <<<<<<<<<<<<<<
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1
 */
      (__pyx_v_p_dst->strides[__pyx_v_new_ndim]) = 0;

      /* "View.MemoryView":759
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1             # <<<<<<<<<<<<<<
 *             new_ndim += 1
 *         else:
 */
      (__pyx_v_p_dst->suboffsets[__pyx_v_new_ndim]) = -1L;

      /* "View.MemoryView":760
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 *         else:
 *             start = index.start or 0
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);

      /* "View.MemoryView":756
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":762
 *             new_ndim += 1
 *         else:
 *             start = index.start or 0             # <<<<<<<<<<<<<<
 *             stop = index.stop or 0
 *             step = index.step or 0
 */
    /*else*/ {
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 762, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 762, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 762, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L7_bool_binop_done:;
      __pyx_v_start = __pyx_t_10;

      /* "View.MemoryView":763
 *         else:
 *             start = index.start or 0
 *             stop = index.stop or 0             # <<<<<<<<<<<<<<
 *             step = index.step or 0
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 763, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 763, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 763, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L9_bool_binop_done:;
      __pyx_v_stop = __pyx_t_10;

      /* "View.MemoryView":764
 *             start = index.start or 0
 *             stop = index.stop or 0
 *             step = index.step or 0             # <<<<<<<<<<<<<<
 * 
 *             have_start = index.start is not None
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 764, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 764, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 764, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L11_bool_binop_done:;
      __pyx_v_step = __pyx_t_10;

      /* "View.MemoryView":766
 *             step = index.step or 0
 * 
 *             have_start = index.start is not None             # <<<<<<<<<<<<<<
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_start = __pyx_t_1;

      /* "View.MemoryView":767
 * 
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None             # <<<<<<<<<<<<<<
 *             have_step = index.step is not None
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 767, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_stop = __pyx_t_1;

      /* "View.MemoryView":768
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None             # <<<<<<<<<<<<<<
 * 
 *             slice_memviewslice(
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 768, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_step = __pyx_t_1;

      /* "View.MemoryView":770
 *             have_step = index.step is not None
 * 
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_v_start, __pyx_v_stop, __pyx_v_step, __pyx_v_have_start, __pyx_v_have_stop, __pyx_v_have_step, 1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(1, 770, __pyx_L1_error)

      /* "View.MemoryView":776
 *                 have_start, have_stop, have_step,
 *                 True)
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);
    }
    __pyx_L6:;

    /* "View.MemoryView":748
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":778
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":779
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":780
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_dtype_func,
 *                                     memview.dtype_is_object)
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(1, 780, __pyx_L1_error) }

    /* "View.MemoryView":781
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 *     else:
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(1, 781, __pyx_L1_error) }

    /* "View.MemoryView":779
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, __pyx_v_memviewsliceobj->to_object_func, __pyx_v_memviewsliceobj->to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 779, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(1, 779, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":778
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  }

  /* "View.MemoryView":784
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":785
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,
 *                                     memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, NULL, NULL, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 784, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "View.MemoryView":784
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(1, 784, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":712
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memview_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_memviewsliceobj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":809
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *__pyx_v_dst, Py_ssize_t __pyx_v_shape, Py_ssize_t __pyx_v_stride, Py_ssize_t __pyx_v_suboffset, int __pyx_v_dim, int __pyx_v_new_ndim, int *__pyx_v_suboffset_dim, Py_ssize_t __pyx_v_start, Py_ssize_t __pyx_v_stop, Py_ssize_t __pyx_v_step, int __pyx_v_have_start, int __pyx_v_have_stop, int __pyx_v_have_step, int __pyx_v_is_slice) {
  Py_ssize_t __pyx_v_new_shape;
  int __pyx_v_negative_step;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":829
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
  __pyx_t_1 = ((!(__pyx_v_is_slice != 0)) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":831
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    __pyx_t_1 = ((__pyx_v_start < 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":832
 * 
 *         if start < 0:
 *             start += shape             # <<<<<<<<<<<<<<
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 */
      __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

      /* "View.MemoryView":831
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    }

    /* "View.MemoryView":833
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    __pyx_t_1 = (0 <= __pyx_v_start);
    if (__pyx_t_1) {
      __pyx_t_1 = (__pyx_v_start < __pyx_v_shape);
    }
    __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":834
 *             start += shape
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)             # <<<<<<<<<<<<<<
 *     else:
 * 
 */
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"Index out of bounds (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 834, __pyx_L1_error)

      /* "View.MemoryView":833
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    }

    /* "View.MemoryView":829
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":837
 *     else:
 * 
 *         negative_step = have_step != 0 and step < 0             # <<<<<<<<<<<<<<
 * 
 *         if have_step and step == 0:
 */
  /*else*/ {
    __pyx_t_1 = ((__pyx_v_have_step != 0) != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_step < 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L6_bool_binop_done:;
    __pyx_v_negative_step = __pyx_t_2;

    /* "View.MemoryView":839
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */
    __pyx_t_1 = (__pyx_v_have_step != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_step == 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_2) {

      /* "View.MemoryView":840
 * 
 *         if have_step and step == 0:
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Step may not be zero (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 840, __pyx_L1_error)

      /* "View.MemoryView":839
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */
    }

    /* "View.MemoryView":843
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
    __pyx_t_2 = (__pyx_v_have_start != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":844
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
      __pyx_t_2 = ((__pyx_v_start < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":845
 *         if have_start:
 *             if start < 0:
 *                 start += shape             # <<<<<<<<<<<<<<
 *                 if start < 0:
 *                     start = 0
 */
        __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

        /* "View.MemoryView":846
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        __pyx_t_2 = ((__pyx_v_start < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":847
 *                 start += shape
 *                 if start < 0:
 *                     start = 0             # <<<<<<<<<<<<<<
 *             elif start >= shape:
 *                 if negative_step:
 */
          __pyx_v_start = 0;

          /* "View.MemoryView":846
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        }

        /* "View.MemoryView":844
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
        goto __pyx_L12;
      }

      /* "View.MemoryView":848
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      __pyx_t_2 = ((__pyx_v_start >= __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":849
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
        __pyx_t_2 = (__pyx_v_negative_step != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":850
 *             elif start >= shape:
 *                 if negative_step:
 *                     start = shape - 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     start = shape
 */
          __pyx_v_start = (__pyx_v_shape - 1);

          /* "View.MemoryView":849
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
          goto __pyx_L14;
        }

        /* "View.MemoryView":852
 *                     start = shape - 1
 *                 else:
 *                     start = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        /*else*/ {
          __pyx_v_start = __pyx_v_shape;
        }
        __pyx_L14:;

        /* "View.MemoryView":848
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      }
      __pyx_L12:;

      /* "View.MemoryView":843
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
      goto __pyx_L11;
    }

    /* "View.MemoryView":854
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":855
 *         else:
 *             if negative_step:
 *                 start = shape - 1             # <<<<<<<<<<<<<<
 *             else:
 *                 start = 0
 */
        __pyx_v_start = (__pyx_v_shape - 1);

        /* "View.MemoryView":854
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
        goto __pyx_L15;
      }

      /* "View.MemoryView":857
 *                 start = shape - 1
 *             else:
 *                 start = 0             # <<<<<<<<<<<<<<
 * 
 *         if have_stop:
 */
      /*else*/ {
        __pyx_v_start = 0;
      }
      __pyx_L15:;
    }
    __pyx_L11:;

    /* "View.MemoryView":859
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
    __pyx_t_2 = (__pyx_v_have_stop != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":860
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
      __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":861
 *         if have_stop:
 *             if stop < 0:
 *                 stop += shape             # <<<<<<<<<<<<<<
 *                 if stop < 0:
 *                     stop = 0
 */
        __pyx_v_stop = (__pyx_v_stop + __pyx_v_shape);

        /* "View.MemoryView":862
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":863
 *                 stop += shape
 *                 if stop < 0:
 *                     stop = 0             # <<<<<<<<<<<<<<
 *             elif stop > shape:
 *                 stop = shape
 */
          __pyx_v_stop = 0;

          /* "View.MemoryView":862
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        }

        /* "View.MemoryView":860
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
        goto __pyx_L17;
      }

      /* "View.MemoryView":864
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      __pyx_t_2 = ((__pyx_v_stop > __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":865
 *                     stop = 0
 *             elif stop > shape:
 *                 stop = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        __pyx_v_stop = __pyx_v_shape;

        /* "View.MemoryView":864
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      }
      __pyx_L17:;

      /* "View.MemoryView":859
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
      goto __pyx_L16;
    }

    /* "View.MemoryView":867
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":868
 *         else:
 *             if negative_step:
 *                 stop = -1             # <<<<<<<<<<<<<<
 *             else:
 *                 stop = shape
 */
        __pyx_v_stop = -1L;

        /* "View.MemoryView":867
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
        goto __pyx_L19;
      }

      /* "View.MemoryView":870
 *                 stop = -1
 *             else:
 *                 stop = shape             # <<<<<<<<<<<<<<
 * 
 *         if not have_step:
 */
      /*else*/ {
        __pyx_v_stop = __pyx_v_shape;
      }
      __pyx_L19:;
    }
    __pyx_L16:;

    /* "View.MemoryView":872
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    __pyx_t_2 = ((!(__pyx_v_have_step != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":873
 * 
 *         if not have_step:
 *             step = 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_step = 1;

      /* "View.MemoryView":872
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    }

    /* "View.MemoryView":877
 * 
 *         with cython.cdivision(True):
 *             new_shape = (stop - start) // step             # <<<<<<<<<<<<<<
 * 
 *             if (stop - start) - step * new_shape:
 */
    __pyx_v_new_shape = ((__pyx_v_stop - __pyx_v_start) / __pyx_v_step);

    /* "View.MemoryView":879
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    __pyx_t_2 = (((__pyx_v_stop - __pyx_v_start) - (__pyx_v_step * __pyx_v_new_shape)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":880
 * 
 *             if (stop - start) - step * new_shape:
 *                 new_shape += 1             # <<<<<<<<<<<<<<
 * 
 *         if new_shape < 0:
 */
      __pyx_v_new_shape = (__pyx_v_new_shape + 1);

      /* "View.MemoryView":879
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    }

    /* "View.MemoryView":882
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    __pyx_t_2 = ((__pyx_v_new_shape < 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":883
 * 
 *         if new_shape < 0:
 *             new_shape = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_new_shape = 0;

      /* "View.MemoryView":882
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    }

    /* "View.MemoryView":886
 * 
 * 
 *         dst.strides[new_ndim] = stride * step             # <<<<<<<<<<<<<<
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset
 */
    (__pyx_v_dst->strides[__pyx_v_new_ndim]) = (__pyx_v_stride * __pyx_v_step);

    /* "View.MemoryView":887
 * 
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape             # <<<<<<<<<<<<<<
 *         dst.suboffsets[new_ndim] = suboffset
 * 
 */
    (__pyx_v_dst->shape[__pyx_v_new_ndim]) = __pyx_v_new_shape;

    /* "View.MemoryView":888
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_dst->suboffsets[__pyx_v_new_ndim]) = __pyx_v_suboffset;
  }
  __pyx_L3:;

  /* "View.MemoryView":891
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
  __pyx_t_2 = (((__pyx_v_suboffset_dim[0]) < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":892
 * 
 *     if suboffset_dim[0] < 0:
 *         dst.data += start * stride             # <<<<<<<<<<<<<<
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 */
    __pyx_v_dst->data = (__pyx_v_dst->data + (__pyx_v_start * __pyx_v_stride));

    /* "View.MemoryView":891
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
    goto __pyx_L23;
  }

  /* "View.MemoryView":894
 *         dst.data += start * stride
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride             # <<<<<<<<<<<<<<
 * 
 *     if suboffset >= 0:
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_suboffset_dim[0]);
    (__pyx_v_dst->suboffsets[__pyx_t_3]) = ((__pyx_v_dst->suboffsets[__pyx_t_3]) + (__pyx_v_start * __pyx_v_stride));
  }
  __pyx_L23:;

  /* "View.MemoryView":896
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":897
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
    __pyx_t_2 = ((!(__pyx_v_is_slice != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":898
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
      __pyx_t_2 = ((__pyx_v_new_ndim == 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":899
 *         if not is_slice:
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 */
        __pyx_v_dst->data = ((((char **)__pyx_v_dst->data)[0]) + __pyx_v_suboffset);

        /* "View.MemoryView":898
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
        goto __pyx_L26;
      }

      /* "View.MemoryView":901
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "             # <<<<<<<<<<<<<<
 *                                      "must be indexed and not sliced", dim)
 *         else:
 */
      /*else*/ {

        /* "View.MemoryView":902
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 *                                      "must be indexed and not sliced", dim)             # <<<<<<<<<<<<<<
 *         else:
 *             suboffset_dim[0] = new_ndim
 */
        __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"All dimensions preceding dimension %d must be indexed and not sliced"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 901, __pyx_L1_error)
      }
      __pyx_L26:;

      /* "View.MemoryView":897
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
      goto __pyx_L25;
    }

    /* "View.MemoryView":904
 *                                      "must be indexed and not sliced", dim)
 *         else:
 *             suboffset_dim[0] = new_ndim             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
    /*else*/ {
      (__pyx_v_suboffset_dim[0]) = __pyx_v_new_ndim;
    }
    __pyx_L25:;

    /* "View.MemoryView":896
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  }

  /* "View.MemoryView":906
 *             suboffset_dim[0] = new_ndim
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":809
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.slice_memviewslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":912
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

static char *__pyx_pybuffer_index(Py_buffer *__pyx_v_view, char *__pyx_v_bufp, Py_ssize_t __pyx_v_index, Py_ssize_t __pyx_v_dim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_suboffset;
  Py_ssize_t __pyx_v_itemsize;
  char *__pyx_v_resultp;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pybuffer_index", 0);

  /* "View.MemoryView":914
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t itemsize = view.itemsize
 *     cdef char *resultp
 */
  __pyx_v_suboffset = -1L;

  /* "View.MemoryView":915
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 *     cdef Py_ssize_t itemsize = view.itemsize             # <<<<<<<<<<<<<<
 *     cdef char *resultp
 * 
 */
  __pyx_t_1 = __pyx_v_view->itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":918
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */
  __pyx_t_2 = ((__pyx_v_view->ndim == 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":919
 * 
 *     if view.ndim == 0:
 *         shape = view.len / itemsize             # <<<<<<<<<<<<<<
 *         stride = itemsize
 *     else:
 */
    if (unlikely(__pyx_v_itemsize == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(1, 919, __pyx_L1_error)
    }
    else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_view->len))) {
      PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
      __PYX_ERR(1, 919, __pyx_L1_error)
    }
    __pyx_v_shape = __Pyx_div_Py_ssize_t(__pyx_v_view->len, __pyx_v_itemsize);

    /* "View.MemoryView":920
 *     if view.ndim == 0:
 *         shape = view.len / itemsize
 *         stride = itemsize             # <<<<<<<<<<<<<<
 *     else:
 *         shape = view.shape[dim]
 */
    __pyx_v_stride = __pyx_v_itemsize;

    /* "View.MemoryView":918
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":922
 *         stride = itemsize
 *     else:
 *         shape = view.shape[dim]             # <<<<<<<<<<<<<<
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 */
  /*else*/ {
    __pyx_v_shape = (__pyx_v_view->shape[__pyx_v_dim]);

    /* "View.MemoryView":923
 *     else:
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]             # <<<<<<<<<<<<<<
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]
 */
    __pyx_v_stride = (__pyx_v_view->strides[__pyx_v_dim]);

    /* "View.MemoryView":924
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    __pyx_t_2 = ((__pyx_v_view->suboffsets != NULL) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":925
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]             # <<<<<<<<<<<<<<
 * 
 *     if index < 0:
 */
      __pyx_v_suboffset = (__pyx_v_view->suboffsets[__pyx_v_dim]);

      /* "View.MemoryView":924
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":927
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  __pyx_t_2 = ((__pyx_v_index < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":928
 * 
 *     if index < 0:
 *         index += view.shape[dim]             # <<<<<<<<<<<<<<
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 */
    __pyx_v_index = (__pyx_v_index + (__pyx_v_view->shape[__pyx_v_dim]));

    /* "View.MemoryView":929
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
    __pyx_t_2 = ((__pyx_v_index < 0) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":930
 *         index += view.shape[dim]
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     if index >= shape:
 */
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 930, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 930, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 930, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(1, 930, __pyx_L1_error)

      /* "View.MemoryView":929
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
    }

    /* "View.MemoryView":927
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  }

  /* "View.MemoryView":932
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
  __pyx_t_2 = ((__pyx_v_index >= __pyx_v_shape) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":933
 * 
 *     if index >= shape:
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     resultp = bufp + index * stride
 */
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 933, __pyx_L1_error)

    /* "View.MemoryView":932
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
  }

  /* "View.MemoryView":935
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     resultp = bufp + index * stride             # <<<<<<<<<<<<<<
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset
 */
  __pyx_v_resultp = (__pyx_v_bufp + (__pyx_v_index * __pyx_v_stride));

  /* "View.MemoryView":936
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":937
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset             # <<<<<<<<<<<<<<
 * 
 *     return resultp
 */
    __pyx_v_resultp = ((((char **)__pyx_v_resultp)[0]) + __pyx_v_suboffset);

    /* "View.MemoryView":936
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  }

  /* "View.MemoryView":939
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 *     return resultp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_resultp;
  goto __pyx_L0;

  /* "View.MemoryView":912
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.pybuffer_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":945
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

static int __pyx_memslice_transpose(__Pyx_memviewslice *__pyx_v_memslice) {
  int __pyx_v_ndim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_r;
  int __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  long __pyx_t_3;
  long __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":946
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:
 *     cdef int ndim = memslice.memview.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 */
  __pyx_t_1 = __pyx_v_memslice->memview->view.ndim;
  __pyx_v_ndim = __pyx_t_1;

  /* "View.MemoryView":948
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 *     cdef Py_ssize_t *shape = memslice.shape             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t *strides = memslice.strides
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->shape;
  __pyx_v_shape = __pyx_t_2;

  /* "View.MemoryView":949
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 *     cdef Py_ssize_t *strides = memslice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->strides;
  __pyx_v_strides = __pyx_t_2;

  /* "View.MemoryView":953
 * 
 *     cdef int i, j
 *     for i in range(ndim / 2):             # <<<<<<<<<<<<<<
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 */
  __pyx_t_3 = __Pyx_div_long(__pyx_v_ndim, 2);
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_4; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":954
 *     cdef int i, j
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i             # <<<<<<<<<<<<<<
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]
 */
    __pyx_v_j = ((__pyx_v_ndim - 1) - __pyx_v_i);

    /* "View.MemoryView":955
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]             # <<<<<<<<<<<<<<
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 */
    __pyx_t_5 = (__pyx_v_strides[__pyx_v_j]);
    __pyx_t_6 = (__pyx_v_strides[__pyx_v_i]);
    (__pyx_v_strides[__pyx_v_i]) = __pyx_t_5;
    (__pyx_v_strides[__pyx_v_j]) = __pyx_t_6;

    /* "View.MemoryView":956
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]             # <<<<<<<<<<<<<<
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 */
    __pyx_t_6 = (__pyx_v_shape[__pyx_v_j]);
    __pyx_t_5 = (__pyx_v_shape[__pyx_v_i]);
    (__pyx_v_shape[__pyx_v_i]) = __pyx_t_6;
    (__pyx_v_shape[__pyx_v_j]) = __pyx_t_5;

    /* "View.MemoryView":958
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_i]) >= 0) != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_j]) >= 0) != 0);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_7) {

      /* "View.MemoryView":959
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")             # <<<<<<<<<<<<<<
 * 
 *     return 1
 */
      __pyx_t_9 = __pyx_memoryview_err(__pyx_builtin_ValueError, ((char *)"Cannot transpose memoryview with indirect dimensions")); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 959, __pyx_L1_error)

      /* "View.MemoryView":958
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    }
  }

  /* "View.MemoryView":961
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "View.MemoryView":945
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.transpose_memslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":978
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

/* Python wrapper */
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":979
 * 
 *     def __dealloc__(self):
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __PYX_XDEC_MEMVIEW((&__pyx_v_self->from_slice), 1);

  /* "View.MemoryView":978
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":981
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":982
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_self->to_object_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":983
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)             # <<<<<<<<<<<<<<
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_v_self->to_object_func(__pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":982
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  }

  /* "View.MemoryView":985
 *             return self.to_object_func(itemp)
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_memoryview_convert_item_to_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 985, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":981
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":987
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":988
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_self->to_dtype_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":989
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)             # <<<<<<<<<<<<<<
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)
 */
    __pyx_t_2 = __pyx_v_self->to_dtype_func(__pyx_v_itemp, __pyx_v_value); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(1, 989, __pyx_L1_error)

    /* "View.MemoryView":988
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":991
 *             self.to_dtype_func(itemp, value)
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  /*else*/ {
    __pyx_t_3 = __pyx_memoryview_assign_item_from_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 991, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":987
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":994
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":995
 *     @property
 *     def base(self):
 *         return self.from_object             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->from_object);
  __pyx_r = __pyx_v_self->from_object;
  goto __pyx_L0;

  /* "View.MemoryView":994
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice___reduce_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__50, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice_2__setstate_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__51, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1001
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice __pyx_v_memviewslice, int __pyx_v_ndim, PyObject *(*__pyx_v_to_object_func)(char *), int (*__pyx_v_to_dtype_func)(char *, PyObject *), int __pyx_v_dtype_is_object) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_TypeInfo *__pyx_t_4;
  Py_buffer __pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  Py_ssize_t *__pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_fromslice", 0);

  /* "View.MemoryView":1009
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  __pyx_t_1 = ((((PyObject *)__pyx_v_memviewslice.memview) == Py_None) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1010
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:
 *         return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "View.MemoryView":1009
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  }

  /* "View.MemoryView":1015
 * 
 * 
 *     result = _memoryviewslice(None, 0, dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     result.from_slice = memviewslice
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_3, 0, Py_None);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryviewslice_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":1017
 *     result = _memoryviewslice(None, 0, dtype_is_object)
 * 
 *     result.from_slice = memviewslice             # <<<<<<<<<<<<<<
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 */
  __pyx_v_result->from_slice = __pyx_v_memviewslice;

  /* "View.MemoryView":1018
 * 
 *     result.from_slice = memviewslice
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)             # <<<<<<<<<<<<<<
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 */
  __PYX_INC_MEMVIEW((&__pyx_v_memviewslice), 1);

  /* "View.MemoryView":1020
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base             # <<<<<<<<<<<<<<
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_memviewslice.memview), __pyx_n_s_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_result->from_object);
  __Pyx_DECREF(__pyx_v_result->from_object);
  __pyx_v_result->from_object = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "View.MemoryView":1021
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 *     result.typeinfo = memviewslice.memview.typeinfo             # <<<<<<<<<<<<<<
 * 
 *     result.view = memviewslice.memview.view
 */
  __pyx_t_4 = __pyx_v_memviewslice.memview->typeinfo;
  __pyx_v_result->__pyx_base.typeinfo = __pyx_t_4;

  /* "View.MemoryView":1023
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 *     result.view = memviewslice.memview.view             # <<<<<<<<<<<<<<
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 */
  __pyx_t_5 = __pyx_v_memviewslice.memview->view;
  __pyx_v_result->__pyx_base.view = __pyx_t_5;

  /* "View.MemoryView":1024
 * 
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data             # <<<<<<<<<<<<<<
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 */
  __pyx_v_result->__pyx_base.view.buf = ((void *)__pyx_v_memviewslice.data);

  /* "View.MemoryView":1025
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim             # <<<<<<<<<<<<<<
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)
 */
  __pyx_v_result->__pyx_base.view.ndim = __pyx_v_ndim;

  /* "View.MemoryView":1026
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None             # <<<<<<<<<<<<<<
 *     Py_INCREF(Py_None)
 * 
 */
  ((Py_buffer *)(&__pyx_v_result->__pyx_base.view))->obj = Py_None;

  /* "View.MemoryView":1027
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 */
  Py_INCREF(Py_None);

  /* "View.MemoryView":1029
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
  __pyx_t_1 = ((((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)->flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1030
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 *         result.flags = PyBUF_RECORDS             # <<<<<<<<<<<<<<
 *     else:
 *         result.flags = PyBUF_RECORDS_RO
 */
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS;

    /* "View.MemoryView":1029
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":1032
 *         result.flags = PyBUF_RECORDS
 *     else:
 *         result.flags = PyBUF_RECORDS_RO             # <<<<<<<<<<<<<<
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 */
  /*else*/ {
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS_RO;
  }
  __pyx_L4:;

  /* "View.MemoryView":1034
 *         result.flags = PyBUF_RECORDS_RO
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape             # <<<<<<<<<<<<<<
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides
 * 
 */
  __pyx_v_result->__pyx_base.view.shape = ((Py_ssize_t *)__pyx_v_result->from_slice.shape);

  /* "View.MemoryView":1035
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_result->__pyx_base.view.strides = ((Py_ssize_t *)__pyx_v_result->from_slice.strides);

  /* "View.MemoryView":1038
 * 
 * 
 *     result.view.suboffsets = NULL             # <<<<<<<<<<<<<<
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 */
  __pyx_v_result->__pyx_base.view.suboffsets = NULL;

  /* "View.MemoryView":1039
 * 
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 */
  __pyx_t_7 = (__pyx_v_result->from_slice.suboffsets + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->from_slice.suboffsets; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_v_suboffset = (__pyx_t_6[0]);

    /* "View.MemoryView":1040
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    __pyx_t_1 = ((__pyx_v_suboffset >= 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":1041
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_result->__pyx_base.view.suboffsets = ((Py_ssize_t *)__pyx_v_result->from_slice.suboffsets);

      /* "View.MemoryView":1042
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break             # <<<<<<<<<<<<<<
 * 
 *     result.view.len = result.view.itemsize
 */
      goto __pyx_L6_break;

      /* "View.MemoryView":1040
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    }
  }
  __pyx_L6_break:;

  /* "View.MemoryView":1044
 *             break
 * 
 *     result.view.len = result.view.itemsize             # <<<<<<<<<<<<<<
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length
 */
  __pyx_t_9 = __pyx_v_result->__pyx_base.view.itemsize;
  __pyx_v_result->__pyx_base.view.len = __pyx_t_9;

  /* "View.MemoryView":1045
 * 
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         result.view.len *= length
 * 
 */
  __pyx_t_7 = (__pyx_v_result->__pyx_base.view.shape + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->__pyx_base.view.shape; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_t_2 = PyInt_FromSsize_t((__pyx_t_6[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "View.MemoryView":1046
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length             # <<<<<<<<<<<<<<
 * 
 *     result.to_object_func = to_object_func
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_result->__pyx_base.view.len); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1046, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_InPlaceMultiply(__pyx_t_2, __pyx_v_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1046, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 1046, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result->__pyx_base.view.len = __pyx_t_9;
  }

  /* "View.MemoryView":1048
 *         result.view.len *= length
 * 
 *     result.to_object_func = to_object_func             # <<<<<<<<<<<<<<
 *     result.to_dtype_func = to_dtype_func
 * 
 */
  __pyx_v_result->to_object_func = __pyx_v_to_object_func;

  /* "View.MemoryView":1049
 * 
 *     result.to_object_func = to_object_func
 *     result.to_dtype_func = to_dtype_func             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  __pyx_v_result->to_dtype_func = __pyx_v_to_dtype_func;

  /* "View.MemoryView":1051
 *     result.to_dtype_func = to_dtype_func
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":1001
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_fromslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1054
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_mslice) {
  struct __pyx_memoryviewslice_obj *__pyx_v_obj = 0;
  __Pyx_memviewslice *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_slice_from_memview", 0);

  /* "View.MemoryView":1057
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1058
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview             # <<<<<<<<<<<<<<
 *         return &obj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(1, 1058, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_obj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":1059
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview
 *         return &obj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, mslice)
 */
    __pyx_r = (&__pyx_v_obj->from_slice);
    goto __pyx_L0;

    /* "View.MemoryView":1057
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  }

  /* "View.MemoryView":1061
 *         return &obj.from_slice
 *     else:
 *         slice_copy(memview, mslice)             # <<<<<<<<<<<<<<
 *         return mslice
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, __pyx_v_mslice);

    /* "View.MemoryView":1062
 *     else:
 *         slice_copy(memview, mslice)
 *         return mslice             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_slice_copy')
 */
    __pyx_r = __pyx_v_mslice;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1054
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.get_slice_from_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_obj);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1065
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_dst) {
  int __pyx_v_dim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  Py_ssize_t *__pyx_v_suboffsets;
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  __Pyx_RefNannySetupContext("slice_copy", 0);

  /* "View.MemoryView":1069
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 * 
 *     shape = memview.view.shape             # <<<<<<<<<<<<<<
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets
 */
  __pyx_t_1 = __pyx_v_memview->view.shape;
  __pyx_v_shape = __pyx_t_1;

  /* "View.MemoryView":1070
 * 
 *     shape = memview.view.shape
 *     strides = memview.view.strides             # <<<<<<<<<<<<<<
 *     suboffsets = memview.view.suboffsets
 * 
 */
  __pyx_t_1 = __pyx_v_memview->view.strides;
  __pyx_v_strides = __pyx_t_1;

  /* "View.MemoryView":1071
 *     shape = memview.view.shape
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets             # <<<<<<<<<<<<<<
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 */
  __pyx_t_1 = __pyx_v_memview->view.suboffsets;
  __pyx_v_suboffsets = __pyx_t_1;

  /* "View.MemoryView":1073
 *     suboffsets = memview.view.suboffsets
 * 
 *     dst.memview = <__pyx_memoryview *> memview             # <<<<<<<<<<<<<<
 *     dst.data = <char *> memview.view.buf
 * 
 */
  __pyx_v_dst->memview = ((struct __pyx_memoryview_obj *)__pyx_v_memview);

  /* "View.MemoryView":1074
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 *     dst.data = <char *> memview.view.buf             # <<<<<<<<<<<<<<
 * 
 *     for dim in range(memview.view.ndim):
 */
  __pyx_v_dst->data = ((char *)__pyx_v_memview->view.buf);

  /* "View.MemoryView":1076
 *     dst.data = <char *> memview.view.buf
 * 
 *     for dim in range(memview.view.ndim):             # <<<<<<<<<<<<<<
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 */
  __pyx_t_2 = __pyx_v_memview->view.ndim;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_dim = __pyx_t_4;

    /* "View.MemoryView":1077
 * 
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]             # <<<<<<<<<<<<<<
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 */
    (__pyx_v_dst->shape[__pyx_v_dim]) = (__pyx_v_shape[__pyx_v_dim]);

    /* "View.MemoryView":1078
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]             # <<<<<<<<<<<<<<
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 * 
 */
    (__pyx_v_dst->strides[__pyx_v_dim]) = (__pyx_v_strides[__pyx_v_dim]);

    /* "View.MemoryView":1079
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object')
 */
    if ((__pyx_v_suboffsets != 0)) {
      __pyx_t_5 = (__pyx_v_suboffsets[__pyx_v_dim]);
    } else {
      __pyx_t_5 = -1L;
    }
    (__pyx_v_dst->suboffsets[__pyx_v_dim]) = __pyx_t_5;
  }

  /* "View.MemoryView":1065
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1082
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *__pyx_v_memview) {
  __Pyx_memviewslice __pyx_v_memviewslice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_copy", 0);

  /* "View.MemoryView":1085
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)             # <<<<<<<<<<<<<<
 *     return memoryview_copy_from_slice(memview, &memviewslice)
 * 
 */
  __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_memviewslice));

  /* "View.MemoryView":1086
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)
 *     return memoryview_copy_from_slice(memview, &memviewslice)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_memoryview_copy_object_from_slice(__pyx_v_memview, (&__pyx_v_memviewslice)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1082
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1089
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_memviewslice) {
  PyObject *(*__pyx_v_to_object_func)(char *);
  int (*__pyx_v_to_dtype_func)(char *, PyObject *);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *(*__pyx_t_3)(char *);
  int (*__pyx_t_4)(char *, PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_copy_from_slice", 0);

  /* "View.MemoryView":1096
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1097
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func             # <<<<<<<<<<<<<<
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 */
    __pyx_t_3 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_object_func;
    __pyx_v_to_object_func = __pyx_t_3;

    /* "View.MemoryView":1098
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func             # <<<<<<<<<<<<<<
 *     else:
 *         to_object_func = NULL
 */
    __pyx_t_4 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_dtype_func;
    __pyx_v_to_dtype_func = __pyx_t_4;

    /* "View.MemoryView":1096
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1100
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 *         to_object_func = NULL             # <<<<<<<<<<<<<<
 *         to_dtype_func = NULL
 * 
 */
  /*else*/ {
    __pyx_v_to_object_func = NULL;

    /* "View.MemoryView":1101
 *     else:
 *         to_object_func = NULL
 *         to_dtype_func = NULL             # <<<<<<<<<<<<<<
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 */
    __pyx_v_to_dtype_func = NULL;
  }
  __pyx_L3:;

  /* "View.MemoryView":1103
 *         to_dtype_func = NULL
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,             # <<<<<<<<<<<<<<
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)
 */
  __Pyx_XDECREF(__pyx_r);

  /* "View.MemoryView":1105
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __pyx_memoryview_fromslice((__pyx_v_memviewslice[0]), __pyx_v_memview->view.ndim, __pyx_v_to_object_func, __pyx_v_to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1089
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy_from_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1111
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

static Py_ssize_t abs_py_ssize_t(Py_ssize_t __pyx_v_arg) {
  Py_ssize_t __pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":1112
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_arg < 0) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1113
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:
 *         return -arg             # <<<<<<<<<<<<<<
 *     else:
 *         return arg
 */
    __pyx_r = (-__pyx_v_arg);
    goto __pyx_L0;

    /* "View.MemoryView":1112
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */
  }

  /* "View.MemoryView":1115
 *         return -arg
 *     else:
 *         return arg             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_get_best_slice_order')
 */
  /*else*/ {
    __pyx_r = __pyx_v_arg;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1111
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1118
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

static char __pyx_get_best_slice_order(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim) {
  int __pyx_v_i;
  Py_ssize_t __pyx_v_c_stride;
  Py_ssize_t __pyx_v_f_stride;
  char __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1123
 *     """
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t f_stride = 0
 * 
 */
  __pyx_v_c_stride = 0;

  /* "View.MemoryView":1124
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0
 *     cdef Py_ssize_t f_stride = 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_f_stride = 0;

  /* "View.MemoryView":1126
 *     cdef Py_ssize_t f_stride = 0
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1127
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1128
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_c_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1129
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
      goto __pyx_L4_break;

      /* "View.MemoryView":1127
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L4_break:;

  /* "View.MemoryView":1131
 *             break
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 */
  __pyx_t_1 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1132
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1133
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_f_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1134
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 */
      goto __pyx_L7_break;

      /* "View.MemoryView":1132
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L7_break:;

  /* "View.MemoryView":1136
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  __pyx_t_2 = ((abs_py_ssize_t(__pyx_v_c_stride) <= abs_py_ssize_t(__pyx_v_f_stride)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1137
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 *         return 'C'             # <<<<<<<<<<<<<<
 *     else:
 *         return 'F'
 */
    __pyx_r = 'C';
    goto __pyx_L0;

    /* "View.MemoryView":1136
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  }

  /* "View.MemoryView":1139
 *         return 'C'
 *     else:
 *         return 'F'             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  /*else*/ {
    __pyx_r = 'F';
    goto __pyx_L0;
  }

  /* "View.MemoryView":1118
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1142
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

static void _copy_strided_to_strided(char *__pyx_v_src_data, Py_ssize_t *__pyx_v_src_strides, char *__pyx_v_dst_data, Py_ssize_t *__pyx_v_dst_strides, Py_ssize_t *__pyx_v_src_shape, Py_ssize_t *__pyx_v_dst_shape, int __pyx_v_ndim, size_t __pyx_v_itemsize) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED Py_ssize_t __pyx_v_src_extent;
  Py_ssize_t __pyx_v_dst_extent;
  Py_ssize_t __pyx_v_src_stride;
  Py_ssize_t __pyx_v_dst_stride;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;

  /* "View.MemoryView":1149
 * 
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 */
  __pyx_v_src_extent = (__pyx_v_src_shape[0]);

  /* "View.MemoryView":1150
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 */
  __pyx_v_dst_extent = (__pyx_v_dst_shape[0]);

  /* "View.MemoryView":1151
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 */
  __pyx_v_src_stride = (__pyx_v_src_strides[0]);

  /* "View.MemoryView":1152
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_dst_stride = (__pyx_v_dst_strides[0]);

  /* "View.MemoryView":1154
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1155
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    __pyx_t_2 = ((__pyx_v_src_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_dst_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }

    /* "View.MemoryView":1156
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):             # <<<<<<<<<<<<<<
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 */
    __pyx_t_2 = (((size_t)__pyx_v_src_stride) == __pyx_v_itemsize);
    if (__pyx_t_2) {
      __pyx_t_2 = (__pyx_v_itemsize == ((size_t)__pyx_v_dst_stride));
    }
    __pyx_t_3 = (__pyx_t_2 != 0);
    __pyx_t_1 = __pyx_t_3;
    __pyx_L5_bool_binop_done:;

    /* "View.MemoryView":1155
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    if (__pyx_t_1) {

      /* "View.MemoryView":1157
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)             # <<<<<<<<<<<<<<
 *        else:
 *            for i in range(dst_extent):
 */
      (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, (__pyx_v_itemsize * __pyx_v_dst_extent)));

      /* "View.MemoryView":1155
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
      goto __pyx_L4;
    }

    /* "View.MemoryView":1159
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 *            for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 */
    /*else*/ {
      __pyx_t_4 = __pyx_v_dst_extent;
      __pyx_t_5 = __pyx_t_4;
      for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
        __pyx_v_i = __pyx_t_6;

        /* "View.MemoryView":1160
 *        else:
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)             # <<<<<<<<<<<<<<
 *                src_data += src_stride
 *                dst_data += dst_stride
 */
        (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, __pyx_v_itemsize));

        /* "View.MemoryView":1161
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride             # <<<<<<<<<<<<<<
 *                dst_data += dst_stride
 *     else:
 */
        __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

        /* "View.MemoryView":1162
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 *                dst_data += dst_stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(dst_extent):
 */
        __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
      }
    }
    __pyx_L4:;

    /* "View.MemoryView":1154
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1164
 *                dst_data += dst_stride
 *     else:
 *         for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *             _copy_strided_to_strided(src_data, src_strides + 1,
 *                                      dst_data, dst_strides + 1,
 */
  /*else*/ {
    __pyx_t_4 = __pyx_v_dst_extent;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "View.MemoryView":1165
 *     else:
 *         for i in range(dst_extent):
 *             _copy_strided_to_strided(src_data, src_strides + 1,             # <<<<<<<<<<<<<<
 *                                      dst_data, dst_strides + 1,
 *                                      src_shape + 1, dst_shape + 1,
 */
      _copy_strided_to_strided(__pyx_v_src_data, (__pyx_v_src_strides + 1), __pyx_v_dst_data, (__pyx_v_dst_strides + 1), (__pyx_v_src_shape + 1), (__pyx_v_dst_shape + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize);

      /* "View.MemoryView":1169
 *                                      src_shape + 1, dst_shape + 1,
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride             # <<<<<<<<<<<<<<
 *             dst_data += dst_stride
 * 
 */
      __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

      /* "View.MemoryView":1170
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride
 *             dst_data += dst_stride             # <<<<<<<<<<<<<<
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,
 */
      __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1142
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

  /* function exit code */
}

/* "View.MemoryView":1172
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

static void copy_strided_to_strided(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize) {

  /* "View.MemoryView":1175
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 *     _copy_strided_to_strided(src.data, src.strides, dst.data, dst.strides,             # <<<<<<<<<<<<<<
 *                              src.shape, dst.shape, ndim, itemsize)
 * 
 */
  _copy_strided_to_strided(__pyx_v_src->data, __pyx_v_src->strides, __pyx_v_dst->data, __pyx_v_dst->strides, __pyx_v_src->shape, __pyx_v_dst->shape, __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1172
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1179
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *__pyx_v_src, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_size;
  Py_ssize_t __pyx_r;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;

  /* "View.MemoryView":1181
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     for shape in src.shape[:ndim]:
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_size = __pyx_t_1;

  /* "View.MemoryView":1183
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 * 
 *     for shape in src.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         size *= shape
 * 
 */
  __pyx_t_3 = (__pyx_v_src->shape + __pyx_v_ndim);
  for (__pyx_t_4 = __pyx_v_src->shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
    __pyx_t_2 = __pyx_t_4;
    __pyx_v_shape = (__pyx_t_2[0]);

    /* "View.MemoryView":1184
 * 
 *     for shape in src.shape[:ndim]:
 *         size *= shape             # <<<<<<<<<<<<<<
 * 
 *     return size
 */
    __pyx_v_size = (__pyx_v_size * __pyx_v_shape);
  }

  /* "View.MemoryView":1186
 *         size *= shape
 * 
 *     return size             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_fill_contig_strides_array')
 */
  __pyx_r = __pyx_v_size;
  goto __pyx_L0;

  /* "View.MemoryView":1179
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1189
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, Py_ssize_t __pyx_v_stride, int __pyx_v_ndim, char __pyx_v_order) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1198
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
  __pyx_t_1 = ((__pyx_v_order == 'F') != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1199
 * 
 *     if order == 'F':
 *         for idx in range(ndim):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
    __pyx_t_2 = __pyx_v_ndim;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_idx = __pyx_t_4;

      /* "View.MemoryView":1200
 *     if order == 'F':
 *         for idx in range(ndim):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 *     else:
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1201
 *         for idx in range(ndim):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }

    /* "View.MemoryView":1198
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1203
 *             stride *= shape[idx]
 *     else:
 *         for idx in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
  /*else*/ {
    for (__pyx_t_2 = (__pyx_v_ndim - 1); __pyx_t_2 > -1; __pyx_t_2-=1) {
      __pyx_v_idx = __pyx_t_2;

      /* "View.MemoryView":1204
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 * 
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1205
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 * 
 *     return stride
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1207
 *             stride *= shape[idx]
 * 
 *     return stride             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 */
  __pyx_r = __pyx_v_stride;
  goto __pyx_L0;

  /* "View.MemoryView":1189
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1210
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_tmpslice, char __pyx_v_order, int __pyx_v_ndim) {
  int __pyx_v_i;
  void *__pyx_v_result;
  size_t __pyx_v_itemsize;
  size_t __pyx_v_size;
  void *__pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  struct __pyx_memoryview_obj *__pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":1221
 *     cdef void *result
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1222
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef size_t size = slice_get_size(src, ndim)             # <<<<<<<<<<<<<<
 * 
 *     result = malloc(size)
 */
  __pyx_v_size = __pyx_memoryview_slice_get_size(__pyx_v_src, __pyx_v_ndim);

  /* "View.MemoryView":1224
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 *     result = malloc(size)             # <<<<<<<<<<<<<<
 *     if not result:
 *         _err(MemoryError, NULL)
 */
  __pyx_v_result = malloc(__pyx_v_size);

  /* "View.MemoryView":1225
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_result != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1226
 *     result = malloc(size)
 *     if not result:
 *         _err(MemoryError, NULL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_err(__pyx_builtin_MemoryError, NULL); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 1226, __pyx_L1_error)

    /* "View.MemoryView":1225
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */
  }

  /* "View.MemoryView":1229
 * 
 * 
 *     tmpslice.data = <char *> result             # <<<<<<<<<<<<<<
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 */
  __pyx_v_tmpslice->data = ((char *)__pyx_v_result);

  /* "View.MemoryView":1230
 * 
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview             # <<<<<<<<<<<<<<
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 */
  __pyx_t_4 = __pyx_v_src->memview;
  __pyx_v_tmpslice->memview = __pyx_t_4;

  /* "View.MemoryView":1231
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1232
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]             # <<<<<<<<<<<<<<
 *         tmpslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_tmpslice->shape[__pyx_v_i]) = (__pyx_v_src->shape[__pyx_v_i]);

    /* "View.MemoryView":1233
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,
 */
    (__pyx_v_tmpslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1235
 *         tmpslice.suboffsets[i] = -1
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,             # <<<<<<<<<<<<<<
 *                               ndim, order)
 * 
 */
  (void)(__pyx_fill_contig_strides_array((&(__pyx_v_tmpslice->shape[0])), (&(__pyx_v_tmpslice->strides[0])), __pyx_v_itemsize, __pyx_v_ndim, __pyx_v_order));

  /* "View.MemoryView":1239
 * 
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1240
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    __pyx_t_2 = (((__pyx_v_tmpslice->shape[__pyx_v_i]) == 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1241
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     if slice_is_contig(src[0], order, ndim):
 */
      (__pyx_v_tmpslice->strides[__pyx_v_i]) = 0;

      /* "View.MemoryView":1240
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    }
  }

  /* "View.MemoryView":1243
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
  __pyx_t_2 = (__pyx_memviewslice_is_contig((__pyx_v_src[0]), __pyx_v_order, __pyx_v_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1244
 * 
 *     if slice_is_contig(src[0], order, ndim):
 *         memcpy(result, src.data, size)             # <<<<<<<<<<<<<<
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 */
    (void)(memcpy(__pyx_v_result, __pyx_v_src->data, __pyx_v_size));

    /* "View.MemoryView":1243
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":1246
 *         memcpy(result, src.data, size)
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  /*else*/ {
    copy_strided_to_strided(__pyx_v_src, __pyx_v_tmpslice, __pyx_v_ndim, __pyx_v_itemsize);
  }
  __pyx_L9:;

  /* "View.MemoryView":1248
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":1210
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.copy_data_to_temp", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1253
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

static int __pyx_memoryview_err_extents(int __pyx_v_i, Py_ssize_t __pyx_v_extent1, Py_ssize_t __pyx_v_extent2) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_extents", 0);

  /* "View.MemoryView":1256
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 *                                                         (i, extent1, extent2))             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err_dim')
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_extent1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_extent2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;

  /* "View.MemoryView":1255
 * cdef int _err_extents(int i, Py_ssize_t extent1,
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %             # <<<<<<<<<<<<<<
 *                                                         (i, extent1, extent2))
 * 
 */
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_4, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __PYX_ERR(1, 1255, __pyx_L1_error)

  /* "View.MemoryView":1253
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1259
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

static int __pyx_memoryview_err_dim(PyObject *__pyx_v_error, char *__pyx_v_msg, int __pyx_v_dim) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_dim", 0);
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1260
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:
 *     raise error(msg.decode('ascii') % dim)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err')
 */
  __pyx_t_2 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyUnicode_Format(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_v_error);
  __pyx_t_3 = __pyx_v_error; __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 1260, __pyx_L1_error)

  /* "View.MemoryView":1259
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_dim", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1263
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

static int __pyx_memoryview_err(PyObject *__pyx_v_error, char *__pyx_v_msg) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err", 0);
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1264
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_msg != NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":1265
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))             # <<<<<<<<<<<<<<
 *     else:
 *         raise error
 */
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_error);
    __pyx_t_4 = __pyx_v_error; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 1265, __pyx_L1_error)

    /* "View.MemoryView":1264
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */
  }

  /* "View.MemoryView":1267
 *         raise error(msg.decode('ascii'))
 *     else:
 *         raise error             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_contents')
 */
  /*else*/ {
    __Pyx_Raise(__pyx_v_error, 0, 0, 0);
    __PYX_ERR(1, 1267, __pyx_L1_error)
  }

  /* "View.MemoryView":1263
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView._err", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1270
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

static int __pyx_memoryview_copy_contents(__Pyx_memviewslice __pyx_v_src, __Pyx_memviewslice __pyx_v_dst, int __pyx_v_src_ndim, int __pyx_v_dst_ndim, int __pyx_v_dtype_is_object) {
  void *__pyx_v_tmpdata;
  size_t __pyx_v_itemsize;
  int __pyx_v_i;
  char __pyx_v_order;
  int __pyx_v_broadcasting;
  int __pyx_v_direct_copy;
  __Pyx_memviewslice __pyx_v_tmp;
  int __pyx_v_ndim;
  int __pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":1278
 *     Check for overlapping memory and verify the shapes.
 *     """
 *     cdef void *tmpdata = NULL             # <<<<<<<<<<<<<<
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 */
  __pyx_v_tmpdata = NULL;

  /* "View.MemoryView":1279
 *     """
 *     cdef void *tmpdata = NULL
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 */
  __pyx_t_1 = __pyx_v_src.memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1281
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)             # <<<<<<<<<<<<<<
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False
 */
  __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_src), __pyx_v_src_ndim);

  /* "View.MemoryView":1282
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False             # <<<<<<<<<<<<<<
 *     cdef bint direct_copy = False
 *     cdef __Pyx_memviewslice tmp
 */
  __pyx_v_broadcasting = 0;

  /* "View.MemoryView":1283
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False             # <<<<<<<<<<<<<<
 *     cdef __Pyx_memviewslice tmp
 * 
 */
  __pyx_v_direct_copy = 0;

  /* "View.MemoryView":1286
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
  __pyx_t_2 = ((__pyx_v_src_ndim < __pyx_v_dst_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1287
 * 
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_src), __pyx_v_src_ndim, __pyx_v_dst_ndim);

    /* "View.MemoryView":1286
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1288
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  __pyx_t_2 = ((__pyx_v_dst_ndim < __pyx_v_src_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1289
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)             # <<<<<<<<<<<<<<
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_dst), __pyx_v_dst_ndim, __pyx_v_src_ndim);

    /* "View.MemoryView":1288
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":1291
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
  __pyx_t_3 = __pyx_v_dst_ndim;
  __pyx_t_4 = __pyx_v_src_ndim;
  if (((__pyx_t_3 > __pyx_t_4) != 0)) {
    __pyx_t_5 = __pyx_t_3;
  } else {
    __pyx_t_5 = __pyx_t_4;
  }
  __pyx_v_ndim = __pyx_t_5;

  /* "View.MemoryView":1293
 *     cdef int ndim = max(src_ndim, dst_ndim)
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 */
  __pyx_t_5 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_5;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1294
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) != (__pyx_v_dst.shape[__pyx_v_i])) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1295
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
      __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) == 1) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":1296
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 *                 broadcasting = True             # <<<<<<<<<<<<<<
 *                 src.strides[i] = 0
 *             else:
 */
        __pyx_v_broadcasting = 1;

        /* "View.MemoryView":1297
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 *                 src.strides[i] = 0             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 */
        (__pyx_v_src.strides[__pyx_v_i]) = 0;

        /* "View.MemoryView":1295
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":1299
 *                 src.strides[i] = 0
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])             # <<<<<<<<<<<<<<
 * 
 *         if src.suboffsets[i] >= 0:
 */
      /*else*/ {
        __pyx_t_6 = __pyx_memoryview_err_extents(__pyx_v_i, (__pyx_v_dst.shape[__pyx_v_i]), (__pyx_v_src.shape[__pyx_v_i])); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 1299, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":1294
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    }

    /* "View.MemoryView":1301
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */
    __pyx_t_2 = (((__pyx_v_src.suboffsets[__pyx_v_i]) >= 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1302
 * 
 *         if src.suboffsets[i] >= 0:
 *             _err_dim(ValueError, "Dimension %d is not direct", i)             # <<<<<<<<<<<<<<
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 */
      __pyx_t_6 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Dimension %d is not direct"), __pyx_v_i); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 1302, __pyx_L1_error)

      /* "View.MemoryView":1301
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */
    }
  }

  /* "View.MemoryView":1304
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  __pyx_t_2 = (__pyx_slices_overlap((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1306
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    __pyx_t_2 = ((!(__pyx_memviewslice_is_contig(__pyx_v_src, __pyx_v_order, __pyx_v_ndim) != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1307
 * 
 *         if not slice_is_contig(src, order, ndim):
 *             order = get_best_order(&dst, ndim)             # <<<<<<<<<<<<<<
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 */
      __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim);

      /* "View.MemoryView":1306
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    }

    /* "View.MemoryView":1309
 *             order = get_best_order(&dst, ndim)
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)             # <<<<<<<<<<<<<<
 *         src = tmp
 * 
 */
    __pyx_t_7 = __pyx_memoryview_copy_data_to_temp((&__pyx_v_src), (&__pyx_v_tmp), __pyx_v_order, __pyx_v_ndim); if (unlikely(__pyx_t_7 == ((void *)NULL))) __PYX_ERR(1, 1309, __pyx_L1_error)
    __pyx_v_tmpdata = __pyx_t_7;

    /* "View.MemoryView":1310
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 *         src = tmp             # <<<<<<<<<<<<<<
 * 
 *     if not broadcasting:
 */
    __pyx_v_src = __pyx_v_tmp;

    /* "View.MemoryView":1304
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  }

  /* "View.MemoryView":1312
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_broadcasting != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1315
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'C', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1316
 * 
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)             # <<<<<<<<<<<<<<
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'C', __pyx_v_ndim);

      /* "View.MemoryView":1315
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
      goto __pyx_L12;
    }

    /* "View.MemoryView":1317
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'F', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1318
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)             # <<<<<<<<<<<<<<
 * 
 *         if direct_copy:
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'F', __pyx_v_ndim);

      /* "View.MemoryView":1317
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    }
    __pyx_L12:;

    /* "View.MemoryView":1320
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    __pyx_t_2 = (__pyx_v_direct_copy != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1322
 *         if direct_copy:
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

      /* "View.MemoryView":1323
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))             # <<<<<<<<<<<<<<
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 */
      (void)(memcpy(__pyx_v_dst.data, __pyx_v_src.data, __pyx_memoryview_slice_get_size((&__pyx_v_src), __pyx_v_ndim)));

      /* "View.MemoryView":1324
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 *             free(tmpdata)
 *             return 0
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

      /* "View.MemoryView":1325
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
      free(__pyx_v_tmpdata);

      /* "View.MemoryView":1326
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "View.MemoryView":1320
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    }

    /* "View.MemoryView":1312
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1328
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = (__pyx_v_order == 'F');
  if (__pyx_t_2) {
    __pyx_t_2 = ('F' == __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim));
  }
  __pyx_t_8 = (__pyx_t_2 != 0);
  if (__pyx_t_8) {

    /* "View.MemoryView":1331
 * 
 * 
 *         transpose_memslice(&src)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&dst)
 * 
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_src)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(1, 1331, __pyx_L1_error)

    /* "View.MemoryView":1332
 * 
 *         transpose_memslice(&src)
 *         transpose_memslice(&dst)             # <<<<<<<<<<<<<<
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_dst)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(1, 1332, __pyx_L1_error)

    /* "View.MemoryView":1328
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1334
 *         transpose_memslice(&dst)
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1335
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)             # <<<<<<<<<<<<<<
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 */
  copy_strided_to_strided((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1336
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 *     free(tmpdata)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1338
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 *     free(tmpdata)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  free(__pyx_v_tmpdata);

  /* "View.MemoryView":1339
 * 
 *     free(tmpdata)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1270
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.memoryview_copy_contents", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1342
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim, int __pyx_v_ndim_other) {
  int __pyx_v_i;
  int __pyx_v_offset;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "View.MemoryView":1346
 *                             int ndim_other) nogil:
 *     cdef int i
 *     cdef int offset = ndim_other - ndim             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_offset = (__pyx_v_ndim_other - __pyx_v_ndim);

  /* "View.MemoryView":1348
 *     cdef int offset = ndim_other - ndim
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1349
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]             # <<<<<<<<<<<<<<
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 */
    (__pyx_v_mslice->shape[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->shape[__pyx_v_i]);

    /* "View.MemoryView":1350
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 */
    (__pyx_v_mslice->strides[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->strides[__pyx_v_i]);

    /* "View.MemoryView":1351
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(offset):
 */
    (__pyx_v_mslice->suboffsets[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->suboffsets[__pyx_v_i]);
  }

  /* "View.MemoryView":1353
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 *     for i in range(offset):             # <<<<<<<<<<<<<<
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 */
  __pyx_t_1 = __pyx_v_offset;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1354
 * 
 *     for i in range(offset):
 *         mslice.shape[i] = 1             # <<<<<<<<<<<<<<
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1
 */
    (__pyx_v_mslice->shape[__pyx_v_i]) = 1;

    /* "View.MemoryView":1355
 *     for i in range(offset):
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_mslice->strides[__pyx_v_i]) = (__pyx_v_mslice->strides[0]);

    /* "View.MemoryView":1356
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_mslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1342
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1364
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_dtype_is_object, int __pyx_v_ndim, int __pyx_v_inc) {
  int __pyx_t_1;

  /* "View.MemoryView":1368
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */
  __pyx_t_1 = (__pyx_v_dtype_is_object != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1369
 * 
 *     if dtype_is_object:
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,             # <<<<<<<<<<<<<<
 *                                            dst.strides, ndim, inc)
 * 
 */
    __pyx_memoryview_refcount_objects_in_slice_with_gil(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_inc);

    /* "View.MemoryView":1368
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */
  }

  /* "View.MemoryView":1364
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

  /* function exit code */
}

/* "View.MemoryView":1373
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  __Pyx_RefNannyDeclarations
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("refcount_objects_in_slice_with_gil", 0);

  /* "View.MemoryView":1376
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 *     refcount_objects_in_slice(data, shape, strides, ndim, inc)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 */
  __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, __pyx_v_shape, __pyx_v_strides, __pyx_v_ndim, __pyx_v_inc);

  /* "View.MemoryView":1373
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "View.MemoryView":1379
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

static void __pyx_memoryview_refcount_objects_in_slice(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("refcount_objects_in_slice", 0);

  /* "View.MemoryView":1383
 *     cdef Py_ssize_t i
 * 
 *     for i in range(shape[0]):             # <<<<<<<<<<<<<<
 *         if ndim == 1:
 *             if inc:
 */
  __pyx_t_1 = (__pyx_v_shape[0]);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1384
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
    __pyx_t_4 = ((__pyx_v_ndim == 1) != 0);
    if (__pyx_t_4) {

      /* "View.MemoryView":1385
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
      __pyx_t_4 = (__pyx_v_inc != 0);
      if (__pyx_t_4) {

        /* "View.MemoryView":1386
 *         if ndim == 1:
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])
 */
        Py_INCREF((((PyObject **)__pyx_v_data)[0]));

        /* "View.MemoryView":1385
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
        goto __pyx_L6;
      }

      /* "View.MemoryView":1388
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 */
      /*else*/ {
        Py_DECREF((((PyObject **)__pyx_v_data)[0]));
      }
      __pyx_L6:;

      /* "View.MemoryView":1384
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":1390
 *                 Py_DECREF((<PyObject **> data)[0])
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                       ndim - 1, inc)
 * 
 */
    /*else*/ {

      /* "View.MemoryView":1391
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 *                                       ndim - 1, inc)             # <<<<<<<<<<<<<<
 * 
 *         data += strides[0]
 */
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_inc);
    }
    __pyx_L5:;

    /* "View.MemoryView":1393
 *                                       ndim - 1, inc)
 * 
 *         data += strides[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_data = (__pyx_v_data + (__pyx_v_strides[0]));
  }

  /* "View.MemoryView":1379
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1399
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item, int __pyx_v_dtype_is_object) {

  /* "View.MemoryView":1402
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1403
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,             # <<<<<<<<<<<<<<
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)
 */
  __pyx_memoryview__slice_assign_scalar(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_itemsize, __pyx_v_item);

  /* "View.MemoryView":1405
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1399
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1409
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

static void __pyx_memoryview__slice_assign_scalar(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_extent;
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;

  /* "View.MemoryView":1413
 *                               size_t itemsize, void *item) nogil:
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t extent = shape[0]
 * 
 */
  __pyx_v_stride = (__pyx_v_strides[0]);

  /* "View.MemoryView":1414
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]
 *     cdef Py_ssize_t extent = shape[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_extent = (__pyx_v_shape[0]);

  /* "View.MemoryView":1416
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1417
 * 
 *     if ndim == 1:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             memcpy(data, item, itemsize)
 *             data += stride
 */
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1418
 *     if ndim == 1:
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)             # <<<<<<<<<<<<<<
 *             data += stride
 *     else:
 */
      (void)(memcpy(__pyx_v_data, __pyx_v_item, __pyx_v_itemsize));

      /* "View.MemoryView":1419
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 *             data += stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(extent):
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }

    /* "View.MemoryView":1416
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1421
 *             data += stride
 *     else:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 */
  /*else*/ {
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1422
 *     else:
 *         for i in range(extent):
 *             _slice_assign_scalar(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                 ndim - 1, itemsize, item)
 *             data += stride
 */
      __pyx_memoryview__slice_assign_scalar(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize, __pyx_v_item);

      /* "View.MemoryView":1424
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 *             data += stride             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1409
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum = {"__pyx_unpickle_Enum", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Enum") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__52, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xb068931, 0x82a3537, 0x6ae9995):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_MemviewEnum_type), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_4 = __pyx_unpickle_Enum__set_state(((struct __pyx_MemviewEnum_obj *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xb068931, 0x82a3537, 0x6ae9995) = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 1) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "BufferFormatFromTypeInfo":1465
 * 
 * @cname('__pyx_format_from_typeinfo')
 * cdef bytes format_from_typeinfo(__Pyx_TypeInfo *type):             # <<<<<<<<<<<<<<
 *     cdef __Pyx_StructField *field
 *     cdef __pyx_typeinfo_string fmt
 */

static PyObject *__pyx_format_from_typeinfo(__Pyx_TypeInfo *__pyx_v_type) {
  __Pyx_StructField *__pyx_v_field;
  struct __pyx_typeinfo_string __pyx_v_fmt;
  PyObject *__pyx_v_part = 0;
  PyObject *__pyx_v_result = 0;
  PyObject *__pyx_v_alignment = NULL;
  PyObject *__pyx_v_parts = NULL;
  PyObject *__pyx_v_extents = NULL;
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_StructField *__pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("format_from_typeinfo", 0);

  /* "BufferFormatFromTypeInfo":1470
 *     cdef bytes part, result
 * 
 *     if type.typegroup == 'S':             # <<<<<<<<<<<<<<
 *         assert type.fields != NULL
 *         assert type.fields.type != NULL
 */
  __pyx_t_1 = ((__pyx_v_type->typegroup == 'S') != 0);
  if (__pyx_t_1) {

    /* "BufferFormatFromTypeInfo":1471
 * 
 *     if type.typegroup == 'S':
 *         assert type.fields != NULL             # <<<<<<<<<<<<<<
 *         assert type.fields.type != NULL
 * 
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      if (unlikely(!((__pyx_v_type->fields != NULL) != 0))) {
        PyErr_SetNone(PyExc_AssertionError);
        __PYX_ERR(1, 1471, __pyx_L1_error)
      }
    }
    #endif

    /* "BufferFormatFromTypeInfo":1472
 *     if type.typegroup == 'S':
 *         assert type.fields != NULL
 *         assert type.fields.type != NULL             # <<<<<<<<<<<<<<
 * 
 *         if type.flags & __PYX_BUF_FLAGS_PACKED_STRUCT:
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      if (unlikely(!((__pyx_v_type->fields->type != NULL) != 0))) {
        PyErr_SetNone(PyExc_AssertionError);
        __PYX_ERR(1, 1472, __pyx_L1_error)
      }
    }
    #endif

    /* "BufferFormatFromTypeInfo":1474
 *         assert type.fields.type != NULL
 * 
 *         if type.flags & __PYX_BUF_FLAGS_PACKED_STRUCT:             # <<<<<<<<<<<<<<
 *             alignment = b'^'
 *         else:
 */
    __pyx_t_1 = ((__pyx_v_type->flags & __PYX_BUF_FLAGS_PACKED_STRUCT) != 0);
    if (__pyx_t_1) {

      /* "BufferFormatFromTypeInfo":1475
 * 
 *         if type.flags & __PYX_BUF_FLAGS_PACKED_STRUCT:
 *             alignment = b'^'             # <<<<<<<<<<<<<<
 *         else:
 *             alignment = b''
 */
      __Pyx_INCREF(__pyx_kp_b__53);
      __pyx_v_alignment = __pyx_kp_b__53;

      /* "BufferFormatFromTypeInfo":1474
 *         assert type.fields.type != NULL
 * 
 *         if type.flags & __PYX_BUF_FLAGS_PACKED_STRUCT:             # <<<<<<<<<<<<<<
 *             alignment = b'^'
 *         else:
 */
      goto __pyx_L4;
    }

    /* "BufferFormatFromTypeInfo":1477
 *             alignment = b'^'
 *         else:
 *             alignment = b''             # <<<<<<<<<<<<<<
 * 
 *         parts = [b"T{"]
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_kp_b__54);
      __pyx_v_alignment = __pyx_kp_b__54;
    }
    __pyx_L4:;

    /* "BufferFormatFromTypeInfo":1479
 *             alignment = b''
 * 
 *         parts = [b"T{"]             # <<<<<<<<<<<<<<
 *         field = type.fields
 * 
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1479, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_kp_b_T);
    __Pyx_GIVEREF(__pyx_kp_b_T);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_kp_b_T);
    __pyx_v_parts = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "BufferFormatFromTypeInfo":1480
 * 
 *         parts = [b"T{"]
 *         field = type.fields             # <<<<<<<<<<<<<<
 * 
 *         while field.type:
 */
    __pyx_t_3 = __pyx_v_type->fields;
    __pyx_v_field = __pyx_t_3;

    /* "BufferFormatFromTypeInfo":1482
 *         field = type.fields
 * 
 *         while field.type:             # <<<<<<<<<<<<<<
 *             part = format_from_typeinfo(field.type)
 *             parts.append(part + b':' + field.name + b':')
 */
    while (1) {
      __pyx_t_1 = (__pyx_v_field->type != 0);
      if (!__pyx_t_1) break;

      /* "BufferFormatFromTypeInfo":1483
 * 
 *         while field.type:
 *             part = format_from_typeinfo(field.type)             # <<<<<<<<<<<<<<
 *             parts.append(part + b':' + field.name + b':')
 *             field += 1
 */
      __pyx_t_2 = __pyx_format_from_typeinfo(__pyx_v_field->type); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1483, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_part, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "BufferFormatFromTypeInfo":1484
 *         while field.type:
 *             part = format_from_typeinfo(field.type)
 *             parts.append(part + b':' + field.name + b':')             # <<<<<<<<<<<<<<
 *             field += 1
 * 
 */
      __pyx_t_2 = PyNumber_Add(__pyx_v_part, __pyx_kp_b__55); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1484, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_field->name); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1484, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = PyNumber_Add(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1484, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyNumber_Add(__pyx_t_5, __pyx_kp_b__55); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1484, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_parts, __pyx_t_4); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 1484, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "BufferFormatFromTypeInfo":1485
 *             part = format_from_typeinfo(field.type)
 *             parts.append(part + b':' + field.name + b':')
 *             field += 1             # <<<<<<<<<<<<<<
 * 
 *         result = alignment.join(parts) + b'}'
 */
      __pyx_v_field = (__pyx_v_field + 1);
    }

    /* "BufferFormatFromTypeInfo":1487
 *             field += 1
 * 
 *         result = alignment.join(parts) + b'}'             # <<<<<<<<<<<<<<
 *     else:
 *         fmt = __Pyx_TypeInfoToFormat(type)
 */
    __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_v_alignment, __pyx_v_parts); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyNumber_Add(__pyx_t_4, __pyx_kp_b__56); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(1, 1487, __pyx_L1_error)
    __pyx_v_result = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "BufferFormatFromTypeInfo":1470
 *     cdef bytes part, result
 * 
 *     if type.typegroup == 'S':             # <<<<<<<<<<<<<<
 *         assert type.fields != NULL
 *         assert type.fields.type != NULL
 */
    goto __pyx_L3;
  }

  /* "BufferFormatFromTypeInfo":1489
 *         result = alignment.join(parts) + b'}'
 *     else:
 *         fmt = __Pyx_TypeInfoToFormat(type)             # <<<<<<<<<<<<<<
 *         if type.arraysize[0]:
 *             extents = [unicode(type.arraysize[i]) for i in range(type.ndim)]
 */
  /*else*/ {
    __pyx_v_fmt = __Pyx_TypeInfoToFormat(__pyx_v_type);

    /* "BufferFormatFromTypeInfo":1490
 *     else:
 *         fmt = __Pyx_TypeInfoToFormat(type)
 *         if type.arraysize[0]:             # <<<<<<<<<<<<<<
 *             extents = [unicode(type.arraysize[i]) for i in range(type.ndim)]
 *             result = (u"(%s)" % u','.join(extents)).encode('ascii') + fmt.string
 */
    __pyx_t_1 = ((__pyx_v_type->arraysize[0]) != 0);
    if (__pyx_t_1) {

      /* "BufferFormatFromTypeInfo":1491
 *         fmt = __Pyx_TypeInfoToFormat(type)
 *         if type.arraysize[0]:
 *             extents = [unicode(type.arraysize[i]) for i in range(type.ndim)]             # <<<<<<<<<<<<<<
 *             result = (u"(%s)" % u','.join(extents)).encode('ascii') + fmt.string
 *         else:
 */
      __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1491, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = __pyx_v_type->ndim;
      __pyx_t_8 = __pyx_t_7;
      for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
        __pyx_v_i = __pyx_t_9;
        __pyx_t_4 = __Pyx_PyInt_FromSize_t((__pyx_v_type->arraysize[__pyx_v_i])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1491, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_2 = __Pyx_PyObject_Unicode(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1491, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_2))) __PYX_ERR(1, 1491, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __pyx_v_extents = ((PyObject*)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "BufferFormatFromTypeInfo":1492
 *         if type.arraysize[0]:
 *             extents = [unicode(type.arraysize[i]) for i in range(type.ndim)]
 *             result = (u"(%s)" % u','.join(extents)).encode('ascii') + fmt.string             # <<<<<<<<<<<<<<
 *         else:
 *             result = fmt.string
 */
      __pyx_t_5 = PyUnicode_Join(__pyx_kp_u__57, __pyx_v_extents); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1492, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = PyUnicode_Format(__pyx_kp_u_s_2, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1492, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyUnicode_AsASCIIString(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1492, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_FromString(__pyx_v_fmt.string); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1492, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = PyNumber_Add(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1492, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(1, 1492, __pyx_L1_error)
      __pyx_v_result = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "BufferFormatFromTypeInfo":1490
 *     else:
 *         fmt = __Pyx_TypeInfoToFormat(type)
 *         if type.arraysize[0]:             # <<<<<<<<<<<<<<
 *             extents = [unicode(type.arraysize[i]) for i in range(type.ndim)]
 *             result = (u"(%s)" % u','.join(extents)).encode('ascii') + fmt.string
 */
      goto __pyx_L7;
    }

    /* "BufferFormatFromTypeInfo":1494
 *             result = (u"(%s)" % u','.join(extents)).encode('ascii') + fmt.string
 *         else:
 *             result = fmt.string             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_FromString(__pyx_v_fmt.string); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1494, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_result = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;
    }
    __pyx_L7:;
  }
  __pyx_L3:;

  /* "BufferFormatFromTypeInfo":1496
 *             result = fmt.string
 * 
 *     return result             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "BufferFormatFromTypeInfo":1465
 * 
 * @cname('__pyx_format_from_typeinfo')
 * cdef bytes format_from_typeinfo(__Pyx_TypeInfo *type):             # <<<<<<<<<<<<<<
 *     cdef __Pyx_StructField *field
 *     cdef __pyx_typeinfo_string fmt
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("BufferFormatFromTypeInfo.format_from_typeinfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_part);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_alignment);
  __Pyx_XDECREF(__pyx_v_parts);
  __Pyx_XDECREF(__pyx_v_extents);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer;

static PyObject *__pyx_tp_new_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *p;
  PyObject *o = __pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer->tp_new(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctreeContainer*)__pyx_vtabptr_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer;
  return o;
}

static void __pyx_tp_dealloc_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  PyObject_GC_Track(o);
  if (likely(__pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer)) __pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer->tp_dealloc(o); else __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer);
}

static int __pyx_tp_traverse_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer(PyObject *o, visitproc v, void *a) {
  int e;
  e = ((likely(__pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer)) ? ((__pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer->tp_traverse) ? __pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer)); if (e) return e;
  return 0;
}

static int __pyx_tp_clear_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer(PyObject *o) {
  if (likely(__pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer)) { if (__pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer->tp_clear) __pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer);
  return 0;
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_max_level(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_9max_level_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_max_level(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_9max_level_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_n_ref(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_5n_ref_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_n_ref(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_5n_ref_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer[] = {
  {"allocate_root", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_1allocate_root, METH_NOARGS, 0},
  {"clear_fileind", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_5clear_fileind, METH_NOARGS, 0},
  {"allocate_domains", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_10allocate_domains, METH_O, 0},
  {"finalize", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_12finalize, METH_VARARGS|METH_KEYWORDS, 0},
  {"add", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_14add, METH_VARARGS|METH_KEYWORDS, 0},
  {"recursively_count", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_16recursively_count, METH_NOARGS, 0},
  {"apply_domain", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_18apply_domain, METH_VARARGS|METH_KEYWORDS, 0},
  {"domain_ind", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_20domain_ind, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_22__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_24__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer[] = {
  {(char *)"max_level", __pyx_getprop_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_max_level, __pyx_setprop_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_max_level, (char *)0, 0},
  {(char *)"n_ref", __pyx_getprop_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_n_ref, __pyx_setprop_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_n_ref, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.particle_oct_container.ParticleOctreeContainer", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_7__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmap __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmap;

static PyObject *__pyx_tp_new_2yt_8geometry_22particle_oct_container_ParticleBitmap(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)o);
  p->__pyx_vtab = __pyx_vtabptr_2yt_8geometry_22particle_oct_container_ParticleBitmap;
  p->masks = Py_None; Py_INCREF(Py_None);
  p->particle_counts = Py_None; Py_INCREF(Py_None);
  p->counts = Py_None; Py_INCREF(Py_None);
  p->max_count = Py_None; Py_INCREF(Py_None);
  p->_last_selector = Py_None; Py_INCREF(Py_None);
  p->_last_return_values = Py_None; Py_INCREF(Py_None);
  p->_cached_octrees = Py_None; Py_INCREF(Py_None);
  p->_last_octree_subset = Py_None; Py_INCREF(Py_None);
  p->_last_oct_handler = Py_None; Py_INCREF(Py_None);
  p->_prev_octree_subset = Py_None; Py_INCREF(Py_None);
  p->_prev_oct_handler = Py_None; Py_INCREF(Py_None);
  p->bitmasks = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)Py_None); Py_INCREF(Py_None);
  p->collisions = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_2yt_8geometry_22particle_oct_container_ParticleBitmap(PyObject *o) {
  struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *p = (struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->masks);
  Py_CLEAR(p->particle_counts);
  Py_CLEAR(p->counts);
  Py_CLEAR(p->max_count);
  Py_CLEAR(p->_last_selector);
  Py_CLEAR(p->_last_return_values);
  Py_CLEAR(p->_cached_octrees);
  Py_CLEAR(p->_last_octree_subset);
  Py_CLEAR(p->_last_oct_handler);
  Py_CLEAR(p->_prev_octree_subset);
  Py_CLEAR(p->_prev_oct_handler);
  Py_CLEAR(p->bitmasks);
  Py_CLEAR(p->collisions);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_2yt_8geometry_22particle_oct_container_ParticleBitmap(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *p = (struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)o;
  if (p->masks) {
    e = (*v)(p->masks, a); if (e) return e;
  }
  if (p->particle_counts) {
    e = (*v)(p->particle_counts, a); if (e) return e;
  }
  if (p->counts) {
    e = (*v)(p->counts, a); if (e) return e;
  }
  if (p->max_count) {
    e = (*v)(p->max_count, a); if (e) return e;
  }
  if (p->_last_selector) {
    e = (*v)(p->_last_selector, a); if (e) return e;
  }
  if (p->_last_return_values) {
    e = (*v)(p->_last_return_values, a); if (e) return e;
  }
  if (p->_cached_octrees) {
    e = (*v)(p->_cached_octrees, a); if (e) return e;
  }
  if (p->_last_octree_subset) {
    e = (*v)(p->_last_octree_subset, a); if (e) return e;
  }
  if (p->_last_oct_handler) {
    e = (*v)(p->_last_oct_handler, a); if (e) return e;
  }
  if (p->_prev_octree_subset) {
    e = (*v)(p->_prev_octree_subset, a); if (e) return e;
  }
  if (p->_prev_oct_handler) {
    e = (*v)(p->_prev_oct_handler, a); if (e) return e;
  }
  if (p->bitmasks) {
    e = (*v)(((PyObject *)p->bitmasks), a); if (e) return e;
  }
  if (p->collisions) {
    e = (*v)(((PyObject *)p->collisions), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_2yt_8geometry_22particle_oct_container_ParticleBitmap(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *p = (struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)o;
  tmp = ((PyObject*)p->masks);
  p->masks = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->particle_counts);
  p->particle_counts = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->counts);
  p->counts = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->max_count);
  p->max_count = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_last_selector);
  p->_last_selector = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_last_return_values);
  p->_last_return_values = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_cached_octrees);
  p->_cached_octrees = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_last_octree_subset);
  p->_last_octree_subset = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_last_oct_handler);
  p->_last_oct_handler = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_prev_octree_subset);
  p->_prev_octree_subset = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_prev_oct_handler);
  p->_prev_oct_handler = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->bitmasks);
  p->bitmasks = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->collisions);
  p->collisions = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_nfiles(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6nfiles_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_nfiles(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6nfiles_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_index_order1(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order1_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_index_order1(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order1_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_index_order2(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order2_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_index_order2(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_12index_order2_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_masks(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5masks_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_masks(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5masks_3__set__(o, v);
  }
  else {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5masks_5__del__(o);
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_particle_counts(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15particle_counts_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_particle_counts(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15particle_counts_3__set__(o, v);
  }
  else {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15particle_counts_5__del__(o);
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_counts(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6counts_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_counts(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6counts_3__set__(o, v);
  }
  else {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6counts_5__del__(o);
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_max_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9max_count_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_max_count(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9max_count_3__set__(o, v);
  }
  else {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9max_count_5__del__(o);
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__last_selector(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_last_selector_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__last_selector(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_last_selector_3__set__(o, v);
  }
  else {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_14_last_selector_5__del__(o);
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__last_return_values(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_return_values_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__last_return_values(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_return_values_3__set__(o, v);
  }
  else {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_return_values_5__del__(o);
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__cached_octrees(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_cached_octrees_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__cached_octrees(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_cached_octrees_3__set__(o, v);
  }
  else {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_cached_octrees_5__del__(o);
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__last_octree_subset(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_octree_subset_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__last_octree_subset(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_octree_subset_3__set__(o, v);
  }
  else {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_last_octree_subset_5__del__(o);
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__last_oct_handler(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_last_oct_handler_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__last_oct_handler(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_last_oct_handler_3__set__(o, v);
  }
  else {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_last_oct_handler_5__del__(o);
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__prev_octree_subset(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_prev_octree_subset_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__prev_octree_subset(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_prev_octree_subset_3__set__(o, v);
  }
  else {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19_prev_octree_subset_5__del__(o);
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__prev_oct_handler(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_prev_oct_handler_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__prev_oct_handler(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_prev_oct_handler_3__set__(o, v);
  }
  else {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17_prev_oct_handler_5__del__(o);
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_bitmasks(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8bitmasks_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_bitmasks(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8bitmasks_3__set__(o, v);
  }
  else {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_8bitmasks_5__del__(o);
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_collisions(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10collisions_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_collisions(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10collisions_3__set__(o, v);
  }
  else {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_10collisions_5__del__(o);
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__used_mi2(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9_used_mi2_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__used_mi2(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9_used_mi2_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_2yt_8geometry_22particle_oct_container_ParticleBitmap[] = {
  {"_bitmask_logicaland", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_3_bitmask_logicaland, METH_VARARGS|METH_KEYWORDS, 0},
  {"_bitmask_intersects", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_5_bitmask_intersects, METH_VARARGS|METH_KEYWORDS, 0},
  {"update_mi2", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_7update_mi2, METH_VARARGS|METH_KEYWORDS, __pyx_doc_2yt_8geometry_22particle_oct_container_14ParticleBitmap_6update_mi2},
  {"_set_coarse_index_data_file", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_11_set_coarse_index_data_file, METH_O, 0},
  {"_set_refined_index_data_file", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_15_set_refined_index_data_file, METH_VARARGS|METH_KEYWORDS, 0},
  {"find_collisions", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_17find_collisions, METH_VARARGS|METH_KEYWORDS, 0},
  {"find_collisions_coarse", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_19find_collisions_coarse, METH_VARARGS|METH_KEYWORDS, 0},
  {"find_uncontaminated", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_21find_uncontaminated, METH_VARARGS|METH_KEYWORDS, 0},
  {"find_contaminated", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_23find_contaminated, METH_VARARGS|METH_KEYWORDS, 0},
  {"find_collisions_refined", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_25find_collisions_refined, METH_VARARGS|METH_KEYWORDS, 0},
  {"calcsize_bitmasks", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_27calcsize_bitmasks, METH_NOARGS, 0},
  {"get_bitmasks", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_29get_bitmasks, METH_NOARGS, 0},
  {"iseq_bitmask", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_31iseq_bitmask, METH_O, 0},
  {"save_bitmasks", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_33save_bitmasks, METH_O, 0},
  {"check_bitmasks", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_35check_bitmasks, METH_NOARGS, 0},
  {"reset_bitmasks", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_37reset_bitmasks, METH_NOARGS, 0},
  {"load_bitmasks", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_39load_bitmasks, METH_O, 0},
  {"print_info", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_41print_info, METH_NOARGS, 0},
  {"count_coarse", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_43count_coarse, METH_O, __pyx_doc_2yt_8geometry_22particle_oct_container_14ParticleBitmap_42count_coarse},
  {"count_refined", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_45count_refined, METH_O, __pyx_doc_2yt_8geometry_22particle_oct_container_14ParticleBitmap_44count_refined},
  {"count_total", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_47count_total, METH_O, __pyx_doc_2yt_8geometry_22particle_oct_container_14ParticleBitmap_46count_total},
  {"check", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_49check, METH_NOARGS, 0},
  {"primary_indices", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_51primary_indices, METH_NOARGS, 0},
  {"file_ownership_mask", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_53file_ownership_mask, METH_O, 0},
  {"finalize", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_55finalize, METH_NOARGS, 0},
  {"get_DLE", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_57get_DLE, METH_NOARGS, 0},
  {"get_DRE", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_59get_DRE, METH_NOARGS, 0},
  {"get_ghost_zones", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_61get_ghost_zones, METH_VARARGS|METH_KEYWORDS, 0},
  {"selector2mask", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_63selector2mask, METH_O, 0},
  {"mask2files", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_65mask2files, METH_O, 0},
  {"mask2filemasks", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_67mask2filemasks, METH_VARARGS|METH_KEYWORDS, 0},
  {"filemasks2addfiles", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_69filemasks2addfiles, METH_O, 0},
  {"identify_file_masks", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_71identify_file_masks, METH_O, 0},
  {"identify_data_files", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_73identify_data_files, METH_VARARGS|METH_KEYWORDS, 0},
  {"mask_to_files", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_75mask_to_files, METH_O, 0},
  {"masks_to_files", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_77masks_to_files, METH_VARARGS|METH_KEYWORDS, 0},
  {"construct_octree", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_79construct_octree, METH_VARARGS|METH_KEYWORDS, 0},
  {"__pyx_fuse_0_coarse_index_data_file", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_0__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_81_coarse_index_data_file, METH_VARARGS|METH_KEYWORDS, 0},
  {"__pyx_fuse_1_coarse_index_data_file", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_1__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_83_coarse_index_data_file, METH_VARARGS|METH_KEYWORDS, 0},
  {"__pyx_fuse_0_refined_index_data_file", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_0__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_87_refined_index_data_file, METH_VARARGS|METH_KEYWORDS, 0},
  {"__pyx_fuse_1_refined_index_data_file", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_fuse_1__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_89_refined_index_data_file, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_93__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_95__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_2yt_8geometry_22particle_oct_container_ParticleBitmap[] = {
  {(char *)"nfiles", __pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_nfiles, __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_nfiles, (char *)0, 0},
  {(char *)"index_order1", __pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_index_order1, __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_index_order1, (char *)0, 0},
  {(char *)"index_order2", __pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_index_order2, __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_index_order2, (char *)0, 0},
  {(char *)"masks", __pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_masks, __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_masks, (char *)0, 0},
  {(char *)"particle_counts", __pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_particle_counts, __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_particle_counts, (char *)0, 0},
  {(char *)"counts", __pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_counts, __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_counts, (char *)0, 0},
  {(char *)"max_count", __pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_max_count, __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_max_count, (char *)0, 0},
  {(char *)"_last_selector", __pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__last_selector, __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__last_selector, (char *)0, 0},
  {(char *)"_last_return_values", __pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__last_return_values, __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__last_return_values, (char *)0, 0},
  {(char *)"_cached_octrees", __pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__cached_octrees, __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__cached_octrees, (char *)0, 0},
  {(char *)"_last_octree_subset", __pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__last_octree_subset, __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__last_octree_subset, (char *)0, 0},
  {(char *)"_last_oct_handler", __pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__last_oct_handler, __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__last_oct_handler, (char *)0, 0},
  {(char *)"_prev_octree_subset", __pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__prev_octree_subset, __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__prev_octree_subset, (char *)0, 0},
  {(char *)"_prev_oct_handler", __pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__prev_oct_handler, __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__prev_oct_handler, (char *)0, 0},
  {(char *)"bitmasks", __pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_bitmasks, __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_bitmasks, (char *)0, 0},
  {(char *)"collisions", __pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_collisions, __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap_collisions, (char *)0, 0},
  {(char *)"_used_mi2", __pyx_getprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__used_mi2, __pyx_setprop_2yt_8geometry_22particle_oct_container_14ParticleBitmap__used_mi2, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmap = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.particle_oct_container.ParticleBitmap", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_22particle_oct_container_ParticleBitmap, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_22particle_oct_container_ParticleBitmap, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_22particle_oct_container_ParticleBitmap, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_22particle_oct_container_ParticleBitmap, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_2yt_8geometry_22particle_oct_container_ParticleBitmap, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_22particle_oct_container_14ParticleBitmap_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_22particle_oct_container_ParticleBitmap, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector;

static PyObject *__pyx_tp_new_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)o);
  p->__pyx_vtab = __pyx_vtabptr_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector;
  p->selector = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)Py_None); Py_INCREF(Py_None);
  p->bitmap = ((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)Py_None); Py_INCREF(Py_None);
  p->refined_ghosts_list = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet *)Py_None); Py_INCREF(Py_None);
  p->select_ewah = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)Py_None); Py_INCREF(Py_None);
  p->ghosts_ewah = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)Py_None); Py_INCREF(Py_None);
  p->ind1_n.data = NULL;
  p->ind1_n.memview = NULL;
  p->ind2_n.data = NULL;
  p->ind2_n.memview = NULL;
  p->neighbors.data = NULL;
  p->neighbors.memview = NULL;
  p->neighbor_list1.data = NULL;
  p->neighbor_list1.memview = NULL;
  p->neighbor_list2.data = NULL;
  p->neighbor_list2.memview = NULL;
  p->file_mask_p.data = NULL;
  p->file_mask_p.memview = NULL;
  p->file_mask_g.data = NULL;
  p->file_mask_g.memview = NULL;
  p->refined_select_bool.data = NULL;
  p->refined_select_bool.memview = NULL;
  p->refined_ghosts_bool.data = NULL;
  p->refined_ghosts_bool.memview = NULL;
  p->coarse_select_bool.data = NULL;
  p->coarse_select_bool.memview = NULL;
  p->coarse_ghosts_bool.data = NULL;
  p->coarse_ghosts_bool.memview = NULL;
  if (unlikely(__pyx_pw_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector(PyObject *o) {
  struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *p = (struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->selector);
  Py_CLEAR(p->bitmap);
  Py_CLEAR(p->refined_ghosts_list);
  Py_CLEAR(p->select_ewah);
  Py_CLEAR(p->ghosts_ewah);
  __PYX_XDEC_MEMVIEW(&p->ind1_n, 1);
  __PYX_XDEC_MEMVIEW(&p->ind2_n, 1);
  __PYX_XDEC_MEMVIEW(&p->neighbors, 1);
  __PYX_XDEC_MEMVIEW(&p->neighbor_list1, 1);
  __PYX_XDEC_MEMVIEW(&p->neighbor_list2, 1);
  __PYX_XDEC_MEMVIEW(&p->file_mask_p, 1);
  __PYX_XDEC_MEMVIEW(&p->file_mask_g, 1);
  __PYX_XDEC_MEMVIEW(&p->refined_select_bool, 1);
  __PYX_XDEC_MEMVIEW(&p->refined_ghosts_bool, 1);
  __PYX_XDEC_MEMVIEW(&p->coarse_select_bool, 1);
  __PYX_XDEC_MEMVIEW(&p->coarse_ghosts_bool, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *p = (struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)o;
  if (p->selector) {
    e = (*v)(((PyObject *)p->selector), a); if (e) return e;
  }
  if (p->bitmap) {
    e = (*v)(((PyObject *)p->bitmap), a); if (e) return e;
  }
  if (p->refined_ghosts_list) {
    e = (*v)(((PyObject *)p->refined_ghosts_list), a); if (e) return e;
  }
  if (p->select_ewah) {
    e = (*v)(((PyObject *)p->select_ewah), a); if (e) return e;
  }
  if (p->ghosts_ewah) {
    e = (*v)(((PyObject *)p->ghosts_ewah), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *p = (struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *)o;
  tmp = ((PyObject*)p->selector);
  p->selector = ((struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->bitmap);
  p->bitmap = ((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->refined_ghosts_list);
  p->refined_ghosts_list = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->select_ewah);
  p->select_ewah = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->ghosts_ewah);
  p->ghosts_ewah = ((struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector[] = {
  {"fill_masks", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_3fill_masks, METH_VARARGS|METH_KEYWORDS, 0},
  {"find_files", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_5find_files, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_7__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_9__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.particle_oct_container.ParticleBitmapSelector", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer;

static PyObject *__pyx_tp_new_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *p;
  PyObject *o = __pyx_ptype_2yt_8geometry_13oct_container_SparseOctreeContainer->tp_new(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctreeContainer*)__pyx_vtabptr_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer;
  p->_index_base_roots.data = NULL;
  p->_index_base_roots.memview = NULL;
  p->_index_base_octs.data = NULL;
  p->_index_base_octs.memview = NULL;
  p->_octs_per_root.data = NULL;
  p->_octs_per_root.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer(PyObject *o) {
  struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *p = (struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_9__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  __PYX_XDEC_MEMVIEW(&p->_index_base_roots, 1);
  __PYX_XDEC_MEMVIEW(&p->_index_base_octs, 1);
  __PYX_XDEC_MEMVIEW(&p->_octs_per_root, 1);
  PyObject_GC_Track(o);
  if (likely(__pyx_ptype_2yt_8geometry_13oct_container_SparseOctreeContainer)) __pyx_ptype_2yt_8geometry_13oct_container_SparseOctreeContainer->tp_dealloc(o); else __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer);
}

static int __pyx_tp_traverse_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer(PyObject *o, visitproc v, void *a) {
  int e;
  e = ((likely(__pyx_ptype_2yt_8geometry_13oct_container_SparseOctreeContainer)) ? ((__pyx_ptype_2yt_8geometry_13oct_container_SparseOctreeContainer->tp_traverse) ? __pyx_ptype_2yt_8geometry_13oct_container_SparseOctreeContainer->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer)); if (e) return e;
  return 0;
}

static int __pyx_tp_clear_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer(PyObject *o) {
  if (likely(__pyx_ptype_2yt_8geometry_13oct_container_SparseOctreeContainer)) { if (__pyx_ptype_2yt_8geometry_13oct_container_SparseOctreeContainer->tp_clear) __pyx_ptype_2yt_8geometry_13oct_container_SparseOctreeContainer->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer);
  return 0;
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_max_level(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_9max_level_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_max_level(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_9max_level_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_n_ref(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_5n_ref_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_n_ref(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_5n_ref_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer__index_base_roots(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_17_index_base_roots_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer__index_base_roots(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_17_index_base_roots_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer__index_base_octs(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_16_index_base_octs_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer__index_base_octs(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_16_index_base_octs_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_overlap_cells(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_13overlap_cells_1__get__(o);
}

static int __pyx_setprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_overlap_cells(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_13overlap_cells_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer[] = {
  {"allocate_domains", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_3allocate_domains, METH_VARARGS|METH_KEYWORDS, 0},
  {"finalize", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_5finalize, METH_NOARGS, 0},
  {"get_index_base_octs", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_7get_index_base_octs, METH_O, 0},
  {"add", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_11add, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_13__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_15__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer[] = {
  {(char *)"max_level", __pyx_getprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_max_level, __pyx_setprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_max_level, (char *)0, 0},
  {(char *)"n_ref", __pyx_getprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_n_ref, __pyx_setprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_n_ref, (char *)0, 0},
  {(char *)"_index_base_roots", __pyx_getprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer__index_base_roots, __pyx_setprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer__index_base_roots, (char *)0, 0},
  {(char *)"_index_base_octs", __pyx_getprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer__index_base_octs, __pyx_setprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer__index_base_octs, (char *)0, 0},
  {(char *)"overlap_cells", __pyx_getprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_overlap_cells, __pyx_setprop_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_overlap_cells, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.particle_oct_container.ParticleBitmapOctreeContainer", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer, /*tp_traverse*/
  __pyx_tp_clear_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__ *__pyx_freelist_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__[8];
static int __pyx_freecount_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__ = 0;

static PyObject *__pyx_tp_new_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__)))) {
    o = (PyObject*)__pyx_freelist_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__[--__pyx_freecount_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__];
    memset(o, 0, sizeof(struct __pyx_obj_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__(PyObject *o) {
  struct __pyx_obj_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__ *p = (struct __pyx_obj_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__)))) {
    __pyx_freelist_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__[__pyx_freecount_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__++] = ((struct __pyx_obj_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__ *p = (struct __pyx_obj_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.particle_oct_container.__pyx_scope_struct____iter__", /*tp_name*/
  sizeof(struct __pyx_obj_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_array __pyx_vtable_array;

static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_array_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_array_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_array;
  p->mode = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_format = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_array___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_array(PyObject *o) {
  struct __pyx_array_obj *p = (struct __pyx_array_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_array___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->mode);
  Py_CLEAR(p->_format);
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_array(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_array(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_array___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_tp_getattro_array(PyObject *o, PyObject *n) {
  PyObject *v = __Pyx_PyObject_GenericGetAttr(o, n);
  if (!v && PyErr_ExceptionMatches(PyExc_AttributeError)) {
    PyErr_Clear();
    v = __pyx_array___getattr__(o, n);
  }
  return v;
}

static PyObject *__pyx_getprop___pyx_array_memview(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(o);
}

static PyMethodDef __pyx_methods_array[] = {
  {"__getattr__", (PyCFunction)__pyx_array___getattr__, METH_O|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_array_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_array_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_array[] = {
  {(char *)"memview", __pyx_getprop___pyx_array_memview, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_array = {
  __pyx_array___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_array, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_array = {
  __pyx_array___len__, /*mp_length*/
  __pyx_array___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_array, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_array = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_array_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_array = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.particle_oct_container.array", /*tp_name*/
  sizeof(struct __pyx_array_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_array, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_array, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_array, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  __pyx_tp_getattro_array, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_array, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_array, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_array, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_array, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_MemviewEnum_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_MemviewEnum_obj *)o);
  p->name = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_Enum(PyObject *o) {
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_Enum(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  if (p->name) {
    e = (*v)(p->name, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_Enum(PyObject *o) {
  PyObject* tmp;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  tmp = ((PyObject*)p->name);
  p->name = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_Enum[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type___pyx_MemviewEnum = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.particle_oct_container.Enum", /*tp_name*/
  sizeof(struct __pyx_MemviewEnum_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_Enum, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_MemviewEnum___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_Enum, /*tp_traverse*/
  __pyx_tp_clear_Enum, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_Enum, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_MemviewEnum___init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_Enum, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_memoryview __pyx_vtable_memoryview;

static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryview_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryview_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_memoryview;
  p->obj = Py_None; Py_INCREF(Py_None);
  p->_size = Py_None; Py_INCREF(Py_None);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  p->view.obj = NULL;
  if (unlikely(__pyx_memoryview___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_memoryview(PyObject *o) {
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_memoryview___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->obj);
  Py_CLEAR(p->_size);
  Py_CLEAR(p->_array_interface);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_memoryview(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  if (p->obj) {
    e = (*v)(p->obj, a); if (e) return e;
  }
  if (p->_size) {
    e = (*v)(p->_size, a); if (e) return e;
  }
  if (p->_array_interface) {
    e = (*v)(p->_array_interface, a); if (e) return e;
  }
  if (p->view.obj) {
    e = (*v)(p->view.obj, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_memoryview(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  tmp = ((PyObject*)p->obj);
  p->obj = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_size);
  p->_size = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_array_interface);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  Py_CLEAR(p->view.obj);
  return 0;
}
static PyObject *__pyx_sq_item_memoryview(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_memoryview(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_memoryview___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_getprop___pyx_memoryview_T(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_shape(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_strides(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_suboffsets(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_ndim(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_itemsize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_nbytes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(o);
}

static PyMethodDef __pyx_methods_memoryview[] = {
  {"is_c_contig", (PyCFunction)__pyx_memoryview_is_c_contig, METH_NOARGS, 0},
  {"is_f_contig", (PyCFunction)__pyx_memoryview_is_f_contig, METH_NOARGS, 0},
  {"copy", (PyCFunction)__pyx_memoryview_copy, METH_NOARGS, 0},
  {"copy_fortran", (PyCFunction)__pyx_memoryview_copy_fortran, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_memoryview[] = {
  {(char *)"T", __pyx_getprop___pyx_memoryview_T, 0, (char *)0, 0},
  {(char *)"base", __pyx_getprop___pyx_memoryview_base, 0, (char *)0, 0},
  {(char *)"shape", __pyx_getprop___pyx_memoryview_shape, 0, (char *)0, 0},
  {(char *)"strides", __pyx_getprop___pyx_memoryview_strides, 0, (char *)0, 0},
  {(char *)"suboffsets", __pyx_getprop___pyx_memoryview_suboffsets, 0, (char *)0, 0},
  {(char *)"ndim", __pyx_getprop___pyx_memoryview_ndim, 0, (char *)0, 0},
  {(char *)"itemsize", __pyx_getprop___pyx_memoryview_itemsize, 0, (char *)0, 0},
  {(char *)"nbytes", __pyx_getprop___pyx_memoryview_nbytes, 0, (char *)0, 0},
  {(char *)"size", __pyx_getprop___pyx_memoryview_size, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_memoryview = {
  __pyx_memoryview___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_memoryview, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_memoryview = {
  __pyx_memoryview___len__, /*mp_length*/
  __pyx_memoryview___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_memoryview, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_memoryview = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_memoryview_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_memoryview = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.particle_oct_container.memoryview", /*tp_name*/
  sizeof(struct __pyx_memoryview_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_memoryview, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_memoryview___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_memoryview, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_memoryview, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_memoryview___str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_memoryview, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_memoryview, /*tp_traverse*/
  __pyx_tp_clear_memoryview, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_memoryview, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_memoryview, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_memoryview, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct__memoryviewslice __pyx_vtable__memoryviewslice;

static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryviewslice_obj *p;
  PyObject *o = __pyx_tp_new_memoryview(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryviewslice_obj *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_memoryview*)__pyx_vtabptr__memoryviewslice;
  p->from_object = Py_None; Py_INCREF(Py_None);
  p->from_slice.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc__memoryviewslice(PyObject *o) {
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_memoryviewslice___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->from_object);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_memoryview(o);
}

static int __pyx_tp_traverse__memoryviewslice(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  e = __pyx_tp_traverse_memoryview(o, v, a); if (e) return e;
  if (p->from_object) {
    e = (*v)(p->from_object, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear__memoryviewslice(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  __pyx_tp_clear_memoryview(o);
  tmp = ((PyObject*)p->from_object);
  p->from_object = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  __PYX_XDEC_MEMVIEW(&p->from_slice, 1);
  return 0;
}

static PyObject *__pyx_getprop___pyx_memoryviewslice_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(o);
}

static PyMethodDef __pyx_methods__memoryviewslice[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets__memoryviewslice[] = {
  {(char *)"base", __pyx_getprop___pyx_memoryviewslice_base, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type___pyx_memoryviewslice = {
  PyVarObject_HEAD_INIT(0, 0)
  "yt.geometry.particle_oct_container._memoryviewslice", /*tp_name*/
  sizeof(struct __pyx_memoryviewslice_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc__memoryviewslice, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Internal class for passing memoryview slices to Python", /*tp_doc*/
  __pyx_tp_traverse__memoryviewslice, /*tp_traverse*/
  __pyx_tp_clear__memoryviewslice, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods__memoryviewslice, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets__memoryviewslice, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new__memoryviewslice, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_particle_oct_container(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_particle_oct_container},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "particle_oct_container",
    __pyx_k_Oct_container_tuned_for_Particl, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_ASCII, __pyx_k_ASCII, sizeof(__pyx_k_ASCII), 0, 0, 1, 1},
  {&__pyx_kp_s_Buffer_view_does_not_expose_stri, __pyx_k_Buffer_view_does_not_expose_stri, sizeof(__pyx_k_Buffer_view_does_not_expose_stri), 0, 0, 1, 0},
  {&__pyx_kp_s_Can_only_create_a_buffer_that_is, __pyx_k_Can_only_create_a_buffer_that_is, sizeof(__pyx_k_Can_only_create_a_buffer_that_is), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_assign_to_read_only_memor, __pyx_k_Cannot_assign_to_read_only_memor, sizeof(__pyx_k_Cannot_assign_to_read_only_memor), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_create_writable_memory_vi, __pyx_k_Cannot_create_writable_memory_vi, sizeof(__pyx_k_Cannot_create_writable_memory_vi), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_index_with_type_s, __pyx_k_Cannot_index_with_type_s, sizeof(__pyx_k_Cannot_index_with_type_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Child_was_already_initialized, __pyx_k_Child_was_already_initialized, sizeof(__pyx_k_Child_was_already_initialized), 0, 0, 1, 0},
  {&__pyx_n_s_Ellipsis, __pyx_k_Ellipsis, sizeof(__pyx_k_Ellipsis), 0, 0, 1, 1},
  {&__pyx_kp_s_Empty_shape_tuple_for_cython_arr, __pyx_k_Empty_shape_tuple_for_cython_arr, sizeof(__pyx_k_Empty_shape_tuple_for_cython_arr), 0, 0, 1, 0},
  {&__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_k_Expected_at_least_d_argument_s_g, sizeof(__pyx_k_Expected_at_least_d_argument_s_g), 0, 0, 1, 0},
  {&__pyx_kp_s_File_03d, __pyx_k_File_03d, sizeof(__pyx_k_File_03d), 0, 0, 1, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor, __pyx_k_Format_string_allocated_too_shor, sizeof(__pyx_k_Format_string_allocated_too_shor), 0, 1, 0, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor_2, __pyx_k_Format_string_allocated_too_shor_2, sizeof(__pyx_k_Format_string_allocated_too_shor_2), 0, 1, 0, 0},
  {&__pyx_kp_s_Function_call_with_ambiguous_arg, __pyx_k_Function_call_with_ambiguous_arg, sizeof(__pyx_k_Function_call_with_ambiguous_arg), 0, 0, 1, 0},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_kp_u_Index_for_a_softening_region, __pyx_k_Index_for_a_softening_region, sizeof(__pyx_k_Index_for_a_softening_region), 0, 1, 0, 0},
  {&__pyx_kp_s_Indirect_dimensions_not_supporte, __pyx_k_Indirect_dimensions_not_supporte, sizeof(__pyx_k_Indirect_dimensions_not_supporte), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_k_Invalid_mode_expected_c_or_fortr, sizeof(__pyx_k_Invalid_mode_expected_c_or_fortr), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_k_Invalid_shape_in_axis_d_d, sizeof(__pyx_k_Invalid_shape_in_axis_d_d), 0, 0, 1, 0},
  {&__pyx_kp_s_Loading_particle_index, __pyx_k_Loading_particle_index, sizeof(__pyx_k_Loading_particle_index), 0, 0, 1, 0},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_k_MemoryView_of_r_at_0x_x, sizeof(__pyx_k_MemoryView_of_r_at_0x_x), 0, 0, 1, 0},
  {&__pyx_kp_s_MemoryView_of_r_object, __pyx_k_MemoryView_of_r_object, sizeof(__pyx_k_MemoryView_of_r_object), 0, 0, 1, 0},
  {&__pyx_kp_s_No_matching_signature_found, __pyx_k_No_matching_signature_found, sizeof(__pyx_k_No_matching_signature_found), 0, 0, 1, 0},
  {&__pyx_kp_s_No_root_found_for, __pyx_k_No_root_found_for, sizeof(__pyx_k_No_root_found_for), 0, 0, 1, 0},
  {&__pyx_kp_u_Non_native_byte_order_not_suppor, __pyx_k_Non_native_byte_order_not_suppor, sizeof(__pyx_k_Non_native_byte_order_not_suppor), 0, 1, 0, 0},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_kp_s_Number_of_bitmasks_conflicts_wit, __pyx_k_Number_of_bitmasks_conflicts_wit, sizeof(__pyx_k_Number_of_bitmasks_conflicts_wit), 0, 0, 1, 0},
  {&__pyx_n_b_O, __pyx_k_O, sizeof(__pyx_k_O), 0, 0, 0, 1},
  {&__pyx_n_s_OSError, __pyx_k_OSError, sizeof(__pyx_k_OSError), 0, 0, 1, 1},
  {&__pyx_kp_s_Octree_does_not_have_roots_along, __pyx_k_Octree_does_not_have_roots_along, sizeof(__pyx_k_Octree_does_not_have_roots_along), 0, 0, 1, 0},
  {&__pyx_kp_s_Octree_must_have_the_same_number, __pyx_k_Octree_must_have_the_same_number, sizeof(__pyx_k_Octree_must_have_the_same_number), 0, 0, 1, 0},
  {&__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_k_Out_of_bounds_on_buffer_access_a, sizeof(__pyx_k_Out_of_bounds_on_buffer_access_a), 0, 0, 1, 0},
  {&__pyx_n_s_ParticleBitmap, __pyx_k_ParticleBitmap, sizeof(__pyx_k_ParticleBitmap), 0, 0, 1, 1},
  {&__pyx_n_s_ParticleBitmapOctreeContainer, __pyx_k_ParticleBitmapOctreeContainer, sizeof(__pyx_k_ParticleBitmapOctreeContainer), 0, 0, 1, 1},
  {&__pyx_n_s_ParticleBitmapSelector, __pyx_k_ParticleBitmapSelector, sizeof(__pyx_k_ParticleBitmapSelector), 0, 0, 1, 1},
  {&__pyx_n_s_ParticleBitmap__coarse_index_dat, __pyx_k_ParticleBitmap__coarse_index_dat, sizeof(__pyx_k_ParticleBitmap__coarse_index_dat), 0, 0, 1, 1},
  {&__pyx_n_s_ParticleBitmap__refined_index_da, __pyx_k_ParticleBitmap__refined_index_da, sizeof(__pyx_k_ParticleBitmap__refined_index_da), 0, 0, 1, 1},
  {&__pyx_n_s_ParticleOctreeContainer, __pyx_k_ParticleOctreeContainer, sizeof(__pyx_k_ParticleOctreeContainer), 0, 0, 1, 1},
  {&__pyx_n_s_ParticleOctreeContainer___iter, __pyx_k_ParticleOctreeContainer___iter, sizeof(__pyx_k_ParticleOctreeContainer___iter), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_Q, __pyx_k_Q, sizeof(__pyx_k_Q), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_kp_u_Smoothing_length_for_particle, __pyx_k_Smoothing_length_for_particle, sizeof(__pyx_k_Smoothing_length_for_particle), 0, 1, 0, 0},
  {&__pyx_kp_b_T, __pyx_k_T, sizeof(__pyx_k_T), 0, 0, 0, 0},
  {&__pyx_kp_s_The_file_format_of_the_index_has, __pyx_k_The_file_format_of_the_index_has, sizeof(__pyx_k_The_file_format_of_the_index_has), 0, 0, 1, 0},
  {&__pyx_kp_s_Total_of_s_s_collisions_3_5f, __pyx_k_Total_of_s_s_collisions_3_5f, sizeof(__pyx_k_Total_of_s_s_collisions_3_5f), 0, 0, 1, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Unable_to_convert_item_to_object, __pyx_k_Unable_to_convert_item_to_object, sizeof(__pyx_k_Unable_to_convert_item_to_object), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_View_MemoryView, __pyx_k_View_MemoryView, sizeof(__pyx_k_View_MemoryView), 0, 0, 1, 1},
  {&__pyx_kp_u__12, __pyx_k__12, sizeof(__pyx_k__12), 0, 1, 0, 0},
  {&__pyx_kp_b__53, __pyx_k__53, sizeof(__pyx_k__53), 0, 0, 0, 0},
  {&__pyx_kp_b__54, __pyx_k__54, sizeof(__pyx_k__54), 0, 0, 0, 0},
  {&__pyx_kp_b__55, __pyx_k__55, sizeof(__pyx_k__55), 0, 0, 0, 0},
  {&__pyx_kp_b__56, __pyx_k__56, sizeof(__pyx_k__56), 0, 0, 0, 0},
  {&__pyx_kp_u__57, __pyx_k__57, sizeof(__pyx_k__57), 0, 1, 0, 0},
  {&__pyx_kp_s__7, __pyx_k__7, sizeof(__pyx_k__7), 0, 0, 1, 0},
  {&__pyx_kp_s__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 0, 1, 0},
  {&__pyx_n_s_add, __pyx_k_add, sizeof(__pyx_k_add), 0, 0, 1, 1},
  {&__pyx_n_s_allocate_buffer, __pyx_k_allocate_buffer, sizeof(__pyx_k_allocate_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_allocate_domains, __pyx_k_allocate_domains, sizeof(__pyx_k_allocate_domains), 0, 0, 1, 1},
  {&__pyx_n_s_allocate_root, __pyx_k_allocate_root, sizeof(__pyx_k_allocate_root), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_argsort, __pyx_k_argsort, sizeof(__pyx_k_argsort), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_astype, __pyx_k_astype, sizeof(__pyx_k_astype), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_s_base_mask, __pyx_k_base_mask, sizeof(__pyx_k_base_mask), 0, 0, 1, 1},
  {&__pyx_n_s_bcoll, __pyx_k_bcoll, sizeof(__pyx_k_bcoll), 0, 0, 1, 1},
  {&__pyx_n_s_bitmap, __pyx_k_bitmap, sizeof(__pyx_k_bitmap), 0, 0, 1, 1},
  {&__pyx_n_s_bitmask_version, __pyx_k_bitmask_version, sizeof(__pyx_k_bitmask_version), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_u_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 1, 0, 1},
  {&__pyx_n_s_calcsize, __pyx_k_calcsize, sizeof(__pyx_k_calcsize), 0, 0, 1, 1},
  {&__pyx_n_s_characteristic_size, __pyx_k_characteristic_size, sizeof(__pyx_k_characteristic_size), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_cmask, __pyx_k_cmask, sizeof(__pyx_k_cmask), 0, 0, 1, 1},
  {&__pyx_n_s_coarse_ghosts, __pyx_k_coarse_ghosts, sizeof(__pyx_k_coarse_ghosts), 0, 0, 1, 1},
  {&__pyx_n_s_coarse_ghosts_list, __pyx_k_coarse_ghosts_list, sizeof(__pyx_k_coarse_ghosts_list), 0, 0, 1, 1},
  {&__pyx_n_s_coarse_index_data_file, __pyx_k_coarse_index_data_file, sizeof(__pyx_k_coarse_index_data_file), 0, 0, 1, 1},
  {&__pyx_n_s_coarse_select_list, __pyx_k_coarse_select_list, sizeof(__pyx_k_coarse_select_list), 0, 0, 1, 1},
  {&__pyx_n_s_code_length, __pyx_k_code_length, sizeof(__pyx_k_code_length), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_kp_s_contiguous_and_direct, __pyx_k_contiguous_and_direct, sizeof(__pyx_k_contiguous_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_contiguous_and_indirect, __pyx_k_contiguous_and_indirect, sizeof(__pyx_k_contiguous_and_indirect), 0, 0, 1, 0},
  {&__pyx_n_s_convert_to_units, __pyx_k_convert_to_units, sizeof(__pyx_k_convert_to_units), 0, 0, 1, 1},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_count_coarse, __pyx_k_count_coarse, sizeof(__pyx_k_count_coarse), 0, 0, 1, 1},
  {&__pyx_n_s_count_refined, __pyx_k_count_refined, sizeof(__pyx_k_count_refined), 0, 0, 1, 1},
  {&__pyx_n_s_count_threshold, __pyx_k_count_threshold, sizeof(__pyx_k_count_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_count_total, __pyx_k_count_total, sizeof(__pyx_k_count_total), 0, 0, 1, 1},
  {&__pyx_n_s_counts, __pyx_k_counts, sizeof(__pyx_k_counts), 0, 0, 1, 1},
  {&__pyx_n_s_cubic, __pyx_k_cubic, sizeof(__pyx_k_cubic), 0, 0, 1, 1},
  {&__pyx_n_s_data_files, __pyx_k_data_files, sizeof(__pyx_k_data_files), 0, 0, 1, 1},
  {&__pyx_n_s_dataset, __pyx_k_dataset, sizeof(__pyx_k_dataset), 0, 0, 1, 1},
  {&__pyx_n_s_defaultdict, __pyx_k_defaultdict, sizeof(__pyx_k_defaultdict), 0, 0, 1, 1},
  {&__pyx_n_s_defaults, __pyx_k_defaults, sizeof(__pyx_k_defaults), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_dmask, __pyx_k_dmask, sizeof(__pyx_k_dmask), 0, 0, 1, 1},
  {&__pyx_n_s_domain_dimensions, __pyx_k_domain_dimensions, sizeof(__pyx_k_domain_dimensions), 0, 0, 1, 1},
  {&__pyx_n_s_domain_id, __pyx_k_domain_id, sizeof(__pyx_k_domain_id), 0, 0, 1, 1},
  {&__pyx_n_s_domain_ind, __pyx_k_domain_ind, sizeof(__pyx_k_domain_ind), 0, 0, 1, 1},
  {&__pyx_n_s_domain_left_edge, __pyx_k_domain_left_edge, sizeof(__pyx_k_domain_left_edge), 0, 0, 1, 1},
  {&__pyx_n_s_domain_right_edge, __pyx_k_domain_right_edge, sizeof(__pyx_k_domain_right_edge), 0, 0, 1, 1},
  {&__pyx_n_s_double, __pyx_k_double, sizeof(__pyx_k_double), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_dtype_is_object, __pyx_k_dtype_is_object, sizeof(__pyx_k_dtype_is_object), 0, 0, 1, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_kp_u_exceeds_max, __pyx_k_exceeds_max, sizeof(__pyx_k_exceeds_max), 0, 1, 0, 0},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_file_hash, __pyx_k_file_hash, sizeof(__pyx_k_file_hash), 0, 0, 1, 1},
  {&__pyx_n_s_file_id, __pyx_k_file_id, sizeof(__pyx_k_file_id), 0, 0, 1, 1},
  {&__pyx_n_s_file_idx, __pyx_k_file_idx, sizeof(__pyx_k_file_idx), 0, 0, 1, 1},
  {&__pyx_n_s_file_list, __pyx_k_file_list, sizeof(__pyx_k_file_list), 0, 0, 1, 1},
  {&__pyx_n_s_file_mask_g, __pyx_k_file_mask_g, sizeof(__pyx_k_file_mask_g), 0, 0, 1, 1},
  {&__pyx_n_s_file_mask_p, __pyx_k_file_mask_p, sizeof(__pyx_k_file_mask_p), 0, 0, 1, 1},
  {&__pyx_n_s_fill_bool, __pyx_k_fill_bool, sizeof(__pyx_k_fill_bool), 0, 0, 1, 1},
  {&__pyx_n_s_fill_masks, __pyx_k_fill_masks, sizeof(__pyx_k_fill_masks), 0, 0, 1, 1},
  {&__pyx_n_s_finalize, __pyx_k_finalize, sizeof(__pyx_k_finalize), 0, 0, 1, 1},
  {&__pyx_n_s_finish, __pyx_k_finish, sizeof(__pyx_k_finish), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_float, __pyx_k_float, sizeof(__pyx_k_float), 0, 0, 1, 1},
  {&__pyx_n_s_float32, __pyx_k_float32, sizeof(__pyx_k_float32), 0, 0, 1, 1},
  {&__pyx_n_s_float64, __pyx_k_float64, sizeof(__pyx_k_float64), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 0, 1, 1},
  {&__pyx_n_u_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 1, 0, 1},
  {&__pyx_n_s_get_DLE, __pyx_k_get_DLE, sizeof(__pyx_k_get_DLE), 0, 0, 1, 1},
  {&__pyx_n_s_get_DRE, __pyx_k_get_DRE, sizeof(__pyx_k_get_DRE), 0, 0, 1, 1},
  {&__pyx_n_s_get_bitmasks, __pyx_k_get_bitmasks, sizeof(__pyx_k_get_bitmasks), 0, 0, 1, 1},
  {&__pyx_n_s_get_pbar, __pyx_k_get_pbar, sizeof(__pyx_k_get_pbar), 0, 0, 1, 1},
  {&__pyx_n_s_get_periodicity, __pyx_k_get_periodicity, sizeof(__pyx_k_get_periodicity), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_k_got_differing_extents_in_dimensi, sizeof(__pyx_k_got_differing_extents_in_dimensi), 0, 0, 1, 0},
  {&__pyx_n_s_hsml, __pyx_k_hsml, sizeof(__pyx_k_hsml), 0, 0, 1, 1},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_ifile, __pyx_k_ifile, sizeof(__pyx_k_ifile), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_in_coll, __pyx_k_in_coll, sizeof(__pyx_k_in_coll), 0, 0, 1, 1},
  {&__pyx_n_s_in_collection, __pyx_k_in_collection, sizeof(__pyx_k_in_collection), 0, 0, 1, 1},
  {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
  {&__pyx_n_s_index_order1, __pyx_k_index_order1, sizeof(__pyx_k_index_order1), 0, 0, 1, 1},
  {&__pyx_n_s_index_order2, __pyx_k_index_order2, sizeof(__pyx_k_index_order2), 0, 0, 1, 1},
  {&__pyx_n_s_indices, __pyx_k_indices, sizeof(__pyx_k_indices), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_int32, __pyx_k_int32, sizeof(__pyx_k_int32), 0, 0, 1, 1},
  {&__pyx_n_s_int64, __pyx_k_int64, sizeof(__pyx_k_int64), 0, 0, 1, 1},
  {&__pyx_n_s_io_handler, __pyx_k_io_handler, sizeof(__pyx_k_io_handler), 0, 0, 1, 1},
  {&__pyx_kp_u_is_negative_with_value, __pyx_k_is_negative_with_value, sizeof(__pyx_k_is_negative_with_value), 0, 1, 0, 0},
  {&__pyx_n_s_isfile, __pyx_k_isfile, sizeof(__pyx_k_isfile), 0, 0, 1, 1},
  {&__pyx_n_s_itemsize, __pyx_k_itemsize, sizeof(__pyx_k_itemsize), 0, 0, 1, 1},
  {&__pyx_kp_s_itemsize_0_for_cython_array, __pyx_k_itemsize_0_for_cython_array, sizeof(__pyx_k_itemsize_0_for_cython_array), 0, 0, 1, 0},
  {&__pyx_n_s_iter, __pyx_k_iter, sizeof(__pyx_k_iter), 0, 0, 1, 1},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_kind, __pyx_k_kind, sizeof(__pyx_k_kind), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_left_edge, __pyx_k_left_edge, sizeof(__pyx_k_left_edge), 0, 0, 1, 1},
  {&__pyx_n_s_log2, __pyx_k_log2, sizeof(__pyx_k_log2), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mask, __pyx_k_mask, sizeof(__pyx_k_mask), 0, 0, 1, 1},
  {&__pyx_n_s_mask2, __pyx_k_mask2, sizeof(__pyx_k_mask2), 0, 0, 1, 1},
  {&__pyx_n_s_mask_threshold, __pyx_k_mask_threshold, sizeof(__pyx_k_mask_threshold), 0, 0, 1, 1},
  {&__pyx_n_s_mask_to_files, __pyx_k_mask_to_files, sizeof(__pyx_k_mask_to_files), 0, 0, 1, 1},
  {&__pyx_n_s_masklevel, __pyx_k_masklevel, sizeof(__pyx_k_masklevel), 0, 0, 1, 1},
  {&__pyx_n_s_masks_to_files, __pyx_k_masks_to_files, sizeof(__pyx_k_masks_to_files), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_max_index_order2, __pyx_k_max_index_order2, sizeof(__pyx_k_max_index_order2), 0, 0, 1, 1},
  {&__pyx_n_s_memview, __pyx_k_memview, sizeof(__pyx_k_memview), 0, 0, 1, 1},
  {&__pyx_n_s_mm_g, __pyx_k_mm_g, sizeof(__pyx_k_mm_g), 0, 0, 1, 1},
  {&__pyx_n_s_mm_s, __pyx_k_mm_s, sizeof(__pyx_k_mm_s), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_n_s_n_ref, __pyx_k_n_ref, sizeof(__pyx_k_n_ref), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_kp_u_ndarray_is_not_C_contiguous, __pyx_k_ndarray_is_not_C_contiguous, sizeof(__pyx_k_ndarray_is_not_C_contiguous), 0, 1, 0, 0},
  {&__pyx_kp_u_ndarray_is_not_Fortran_contiguou, __pyx_k_ndarray_is_not_Fortran_contiguou, sizeof(__pyx_k_ndarray_is_not_Fortran_contiguou), 0, 1, 0, 0},
  {&__pyx_n_s_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_nfiles, __pyx_k_nfiles, sizeof(__pyx_k_nfiles), 0, 0, 1, 1},
  {&__pyx_n_s_ngz, __pyx_k_ngz, sizeof(__pyx_k_ngz), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_none, __pyx_k_none, sizeof(__pyx_k_none), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_nsub_mi, __pyx_k_nsub_mi, sizeof(__pyx_k_nsub_mi), 0, 0, 1, 1},
  {&__pyx_n_s_num_root, __pyx_k_num_root, sizeof(__pyx_k_num_root), 0, 0, 1, 1},
  {&__pyx_n_s_num_zones, __pyx_k_num_zones, sizeof(__pyx_k_num_zones), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_o, __pyx_k_o, sizeof(__pyx_k_o), 0, 0, 1, 1},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
  {&__pyx_n_s_open, __pyx_k_open, sizeof(__pyx_k_open), 0, 0, 1, 1},
  {&__pyx_n_s_order, __pyx_k_order, sizeof(__pyx_k_order), 0, 0, 1, 1},
  {&__pyx_n_s_order1, __pyx_k_order1, sizeof(__pyx_k_order1), 0, 0, 1, 1},
  {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
  {&__pyx_n_s_out, __pyx_k_out, sizeof(__pyx_k_out), 0, 0, 1, 1},
  {&__pyx_n_s_out_collection, __pyx_k_out_collection, sizeof(__pyx_k_out_collection), 0, 0, 1, 1},
  {&__pyx_n_s_pack, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
  {&__pyx_n_s_path, __pyx_k_path, sizeof(__pyx_k_path), 0, 0, 1, 1},
  {&__pyx_n_s_periodicity, __pyx_k_periodicity, sizeof(__pyx_k_periodicity), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pos, __pyx_k_pos, sizeof(__pyx_k_pos), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_print_info, __pyx_k_print_info, sizeof(__pyx_k_print_info), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_getbuffer, __pyx_k_pyx_getbuffer, sizeof(__pyx_k_pyx_getbuffer), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Enum, __pyx_k_pyx_unpickle_Enum, sizeof(__pyx_k_pyx_unpickle_Enum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_q, __pyx_k_q, sizeof(__pyx_k_q), 0, 0, 1, 1},
  {&__pyx_n_s_quartic, __pyx_k_quartic, sizeof(__pyx_k_quartic), 0, 0, 1, 1},
  {&__pyx_n_s_quintic, __pyx_k_quintic, sizeof(__pyx_k_quintic), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_rb, __pyx_k_rb, sizeof(__pyx_k_rb), 0, 0, 1, 1},
  {&__pyx_n_s_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_refined_index_data_file, __pyx_k_refined_index_data_file, sizeof(__pyx_k_refined_index_data_file), 0, 0, 1, 1},
  {&__pyx_n_s_right_edge, __pyx_k_right_edge, sizeof(__pyx_k_right_edge), 0, 0, 1, 1},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_kp_u_s_2, __pyx_k_s_2, sizeof(__pyx_k_s_2), 0, 1, 0, 0},
  {&__pyx_n_s_save_bitmasks, __pyx_k_save_bitmasks, sizeof(__pyx_k_save_bitmasks), 0, 0, 1, 1},
  {&__pyx_n_s_selector, __pyx_k_selector, sizeof(__pyx_k_selector), 0, 0, 1, 1},
  {&__pyx_n_s_selector_mask, __pyx_k_selector_mask, sizeof(__pyx_k_selector_mask), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_kp_s_self__ptr_octs_per_root_self_oct, __pyx_k_self__ptr_octs_per_root_self_oct, sizeof(__pyx_k_self__ptr_octs_per_root_self_oct), 0, 0, 1, 0},
  {&__pyx_kp_s_self_dom_offsets_self_oct_list_s, __pyx_k_self_dom_offsets_self_oct_list_s, sizeof(__pyx_k_self_dom_offsets_self_oct_list_s), 0, 0, 1, 0},
  {&__pyx_kp_s_self_file_markers_cannot_be_conv, __pyx_k_self_file_markers_cannot_be_conv, sizeof(__pyx_k_self_file_markers_cannot_be_conv), 0, 0, 1, 0},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_sort, __pyx_k_sort, sizeof(__pyx_k_sort), 0, 0, 1, 1},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_kp_s_strided_and_direct, __pyx_k_strided_and_direct, sizeof(__pyx_k_strided_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_direct_or_indirect, __pyx_k_strided_and_direct_or_indirect, sizeof(__pyx_k_strided_and_direct_or_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_indirect, __pyx_k_strided_and_indirect, sizeof(__pyx_k_strided_and_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_strip, __pyx_k_strip, sizeof(__pyx_k_strip), 0, 0, 1, 1},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_sub_mi1, __pyx_k_sub_mi1, sizeof(__pyx_k_sub_mi1), 0, 0, 1, 1},
  {&__pyx_n_s_sub_mi2, __pyx_k_sub_mi2, sizeof(__pyx_k_sub_mi2), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_total_particles, __pyx_k_total_particles, sizeof(__pyx_k_total_particles), 0, 0, 1, 1},
  {&__pyx_n_s_u8, __pyx_k_u8, sizeof(__pyx_k_u8), 0, 0, 1, 1},
  {&__pyx_n_s_uint32, __pyx_k_uint32, sizeof(__pyx_k_uint32), 0, 0, 1, 1},
  {&__pyx_n_s_uint64, __pyx_k_uint64, sizeof(__pyx_k_uint64), 0, 0, 1, 1},
  {&__pyx_n_s_uint8, __pyx_k_uint8, sizeof(__pyx_k_uint8), 0, 0, 1, 1},
  {&__pyx_kp_s_unable_to_allocate_array_data, __pyx_k_unable_to_allocate_array_data, sizeof(__pyx_k_unable_to_allocate_array_data), 0, 0, 1, 0},
  {&__pyx_kp_s_unable_to_allocate_shape_and_str, __pyx_k_unable_to_allocate_shape_and_str, sizeof(__pyx_k_unable_to_allocate_shape_and_str), 0, 0, 1, 0},
  {&__pyx_n_s_units, __pyx_k_units, sizeof(__pyx_k_units), 0, 0, 1, 1},
  {&__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_k_unknown_dtype_code_in_numpy_pxd, sizeof(__pyx_k_unknown_dtype_code_in_numpy_pxd), 0, 1, 0, 0},
  {&__pyx_n_s_unpack, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {&__pyx_n_s_verbose, __pyx_k_verbose, sizeof(__pyx_k_verbose), 0, 0, 1, 1},
  {&__pyx_n_s_wb, __pyx_k_wb, sizeof(__pyx_k_wb), 0, 0, 1, 1},
  {&__pyx_n_s_wendland2, __pyx_k_wendland2, sizeof(__pyx_k_wendland2), 0, 0, 1, 1},
  {&__pyx_n_s_wendland4, __pyx_k_wendland4, sizeof(__pyx_k_wendland4), 0, 0, 1, 1},
  {&__pyx_n_s_wendland6, __pyx_k_wendland6, sizeof(__pyx_k_wendland6), 0, 0, 1, 1},
  {&__pyx_n_s_where, __pyx_k_where, sizeof(__pyx_k_where), 0, 0, 1, 1},
  {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
  {&__pyx_n_s_yield_coordinates, __pyx_k_yield_coordinates, sizeof(__pyx_k_yield_coordinates), 0, 0, 1, 1},
  {&__pyx_n_s_yt_funcs, __pyx_k_yt_funcs, sizeof(__pyx_k_yt_funcs), 0, 0, 1, 1},
  {&__pyx_n_s_yt_geometry_particle_oct_contain, __pyx_k_yt_geometry_particle_oct_contain, sizeof(__pyx_k_yt_geometry_particle_oct_contain), 0, 0, 1, 1},
  {&__pyx_kp_s_yt_geometry_particle_oct_contain_2, __pyx_k_yt_geometry_particle_oct_contain_2, sizeof(__pyx_k_yt_geometry_particle_oct_contain_2), 0, 0, 1, 0},
  {&__pyx_n_s_zeros, __pyx_k_zeros, sizeof(__pyx_k_zeros), 0, 0, 1, 1},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 101, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 243, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 385, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 642, __pyx_L1_error)
  __pyx_builtin_open = __Pyx_GetBuiltinName(__pyx_n_s_open); if (!__pyx_builtin_open) __PYX_ERR(0, 1026, __pyx_L1_error)
  __pyx_builtin_OSError = __Pyx_GetBuiltinName(__pyx_n_s_OSError); if (!__pyx_builtin_OSError) __PYX_ERR(0, 1058, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 1240, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 1240, __pyx_L1_error)
  __pyx_builtin_sum = __Pyx_GetBuiltinName(__pyx_n_s_sum); if (!__pyx_builtin_sum) __PYX_ERR(0, 1397, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(2, 272, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(2, 1037, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(5, 137, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(1, 149, __pyx_L1_error)
  __pyx_builtin_Ellipsis = __Pyx_GetBuiltinName(__pyx_n_s_Ellipsis); if (!__pyx_builtin_Ellipsis) __PYX_ERR(1, 406, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(1, 615, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "yt/geometry/particle_oct_container.pyx":357
 *         for i in range(2,3):
 *             if level_offset[i] != level_offset[0]:
 *                 raise Exception("Octree must have the same number of roots in each dimension for this.")             # <<<<<<<<<<<<<<
 *         # Get root for index
 *         index = (mi >> ((order - level_offset[0])*3))
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_Octree_must_have_the_same_number); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "yt/geometry/particle_oct_container.pyx":380
 *     def apply_domain(self, int domain_id, BoolArrayCollection mask,
 *                      int masklevel):
 *         cdef SelectorObject selector = AlwaysSelector(None)             # <<<<<<<<<<<<<<
 *         ind = self.domain_ind(selector, mask = mask, masklevel = masklevel)
 *         for i in range(self.nocts):
 */
  __pyx_tuple__4 = PyTuple_Pack(1, Py_None); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.dom_offsets,self.oct_list,self.root_mesh cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dom_offsets,self.oct_list,self.root_mesh cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_self_dom_offsets_self_oct_list_s); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "(tree fragment)":4
 *     raise TypeError("self.dom_offsets,self.oct_list,self.root_mesh cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.dom_offsets,self.oct_list,self.root_mesh cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_self_dom_offsets_self_oct_list_s); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "yt/geometry/particle_oct_container.pyx":540
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def _coarse_index_data_file(self,             # <<<<<<<<<<<<<<
 *                                 np.ndarray[floating, ndim=2] pos,
 *                                 np.ndarray[floating, ndim=1] hsml,
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_No_matching_signature_found); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_Function_call_with_ambiguous_arg); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "yt/geometry/particle_oct_container.pyx":570
 *         cdef np.float64_t dds[3]
 *         cdef np.float64_t radius
 *         cdef np.uint8_t[:] mask = self.masks[:, file_id]             # <<<<<<<<<<<<<<
 *         cdef np.uint64_t[:] particle_counts = self.particle_counts
 *         cdef np.uint64_t msize = (1 << (self.index_order1 * 3))
 */
  __pyx_slice__11 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__11)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__11);
  __Pyx_GIVEREF(__pyx_slice__11);

  /* "yt/geometry/particle_oct_container.pyx":753
 *             sorted_order = np.argsort(morton_indices).astype(np.uint64, copy=False)
 *         else:
 *             sorted_order = np.argsort(hsml)[::-1].astype(np.uint64, copy=False)             # <<<<<<<<<<<<<<
 *         for sorted_ind in range(sorted_order.shape[0]):
 *             p = sorted_order[sorted_ind]
 */
  __pyx_slice__17 = PySlice_New(Py_None, Py_None, __pyx_int_neg_1); if (unlikely(!__pyx_slice__17)) __PYX_ERR(0, 753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__17);
  __Pyx_GIVEREF(__pyx_slice__17);
  __pyx_slice__18 = PySlice_New(Py_None, Py_None, __pyx_int_neg_1); if (unlikely(!__pyx_slice__18)) __PYX_ERR(0, 753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__18);
  __Pyx_GIVEREF(__pyx_slice__18);

  /* "yt/geometry/particle_oct_container.pyx":1066
 *             ver = 0 # Original bitmaps had number of files first
 *         if ver != _bitmask_version:
 *             raise OSError("The file format of the index has changed since "             # <<<<<<<<<<<<<<
 *                           "this file was created. It will be replaced with an "
 *                           "updated version.")
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_The_file_format_of_the_index_has); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "yt/geometry/particle_oct_container.pyx":1175
 *         cdef int i
 *         cdef np.ndarray[np.float64_t, ndim=1] DLE
 *         DLE = np.zeros(3, dtype='float64')             # <<<<<<<<<<<<<<
 *         for i in range(3):
 *             DLE[i] = self.left_edge[i]
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_int_3); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.file_markers cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.file_markers cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_s_self_file_markers_cannot_be_conv); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "(tree fragment)":4
 *     raise TypeError("self.file_markers cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.file_markers cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_kp_s_self_file_markers_cannot_be_conv); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "yt/geometry/particle_oct_container.pyx":2133
 *             # noct = o.children[cind(ind[0],ind[1],ind[2])]
 *             if o.children[cind(ind[0],ind[1],ind[2])] != NULL:
 *                 raise Exception('Child was already initialized...')             # <<<<<<<<<<<<<<
 *             noct = self.allocate_oct()
 *             noct.domain = o.domain
 */
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_kp_s_Child_was_already_initialized); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 2133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self._ptr_octs_per_root,self.oct_list,self.root_mesh,self.root_nodes,self.tree_root cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._ptr_octs_per_root,self.oct_list,self.root_mesh,self.root_nodes,self.tree_root cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_s_self__ptr_octs_per_root_self_oct); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "(tree fragment)":4
 *     raise TypeError("self._ptr_octs_per_root,self.oct_list,self.root_mesh,self.root_nodes,self.tree_root cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._ptr_octs_per_root,self.oct_list,self.root_mesh,self.root_nodes,self.tree_root cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_s_self__ptr_octs_per_root_self_oct); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_C_contiguous); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(2, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_ARRAY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_Fortran_contiguou); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(2, 276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":306
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
  __pyx_tuple__30 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(2, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":855
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
  __pyx_tuple__31 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(2, 855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":879
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor_2); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(2, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1037
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__33 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(2, 1037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "../_h_env_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_placehold_plac/lib/python3.8/site-packages/Cython/Includes/numpy/__init__.pxd":1043
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__34 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(2, 1043, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "View.MemoryView":134
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
  __pyx_tuple__35 = PyTuple_Pack(1, __pyx_kp_s_Empty_shape_tuple_for_cython_arr); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(1, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "View.MemoryView":137
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
  __pyx_tuple__36 = PyTuple_Pack(1, __pyx_kp_s_itemsize_0_for_cython_array); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(1, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "View.MemoryView":149
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__37 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_shape_and_str); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(1, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);

  /* "View.MemoryView":177
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */
  __pyx_tuple__38 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_array_data); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "View.MemoryView":193
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
  __pyx_tuple__39 = PyTuple_Pack(1, __pyx_kp_s_Can_only_create_a_buffer_that_is); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(1, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__40 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__41 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);

  /* "View.MemoryView":420
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
  __pyx_tuple__42 = PyTuple_Pack(1, __pyx_kp_s_Cannot_assign_to_read_only_memor); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(1, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);

  /* "View.MemoryView":497
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
  __pyx_tuple__43 = PyTuple_Pack(1, __pyx_kp_s_Unable_to_convert_item_to_object); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(1, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);

  /* "View.MemoryView":522
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
  __pyx_tuple__44 = PyTuple_Pack(1, __pyx_kp_s_Cannot_create_writable_memory_vi); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(1, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);

  /* "View.MemoryView":572
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
  __pyx_tuple__45 = PyTuple_Pack(1, __pyx_kp_s_Buffer_view_does_not_expose_stri); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(1, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);

  /* "View.MemoryView":579
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
  __pyx_tuple__46 = PyTuple_New(1); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(1, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_tuple__46, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_tuple__46);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__47 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__48 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);

  /* "View.MemoryView":705
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__49 = PyTuple_Pack(1, __pyx_kp_s_Indirect_dimensions_not_supporte); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(1, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__50 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__51 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);
  __pyx_tuple__52 = PyTuple_Pack(3, __pyx_int_184977713, __pyx_int_136983863, __pyx_int_112105877); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);

  /* "yt/geometry/particle_oct_container.pyx":79
 * DEF RefinedExternalGhosts = 1
 * 
 * _bitmask_version = np.uint64(5)             # <<<<<<<<<<<<<<
 * 
 * from ..utilities.lib.ewah_bool_wrap cimport (
 */
  __pyx_tuple__58 = PyTuple_Pack(1, __pyx_int_5); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);

  /* "yt/geometry/particle_oct_container.pyx":540
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def _coarse_index_data_file(self,             # <<<<<<<<<<<<<<
 *                                 np.ndarray[floating, ndim=2] pos,
 *                                 np.ndarray[floating, ndim=1] hsml,
 */
  __pyx_tuple__59 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_pos, __pyx_n_s_hsml, __pyx_n_s_file_id); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);
  __pyx_codeobj__60 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__59, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_yt_geometry_particle_oct_contain_2, __pyx_n_s_coarse_index_data_file, 540, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__60)) __PYX_ERR(0, 540, __pyx_L1_error)

  /* "yt/geometry/particle_oct_container.pyx":669
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     def _refined_index_data_file(self,             # <<<<<<<<<<<<<<
 *                                  BoolArrayCollection in_collection,
 *                                  np.ndarray[floating, ndim=2] pos,
 */
  __pyx_tuple__61 = PyTuple_Pack(13, __pyx_n_s_self, __pyx_n_s_in_collection, __pyx_n_s_pos, __pyx_n_s_hsml, __pyx_n_s_mask, __pyx_n_s_sub_mi1, __pyx_n_s_sub_mi2, __pyx_n_s_file_id, __pyx_n_s_nsub_mi, __pyx_n_s_count_threshold, __pyx_n_s_mask_threshold, __pyx_n_s_in_coll, __pyx_n_s_out_collection); if (unlikely(!__pyx_tuple__61)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__61);
  __Pyx_GIVEREF(__pyx_tuple__61);
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(11, 0, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__61, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_yt_geometry_particle_oct_contain_2, __pyx_n_s_refined_index_data_file, 669, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) __PYX_ERR(0, 669, __pyx_L1_error)

  /* "View.MemoryView":287
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_tuple__63 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct_or_indirect); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(1, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);

  /* "View.MemoryView":288
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_tuple__64 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct); if (unlikely(!__pyx_tuple__64)) __PYX_ERR(1, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);

  /* "View.MemoryView":289
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__65 = PyTuple_Pack(1, __pyx_kp_s_strided_and_indirect); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(1, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);

  /* "View.MemoryView":292
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_tuple__66 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_direct); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(1, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);

  /* "View.MemoryView":293
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__67 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_indirect); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(1, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__68 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);
  __pyx_codeobj__69 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Enum, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__69)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_9 = PyInt_FromLong(9); if (unlikely(!__pyx_int_9)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_112105877 = PyInt_FromLong(112105877L); if (unlikely(!__pyx_int_112105877)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_136983863 = PyInt_FromLong(136983863L); if (unlikely(!__pyx_int_136983863)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_184977713 = PyInt_FromLong(184977713L); if (unlikely(!__pyx_int_184977713)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  generic = Py_None; Py_INCREF(Py_None);
  strided = Py_None; Py_INCREF(Py_None);
  indirect = Py_None; Py_INCREF(Py_None);
  contiguous = Py_None; Py_INCREF(Py_None);
  indirect_contiguous = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_t_1 = PyImport_ImportModule("yt.geometry.oct_container"); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_container", "OctreeContainer", sizeof(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_13oct_container_OctreeContainer = (struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctreeContainer*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_13oct_container_OctreeContainer)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer = &__pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer.__pyx_base = *__pyx_vtabptr_2yt_8geometry_13oct_container_OctreeContainer;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer.__pyx_base.get_domain_offset = (__pyx_t_5numpy_int64_t (*)(struct __pyx_obj_2yt_8geometry_13oct_container_OctreeContainer *, int))__pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_get_domain_offset;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer.visit_free = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *))__pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_visit_free;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer.visit_clear = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *))__pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_visit_clear;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer.visit_assign = (PyObject *(*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_int64_t *, int, int *))__pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_visit_assign;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer.allocate_oct = (struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *(*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *))__pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_allocate_oct;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer.refine_oct = (struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *(*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_uint64_t, int, __pyx_t_5numpy_uint8_t))__pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_refine_oct;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer.filter_particles = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_uint64_t *, __pyx_t_5numpy_int64_t, int, __pyx_t_5numpy_uint8_t))__pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_filter_particles;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer.visit = (PyObject *(*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_int64_t *, struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_visit *__pyx_optional_args))__pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_visit;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer.get_from_index = (struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *(*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer *, __pyx_t_5numpy_uint64_t, struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_get_from_index *__pyx_optional_args))__pyx_f_2yt_8geometry_22particle_oct_container_23ParticleOctreeContainer_get_from_index;
  __pyx_type_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer.tp_base = __pyx_ptype_2yt_8geometry_13oct_container_OctreeContainer;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer) < 0) __PYX_ERR(0, 90, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer.tp_dictoffset && __pyx_type_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer.tp_dict, __pyx_vtabptr_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer) < 0) __PYX_ERR(0, 90, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ParticleOctreeContainer, (PyObject *)&__pyx_type_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer) < 0) __PYX_ERR(0, 90, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer) < 0) __PYX_ERR(0, 90, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer = &__pyx_type_2yt_8geometry_22particle_oct_container_ParticleOctreeContainer;
  __pyx_vtabptr_2yt_8geometry_22particle_oct_container_ParticleBitmap = &__pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmap;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmap._update_mi2 = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *, __pyx_t_5numpy_uint64_t))__pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap__update_mi2;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmap.__pyx___set_coarse_index_data_file = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *, __pyx_t_5numpy_uint64_t))__pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___set_coarse_index_data_file;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmap.__pyx___fill_refined_ranges = (__pyx_t_5numpy_int64_t (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_float64_t *, __pyx_t_2yt_9utilities_3lib_15ewah_bool_array_bool_array &))__pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___fill_refined_ranges;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmap.__pyx___set_refined_index_data_file = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *, PyArrayObject *, PyArrayObject *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_int64_t))__pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___set_refined_index_data_file;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmap.__pyx_fuse_0__coarse_index_data_file = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *, PyArrayObject *, PyArrayObject *, __pyx_t_5numpy_uint64_t))__pyx_fuse_0__pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___coarse_index_data_file;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmap.__pyx_fuse_1__coarse_index_data_file = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *, PyArrayObject *, PyArrayObject *, __pyx_t_5numpy_uint64_t))__pyx_fuse_1__pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___coarse_index_data_file;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmap.__pyx_fuse_0__refined_index_data_file = (struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *(*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, PyArrayObject *, PyArrayObject *, PyArrayObject *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint8_t))__pyx_fuse_0__pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___refined_index_data_file;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmap.__pyx_fuse_1__refined_index_data_file = (struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *(*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmap *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection *, PyArrayObject *, PyArrayObject *, PyArrayObject *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint8_t))__pyx_fuse_1__pyx_f_2yt_8geometry_22particle_oct_container_14ParticleBitmap___refined_index_data_file;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmap) < 0) __PYX_ERR(0, 433, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmap.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmap.tp_dictoffset && __pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmap.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmap.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmap.tp_dict, __pyx_vtabptr_2yt_8geometry_22particle_oct_container_ParticleBitmap) < 0) __PYX_ERR(0, 433, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ParticleBitmap, (PyObject *)&__pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmap) < 0) __PYX_ERR(0, 433, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmap) < 0) __PYX_ERR(0, 433, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleBitmap = &__pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmap;
  __pyx_vtabptr_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector = &__pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.is_refined = (int (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_is_refined;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.is_refined_files = (int (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_is_refined_files;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.add_coarse = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t, struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_coarse *__pyx_optional_args))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_coarse;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.set_files_coarse = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_files_coarse;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.add_refined = (int (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, struct __pyx_opt_args_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_refined *__pyx_optional_args))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_refined;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.set_files_refined = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_files_refined;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.add_neighbors_coarse = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_neighbors_coarse;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.set_files_neighbors_coarse = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_files_neighbors_coarse;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.add_neighbors_refined = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_neighbors_refined;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.set_files_neighbors_refined = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_files_neighbors_refined;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.set_coarse_list = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_coarse_list;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.set_refined_list = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_refined_list;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.set_coarse_bool = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_coarse_bool;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.set_refined_bool = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_set_refined_bool;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.push_refined_bool = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_push_refined_bool;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.add_ghost_zones = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *, struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed *))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_add_ghost_zones;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.fill_subcells_mi1 = (int (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t *))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_fill_subcells_mi1;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.fill_subcells_mi2 = (int (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t *))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_fill_subcells_mi2;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.recursive_morton_mask = (int (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_int32_t, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t *))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_recursive_morton_mask;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.recursive_morton_files = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector *, __pyx_t_5numpy_int32_t, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_float64_t *, __pyx_t_5numpy_uint64_t))__pyx_f_2yt_8geometry_22particle_oct_container_22ParticleBitmapSelector_recursive_morton_files;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector) < 0) __PYX_ERR(0, 1442, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.tp_dictoffset && __pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector.tp_dict, __pyx_vtabptr_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector) < 0) __PYX_ERR(0, 1442, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ParticleBitmapSelector, (PyObject *)&__pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector) < 0) __PYX_ERR(0, 1442, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector) < 0) __PYX_ERR(0, 1442, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector = &__pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapSelector;
  __pyx_ptype_2yt_8geometry_13oct_container_SparseOctreeContainer = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_container", "SparseOctreeContainer", sizeof(struct __pyx_obj_2yt_8geometry_13oct_container_SparseOctreeContainer), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_13oct_container_SparseOctreeContainer) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_13oct_container_SparseOctreeContainer = (struct __pyx_vtabstruct_2yt_8geometry_13oct_container_SparseOctreeContainer*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_13oct_container_SparseOctreeContainer->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_13oct_container_SparseOctreeContainer)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer = &__pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer.__pyx_base = *__pyx_vtabptr_2yt_8geometry_13oct_container_SparseOctreeContainer;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer.visit_assign = (PyObject *(*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_int64_t *, int, int *, __pyx_t_5numpy_int64_t))__pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_visit_assign;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer.allocate_oct = (struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *(*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *))__pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_allocate_oct;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer._get_index_base_octs = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *, __Pyx_memviewslice, __Pyx_memviewslice))__pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer__get_index_base_octs;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer.visit_free = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, int))__pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_visit_free;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer.recursive_add = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, PyArrayObject *, int, int *, int, int *))__pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_recursive_add;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer.refine_oct = (struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *(*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_uint64_t, int))__pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_refine_oct;
  __pyx_vtable_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer.filter_particles = (void (*)(struct __pyx_obj_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer *, struct __pyx_t_2yt_8geometry_12oct_visitors_Oct *, __pyx_t_5numpy_uint64_t *, __pyx_t_5numpy_int64_t, int))__pyx_f_2yt_8geometry_22particle_oct_container_29ParticleBitmapOctreeContainer_filter_particles;
  __pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer.tp_base = __pyx_ptype_2yt_8geometry_13oct_container_SparseOctreeContainer;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer) < 0) __PYX_ERR(0, 1965, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer.tp_dictoffset && __pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer.tp_dict, __pyx_vtabptr_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer) < 0) __PYX_ERR(0, 1965, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ParticleBitmapOctreeContainer, (PyObject *)&__pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer) < 0) __PYX_ERR(0, 1965, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer) < 0) __PYX_ERR(0, 1965, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer = &__pyx_type_2yt_8geometry_22particle_oct_container_ParticleBitmapOctreeContainer;
  if (PyType_Ready(&__pyx_type_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__) < 0) __PYX_ERR(0, 152, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__.tp_dictoffset && __pyx_type_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__ = &__pyx_type_2yt_8geometry_22particle_oct_container___pyx_scope_struct____iter__;
  __pyx_vtabptr_array = &__pyx_vtable_array;
  __pyx_vtable_array.get_memview = (PyObject *(*)(struct __pyx_array_obj *))__pyx_array_get_memview;
  if (PyType_Ready(&__pyx_type___pyx_array) < 0) __PYX_ERR(1, 106, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_array.tp_print = 0;
  #endif
  if (__Pyx_SetVtable(__pyx_type___pyx_array.tp_dict, __pyx_vtabptr_array) < 0) __PYX_ERR(1, 106, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_array) < 0) __PYX_ERR(1, 106, __pyx_L1_error)
  __pyx_array_type = &__pyx_type___pyx_array;
  if (PyType_Ready(&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(1, 280, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_MemviewEnum.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_MemviewEnum.tp_dictoffset && __pyx_type___pyx_MemviewEnum.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_MemviewEnum.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(1, 280, __pyx_L1_error)
  __pyx_MemviewEnum_type = &__pyx_type___pyx_MemviewEnum;
  __pyx_vtabptr_memoryview = &__pyx_vtable_memoryview;
  __pyx_vtable_memoryview.get_item_pointer = (char *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_get_item_pointer;
  __pyx_vtable_memoryview.is_slice = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_is_slice;
  __pyx_vtable_memoryview.setitem_slice_assignment = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_slice_assignment;
  __pyx_vtable_memoryview.setitem_slice_assign_scalar = (PyObject *(*)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_setitem_slice_assign_scalar;
  __pyx_vtable_memoryview.setitem_indexed = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_indexed;
  __pyx_vtable_memoryview.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryview_convert_item_to_object;
  __pyx_vtable_memoryview.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryview_assign_item_from_object;
  if (PyType_Ready(&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(1, 331, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryview.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryview.tp_dictoffset && __pyx_type___pyx_memoryview.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryview.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryview.tp_dict, __pyx_vtabptr_memoryview) < 0) __PYX_ERR(1, 331, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(1, 331, __pyx_L1_error)
  __pyx_memoryview_type = &__pyx_type___pyx_memoryview;
  __pyx_vtabptr__memoryviewslice = &__pyx_vtable__memoryviewslice;
  __pyx_vtable__memoryviewslice.__pyx_base = *__pyx_vtabptr_memoryview;
  __pyx_vtable__memoryviewslice.__pyx_base.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryviewslice_convert_item_to_object;
  __pyx_vtable__memoryviewslice.__pyx_base.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryviewslice_assign_item_from_object;
  __pyx_type___pyx_memoryviewslice.tp_base = __pyx_memoryview_type;
  if (PyType_Ready(&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(1, 967, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryviewslice.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryviewslice.tp_dictoffset && __pyx_type___pyx_memoryviewslice.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryviewslice.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryviewslice.tp_dict, __pyx_vtabptr__memoryviewslice) < 0) __PYX_ERR(1, 967, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(1, 967, __pyx_L1_error)
  __pyx_memoryviewslice_type = &__pyx_type___pyx_memoryviewslice;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(6, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(2, 206, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(2, 229, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(2, 233, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(2, 242, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(2, 917, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("yt.geometry.selection_routines"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject = __Pyx_ImportType(__pyx_t_1, "yt.geometry.selection_routines", "SelectorObject", sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_SelectorObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject) __PYX_ERR(7, 24, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_SelectorObject*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_18selection_routines_SelectorObject->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_18selection_routines_SelectorObject)) __PYX_ERR(7, 24, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_AlwaysSelector = __Pyx_ImportType(__pyx_t_1, "yt.geometry.selection_routines", "AlwaysSelector", sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_AlwaysSelector), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_18selection_routines_AlwaysSelector) __PYX_ERR(7, 76, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_18selection_routines_AlwaysSelector = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_AlwaysSelector*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_18selection_routines_AlwaysSelector->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_18selection_routines_AlwaysSelector)) __PYX_ERR(7, 76, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_OctreeSubsetSelector = __Pyx_ImportType(__pyx_t_1, "yt.geometry.selection_routines", "OctreeSubsetSelector", sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_OctreeSubsetSelector), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_18selection_routines_OctreeSubsetSelector) __PYX_ERR(7, 79, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_18selection_routines_OctreeSubsetSelector = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_OctreeSubsetSelector*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_18selection_routines_OctreeSubsetSelector->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_18selection_routines_OctreeSubsetSelector)) __PYX_ERR(7, 79, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_18selection_routines_BooleanSelector = __Pyx_ImportType(__pyx_t_1, "yt.geometry.selection_routines", "BooleanSelector", sizeof(struct __pyx_obj_2yt_8geometry_18selection_routines_BooleanSelector), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_18selection_routines_BooleanSelector) __PYX_ERR(7, 83, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanSelector = (struct __pyx_vtabstruct_2yt_8geometry_18selection_routines_BooleanSelector*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_18selection_routines_BooleanSelector->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_18selection_routines_BooleanSelector)) __PYX_ERR(7, 83, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("yt.utilities.lib.allocation_container"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2yt_9utilities_3lib_20allocation_container_ObjectPool = __Pyx_ImportType(__pyx_t_1, "yt.utilities.lib.allocation_container", "ObjectPool", sizeof(struct __pyx_obj_2yt_9utilities_3lib_20allocation_container_ObjectPool), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_9utilities_3lib_20allocation_container_ObjectPool) __PYX_ERR(8, 20, __pyx_L1_error)
  __pyx_vtabptr_2yt_9utilities_3lib_20allocation_container_ObjectPool = (struct __pyx_vtabstruct_2yt_9utilities_3lib_20allocation_container_ObjectPool*)__Pyx_GetVtable(__pyx_ptype_2yt_9utilities_3lib_20allocation_container_ObjectPool->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_9utilities_3lib_20allocation_container_ObjectPool)) __PYX_ERR(8, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("yt.geometry.oct_container"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2yt_8geometry_13oct_container_OctObjectPool = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_container", "OctObjectPool", sizeof(struct __pyx_obj_2yt_8geometry_13oct_container_OctObjectPool), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_13oct_container_OctObjectPool) __PYX_ERR(9, 47, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_13oct_container_OctObjectPool = (struct __pyx_vtabstruct_2yt_8geometry_13oct_container_OctObjectPool*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_13oct_container_OctObjectPool->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_13oct_container_OctObjectPool)) __PYX_ERR(9, 47, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_13oct_container_RAMSESOctreeContainer = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_container", "RAMSESOctreeContainer", sizeof(struct __pyx_obj_2yt_8geometry_13oct_container_RAMSESOctreeContainer), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_13oct_container_RAMSESOctreeContainer) __PYX_ERR(9, 93, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_13oct_container_RAMSESOctreeContainer = (struct __pyx_vtabstruct_2yt_8geometry_13oct_container_RAMSESOctreeContainer*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_13oct_container_RAMSESOctreeContainer->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_13oct_container_RAMSESOctreeContainer)) __PYX_ERR(9, 93, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("yt.geometry.oct_visitors"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2yt_8geometry_12oct_visitors_OctVisitor = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "OctVisitor", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_OctVisitor), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_OctVisitor) __PYX_ERR(10, 33, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_OctVisitor = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_OctVisitor*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_OctVisitor->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_OctVisitor)) __PYX_ERR(10, 33, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_CountTotalOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "CountTotalOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_CountTotalOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_CountTotalOcts) __PYX_ERR(10, 59, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_CountTotalOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CountTotalOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_CountTotalOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_CountTotalOcts)) __PYX_ERR(10, 59, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_CountTotalCells = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "CountTotalCells", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_CountTotalCells), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_CountTotalCells) __PYX_ERR(10, 62, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_CountTotalCells = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CountTotalCells*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_CountTotalCells->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_CountTotalCells)) __PYX_ERR(10, 62, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_MarkOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "MarkOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_MarkOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_MarkOcts) __PYX_ERR(10, 65, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_MarkOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_MarkOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_MarkOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_MarkOcts)) __PYX_ERR(10, 65, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_MaskOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "MaskOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_MaskOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_MaskOcts) __PYX_ERR(10, 69, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_MaskOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_MaskOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_MaskOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_MaskOcts)) __PYX_ERR(10, 69, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_IndexOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "IndexOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_IndexOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_IndexOcts) __PYX_ERR(10, 72, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_IndexOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IndexOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_IndexOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_IndexOcts)) __PYX_ERR(10, 72, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_MaskedIndexOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "MaskedIndexOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_MaskedIndexOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_MaskedIndexOcts) __PYX_ERR(10, 75, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_MaskedIndexOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_MaskedIndexOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_MaskedIndexOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_MaskedIndexOcts)) __PYX_ERR(10, 75, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_IndexMaskMapOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "IndexMaskMapOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_IndexMaskMapOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_IndexMaskMapOcts) __PYX_ERR(10, 79, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_IndexMaskMapOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IndexMaskMapOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_IndexMaskMapOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_IndexMaskMapOcts)) __PYX_ERR(10, 79, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_ICoordsOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "ICoordsOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_ICoordsOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_ICoordsOcts) __PYX_ERR(10, 85, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_ICoordsOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_ICoordsOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_ICoordsOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_ICoordsOcts)) __PYX_ERR(10, 85, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_IResOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "IResOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_IResOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_IResOcts) __PYX_ERR(10, 88, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_IResOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IResOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_IResOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_IResOcts)) __PYX_ERR(10, 88, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_FCoordsOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "FCoordsOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_FCoordsOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_FCoordsOcts) __PYX_ERR(10, 91, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_FCoordsOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FCoordsOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_FCoordsOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_FCoordsOcts)) __PYX_ERR(10, 91, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_FWidthOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "FWidthOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_FWidthOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_FWidthOcts) __PYX_ERR(10, 94, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_FWidthOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FWidthOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_FWidthOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_FWidthOcts)) __PYX_ERR(10, 94, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_CopyArrayI64 = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "CopyArrayI64", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_CopyArrayI64), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_CopyArrayI64) __PYX_ERR(10, 97, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_CopyArrayI64 = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CopyArrayI64*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_CopyArrayI64->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_CopyArrayI64)) __PYX_ERR(10, 97, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_CopyArrayF64 = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "CopyArrayF64", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_CopyArrayF64), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_CopyArrayF64) __PYX_ERR(10, 101, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_CopyArrayF64 = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CopyArrayF64*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_CopyArrayF64->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_CopyArrayF64)) __PYX_ERR(10, 101, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_CopyFileIndArrayI8 = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "CopyFileIndArrayI8", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_CopyFileIndArrayI8), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_CopyFileIndArrayI8) __PYX_ERR(10, 105, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_CopyFileIndArrayI8 = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CopyFileIndArrayI8*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_CopyFileIndArrayI8->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_CopyFileIndArrayI8)) __PYX_ERR(10, 105, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_IdentifyOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "IdentifyOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_IdentifyOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_IdentifyOcts) __PYX_ERR(10, 110, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_IdentifyOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_IdentifyOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_IdentifyOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_IdentifyOcts)) __PYX_ERR(10, 110, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_AssignDomainInd = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "AssignDomainInd", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_AssignDomainInd), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_AssignDomainInd) __PYX_ERR(10, 113, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_AssignDomainInd = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_AssignDomainInd*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_AssignDomainInd->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_AssignDomainInd)) __PYX_ERR(10, 113, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_FillFileIndicesO = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "FillFileIndicesO", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_FillFileIndicesO), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_FillFileIndicesO) __PYX_ERR(10, 116, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_FillFileIndicesO = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FillFileIndicesO*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_FillFileIndicesO->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_FillFileIndicesO)) __PYX_ERR(10, 116, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_FillFileIndicesR = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "FillFileIndicesR", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_FillFileIndicesR), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_FillFileIndicesR) __PYX_ERR(10, 121, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_FillFileIndicesR = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_FillFileIndicesR*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_FillFileIndicesR->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_FillFileIndicesR)) __PYX_ERR(10, 121, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_CountByDomain = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "CountByDomain", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_CountByDomain), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_CountByDomain) __PYX_ERR(10, 126, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_CountByDomain = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_CountByDomain*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_CountByDomain->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_CountByDomain)) __PYX_ERR(10, 126, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_StoreOctree = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "StoreOctree", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_StoreOctree), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_StoreOctree) __PYX_ERR(10, 129, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_StoreOctree = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_StoreOctree*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_StoreOctree->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_StoreOctree)) __PYX_ERR(10, 129, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_LoadOctree = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "LoadOctree", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_LoadOctree), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_LoadOctree) __PYX_ERR(10, 132, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_LoadOctree = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_LoadOctree*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_LoadOctree->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_LoadOctree)) __PYX_ERR(10, 132, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_MortonIndexOcts = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "MortonIndexOcts", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_MortonIndexOcts), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_MortonIndexOcts) __PYX_ERR(10, 138, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_MortonIndexOcts = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_MortonIndexOcts*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_MortonIndexOcts->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_MortonIndexOcts)) __PYX_ERR(10, 138, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_StoreIndex = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "StoreIndex", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_StoreIndex), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_StoreIndex) __PYX_ERR(10, 149, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_StoreIndex = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_StoreIndex*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_StoreIndex->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_StoreIndex)) __PYX_ERR(10, 149, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "BaseNeighbourVisitor", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor) __PYX_ERR(10, 153, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_BaseNeighbourVisitor)) __PYX_ERR(10, 153, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_NeighbourCellIndexVisitor = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "NeighbourCellIndexVisitor", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_NeighbourCellIndexVisitor), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_NeighbourCellIndexVisitor) __PYX_ERR(10, 169, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_NeighbourCellIndexVisitor = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_NeighbourCellIndexVisitor*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_NeighbourCellIndexVisitor->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_NeighbourCellIndexVisitor)) __PYX_ERR(10, 169, __pyx_L1_error)
  __pyx_ptype_2yt_8geometry_12oct_visitors_NeighbourCellVisitor = __Pyx_ImportType(__pyx_t_1, "yt.geometry.oct_visitors", "NeighbourCellVisitor", sizeof(struct __pyx_obj_2yt_8geometry_12oct_visitors_NeighbourCellVisitor), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_12oct_visitors_NeighbourCellVisitor) __PYX_ERR(10, 173, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_12oct_visitors_NeighbourCellVisitor = (struct __pyx_vtabstruct_2yt_8geometry_12oct_visitors_NeighbourCellVisitor*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_12oct_visitors_NeighbourCellVisitor->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_12oct_visitors_NeighbourCellVisitor)) __PYX_ERR(10, 173, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("yt.geometry.particle_deposit"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2yt_8geometry_16particle_deposit_ParticleDepositOperation = __Pyx_ImportType(__pyx_t_1, "yt.geometry.particle_deposit", "ParticleDepositOperation", sizeof(struct __pyx_obj_2yt_8geometry_16particle_deposit_ParticleDepositOperation), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_8geometry_16particle_deposit_ParticleDepositOperation) __PYX_ERR(5, 139, __pyx_L1_error)
  __pyx_vtabptr_2yt_8geometry_16particle_deposit_ParticleDepositOperation = (struct __pyx_vtabstruct_2yt_8geometry_16particle_deposit_ParticleDepositOperation*)__Pyx_GetVtable(__pyx_ptype_2yt_8geometry_16particle_deposit_ParticleDepositOperation->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_8geometry_16particle_deposit_ParticleDepositOperation)) __PYX_ERR(5, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("yt.utilities.lib.ewah_bool_wrap"); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks = __Pyx_ImportType(__pyx_t_1, "yt.utilities.lib.ewah_bool_wrap", "FileBitmasks", sizeof(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks) __PYX_ERR(11, 16, __pyx_L1_error)
  __pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks = (struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks*)__Pyx_GetVtable(__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_FileBitmasks)) __PYX_ERR(11, 16, __pyx_L1_error)
  __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection = __Pyx_ImportType(__pyx_t_1, "yt.utilities.lib.ewah_bool_wrap", "BoolArrayCollection", sizeof(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection) __PYX_ERR(11, 58, __pyx_L1_error)
  __pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection = (struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection*)__Pyx_GetVtable(__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollection)) __PYX_ERR(11, 58, __pyx_L1_error)
  __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed = __Pyx_ImportType(__pyx_t_1, "yt.utilities.lib.ewah_bool_wrap", "BoolArrayCollectionUncompressed", sizeof(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed) __PYX_ERR(11, 98, __pyx_L1_error)
  __pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed = (struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed*)__Pyx_GetVtable(__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_BoolArrayCollectionUncompressed)) __PYX_ERR(11, 98, __pyx_L1_error)
  __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet = __Pyx_ImportType(__pyx_t_1, "yt.utilities.lib.ewah_bool_wrap", "SparseUnorderedBitmaskSet", sizeof(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet) __PYX_ERR(11, 123, __pyx_L1_error)
  __pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet = (struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet*)__Pyx_GetVtable(__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskSet)) __PYX_ERR(11, 123, __pyx_L1_error)
  __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector = __Pyx_ImportType(__pyx_t_1, "yt.utilities.lib.ewah_bool_wrap", "SparseUnorderedBitmaskVector", sizeof(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector) __PYX_ERR(11, 132, __pyx_L1_error)
  __pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector = (struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector*)__Pyx_GetVtable(__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedBitmaskVector)) __PYX_ERR(11, 132, __pyx_L1_error)
  __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet = __Pyx_ImportType(__pyx_t_1, "yt.utilities.lib.ewah_bool_wrap", "SparseUnorderedRefinedBitmaskSet", sizeof(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet) __PYX_ERR(11, 144, __pyx_L1_error)
  __pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet = (struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet*)__Pyx_GetVtable(__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskSet)) __PYX_ERR(11, 144, __pyx_L1_error)
  __pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector = __Pyx_ImportType(__pyx_t_1, "yt.utilities.lib.ewah_bool_wrap", "SparseUnorderedRefinedBitmaskVector", sizeof(struct __pyx_obj_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector) __PYX_ERR(11, 153, __pyx_L1_error)
  __pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector = (struct __pyx_vtabstruct_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector*)__Pyx_GetVtable(__pyx_ptype_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector->tp_dict); if (unlikely(!__pyx_vtabptr_2yt_9utilities_3lib_14ewah_bool_wrap_SparseUnorderedRefinedBitmaskVector)) __PYX_ERR(11, 153, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("yt.geometry.oct_container"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "ORDER_MAX", (void **)&__pyx_vp_2yt_8geometry_13oct_container_ORDER_MAX, "int") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("yt.geometry.grid_visitors"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "check_child_masked", (void (**)(void))&__pyx_f_2yt_8geometry_13grid_visitors_check_child_masked, "__pyx_t_5numpy_uint8_t (struct __pyx_t_2yt_8geometry_13grid_visitors_GridVisitorData *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("yt.utilities.lib.geometry_utils"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "morton_neighbors_coarse", (void (**)(void))&__pyx_f_2yt_9utilities_3lib_14geometry_utils_morton_neighbors_coarse, "__pyx_t_5numpy_uint32_t (__pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, int *, __pyx_t_5numpy_uint32_t, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "morton_neighbors_refined", (void (**)(void))&__pyx_f_2yt_9utilities_3lib_14geometry_utils_morton_neighbors_refined, "__pyx_t_5numpy_uint32_t (__pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, __pyx_t_5numpy_uint64_t, int *, __pyx_t_5numpy_uint32_t, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice, __Pyx_memviewslice)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initparticle_oct_container(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initparticle_oct_container(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_particle_oct_container(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_particle_oct_container(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_particle_oct_container(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  static PyThread_type_lock __pyx_t_4[8];
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'particle_oct_container' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_particle_oct_container(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("particle_oct_container", __pyx_methods, __pyx_k_Oct_container_tuned_for_Particl, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_yt__geometry__particle_oct_container) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "yt.geometry.particle_oct_container")) {
      if (unlikely(PyDict_SetItemString(modules, "yt.geometry.particle_oct_container", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_variable_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "yt/geometry/particle_oct_container.pyx":28
 * )
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * 
 * cimport cython
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":61
 * )
 * 
 * from collections import defaultdict             # <<<<<<<<<<<<<<
 * 
 * from yt.funcs import get_pbar
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_defaultdict);
  __Pyx_GIVEREF(__pyx_n_s_defaultdict);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_defaultdict);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_defaultdict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_defaultdict, __pyx_t_1) < 0) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/particle_oct_container.pyx":63
 * from collections import defaultdict
 * 
 * from yt.funcs import get_pbar             # <<<<<<<<<<<<<<
 * 
 * from particle_deposit cimport gind
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_get_pbar);
  __Pyx_GIVEREF(__pyx_n_s_get_pbar);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_get_pbar);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_yt_funcs, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_get_pbar); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_pbar, __pyx_t_2) < 0) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":70
 * from ..utilities.lib.ewah_bool_wrap cimport BoolArrayCollection
 * 
 * import os             # <<<<<<<<<<<<<<
 * import struct
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_os, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_1) < 0) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":71
 * 
 * import os
 * import struct             # <<<<<<<<<<<<<<
 * 
 * # If set to 1, ghost cells are added at the refined level regardless of if the
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_struct, __pyx_t_1) < 0) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":79
 * DEF RefinedExternalGhosts = 1
 * 
 * _bitmask_version = np.uint64(5)             # <<<<<<<<<<<<<<
 * 
 * from ..utilities.lib.ewah_bool_wrap cimport (
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_uint64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__58, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bitmask_version, __pyx_t_1) < 0) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "yt/geometry/particle_oct_container.pyx":219
 *     @cython.cdivision(True)
 *     def add(self, np.ndarray[np.uint64_t, ndim=1] indices,
 *             np.uint8_t order = ORDER_MAX):             # <<<<<<<<<<<<<<
 *         #Add this particle to the root oct
 *         #Then if that oct has children, add it to them recursively
 */
  __pyx_k_ = __pyx_v_2yt_8geometry_13oct_container_ORDER_MAX;

  /* "yt/geometry/particle_oct_container.pyx":339
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     cdef Oct *get_from_index(self, np.uint64_t mi, np.uint8_t order = ORDER_MAX,             # <<<<<<<<<<<<<<
 *                              int max_level = 99):
 *         cdef Oct *cur
 */
  __pyx_k__2 = __pyx_v_2yt_8geometry_13oct_container_ORDER_MAX;

  /* "yt/geometry/particle_oct_container.pyx":430
 *                 _mask_children(mask, cur.children[cind(i,j,k)])
 * 
 * cdef np.uint64_t ONEBIT=1             # <<<<<<<<<<<<<<
 * cdef np.uint64_t FLAG = ~(<np.uint64_t>0)
 * 
 */
  __pyx_v_2yt_8geometry_22particle_oct_container_ONEBIT = 1;

  /* "yt/geometry/particle_oct_container.pyx":431
 * 
 * cdef np.uint64_t ONEBIT=1
 * cdef np.uint64_t FLAG = ~(<np.uint64_t>0)             # <<<<<<<<<<<<<<
 * 
 * cdef class ParticleBitmap:
 */
  __pyx_v_2yt_8geometry_22particle_oct_container_FLAG = (~((__pyx_t_5numpy_uint64_t)0));

  /* "yt/geometry/particle_oct_container.pyx":540
 *     @cython.wraparound(False)
 *     @cython.cdivision(True)
 *     def _coarse_index_data_file(self,             # <<<<<<<<<<<<<<
 *                                 np.ndarray[floating, ndim=2] pos,
 *                                 np.ndarray[floating, ndim=1] hsml,
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_FusedFunction_New(&__pyx_fuse_0__pyx_mdef_2yt_8geometry_22particle_oct_container_14ParticleBitmap_81_coarse_index_data_file, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ParticleBitmap__coarse_index_dat, NULL, __pyx_n_s_yt_geometry_particle_oct_contain, __pyx_d, ((PyObject *)__pyx_codeobj__60)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_float, __pyx_t_2) < 0) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_FusedFunction_New(&__pyx_fuse_1__pyx_mdef_2yt_8geometry_22particle_oct_container_14ParticleBitmap_83_coarse_index_data_file, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ParticleBitmap__coarse_index_dat, NULL, __pyx_n_s_yt_geometry_particle_oct_contain, __pyx_d, ((PyObject *)__pyx_codeobj__60)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_double, __pyx_t_2) < 0) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_FusedFunction_New(&__pyx_mdef_2yt_8geometry_22particle_oct_container_14ParticleBitmap_9_coarse_index_data_file, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ParticleBitmap__coarse_index_dat, NULL, __pyx_n_s_yt_geometry_particle_oct_contain, __pyx_d, ((PyObject *)__pyx_codeobj__60)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_2)->__signatures__ = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleBitmap->tp_dict, __pyx_n_s_coarse_index_data_file, __pyx_t_2) < 0) __PYX_ERR(0, 540, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleBitmap);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "yt/geometry/particle_oct_container.pyx":677
 *                                  np.ndarray[np.uint64_t, ndim=1] sub_mi2,
 *                                  np.uint64_t file_id, np.int64_t nsub_mi,
 *                                  np.uint64_t count_threshold = 128,             # <<<<<<<<<<<<<<
 *                                  np.uint8_t mask_threshold = 2):
 *         self._used_mi2 = 1
 */
  __pyx_t_2 = __Pyx_PyInt_From_long(0x80); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "yt/geometry/particle_oct_container.pyx":678
 *                                  np.uint64_t file_id, np.int64_t nsub_mi,
 *                                  np.uint64_t count_threshold = 128,
 *                                  np.uint8_t mask_threshold = 2):             # <<<<<<<<<<<<<<
 *         self._used_mi2 = 1
 *         if in_collection is None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "yt/geometry/particle_oct_container.pyx":669
 *     @cython.cdivision(True)
 *     @cython.initializedcheck(False)
 *     def _refined_index_data_file(self,             # <<<<<<<<<<<<<<
 *                                  BoolArrayCollection in_collection,
 *                                  np.ndarray[floating, ndim=2] pos,
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_k__13 = 0x80;
  __pyx_k__14 = 2;
  __pyx_k__15 = 0x80;
  __pyx_k__16 = 2;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_FusedFunction_New(&__pyx_fuse_0__pyx_mdef_2yt_8geometry_22particle_oct_container_14ParticleBitmap_87_refined_index_data_file, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ParticleBitmap__refined_index_da, NULL, __pyx_n_s_yt_geometry_particle_oct_contain, __pyx_d, ((PyObject *)__pyx_codeobj__62)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_t_3);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_float, __pyx_t_2) < 0) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_FusedFunction_New(&__pyx_fuse_1__pyx_mdef_2yt_8geometry_22particle_oct_container_14ParticleBitmap_89_refined_index_data_file, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ParticleBitmap__refined_index_da, NULL, __pyx_n_s_yt_geometry_particle_oct_contain, __pyx_d, ((PyObject *)__pyx_codeobj__62)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_t_3);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_double, __pyx_t_2) < 0) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_FusedFunction_New(&__pyx_mdef_2yt_8geometry_22particle_oct_container_14ParticleBitmap_13_refined_index_data_file, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ParticleBitmap__refined_index_da, NULL, __pyx_n_s_yt_geometry_particle_oct_contain, __pyx_d, ((PyObject *)__pyx_codeobj__62)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_t_3);
  ((__pyx_FusedFunctionObject *) __pyx_t_2)->__signatures__ = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleBitmap->tp_dict, __pyx_n_s_refined_index_data_file, __pyx_t_2) < 0) __PYX_ERR(0, 669, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_2yt_8geometry_22particle_oct_container_ParticleBitmap);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "yt/geometry/particle_oct_container.pyx":1
 * # distutils: include_dirs = LIB_DIR_EWAH             # <<<<<<<<<<<<<<
 * # distutils: language = c++
 * # distutils: extra_compile_args = CPP14_FLAG
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":210
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(array self):
 */
  __pyx_t_3 = __pyx_capsule_create(((void *)(&__pyx_array_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_array_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_3) < 0) __PYX_ERR(1, 210, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_array_type);

  /* "View.MemoryView":287
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__63, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XGOTREF(generic);
  __Pyx_DECREF_SET(generic, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":288
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__64, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XGOTREF(strided);
  __Pyx_DECREF_SET(strided, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":289
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__65, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XGOTREF(indirect);
  __Pyx_DECREF_SET(indirect, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":292
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__66, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XGOTREF(contiguous);
  __Pyx_DECREF_SET(contiguous, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":293
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__67, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XGOTREF(indirect_contiguous);
  __Pyx_DECREF_SET(indirect_contiguous, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":317
 * 
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0             # <<<<<<<<<<<<<<
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [
 *     PyThread_allocate_lock(),
 */
  __pyx_memoryview_thread_locks_used = 0;

  /* "View.MemoryView":318
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [             # <<<<<<<<<<<<<<
 *     PyThread_allocate_lock(),
 *     PyThread_allocate_lock(),
 */
  __pyx_t_4[0] = PyThread_allocate_lock();
  __pyx_t_4[1] = PyThread_allocate_lock();
  __pyx_t_4[2] = PyThread_allocate_lock();
  __pyx_t_4[3] = PyThread_allocate_lock();
  __pyx_t_4[4] = PyThread_allocate_lock();
  __pyx_t_4[5] = PyThread_allocate_lock();
  __pyx_t_4[6] = PyThread_allocate_lock();
  __pyx_t_4[7] = PyThread_allocate_lock();
  memcpy(&(__pyx_memoryview_thread_locks[0]), __pyx_t_4, sizeof(__pyx_memoryview_thread_locks[0]) * (8));

  /* "View.MemoryView":551
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_memoryview_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_3) < 0) __PYX_ERR(1, 551, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_memoryview_type);

  /* "View.MemoryView":997
 *         return self.from_object
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem((PyObject *)__pyx_memoryviewslice_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_3) < 0) __PYX_ERR(1, 997, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_memoryviewslice_type);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum, NULL, __pyx_n_s_View_MemoryView); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Enum, __pyx_t_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "BufferFormatFromTypeInfo":1465
 * 
 * @cname('__pyx_format_from_typeinfo')
 * cdef bytes format_from_typeinfo(__Pyx_TypeInfo *type):             # <<<<<<<<<<<<<<
 *     cdef __Pyx_StructField *field
 *     cdef __pyx_typeinfo_string fmt
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init yt.geometry.particle_oct_container", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init yt.geometry.particle_oct_container");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* IsLittleEndian */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case '?': return "'bool'";
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case '?': case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number, ndim;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    ndim = ctx->head->field->type->ndim;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case '?': case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if ((ctx->enc_type == *ts) && (got_Z == ctx->is_complex) &&
            (ctx->enc_packmode == ctx->new_packmode) && (!ctx->is_valid_array)) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* BufferGetAndValidate */
  static CYTHON_INLINE void __Pyx_SafeReleaseBuffer(Py_buffer* info) {
  if (unlikely(info->buf == NULL)) return;
  if (info->suboffsets == __Pyx_minusones) info->suboffsets = NULL;
  __Pyx_ReleaseBuffer(info);
}
static void __Pyx_ZeroBuffer(Py_buffer* buf) {
  buf->buf = NULL;
  buf->obj = NULL;
  buf->strides = __Pyx_zeros;
  buf->shape = __Pyx_zeros;
  buf->suboffsets = __Pyx_minusones;
}
static int __Pyx__GetBufferAndValidate(
        Py_buffer* buf, PyObject* obj,  __Pyx_TypeInfo* dtype, int flags,
        int nd, int cast, __Pyx_BufFmt_StackElem* stack)
{
  buf->buf = NULL;
  if (unlikely(__Pyx_GetBuffer(obj, buf, flags) == -1)) {
    __Pyx_ZeroBuffer(buf);
    return -1;
  }
  if (unlikely(buf->ndim != nd)) {
    PyErr_Format(PyExc_ValueError,
                 "Buffer has wrong number of dimensions (expected %d, got %d)",
                 nd, buf->ndim);
    goto fail;
  }
  if (!cast) {
    __Pyx_BufFmt_Context ctx;
    __Pyx_BufFmt_Init(&ctx, stack, dtype);
    if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
  }
  if (unlikely((size_t)buf->itemsize != dtype->size)) {
    PyErr_Format(PyExc_ValueError,
      "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "d byte%s) does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "d byte%s)",
      buf->itemsize, (buf->itemsize > 1) ? "s" : "",
      dtype->name, (Py_ssize_t)dtype->size, (dtype->size > 1) ? "s" : "");
    goto fail;
  }
  if (buf->suboffsets == NULL) buf->suboffsets = __Pyx_minusones;
  return 0;
fail:;
  __Pyx_SafeReleaseBuffer(buf);
  return -1;
}

/* PyFunctionFastCall */
  #if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#if defined(__Pyx_CyFunction_USED) && defined(NDEBUG)
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyCFunctionFastCall */
  #if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCallOneArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
  #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyIntBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyDictVersioning */
  #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
  #if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCall2Args */
  static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* WriteUnraisableException */
  static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* GetItemInt */
  static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ExtTypeTest */
  static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* BufferFallbackError */
  static void __Pyx_RaiseBufferFallbackError(void) {
  PyErr_SetString(PyExc_ValueError,
     "Buffer acquisition failed on assignment; and then reacquiring the old buffer failed too!");
}

/* PyIntBinop */
  #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* MemviewSliceInit */
  static int
__Pyx_init_memviewslice(struct __pyx_memoryview_obj *memview,
                        int ndim,
                        __Pyx_memviewslice *memviewslice,
                        int memview_is_new_reference)
{
    __Pyx_RefNannyDeclarations
    int i, retval=-1;
    Py_buffer *buf = &memview->view;
    __Pyx_RefNannySetupContext("init_memviewslice", 0);
    if (unlikely(memviewslice->memview || memviewslice->data)) {
        PyErr_SetString(PyExc_ValueError,
            "memviewslice is already initialized!");
        goto fail;
    }
    if (buf->strides) {
        for (i = 0; i < ndim; i++) {
            memviewslice->strides[i] = buf->strides[i];
        }
    } else {
        Py_ssize_t stride = buf->itemsize;
        for (i = ndim - 1; i >= 0; i--) {
            memviewslice->strides[i] = stride;
            stride *= buf->shape[i];
        }
    }
    for (i = 0; i < ndim; i++) {
        memviewslice->shape[i]   = buf->shape[i];
        if (buf->suboffsets) {
            memviewslice->suboffsets[i] = buf->suboffsets[i];
        } else {
            memviewslice->suboffsets[i] = -1;
        }
    }
    memviewslice->memview = memview;
    memviewslice->data = (char *)buf->buf;
    if (__pyx_add_acquisition_count(memview) == 0 && !memview_is_new_reference) {
        Py_INCREF(memview);
    }
    retval = 0;
    goto no_fail;
fail:
    memviewslice->memview = 0;
    memviewslice->data = 0;
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}
#ifndef Py_NO_RETURN
#define Py_NO_RETURN
#endif
static void __pyx_fatalerror(const char *fmt, ...) Py_NO_RETURN {
    va_list vargs;
    char msg[200];
#if PY_VERSION_HEX >= 0x030A0000 || defined(HAVE_STDARG_PROTOTYPES)
    va_start(vargs, fmt);
#else
    va_start(vargs);
#endif
    vsnprintf(msg, 200, fmt, vargs);
    va_end(vargs);
    Py_FatalError(msg);
}
static CYTHON_INLINE int
__pyx_add_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)++;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE int
__pyx_sub_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)--;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE void
__Pyx_INC_MEMVIEW(__Pyx_memviewslice *memslice, int have_gil, int lineno)
{
    int first_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None))
        return;
    if (unlikely(__pyx_get_slice_count(memview) < 0))
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    first_time = __pyx_add_acquisition_count(memview) == 0;
    if (unlikely(first_time)) {
        if (have_gil) {
            Py_INCREF((PyObject *) memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_INCREF((PyObject *) memview);
            PyGILState_Release(_gilstate);
        }
    }
}
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *memslice,
                                             int have_gil, int lineno) {
    int last_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None)) {
        memslice->memview = NULL;
        return;
    }
    if (unlikely(__pyx_get_slice_count(memview) <= 0))
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    last_time = __pyx_sub_acquisition_count(memview) == 1;
    memslice->data = NULL;
    if (unlikely(last_time)) {
        if (have_gil) {
            Py_CLEAR(memslice->memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_CLEAR(memslice->memview);
            PyGILState_Release(_gilstate);
        }
    } else {
        memslice->memview = NULL;
    }
}

/* BufferIndexError */
  static void __Pyx_RaiseBufferIndexError(int axis) {
  PyErr_Format(PyExc_IndexError,
     "Out of bounds on buffer access (axis %d)", axis);
}

/* DictGetItem */
  #if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* UnicodeAsUCS4 */
  static CYTHON_INLINE Py_UCS4 __Pyx_PyUnicode_AsPy_UCS4(PyObject* x) {
   Py_ssize_t length;
   #if CYTHON_PEP393_ENABLED
   length = PyUnicode_GET_LENGTH(x);
   if (likely(length == 1)) {
       return PyUnicode_READ_CHAR(x, 0);
   }
   #else
   length = PyUnicode_GET_SIZE(x);
   if (likely(length == 1)) {
       return PyUnicode_AS_UNICODE(x)[0];
   }
   #if Py_UNICODE_SIZE == 2
   else if (PyUnicode_GET_SIZE(x) == 2) {
       Py_UCS4 high_val = PyUnicode_AS_UNICODE(x)[0];
       if (high_val >= 0xD800 && high_val <= 0xDBFF) {
           Py_UCS4 low_val = PyUnicode_AS_UNICODE(x)[1];
           if (low_val >= 0xDC00 && low_val <= 0xDFFF) {
               return 0x10000 + (((high_val & ((1<<10)-1)) << 10) | (low_val & ((1<<10)-1)));
           }
       }
   }
   #endif
   #endif
   PyErr_Format(PyExc_ValueError,
                "only single character unicode strings can be converted to Py_UCS4, "
                "got length %" CYTHON_FORMAT_SSIZE_T "d", length);
   return (Py_UCS4)-1;
}

/* object_ord */
  static long __Pyx__PyObject_Ord(PyObject* c) {
    Py_ssize_t size;
    if (PyBytes_Check(c)) {
        size = PyBytes_GET_SIZE(c);
        if (likely(size == 1)) {
            return (unsigned char) PyBytes_AS_STRING(c)[0];
        }
#if PY_MAJOR_VERSION < 3
    } else if (PyUnicode_Check(c)) {
        return (long)__Pyx_PyUnicode_AsPy_UCS4(c);
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    } else if (PyByteArray_Check(c)) {
        size = PyByteArray_GET_SIZE(c);
        if (likely(size == 1)) {
            return (unsigned char) PyByteArray_AS_STRING(c)[0];
        }
#endif
    } else {
        PyErr_Format(PyExc_TypeError,
            "ord() expected string of length 1, but %.200s found", Py_TYPE(c)->tp_name);
        return (long)(Py_UCS4)-1;
    }
    PyErr_Format(PyExc_TypeError,
        "ord() expected a character, but string of length %zd found", size);
    return (long)(Py_UCS4)-1;
}

/* SetItemInt */
  static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* IterFinish */
  static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* PyObjectGetMethod */
  static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
  static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
  static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
  static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
  static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* RaiseNoneIterError */
  static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
  static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
  static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
  static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* ObjectGetItem */
  #if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* JoinPyUnicode */
  static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      CYTHON_UNUSED Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind;
    Py_ssize_t i, char_pos;
    void *result_udata;
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely(char_pos + ulength < 0))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + char_pos * result_ukind, udata, (size_t) (ulength * result_ukind));
        } else {
            #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    result_ulength++;
    value_count++;
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* GetTopmostException */
  #if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
  #if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
  #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* GetAttr */
  static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* HasAttr */
  static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* BytesEquals */
  static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
  static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* DivInt[Py_ssize_t] */
  static CYTHON_INLINE Py_ssize_t __Pyx_div_Py_ssize_t(Py_ssize_t a, Py_ssize_t b) {
    Py_ssize_t q = a / b;
    Py_ssize_t r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* decode_c_string */
  static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* GetAttr3 */
  static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* SwapException */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* Import */
  static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* FastTypeChecks */
  #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* None */
  static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* DivInt[long] */
  static CYTHON_INLINE long __Pyx_div_long(long a, long b) {
    long q = a / b;
    long r = a - q*b;
    q -= ((r != 0) & ((r ^ b) < 0));
    return q;
}

/* ImportFrom */
  static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* StringJoin */
  #if !CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    return PyObject_CallMethodObjArgs(sep, __pyx_n_s_join, values, NULL);
}
#endif

/* CallNextTpDealloc */
  static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_dealloc != current_tp_dealloc)
        type = type->tp_base;
    while (type && type->tp_dealloc == current_tp_dealloc)
        type = type->tp_base;
    if (type)
        type->tp_dealloc(obj);
}

/* CallNextTpTraverse */
  static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_traverse != current_tp_traverse)
        type = type->tp_base;
    while (type && type->tp_traverse == current_tp_traverse)
        type = type->tp_base;
    if (type && type->tp_traverse)
        return type->tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
  static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_clear != current_tp_clear)
        type = type->tp_base;
    while (type && type->tp_clear == current_tp_clear)
        type = type->tp_base;
    if (type && type->tp_clear)
        type->tp_clear(obj);
}

/* PyObject_GenericGetAttrNoDict */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* TypeImport */
  #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
  static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* SetVTable */
  static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
  static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
  static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* FetchCommonType */
  static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunctionShared */
  #include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
#if PY_MAJOR_VERSION < 3
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
#endif
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* FusedFunction */
  static PyObject *
__pyx_FusedFunction_New(PyMethodDef *ml, int flags,
                        PyObject *qualname, PyObject *closure,
                        PyObject *module, PyObject *globals,
                        PyObject *code)
{
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_FusedFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        __pyx_FusedFunctionObject *fusedfunc = (__pyx_FusedFunctionObject *) op;
        fusedfunc->__signatures__ = NULL;
        fusedfunc->type = NULL;
        fusedfunc->self = NULL;
        PyObject_GC_Track(op);
    }
    return op;
}
static void
__pyx_FusedFunction_dealloc(__pyx_FusedFunctionObject *self)
{
    PyObject_GC_UnTrack(self);
    Py_CLEAR(self->self);
    Py_CLEAR(self->type);
    Py_CLEAR(self->__signatures__);
    __Pyx__CyFunction_dealloc((__pyx_CyFunctionObject *) self);
}
static int
__pyx_FusedFunction_traverse(__pyx_FusedFunctionObject *self,
                             visitproc visit,
                             void *arg)
{
    Py_VISIT(self->self);
    Py_VISIT(self->type);
    Py_VISIT(self->__signatures__);
    return __Pyx_CyFunction_traverse((__pyx_CyFunctionObject *) self, visit, arg);
}
static int
__pyx_FusedFunction_clear(__pyx_FusedFunctionObject *self)
{
    Py_CLEAR(self->self);
    Py_CLEAR(self->type);
    Py_CLEAR(self->__signatures__);
    return __Pyx_CyFunction_clear((__pyx_CyFunctionObject *) self);
}
static PyObject *
__pyx_FusedFunction_descr_get(PyObject *self, PyObject *obj, PyObject *type)
{
    __pyx_FusedFunctionObject *func, *meth;
    func = (__pyx_FusedFunctionObject *) self;
    if (func->self || func->func.flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(self);
        return self;
    }
    if (obj == Py_None)
        obj = NULL;
    meth = (__pyx_FusedFunctionObject *) __pyx_FusedFunction_New(
                    ((PyCFunctionObject *) func)->m_ml,
                    ((__pyx_CyFunctionObject *) func)->flags,
                    ((__pyx_CyFunctionObject *) func)->func_qualname,
                    ((__pyx_CyFunctionObject *) func)->func_closure,
                    ((PyCFunctionObject *) func)->m_module,
                    ((__pyx_CyFunctionObject *) func)->func_globals,
                    ((__pyx_CyFunctionObject *) func)->func_code);
    if (!meth)
        return NULL;
    if (func->func.defaults) {
        PyObject **pydefaults;
        int i;
        if (!__Pyx_CyFunction_InitDefaults((PyObject*)meth,
                                      func->func.defaults_size,
                                      func->func.defaults_pyobjects)) {
            Py_XDECREF((PyObject*)meth);
            return NULL;
        }
        memcpy(meth->func.defaults, func->func.defaults, func->func.defaults_size);
        pydefaults = __Pyx_CyFunction_Defaults(PyObject *, meth);
        for (i = 0; i < meth->func.defaults_pyobjects; i++)
            Py_XINCREF(pydefaults[i]);
    }
    Py_XINCREF(func->func.func_classobj);
    meth->func.func_classobj = func->func.func_classobj;
    Py_XINCREF(func->__signatures__);
    meth->__signatures__ = func->__signatures__;
    Py_XINCREF(type);
    meth->type = type;
    Py_XINCREF(func->func.defaults_tuple);
    meth->func.defaults_tuple = func->func.defaults_tuple;
    if (func->func.flags & __Pyx_CYFUNCTION_CLASSMETHOD)
        obj = type;
    Py_XINCREF(obj);
    meth->self = obj;
    return (PyObject *) meth;
}
static PyObject *
_obj_to_str(PyObject *obj)
{
    if (PyType_Check(obj))
        return PyObject_GetAttr(obj, __pyx_n_s_name_2);
    else
        return PyObject_Str(obj);
}
static PyObject *
__pyx_FusedFunction_getitem(__pyx_FusedFunctionObject *self, PyObject *idx)
{
    PyObject *signature = NULL;
    PyObject *unbound_result_func;
    PyObject *result_func = NULL;
    if (self->__signatures__ == NULL) {
        PyErr_SetString(PyExc_TypeError, "Function is not fused");
        return NULL;
    }
    if (PyTuple_Check(idx)) {
        PyObject *list = PyList_New(0);
        Py_ssize_t n = PyTuple_GET_SIZE(idx);
        PyObject *sep = NULL;
        int i;
        if (unlikely(!list))
            return NULL;
        for (i = 0; i < n; i++) {
            int ret;
            PyObject *string;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyObject *item = PyTuple_GET_ITEM(idx, i);
#else
            PyObject *item = PySequence_ITEM(idx, i);  if (unlikely(!item)) goto __pyx_err;
#endif
            string = _obj_to_str(item);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
            Py_DECREF(item);
#endif
            if (unlikely(!string)) goto __pyx_err;
            ret = PyList_Append(list, string);
            Py_DECREF(string);
            if (unlikely(ret < 0)) goto __pyx_err;
        }
        sep = PyUnicode_FromString("|");
        if (likely(sep))
            signature = PyUnicode_Join(sep, list);
__pyx_err:
;
        Py_DECREF(list);
        Py_XDECREF(sep);
    } else {
        signature = _obj_to_str(idx);
    }
    if (!signature)
        return NULL;
    unbound_result_func = PyObject_GetItem(self->__signatures__, signature);
    if (unbound_result_func) {
        if (self->self || self->type) {
            __pyx_FusedFunctionObject *unbound = (__pyx_FusedFunctionObject *) unbound_result_func;
            Py_CLEAR(unbound->func.func_classobj);
            Py_XINCREF(self->func.func_classobj);
            unbound->func.func_classobj = self->func.func_classobj;
            result_func = __pyx_FusedFunction_descr_get(unbound_result_func,
                                                        self->self, self->type);
        } else {
            result_func = unbound_result_func;
            Py_INCREF(result_func);
        }
    }
    Py_DECREF(signature);
    Py_XDECREF(unbound_result_func);
    return result_func;
}
static PyObject *
__pyx_FusedFunction_callfunction(PyObject *func, PyObject *args, PyObject *kw)
{
     __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    int static_specialized = (cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD &&
                              !((__pyx_FusedFunctionObject *) func)->__signatures__);
    if (cyfunc->flags & __Pyx_CYFUNCTION_CCLASS && !static_specialized) {
        return __Pyx_CyFunction_CallAsMethod(func, args, kw);
    } else {
        return __Pyx_CyFunction_Call(func, args, kw);
    }
}
static PyObject *
__pyx_FusedFunction_call(PyObject *func, PyObject *args, PyObject *kw)
{
    __pyx_FusedFunctionObject *binding_func = (__pyx_FusedFunctionObject *) func;
    Py_ssize_t argc = PyTuple_GET_SIZE(args);
    PyObject *new_args = NULL;
    __pyx_FusedFunctionObject *new_func = NULL;
    PyObject *result = NULL;
    PyObject *self = NULL;
    int is_staticmethod = binding_func->func.flags & __Pyx_CYFUNCTION_STATICMETHOD;
    int is_classmethod = binding_func->func.flags & __Pyx_CYFUNCTION_CLASSMETHOD;
    if (binding_func->self) {
        Py_ssize_t i;
        new_args = PyTuple_New(argc + 1);
        if (!new_args)
            return NULL;
        self = binding_func->self;
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_INCREF(self);
#endif
        Py_INCREF(self);
        PyTuple_SET_ITEM(new_args, 0, self);
        for (i = 0; i < argc; i++) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyObject *item = PyTuple_GET_ITEM(args, i);
            Py_INCREF(item);
#else
            PyObject *item = PySequence_ITEM(args, i);  if (unlikely(!item)) goto bad;
#endif
            PyTuple_SET_ITEM(new_args, i + 1, item);
        }
        args = new_args;
    } else if (binding_func->type) {
        if (argc < 1) {
            PyErr_SetString(PyExc_TypeError, "Need at least one argument, 0 given.");
            return NULL;
        }
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        self = PyTuple_GET_ITEM(args, 0);
#else
        self = PySequence_ITEM(args, 0);  if (unlikely(!self)) return NULL;
#endif
    }
    if (self && !is_classmethod && !is_staticmethod) {
        int is_instance = PyObject_IsInstance(self, binding_func->type);
        if (unlikely(!is_instance)) {
            PyErr_Format(PyExc_TypeError,
                         "First argument should be of type %.200s, got %.200s.",
                         ((PyTypeObject *) binding_func->type)->tp_name,
                         Py_TYPE(self)->tp_name);
            goto bad;
        } else if (unlikely(is_instance == -1)) {
            goto bad;
        }
    }
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
    Py_XDECREF(self);
    self = NULL;
#endif
    if (binding_func->__signatures__) {
        PyObject *tup;
        if (is_staticmethod && binding_func->func.flags & __Pyx_CYFUNCTION_CCLASS) {
            tup = PyTuple_Pack(3, args,
                               kw == NULL ? Py_None : kw,
                               binding_func->func.defaults_tuple);
            if (unlikely(!tup)) goto bad;
            new_func = (__pyx_FusedFunctionObject *) __Pyx_CyFunction_CallMethod(
                func, binding_func->__signatures__, tup, NULL);
        } else {
            tup = PyTuple_Pack(4, binding_func->__signatures__, args,
                               kw == NULL ? Py_None : kw,
                               binding_func->func.defaults_tuple);
            if (unlikely(!tup)) goto bad;
            new_func = (__pyx_FusedFunctionObject *) __pyx_FusedFunction_callfunction(func, tup, NULL);
        }
        Py_DECREF(tup);
        if (unlikely(!new_func))
            goto bad;
        Py_XINCREF(binding_func->func.func_classobj);
        Py_CLEAR(new_func->func.func_classobj);
        new_func->func.func_classobj = binding_func->func.func_classobj;
        func = (PyObject *) new_func;
    }
    result = __pyx_FusedFunction_callfunction(func, args, kw);
bad:
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
    Py_XDECREF(self);
#endif
    Py_XDECREF(new_args);
    Py_XDECREF((PyObject *) new_func);
    return result;
}
static PyMemberDef __pyx_FusedFunction_members[] = {
    {(char *) "__signatures__",
     T_OBJECT,
     offsetof(__pyx_FusedFunctionObject, __signatures__),
     READONLY,
     0},
    {0, 0, 0, 0, 0},
};
static PyMappingMethods __pyx_FusedFunction_mapping_methods = {
    0,
    (binaryfunc) __pyx_FusedFunction_getitem,
    0,
};
static PyTypeObject __pyx_FusedFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "fused_cython_function",
    sizeof(__pyx_FusedFunctionObject),
    0,
    (destructor) __pyx_FusedFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    0,
    0,
    0,
    &__pyx_FusedFunction_mapping_methods,
    0,
    (ternaryfunc) __pyx_FusedFunction_call,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __pyx_FusedFunction_traverse,
    (inquiry) __pyx_FusedFunction_clear,
    0,
    0,
    0,
    0,
    0,
    __pyx_FusedFunction_members,
    __pyx_CyFunction_getsets,
    &__pyx_CyFunctionType_type,
    0,
    __pyx_FusedFunction_descr_get,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_FusedFunction_init(void) {
    __pyx_FusedFunctionType_type.tp_base = __pyx_CyFunctionType;
    __pyx_FusedFunctionType = __Pyx_FetchCommonType(&__pyx_FusedFunctionType_type);
    if (__pyx_FusedFunctionType == NULL) {
        return -1;
    }
    return 0;
}

/* CLineInTraceback */
  #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
  static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
  #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) return __pyx_pw_5numpy_7ndarray_1__getbuffer__(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_array_type)) return __pyx_array_getbuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_memoryview_type)) return __pyx_memoryview_getbuffer(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
        else if (__Pyx_TypeCheck(obj, __pyx_ptype_5numpy_ndarray)) __pyx_pw_5numpy_7ndarray_3__releasebuffer__(obj, view);
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


  /* MemviewSliceIsContig */
  static int
__pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim)
{
    int i, index, step, start;
    Py_ssize_t itemsize = mvs.memview->view.itemsize;
    if (order == 'F') {
        step = 1;
        start = 0;
    } else {
        step = -1;
        start = ndim - 1;
    }
    for (i = 0; i < ndim; i++) {
        index = start + step * i;
        if (mvs.suboffsets[index] >= 0 || mvs.strides[index] != itemsize)
            return 0;
        itemsize *= mvs.shape[index];
    }
    return 1;
}

/* OverlappingSlices */
  static void
__pyx_get_array_memory_extents(__Pyx_memviewslice *slice,
                               void **out_start, void **out_end,
                               int ndim, size_t itemsize)
{
    char *start, *end;
    int i;
    start = end = slice->data;
    for (i = 0; i < ndim; i++) {
        Py_ssize_t stride = slice->strides[i];
        Py_ssize_t extent = slice->shape[i];
        if (extent == 0) {
            *out_start = *out_end = start;
            return;
        } else {
            if (stride > 0)
                end += stride * (extent - 1);
            else
                start += stride * (extent - 1);
        }
    }
    *out_start = start;
    *out_end = end + itemsize;
}
static int
__pyx_slices_overlap(__Pyx_memviewslice *slice1,
                     __Pyx_memviewslice *slice2,
                     int ndim, size_t itemsize)
{
    void *start1, *end1, *start2, *end2;
    __pyx_get_array_memory_extents(slice1, &start1, &end1, ndim, itemsize);
    __pyx_get_array_memory_extents(slice2, &start2, &end2, ndim, itemsize);
    return (start1 < end2) && (start2 < end1);
}

/* Capsule */
  static CYTHON_INLINE PyObject *
__pyx_capsule_create(void *p, CYTHON_UNUSED const char *sig)
{
    PyObject *cobj;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, NULL);
#else
    cobj = PyCObject_FromVoidPtr(p, NULL);
#endif
    return cobj;
}

/* TypeInfoCompare */
  static int
__pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b)
{
    int i;
    if (!a || !b)
        return 0;
    if (a == b)
        return 1;
    if (a->size != b->size || a->typegroup != b->typegroup ||
            a->is_unsigned != b->is_unsigned || a->ndim != b->ndim) {
        if (a->typegroup == 'H' || b->typegroup == 'H') {
            return a->size == b->size;
        } else {
            return 0;
        }
    }
    if (a->ndim) {
        for (i = 0; i < a->ndim; i++)
            if (a->arraysize[i] != b->arraysize[i])
                return 0;
    }
    if (a->typegroup == 'S') {
        if (a->flags != b->flags)
            return 0;
        if (a->fields || b->fields) {
            if (!(a->fields && b->fields))
                return 0;
            for (i = 0; a->fields[i].type && b->fields[i].type; i++) {
                __Pyx_StructField *field_a = a->fields + i;
                __Pyx_StructField *field_b = b->fields + i;
                if (field_a->offset != field_b->offset ||
                    !__pyx_typeinfo_cmp(field_a->type, field_b->type))
                    return 0;
            }
            return !a->fields[i].type && !b->fields[i].type;
        }
    }
    return 1;
}

/* MemviewSliceValidateAndInit */
  static int
__pyx_check_strides(Py_buffer *buf, int dim, int ndim, int spec)
{
    if (buf->shape[dim] <= 1)
        return 1;
    if (buf->strides) {
        if (spec & __Pyx_MEMVIEW_CONTIG) {
            if (spec & (__Pyx_MEMVIEW_PTR|__Pyx_MEMVIEW_FULL)) {
                if (unlikely(buf->strides[dim] != sizeof(void *))) {
                    PyErr_Format(PyExc_ValueError,
                                 "Buffer is not indirectly contiguous "
                                 "in dimension %d.", dim);
                    goto fail;
                }
            } else if (unlikely(buf->strides[dim] != buf->itemsize)) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
        if (spec & __Pyx_MEMVIEW_FOLLOW) {
            Py_ssize_t stride = buf->strides[dim];
            if (stride < 0)
                stride = -stride;
            if (unlikely(stride < buf->itemsize)) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
    } else {
        if (unlikely(spec & __Pyx_MEMVIEW_CONTIG && dim != ndim - 1)) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not contiguous in "
                         "dimension %d", dim);
            goto fail;
        } else if (unlikely(spec & (__Pyx_MEMVIEW_PTR))) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not indirect in "
                         "dimension %d", dim);
            goto fail;
        } else if (unlikely(buf->suboffsets)) {
            PyErr_SetString(PyExc_ValueError,
                            "Buffer exposes suboffsets but no strides");
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_check_suboffsets(Py_buffer *buf, int dim, CYTHON_UNUSED int ndim, int spec)
{
    if (spec & __Pyx_MEMVIEW_DIRECT) {
        if (unlikely(buf->suboffsets && buf->suboffsets[dim] >= 0)) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer not compatible with direct access "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    if (spec & __Pyx_MEMVIEW_PTR) {
        if (unlikely(!buf->suboffsets || (buf->suboffsets[dim] < 0))) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer is not indirectly accessible "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_verify_contig(Py_buffer *buf, int ndim, int c_or_f_flag)
{
    int i;
    if (c_or_f_flag & __Pyx_IS_F_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = 0; i < ndim; i++) {
            if (unlikely(stride * buf->itemsize != buf->strides[i]  &&  buf->shape[i] > 1)) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not fortran contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    } else if (c_or_f_flag & __Pyx_IS_C_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = ndim - 1; i >- 1; i--) {
            if (unlikely(stride * buf->itemsize != buf->strides[i]  &&  buf->shape[i] > 1)) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not C contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    }
    return 1;
fail:
    return 0;
}
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj)
{
    struct __pyx_memoryview_obj *memview, *new_memview;
    __Pyx_RefNannyDeclarations
    Py_buffer *buf;
    int i, spec = 0, retval = -1;
    __Pyx_BufFmt_Context ctx;
    int from_memoryview = __pyx_memoryview_check(original_obj);
    __Pyx_RefNannySetupContext("ValidateAndInit_memviewslice", 0);
    if (from_memoryview && __pyx_typeinfo_cmp(dtype, ((struct __pyx_memoryview_obj *)
                                                            original_obj)->typeinfo)) {
        memview = (struct __pyx_memoryview_obj *) original_obj;
        new_memview = NULL;
    } else {
        memview = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                            original_obj, buf_flags, 0, dtype);
        new_memview = memview;
        if (unlikely(!memview))
            goto fail;
    }
    buf = &memview->view;
    if (unlikely(buf->ndim != ndim)) {
        PyErr_Format(PyExc_ValueError,
                "Buffer has wrong number of dimensions (expected %d, got %d)",
                ndim, buf->ndim);
        goto fail;
    }
    if (new_memview) {
        __Pyx_BufFmt_Init(&ctx, stack, dtype);
        if (unlikely(!__Pyx_BufFmt_CheckString(&ctx, buf->format))) goto fail;
    }
    if (unlikely((unsigned) buf->itemsize != dtype->size)) {
        PyErr_Format(PyExc_ValueError,
                     "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "u byte%s) "
                     "does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "u byte%s)",
                     buf->itemsize,
                     (buf->itemsize > 1) ? "s" : "",
                     dtype->name,
                     dtype->size,
                     (dtype->size > 1) ? "s" : "");
        goto fail;
    }
    if (buf->len > 0) {
        for (i = 0; i < ndim; i++) {
            spec = axes_specs[i];
            if (unlikely(!__pyx_check_strides(buf, i, ndim, spec)))
                goto fail;
            if (unlikely(!__pyx_check_suboffsets(buf, i, ndim, spec)))
                goto fail;
        }
        if (unlikely(buf->strides && !__pyx_verify_contig(buf, ndim, c_or_f_flag)))
            goto fail;
    }
    if (unlikely(__Pyx_init_memviewslice(memview, ndim, memviewslice,
                                         new_memview != NULL) == -1)) {
        goto fail;
    }
    retval = 0;
    goto no_fail;
fail:
    Py_XDECREF(new_memview);
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_float(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 2,
                                                 &__Pyx_TypeInfo_float, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_double(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 2,
                                                 &__Pyx_TypeInfo_double, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_int64_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_nn___pyx_t_5numpy_int64_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint8_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint8_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn___pyx_t_5numpy_uint64_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* Print */
  #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static PyObject *__Pyx_GetStdout(void) {
    PyObject *f = PySys_GetObject((char *)"stdout");
    if (!f) {
        PyErr_SetString(PyExc_RuntimeError, "lost sys.stdout");
    }
    return f;
}
static int __Pyx_Print(PyObject* f, PyObject *arg_tuple, int newline) {
    int i;
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    for (i=0; i < PyTuple_GET_SIZE(arg_tuple); i++) {
        PyObject* v;
        if (PyFile_SoftSpace(f, 1)) {
            if (PyFile_WriteString(" ", f) < 0)
                goto error;
        }
        v = PyTuple_GET_ITEM(arg_tuple, i);
        if (PyFile_WriteObject(v, f, Py_PRINT_RAW) < 0)
            goto error;
        if (PyString_Check(v)) {
            char *s = PyString_AsString(v);
            Py_ssize_t len = PyString_Size(v);
            if (len > 0) {
                switch (s[len-1]) {
                    case ' ': break;
                    case '\f': case '\r': case '\n': case '\t': case '\v':
                        PyFile_SoftSpace(f, 0);
                        break;
                    default:  break;
                }
            }
        }
    }
    if (newline) {
        if (PyFile_WriteString("\n", f) < 0)
            goto error;
        PyFile_SoftSpace(f, 0);
    }
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
}
#else
static int __Pyx_Print(PyObject* stream, PyObject *arg_tuple, int newline) {
    PyObject* kwargs = 0;
    PyObject* result = 0;
    PyObject* end_string;
    if (unlikely(!__pyx_print)) {
        __pyx_print = PyObject_GetAttr(__pyx_b, __pyx_n_s_print);
        if (!__pyx_print)
            return -1;
    }
    if (stream) {
        kwargs = PyDict_New();
        if (unlikely(!kwargs))
            return -1;
        if (unlikely(PyDict_SetItem(kwargs, __pyx_n_s_file, stream) < 0))
            goto bad;
        if (!newline) {
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                goto bad;
            if (PyDict_SetItem(kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                goto bad;
            }
            Py_DECREF(end_string);
        }
    } else if (!newline) {
        if (unlikely(!__pyx_print_kwargs)) {
            __pyx_print_kwargs = PyDict_New();
            if (unlikely(!__pyx_print_kwargs))
                return -1;
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                return -1;
            if (PyDict_SetItem(__pyx_print_kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                return -1;
            }
            Py_DECREF(end_string);
        }
        kwargs = __pyx_print_kwargs;
    }
    result = PyObject_Call(__pyx_print, arg_tuple, kwargs);
    if (unlikely(kwargs) && (kwargs != __pyx_print_kwargs))
        Py_DECREF(kwargs);
    if (!result)
        return -1;
    Py_DECREF(result);
    return 0;
bad:
    if (kwargs != __pyx_print_kwargs)
        Py_XDECREF(kwargs);
    return -1;
}
#endif

/* MemviewDtypeToObject */
  static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5numpy_uint8_t(const char *itemp) {
    return (PyObject *) __Pyx_PyInt_From_npy_uint8(*(__pyx_t_5numpy_uint8_t *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5numpy_uint8_t(const char *itemp, PyObject *obj) {
    __pyx_t_5numpy_uint8_t value = __Pyx_PyInt_As_npy_uint8(obj);
    if ((value == ((npy_uint8)-1)) && PyErr_Occurred())
        return 0;
    *(__pyx_t_5numpy_uint8_t *) itemp = value;
    return 1;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_nn___pyx_t_5numpy_uint32_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 2,
                                                 &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint32_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
  static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_dsds_nn___pyx_t_5numpy_uint64_t(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 2,
                                                 &__Pyx_TypeInfo_nn___pyx_t_5numpy_uint64_t, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* MemviewDtypeToObject */
  static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5numpy_int64_t(const char *itemp) {
    return (PyObject *) __Pyx_PyInt_From_npy_int64(*(__pyx_t_5numpy_int64_t *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5numpy_int64_t(const char *itemp, PyObject *obj) {
    __pyx_t_5numpy_int64_t value = __Pyx_PyInt_As_npy_int64(obj);
    if ((value == ((npy_int64)-1)) && PyErr_Occurred())
        return 0;
    *(__pyx_t_5numpy_int64_t *) itemp = value;
    return 1;
}

/* MemviewDtypeToObject */
  static CYTHON_INLINE PyObject *__pyx_memview_get_nn___pyx_t_5numpy_uint64_t(const char *itemp) {
    return (PyObject *) __Pyx_PyInt_From_npy_uint64(*(__pyx_t_5numpy_uint64_t *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_nn___pyx_t_5numpy_uint64_t(const char *itemp, PyObject *obj) {
    __pyx_t_5numpy_uint64_t value = __Pyx_PyInt_As_npy_uint64(obj);
    if ((value == ((npy_uint64)-1)) && PyErr_Occurred())
        return 0;
    *(__pyx_t_5numpy_uint64_t *) itemp = value;
    return 1;
}

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* MemviewSliceCopyTemplate */
  static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object)
{
    __Pyx_RefNannyDeclarations
    int i;
    __Pyx_memviewslice new_mvs = { 0, 0, { 0 }, { 0 }, { 0 } };
    struct __pyx_memoryview_obj *from_memview = from_mvs->memview;
    Py_buffer *buf = &from_memview->view;
    PyObject *shape_tuple = NULL;
    PyObject *temp_int = NULL;
    struct __pyx_array_obj *array_obj = NULL;
    struct __pyx_memoryview_obj *memview_obj = NULL;
    __Pyx_RefNannySetupContext("__pyx_memoryview_copy_new_contig", 0);
    for (i = 0; i < ndim; i++) {
        if (unlikely(from_mvs->suboffsets[i] >= 0)) {
            PyErr_Format(PyExc_ValueError, "Cannot copy memoryview slice with "
                                           "indirect dimensions (axis %d)", i);
            goto fail;
        }
    }
    shape_tuple = PyTuple_New(ndim);
    if (unlikely(!shape_tuple)) {
        goto fail;
    }
    __Pyx_GOTREF(shape_tuple);
    for(i = 0; i < ndim; i++) {
        temp_int = PyInt_FromSsize_t(from_mvs->shape[i]);
        if(unlikely(!temp_int)) {
            goto fail;
        } else {
            PyTuple_SET_ITEM(shape_tuple, i, temp_int);
            temp_int = NULL;
        }
    }
    array_obj = __pyx_array_new(shape_tuple, sizeof_dtype, buf->format, (char *) mode, NULL);
    if (unlikely(!array_obj)) {
        goto fail;
    }
    __Pyx_GOTREF(array_obj);
    memview_obj = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                    (PyObject *) array_obj, contig_flag,
                                    dtype_is_object,
                                    from_mvs->memview->typeinfo);
    if (unlikely(!memview_obj))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview_obj, ndim, &new_mvs, 1) < 0))
        goto fail;
    if (unlikely(__pyx_memoryview_copy_contents(*from_mvs, new_mvs, ndim, ndim,
                                                dtype_is_object) < 0))
        goto fail;
    goto no_fail;
fail:
    __Pyx_XDECREF(new_mvs.memview);
    new_mvs.memview = NULL;
    new_mvs.data = NULL;
no_fail:
    __Pyx_XDECREF(shape_tuple);
    __Pyx_XDECREF(temp_int);
    __Pyx_XDECREF(array_obj);
    __Pyx_RefNannyFinishContext();
    return new_mvs;
}

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE npy_uint8 __Pyx_PyInt_As_npy_uint8(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const npy_uint8 neg_one = (npy_uint8) -1, const_zero = (npy_uint8) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(npy_uint8) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(npy_uint8, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_uint8) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_uint8) 0;
                case  1: __PYX_VERIFY_RETURN_INT(npy_uint8, digit, digits[0])
                case 2:
                    if (8 * sizeof(npy_uint8) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) >= 2 * PyLong_SHIFT) {
                            return (npy_uint8) (((((npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_uint8) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) >= 3 * PyLong_SHIFT) {
                            return (npy_uint8) (((((((npy_uint8)digits[2]) << PyLong_SHIFT) | (npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_uint8) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) >= 4 * PyLong_SHIFT) {
                            return (npy_uint8) (((((((((npy_uint8)digits[3]) << PyLong_SHIFT) | (npy_uint8)digits[2]) << PyLong_SHIFT) | (npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (npy_uint8) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(npy_uint8) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint8, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_uint8) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint8, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_uint8) 0;
                case -1: __PYX_VERIFY_RETURN_INT(npy_uint8, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(npy_uint8,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(npy_uint8) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_uint8) (((npy_uint8)-1)*(((((npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(npy_uint8) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_uint8) ((((((npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(npy_uint8) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_uint8) (((npy_uint8)-1)*(((((((npy_uint8)digits[2]) << PyLong_SHIFT) | (npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_uint8) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_uint8) ((((((((npy_uint8)digits[2]) << PyLong_SHIFT) | (npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(npy_uint8) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_uint8) (((npy_uint8)-1)*(((((((((npy_uint8)digits[3]) << PyLong_SHIFT) | (npy_uint8)digits[2]) << PyLong_SHIFT) | (npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_uint8) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint8, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint8) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_uint8) ((((((((((npy_uint8)digits[3]) << PyLong_SHIFT) | (npy_uint8)digits[2]) << PyLong_SHIFT) | (npy_uint8)digits[1]) << PyLong_SHIFT) | (npy_uint8)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(npy_uint8) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint8, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_uint8) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint8, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            npy_uint8 val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (npy_uint8) -1;
        }
    } else {
        npy_uint8 val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_uint8) -1;
        val = __Pyx_PyInt_As_npy_uint8(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_uint8");
    return (npy_uint8) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_uint8");
    return (npy_uint8) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE npy_uint64 __Pyx_PyInt_As_npy_uint64(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const npy_uint64 neg_one = (npy_uint64) -1, const_zero = (npy_uint64) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(npy_uint64) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(npy_uint64, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_uint64) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_uint64) 0;
                case  1: __PYX_VERIFY_RETURN_INT(npy_uint64, digit, digits[0])
                case 2:
                    if (8 * sizeof(npy_uint64) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) >= 2 * PyLong_SHIFT) {
                            return (npy_uint64) (((((npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_uint64) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) >= 3 * PyLong_SHIFT) {
                            return (npy_uint64) (((((((npy_uint64)digits[2]) << PyLong_SHIFT) | (npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_uint64) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) >= 4 * PyLong_SHIFT) {
                            return (npy_uint64) (((((((((npy_uint64)digits[3]) << PyLong_SHIFT) | (npy_uint64)digits[2]) << PyLong_SHIFT) | (npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (npy_uint64) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(npy_uint64) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint64, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_uint64) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint64, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_uint64) 0;
                case -1: __PYX_VERIFY_RETURN_INT(npy_uint64, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(npy_uint64,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(npy_uint64) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_uint64) (((npy_uint64)-1)*(((((npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(npy_uint64) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_uint64) ((((((npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(npy_uint64) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_uint64) (((npy_uint64)-1)*(((((((npy_uint64)digits[2]) << PyLong_SHIFT) | (npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_uint64) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_uint64) ((((((((npy_uint64)digits[2]) << PyLong_SHIFT) | (npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(npy_uint64) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_uint64) (((npy_uint64)-1)*(((((((((npy_uint64)digits[3]) << PyLong_SHIFT) | (npy_uint64)digits[2]) << PyLong_SHIFT) | (npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_uint64) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint64, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint64) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_uint64) ((((((((((npy_uint64)digits[3]) << PyLong_SHIFT) | (npy_uint64)digits[2]) << PyLong_SHIFT) | (npy_uint64)digits[1]) << PyLong_SHIFT) | (npy_uint64)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(npy_uint64) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint64, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_uint64) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint64, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            npy_uint64 val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (npy_uint64) -1;
        }
    } else {
        npy_uint64 val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_uint64) -1;
        val = __Pyx_PyInt_As_npy_uint64(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_uint64");
    return (npy_uint64) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_uint64");
    return (npy_uint64) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE npy_int64 __Pyx_PyInt_As_npy_int64(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const npy_int64 neg_one = (npy_int64) -1, const_zero = (npy_int64) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(npy_int64) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(npy_int64, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_int64) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_int64) 0;
                case  1: __PYX_VERIFY_RETURN_INT(npy_int64, digit, digits[0])
                case 2:
                    if (8 * sizeof(npy_int64) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) >= 2 * PyLong_SHIFT) {
                            return (npy_int64) (((((npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_int64) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) >= 3 * PyLong_SHIFT) {
                            return (npy_int64) (((((((npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_int64) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) >= 4 * PyLong_SHIFT) {
                            return (npy_int64) (((((((((npy_int64)digits[3]) << PyLong_SHIFT) | (npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (npy_int64) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(npy_int64) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_int64) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_int64) 0;
                case -1: __PYX_VERIFY_RETURN_INT(npy_int64, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(npy_int64,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(npy_int64) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_int64) (((npy_int64)-1)*(((((npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(npy_int64) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_int64) ((((((npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(npy_int64) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_int64) (((npy_int64)-1)*(((((((npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_int64) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_int64) ((((((((npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(npy_int64) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_int64) (((npy_int64)-1)*(((((((((npy_int64)digits[3]) << PyLong_SHIFT) | (npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_int64) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int64, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int64) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_int64) ((((((((((npy_int64)digits[3]) << PyLong_SHIFT) | (npy_int64)digits[2]) << PyLong_SHIFT) | (npy_int64)digits[1]) << PyLong_SHIFT) | (npy_int64)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(npy_int64) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_int64) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int64, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            npy_int64 val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (npy_int64) -1;
        }
    } else {
        npy_int64 val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_int64) -1;
        val = __Pyx_PyInt_As_npy_int64(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_int64");
    return (npy_int64) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_int64");
    return (npy_int64) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
  static CYTHON_INLINE npy_uint32 __Pyx_PyInt_As_npy_uint32(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const npy_uint32 neg_one = (npy_uint32) -1, const_zero = (npy_uint32) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(npy_uint32) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(npy_uint32, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_uint32) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_uint32) 0;
                case  1: __PYX_VERIFY_RETURN_INT(npy_uint32, digit, digits[0])
                case 2:
                    if (8 * sizeof(npy_uint32) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint32, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint32) >= 2 * PyLong_SHIFT) {
                            return (npy_uint32) (((((npy_uint32)digits[1]) << PyLong_SHIFT) | (npy_uint32)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_uint32) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint32, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint32) >= 3 * PyLong_SHIFT) {
                            return (npy_uint32) (((((((npy_uint32)digits[2]) << PyLong_SHIFT) | (npy_uint32)digits[1]) << PyLong_SHIFT) | (npy_uint32)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_uint32) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint32, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint32) >= 4 * PyLong_SHIFT) {
                            return (npy_uint32) (((((((((npy_uint32)digits[3]) << PyLong_SHIFT) | (npy_uint32)digits[2]) << PyLong_SHIFT) | (npy_uint32)digits[1]) << PyLong_SHIFT) | (npy_uint32)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (npy_uint32) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(npy_uint32) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint32, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_uint32) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint32, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_uint32) 0;
                case -1: __PYX_VERIFY_RETURN_INT(npy_uint32, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(npy_uint32,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(npy_uint32) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint32, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint32) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_uint32) (((npy_uint32)-1)*(((((npy_uint32)digits[1]) << PyLong_SHIFT) | (npy_uint32)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(npy_uint32) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint32, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint32) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_uint32) ((((((npy_uint32)digits[1]) << PyLong_SHIFT) | (npy_uint32)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(npy_uint32) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint32, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint32) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_uint32) (((npy_uint32)-1)*(((((((npy_uint32)digits[2]) << PyLong_SHIFT) | (npy_uint32)digits[1]) << PyLong_SHIFT) | (npy_uint32)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_uint32) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint32, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint32) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_uint32) ((((((((npy_uint32)digits[2]) << PyLong_SHIFT) | (npy_uint32)digits[1]) << PyLong_SHIFT) | (npy_uint32)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(npy_uint32) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint32, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint32) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_uint32) (((npy_uint32)-1)*(((((((((npy_uint32)digits[3]) << PyLong_SHIFT) | (npy_uint32)digits[2]) << PyLong_SHIFT) | (npy_uint32)digits[1]) << PyLong_SHIFT) | (npy_uint32)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_uint32) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_uint32, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_uint32) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_uint32) ((((((((((npy_uint32)digits[3]) << PyLong_SHIFT) | (npy_uint32)digits[2]) << PyLong_SHIFT) | (npy_uint32)digits[1]) << PyLong_SHIFT) | (npy_uint32)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(npy_uint32) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint32, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_uint32) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_uint32, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            npy_uint32 val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (npy_uint32) -1;
        }
    } else {
        npy_uint32 val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_uint32) -1;
        val = __Pyx_PyInt_As_npy_uint32(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_uint32");
    return (npy_uint32) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_uint32");
    return (npy_uint32) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int64(npy_int64 value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const npy_int64 neg_one = (npy_int64) -1, const_zero = (npy_int64) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_int64) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_int64) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int64) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_int64) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int64) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(npy_int64),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
  static CYTHON_INLINE npy_int32 __Pyx_PyInt_As_npy_int32(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const npy_int32 neg_one = (npy_int32) -1, const_zero = (npy_int32) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(npy_int32) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(npy_int32, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (npy_int32) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_int32) 0;
                case  1: __PYX_VERIFY_RETURN_INT(npy_int32, digit, digits[0])
                case 2:
                    if (8 * sizeof(npy_int32) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) >= 2 * PyLong_SHIFT) {
                            return (npy_int32) (((((npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_int32) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) >= 3 * PyLong_SHIFT) {
                            return (npy_int32) (((((((npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_int32) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) >= 4 * PyLong_SHIFT) {
                            return (npy_int32) (((((((((npy_int32)digits[3]) << PyLong_SHIFT) | (npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (npy_int32) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(npy_int32) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int32, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_int32) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int32, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (npy_int32) 0;
                case -1: __PYX_VERIFY_RETURN_INT(npy_int32, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(npy_int32,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(npy_int32) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_int32) (((npy_int32)-1)*(((((npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(npy_int32) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 2 * PyLong_SHIFT) {
                            return (npy_int32) ((((((npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(npy_int32) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_int32) (((npy_int32)-1)*(((((((npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(npy_int32) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 3 * PyLong_SHIFT) {
                            return (npy_int32) ((((((((npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(npy_int32) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_int32) (((npy_int32)-1)*(((((((((npy_int32)digits[3]) << PyLong_SHIFT) | (npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(npy_int32) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(npy_int32, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(npy_int32) - 1 > 4 * PyLong_SHIFT) {
                            return (npy_int32) ((((((((((npy_int32)digits[3]) << PyLong_SHIFT) | (npy_int32)digits[2]) << PyLong_SHIFT) | (npy_int32)digits[1]) << PyLong_SHIFT) | (npy_int32)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(npy_int32) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int32, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(npy_int32) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(npy_int32, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            npy_int32 val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (npy_int32) -1;
        }
    } else {
        npy_int32 val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (npy_int32) -1;
        val = __Pyx_PyInt_As_npy_int32(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to npy_int32");
    return (npy_int32) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to npy_int32");
    return (npy_int32) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_int32(npy_int32 value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const npy_int32 neg_one = (npy_int32) -1, const_zero = (npy_int32) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_int32) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_int32) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int32) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_int32) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_int32) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(npy_int32),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_uint64(npy_uint64 value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const npy_uint64 neg_one = (npy_uint64) -1, const_zero = (npy_uint64) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_uint64) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_uint64) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_uint64) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_uint64) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_uint64) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(npy_uint64),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_Py_intptr_t(Py_intptr_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_intptr_t neg_one = (Py_intptr_t) -1, const_zero = (Py_intptr_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(Py_intptr_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(Py_intptr_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(Py_intptr_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(Py_intptr_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(Py_intptr_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(Py_intptr_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_uint8(npy_uint8 value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const npy_uint8 neg_one = (npy_uint8) -1, const_zero = (npy_uint8) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_uint8) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_uint8) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_uint8) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_uint8) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_uint8) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(npy_uint8),
                                     little, !is_unsigned);
    }
}

/* TypeInfoToFormat */
  static struct __pyx_typeinfo_string __Pyx_TypeInfoToFormat(__Pyx_TypeInfo *type) {
    struct __pyx_typeinfo_string result = { {0} };
    char *buf = (char *) result.string;
    size_t size = type->size;
    switch (type->typegroup) {
        case 'H':
            *buf = 'c';
            break;
        case 'I':
        case 'U':
            if (size == 1)
                *buf = (type->is_unsigned) ? 'B' : 'b';
            else if (size == 2)
                *buf = (type->is_unsigned) ? 'H' : 'h';
            else if (size == 4)
                *buf = (type->is_unsigned) ? 'I' : 'i';
            else if (size == 8)
                *buf = (type->is_unsigned) ? 'Q' : 'q';
            break;
        case 'P':
            *buf = 'P';
            break;
        case 'C':
         {
            __Pyx_TypeInfo complex_type = *type;
            complex_type.typegroup = 'R';
            complex_type.size /= 2;
            *buf++ = 'Z';
            *buf = __Pyx_TypeInfoToFormat(&complex_type).string[0];
            break;
         }
        case 'R':
            if (size == 4)
                *buf = 'f';
            else if (size == 8)
                *buf = 'd';
            else
                *buf = 'g';
            break;
    }
    return result;
}

/* BytesContains */
  static CYTHON_INLINE int __Pyx_BytesContains(PyObject* bytes, char character) {
    const Py_ssize_t length = PyBytes_GET_SIZE(bytes);
    char* char_start = PyBytes_AS_STRING(bytes);
    return memchr(char_start, (unsigned char)character, (size_t)length) != NULL;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_npy_uint32(npy_uint32 value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const npy_uint32 neg_one = (npy_uint32) -1, const_zero = (npy_uint32) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(npy_uint32) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(npy_uint32) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_uint32) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(npy_uint32) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(npy_uint32) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(npy_uint32),
                                     little, !is_unsigned);
    }
}

/* PrintOne */
  #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static int __Pyx_PrintOne(PyObject* f, PyObject *o) {
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    if (PyFile_SoftSpace(f, 0)) {
        if (PyFile_WriteString(" ", f) < 0)
            goto error;
    }
    if (PyFile_WriteObject(o, f, Py_PRINT_RAW) < 0)
        goto error;
    if (PyFile_WriteString("\n", f) < 0)
        goto error;
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
    /* the line below is just to avoid C compiler
     * warnings about unused functions */
    return __Pyx_Print(f, NULL, 0);
}
#else
static int __Pyx_PrintOne(PyObject* stream, PyObject *o) {
    int res;
    PyObject* arg_tuple = PyTuple_Pack(1, o);
    if (unlikely(!arg_tuple))
        return -1;
    res = __Pyx_Print(stream, arg_tuple, 1);
    Py_DECREF(arg_tuple);
    return res;
}
#endif

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum NPY_TYPES neg_one = (enum NPY_TYPES) -1, const_zero = (enum NPY_TYPES) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum NPY_TYPES) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum NPY_TYPES) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum NPY_TYPES),
                                     little, !is_unsigned);
    }
}

/* ImportNumPyArray */
  static PyObject* __Pyx__ImportNumPyArray(void) {
    PyObject *numpy_module, *ndarray_object = NULL;
    numpy_module = __Pyx_Import(__pyx_n_s_numpy, NULL, 0);
    if (likely(numpy_module)) {
        ndarray_object = PyObject_GetAttrString(numpy_module, "ndarray");
        Py_DECREF(numpy_module);
    }
    if (unlikely(!ndarray_object)) {
        PyErr_Clear();
    }
    if (unlikely(!ndarray_object || !PyObject_TypeCheck(ndarray_object, &PyType_Type))) {
        Py_XDECREF(ndarray_object);
        Py_INCREF(Py_None);
        ndarray_object = Py_None;
    }
    return ndarray_object;
}
static CYTHON_INLINE PyObject* __Pyx_ImportNumPyArrayTypeIfAvailable(void) {
    if (unlikely(!__pyx_numpy_ndarray)) {
        __pyx_numpy_ndarray = __Pyx__ImportNumPyArray();
    }
    Py_INCREF(__pyx_numpy_ndarray);
    return __pyx_numpy_ndarray;
}

/* CIntFromPy */
  static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const char neg_one = (char) -1, const_zero = (char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(char, digit, digits[0])
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 2 * PyLong_SHIFT) {
                            return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 3 * PyLong_SHIFT) {
                            return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 4 * PyLong_SHIFT) {
                            return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(char, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (char) -1;
        }
    } else {
        char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyInt_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* PyObjectCallMethod1 */
  static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
  #include <structmember.h>
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (Py_REFCNT(self) > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
  static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
  #ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
  static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
  static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* VoidPtrImport */
  #ifndef __PYX_HAVE_RT_ImportVoidPtr
#define __PYX_HAVE_RT_ImportVoidPtr
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* FunctionImport */
  #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
  static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
